#include "PGdec.h"
#include "PGstr.h"

#include "options.h"
#include "oh.h"
#include "vov.h"

double defaultPeakCurrent();
struct Terminal *qlast;

/* The algorithm is as follows :
**     for each power or ground net {
**         find the corresponding pad pin;
**         every pterm or aterm under the same subnet as the pad pin
**                             becomes children of that pin;
**         for every pterm find the other subnet containing it,
**         The terminals under that subnet become children of that pterm;
**     }
*/

void
PGWarning(es)
    char *es;
{
  FPRINTF(stderr,"PGcurrent: WARNING: %s\n" , es);
  return;
}

void
PGMessage(es)
    char *es;
{
  FPRINTF(stderr,"PGcurrent: MESSAGE: %s\n" , es);
  return;
}

void
PGBug(es)
    char *es;
{
  FPRINTF(stderr,"PGcurrent: INTERNAL INCONSISTENCY:\n** %s\n" , es);
  FPRINTF(stderr,"\t** continuing anyway\n");
  return;
}


void
PGFatalOctError(es)
    char *es;
{
  FPRINTF(stderr,"PGcurrent: FATAL OCT ERROR: %s\n" , es);
  octError("");
  VOVend(-1);
  /*NOTREACHED*/
}

static octObject GlobalChipContents;
static octObject GlobalChipInterface;	/* the interface facet of unsapced */


struct Terminal *termtreeptr; /* root of the terminal tree for width calc */

int 	debug = 0;
char   errmsg[256];		/* error message string */


optionStruct optionList[] = {
    { "d",	0,		"debug mode" },
    { OPT_RARG,	"cell:view",	"input cell and view" },
    { 0,	0,		0 }
};

main(argc, argv)
    int argc;
    char **argv;
{
    extern int	optind;
    extern char	*optarg;
    struct Terminal *qpop();
	
	
    int groundflag, supplyflag;

    octObject oPorGNet;
    octObject oNetType;
    octObject CDbag;
    char* netType;
    octGenerator oGenNet;

    octStatus status1;
    int option;

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	case 'd':
	    debug = 1;
	    break;
	}
    }
    if ( argc <= 1 ) optUsage();

    octBegin();
    /* open the cd_after_gr view of the chip */
    VOVbegin( argc, argv );

    ohUnpackDefaults( &GlobalChipContents,  "a", "cell:cd_after_gr:contents" );
    ohUnpackDefaults( &GlobalChipInterface, "a", "cell:cd_after_gr:interface" );
    ohUnpackFacetName( &GlobalChipContents, argv[optind] );
    
    GlobalChipInterface.contents.facet.cell = GlobalChipContents.contents.facet.cell;
    GlobalChipInterface.contents.facet.view = GlobalChipContents.contents.facet.view;
    if ( octOpenFacet(&GlobalChipContents) != OCT_OLD_FACET ) {
	PGFatalOctError("opening the cell contents facet");
    }
    VOVinputFacet( &GlobalChipContents );
    VOVoutputFacet( &GlobalChipContents );

    if ( octOpenFacet(&GlobalChipInterface) != OCT_OLD_FACET ) {
	PGFatalOctError("opening the cell interface facet");
    }

    VOVinputFacet( &GlobalChipInterface );
    VOVoutputFacet( &GlobalChipInterface );
    /* compute the default peak current */

    (void)defaultPeakCurrent(&GlobalChipContents);

    /* find all the nets in the chip */

    groundflag = NO;
    supplyflag = NO;

    if ( ohGetByBagName( &GlobalChipContents, &CDbag, "CDChanDef" ) != OCT_OK ){
	PGFatalOctError( "Cannot run PGcurrent. Missing CDChanDef bag.\n\
You must perform global routing before PGcurrent" );
    }
    if ( octInitGenContents(&GlobalChipContents,OCT_NET_MASK,&oGenNet) < OCT_OK ) {
	PGFatalOctError("generating the nets ");
    }

    oNetType.type = OCT_PROP;
    oNetType.contents.prop.name = "NETTYPE";
    oNetType.contents.prop.type = OCT_STRING;

    while ((status1 = octGenerate(&oGenNet,&oPorGNet)) == OCT_OK ) {
	/* check whether the net is P or G net */
	if (octGetByName(&oPorGNet,&oNetType) < OCT_OK) {
	    continue;
	}
	netType = oNetType.contents.prop.value.string;
	if ((strcmp(netType, "GROUND")!=0) && (strcmp(netType, "SUPPLY")!=0) )       {
	    continue;
	}

	if ( strcmp(netType, "GROUND")==0 ) {
	    groundflag = YES;
	}
	if ( strcmp(netType, "SUPPLY")==0 ) {
	    supplyflag = YES;
	}

	/* find all the actual terminals on this net */
    
	PGprocessNet( &oPorGNet, netType );
    }

    if ( supplyflag == NO ) {
	PGMessage("******** no supply nets *********");
    }
    if ( groundflag == NO ) {
	PGMessage("******** no ground nets *********");
    }

	
    OH_ASSERT(status1);

    if (debug) {
	(void)sprintf(errmsg,"Flushing %s", ohFormatName(&GlobalChipContents));
	PGMessage(errmsg);
    }
    OH_FAIL(octFlushFacet(&GlobalChipContents), "flushing", &GlobalChipContents);
    if (debug) {
	(void)sprintf(errmsg,"Flushing %s", ohFormatName(&GlobalChipInterface));
	PGMessage(errmsg);
    }
    OH_FAIL(octFlushFacet(&GlobalChipInterface), "flushing", &GlobalChipInterface);

    if (debug) PGMessage("DONE");

    octEnd();
    VOVend(0);
}


int PGprocessNet( net, type )
    octObject* net;
    char* type;
{
    struct Terminal *curpterm;	/* the psuedo terminal to be processed */
    struct Terminal *newterm;	/* the child terminal (psuedo or actual) */
    struct Terminal *qpop();
	
    struct node *tempnode;	/*dummy variable, never used(might be deleted later)*/
	
    int numpgpad = 0;

    octObject oPorGPadT;
    octObject oSubNet;
    octObject oPorGTerm;
    octObject oActualTerm;
    octObject MasterTerm;
    octObject InterfaceTerm;
    octObject Direction;
    octGenerator oGenSNet;
    octGenerator oGenTerm;

    int updatePEAKCURRENT = 0;
    octStatus  status2, status3, status4, status5, status6, status7;

    Direction.type = OCT_PROP;
    Direction.contents.prop.name = "DIRECTION";
    Direction.contents.prop.type = OCT_STRING;

    if ( octInitGenContents( net ,OCT_TERM_MASK, &oGenTerm) < OCT_OK ) {
	PGFatalOctError("generating the actual terminals from nets");
    }


    while ((status2=octGenerate(&oGenTerm,&oPorGTerm)) == OCT_OK ) {
	if ( oPorGTerm.contents.term.instanceId == 0 ) {
	    /* found a formal terminal. this is a pad if its direction is IN */
	    InterfaceTerm.contents.term.name = oPorGTerm.contents.term.name;
	    InterfaceTerm.type = OCT_TERM;
	    OH_ASSERT(octGetByName(&GlobalChipInterface,&InterfaceTerm));
	    if (octGetByName(&InterfaceTerm,&Direction)==OCT_OK &&
		strcmp(Direction.contents.prop.value.string,"INPUT")==0) {
		(void) sprintf(errmsg,"Found root: %s (FT) for %s net %s",
			       InterfaceTerm.contents.term.name,
			       type,
			       net->contents.net.name);
		PGMessage(errmsg);
		OH_ASSERT(octGenFirstContent(&oPorGTerm, OCT_TERM_MASK, &oActualTerm));
		numpgpad++;
		oPorGPadT = oActualTerm;
		updatePEAKCURRENT = 1;
	    }
	} else {		/* an actual terminal */
	
	    /* check whether this is a pad terminal
	     * if the formal terminal in the interface of the master has direction out 
	     */
	    (void) getMasterTerm(&oPorGTerm,&MasterTerm);
	    if (octGetByName(&MasterTerm,&Direction)==OCT_OK &&
		strcmp(Direction.contents.prop.value.string,"OUTPUT")==0) {
		(void) sprintf(errmsg,"Found root: %s (IT) for %s net %s",
			       MasterTerm.contents.term.name,
			       type,
			       net->contents.net.name);
		PGMessage(errmsg);
		numpgpad++;
		oPorGPadT = oPorGTerm;
		updatePEAKCURRENT = 0;
	    }

	}
    }				/* End generation of terminals. */

    if( status2 != OCT_GEN_DONE ) PGFatalOctError("terminal generation from net");

    if ( numpgpad == 0 ) {
	(void) sprintf(errmsg,"No pads associated with net %s of type %s",
		       net->contents.net.name, type);
	PGWarning(errmsg);
	return -1;
    }
    if ( numpgpad > 1 ) {
	(void) sprintf(errmsg,"%d pads associated with net %s of type %s",
		       numpgpad,
		       net->contents.net.name, type );
	PGMessage(errmsg);
	(void) sprintf(errmsg,"  continuing assuming %s is the pad",
		       oPorGPadT.contents.term.name);
	PGMessage(errmsg);
    }

    /* find the subnet containing the pad terminal */
    OH_ASSERT(octInitGenContainers(&oPorGPadT,OCT_NET_MASK,&oGenSNet));

    while ((status3=octGenerate(&oGenSNet,&oSubNet))==OCT_OK) {
	if (strcmp(oSubNet.contents.net.name , "CDSubnet") == 0 ) {
	    break;
	} else {
	    (void)sprintf(errmsg,"Strange net attached to pad:%s",ohFormatName(&oSubNet));
	    PGMessage( errmsg );
	}
    }

    if ( status3 != OCT_OK ) {
	PGFatalOctError("Cannot find CDsubnet. Did you run the global router?");
    }

    /* allocate memory for the root of the tree to be traversed */
    termtreeptr = (struct Terminal *)calloc(1,sizeof(struct Terminal));
    termtreeptr->termname = oPorGPadT.objectId;
    curpterm = termtreeptr;

    /* initialize the binary search tree and insert the first subnet*/
    firstnode = NULL;
    (void) InsertSNetId(oSubNet.objectId,&tempnode);

    /* initialize the queue to empty */
    qlast = NULL;

    while ( 1 ) {

	if ( debug == 1 ) {
	    octObject obj;
	    ohGetById( &obj, curpterm->termname );
	    PRINTF("PGcurrent:curpterm=%s,subnet=%s\naterms: ",
		   ohFormatName(&obj), ohFormatName(&oSubNet));
	}

	OH_ASSERT(octInitGenContents(&oSubNet,OCT_TERM_MASK, &oGenTerm));

	while ((status4=octGenerate(&oGenTerm,&oPorGTerm)) == OCT_OK ) {
	    /* the pad pin can not be its own child */
	    if (!octIdsEqual(curpterm->termname, oPorGTerm.objectId)) {
		/* allocate memory space for the actual terminal
		   and make it a son of the psuedo terminal  */
		newterm=(struct Terminal *) calloc(1,sizeof(struct Terminal));
		newterm->termname = oPorGTerm.objectId;
		if ( debug == 1 ) {
		    PRINTF( ohFormatName( &oPorGTerm ) );
		}

		InsertInG(newterm, curpterm);
	    }
	}
	OH_ASSERT( status4 );


	if ( debug == 1 ) {
	    PRINTF("\npterms: ");
	}

	OH_ASSERT(octInitGenContents(&oSubNet,OCT_BAG_MASK, &oGenTerm));
	while ((status5=octGenerate(&oGenTerm,&oPorGTerm))== OCT_OK ) {
	    if (strcmp(oPorGTerm.contents.bag.name,"CDPseudoTerm")==0 &&
		!octIdsEqual(curpterm->termname, oPorGTerm.objectId)) {
		/* allocate memory space for the psuedo terminal
		   and make it a son of the parent psuedo terminal and
		   then add it to the queue for breadth first search */
		newterm = (struct Terminal *) calloc(1,sizeof(struct Terminal));
		newterm->termname = oPorGTerm.objectId;
		InsertInG(newterm, curpterm);
						
		if ( debug == 1 ) {
		    PRINTF("%s, ", octFormatId(oPorGTerm.objectId));
		}
		/* push the pterm in the queue */
		qpush(newterm);
	    }
	}
	OH_ASSERT( status5 );

	if ( debug == 1 ) {
	    PRINTF("\n");
	}

	/* find the subnet containing the psuedo terminal */
	oPorGTerm.objectId = curpterm->termname;
	OH_ASSERT(octGetById(&oPorGTerm)); 

	OH_ASSERT(octInitGenContainers(&oPorGTerm,OCT_NET_MASK,&oGenSNet));
	while ((status6=octGenerate(&oGenSNet,&oSubNet))==OCT_OK) {
	    if (strcmp(oSubNet.contents.net.name , "CDSubnet") == 0 ) {
		if (findkey(oSubNet.objectId,&tempnode) ==NO) {
		    /* it is not in the search tree yet */
		    (void) InsertSNetId(oSubNet.objectId,&tempnode) ;
		    /* The subnet's terminals are children
		       of the current pTerm */
		    break;
		}
	    }
	}

	OH_ASSERT(status6);
      
	/* all the parent subnets have been processed */

	if ((curpterm = qpop()) == NULL ) { /* queue is empty */
	    break;
	}
	oPorGTerm.objectId = curpterm->termname;
	OH_ASSERT(octGetById(&oPorGTerm));

	/* find the subnet containing the psuedo terminal */
	OH_ASSERT(octInitGenContainers(&oPorGTerm,OCT_NET_MASK,&oGenSNet));

	while ((status7=octGenerate(&oGenSNet,&oSubNet))==OCT_OK) {
	    if (strcmp(oSubNet.contents.net.name ,"CDSubnet") == 0){
		if (findkey(oSubNet.objectId,&tempnode) ==NO) {
		    /* it is not in the search tree yet */
		    (void)InsertSNetId(oSubNet.objectId,&tempnode) ;
		    /* The subnet's terminals are children of the current pTerm */
		    break;
		}
	    }
	}
	/* a pterm is contained by at least two subnets */
	OH_ASSERT( status7 );
      
    }

    /* Now it is the time to process the tree */
    (void)PostorderGT( termtreeptr );
    
    {
	double pkcr = -termtreeptr->powreq;
	octObject peakCurrent;
	(void)sprintf(errmsg,"PEAKCURRENT at root: %.3g Amperes", pkcr );
	PGMessage( errmsg );
	if ( updatePEAKCURRENT ) {
	    ohCreateOrModifyPropReal( &InterfaceTerm, &peakCurrent, "PEAKCURRENT", pkcr );
	    PGMessage("PEAKCURRENT updated in interface facet.");
	}
    }
    return 1;
}

/* insert the Terminal in the queue */
qpush(tempterm) 
    struct Terminal *tempterm;
{

  if ( qlast == NULL ) {
    qlast = tempterm;
    qlast->qnext = qlast->qprev = qlast;
  }
  else {
    tempterm->qprev = qlast;
    tempterm->qnext = qlast->qnext;
    qlast->qnext->qprev = tempterm;
    qlast->qnext = tempterm;
  }
}


/* pop the last Terminal in the queue */
struct Terminal *qpop()
{
  struct Terminal *tempterm;

  if ( qlast == NULL ) {
    return NULL;
  }

  tempterm = qlast;
  if ( qlast->qnext == qlast ) {
    qlast = NULL;
  }
  else {
    qlast = qlast->qprev;
    qlast->qnext = qlast->qnext->qnext;
    qlast->qnext->qprev = qlast;
  }
  return  tempterm;
}


InsertInG(newson, parent)
struct Terminal *newson;
struct Terminal *parent;
{
  struct ChildPtr *tempchild;

  if (parent == NULL) {
    termtreeptr = newson;
  }
  else {
    tempchild = (struct ChildPtr *) calloc(1,sizeof(struct ChildPtr));
    tempchild->sonptr = newson;
    tempchild->nextson = parent->leftmostson;
    parent->leftmostson = parent->currentson = tempchild;
  }
}





int getMasterTerm(it,mt)
    octObject *it;		/* instance terminal */
    octObject *mt;		/* master terminal */
{
    octObject Instance;
    octObject Master;

  
    Instance.objectId = it->contents.term.instanceId;
    OH_ASSERT(octGetById(&Instance) );
    Master.type = OCT_FACET;
    Master.contents.facet.facet = "interface";
    Master.contents.facet.mode = "r";
    if (octOpenMaster(&Instance,&Master)< OCT_OK) {
	PGFatalOctError("cannot open a master");
    }

    mt->type = OCT_TERM;
    mt->contents.term.name = it->contents.term.name;
    OH_ASSERT(octGetByName(&Master,mt));

    return 1;
}


