#include "PGdec.h"
#include "PGstr.h"
#include "oh.h"

extern int debug;		/* the debug flag */
extern char errmsg[];		/* Error message string */
extern void PGBug();
extern void PGFatalOctError();

struct Terminal *stckptr; /* access to the stack */

/* postorder traversing for a general tree */

int PostorderGT(treeptr)
    struct Terminal *treeptr;
{
    struct Terminal *tempptr;
    struct Terminal *pop();
    struct ChildPtr *childptr;

    stckptr = NULL;

    /* this function prints the tree information in postorder */
    /* algorithm is modified version of hw8 for cs160 */

    if ( treeptr == NULL ) {
	FPRINTF(stderr,"PGcurrent:tree is empty\n");
	return -1;
    }

    /* originally all the currentson fields are the same as leftmostson */
    tempptr = treeptr;
    while ( 1 ) {
	while ( tempptr != NULL ) {
	    childptr = tempptr->currentson;
	    push(tempptr);
	    if ( childptr != NULL ) {
		tempptr = childptr->sonptr;
	    } else {
		tempptr = NULL;
	    }
	}

	while ( 1 ) {
	    tempptr = pop();

	    if ( tempptr == NULL ) {
		return 0;
	    }

	    if (tempptr->currentson == NULL) {
		visit( tempptr );
	    } else {
		childptr = tempptr->currentson;
		push( tempptr );
		if ( childptr != NULL ) {
		    tempptr = childptr->sonptr;
		} else {
		    tempptr = NULL;
		}
		break;
	    }
	}
    }
}   

/* push into the stack */
push(elmptr)
    struct Terminal *elmptr;
{
    elmptr->nextterm = stckptr;
    stckptr = elmptr;
    if ( elmptr->currentson != NULL ) {
	elmptr->currentson = elmptr->currentson->nextson;
    }
}

struct Terminal *pop()
{
    struct Terminal *newptr;

    if ( stckptr == NULL ) {
	return ( NULL );
    } else {
	newptr = stckptr;
	stckptr = stckptr->nextterm;
	return(newptr);
    }
}

int  visit( elmptr )
    struct Terminal *elmptr;
{
  octObject PowerTerm, oFormal;
  octObject PeakCurrent;
  struct ChildPtr *tempptr;
  extern double defaultPeakCurrent();
  octStatus s;



  OH_ASSERT(ohGetById( &PowerTerm, elmptr->termname ));
  if ( PowerTerm.type == OCT_TERM) {
      if ( elmptr->leftmostson == NULL ) {
	  /* this is an actual terminal*/
	  if (PowerTerm.contents.term.instanceId == 0) {
	      (void) sprintf(errmsg,"The formal term %s cannot be a leaf of the tree",
			     PowerTerm.contents.term.name);
	      PGBug(errmsg);
	      return -1;
	  }
	  
	  /* Set default value. */
	  PeakCurrent.type = OCT_PROP;
	  PeakCurrent.contents.prop.type = OCT_REAL;
	  PeakCurrent.contents.prop.value.real = defaultPeakCurrent((octObject*) 0);
	  s = ohGetByPropName( &PowerTerm, &PeakCurrent, "PEAKCURRENT" );
	  if ( s != OCT_OK ) {
	      if ( s != OCT_NOT_FOUND ) {
		  PGFatalOctError("getting PEAKCURRENT from actual" );
	      } 
	      OH_ASSERT(ohFindFormal( &oFormal, &PowerTerm));
	      s = ohGetByPropName( &oFormal, &PeakCurrent, "PEAKCURRENT" );
	      if ( s == OCT_NOT_FOUND ) {
		  if (debug) {
		      FPRINTF(stderr,"PGcurrent: no %s for %s\n\tAssuming default %.4f.",
			      PeakCurrent.contents.prop.name,
			      ohFormatName( &oFormal ),
			      PeakCurrent.contents.prop.value.real );
		  }
	      } else if ( s != OCT_OK ) {
		  PGFatalOctError("getting PEAKCURRENT from formal" );
	      }
	  }
	  elmptr->powreq = PeakCurrent.contents.prop.value.real;
      } else  {
	  /* this is the Power or Ground Pad */
	  tempptr = elmptr->leftmostson;
	  elmptr->powreq = 0.;
	  while ( tempptr != NULL ) {
	      elmptr->powreq -= tempptr->sonptr->powreq;
	      tempptr = tempptr->nextson;
	  }
	  PeakCurrent.contents.prop.value.real = elmptr->powreq;
      }
  } else if ( PowerTerm.type == OCT_BAG ) {
      /* this a psuedo terminal */
      tempptr = elmptr->leftmostson;
      elmptr->powreq = 0.;
      while ( tempptr != NULL ) {
	  if (debug) PRINTF("PGcurrent: pseudo term %.4g\n", tempptr->sonptr->powreq);
	  elmptr->powreq += tempptr->sonptr->powreq;
	  tempptr = tempptr->nextson;
      }

       PeakCurrent.contents.prop.value.real = elmptr->powreq;
  } else {
      PGMessage("visit: Unknown object" );
  }

  PeakCurrent.contents.prop.name = "PEAKCURRENT";
  PeakCurrent.contents.prop.type = OCT_REAL;
  OH_ASSERT( octCreateOrModify(&PowerTerm, &PeakCurrent) );
  if (debug) PRINTF("PGcurrent:%s:: PEAKCURRENT %.4f\n", ohFormatName(&PowerTerm), 
		    PeakCurrent.contents.prop.value.real);
  return 0;

}






