#include "oct.h"
#include "tap.h"
#include "oh.h"

/* #include "strings.h" */

#define OCT_LAMBDA 20
extern char errmsg[];

#define NUM_LAYERS 2
static char* layerNames[] = {"MET1","MET2"};

double defaultPeakCurrent(facet)
    octObject *facet;
{
    static int known = 0;
    static double def = 0.1e-4;	/* amperes per lambda */
    int i;			/* loop counter  */
    FILE *fp,*fopen();
    char buff[500],dummy[25],name[40];
    double value;
    double cur_denTable[NUM_LAYERS];
    int    minwidth[NUM_LAYERS];
    char *TECH_DIR;
    double   coordSize = 1e-6;
    octCoord lambda = 20;	/* Default */
    

    if (known) return def;
    if (facet == (octObject *) 0) return def;


    if ( tapGetProp( facet, TAP_LAMBDA ) ) {
	lambda = TAP_LAMBDA->value.integer;
    }
    if ( tapGetProp( facet, TAP_COORD_SIZE ) ) {
	coordSize = TAP_COORD_SIZE->value.real;
    }

    /* get min width for each layer */

    for (i = 0; i < NUM_LAYERS ; i++) {
	tapLayerListElement layerList[1];
	
	layerList[0].layer.type = OCT_LAYER;
	layerList[0].layer.contents.layer.name = layerNames[i];
	
	OH_ASSERT( octGetOrCreate( facet, &layerList[0].layer) );

	minwidth[i] = tapGetMinWidth(&layerList[0].layer) / lambda ;
	if ( tapGetLyrProp( 1, layerList, TAP_MAX_CURRENT_DENSITY ) ) {
	    cur_denTable[i] = TAP_MAX_CURRENT_DENSITY->value.real * coordSize * lambda;
	} else {
	    cur_denTable[i] = 0.0015; /* Reasonable default. */
	}
	sprintf(errmsg,"%s current density %lf (amps/lambda), min width %d",
		layerNames[i],value, minwidth[i]);
    }

    /*
     * Now find the minimum of the max current carried
     * by a minimum width wire on each layer
     */
    def = 1e30;			/* huge number for the current density(amps)/lambda */
    for (i = 0; i < NUM_LAYERS ; i++) {
	double tmp = cur_denTable[i] * minwidth[i];
	def = (tmp < def) ? tmp : def;
    }
    known = 1;
    /* Now divide this number by a fudge factor */
    def /= 3.0;
    (void) sprintf(errmsg,"Default PEAKCURRENT is %.5lf Amperes",def);
    PGMessage(errmsg);
    return def;
}

