.DS








.DE
.nr PS +8
.nr VS +8
.LP
.ce
.B "The TimberWolfSC Standard Cell Placement"
.sp
.ce
.B "and Global Routing Package"
.nr PS -2
.nr VS -2
.LP
.sp 5
.ce
.B "User's Guide for Version 4.2c"
.LP
.sp 3
.nr PS -2
.nr VS -2
.LP
.ce
.B "1987 October 14"
.sp 2
.ce
.B "Carl Sechen"
.ce
.B "Kai-Win Lee"
.ce
.B "Bill Swartz"
.ce
.B "Dahe Chen"
.ce
.B "Mindy Lee"
.ce
.B "Yale University"
.nr PS -4
.nr VS -4
.LP
.bp
.ce
.B "Table of Contents"
.sp 7
.PP
1.  Introduction
.sp 4
.PP
2.  Compilation of TimberWolfSC
.sp 4
.PP
3.  TimberWolfSC Input Files
.sp 4
.PP
4.  Crash Recovery 
.sp 4
.PP
5.  TimberWolfSC Output Files
.sp 4
.PP
6.  Recommendations 
.sp 4
.PP
7.  References
.bp
.ce
.B "1.  Introduction"
.sp
.PP
TimberWolfSC is a complete standard cell placement and global
routing package.
The TimberWolfSC input format can be generated in a straight forward
manner from any standard netlist.  TimberWolfSC generates the
rows into which the standard cells are placed.
Furthermore, TimberWolfSC places the pads and macro blocks.
.PP
TimberWolfSC attempts to find a placement of the standard cells
such that the total estimated wire length is minimized.
Furthermore, final standard cell placement adjustments are
performed so as to minimize the total number of wiring tracks
required to route the chip.  The adjustments include cell orientation
changes and neighboring cell interchanges.
Pads and macro blocks are not moved after their initial placement.
TimberWolfSC can be used as part of the ThunderBird [5] standard
cell layout system, which also performs detailed routing.
.PP
The algorithm employed by TimberWolfSC is the simulated annealing
technique[1].
Early versions, and subsequently version 3.2, of TimberWolfSC
have been described in detail[2][3].
\fBVersion 4.2\fR contains numerous algorithmic[6] and
data structural enhancements.
The new algorithm requires 9 to 48 times less cpu time while
still achieving placements of equal, and usually better, quality.
In fact, it is now possible to achieve the best possible
results from TimberWolfSC for a 3000-cell circuit in less
than 24 hours on a MicroVAX II desktop workstation.
For circuits on the order of 500 cells, less than 2 hours are
required on a MicroVAX II.
Furthermore, the memory requirements have been reduced by
more than a factor of two.  A 3000-cell circuit
will use typically less than 2.5 megabytes during the 
basic placement stage and
will use typically less than 8 megabytes during global
routing.
.PP
Version 4.2b now features near-perfect row length control,
even if thousands of feed-through cells must be added.
Also, if your standard cells have nonuniform heights,
you may direct TimberWolfSC to take
the cell heights into account during the
area minimization stage (stage 3).
.sp 3
.ce 
.B "2.  Compilation of TimberWolfSC"
.sp
.PP
The distribution tape includes all the files necessary to
compile TimberWolfSC version 4.2.  TimberWolfSC is written in C.
It has been successfully compiled under
the following systems: VAX/UNIX, VAX/VMS, IBM/UTS, Amdahl/UTS,
Prime/Primos, and several other systems supporting the C
language.
The distribution includes a file named:
.B "makefile,"
which is a UNIX shell script.
UNIX users may proceed as follows:
First create a directory for the TimberWolfSC source and
object directories.  For example, in your home directory,
create a directory called \fBtwsc\fR.  Move the source
code to a subdirectory called \fBtwsc/src\fR.
Then type
\fBmake install\fR, followed by \fBmake\fR.
This will result in the generation of the
object files (placing them in a directory called
\fBtwsc/obj\fR) and the linking of TimberWolfSC.  The executable file
.B TimberWolfSC
is created and placed in the \fBtwsc\fR directory.
TimberWolfSC can then be executed by issuing the command
.B
TimberWolfSC circuitName,
.R
where 
.B circuitName
is a command line argument specifying the name of the circuit
for which TimberWolfSC is to generate the placement.
.PP
There are a few conditional compilation directives.  These are
issued at the top of the file standard.h.  For example, if
you wish to compile TimberWolfSC for a VAX/VMS system, you
must select
.LP
.B "#define VMS"
.LP
For all other systems, this line must be commented out as follows:
.LP
.B "/* #define VMS */"
.PP
Similarly, if you prefer that feed-through cells be given
distinct \fIinstance name\fR, select
.LP
.B "#define FEED_INSTANCES"
.LP
.bp
.ce
.B "3.  TimberWolfSC Input Files"
.sp 2
.PP
The term
.I
working directory
.R
will refer to the directory of the file system to which the user
is currently attached.  TimberWolfSC requires the presence of 4
input files in the working directory.  The executable file
.B TimberWolfSC
need not reside in the working directory, however TimberWolfSC
searches for the input files in the working directory.
The 4 input files which must be prepared by the user are:
.B 
circuitName.blk, circuitName.cel, circuitName.net
.R
and
.B 
circuitName.par\ .
.R
The prefix to the input files (circuitName) is the name of the
circuit as entered on the command line when issuing the
command:
.LP
.B "TimberWolfSC circuitName\ "
.PP
The file 
.B 
circuitName.blk
.R
contains information regarding the number and characteristics
of the blocks ( rows ) into which the standard cells are to be placed.
Each line entered in this file is as follows:
.LP
.B "block height [integer] class [integer] {mirror} {nofeeds} { relative_length [ float ] }"
.LP
The number of such lines in circuitName.blk determines the number
of rows for the placement.  The integer following the keyword
height specifies the height of the row.  The usual 
(and strongly suggested) metric is
microns.  The integer following class specifies an exchange
class for the block (row).  Blocks with the same exchange
class may have their cells interchanged.  For the case of
blocks of two different heights, the user may wish to prevent
shorter cells from being placed in the taller blocks.  This may
be accomplished by giving the shorter blocks class 1 and the
taller blocks class 2.  The presence of the optional keyword 
.B mirror
implies that this block should have its cells y-coordinates
mirrored with respect to blocks not having the keyword mirror.
.PP
For example, if the user wishes to have 12 blocks ( rows ) with
all blocks having height 50 microns and with every other row
mirrored, the contents of circuitName.blk would be given as follows:
.sp
.DS
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
\fBblock height\fR 50 \fBclass\fR 1
\fBblock height\fR 50 \fBclass\fR 1 \fBmirror\fR
.DE
.sp
.PP
The presence of the optional keyword
.B nofeeds
implies that TimberWolfSC is not to insert any feed-through cells
into the particular row nor is TimberWolfSC to use any implicit
feed throughs (jumpers built into the standard cells) in this
particular row.  Basically the presence of this keyword implies
that any connections that need to pass through this row must go
around it.
.PP
The presence of the optional keyword
.B relative_length
followed by a floating point number between 0 and 1
enables the user to specify the desired variances in the row
lengths.
The nominal (longest) rows are assumed to have a relative length of 1.0.
Hence, to specify that a row should have a relative length
which is 70% of a nominal row, do the following:
.DS
\fBblock height\fR 50 \fBclass\fR 1 \fBrelative_length\fR 0.7
.DE
For example, if a macro block is to reside in the upper right 
corner of the chip, the user may want to restrict the lengths
of the upper rows to allow space for the macro block. One way to
accomplish this is to use the keyword \fBrelative_length\fR followed
by a suitable fraction (perhaps 0.7) for those rows which would
otherwise overlap the area designated for the macro block.
(Of course, another way to do this is to simply include the
macro block in the cell list, as will be presented later).
The user should note that the current implementation of TimberWolfSC
aligns the left edges of all of the rows unless the presence of
a macro block on the left side of the chip dictates otherwise.
.sp 3
.PP
The file 
.B circuitName.cel
contains the descriptions of the standard cells, macro blocks, and
pads, as well as the netlist.  The descriptions of the standard
cells must precede the descriptions of the macro blocks and pads.
Each entry in circuitName.cel for standard cells
takes on the following form:
.sp
.DS
\fBcell\fR [ \fIstring\fR ] [ string ]
{ { \fBinitially\fR [\fBfixed\fR, \fBnonfixed\fR] [integer] \fBfrom\fR [\fBleft\fR, \fBright\fR] \fBof\fR \fBblock\fR [integer] } }
{ \fBnomirror\fR }
\fBleft\fR [ integer ] \fBright\fR [ integer ] \fBbottom\fR [ integer ] \fBtop\fR [ integer ]
{ \fBpin\fR \fBname\fR [ string ] \fBsignal\fR [ string ] \fBlayer\fR [ integer ] [ integer ] [ integer ] }
{ \fBequiv\fR \fBname\fR [ string ] \fBlayer\fR [ integer ] [ integer ] [ integer ] }
{ \fBpin\fR \fBname\fR [ string ] \fBsignal\fR [ string ] \fBlayer\fR [ integer ] [ integer ] [ integer ] }
{ \fBunequiv\fR \fBname\fR [ string ] \fBlayer\fR [ integer ] [ integer ] [ integer ] }
.DE
.sp 1
.PP
The keyword 
.B cell 
begins the description of a standard cell.  The \fIstring\fR following
the keyword cell is ignored completely by TimberWolfSC.
It can be used for any purpose by the user.
For example, often
the n-th cell description has this field
set to n.  However, others have used this string as an \fIinstance\fR
name, with the subsequent string field
containing the \fImaster\fR name.
In any event, TimberWolfSC ignores the first string.
The second string following the keyword cell must
be the \fIname\fR of the cell.
TimberWolfSC outputs the placement
information in terms of the these cell names.
.PP
The next line in the description of a standard cell is optional.
This line may take two different forms.  The form beginning
with the keyword
.B initially
is used to give a standard cell an initial placement into a
particular row.
Following 
.B initially
the user may select either
.B fixed
if the cell is to remain fixed at the specified location, or
.B nonfixed
if the cell is free to move.
The integer following
.B fixed
or 
.B nonfixed
represents how far from the left or right end of a row the
cell should be initially placed.
Following the keyword
.B "from,"
the user selects either
.B left
or 
.B right
to indicate whether the integer represents a distance from the
left or right end of the row.
If the user selects left, then the distance is the amount 
the left edge of the cell is to be placed from the left end
of the block (row).
On the other hand, if the user selects right, then the distance is
the amount the right edge of the cell is to be placed from the
right end of the block.
Following the keywords
.B of
and 
.B block
is an integer specifying the block ( row ) into which the cell
is to be initially placed.  This integer represents a row number,
in which the rows are numbered starting from the bottom of
the layout.
.PP
The keyword
.B nomirror
is optional and specifies that a cell in a row cannot have its
x-coordinates mirrored.  By default, mirroring is allowed.
.PP
The line beginning with the keyword
.B left
is required.
Each keyword 
.B
left, right, bottom,
.R
and
.B top
is followed by an integer representing the distance of each edge
from the exact center of the cell.  If a cell has an odd width,
the x-center is truncated to form an integer.  Note that the integers
following left and bottom are necessarily negative and that the
integers following right and top are necessarily positive.
Furthermore, the following relationships must hold:
.sp
.DS
right - | left | = 0 or 1
right + left = 0 or 1
top - | bottom | = 0 or 1
top - bottom = 0 or 1
.DE
.sp
.PP
The remainder of the description of a cell indicates the position
of the pins and the signal or net names associated with each pin.
Note that all pins associated with power and ground should not
be given to TimberWolfSC.  This is because all cells and pads are
usually connected to power and ground, hence cell placement
cannot affect the wire length of the power and ground nets.
Hence giving TimberWolfSC the pins associated with power and ground
cannot improve the placement and instead can dramatically 
increase the run time since so many additional pins have to be
considered when evaluating a new placement.
.PP
Each pin description begins with the keyword
.B pin
and is followed by the keyword
.B name
which is in turn followed by a string representing the name of the
pin.
Following the keyword 
.B signal
is a string representing the name of the signal or net to which
this pin is to be connected.
.PP
The keyword \fBlayer\fR is an optional keyword following the
net name.  It \fImust\fP be present if the keyword 
\fBpin_layers_given\fP is present in the \fBcircuitName.par\fP file.
The keyword \fBlayer\fP is followed by an integer representing
the layer number of the pin.  Legal values are currently 1, 2, or 3.
(For example, poly can be represented by 1, metal1 can be represented 
by 2, and metal2 can be represented by 3.)
The next pair of integers represent
the location of the pin relative to the center of the cell.
.PP
The description of an equivalent (or internally connected) pin 
begins with the keyword
.B equiv
and is followed by the keyword
.B name
which in turn is followed by a string representing the name of the
equivalent pin.  
The pin name may be followed by the optional keyword \fBlayer\fP
and the corresponding layer number.
Next are two integers specifying
the location of the equivalent pin relative to the center of the
cell.  Note that pins described by 
.B equiv
are assumed to be connected to the same signal or net name as the
last entered pin whose description begins with the keyword
.B
pin.
.R
.PP
If only one of an electrically-equivalent pair of pins can be
used (for example, because the resistance of the poly line
connecting them is too high), the description of the
equivalent pin must begin with the keyword \fBunequiv\fP.
The keyword \fBunequiv\fP and is followed by the keyword
.B name
which in turn is followed by a string representing the name of the
un-equivalent pin. 
The pin name may be followed by the optional keyword \fBlayer\fP
and the corresponding layer number.
Next are two integers specifying
the location of the un-equivalent pin relative to the center of the
cell.  Note that pins described by 
.B unequiv
are assumed to be connected to the same signal or net name as the
last entered pin whose description begins with the keyword
.B
pin.
.R
Following a pin description beginning with \fBpin\fP,
any number of pin descriptions beginning with
.B equiv
may follow.
However, \fIonly one\fP \fBunequiv\fP may be specified following
a pin description beginning with the keyword \fBpin\fP.
.PP
An example of a standard cell description in which the cell is
to be preplaced and mirroring is not allowed is shown below.
.sp
.DS
cell 78 ABcell8
initially fixed 0 from left of block 3
nomirror
left -50 right 50 bottom -25 top 25
pin name pin2 signal A 0 25
equiv name pin2.1 0 -25

.DE
.PP
This is an example of the description of a non-preplaced standard cell
in which mirroring is allowed.
.sp
.DS
cell 79 ABcell11
left -50 right 50 bottom -25 top 25
pin name pin5 signal C 0 25
equiv name pin5.1 0 -25

.DE
.sp
.PP
TimberWolfSC now supports implicit feed throughs (or 
built-into-the-cell jumpers), common in double metal circuits.
The user must enter the signal name 
.B "TW_PASS_THRU"
for any such pins.
If implicit feed-through paths are present, TimberWolfSC will use such
a path instead of inserting a feed-through cell whenever possible.
An example follows:
.DS

cell 1 cell1
....
....
pin name 1 signal TW_PASS_THRU 30 -52
equiv name 2 30 53
pin name 3 signal TW_PASS_THRU -30 -52
equiv name 4 -30 53

.DE
.sp 3
.PP
The macro block and pad description format is now presented.
Both macro blocks and pads are described beginning with keyword
.B
pad.
.R
The general form is show below:
.sp
.DS
\fBpad\fR [ integer ] [ string ] \fBorient\fR [ integer ]
\fBpadside\fR [ \fB\fRL, \fBB\fR, \fBR\fR, \fBT\fR, \fBML\fR, \fBMB\fR, \fBMR\fR, \fBMT\fR, \fBMUL\fR, \fBMUR\fR, \fBMLL\fR, \fBMLR\fR, \fBMM\fR, \fBMTT\fR, \fBMBB\fR ]
\fBleft\fR [ integer ] \fBright\fR [ integer ] \fBbottom\fR [ integer ] \fBtop\fR [ integer ] 
{ \fBsidespace\fR [ float ] }
{ \fBpin\fR \fBname\fR [ string ] \fBsignal\fR [ string ] \fBlayer\fP [ integer ] [ integer ] [ integer ] }
{ \fBequiv\fR \fBname\fR [ string ] \fBlayer\fP [ integer ] [ integer ] [ integer ] }

.DE
.PP
The keyword 
.B pad 
begins the description of a macro block or pad.
The \fIstring\fR following
the keyword cell is ignored completely by TimberWolfSC.
It can be used for any purpose by the user.
For example, often
the n-th pad description has this field
set to n.  However, others have used this string as an \fIinstance\fR
name, with the subsequent string field
containing the \fImaster\fR name.
In any event, TimberWolfSC ignores the first string.
The second string following the keyword pad must
be the \fIname\fR of the macro block or pad.
TimberWolfSC outputs the placement
information in terms of the these macro block and pad names.
.PP
The keyword 
.B orient
is followed by an integer representing the orientation of the macro
block or pad.  The geometry and pin locations for a pad or macro
block are assumed to be for orientation 0.  If a pad is to be
placed on a given side of the chip, the user may either enter 
the geometry appropriate for a pad on this particular side and
call that orientation 0, or the user may enter the geometry
for all pads in the same orientation (0) and simply change
the orientation integer for pads on different sides.
.PP
For example, if all pads are given in orientation 0
in the netlist ( such that the pins are on the top of the pad ),
then pads to be placed on the left side should have orientation 7,
pads on the top side should have orientation 3 and pads on the
right side should have orientation 6.
On the other hand, the user may rotate the geometry as appropriate
such that the pad pins face the core and enter the pad descriptions
with orientation 0.
.PP
The TimberWolfSC orientation numbering scheme is as follows:
Orientation 0 is that description of the geometry appearing for
pads and macro blocks.  The other 7 possible orientations are
numbered as follows: (1) Orientation 1 represents a mirror of the
cell's y coordinates, that is, a mirror about the x-axis with respect
to orientation 0. (2) Orientation 2 represents a mirror of the
cell's x coordinates, that is, a mirror about the y-axis with respect
to orientation 0. (3) Orientation 3 represents a 180 degree
rotation of the cell's coordinates with respect to orientation 0.
(4)  Orientation 4 represents a combination of a mirror of the
cell's x coordinates followed by a 90 degree rotation of the cell
with respect to orientation 0. (5) Orientation 5 represents
a combination of a mirror of the cell's x coordinates followed by
a -90 degree rotation of the cell with respect to orientation 0.
(6) Orientation 6 represents a 90 degree rotation of the cell
with respect to orientation 0. (7) Orientation 7 represents
a -90 degree rotation of the cell with respect to orientation 0.
.PP
The keyword 
.B padside
is followed by a string which must be one of the following:
.LP
.B
L, B, R, T, ML, MB, MR, MT, MUL, MUR, MLL, MLR, MM, MTT,
.R
or
.B MBB
.PP
The string
.B L
implies that the pad description is that of a pad which must be
placed on the left side of the chip.  All pads to be placed on the
left side must appear sequentially in circuitName.cel and must be
entered in their bottom to top order.  That is, the user must
.I place
the pads in their desired order on each side.  TimberWolfSC will
not change the pad ordering.
.PP
The string
.B B
implies that the pad description is that of a pad which must be
placed on the bottom side of the chip.  All pads to be placed on the
bottom side must appear sequentially in circuitName.cel and must
be entered in their left to right order.  Again, TimberWolfSC will not
change the pad ordering.
.PP
The string
.B R
implies that the pad description is that of a pad which must be
placed on the right side of the chip.  All pads to be placed on the
right side must appear sequentially in circuitName.cel and must
be entered in their bottom to top order.
.PP
The string
.B T
implies that the pad description is that of a pad which must be
placed on the top side of the chip.  All pads to be placed on the
top side must appear sequentially in circuitName.cel and must
be entered in their left to right order.
.PP
All of the possible strings beginning with the letter
.B M
imply that the
.I pad
description is that for a macro block.
You may enter more than one macro with padside \fBMTT\fR and more than
one macro with padside \fBMBB\fR (they will be placed in left
to right order).  For all other padsides beginning with
the letter \fBM\fR, only one macro can be specified
with that particular padside.
.PP
The string 
.B ML
implies that the 
.I pad 
description is for a macro block which is to be placed in the
center of the left edge of the core region of the chip.
.PP
The string 
.B MB
implies that the 
.I pad 
description is for a macro block which is to be placed in the
center of the bottom edge of the core region of the chip.
.PP
The string 
.B MR
implies that the 
.I pad 
description is for a macro block which is to be placed in the
center of the right edge of the core region of the chip.
.PP
The string 
.B MT
implies that the 
.I pad 
description is for a macro block which is to be placed in the
center of the top edge of the core region of the chip.
.PP
The string 
.B MUL
implies that the 
.I pad 
description is for a macro block which is to be placed in the
upper left corner of the core region of the chip.
.PP
The string 
.B MUR
implies that the 
.I pad 
description is for a macro block which is to be placed in the
upper right corner of the core region of the chip.
.PP
The string 
.B MLL
implies that the 
.I pad 
description is for a macro block which is to be placed in the
lower left corner of the core region of the chip.
.PP
The string 
.B MLR
implies that the 
.I pad 
description is for a macro block which is to be placed in the
lower right corner of the core region of the chip.
.PP
The string 
.B MM
implies that the 
.I pad 
description is for a macro block which is to be placed in the
center of the core region of the chip.
.PP
The string 
.B MTT
implies that the 
.I pad 
description is for a macro block which is to be placed in the
upper portion of the core region of the chip, however all rows
must appear below this macro block.
.PP
The string 
.B MBB
implies that the 
.I pad 
description is for a macro block which is to be placed in the
lower portion of the core region of the chip, however all rows
must appear above this macro block.
.PP
For the macro blocks whose padside
is one of the strings
.B
ML, MB, MR, MT, MUL, MUR, MLL,
.R
or
.B 
MLR,
.R
the user must ensure that these macro blocks do not overlap.
If macro blocks overlap, or nearly overlap, it is up to the user
to compose these macro blocks into fewer, larger macro blocks.
.PP
Macros with padside \fBMUL\fR, \fBML\fR, \fBMLL\fR, 
\fBMUR\fR, \fBMR\fR, and \fBMLR\fR (that is,
the macros on the left and right sides of the chip) 
must have 1 / 3 of a channel
height (row separation) space above and below the macro.
Otherwise, TimberWolfSC will knock out a section of the row above
or below (whichever is applicable).  This is hard coded and
you may want to contact the author is you aren't pleased.
The macros with padside MT, MM, or MB must have 2 / 3 of a
channel above and below the macro.
If you aren't pleased with the configuration, note that it is
possible to change the row separation.  
This has the effect of increasing or
decreasing (depending directly on how you change the
row separation) the 
required space above or below the macro.  Above all, the row
separation should be set close to the expected average 
channel separation.
.PP
The line beginning with the keyword
.B left
is required.
Each keyword 
.B
left, right, bottom,
.R
and
.B top
is followed by an integer representing the distance of each edge
from the exact center of the pad or macro block.  
If a pad or macro block has an odd width,
the x-center is truncated to form an integer.  Note that the integers
following left and bottom are necessarily negative and that the
integers following right and top are necessarily positive.
Furthermore, the following relationships must hold:
.sp
.DS
right - | left | = 0 or 1
right + left = 0 or 1
top - | bottom | = 0 or 1
top - bottom = 0 or 1
.DE
.sp
.PP
The optional keyword
.B sidespace 
followed by a floating point number allows the user to place
pads at a particular relative position or allows the user to
specify the space between the left and/or right sides
of a macro block and adjacent rows.
For left or right side pads (padside \fBL\fR or \fBR\fR),
the floating point number represents a decimal fraction which
specifies the fraction of the bottom to top span of the chip
to which the center of the pad is to be placed.
For example, if the vertical span of the chip is 10000 microns,
and if the sidespace floating point number is .10 for a pad to
be placed on the left side, then the pad will have its y-center
placed at 1000 microns from the bottom edge of the chip.
For bottom or top side pads (padside \fBB\fR or \fBT\fR),
the floating point number represents a decimal fraction which
specifies the fraction of the left to right span of the chip
to which the center of the pad is to be placed.
For example, if the horizontal span of the chip is 10000 microns,
and if the sidespace floating point number is .70 for a pad to
be placed on the top side, then the pad will have its x-center
placed at 7000 microns from the left edge of the chip.
.PP
Note that if the 
.B sidespace
keyword is present for a pad on a given side, then all pads
for that side are assumed to have been given an initial placement
via the 
.B sidespace
parameter.
If the 
.B sidespace 
keyword is not present for any pads on a given side, by default
TimberWolfSC will uniformly distribute the pads along the given side.
In summary, if the user wishes to specify a particular location
for a pad on a given side (in terms of a fraction of the vertical or 
horizontal span), then he must do so for all pads on that same side.
If the 
.B sidespace
keyword is not present for any of the pads on a given side, then
TimberWolfSC will uniformly distribute the pads for that side in the
order given in the .cel file.
.sp
.PP
For macro blocks, the 
.B sidespace 
keyword takes on a different meaning.  
The floating point number determines the spacing between the left
and right sides of the macro block and the adjacent rows 
(on the left and right sides, respectively).
The actual spacing in microns is computed by multiplying the floating
point number by the average row height.
For example, if the row height is 200 microns, and if the sidespace
floating point number is 0.5, then TimberWolfSC will ensure that
100 microns of space is available between the left and/or right sides
of the macro block and any adjacent rows.
If the sidespace keyword is not present, then the default spacing
is equal to the row separation.
It only makes sense to include this keyword for macro blocks with
padside \fBMUL\fR,\fBML\fR,\fBMLL\fR,\fBMUR\fR,\fBMR\fR,\fBMLR\fR,
\fBMT\fR,\fBMM\fR,or \fBMB\fR.
.sp
.PP
Examples of pad descriptions will now be given.
Three pads which are to appear in bottom to top order
on the left side of the core are given below:
.sp
.DS
pad 1 TWpad1 orient 7
padside L
left -100 right 100 bottom -100 top 100
pin name p1 signal A 0 100

pad 2 TWpad2 orient 7
padside L
left -100 right 100 bottom -100 top 100
pin name p2 signal B 0 100

pad 3 TWpad3 orient 7
padside L
left -100 right 100 bottom -100 top 100
pin name p3 signal C 0 100

.DE
.PP
An example of a macro block which is to be placed across
the top region of the core (with all rows below the macro block)
is show below:
.sp
.DS
pad 11 TWmac3 orient 0
padside MTT
left -2000 right 2000 bottom -300 top 300
pin name p7  signal D -1000 -300
pin name p8  signal E 1000 -300
pin name p9  signal F 0 -300
pin name p10 signal G -1000 300
pin name p11 signal H 1000 300
pin name p12 signal I -2000 -200
pin name p13 signal J -2000 200
pin name p14 signal K 2000 -200
pin name p15 signal L 2000 200

.DE
.sp 3
.PP
The file 
.B circuitName.net
contains the net-weighting data.
The first line in this file is required and is of the following form:
.LP
\fBallnets\fR \fBHVweights\fR [ float ] [ float ]
.PP
This line specifies the default horizontal and vertical weighting
factor for every net.  The first floating point number is the
horizontal weight and the second is the vertical weight. If equal
weighting is desired, setting both floating point numbers to 1.0
is the recommended approach.
Note that the wiring cost for a net is the sum of two parts:
(1) The horizontal span of the net times the horizontal weighting
factor and (2) the vertical span of the net times the vertical
weighting factor.
.PP
The default weighting factors may be overridden by entering one or
more lines of the following format:
.LP
\fBnet\fR [ net_name ] \fBHVweights\fR [ float ] [ float ] { \fBnofeeds\fR } { \fBignore\fR }
.PP
The keyword 
.B net
is followed by a string which must be the name of a net specified
in circuitName.cel\ .
The two floating point numbers following the keyword
.B HVweights
again specify the horizontal and vertical weighting factors,
respectively, for this particular net.
If the optional keyword 
.B nofeeds
is present,
TimberWolfSC will not insert any feed-through cells for this net,
whether or not any are needed.
Hence this net will be routed around the rows if necessary.
.PP
If the optional keyword 
.B ignore
is present,
TimberWolfSC will ignore this net during the first stage of the
placement.
However, this net \fIwill\fR be globally routed.
Nets which connect to almost every cell, for example, clock
and power/ground lines, have no influence on placement quality.
However, because they connect to virtually every cell, each
incremental placement change in the simulated annealing 
algorithm requires reevaluation of the horizontal and vertical
spans of these nets. The consequence is a dramatic increase in
the run time of algorithm. Hence, the intent of this
\fBignore\fR option is simply to speed up TimberWolfSC
without sacrificing placement quality, while still having
all nets subjected to the global routing and placement
refinement steps.
.PP
Typical weighting schemes for single metal circuits:
.sp
.DS
 HVweights   1.0  1.0  --------vast majority of nets
 HVweights   3.0  3.0  --------somewhat critical nets
 HVweights   5.0  5.0  --------a few rather critical nets
 HVweights 10.0 10.0  --------one or two highly critical nets

.DE
.sp 3
.PP
The file 
.B circuitName.par
contains parameter specifications for TimberWolfSC.
The format for this file is as follows:
.LP
.DS
{ \fBcost_only\fR }
{ \fBfast\fR }
{ \fBpin_layers_given\fR }
\fBrowSep\fR [ float ]
{ \fBaddFeeds\fR
  \fBfeedThruWidth\fR [ integer ] { \fBlayer\fP } [ integer ]
  \fBimplicit.feed.thru.range\fR [ float ] }
{ \fBdo.global.route\fR
  { \fBdo.global.route.cell.swaps\fR { \fBfull\fR }
    { \fBtrack.spacing\fR [ integer ]  } } }

.DE
.PP
The keyword
.B cost_only
is an optional entry in the file circuitName.par\ .
Its presence will result in TimberWolfSC reading the input files,
generating an initial placement, computing the initial cost,
generating the output files, and then graceful death.
Including this keyword is highly recommended on the first run
on the input files.  Any errors will be directed to the output
file called circuitName.out\ .
.PP
The keyword
.B fast
is an optional entry in the file circuitName.par\ .
The inclusion of this keyword
will cause TimberWolfSC to be executed 
5 times faster.  However,
it is likely that your circuit area and total wire length will
be 5 to 10 percent higher than for a full run.  Nonetheless,
for chip-planning applications, the fast run is usually sufficient.
.PP
The keyword
.B pin_layers_given
is an optional entry in the file circuitName.par\ .
Version 4.2 will normally expect
to have a layer specification for each pin in the circuitName.cel
file.  However, for
upward compatibility reasons,
the \fIexclusion\fP of the keyword \fBpin_layers_given\fP 
(implying that pin layers are \fInot\fP given) enables
the use of circuitName.cel files which were generated for
previous versions.
.sp
.PP
The keyword
.B rowSep
is followed by a floating point number representing the desired
amount of separation between rows.  The amount of separation
between rows is this number times the average height of the rows.
This is, if you want the row separation equal to the average row
height, then this number should be 1.0.  On the other hand, if
you want the row separation to be twice the height of the rows,
then this number should be 2.0.  Normally,
a value of 1.0 is appropriate.
If you have macro blocks on the chip, the row separation is 
often defined by how many rows should be \fIcut out\fR by
the macro blocks.
.sp
.PP
The presence of the optional keyword
.B addFeeds
will result in TimberWolfSC adding the necessary number of 
feed-through cells and/or utilizing the required number of implicit
feed throughs so as to be able to complete the routing.
.PP
The keyword
.B feedThruWidth 
followed by an integer is required only if the keyword
.B addFeeds
is present.
The integer informs TimberWolfSC of the width of the feed-through
cells that are to be inserted should such be necessary.
The keyword \fBlayer\fP, followed by the appropriate layer 
number for the
feed-through-cell pins, must follow if the keyword
\fBpin_layers_given\fP is also present.
.PP
The keyword
.B implicit.feed.thru.range
followed by a floating point number
is also required if the keyword
.B addFeeds
is present.
The floating point number is multiplied by the row span of the
chip to give a maximum search range for finding an implicit feed.
When TimberWolfSC determines that a feed-through path is necessary
near a particular location within a row, the program first
searches the cells in the row for the nearest implicit feed through
with the search range.
If none can be found within the range, TimberWolfSC inserts an
actual feed-through cell.  Clearly it does no good to make the
floating point number larger than 1.0 and a recommended value
is approximately .25\ .
A large floating point number can result in long wires in the channels
reaching out for the implicit feeds.
.PP
The presence of the optional keyword
.B do.global.route
will result in TimberWolfSC completing a global routing step.
Following this step, pins have been selected for interconnection
in particular channels so as to minimize the total number of 
wiring tracks required.
TimberWolfSC optimizes
net segment placement (which side of a row to place a net segment,
when it is switchable) so as to reduce the number of wiring
tracks required.  This keyword is optional, however since the
global route is rather fast and does a good job, it is recommended
that you consider it "required".
In this step, the pins for each net are assigned a
.I group
number.
Pins with group number 0 are not to be connected and pins with
the same positive group number are to be connected.
More details are available in section 5 on the presentation
of the circuitName.pin file.
If the keyword
.B addFeeds
is not present, the global routing step is much less meaningful.
This is because TimberWolfSC is forced to route many of the connections
around the rows.
.PP
If you select \fB do.global.route.cell.swaps\fR,
then you must have selected
.B "do.global.route ."
If so instructed, TimberWolfSC will perform
neighboring cell swaps and cell orientation changes
if the number of tracks is reduced
(intuitively, the wire length shouldn't increase 
appreciably if the number
of routing tracks required decreases).
By default, the number of such placement refinements that will
be attempted is equal to 0.3 times the number of standard cells
and feed-through cells.
If the optional keyword \fBfull\fR is selected,
the number of attempts will be increased to 0.5 times
the number of standard and feed-through cells.
Normally the default setting is adequate.  However, if you notice
(from the circuitName.out file)
that the frequency of routing-track reductions is quite high
near the end of the run, include the keyword \fBfull\fR on the
next run.  Of course, you can always restart a run from the
point at which it completed if you feel that more optimization
is necessary.
.PP
If the keyword \fBtrack.spacing\fR is present, TimberWolfSC
will take the heights of the cells into account during stage 3
(the cell swapping phase).  This is only useful if your
standard cells have varying heights, and it is \fIvery\fR effective
for that case.  The integer following the keyword
\fBtrack.spacing\fR specifies the center-to-center
track separation for routing in the horizontal channels.
.sp 3
.ce
.B "4.  Crash Recovery"
.sp
.PP
TimberWolfSC has crash recovery.  
Two checkpoint files are created during the course of a TimberWolfSC
run.  The first of these is named
.B circuitName.sav
and the second is 
.B "circuitName.sv2\ ."
The first checkpoint file is updated after approximately 10 minutes
of cpu time has elapsed since the last checkpoint, 
independent of the computer system, during stage 1.
The second check point file begins to be created with the
completion of iteration 121,
when the feed throughs have been inserted.
The second check point file also is updated periodically during
the global route cell swapping portion of the run.
.PP
The instructions for restarting a TimberWolfSC run are as follows:
The keyword \fBrestart\fP must be present in the circuitName.par file.
If only the checkpoint file
.B circuitName.sav
is found in your working directory,
then copy this file to a file named
.B "circuitName.res\ ."
Whenever TimberWolfSC is invoked, it first checks the working directory
for the presence of a file called
.B
circuitName.res\ .
.R
If found, it will resume the run from the state stored in that file
(hence the need to delete old circuitName.res files from your
working directory).
When TimberWolfSC is restarted, you may insert
.B cost_only
in circuitName.par to obtain information in circuitName.out about
the stored state (including placement output files), or without
cost_only in which case the run continues.
Hence, to restart a run, simply copy circuitName.sav to circuitName.res
and execute TimberWolfSC.
.PP
Furthermore, if the file 
.B circuitName.sv2
was created in your working directory,
you must
.B also
copy this file to a file named
.B circuitName.rs2
in order for the run to be restarted.
That is, if TimberWolfSC has advanced to the point in which 
feed-through cells have been inserted (beyond iteration 121) and
possibly if it was in the midst of performing global route cell
swaps, the user must copy both 
.B circuitName.sav
to 
.B circuitName.res
and
.B circuitName.sv2
to 
.B circuitName.rs2
in order to restart the run from the point it was killed.
.PP
As an example, if the current run is in the midst of global route
cell swaps and from the circuitName.out file you are satisfied
with the current number of routing tracks shown there (and you
don't want to wait any longer), you may kill the run at that
point.  Then, to generate the placement and global routing
output files, simply copy
.B circuitName.sav
to
.B circuitName.res
and 
.B circuitName.sv2
to 
.B circuitName.rs2
and restart the run with 
.B cost_only
and 
.B restart
included in the 
.B circuitName.par 
file and with 
.B do.global.route.cell.swaps
removed from the
.B circuitName.par
file.
On the other hand, if you want to continue the run, eliminate
.B cost_only
from the .par file and leave
.B do.global.route.cell.swaps
in the .par file.
.sp 3
.ce
.B "5.  TimberWolfSC Output Files"
.sp
.PP
All error messages are directed to the file 
.B "circuitName.out\ ."
Whenever you run a circuit for the first time (and possibly
always), check 
.B circuitName.out 
to see if the run was complete.
The file circuitName.out has numerous statistical data
concerning a TimberWolfSC run.  Statistical data is also printed
in this file following each iteration of TimberWolfSC.
The information available in this file is largely self explanatory.
The user should become familiar with its contents.
.PP
When a TimberWolfSC run has advanced to the global routing step,
additional information is presented in
.B "circuitName.out\ ."
TimberWolfSC prints out data in the .out file concerning
the status of the global routing step and the cell swapping step.
Also, it gives the user the density of each channel and the positions
along each channel (intervals) in which the density is achieved.
From this data, you can see how many tracks TimberWolfSC has
eliminated from global routing with or without cell swaps.
.PP
TimberWolfSC also prints out iteration numbers to the standard
output device up through iteration 124.
At this point, the cell placement optimization based on
minimizing total weighted wire length is complete.
.PP
TimberWolfSC creates two files containing the general placement 
information.  The first of these is 
.B circuitName.pl1 
and the second is 
.B 
circuitName.pl2\ .  
.R
The file circuitName.pl1 contains the placement information for
each standard cell, pad, and macro block.
Meanwhile, the file circuitName.pl2 contains the placement 
information for each row or column block, and for the pads
and macro blocks.  Note that circuitName.pl2 does not contain
standard cell placement information.
The intention is that the purpose of this latter file is to aid
the user in determining the proper configuration of the circuit,
i.e. number of rows, macro block placement, and pad placement.
.PP
The format for each entry in these two 
files is the same, and is as shown below:
.LP
.DS
[ string ] [integer] [integer] [integer] [integer] [integer] [integer]

.DE
.LP
For circuitName.pl1,
the string is the name of a standard cell, macro block, or pad.
For circuitName.pl2, the string is the name of a macro block or
pad, or the row or column block number.
.PP
The pair of integers following the string represent the 
x,y\ -\ coordinates of the lower left corner of the standard cell,
macro block, pad, or row or column block.
The third and fourth integers represent the x,y\ -\ coordinates of
the upper right corner of the standard cell,
macro block, pad, or row or column block.
.PP
The fifth integer following the string represents the orientation
of the standard cell, macro block, or pad.  This field is set to 0
for row and columns blocks.
All standard cells were presumed to have been entered in orientation 0
in circuitName.cel\ . The orientation number given, if different from 0,
represents a change in a standard cell orientation.
Similarly, all macro block and pad geometry were presumed to 
have been entered in orientation 0
in circuitName.cel\ . The orientation number given, if different from 0,
represents a change in a macro block or pad orientation.
The orientation numbering scheme was presented above during the
description of the contents of the file circuitName.cel\ .  However,
due to its immediate relevance here, it will be repeated.
.PP
The TimberWolfSC orientation numbering scheme is as follows:
Orientation 0 is that description of the geometry appearing for
cells, pads and macro blocks.  The other 7 possible orientations are
numbered as follows: (1) Orientation 1 represents a mirror of the
cell's y coordinates, that is, a mirror about the x-axis with respect
to orientation 0. (2) Orientation 2 represents a mirror of the
cell's x coordinates, that is, a mirror about the y-axis with respect
to orientation 0. (3) Orientation 3 represents a 180 degree
rotation of the cell's coordinates with respect to orientation 0.
(4)  Orientation 4 represents a combination of a mirror of the
cell's x coordinates followed by a 90 degree rotation of the cell
with respect to orientation 0. (5) Orientation 5 represents
a combination of a mirror of the cell's x coordinates followed by
a -90 degree rotation of the cell with respect to orientation 0.
(6) Orientation 6 represents a 90 degree rotation of the cell
with respect to orientation 0. (7) Orientation 7 represents
a -90 degree rotation of the cell with respect to orientation 0.
.PP
The sixth integer following the string represents the block number
to which a standard cell belongs.
This field is set to zero for all macro blocks, pads, and row or
column blocks.
.PP
The first entries in the file circuitName.pl1 are the standard
cell placement information.
These entries are sorted by block number.
That is, the standard cells for block number 1 appear first,
then the standard cells for block number 2 appear, and so on.
Furthermore, the standard cells for each block are sorted in
left to right order.
Following the standard cell placement information, the pad and
macro block placement information appears.
Note that the entries for macro blocks and pads have the block
integer set to 0 (the macro blocks and pads are not placed
in a row or column block).
.PP
The first entries in the file circuitName.pl2 are the row and
column block placement information.
These entries are sorted by block number.
Following the row and column block placement information, the pad and
macro block placement information appears.
Note that the entries in this file have the block
integer set to 0. 
.sp
.PP
The global routing information is presented in an output file
named
.B "circuitName.pin\ ."
Each line in the 
.B circuitName.pin 
file consists of 9 fields of information concerning an 
.I active
pin (a pin which will participate in the minimum
area global route of the circuit) in the following format:
.DS L

[ [string] [integer] [string] [string] [integer] [integer] [integer] [integer] [integer] ]

.DE
.LP
Pins which are \fIinactive\fR, that is, pins which are not used
in the minimum area global route, are not included in this file.
.PP
The fields for each active pin are now presented.
The first field represents the name of the net attached to the pin.
The second field is an integer which represents the 
.I "group number"
for the pin.
That is, each net has its constituent pins broken down into groups
such that the pins with a common group number are to be interconnected.
Two pins belonging to the same net but with different group numbers
are 
.B not
to be interconnected.
Hence, a channel or detailed router is 
.B not
to be passed the net name for a pin, but rather the group number
in order to achieve the minimum area layout.
For clarity, the user may wish to pass both the net name and
the group number (in the fashion: netName_groupNumber) so that
the label for a pin produced on a plot contains the netName.
In any event, the group number must be passed to the channel or
detailed router in order to achieve the routing density reported
by TimberWolfSC.
.PP
The third field is a string representing the name of the cell to
which the pins belongs.
The fourth field is a string which represents the name of the pin.
The fifth and sixth field are a pair of integers representing the
x and y coordinates of the location of the pin.
.PP
The seventh field represents the channel number to which the pin
belongs. Horizontal channels are numbered (starting at 1) from 
the channel below the first row.
The left side channel (a vertical channel) is designated as
channel number -1.
The right side channel (also a vertical channel) is designated as
channel number -2.
The so-called \fItop\fP of the left side channel corresponds to the
pin locations of the left side pads and the
\fIbottom\fP of the left side channel is adjacent to the left ends
of the vertical channels.
The so-called \fItop\fP of the right side channel corresponds to the
pin locations of the right side pads and the
\fIbottom\fP of the right side channel is adjacent to the right ends
of the vertical channels.
.PP
The eighth field is a location field. It is an integer 
which takes on one of five
possible values (-2 , -1 , 0 , 1, 2). 
A net which must leave a horizontal channel and enter a vertical
channel does so by means of a pseudo pin at either the left
end or right end of a horizontal channel.
A pseudo pin at the left end of a horizontal channel has this
location field set to -2
and a pseudo pin at the right end of a horizontal channel has this
location field set to 2.
The name of a pseudo pin is: \fBPSEUDO_PIN\fP and it is
said to lie on a cell named: \fBPSEUDO_CELL\fP.
Each left- or right-side pad pin actually appears three times!  
Once as a pseudo
pin for a horizontal channel, once as a pseudo pin on the
\fBbottom\fP (designated by -1) of a side channel, 
and once as a \fIreal\fP pin
on the \fBtop\fP (designated by 1) of the same side channel.
.PP
If a pin is located
at the \fItop\fP of a channel this field is set to 1.
If a pin is located
at the \fIbottom\fP of a channel this field is set to -1.
A pin which is
neither at the top or bottom of a channel, for example,
a pin on an internal macro block,
currently has this field set to 0.
.PP
Pins on the top side pads have location 1 for the channel which
has a number equal to the number of rows plus one.
Pins on the bottom side pads have location -1 of channel number 1.
.sp
.PP
The ninth and final field is an integer which is always 0
(for upward compatibility reasons).
.PP
In summary, the nine fields for an 
.I active 
pin are:
.DS

1. Name of net to which the pin belongs.
2. Group number of the pin.
3. Cell name to which the pin belongs.
4. The name of the pin.
5. The x coordinate of the pin location.
6. The y coordinate of the pin location.
7. The channel to which the pin belongs.
8. Whether the pin is at the top (1), bottom (-1) or 
     neither (0) of the channel.
9. 0 

.DE
.sp 3
.ce
.B "6.  Recommendations"
.sp 2
.PP
The first time you run TimberWolfSC on a circuit it is important
to verify that your input syntax is correct and that the circuit
configuration generated by TimberWolfSC is a desirable one.
Consequently, you should run TimberWolfSC with the 
.B circuitName.par 
file shown below.
.DS

cost_only
rowSep 1.0

.DE
.PP
By plotting the placement information in the 
.B circuitName.pl2 
file you can observe the placement of the rows, pads, and macro
blocks.  If the configuration is to your liking, you are now 
ready for a full run.
.PP
Now that the basic configuration of the chip is to your
liking, the next step is to get an area estimate.
Since the request
.B do.global.route
is cheap in terms of cpu time and eliminates
many tracks, it is highly recommended that this keyword be present.
Then the circuitName.par file should resemble the following:
.DS

rowSep 1.0
implicit.feed.thru.range 0.25
feedThruWidth 15 layer 1
addFeeds
do.global.route

.DE
.LP
If you want a very \fIfast\fR TimberWolfSC run, then add the keyword
\fBfast\fR to the above circuitName.par file.  TimberWolfSC will
be executed 5 times faster if this keyword is present.  However,
it is likely that your circuit area and total wire length will
be 5 to 10 percent higher than for a full run.  Nonetheless,
for chip-planning applications, the fast run is usually sufficient.
.PP
The circuitName.par for a production run 
is now presented.  Basically, the user has decided that this is it
and he wants the circuit as small as possible.  Hence the .par
should look like:
.DS

rowSep 1.0
implicit.feed.thru.range 0.25
feedThruWidth 15 layer 1
addFeeds
do.global.route
do.global.route.cell.swaps

.DE
.sp
.PP
TimberWolfSC may also be used for gate array placement
in which the gates are arranged in rows.  TimberWolfSC carefully
controls the final row lengths to ensure feasible placements.
If the gate array has routing "gaps" in the rows, the user
can insert fixed dummy cells of the appropriate width and at the
appropriate location in the rows.
.sp
Questions, comments, and recommendations are welcome and
can be directed to:
.sp
Carl Sechen
.sp 0
Dept. of EE
.sp 0
Yale University
.sp 0
15 Prospect St.
.sp 0
New Haven, CT 06520
.sp 0
(203) 432-4259
.sp 0
sechen@yale.arpa
.bp
.ce
.B "7.  References"
.sp 2
.IP "1."
S. Kirkpatrick, C. Gelatt and M. Vecchi, \fIOptimization by Simulated
Annealing\fR, \fBScience\fR, vol. 220, n. 4598, pp. 671-680,
May 13, 1983.
Watson Research Center, Yorktown Heights, NY, 1982.
.IP "2."
C. Sechen and A. Sangiovanni-Vincentelli, \fIThe TimberWolf Placement and
Routing Package\fR, \fBProc. 1984 Custom Integrated Circuits Conference\fR,
Rochester, NY, May 1984.
.IP "3."
C. Sechen and A. Sangiovanni-Vincentelli, \fIThe TimberWolf Placement and
Routing Package\fR, \fBJ. of Solid State Circuits\fR, April 1985.
.IP "4."
C. Sechen and A. Sangiovanni-Vincentelli, \fITimberWolf3.2: A New
Standard Cell
Placement and Global Routing Package\fR, \fBProc. of the 1986
Design Automation Conference\fR, Las Vegas, Nevada, June 29-July 2, 1986.
.IP "5."
D. Braun, C. Sechen and A. Sangiovanni-Vincentelli, \fIThunderBird: A Complete
Standard Cell Layout System\fR,
Placement and Global Routing Package\fR, \fBProc. of the 1986
Custom Integrated Circuits Conference\fR, Rochester, N.Y. , May 12-14, 1986.
.IP "6."
C. Sechen and K. Lee, \fIAn Improved Simulated Annealing Algorithm
for Row-Based Placement\fR, \fBProc. Int. Conf. on Comp.
Aided Design,\fR,
Santa Clara, CA, Nov. 1987.

