#include "port.h"

#define systype "bsd4.2"
/*
  *--------------------------------------------------------------
  * C MEMORY MANAGEMENT SYSTEM
  *          Description : These Routines replace malloc/calloc
  *                        & free calls.  AVOID defeciencies in
  *                        APOLLO heap management routines.
  *
  * CODED BY : Bill Swartz
  *--------------------------------------------------------------
*/

#define  mp_block_cnt      256
#define  region_mask       0xFFFF8000
#define  region_overhead   8
#define  region_size       (4095 * 1024) 
#define  expected_size     (256  * 1024) 
#define  waste_allowed     28
#define  TRUE              1
#define  FALSE             0
#define  FOREVER           1
#define  ABS(value)        ( (value)>=0 ? (value) : -(value) )
#define  status_$ok        0

#define  block_overhead    (sizeof(header_t) + sizeof(trailer_t))
#define  double_word_mask  0xFFFFFFFC

#define  heap_error_base   0x40060000
#define  heap_bad_block    (heap_error_base + 1)
#define  heap_no_mem       (heap_error_base + 2)

#define  char_p            char *
typedef  int               INT;

typedef  struct {
            int              size;
         }  header_t;

typedef  struct {
            int              all;
         }  status_$t;

typedef  struct block_type {
            header_t         head;
            union {
               struct {
                  struct block_type   *link;
                  struct block_type   *back;
               } zero_case;
               int              data;
            } case_of_two;
         }  block_t, *block_p;

typedef  int              region_head_t, *region_head_p;

typedef  struct {
            int             size;
         }  trailer_t, *trailer_p;

typedef  struct {
            block_t          avail;
            block_p          alloc_rover;
            region_head_p    highest_region;
            int              cnt_blks;
            int              max_blks;
         }  zone_t;

static  zone_t       zone ;
static status_$t  status;
static char heapNotStarted = TRUE;
static int totalAllocation = 0 ;
static int inUse = 0 ;
extern SIGNAL_FN cleanup() ;

/* ++++++++++++++++++ HEAP MANAGEMENT ROUTINES ++++++++++++++ */
/* ---------heap_dispose--------- */
status_$t  heapDispose (ptr)
char_p        ptr;
{
   status_$t     st;
   trailer_p     tail;
   block_p       headPtr;
   block_p       check;
   block_p       lower;
   block_p       higher;
   block_p       temp;

/*  Begin procedure heap dispose.  */

   headPtr = (block_p) ((int) ptr - sizeof(header_t));
   inUse += headPtr->head.size ;
   tail = (trailer_p) ( ((int)headPtr) - headPtr->head.size - sizeof(trailer_t));
   if ((headPtr->head.size > 0) || (headPtr->head.size != tail->size)) {
     st.all = heap_bad_block;
     return (st);
   }

   headPtr->head.size *= -1;
   tail = (trailer_p) ( ((int)headPtr) - sizeof(trailer_t));
   if (tail->size > 0) {
      check = (block_p) ( ((int)headPtr) - tail->size);
      higher = check->case_of_two.zero_case.link;
      lower = check->case_of_two.zero_case.back;
      higher->case_of_two.zero_case.back = lower;
      lower->case_of_two.zero_case.link = higher;
      check->head.size += headPtr->head.size;
      headPtr = check;
   }

  check = (block_p) ( ((int)headPtr) + headPtr->head.size);
  if (check->head.size > 0) {
      if (zone.alloc_rover == (block_p) &(check->head.size)) {
         zone.alloc_rover = (block_p) &(headPtr->head.size);
      }
      higher = check->case_of_two.zero_case.link;
      lower = check->case_of_two.zero_case.back;
      higher->case_of_two.zero_case.back = lower;
      lower->case_of_two.zero_case.link = higher;
      headPtr->head.size += check->head.size;
   }
   tail = (trailer_p) ( ((int)headPtr) + headPtr->head.size - sizeof(trailer_t));
   tail->size = headPtr->head.size;
   headPtr->case_of_two.zero_case.link = zone.avail.case_of_two.zero_case.link;
   headPtr->case_of_two.zero_case.back = (block_p) &(zone.avail);
   temp =  (block_p) &(headPtr->head);
   zone.avail.case_of_two.zero_case.link->case_of_two.zero_case.back = temp ;
   zone.avail.case_of_two.zero_case.link = temp ;

   st.all = status_$ok;
   return (st);
}  /*  End of procedure heap dispose.  */


/* ---------- Utility Routines ---------- */

status_$t  allocateRegion (min_size)
int            min_size;
{
typedef  int   *tag_p;

   status_$t      st;
   block_p        head;
   trailer_p      tail;
   int            allocation;
   int            pageSize;
   int            i;
   int            *memory;

/*  Begin procedure allocate region.  */
   pageSize = getpagesize() ;
   if( min_size < expected_size ){
	min_size = expected_size ;
   } 
   allocation = min_size + 2*sizeof(int);
   allocation += pageSize - (allocation % pageSize) ;
   totalAllocation += allocation ;
   head = (block_p) sbrk(0) ;
   printf("memory management system invoked\n") ;
   printf("old starting memory address:%0x Page size = %d\n",head,pageSize) ;
   head = (block_p) sbrk(allocation) ; 
   printf("new starting memory address:%0x with allocation:%d bytes\n",head,
	allocation) ;
   printf("user requested memory space = %d bytes approx. %d pages\n",inUse,inUse/pageSize) ;
   printf("new memory space = %d pages of %d bytes\n",totalAllocation / pageSize, pageSize) ;
   printf("total current allocation = %d\n\n",totalAllocation) ;
   /* head = (block_p) malloc(allocation) */ ;
   if ( !(head) || head == (block_p) -1) {
      cleanupHandler(heap_no_mem);
   }
   /* initialize all bytes of memory to 1 to catch bugs */
   memory = (int *) head ;
   for( i=0;i<allocation/4;i++){
	memory[i] = -1 ;
   }

   inUse += allocation - 2*sizeof(int) ;
   tail = (trailer_p) ((int) head + allocation - sizeof(int));
   if (zone.highest_region != NULL) {
      if (zone.highest_region + ABS(*zone.highest_region) == (int *) head) {
         *zone.highest_region = *zone.highest_region - allocation;
         head = (block_p) ((int) head - sizeof(int));
      } else {
         zone.highest_region = (int *) head;
         *zone.highest_region = -allocation;
         head = (block_p) ((int) head + sizeof(int));
         allocation = allocation - 2*sizeof(int);
      }
   } else {
      zone.highest_region = (int *) head;
      *zone.highest_region = -allocation;
      head = (block_p) ((int) head + sizeof(int));
      allocation = allocation - 2*sizeof(int);
   }

   tail->size = *zone.highest_region;
   tail = (trailer_p) ((int) head + allocation - sizeof(int));
   head->head.size = tail->size = -allocation;
   head = (block_p) ((int) head + sizeof(int));
   st = heapDispose (head);
   zone.alloc_rover = zone.avail.case_of_two.zero_case.link;
   st.all = status_$ok;
   return (st);
}  /*  End procedure allocate region.  */


void  moveMem(src, dst, cnt)
int       *src;
int       *dst;
int       cnt;
{
   int       i;
   int       max_word;

/*  Begin procedure moveMem.  */
   max_word = ((cnt + 3) / 4) - 1;
   for (i=0; i<max_word; i++) {
      dst[i] = src[i];
   }
}  /*  End of procedure moveMem.  */


/* -------- heap_init -------- */
status_$t  heapInit (initial_size)
int           initial_size;
{
   status_$t     st;
   block_p       block;
   int           allocation;

/*  Begin procedure heap init.  */

   /* initialize zone record */
   zone.avail.head.size = 0 ;
   zone.avail.case_of_two.zero_case.link = &(zone.avail) ;
   zone.avail.case_of_two.zero_case.back = &(zone.avail) ;
   zone.alloc_rover = &(zone.avail) ;
   zone.highest_region = NULL ;
   zone.cnt_blks = 0 ;
   zone.max_blks = 0 ;

   if (initial_size > 0) {
      st = allocateRegion (initial_size);
      if (st.all != status_$ok){
         return (st);
      } else {
         heapNotStarted = FALSE;
      }
   }
   st.all = status_$ok;
   return (st);
}  /*  End of procedure heap init.  */


/* -------- heap_new --------- */
status_$t  heapNew (ptr, req_size)
char_p        *ptr;
int           req_size;
{
   status_$t      st;
   int            excess;
   int            allocation;
   int            block_size;
   block_p        new_block;
   block_p        check;
   trailer_p      trailer;
   region_head_p  region;

/*  Begin procedure heap new.  */
   if (req_size <= 0) {
      *ptr = NULL;
      st.all = status_$ok;
      return (st);
   } else if (heapNotStarted){
      heapInit(region_size);
   }
   block_size = (req_size + 3 + block_overhead) & double_word_mask;
   inUse += block_size ;
   check = zone.alloc_rover;
   do {
      if (check->head.size < block_size) {
         check = check->case_of_two.zero_case.link;
      } else {
         excess = check->head.size - block_size;
         zone.alloc_rover = check->case_of_two.zero_case.link;
         if (excess < waste_allowed) {
            check->case_of_two.zero_case.back->case_of_two.zero_case.link =
	        check->case_of_two.zero_case.link;
            check->case_of_two.zero_case.link->case_of_two.zero_case.back =
		check->case_of_two.zero_case.back;
            block_size = check->head.size;
         } else {
            check->head.size = excess;
            trailer = (trailer_p) ((int) check + excess - sizeof(trailer_t));
            trailer->size = excess;
            check = (block_p) ((int) check + excess);
         }
         check->head.size = -block_size;
         trailer = (trailer_p) ((int) check + block_size - sizeof(trailer_t));
         trailer->size = -block_size;
         *ptr = (char_p) &(check->case_of_two.data);
         st.all = status_$ok;
         return (st);
      }
      if (check == zone.alloc_rover) {
         st = allocateRegion (block_size);
         if (st.all != status_$ok)
            return (st);
         check = zone.alloc_rover;
      }
   } while (FOREVER);
}  /*  End procedure heap new.  */

/* ++++++++++++++++++ END HEAP MANAGEMENT ROUTINES ++++++++++++++ */

/*  SUBSTITION CALLS FOR C RUN TIME LIBRARY */

    void safefree(ptr)
    char *ptr;
    {
       status = heapDispose (ptr);
       if (status.all!=status_$ok)
           cleanupHandler(status);
    }

    void safecfree(ptr)
    char *ptr;
    {
       status = heapDispose (ptr);
       if (status.all!=status_$ok){
           cleanupHandler(status);
       }
    }


    char *testmalloc(bytes)
    long bytes;
    {
       char *ptr;
       char *i;
       if (bytes<5) {
          bytes = 8;
          }
       status = heapNew (&ptr, bytes);
       if (status.all!=status_$ok){
           cleanupHandler(status);
       }
       return(ptr);
    }

    char *testcalloc(num_entries, bytes)
    long num_entries;
    long bytes;
    {
       char *ptr;
       long *i;
       long k,l;

       k = bytes*num_entries;
       if (k<5) {
          k = 8;
          }
       status = heapNew (&ptr, k);
       if (status.all!=status_$ok){
           cleanupHandler(status);
       }
       l = (k+3)>>2;
       for(i=(long *)ptr;l>0;i++, l--){
          *i=(long)0;
          }
       return(ptr);
    }

    char *testrealloc(ptr, bytes)
    char *ptr;
    long bytes;
    {
       char               *ptr2;
       int                oldSize;
       int                i;
       block_p            headPtr;
       trailer_p          tail;

       /* allocate new memory */
       ptr2 = testmalloc(bytes);

       /* get current size of ptr */
       headPtr = (block_p) ((int) ptr - sizeof(header_t));
       tail = (trailer_p) ( ((int)headPtr) - headPtr->head.size
	       - sizeof(trailer_t));
       if ((headPtr->head.size > 0) || (headPtr->head.size != tail->size)) {
          cleanupHandler(heap_bad_block);
       }
       /* copy only the smaller amount of two blocks */
       oldSize = - (headPtr->head.size) ;
       if( oldSize < bytes ){
	  bytes = oldSize ;
       }	

       for (i=0; i<bytes; i++){
           ptr2[i] = ptr[i];
       }


       safefree(ptr);

       return(ptr2);
    }

SIGNAL_FN cleanup(sigNum)
int sigNum ;
{

    cleanupHandler(-sigNum) ;

}

int cleanupHandler(status)
int status ;
{

    if (status > 0){
	printf("We have trapped our own error\n") ;
	switch(status){
	    case heap_bad_block:
               printf("Error = %0x Bad heap block\n",status) ;
	       break; 
	    case heap_no_mem:
               printf("Error = %0x No more memory to allocate\n",status) ;
	       break; 
	    default:
               printf("Error = %0x Unrecognized error code.\n",status) ;
	}

    } else {
	status = -status ;
	if( status == SIGINT ){
       	    printf("\nTimberwolfSC terminated by user\n\n") ;
	} else {
	    printf("\nSystem has found an error, we were unable to trap\n") ;
	    status = -status ;
	    printf("\nSystem status = %d\n",status) ;
	}
    }

    printf("Generating core dump for user traceback...\n") ;
    fflush(stdout) ;
    signal(SIGILL,SIG_DFL);
    abort() ;

}

int getCurMemUse()
{
    return(inUse) ;
}
