#include "standard.h"

#ifndef VMS
#include <sys/time.h>
#include <sys/resource.h>
#endif

print_stats(fp)
FILE *fp;
{
#if !defined(VMS) && !defined(hpux)
    struct rusage rusage;
    struct rlimit rlp;
    int text, data, vm_used, vm_limit, vm_soft_limit;
    double user, system, scale;
    char hostname[256];
    caddr_t p, sbrk();

    /* Get the hostname */
    gethostname(hostname, 256);
    hostname[256] = '\0';		/* just in case */

    /* Get the brk() value */
    p = sbrk(0);
    vm_used = ((int) p) / 1024.0 + 0.5;

    /* Get virtual memory limits */
    getrlimit(RLIMIT_DATA, &rlp);
    vm_limit = rlp.rlim_max / 1024.0 + 0.5;
    vm_soft_limit = rlp.rlim_cur / 1024.0 + 0.5;

    /* get usage stats */
    getrusage(RUSAGE_SELF, &rusage);
    user = rusage.ru_utime.tv_sec + rusage.ru_utime.tv_usec/1.0e6;
    system = rusage.ru_stime.tv_sec + rusage.ru_stime.tv_usec/1.0e6;
    scale = (user + system)*100.0;
    if (scale == 0.0) scale = 0.001;

    fprintf(fp, "\n\n");
    fprintf(fp, "Runtime Statistics\n");
    fprintf(fp, "-------------------------\n");
    fprintf(fp, "Machine name: %s\n\n", hostname);

    fprintf(fp, "User time   %6.1f seconds\n", user);
    fprintf(fp, "System time %6.1f seconds\n\n", system);

    text = rusage.ru_ixrss / scale + 0.5;
    data = (rusage.ru_idrss + rusage.ru_isrss) / scale + 0.5;
    fprintf(fp, "Average resident text size       = %5dK\n", text);
    fprintf(fp, "Average resident data+stack size = %5dK\n", data);
    fprintf(fp, "Maximum resident size            = %5dK\n", 
	rusage.ru_maxrss/2);
    fprintf(fp, "Virtual memory size              = %5dK\n", vm_used);
    fprintf(fp, "Virtual memory limit             = %5dK (%dK)\n\n", 
	vm_soft_limit, vm_limit);

    fprintf(fp, "Major page faults = %d\n", rusage.ru_majflt);
    fprintf(fp, "Minor page faults = %d\n", rusage.ru_minflt);
    fprintf(fp, "Swaps = %d\n\n", rusage.ru_nswap);

    fprintf(fp, "Input blocks = %d\n", rusage.ru_inblock);
    fprintf(fp, "Output blocks = %d\n\n", rusage.ru_oublock);

    fprintf(fp, "Context switch (voluntary) = %d\n", rusage.ru_nvcsw);
    fprintf(fp, "Context switch (involuntary) = %d\n", rusage.ru_nivcsw);
#else
    fprintf(fp, "Usage statistics not available\n");
#endif
}


double cpu_time()
{
#if !defined(VMS) && !defined(hpux)
    struct rusage rusage;
    double user, system;

    getrusage(RUSAGE_SELF, &rusage);
    user = rusage.ru_utime.tv_sec + rusage.ru_utime.tv_usec/1.0e6;
    system = rusage.ru_stime.tv_sec + rusage.ru_stime.tv_usec/1.0e6;
    return user+system;
#else
    return 0.0;
#endif
}
