/*
 * addObject.c
 *
 * Handles data structure maintenance of application "add thing" requests.
 */

#include "aceInternal.h"

/*
 * ace_create_matrix
 *
 * Called once per annealing, this routine sets up a matrix of the
 * desired size.
 * Input sanity check and first-call check already done.
 */
aceId
ace_create_matrix( rows, cols)
int rows, cols;
{
    aceRowCol *ace_new_rowcols();
    aceObject *oObj;
    acePlane *ace_new_plane(), *plane;
    int i;

    AceG.matrix.status |= ACE_MS_CREATED;
    AceG.matrix.size.row = rows;
    AceG.matrix.size.col = cols;

    AceG.rcC = rows + cols;

    /* allocate the RowCol edge arrays					*/
    AceG.matrix.rowRCs = ace_new_rowcols( rows);
    AceG.matrix.colRCs = ace_new_rowcols( cols);

    /* allocate the plane struct for each edge				*/
    plane = ace_new_plane();
    plane->range.upper = rows - 1;
    /* default net capacities						*/
    plane->netCap = ACE_DEFAULT_NET_CAPACITY;
    plane->cellCap = ACE_DEFAULT_CELL_CAPACITY;

    /* add plane as object						*/
    oObj = ace_new_object( plane );
    oObj->type = ACE_OBJ_PLANE;
    plane->id = ace_add_object( oObj);
    AceG.matrix.rowPlanes = plane;

    /* initialize the planeId fields					*/
    for( i = rows - 1; i >= 0; --i){
	oObj = ace_new_object(  &(AceG.matrix.rowRCs[ i ]) );
	oObj->type = ACE_OBJ_ROWCOL;
	AceG.matrix.rowRCs[ i ].id = ace_add_object( oObj );
	AceG.matrix.rowRCs[ i ].rowcol = i;
	AceG.matrix.rowRCs[ i ].isRow = TRUE;
	AceG.matrix.rowRCs[ i ].plane = plane;
    }

    plane = ace_new_plane();
    plane->range.upper = cols - 1;
    plane->netCap = ACE_DEFAULT_NET_CAPACITY;
    plane->cellCap = ACE_DEFAULT_CELL_CAPACITY;
    oObj = ace_new_object( plane );
    oObj->type = ACE_OBJ_PLANE;
    plane->id = ace_add_object( oObj);
    AceG.matrix.colPlanes = plane;

    /* initialize the planeId fields					*/
    for( i = cols - 1; i >= 0; --i){
	oObj = ace_new_object(  &(AceG.matrix.colRCs[ i ]) );
	oObj->type = ACE_OBJ_ROWCOL;
	AceG.matrix.colRCs[ i ].id = ace_add_object( oObj );
	AceG.matrix.colRCs[ i ].rowcol = i;
	AceG.matrix.colRCs[ i ].isRow = FALSE;
	AceG.matrix.colRCs[ i ].plane = plane;
    }

    AceG.matrix.rowPlanes->isRow = TRUE;
    AceG.matrix.colPlanes->isRow = FALSE;
    AceG.planeC = 2;

    /* there is only one MATRIX-type object				*/
    oObj = ace_new_object( &AceG.matrix );
    oObj->type = ACE_OBJ_MATRIX;

    AceG.matrix.id = ace_add_object( oObj);
    return( AceG.matrix.id );
}

/*
 * ace_add_master
 *
 * Register a new cell type (master).
 */
aceId
ace_add_master( name, path, rowCap, colCap)
char *name, *path;
int rowCap, colCap;
{
    aceMaster *master, *ace_new_master();
    aceObject *oMaster;

    master = ace_new_master();
    master->name = util_strsav( name);
    /* path is optional							*/
    if( path == NIL(char)){
	master->path = NIL(char);
    } else {
	master->path = util_strsav( path);
    }

    master->cellCap.row = rowCap;
    master->cellCap.col = colCap;

    oMaster = ace_new_object( master );
    oMaster->type = ACE_OBJ_MASTER;
    master->id = ace_add_object( oMaster );
    return( master->id );
}

/*
 * ace_add_cell
 *
 * Add an instance of oMaster at loc.
 * The arguments have already been checked for validity (ie: master exists
 *   and loc is in-bounds).
 */
aceId
ace_add_cell( oMaster, loc)
aceObject *oMaster;
aceCoord *loc;
{
    aceCell *cell, *ace_new_cell();
    aceObject *oCell;

    /* keep stats							*/
    ++AceG.cellC;

    /* initialize cell							*/
    cell = ace_new_cell();
    cell->master = (aceMaster *)(oMaster);
    cell->masterId = cell->master->id;
    cell->loc = *loc;

    /* add it as an object						*/
    oCell = ace_new_object( cell );
    oCell->type = ACE_OBJ_CELL;
    cell->id = ace_add_object( oCell);

    /* add it to the topographical net data structure			*/
    ace_insert_cell( cell, FALSE);
    ace_insert_cell( cell, TRUE);

    return( cell->id);
}

/*
 * ace_insert_cell
 *
 * This routine handles internal semantics of adding a cell at a point.
 * It places it in the correct net, creating nets if necessary.
 */
void
ace_insert_cell( cell, isRow)
aceCell *cell;
short isRow;
{
    aceRowCol *rc;
    aceNet *net, *ace_new_net();
    aceObject *oNet;
    aceCell **ace_new_cell_ptrs();
    int pos;

    if( isRow){
	rc = ACE_ROW_RC( cell->loc.row);
    } else {
	rc = ACE_COL_RC( cell->loc.col);
    }

    if( rc->netlist == NIL(aceNet)){
	/* keep stats							*/
	++AceG.netC;

	/* first net in this rowcol					*/
	net = ace_new_net();

	/* initialize the non-zero net fields				*/
	net->isRow = isRow;
	net->rowcol = isRow ? cell->loc.row : cell->loc.col;
	net->range.upper = ( isRow ?
			    AceG.matrix.size.col : AceG.matrix.size.row) - 1;
	net->cells = ace_new_cell_ptrs();

	/* add the net to the rowcol					*/
	rc->netlist = net;
	++(rc->count);

	/* add the net as an object					*/
	oNet = ace_new_object( net );
	oNet->type = ACE_OBJ_NET;
	net->id = ace_add_object( oNet);
    } else {
	/* search for net that contains the cell location		*/
	/*** assume that during set up, there is no overlapping and there is
	 * full net coverage (all locs on a rowcol are part of a unique net).
	 */
	pos = isRow ? cell->loc.col : cell->loc.row;
	ACE_GET_RCNET( rc, pos, net);
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
			"couldn't find %s net at (%d, %d)",
			isRow ? "horizontal" : "vertical",
			cell->loc.row, cell->loc.col);
	}
	/* make sure there's enough memory for new cell			*/
	if(( net->cells = REALLOC( aceCell *, net->cells, net->cellCount + 1))
		== NIL(aceCell *)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM,
			"no memory to realloc cell ptr array");
	}
    }
    net->cells[ (net->cellCount)++ ] = cell;

    if( isRow){
	cell->rowNet = net;
    } else {
	cell->colNet = net;
    }
}

/*
 * ace_object_type_string
 *
 * return the ascii name of the given type
 */
char *
ace_object_type_string( ot)
aceObjType ot;
{
    switch( ot){
	case ACE_OBJ_MATRIX:
		return "Matrix";
		break;
	case ACE_OBJ_PLANE:
		return "Plane";
		break;
	case ACE_OBJ_ROWCOL:
		return "Rowcol";
		break;
	case ACE_OBJ_NET:
		return "Net";
		break;
	case ACE_OBJ_MASTER:
		return "Master";
		break;
	case ACE_OBJ_CELL:
		return "Cell";
		break;
	case ACE_OBJ_UNKNOWN:
	default:
		return "Unknown";
		break;
    }
    return( NIL(char));
}
