/*
 * apAddCells.c
 *
 * Application interface routines to set up the initial matrix.
 * This is the heart of the procedural interface for set up.
 */

#include "aceInternal.h"

/*
 * aceAddCell
 *
 * Add a cell to a loction in the matrix or a plane. Return the cell's id.
 */
aceId
aceAddCell( master, row, col)
aceId master;			/* instance of what			*/
int row, col;			/* where (in full matrix coords)	*/
{
    aceObject *oMaster;
    aceId cellId, ace_add_cell();
    aceCoord loc;

    /* error handler set up						*/
    ACE_START( "aceAddCell");

    /* check arguments							*/
    if(( oMaster = ace_get_object( master)) == NIL(aceObject)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ID,
		"`master' is an corrupted id");
    }
    if( oMaster->type != ACE_OBJ_MASTER){
	errRaise( ACE_PKG_NAME, ACE_BAD_TYPE,
		"`master' id refers to an ace %s, not an ace master",
		ace_object_type_string( oMaster->type));
    }

    if(( row >= AceG.matrix.size.row) || (col >= AceG.matrix.size.col)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"(%d, %d) argument outside full matrix (%d, %d)",
		row, col, AceG.matrix.size.row, AceG.matrix.size.col);
    }

    /* now do the add							*/
    loc.row = row;
    loc.col = col;
    cellId = ace_add_cell( oMaster, &loc);

    ACE_END();

    return( cellId);
}

/*
 * aceAddMaster
 *
 * Register a new master.
 * Return a reference id number.
 */
aceId
aceAddMaster( name, path, rowCapacity, colCapacity)
char *name;			/* length 1 string name			*/
char *path;			/* optional path to master description	*/
int rowCapacity;		/* amount of row capacity used by an inst */
int colCapacity;		/* amount of col capacity used by an inst */
{
    char *errMsg;
    aceId id, ace_add_master();

    ACE_START( "aceAddMaster");

    /* check all arguments						*/
    errMsg = NIL(char);
    if( name == NIL(char)){
	errMsg = "`name' must be non-nil";
    } else if( strlen( name) != 1){
	errMsg = "`name' must be a length 1 string";
    } else if( rowCapacity < 0){
	errMsg = "`rowCapacity' must be non-negative";
    } else if( colCapacity < 0){
	errMsg = "`colCapacity' must be non-negative";
    }

    if( errMsg != NIL(char)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "%s", errMsg);
    }

    id = ace_add_master( name, path, rowCapacity, colCapacity);

    ACE_END();

    return( id );
}

/*
 * aceSplitNet
 *
 * Break the implict net connection in the specified direction at
 * the specified location of the specified thing.
 * Break is inserted bewteen breakpt and breakpt + 1.
 */
/*ARGSUSED*/
void
aceSplitNet( id, breakpt)
aceId id;			/* id of net to be broken		*/
int breakpt;			/* break point (in matrix coords)	*/
{
    ACE_START( "aceSplitNet");
    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR, "not implemented");
    ACE_END();
}

#define ACE_TRANS_ARRAY_SIZE 20
static aceCoord AceTransArray[ ACE_TRANS_ARRAY_SIZE ];

/*
 * aceTranslate
 *
 * translate from plane coords to matrix coords.
 */
/*ARGSUSED*/
aceCoord *
aceTranslate( id, row, col)
aceId id;			/* id of plane				*/
int row, col;			/* loc in plane coords			*/
{
    static index = 0;
    aceCoord *loc;
    acePlane *plane;

    ACE_START( "aceTranslate");
    loc = &(AceTransArray[ (index++) % ACE_TRANS_ARRAY_SIZE]);

    plane = (acePlane *)ace_get_object( id );
    if( plane->isRow ){
	loc->row = row + plane->range.lower;
	loc->col = col;
    } else {
	loc->row = row;
	loc->col = col + plane->range.lower;
    }

    ACE_END();
    return( loc );
}
