/*
 * apAnneal.c
 *
 * These routines implement an interface for the application to set annealing
 * parameters and to start the annealing process.
 */

#include "aceInternal.h"

/*
 * aceAnnealMatrix
 *
 * This routine starts the annealing process. Basically, once this routine
 * is called, ACE will run its simulated annealing algorithm until its done.
 * Then control will be returned to the application.
 */
/*ARGSUSED*/
void
aceAnnealMatrix( fastFlag )
unsigned fastFlag;      /* do fast annealing (quenching?) if set */
{
    ACE_START( "aceAnnealMatrix");

    /* run the annealing engine						*/
    ace_anneal( fastFlag );

    /* run the the constraint checking routines				*/
    ace_ucf_check_cnstr();

    /* run post-annealing cleanup before returning to the application	*/
    ace_clean_up();

    ACE_END();
}

/* for seeding the random number generator                              */
extern long time();

/*
 * aceRandomSeed
 *
 * This routine randomly seeds the random number generator.
 * The default is to seed it with 0 every time.
 */
void
aceRandomSeed()
{
    ACE_START( "aceRandomSeed");

    /* seed the random number generator                                 */
    (void) srandom( (int) time(( long *) 0));

    ACE_END();
}
