/*
 * apSetDefaults.c
 *
 * Application access to various parameters and functions used by ace.
 */

#include "aceInternal.h"

static aceDefaults apDefaults;

/*
 * aceGetDefaults
 *
 * Return a pointer to the defaults structure so the application
 * can access and alter it.  The values in the structure are ace's own
 * defaults if aceSetDefaults hasn't been called, or the set values set
 * by aceSetDefaults if it has been called.
 */
aceDefaults *
aceGetDefaults()
{
    ACE_START( "aceGetDefaults");

    ace_get_defaults( &apDefaults);

    ACE_END();

    return( &apDefaults);
}

/*
 * aceSetDefaults
 *
 * Set entries in the real defaults structure equal to the applications's copy.
 */
void
aceSetDefaults( def )
aceDefaults *def;	/* ptr to application's copy of defaults */
{
    ACE_START( "aceSetDefaults");

    if( def == NIL(aceDefaults)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "`def' is a NIL pointer");
    }

    if( def != &apDefaults){
	apDefaults = *def;
    }

    ace_set_defaults( def );

    ACE_END();
}
