/*
 * apFree.c
 *
 * The application function to free all memory associated with a matrix.
 */

#include "aceInternal.h"

/*
 * aceFreeMatrix
 *
 * This function frees every piece of memory associated with a matrix.
 * Since only one matrix is allowed at once, this function basically
 * turns off the ace package.
 */
/*ARGSUSED*/
void
aceFreeMatrix()
{
    int i;

    ACE_START( "aceFreeMatrix");
    /* free the memory associated with each object.  This includes every
     * net, cell, master, and plane. Also, this frees the object bookkeeping
     * memory space.
     */
    ace_free_objects();

    /* now take care of random stuff					*/
    /* Free both edge arrays						*/
    FREE( AceG.matrix.rowRCs);
    FREE( AceG.matrix.colRCs);

    for( i = AceG.matrix.size.row - 1; i >= 0; --i){
	FREE( AceG.matrix.map[i]);
    }
    FREE( AceG.matrix.map);

    /* if there were any UCFs, free them				*/
    ace_ucf_free_array();
    ace_ucf_efreeq();

    /* free the rmOverlap move scratch space
    ace_rmo_free();
    */

    /* reset globals							*/
    (void)memset((char *)&AceG, 0, sizeof( aceGlobals));

    ACE_END();
}
