/*
 * apTermNet.c
 *
 * Allow application to specify nets that must be routed to the edge of the cell
 * ( i.e., nets containing I/O terminals of the cell).
 */

#include "aceInternal.h"

/*
 * aceRegisterTermNet
 *
 * Make the specified net a terminal net.
 * `netId' specifies the net.
 * `mask' is one or both of the bits ACE_TERM_LOWER_VE and ACE_TERM_TWO_STATE.
 */
void
aceRegisterTermNet( netId, mask)
aceId netId;		/* id of net					*/
unsigned mask;		/* edge mask					*/
{
    aceObject *oNet;
    aceNet *net;
    static int size = 8;

    ACE_START( "aceRegisterTermNet");

    /* check args							*/
    if(( oNet = ace_get_object( netId )) == NIL(aceObject)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "netId corrupt");
    }
    if( oNet->type != ACE_OBJ_NET){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "netId isn't the id of a net");
    }
    if( mask & ~( ACE_TERM_LOWER_VE | ACE_TERM_TWO_STATE)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"illegal bits set in mask (0x%x)", mask);
    }

    /* args seem to be OK, so `just do it'.				*/
    net = (aceNet *)oNet;

    /* check to see if this net has already been made a term		*/
    if( net->termMask == 0){
	if( !AceG.matrix.termNetC ){
	    /* allocate the initial array of net pointers		*/
	    AceG.matrix.termNets = (aceNet **) ALLOC( aceNet *, size);
	} else if( AceG.matrix.termNetC == size ){
	    size <<= 1;
	    AceG.matrix.termNets = (aceNet **) REALLOC( aceNet *,
						AceG.matrix.termNets, size );
	}
	if( AceG.matrix.termNets == NIL(aceNet *)){
	    errRaise( ACE_PKG_NAME, ACE_NO_MEM, "no memory for termNet array");
	}
	/* add net into array						*/
	AceG.matrix.termNets[ AceG.matrix.termNetC ] = net;
	++AceG.matrix.termNetC;

	/* update RC							*/
	if( net->isRow ){
	    ++AceG.matrix.rowRCs[ net->rowcol ].termNetC;
	} else {
	    ++AceG.matrix.colRCs[ net->rowcol ].termNetC;
	}
    } else {
	(void)fprintf( stderr, "%s: Warning in aceRegisterTermNet: %s",
		ACE_PKG_NAME,
		"this net registered as a terminal multiple times.\n");
    }
    /* update net							*/
    net->termMask |= mask | ACE_TERM_IS_TERM;

    ACE_END();
}
