/*
 * apUCF.c
 *
 * Application functions to allow registration of User Constraint Functoins.
 */

#include "aceInternal.h"

/*
 * aceRegisterUCF
 *
 * Register a UCF. Specify the number of arguments, and the ids of
 * the arguments (in a user-memory array; ace makes its own copy).
 * For efficiency, set up the nets that, when moved, trigger a call
 * to this UCF registration (net dependancies).
 */
void
aceRegisterUCF( cnstr_fn, check_fn, eventMask,
		argCount, argIds, dependCount, dependIds)
double (* cnstr_fn)();	/* address of constraint function */
int (* check_fn)();	/* address of constraint checking function */
unsigned eventMask;	/* events that trigger call to this UCF */
int argCount;		/* number of entries in argIds[] */
aceId argIds[];		/* Ids of arguments to this UCF */
int dependCount;	/* number of entries in dependIds[] */
aceId dependIds[];	/* Ids of objects that trigger call to this UCF */
{
    char *errMsg;
    int ucfIndex;

    ACE_START( "aceRegisterUCF");

    /* a few sanity checks on the arguments				*/
    errMsg = NIL(char);
    if( cnstr_fn == (double (*)()) 0){
	errMsg = "`cnstr_fn' is a nil function pointer";
    } else if(( eventMask & ACE_UCFE_ALL) == 0){
	errMsg = "`eventMask' specifies no events";
    } else if( argCount < 0){
	errMsg = "`argCount' is negative";
    } else if(( argIds == NIL(aceId)) && (argCount != 0)){
	errMsg = "`argArray' is nil";
    } else if( dependCount < 0){
	errMsg = "`dependCount' is negative";
    } else if(( dependIds == NIL(aceId)) && (dependCount != 0)){
	errMsg = "`dependArray' is nil";
    }
    if( errMsg != NIL(char)){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "%s", errMsg);
    }

    /* store the UCF internally						*/
    ucfIndex = ace_add_ucf( cnstr_fn, check_fn, eventMask, argCount, argIds);

    /* update UCF move dependancies					*/
#ifdef ACE_PARANOID
    if( dependCount == 0){
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"dependCount == 0, UCF will never be called");
    }
#endif

    ace_make_depend( dependCount, dependIds, ucfIndex);

    ACE_END();
}
