/*
 * apCFlib.c
 *
 * A library of constraint functions, provided both for use
 * and for example.
 */

#include "ace.h"

/* penalty constants 							*/
#define ADJ_FIXED_COST	150.0
#define ADJ_DIST_COST	30.0

/*
 * aceUCFAdjNets
 *
 * A penalty function for keeping two nets adjacent.
 */
/*ARGSUSED*/
double
aceUCFAdjNets( count, array)
int count;
aceObject *(array[]);
{
    int dist;

#ifdef ACE_PARANOID
    if( count != 2){
	errRaise( ACE_PKG_NAME, -1, "aceUCFAdjNets: %d nets", count);
    }
#endif
    
    /* find the net's and there distance from each other		*/
    /*** Actually, there are two possible approaces. One is to have this be
     *   a two-valued function: 0 if adjacent, X if not adjacent. The other
     *   way is to have the penalty be bigger for lager separations of the
     *   nets. This makes sence if moves have a range limit, because
     *   then many moves may be required to put the nets back together again,
     *   plus, if the nets were separated, a move that made them more separated
     *   would be discouraged (a good thing, one would suspect).
     */
    dist = ((aceApNet *)array[0])->rowcol - ((aceApNet *)array[1])->rowcol;

#ifdef ACE_PARANOID
    if(( ((aceApNet *)array[0])->isRow && !((aceApNet *)array[1])->isRow) ||
       ( !((aceApNet *)array[0])->isRow && ((aceApNet *)array[1])->isRow)){
	errRaise( ACE_PKG_NAME, -1, "aceUCFAdjNets: orthogonal nets");
    }
#endif

    if( dist < 0){
	dist = -dist;
    }

    if( dist == 1) return( 0.0);
    if( dist == 0) return( ADJ_FIXED_COST );

    return( ADJ_FIXED_COST + dist * ADJ_DIST_COST);
}

#define LEN_COST 10.0

/*
 * aceUCFNetLen
 *
 * Make the lenght of a particular net, or set of nets, more costly.
 */
/*ARGSUSED*/
double
aceUCFNetLength( count, array)
int count;
aceObject *(array[]);
{
    int i, len;

    len = 0;
    for( i = count - 1; i >= 0; --i){
	len += ((aceApNet *)array[ i ])->upper -
	       ((aceApNet *)array[ i ])->lower;
    }
    len += count;
    return( LEN_COST * len);
}

#define ORDER_COST 40.0

/*
 * aceUCFNetOrder
 *
 * Constrain nets to be in a certain order.
 * This isn't too effecient now, but its an inherantly difficult
 * constraint to evaluate exactly.
 * Alg: given proper order is location array[i] < array[i+1] for i < count - 1
 *      for each array[i], 0 <= i <= count - 2
 *	    for each array[j], i < j <= count - 1
 *		penalty if loc array[j] < loc array[i]
 */
/*ARGSUSED*/
double
aceUCFNetOrder( count, array)
int count;
aceObject *(array[]);
{
    int i, j, pen;

    pen = 0;
    for( i = count - 2; i <= 0; --i){
	for( j = i + 1; j < count; ++j){
	    if( ((aceApNet *)array[ j ])->rowcol <
		((aceApNet *)array[ i ])->rowcol){
		++pen;
	    }
	}
    }
    return( pen * ORDER_COST);
}
