/*
 * cleanUp.c
 *
 * Clean up the matrix...
 **Get rid of any overlap,
 **check that constraints are satisfied,
 * remove any empty rows or columns,
 * and shift plane, net, and cell indicies around to make
 *	the matrix look contiguous.
 */

#include "aceInternal.h"

/*
 * ace_clean_up
 *
 * do said cleanup procdure
 */
void
ace_clean_up()
{
    acePlane *p;
    aceNet *net;
    int rc, rcoff, i;
    aceCoord loc;

    /* remove all empty rows and columns				*/
    if( ace_is_overlap( TRUE, 0, AceG.matrix.size.row - 1, &loc ) ){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_clean_up: found row overlap [%d,%d]", loc.row, loc.col);
    }
    if( ace_is_overlap( FALSE, 0, AceG.matrix.size.col - 1, &loc ) ){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_clean_up: found col overlap [%d,%d]", loc.row, loc.col);
    }
    for( p = AceG.matrix.rowPlanes; p != NIL(acePlane); p = p->next){
	(void)ace_rm_overlap_move( p );
    }
    for( p = AceG.matrix.colPlanes; p != NIL(acePlane); p = p->next){
	(void)ace_rm_overlap_move( p );
    }
    /* set all nets and cells to their locs without dead and empty rcs	*/
    p = AceG.matrix.rowPlanes;
    rcoff = p->deadC + p->emptyC;
    for( p = p->next; p != NIL(acePlane); p = p->next){
	for( rc = p->range.lower; rc <= p->range.upper; ++rc){
	    for( net = AceG.matrix.rowRCs[ rc ].netlist; net != NIL(aceNet);
			net = net->next){
		net->rowcol -= rcoff;
		for( i = net->cellCount - 1; i >= 0; --i){
		    net->cells[ i ]->loc.row -= rcoff;
		}
	    }
	}
	p->range.upper -= rcoff;
	p->range.lower -= rcoff;
	rcoff += p->deadC + p->emptyC;
    }

    p = AceG.matrix.colPlanes;
    rcoff = p->deadC + p->emptyC;
    for( p = p->next; p != NIL(acePlane); p = p->next){
	for( rc = p->range.lower; rc <= p->range.upper; ++rc){
	    for( net = AceG.matrix.colRCs[ rc ].netlist; net != NIL(aceNet);
			net = net->next){
		net->rowcol -= rcoff;
		for( i = net->cellCount - 1; i >= 0; --i){
		    net->cells[ i ]->loc.col -= rcoff;
		}
	    }
	}
	p->range.upper -= rcoff;
	p->range.lower -= rcoff;
	rcoff += p->deadC + p->emptyC;
    }
}
