/*
 * defaults.c
 *
 * Initialize ace's default values.
 */

#include "aceInternal.h"

static aceDefaults AceDefValues;

/* TRUE if AceDefValues has been assigned to				*/
static short defSet = FALSE;

/*
 * ace_get_defaults
 *
 * Copy ace's real defaults into the given structure.
 */
void
ace_get_defaults( def )
aceDefaults *def;
{
    ace_set_own_defaults();

    *def = AceDefValues;
}

/*
 * ace_get_default_ptr
 *
 * This is used internally by various ace modules to access the defaults
 * structure during initialization after control has been passed to
 * ace to begin the annealing.
 */
aceDefaults *
ace_get_default_ptr()
{
    ace_set_own_defaults(); /* make sure its been initialized		*/

    return( &AceDefValues );
}

/*
 * ace_set_defaults
 *
 * Copy the given defaults structure over the the real defaults.
 */
void
ace_set_defaults( def )
aceDefaults *def;
{
    if( def->cost.netLength.row >= 0){
	AceDefValues.cost.netLength.row = def->cost.netLength.row;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.netLength.row");
    }
    if( def->cost.netLength.col >= 0){
	AceDefValues.cost.netLength.col = def->cost.netLength.col;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.netLength.col");
    }
    if( def->cost.usedRC.row >= 0){
	AceDefValues.cost.usedRC.row = def->cost.usedRC.row;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.usedRC.row");
    }
    if( def->cost.usedRC.col >= 0){
	AceDefValues.cost.usedRC.col = def->cost.usedRC.col;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.usedRC.col");
    }
    if( def->cost.usedBias.row >= 0){
	AceDefValues.cost.usedBias.row = def->cost.usedBias.row;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.usedBias.row");
    }
    if( def->cost.usedBias.col >= 0){
	AceDefValues.cost.usedBias.col = def->cost.usedBias.col;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.usedBias.col");
    }
    if( def->cost.netCapacity.row >= 0){
	AceDefValues.cost.netCapacity.row = def->cost.netCapacity.row;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.netCapacity.row");
    }
    if( def->cost.netCapacity.col >= 0){
	AceDefValues.cost.netCapacity.col = def->cost.netCapacity.col;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.netCapacity.col");
    }
    if( def->cost.cellCapacity.row >= 0){
	AceDefValues.cost.cellCapacity.row = def->cost.cellCapacity.row;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.cellCapacity.row");
    }
    if( def->cost.cellCapacity.col >= 0){
	AceDefValues.cost.cellCapacity.col = def->cost.cellCapacity.col;
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"negative cost param: cost.cellCapacity.col");
    }

    *(AceDefValues.probFnVec) = *(def->probFnVec);
    AceDefValues.initTemp = def->initTemp;
    AceDefValues.updateTemp = def->updateTemp;
    AceDefValues.initStopInner = def->initStopInner;
    AceDefValues.stopInner = def->stopInner;
    AceDefValues.stopOuter = def->stopOuter;

    defSet = TRUE;
}

/*
 * ace_set_own_defaults
 *
 * Copy the starting values into the defaults structure.
 * This will execute the assignment to AceDefValues only once.
 */
void
ace_set_own_defaults()
{
    if( defSet ) return;

    defSet = TRUE;
    AceDefValues.cost.usedRC.row = 12.0;
    AceDefValues.cost.usedRC.col = 12.0;
    AceDefValues.cost.netLength.row = 2.0;
    AceDefValues.cost.netLength.col = 2.0;
    AceDefValues.cost.usedBias.row = 10.0;
    AceDefValues.cost.usedBias.col = 10.0;
    AceDefValues.cost.netCapacity.row = 60.0;
    AceDefValues.cost.netCapacity.col = 60.0;
    AceDefValues.cost.cellCapacity.row = 60.0;
    AceDefValues.cost.cellCapacity.col = 60.0;

    AceDefValues.probFnVec[ ACE_MOVE_SWAP ] = ace_move_prob_swap;
    AceDefValues.probFnVec[ ACE_MOVE_DISPLACE ] = ace_move_prob_disp;
    AceDefValues.probFnVec[ ACE_MOVE_FLIP_TERM ] = ace_move_prob_flip_term;
    AceDefValues.probFnVec[ ACE_MOVE_RM_OVERLAP ] = ace_move_prob_rm_overlap;

    ace_setup_scheduler( &AceDefValues );
}
