/*
 * generate.c
 *
 * Routines to generate specific contents of specific aceObject types.
 */

#include "aceInternal.h"

/*
 * ace_gen_plane_rowcols
 *
 * Generate the rowcols of a plane.
 */
int
ace_gen_plane_rowcols( gen, obj)
aceGenerator *gen;
aceObject **obj;
{
    aceGenState *state;
    acePlane *plane;
    aceRowCol *rc;

    ACE_START( "aceGenerate");
    state = (aceGenState *)gen->state;
    plane = (acePlane *)(state->obj);

    if( state->index1 == -1){
	/* first call							*/
	state->index1 = plane->range.lower;
    }

    /* generate next state						*/
    state->index1++;

    /* find next net, or find next rowcol and possibly finish generation */
    if( state->index1 > plane->range.upper){
	gen->free_fn( gen );
	gen->state = NIL(char);
	ACE_END();
	return( FALSE );
    }
    /* generate rowcol with this index					*/
    if( plane->isRow){
	rc = &( AceG.matrix.rowRCs[ state->index1 ]);
    } else {
	rc = &( AceG.matrix.colRCs[ state->index1 ]);
    }
    *obj = (aceObject *)rc;
    ACE_END();
    return( TRUE );
}

/*
 * ace_gen_rowcol_nets
 *
 * Generate the rowcols of a plane.
 */
int
ace_gen_rowcol_nets( gen, obj)
aceGenerator *gen;
aceObject **obj;
{
    aceGenState *state;
    aceRowCol *rc;
    aceNet *net;
    int i;

    ACE_START( "aceGenerate");
    state = (aceGenState *)gen->state;
    rc = (aceRowCol *)(state->obj);

    if( state->index1 == -1){
	/* first call							*/
	state->index1 = 0;
    }

    /* generate next state						*/
    state->index1++;

    /* check if generation is done					*/
    if( state->index1 > rc->count){
	gen->free_fn( gen );
	gen->state = NIL(char);
	ACE_END();
	return( FALSE );
    }

    /* generate next net						*/
    net = rc->netlist;
    for( i = state->index1 - 1; i > 0; i--){
	net = net->next;
	if( net == NIL(aceNet)){
	    errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
		"in ace_gen_rowcol_nets: net is NIL");
	}
    }
    *obj = (aceObject *)net;
    ACE_END();
    return( TRUE );
}

/*
 * ace_gen_net_cells
 *
 * Generate the cells of a net.
 */
int
ace_gen_net_cells( gen, obj)
aceGenerator *gen;
aceObject **obj;
{
    aceGenState *state;

    ACE_START( "aceGenerate");
    state = (aceGenState *)(gen->state);
    if( state->index1 == -1){
	/* first call							*/
	state->index1 = ((aceNet *)(state->obj))->cellCount;
    }
    /* produce next state						*/
    --( state->index1 );

    /* evaluate the state						*/
    if( state->index1 < 0){
	gen->free_fn( gen );
	gen->state = NIL(char);
	ACE_END();
	return( FALSE );
    } else {
	*obj = (aceObject *)(((aceNet *)(state->obj))->cells[ state->index1 ]);
	ACE_END();
	return( TRUE );
    }
}

/*
 * ace_gen_matrix_planes
 *
 * Generate the planes of a matrix.
 */
int
ace_gen_matrix_planes( gen, obj)
aceGenerator *gen;
aceObject **obj;
{
    aceGenState *state;
    aceMatrix *matrix;
    int i;
    acePlane *plane;

    ACE_START( "aceGenerate");

    state = (aceGenState *)gen->state;
    matrix = (aceMatrix *)(state->obj);

    if( state->index1 == -1){
	/* first call							*/
	state->index1 = 0;
	state->index2 = -1;
    }

    /* generate next state						*/
    state->index2++;

    /* get the plane							*/
    plane = matrix->rowPlanes;
    for( i = state->index2; ( i > 0) && (plane != NIL(acePlane)); i--){
	plane = plane->next;
    }
    if( plane != NIL(acePlane)){
	*obj = (aceObject *)plane;
    } else {
	plane = matrix->colPlanes;
	for( ; ( i > 0) && (plane != NIL(acePlane)); i--){
	    plane = plane->next;
	}
	if( plane != NIL(acePlane)){
	    *obj = (aceObject *)plane;
	} else {
	    return( FALSE );
	}
    }

    ACE_END();
    return( TRUE );
}

/*
 * ace_free_gen_status
 *
 * Free the memory of the status structure of a generator.
 */
void
ace_free_gen_status( gen)
aceGenerator *gen;
{
    if( gen != NIL(aceGenerator)){
	if( gen->state != NIL(char)){
	    FREE( gen->state );
	}
    } else {
	errRaise( ACE_PKG_NAME, ACE_BAD_ARG, "ace_free_gen_status: gen is NIL");
    }
}
