/*
 * move.c
 *
 * Functions to accept and reject moves
 *
 * WARNING: the undo functions [ eg, ace_replace_net, ace_unflip_term ] work
 *  closely with the delta cost functions [ eg, ace_displace_net,
 *  ace_flip_term ].  A change to one usually means a change to the other.
 */

#include "aceInternal.h"

/*
 * ace_accept_move
 *
 * This function is fairly trivial since to propose a move in ACE, the
 * state data structure must be changed to the next proposed state,
 * so to accept the move is to leave to state structure in its current state.
 * Also, this functions is used to clean up the UCF structures for the next
 * sate.
 */
void
ace_accept_move()
{
#ifndef ACE_NO_UCF
    aceUCFEvent *e;

    /* reset the `called' field and put the event in the free queue	*/
    while(( e = ace_ucf_egetq( &AceG.ucfCalledQ)) != NIL(aceUCFEvent)){
	AceG.ucfArray[ e->index ].called = FALSE;
	ace_ucf_free_e( e );
    }
#endif /* ACE_NO_UCF */
    AceG.comc.usedRC.row += AceG.comdc.usedRC.row;
    AceG.comc.usedRC.col += AceG.comdc.usedRC.col;
    AceG.comc.netLen.row += AceG.comdc.netLen.row;
    AceG.comc.netLen.col += AceG.comdc.netLen.col;
    AceG.comc.usedBias.row += AceG.comdc.usedBias.row;
    AceG.comc.usedBias.col += AceG.comdc.usedBias.col;
    AceG.comc.netCap.row += AceG.comdc.netCap.row;
    AceG.comc.netCap.col += AceG.comdc.netCap.col;
    AceG.comc.cellCap.row += AceG.comdc.cellCap.row;
    AceG.comc.cellCap.col += AceG.comdc.cellCap.col;

    /* update the ucfCost to the value due to this move			*/
    AceG.ucfCost = AceG.ucfNewCost;
}

/*
 * ace_reject_move
 *
 * This function undoes changes in the state data structure made
 * to evaluate the proposed move that was rejected.
 */
void
ace_reject_move()
{
#ifndef ACE_NO_UCF
    aceUCFEvent *e;
#endif /* ACE_NO_UCF */

    if( AceG.move.index == ACE_MOVE_DISPLACE){
	ace_replace_net( AceG.move.net0, -AceG.move.displacement, 0, TRUE);
    } else if( AceG.move.index == ACE_MOVE_SWAP){
	ace_replace_net( AceG.move.net1, AceG.move.displacement, 1, FALSE);
	ace_replace_net( AceG.move.net0, -AceG.move.displacement, 0, FALSE);
    } else if( AceG.move.index == ACE_MOVE_FLIP_TERM){
	ace_unflip_term_net( AceG.move.net0 );
    }

#ifndef ACE_NO_UCF
    while(( e = ace_ucf_egetq( &AceG.ucfCalledQ)) != NIL(aceUCFEvent)){
	/* clear the called field					*/
	AceG.ucfArray[ e->index ].called = FALSE;
	/* restore the previous last returned field			*/
	AceG.ucfArray[ e->index ].lastRet =
					AceG.ucfArray[ e->index ].restoreSave;
	/* put the event structure in the free queue			*/
	ace_ucf_free_e( e );
    }
#endif /* ACE_NO_UCF */
    /* set ucfNewCost back to to value before this move			*/
    AceG.ucfNewCost = AceG.ucfCost;
}
