/*
 * start.c
 *
 * Do one-time initialization and checking.
 */

#include "aceInternal.h"

/*
 * ace_start
 *
 * Do one-time initialization.
 */
void
ace_start()
{
    char *name;
    int depth;

    /* save these before zeroing everything				*/
    name = AceG.errName;
    depth = AceG.errDepth;

    /* zero out global and set started status bit			*/
    (void)memset((char *)&AceG, 0, sizeof( aceGlobals));
    AceG.matrix.status = ACE_MS_STARTED;

    /* reset them							*/
    AceG.errName = name;
    AceG.errDepth = depth;

    /* check data structure definitions					*/
    ace_check_ap_structs();

    /* set some default information					*/
    ace_set_own_defaults();

    /* seed the random number generator					*/
    (void)srandom( (int) 0);
}

/*
 * ace_check_ap_structs
 *
 * Make sure the offsets to fields in application data structures
 * are the same as offsets to the same fields in internal data structures.
 */
void
ace_check_ap_structs()
{
    aceNet net;
    aceApNet apnet;
    aceCell cell;
    aceApCell apcell;
    acePlane plane;
    aceApPlane applane;
    aceRowCol rc;
    aceApRowCol aprc;
    aceMaster mast;
    aceApMaster apmast;
    aceMatrix matrix;
    aceObject object;

    if(( (int)&net - (int)&( net.objNext)) !=
       ( (int)&apnet - (int)&( apnet.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.objType)) !=
       ( (int)&apnet - (int)&( apnet.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.id)) !=
       ( (int)&apnet - (int)&( apnet.id))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.termMask)) !=
       ( (int)&apnet - (int)&( apnet.termMask))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.isRow)) !=
       ( (int)&apnet - (int)&( apnet.isRow))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.rowcol)) !=
       ( (int)&apnet - (int)&( apnet.rowcol))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.range.lower)) !=
       ( (int)&apnet - (int)&( apnet.lower))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.range.upper)) !=
       ( (int)&apnet - (int)&( apnet.upper))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

    if(( (int)&cell - (int)&( cell.objNext)) !=
       ( (int)&apcell - (int)&( apcell.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.objType)) !=
       ( (int)&apcell - (int)&( apcell.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.id)) !=
       ( (int)&apcell - (int)&( apcell.id))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.loc.row)) !=
       ( (int)&apcell - (int)&( apcell.loc.row))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.loc.col)) !=
       ( (int)&apcell - (int)&( apcell.loc.col))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.masterId)) !=
       ( (int)&apcell - (int)&( apcell.masterId))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

    if(( (int)&plane - (int)&( plane.objNext)) !=
       ( (int)&applane - (int)&( applane.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.objType)) !=
       ( (int)&applane - (int)&( applane.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.id)) !=
       ( (int)&applane - (int)&( applane.id))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.range.lower)) !=
       ( (int)&applane - (int)&( applane.lower))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.range.upper)) !=
       ( (int)&applane - (int)&( applane.upper))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.isRow)) !=
       ( (int)&applane - (int)&( applane.isRow))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

    if(( (int)&rc - (int)&( rc.objNext)) !=
       ( (int)&aprc - (int)&( aprc.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.objType)) !=
       ( (int)&aprc - (int)&( aprc.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.id)) !=
       ( (int)&aprc - (int)&( aprc.id))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.rowcol)) !=
       ( (int)&aprc - (int)&( aprc.rowcol))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.isRow)) !=
       ( (int)&aprc - (int)&( aprc.isRow))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.count)) !=
       ( (int)&aprc - (int)&( aprc.netcount))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

    if(( (int)&mast - (int)&( mast.objNext)) !=
       ( (int)&apmast - (int)&( apmast.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.objType)) !=
       ( (int)&apmast - (int)&( apmast.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.id)) !=
       ( (int)&apmast - (int)&( apmast.id))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.name)) !=
       ( (int)&apmast - (int)&( apmast.name))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.path)) !=
       ( (int)&apmast - (int)&( apmast.path))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.cellCap.row)) !=
       ( (int)&apmast - (int)&( apmast.cellCap.row))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

    /* check object fields						*/
    if(( (int)&net - (int)&( net.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&net - (int)&( net.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&cell - (int)&( cell.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&plane - (int)&( plane.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&rc - (int)&( rc.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&mast - (int)&( mast.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&matrix - (int)&( matrix.objType)) !=
       ( (int)&object - (int)&( object.type))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }
    if(( (int)&matrix - (int)&( matrix.objNext)) !=
       ( (int)&object - (int)&( object.next))){
	errRaise( ACE_PKG_NAME, ACE_GENERAL_ERR,
	    "ace_check_ap_structs: struct offsets not equal");
    }

}
