/*
 * trace.c
 *
 * Output periodic trace info.
 */

#include "aceInternal.h"

#ifdef ACE_COMP_TRACE

static void ace_component_hdr();
static void ace_component_line();
static void ace_move_gen_hdr();
static void ace_move_gen_line();

/*
 * ace_trace
 *
 * Print out the desired trace data as specified by AceG.traceMask;
 * This routine is only called when a specfied frequency event happens
 * (ie, change temp state, gen move, etc.).  The cmask arg tells
 * why ace_trace was called (ie, it is one of the ACE_TRACEF_XXX bits).
 */
void
ace_trace( cmask )
unsigned cmask;		/* context mask to describe when trace is called */
{
    int i;
    char *name;
    long time;

    switch( cmask ){
	case ACE_TRACEF_INITIAL:
	    /* print out info about the state just completed		*/
	    (void)fprintf( AceG.traceFp,
			"Initial state:");
	    if( AceG.traceMask & ACE_TRACE_COST){
		(void)fprintf( AceG.traceFp, " Temp = %.2f", AceG.temp);
	    }
	    if( AceG.traceMask & ACE_TRACE_TEMP){
		(void)fprintf( AceG.traceFp, " Cost = %.2f", AceG.cost);
	    }
	    (void)fprintf( AceG.traceFp, "\n");
	    if( AceG.traceMask & ACE_TRACE_COM_COST){
		ace_component_hdr();
		ace_component_line( "used rowcols ", AceG.comc.usedRC.row,
		    (double)AceG.comc.usedRC.row * AceG.aparams.usedRC.row,
		    AceG.comc.usedRC.col,
		    (double)AceG.comc.usedRC.col * AceG.aparams.usedRC.col);
		ace_component_line( "rowcol bias  ", AceG.comc.usedBias.row,
		    (double)AceG.comc.usedBias.row * AceG.aparams.usedBias.row,
		    AceG.comc.usedBias.col,
		    (double)AceG.comc.usedBias.col * AceG.aparams.usedBias.col);
		ace_component_line( "net length   ", AceG.comc.netLen.row,
		    (double)AceG.comc.netLen.row * AceG.aparams.netLength.row,
		    AceG.comc.netLen.col,
		    (double)AceG.comc.netLen.col * AceG.aparams.netLength.col);
		ace_component_line( "net capacity ", AceG.comc.netCap.row,
		    (double)AceG.comc.netCap.row*AceG.aparams.netCapacity.row,
		    AceG.comc.netCap.col,
		    (double)AceG.comc.netCap.col*AceG.aparams.netCapacity.col);
		ace_component_line( "cell capacity", AceG.comc.cellCap.row,
		   (double)AceG.comc.cellCap.row*AceG.aparams.cellCapacity.row,
		    AceG.comc.cellCap.col,
		   (double)AceG.comc.cellCap.col*AceG.aparams.cellCapacity.col);
		ace_component_line( "ucf cost sum ", AceG.ucfCount,
		   AceG.ucfCost, 0, 0.0);
	    }
	    (void)fflush( AceG.traceFp);
	    break;
	case ACE_TRACEF_STATE:
	    /* print out info about the current state			*/
	    (void)fprintf( AceG.traceFp, "State %d:", AceG.astat.numStates);
	    if( AceG.traceMask & ACE_TRACE_TEMP){
		(void)fprintf( AceG.traceFp, " Temp = %.2f", AceG.temp);
	    }
	    if( AceG.traceMask & ACE_TRACE_COST){
		(void)fprintf( AceG.traceFp,
			" Cost = %.2f, DeltaCost = %.2f", AceG.cost,
			AceG.cost - AceG.astat.thisState.startCost);
	    }
	    (void)fprintf( AceG.traceFp, "\n");
	    if( AceG.traceMask & ACE_TRACE_COM_COST){
		ace_component_hdr();
		ace_component_line( "used rowcols ", AceG.comc.usedRC.row,
		    (double)AceG.comc.usedRC.row * AceG.aparams.usedRC.row,
		    AceG.comc.usedRC.col,
		    (double)AceG.comc.usedRC.col * AceG.aparams.usedRC.col);
		ace_component_line( "rowcol bias  ", AceG.comc.usedBias.row,
		    (double)AceG.comc.usedBias.row * AceG.aparams.usedBias.row,
		    AceG.comc.usedBias.col,
		    (double)AceG.comc.usedBias.col * AceG.aparams.usedBias.col);
		ace_component_line( "net length   ", AceG.comc.netLen.row,
		    (double)AceG.comc.netLen.row * AceG.aparams.netLength.row,
		    AceG.comc.netLen.col,
		    (double)AceG.comc.netLen.col * AceG.aparams.netLength.col);
		ace_component_line( "net capacity ", AceG.comc.netCap.row,
		    (double)AceG.comc.netCap.row*AceG.aparams.netCapacity.row,
		    AceG.comc.netCap.col,
		    (double)AceG.comc.netCap.col*AceG.aparams.netCapacity.col);
		ace_component_line( "cell capacity", AceG.comc.cellCap.row,
		   (double)AceG.comc.cellCap.row*AceG.aparams.cellCapacity.row,
		    AceG.comc.cellCap.col,
		   (double)AceG.comc.cellCap.col*AceG.aparams.cellCapacity.col);
		ace_component_line( "ucf cost sum ", AceG.ucfCount,
		   AceG.ucfCost, 0, 0.0);
	    }
	    if( AceG.traceMask & ACE_TRACE_GEN_MOVE){
		(void)fprintf( AceG.traceFp,
			"\t%d moves generated [ ",
			AceG.astat.thisState.numGen);
		if( AceG.traceMask & ACE_TRACE_ACC_MOVE){
		    (void)fprintf( AceG.traceFp, "%d accepted ",
			AceG.astat.thisState.numAcc);
		}
		if( AceG.traceMask & ACE_TRACE_REJ_MOVE){
		    (void)fprintf( AceG.traceFp, "%d rejected ",
			AceG.astat.thisState.numRej);
		}
		(void)fprintf( AceG.traceFp, "]\n");
	    }
	    if( AceG.traceMask & ACE_TRACE_MOVE_RESULTS){
		ace_move_gen_hdr();
		for( i = 0; i < ACE_MOVE_COUNT; ++i){
		    switch ( i ){
			case 0:
			    name = "swap nets    ";
			    break;
			case 1:
			    name = "displace net ";
			    break;
			case 2:
			    name = "flip term    ";
			    break;
			case 3:
			    name = "rm overlap   ";
			    break;
		    }
		    if(( AceG.mstat.thisState.move[ i ].accCnt + 
			    AceG.mstat.thisState.move[ i ].rejCnt) != 0){
			ace_move_gen_line( name,
			    (double)(( AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt) * 100 ) /
			    (double) AceG.astat.thisState.numGen,

			    (double)(AceG.mstat.thisState.move[i].accCnt * 100)/
			    (double) ( AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt),

			    (double)(AceG.mstat.thisState.move[i].rejCnt * 100)/
			    (double) (AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt),

			    AceG.mstat.thisState.move[ i ].accDC,
			    AceG.mstat.thisState.move[ i ].rejDC);
		    }
		}
	    }
	    if( AceG.traceMask & ACE_TRACE_CPU_TIME){
		time = util_cpu_time();
		(void)fprintf( AceG.traceFp, "\t%s elapsed CPU time\n",
			util_print_time( time - AceG.astat.thisState.cpuTime));
		AceG.astat.thisState.cpuTime = time;
	    }
	    (void)fflush( AceG.traceFp);
	    break;
	case ACE_TRACEF_FINAL:
	    /* print out info about the final state			*/
	    (void)fprintf( AceG.traceFp,
			"Final state:");
	    if( AceG.traceMask & ACE_TRACE_COST){
		(void)fprintf( AceG.traceFp, " Temp = %.2f", AceG.temp);
	    }
	    if( AceG.traceMask & ACE_TRACE_TEMP){
		(void)fprintf( AceG.traceFp, " Cost = %.2f DeltaCost = %.2f",
			AceG.cost, AceG.cost - AceG.astat.allStates.startCost);
	    }
	    (void)fprintf( AceG.traceFp, "\n");

	    if( AceG.traceMask & ACE_TRACE_COM_COST){
		ace_component_hdr();
		ace_component_line( "used rowcols ", AceG.comc.usedRC.row,
		    (double)AceG.comc.usedRC.row * AceG.aparams.usedRC.row,
		    AceG.comc.usedRC.col,
		    (double)AceG.comc.usedRC.col * AceG.aparams.usedRC.col);
		ace_component_line( "rowcol bias  ", AceG.comc.usedBias.row,
		    (double)AceG.comc.usedBias.row * AceG.aparams.usedBias.row,
		    AceG.comc.usedBias.col,
		    (double)AceG.comc.usedBias.col * AceG.aparams.usedBias.col);
		ace_component_line( "net length   ", AceG.comc.netLen.row,
		    (double)AceG.comc.netLen.row * AceG.aparams.netLength.row,
		    AceG.comc.netLen.col,
		    (double)AceG.comc.netLen.col * AceG.aparams.netLength.col);
		ace_component_line( "net capacity ", AceG.comc.netCap.row,
		    (double)AceG.comc.netCap.row*AceG.aparams.netCapacity.row,
		    AceG.comc.netCap.col,
		    (double)AceG.comc.netCap.col*AceG.aparams.netCapacity.col);
		ace_component_line( "cell capacity", AceG.comc.cellCap.row,
		   (double)AceG.comc.cellCap.row*AceG.aparams.cellCapacity.row,
		    AceG.comc.cellCap.col,
		   (double)AceG.comc.cellCap.col*AceG.aparams.cellCapacity.col);
		ace_component_line( "ucf cost sum ", AceG.ucfCount,
		   AceG.ucfCost, 0, 0.0);
	    }

	    if( AceG.traceMask & ACE_TRACE_GEN_MOVE){
		(void)fprintf( AceG.traceFp,
			"\t%.2d moves generated [ ",
			AceG.astat.allStates.numGen);
		if( AceG.traceMask & ACE_TRACE_ACC_MOVE){
		    (void)fprintf( AceG.traceFp, "%d accepted ",
			AceG.astat.allStates.numAcc);
		}
		if( AceG.traceMask & ACE_TRACE_REJ_MOVE){
		    (void)fprintf( AceG.traceFp, "%d rejected ",
			AceG.astat.allStates.numRej);
		}
		(void)fprintf( AceG.traceFp, "]\n");
	    }

	    if( AceG.traceMask & ACE_TRACE_MOVE_RESULTS){
		ace_move_gen_hdr();
		for( i = 0; i < ACE_MOVE_COUNT; ++i){
		    switch ( i ){
			case 0:
			    name = "swap nets    ";
			    break;
			case 1:
			    name = "displace net ";
			    break;
			case 2:
			    name = "flip term    ";
			    break;
			case 3:
			    name = "rm overlap   ";
			    break;
		    }
		    if(( AceG.mstat.thisState.move[ i ].accCnt + 
			    AceG.mstat.thisState.move[ i ].rejCnt) != 0){
			ace_move_gen_line( name,
			    (double)(( AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt) * 100 ) /
			    (double) AceG.astat.thisState.numGen,

			    (double)(AceG.mstat.thisState.move[i].accCnt * 100)/
			    (double) ( AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt),

			    (double)(AceG.mstat.thisState.move[i].rejCnt * 100)/
			    (double) (AceG.mstat.thisState.move[ i ].accCnt +
			    AceG.mstat.thisState.move[ i ].rejCnt),

			    AceG.mstat.thisState.move[ i ].accDC,
			    AceG.mstat.thisState.move[ i ].rejDC);
		    }
		}
	    }
	    if( AceG.traceMask & ACE_TRACE_CPU_TIME){
		(void)fprintf( AceG.traceFp, "\t%s total CPU time\n",
			util_print_time( util_cpu_time() -
					AceG.astat.allStates.cpuTime));
	    }
	    (void)fflush( AceG.traceFp);
	    break;
	case ACE_TRACEF_ACC_MOVE:
	case ACE_TRACEF_REJ_MOVE:
	    /* print out (brief!) info about the move just generated	*/
	    if( AceG.astat.thisState.numGen == 1){
		/* flush all last state's data				*/
		(void)fflush( AceG.traceFp);
	    }
	    /* add trace data for this move				*/
	    if( AceG.move.index == ACE_MOVE_DISPLACE){
		(void)fprintf( AceG.traceFp, "displace %s net at %d by %d",
		    AceG.move.net0->isRow ? "horizontal" : "vertical",
		    (cmask & ACE_TRACEF_ACC_MOVE) ?
			(AceG.move.net0->rowcol - AceG.move.displacement) :
			(AceG.move.net0->rowcol),
		    AceG.move.displacement);
	    } else if( AceG.move.index == ACE_MOVE_SWAP){
		(void)fprintf( AceG.traceFp, "%s swap nets at %d and %d",
		    AceG.move.net0->isRow ? "horizontal" : "vertical",
		    (cmask & ACE_TRACEF_ACC_MOVE) ?
			(AceG.move.net0->rowcol - AceG.move.displacement) :
			(AceG.move.net0->rowcol),
		    (cmask & ACE_TRACEF_ACC_MOVE) ?
			(AceG.move.net1->rowcol - AceG.move.displacement) :
			(AceG.move.net1->rowcol) );
	    } else if( AceG.move.index == ACE_MOVE_RM_OVERLAP){
		(void)fprintf( AceG.traceFp, "remove overlap");
	    }
	    (void)fprintf( AceG.traceFp, ": DC is %.2f", AceG.delC);
	    if( cmask & ACE_TRACEF_ACC_MOVE){
		(void)fprintf( AceG.traceFp, ": accepted\n");
	    } else {
		(void)fprintf( AceG.traceFp, ": rejected\n");
	    }
	    break;
	default:
	    /* bad call to ace_trace					*/
	    errRaise( ACE_PKG_NAME, ACE_BAD_ARG,
		"ace_trace: unknown context mask 0x%x", cmask);
	    break;
    }
}

/*
 * ace_component_hdr
 *
 * Print out the header for the cost component trace table.
 */
static void
ace_component_hdr()
{
    (void)fprintf( AceG.traceFp,
	"\tComponents   : Row count,     cost  | Col count,     cost\n");
}


/*
 * ace_component_table
 *
 * Print out lines in the cost component trace table.
 */
static void
ace_component_line( name, rcount, rcost, ccount, ccost)
char *name;
int rcount, ccount;
double rcost, ccost;
{
    (void)fprintf( AceG.traceFp,
	"\t%s|     %5d, %10.2f|     %5d, %10.2f\n",
	name, rcount, rcost, ccount, ccost);
}

/*
 * ace_move_gen_hrd
 *
 * Print the move generation table header line.
 */
static void
ace_move_gen_hdr()
{
    (void)fprintf( AceG.traceFp,
       "\tMoves        :   Gen%%   Acc%%   Rej%%      AccDC       RejDC\n");
}

/*
 * ace_move_gen_line
 *
 * Print a line in the move generation table.
 */
static void
ace_move_gen_line( name, genp, accp, rejp, accdc, rejdc)
char *name;
double genp, accp, rejp, accdc, rejdc;
{
    (void)fprintf( AceG.traceFp,
	"\t%s|  %5.1f  %5.1f  %5.1f  %10.2f  %10.2f\n",
	name, genp, accp, rejp, accdc, rejdc);
}

/*
 * ace_init_trace_time
 *
 * Initialize the cpu time trace.
 */
void
ace_init_trace_time()
{
    AceG.astat.allStates.cpuTime = AceG.astat.thisState.cpuTime =
		util_cpu_time();
}

#endif /* ACE_COMP_TRACE */
