#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "st.h"



/*
 *   MACRO DEFINITIONS
 *
 *   Macros are distinguished by using solely capital letters in their
 *   identifiers.  This contrasts with C defined identifiers which are strickly 
 *   lower case, and program variable and procedure names which use both upper 
 *   and lower case.
 */

/* Begin macros. */

/* Boolean data type */
#define  BOOLEAN	int
#define  FALSE		0
#define  TRUE		1
#define  NOT		!
#define  AND		&&
#define  OR		||

#define YES  	1
#define NO	0
#ifndef NULL
#define NULL	0
#endif
#define OK   1
#define NOT_OK 0
#define LEFT 0
#define RIGHT 1
#define TOP 2
#define BOTTOM 3

char _errBuffer[64];
#define OCT_ASSERT(fct) \
    if ((fct) < OCT_OK) {\
	(void) sprintf(_errBuffer,"An OCT assertion failed: %s line %d", \
		__FILE__, __LINE__ );\
	octError(_errBuffer);\
	exit(-1);\
    }

#define NIL(type)   ((type *) 0)

/* Macro commands */
/* Macro functions that return the maximun or minimum independent of type. */
#define  MAX(a,b)           ((a) > (b) ? (a) : (b))
#define  MIN(a,b)           ((a) < (b) ? (a) : (b))

/* Macro function that returns the absolute value of a floating point number. */
#define  ABS(a)             ((a) < 0.0 ? -(a) : (a))

/* Macro procedure that swaps two entities. */
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}


/* Memory allocation and dealocation macros. */
#define MALLOC(type,number)  (type *)malloc((unsigned)(sizeof(type)*(number)))
#define CALLOC(type,number)  (type *)calloc(sizeof(type),(unsigned)(number))
#define REALLOC(ptr,type,number)  \
           ptr = (type *)realloc((char *)ptr,(unsigned)(sizeof(type)*(number)))
#define FREE(pointer)  if ((pointer) != NULL)  \
	   { free((char *)pointer); (pointer)=NULL; }

#define strsav(s)		strcpy(MALLOC(char,strlen(s)+1), s)



/*
 *  GLOBAL VARIABLE DECLARATION 
 *
 *  This variable is global to all matrix routines.  The small g prefix 
 *  denotes this.  Making this variable global greatly reduces the amount of 
 *  parameter passing required.
 *
 *  >>> Global variable:
 */

/* Begin global declarations. */

extern octObject	Chip;
extern char		buf[];
extern st_table *st_masters;

extern short verbose;

typedef struct dir {
    char *path;
    struct dir *next;
} dirType;

typedef struct viewList {
    char *view;
    char *name;
    short validInterface;
    short IOequivalent;		/* to the original (the first) view */
    short soft;
    short currentView;		/* flag TRUE if this is the current view */
    double area;
    octCoord w;			/* width */
    octCoord h;			/* height */
    double HoverW;		/* aspect ratio */
    double aspectRatio;		/* obsolete */
    double totalRMSpower;	/* future expansion */
    double totalRMScurrent;	/* future expansion */
    struct viewList *prec;
    struct viewList *next;
} viewListType;

typedef struct termElement {
  char *name;
  char *type;
  octObject *Term;
  struct termElement *prec;
  struct termElement *next;
  struct termElement *equivPrec; /* pointer to electrically equiv pin */
  struct termElement *equivNext; /* pointer to electrically equiv pin */
} termElement;




extern	octObject	LayerPlace; 
extern	octObject	LayerMET1; 

extern	octObject	Mobility;
extern  octObject       FloorplanBag; /* This is a bag */
extern  octObject	ModuleGenerator;
extern	octObject	MinHeightOverWidth;
extern  octObject	MaxHeightOverWidth;
extern  octObject       FloatingEdge;
extern  octObject	TermEdge;
extern	octObject	TermType;
extern	octObject	SideRestrictions;
extern	octObject	JoinedTerms;
extern	octObject	OrderTerms;
extern	octObject	PermTerms;
extern	octObject	NetType;
extern	octObject	Criticality;

extern	octObject	AvailableArea;


extern char *TermTypeString[]; 
extern char *NetTypeString[]; 


#define  VEM_MIN_FEATURE 20 
