
#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "errtrap.h"
#include "oh.h"
#include "nle.h"
#include "vov.h"
#include <sys/dir.h>

int isSoft( instance )
    octObject* instance;
{
    octObject master;
    octObject bag;
    OH_ASSERT( ohOpenMaster( &master, instance, "interface", "r" ) );
    VOVinputFacet( &master );
    return ( ohGetByBagName( &master, &bag, "FLOORPLAN" ) == OCT_OK );
}


void listInstances( chipP, prefix, softOnly)
    octObject* chipP;
    char* prefix;
    int softOnly;
{
    octObject bag;
    octGenerator genInst;
    octObject instance;
    char  cellview[1024];

    if ( ohGetByBagName( chipP, &bag, "INSTANCES" ) != OCT_OK ) {
	errRaise( "alba", 1, "No INSTANCES bag" );
    }
    OH_ASSERT( octInitGenContents( &bag, OCT_INSTANCE_MASK, &genInst ) );
    
    if ( prefix == 0 ) prefix = "";
    while ( octGenerate( &genInst, &instance ) == OCT_OK ) {
	if ( softOnly && !isSoft( &instance ) ) continue;

	sprintf( cellview, "%s:%s:%s",
		instance.contents.instance.master,
		instance.contents.instance.view,
		instance.contents.instance.facet
		);
	printf( prefix, cellview );
	printf( "\n" );
    }
}


alba_checkInstances()
{
  octObject		InstanceBag, *InstanceContainer;


  if (verbose) printf("---------  Checking instances  -------------\n");
  /* look for the instance bag */
  InstanceBag.type = OCT_BAG;
  InstanceBag.contents.bag.name = "INSTANCES";
  if (octGetByName(&Chip,&InstanceBag) == OCT_NOT_FOUND) {
    octError("looking for INSTANCES bag:\n");
    (void) sprintf(buf, "Warning: The INSTANCES bag is missing.\n");
    Message(buf);
    (void) sprintf(buf,"   Continuing generating instances from facet\n");
    Message(buf);
    InstanceContainer = &Chip;
  } else {
    InstanceContainer = &InstanceBag;
  }
  if (albaCheckInstances(InstanceContainer)!=NLE_OK) {
    Warning("Problems checking instances");
  }
}

int
albaCheckInstances(Bag)
    octObject *Bag;
{
  octObject		Instance;
  octGenerator		genInst;


  OCT_ASSERT(octInitGenContents(Bag,  OCT_INSTANCE_MASK, &genInst));

  while (octGenerate(&genInst,&Instance)==OCT_OK) {
    OCT_ASSERT(alba_checkViews(&Instance));
  }

  return(NLE_OK);
}

/*
 *    alba_checkViews(Instance)
 *    
 *    RETURNS 
 *    SIDE_EFFECTS
 *      
 *    GLOBAL VARIABLES 
 */
octStatus
alba_checkViews(Instance)
     octObject	*Instance;
{
  DIR 		*dirp;		/* directory pointer */
  struct direct *dp;		/* directory entry */
  octObject   	Master, View, Interface;
  octCoord    	h,w;
  struct octBox BBox;
  short 	IOequivalent();
  char  	 *selectNewView();
  char  	command[256];
  viewListType	*firstViewPtr = NULL;
  viewListType  *lastViewPtr  = NULL;
  viewListType	*viewPtr;

  if (verbose ) printf(" ****   %12s  of %s:%s\n",
		       Instance->contents.instance.name,
		       Instance->contents.instance.master,
		       Instance->contents.instance.view);

  Master.type = OCT_FACET;
  Master.contents.facet.cell    = Instance->contents.instance.master;
  Master.contents.facet.view    = Instance->contents.instance.view;
  Master.contents.facet.facet   = "contents";
  Master.contents.facet.version = Instance->contents.instance.version;
  Master.contents.facet.mode = "r";

  if (verbose ) printf(" ****   %12s  of %s:%s\n",
		       Instance->contents.instance.name,
		       Instance->contents.instance.master,
		       Instance->contents.instance.view);

  if (octOpenFacet(&Master) < OCT_OK) FatalError("opening master contents");
  VOVinputFacet( &Master );
  View.type = OCT_FACET;
  View.contents.facet.cell    = Instance->contents.instance.master;
  View.contents.facet.facet   = "contents";
  View.contents.facet.version = Instance->contents.instance.version;
  View.contents.facet.mode = "r";
  Interface.type = OCT_FACET;
  Interface.contents.facet.cell    = Instance->contents.instance.master;
  Interface.contents.facet.facet   = "interface";
  Interface.contents.facet.version = Instance->contents.instance.version;
  Interface.contents.facet.mode = "r";

  /*  read the directory of the Master listing all the valid views */
  if ((dirp = opendir(Instance->contents.instance.master)) == NULL) {
    (void) sprintf(buf,"Error: Could not  open directory %s\n",
	    Instance->contents.instance.master);
    Warning(buf);
    return(OCT_OK);
  }
  /* skip the first two entries, "." and ".." */
  dp = readdir(dirp);  dp = readdir(dirp);
  /* check each entry in this directory to see if it is a legal VIEW */
  while ((dp = readdir(dirp)) != NULL) {
    if (verbose) printf("Checking %s:%s\n",Instance->contents.instance.master , dp->d_name);
    View.contents.facet.view = dp->d_name;
    Interface.contents.facet.view = dp->d_name;
    if (octOpenFacet(&View) < OCT_OK) {
      (void) sprintf(buf,"Extraneous directory %s/%s\n",
	      Instance->contents.instance.master, dp->d_name);
      Message(buf);
      if (strcmp(View.contents.facet.view,Instance->contents.instance.view)==0)
	/* because it is the same as Master */
	FatalError("This view must be correct");
    } else {
      /* found a legal OCT view */
      viewPtr = MALLOC(viewListType,1);
      viewPtr->next = NULL;
      viewPtr->prec = NULL;
      viewPtr->name = strsav(dp->d_name);
      viewPtr->currentView = (strcmp(viewPtr->name,Instance->contents.instance.view)==0)?TRUE:FALSE;
      viewPtr->IOequivalent = IOequivalent(&Master,&View);
      /* check the interface (must be opened first) */
      if (octOpenFacet(&Interface) < OCT_OK) {
	/*  PROBLEM OPENING THE INTERFACE */
	/*  try to create the interface using vulcan */
	viewPtr->validInterface = FALSE;
	(void) sprintf(command, "vulcan -k %s:%s",
		Interface.contents.facet.cell,
		Interface.contents.facet.view);
	if (verbose) {	/* report the error */
	  octError("\n");
	  (void) sprintf(buf,"Creating interface facet with: %s\n", command);
	  Message(buf);
	}
	if (system(command) == 127) {
	  (void) sprintf(buf,"\nCould not execute the command %s\n", command);
	  Message(buf);
	} else {
	  /* assume Vulcan is always successfull */
	  viewPtr->validInterface == TRUE;
	}
      } else {
	viewPtr->validInterface = IOequivalent(&View, &Interface);
      }
      /*
       *  get all the information from the interface
       */
      if (octGetByName(&Interface,&FloorplanBag) == OCT_NOT_FOUND) {
	viewPtr->soft = FALSE;
      } else {
	viewPtr->soft = TRUE;
      }
      if (octBB(&Interface,&BBox) == OCT_OK) {
	viewPtr->w = w = (BBox.upperRight.x - BBox.lowerLeft.x) / VEM_MIN_FEATURE;
	viewPtr->h = h = (BBox.upperRight.y - BBox.lowerLeft.y) / VEM_MIN_FEATURE;
	viewPtr->area = w * h ;
	viewPtr->HoverW = viewPtr->aspectRatio = (double) h / (double) w;
      } else {
	viewPtr->area = 0.0;
	viewPtr->w = viewPtr->h = 0;
	viewPtr->HoverW = viewPtr->aspectRatio = 0.0;
      }
      /* append at the end of the list */
      if (firstViewPtr == NULL) {
	firstViewPtr = lastViewPtr = viewPtr;
      } else {
	viewPtr->prec = lastViewPtr;
	lastViewPtr->next = viewPtr;
	lastViewPtr = viewPtr;
      }
    }
  }
  closedir(dirp);

  return(OCT_OK);
}


short
IOequivalent(FacetA, FacetB)
     octObject *FacetA, *FacetB;
{
  termElement *termsOfA, *termsOfB;
  termElement *a, *b;
  termElement *equivA, *equivB, *firstA , *firstB;
  termElement *listFormalTerminals();
  short IOeq = TRUE;
  short found;

  termsOfA = listFormalTerminals(FacetA);
  termsOfB = listFormalTerminals(FacetB);

  a = termsOfA;
  b = termsOfB;
  
  while( a != NULL AND b != NULL) {
    if (strcmp(a->name, b->name)==0) {
  /*       if (verbose) printf("Term \t%12s\tOK\n", a->name); */
      a = a->next;
      b = b->next;
    } else {
      /*
       ** may be either a or b are not terminals of type SUPPLY or GROUND
       ** in which case they can be ignored at this time
       */
      if (strcmp(a->type,"SUPPLY")== 0 OR strcmp(a->type,"GROUND")==0 ) {
	a = a->next;
	continue;		/* while loop */
      }
      if (strcmp(b->type,"SUPPLY")== 0 OR strcmp(b->type,"GROUND")==0 ) {
	b = b->next;
	continue;		/* while loop */
      }
      
      /* may be the two terminals have equivalent pins */
      found = FALSE;
      firstB = NULL;
      equivB = b;
      while(found==FALSE && equivB != firstB && equivB != NULL) {
	if (firstB==NULL) firstB = equivB;
	firstA = NULL;
	equivA = a;
	while (found == FALSE && equivA!=firstA && equivA != NULL) {
	  if (firstA==NULL) firstA = equivA;
	  if (strcmp(equivA->name,equivB->name)==0)    found = TRUE ;
	  if (found == FALSE) equivA=equivA->equivNext;
	}
	if (found==FALSE) equivB=equivB->equivNext;
      }
      if (found) {
	if (verbose) {
	  (void) sprintf(buf,"Term %s of %s:%s <--> Term %s of %s:%s\n", a->name,
		 FacetA->contents.facet.view,
		 FacetA->contents.facet.facet,
		 b->name,
		 FacetB->contents.facet.view,
		 FacetB->contents.facet.facet);
	  Message(buf);
	  (void) sprintf(buf,"\t by virtue of electrically equivalent terminals\n");
	  Message(buf);
	}
	a = a->next;
	b = b->next;
      } else if (strcmp(a->name, b->name)<0) {
	(void) sprintf(buf,"Term %12s of %s:%s is NOT in %s:%s\n",
	       a->name,
	       FacetA->contents.facet.view,
	       FacetA->contents.facet.facet,
	       FacetB->contents.facet.view,
	       FacetB->contents.facet.facet);
	Message(buf);
	IOeq = FALSE;
	a = a->next;
      } else {
	(void) sprintf(buf,"Term %12s of  %s:%s is NOT in %s:%s\n",
	       b->name,
	       FacetB->contents.facet.view,
	       FacetB->contents.facet.facet,
	       FacetA->contents.facet.view,
	       FacetA->contents.facet.facet);
	Message(buf);
	IOeq = FALSE;
	b = b->next;
      }
    }
  }
  /* are there any terminals left over ? */
  while (a != NULL) {
    (void) sprintf(buf,"Term %12s of %s:%s is NOT in %s:%s\n",
	   a->name,
	   FacetA->contents.facet.view,
	   FacetA->contents.facet.facet,
	   FacetB->contents.facet.view,
	   FacetB->contents.facet.facet);
    Message(buf);
    IOeq = FALSE;
    a = a->next;
  }
  while (b != NULL) {
    (void) sprintf(buf,"Term %12s of  %s:%s is NOT in %s:%s\n",
	   b->name,
	   FacetB->contents.facet.view,
	   FacetB->contents.facet.facet,
	   FacetA->contents.facet.view,
	   FacetA->contents.facet.facet);
    Message(buf);
    IOeq = FALSE;
    b = b->next;
  }
  return(IOeq);
}


/*  returns list of all formal terminals sorted by name */
termElement *
listFormalTerminals(Facet)
     octObject *Facet;
{
  termElement *firstTermPtr = NULL;
  termElement *lastTermPtr = NULL;
  termElement *newTermPtr, *termPtr;
  octGenerator genTerm;
  octObject Term;
  octObject TermType;

  TermType.type = OCT_PROP;
  TermType.contents.prop.name = "TERMTYPE";

  OCT_ASSERT(octInitGenContents(Facet,OCT_TERM_MASK,&genTerm));
  while (octGenerate(&genTerm,&Term) == OCT_OK) {
    if (Term.contents.term.instanceId == 0) {
      newTermPtr = MALLOC(termElement,1);
      newTermPtr->next = NULL;
      newTermPtr->prec = NULL;
      newTermPtr->equivNext = NULL;
      newTermPtr->equivPrec = NULL;
      /* fill up all the fields in the record of this terminal */
      newTermPtr->name = Term.contents.term.name;
      if (octGetByName(&Term,&TermType) == OCT_NOT_FOUND) {
	/* default type is SIGNAL */
	newTermPtr->type = "SIGNAL";
      } else {
	newTermPtr->type = strsav(TermType.contents.prop.value.string);
      }
      /* insert in the list in alphabetical order */
      if (firstTermPtr == NULL) {
	firstTermPtr =  lastTermPtr = newTermPtr;
      } else if (strcmp(firstTermPtr->name,newTermPtr->name)>=0) {
	/*  insert at the beginning of the list */
	newTermPtr->next = firstTermPtr;
	firstTermPtr->prec = newTermPtr;
	firstTermPtr= newTermPtr;
      } else if  (strcmp(lastTermPtr->name,newTermPtr->name)<0) {
	/* insert at the end of the list */
	newTermPtr->prec = lastTermPtr;
	newTermPtr->prec->next = newTermPtr;
	lastTermPtr = newTermPtr;
      } else {
	termPtr = firstTermPtr;
	while (termPtr != NULL AND strcmp(termPtr->name, newTermPtr->name)<0){
	  termPtr = termPtr->next;
	}
	if (termPtr == NULL) abort(); /* debug stoppoint */
	  
	newTermPtr->next = termPtr;
	newTermPtr->prec = termPtr->prec;
	newTermPtr->next->prec = newTermPtr;
	newTermPtr->prec->next = newTermPtr;
      }
    }
  }
  findEquivalentTerms(Facet, firstTermPtr);
  return(firstTermPtr);
}




findEquivalentTerms(Facet, firstTermPtr)
     octObject *Facet;
     termElement *firstTermPtr;
{
  octObject JoinedTerms, Joined;
  octObject Term;
  octGenerator genBag,genTerm;
  termElement *last, *this;
  
  JoinedTerms.type = OCT_BAG;
  JoinedTerms.contents.bag.name = "JOINEDTERMS";
    
  if (octGetByName(Facet,&JoinedTerms) == OCT_OK) {
    OCT_ASSERT(octInitGenContents(&JoinedTerms, OCT_BAG_MASK, &genBag));
    while (octGenerate(&genBag,&Joined) == OCT_OK) {
      if (strcmp(Joined.contents.bag.name, "JOINED") != 0) {
	(void) sprintf(buf,"Extraneous bag %s found in JOINEDTERMS bag\n", Joined.contents.bag.name);
	Message(buf);
      } else {
	OCT_ASSERT(octInitGenContents(&Joined,OCT_TERM_MASK,&genTerm));
	last = NULL;
	while (octGenerate(&genTerm,&Term) == OCT_OK) {
	  if (Term.contents.term.instanceId != 0) FatalError("actual terminal in JOINED bag");
	  /* find this terminal in the list of the formal terminals */
	  this = firstTermPtr;
	  while (this != NULL AND strcmp(this->name,Term.contents.term.name)!=0) this= this->next;

	  if (this == NULL) {
	    (void) sprintf(buf,"No formal terminal with name %s\n",Term.contents.term.name);
	    FatalError(buf);
	  } else if (this->equivNext != NULL) {
	    (void) sprintf(buf,"Error: term %s of %s:%s is contained in a joined bag more than once\n",
		    this->name, Facet->contents.facet.cell, Facet->contents.facet.view);
	    Message(buf);
	  } else {
	    if (last == NULL) {
	      last = this;
	      this->equivNext = this->equivPrec = this;	/* point to itself */
	    } else {		/* add at the end of the list */
	      this->equivPrec = last;
	      this->equivNext = last->equivPrec; /* the list is a ring */
	      this->equivPrec->equivNext = this;
	      this->equivNext->equivPrec = this;
	      last = this;
	    }
	  }
	}
      }
    }
  }
}
