#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "vov.h"

#define  CLEAN_STDIN()     while (getchar() != '\n')


/*
 ** structure for linked list of terminal names
 */
struct termlist {
  char 		  *name;
  struct termlist *next;
};

struct termlist *termsWithoutImplFirst = NULL; /* terminals without implementation */
struct termlist *termsWithoutImplLast = NULL; /* terminals without implementation */


appendTermWithoutImpl(name)
     char *name;
{
  struct termlist *newTerm;
  
  newTerm = CALLOC (struct termlist , 1);
  newTerm->name = strsav(name);
  newTerm->next = NULL;
  if (termsWithoutImplFirst == NULL) {
    termsWithoutImplFirst = termsWithoutImplLast = newTerm;
  } else {
    termsWithoutImplLast->next = newTerm;
    termsWithoutImplLast = newTerm;
  }
}
 
freeTermWithoutImpl()
{
  struct termlist *tptr;
  
  tptr = termsWithoutImplFirst;
  while (tptr != NULL) {
    free(tptr->name);
    free((char *) tptr);
    tptr = tptr->next;
  }
  termsWithoutImplFirst = termsWithoutImplLast = NULL;
}

void alba_checkTerminalImplementation(Instance)
     octObject *Instance;
{
  octObject		MasterCon, Interface;
  octObject		Term,Box;
  octGenerator		genBox, genTerm;
  struct termlist	*termListPtr;
  int			totTerms = 0; 	/* total number of terminals */
  int			totTermsWithoutImpl = 0;

  if (verbose) printf("Checking terminal implementations of %s:%s\n",
		      Instance->contents.instance.master,
		      Instance->contents.instance.view);
  MasterCon.type = OCT_FACET;
  MasterCon.contents.facet.cell = Instance->contents.instance.master;
  MasterCon.contents.facet.view = Instance->contents.instance.view;
  MasterCon.contents.facet.version = Instance->contents.instance.version;
  MasterCon.contents.facet.facet = "contents";
  MasterCon.contents.facet.mode = "r";
  Interface.type = OCT_FACET;
  Interface.contents.facet.cell = Instance->contents.instance.master;
  Interface.contents.facet.view = Instance->contents.instance.view;
  Interface.contents.facet.version = Instance->contents.instance.version;
  Interface.contents.facet.facet = "interface";
  Interface.contents.facet.mode = "r";

  VOVinputFacet( &MasterCon );
  if (octOpenFacet(&MasterCon) < OCT_OK) {
    octError("ALBA:opening contents\n");
  }
  VOVinputFacet( &Interface );
  if (octOpenFacet(&Interface) < OCT_OK) {
    octError("ALBA:opening interface\n");
  }
  /* get all the formal terminals from the contents */

  OCT_ASSERT(octInitGenContents(&Interface,OCT_TERM_MASK, &genTerm));
  while (octGenerate(&genTerm,&Term) == OCT_OK) {
    if (Term.contents.term.instanceId != 0) {
      continue;			/* while loop */
    }
    /*  only formal terminals get here */
    totTerms++;			/* count all the formal terminals */
    OCT_ASSERT(octInitGenContents(&Term,OCT_BOX_MASK,&genBox));
    if (octGenerate(&genBox,&Box)!=OCT_OK) {
      appendTermWithoutImpl(Term.contents.term.name);
      totTermsWithoutImpl++;	/* count all the terminals without implementaion */
    }
  }
  if (totTermsWithoutImpl == 0 ) {
    if (verbose) printf(" All terminals in %s:%s are OK\n",
			MasterCon.contents.facet.cell,
			MasterCon.contents.facet.view);
  } else {
    if (totTermsWithoutImpl == totTerms) {
      printf ("All the terminals in %s:%s need an implementation\n",
	      MasterCon.contents.facet.cell,
	      MasterCon.contents.facet.view);
    } else {
      printf ("The following terminals need an implementation:\n");
      termListPtr = termsWithoutImplFirst;
      while (termListPtr != NULL) {
	printf("\t%s", termListPtr->name);
        termListPtr= termListPtr->next;
      }
      printf("\n");
    }
    /*
     ** decide whether the terminals should be implemented
     */

  }
  freeTermWithoutImpl();
}








