/*
 *  FILE: main.c
 *
 *  DESCRIPTION: main routine for the program alba
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR: Andrea Casotto
 *  MODIFY DATE:  March 1988
 *
 *  FUNCTIONS: 
 *  usage()
 *  main(argv,argc)
 *  FatalError(es)
 *  FatalOctError(es)
 *  Warning(es)
 *  Message(es)
 */
#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "oh.h"
#include "options.h"
#include "vov.h"
/*
 *   >>>  GLOBAL VARIABLES DECLARATION
 */


short	verbose = FALSE;
short   postFlpFlag = FALSE;
int     optimizeFlag = FALSE;
int     doNotCallModuleGenerators = FALSE;
char    *modGen = NULL;

st_table		*st_masters;

octObject	Chip;		/* The chip on which alba runs */
char		buf[256];

/* ARGSUSED */

optionStruct optionList[] = {
    {"Q", "",                  "Post process Floorplan (no call to module generators)"},
    {"P", "",                  "Post process Floorplan"},
    {"R", "",                  "Post process Floorplan (OPTIMIZING)"},
    {"M", "name",              "(use with -P) Specify module generator you want to use"},
    {"S", "",                  "Symmetrize"},
    {"n", "name",	       "Output view of the module generators"},
    {"p", "",                  "Process partitions"},
    {"f", "format",	       "Specify format string for instance listing" },
    {OPT_CONT, "",     "Default is \"%s\"" },
    {"l", 0,	       "List masters" },
    {"s", 0,           "Limit list to only soft cells" },
    {"O", "oldviewname",       "Old view name" },
    {"N", "newviewname",       "New view name" },
    {"v", "",                  "Verbose mode"},
    {"o", "cell:view",         "Specify output cell and view"},
    {"r", 0,		       "Rip-up all routing" },
    {OPT_RARG,             "","cell:view"},
    { 0, 0 , 0}
};



main(argc,argv)
    int argc;
    char *argv[];
{
    char  *outputName = NIL(char);
    char  *inputName = NIL(char);
    int   c;
    int   doRipup = FALSE;
    int	  doPartitions = FALSE;
    char* MG_outputView = "alba";
    char* oldViewName = 0;
    char* newViewName = 0;
    int listOnly = 0;
    int softOnly = 0;
    int symmOnly = 0;
    char* listFormat = "%s";

    extern void  symmetrize();
    extern char *version();

    
    /* parse command line options*/
    while ((c = optGetOption(argc,argv)) != EOF) {
	switch(c) {
	case 'l':
	    listOnly = 1;
	    VOVdisable();
	    break;
	case 's':
	    softOnly = 1;
	    break;
	case 'f':
	    listFormat = optarg;
	    break;
	case 'R':
	    postFlpFlag = TRUE;
	    optimizeFlag = TRUE;
	    break;
	case 'S':
	    symmOnly = 1;
	    break;
	case 'P':
	    postFlpFlag = TRUE;
	    break;
	case 'Q':
	    doNotCallModuleGenerators = TRUE;
	    postFlpFlag = TRUE;
	    break;
	case 'n':
	    MG_outputView = optarg;
	    break;
	case 'r':
	    doRipup = TRUE;
	    break;
	case 'p':
	    doPartitions = TRUE;
	    break;
	case 'o':
	    outputName = strsav(optarg);
	    break;
	case 'M':
	    modGen = optarg;
	    break;
	case 'v':
	    verbose = TRUE;
	    VOVredirect( 0 );
	    break;
	case 'O':
	    oldViewName = optarg;
	    break;
	case 'N':
	    newViewName = optarg;
	    break;
	default:
	    optUsage();
	    break;
	}
    }

    /* the remaining arguments are argv[optind ... argc-1] */
    if (optind >= argc) {
	optUsage();
    } else {
	inputName  = strsav(argv[optind]);
    }

    octBegin();
    VOVbegin( argc, argv );


    alba_initPlacementObjects();

    if ( listOnly ) {
	extern void listInstances( );

	ohUnpackDefaults( &Chip, "r", "::contents" );
	ohUnpackFacetName( &Chip, inputName );
	OH_ASSERT(octOpenFacet( &Chip ));
	listInstances( &Chip, listFormat, softOnly );
	VOVend ( 0 );
    } else {
	ohWorkFacet( &Chip, inputName, outputName );
    }

    (void)fprintf(stderr,"%s: Routines for floorplanning and placement. %s\n",
		  argv[0], version());



    if (doRipup) {
	ripupAllNets( &Chip );
    } else if ( newViewName ) {
	if ( oldViewName ) {
	    changeInstanceView( &Chip, oldViewName, newViewName );
	} else {
	    FatalError("When using option -N you must also use option -O" );
	}
    } else if (doPartitions) {
	processPartitions();
    } else if (postFlpFlag == TRUE) {
	postFlp( MG_outputView, optimizeFlag, doNotCallModuleGenerators);
    } else if( symmOnly ) {
	symmetrize( &Chip );
    } else {
	st_masters = st_init_table(strcmp, st_strhash);
	alba_checkMasters();
	alba_checkInstances();
	alba_unconnectedTerminals();
	st_free_table(st_masters);  
    }

  
    OCT_ASSERT(octFlushFacet(&Chip));
    octEnd();
    VOVend(0);			/* Successful. */
}



FatalError(es)
     char *es;			/* error string */
{
  (void) fprintf(stderr,"Fatal Error: %s\n",es);
  exit(-1);
}


Warning(es)			/* continuable error */
     char *es;			/* error string */
{
  (void) fprintf(stderr, "alba:warning: %s\n", es);
}

Message(es)			
     char *es;			/* error string */
{
  (void) fprintf(stderr, "alba:message: %s\n", es);
}

