#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "nle.h"
#include "oh.h"
#include "vov.h"

processPartitions()
{
    octObject	NewCell, NewCellInterface;
    octObject	NewInstance;
    octObject	FloorplanBag;
    octObject	PartitionBag;
    octObject	NameProp;
    octObject	ModGenProp;
    octObject	PartitionType;
    octGenerator  genBag;
    int		hardFlag = FALSE;


    FloorplanBag.type = OCT_BAG;
    FloorplanBag.contents.bag.name = "FLOORPLAN";

    PartitionBag.type = OCT_BAG;
    PartitionBag.contents.bag.name = "PARTITION";
  
    NameProp.type = OCT_PROP;
    NameProp.contents.prop.name = "PARTITION_NAME";
    NameProp.contents.prop.type = OCT_STRING;


    PartitionType.type = OCT_PROP;
    PartitionType.contents.prop.name = "PARTITION_TYPE";
    PartitionType.contents.prop.type = OCT_STRING;


    ModGenProp.type = OCT_PROP;
    ModGenProp.contents.prop.name = "MODULE_GENERATOR";
    ModGenProp.contents.prop.type = OCT_STRING;
  
    OCT_ASSERT(octInitGenContents(&Chip,OCT_BAG_MASK,&genBag));
    while (octGenerate(&genBag,&PartitionBag)==OCT_OK) {
	if (strncmp(PartitionBag.contents.bag.name, "PARTITION", 9) != 0) {
	    continue;		/* Not a partition bag */
	}
	OCT_ASSERT(octGetByName(&PartitionBag, &NameProp));
	if (octGetByName(&PartitionBag, &ModGenProp)!=OCT_OK) {
	    ModGenProp.contents.prop.value.string = "";
	}
	if (octGetByName(&PartitionBag, &PartitionType)==OCT_OK) {
	    if (strcmp(PartitionType.contents.prop.value.string, "HARD")==0) {
		hardFlag = TRUE;
	    }
	}


	NewCell.objectId = oct_null_id;
	
	ohUnpackDefaults( &NewCell, "w", ":alba:contents" );
	ohUnpackFacetName( &NewCell, NameProp.contents.prop.value.string );
    

	printf("Creating new %s macro %s\n\tto be generated by %s.\n" ,
	       hardFlag ? "HARD" : "SOFT",
	       ohFormatName( &NewCell ),
	       ModGenProp.contents.prop.value.string );


	VOVoutputFacet( &NewCell );
	OH_ASSERT(octOpenFacet(&NewCell));
    
	NewInstance.type = OCT_INSTANCE;
	NewInstance.contents.instance.name = NameProp.contents.prop.value.string;
    

	if (nleMakeMacro(&NewCell,
			 &PartitionBag ,
			 &NewInstance ,
			 hardFlag ? NLE_HARD : NLE_SOFT
			 ) != NLE_OK)
	  FatalError("During Merging");
      
	OCT_ASSERT(octDelete(&PartitionBag));

	NewCellInterface.type = OCT_FACET;
	NewCellInterface.contents.facet = NewCell.contents.facet;
	NewCellInterface.contents.facet.facet = "interface";
	NewCellInterface.contents.facet.mode = "a";

	VOVoutputFacet( &NewCellInterface );
	OH_ASSERT(octOpenFacet(&NewCellInterface) );


	OCT_ASSERT(octGetOrCreate(&NewCellInterface,&FloorplanBag));
	OCT_ASSERT(octCreateOrModify(&FloorplanBag,&ModGenProp));

	OCT_ASSERT(octFlushFacet(&NewCell));
	OCT_ASSERT(octFlushFacet(&NewCellInterface));

    }

}
