/*
 *  FILE: postFlp.c
 *
 *  DESCRIPTION: routines to process the soft cells after floorplanning
 *
 *
 *  CREATION DATE: 11/03/87		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS:
 */

#include "copyright.h"
#include "port.h"
#include "alba.h"
#include "nle.h"
#include "oh.h"
#include "vov.h"

#define MG_NO_MODULE_GENERATOR 	0
#define MG_WOLFE 		1 
#define MG_GEM			2
#define MG_PADP			3

void changeInstanceView( chip, oldViewName, newViewName )
    octObject* chip;
    char* oldViewName;
    char* newViewName;
{
    octObject bag;
    octGenerator genInst;
    octObject instance, newinstance;
    OH_ASSERT( ohGetByBagName( chip, &bag, "INSTANCES" ) );
    OH_ASSERT( octInitGenContents( &bag, OCT_INSTANCE_MASK, &genInst ) );
    
    while ( octGenerate( &genInst, &instance ) == OCT_OK ) {
	if ( oldViewName && strcmp( instance.contents.instance.view, oldViewName ) ) {
	    continue;		/* Skip this instance. */
	}
	
	ohGetById( &newinstance, instance.objectId );
	newinstance.contents.instance.view = newViewName;
	if ( verbose ) {
	    printf( "Replacing %s with %s\n" , ohFormatName( &instance ), ohFormatName( &newinstance ) );
	}
	if (nleModifyInstanceView( &instance, &newinstance) != NLE_OK) {
	    FatalError( "Failed instance replacing. Must exit." );
	}
	{
	    octObject master;
	    OH_ASSERT( ohOpenMaster(&master, &newinstance, "interface", "r" ) );
	    VOVinputFacet( &master );
	}

    }
}


int
postFlp( outputView, optimizeFlag	, doNotCallModuleGenerators)
    char* outputView;
    int optimizeFlag;
    int doNotCallModuleGenerators;
{
    octObject		Instance;
    octObject		NewInstance;
    octObject		Master;
    octObject		instanceBag;
    octObject		softCellBag;
    octGenerator		genInst;
    char			options[256];
    char			command[256];
    int			systemStatus;
    int			moduleGenerator;

    Message("Post Floorplanning Processing");

    ohCreateBag( &Chip, &softCellBag, "SOFT_CELLS_BAG" );
    octMarkTemporary( &softCellBag );
    OH_ASSERT( ohGetByBagName( &Chip, &instanceBag, "INSTANCES" ) );
    OH_ASSERT(octInitGenContents( &instanceBag, OCT_INSTANCE_MASK, &genInst));
    while (octGenerate(&genInst,&Instance)==OCT_OK) {
	if ( isSoft( &Instance ) ) {
	    octAttach( &softCellBag,  &Instance);
	}
    }

    OH_ASSERT(octInitGenContents( &softCellBag, OCT_INSTANCE_MASK, &genInst));
    while (octGenerate(&genInst,&Instance)==OCT_OK) {
	Master.type = OCT_FACET;
	Master.contents.facet.facet = "interface";
	Master.contents.facet.mode = "r";
	OH_ASSERT( ohOpenMaster( &Master, &Instance, "interface", "r" ) );
	VOVinputFacet( &Master );

	if (verbose) printf("postFlp: Processing %s\n", ohFormatName( &Master ));
	if (isSoft(&Instance)) {
	    moduleGenerator = getModuleGenerator(&Master) ;
	    switch (moduleGenerator) {
	    case MG_WOLFE:
		(void) sprintf(options, "%s -o %s:%s",
			       optimizeFlag ? "" : "-f", Master.contents.facet.cell , outputView);
		(void) sprintf(command, "wolfe %s %s:%s" , options ,
			       Master.contents.facet.cell, Master.contents.facet.view);
		if (doNotCallModuleGenerators == FALSE) {
		    printf("System call: %s\n", command);
		    systemStatus = system(command);
		    printf("Returned systemStatus is %d\n", systemStatus);
		    if (systemStatus == 127) FatalError("failed system call");
		    if (systemStatus != 0) {
			Warning("Module Generator failed");
			continue;
		    }
		}
		NewInstance.type = OCT_INSTANCE;
		(NewInstance.contents.instance = Instance.contents.instance) ;
		(NewInstance.contents.instance.transform = Instance.contents.instance.transform) ;
		NewInstance.contents.instance.view = outputView;
		if (nleModifyInstanceView( &Instance, &NewInstance) != NLE_OK) {
		    Warning("Instance modification failed");
		}
		break;
	    case MG_GEM:
		(void) sprintf(command, "gem %s",Master.contents.facet.cell);
		if (doNotCallModuleGenerators == FALSE) {
		    printf("System call: %s\n", command);
		    systemStatus = system(command);
		    printf("Returned systemStatus is %d\n", systemStatus);
		    if (systemStatus == 127) FatalError("failed system call");
		    if (systemStatus != 0) {
			Warning("Module Generator failed");
			continue;
		    }
		}
		NewInstance.type = OCT_INSTANCE;
		(NewInstance.contents.instance = Instance.contents.instance) ;
		(NewInstance.contents.instance.transform = Instance.contents.instance.transform) ;
		NewInstance.contents.instance.view = outputView;
		if (nleModifyInstanceView( &Instance, &NewInstance) != NLE_OK) {
		    Warning("Instance modification failed");
		}
		break;
	    case MG_PADP:
		outputView = "padpMG";
		(void) sprintf(options,"-g -o %s:%s %s",
			       Master.contents.facet.cell, outputView, (verbose ? "-v" : ""));
		(void) sprintf(command, "padplace %s %s:%s" , options ,
			       Master.contents.facet.cell, Master.contents.facet.view);
		if (doNotCallModuleGenerators == FALSE) {
		    printf("System call: %s\n", command);
		    systemStatus = system(command);
		    printf("Returned systemStatus is %d\n", systemStatus);
		    if (systemStatus == 127) FatalError("failed system call");
		    if (systemStatus != 0) {
			Warning("Module Generator failed");
			continue;
		    }
		}
		NewInstance.type = OCT_INSTANCE;
		(NewInstance.contents.instance = Instance.contents.instance) ;
		(NewInstance.contents.instance.transform = Instance.contents.instance.transform) ;
		NewInstance.contents.instance.view = outputView;
		if (nleModifyInstanceView( &Instance, &NewInstance) != NLE_OK) {
		    Warning("Instance modification failed");
		}
		break;
	    default:
		(void) sprintf(buf,"No module generator specified for soft cell %s:%s",
			       Master.contents.facet.cell, Master.contents.facet.view);
		Warning(buf);
		break;
	    }
	}
    }
    return OK;
}


int
oldIsSoft( InterfaceP )
    octObject *InterfaceP;
{
    octObject FloorplanBag;

    FloorplanBag.type = OCT_BAG;
    FloorplanBag.contents.bag.name = "FLOORPLAN";

    if (octGetByName(InterfaceP,&FloorplanBag)==OCT_OK) {
	return(1);
    } else {
	return(0);
    }
}


int
getModuleGenerator(InterfaceP)
    octObject *InterfaceP;
{
    extern char *modGen;
    char *mg;			/* short for module generator */
    octObject 	ModuleGenerator;
    octObject 	FloorplanBag;

    if (modGen != NULL) {
	mg = modGen;
    } else {
	OH_ASSERT( ohGetByBagName( InterfaceP, &FloorplanBag, "FLOORPLAN" ) );
	OH_ASSERT( ohGetOrCreatePropStr( &FloorplanBag, &ModuleGenerator, "MODULE_GENERATOR", "" ) );
	mg = ModuleGenerator.contents.prop.value.string;
    }
    if (strcmp(mg,"")==0) {
	return(MG_NO_MODULE_GENERATOR);
    } else if (strcmp(mg,"wolfe")==0) {
	return(MG_WOLFE);
    } else if (strcmp(mg,"gem")==0) {
	return(MG_GEM);
    } else if (strncmp(mg,"padplace",4)==0) {
	return(MG_PADP);
    } else {
	(void) sprintf(buf,"Unknown module generator %s\n", mg);
	FatalError(buf);
    }
    return MG_WOLFE;		/* Actually this should never be reached */
}
    
