/*
 *   alba_routines.c
 * >>>> ROUTINES
 *  alba_copyFacets ( cellName , oldView , newView )
 *  alba_removeBusses ( cellName , view )
 *  alba_checkInstances(cellName, view)
 *  alba_checkViews(Instance)
 *  alba_modifyInstanceView(Chip,Instance, newView)
 *  alba_unconnectedTerminals(cellName, viewName)
 *  alba_checkAreaInformation(Instance)
 */

#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "alba.h"


alba_copyFacets ( cellName , oldView , newView )
    char *cellName, *oldView ,*newView;
{
  octObject  	Input, Output;

  (void) sprintf(buf,"Copying %s:%s into %s:%s", cellName, oldView , cellName, newView);
  Message(buf);

  Input.type = OCT_FACET;
  Input.contents.facet.cell = cellName;
  Input.contents.facet.view = oldView;
  Input.contents.facet.facet = "contents";
  Input.contents.facet.version = OCT_CURRENT_VERSION;

  Output.type = OCT_FACET;
  Output.contents.facet.cell = cellName;
  Output.contents.facet.view = newView;
  Output.contents.facet.facet = "contents";
  Output.contents.facet.version = OCT_CURRENT_VERSION;

  if (octCopyFacet(&Output,&Input) != OCT_OK) FatalError("Copying contents facet");
  Input.contents.facet.facet = "interface";
  Output.contents.facet.facet = "interface";
  if (octCopyFacet(&Output,&Input) != OCT_OK) FatalError("Copying interface facet");
}




/*
 *  looks for unconnected terminals in the instances
 *  and tries to connect them to the right net
 */
alba_unconnectedTerminals()
{
  octObject     Instance;
  octObject     Term, FormalTerm;
  octObject     Net;
  octObject     VddNet,GNDNet; /* nets */
  octObject     TermType;	/* oct property */
  
  octGenerator  genInst, genNet, genTerm, genFormalTerm;
  char 		*termType;		/* type of a terminal */
  char		formalTermName[256];

  if (verbose ) printf("\n--- Unconnected terminals\n");

  TermType.type = OCT_PROP;
  TermType.contents.prop.name = "TERMTYPE";
  VddNet.type = OCT_NET;
  VddNet.contents.net.name = "Vdd";
  GNDNet.type = OCT_NET;
  GNDNet.contents.net.name = "GND";

  /*
   * generate all instances from Chip and all the terminals
   * from each instance. Check each terminal to see if it is
   * connected to any net
   */
  OCT_ASSERT(octInitGenContents(&Chip, OCT_INSTANCE_MASK, &genInst));
  while (octGenerate(&genInst,&Instance)==OCT_OK) {
    OCT_ASSERT(octInitGenContents(&Instance,OCT_TERM_MASK,&genTerm));
    while (octGenerate(&genTerm, &Term) == OCT_OK) {
      /*
       *  read property termType
       */
      if (octGetByName(&Term,&TermType)==OCT_NOT_FOUND) {
	termType = "UNSPECIFIED";
      } else {
	termType = TermType.contents.prop.value.string;
      }
      /* is the terminal connected to a formal terminal ? */
      OCT_ASSERT(octInitGenContainers(&Term,OCT_TERM_MASK,&genFormalTerm));
      if (octGenerate(&genFormalTerm,&FormalTerm)== OCT_OK AND FormalTerm.contents.term.instanceId==0) {
	(void) sprintf(formalTermName, "FORMAL:%s" ,FormalTerm.contents.term.name);
      } else {
	  (void) sprintf(formalTermName,"---");
      }
      /*
       *  is this terminal connected to a net ?
       */
      OCT_ASSERT(octInitGenContainers(&Term,OCT_NET_MASK,&genNet));
      if (octGenerate(&genNet,&Net)!= OCT_OK) {
	if (verbose ) (void) sprintf(buf,"Term %s\t of %s\t %s %s\n",
			      Term.contents.term.name,
			      Instance.contents.instance.name,
			      termType,
			      formalTermName);
	Message(buf);
      }
    }
  }
}
