/*
 * anchor.c
 *
 * main() subroutine is defined here.  It controls the flow of the
 * module generator, envoking the various steps in order.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"
#include "vov.h"

/* global variables */

/* values read in from the input pla(5CAD) file, used to initialize ace	*/
s_signal *InSigs, *OutSigs;	/* name, aceId and used bit of I/O signals */
int InSigCount, OutSigCount;	/* number of I/O signals		*/
s_signal *ProdSigs;		/* name, aceId and used bit of prod terms */
int ProdCount;			/* number of product terms		*/
char **PlaTT;			/* the truth table data of input pla	*/
int CellCount;			/* number of transistors in the pla core */

s_cells		Cells;		/* contains cell names			*/
s_ruleNames	RuleNames;	/* names of the connection rules	*/
char		**Core;	    	/* sanitized personality matrix		*/
char		**GivenCore;    /* given personality matrix		*/
s_options	Options;	/* command line options			*/
s_arrays	Arrays;		/* the various mkarray mkId's		*/
octObject 	oFacet;		/* the output PLA facet			*/
octObject 	oPalette;	/* the palette of instances		*/

s_input		**Inputs;	/* parameters of sanitized matrix	*/
s_output	**Outputs;
int 		Rows;		/* number of rows in the PLA		*/
int		AndCols;	/* number of columns in the AND plane	*/
int		OrCols;		/* number of columns in the OR plane	*/
int		OddEvenBreak;	/* split in OR plane between fold types	*/

/***parameters of given matrix that aren't being used */
s_input		**GivenInputs;
s_output	**GivenOutputs;
int 		GivenRows;
int		GivenAndCols;
int		GivenOrCols;

main(argc, argv)
int argc;
char *argv[];
{
    anchorVersion( argv[ 0 ]);

    /* set Options defaults						*/
    optionsDefaults();

    if ( argc > 1 ) VOVbegin(argc,argv);

    if( Options.debug){
	/* generate core on error					*/
	errCore(TRUE);
    }

    /* set other defaults						*/
    setDefaults();
    setCellAndRuleNameDefaults();

    /* parse command line */
    parseComLine( argc, argv );
    doInheritedDefaults();

    /* initialize ace and run the annealing				*/
    if( Options.forceAnneal || needToAnneal() ){
	readPla( Options.inPlaFile);
	annealPla();
	writePM( Options.outPMFile);
    } else {
	readPM( Options.inPMFile );
    }

    if( Options.noLayout ) VOVend( 1 );

    octBegin();
    openFacet(Options.plaFacet); 

    /* dump PLA(5) output */
    if(( Options.debug > 1) && (Options.plaVerifyFile != NIL(char))){
	printPLA( GivenRows, GivenOrCols, GivenCore, GivenOutputs);
    }

    /* fix any problems with the personality matrix			*/
    sanitizeMatrix();
    if( Options.debug > 0 )printTrue();

    /* dump PLA(5) output */
    if( Options.plaVerifyFile != NIL(char)){
	printPLA( Rows, OrCols, Core, Outputs);
    }

    /* open the palette							*/
    if( Options.localPaletteName != NIL( char )){
	openLocalPalette( Options.localPaletteName );
    } else {
	openPalette( Options.paletteName );
    }

    /* create the array and subArrays					*/
    createArrays();

    /* read the wiring rules from oPalette				*/
    mkReadRules( Arrays.full, &oPalette );

    /* instanciate and wire the buffers 				*/
    createIOBuffers();

    /* instanciate and wire the core of the PLA				*/
    createPlaCore();

    /* wiring the perimeter of the array				*/
    wirePerimeter();

    /* solve the array							*/
    mkPlaceArray( Arrays.full );

    /* close oFacet and exit						*/ 
    moveInterfaceProps();
    OH_ASSERT( octCloseFacet( &oFacet ));
    octEnd();
    VOVend(0);
}

/*
 * needToAnneal
 *
 * If the personality matrix file doesn't exist, must anneal
 */
needToAnneal()
{
    FILE *fp;

    if(( fp = VOVfopen( Options.inPMFile, "r")) == NIL(FILE)){
	return( TRUE );
    } else {
	(void)fclose( fp );
	return( FALSE );
    }
}
