#include "copyright.h"
#include "port.h"
#include "anchor.h"
extern s_options Options;
extern int AndCols, OrCols, Rows;
extern s_arrays Arrays;
extern octObject oFacet;
/*
 * createArrays
 *
 * creates the array and subarray structure.  The subarray structure is
 * set up so that most changes to the architecture of the PLA can be 
 * made here, without modification to the instanciation and wiring 
 * routines.
 *
 */
createArrays()
{
    int andPlaneCols, orPlaneCols, coreRows, fullRows, fullCols;
    struct octBox *area; 	/* areas in the full array		*/
    struct octBox *subArea;  	/* areas in subArrays			*/
    int row, col;
    int coreBase;  	/* the first row of the AND and OR planes	*/
    int andBase;	/* the first column of the AND plane		*/
    int orBase;		/* the first column of the OR plane		*/

    /* open the arrays 							*/

    mkOpenArray( &Arrays.full, &oFacet );
    mkOpenSubarray( &Arrays.andPullup, Arrays.full );
    mkOpenSubarray( &Arrays.andPlaneGnd, Arrays.full );
    mkOpenSubarray( &Arrays.andTran, Arrays.full );
    mkOpenSubarray( &Arrays.andCont, Arrays.full );
    mkOpenSubarray( &Arrays.andGnd, Arrays.full );
    mkOpenSubarray( &Arrays.andOrInterface, Arrays.full );
    mkOpenSubarray( &Arrays.orPullup, Arrays.full );
    mkOpenSubarray( &Arrays.orPower, Arrays.full );
    mkOpenSubarray( &Arrays.orEvenPlaneGnd, Arrays.full );
    mkOpenSubarray( &Arrays.orOddPlaneGnd, Arrays.full );
    mkOpenSubarray( &Arrays.orTran, Arrays.full );
    mkOpenSubarray( &Arrays.orEvenGnd, Arrays.full );
    mkOpenSubarray( &Arrays.orOddGnd, Arrays.full );
    mkOpenSubarray( &Arrays.orEvenCont, Arrays.full );
    mkOpenSubarray( &Arrays.orOddCont, Arrays.full );
    mkOpenSubarray( &Arrays.inputTerms, Arrays.full );
    mkOpenSubarray( &Arrays.outputTerms, Arrays.full );
    mkOpenSubarray( &Arrays.input, Arrays.full );
    mkOpenSubarray( &Arrays.output, Arrays.full );
    mkOpenSubarray( &Arrays.vdd, Arrays.full );
    mkOpenSubarray( &Arrays.vss, Arrays.full );
    mkOpenSubarray( &Arrays.inPhi, Arrays.full );
    mkOpenSubarray( &Arrays.inPhiBar, Arrays.full );
    mkOpenSubarray( &Arrays.outPhi, Arrays.full );

    /* array options							*/
    if( Options.printGraph ) mkSetPrintGraph( Arrays.full, 1 );

    /* size the arrays 							*/

    coreRows = 2 * Rows + 1;
    andPlaneCols = 2 * AndCols + 1;
    orPlaneCols = 2 * OrCols;
    fullRows = coreRows + 18;
    fullCols = andPlaneCols + orPlaneCols + 11;
    coreBase = 9;
    andBase = 4;
    orBase = andBase + andPlaneCols + 3;

    mkAddArrayArea( Arrays.full, mkPoint( fullRows - 1, fullCols - 1 ));
    mkAddArrayArea( Arrays.andPullup, mkPoint( Rows - 1, 1 ));
    mkAddArrayArea( Arrays.andPlaneGnd, mkPoint( Rows, AndCols ));
    mkAddArrayArea( Arrays.andTran, mkPoint( Rows - 1, AndCols - 1 ));
    mkAddArrayArea( Arrays.andCont, mkPoint( Rows - 1, AndCols - 1 ));
    mkAddArrayArea( Arrays.andGnd, mkPoint( Rows - 1, AndCols - 1 ));
    mkAddArrayArea( Arrays.andOrInterface, mkPoint( Rows - 1, 0 ));
    mkAddArrayArea( Arrays.orPullup, mkPoint( 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.orPower, mkPoint( 1, OrCols + 1 ));
    mkAddArrayArea( Arrays.orOddPlaneGnd, mkPoint( Rows, OrCols + 1 ));
    mkAddArrayArea( Arrays.orEvenPlaneGnd, mkPoint( Rows, OrCols + 1));
    mkAddArrayArea( Arrays.orTran, mkPoint( Rows - 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.orEvenGnd, mkPoint( Rows - 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.orOddGnd, mkPoint( Rows - 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.orEvenCont, mkPoint( Rows - 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.orOddCont, mkPoint( Rows - 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.input, mkPoint( 11, AndCols - 1 ));
    mkAddArrayArea( Arrays.output, mkPoint( 11, OrCols - 1 ));
    mkAddArrayArea( Arrays.inputTerms, mkPoint( 1, AndCols - 1 ));
    mkAddArrayArea( Arrays.outputTerms, mkPoint( 1, OrCols - 1 ));
    mkAddArrayArea( Arrays.vdd, mkPoint( 1, 2 ));
    mkAddArrayArea( Arrays.vss, mkPoint( 3, 2 ));
    mkAddArrayArea( Arrays.inPhi, mkPoint( 1, 0 ));
    mkAddArrayArea( Arrays.inPhiBar, mkPoint( 1, 0 ));
    mkAddArrayArea( Arrays.outPhi, mkPoint( 1, 0 ));

    /* linking the andPullup subArray					*/
    /* linking the AND plane subArrays					*/
    /* this relies on the fact that linking a subArray links row	*/
    /* and columns in the subarray to the array.  The rows and columns	*/
    /* are explicitly linked, not each individual area			*/


    for( row = 0; row < Rows; row++ ){
	subArea = mkRow( row );

	area = mkRow( coreBase + 2 * row );
	mkLinkSubarray( Arrays.andPlaneGnd, subArea, area );

	area = mkRow( coreBase + 2 * row + 1 );
	mkLinkSubarray( Arrays.andPullup, subArea, area );
	mkLinkSubarray( Arrays.andGnd, subArea, area );
	mkLinkSubarray( Arrays.andTran, subArea, area );
	mkLinkSubarray( Arrays.andCont, subArea, area );
    }
    mkLinkSubarray( Arrays.andPlaneGnd, mkRow( Rows ), 
			       mkRow( coreBase + 2 * Rows ));

    for( col = 0; col < AndCols; col++ ){
	subArea = mkCol( col );
	area = mkCol( andBase + 2 * col + 1 );
	mkLinkSubarray( Arrays.andTran, subArea, area );
	if( col%2 ){

	    area = mkCol( andBase + 2 * col + 2 );
	    mkLinkSubarray( Arrays.andPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.andGnd, subArea, area );

	    area = mkCol( andBase + 2 * col );
	    mkLinkSubarray( Arrays.andCont, subArea, area );
	} else {

	    area = mkCol( andBase + 2 * col );
	    mkLinkSubarray( Arrays.andPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.andGnd, subArea, area );

	    area = mkCol( andBase + 2 * col + 2 );
	    mkLinkSubarray( Arrays.andCont, subArea, area );
	}
    }
    /* the last column of the andPLaneGnd is the center vss column	*/
    mkLinkSubarray( Arrays.andPlaneGnd, mkCol( AndCols ), mkCol( orBase - 2 ));
    /* linking the andOrInterface subArray				*/

    for( row = 0; row < Rows; row++ ){
	subArea = mkPoint( row, 0 );
	area = mkPoint( coreBase + 2 * row + 1,  orBase - 1 );
	mkLinkSubarray( Arrays.andOrInterface, subArea, area );
    }

    /* linking the OR plane subArrays					*/
    /* this relies on the fact that linking a subArray links row	*/
    /* and columns in the subarray to the array.  The rows and columns	*/
    /* are explicitly linked, not each individual area			*/

    for( row = 0; row < Rows; row++ ){
	subArea = mkRow( row );
	area = mkRow( coreBase + 2 * row + 1 );
	mkLinkSubarray( Arrays.orTran, subArea, area );
	if( row%2 ){
	    area = mkRow( coreBase + 2 * row + 2 );
	    mkLinkSubarray( Arrays.orOddPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.orOddGnd, subArea, area );
	    mkLinkSubarray( Arrays.orEvenCont, subArea, area );

	    area = mkRow( coreBase + 2 * row );
	    mkLinkSubarray( Arrays.orEvenPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.orEvenGnd, subArea, area );
	    mkLinkSubarray( Arrays.orOddCont, subArea, area );
	} else {
	    area = mkRow( coreBase + 2 * row );
	    mkLinkSubarray( Arrays.orOddPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.orOddGnd, subArea, area );
	    mkLinkSubarray( Arrays.orEvenCont, subArea, area );

	    area = mkRow( coreBase + 2 * row + 2 );
	    mkLinkSubarray( Arrays.orOddCont, subArea, area );
	    mkLinkSubarray( Arrays.orEvenPlaneGnd, subArea, area );
	    mkLinkSubarray( Arrays.orEvenGnd, subArea, area );
	}
    }
    area = mkRow( coreBase + 2 * row + 1 );
    for( col = 0; col < OrCols; col++ ){
	subArea = mkCol( col );
	area = mkCol ( orBase + 2 * col);
	mkLinkSubarray( Arrays.orPower, subArea, area );
	mkLinkSubarray( Arrays.orOddPlaneGnd, subArea, area );
	mkLinkSubarray( Arrays.orEvenPlaneGnd, subArea, area );

	area = mkCol( orBase + 2 * col + 1 );
	mkLinkSubarray( Arrays.orPullup, subArea, area );
	mkLinkSubarray( Arrays.orTran, subArea, area );
	mkLinkSubarray( Arrays.orEvenGnd, subArea, area );
	mkLinkSubarray( Arrays.orOddGnd, subArea, area );
	mkLinkSubarray( Arrays.orEvenCont, subArea, area );
	mkLinkSubarray( Arrays.orOddCont, subArea, area );
    }
    mkLinkSubarray( Arrays.orPower, mkCol( OrCols  ),
			       mkCol( orBase + 2 * OrCols ));
    mkLinkSubarray( Arrays.orOddPlaneGnd, mkCol( OrCols  ),
			       mkCol( orBase + 2 * OrCols ));
    mkLinkSubarray( Arrays.orEvenPlaneGnd, mkCol( OrCols ),
			       mkCol( orBase + 2 * OrCols ));

    /* linking the input						*/

    for( col = 0; col < AndCols; col++ ){
	for( row = 0; row < 6; row++ ){
	    subArea = mkPoint( row, col );
	    area = mkPoint( row, andBase + 2 * col + 1 );
	    mkLinkSubarray( Arrays.input, subArea, area );

	    subArea = mkPoint( row + 6, col );
	    area = mkPoint( fullRows - 6 + row, andBase + 2 * col + 1 );
	    mkLinkSubarray( Arrays.input, subArea, area );
	}

	area = mkCol( andBase + 2 * col + 1 );
	mkLinkSubarray( Arrays.inputTerms, mkCol( col ), area );
	mkLinkSubarray( Arrays.inputTerms, mkCol( col ), area );
    }

    /* linking the output						*/

    for( col = 0; col < OrCols; col++ ){
	for( row = 0; row < 6; row++ ){
	    subArea = mkPoint( row, col );
	    area = mkPoint( row, orBase + 2 * col + 1 );
	    mkLinkSubarray( Arrays.output, subArea, area );

	    subArea = mkPoint( row + 6, col );
	    area = mkPoint( fullRows - 6 + row, orBase + 2 * col + 1 );
	    mkLinkSubarray( Arrays.output, subArea, area );
	}
	area = mkCol( orBase + 2 * col + 1 );
	mkLinkSubarray( Arrays.outputTerms, mkCol( col ), area );
    }
    /* linking the vss and clock arrays					*/
    mkLinkSubarray( Arrays.inPhiBar, mkRow( 0 ), mkRow( 0 ));
    mkLinkSubarray( Arrays.inPhi, mkRow( 0 ), mkRow( 1 ));
    mkLinkSubarray( Arrays.vss, mkRow( 0 ), mkRow( 2 ));
    mkLinkSubarray( Arrays.vdd, mkRow( 0 ), mkRow( 3 ));
    mkLinkSubarray( Arrays.vss, mkRow( 1 ), mkRow( 4 ));
    mkLinkSubarray( Arrays.outPhi, mkRow( 0 ), mkRow( 4 ));
    mkLinkSubarray( Arrays.inputTerms, mkRow( 0 ), mkRow( 6 ));
    mkLinkSubarray( Arrays.outputTerms, mkRow( 0 ), mkRow( 6 ));
    mkLinkSubarray( Arrays.orPullup, mkRow( 0 ), mkRow( 7 ));
    mkLinkSubarray( Arrays.orPower, mkRow( 0 ), mkRow( 8 ));

    mkLinkSubarray( Arrays.inPhiBar, mkRow( 1 ), mkRow( fullRows - 1 ));
    mkLinkSubarray( Arrays.inPhi, mkRow( 1 ), mkRow( fullRows - 2 ));
    mkLinkSubarray( Arrays.vss, mkRow( 3 ), mkRow( fullRows - 3 ));
    mkLinkSubarray( Arrays.vdd, mkRow( 1 ), mkRow( fullRows - 4 ));
    mkLinkSubarray( Arrays.vss, mkRow( 2 ), mkRow( fullRows - 5 ));
    mkLinkSubarray( Arrays.outPhi, mkRow( 1 ), mkRow( fullRows - 5 ));
    mkLinkSubarray( Arrays.inputTerms, mkRow( 1 ), mkRow( fullRows - 7 ));
    mkLinkSubarray( Arrays.outputTerms, mkRow( 1 ), mkRow( fullRows - 7 ));
    mkLinkSubarray( Arrays.orPullup, mkRow( 1 ), mkRow( fullRows - 8 ));
    mkLinkSubarray( Arrays.orPower, mkRow( 1 ), mkRow( fullRows - 9 ));

    mkLinkSubarray( Arrays.andPullup, mkCols( 0, 1 ), mkCols( 2, 3 ));
    mkLinkSubarray( Arrays.inPhiBar, mkCol( 0 ), mkCol( 0 ));
    mkLinkSubarray( Arrays.inPhi, mkCol( 0 ), mkCol( 1 ));
    mkLinkSubarray( Arrays.vdd, mkCol( 0 ), mkCol( 2 ));
    mkLinkSubarray( Arrays.vss, mkCol( 0 ), mkCol( 3 ));
    mkLinkSubarray( Arrays.vdd, mkCol( 1 ), mkCol( orBase - 3 ));
    mkLinkSubarray( Arrays.vss, mkCol( 1 ), mkCol( orBase - 2 ));
    mkLinkSubarray( Arrays.vdd, mkCol( 2 ), mkCol( fullCols - 3 ));
    mkLinkSubarray( Arrays.vss, mkCol( 2 ), mkCol( fullCols - 2 ));
    mkLinkSubarray( Arrays.outPhi, mkCol( 0 ), mkCol( fullCols - 1 ));
}
