/*
 * defaults.c
 *
 * Set up defaults layout generation.
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"

#define CADROOT "$OCTTOOLS"

extern s_options	Options;
extern s_cells 		Cells;
extern s_ruleNames	RuleNames;
extern octObject	oFacet, oPalette;

static char cellPath[] = "/tech/scmos/anchor/";

/*
 * setDefaults
 *
 * Set default values for various global structures
 */
setDefaults( )
{
    oFacet.contents.facet.cell = NIL(char);
    oFacet.contents.facet.view = "unspaced";
    oFacet.contents.facet.facet = "contents";
    oFacet.contents.facet.version = OCT_CURRENT_VERSION;

    oPalette.contents.facet.view = "symbolic";
    oPalette.contents.facet.facet = "contents";
    oPalette.contents.facet.version = OCT_CURRENT_VERSION;
}

/*
 * setCellAndRuleNameDefaults
 *
 * Set cell names and views; set rule names.
 */
setCellAndRuleNameDefaults()
{
    int	pathLen;
    char *newStr;

    pathLen = strlen( cellPath ) + strlen( CADROOT);

    Cells.andPullup.name = "andPullup";
    newStr = ALLOC( char, pathLen + strlen(Cells.andPullup.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andPullup.name);
    Cells.andPullup.master = newStr;
    Cells.andPullup.view = "physical";

    Cells.andPlaneGnd.name = "andPlaneGnd";
    newStr = ALLOC( char, pathLen + strlen(Cells.andPlaneGnd.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andPlaneGnd.name);
    Cells.andPlaneGnd.master = newStr;
    Cells.andPlaneGnd.view = "physical";

    Cells.andTran.name = "andTran";
    newStr = ALLOC( char, pathLen + strlen(Cells.andTran.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andTran.name);
    Cells.andTran.master = newStr;
    Cells.andTran.view = "physical";

    Cells.andGnd.name = "andGnd";
    newStr = ALLOC( char, pathLen + strlen(Cells.andGnd.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andGnd.name);
    Cells.andGnd.master = newStr;
    Cells.andGnd.view = "physical";

    Cells.andCont.name = "andCont";
    newStr = ALLOC( char, pathLen + strlen(Cells.andCont.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andCont.name);
    Cells.andCont.master = newStr;
    Cells.andCont.view = "physical";

    Cells.andOrInterface.name = "andOrInt";
    newStr = ALLOC( char, pathLen + strlen(Cells.andOrInterface.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.andOrInterface.name);
    Cells.andOrInterface.master = newStr;
    Cells.andOrInterface.view = "physical";

    Cells.orPullup.name = "orPullup";
    newStr = ALLOC( char, pathLen + strlen(Cells.orPullup.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.orPullup.name);
    Cells.orPullup.master = newStr;
    Cells.orPullup.view = "physical";

    Cells.orPlaneGnd.name = "orPlaneGnd";
    newStr = ALLOC( char, pathLen + strlen(Cells.orPlaneGnd.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.orPlaneGnd.name);
    Cells.orPlaneGnd.master = newStr;
    Cells.orPlaneGnd.view = "physical";

    Cells.orTran.name = "orTran";
    newStr = ALLOC( char, pathLen + strlen(Cells.orTran.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.orTran.name);
    Cells.orTran.master = newStr;
    Cells.orTran.view = "physical";

    Cells.orGnd.name = "orGnd";
    newStr = ALLOC( char, pathLen + strlen(Cells.orGnd.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.orGnd.name);
    Cells.orGnd.master = newStr;
    Cells.orGnd.view = "physical";

    Cells.orCont.name = "orCont";
    newStr = ALLOC( char, pathLen + strlen(Cells.orCont.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.orCont.name);
    Cells.orCont.master = newStr;
    Cells.orCont.view = "physical";

    Cells.input.name = "input";
    newStr = ALLOC( char, pathLen + strlen(Cells.input.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.input.name);
    Cells.input.master = newStr;
    Cells.input.view = "physical";

    Cells.input_clk.name = "input.clk";
    newStr = ALLOC( char, pathLen + strlen(Cells.input_clk.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.input_clk.name);
    Cells.input_clk.master = newStr;
    Cells.input_clk.view = "physical";

    Cells.output1.name = "output1";
    newStr = ALLOC( char, pathLen + strlen(Cells.output1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.output1.name);
    Cells.output1.master = newStr;
    Cells.output1.view = "physical";

    Cells.output1_clk.name = "output1.clk";
    newStr = ALLOC( char, pathLen + strlen(Cells.output1_clk.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.output1_clk.name);
    Cells.output1_clk.master = newStr;
    Cells.output1_clk.view = "physical";

    Cells.output2.name = "output2";
    newStr = ALLOC( char, pathLen + strlen(Cells.output2.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.output2.name);
    Cells.output2.master = newStr;
    Cells.output2.view = "physical";

    Cells.output2_clk.name = "output2.clk";
    newStr = ALLOC( char, pathLen + strlen(Cells.output2_clk.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.output2_clk.name);
    Cells.output2_clk.master = newStr;
    Cells.output2_clk.view = "physical";

    Cells.pom1.name = "pom1";
    newStr = ALLOC( char, pathLen + strlen(Cells.pom1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.pom1.name);
    Cells.pom1.master = newStr;
    Cells.pom1.view = "physical";

    Cells.pom1_1.name = "pom1_1";
    newStr = ALLOC( char, pathLen + strlen(Cells.pom1_1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.pom1_1.name);
    Cells.pom1_1.master = newStr;
    Cells.pom1_1.view = "physical";

    Cells.pom1_2.name = "pom1_2";
    newStr = ALLOC( char, pathLen + strlen(Cells.pom1_2.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.pom1_2.name);
    Cells.pom1_2.master = newStr;
    Cells.pom1_2.view = "physical";

    Cells.m1m1.name = "m1m1";
    newStr = ALLOC( char, pathLen + strlen(Cells.m1m1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.m1m1.name);
    Cells.m1m1.master = newStr;
    Cells.m1m1.view = "physical";

    Cells.m1m1_1.name = "m1m1_1";
    newStr = ALLOC( char, pathLen + strlen(Cells.m1m1_1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.m1m1_1.name);
    Cells.m1m1_1.master = newStr;
    Cells.m1m1_1.view = "physical";

    Cells.m2m2.name = "m2m2";
    newStr = ALLOC( char, pathLen + strlen(Cells.m2m2.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.m2m2.name);
    Cells.m2m2.master = newStr;
    Cells.m2m2.view = "physical";

    Cells.m2m2_1.name = "m2m2_1";
    newStr = ALLOC( char, pathLen + strlen(Cells.m2m2_1.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.m2m2_1.name);
    Cells.m2m2_1.master = newStr;
    Cells.m2m2_1.view = "physical";

    Cells.vdd_via.name = "vdd_via";
    newStr = ALLOC( char, pathLen + strlen(Cells.vdd_via.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.vdd_via.name);
    Cells.vdd_via.master = newStr;
    Cells.vdd_via.view = "physical";

    Cells.vss_via.name = "vss_via";
    newStr = ALLOC( char, pathLen + strlen(Cells.vss_via.name) + 1);
    sprintf( newStr, "%s%s%s", CADROOT, cellPath, Cells.vss_via.name);
    Cells.vss_via.master = newStr;
    Cells.vss_via.view = "physical";

    RuleNames.orOutput = "OR_OUTPUT";
    RuleNames.productTerm = "PRODUCT_TERM";
    RuleNames.andTranDiffusion = "AND_TRAN_DIFFUSION";
    RuleNames.andGndDiffusion = "AND_GND_DIFFUSION";
    RuleNames.orTranDiffusion = "OR_TRAN_DIFFUSION";
    RuleNames.orGndDiffusion = "OR_GND_DIFFUSION";
    RuleNames.orGround = "OR_GROUND_RULE";
    RuleNames.andGround = "AND_GROUND_RULE";
    RuleNames.literal = "LITERAL_RULE";
    RuleNames.andPullup = "AND_PULLUP";
    RuleNames.orPullup = "OR_PULLUP";
    RuleNames.buffer = "BUFFER";
    RuleNames.cPower = "C_POWER";
}

/*
 * optionsDefaults
 *
 * Set options structure to default values before parsing the 
 * command line.
 */
optionsDefaults()
{
    Options.technology = "scmos";
    Options.paletteName = "anchor";
    /*** little hack cuz tap isn't set up right				*/
    Options.localPaletteName = NIL(char);
    Options.inPlaFile = NIL(char);
    Options.inPMFile = NIL(char);
    Options.outPMFile = NIL(char);
    Options.plaFacet = NIL(char);
    Options.printGraph = FALSE;
    Options.andGndLineSep = 4;
    Options.orGndLineSep = 4;
    Options.wire = TRUE;
    Options.debug = 0;
    Options.clocking = 0;
    Options.extraTerms = 1;
    Options.plaVerifyFile = NIL(char);
    Options.fastAnneal = FALSE;
    Options.fold = TRUE;
    Options.forceAnneal = FALSE;
    Options.noLayout = FALSE;
}

/*
 * doInheritedDefaults
 *
 * Calculate any necessary names that haven't been specified by starting
 * defaults or command line options.
 */
doInheritedDefaults()
{
    int len = 0;
    octObject oDummy;

    if( Options.inPlaFile == NIL(char)){
	OH_ASSERT_DESCR(ohUnpackFacetName( &oDummy, Options.plaFacet),
		"couldn't unpack the facet name");
	len = strlen( oDummy.contents.facet.cell );
	Options.inPlaFile = ALLOC( char, len + 5);
	if( Options.inPlaFile == NIL(char)){
	    errRaise( optProgName, 1, "no memory for inPlaFile name");
	}
	(void)sprintf( Options.inPlaFile, "%s.pla", oDummy.contents.facet.cell);
    }

    if( Options.outPMFile == NIL(char)){
	if( len == 0){
	    OH_ASSERT_DESCR(ohUnpackFacetName( &oDummy, Options.plaFacet),
		    "couldn't unpack the facet name");
	    len = strlen( oDummy.contents.facet.cell );
	}
	Options.outPMFile = ALLOC( char, len + 4);
	if( Options.outPMFile == NIL(char)){
	    errRaise( optProgName, 1, "no memory for outPMFile name");
	}
	(void)sprintf( Options.outPMFile, "%s.pm", oDummy.contents.facet.cell);
    }

    if( Options.inPMFile == NIL(char)){
	Options.inPMFile = Options.outPMFile;
    }
}
