#include "copyright.h"
#include "port.h"
#include "anchor.h"
#include "vov.h"

extern octObject oPalette;
extern octObject oFacet;
extern s_options Options;

/* to return different objects from these routines			*/
#define OBJ_ARRAY_SIZE 10
static octObject ObjArray[ OBJ_ARRAY_SIZE ];
static int ObjCount = 0;
/*
 * openFacet
 *
 * opens oFacet with the name facetname
 * facetname is of the form cell[:view[:facet[:version]]]
 * It is assumed that the default values of cell view version
 * are already set ( in octopusDefaults() ).
 */
openFacet( facetName )
char	*facetName;
{
    octObject oObj;

    OH_ASSERT_DESCR(ohUnpackFacetName( &oFacet, facetName ),
			"couldn't unpack the facet name");
    oFacet.type = OCT_FACET;
    oFacet.contents.facet.mode = "w";
    VOVoutputFacet( &oFacet );
    if( octOpenFacet( &oFacet ) < OCT_OK ) {
	errRaise( optProgName, -1,
		 "Could not open %s:%s:%s: %s.",
		 oFacet.contents.facet.cell,
		 oFacet.contents.facet.view,
		 oFacet.contents.facet.facet,
		 octErrorString());
    }
    OH_ASSERT( ohCreatePropStr( &oFacet, &oObj, "VIEWTYPE", "SYMBOLIC" ));
    OH_ASSERT( ohCreatePropStr( &oFacet, &oObj, "EDITSTYLE", "SYMBOLIC" ));
    OH_ASSERT( ohCreatePropStr( &oFacet, &oObj, "CELLCLASS", "MODULE" ));
    OH_ASSERT( ohCreatePropStr( &oFacet, &oObj, "CELLTYPE", "MODULE" ));
    OH_ASSERT( ohCreatePropStr( &oFacet, &oObj, "TECHNOLOGY", 
				 Options.technology ));
    OH_ASSERT( ohCreateBag( &oFacet, &oObj, "INSTANCES" ));
    OH_ASSERT( ohCreateBag( &oFacet, &oObj, "CONNECTORS" ));
}
/*
 * openPalette
 *
 * sets oPalette to the palette named name in technology
 */
openPalette( name )
char *name;
{

    if( tapOpenPalette( &oFacet, name, &oPalette, "r" ) < OCT_OK ){
	errRaise( optProgName, -1, "Could not open palette %s.",
			 name);
    }
}
/* 
 * openLocalPalette
 *
 * sets palette to the openFacet of name.  The default view and 
 * facet are :symbolic:contents
 */
openLocalPalette( name )
char	*name; /* name of local palette in form cell[:view[:facet[:version]]] */
{
    OH_ASSERT_DESCR(ohUnpackFacetName( &oPalette, name ),
			"couldn't unpack the local palette name");
    oPalette.type = OCT_FACET;
    oPalette.contents.facet.mode = "r";
    if( octOpenFacet( &oPalette ) < OCT_OK ){
	errRaise( optProgName, -1, "Could not open %s:%s:%s: %s.",
			 oPalette.contents.facet.cell,
			 oPalette.contents.facet.view,
			 oPalette.contents.facet.facet,
			 octErrorString());
    }
}

/*
 * getInst
 *
 * gets an instance from oPalette by name, sets the transform,
 * attaches it to oFacet and the INSTANCES or CONNECTORS bag
 * The translation is set in Mkarray
 */
octObject *
getInst( name, transform )
char *name;
enum octTransformType transform;
{
    static int first = FALSE;
    static octObject oConnectors;
    static octObject oInstances;
    octObject *oInst;
    octObject oBag;
    int isInstance = FALSE;
    octStatus octRet;
    static int instNum = 0;
    static char nameBuffer[128];

    if( first++ == 0 ){
	/* get oConnectors and oInstances from oFacet by name		*/
	OH_ASSERT( ohGetByBagName(&oFacet, &oConnectors, "CONNECTORS"));
	OH_ASSERT( ohGetByBagName(&oFacet, &oInstances, "INSTANCES"));
    }
    oInst = &ObjArray[ (++ObjCount)%OBJ_ARRAY_SIZE ];
    /* get the instance from oPalette by name				*/
    if( ohGetByInstName(&oPalette, oInst, name) < OCT_OK ){
	errRaise( optProgName, -1,
			"Couldn't get %s from palette %s:%s: %s.",
			oInst->contents.instance.name,
			oPalette.contents.facet.cell,
			oPalette.contents.facet.view,
			octErrorString());
    }

    /* getContainerByName to find if INSTANCE				*/
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = "INSTANCES";
    octRet = octGetContainerByName( oInst, &oBag);
    if( octRet == OCT_OK) {
	isInstance = TRUE;
    } else if ( octRet != OCT_NOT_FOUND) {
	ohOctFailed(__FILE__, __LINE__, NIL(char), NIL(octObject));
    } 

    /* set the instance's transform					*/
    oInst->contents.instance.transform.transformType = transform;

    /* create the instance in oFacet					*/
    OH_ASSERT( octCreate( &oFacet, oInst));

    /* attach the instance to the INSTANCES or CONNECTORS bag		*/
    if ( isInstance ){
	OH_ASSERT( octAttach( &oInstances, oInst));
    } else {
	OH_ASSERT( octAttach( &oConnectors, oInst));
    }

    /* uniquifying the instance name if debug > 0 			*/
    if( Options.debug > 0 ){
	sprintf( nameBuffer,"%s.%d", oInst->contents.instance.name, ++instNum );
	oInst->contents.instance.name = nameBuffer;
	OH_ASSERT( octModify( oInst ));
    }


    return( oInst );
}

/* 
 * promoteAterm
 *
 * promotes the actual terminal named atermName to a formal terminal
 * named ftermName.  A formal terminal and net named ftermName are 
 * created in oFacet.
 *
 */
promoteAterm( oInst, atermName, ftermName, termType, direction )
octObject *oInst;
char *atermName;
char *ftermName;
char *termType;		/* value of the TERMTYPE property		*/
char *direction;	/* value of the DIRECTION property		*/
{
    octObject oNet;
    octObject oAterm;
    octObject *oFterm;
    octObject oBag;	/* for call to find INSTANCES bag */
    octObject oProp;
    octStatus octRet;
    
    /* get oAterm by name from oInst					*/
    OH_ASSERT( ohGetByTermName( oInst, &oAterm, atermName));

    /* create oFterm in oFacet						*/
    oFterm = &ObjArray[ (++ObjCount)%OBJ_ARRAY_SIZE ];
    OH_ASSERT( ohCreateTerm( &oFacet, oFterm, ftermName));

    if( termType != NIL( char ) && *termType != '\0')
	OH_ASSERT( ohCreatePropStr( oFterm, &oProp, "TERMTYPE", termType ));
    if( direction != NIL( char ) && *direction != '\0')
	OH_ASSERT( ohCreatePropStr( oFterm, &oProp, "DIRECTION", direction ));

    /* create oNet in oFacet						*/
    OH_ASSERT( ohCreateNet( &oFacet, &oNet, ftermName));

    /* attach oAterm to oFterm						*/
    OH_ASSERT( octAttach( oFterm, &oAterm));

    /* attach oFterm to oNet						*/
    OH_ASSERT( octAttach( &oNet, oFterm));

    /* if oInst is an INSTANCE, attach oAterm to oNet			*/
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = "INSTANCES";
    octRet = octGetContainerByName( oInst, &oBag);
    if( octRet == OCT_OK) {
	/* attach oAterm to oNet					*/
	OH_ASSERT( octAttach( &oNet, &oAterm));
    } else if ( octRet != OCT_NOT_FOUND) {
	ohOctFailed(__FILE__, __LINE__, NIL(char), NIL(octObject));
    }

    /*
    return( oFterm );
    */
}

/*
 * moveInterfaceProps
 *
 * move the interface properties from the contents to the interface facets
 */
moveInterfaceProps()
{
    octObject oInterface;
    octObject oTerm;		/* contents formal terminals		*/
    octObject oITerm;		/* interface formal terminals		*/
    octObject oProp;		/* properties on the formal terms	*/
    octObject oNet;		/* Vdd and GND nets			*/
    octGenerator gTerm;

    oInterface = oFacet;
    oInterface.contents.facet.mode = "w";
    oInterface.contents.facet.facet = "interface";
    VOVoutputFacet( &oInterface );
    if( octOpenFacet( &oInterface ) < OCT_OK ){
	errRaise( optProgName, -1, "Could not open %s:%s:%s: %s.",
			 oInterface.contents.facet.cell,
			 oInterface.contents.facet.view,
			 oInterface.contents.facet.facet,
			 octErrorString());
    }

    /* add properties - RLS */
    OH_ASSERT( ohCreatePropStr( &oInterface, &oProp, "VIEWTYPE", "SYMBOLIC" ));
    OH_ASSERT( ohCreatePropStr( &oInterface, &oProp, "EDITSTYLE", "SYMBOLIC" ));
    OH_ASSERT( ohCreatePropStr( &oInterface, &oProp, "CELLCLASS", "MODULE" ));
    OH_ASSERT( ohCreatePropStr( &oInterface, &oProp, "CELLTYPE", "MODULE" ));
    OH_ASSERT( ohCreatePropStr( &oInterface, &oProp, "TECHNOLOGY", 
				 Options.technology ));

    oITerm.type = OCT_TERM;
    OH_ASSERT( octInitGenContents( &oFacet, OCT_TERM_MASK, &gTerm ));
    while( octGenerate( &gTerm, &oTerm ) == OCT_OK ){
	oITerm.contents.term = oTerm.contents.term;
	if( octGetByName( &oInterface, &oITerm ) != OCT_OK ){
	    errRaise( optProgName, -1,
			    "Getting term %s from %s:%s:%s: %s.",
			     oITerm.contents.term.name,
			     oInterface.contents.facet.cell,
			     oInterface.contents.facet.view,
			     oInterface.contents.facet.facet,
			     octErrorString());
	}
	if( ohGetByPropName( &oTerm, &oProp, "TERMTYPE" ) == OCT_OK ){
	    OH_ASSERT( octDelete( &oProp ));
	    OH_ASSERT( octCreate( &oITerm, &oProp ));
	}
	if( ohGetByPropName( &oTerm, &oProp, "DIRECTION" ) == OCT_OK ){
	    OH_ASSERT( octDelete( &oProp ));
	    OH_ASSERT( octCreate( &oITerm, &oProp ));
	}
    } 

    /* mark the supply and ground nets in oFacet			*/
    if( ohGetByTermName( &oFacet, &oTerm, "Vdd") == OCT_OK){
	if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet) == OCT_OK){
	    OH_ASSERT( ohCreatePropStr( &oNet, &oProp, "NETTYPE", "SUPPLY" ));
	}
    }
    if( ohGetByTermName( &oFacet, &oTerm, "GND") == OCT_OK){
	if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet) == OCT_OK){
	    OH_ASSERT( ohCreatePropStr( &oNet, &oProp, "NETTYPE", "GROUND" ));
	}
    }

    OH_ASSERT( octFlushFacet( &oInterface ));
}
