/*
 * parse commmand line and set values in options structure.
 */


#include "copyright.h"
#include "port.h"
#include "anchor.h"

optionStruct optionList[] = {
    { "a",	"", "Force annealing. By default, if a file named `cell'.pm exists (where `cell' is the output cell name), it will be used to generate the layout, and the annealing step will be skipped."},
    { "A",	"", "Force annealing and don't generate layout.  A personality matrix file is written (name determined by -m option or its default)."},
    { "C",	"buffer","Specify clocking for buffers."},
    { OPT_CONT, 0,"\n    legal types are 'in', for clocked input buffer."},
    { OPT_CONT, 0,"\n                    'out', for clocked output buffer."},
    { OPT_CONT, 0,"\n                    'pre', for clocked precharged output buffer."},
    { "d",	"level",	      "Set the level of debugging."},
    { OPT_CONT, 0,"\n    0   : run silently and quickly ( default )."},
    { OPT_CONT, 0,"\n    1-n : do various special things depending on value."},
    { "f",	0, "Use fast annealing schedule."},
    { "F",	0, "Don't do any folding. Folding is done by default."},
    { "g",	"", "Print the mkarray constraint graphs."},
    { "i",	"filename", "Name of the input pla(5OCTTOOLS) format file.  MisII and espresso can both generate this type of file."},
    { OPT_CONT, 0, "\tDefaults to `cell'.pla (where `cell' is the output cell name)"},
    { "m",	"filename", "Set personality matrix file name. Used either to read or write the annealed personality matrix. Defaults to `cell'.pm, where `cell' is the output cell name"},
    { "r",	0, "randomly seed the random number generator for random annealing results"},
    { "p",	"palette_name", "Palette name known to tap(3OCTTOOLS)."},
    { "P",	"user_palette[:view[:facet[:version]]]",
	"Name of user palette. Overrides the '-p' option."},
    { "s",	"separation", "Give ground line separation in AND plane."},
    { "S",	"separation", "Give ground line separation in OR plane."},
    { "t",	"level", "Set the type and frequency of tracing for ace."},
    { "T",	"technology", "Name of the technology."},
    { "v",	"filename", "Output a pla(5CAD) format file for verification/comparison with the given pla(5OCTTOOLS) file."},
    { "w",	"",           "Suppress wiring."},
    { OPT_RARG, "cell[:view[:facet[:version]]]",
	"Name, view, facet, and version of the output cell."},

    { OPT_DESC, 0, "\nAnchor is a programmable logic array (PLA) module generator for the Oct CAD environment.  It takes a pla(5OCTTOOLS) PLA description, optionally folds it, and generates a NAND-NAND XXX style PLA.  For more information, see anchor(1OCTTOOLS)."},
    { 0,        0,               0}
};

/* global variables set by the command line				*/

extern s_options Options;

/*
 * parseComLine
 *
 * parse the command line 
 */
parseComLine( argc, argv )
int argc;
char **argv;
{

    extern char	*optarg;
    int		option, level;
    char	*mapToLower();
    char 	*util_strsav();

    while(( option = optGetOption( argc,argv)) != EOF ){
	switch ( option ) {
	case 'a':
	    Options.forceAnneal = TRUE;
	    break;
	case 'A':
	    Options.forceAnneal = TRUE;
	    Options.noLayout = TRUE;
	    break;
	case 'f':
	    Options.fastAnneal = TRUE;
	    break;
	case 'F':
	    Options.fold = FALSE;
	    break;
	case 'T':
	    Options.technology = util_strsav( optarg );
	    break;
	case 'r':
	    aceRandomSeed();
	    break;
	case 'p':
	    Options.paletteName = util_strsav( optarg );
	    break;
	case 'P':
	    Options.localPaletteName = util_strsav( optarg );
	    break;
	case 'C':
	    if (!strcmp (optarg, "in")) {
		Options.clocking |= CLOCKED_INPUT;
	    } else if (!strcmp (optarg, "out")) {
		Options.clocking |= CLOCKED_OUTPUT;
		/*
		  } else if (!strcmp (optarg, "pre")) {
		  Options.clocking |= CLOCKED_PRE_OUT;
		  */
	    } else {
		optUsage();
	    }
	    break;
	case 'd':
	    Options.debug = atoi( optarg );
	    break;
	case 'g':
	    Options.printGraph = TRUE;
	    break;
	case 'i':
	    Options.inPlaFile = util_strsav( optarg );
	    break;
	case 'm':
	    Options.outPMFile = util_strsav( optarg );
	    Options.inPMFile = Options.outPMFile;
	    break;
	case 'v':
	    Options.plaVerifyFile = util_strsav( optarg );
	    break;
	case 's':
	    Options.andGndLineSep = atoi (optarg );
	    break;
	case 'S':
	    Options.orGndLineSep = atoi (optarg );
	    break;
	case 'w':
	    Options.wire = FALSE;
	    break;
	case 't':
	    level = atoi( optarg );
	    if( level == 0){
		Options.traceLevel = 0;
	    } else if( level == 1){
		Options.traceLevel = ACE_TRACE_COST | ACE_TRACE_TEMP |
		  ACE_TRACE_CPU_TIME | ACE_TRACEF_FINAL |
		    ACE_TRACEF_INITIAL;
	    } else if( level == 2){
		Options.traceLevel = ACE_TRACE_COST | ACE_TRACE_TEMP |
		  ACE_TRACEF_FINAL | ACE_TRACEF_INITIAL |
		    ACE_TRACEF_STATE;
	    } else if( level == 3){
		Options.traceLevel = ACE_TRACE_COST | ACE_TRACE_TEMP |
		  ACE_TRACE_GEN_MOVE | ACE_TRACE_CPU_TIME |
		    ACE_TRACEF_FINAL | ACE_TRACE_MOVE_RESULTS |
		      ACE_TRACEF_INITIAL | ACE_TRACEF_STATE;
	    } else if( level == 4){
		Options.traceLevel = ACE_TRACE_COST | ACE_TRACE_TEMP |
		  ACE_TRACE_GEN_MOVE | ACE_TRACE_MOVE_RESULTS |
		    ACE_TRACE_COM_COST | ACE_TRACE_CPU_TIME |
		      ACE_TRACEF_FINAL | ACE_TRACEF_INITIAL |
			ACE_TRACEF_STATE;
	    } else if( level == 5){
		Options.traceLevel = ACE_TRACE_COST | ACE_TRACE_TEMP |
		  ACE_TRACE_GEN_MOVE | ACE_TRACEF_FINAL |
		    ACE_TRACEF_INITIAL | ACE_TRACEF_STATE |
		      ACE_TRACEF_GEN_MOVE;
	    } else {
		fprintf( stderr, "%s: unknown level %d. Use 0 through 5.\n",
			optProgName, level);
		optUsage();
	    }
	    break;
	default:
	    optUsage();
	    break;
	}
    }
    if(( argc == 1 )||( optind == argc )){
	optUsage();
	exit(-1);
    }
    Options.plaFacet = util_strsav (argv[ optind ] );
}
