/*
 * pmIO.c
 *
 * Ponality matrix read/write routines.
 * This format is used for intermediate storage of the annealed matrix.
 * Backwards compatibility with genie/octopus isn't guarenteed (especially
 * in light of the sanitizeMatrix issues).
 */

#include "copyright.h"
#include "port.h"
#include "anchor.h"
#include "vov.h"

extern int	Rows;	/* dimentions given in personality matrix */
extern int	AndCols, OrCols;
s_input		**Inputs;
s_output	**Outputs;
extern char	**Core, **allocCore();


/*
 * readPM
 *
 * Read a personality matrix file of the type written by writePM().
 */
readPM(fileName)
char	*fileName;
{
    FILE	*fopen(), *ifp;
    int		allCols, i, j, c, maxAndCol;
    char	matrixType[ NAMESIZE ];


    /* open input file */
    if (( ifp = VOVfopen( fileName, "r" )) == NIL(FILE)) {
	errRaise( optProgName, 1, "Could not open \"%s\".", fileName);
    }
   
    /* read the first line, which consists of a string and 2 numbers which
     * define the size of the array.
     */
    if( fscanf(ifp, "%s %d %d", matrixType, &Rows, &allCols) != 3 ){
	errRaise( optProgName, 1, "%s %s",
		"The first line of personality matrix should",
		"have a name and two integers.");
    }

    /*** here we could check the matrixType				*/

    /* allocate memory for Core matrix				*/ 
    Core = allocCore(Rows, allCols);
   
    /* read characters in personality matrix				*/
    maxAndCol = -1;
    for( i = Rows - 1; i >= 0; --i){
	/* get rid of white space */
	while( isspace( c = getc(ifp) )) ;
	if( c == '(' ){
	    errRaise( optProgName, -1, "%s %s",
		    "Not enough rows in personality matrix character array.");
	}
	(void)ungetc( c, ifp);
	/* read a line of the character array				*/
	for( j=0; j < allCols; ++j){
	    if(((c = getc(ifp)) == DOT) || (c == SUM) || (c == PRODUCT)){
		Core[ i ][ j ] = c;
		if(( c == SUM )  && ( j > maxAndCol)) maxAndCol = j;
	    } else if( c == '\n'){
		errRaise( optProgName, 1,
			"personality matrix not wide enough.");
	    } else if( c == EOF){
		errRaise( optProgName, 1,
			"unexpected EOF while reading personality matrix.");
	    } else {
		errRaise( optProgName, 1,
			"Unrecognizable character '%c' in personality matrix.",
			(char) c);
	    }
	}	/* finish reading a line */

	/* read to the end of the line */
	while((c = getc(ifp)) != '\n'){
	    switch( c ){
		case EOF: 
		    errRaise( optProgName, 1,
			"Unexpected EOF while reading personality matrix.");
		    break;
		case ' ':
		case '\t': break;
		default: 
		    errRaise( optProgName, 1,
		      "Unrecognizable char '%c' in personality matrix[%d,%d].",
				c, j, i);
	    }
	}
    } /* end reading core matrix					*/
    AndCols = maxAndCol + 1;
    OrCols = allCols - AndCols;

    if(( AndCols <= 0) || (OrCols <= 0)){
	errRaise( optProgName, 1, "readPM: bad AndCols or OrCols number");
    }

    /* remove white space between matrix and row definitions		*/
    while( isspace( c = getc(ifp) )) ;

    if( c != '(' ){
	errRaise( optProgName, -1, "Too many lines in matrix core.");
    } else {
	(void)ungetc(c, ifp);
    }

    /* read the signal names and locations				*/
    readSignals( ifp );

    (void)fclose( ifp );
}


/*
 * readSignals
 *
 * Read the row and column signal definitions
 * ifp is assumed to be at the begining of the first (row ...) definition.
 */
readSignals( ifp )
FILE	*ifp; /* input file pointer					*/
{
    char keyword[ NAMESIZE ], sigName[ STRLEN ];
    int	index, start, end;
    int	c;
    s_input *getInputByName(), *sig1, *sig2; /* for assigning complements*/

    /* allocate memory for Inputs and Outputs arrays			*/
    Inputs  = ALLOC( s_input *, AndCols);
    memset((char *)Inputs, 0, sizeof(s_input *) * AndCols);

    Outputs = ALLOC( s_output *, OrCols);
    memset((char *)Outputs, 0, sizeof(s_output *) * OrCols);

    /* read column definitions						*/
    while( fscanf(ifp, " ( %s", keyword) == 1){
	if(strcmp( keyword, "col") == 0){
	    if( fscanf( ifp, "%d", &index) != 1){
		errRaise( optProgName, 1, "bad format of col line");
	    }
	    while( scanSignal( ifp, sigName, &start, &end)){
		/* append to either inputs or outputs list at index	*/
		if( index < AndCols){
		    /* process this row as inputs			*/
		    appendInput(index, sigName, start, end);
		} else {
		    /* process this row as outputs			*/
		    appendOutput(index, sigName, start, end);
		}
	    }
	} else if( strcmp(keyword, "complement") == 0){
	    if((fscanf( ifp, "%s", sigName) == 1) && (strcmp(")", sigName))){
		sig1 = getInputByName(sigName);
		if((fscanf( ifp,"%s", sigName) == 1) && (strcmp(")", sigName))){
		    /* have two signals					*/
		    sig1->phase = POS;
		    sig2 = getInputByName(sigName);
		    sig2->phase = NEG;
		    sig1->comp = sig2;
		    sig2->comp = sig1;
		} else {
		    /* only one signal					*/
		    sig1->phase = NEG;
		}
	    }
	} else if(strcmp (keyword, "row") == 0){
	    /* Ignore it and read to end of line.			 */
	} else {
	    /* unknown keyword						*/
	    errRaise( optProgName, -1,
			"Unknown keyword \"%s\" in personality matrix file.",
			keyword);
	}
	while( ((c = getc(ifp)) != '\n') && (c != EOF)) ;
	ungetc( c, ifp );
    }
}

/*
 * scanSignal
 *
 * read in one signal.
 */
scanSignal( ifp, name, start, end)
FILE *ifp;
char *name;
int *start, *end;
{
    if( fscanf( ifp, " ( %s %d %d )", name, start, end) == 3){
	return( TRUE );
    } else {
	return( FALSE );
    }
}


/*
 * appendInput
 *
 * Add a new input on to the list of inputs at Inputs[column].
 */
appendInput( column, sigName, start, end)
int	column;
char	*sigName;
int	start, end;
{
    s_input	*getInputByName(), *input;

    if((input = getInputByName(sigName)) == NIL(s_input)){
	errRaise( optProgName, 1,
		"appendInput: Could not get input siginal \"%s\".", sigName);
    }

    input->col = column;
    input->bottom = start;
    input->top = end;

    if( input->comp == NIL(s_input)){
	input->comp = getInputByName( formSigNameComp( input->name ));
	input->comp->col = -1;
	input->comp->comp = input;
    }

    addInput( input );
}

/* appendOutput
 *
 * Add a new output.
 */
appendOutput( column, sigName, start, end)
int	column;
char	*sigName;
int	start, end;
{
    s_output	*newOutput(), *output;

    output = newOutput ( sigName );

    output->col = column;
    output->bottom = start;
    output->top = end;

    addOutput( output );
}

/*
 * writePM
 *
 * Write out a personality matrix.
 */
writePM( fileName )
char *fileName;
{
    int i, j;
    FILE *fp;
    s_input *in;

    if(( fp = VOVfopen( fileName, "w")) == NIL(FILE)){
	errRaise( optProgName, 1,
		"couldn't open output personality matrix file \"%s\".",
		fileName);
    }

    /* write header							*/
    fprintf( fp, "pla %d %d\n", Rows, AndCols + OrCols);

    /* write out the core of the matrix					*/
    for( i = Rows - 1; i >= 0; --i){
	for( j = 0; j < AndCols + OrCols; ++j){
	    fprintf( fp, "%c", Core[i][j]);
	}
	fprintf( fp, "\n");
    }

    /* write input columns						*/
    for( i = 0; i < AndCols; ++i){
	fprintf( fp, "( col %d", i);
	if( Inputs[i] != NIL(s_input)){
	    fprintf( fp, " ( %s %d %d)", Inputs[i]->name, Inputs[i]->bottom,
			Inputs[i]->top);
	    if( Inputs[i]->next != NIL(s_input)){
		fprintf( fp, " ( %s %d %d)", Inputs[i]->next->name,
			Inputs[i]->next->bottom, Inputs[i]->next->top);
	    }
	}
	fprintf( fp, ")\n");
    }
    /* write output columns						*/
    for( i = 0; i < OrCols; ++i){
	fprintf( fp, "( col %d", AndCols + i);
	if( Outputs[i] != NIL(s_output)){
	    fprintf( fp, " ( %s %d %d)", Outputs[i]->name, Outputs[i]->bottom,
			Outputs[i]->top);
	    if( Outputs[i]->next != NIL(s_output)){
		fprintf( fp, " ( %s %d %d)", Outputs[i]->next->name,
			Outputs[i]->next->bottom, Outputs[i]->next->top);
	    }
	}
	fprintf( fp, ")\n");
    }
    /* write complements						*/
    for( i = AndCols - 1; i >= 0; --i){
	for( in = Inputs[ i ]; in != NIL(s_input); in = in->next){
	    if(( in->phase == POS) && ( in->comp != NIL(s_input)) &&
	       ( in->comp->col != -1)){
		fprintf( fp, "( complement %s %s )\n",
			in->name, in->comp->name);
	    } else if(( in->phase == NEG) &&
		      (( in->comp == NIL(s_input)) || ( in->comp->col < 0))){
		fprintf( fp, "( complement %s )\n", in->name);
	    }
	}
    }
    (void)fclose( fp );
}
