/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSMain.c

Main routine for channel definition slicing structure.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSMain.c,v $
 * Revision 1.2  90/02/07  16:10:12  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:45  octtools
 * Initial revision
 * 
 * Revision 1.7  88/05/07  19:47:51  mitch
 * Fixed bug with disappearing real rooms
 * 
 * Revision 1.6  88/04/15  16:10:00  mitch
 * Fix updating of oct facet to check if there is already a channel
 * definition tree.  Also check for old corner bags.
 * Fix CDSOctPurge was leaving one cornerEdge for each instance.
 * 
 * Revision 1.5  88/04/15  14:08:38  mitch
 * Put in channel path for used external channels.
 * 
 * Revision 1.4  88/04/08  11:29:41  mitch
 * Put in octBegin and octEnd into main procedure.
 * 
 * Revision 1.3  88/04/05  16:16:24  mitch
 * Fixed bug in initializing CDSMaxNumCh used in printing the sequence file.
 * 
 * Revision 1.2  88/04/05  12:26:10  mitch
 * Made all operations off of one oct facet instead of several like cdmerge
 * and cdexpandxxx views.  This was done so that atlasRpc would work.
 * Added code to try to resolve instance overlaps by expansion.
 * Also z-shaped channel handling was made more robust.
 * 
 * Revision 1.1  88/03/31  13:53:49  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#ifndef lint
static char RCSId[] = "@(#) $Header: /octtools/release-3.0/src/atlas/RCS/CDSMain.c,v 1.2 90/02/07 16:10:12 octtools Exp Locker: octtools $";
#endif
#include "copyright.h"
#include "port.h"
#include "options.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"
#include "vov.h"

/* Main procedure for channel definition
----------------------------------------------------------------------------*/
main(argc, argv)
    int		argc;
    String	argv[];
{
    int		rc;
    int		CDSMainFromFacet();
    octObject	atlasMain;
    void	dummyPgm();

    CDSInitParm(argc, argv);

    VOVbegin( argc, argv );

    MIPrefix = "atlas";

    {
	char	month[10];
	int	date, year;

	printf("%s: Channel definition and ordering (Rev 2.0", MIPrefix);
#ifdef CUR_DATE
	printf(" compiled %s)\n", CUR_DATE);
#else				/* CUR_DATE */
	printf(")\n");
#endif				/* CUR_DATE */
    }

    MIModStatStart("Atlas", "debugAtlas", TRUE);

    MI_STACK_ALLOC(10000);
    MI_QUEUE_ALLOC(10000);

    CDSOctMode = TRUE;

    if (CDSOctMode) {
	CDSOctCopyAndOpen(CDSParm.cellName, CDSParm.inputView, CDSParm.outputView,
			  &atlasMain);
    }

    rc = CDSMainFromFacet(&atlasMain, dummyPgm);

    if (rc!=99) {
	OH_ASSERT(octCloseFacet(&atlasMain));
    }

    MI_STACK_FREE();
    MI_QUEUE_FREE();

    MIModStatEnd();
    VOVend(rc);
}				/* ............................. End of main procedure */

void dummyPgm() {MIerror("Compile and load libraries incompatible\n");}

optionStruct optionList[] = {
    { "s",	"","show channel cuts"},
    { "I",	"","merge ios only"},
    { "i",	"","detect and automatically merge ios around perimeter"},
    { "z",	"","use z-shaped channels"},
    { "b",	"","ignore blockages in PLACE layer"},
    { "e",	"expand_factor",
		    "expansion factor number > 1.0 (default=1.2)"},
    { "o",	"output_view","output view name (default is \"inputView\".cd)"},
    { OPT_RARG,	"cell:view","input facet for channel definition"},
    { 0,	0, 0}
};
/* Init parms
----------------------------------------------------------------------------*/
CDSInitParm(argc, argv)

int		argc;
String		argv[];
{
    int		option;
    String		strchr();
    char		*outViewSuffix = ".cd";

    CDSParm.cellName     = NULL;
    CDSParm.inputView    = NULL;
    CDSParm.outputView   = NULL;
    CDSParm.mode         = "atlas";
    CDSParm.showChanFlg  = FALSE;
    CDSParm.mergeIoFlg   = FALSE;
    CDSParm.mergeIoOnlyFlg = FALSE;
    CDSParm.rpcFlg       = FALSE;
    CDSParm.zChanFlg     = FALSE;
    CDSParm.useBlkFlg    = TRUE;
    CDSParm.expandFactor = 1.2;

    while ((option=optGetOption(argc, argv))!=EOF) {
	switch(option) {
	case 's':	CDSParm.showChanFlg = TRUE;
	    break;
	case 'I':	CDSParm.mergeIoOnlyFlg = TRUE;
	    break;
	case 'i':	CDSParm.mergeIoFlg = TRUE;
	    break;
	case 'z':	CDSParm.zChanFlg = TRUE;
	    break;
	case 'b':	CDSParm.useBlkFlg = FALSE;
	    break;
	case 'e':	sscanf(optarg, "%lf", &CDSParm.expandFactor);
	    break;
	case 'o':	CDSParm.outputView = STR_COPY(optarg);
	    break;
	default:	optUsage();
	}
    }

    if (optind+1!=argc) optUsage();

    CDSParm.cellName = STR_COPY(argv[optind]);
    CDSParm.inputView = strchr(CDSParm.cellName,':');
    if (CDSParm.inputView) {
	*CDSParm.inputView = '\0';
	CDSParm.inputView++;
    }
    if (!CDSParm.inputView || !(*CDSParm.inputView))
      optUsage();

    if (!CDSParm.outputView) {
	CDSParm.outputView =
	  ALLOC_ARRAY(strlen(CDSParm.inputView)+strlen(outViewSuffix)+1, char);
	(void) strcpy(CDSParm.outputView, CDSParm.inputView);
	(void) strcpy(CDSParm.outputView+strlen(CDSParm.inputView), outViewSuffix);
    }

    MSG(1, ("Cell name <%s>, inView <%s> and outView <%s>\n",
	    CDSParm.cellName, CDSParm.inputView, CDSParm.outputView));

}				/* ............................. End of procedure CDSInitParm */
