/*LINTLIBRARY*/
/*
MIUtil.c:

General C-language utility routines (to be used in conjunction with MIInc.h)

MIModStatStart		Initialize statistics and debug environment.
MIModStatEnd		Print out statistics and free up debug environment.
MIInitEnv		Changing of debug environment (interrupt handler).
MIFindSubIdx		Find subroutine index (if it does not exist then add new
			entry to module table).
MIPrintRev		Print out revision and date stamp.
MIfopen			fopen with system error reporting.
MIprintf		printf with (void) fflush.
MIfprintf		fprintf with (void) fflush.
MITokenize		Tokenize a string.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	MIUtil.c,v $
 * Revision 1.6  90/02/07  16:10:46  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.6  89/03/13  13:43:28  octtools
 * signal.h
 * 
 * Revision 1.5  89/03/12  21:55:47  octtools
 * SIGNAL_FN
 * 
 * Revision 1.4  89/03/12  14:31:03  octtools
 * SIGNAL_FN
 * 
 * Revision 1.3  89/02/08  20:48:15  octtools
 * got rid of sbrk, was not working on PMAX
 * 
 * Revision 1.2  89/02/08  15:53:44  octtools
 * put in portable varargs code
 * 
 * Revision 1.1  89/01/26  14:20:19  octtools
 * Initial revision
 * 
 * Revision 1.3  88/04/24  16:56:29  mitch
 * Put in package prefix
 * 
 * Revision 1.1  88/03/31  15:11:42  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <varargs.h>
#include "uprintf.h"
#define MIInc_MAIN
#include "MIInc.h"

SIGNAL_FN	(*MISaveSig)();
Boolean		MISigSavedFlg = FALSE;
FILE		*MIEnvFile;

long		MIModStatTime;
extern int	errno;			/* System error number */
extern String	sys_errlist[];		/* SysErr reason (indexed by errno) */

/* Initialize module statistics and control array.
   Reset interrupt handler to MIInitEnv.
----------------------------------------------------------------------------*/
MIModStatStart(name, debugFileName, prtHeaderFlag)

String		name;
String		debugFileName;
Boolean		prtHeaderFlag;
{
short		m;
long		tloc;
extern long	time();
extern char	*ctime();
char		tm[128];
short		MIFindModIdx();
SIGNAL_FN	MIInitEnv();

#ifdef	lint
String		MITokenize();		/* Reference MITokenize & MIPrintRev */
MI_REV_STAMP((String)MITokenize((String)NULL,(String)NULL,(String)NULL), RCSId)
#endif

MIPrtHeaderFlag = prtHeaderFlag;
tloc = time((long*)0);
if (MIPrtHeaderFlag) {
    (void) strcpy(tm, ctime(&tloc));
    MIfprintf(MIDBG,"-- Start \"%s\" on %s", name, tm);
}

if (MIAllocDebugFlag) {
    MIAllocFile = MIfopen("allo","w","alloc/free debug");
    if (!MIAllocFile)
	MIAllocDebugFlag = FALSE;
}

MIModStat   = ALLOC_ARRAY(MI_MOD_STAT_DIM, MIModStatDef);
MITraceBack = ALLOC_ARRAY(MI_TRACE_BACK_DIM, short);
MINumMod    = MINumTraceBack = 0;
m = MIFindModIdx(name, FALSE);
DASSERT(m==0);

if (debugFileName) {
    if (MIEnvFile=MIfopen(debugFileName,"r","run-time debug")) {
					/* Save default interrupt hander */
	MISaveSig = signal(SIGINT, MIInitEnv);
	MISigSavedFlg = TRUE;
	MIInitEnv();
    }
    else MISigSavedFlg = FALSE;
}
else MISigSavedFlg = FALSE;

MIModStatTime = util_cpu_time();

} /* ............................. End of procedure MIModStatStart */
/* Print out runtime statistics of all called subroutines.
   Reset interrupt handler to default.
----------------------------------------------------------------------------*/
MIModStatEnd()
{
short		m;
long		tloc;
extern long	time();
extern char	*ctime();
char		tm[128];

if (MISigSavedFlg)			/* Restore interrupt handler */
    MISaveSig = signal(SIGINT, MISaveSig);
RANGE(0, MINumTraceBack, 1);

for (m=1; m<MINumMod; m++) {		/* Cpu times of each module */
    if (MIModStat[m].timeFlag && MIModStat[m].numTime>0) {
	MIMSG(-1, (MIDBG,"Cpu time for module %-16s :%8.3f sec%10dx",
	    MIModStat[m].name, ((double) MIModStat[m].cTime)/1000.0,
	    MIModStat[m].numTime));
    }
    else if (MIModStat[m].numTime>0) {
	MIMSG( 1, (MIDBG,"Cpu time for module %-16s :%8.3f sec%10dx",
	    MIModStat[m].name, ((double) MIModStat[m].cTime)/1000.0,
	    MIModStat[m].numTime));
    }
    if (MIModStat[m].nameAlloFlag)
	FREE_STRING(MIModStat[m].name);
}

MIModStatTime = util_cpu_time() - MIModStatTime;

MIMSG(0, (MIDBG,
    "time u+s=%.2f\n", ((double) ModModStatTime) / 1000.0));

tloc = time((long*)0);
if (MIPrtHeaderFlag) {
    MIfprintf(MIDBG,"\nTotal Cpu Time = %.2f sec\n",
	      ((double) util_cpu_time()) / 1000.0);
    (void) strcpy(tm, ctime(&tloc));
    MIfprintf(MIDBG,"-- End \"%s\" on %s", MIModStat[0].name, tm);
}
					/* Final clean up */
if (MIAllocDebugFlag) {
    (void) fclose(MIAllocFile);
    MIAllocDebugFlag = FALSE;
}

FREE_ARRAY(MI_MOD_STAT_DIM, MIModStat, MIModStatDef);
FREE_ARRAY(MI_TRACE_BACK_DIM, MITraceBack, short);

} /* ............................. End of procedure MIModStatEnd */
/* Initialization of runtime environment (interrupt handler).
----------------------------------------------------------------------------*/
SIGNAL_FN MIInitEnv()
{
FILE		*newEnvFile;
short		m, t;
int		debug;
int		threshold;
short		MIFindModIdx();

while(1) {
    if (MIEnvFile==stdin) {		/* Prompt (only in interactive mode) */
	(void) fflush(stdout);
	fprintf(stderr,"\ninitEnv> ");
	(void) fflush(stderr);
    }
    if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;

    if	    (!strcmp (MIBuf,"/*")) {	/* Skip comments */
	while(fscanf(MIEnvFile,"%s", MIBuf)!=EOF)
	    if (!strcmp(MIBuf,"*/")) break;
    }
    else if (!strncmp(MIBuf,"debug",1)) {
	if (fscanf(MIEnvFile,"%s %d", MIBuf, &debug)==EOF) break;
	if (!strcmp(MIBuf,"default"))
	    MIDefaultDebugFlag = debug;
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].debugFlag = debug;
	}
    }
    else if (!strncmp(MIBuf,"trace",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	if (!strcmp(MIBuf,"default")) {
	    MIDefaultTraceFlag = !MIDefaultTraceFlag;
	    if (fscanf(MIEnvFile,"%d", &threshold)==EOF) break;
	    MIModThreshold = threshold;
	}
	else if (!strcmp(MIBuf,"alloc")) {
	    if (MIAllocDebugFlag) {
		(void) fclose(MIAllocFile);
		MIAllocDebugFlag = FALSE;
	    }
	    else {
		MIAllocFile = MIfopen("allo","w","alloc/free debug");
		if (MIAllocFile)
		    MIAllocDebugFlag = TRUE;
	    }
	}
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].traceFlag = !MIModStat[m].traceFlag;
	}
    }
    else if (!strncmp(MIBuf,"time",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	if (!strcmp(MIBuf,"default"))
	    MIDefaultTimeFlag = !MIDefaultTimeFlag;
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].timeFlag = !MIModStat[m].timeFlag;
	}
    }
    else if (!strcmp(MIBuf,"abort")) {
	abort();
    }
    else if (!strncmp(MIBuf,"status",3)) {
	fprintf(stderr,"defaults: debugLevel=%d, %s traced(t), %s timed(.)\n",
	    MIDefaultDebugFlag,
	    MIDefaultTraceFlag ? "" : " not",
	    MIDefaultTimeFlag  ? "" : " not");
	for (m=0; m<MINumMod; m++) {
	    if (m>0 && m%2==0) fprintf(stderr,"\n");
	    fprintf(stderr,"%-20s %c%c %2df%7dx%c   ",
		MIModStat[m].name,
		(MIModStat[m].stopFlag  ? 's' : ' '),
		(MIModStat[m].traceFlag ? 't' : ' '),
		MIModStat[m].debugFlag, MIModStat[m].numTime,
		(MIModStat[m].timeFlag ? '.' : ' '));
	}
    }
    else if (!strncmp(MIBuf,"stop",3)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	m = MIFindModIdx(MIBuf, TRUE);
	MIModStat[m].stopFlag = !MIModStat[m].stopFlag;
    }
    else if (!strncmp(MIBuf,"where",1)) {
	if (MITraceBackFlag) 
	    for (t=MINumTraceBack-1; t>=0; t--) {
		m = MITraceBack[t];
		fprintf(stderr,"%s\n", MIModStat[m].name);
	    }
	else if (0<=MICurrentMod && MICurrentMod<MINumMod)
	    fprintf(stderr,"%s", MIModStat[MICurrentMod].name);
	else
	    fprintf(stderr,"ERROR Cannot determine at this time");
    }
    else if (!strncmp(MIBuf,"file",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	newEnvFile = MIfopen(MIBuf,"r","debug setup");
	if (newEnvFile) {
            if (MIEnvFile!=stdin)
		(void) fclose(MIEnvFile);
	    MIEnvFile = newEnvFile;
	}
    }
    else if (!strncmp(MIBuf,"nosignal",1)) {
	if (MISigSavedFlg) {
	    MISaveSig = signal(SIGINT, MISaveSig);
	    MISigSavedFlg = FALSE;
	}
    }
    else if (!strncmp(MIBuf,"foreground",2)) {
	if (MIEnvFile!=stdin) {
	    (void) fclose(MIEnvFile);
	    MIEnvFile = stdin;
	}
    }
    else if (!strncmp(MIBuf,"continue",1)) {
	if (MIEnvFile!=stdin) {
	    (void) fclose(MIEnvFile);
	    MIEnvFile = stdin;
	}
	break;
    }
    else if (!strcmp(MIBuf,"quit") || !strcmp(MIBuf,"exit"))
	exit(-1);
    else if (!strncmp(MIBuf,"help",1) ||
    	     !strncmp(MIBuf,"?",1)) {
	fprintf(stderr,"Summary of Commands:\n");
	fprintf(stderr,"  debug <subrout or \"default\"> <option>\n");
	fprintf(stderr,"  trace <subrout or \"default\" or \"alloc\"> <threshold>\n");
	fprintf(stderr,"  time  <subrout or \"default\">\n");
	fprintf(stderr,"  status\n");
	fprintf(stderr,"  stop <module name>\n");
	fprintf(stderr,"  file <input file name>\n");
	fprintf(stderr,"  foreground\n");
	fprintf(stderr,"  where\n");
	fprintf(stderr,"  continue or ctrl-d\n");
	fprintf(stderr,"  quit, exit or ctrl-c");
    }
    else
	if (MIEnvFile==stdin)
	    fprintf(stderr,"ERROR Cannot interpret command <%s>", MIBuf);
	else
	    MIfprintf(MIERR,"ERROR Cannot interpret command <%s>\n", MIBuf);
}

if (MIEnvFile!=stdin)			/* Input file must be stdin */
    MIEnvFile = stdin;
} /* ............................. End of procedure MIInitEnv */
/* Returns subroutine index (if it does not exist then add a new entry into
   the module table).
----------------------------------------------------------------------------*/
short
MIFindModIdx(name, nameAlloFlag)

String		name;			/* name of module */
Boolean		nameAlloFlag;		/* Alloc separate storage for name ? */
{
short		m;

for (m=0; m<MINumMod; m++)		/* Search for name */
    if (!strcmp(name, MIModStat[m].name)) break;

if (m==MINumMod) {			/* Name not found - add new entry */
    URANGE(MINumMod, MI_MOD_STAT_DIM);
    if (!nameAlloFlag)
    then MIModStat[MINumMod].name = name;
    else MIModStat[MINumMod].name = MI_STR_COPY(name);
    MIModStat[MINumMod].nameAlloFlag = nameAlloFlag;
    MIModStat[MINumMod].cTime        = 0;
    MIModStat[MINumMod].numTime      = 0;
    MIModStat[MINumMod].stopFlag     = FALSE;
    MIModStat[MINumMod].debugFlag    = MIDefaultDebugFlag;
    MIModStat[MINumMod].traceFlag    = MIDefaultTraceFlag;
    MIModStat[MINumMod].timeFlag     = MIDefaultTimeFlag;
    for (; *name; name++)
	if (!strncmp(name,"Main",4)) {
	    MIModStat[MINumMod].timeFlag = TRUE;
	    break;
	}
    MINumMod++;
}

return(m);				/* Return index of module entry */

} /* ............................. End of procedure MIFindModIdx */
/* Called version of subroutine start
----------------------------------------------------------------------------*/
MIModStart(name, modNum, debugFlag, startTime)

String		name;
short		*modNum;
short		**debugFlag;
long 		*startTime;
{
short		MIFindModIdx();

if (*modNum<0) {
    *modNum    = MIFindModIdx(name, FALSE);
    *debugFlag = &(MIModStat[*modNum].debugFlag);
}

if (MITraceBackFlag)
then MITraceBack[MINumTraceBack++] = *modNum;
else MICurrentMod = *modNum;

if (MIModStat[*modNum].timeFlag)
    *startTime = util_cpu_time();
if (MIModStat[*modNum].traceFlag && MIModStat[*modNum].numTime<MIModThreshold)
    MIfprintf(MIDBG,"--Start %s\n", MIModStat[*modNum].name);
if (MIModStat[*modNum].stopFlag) {
    MI_BEEP();
    MIInitEnv();
}
} /* ............................. End of procedure MIModStart */

/* Called version of subroutine end
----------------------------------------------------------------------------*/
MIModEnd(modNum, startTime)

short		modNum;
long 		*startTime;
{

if (MIModStat[modNum].timeFlag) {
    MIModStat[modNum].cTime += util_cpu_time();
}
if (MIModStat[modNum].traceFlag && MIModStat[modNum].numTime<MIModThreshold)
    MIfprintf(MIDBG,"--End   %s\n", MIModStat[modNum].name);
MIModStat[modNum].numTime++;
if (MITraceBackFlag)
    MINumTraceBack--;
} /* ............................. End of procedure MIModEnd */
/* Print header and append revision and date stamp.
----------------------------------------------------------------------------*/
MIPrintRev(header, revStamp)

String		header;			/* Header to be printed with revStamp */
String		revStamp;		/* RCSID or user imput revStamp */
{
String		revAndDate;
String		r;
int		i;
Boolean		unknownFlg = TRUE;

revAndDate = MIBuf;
if (revStamp!=NULL && strcmp(revStamp+6,"Header$"))
    if (strncmp(revStamp+6, "Header:",7)) {
	(void) strcpy(revAndDate, revStamp);
	unknownFlg = FALSE;
    }
    else {
	i = sscanf(revStamp,"%*s %*s %*s %s", revAndDate);
	if (i==1) {
	    r = revAndDate + strlen(revAndDate);
	    *(r++) = ' ';
	    i = sscanf(revStamp,"%*s %*s %*s %*s %s", r);
	    if (i==1) {
		r = revAndDate + strlen(revAndDate);
		*(r++) = ' ';
		i = sscanf(revStamp,"%*s %*s %*s %*s %*s %s", r);
	    }
	}
	unknownFlg = (i!=1);
    }
if (unknownFlg)
    (void) strcpy(revAndDate,"UNKNOWN ??/??/??");

MIfprintf(MIDBG,"%s (rev %s)\n", header, revAndDate);

} /* ............................. End of procedure MIPrintRev */
/* Just like printf except with (void) fflush.
----------------------------------------------------------------------------*/
FILE *
MIfopen(fileName, mode, fileType)

String		fileName;
String		mode;			/* Supports "r", "w" and "a" */
String		fileType;
{
FILE		*f;

f = fopen(fileName, mode);
if (!f && strcmp(fileType,"run-time debug"))
    MIfprintf(MIERR,"ERROR Cannot %s %s file %s because %s\n",
	(*mode=='r' ? "read" :
	(*mode=='w' ? "write" : 
	(*mode=='a' ? "append" : "!$%?"))),
	(fileType ? fileType : "temp"),
	fileName, sys_errlist[errno]);
return(f);
} /* ............................. End of procedure MIfopen */

static char Message[4096];

/* Just like printf except with (void) fflush.
----------------------------------------------------------------------------*/
#ifdef lint
/*ARGSUSED*/
/*VARARGS1*/
#undef va_alist
#undef va_dcl
#define va_alist	fmt
#define va_dcl		char *fmt;
#endif

MIprintf(va_alist)
va_dcl
{
    va_list ap;
    char *format;

    va_start(ap);
    format = va_arg(ap, char *);
    (void) uprintf(Message, format, &ap);
    va_end(ap);
    printf("%s", Message);

    (void) fflush(stdout);
} /* ............................. End of procedure MIprintf */

/* Just like fprintf except with (void) fflush.
----------------------------------------------------------------------------*/
#ifdef lint
/*ARGSUSED*/
/*VARARGS2*/
#undef va_alist
#undef va_dcl
#define va_alist	iop, fmt
#define va_dcl		FILE *iop; char *fmt;
#endif

MIfprintf(va_alist)
va_dcl
{
    va_list ap;
    FILE *iop;
    char *format;

    va_start(ap);
    iop = va_arg(ap, FILE *);
    format = va_arg(ap, char *);
    (void) uprintf(Message, format, &ap);
    va_end(ap);

    if (iop==stderr && MIPrefix) {
	fprintf(stderr,"%s: ", MIPrefix);
    }
    fprintf(iop, "%s", Message);
    (void) fflush(iop);

} /* ............................. End of procedure MIfprintf */
/* Tokenize input string "buf"; fill "tok" with next available token delimited
   by any character in "delimiters" string.  Return character ptr to next
   delimiter after token.
----------------------------------------------------------------------------*/
String
MITokenize(buf, tok, delimiters)

String		buf;
String		tok;
String		delimiters;
{
String		t, d;
Boolean		skipWhiteSpaceFlg = TRUE;

for (d=delimiters; (*d) && *d!=' '; d++);
if (*d) skipWhiteSpaceFlg = FALSE;

if (skipWhiteSpaceFlg)			/* Skip over white space */
    for ( ; *buf==' ' || *buf=='\t'; buf++);

for (t=tok; *buf; buf++) {
    for (d=delimiters; (*d) && *d!=*buf; d++);
    if (*d) {
	if (t==tok) 
	    *(t++) = *(buf++);
	break;
    }
    else *(t++) = *buf;
}

if (t!=tok && skipWhiteSpaceFlg) {	/* Trim white space */
    for (t--; *t==' '; t--);
    t++;
}

*t = '\0';

return(buf);
} /* ............................. End of procedure MITokenize */

/* Interactively check if user wants to continue.
----------------------------------------------------------------------------*/
MIContinue()
{
char		stuff[40];
Boolean		response;

MIprintf("Continue ? ");
(void) scanf("%s", stuff);
response = (*stuff!='n');

if (!response) exit(4);
}
