/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

atlasRpc.c

Main routine for rcp channel definition structure.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	atlasRpc.c,v $
 * Revision 1.3  90/02/07  16:10:47  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.2  89/02/07  11:42:00  octtools
 * fixed some lint errors
 * 
 * Revision 1.1  89/01/27  22:39:41  octtools
 * Initial revision
 * 
 * Revision 1.4  88/04/25  13:08:39  mitch
 * Reset CDSParm.mergeIoFlg=TRUE after channel definition.
 * Made more robust when the channel path does not exist.
 * 
 * Revision 1.3  88/04/05  16:21:37  mitch
 * Fixed bug in initializing CDSMaxNumCh used in printing the sequence file.
 * 
 * Revision 1.2  88/04/05  15:33:34  mitch
 * Used update procedures in atlas so that atlasRpc now works.
 * Added common routine CDSInitAtlasChanId to speed up show channel programs.
 * Added atlasRpcShow3 which sequences through the channels.
 * 
 * Revision 1.1  88/03/31  14:21:20  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "rpc.h"
#include "oct.h"
#include "oh.h"
#include "MIInc.h"
#include "CDSInc.h"

octObject	atlasFacet;
int		atlasCurChan, atlasLastChan;
struct octBox	atlasBBox;
octId		*atlasChanId;
int		atlasNumChan;
int		CDSInitAtlasChanId();
/* Main procedure for channel definition
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcMain(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
int		rc, io;
int		zChanIter;
int		overlapIter;
int		CDSCheckEdgeOverlap();
int		CDSlice();
int		CDSMergeIo();
Boolean		expFlg;
octObject	octChan;

MI_MOD_START("atlasRpcMain");

OH_ASSERT(ohGetById(&atlasFacet, spot->facet));
if (ohGetByBagName(&atlasFacet, &octChan, "CDChanDef")==OCT_OK) {
    VEMMSG("Channels already defined\n");
    goto atlasRpcMainEnd;
}
VEMMSG("Starting channel definition and ordering\n");

for (io=0; io<4; io++)
    CDSIo[io].mergedInstName = NULL;

for (zChanIter=0; zChanIter<=5; zChanIter++) {
    for (overlapIter=0; overlapIter<=1; overlapIter++) {
	CDSOctInput(&atlasFacet);
	CDSInit();
	rc = CDSCheckEdgeOverlap();
	if (rc==0) break;
	MIprintf("Trying to resolve overlaps by expanding all instances\n");
	CDSExpandInst(&atlasFacet, 1.1);
	CDSCleanUp("CDSEdge");
	CDSCleanUp("CDSObj");
    }
    if (rc!=0) {
	VEMMSG("ERROR Could not resolve instance overlaps\n");
	goto atlasRpcMainEnd;
    } 

    CDSBldMaxEdge();
    CDSCleanUp("CDSEdge");
    CDSBldEdgeGraph();
    CDSInitGNObj();
    if (*MIDebugFlag>=0)
	CDSPrintNGraph(CDSGNode);
    rc = CDSlice();
    expFlg = ((rc!=0 && !CDSParm.zChanFlg) && zChanIter<5);

    if (CDSParm.mergeIoFlg) {
	CDSParm.mergeIoFlg = FALSE;
	if (CDSMergeIo(&atlasFacet, 1.2, expFlg)==0) {
	    VEMMSG("Merged ios - starting pass 2 of atlas ...\n");
	    CDSCleanUp("all");
	    continue;
	}
    }

    if (!expFlg) break;
    CDSExpandInst(&atlasFacet, 1.2);	/* Expand instances */
    CDSCleanUp("all");
}

CDSHead.numChan = 0;
CDSChanOrder(CDSTree);
if (*MIDebugFlag>=0)
    CDSPrintTree();
if (CDSParm.zChanFlg)
then CDSInitObjCornNew();
else CDSInitObjCorn();
if (rc==1 && CDSParm.zChanFlg) rc = 0;

CDSOctOutput(&atlasFacet, FALSE);
CDSCleanUp("all");
atlasChanId = NULL;

if (rc==1)
    VEMMSG("WARNING There are \"L\" and/or \"Z\" shaped channels\n");
else if (rc>1)
    VEMMSG("ERROR Something screwed up in atlas\n");

VEMMSG("Channel definition and ordering done!\n");
atlasRpcMainEnd:
CDSParm.mergeIoFlg = TRUE;
MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcMain procedure */
/* Show channel definition stuff
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcShow(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
octObject	octChan;
octObject	octPath;
octObject	octLayer;		/* Stuff to show slices in oct */
int32		numPts;
int		i;

MI_MOD_START("atlasRpcShow");
					/* Get atlas facet & init atlasChanId */
OH_ASSERT(ohGetById(&atlasFacet, spot->facet));
if (CDSInitAtlasChanId(&atlasFacet)!=0)
    goto atlasRpcShowEnd;

VEMMSG("Showing all channels\n");

for (i=0; i<=atlasNumChan; i++) {
    octChan.objectId = atlasChanId[i];
    if (octIdIsNull(octChan.objectId)) continue;

    OH_ASSERT(octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath));
    numPts = 0;
    OH_ASSERT(octGetPoints(&octPath, &numPts, (struct octPoint *)NULL));
    OH_ASSERT(ohGetOrCreateLayer(&atlasFacet, &octLayer,
	numPts==2 ? "MET1" : "POLY"));
    OH_ASSERT(octAttach(&octLayer, &octPath));
}

if (atlasBBox.upperRight.x<atlasBBox.lowerLeft.x)
    OH_ASSERT(octBB(&atlasFacet, &atlasBBox));
(void) vemWnQRegion(&atlasFacet, &atlasBBox);

atlasRpcShowEnd:
MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcShow procedure */
/* Show channel definition stuff (one channel at a time)
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcShow2(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
octObject	octChan;
octObject	octPath;
octObject	octLayer;		/* Stuff to show slices in oct */
octGenerator	genLayer;
RPCArg		*arg;

MI_MOD_START("atlasRpcShow2");

if (lsLength(cmdList)!=1) {		/* Get argument = channel number */
    VEMMSG("error: enter \"channel number\" show-channels-one-by-one\n");
    goto atlasRpcShow2End;
}
lsFirstItem(cmdList, (Pointer *) &arg, (lsHandle *) NULL);
if (arg->argType!=VEM_TEXT_ARG) {
    VEMMSG("error: enter channel number in double quotes\n");
    goto atlasRpcShow2End;
}
(void) sscanf(arg->argData.string, "%d", &atlasCurChan);
atlasCurChan = MIN(MAX(0, atlasCurChan), atlasNumChan);
if (atlasLastChan==atlasCurChan)
    atlasLastChan = -1;

					/* Get atlas facet & init atlasChanId */
OH_ASSERT(ohGetById(&atlasFacet, spot->facet));
if (CDSInitAtlasChanId(&atlasFacet)!=0)
    goto atlasRpcShow2End;

sprintf(MIBuf,"Showing channel %d (total number of channels = %d)\n",
    atlasCurChan, atlasNumChan);
VEMMSG(MIBuf);

octChan.objectId = atlasChanId[atlasCurChan];
if (!octIdIsNull(octChan.objectId)) {	/* Show current channel */
    OH_ASSERT(octGetById(&octChan));
    if (octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath)==OCT_OK) {
	octInitGenContainers(&octPath, OCT_LAYER_MASK, &genLayer);
	while (octGenerate(&genLayer, &octLayer)==OCT_OK)
	    OH_ASSERT(octDetach(&octLayer, &octPath));
	OH_ASSERT(ohGetOrCreateLayer(&atlasFacet, &octLayer, "POLY"));
	OH_ASSERT(octAttach(&octLayer, &octPath));
    }
}

octChan.objectId = (atlasLastChan < 0 ? oct_null_id : atlasChanId[atlasLastChan]);
if (!octIdIsNull(octChan.objectId)) {	/* Unhilite previous channel */
    OH_ASSERT(octGetById(&octChan));
    if (octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath)==OCT_OK) {
	octInitGenContainers(&octPath, OCT_LAYER_MASK, &genLayer);
	while (octGenerate(&genLayer, &octLayer)==OCT_OK)
	    OH_ASSERT(octDetach(&octLayer, &octPath));
	OH_ASSERT(ohGetOrCreateLayer(&atlasFacet, &octLayer, "MET1"));
	OH_ASSERT(octAttach(&octLayer, &octPath));
    }
}

atlasLastChan = atlasCurChan;

if (atlasBBox.upperRight.x<atlasBBox.lowerLeft.x)
    OH_ASSERT(octBB(&atlasFacet, &atlasBBox));
(void) vemWnQRegion(&atlasFacet, &atlasBBox);

atlasRpcShow2End:
MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcShow2 procedure */
/* Show channel definition stuff (next channel)
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcShow3(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
octObject	octChan;
octObject	octPath;
octObject	octLayer;		/* Stuff to show slices in oct */
octGenerator	genLayer;
struct octBox	pathBBox;

MI_MOD_START("atlasRpcShow3");

atlasCurChan++;				/* Get next channel number */
atlasCurChan = MIN(MAX(0, atlasCurChan), atlasNumChan);
if (atlasLastChan==atlasCurChan)
    atlasLastChan = -1;

					/* Get atlas facet & init atlasChanId */
OH_ASSERT(ohGetById(&atlasFacet, spot->facet));
if (CDSInitAtlasChanId(&atlasFacet)!=0)
    goto atlasRpcShow3End;

sprintf(MIBuf,"Showing channel %d (total number of channels = %d)\n",
    atlasCurChan, atlasNumChan);
VEMMSG(MIBuf);

octChan.objectId = atlasChanId[atlasCurChan];
if (!octIdIsNull(octChan.objectId)) {	/* Show current channel */
    OH_ASSERT(octGetById(&octChan));
    if (octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath)==OCT_OK) {
	octInitGenContainers(&octPath, OCT_LAYER_MASK, &genLayer);
	while (octGenerate(&genLayer, &octLayer)==OCT_OK)
	    OH_ASSERT(octDetach(&octLayer, &octPath));
	OH_ASSERT(ohGetOrCreateLayer(&atlasFacet, &octLayer, "POLY"));
	OH_ASSERT(octAttach(&octLayer, &octPath));
	OH_ASSERT(octBB(&octPath, &pathBBox));
	(void) vemWnQRegion(&atlasFacet, &pathBBox);
    }
}

octChan.objectId = (atlasLastChan<0 ? oct_null_id : atlasChanId[atlasLastChan]);
if (!octIdIsNull(octChan.objectId)) {	/* Unhilite previous channel */
    OH_ASSERT(octGetById(&octChan));
    if (octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath)==OCT_OK) {
	octInitGenContainers(&octPath, OCT_LAYER_MASK, &genLayer);
	while (octGenerate(&genLayer, &octLayer)==OCT_OK)
	    OH_ASSERT(octDetach(&octLayer, &octPath));
	OH_ASSERT(ohGetOrCreateLayer(&atlasFacet, &octLayer, "MET1"));
	OH_ASSERT(octAttach(&octLayer, &octPath));
	OH_ASSERT(octBB(&octPath, &pathBBox));
	(void) vemWnQRegion(&atlasFacet, &pathBBox);
    }
}

atlasLastChan = atlasCurChan;

atlasRpcShow3End:
MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcShow3 procedure */
/* Clear shown channel definition stuff
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcShowC(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
int		i;
octObject	octChan;
octObject	octPath;
octObject	octLayer;		/* Stuff to show slices in oct */
octGenerator	genLayer;

MI_MOD_START("atlasRpcShowC");

					/* Get atlas facet & init atlasChanId */
OH_ASSERT(ohGetById(&atlasFacet, spot->facet));
if (CDSInitAtlasChanId(&atlasFacet)!=0)
    goto atlasRpcShowCEnd;

VEMMSG("Removing channel marks\n");

for (i = 0; i <= atlasNumChan; i++) {
    octChan.objectId = atlasChanId[i];
    if (octIdIsNull(octChan.objectId)) continue;

    OH_ASSERT(octGetById(&octChan));
    if (octGenFirstContent(&octChan, OCT_PATH_MASK, &octPath)==OCT_OK) {
	octInitGenContainers(&octPath, OCT_LAYER_MASK, &genLayer);
	while (octGenerate(&genLayer, &octLayer)==OCT_OK)
	    OH_ASSERT(octDetach(&octLayer, &octPath));
    }
}
atlasLastChan = -1;
atlasCurChan  =  0;

if (atlasBBox.upperRight.x<atlasBBox.lowerLeft.x)
    OH_ASSERT(octBB(&atlasFacet, &atlasBBox));
(void) vemWnQRegion(&atlasFacet, &atlasBBox);

atlasRpcShowCEnd:
MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcShowC procedure */
/* Clear channel definition stuff
----------------------------------------------------------------------------*/
/*ARGSUSED*/
atlasRpcClear(spot, cmdList, uow)

RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
int		CDSOctPurge();

MI_MOD_START("atlasRpcClear");

OH_ASSERT(ohGetById(&atlasFacet, spot->facet));

if (CDSOctPurge(&atlasFacet)==4) {
    VEMMSG("No channels defined yet\n");
}
else {
    if (atlasChanId)
	FREE_ARRAY(atlasNumChan+1, atlasChanId, octId);
    VEMMSG("Channel definition purged!\n");
}

MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcClear procedure */
/* Init atlasChanId array
----------------------------------------------------------------------------*/
int
CDSInitAtlasChanId(atlasFacet)

octObject	*atlasFacet;
{
octObject	octChan;
octObject	octSubChan;
octObject	octECProp, octChOrd;
octGenerator	genSubChan;
int		i;
int		stkRef;
Boolean		leafCellFlg;
int		rc = 0;

MI_MOD_START("CDSInitAtlasChanId");

if (ohGetByBagName(atlasFacet, &octChan, "CDChanDef")!=OCT_OK) {
    VEMMSG("No channels defined yet\n");
    rc = 1;
    goto CDSInitAtlasChanIdEnd;
}

if (atlasChanId) goto CDSInitAtlasChanIdEnd;

OH_ASSERT(ohGetByPropName(&octChan, &octChOrd, "CDChanOrder"));
atlasNumChan = octChOrd.contents.prop.value.integer;

atlasChanId = ALLOC_ARRAY(atlasNumChan+1, octId);
for (i=0; i<=atlasNumChan; i++)
    atlasChanId[i] = oct_null_id;

stkRef = MIStackPtr;
MI_STACK_PUSH(octChan.objectId);

while (MIStackPtr!=stkRef) {		/* Traverse channel tree */
    MI_STACK_POP(octId, octChan.objectId);

    octInitGenContents(&octChan, OCT_BAG_MASK, &genSubChan);
    for (leafCellFlg=TRUE; octGenerate(&genSubChan, &octSubChan)==OCT_OK; ) {
	if (strcmp(octSubChan.contents.bag.name, "CDChanDef")) continue;
	MI_STACK_PUSH(octSubChan.objectId);
	leafCellFlg = FALSE;
    }
    if (leafCellFlg) continue;
    if (ohGetByPropName(&octChan, &octECProp, "CDOuterChannel")==OCT_OK)
	continue;

    OH_ASSERT(ohGetByPropName(&octChan, &octChOrd, "CDChanOrder"));
    i = octChOrd.contents.prop.value.integer;
    DRANGE(0, i, atlasNumChan+1);
    atlasChanId[i] = octChan.objectId;
}

CDSInitAtlasChanIdEnd:
MI_MOD_END();
return(rc);
} /* ............................. End of CDSInitAtlasChanId procedure */
