/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSBldNode.c

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSBldNode.c,v $
 * Revision 1.1  90/02/07  16:12:06  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:41  octtools
 * Initial revision
 * 
 * Revision 1.1  88/03/31  11:37:34  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"
#include "CDSInc.h"

/* Build edge graph given maximal tile graph (trims stubs)
----------------------------------------------------------------------------*/
CDSBldEdgeGraph()
{
CDSGEdgeDef	*ge, *ge2;
CDSGNodeDef	*n, *lNode, *pNode = NULL;
CDSGNodeDef	*sNode;

short		k;
char		d, d2;
int		l, l2;

MI_MOD_START("CDSBldEdgeGraph");

DASSERT(CDS_V==CDS_X && CDS_H==CDS_Y);
CDSHead.numGNode = 0;
for (d=CDS_V; ; d=CDS_H) {
    MSG(1, ("Generating node graph for %s\n", CDS_DIR_TOSTR(d)));
    d2 = !d;
    for (ge=CDSGEdge[d]; ge; ge=ge->next) {
	MSG(3, ("On edge l=%d (%d %d)\n",
	    ge->loc, ge->range[CDS_LO], ge->range[CDS_HI]));
	l = ge->loc;
	lNode = NULL;
	sNode = CDSGNode;
	for (ge2=CDSGEdge[d2]; ge2; ge2=ge2->next) {
	    l2 = ge2->loc;
	    if (ge->range[CDS_LO]>l2) continue;
	    if (ge->range[CDS_HI]<l2) continue;
					/* This statement assumed gEdges were
	    if (ge->range[CDS_HI]<l2) break; ordered but they are not */

	    if (ge2->range[CDS_LO]>l ||
		ge2->range[CDS_HI]<l) continue;
					/* Intersection found */
	    if (d==CDS_V) {
		n = ALLOC_TYPE(CDSGNodeDef);
		if (!pNode)
		then CDSGNode    = n;
		else pNode->next = n;
		n->id   = CDSHead.numGNode++;
		n->type = '*';
		n->loc[CDS_X] = l;
		n->loc[CDS_Y] = l2;
		for (k=0; k<4; k++) {
		    n->bch[k].node   = NULL;
		    n->bch[k].type   = '*';
		    n->bch[k].treeId = CDS_NULL_TREEID;
		    n->bch[k].objNum = CDS_NBOBJ_NOT_SET;
		}
		n->bch[CDS_S].node = lNode;
		if (lNode)
		    lNode->bch[CDS_N].node = n;
		pNode = lNode = n;
		MSG(4, ("Found X at (%d %d)\n", l, l2));
	    }
	    else {
				    /* Find node */
		for (n=sNode; n; n=n->next)
		    if (n->loc[CDS_X]==l2 &&
			n->loc[CDS_Y]==l) break;
		MSG(4, ("Found node %d at (%d %d)\n", CDS_ID(n), l2, l));
		DASSERT(n);
		n->bch[CDS_W].node = lNode;
		if (lNode)
		    lNode->bch[CDS_E].node = n;
		lNode = n;
		sNode = n;
	    }
	}
    }
    if (d==CDS_H) break;
    pNode->next = NULL;			/* Terminate node list */
}

CDSInitExtEdgeGraph();

MI_MOD_END();
} /* ............................. End of procedure CDSBldEdgeGraph */
/* Initialize the external edges
----------------------------------------------------------------------------*/
CDSInitExtEdgeGraph()
{
CDSGNodeDef	*g, *gn;

short		k, rpd;
short		t;

MI_MOD_START("CDSInitExtEdgeGraph");

gn  = NULL;
rpd = CDS_W;
for (g=CDSGNode; gn!=CDSGNode; g=gn) {	/* Clockwise around exterior */
    for (k=CDS_NEXTDIR(rpd,CDS_CW);	/* Get next exterior node */
	 !g->bch[k].node; k=CDS_NEXTDIR(k,CDS_CW));
    gn = g->bch[k].node;
    MSG(3, ("Exterior node=%d (%d %d) nextdir=%s\n",
	CDS_ID(g), g->loc[CDS_X], g->loc[CDS_Y], CDS_CDIR_TOSTR(k)));

    rpd = CDS_REVDIR(k);

    switch (k) {
    case CDS_N: t = CDS_W_EXT_TREEID; break;
    case CDS_E: t = CDS_N_EXT_TREEID; break;
    case CDS_S: t = CDS_E_EXT_TREEID; break;
    default   : t = CDS_S_EXT_TREEID; break;
    }

    g ->type = 'e';
    g ->bch[k  ].type   = 'e';
    g ->bch[k  ].treeId = t;
    gn->bch[rpd].type   = 'e';
    gn->bch[rpd].treeId = t;
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitExtEdgeGraph */
