/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSInit.c

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSInit.c,v $
 * Revision 1.1  90/02/07  16:12:11  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:44  octtools
 * Initial revision
 * 
 * Revision 1.2  88/04/05  12:24:35  mitch
 * Changed CDSCheckEdgeOverlap to return nonzero if overlaps are found
 * instead of exit(2)
 * 
 * Revision 1.1  88/03/31  11:38:23  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"
#include "CDSInc.h"
#include "oh.h"

/* Initialize Edge list and first channel tree node from object list.
----------------------------------------------------------------------------*/
CDSInit()
{
short		o;
short		b;
int		e;
char		d, d2, hl;
int		*loc, *pLoc;
CDSEdgeDef	*edge;

int		minXY, xy;
short		minXYBdyIdx;

static Boolean	firstTimeFlg = FALSE;

MI_MOD_START("CDSInit");

CDSEmptyRoomDimX = 320;
CDSEmptyRoomDimY = 320;

CDSInitExtObj();			/* Initialize external object */

MSG(1, ("NumObj = %d\n", CDSHead.numObj));
CDSHead.numEdge = 0;
for (o=0; o<CDSHead.numObj; o++) {
    CDSObj[o].numCorn = 0;
    CDSObj[o].corn    = NULL;
    DRANGE(4,CDSObj[o].numBdy,500);
    CDSHead.numEdge += CDSObj[o].numBdy;
    MSG(1, ("Obj %d: numBdy=%d\n", o, CDSObj[o].numBdy));
    for (b=0; b<CDSObj[o].numBdy; b++)
	MSG(1, ("Bdy %d: (%d %d)\n",
	    b, CDSObj[o].bdy[b][CDS_X], CDSObj[o].bdy[b][CDS_Y]));
}
CDSEdgeArray = ALLOC_ARRAY(CDSHead.numEdge, CDSEdgeDef);
MSG(1, ("Number of edges=%d\n", CDSHead.numEdge));

e = 0;
DASSERT(CDS_V==CDS_X && CDS_H==CDS_Y);
for (o=0; o<CDSHead.numObj; o++) {
    pLoc  = CDSObj[o].bdy[CDSObj[o].numBdy-1];
    minXY = MAXINT;
    for (b=0; b<CDSObj[o].numBdy; b++, e++) {
	loc = CDSObj[o].bdy[b];
	DASSERT(loc[CDS_X]!=pLoc[CDS_X] || loc[CDS_Y]!=pLoc[CDS_Y]);
	if (loc[CDS_X]==pLoc[CDS_X])
	then d = CDS_V;
	else d = CDS_H;
	d2 = !d;
	hl = (loc[d2]<pLoc[d2] ? CDS_LO : CDS_HI);

	edge = &CDSEdgeArray[e];
	edge->id         = e;
	edge->objNum     = o;
	edge->bdyIdx     = b;
	edge->dir        = d;
	edge->hiLo       = (d==CDS_V ? hl : !hl);
	edge->loc        = loc [d];
	edge->doneFlg    = FALSE;
	edge->queueFlg	 = FALSE;
	edge->range[!hl] = pLoc[d2];
	edge->range[ hl] = loc [d2];
	DASSERT(edge->range[CDS_LO]<edge->range[CDS_HI]);

	pLoc = loc;
	MSG(3, ("Edge %d: %s %s o=%d l=%d (%d %d)\n",
	    e, CDS_DIR_TOSTR(edge->dir), CDS_HILO_TOSTR(edge->hiLo),
	    edge->objNum, edge->loc, edge->range[CDS_LO], edge->range[CDS_HI]));

	xy = loc[CDS_X]+loc[CDS_Y];	/* Init CDSObj[].minLoc */
	if (xy<minXY) {
	    minXY       = xy;
	    minXYBdyIdx = b;
	}
    }

    CDSObj[o].minLocBdyIdx = minXYBdyIdx;
}

CDSEdge[CDS_V] = NULL;			/* Link up vert & hor edges */
CDSEdge[CDS_H] = NULL;
for (e=0; e<CDSHead.numEdge; e++) {
    d = CDSEdgeArray[e].dir;
    CDSEdgeArray[e].next = CDSEdge[d];
    CDSEdgeArray[e].prev = NULL;
    CDSEdge[d] = &CDSEdgeArray[e];
}

CDSOrderEdge(&CDSEdge[CDS_V]);
CDSOrderEdge(&CDSEdge[CDS_H]);

if (firstTimeFlg) {
    firstTimeFlg = FALSE;
    for (d=0; d<4; d++)
	CDSIoFlag[d] = FALSE;
}

MI_MOD_END();
} /* ............................. End of Procedure CDSInit */
/* Check for overlapping objects.  Go through each edge and check if it
intersects any other edge of another object.
----------------------------------------------------------------------------*/
int
CDSCheckEdgeOverlap()
{
    CDSEdgeDef	*e, *e2;
    short		d, d2;
    short		o, o2;
    short		b;
    int		l, l2;
    CDSLocDef	range;
    int		numIntersect = 0;

    int		k, k2;
    Boolean		*objStatus;
    int		rc = 0;

    MI_MOD_START("CDSCheckEdgeOverlap");

    objStatus = ALLOC_ARRAY(CDSHead.numObj*CDSHead.numObj, Boolean);
    for (k=0; k<CDSHead.numObj*CDSHead.numObj; k++)
      objStatus[k] = 0;

    for (d=CDS_V; ; d=CDS_H) {
	d2 = !d;
	for (e=CDSEdge[d]; e; e=e->next) {
	    l = e->loc;
	    o = e->objNum;
	    CDS_COPY_LOC(range, e->range);

	    for (e2=CDSEdge[d2]; e2; e2=e2->next) {
		l2 = e2->loc;
		o2 = e2->objNum;
		if (l2<range[CDS_LO] || o2==o) continue;
		if (l2>range[CDS_HI]) break;

		if (e2->range[CDS_LO]>l) {
		    if (l2==range[CDS_LO] || l2==range[CDS_HI]) continue;
		    if (e->hiLo==CDS_HI || d==CDS_H) continue;
		    k = o*CDSHead.numObj+o2;
		    if (objStatus[k]==0)
		      objStatus[k] = 1;
		    else if (objStatus[k]==2) {
			k2 = o2*CDSHead.numObj+o;
			objStatus[k] = objStatus[k2] = 3;
			numIntersect++;
		    }
		    continue;
		}

		if (e2->range[CDS_HI]<l) {
		    if (l2==range[CDS_LO] || l2==range[CDS_HI]) continue;
		    if (e->hiLo==CDS_LO || d==CDS_H) continue;
		    k = o*CDSHead.numObj+o2;
		    if (objStatus[k]==0)
		      objStatus[k] = 2;
		    else if (objStatus[k]==1) {
			k2 = o2*CDSHead.numObj+o;
			objStatus[k] = objStatus[k2] = 3;
			numIntersect++;
		    }
		    continue;
		}

		if (e2->hiLo==CDS_LO && l2==    range[CDS_HI] ||
		    e2->hiLo==CDS_HI && l2==    range[CDS_LO] ||
		    e ->hiLo==CDS_LO && l ==e2->range[CDS_HI] ||
		    e ->hiLo==CDS_HI && l ==e2->range[CDS_LO]) continue;

		k  =  o*CDSHead.numObj+o2;
		k2 = o2*CDSHead.numObj+o;
		if (objStatus[k]<3 && objStatus[k2]<3) {
		    objStatus[k ] = 4;
		    objStatus[k2] = 4;
		    numIntersect++;
		}

		MSG(6, ("ERROR edge%d (%d %d) (%d %d) of obj%d intersects\n",
			CDS_ID(e),
			(d==CDS_V ? l : e->range[CDS_LO]),
			(d==CDS_H ? l : e->range[CDS_LO]),
			(d==CDS_V ? l : e->range[CDS_HI]),
			(d==CDS_H ? l : e->range[CDS_HI]), o));
		MSG(6, ("      edge%d (%d %d) (%d %d) of obj%d\n",
			CDS_ID(e2),
			(d==CDS_H ? l2 : e2->range[CDS_LO]),
			(d==CDS_V ? l2 : e2->range[CDS_LO]),
			(d==CDS_H ? l2 : e2->range[CDS_HI]),
			(d==CDS_V ? l2 : e2->range[CDS_HI]), o2));
	    }
	}
	if (d==CDS_H) break;
    }

    if (numIntersect) {
	MIerror("%d overlaps found\n", numIntersect);
	for (o=0; o<CDSHead.numObj; o++) {
	    k = o*CDSHead.numObj+o+1;
	    for (o2=o+1; o2<CDSHead.numObj; o2++, k++) {
		if (CDSOctMode) {
		    if (objStatus[k]==4 || objStatus[k]==3) {
			octObject inst1, inst2;
			char *n1,*n2;
			n1 = n2 = "(unknown)";
			if ( ohGetById( &inst1, CDSObj[o].dbRefId ) == OCT_OK) {
			    n1 = ohFormatName( &inst1 );
			}
			if ( ohGetById( &inst2, CDSObj[o2].dbRefId ) == OCT_OK)  {
			    n2 = ohFormatName( &inst2 );
			}
			fprintf( stderr, "ERROR %s overlaps %s\n", n1, n2 );
		    }
		} else {
		    switch (objStatus[k]) {
		    case 4 : MIerror("Obj %d overlaps obj %d\n", o, o2);
		    case 0 : break;
		    case 3 : MIerror("Obj %d inside obj %d\n", o2, o);
		    default: break;
		    }
		}
		if (objStatus[k]==4 || objStatus[k]==3) {
		    MSG(0, ("Bdy of obj %d:\n", o));
		    for (b=0; b<CDSObj[o].numBdy; b++)
		      MSG(0, ("Bdy %d: (%d %d)\n",
			      b, CDSObj[o].bdy[b][CDS_X], CDSObj[o].bdy[b][CDS_Y]));
		}
		if (objStatus[k]==4 || objStatus[k]==3) {
		    MSG(0, ("Bdy of obj %d:\n", o2))
		      for (b=0; b<CDSObj[o2].numBdy; b++)
			MSG(0, ("Bdy %d: (%d %d)\n",
				b, CDSObj[o2].bdy[b][CDS_X], CDSObj[o2].bdy[b][CDS_Y]));
		}
	    }
	}
	rc = 4;
	fprintf(stderr,  "Atlas suggests that you change your placement\n");

    }


    FREE_ARRAY(CDSHead.numObj*CDSHead.numObj, objStatus, Boolean);

    MI_MOD_END();
    return(rc);
}				/* ............................. End of procedure CDSCheckEdgeOverlap */
/* Order edges for I don't know what
----------------------------------------------------------------------------*/
CDSOrderEdge(startEdge)

CDSEdgeDef	**startEdge;		/* Edge list to be ordered */
{
char		d;			/* Direction of edges */
int		numEdge = 0;
int		e, ee;
int		*edgeOrder;
CDSEdgeDef	*edge, *pEdge, *sEdge;
int		CDSQsortOrderEdge();

MI_MOD_START("CDSOrderEdge");

sEdge = *startEdge;
d = sEdge->dir;
MSG(1, ("Sorting %s edges\n", CDS_DIR_TOSTR(d)));
for (edge=sEdge; edge; edge=edge->next) {
    DRANGE(0,edge->id,CDSHead.numEdge);
    DASSERT(edge->dir==d);
    numEdge++;
    DURANGE(numEdge,CDSHead.numEdge);
}

edgeOrder = ALLOC_ARRAY(numEdge, int);

for (edge=sEdge, e=0; edge; edge=edge->next, e++)
    edgeOrder[e] = edge->id;

qsort((char *)edgeOrder, numEdge, sizeof(int), CDSQsortOrderEdge);

pEdge = NULL;
for (ee=0; ee<numEdge; ee++) {
    e = edgeOrder[ee];
    DRANGE(0,e,CDSHead.numEdge);
    edge = &CDSEdgeArray[e];
    MSG(3, ("Edge %d: %s o=%d l=%d (%d %d)\n",
	e, CDS_HILO_TOSTR(edge->hiLo),
	edge->objNum, edge->loc, edge->range[CDS_LO], edge->range[CDS_HI]));
    if (!pEdge) {
	*startEdge = edge;
	edge->prev = NULL;
    }
    else {
	DASSERT(pEdge->loc<=edge->loc);
	pEdge->next = edge;
	edge ->prev = pEdge;
    }
    pEdge = edge;
}
edge->next = NULL;

FREE_ARRAY(numEdge, edgeOrder, int);

MI_MOD_END();
} /* ............................. End of procedure CDSOrderEdge */
/* Edge ordering for qsort (see CDSOrderEdge)
----------------------------------------------------------------------------*/
int
CDSQsortOrderEdge(e1, e2)

int		*e1, *e2;
{
int		d;

d = CDSEdgeArray[*e1].loc -CDSEdgeArray[*e2].loc;
if (d) return(d);
d = CDSEdgeArray[*e2].hiLo-CDSEdgeArray[*e1].hiLo;
if (d) return(d);
if (CDSEdgeArray[*e1].range[CDS_LO]>
    CDSEdgeArray[*e2].range[CDS_HI]) return( 1);
if (CDSEdgeArray[*e1].range[CDS_HI]<
    CDSEdgeArray[*e2].range[CDS_LO]) return(-1);
return(-1);
} /* ............................. End of procedure CDSQsortOrderEdge */

/* Initialize external object boundary.
----------------------------------------------------------------------------*/
CDSInitExtObj()
{
    short		o, b;
    char		d;
    CDSLocDef	minLoc, maxLoc;

    MI_MOD_START("CDSInitExtObj");

    o = CDSHead.numObj-1;	/* Initialize last obj */
    CDSObj[o].dbRef  = (RefAny)NULL;
    CDSObj[o].dbRefId  = oct_null_id;
    
    CDSObj[o].numBdy = 4;
    CDSObj[o].bdy = ALLOC_ARRAY(CDSObj[o].numBdy, CDSLocDef);

    for (d=CDS_X; ; d=CDS_Y) {	/* Find max bounds of all objs */
	minLoc[d] =  MAXINT;
	maxLoc[d] = -MAXINT;
	for (o=0; o<CDSHead.numObj-1; o++)
	  for (b=0; b<CDSObj[o].numBdy; b++) {
	      minLoc[d] = MIN(minLoc[d], CDSObj[o].bdy[b][d]);
	      maxLoc[d] = MAX(maxLoc[d], CDSObj[o].bdy[b][d]);
	  }
	minLoc[d] -= 10;
	maxLoc[d] += 10;
	/* This boundary is CLOCKWISE ! */
	CDSObj[o].bdy[0][d] = minLoc[d];
	CDSObj[o].bdy[2][d] = maxLoc[d];
	CDSObj[o].bdy[1][d] = (d==CDS_X ? minLoc[d] : maxLoc[d]);
	CDSObj[o].bdy[3][d] = (d==CDS_Y ? minLoc[d] : maxLoc[d]);

	if (d==CDS_Y) break;
    }

    CDSExtBox.ll.x = CDSObj[o].bdy[0][CDS_X];
    CDSExtBox.ll.y = CDSObj[o].bdy[0][CDS_Y];
    CDSExtBox.ur.x = CDSObj[o].bdy[2][CDS_X];
    CDSExtBox.ur.y = CDSObj[o].bdy[2][CDS_Y];

    MI_MOD_END();
}				/* ............................. End of procedure CDSInitExtObj */
