/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSMainFromFacet.c

CDSMainFromFacet	Main routine for channel definition slicing structure.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSMainFromFacet.c,v $
 * Revision 1.1  90/02/07  16:12:14  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:45  octtools
 * Initial revision
 * 
 * Revision 1.7  88/05/07  19:47:51  mitch
 * Fixed bug with disappearing real rooms
 * 
 * Revision 1.6  88/04/15  16:10:00  mitch
 * Fix updating of oct facet to check if there is already a channel
 * definition tree.  Also check for old corner bags.
 * Fix CDSOctPurge was leaving one cornerEdge for each instance.
 * 
 * Revision 1.5  88/04/15  14:08:38  mitch
 * Put in channel path for used external channels.
 * 
 * Revision 1.4  88/04/08  11:29:41  mitch
 * Put in octBegin and octEnd into main procedure.
 * 
 * Revision 1.3  88/04/05  16:16:24  mitch
 * Fixed bug in initializing CDSMaxNumCh used in printing the sequence file.
 * 
 * Revision 1.2  88/04/05  12:26:10  mitch
 * Made all operations off of one oct facet instead of several like cdmerge
 * and cdexpandxxx views.  This was done so that atlasRpc would work.
 * Added code to try to resolve instance overlaps by expansion.
 * Also z-shaped channel handling was made more robust.
 * 
 * Revision 1.1  88/03/31  13:53:49  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#ifndef lint
static char RCSId[] = "@(#) $Header: /octtools/release-3.0/src/atlast/RCS/CDSMainFromFacet.c,v 1.1 90/02/07 16:12:14 octtools Exp Locker: octtools $";
#endif
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "options.h"
#include "miutil.h"
#include "CDSInc.h"

/* Main procedure for channel definition
Return code = 0  means no z-shaped channels
	    > 0  means z-shaped channels
	    = 99 means overlapping instances
----------------------------------------------------------------------------*/
int
CDSMainFromFacet(mainFacet, convertPgm)

octObject	*mainFacet;
void		(*convertPgm)();
{
int		rc, io, cornRc;
int		zChanIter;
int		overlapIter;
int		CDSCheckEdgeOverlap();
int		CDSlice();
int		CDSMergeIo();
Boolean		expFlg;
octObject	octProp;

CDSOctMode = TRUE;
CDSEmptyRoomFlg = FALSE;

for (io=0; io<4; io++)
    CDSIo[io].mergedInstName = NULL;

CDSDoItAgain:
for (zChanIter=0; zChanIter<=5; zChanIter++) {
    for (overlapIter=0; overlapIter<=1; overlapIter++) {
	if (CDSOctMode)
	    CDSOctInput(mainFacet);
	else {
	    DASSERT(zChanIter==0 && overlapIter==0);
	    CDSPreInit(CDSParm.cellName);
	}
	CDSInit();
	rc = CDSCheckEdgeOverlap();
	if (rc==0) break;
	MSG(-1, ("Trying to resolve overlaps by expanding all instances\n"));
	CDSExpandInst(mainFacet, 1.1);
	CDSCleanUp("CDSEdge");
	CDSCleanUp("CDSObj");
    }
    if (rc!=0) {
	MIerror("Could not resolve instance overlaps\n");
	rc = 99;
	goto CDSMainFromFacetEnd;
    }

    CDSBldMaxEdge();
    CDSCleanUp("CDSEdge");
    CDSBldEdgeGraph();
    CDSInitGNObj();
    if (*debugFlag>=0)
	CDSPrintNGraph(CDSGNode);
    rc = CDSlice();
    expFlg = ((rc!=0 && !CDSParm.zChanFlg) && zChanIter<5);
    if (expFlg) {
	MSG(-1, ("Chip has z-shaped channels, expanding instances by %4.2fx\n",
	    CDSParm.expandFactor));
    }
    else if (rc!=0 && (zChanIter>0 && zChanIter<5)) {
	MSG(-1, ("Z-shaped channels of chip have been eliminated by expansion\n"));
    }

    if (CDSParm.mergeIoFlg || CDSParm.mergeIoOnlyFlg) {
	CDSParm.mergeIoFlg = FALSE;
	rc = CDSMergeIo(mainFacet, CDSParm.expandFactor,
		expFlg && !CDSParm.mergeIoOnlyFlg);
	if (rc==0) {
	    CDSCleanUp("all");
	    if (CDSParm.mergeIoOnlyFlg) {
		MSG(-1, ("Merged ios ...\n"));
		goto CDSMainFromFacetEnd;
	    }
	    MSG(-1, ("Merged ios - starting pass 2 of atlas ...\n"));
	    continue;
	}
	else if (rc==16) {
	    MIerror("cannot merge ios in rpc mode\n");
	    goto CDSMainFromFacetEnd;
	}
    }

    if (!expFlg) break;
    					/* Expand instances */
    CDSExpandInst(mainFacet, CDSParm.expandFactor);
    CDSCleanUp("all");
}
if (!strcmp(CDSParm.mode,"hgInter")) {
    CDSCleanUp("all");			/* Do everything one more time */
    if (CDSOctMode)
	CDSOctInput(mainFacet);
    printf("Add empty rooms - starting pass 3 of atlas ...\n");
    CDSInit();
    CDSBldMaxEdge();
    CDSCleanUp("CDSEdge");
    CDSBldEdgeGraph();
    CDSEmptyRoomFlg = TRUE;		/* With empty rooms */
    CDSInitGNObj();
    if (*debugFlag>=0)
	CDSPrintNGraph(CDSGNode);
    rc = CDSlice();
}

CDSHead.numChan = 0;
CDSChanOrder(CDSTree);
if (*debugFlag>=0)
    CDSPrintTree();
if (CDSParm.zChanFlg) {
/* then CDSInitObjCornNew2(); 		Not debugged yet */
    cornRc = CDSInitObjCornNew();
    if (cornRc!=0) {
	CDSExpandInst(mainFacet, CDSParm.expandFactor);
	CDSCleanUp("all");
	goto CDSDoItAgain;
    }
}
else
    CDSInitObjCorn();

if (rc==1 && CDSParm.zChanFlg) rc = 0;

if (CDSEmptyRoomFlg)
    CDSOctAddEmptyRooms(mainFacet);

if (!strcmp(CDSParm.mode,"hgInter"))
    convertPgm(mainFacet);
else if (CDSOctMode)
    CDSOctOutput(mainFacet, CDSParm.showChanFlg);

OH_ASSERT(ohGetOrCreatePropInt(mainFacet, &octProp, "atlasNumberOfChannels", CDSMaxNumCh+1));

if (!strcmp(CDSParm.mode,"hgInter"))
    CDSCleanUp("CDSHg");
else
    CDSCleanUp("all");

CDSMainFromFacetEnd:
return(rc);
} /* ............................. End of CDSMainFromFacet */
