/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSObjRing.c

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSObjRing.c,v $
 * Revision 1.1  90/02/07  16:12:21  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:50  octtools
 * Initial revision
 * 
 * Revision 1.1  88/03/31  11:38:57  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"
#include "CDSInc.h"
#define WITHOUT_EMPTY_ROOMS

#ifdef WITHOUT_EMPTY_ROOMS
/* Initialize gNodes
----------------------------------------------------------------------------*/
CDSInitGNObj()
{
int		numTNode = 0;
CDSGNodeDef	**tNode;
CDSGNodeDef	*gn, *ngn;
int		*loc;
int		CDSQsortTNode();
short		nullObjNum = CDS_NBOBJ_EXTERNAL-1;
short		o;
int		n;

MI_MOD_START("CDSInitGNObj");
					/* Init tNode[] */
tNode = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);
for (gn=CDSGNode; gn; gn=gn->next) {	/* Fill with nodes with North bch */
    if (!gn->bch[CDS_N].node) continue;
    tNode[numTNode++] = gn;
}
					/* Sort lowest x then y first */
qsort((char *)tNode, numTNode, sizeof(CDSGNodeDef*), CDSQsortTNode);

for (o=0; o<CDSHead.numObj-1; o++) {	/* Set obj ring for each obj */
					/* Get lo starting Norch bch */
    loc = CDSObj[o].bdy[CDSObj[o].minLocBdyIdx];

    for (n=numTNode-1; n>=0; n--) {
	gn = tNode[n];
	if (gn->loc[CDS_X]<=loc[CDS_X] &&
	    gn->loc[CDS_Y]<=loc[CDS_Y] &&
	    gn->bch[CDS_N].node->loc[CDS_Y]>loc[CDS_Y]) break;
    }
    DASSERT(n>=0);
    ngn = gn->bch[CDS_N].node;

    CDSInitNObjRing(gn, ngn, o);	/* Init ring bchs with obj num */
}

gn = CDSGNode->bch[CDS_N].node;		/* Init external obj ring */
CDSInitNObjRing(gn, CDSGNode, (short)CDS_NBOBJ_EXTERNAL);

for (gn=CDSGNode; gn; gn=gn->next) {	/* Init null obj ring bchs */
    if (!gn->bch[CDS_N].node) continue;
    if (gn->bch[CDS_N].objNum==CDS_NBOBJ_NOT_SET)
	CDSInitNObjRing(gn, gn->bch[CDS_N].node, nullObjNum--);
}

MSG(1, ("%d empty rooms\n", -nullObjNum+CDS_NBOBJ_EXTERNAL-1));

FREE_ARRAY(CDSHead.numGNode, tNode, CDSGNodeDef*);

MI_MOD_END();
} /* ............................. End of procedure CDSInitGNObj */
#else /* WITHOUT_EMPTY_ROOMS */
/* Initialize gNodes
----------------------------------------------------------------------------*/
CDSInitGNObj()
{
    int		numTNode = 0;
    CDSGNodeDef	**tNode;
    CDSGNodeDef	*gn, *ngn;
    CDSGNodeDef	*g, *gStart, *gll;
    short	rpd, k;
    int		*loc;
    int		CDSQsortTNode();
    short	nullObjNum = CDS_NBOBJ_EXTERNAL-1;
    short	o;
    int		n;
    int		numEmptyObj = 0;
    CDSObjDef	*obj, *oldObjList;
    CDSObjDef	*emptyObjTail, *emptyObjHead;
    int		llx, lly;
    Boolean	extFlg;
    short	extDir, ioDir;
#define		DIM_EXTOBJ 4
    short	numExtObj, eo;
    short	extObj[DIM_EXTOBJ];
    short	doubleExtObj;

    MI_MOD_START("CDSInitGNObj");
    /* Init tNode[] */
    tNode = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);
    for (gn=CDSGNode; gn; gn=gn->next) { /* Fill with nodes with North bch */
	if (!gn->bch[CDS_N].node) continue;
	tNode[numTNode++] = gn;
    }
    /* Sort lowest x then y first */
    qsort((char *)tNode, numTNode, sizeof(CDSGNodeDef*), CDSQsortTNode);

    for (o=0; o<CDSHead.numObj-1; o++) { /* Set obj ring for each obj */
	/* Get lo starting Norch bch */
	loc = CDSObj[o].bdy[CDSObj[o].minLocBdyIdx];

	for (n=numTNode-1; n>=0; n--) {
	    gn = tNode[n];
	    if (gn->loc[CDS_X]<=loc[CDS_X] &&
		gn->loc[CDS_Y]<=loc[CDS_Y] &&
		gn->bch[CDS_N].node->loc[CDS_Y]>loc[CDS_Y]) break;
	}
	DASSERT(n>=0);
	ngn = gn->bch[CDS_N].node;

	CDSInitNObjRing(gn, ngn, o); /* Init ring bchs with obj num */
    }

    gn = CDSGNode->bch[CDS_N].node; /* Init external obj ring */
    CDSInitNObjRing(gn, CDSGNode, (short)CDS_NBOBJ_EXTERNAL);

    for (gn=CDSGNode; gn; gn=gn->next) { /* Init null obj ring bchs */
	ngn = gn->bch[CDS_N].node;
	if (!ngn) continue;
	if (gn->bch[CDS_N].objNum!=CDS_NBOBJ_NOT_SET) continue;

	numExtObj = 0;
	doubleExtObj = -1;

	extFlg = FALSE;
	gStart = gn;
	gn  = NULL;
	rpd = CDS_NEXTDIR(CDS_N, CDS_CW);
	gll = gStart;

	for (g=gStart; gn!=gStart; g=gn) { /* Go around graph face */
	    for (k=CDS_NEXTDIR(rpd,CDS_CCW); /* Get next internal node */
		 !g->bch[k].node; k=CDS_NEXTDIR(k,CDS_CCW));
	    gn = g->bch[k].node;
	    DASSERT(g->bch[k].objNum==CDS_NBOBJ_NOT_SET);
	    rpd = CDS_REVDIR(k);
	    /* Check if on external boundary */
	    o = gn->bch[rpd].objNum;
	    if (o==CDS_NBOBJ_EXTERNAL) {
		extFlg = TRUE;
		extDir = k;
	    }
	    /* Check if bordering object is on */
	    /*    two sides of empty? room */
	    else if (o>=0 && numExtObj<DIM_EXTOBJ) {
		for (eo=0; eo<numExtObj; eo++)
		  if (extObj[eo]==o) break;
		if (eo<numExtObj)
		  doubleExtObj = o;
		else
		  extObj[numExtObj++] = o;
	    }
	    /* Get lower left corner */
	    if (gn ->loc[CDS_X]+gn ->loc[CDS_Y]<=
		gll->loc[CDS_X]+gll->loc[CDS_Y])
	      gll = gn;
	}
	DASSERT(gll->bch[CDS_N].node);
	DASSERT(gll->bch[CDS_E].node);

#ifdef CHECK_EXTIO
	if (extFlg) {
	    switch(extDir) {
	    case CDS_N: ioDir = CDS_IO_LEFT;   break;
	    case CDS_S: ioDir = CDS_IO_RIGHT;  break;
	    case CDS_E: ioDir = CDS_IO_TOP;    break;
		default   : ioDir = CDS_IO_BOTTOM; break;
	    }
	    if (!CDSIoFlag[ioDir] && CDSEmptyRoomFlg) {
		MSG(-1, ("Adding empty room at %d %d on ext side %s\n",
			 gn->loc[CDS_X], gn->loc[CDS_Y], CDS_IO_TOSTR(ioDir)));
		extFlg = FALSE;
	    }
	}
#endif

	if (doubleExtObj>=0) {
	    if (CDSEmptyRoomFlg)
	      MSG(-1, ("Found empty room with obj %d adjacent to 2 sides\n",
		       doubleExtObj));
	    /* Make sure not rectangular */
	    DASSERT(CDSObj[doubleExtObj].numBdy>4);
	    extFlg = TRUE;
	}

	if (!CDSEmptyRoomFlg || extFlg) {
	    CDSInitNObjRing(gn, gn->bch[CDS_N].node, nullObjNum--);
	    if (CDSEmptyRoomFlg)
	      MSG(-1, ("Not adding empty room at %d %d\n",
		       gn->loc[CDS_X], gn->loc[CDS_Y]));
	}
	else {
	    CDSInitNObjRing(gn, gn->bch[CDS_N].node,
			    CDSHead.numObj+numEmptyObj-1);

	    llx = (gll->loc[CDS_X]+gll->bch[CDS_E].node->loc[CDS_X])/2;
	    lly = (gll->loc[CDS_Y]+gll->bch[CDS_N].node->loc[CDS_Y])/2;
	    llx = llx-CDSEmptyRoomDimX/2;
	    lly = lly-CDSEmptyRoomDimY/2;
	    if ((llx<=gll->loc[CDS_X]) || (lly<=gll->loc[CDS_Y]))
	      MIerror("Empty room at (%d %d) too large\n",
		      llx+CDSEmptyRoomDimX/2, lly+CDSEmptyRoomDimY/2);

	    obj = ALLOC_TYPE(CDSObjDef);
	    obj->dbRef = NULL;
	    obj->dbRefId = oct_null_id;
	    if (numEmptyObj)
	      emptyObjTail->dbRef = (RefAny)obj;
	    else
	      emptyObjHead = obj;
	    emptyObjTail = obj;
	    numEmptyObj++;
	    nullObjNum--;

	    obj->ref = NULL;
	    obj->numBdy = 4;
	    obj->bdyOrient = CDS_CCW;
	    obj->minLocBdyIdx = 0;
	    obj->bdy = ALLOC_ARRAY(obj->numBdy, CDSLocDef);

	    obj->bdy[0][CDS_X] = llx;
	    obj->bdy[0][CDS_Y] = lly;
	    obj->bdy[1][CDS_X] = llx+CDSEmptyRoomDimX;
	    obj->bdy[1][CDS_Y] = lly;
	    obj->bdy[2][CDS_X] = llx+CDSEmptyRoomDimX;
	    obj->bdy[2][CDS_Y] = lly+CDSEmptyRoomDimY;
	    obj->bdy[3][CDS_X] = llx;
	    obj->bdy[3][CDS_Y] = lly+CDSEmptyRoomDimY;
	    obj->ioPadFlag = FALSE;
	}
    }

    if (numEmptyObj>0) {
	oldObjList = CDSObj;
	CDSObj = ALLOC_ARRAY(CDSHead.numObj+numEmptyObj, CDSObjDef);

	for (o=0; o<CDSHead.numObj-1; o++) /* Copy old stuff */
	  CDSObj[o] = oldObjList[o];
	/* Copy empty room objs */
	for (obj=emptyObjHead; obj; obj=emptyObjHead, o++) {
	    CDSObj[o] = *obj;
	    CDSObj[o].dbRef = NULL;
	    CDSObj[o].dbRefId  = oct_null_id;
	    emptyObjHead = (CDSObjDef *)obj->dbRef;
	    FREE_TYPE(obj, CDSObjDef);
	}
	/* Copy external object */
	CDSObj[o] = oldObjList[CDSHead.numObj-1];

	FREE_ARRAY(CDSHead.numObj, oldObjList, CDSObjDef);
	CDSHead.numObj += numEmptyObj;
    }
    else
      CDSEmptyRoomFlg = FALSE;

    MSG(1, ("%d empty rooms\n", -nullObjNum+CDS_NBOBJ_EXTERNAL-1));

    FREE_ARRAY(CDSHead.numGNode, tNode, CDSGNodeDef*);

    MI_MOD_END();
}				/* ............................. End of procedure CDSInitGNObj */
#endif /* WITHOUT_EMPTY_ROOMS */
/* Sort North branch nodes by lowest x then lowest y first.
----------------------------------------------------------------------------*/
int
CDSQsortTNode(n1, n2)

CDSGNodeDef	**n1, **n2;
{
int		l1, l2;

l1 = (*n1)->loc[CDS_X];
l2 = (*n2)->loc[CDS_X];
if (l1!=l2) return(l1-l2);
l1 = (*n1)->loc[CDS_Y];
l2 = (*n2)->loc[CDS_Y];
DASSERT(l1!=l2);
return(l1-l2);
} /* ............................. End of procedure CDSQsortTNode */
/* Starting from gStart and going towards gNext, initialize internal node ring
objNum.  Go around clockwise for northbound vector (internal ring), counter-
clockwise for southbound vector (external ring).
----------------------------------------------------------------------------*/
CDSInitNObjRing(gStart, gNext, objNum)

CDSGNodeDef	*gStart, *gNext;
short		objNum;
{
short		rpd, d, k;
CDSGNodeDef	*g, *gn;

MI_MOD_START("CDSInitNObjRing");

MSG(1, ("Init node ring objNum: gs=%d (%d %d), gn=%d (%d %d), o=%d\n",
    CDS_ID(gStart), gStart->loc[CDS_X], gStart->loc[CDS_Y],
    CDS_ID(gNext ), gNext ->loc[CDS_X], gNext ->loc[CDS_Y], objNum));

CDS_GETDIR(d, gStart->loc, gNext->loc);

gn = NULL;
rpd = CDS_NEXTDIR(d, CDS_CW);
for (g=gStart; gn!=gStart; g=gn) {	/* Go around graph face */
    for (k=CDS_NEXTDIR(rpd,CDS_CCW);	/* Get next internal node */
	!g->bch[k].node; k=CDS_NEXTDIR(k,CDS_CCW));

    gn  = g->bch[k].node;
    MSG(3, ("Next node=%d (%d %d) dir=%s, o=%d\n",
        CDS_ID(gn), gn->loc[CDS_X], gn->loc[CDS_Y],
	CDS_CDIR_TOSTR(k), gn->bch[k].objNum));
    DASSERT(g->bch[k].objNum==CDS_NBOBJ_NOT_SET);
    g->bch[k].objNum = objNum;

    DASSERT(g!=gStart || gn==gNext);
    rpd = CDS_REVDIR(k);
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitNObjRing */
