/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSUtil.c

CDSCleanUp		Free atlas internal data structures

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSUtil.c,v $
 * Revision 1.1  90/02/07  16:12:29  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:54  octtools
 * Initial revision
 * 
 * Revision 1.5  88/04/15  14:11:13  mitch
 * Free slice for external channels.
 * 
 * Revision 1.2  88/04/05  12:32:21  mitch
 * Added CDSPrintSeqFile and added new function for CDSCleanUp to handle
 * "CDSObj" mode.
 * 
 * Revision 1.1  88/03/31  13:55:01  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "miutil.h"
#include "CDSInc.h"

/* Clean up all temporary arrays in CDS routines.
----------------------------------------------------------------------------*/
CDSCleanUp(mode)

String		mode;
{
CDSGNodeDef	*gn, *gnNext;
CDSGEdgeDef	*ge, *geNext;
short		o, d;
CDSTreeDef	*t;
CDSBlkDef	*b, *bn;

MI_MOD_START("CDSCleanUp");

if (!strcmp(mode, "CDSEdge")) {
    if (CDSEdgeArray) {
	FREE_ARRAY(CDSHead.numEdge, CDSEdgeArray, CDSEdgeDef);
	CDSEdgeArray = NULL;
    }
    goto CDSEndCleanUp;
}

if (!strcmp(mode, "CDSHg")) {
    for (o=0; o<CDSHead.numObj; o++) {
	if (!CDSObj[o].bdy) continue;
	FREE_ARRAY(CDSObj[o].numBdy, CDSObj[o].bdy, CDSLocDef);
	CDSObj[o].bdy = NULL;
    }
    for (gn=CDSGNode; gn; gn=gnNext) {
	gnNext = gn->next;
	FREE_TYPE(gn, CDSGNodeDef);
    }
    CDSGNode = NULL;
    for (d=CDS_V; ; d=CDS_H) {
	for (ge=CDSGEdge[d]; ge; ge=geNext) {
	    geNext = ge->next;
	    FREE_TYPE(ge, CDSGEdgeDef);
	}
	CDSGEdge[d] = NULL;
	if (d==CDS_H) break;
    }
    goto CDSEndCleanUp;
}

for (o=0; o<CDSHead.numObj; o++) {
    if (CDSObj[o].dbRef) {
	FREE_STRING(CDSObj[o].dbRef);
	CDSObj[o].dbRef = NULL;
    }
    FREE_ARRAY(CDSObj[o].numBdy, CDSObj[o].bdy, CDSLocDef);
    if (CDSObj[o].numCorn) {
	FREE_ARRAY(CDSObj[o].numCorn, CDSObj[o].corn, CDSCornDef);
	CDSObj[o].numCorn = 0;
    }
}
FREE_ARRAY(CDSHead.numObj, CDSObj, CDSObjDef);
if (!strcmp(mode, "CDSObj")) goto CDSEndCleanUp;

for (gn=CDSGNode; gn; gn=gnNext) {
    gnNext = gn->next;
    FREE_TYPE(gn, CDSGNodeDef);
}
CDSGNode = NULL;

for (d=CDS_V; ; d=CDS_H) {
    for (ge=CDSGEdge[d]; ge; ge=geNext) {
	geNext = ge->next;
	FREE_TYPE(ge, CDSGEdgeDef);
    }
    CDSGEdge[d] = NULL;
    if (d==CDS_H) break;
}

if (CDSTree) {
    MI_STACK_PUSH(CDSTree);
    while (MIStackPtr) {
	MI_STACK_POP(CDSTreeDef*, t);
	if (t->child[CDS_LO])
	    MI_STACK_PUSH(t->child[CDS_LO]);
	if (t->child[CDS_HI])
	    MI_STACK_PUSH(t->child[CDS_HI]);
	if (t->type>0)
	    FREE_ARRAY(t->type, t->slice, CDSLocDef);
	if (t->type==CDS_TREE_EXT_DONE)
	    FREE_ARRAY(2, t->slice, CDSLocDef);
	FREE_TYPE(t, CDSTreeDef);
    }
    CDSTree = NIL(CDSTreeDef);
}

for (b=CDSBlk; b; b=bn) {
    bn = b->next;
    FREE_TYPE(b, CDSBlkDef);
}
CDSBlk = NIL(CDSBlkDef);

CDSEndCleanUp: NOTHING
MI_MOD_END();
} /* ............................. End of procedure CDSCleanUp */
