/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSlice.c

Slice channels for cut lines

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CDSlice.c,v $
 * Revision 1.1  90/02/07  16:12:31  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:55  octtools
 * Initial revision
 * 
 * Revision 1.5  88/04/15  14:10:52  mitch
 * Put in channel path for used external channels.
 * 
 * Revision 1.1  88/03/31  11:39:30  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"
#include "CDSInc.h"

/* Slice channels for cut lines
----------------------------------------------------------------------------*/
int
CDSlice()
{
#define		CDS_DIM_SLICENODE 2000
CDSTreeDef	*t, *tc;
short		numSliceNode, s;
CDSGNodeDef	*sliceNode[CDS_DIM_SLICENODE];
short		d, pd, rpd, k, hl;
short		ori;
char		et;
char		chanDir;
CDSGNodeDef	*gn, *gn2, *pgn, *pgn2, *gnf, *gna,
		*gnts, *gnt;

short		numTSlice;
CDSLocDef	tSliceLoc[CDS_DIM_SLICENODE];
int		CDSFindBestSlice();

int		xy, minXY;
CDSGNodeDef	*minGn;
int		rc = 0;
int		rct;

Boolean		ioHiLoFlg;		/* Variables for io ring slices */
short		io;

MI_MOD_START("CDSlice");

CDSInitSlice();

for (io=0; io<4; io++)
    CDSIo[io].sliceTriedFlg = FALSE;

while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    DASSERT(t->type==CDS_TREE_NOT_DONE);

    if (*debugFlag>=4)
	CDSPrintNGraph(t->node);

    numSliceNode = CDS_DIM_SLICENODE;
    rct = CDSFindBestSlice(t->node, &numSliceNode, sliceNode,
			   &ioHiLoFlg, &io, &chanDir, t->ioFlag);
    rc  = MAX(rct, rc);
    MSG(2, ("Find best slice rc=%d, last rc=%d\n", rct, rc));

    if (numSliceNode==0) {		/* Leaf channel */
	DASSERT(t->node->bch[CDS_N].node);
	t->type = CDS_TREE_ELEMENT;
	t->objNum = t->node->bch[CDS_N].objNum;
	t->child[CDS_LO] = NULL;
	t->child[CDS_HI] = NULL;
	continue;
    }
    else if (numSliceNode== -1) {	/* Merged empty rooms - retry */
	MI_STACK_PUSH(t);
	continue;
    }
    t->node = NULL;
					/* Real channel */
    DRANGE(2,numSliceNode,CDS_DIM_SLICENODE);
    DASSERT(t->id!=CDS_NULL_TREEID);

    CDS_GETDIR(d, sliceNode[0]->loc, sliceNode[1]->loc);
    ori = ((d==CDS_E || d==CDS_S) ? CDS_CW : CDS_CCW);
    pd  = CDS_NEXTDIR(d, ori);
    pgn = pgn2 = NULL;
    numTSlice = 0;

    for (s=0; s<numSliceNode; s++) {	/* Unzip node graph along sliceNode */
	gn  = sliceNode[s];

	gn2 = ALLOC_TYPE(CDSGNodeDef);	/* Create new node copy of gn */
	gn2->next = CDSGNode;
	gn2->id   = CDSHead.numGNode++;
	gn2->type = gn->type = 'e';
	gn2->loc[CDS_X] = gn->loc[CDS_X];
	gn2->loc[CDS_Y] = gn->loc[CDS_Y];
	for (k=0; k<4; k++) {
	    gn2->bch[k].type   = '*';
	    gn2->bch[k].node   = NULL;
	    gn2->bch[k].treeId = gn->bch[k].treeId;
	    gn2->bch[k].objNum = gn->bch[k].objNum;
	}
	CDSGNode = gn2;

	if (s<numSliceNode-1)		/* Not last slice node */
	    gnf = sliceNode[s+1];	/* Go forwards */
	else {
	    gnf = sliceNode[s-1];	/* Go backwards */
	    ori = CDS_REVORI(ori);
	}
					/* gn2 is on the side of gn to gnf */
					/*   in the ori direction */
	rpd = CDS_REVDIR(pd);
	CDS_GETDIR(d, gn->loc, gnf->loc);
	for (k=CDS_NEXTDIR(d,ori); k!=rpd; k=CDS_NEXTDIR(k,ori)) {
	    MSG(3, ("Reattaching from node %d (%d %d) in %s\n",
		CDS_ID(gn), gn->loc[CDS_X], gn->loc[CDS_Y],
		CDS_CDIR_TOSTR(k)));
	    gna = gn->bch[k].node;
	    et  = gn->bch[k].type;
	    if (gna) {
		gn2->bch[k].type = et;
		gn2->bch[k].node = gna;
		gn ->bch[k].node = NULL;
		gn ->bch[k].type = '*';
		DASSERT(gna->bch[CDS_REVDIR(k)].node==gn);
		gna->bch[CDS_REVDIR(k)].node = gn2;

		if (et=='e') break;	/* Escape for 1st slice pt */
	    }
	}

        if (pgn) {			/* Connect up nodes along slice bdy */
	    DASSERT(gn ->bch[rpd].node==pgn);
	    DASSERT(pgn->bch[ pd].node== gn);
	    gn  ->bch[rpd].type   = 'e';
	    gn  ->bch[rpd].treeId = t->id;
	    pgn ->bch[ pd].type   = 'e';
	    pgn ->bch[ pd].treeId = t->id;
	    gn2 ->bch[rpd].type   = 'e';
	    gn2 ->bch[rpd].treeId = t->id;
	    pgn2->bch[ pd].type   = 'e';
	    pgn2->bch[ pd].treeId = t->id;
	    gn2 ->bch[rpd].node = pgn2;
	    pgn2->bch[ pd].node = gn2;
	}
					/* Flush redundant pts from slice */
	if (pd!=d || s==numSliceNode-1) {
	    CDS_COPY_LOC(tSliceLoc[numTSlice], sliceNode[s]->loc);
	    numTSlice++;
	}

	pd   = d;
	pgn  = gn;
	pgn2 = gn2;
    }

    t->type  = numTSlice;		/* Copy slice from tSlice */
    t->slice = ALLOC_ARRAY(numTSlice, CDSLocDef);
    t->dir   = chanDir;
    for (s=0; s<numTSlice; s++)
	CDS_COPY_LOC(t->slice[s], tSliceLoc[s]);

    for (hl=CDS_LO; ; hl=CDS_HI) {	/* Create two new channels on tree */
					/* Low side gn2 gets pushed first */
					/*   i.e. processed last */
	tc = ALLOC_TYPE(CDSTreeDef);
	t->child[hl] = tc;
	tc->id       = CDSHead.numTree++;
	tc->type     = CDS_TREE_NOT_DONE;
	tc->order    = CDS_UNORDERED;
	tc->parentId = t->id;
	tc->dir      = '*';
	tc->ioFlag   = t->ioFlag;
					/* Find low extern corner of graph */
	if (hl==CDS_LO)
	then gnt = gnts = gn2;
	else gnt = gnts = gn;
	MSG(3, ("Starting at gn=%d (%d %d)\n",
	    CDS_ID(gnt), gnt->loc[CDS_X], gnt->loc[CDS_Y]));

	minXY = MAXINT;			/* Go around external part of graph */
	for (d=CDS_N; ; d=CDS_NEXTDIR(d,CDS_CW)) {
	    for ( ; gnt->bch[d].type!='e'; d=CDS_NEXTDIR(d,CDS_CCW));
	    MSG(4, ("On gnt=%d (%d %d) nextDir=%s\n",
		CDS_ID(gnt), gnt->loc[CDS_X], gnt->loc[CDS_Y],
		CDS_CDIR_TOSTR(d)));
	    xy = gnt->loc[CDS_X]+gnt->loc[CDS_Y];
	    if (xy<minXY) {
		minXY = xy;
		minGn = gnt;
	    }
	    gnt = gnt->bch[d].node;
	    if (gnt==gnts) break;
	}
	DASSERT(minGn->bch[CDS_N].node);
	tc->node = minGn;
					/* Push children on stack */
	MSG(2, ("Pushing child %d: %s, node=%d\n",
	    tc->id, CDS_HILO_TOSTR(hl), tc->node->id));
	if (io>=0) {
	    DASSERT(CDSIo[io].sliceFoundFlg);
	    if (hl==CDS_LO) {
		if (ioHiLoFlg)
		    CDSIo[io].bodyGNode = tc->node;
		else {
		    CDSIo[io].sideTree  = tc;
		    tc->ioFlag = TRUE;
		}
	    }
	    else {
		if (!ioHiLoFlg)
		    CDSIo[io].bodyGNode = tc->node;
		else {
		    CDSIo[io].sideTree  = tc;
		    tc->ioFlag = TRUE;
		}
	    }
	}
	MI_STACK_PUSH(tc);

	if (hl==CDS_HI) break;
    }
}

MI_MOD_END();
return(rc);
} /* ............................. End of procedure CDSlice */
/* Do the first four external channel slices.
----------------------------------------------------------------------------*/
CDSInitSlice()
{
CDSTreeDef	*t, *tp;
short		k;
CDSLocDef	*extBdy;

MI_MOD_START("CDSInitSlice");

CDSHead.numTree = 0;			/* Init top node on tree */

tp = NULL;
extBdy = CDSObj[CDSHead.numObj-1].bdy;

for (k=0; k<5; k++) {
    t = ALLOC_TYPE(CDSTreeDef);
    if (k==0)
	CDSTree = t;
    t->id    = CDSHead.numTree++;
    t->node  = CDSGNode;
    t->type  = CDS_TREE_EXT_DONE;
    t->order = CDS_UNORDERED;
    t->dir   = '*';
    t->ioFlag = FALSE;
    t->useExtFlag = FALSE;
    t->child[CDS_LO] = NULL;
    t->child[CDS_HI] = NULL;
    t->slice = ALLOC_ARRAY(2, CDSLocDef);

    t->slice[0][CDS_X] = extBdy[0][CDS_X];
    t->slice[0][CDS_Y] = extBdy[0][CDS_Y];
    t->slice[1][CDS_X] = extBdy[2][CDS_X];
    t->slice[1][CDS_Y] = extBdy[2][CDS_Y];
    if      (k==CDS_W_EXT_TREEID) t->slice[1][CDS_X] = t->slice[0][CDS_X];
    else if (k==CDS_E_EXT_TREEID) t->slice[0][CDS_X] = t->slice[1][CDS_X];
    else if (k==CDS_S_EXT_TREEID) t->slice[1][CDS_Y] = t->slice[0][CDS_Y];
    else if (k==CDS_N_EXT_TREEID) t->slice[0][CDS_Y] = t->slice[1][CDS_Y];

    if (tp) {
	t->parentId = tp->id;
	switch(k-1) {			/* Channels in reverse order in which */
					/*   they are routed i.e. N is first */
					/*   and W is last */
	case CDS_W_EXT_TREEID:
	    tp->child[CDS_HI] = t;
	    tp->dir = 'v';		/* Parent is a vertical chan */
	    break;
	case CDS_E_EXT_TREEID:
	    tp->child[CDS_LO] = t;
	    tp->dir = 'v';
	    break;
	case CDS_S_EXT_TREEID:
	    tp->child[CDS_HI] = t;
	    tp->dir = 'h';		/* Parent is a horizontal chan */
	    break;
	case CDS_N_EXT_TREEID:
	    tp->child[CDS_LO] = t;
	    tp->dir = 'h';
	    break;
	default: MIerror("in building external tree\n");
	}
    }
    else
	t->parentId = CDS_NULL_TREEID;
    tp = t;
}

t->type = CDS_TREE_NOT_DONE;
MI_STACK_PUSH(t);

MI_MOD_END();
} /* ............................. End of procedure CDSInitSlice */
/* Order channels:  1st channel to be routed has order=0; last channel to be
routed has order numChan-1.
----------------------------------------------------------------------------*/
CDSChanOrder(t)

CDSTreeDef	*t;
{
Boolean		leafCellFlg = TRUE;

MI_MOD_START("CDSChanOrder");

MSG(2, ("On chan node=%d\n", t->id));

if (t->child[CDS_LO]) {
    CDSChanOrder(t->child[CDS_LO]);
    leafCellFlg = FALSE;
}

if (t->child[CDS_HI]) {
    CDSChanOrder(t->child[CDS_HI]);
    leafCellFlg = FALSE;
}

if (!leafCellFlg)			/* Is a channel */
    t->order = CDSHead.numChan++;

MI_MOD_END();
} /* ............................. End of procedure CDSChanOrder */
