/*
miutil.c:

General C-language utility routines (to be used in conjunction with MIInc.h)

MIModStatStart		Initialize statistics and debug environment.
MIModStatEnd		Print out statistics and free up debug environment.
MIInitEnv		Changing of debug environment (interrupt handler).
MIFindSubIdx		Find subroutine index (if it does not exist then add new
			entry to module table).
MIfopen			fopen with system error reporting.
MIerror			Print error messsage
MITokenize		Tokenize a string.
MIContinue		Prompt user to continue running.
MIMaxMem		Get maximum memory usage in megabytes.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	miutil.c,v $
 * Revision 1.1  90/02/07  16:12:35  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.3  88/04/24  16:56:29  mitch
 * Put in package prefix
 * 
 * Revision 1.1  88/03/31  15:11:42  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include <signal.h>
#define MIUTIL_MAIN
#include "miutil.h"
#include "varargs.h"
#include "vov.h"

#ifdef sprite
#define vtimes	tms
static void tms(struct tms *buffer, struct tms *buff2)
{	times(buffer); }
#endif

#define	MI_MOD_STAT_DIM   200
#define	MI_TRACE_BACK_DIM 500

static short	MINumMod;
static Boolean	MIDefaultTraceFlag = FALSE;
static Boolean	MIDefaultTimeFlag  = FALSE;
static Boolean	MISigSavedFlg = FALSE;
static Boolean	MIPrtHeaderFlag;
static FILE	*MIEnvFile;
static struct vtimes MIModStatVm;
void		(*MISaveSig)();

extern int	errno;			/* System error number */
extern String	sys_errlist[];		/* SysErr reason (indexed by errno) */

/* Initialize module statistics and control array.
   Reset interrupt handler to MIInitEnv.
----------------------------------------------------------------------------*/
MIModStatStart(name, debugFileName, prtHeaderFlag)

String		name;
String		debugFileName;
Boolean		prtHeaderFlag;
{
short		m;
long		tloc, time();
String		ctime();
short		MIFindModIdx();
FILE		*MIfopen();

MIPrtHeaderFlag = prtHeaderFlag;
tloc = time((long*)0);
if (MIPrtHeaderFlag) {
    MSG(0, ("-- Start \"%s\" on %s", name, ctime(&tloc)));
    MSG(0, ("Initial memory size = %.2fM\n", MIMaxMem()));
}

if (MIAllocDebugFlag) {
    MIAllocFile = MIfopen("allo","w","alloc/free debug");
    if (!MIAllocFile)
	MIAllocDebugFlag = FALSE;
}

MIModStat   = ALLOC_ARRAY(MI_MOD_STAT_DIM, MIModStatDef);
MITraceBack = ALLOC_ARRAY(MI_TRACE_BACK_DIM, short);
MINumMod    = MINumTraceBack = 0;
m = MIFindModIdx(name, FALSE);
DASSERT(m==0);

if (debugFileName) {
    if (MIEnvFile=MIfopen(debugFileName,"r","run-time debug")) {
					/* Save default interrupt hander */
	MISaveSig = signal(SIGINT, MIInitEnv);
	MISigSavedFlg = TRUE;
#if defined(hpux)
	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
	setvbuf(stderr, NULL, _IOLBF, BUFSIZ);
#else
	setlinebuf(stdout);
	setlinebuf(stderr);
#endif
	MIInitEnv();
    }
    else MISigSavedFlg = FALSE;
}
else MISigSavedFlg = FALSE;

#if !defined(hpux)
vtimes(&MIModStatVm, NIL(struct vtimes));
#endif

} /* ............................. End of procedure MIModStatStart */
/* Print out runtime statistics of all called subroutines.
   Reset interrupt handler to default.
----------------------------------------------------------------------------*/
MIModStatEnd()
{
short		m;
long		tloc, time();
String		ctime();
struct vtimes	end_vm;

if (MISigSavedFlg)			/* Restore interrupt handler */
    MISaveSig = signal(SIGINT, MISaveSig);
RANGE(0, MINumTraceBack, 1);

for (m=1; m<MINumMod; m++) {		/* Cpu times of each module */
    if (MIModStat[m].timeFlag && MIModStat[m].numTime>0) {
	MSG(0, ("Cpu time for module %-16s :%8.3f min%10dx\n",
	    MIModStat[m].name, (float)MIModStat[m].cTime/3600.0,
	    MIModStat[m].numTime));
    }
    else if (MIModStat[m].numTime>0) {
	MSG(1, ("Cpu time for module %-16s :%8.3f min%10dx\n",
	    MIModStat[m].name, (float)MIModStat[m].cTime/3600.0,
	    MIModStat[m].numTime));
    }
    if (MIModStat[m].nameAlloFlag)
	FREE_STRING(MIModStat[m].name);
}

#if !defined(hpux)
vtimes(&end_vm, NIL(struct vtimes));	/* Stats of run */
#endif
#ifndef sprite
MIModStatVm.vm_utime  = end_vm.vm_utime -MIModStatVm.vm_utime;
MIModStatVm.vm_stime  = end_vm.vm_stime -MIModStatVm.vm_stime;
MIModStatVm.vm_majflt = end_vm.vm_majflt-MIModStatVm.vm_majflt;
MIModStatVm.vm_minflt = end_vm.vm_minflt-MIModStatVm.vm_minflt;
MIModStatVm.vm_nswap  = end_vm.vm_nswap -MIModStatVm.vm_nswap;
MIModStatVm.vm_inblk  = end_vm.vm_inblk -MIModStatVm.vm_inblk;
MIModStatVm.vm_oublk  = end_vm.vm_oublk -MIModStatVm.vm_oublk;

MSG(1, ("time u=%.2f s=%.2f, page maj=%d min=%d, swap=%d, i/o r=%d w=%d\n",
    (float)MIModStatVm.vm_utime/3600.0, (float)MIModStatVm.vm_stime/3600.0,
    MIModStatVm.vm_majflt, MIModStatVm.vm_minflt, MIModStatVm.vm_nswap,
    MIModStatVm.vm_inblk , MIModStatVm.vm_oublk));

tloc = time((long*)0);
if (MIPrtHeaderFlag) {
    MSG(0, ("Total Cpu Time = %.2f min\n",
	(float)(MIModStatVm.vm_utime+MIModStatVm.vm_stime)/3600.0));
    MSG(0, ("Maximum memory size = %.2fM\n", MIMaxMem()));
    MSG(0, ("-- End \"%s\" on %s", MIModStat[0].name, ctime(&tloc)));
}
#else
MIModStatVm.tms_utime  = end_vm.tms_utime -MIModStatVm.tms_utime;
MIModStatVm.tms_stime  = end_vm.tms_stime -MIModStatVm.tms_stime;

MSG(1, ("time u=%.2f s=%.2f\n",
    (float)MIModStatVm.tms_utime/3600.0,
    (float)MIModStatVm.tms_stime/3600.0));

tloc = time((long*)0);
if (MIPrtHeaderFlag) {
    MSG(0, ("Total Cpu Time = %.2f min\n",
	(float)(MIModStatVm.tms_utime+MIModStatVm.tms_stime)/3600.0));
    MSG(0, ("-- End \"%s\" on %s", MIModStat[0].name, ctime(&tloc)));
}
#endif

					/* Final clean up */
if (MIAllocDebugFlag) {
    (void) fclose(MIAllocFile);
    MIAllocDebugFlag = FALSE;
}

(void) fflush(stdout);
(void) fflush(stderr);

FREE_ARRAY(MI_MOD_STAT_DIM, MIModStat, MIModStatDef);
FREE_ARRAY(MI_TRACE_BACK_DIM, MITraceBack, short);

} /* ............................. End of procedure MIModStatEnd */
/* Initialization of runtime environment (interrupt handler).
----------------------------------------------------------------------------*/
void
MIInitEnv()
{
FILE		*newEnvFile, *MIfopen();
short		m, t;
int		debug;
int		threshold;
short		MIFindModIdx();

while(1) {
    if (MIEnvFile==stdin) {		/* Prompt (only in interactive mode) */
	fflush(stdout);
	fprintf(stderr,"\ninitEnv> ");
	fflush(stderr);
    }
    if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;

    if	    (!strcmp (MIBuf,"/*")) {	/* Skip comments */
	while(fscanf(MIEnvFile,"%s", MIBuf)!=EOF)
	    if (!strcmp(MIBuf,"*/")) break;
    }
    else if (!strncmp(MIBuf,"debug",1)) {
	if (fscanf(MIEnvFile,"%s %d", MIBuf, &debug)==EOF) break;
	if (!strcmp(MIBuf,"default"))
	    MIDefaultDebugFlag = debug;
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].debugFlag = debug;
	}
    }
    else if (!strncmp(MIBuf,"trace",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	if (!strcmp(MIBuf,"default")) {
	    MIDefaultTraceFlag = !MIDefaultTraceFlag;
	    if (fscanf(MIEnvFile,"%d", &threshold)==EOF) break;
	    MIModThreshold = threshold;
	}
	else if (!strcmp(MIBuf,"alloc")) {
	    if (MIAllocDebugFlag) {
		(void) fclose(MIAllocFile);
		MIAllocDebugFlag = FALSE;
	    }
	    else {
		MIAllocFile = MIfopen("allo","w","alloc/free debug");
		if (MIAllocFile)
		    MIAllocDebugFlag = TRUE;
	    }
	}
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].traceFlag = !MIModStat[m].traceFlag;
	}
    }
    else if (!strncmp(MIBuf,"time",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	if (!strcmp(MIBuf,"default"))
	    MIDefaultTimeFlag = !MIDefaultTimeFlag;
	else {
	    m = MIFindModIdx(MIBuf, TRUE);
	    MIModStat[m].timeFlag = !MIModStat[m].timeFlag;
	}
    }
    else if (!strcmp(MIBuf,"abort")) {
	abort();
    }
    else if (!strncmp(MIBuf,"status",3)) {
	fprintf(stderr,"defaults: debugLevel=%d, %s traced(t), %s timed(.)\n",
	    MIDefaultDebugFlag,
	    MIDefaultTraceFlag ? "" : " not",
	    MIDefaultTimeFlag  ? "" : " not");
	for (m=0; m<MINumMod; m++) {
	    if (m>0 && m%2==0) fprintf(stderr,"\n");
	    fprintf(stderr,"%-20s %c%c %2df%7dx%c   ",
		MIModStat[m].name,
		(MIModStat[m].stopFlag  ? 's' : ' '),
		(MIModStat[m].traceFlag ? 't' : ' '),
		MIModStat[m].debugFlag, MIModStat[m].numTime,
		(MIModStat[m].timeFlag ? '.' : ' '));
	}
    }
    else if (!strncmp(MIBuf,"stop",3)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	m = MIFindModIdx(MIBuf, TRUE);
	MIModStat[m].stopFlag = !MIModStat[m].stopFlag;
    }
    else if (!strncmp(MIBuf,"where",1)) {
	if (MITraceBackFlag) 
	    for (t=MINumTraceBack-1; t>=0; t--) {
		m = MITraceBack[t];
		fprintf(stderr,"%s\n", MIModStat[m].name);
	    }
	else if (0<=MICurrentMod && MICurrentMod<MINumMod)
	    fprintf(stderr,"%s", MIModStat[MICurrentMod].name);
	else
	    fprintf(stderr,"ERROR Cannot determine at this time");
    }
    else if (!strncmp(MIBuf,"file",2)) {
	if (fscanf(MIEnvFile,"%s", MIBuf)==EOF) break;
	newEnvFile = MIfopen(MIBuf,"r","debug setup");
	if (newEnvFile) {
            if (MIEnvFile!=stdin)
		(void) fclose(MIEnvFile);
	    MIEnvFile = newEnvFile;
	}
    }
    else if (!strncmp(MIBuf,"nosignal",1)) {
	if (MISigSavedFlg) {
	    MISaveSig = signal(SIGINT, MISaveSig);
	    MISigSavedFlg = FALSE;
	}
    }
    else if (!strncmp(MIBuf,"foreground",2)) {
	if (MIEnvFile!=stdin) {
	    (void) fclose(MIEnvFile);
	    MIEnvFile = stdin;
	}
    }
    else if (!strncmp(MIBuf,"continue",1)) {
	if (MIEnvFile!=stdin) {
	    (void) fclose(MIEnvFile);
	    MIEnvFile = stdin;
	}
	break;
    }
    else if (!strcmp(MIBuf,"quit") || !strcmp(MIBuf,"exit"))
	exit(-1);
    else if (!strncmp(MIBuf,"help",1) ||
    	     !strncmp(MIBuf,"?",1)) {
	fprintf(stderr,"Summary of Commands:\n");
	fprintf(stderr,"  debug <subrout or \"default\"> <option>\n");
	fprintf(stderr,"  trace <subrout or \"default\" or \"alloc\"> <threshold>\n");
	fprintf(stderr,"  time  <subrout or \"default\">\n");
	fprintf(stderr,"  status\n");
	fprintf(stderr,"  stop <module name>\n");
	fprintf(stderr,"  file <input file name>\n");
	fprintf(stderr,"  foreground\n");
	fprintf(stderr,"  where\n");
	fprintf(stderr,"  continue or ctrl-d\n");
	fprintf(stderr,"  quit, exit or ctrl-c");
    }
    else MIerror("Cannot interpret command <%s>\n", MIBuf);
}

if (MIEnvFile!=stdin)			/* Input file must be stdin */
    MIEnvFile = stdin;
} /* ............................. End of procedure MIInitEnv */
/* Returns subroutine index (if it does not exist then add a new entry into
   the module table).
----------------------------------------------------------------------------*/
short
MIFindModIdx(name, nameAlloFlag)

String		name;			/* name of module */
Boolean		nameAlloFlag;		/* Alloc separate storage for name ? */
{
short		m;

for (m=0; m<MINumMod; m++)		/* Search for name */
    if (!strcmp(name, MIModStat[m].name)) break;

if (m==MINumMod) {			/* Name not found - add new entry */
    URANGE(MINumMod, MI_MOD_STAT_DIM);
    if (!nameAlloFlag)
    then MIModStat[MINumMod].name = name;
    else MIModStat[MINumMod].name = STR_COPY(name);
    MIModStat[MINumMod].nameAlloFlag = nameAlloFlag;
    MIModStat[MINumMod].cTime        = 0;
    MIModStat[MINumMod].numTime      = 0;
    MIModStat[MINumMod].stopFlag     = FALSE;
    MIModStat[MINumMod].debugFlag    = MIDefaultDebugFlag;
    MIModStat[MINumMod].traceFlag    = MIDefaultTraceFlag;
    MIModStat[MINumMod].timeFlag     = MIDefaultTimeFlag;
    for (; *name; name++)
	if (!strncmp(name,"Main",4)) {
	    MIModStat[MINumMod].timeFlag = TRUE;
	    break;
	}
    MINumMod++;
}

return(m);				/* Return index of module entry */

} /* ............................. End of procedure MIFindModIdx */
/* Called version of subroutine start
----------------------------------------------------------------------------*/
MIModStart(name, modNum, debugFlag, start_vm)

String		name;
short		*modNum;
short		**debugFlag;
struct vtimes	*start_vm;
{
short		MIFindModIdx();

if (*modNum<0) {
    *modNum    = MIFindModIdx(name, FALSE);
    *debugFlag = &(MIModStat[*modNum].debugFlag);
}

if (MITraceBackFlag)
then MITraceBack[MINumTraceBack++] = *modNum;
else MICurrentMod = *modNum;

#if !defined(hpux)
if (MIModStat[*modNum].timeFlag)
    vtimes(start_vm, NIL(struct vtimes));
#endif
if (MIModStat[*modNum].traceFlag && MIModStat[*modNum].numTime<MIModThreshold)
    printf("--Start %s\n", MIModStat[*modNum].name);
if (MIModStat[*modNum].stopFlag) {
    MIInitEnv();
}
} /* ............................. End of procedure MIModStart */

/* Called version of subroutine end
----------------------------------------------------------------------------*/
MIModEnd(modNum, start_vm)

short		modNum;
struct vtimes	*start_vm;
{
struct vtimes	end_vm;

if (MIModStat[modNum].timeFlag) {
#if !defined(hpux)
    vtimes(&end_vm, NIL(struct vtimes));
#endif
#ifndef sprite
    MIModStat[modNum].cTime += end_vm.vm_utime-start_vm->vm_utime+
			       end_vm.vm_stime-start_vm->vm_stime;
#else
    MIModStat[modNum].cTime += end_vm.tms_utime-start_vm->tms_utime+
			       end_vm.tms_stime-start_vm->tms_stime;
#endif
}
if (MIModStat[modNum].traceFlag && MIModStat[modNum].numTime<MIModThreshold)
    printf("--End   %s\n", MIModStat[modNum].name);
MIModStat[modNum].numTime++;
if (MITraceBackFlag)
    MINumTraceBack--;
} /* ............................. End of procedure MIModEnd */
/* Just like printf except with fflush.
----------------------------------------------------------------------------*/
FILE *
MIfopen(fileName, mode, fileType)

String		fileName;
String		mode;			/* Supports "r", "w" and "a" */
String		fileType;
{
    FILE		*f;

    f = VOVfopen(fileName, mode);
    if (!f && strcmp(fileType,"run-time debug"))
      MIerror("Cannot %s %s file %s because %s\n",
	      (*mode=='r' ? "read" :
	       (*mode=='w' ? "write" : 
		(*mode=='a' ? "append" : "!$%?"))),
	      (fileType ? fileType : "temp"),
	      fileName, sys_errlist[errno]);
    return(f);
}				/* ............................. End of procedure MIfopen */

char Message[1024];
/* Just like printf except with (void) fflush.
----------------------------------------------------------------------------*/
#ifdef lint
/*ARGSUSED*/
/*VARARGS1*/
#undef va_alist
#undef va_dcl
#define va_alist	fmt
#define va_dcl		char *fmt;
#endif

#ifdef UNUSED
MIerror(va_alist)
    va_dcl
{
    va_list ap;
    char *format;

    va_start(ap);
    format = va_arg(ap, char *);
    (void) uprintf(Message, format, &ap);
    va_end(ap);
    if (MIPrefix)
	fprintf(stderr,"%s: ERROR ", MIPrefix);
    else
	fprintf(stderr,"ERROR ");
    fprintf(stderr,"%s", Message);

#else /* UNUSED */
#if defined(__STDC__)||defined(sprite)
MIerror(va_alist)
va_dcl
{
	va_list va;
	char *fmt;
	va_start(va);
	fmt = va_arg(va, char *);
	fprintf(stderr,"ERROR ");
	vfprintf(stderr, fmt, va);
	va_end(va);
#else /* __STDC__ */
MIerror(fmt, args)

char		*fmt;
{
fprintf(stderr,"ERROR ");
_doprnt(fmt, &args, stderr);
#endif /* __STDC__ */

#endif /* UNUSED */

    if (MIErrorLevel==1) {
	printf("terminate or continue (t or c) ? ");
	scanf("%s", Message);
	if (*Message=='t')
	    abort();
    }
    else if (MIErrorLevel>1)
	abort();
    (void) fflush(stderr);

} /* ............................. End of procedure MIerror */
/* Tokenize input string "buf"; fill "tok" with next available token delimited
   by any character in "delimiters" string.  Return character ptr to next
   delimiter after token.
----------------------------------------------------------------------------*/
String
MITokenize(buf, tok, delimiters)

String		buf;
String		tok;
String		delimiters;
{
String		t, d;
Boolean		skipWhiteSpaceFlg = TRUE;

for (d=delimiters; (*d) && *d!=' '; d++);
if (*d) skipWhiteSpaceFlg = FALSE;

if (skipWhiteSpaceFlg)			/* Skip over white space */
    for ( ; *buf==' ' || *buf=='\t'; buf++);

for (t=tok; *buf; buf++) {
    for (d=delimiters; (*d) && *d!=*buf; d++);
    if (*d) {
	if (t==tok) 
	    *(t++) = *(buf++);
	break;
    }
    else *(t++) = *buf;
}

if (t!=tok && skipWhiteSpaceFlg) {	/* Trim white space */
    for (t--; *t==' '; t--);
    t++;
}

*t = '\0';

return(buf);
} /* ............................. End of procedure MITokenize */

/* Interactively check if user wants to continue.
----------------------------------------------------------------------------*/
MIContinue()
{
char		stuff[40];
Boolean		response;

printf("Continue ? ");
(void) scanf("%s", stuff);
response = (*stuff!='n');

if (!response) exit(4);
}

/* Get maximum memory usage up till now in megabytes.
----------------------------------------------------------------------------*/
double
MIMaxMem()
{
double		maxMemInMeg;
char		*sbrk();
#ifndef _IBMR2
#ifdef lint
extern int	edata;
#else
extern char	*edata;
#endif

maxMemInMeg = (unsigned)sbrk(0)-(unsigned)(&edata);
maxMemInMeg = maxMemInMeg/1048576.0;
#else
maxMemInMeg = 0.0;
#endif

return(maxMemInMeg);
} /* ............................. End of procedure MIMaxMem */
