/*
miutil.h:

General C-language support defines, varables, and macros.

Author(s):		Mitsuru Igusa
----------------------------------------------------------------------------*/
#include <sys/vtimes.h>

typedef char	Boolean;
typedef char	*RefAny;
typedef char	*String;
typedef char	Byte;

#ifndef PORT_H
extern char	*malloc();
extern char	*strcpy();
extern int	strlen();
extern int	strcmp();
extern void	exit();
extern void	free();
extern void	abort();
#endif /* PORT_H */

extern FILE	*MIfopen();
extern void	MIInitEnv();
extern double	MIMaxMem();

#define TRUE 1
#define FALSE 0
#define ABS(A) ((A)>0?(A):-(A))

#ifndef MAX
#define MAX(A,B) ((A)>(B)?(A):(B))
#endif

#ifndef MIN
#define MIN(A,B) ((A)<(B)?(A):(B))
#endif

#ifdef MIUTIL_MAIN
short		MIDefaultDebugFlag = -1;
short		MIErrorLevel = 0;
short		*debugFlag = &MIDefaultDebugFlag;
Boolean		MIAllocDebugFlag = FALSE;
char		MIBuf[1000];
String		MIPrefix = NULL;
#endif /* MIUTIL_MAIN */
extern short	MIDefaultDebugFlag;
extern short	MIErrorLevel;
extern short	*debugFlag;
extern Boolean	MIAllocDebugFlag;
extern char	MIBuf[];
extern String	MIPrefix;

#ifndef NIL
#define NIL(type) ((type *) 0)
#endif /* NIL */
#define then
#define NOTHING {}
#define MAXSHORT (32767)
#define MAXINT   (2147483647)
#ifndef MAXFLOAT
#define MAXFLOAT (1.0e30)
#endif /* MAXFLOAT */
#define MINMAX(MIN,MAX,A,B) {if ((A)<(B)) {MIN=(A); MAX=(B);} \
			     else	  {MIN=(B); MAX=(A);}}
#define DELTA(A,B) ((A)>(B)?(A)-(B):(B)-(A))
#define ROUND(A) ((A)>=0?((A)+.5):((A)-.5))
#define STR_COPY(S) strcpy(ALLOC_ARRAY(strlen(S)+1, char),S)
#define CONV_CHAR_TO_INT(c) (int)((c)&0x80 ? ((c)&0x7f)+0x80 : (c))
#define CONV_BOOLEAN_TO_CHAR(b) (char)((b) ? 'y' : 'n')
#define CONV_CHAR_TO_BOOLEAN(c) ((c)=='y' || (c)=='Y')
#define PRT_BOOLEAN(x) ((x) ? "Y" : "N")
#define PRT_STRING(x)  ((x) ? (x) : "NULL")
#define MI_MAXMEM(level, msg) { \
    if (*debugFlag>=level) { \
	if (MIPrefix) \
	    printf("%s: ", MIPrefix); \
	printf("Maximum memory size (%s) = %.2fM\n", msg, MIMaxMem()); \
    } \
}
/* miutil.h: General globals for Mitch's routines
----------------------------------------------------------------------------*/
#ifdef MIUTIL_MAIN
String MIAssertFmt  = "Line %d of file \"%s\" Assert failed\n";
String MIRangeFmt   = "Line %d of file \"%s\" Variable out of range lo=%d value=%d hi=%d\n";
String MIURangeFmt  = "Line %d of file \"%s\" Variable exceeded range value=%d limit=%d\n";
String MIAssertSFmt = "Line %d of file \"%s\" Variable %s=%d but should be %d\n";
String MIAllocChkFmt= "Line %d of file \"%s\" Could not allocate variable %s\nCurrent memory = %.2fM and trying to allocate %d bytes\n";
String MIAllocFmt   = "%8d + %7d %4d %s\n";
String MIFreeFmt    = "%8d - %7d %4d %s\n";
#endif /* MIUTIL_MAIN */
extern String	MIAssertFmt;
extern String	MIRangeFmt;
extern String	MIURangeFmt;
extern String	MIAssertSFmt;
extern String	MIAllocChkFmt;
extern String	MIAllocFmt;
extern String	MIFreeFmt;
FILE		*MIAllocFile;
String		MIAllocPtr;

#define ASSERT(cond99) \
    { if (!(cond99)) \
	MIerror(MIAssertFmt,__LINE__,__FILE__); }
#define FASSERT(cond99) \
    { ASSERT(cond99) exit(99); }
#define RANGE(lo99, val99, hi99) \
    { if ((val99)<(lo99) || (hi99)<=(val99)) \
    	MIerror(MIRangeFmt,__LINE__,__FILE__,lo99,val99,hi99); } 
#define URANGE(val99, limit99) \
    { if ((limit99)<=(val99)) \
    	MIerror(MIURangeFmt,__LINE__,__FILE__,val99,limit99); } 
#define ASSERTSAME(val99, val98) \
    { if ((val99)!=(val98)) \
    	MIerror(MIAssertSFmt,__LINE__,__FILE__,"val99",val99,val98); } 
#define ALLOC_CHK(ptr, number, type) \
    { if (!(ptr)) { \
	char	*sbrk(); \
	MIerror(MIAllocChkFmt,__LINE__,__FILE__,"ptr", \
	    ((unsigned)sbrk(0))/1048576., (number)*sizeof(type)); \
	abort(); \
    } }

#ifdef MI_DEBUG
#   define MSG(level, msg)	{if (*debugFlag>=level) { \
				    if (MIPrefix) \
					printf("%s: ", MIPrefix); \
				    printf msg; } }
#   define DASSERT(cond)	 ASSERT(cond)
#   define DRANGE(lo, value, hi) RANGE(lo, value, hi)
#   define DURANGE(value, limit) URANGE(value, limit)
#   define DASSERTSAME(v1, v2)   ASSERTSAME(v1, v2)
#   define ALLOC_ARRAY(number, type) \
	   (MIAllocDebugFlag ? \
		(fprintf(MIAllocFile, MIAllocFmt, \
		    (MIAllocPtr = malloc((unsigned)(number)*sizeof(type))), \
		    number*sizeof(type), __LINE__, __FILE__), \
		 (type *) MIAllocPtr) : \
		((type *) malloc((unsigned)(number)*sizeof(type))) )
#   define FREE_ARRAY(number, ptr, type) { \
	    free((char *)ptr); \
	    if (MIAllocDebugFlag) \
		fprintf(MIAllocFile, MIFreeFmt, \
		    ptr, number*sizeof(type), __LINE__, __FILE__); }
#else /* MI_DEBUG */
#   define MSG(level, msg)		 NOTHING
#   define DASSERT(cond)		 NOTHING
#   define DRANGE(lo, value, hi)	 NOTHING
#   define DURANGE(value, limit)	 NOTHING
#   define DASSERTSAME(v1, v2)   ASSERTSAME(v1, v2)
#   define ALLOC_ARRAY(number, type)	 (type *) malloc((unsigned)(number)*sizeof(type))
#   define FREE_ARRAY(number, ptr, type) free((char *)(ptr))
#endif /* MI_DEBUG */
#define ALLOC_TYPE(type) 		 ALLOC_ARRAY(1, type)
#define FREE_TYPE(ptr, type)		 FREE_ARRAY(1, ptr, type)
#define FREE_STRING(str)		 FREE_ARRAY(strlen(str)+1, str, char)
/* miutil.h: Variables and macros to time subroutines.
----------------------------------------------------------------------------*/
typedef struct	MIModStatDef	MIModStatDef;

struct MIModStatDef {
    String	name;		/* Name of module */
    short	debugFlag;	/* Debug number */
    Boolean	nameAlloFlag;	/* Did a copy of the name get allocated ? */
    Boolean	stopFlag;	/* Stop in module flag */
    Boolean	traceFlag;	/* Trace entry/exit of module ? */
    Boolean	timeFlag;	/* Time module ? */
    int		cTime;		/* Cpu time */
    int		numTime;	/* Number of module invokations */
};

#ifdef MIUTIL_MAIN
Boolean		MITraceBackFlag	   = FALSE;
int		MIModThreshold     = 25;
#endif /* MIUTIL_MAIN */
extern Boolean	MITraceBackFlag;
extern int	MIModThreshold;

MIModStatDef	*MIModStat;
short		MINumTraceBack;
short		*MITraceBack;
short		MICurrentMod;

#ifdef MI_DEBUG_INLINE
/* Use this macro to time any subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_START(name99) \
\
/* String	name99; */ \
{ \
static MIModStatDef *modStat99 = NULL; \
static short	modNum99; \
static short	*debugFlag; \
struct vtimes	start_vm99, end_vm99; \
short		MIFindModIdx(); \
\
if (!modStat99) { \
    modNum99  = MIFindModIdx(name99, FALSE); \
    modStat99 = &MIModStat[modNum99]; \
    debugFlag = &(modStat99->debugFlag); \
} \
if (MITraceBackFlag) \
then MITraceBack[MINumTraceBack++] = modNum99; \
else MICurrentMod = modNum99; \
if (modStat99->timeFlag) \
    vtimes(&start_vm99, (struct vtimes *)NULL); \
if (modStat99->traceFlag && modStat99->numTime<MIModThreshold) \
    printf("--Start %s\n", modStat99->name); \
if (modStat99->stopFlag) { \
    MI_BEEP() \
    MIInitEnv(); \
}

/* ............................... End of macro MI_MOD_START */

/* Use this macro to end any timed subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_END() \
\
if (modStat99->timeFlag) { \
    vtimes(&end_vm99, (struct vtimes *)NULL); \
    modStat99->cTime += end_vm99.vm_utime-start_vm99.vm_utime+ \
			end_vm99.vm_stime-start_vm99.vm_stime; \
} \
if (modStat99->traceFlag && modStat99->numTime<MIModThreshold) \
    printf("--End   %s\n", modStat99->name); \
modStat99->numTime++; \
if (MITraceBackFlag) \
    MINumTraceBack--; \
} /* ............................. End of macro MI_MOD_END */

#else /* MI_DEBUG_INLINE */
#ifdef MI_DEBUG
/* Use this macro to time any subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_START(name99) \
\
/* String	name99; */ \
{ \
static short	modNum99 = -1; \
static short	*debugFlag; \
struct vtimes	start_vm99; \
\
MIModStart(name99, &modNum99, &debugFlag, &start_vm99);
/* ............................... End of macro MI_MOD_START */

/* Use this macro to end any timed subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_END() \
\
MIModEnd(modNum99, &start_vm99); \
} /* ............................. End of macro MI_MOD_END */
#else /* MI_DEBUG */

#define MI_MOD_START(name)
#define MI_MOD_END() NOTHING

#endif /* MI_DEBUG */
#endif /* MI_DEBUG_INLINE */
/* Variables and macros to implement a FIFO queue.
----------------------------------------------------------------------------*/
int		MIQueuePtr;	/* Queue pointer */
int		MIQueueMax;	/* Maximum depth of queue */
int		MIQueueMaxM1;	/* Maximum depth of queue - 1 */
int		MIQueueNum;	/* Number of pointers queued */
RefAny		*MIQueue;	/* Pointer queue */

/* Allocate an n deep pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE_ALLOC(n) \
\
/* int		n;	 * Size of queue desired */ \
{ \
MIQueue = ALLOC_ARRAY(n, RefAny); \
MIQueueMax   = n; \
MIQueueMaxM1 = n-1; \
MIQueuePtr   = 0; \
MIQueueNum   = 0; \
} /* ............................. End of macro MI_QUEUE_ALLOC */

/* Free pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE_FREE() \
\
FREE_ARRAY(MIQueueMax, MIQueue, RefAny)
/* ............................... End of macro MI_QUEUE_FREE */

/* Put p onto pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE(p) \
\
/* int		p;	 * Pointer to be pushed onto queue */ \
{ \
URANGE(MIQueueNum,MIQueueMax) \
MIQueueNum++; \
MIQueue[MIQueuePtr] = (RefAny)(p); \
if (MIQueuePtr<MIQueueMaxM1) \
     MIQueuePtr++; \
else MIQueuePtr = 0; \
} /* ............................. End of macro MI_QUEUE */

/* Pop p from pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_DEQUEUE(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be dequeued */ \
{ \
int		lastQ; \
\
DASSERT(MIQueueNum) \
if (MIQueuePtr>=MIQueueNum) \
     lastQ = MIQueuePtr-MIQueueNum; \
else lastQ = MIQueuePtr-MIQueueNum+MIQueueMax; \
(p) = (type)MIQueue[lastQ]; \
MIQueueNum--; \
} /* ............................. End of macro MI_DEQUEUE */
/* Variables and macros to implement a push down stack.
----------------------------------------------------------------------------*/
int		MIStackPtr;	/* Stack pointer */
int		MIStackMax;	/* Maximum depth of stack */
RefAny		*MIStack;	/* Pointer stack */

/* Allocate an n deep pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_ALLOC(n) \
\
/* int		n;	 * Size of stack desired */ \
{ \
MIStack = ALLOC_ARRAY(n, RefAny); \
MIStackMax = n; \
MIStackPtr = 0; \
} /* ............................. End of macro MI_STACK_ALLOC */

/* Free pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_FREE() \
\
FREE_ARRAY(MIStackMax, MIStack, RefAny)
/* ............................... End of macro MI_STACK_FREE */

/* Push p onto pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_PUSH(p) \
\
/* int		p;	 * Pointer to be pushed onto stack */ \
{ \
URANGE(MIStackPtr,MIStackMax) \
MIStack[MIStackPtr++] = (RefAny)(p); \
} /* ............................. End of macro MI_STACK_PUSH */

/* Pop p from pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_POP(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be popped from stack */ \
{ \
DASSERT(MIStackPtr) \
(p) = (type)MIStack[--MIStackPtr]; \
} /* ............................. End of macro MI_STACK_POP */

/* Get last p from pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_LAST(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be popped from stack */ \
(p) = (type)MIStack[MIStackPtr];
/* ............................... End of macro MI_STACK_LAST */

/* ............................... End of include miutil.h */
