#include "electra.h"

/* declare the move generation functions here				*/
extern int genNewEdge();
extern int genNewTermloc();
extern int resizeNet();

/* declare the move acceptance functions here				*/
extern int accMove();

#define NUMFN 7
/* the array of generation functions					*/
int (*genFn[NUMFN])() = {
	genNewEdge, 
	genNewEdge,
	genNewEdge,
	genNewTermloc, 
	genNewTermloc, 
	genNewTermloc,
	resizeNet};

int (*accFn[NUMFN])() = {
	accMove, 
	accMove, 
	accMove, 
	accMove, 
	accMove, 
	accMove, 
	accMove };
double Temp;
double Cost;
double DelC;

anneal( verbose )
int verbose;
{

    s_net *net;
    int numGen;		/* number of states generated at each temp 	*/
    int numAcc;		/* number of states accepted at each temp 	*/
    int numRej;		/* number of states rejected at each temp 	*/
    int index;		/* index into array of gen and acc functions	*/
    double initialTemp(),updateTemp();
    double costFn();
    double delCostFn();
    long random();

    if( NumToRoute == 0 )return;
    Temp = initialTemp(); 
    Cost = costFn();
    if( verbose ){
	fprintf(stderr,"Starting annealing - initial cost = %1.3g\n",Cost);
	fprintf(stderr,"                     initial temp = %1.2g\n",Temp);
    }
    while ( !stopCriterion()) {
	numGen = 0;
	numAcc = 0;
	numRej = 0;
	while ( !innerLoopCriterion( numGen, numAcc )){

	    /* generate a new move					*/
	    index = randomInt( 0, NUMFN - 1 );
	    net = NetsToRoute[ randomInt( 0, NumToRoute - 1 ) ];
	    (*genFn[ index ])( net );
	    numGen++ ;
	    /* accept or reject the move				*/
	    DelC = delCostFn( net );
	    if (( Move != INVALID )&&
		((DelC <= 0)||
		(( Temp > 0 )&&
		 ((double)random())/(double)2147483647 < exp( -DelC / Temp )))){
		(*accFn[ index ])( net );
		Cost += DelC;
#ifdef PARANOID
		if( Cost != costFn() ){
		    fprintf(stderr,"Cost function error: numgen = %d\n",numGen);
		    Cost = costFn();
		}
#endif
		numAcc++;
	    } else {
		numRej++;
	    }
	}
	if( verbose ){
	    fprintf(stderr,
		    "At T = %1.2g accumulated cost = %1.3g real cost = %1.3g\n",
		    Temp, Cost, costFn());
	    fprintf( stderr,
		     "%d states accepted %d states rejected\n",
		     numAcc, numRej); 
	}
	/*
	if( numAcc / numGen >= .5 )Temp *= 0.8;
	else 			   Temp *= 0.9;
	*/
	Temp *= 0.9;
    }
    if( verbose ){
	fprintf(stderr,"Finished annealing - final cost = %1.3g\n", Cost);
        printCosts();
    }
    return;
} /* end of anneal */

/*
 * innerLoopCriterion()
 *
 * Returns true if the annealing should break out of the inner loop and
 * update the temperature, FALSE if the inner loop should continue
 */
innerLoopCriterion( numGen, numAcc )
int numGen; 		/* number of states which have been generated 	*/
{
    static int maxStates;	/* number of states at a given temp 	*/
    static int maxAcc;		/* number of accepted states at a term  */

    /* the temp has changed if numGen == 0 - reset MaxStates	*/
    if( numGen == 0 ){
	maxStates = numStates();
	maxAcc = (int)(maxStates * .6) ;
    }
    if ( numAcc >= maxAcc ) return( TRUE );
    if ( numGen < maxStates ) return(FALSE);
    return(TRUE);
} /* end of innerLoopCriterion */

/*
 * stoppingCriterion()
 *
 * Returns TRUE if the annealing is finished and should break out of the 
 * outer loop, FALSE if it should continue
 *
 * Presently returns TRUE if the cost hasn't changed for four temperatures.
 */
stopCriterion()
{
    static double cost1 = -1.0;
    static double cost2 = -2.0;
    static double cost3 = -3.0;
    if(( Cost == 0)||((Cost == cost1)&&(Cost == cost2)&&(Cost == cost3))){
	return(TRUE);
    } 
    cost3 = cost2;
    cost2 = cost1;
    cost1 = Cost;
    return(FALSE);
} /* end of stopCriterion */

/*
 * randomInt
 *
 * return a pseudo random int between lower and upper
 */
int
randomInt( lower, upper )
int lower;
int upper;
{
    if( lower == upper )return( lower );
    return(( random()%( upper + 1 - lower )) + lower );
}
/*
 * initialTemp()
 *
 * calculates the initial temperature
 */
double
initialTemp()
{

    return( 20 * ( double )NumToRoute );
}
int 
numStates()
{
    return( 200 * NumToRoute); 
}
