#include "electra.h"
int hCost;		/* the vertical netlength cost			*/
int vCost;		/* the horizontal netlength cost		*/
int sCost;		/* the cost of stretching signals		*/
int cCost;		/* the cost of violating bestTerm constraints 	*/
int tCost;		/* the cost of bad POLYM2 term locations	*/
int m1Cost;		/* the cost of M1 overlap - 3 vertical tracks 	*/
int m2Cost;		/* the cost of M2 overlap - 2 horizontal tracks	*/
double hFactor = 1;	/* relative weight of hCost			*/
double vFactor = .5;	/* relative weight of vCost			*/
double sFactor = 3.0;	/* relative weight of sCost			*/
double cFactor = 2.5;	/* relative weight of cCost			*/
double tFactor = 20;	/* relative weight of tCost			*/
double m1Factor = 400;	/* relative weight of m1Cost			*/
double m2Factor = 400;	/* relative weight of m2Cost			*/
int Perimeter = 0;
int HalfPerimeter = 0;
/*
 * costFn
 *
 * returns the total cost of a configuration
 */
double 
costFn()
{
    octPoint termloc;
    int edgeloc;
    s_net *net;
    double cost;
    int row, col;
    int tmpCost;
    int num;
    s_network *network;


    if( Perimeter == 0 ){
	HalfPerimeter = NumRows + NumCols;
	Perimeter = 2 * HalfPerimeter;
    }
    hCost = vCost = sCost = cCost = tCost = m1Cost = m2Cost = 0;
    for( num = NumToRoute - 1; num >= 0; num-- ){
	net = NetsToRoute[ num ];
	termloc = net->termloc;
	/* the netlength cost						*/

	switch( net->edge ){
	    case T_EDGE:
		vCost += NumRows - 1 - termloc.y;
		break;
	    case B_EDGE:
		vCost += termloc.y;
		break;
	    case L_EDGE:
		hCost += termloc.x;
		break;
	    case R_EDGE:
		hCost += NumCols - 1 - termloc.x;
		break;
	}
	/* the cost of stretching signals				*/
	sCost += net->signal->maxLeft - net->signal->left;
	sCost += net->signal->right - net->signal->minRight;

	/* the terminal constraint violation cost			*/
	if( net->status&CONSTRAINED ){
	    switch( net->edge ){
		case T_EDGE:
		    edgeloc = termloc.x + 1;
		    break;
		case B_EDGE:
		    edgeloc = HalfPerimeter + NumCols - termloc.x; 
		    break;
		case L_EDGE:
		    edgeloc = HalfPerimeter + NumCols + termloc.y + 1; 
		    break;
		case R_EDGE:
		    edgeloc = HalfPerimeter - termloc.y; 
		    break;
	    }
	    tmpCost = edgeloc - net->bestloc;
	    if( tmpCost < 0 ) tmpCost = -tmpCost;
	    if( tmpCost > HalfPerimeter ) cCost += Perimeter - tmpCost;
	    else                          cCost += tmpCost;
	}

	/* the terminal violation ( contacts on networks ) cost		*/
	network = Matrix[ termloc.y ][ termloc.x ].network;
	if(( network != NIL(s_network))&&
	   ( network != net->network )&&
	   ( termloc.y == network->firstTran ))tCost++;
    }
    tCost += NetworksWithBadTerms;

    /* calculate the overlap cost					*/

    for( row = 0; row < NumRows; row++ ){
	for( col = 0; col < NumCols; col++ ){
	    if(( num = Matrix[row][col].metal1Mask ) > 3 ) m1Cost += num - 3;
	    if(( num = Matrix[row][col].metal2Mask ) > 2 ) m2Cost += num - 2;
	}
    }

    /* scale the costs by the appropriae factors and add them together	*/
    cost = hCost * hFactor + vCost * vFactor  + cCost * cFactor + 
	   tCost * tFactor + m1Cost * m1Factor + m2Cost * m2Factor + 
	   sCost * sFactor;
    return( cost );
}

/*
 * delCostFn
 *
 * return the change in cost function due to changing a net
 */
double
delCostFn( net )
s_net *net;
{

    double cost;
    int row, col;
    int tmpCost;
    octPoint termloc;
    int edgeloc;
    s_network *network;


    sCost = hCost = vCost = cCost = tCost = m1Cost = m2Cost = 0;
    termloc = net->termloc;
    if(( NewLeft == net->signal->left )&&
       ( NewRight == net->signal->right )&&
       ( NewEdge == net->edge )&&
       ( NewTermloc.x == termloc.x )&&
       ( NewTermloc.y == termloc.y ))return( 0.0 );


    if( Move == RESIZE ){
	row = net->signal->row;
	sCost = net->signal->left - NewLeft + NewRight - net->signal->right;
	if( net->signal->numTerms > 0 ){
	    if(( net->edge&L_EDGE ) == 0 ){
		for( col = NewLeft; col < net->signal->left; col++ ){
		    if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
		}
		for( col = net->signal->left; col < NewLeft; col++ ){
		    if( Matrix[row][col].metal2Mask > 2 )m2Cost--;
		}
	    }
	    if(( net->edge&R_EDGE ) == 0 ){
		for( col = NewRight + 1; col <= net->signal->right; col++ ){
		    if( Matrix[row][col].metal2Mask > 2 )m2Cost--;
		}
		for( col = net->signal->right + 1; col <= NewRight; col++ ){
		    if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
		}
	    }
	} else if( net->edge&L_EDGE ){
	    for( col = termloc.x + 1; col <= NewTermloc.x; col++ ){
		if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
	    }
	    for( col = NewTermloc.x + 1; col <= termloc.x; col++ ){
		if( Matrix[row][col].metal2Mask > 2 )m2Cost--;
	    }
	} else if( net->edge&R_EDGE ){
	    for( col = termloc.x; col < NewTermloc.x; col++ ){
		if( Matrix[row][col].metal2Mask > 2 )m2Cost--;
	    }
	    for( col = NewTermloc.x; col < termloc.x; col++ ){
		if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
	    }
	} 
	if( net->edge&L_EDGE ){
	    hCost = NewTermloc.x - termloc.x;
	} else if( net->edge&R_EDGE ){
	    hCost = termloc.x - NewTermloc.x;
	} 
	if( termloc.x != NewTermloc.x ){
	    if((( net->network == NIL( s_network ))||
		( termloc.x != net->network->col ))&&
	       (( net->edge&TB_EDGE )||
		( IS_TRAN( termloc.y, termloc.x ) == FALSE ))){
		if( Matrix[termloc.y][termloc.x].metal1Mask > 3 )m1Cost--;
	    }
	    if((( net->network == NIL( s_network ))||
		( NewTermloc.x != net->network->col ))&&
	       (( net->edge&TB_EDGE )||
		( IS_TRAN( termloc.y, NewTermloc.x ) == FALSE ))){
		if( Matrix[termloc.y][NewTermloc.x].metal1Mask >= 3 )m1Cost++;
	    }
	}
    } else {
	/* cost evaluation for NEWEDGE and NEWTERMLOC moves		*/
	/* calculate the hCost, vCost and overlap costs			*/
	switch( net->edge ){
	    case T_EDGE:
		vCost -= NumRows - 1 - termloc.y;
		col = termloc.x;
		for( row = termloc.y + 1; row < NumRows; row++ ){
		    if( Matrix[row][col].metal1Mask-- > 3 )m1Cost--;
		}
		break;
	    case B_EDGE:
		vCost -= termloc.y;
		col = termloc.x;
		for( row = termloc.y - 1; row >= 0; row-- ){
		    if( Matrix[row][col].metal1Mask-- > 3 )m1Cost--;
		}
		break;
	    case L_EDGE:
		hCost -= termloc.x;
		row = termloc.y;
		for( col = termloc.x - 1; col >= 0; col-- ){
		    if( Matrix[row][col].metal2Mask-- > 2 )m2Cost--;
		}
		break;
	    case R_EDGE:
		hCost -= NumCols - 1 - termloc.x;
		row = termloc.y;
		for( col = termloc.x + 1; col < NumCols; col++ ){
		    if( Matrix[row][col].metal2Mask-- > 2 )m2Cost--;
		}
		break;
	}
	if((( net->network == NIL( s_network ))||
	    ( termloc.x != net->network->col ))&&
	   (( net->edge&TB_EDGE )||
	    ( IS_TRAN( termloc.y, termloc.x ) == FALSE ))){
	   if( Matrix[ termloc.y ][ termloc.x ].metal1Mask-- > 3 )m1Cost--;
	}
	if(( net->edge&LR_EDGE )&&( net->signal->numTerms == 0 )){
	    if( Matrix[ termloc.y ][ termloc.x ].metal2Mask-- > 2 )m2Cost--;
	}
	switch( NewEdge ){
	    case T_EDGE:
		vCost += NumRows - 1 - NewTermloc.y;
		col = NewTermloc.x;
		for( row = NewTermloc.y + 1; row < NumRows; row++ ){
		    if( Matrix[row][col].metal1Mask >= 3 )m1Cost++;
		}
		break;
	    case B_EDGE:
		vCost += NewTermloc.y;
		col = NewTermloc.x;
		for( row = NewTermloc.y - 1; row >= 0; row-- ){
		    if( Matrix[row][col].metal1Mask >= 3 )m1Cost++;
		}
		break;
	    case L_EDGE:
		hCost += NewTermloc.x;
		row = NewTermloc.y;
		for( col = NewTermloc.x - 1; col >= 0; col-- ){
		    if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
		}
		break;
	    case R_EDGE:
		hCost += NumCols - 1 - NewTermloc.x;
		row = NewTermloc.y;
		for( col = NewTermloc.x + 1; col < NumCols; col++ ){
		    if( Matrix[row][col].metal2Mask >= 2 )m2Cost++;
		}
		break;
	}
	if((( net->network == NIL( s_network ))||
	    ( NewTermloc.x != net->network->col ))&&
	   (( NewEdge&TB_EDGE )||
	    ( IS_TRAN( NewTermloc.y, NewTermloc.x ) == FALSE ))){
	   if( Matrix[ NewTermloc.y ][ NewTermloc.x ].metal1Mask >= 3 )m1Cost++;
	}
	if(( NewEdge&LR_EDGE )&&( net->signal->numTerms == 0 )){
	   if( Matrix[ NewTermloc.y ][ NewTermloc.x ].metal2Mask >= 2 )m2Cost++;
	}

	switch( net->edge ){
	    case T_EDGE:
		col = termloc.x;
		for( row = termloc.y + 1; row < NumRows; row++ ){
		    Matrix[row][col].metal1Mask++;
		}
		break;
	    case B_EDGE:
		col = termloc.x;
		for( row = termloc.y - 1; row >= 0; row-- ){
		    Matrix[row][col].metal1Mask++;
		}
		break;
	    case L_EDGE:
		row = termloc.y;
		for( col = termloc.x - 1; col >= 0; col-- ){
		    Matrix[row][col].metal2Mask++;
		}
		break;
	    case R_EDGE:
		row = termloc.y;
		for( col = termloc.x + 1; col < NumCols; col++ ){
		    Matrix[row][col].metal2Mask++;
		}
		break;
	}
	if((( net->network == NIL( s_network ))||
	    ( termloc.x != net->network->col ))&&
	   (( net->edge&TB_EDGE )||
	    ( IS_TRAN( termloc.y, termloc.x ) == FALSE ))){
	   Matrix[ termloc.y ][ termloc.x ].metal1Mask++;
	}
	if(( net->edge&LR_EDGE )&&( net->signal->numTerms == 0 )){
	    Matrix[ termloc.y ][ termloc.x ].metal2Mask++;
	}
    }
    /* the constraint violation cost 				*/
    if(( net->status&CONSTRAINED )&&
       (( net->edge != NewEdge )|| 
	( termloc.x != NewTermloc.x )||
	( termloc.y != NewTermloc.y ))){
	switch( net->edge ){
	    case T_EDGE:
		edgeloc = termloc.x + 1;
		break;
	    case B_EDGE:
		edgeloc = HalfPerimeter + NumCols - termloc.x; 
		break;
	    case L_EDGE:
		edgeloc = HalfPerimeter + NumCols + termloc.y + 1; 
		break;
	    case R_EDGE:
		edgeloc = HalfPerimeter - termloc.y; 
		break;
	}
	tmpCost = edgeloc - net->bestloc;
	if( tmpCost < 0 ) tmpCost = -tmpCost;
	if( tmpCost > HalfPerimeter ) cCost -= ( Perimeter - tmpCost );
	else                          cCost -= tmpCost;
	switch( NewEdge ){
	    case T_EDGE:
		edgeloc = NewTermloc.x + 1;
		break;
	    case B_EDGE:
		edgeloc = HalfPerimeter + NumCols - NewTermloc.x; 
		break;
	    case L_EDGE:
		edgeloc = HalfPerimeter + NumCols + NewTermloc.y + 1; 
		break;
	    case R_EDGE:
		edgeloc = HalfPerimeter - NewTermloc.y; 
		break;
	}
	tmpCost = edgeloc - net->bestloc;
	if( tmpCost < 0 ) tmpCost = -tmpCost;
	if( tmpCost > HalfPerimeter ) cCost += ( Perimeter - tmpCost );
	else                          cCost += tmpCost;
    }
    if(( termloc.x != NewTermloc.x )||( termloc.y != NewTermloc.y )){
	network = Matrix[ termloc.y ][ termloc.x ].network;
	if(( network != NIL( s_network ))&&(network != net->network )){
	    if( termloc.y == network->firstTran )tCost--;
	    if(( network->numBadTerms == 1 )&&
	       ((( termloc.y <= network->lastTran )&&
		 ( termloc.y >= network->firstTran ))||
		(( termloc.y >= network->lastTran )&&
		 ( termloc.y <= network->firstTran )))&&
	       ( IS_TRAN( termloc.y, termloc.x )))tCost--;
	}
	network = Matrix[ NewTermloc.y ][ NewTermloc.x ].network;
	if(( network != NIL( s_network ))&&(network != net->network )){
	    if( NewTermloc.y == network->firstTran )tCost++;
	    if(( network->numBadTerms == 0 )&&
	       ((( NewTermloc.y <= network->lastTran )&&
		 ( NewTermloc.y >= network->firstTran ))||
		(( NewTermloc.y >= network->lastTran )&&
		 ( NewTermloc.y <= network->firstTran )))&&
	       ( IS_TRAN( NewTermloc.y, NewTermloc.x )))tCost++;
	}
    }
    /* scale the costs by the appropriate factors and add them together	*/
    cost = hCost * hFactor + vCost * vFactor  + cCost * cFactor + 
	   tCost * tFactor + m1Cost * m1Factor + m2Cost * m2Factor + 
	   sCost * sFactor;
    return( cost );
}
/*
 * printCosts
 *
 */
printCosts()
{
    fprintf(stderr, "The cost factors are:\n");
    fprintf(stderr, "\thorizontal net length cost = %1.4g\n", hCost * hFactor );
    fprintf(stderr, "\tvertical net length cost = %1.4g\n", vCost * vFactor );
    fprintf(stderr, "\tsignal length cost = %1.4g\n", sCost * sFactor );
    fprintf(stderr, "\tconstraint cost = %1.4g\n", cCost * cFactor );
    fprintf(stderr, "\tterminal cost = %1.4g\n", tCost * tFactor );
    fprintf(stderr, "\tmetal1 overlap cost = %1.4g\n", m1Cost * m1Factor );
    fprintf(stderr, "\tmetal2 overlap cost = %1.4g\n", m2Cost * m2Factor );
}
