#include "electra.h"
/*
 * createInterface
 *
 * creates the interface facet 
 * creates the DIRECTION, TERMTYPE, and PEAKCURRENT properties on the 
 * formal terms
 */
createInterface()
{
    octObject oInterface;
    octObject oTerm;
    octObject oProp;
    s_net *net;

    oInterface = oFacet;
    oInterface.contents.facet.facet = "interface";
    oInterface.contents.facet.mode = "w";
    if( octOpenFacet( &oInterface ) < OCT_OK ){
	(void) sprintf( ErrBuffer, "Could not open interface facet %s:%s:%s:",
			    oInterface.contents.facet.cell,
			    oInterface.contents.facet.view,
			    oInterface.contents.facet.facet );
	octError( ErrBuffer );
	fprintf(stderr,"\n No interface facet will be created\n");
	return;
    }
    oTerm.type = OCT_TERM;
    oProp.type = OCT_PROP;
    for( net = Netlist; net != NIL( s_net ); net = net->next ){
	if(( net->status&EXTERNAL ) == 0 )continue;
	oTerm.contents.term.name = net->termName;
	octGetByName( &oInterface, &oTerm );

	oProp.contents.prop.type = OCT_STRING;
	oProp.contents.prop.name = "TERMTYPE";
	oProp.contents.prop.value.string = "SIGNAL";
	octCreate( &oTerm, &oProp );

	oProp.contents.prop.name = "DIRECTION";
	if( net->network == NIL( s_network )){
	    oProp.contents.prop.value.string = "INPUT";
	} else {
	    oProp.contents.prop.value.string = "OUTPUT";
	}
	octCreate( &oTerm, &oProp );
    }
    oTerm.contents.term.name = "GND";
    octGetByName( &oInterface, &oTerm );

    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.name = "TERMTYPE";
    oProp.contents.prop.value.string = "GROUND";
    octCreate( &oTerm, &oProp );

    oProp.contents.prop.name = "DIRECTION";
    oProp.contents.prop.value.string = "INPUT";
    octCreate( &oTerm, &oProp );

    oProp.contents.prop.type = OCT_REAL;
    oProp.contents.prop.name = "PEAKCURRENT";
    oProp.contents.prop.value.real = 1e-3;
    octCreate( &oTerm, &oProp );

    oTerm.contents.term.name = "Vdd";
    octGetByName( &oInterface, &oTerm );

    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.name = "TERMTYPE";
    oProp.contents.prop.value.string = "SUPPLY";
    octCreate( &oTerm, &oProp );

    oProp.contents.prop.name = "DIRECTION";
    oProp.contents.prop.value.string = "INPUT";
    octCreate( &oTerm, &oProp );

    oProp.contents.prop.type = OCT_REAL;
    oProp.contents.prop.name = "PEAKCURRENT";
    oProp.contents.prop.value.real = 1e-3;
    octCreate( &oTerm, &oProp );

    octCloseFacet( &oInterface );
}
