#include "electra.h"

char ErrBuffer[512];		/* a buffer for error messages		*/
s_options Options;

int NumRows, NumCols;		/* number of rows and cols in Matrix	*/
int TotalRows, TotalCols;	/* number of rows and columns in FullArray */

s_matrix **Matrix;		/* the personality matrix		*/
s_charEntry charTable[128];	/* maps charactors to instance names 	*/

s_network **NetworkArray;	/* the networks in the array ( by col )	*/
s_signal **SignalArray;		/* the signals in the array ( by row )	*/
s_net *Netlist;			/* list of nets 			*/

octObject oFacet;		/* the output facet			*/
octObject oPalette;		/* the palette of transistors		*/
octObject oInstBag;		/* the INSTNACES bag			*/
octObject oConnBag;		/* the CONNECTORS bag			*/

mkId FullArray;			/* the top level layout array		*/
mkId NTrans, PTrans;		/* NMOS and PMOS tracks 		*/
mkId NandOutput, NorOutput;	/* nand and nor output tracks		*/
mkId VddArray, VssArray;	/* VSS and VDD tracks			*/

int NumToRoute;			/* number of nets to be routed 		*/
s_net **NetsToRoute;		/* array of nets to be routed		*/
int Move;			/* the current move during annealing	*/
unsigned NewEdge;		/* the new edge 			*/
octPoint NewTermloc;		/* the new term location		*/
int NewLeft;			/* the new left edge of a signal	*/
int NewRight;			/* the new right edge of a signal	*/
int NetworksWithBadTerms;	/* number of networks with POLYM2 terms */
				/* in their output column		*/
unsigned *TEdgeMask;		/* mask of term locations on top edge 	*/
unsigned *BEdgeMask;		/* mask of term locations on bottom edge */

main( argc , argv )
int argc;
char **argv;
{
    if( Debug > 0 )printVersion( stderr, argv[ 0 ] );
    octBegin();

    setDefaults();
    parseComLine( argc , argv );

    fprintf(stderr,"Parsing input and tech files\n");
    parseTech();
    parseIn();

    fprintf(stderr,"Building connectivity graphs\n");
    finishMatrix();
    if ( Debug > 0 ){
	checkSignals();
	checkNetworks();
	printStats();
    }
    netlist();
    if( FloorPlannerPrep == 1 ){
	fprintf(stderr,"Preparing the cell for floorplanning\n");
	floorPlanPrep();
	exit(0);
    }
    if( Size ){
	fprintf(stderr,"Sizing the transistors\n");
	sizeTrans();
    }

    fprintf(stderr,"Reading terminal constraints\n");
    readTermConstraints();

    newFacet();
    if( Options.localPaletteName != NIL( char ))
	openLocalPalette( Options.localPaletteName );
    else 
	openPalette( Options.paletteName );

    fprintf(stderr,"Specifying the architecture\n");
    specArch();

    fprintf(stderr,"Creating instances\n");
    makeSignals();
    makeNetworks();

    if( Wire == TRUE ){
	fprintf(stderr,"Routing the array\n");
	routeTerms();
    }

    if( FloorPlannerPrep == 2 ){
	fprintf(stderr,"Preparing the cell for floorplanning\n");
	floorPlanPrep();
	exit(0);
    }

    powerTerms();

    if( Wire == TRUE ){
	fprintf(stderr,"Wiring the array\n");
	assignSignalTracks();
	mkReadRules( FullArray, &oPalette );
	networkTerms();
	signalTerms();
	wireNetworks();
	wireSignals();
    }

    fprintf(stderr,"Placing the array\n");
    mkPlaceArray( FullArray );

    fprintf(stderr,"Writing the cell\n");
    octFlushFacet( &oFacet );

    fprintf(stderr,"Creating the interface facet\n");
    createInterface();
    octEnd();

    fprintf(stderr,"GoodBye\n");
    exit( 0 );
}
