/*
 * electra.h
 *
 * the header file for the programs electra, electraRules and electraTrans
 */
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "mkarray.h"
#include "utility.h"

/* general stuff */
#define TRUE 1
#define FALSE 0

/* default output facet cell, view, facet, and version */

#define DEFAULT_CELLLIB "$OCTTOOLS/tech/scmos/gem"
#define DEFAULT_RULEFACET "rules:symbolic"
#define DEFAULT_OUT_VIEW "unspaced"
#define DEFAULT_IN_VIEW "optlogic"
#define DEFAULT_VERSION OCT_CURRENT_VERSION
/* default connector and rule names */

#define POWERTERM "m1m1.3x12"
#define POLYTERM "popo.2x2"
#define POLYM1TERM "copo.4x4"
#define M1TERM "m1m1.3x3"
#define M1M2TERM "com2.4x4"
#define M2TERM "m2m2.3x3"
#define M2FTERM "m2m2.4x4"
#define M1FTERM "m1m1.4x4"
#define POLYM2TERM "pom2.4x9"

#define NAND_RULE_NAME "NAND_RULE"
#define NOR_RULE_NAME "NOR_RULE"
#define SIGNAL_RULE_NAME "SIGNAL_RULE"
#define POWERTRACK_RULE_NAME "POWERTRACK_RULE"
#define POWER_RULE_NAME "POWER_RULE"
#define FIRSTTRAN_RULE_NAME "FIRSTTRAN_RULE"
#define OUTPUT_RULE_NAME "OUTPUT_RULE"

/* default transitor terminal names used in electraRules and electraTrans */
#define GATE "GATE"
#define SOURCE "SOURCE"
#define DRAIN "DRAIN"

/* defined constants */

/* cmos transistor types */
enum fetType { NFET, PFET };

/* instance types		*/
#define CONNECTOR 0
#define INSTANCE 1

/* types of signals and networks */
#define UNDEFINED -1
#define NAND 1        	/* networks contains an 's'*/
#define NOR 2       	/* networks contains an 'p'*/

/* values of matrix.type				*/
#define DOT 1
#define TRAN 2
#define TERM 3

/* valuse of net->status and net->edge */
#define EXTERNAL (unsigned)1<<0
#define R_EDGE (unsigned)1<<1
#define L_EDGE (unsigned)1<<2
#define T_EDGE (unsigned)1<<3
#define B_EDGE (unsigned)1<<4
#define LR_EDGE (L_EDGE|R_EDGE)
#define TB_EDGE (T_EDGE|B_EDGE)
#define ANY_EDGE (R_EDGE|L_EDGE|T_EDGE|B_EDGE)
#define ROUTED (unsigned)1<<5
#define UPPER (unsigned)1<<6
#define LOWER (unsigned)1<<7
#define LEFT (unsigned)1<<8
#define RIGHT (unsigned)1<<9
#define CENTER (unsigned)1<<10
#define CONSTRAINED (unsigned)1<<11

/* types of moves ( extern int Move ) done in the annealing	*/
#define INVALID 0 	/* bad move - never accepted 		*/
#define NEWEDGE 1	/* generate a new edge and termloc	*/
#define NEWTERMLOC 2	/* generate a new termloc		*/
#define RESIZE 3	/* resize the nets signal		*/

/* variations on the basic transistor */
#define NO_CONTS (unsigned)0
#define SOURCE_CONTS (unsigned)1<<0
#define DRAIN_CONTS  (unsigned)1<<1
#define SKIP_CONTS   (unsigned)1<<2
#define SOURCE_SPEC 's'
#define DRAIN_SPEC 'd'
#define SKIP_SPEC 'k'

/* types of status */
#define PLACED 1

/* whether a tran or dot */
#define IS_TRAN(row,col) \
	(( Matrix[row][col].letter != '.' )&&\
	 ( Matrix[row][col].letter != 's' )&&\
	 ( Matrix[row][col].letter != 'p' ))

#define IS_TERM(row,col) \
	(( Matrix[row][col].letter == 's' )||\
	 ( Matrix[row][col].letter == 'p' ))

/* data types */

typedef struct matrix {
    int type;
    char letter;
    struct network *network;
    struct signal *signal;
    unsigned metal1Mask;
    unsigned metal2Mask;
} s_matrix;

typedef struct network {
    int type;
    int col;
    int top;
    int bottom;
    int firstTran;
    int lastTran;
    int numBadTerms;
    struct network *next;
    struct network *prev;
    struct net *net;
} s_network;

typedef struct signal {
    int row;
    int left;
    int right;
    int maxLeft;
    int minRight;
    struct net *net;
    struct signal *next;
    struct signal *prev;
    int numRuns;
    struct polyrun *polyrun;
    int numTerms;
    int numTrans;
    struct termList *term;
} s_signal;

typedef struct net {
    char *name;
    char *termName;
    unsigned status;
    unsigned edge;
    octPoint termloc;
    int bestloc;
    octId netId;
    struct net *next;
    struct network *network;
    struct signal *signal;
} s_net;

typedef struct charEntry {
    int set;
    char *nmos;
    char *pmos;
} s_charEntry;

typedef struct tBuffEntry {
    int net;
    int x;
    int y;
    int layer;
} s_tBuffEntry;

typedef struct polyrun {
    int left;
    int right;
} s_polyrun;

typedef struct termList {
    octId objectId;
    int col;
} s_termList;

typedef struct options {
    int debug;
    char *technology;
    char *paletteName;
    char *localPaletteName;
} s_options;

/* command line variables defined and set in parseComLine.c		*/

extern char *Celllib;
extern int Debug;
extern int FloorPlannerPrep;
extern int PrintGraph;
extern char *InfileName;
extern int NoTermConstraints;
extern int PolyTerms;
extern char *RuleFacetName;
extern int Size;
extern char *TechFileName;
extern int Wire;
extern char *OutFacetName;
extern char *InFacetName;

/* global variables defined in electra.c				*/

extern char ErrBuffer[];

extern int NumRows, NumCols;
extern int TotalRows, TotalCols;

extern struct matrix **Matrix;
extern struct charEntry charTable[];

extern struct network **NetworkArray;
extern struct signal **SignalArray;
extern struct net *Netlist;

extern octObject oFacet;
extern octObject oInstBag;
extern octObject oConnBag;

extern mkId FullArray;
extern mkId NTrans, PTrans;
extern mkId NandOutput, NorOutput;
extern mkId VddArray, VssArray;

extern int NumToRoute;
extern struct net **NetsToRoute;
extern int Move;
extern unsigned NewEdge;
extern octPoint NewTermloc;
extern int NewLeft;
extern int NewRight;
extern int NetworksWithBadTerms;
extern unsigned *TEdgeMask;
extern unsigned *BEdgeMask;

extern octId createPolyM2Term();
