#include "electra.h"
#define WIDTHFACTOR 800
#define HEIGHTFACTOR 300
/*
 * floorPlanPrep
 *
 * annotates the interfacet facet of InFacet to make the terminals 
 * floating and give an initial estimate of the area of the cell
 */
floorPlanPrep()
{
    octObject oInterface;
    octObject oContents;
    octObject oFloorPlan;
    octObject oProp;
    octObject oLayer;
    octObject oBox;
    octObject oTerm;
    octGenerator gTerm;
    s_net *net;

    ohUnpackDefaults( &oInterface, "a", ":logic:interface" );
    ohUnpackFacetName( &oInterface, InFacetName );
    if( octOpenFacet( &oInterface ) < OCT_OK ){
	fprintf(stderr,"Could not open %s: %s\n",
		       ohFormatName( &oInterface ),
		       octErrorString());
	exit(1);
    }
    ohUnpackDefaults( &oContents, "a", ":logic:contents" );
    ohUnpackFacetName( &oContents, InFacetName );
    if( octOpenFacet( &oContents ) < OCT_OK ){
	fprintf(stderr,"Could not open %s: %s\n",
		       ohFormatName( &oContents ),
		       octErrorString());
	exit(1);
    }
    /* specify the FLOORPLAN bag					*/
    oFloorPlan.type = OCT_BAG;
    oFloorPlan.contents.bag.name = "FLOORPLAN";
    octGetOrCreate( &oInterface, &oFloorPlan );

    oProp.type = OCT_PROP;
    oProp.contents.prop.name = "MODULE_GENERATOR";
    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.value.string = "gem";
    octGetOrCreate( &oFloorPlan, &oProp );

    oProp.contents.prop.name = "LEGAL_TRANSFORMATIONS";
    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.value.string = "THIS_ASPECT_RATIO";
    octGetOrCreate( &oFloorPlan, &oProp );
    octAttachOnce( &oInterface, &oProp );

    /* specify that terminals can float on all edges			*/
    oProp.type = OCT_PROP;
    oProp.contents.prop.name = "TERM_FLOATING_EDGES";
    oProp.contents.prop.type = OCT_STRING;
    if( FloorPlannerPrep == 1 ){
	oProp.contents.prop.value.string = "";
	octInitGenContents( &oInterface, OCT_TERM_MASK, &gTerm );
	while( octGenerate( &gTerm, &oTerm ) == OCT_OK ){
	    octCreateOrModify( &oTerm, &oProp );
	}
    } else {
	oTerm.type = OCT_TERM;
	for( net = Netlist; net != NIL( s_net ); net = net->next ){
	    if(( net->edge&ANY_EDGE ) == 0 )continue;
	    oTerm.contents.term.name = net->termName;
	    if( octGetByName( &oInterface, &oTerm ) < OCT_OK ){
		fprintf( stderr, "WARNING - Could not find term %s %s\n",
			 net->termName,
			"in the interface facet - it will not be constrained");
		continue;
	    }
	    switch( net->edge ){
		case T_EDGE:
		    if( net->termloc.x > .75 * NumRows ){
			oProp.contents.prop.value.string = "TOP RIGHT";
		    } else if( net->termloc.x < .25 * NumRows ){
			oProp.contents.prop.value.string = "TOP LEFT";
		    } else {
			oProp.contents.prop.value.string = "TOP";
		    }
		    break;
		case B_EDGE:
		    if( net->termloc.x > .75 * NumRows ){
			oProp.contents.prop.value.string = "BOTTOM RIGHT";
		    } else if( net->termloc.x < .25 * NumRows ){
			oProp.contents.prop.value.string = "BOTTOM LEFT";
		    } else {
			oProp.contents.prop.value.string = "BOTTOM";
		    }
		    break;
		case L_EDGE:
		    if( net->termloc.y > .75 * NumCols ){
			oProp.contents.prop.value.string = "LEFT TOP";
		    } else if( net->termloc.y < .25 * NumCols ){
			oProp.contents.prop.value.string = "LEFT BOTTOM";
		    } else {
			oProp.contents.prop.value.string = "LEFT";
		    }
		    break;
		case R_EDGE:
		    if( net->termloc.y > .75 * NumCols ){
			oProp.contents.prop.value.string = "RIGHT TOP";
		    } else if( net->termloc.y < .25 * NumCols ){
			oProp.contents.prop.value.string = "RIGHT BOTTOM";
		    } else {
			oProp.contents.prop.value.string = "RIGHT";
		    }
		    break;
	    }
	    octCreateOrModify( &oTerm, &oProp );
	}
    }

    /* an initial guess of the size of the module			*/
    oLayer.type = OCT_LAYER;
    oLayer.contents.layer.name = "MET1";

    oBox.type = OCT_BOX;
    oBox.contents.box.lowerLeft.x = oBox.contents.box.lowerLeft.y = 0;
    oBox.contents.box.upperRight.x = WIDTHFACTOR * NumCols;
    oBox.contents.box.upperRight.y = HEIGHTFACTOR * NumRows;

    octGetOrCreate( &oInterface, &oLayer );
    octCreate( &oLayer, &oBox );

    octGetOrCreate( &oContents, &oLayer );
    octCreate( &oLayer, &oBox );

    octCloseFacet( &oInterface );
    octCloseFacet( &oContents );
}
