#include "electra.h"
#define ALWAYS_BREAK 2	/* empty space to always start a polyrun 	*/
#define SHOULD_BREAK 5	/* size of signal such that it shouldBreak	*/
#define CAN_BREAK 1	/* empty space to start a polyrun if shouldBreak*/
/*
 * wireSignals
 *
 * creates the implementation of the signals.
 *
 * The signal is scanned from left to right.  If more than ALWAYS_BREAK
 * spaces are encountered without a tran, the signal goes up to MET2.
 * The outputs of gates are added to the signal, either a POLY_M1 
 * contact or a M1M2 contact, depending if the signal is currently
 * running in poly or not.  
 *
 * This is broken into wireSignals and wireSignal to reduce the depth
 * of the routine.
 */
makeSignals()
{
    int row;
    s_signal *signal;

    /* call wireSignal on every signal 					*/
    for( row = NumRows - 1; row >= 0; row-- ){
	for( signal = SignalArray[ row ]; signal != NIL( s_signal );
	     signal = signal->next ){
	    makeSignal( signal );
	}
    }
}
/*
 * makeSignal
 *
 * wire a signal
 */
makeSignal( signal )
s_signal *signal;
{
    int row, col;
    int left, right;
    int hasM2 = FALSE;
    int firstTerm = TRUE;
    octObject *oInst;
    octBox area;
    octObject *newTerm();
    octObject oTerm, oNet;
    int shouldBreak = FALSE;
    int firstPolyM2Term;

    row = signal->row;

    right = signal->left;
    if( signal->right + 1 - signal->left >= SHOULD_BREAK )shouldBreak = TRUE;
    for( col = signal->left; col <= signal->right; col++ ){
	if(( IS_TRAN( row, col ))||( IS_TERM( row, col ))){
	    right = col;
	} else if(((shouldBreak)&&((col - right) >= CAN_BREAK ))||
		  (( col - right) >= ALWAYS_BREAK )){
	    hasM2 = TRUE;
	    break;
	}
    }
    if( hasM2 ){

	if( signal->numTrans == 0 )return;

	/* traverse the signal looking for poly runs			*/
	for( col = signal->left; col <= signal->right; col++ ){

	    /* the network output terminal is always a M1M2 term	*/
	    if( IS_TERM( row, col )){
		if( signal->numTerms++ == 0 ){
		    signal->term = ALLOC( s_termList, signal->numTerms );
		} else {
		    signal->term = REALLOC( s_termList, signal->term,
					      signal->numTerms );
		}
		oInst = newTerm( M1M2TERM );
	        if(( signal->net->status&EXTERNAL) != EXTERNAL ){
		    octGenFirstContent( oInst, OCT_TERM_MASK, &oTerm );
		    oNet.objectId = signal->net->netId;
		    octGetById( &oNet );
		    octAttach( &oNet, &oTerm );
		}
		area.lowerLeft.x = area.upperRight.x = col;
		area.lowerLeft.y = 3 * row + 2;
		area.upperRight.y = 3 * row + 4;
		if( Matrix[ row ][ col ].network->type == NAND ){
		    mkAttach( NandOutput, &area, oInst );
		} else {
		    mkAttach( NorOutput, &area, oInst );
		}
		signal->term[ signal->numTerms - 1].objectId = oInst->objectId;
		signal->term[ signal->numTerms - 1].col = col;
		firstTerm = FALSE;

	    /* this is the start of a poly run				*/
	    } else if( IS_TRAN( row, col )){
		if( signal->numRuns++ == 0 ){
		    signal->polyrun = ALLOC( s_polyrun, signal->numRuns );
		} else {
		    signal->polyrun = REALLOC( s_polyrun, signal->polyrun,
					       signal->numRuns );
		}
		/* find the other side of the run			*/
		left = right = col;
		while( ++col <= signal->right ){
		    if( IS_TRAN( row, col )){
			right = col;
		    } else if(((shouldBreak)&&((col - right) >= CAN_BREAK ))||
			      (( col - right) >= ALWAYS_BREAK )){
			break;
		    }
		}
		col = right;

		/* left and right are the boundary of the poly run	*/
		/* col is ready for the next iteration 			*/
		/* specify a POLYM2 term to connect the polyrun to M2	*/

		if( firstTerm ){
		    /* the POLYM2 term goes to the right of the polyrun	*/
		    firstTerm = FALSE;
		    newPolyM2Term( signal, ++right, FALSE );
		    firstPolyM2Term = right;
		    Matrix[ row ][ right ].metal1Mask = 1;
		    signal->polyrun[ 0 ].left = left;
		    signal->polyrun[ 0 ].right = right;
		} else if(( signal->numRuns == 2 )&&
			  ( left == firstPolyM2Term + 1 )){
		    signal->numRuns = 1;
		    signal->polyrun[ 0 ].right = right;
		} else {
		    /* the POLYM2 term goes to the left of the polyrun	*/
		    newPolyM2Term( signal, --left, FALSE );
		    Matrix[ row ][ left ].metal1Mask = 1;
		    signal->polyrun[ signal->numRuns - 1 ].left = left;
		    signal->polyrun[ signal->numRuns - 1 ].right = right;
		}
	    }
	} 
	/* set the m2 mask						*/
	for( col = signal->left; col <= signal->right; col++ ){
	    Matrix[row][col].metal2Mask = 1;
	}
    } else {

	/* place a POLYM1 term to join the network and signal if necesary */
	if( signal->numTrans != 0 ){
	    for( col = signal->left; col <= signal->right; col++ ){
		if( IS_TERM( row, col )){
		    oInst = newTerm( POLYM1TERM );
		    if(( signal->net->status&EXTERNAL) != EXTERNAL ){
			octGenFirstContent( oInst, OCT_TERM_MASK, &oTerm );
			oNet.objectId = signal->net->netId;
			octGetById( &oNet );
			octAttach( &oNet, &oTerm );
		    }
		    area.lowerLeft.x = area.upperRight.x = col;
		    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
		    if( Matrix[ row ][ col ].network->type == NAND ){
			mkAttach( NandOutput, &area, oInst );
		    } else {
			mkAttach( NorOutput, &area, oInst );
		    }
		    break;
		}
	    }
	}
    }
}
