#include "electra.h"
/*
 * finishMatrix
 *
 * at this point, every letter field in the Matrix is filled out and
 * the declared signals and networks are allocated and attached to the
 * first Matrix element which they contain.  This procedure allocates
 * any nondeclared signals and networks and attaches every non DOT 
 * element in the Matrix to a signal and a network
 *
 * this is the algorithm:
 *    for each row{
 *       find the first non DOT column
 *       allocate a signal if necessary;
 *       set lastsig to the signal
 *       while not at the end of the row {
 *           go to the next non DOT element
 *           if its signal == NULL set it to lastsig
 *           otherwise set lastsig to its signal
 *       }
 *    }
 *    do the analog for each column
 */
finishMatrix()
{
    int row, col;
    s_signal *lastsig;
    s_network *lastnetwork;
    s_signal *allocSignal();
    s_signal *signal;
    s_network *allocNetwork();
    s_network *network;
    for( row = NumRows - 1; row >= 0; row-- ){
	col = 0;
	if ( Matrix[row][col].signal == NULL ){
	    Matrix[row][col].signal = allocSignal( row );
	    Matrix[row][col].signal->row = row;
	    Matrix[row][col].signal->left = col;
	    Matrix[row][col].signal->right = col;
	} 
	lastsig = Matrix[ row ][ col ].signal;
	while ( ++col < NumCols ){
	    if ( Matrix[row][col].signal == NULL ){
		Matrix[row][col].signal = lastsig;
	    } else {
		lastsig = Matrix[row][col].signal;
	    }
	    Matrix[row][col].signal->right = col;
	}
	/* pruning the signals */
	for( signal = SignalArray[row]; signal != NIL(s_signal);
	     signal = signal->next ){
	    for( col = signal->left; col <= signal->right; col++ ){
		if( Matrix[row][col].type == DOT ){
		    Matrix[row][col].signal = NIL(s_signal);
		    signal->left++;
		} else {
		    break;
		}
	    }
	    for( col = signal->right; col >= signal->left; col-- ){
		if( Matrix[row][col].type == DOT ){
		    Matrix[row][col].signal = NIL(s_signal);
		    signal->right--;
		} else {
		    break;
		}
	    }
	    /* counting tbe number of transistors on the signal		*/
	    for( col = signal->left; col <= signal->right; col++ ){
		if( Matrix[row][col].type == TRAN )signal->numTrans++;
	    }
	}
    }
    /* the dual for the networks */
    for( col = NumCols - 1; col >= 0; col-- ){
	row = NumRows - 1;
	if ( Matrix[row][col].network == NULL ){
	    Matrix[row][col].network = allocNetwork( col );
	    Matrix[row][col].network->col = col;
	    Matrix[row][col].network->top = row;
	    Matrix[row][col].network->bottom = row;
	}
	lastnetwork = Matrix[row][col].network;
	while ( --row >= 0 ){
	    if ( Matrix[row][col].network == NULL ){
		Matrix[row][col].network = lastnetwork;
	    } else {
		networkType( lastnetwork );
		lastnetwork = Matrix[row][col].network;
	    }
	    Matrix[row][col].network->bottom = row;
	}
	networkType( lastnetwork );
	/* pruning the networks */
	for( network = NetworkArray[col]; network != NIL(s_network);
	     network = network->next ){
	    for( row = network->bottom; row <= network->top; row++ ){
		if( Matrix[row][col].type == DOT ){
		    Matrix[row][col].network = NIL(s_network);
		    network->bottom++;
		} else {
		    break;
		}
	    }
	    for( row = network->top; row >= network->bottom; row-- ){
		if( Matrix[row][col].type == DOT ){
		    Matrix[row][col].network = NIL(s_network);
		    network->top--;
		} else {
		    break;
		}
	    }
	    /* setting the network firstTran and lastTran fields	*/
	    /* this is dependant upon the network column because	*/
	    /* networks with odd columns are upside down		*/
	    for ( row = network->top ; row >= network->bottom ; row-- ){
		if( Matrix[row][col].type == TRAN ){
		    if( network->col%2 ) network->lastTran = row; 
		    else         network->firstTran = row;
		    break;
		}
	    }
	    for ( row = network->bottom ; row <= network->top ; row++ ){
		if( Matrix[row][col].type == TRAN ){
		    if( network->col%2 ) network->firstTran = row;
		    else         network->lastTran = row; 
		    break;
		}
	    }
	}
    }
}
/*
 * networkType
 * 
 * This procedure determines if the networks in a column are NAND or NOR
 */
networkType(network)
s_network *network;
{
    int row, col;
    col = network->col;
    for( row = network->top; row >= network->bottom ; row-- ){
	if ( Matrix[row][col].type == TERM ){
	    if ( Matrix[row][col].letter == 's' ){
		network->type = NAND;
	    } else {
		network->type = NOR;
	    }
	    return;
	}
    }
    networkError( network );
}
