#include "electra.h"
extern octObject oPalette;
extern octObject oFacet;
extern s_options Options;

/* to return different objects from these routines			*/
#define OBJ_ARRAY_SIZE 10
octObject ObjArray[ OBJ_ARRAY_SIZE ];
int ObjCount = 0;
newFacet()
{
    octObject oProp, *oNet;
    octObject *newNet();
    int row;
    s_signal *signal;

    /* to set the default output facet to the input facet		*/
    oFacet.contents.facet.view = DEFAULT_OUT_VIEW;

    ohUnpackDefaults( &oFacet, "w", ":logic:contents" );
    ohUnpackFacetName( &oFacet, InFacetName );
    ohUnpackFacetName( &oFacet, OutFacetName );
    if( octOpenFacet( &oFacet ) < OCT_OK ){
	fprintf(stderr,"Could not open %s: %s\n",
       ohFormatName( &oFacet ),
       octErrorString());
       exit(1);
    }
    /* creating the nets						*/
    for( row = 0; row < NumRows; row++ ){
	for( signal = SignalArray[row]; signal != NIL( s_signal );
	     signal = signal->next ){
	    if( signal->net->name != NIL( char )){
		oNet = newNet( signal->net->name );
		signal->net->netId = oNet->objectId;
	    } else if( signal->net->termName != NIL( char )){
		oNet = newNet( signal->net->termName );
		signal->net->netId = oNet->objectId;
	    } 
	}
    }
    /* creating all of the necessary turd properties			*/
    oProp.type = OCT_PROP;
    oProp.contents.prop.type = OCT_STRING;

    oProp.contents.prop.name = "VIEWTYPE";
    oProp.contents.prop.value.string = "SYMBOLIC";
    octCreate(&oFacet, &oProp );

    oProp.contents.prop.name = "CELLCLASS";
    oProp.contents.prop.value.string = "MODULE";
    octCreate(&oFacet, &oProp );

    oProp.contents.prop.name = "CELLTYPE";
    oProp.contents.prop.value.string = "COMBINATIONAL";
    octCreate(&oFacet, &oProp );

    oProp.contents.prop.name = "EDITSTYLE";
    oProp.contents.prop.value.string = "SYMBOLIC";
    octCreate(&oFacet, &oProp );

    oProp.contents.prop.name = "TECHNOLOGY";
    oProp.contents.prop.value.string = "scmos";
    octCreate(&oFacet, &oProp );

    oInstBag.type = OCT_BAG;
    oInstBag.contents.bag.name = "INSTANCES";
    octCreate(&oFacet, &oInstBag );

    oConnBag.type = OCT_BAG;
    oConnBag.contents.bag.name = "CONNECTORS";
    octCreate(&oFacet, &oConnBag );
}
/*
 * openFacet
 *
 * opens oFacet with the name facetname
 * facetname is of the form cell[:view[:facet[:version]]]
 * It is assumed that the default values of cell view version
 * are already set ( in octopusDefaults() ).
 */
openFacet( facetName )
char	*facetName;
{
    octObject oObj;

    oFacet.type = OCT_FACET;
    ohUnpackFacetName( &oFacet, facetName );
    oFacet.type = OCT_FACET;
    oFacet.contents.facet.mode = "w";
   if( octOpenFacet( &oFacet ) < OCT_OK ){
	fprintf( stderr, "Fatal error: Could not open %s:%s:%s: %s\n",
			 oFacet.contents.facet.cell,
			 oFacet.contents.facet.view,
			 oFacet.contents.facet.facet,
			 octErrorString());
	exit(1);
    }
    ohCreatePropStr( &oFacet, &oObj, "VIEWTYPE", "SYMBOLIC" );
    ohCreatePropStr( &oFacet, &oObj, "EDITSTYLE", "SYMBOLIC" );
    ohCreatePropStr( &oFacet, &oObj, "CELLCLASS", "MODULE" );
    ohCreatePropStr( &oFacet, &oObj, "CELLTYPE", "COMBINATIONAL" );
    ohCreatePropStr( &oFacet, &oObj, "TECHNOLOGY", Options.technology );
    ohCreateBag( &oFacet, &oObj, "INSTANCES" );
    ohCreateBag( &oFacet, &oObj, "CONNECTORS" );
}
/*
 * openPalette
 *
 * sets oPalette to the palette named name in technology
 */
openPalette( name )
char *name;
{

    if( tapOpenPalette( &oFacet, name, &oPalette, "r" ) == 0 ){
	fprintf( stderr, "Fatal error: Could not open palette %s: %s\n",
			 name, octErrorString());
	exit(1);
    }
}
/* 
 * openLocalPalette
 *
 * sets palette to the openFacet of name.  The default view and 
 * facet are :symbolic:contents
 */
openLocalPalette( name )
char	*name; /* name of local palette in form cell[:view[:facet[:version]]] */
{
    ohUnpackFacetName( &oPalette, name );
    oPalette.type = OCT_FACET;
    oPalette.contents.facet.mode = "r";
    if( octOpenFacet( &oPalette ) < OCT_OK ){
	fprintf( stderr, "Fatal error: Could not open %s:%s:%s: %s\n",
			 oPalette.contents.facet.cell,
			 oPalette.contents.facet.view,
			 oPalette.contents.facet.facet,
			 octErrorString());
	exit(1);
    }
}

/*
 * getInst
 *
 * gets an instance from oPalette by name, sets the transform,
 * attaches it to oFacet and the INSTANCES or CONNECTORS bag
 * The translation is set in Mkarray
 */
octObject *
getInst( name, transform, isInstance )
char *name;
enum octTransformType transform;
int isInstance;
{
    static octObject oConnectors;
    static octObject oInstances;
    static int instNum = 0;
    octObject *oInst;
    static char nameBuffer[128];

    if( instNum++ == 0 ){
	/* get oConnectors and oInstances from oFacet by name		*/
	ohGetByBagName(&oFacet, &oConnectors, "CONNECTORS" );
	ohGetByBagName(&oFacet, &oInstances, "INSTANCES" );
    }
    oInst = &ObjArray[ (++ObjCount)%OBJ_ARRAY_SIZE ];
    /* get the instance from oPalette by name				*/
    if( ohGetByInstName(&oPalette, oInst, name) < OCT_OK ){
	fprintf(stderr,"Fatal error: couldn't get %s from palette %s:%s: %s\n",
			oInst->contents.instance.name,
			oPalette.contents.facet.cell,
			oPalette.contents.facet.view,
			octErrorString());
	exit( 1 );
    }

#ifdef notdef
    /* getContainerByName to find if INSTANCE				*/
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = "INSTANCES";
    octRet = octGetContainerByName( oInst, &oBag);
    if( octRet == OCT_OK) {
	isInstance = TRUE;
    } else if ( octRet != OCT_NOT_FOUND) {
	ohOctFailed(__FILE__, __LINE__, NIL(char), NIL(octObject));
    } 
#endif

    /* set the instance's transform					*/
    oInst->contents.instance.transform.transformType = transform;

    /* create the instance in oFacet					*/
    octCreate( &oFacet, oInst );

    /* attach the instance to the INSTANCES or CONNECTORS bag		*/
    if ( isInstance ){
	octAttach( &oInstances, oInst );
    } else {
	octAttach( &oConnectors, oInst );
    }

    /* uniquifying the instance name if debug > 0 			*/
    if( Options.debug > 0 ){
	(void) sprintf( nameBuffer,"%s.%d", oInst->contents.instance.name, ++instNum );
	oInst->contents.instance.name = nameBuffer;
	octModify( oInst );
    }


    return( oInst );
}

/* 
 * promoteAterm
 *
 * promotes the actual terminal named atermName to a formal terminal
 * named ftermName.  A formal terminal and net named ftermName are 
 * created in oFacet.
 *
 */
promoteAterm( oInst, atermName, ftermName, termType, direction )
octObject *oInst;
char *atermName;
char *ftermName;
char *termType;		/* value of the TERMTYPE property		*/
char *direction;	/* value of the DIRECTION property		*/
{
    octObject oNet;
    octObject oAterm;
    octObject *oFterm;
    octStatus octRet;
    octObject oBag;	/* for call to find INSTANCES bag */
    octObject oProp;
    
    /* get oAterm by name from oInst					*/
    ohGetByTermName( oInst, &oAterm, atermName );

    /* create oFterm in oFacet						*/
    oFterm = &ObjArray[ (++ObjCount)%OBJ_ARRAY_SIZE ];
    ohCreateTerm( &oFacet, oFterm, ftermName );

    if( termType != NIL( char ) && *termType != '\0')
	ohCreatePropStr( oFterm, &oProp, "TERMTYPE", termType );
    if( direction != NIL( char ) && *direction != '\0')
	ohCreatePropStr( oFterm, &oProp, "DIRECTION", direction );

    /* create oNet in oFacet						*/
    ohCreateNet( &oFacet, &oNet, ftermName );

    /* attach oAterm to oFterm						*/
    octAttach( oFterm, &oAterm );

    /* attach oFterm to oNet						*/
    octAttach( &oNet, oFterm );

    /* if oInst is an INSTANCE, attach oAterm to oNet			*/
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = "INSTANCES";
    octRet = octGetContainerByName( oInst, &oBag);
    if( octRet == OCT_OK) {
	/* attach oAterm to oNet					*/
	octAttach( &oNet, &oAterm );
    } else if ( octRet != OCT_NOT_FOUND) {
	ohOctFailed(__FILE__, __LINE__, NIL(char), NIL(octObject));
    }

    /*
    return( oFterm );
    */
}

/*
 * moveInterfaceProps
 *
 * move the interface properties from the contents to the interface facets
 */
moveInterfaceProps()
{
    octObject oInterface;
    octObject oTerm;		/* contents formal terminals		*/
    octObject oITerm;		/* interface formal terminals		*/
    octObject oProp;		/* properties on the formal terms	*/
    octObject oNet;		/* Vdd and GND nets			*/
    octGenerator gTerm;

    oInterface = oFacet;
    oInterface.contents.facet.mode = "w";
    oInterface.contents.facet.facet = "interface";
    if( octOpenFacet( &oInterface ) < OCT_OK ){
	fprintf( stderr, "Fatal error: Could not open %s:%s:%s: %s\n",
			 oInterface.contents.facet.cell,
			 oInterface.contents.facet.view,
			 oInterface.contents.facet.facet,
			 octErrorString());
	exit(1);
    }
    oITerm.type = OCT_TERM;
    octInitGenContents( &oFacet, OCT_TERM_MASK, &gTerm );
    while( octGenerate( &gTerm, &oTerm ) == OCT_OK ){
	oITerm.contents.term = oTerm.contents.term;
	if( octGetByName( &oInterface, &oITerm ) != OCT_OK ){
	    fprintf( stderr, "Fatal error: %s line %d: Getting term %s from %s:%s:%s: %s\n",
			     __FILE__, __LINE__,
			     oITerm.contents.term.name,
			     oInterface.contents.facet.cell,
			     oInterface.contents.facet.view,
			     oInterface.contents.facet.facet,
			     octErrorString());
	    exit(1);
	}
	if( ohGetByPropName( &oTerm, &oProp, "TERMTYPE" ) == OCT_OK ){
	    octDelete( &oProp );
	    octCreate( &oITerm, &oProp );
	}
	if( ohGetByPropName( &oTerm, &oProp, "DIRECTION" ) == OCT_OK ){
	    octDelete( &oProp );
	    octCreate( &oITerm, &oProp );
	}
    } 

    /* mark the supply and ground nets in oFacet			*/
    if( ohGetByTermName( &oFacet, &oTerm, "Vdd") == OCT_OK){
	if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet) == OCT_OK){
	    ohCreatePropStr( &oNet, &oProp, "NETTYPE", "SUPPLY" );
	}
    }
    if( ohGetByTermName( &oFacet, &oTerm, "GND") == OCT_OK){
	if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet) == OCT_OK){
	    ohCreatePropStr( &oNet, &oProp, "NETTYPE", "GROUND" );
	}
    }

    octFlushFacet( &oInterface );
}
/*
 * newTran
 *
 */
octObject *
newTran( name, context, transform )
char *name;
int context;
enum octTransformType transform;
{
    char nameBuf[500];
    octObject *getInst();
    switch (context){
	case NO_CONTS:
	    (void) strcpy(nameBuf,name );
	    break;
	case DRAIN_CONTS:
	    (void) sprintf(nameBuf,"%s.%c",name,DRAIN_SPEC);
	    break;
	case SKIP_CONTS:
	    (void) sprintf(nameBuf,"%s.%c",name,SKIP_SPEC);
	    break;
	case SOURCE_CONTS:
	    (void) sprintf(nameBuf,"%s.%c",name,SOURCE_SPEC);
	    break;
	case DRAIN_CONTS|SKIP_CONTS:
	    (void) sprintf(nameBuf,"%s.%c%c",name,DRAIN_SPEC,SKIP_SPEC);
	    break;
	case DRAIN_CONTS|SOURCE_CONTS:
	    (void) sprintf(nameBuf,"%s.%c%c",name,DRAIN_SPEC,SOURCE_SPEC);
	    break;
    }
    return( getInst( nameBuf, transform, 1 ));
}
/*
 * newTerm
 *
 * Make a new terminal of type name
 */
octObject *
newTerm( name )
char *name;
{
    octObject *getInst();

    return( getInst( name, OCT_NO_TRANSFORM, 0 ));
}
/*
 * newFterm
 *
 * promotes a CONNECTOR instance to a formal terminal
 */
newFterm( net, oInst)
s_net *net;
octObject *oInst;
{
    octObject oNet;
    octObject oFterm;
    octObject oAterm;

    oNet.objectId = net->netId;
    octGetById( &oNet );

    octGenFirstContent( oInst, OCT_TERM_MASK, &oAterm );

    oFterm.type = OCT_TERM;
    oFterm.contents.term.name = net->termName;
    oFterm.contents.term.instanceId = oct_null_id;
    oFterm.contents.term.width = 1;
    octCreate( &oFacet, &oFterm );

    octAttach( &oNet, &oFterm );
    octAttach( &oFterm, &oAterm );
}
/*
 * mkNet
 * 
 * Creates a net 
 */
octObject *
newNet( name )
char *name;
{
    static octObject oNet;

    oNet.type = OCT_NET;
    oNet.contents.net.name = name;
    oNet.contents.net.width = 1;
    octCreate( &oFacet, &oNet );

    return ( &oNet );
}
