#include "electra.h"
#include "options.h"

optionStruct optionList[] = {
    { "T",	"technology", "name of the technology ( \"scmos\" default )"},
    { "n",	"size",       "(integer) size of an NFET in the palette"},
    { "p",	"size",       "(integer) size of a PFET in the palette"},
    { OPT_RARG,	"cell:view",  "cell and view of the palette"},
    { 0, 0, 0 }};

typedef struct {
    octObject **nfet;		/* array of nmos transistors		*/
    octObject **pfet;		/* array of pmos transistors		*/
    octObject oPower;		/* m1 term in power columns		*/
    octObject oOutput;		/* m1 term in output columns		*/
    octObject oPolyM1;		/* poly m1 connectors in output columns	*/
    octObject oPoly;		/* poly terms in signals		*/
    octObject oPolyM2;		/* poly - m2 terms in signals		*/
    octObject oM1M2;		/* vias					*/
    octObject oM2term;		/* metal2 connectors			*/
    octObject oM1Fterm;		/* metal1 formal terms			*/
    octObject oM2Fterm;		/* metal2 formal terms			*/
} s_cellObjs;

typedef struct paletteOptions {
    char *paletteName;
    char *technology;
    char *nfetName;
    char *pfetName;
    int numNfets;
    int numPfets;
    int *nfetWidth;
    int *nfetLength;
    int *pfetWidth;
    int *pfetLength;
} s_paletteOptions;

#define TRAN_ 0
#define TRAN_D 1
#define TRAN_K 2
#define TRAN_S 3
#define TRAN_DK 4
#define TRAN_DS 5
#define PURPOSE "purpose"

s_paletteOptions Options;
s_cellObjs Cells;

/*
 * createPalette
 *
 * a program to create the palette for electra.
 */

octObject oPalette;		/* the rule facet 			*/

main( argc, argv )
int argc;
char **argv;
{

    octObject oRule;
    octObject oConnect;
    octObject oTgroup;
    int fetId;
    int type;

    parseComLine(argc, argv);
    openPalette();
    createFets( NFET, &Cells.nfet, Options.numNfets,
		Options.nfetWidth, Options.nfetLength );
    createFets( PFET, &Cells.pfet, Options.numPfets,
		Options.pfetWidth, Options.pfetLength );
    createConnectors();
    createLayers();

    /* signal rules */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, SIGNAL_RULE_NAME );
    newBag( &oRule, &oConnect, CONNECTBAG );
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );

    for( fetId = 0; fetId < Options.numNfets; fetId++ ){
	for( type = TRAN_; type <= TRAN_DS; type++ ){
	    attachTermByName( &oTgroup, &Cells.nfet[type][fetId], GATE );
	}
    }
    for( fetId = 0; fetId < Options.numPfets; fetId++ ){
	for( type = TRAN_; type <= TRAN_DS; type++ ){
	    attachTermByName( &oTgroup, &Cells.pfet[type][fetId], GATE );
	}
    }

    octAttach( &oTgroup, &Cells.oPoly );
    octAttach( &oTgroup, &Cells.oPolyM1 );
    octAttach( &oTgroup, &Cells.oPolyM2 );

    /* power rules */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, POWER_RULE_NAME );
    newBag( &oRule, &oConnect, CONNECTBAG );
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < Options.numNfets; fetId++ ){
	attachTermByName( &oTgroup, &Cells.nfet[TRAN_DS][fetId], SOURCE );
	attachTermByName( &oTgroup, &Cells.nfet[TRAN_S][fetId], SOURCE );
    }
    for( fetId = 0; fetId < Options.numPfets; fetId++ ){
	attachTermByName( &oTgroup, &Cells.pfet[TRAN_DS][fetId], SOURCE );
	attachTermByName( &oTgroup, &Cells.pfet[TRAN_S][fetId], SOURCE );
    }
    octAttach( &oConnect, &Cells.oPower );

    newBag( &oRule, &oConnect, CONNECTBAG );
    octAttach( &oConnect, &Cells.oPower );
    octAttach( &oConnect, &oTgroup );

    /* power track rule */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, POWERTRACK_RULE_NAME );
    newBag( &oRule, &oConnect, CONNECTBAG );
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    octAttach( &oTgroup, &Cells.oPower );
    octAttach( &oTgroup, &Cells.oOutput );
    octAttach( &oTgroup, &Cells.oM1M2 );

    /* nand rules */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, NAND_RULE_NAME );
    networkRule( &oRule, Cells.nfet, Options.numNfets, Cells.pfet, Options.numPfets );

    /* nor rules */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, NOR_RULE_NAME );
    networkRule( &oRule, Cells.pfet, Options.numPfets, Cells.nfet, Options.numNfets );

    /* firsttran rules */
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, FIRSTTRAN_RULE_NAME );
    newBag( &oRule, &oConnect, CONNECTBAG );
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );

    for( fetId = 0; fetId < Options.numNfets; fetId++ ){
	attachTermByName( &oTgroup, &Cells.nfet[TRAN_D][fetId], DRAIN );
	attachTermByName( &oTgroup, &Cells.nfet[TRAN_DK][fetId], DRAIN );
	attachTermByName( &oTgroup, &Cells.nfet[TRAN_DS][fetId], DRAIN );
    }
    for( fetId = 0; fetId < Options.numPfets; fetId++ ){
	attachTermByName( &oTgroup, &Cells.pfet[TRAN_D][fetId], DRAIN );
	attachTermByName( &oTgroup, &Cells.pfet[TRAN_DK][fetId], DRAIN );
	attachTermByName( &oTgroup, &Cells.pfet[TRAN_DS][fetId], DRAIN );
    }
    octAttach( &oConnect, &Cells.oOutput );

    newBag( &oRule, &oConnect, CONNECTBAG );
    octAttach( &oConnect, &Cells.oOutput );
    octAttach( &oConnect, &oTgroup );

    /* output column							*/
    newBag( &oPalette, &oRule, RULEBAG );
    newStringProp( &oRule, RULENAMEPROP, OUTPUT_RULE_NAME );
    newBag( &oRule, &oConnect, CONNECTBAG );
    newIntProp( &oConnect, MK_ORIENTATION_PROP, MK_VER_SEG );
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    octAttach( &oTgroup, &Cells.oOutput );
    octAttach( &oTgroup, &Cells.oPolyM1 );
    octAttach( &oTgroup, &Cells.oM1M2 );
    octAttach( &oTgroup, &Cells.oM1Fterm );


    octCloseFacet( &oPalette );
    exit( 0 );
}

parseComLine( argc, argv )
int argc;
char **argv;
{
    extern char *optarg;
    int option;
    Options.technology = "scmos";
    Options.nfetName = "nfet";
    Options.pfetName = "pfet";
    Options.numNfets = 0;
    Options.numPfets = 0;
    Options.nfetWidth = NIL( int );
    Options.nfetLength = NIL( int );
    Options.pfetWidth = NIL( int );
    Options.pfetLength = NIL( int );

    if( argc == 1 ){
	optUsage();
	exit(1);
    }
    while((option = optGetOption( argc , argv )) != EOF ){
	switch( option ) {
	    case 'T':
		Options.technology = optarg;
		break;
	    case 'n':
		Options.numNfets++;
		Options.nfetWidth = REALLOC( int, Options.nfetWidth, 
					     Options.numNfets );
		Options.nfetLength = REALLOC( int, Options.nfetLength, 
					      Options.numNfets );
		if( sscanf( optarg, "%dx%d", 
			    &Options.nfetWidth[ Options.numNfets - 1 ],
			    &Options.nfetLength[ Options.numNfets - 1 ]) != 2 ){
		    fprintf( stderr, "Bad arguement to -n option - %s\n",
			     "expected form is \"WxL\" with W and L integers");
		    exit( 1 );
		} 
		break;
	    case 'p':
		Options.numPfets++;
		Options.pfetWidth = REALLOC( int, Options.pfetWidth, 
					     Options.numPfets );
		Options.pfetLength = REALLOC( int, Options.pfetLength, 
					      Options.numPfets );
		if( sscanf( optarg, "%dx%d", 
			    &Options.pfetWidth[ Options.numPfets - 1 ],
			    &Options.pfetLength[ Options.numPfets - 1 ]) != 2 ){
		    fprintf( stderr, "Bad arguement to -p option - %s\n",
			     "expected form is \"WxL\" with W and L integers");
		    exit( 1 );
		} 
		break;
	    default:
		optUsage();
		break;
	}
    }
    Options.paletteName = argv[ optind ];
}
/*
 * network rule
 *
 * construct a network rule
 */
networkRule( oRule, sfet, numSfet, pfet, numPfets )
octObject *oRule;		/* the rule to be constructed		*/
octObject **sfet;		/* array of series transistors		*/
int numSfet;			/* number of series transistors		*/
octObject **pfet;	 	/* array of parallel transistors	*/
int numPfets;			/* number of parallel transistors	*/
{
    octObject oConnect;
    octObject oTgroup;
    int fetId;

    /* series trans 							*/
    newBag( oRule, &oConnect, CONNECTBAG );
    newStringProp( &oConnect, PURPOSE, "series trans");

    /* series sources							*/
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numSfet; fetId++ ){
	attachTermByName( &oTgroup, &sfet[TRAN_][fetId], SOURCE );
	attachTermByName( &oTgroup, &sfet[TRAN_D][fetId], SOURCE );
	attachTermByName( &oTgroup, &sfet[TRAN_DK][fetId], SOURCE );
	attachTermByName( &oTgroup, &sfet[TRAN_K][fetId], SOURCE );
    }
    /* series drains							*/
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numSfet; fetId++ ){
	attachTermByName( &oTgroup, &sfet[TRAN_][fetId], DRAIN );
	attachTermByName( &oTgroup, &sfet[TRAN_D][fetId], DRAIN );
	attachTermByName( &oTgroup, &sfet[TRAN_DK][fetId], DRAIN );
	attachTermByName( &oTgroup, &sfet[TRAN_DS][fetId], DRAIN );
	attachTermByName( &oTgroup, &sfet[TRAN_K][fetId], DRAIN );
	attachTermByName( &oTgroup, &sfet[TRAN_S][fetId], DRAIN );
    }

    /* parallel trans 							*/
    newBag( oRule, &oConnect, CONNECTBAG );
    newStringProp( &oConnect, PURPOSE, "parallel sources");
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numPfets; fetId++ ){
	attachTermByName( &oTgroup, &pfet[TRAN_DS][fetId], SOURCE );
	attachTermByName( &oTgroup, &pfet[TRAN_S][fetId], SOURCE );
    }

    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numPfets; fetId++ ){
	attachTermByName( &oTgroup, &pfet[TRAN_][fetId], SOURCE );
	attachTermByName( &oTgroup, &pfet[TRAN_D][fetId], SOURCE );
    }

    newBag( oRule, &oConnect, CONNECTBAG );
    newStringProp( &oConnect, PURPOSE, "parallel drains");
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numPfets; fetId++ ){
	attachTermByName( &oTgroup, &pfet[TRAN_D][fetId], DRAIN );
	attachTermByName( &oTgroup, &pfet[TRAN_DS][fetId], DRAIN );
    }

    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numPfets; fetId++ ){
	attachTermByName( &oTgroup, &pfet[TRAN_][fetId], DRAIN );
	attachTermByName( &oTgroup, &pfet[TRAN_S][fetId], DRAIN );
    }

    /* output columns							*/
    newBag( oRule, &oConnect, CONNECTBAG );
    newStringProp( &oConnect, PURPOSE, "parallel drains to output column");
    newBag( &oConnect, &oTgroup, TERMGROUPBAG );
    for( fetId = 0; fetId < numPfets; fetId++ ){
	attachTermByName( &oTgroup, &pfet[TRAN_D][fetId], DRAIN );
	attachTermByName( &oTgroup, &pfet[TRAN_DS][fetId], DRAIN );
    }
    octAttach( &oConnect, &Cells.oOutput );

    newBag( oRule, &oConnect, CONNECTBAG );
    newStringProp( &oConnect, PURPOSE, "output column to parallel drains");
    octAttach( &oConnect, &Cells.oOutput );
    octAttach( &oConnect, &oTgroup );
}

createFets(  type, fet, numFets, fetWidth, fetLength )
enum fetType type;
octObject ***fet;
int numFets;
int *fetWidth;
int *fetLength;
{
    int fetId;
    octObject oInst;
    char *master;

    master = ALLOC( char, strlen( oPalette.contents.facet.cell ) +
		          strlen( Options.nfetName ) + 
		          strlen( Options.pfetName ) + 50 );

    oInst.type = OCT_INSTANCE;
    oInst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    oInst.contents.instance.transform.translation.x = 0;
    oInst.contents.instance.transform.translation.y = 0;
    oInst.contents.instance.name = 
	master + strlen( oPalette.contents.facet.cell ) + 1;
    oInst.contents.instance.master = master;
    oInst.contents.instance.view = "physical";
    oInst.contents.instance.facet = "contents";
    oInst.contents.instance.version = OCT_CURRENT_VERSION;

    (void)(void) sprintf( master,"%s/", oPalette.contents.facet.cell );

    (*fet) = ALLOC( octObject *, 6 );
    (*fet)[TRAN_] = ALLOC( octObject, numFets );
    (*fet)[TRAN_D] = ALLOC( octObject, numFets );
    (*fet)[TRAN_K] = ALLOC( octObject, numFets );
    (*fet)[TRAN_S] = ALLOC( octObject, numFets );
    (*fet)[TRAN_DK] = ALLOC( octObject, numFets );
    (*fet)[TRAN_DS] = ALLOC( octObject, numFets );

    for( fetId = 0; fetId < numFets; fetId++ ){
	(*fet)[TRAN_][fetId] = (*fet)[TRAN_D][fetId] = oInst;
	(*fet)[TRAN_K][fetId] = (*fet)[TRAN_S][fetId] = oInst;
	(*fet)[TRAN_DK][fetId] = (*fet)[TRAN_DS][fetId] = oInst;
    }

    for( fetId = 0; fetId < numFets; fetId++ ){

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_][fetId], type, NO_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_][fetId] );

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d.d",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_D][fetId], type, DRAIN_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_D][fetId] );

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d.k",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_K][fetId], type, SKIP_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_K][fetId] );

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d.s",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_S][fetId], type, SOURCE_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_S][fetId] );

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d.dk",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_DK][fetId], type, DRAIN_CONTS|SKIP_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_DK][fetId] );

	(void)(void) sprintf( oInst.contents.instance.name, "%s.%dx%d.ds",
		       ( type == NFET ? Options.nfetName : Options.pfetName ), 
		       fetWidth[ fetId ], fetLength[ fetId ] );
	createTran( &(*fet)[TRAN_DS][fetId], type, DRAIN_CONTS|SOURCE_CONTS, 
		    fetWidth[fetId], fetLength[ fetId ], Options.technology );
	octCreate( &oPalette, &(*fet)[TRAN_DS][fetId] );
    }
    FREE( master );
}

createConnectors()
{
    octObject oInst;
    char *master;

    master = ALLOC( char, strlen( oPalette.contents.facet.cell ) +
		          strlen( Options.nfetName ) + 
		          strlen( Options.pfetName ) + 50 );

    oInst.type = OCT_INSTANCE;
    oInst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    oInst.contents.instance.transform.translation.x = 0;
    oInst.contents.instance.transform.translation.y = 0;
    oInst.contents.instance.name = 
	master + strlen( oPalette.contents.facet.cell ) + 1;
    oInst.contents.instance.master = master;
    oInst.contents.instance.view = "physical";
    oInst.contents.instance.facet = "contents";
    oInst.contents.instance.version = OCT_CURRENT_VERSION;

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, M1TERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oOutput );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, POWERTERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oPower );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, POLYTERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oPoly );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, POLYM1TERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oPolyM1 );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, POLYM2TERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oPolyM2 );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, M1M2TERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oM1M2 );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, M1FTERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oM1Fterm );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, M2FTERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oM2Fterm );

    (void) sprintf( master,"%s/%s", oPalette.contents.facet.cell, M2TERM );
    octCreate( &oPalette, &oInst );
    octGenFirstContent( &oInst, OCT_TERM_MASK, &Cells.oM2term );

    FREE( master );
}

/*
 * createLayers
 *
 * create the layers and horizontal and vertical precedence
 */
createLayers()
{
    octObject oHorBag;
    octObject oVerBag;
    octObject oMet2;
    octObject oMet1;
    octObject oPoly;
    octObject oNDif;
    octObject oPDif;
    octObject oProp;

    oHorBag.type = oVerBag.type = OCT_BAG;
    oHorBag.contents.bag.name = HORIZONTAL_LAYERS;
    oVerBag.contents.bag.name = VERTICAL_LAYERS;

    octCreate( &oPalette, &oHorBag );
    octCreate( &oPalette, &oVerBag );

    oMet2.type = oMet1.type = oPoly.type = oNDif.type = oPDif.type = OCT_LAYER;
    oMet2.contents.layer.name = "MET2";
    oMet1.contents.layer.name = "MET1";
    oPoly.contents.layer.name = "POLY";
    oNDif.contents.layer.name = "NDIF";
    oPDif.contents.layer.name = "PDIF";

    octCreate( &oPalette, &oMet2 );
    octCreate( &oPalette, &oMet1 );
    octCreate( &oPalette, &oPoly );
    octCreate( &oPalette, &oNDif );
    octCreate( &oPalette, &oPDif );

    oProp.type = OCT_PROP;
    oProp.contents.prop.name = MK_WIDTH_PROP;
    oProp.contents.prop.type = OCT_INTEGER;
    oProp.contents.prop.value.integer = MK_MAX_WIDTH;

    octCreate( &oNDif, &oProp );
    octCreate( &oPDif, &oProp );

    octAttach( &oHorBag, &oMet2 );
    octAttach( &oHorBag, &oMet1 );
    octAttach( &oHorBag, &oPoly );
    octAttach( &oHorBag, &oNDif );
    octAttach( &oHorBag, &oPDif );

    octAttach( &oVerBag, &oMet1 );
    octAttach( &oVerBag, &oMet2 );
    octAttach( &oVerBag, &oPoly );
    octAttach( &oVerBag, &oNDif );
    octAttach( &oVerBag, &oPDif );
}

/*
 * openPalette
 *
 * open the palette for write
 */
openPalette()
{
    octObject oProp;

    ohUnpackDefaults( &oPalette, "w", ":symbolic:contents" );
    ohUnpackFacetName( &oPalette, Options.paletteName );
    if( octOpenFacet( &oPalette ) < OCT_OK ){
	fprintf( stderr, "Could not open %s: %s\n",
		 ohFormatName( &oPalette ), octErrorString());
	exit(1);
    }
    ohCreatePropStr( &oPalette, &oProp,
	                         "TECHNOLOGY", Options.technology );
    ohCreatePropStr( &oPalette, &oProp, "VIEWTYPE", "symbolic" );
    ohCreatePropStr( &oPalette, &oProp, "EDITSTYLE", "symbolic" );
}
newBag( oContainer, oBag, name )
octObject *oContainer;
octObject *oBag;
char *name;
{
    oBag->type = OCT_BAG;
    oBag->contents.bag.name = name;
    octCreate( oContainer, oBag );
}
newStringProp( oContainer, name, value )
octObject *oContainer;
char *name;
char *value;
{
    octObject oProp;
    oProp.type = OCT_PROP;
    oProp.contents.prop.name = name;
    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.value.string = value;
    octCreate( oContainer, &oProp );
}
newIntProp( oContainer, name, value )
octObject *oContainer;
char *name;
int value;
{
    octObject oProp;
    oProp.type = OCT_PROP;
    oProp.contents.prop.name = name;
    oProp.contents.prop.type = OCT_INTEGER;
    oProp.contents.prop.value.integer = value;
    octCreate( oContainer, &oProp );
}
attachTermByName( oContainer, oInst, name )
octObject *oContainer;
octObject *oInst;
char *name;
{
    octObject oTerm;

    oTerm.type = OCT_TERM;
    oTerm.contents.term.name = name;
    octGetByName( oInst, &oTerm );
    octAttach( oContainer, &oTerm );
}
