#include "electra.h"
#include "options.h"
extern s_options Options;
optionStruct optionList[] = {
    { "i",     "input file", "name of the personality matrix"},
    { "T",      "technology", "name of the technology ( \"scmos\" default )"},
    { "p",     "palette-name", "palette name known to tap"},
    { "P",     "user-palette[:view[:facet[:version]]]",
	       "name of user palette. Overrides -p"},
    { "f",      "",           "quickly prepares the input for floorplanning"},
    { "F",      "",           "prepares the input for floorplanning"},
    { "g",      "",           "prints the mkarray constraint graphs"},
    { "n",      "",           "no sparcs terminal contstraints are added"},
    { "p",      "",           "allow polysilicon formal terms"},
    { "s",      "",           "size the transistors"},
    { "w",      "",           "no wiring is done"},
    { "c",     "cell library","location of the cell library"},
    { "d",     "debug level","sets different levels if debug"},
    { OPT_CONT, 0,"\n    0 - run silently"},
    { OPT_CONT, 0,"\n    1 - print each step in the process ( default )"},
    { OPT_CONT, 0,"\n    2 - print details of each step ( verbose )"},
    { OPT_CONT, 0,"\n    3 - sets debugging mode, results may vary"},
    { "r",     "rule facet", "name of the rule facet ( in the cell library )"},
    { "t",     "tech file",  "name of the tech file ( in the cell library )"},
    { "o",     "cell:view",  "name and view of the output facet"},
    { OPT_RARG, "cell:view",  "name and view of the input facet"},
    { 0,        0,               0}
};

/* global variables set by the command line				*/

char *Celllib = DEFAULT_CELLLIB;
int Debug = 1;
int FloorPlannerPrep = 0;
int PrintGraph = FALSE;
char *InfileName = "final";
int NoTermConstraints = FALSE;
int PolyTerms = FALSE;
char *RuleFacetName = DEFAULT_RULEFACET;
int Size = FALSE;
char *TechFileName = "electra.tech";
int Wire = TRUE;
char *OutFacetName = ":unspaced";
char *InFacetName = "";

/*
 * parseComLine
 *
 * parse the command line 
 */
parseComLine( argc, argv )
int argc;
char **argv;
{

    extern char *optarg;
    int option;

    while(( option = optGetOption( argc,argv)) != EOF ){
	switch ( option ) {
	    case 'T':
		Options.technology = optarg;
		break;
	    case 'p':
		Options.paletteName = optarg;
		break;
	    case 'P':
		Options.localPaletteName = optarg;
		break;
	    case 'c':
		Celllib = util_strsav( optarg );
		break;
	    case 'd':
		Debug = atoi( optarg );
		break;
	    case 'f':
		FloorPlannerPrep = 1;
		break;
	    case 'F':
		FloorPlannerPrep = 2;
		break;
	    case 'g':
		PrintGraph = TRUE;
		break;
	    case 'i':
		InfileName = util_strsav( optarg );
		break;
	    case 'n':
		NoTermConstraints = TRUE;
		break;
	    case 'l':
		PolyTerms = TRUE;
		break;
	    case 'o':
		OutFacetName = util_strsav( optarg );
		break;
	    case 'r':
		RuleFacetName = util_strsav( optarg );
		break;
	    case 's':
		Size = TRUE;
		break;
	    case 't':
		TechFileName = util_strsav( optarg );
		break;
	    case 'w':
		Wire = FALSE;
		break;
	    default:
		optUsage();
		break;
	}
    }
    if( argc == 1 ){
	optUsage();
	exit(-1);
    }
    InFacetName = argv[ optind ];
}
