#include "electra.h"
#define MAX_POLY_RUN_TO_EDGE 2
/*
 * routePoly
 *
 * These routines route to the left and right side of the array
 * in POLY.  They return TRUE if the route was completed and 
 * mark net->status as ROUTED.  The route is completed if the
 * distance from the edge of the signal to the edge of the array 
 * is >= than MAX_POLY_RUN_TO_EDGE.
 */
routePolyLeft( signal )
s_signal *signal;
{
    octObject *oTerm;		/* the formal term			*/
    octObject *newTerm();	/* returns an instance			*/
    octBox area;
    int row, col;
    s_network *network;

    if(( signal->net->status&ROUTED) == ROUTED )return;

    /* checking conditions where it is not favorable to route in POLY	*/
    if(( signal->net->network != NIL( s_network ))||
       ( signal->numTerms > 0 )||
       ( signal->left >= MAX_POLY_RUN_TO_EDGE )) return;

    /* to prevent false transistors on another network			*/
    row = signal->row;
    for( col = 0; col < signal->left; col++ ){
	if( Matrix[ row ][ col ].network != NIL(s_network ))return;
    }

    /* the signal can be routed						*/
    signal->net->status |= ROUTED;
    signal->left = 0;
    oTerm = newTerm( POLYTERM );
    newFterm( signal->net, oTerm );

    area.lowerLeft.x = area.upperRight.x = 0;
    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
    mkAttach( FullArray, &area, oTerm );
    constrainTerm( L_EDGE, oTerm );

    /* put an output term in the signal if necessary			*/
    network = signal->net->network;
    if(( network != NIL(s_network ))&&( signal->numTrans == 0 )){
	oTerm = newTerm( POLYM1TERM );
	area.lowerLeft.x = area.upperRight.x = 4 * network->col + 3;
	mkAttach( FullArray, &area, oTerm );
    }
}
routePolyRight( signal )
s_signal *signal;
{
    octObject *oTerm;		/* the formal term			*/
    octObject *newTerm();	/* returns an instance			*/
    octBox area;
    int row, col;
    s_network *network;

    if(( signal->net->status&ROUTED) == ROUTED )return;

    /* checking conditions where it is not favorable to route in POLY	*/
    if(( signal->net->network != NIL( s_network ))||
       ( signal->numTerms > 0 )||
       ( NumCols - signal->right > MAX_POLY_RUN_TO_EDGE )) return;


    /* to prevent false transistors on another network			*/
    row = signal->row;
    for( col = NumCols - 1; col > signal->right; col-- ){
	if( Matrix[ row ][ col ].network != NIL(s_network ))return;
    }

    /* the signal can be routed						*/
    signal->net->status |= ROUTED;
    signal->right = NumCols - 1;
    oTerm = newTerm( POLYTERM );
    newFterm( signal->net, oTerm );

    area.lowerLeft.x = area.upperRight.x = TotalCols - 1;
    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
    mkAttach( FullArray, &area, oTerm );
    constrainTerm( R_EDGE, oTerm );

    /* put an output term in the signal if necessary			*/
    network = signal->net->network;
    if(( network != NIL(s_network ))&&( signal->numTrans == 0 )){
	oTerm = newTerm( POLYM1TERM );
	area.lowerLeft.x = area.upperRight.x = 4 * network->col + 3;
	mkAttach( FullArray, &area, oTerm );
    }
}
