#include "electra.h"
/*
 * signalTerms
 *
 * Wire the signals to the edge of the array
 */
signalTerms()
{
    unsigned *edgeMask;
    octBox area;
    s_signal *signal;
    s_net *net;
    octPoint termloc;
    int row, col;
    int index;
    octObject oAterm;
    octObject *oFterm;
    octObject *newTerm();

    for( row = 0; row < NumRows; row++ ){
	for( signal = SignalArray[row]; signal != NIL( s_signal );
	     signal = signal->next ){

	    net = signal->net;
	    termloc = net->termloc;

	    if(((net->edge&ANY_EDGE ) == 0 )|| 
	       ( termloc.y != signal->row )||
	       (( net->network != NIL( s_network ))&&
		( net->network->col == termloc.x )))continue;

	    if( net->edge&TB_EDGE){

		if(net->edge&T_EDGE)edgeMask = TEdgeMask;
		else                edgeMask = BEdgeMask;


		/* creating and placing the actual term in the signal	*/
		if( IS_TRAN( termloc.y, termloc.x )){

#ifdef LASTGASP
		    lastGasp( net );
#else
		    errorTerm( net );
#endif
		    continue;

		} else if(( edgeMask[ termloc.x ]&LEFT ) != LEFT ){

		    /* wiring on the left metal1 track			*/
		    edgeMask[ termloc.x ] |= LEFT;
		    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 2;

		} else if(( edgeMask[ termloc.x ]&CENTER ) != CENTER ){

		    /* wiring on the output column metal1 track		*/
		    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 3;
		    edgeMask[ termloc.x ] |= CENTER;

		} else if(( edgeMask[ termloc.x ]&RIGHT ) != RIGHT ){

		    /* wiring on the right metal1 track			*/
		    area.lowerLeft.x = 5 * termloc.x + 4;
		    area.upperRight.x = area.lowerLeft.x + 1;
		    edgeMask[ termloc.x ] |= RIGHT;

		} else {

		    /* every terminal location is blocked */
		    errorTerm( net );
		    continue;

		}
		/* creating the actual terminal - if the signal has 	*/
		/* metal2, the term is POLYM2 and has to inserted in	*/
		/* the signal->term[] array.  Otherwise the term is	*/
		/* polym1 and does not have to inserted into 		*/
		/* signal->term 					*/

		if( signal->numTerms > 0 ){

		    /* connect to metal 2				*/
		    oAterm = *newTerm( M1M2TERM );
		    signal->term = REALLOC( s_termList , 
					    signal->term, ++signal->numTerms );
		    for( index = signal->numTerms - 1; index >= 0; index-- ){
			if(( index == 0 )||
			   ( signal->term[ index - 1 ].col <= termloc.x )){
			    signal->term[ index ].col = termloc.x;
			    signal->term[ index ].objectId = oAterm.objectId;
			    break;
			} else {
			    signal->term[ index ] = signal->term[ index - 1];
			}
		    }
		    if( net->edge&UPPER ){
			area.lowerLeft.y =  
			area.upperRight.y = 3 * termloc.y + 4; 
		    } else {
			area.lowerLeft.y = 
			area.upperRight.y = 3 * termloc.y + 2;  
		    }

		    mkAttach( FullArray, &area, &oAterm );
		} else {
		    oAterm = *newTerm( POLYM1TERM );
		    area.lowerLeft.y = area.upperRight.y = 3 * signal->row + 3;
		    mkAttach( FullArray, &area, &oAterm );
		}

		/* creating and placing the formal term			*/
		oFterm = newTerm( M1FTERM );
		newFterm( net, oFterm );
		if( net->edge&T_EDGE ){
		    area.lowerLeft.y = area.upperRight.y = TotalRows - 1;
		    constrainTerm( T_EDGE, oFterm );
		} else {
		    area.lowerLeft.y = area.upperRight.y = 0;
		    constrainTerm( B_EDGE, oFterm );
		}
		mkAttach( FullArray, &area, oFterm );
		mkNewSegment( FullArray, oFterm, &oAterm, NIL(char),
					 MK_MIN_WIDTH, MK_VER_SEG );
	    } else {
		/* these terms are wired to the edge in wireLRTerms	*/
		if( signal->numTerms == 0 ){
		    newPolyM2Term( signal, signal->net->termloc.x, TRUE );
		}
		if( net->edge&L_EDGE ){
		    oAterm.objectId = signal->term[ 0 ].objectId;
		    octGetById( &oAterm );
		    col = signal->term[ 0 ].col;
		    wireToLeftEdge( net, &oAterm, row, col );
		} else {
		    oAterm.objectId = signal->term[ signal->numTerms - 1 ].objectId;
		    octGetById( &oAterm );
		    col = signal->term[ signal->numTerms - 1 ].col;
		    wireToRightEdge( net, &oAterm, termloc.y, termloc.x );
		}
	    }
	}
    }
}

#ifdef notdef
/*
 * lastGasp
 *
 * tries to route a terminal to the TOP or BOTTOM of the array, if not
 * possible, calls errorTerm
 *
 */
lastGasp( net )
s_net *net;
{
    octPoint termloc;
    s_signal *signal;
    int row, index;
    int blockage;
    unsigned *edgeMask;
    octBox area;
    octObject oAterm1;
    octObject oAterm2;
    octObject *oFterm;
    octObject *newTerm();


    termloc = net->termloc;
    signal = net->signal;

    if( net->edge&T_EDGE ){
	/* check for metal1 blockages 					*/
	if( Matrix[ NumRows - 1 ][ termloc.x ].metal1Mask > 3 )errorTerm( net );
	for( blockage = NumRows - 1; blockage >= termloc.y; blockage-- ){
	    if( Matrix[ blockage ][ termloc.x ].metal1Mask > 3 )break;
	}
	/* if no blockage, run the term in the unused output track 	*/
	if( blockage < termloc.y ){
	    edgeMask[ termloc.x ] |= CENTER;
	    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 3;
	    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
	    oAterm2 = *newTerm( POLYM1TERM );
	    mkAttach( FullArray, &area, &oAterm2 );

	/* else, try to run past the blockage in poly			*/
	} else {
	    /* check to see if there is a poly blockage			*/
	    for( row = termloc.y; row < blockage; row++ ){
		signal = Matrix[ row ][ termloc.x ].signal;
		for( index = 0; index < signal->numRuns; index++ ){
		    if( signal->polyrun[ index ].left > termloc.x )break;
		    if( signal->polyrun[ index ].right > termloc.x ){
			/* give up - its blocked			*/
			errorTerm( net );
		    }
		}
	    }

	    /* wiring on the output column metal1 track		*/
	    if(( edgeMask[ termloc.x ]&CENTER ) == CENTER )errorTerm( net );
	    edgeMask[ termloc.x ] |= CENTER;
	    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 3;
	    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
	    oAterm1 = *newTerm( POLYTERM );
	    mkAttach( FullArray, &area, &oAterm1 );

	    area.lowerLeft.y = area.upperRight.y = 3 * (blockage + 1)  + 2;
	    oAterm2 = *newTerm( POLYM1TERM );
	    mkAttach( FullArray, &area, &oAterm2 );
	    mkNewSegment( FullArray, &oAterm1, &oAterm2, NIL(char),
				     MK_MIN_WIDTH, MK_VER_SEG );
	}

	oFterm = newTerm( M1FTERM );
	newFterm( net, oFterm );
	constrainTerm( T_EDGE, oFterm );
	area.lowerLeft.y = area.upperRight.y = TotalRows - 1;
	mkAttach( FullArray, &area, oFterm );
	mkNewSegment( FullArray, oFterm, &oAterm2, NIL(char),
				 MK_MIN_WIDTH, MK_VER_SEG );
    } else if( net->edge&B_EDGE ){
	/* check for metal1 blockages 					*/
	if( Matrix[ 0 ][ termloc.x ].metal1Mask > 3 )errorTerm( net );
	for( blockage = 0; blockage <= termloc.y; blockage++ ){
	    if( Matrix[ blockage ][ termloc.x ].metal1Mask > 3 )break;
	}
	/* if no blockage, run the term in the unused output track 	*/
	if( blockage > termloc.y ){
	    edgeMask[ termloc.x ] |= CENTER;
	    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 3;
	    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
	    oAterm2 = *newTerm( POLYM1TERM );
	    mkAttach( FullArray, &area, &oAterm2 );

	/* else, try to run past the blockage in poly			*/
	} else {
	    /* check to see if there is a poly blockage			*/
	    for( row = termloc.y; row > blockage; row-- ){
		signal = Matrix[ row ][ termloc.x ].signal;
		for( index = 0; index < signal->numRuns; index++ ){
		    if( signal->polyrun[ index ].left > termloc.x )break;
		    if( signal->polyrun[ index ].right > termloc.x ){
			/* give up - its blocked			*/
			errorTerm( net );
		    }
		}
	    }

	    /* wiring on the output column metal1 track		*/
	    if(( edgeMask[ termloc.x ]&CENTER ) == CENTER )errorTerm( net );
	    edgeMask[ termloc.x ] |= CENTER;
	    area.lowerLeft.x = area.upperRight.x = 5 * termloc.x + 3;
	    area.lowerLeft.y = area.upperRight.y = 3 * row + 3;
	    oAterm1 = *newTerm( POLYTERM );
	    mkAttach( FullArray, &area, &oAterm1 );

	    area.lowerLeft.y = area.upperRight.y = 3 * (blockage + 1)  + 4;
	    oAterm2 = *newTerm( POLYM1TERM );
	    mkAttach( FullArray, &area, &oAterm2 );
	    mkNewSegment( FullArray, oFterm, &oAterm2, NIL(char),
		          MK_MIN_WIDTH, MK_VER_SEG );
	}

	oFterm = newTerm( M1FTERM );
	newFterm( net, oFterm );
	constrainTerm( T_EDGE, oFterm );
	area.lowerLeft.y = area.upperRight.y = 0;
	mkAttach( FullArray, &area, oFterm );
	mkNewSegment( FullArray, &oAterm1, &oAterm2, NIL(char),
		      MK_MIN_WIDTH, MK_VER_SEG );
    
    } else {
	errorTerm( net );
    }
}
#endif

/*
 * errorTerm
 *
 * Places a terminal on a signal in case of error
 * This error is typically caused by a terminal constraint violation
 */
errorTerm( net )
s_net *net;
{
    octBox area;
    octObject *oFterm;
    octObject *newTerm();
    octPoint termloc;
    int index;
    s_signal *signal;

    /* scream								*/
    termloc = net->termloc;
    signal = net->signal;

    if( net->edge&T_EDGE ){
	fprintf( stderr,
		 "WARNING: failed to connect term %s to the top edge\n", 
		 net->termName );
    } else if( net->edge&B_EDGE ){
	fprintf( stderr,
		 "WARNING: failed to connect term %s to the bottom edge\n", 
		 net->termName );
    } else if( net->edge&L_EDGE ){
	fprintf( stderr,
		 "WARNING: failed to connect term %s to the left edge\n", 
		 net->termName );
    } else if( net->edge&R_EDGE ){
	fprintf( stderr,
		 "WARNING: failed to connect term %s to the right edge\n", 
		 net->termName );
    } else {
	fprintf( stderr,
		 "WARNING: failed to connect term %s to an edge\n", 
		 net->termName );
    }

    /* place a formal terminal at a rational place in the array		*/

    if(( net->network != NIL( s_network ))&&
       ( termloc.x == net->network->col )){
	oFterm = newTerm( M1TERM );
	area.lowerLeft.y = area.upperRight.y = 3 * termloc.y + 2;
	area.lowerLeft.x = area.upperRight.x = termloc.x;
	if( net->network->type == NAND ){
	    mkAttach( NandOutput, &area, oFterm );
	} else {
	    mkAttach( NorOutput, &area, oFterm );
	}
	newFterm( net, oFterm );
	return;
    } 
    if( net->signal->numTerms > 0 ){
	oFterm = newTerm( M2TERM );
	signal->term = REALLOC( s_termList, signal->term, ++signal->numTerms );
	for( index = signal->numTerms - 1; index >= 0; index-- ){
	    if(( index == 0 )||
	       ( signal->term[ index - 1 ].col <= termloc.x )){

		signal->term[ index ].col = termloc.x;
		signal->term[ index ].objectId = oFterm->objectId;
		break;
	    } else {

		signal->term[ index ] = signal->term[ index - 1 ];
	    }
	}
	if( net->edge&UPPER ){
	    area.lowerLeft.y = area.upperRight.y = 3 * termloc.y + 4;
	} else {
	    area.lowerLeft.y = area.upperRight.y = 3 * termloc.y + 2;
	}
    } else {
	oFterm = newTerm( POLYTERM );
	area.lowerLeft.y = area.upperRight.y = 3 * termloc.y + 3;
    }
    area.lowerLeft.x = area.upperRight.x = termloc.x;
    if(( Matrix[ termloc.y ][ termloc.x ].network == NIL( s_network ))||
       ( Matrix[ termloc.y ][ termloc.x ].network->type == NAND )){
	mkAttach( NorOutput, &area, oFterm );
    } else {
	mkAttach( NandOutput, &area, oFterm );
    }
    newFterm( net, oFterm );
}
