#include "electra.h"
/*
 */
readTermConstraints()
{
    octObject oContents;
    octObject oInterface;
    octObject oNet;
    octObject oFterm;
    octObject oEdge;
    octObject oRelPos;
    octGenerator gFterm;
    s_net *net;
    s_net *getNet();


    ohUnpackDefaults( &oContents, "a", ":logic:contents" );
    ohUnpackFacetName( &oContents, InFacetName );
    if( octOpenFacet( &oContents ) < OCT_OK ){
	fprintf(stderr,"Could not open %s: %s - %s\n",
	ohFormatName( &oContents ),
	octErrorString(),
	"no formal terminals will be created" );
	return;
    }
    ohUnpackDefaults( &oInterface, "a", ":logic:interface" );
    ohUnpackFacetName( &oInterface, InFacetName );
    if( octOpenFacet( &oInterface ) < OCT_OK ){
	fprintf(stderr,"Could not open %s: %s - %s\n",
	ohFormatName( &oInterface ),
	octErrorString(),
	"no terminal constraints will be read" );
	return;
    }

    oEdge.type = OCT_PROP;
    oEdge.contents.prop.name = "TERM_EDGE";
    oRelPos.type = OCT_PROP;
    oRelPos.contents.prop.name = "TERM_RELATIVE_POSITION";

    octInitGenContents( &oContents, OCT_TERM_MASK, &gFterm );
    while( octGenerate( &gFterm, &oFterm ) == OCT_OK ){

	if (!octIdIsNull(oFterm.contents.term.instanceId)) continue;
	octGenFirstContainer( &oFterm, OCT_NET_MASK, &oNet );

	net = getNet( oNet.contents.net.name );
	if( net == NIL(s_net)){
	    (void) sprintf(ErrBuffer,"Could not match formal term %s to a net\n",
			       oFterm.contents.term.name );
	    internalError( ErrBuffer);
	}
	net->status |= EXTERNAL;
	net->termName = util_strsav( oFterm.contents.term.name );

	if (octIdIsNull(oInterface.objectId)) continue;
	octGetByName( &oInterface, &oFterm );

	if(( octGetByName( &oFterm, &oEdge ) == OCT_OK )&&
	   ( oEdge.contents.prop.type == OCT_STRING )&&
	   ( octGetByName( &oFterm, &oRelPos ) == OCT_OK )&&
	   ( oRelPos.contents.prop.type == OCT_REAL )){
	    if( strcmp( oEdge.contents.prop.value.string, "RIGHT") == 0){

		net->status |= CONSTRAINED;
		net->bestloc =  NumRows + NumCols - 
		    (int)(NumRows * oRelPos.contents.prop.value.real );

	    } else if( strcmp( oEdge.contents.prop.value.string, "LEFT") == 0){

		net->status |= CONSTRAINED;
		net->bestloc =  NumRows + 2 * NumCols +
		    (int)(NumRows * oRelPos.contents.prop.value.real );

	    } else if( strcmp( oEdge.contents.prop.value.string, "TOP") == 0){

		net->status |= CONSTRAINED;
		net->bestloc =  
		    (int)(NumCols * oRelPos.contents.prop.value.real );

	    } else if( strcmp( oEdge.contents.prop.value.string,"BOTTOM") == 0){
		net->status |= CONSTRAINED;
		net->bestloc =  NumRows + 2 * NumCols - 
		    (int)(NumCols * oRelPos.contents.prop.value.real );
	    } else {
		if( Debug > 0 ){
		    fprintf(stderr,
			"Could not understand % constraint on term %s\n",
			oEdge.contents.prop.value.string,
			oFterm.contents.term.name);
		}
		continue;
	    }
	    if( Debug > 0 ){
		fprintf(stderr,"Term %s constrained to %1.2g on the %s edge\n",
			       oFterm.contents.term.name,
			       oRelPos.contents.prop.value.real,
			       oEdge.contents.prop.value.string );
	    }
	}
    }
    octCloseFacet( &oContents );
    if (!octIdIsNull(oInterface.objectId))
	octCloseFacet( &oInterface );
}

/* 
 * constrainTerm
 *
 * write a terminal constraint so that SPARCS puts the term in the 
 * edge of the array
 */
constrainTerm( edge, oInst )
unsigned edge;
octObject *oInst;
{
    static unsigned edgeBagMask; 
    static octObject oLeft;
    static octObject oRight;
    static octObject oTop;
    static octObject oBottom;
    static octObject oMin;
    static octObject oMax;
    octObject oConstraint;

    oMin.type = OCT_PROP;
    oMin.contents.prop.name = "MIN";
    oMin.contents.prop.type = OCT_INTEGER;
    oMin.contents.prop.value.integer = 0;

    oMax = oMin;
    oMax.contents.prop.name = "MAX";

    oConstraint.type = OCT_BAG;
    if(edge&L_EDGE){
	if( (edgeBagMask&L_EDGE) != L_EDGE ){
	    edgeBagMask |= L_EDGE;
	    oConstraint.contents.bag.name = "XCONSTRAINTS";
	    octGetOrCreate( &oFacet, &oConstraint );
	    oLeft.type = OCT_BAG;
	    oLeft.contents.bag.name = "LEFT";
	    octCreate( &oConstraint, &oLeft );
	}
	oConstraint.contents.bag.name = "CONSTRAINT";
	octCreate( &oLeft, &oConstraint );
	octAttach( &oConstraint, oInst );
	octCreate( &oConstraint, &oMin );
	octCreate( &oConstraint, &oMax );
    }
    if(edge&R_EDGE){
	if( (edgeBagMask&R_EDGE) != R_EDGE ){
	    edgeBagMask |= R_EDGE;
	    oConstraint.contents.bag.name = "XCONSTRAINTS";
	    octGetOrCreate( &oFacet, &oConstraint );
	    oRight.type = OCT_BAG;
	    oRight.contents.bag.name = "RIGHT";
	    octCreate( &oConstraint, &oRight );
	}
	oConstraint.contents.bag.name = "CONSTRAINT";
	octCreate( &oRight, &oConstraint );
	octAttach( &oConstraint, oInst );
	octCreate( &oConstraint, &oMin );
	octCreate( &oConstraint, &oMax );
    }
    if(edge&T_EDGE){
	if( (edgeBagMask&T_EDGE) != T_EDGE ){
	    edgeBagMask |= T_EDGE;
	    oConstraint.contents.bag.name = "YCONSTRAINTS";
	    octGetOrCreate( &oFacet, &oConstraint );
	    oTop.type = OCT_BAG;
	    oTop.contents.bag.name = "TOP";
	    octCreate( &oConstraint, &oTop );
	}
	oConstraint.contents.bag.name = "CONSTRAINT";
	octCreate( &oTop, &oConstraint );
	octAttach( &oConstraint, oInst );
	octCreate( &oConstraint, &oMin );
	octCreate( &oConstraint, &oMax );
    }
    if(edge&B_EDGE){
	if( (edgeBagMask&B_EDGE) != B_EDGE ){
	    edgeBagMask |= B_EDGE;
	    oConstraint.contents.bag.name = "YCONSTRAINTS";
	    octGetOrCreate( &oFacet, &oConstraint );
	    oBottom.type = OCT_BAG;
	    oBottom.contents.bag.name = "BOTTOM";
	    octCreate( &oConstraint, &oBottom );
	}
	oConstraint.contents.bag.name = "CONSTRAINT";
	octCreate( &oBottom, &oConstraint );
	octAttach( &oConstraint, oInst );
	octCreate( &oConstraint, &oMin );
	octCreate( &oConstraint, &oMax );
    }
}
