#include "electra.h"
/*
 * assignTracks
 *
 * Assigns signal tracks to be UPPER or LOWER.  Some signals are constrained
 * because they have terminals located at the output node of a network.  
 * In this case, if the network->col is even, the signal must be LOWER
 * and visa versa.
 *
 * This routine also calls creatPolyM2Term on the signals to create the
 * polyM2Terms specified in makeSignals
 */
assignSignalTracks()
{
    s_net *net;
    s_signal *signal;
    s_network *network;
    s_signal *sigptr;
    unsigned mask;
    int row, col, index;
    int left, right; 		/* bounds for setting the metal2mask	*/

    /* assign tracks for constrained signals				*/
    for( net = Netlist; net != NIL( s_net ); net = net->next ){
	network = Matrix[ net->termloc.y ][ net->termloc.x ].network;
	if(( network != NIL(s_network))&&
	   ( network != net->network )&&
	   ( net->termloc.y == network->firstTran )){
	    if(( net->termloc.x%2 ) == 0 ) 
		net->edge |= LOWER;
	    else        
		net->edge |= UPPER;
	}
    }
    for( row = NumRows - 1; row >= 0; row-- ){

	/* assign tracks to every signal in the row			*/

	for( signal = SignalArray[ row ]; signal != NIL( s_signal );
	     signal = signal->next ){
	    /* determine the track for the main part or the signal	*/
	    if( signal->numTrans == 0 )continue;
	    if(( signal->net->edge&(UPPER|LOWER)) == 0 ){
		if( signal->net->edge&L_EDGE ){

		    /* use the unused track of the prev signal		*/
		    /* else the track is LOWER				*/
		    for( sigptr = signal->prev; sigptr != NIL( s_signal );
			 sigptr = sigptr->prev ){
			if( sigptr->net->edge&LOWER ){
			    signal->net->edge |= UPPER;
			    break;
			} else if( sigptr->net->edge&UPPER ){
			    signal->net->edge |= LOWER;
			    break;
			}
		    }
		    if(( signal->net->edge&(UPPER|LOWER)) == 0 ){
			signal->net->edge |= LOWER;
		    }

		} else if( signal->net->edge&R_EDGE ){

		    /* use the unused track of the next assigned signal */
		    /* use the track of the prev signal 		*/
		    /* else the track is LOWER				*/

		    for( sigptr = signal->next; sigptr != NIL( s_signal );
			 sigptr = sigptr->next ){
			if( sigptr->net->edge&LOWER ){
			    signal->net->edge |= UPPER;
			    break;
			} else if( sigptr->net->edge&UPPER ){
			    signal->net->edge |= LOWER;
			    break;
			}
		    } 
		    if( sigptr == NIL( s_signal )){
			if(( signal->prev != NIL( s_signal ))&&
			   ( signal->prev->net->edge&UPPER )){
			    signal->net->edge |= UPPER;
			} else {
			    signal->net->edge |= LOWER;
			}
		    }
		} else if ( signal->numTerms > 0 ) {

		    /* this is the nasty case - look at the code	*/

		    if(( signal->prev != NIL( s_signal ))&&
		       ( signal->prev->net->edge&R_EDGE )){
		       if( signal->prev->net->edge&UPPER ){
			    signal->net->edge |= LOWER;
		       } else {
			    signal->net->edge |= UPPER;
		       }
		    } else if(( signal->next != NIL( s_signal ))&&
		              ( signal->next->net->edge&(L_EDGE|UPPER))){
			signal->net->edge |= LOWER;
		    } else if(( signal->next != NIL( s_signal ))&&
		              ( signal->next->net->edge&(L_EDGE|LOWER))){
			signal->net->edge |= UPPER;
		    } else if(( signal->prev != NIL( s_signal ))&&
		              ( signal->prev->net->edge&UPPER )){
			signal->net->edge |= UPPER;
		    } else {
			signal->net->edge |= LOWER;
		    }
		}
	    }
	    /* call createPolyM2Term on all POLYM2 terms in the signal	*/
	    index = signal->numTerms;
	    while( --index >= 0 ){
		if (!octIdIsNull(signal->term[ index ].objectId)) continue;
		signal->term[ index ].objectId = 
		    createPolyM2Term( signal->row,
				      signal->term[ index ].col,
				      signal->net->edge&UPPER );
	    }
	}

	/* update the metalMatrix blockages				*/

	for( signal = SignalArray[ row ]; signal != NIL( s_signal );
	     signal = signal->next ){
	    mask = signal->net->edge&(UPPER|LOWER );
	    if( signal->numTerms > 0 ){
		left = signal->term[0].col;
		right = signal->term[ signal->numTerms - 1 ].col;
	    } else if( signal->net->edge&L_EDGE ){
		left = signal->left;
		right = signal->net->termloc.x;
	    } else if( signal->net->edge&R_EDGE ){
		left = signal->net->termloc.x;
		right = signal->right;
	    } else {
		continue;
	    }
	    for( col = left; col <= right; col++ ){
		Matrix[row][col].metal2Mask |= mask;
	    }
	}
    }
}
