#include "electra.h"
/*
 * wireSignals
 *
 * calls mkWire to wire the signals
 *
 */
wireSignals()
{
    int row;
    int index;
    s_signal *signal;
    s_polyrun *polyrun;
    octBox area;
    octObject oTerm1, oTerm2;

    /* wire the poly signals by pattern					*/
    for( row = NumRows - 1; row >= 0; row-- ){
	for( signal = SignalArray[ row ]; signal != NIL( s_signal );
	     signal = signal->next ){
		
	    /* wire the poly by pattern					*/
	    area.lowerLeft.y = area.upperRight.y = 3 * signal->row + 3;
	    if( signal->numRuns == 0 ){

		/* this signal is run totally in poly			*/

		if( signal->left == 0 ){
		    area.lowerLeft.x = 0;
		} else {
		    area.lowerLeft.x = 5 * signal->left + 2;
		}
		if( signal->right == NumCols - 1 ){
		    area.upperRight.x = TotalCols - 1;
		} else {
		    area.upperRight.x = 5 * signal->right + 5;
		}
		mkWireArea( FullArray, &area, SIGNAL_RULE_NAME );

	    } else {

		/* wire each polyrun					*/

		for( index = 0; index < signal->numRuns; index++ ){
		    polyrun = &signal->polyrun[ index ];
		    if( polyrun->left == 0 ){
			area.lowerLeft.x = 0;
		    } else {
			area.lowerLeft.x = 5 * polyrun->left + 2;
		    }
		    if( polyrun->right == NumCols - 1 ){
			area.upperRight.x = TotalCols - 1;
		    } else {
			area.upperRight.x = 5 * polyrun->right + 5;
		    }
		    mkWireArea( FullArray, &area, SIGNAL_RULE_NAME );
		}
	    }

	    /* create segments between any terms 			*/

	    if( signal->numTerms > 0 ){

		oTerm1.objectId = signal->term[0].objectId;
		octGetById( &oTerm1 );
		for( index = 1; index < signal->numTerms; index++ ){
		    oTerm2.objectId = signal->term[ index ].objectId;
		    octGetById( &oTerm2 );
		    mkNewSegment( FullArray, &oTerm1, &oTerm2, NIL( char ), 
				  MK_MIN_WIDTH, MK_HOR_SEG);
		    oTerm1 = oTerm2;
		}
	    }
	}
    }
}
