#!/bin/csh -f
if ($#argv < 1) goto usage

# Prepare a banner.
set B = "gem: "


if ( $?OCTTOOLS == 0 ) then
	echo "Environment variable OCTTOOLS must be defined"
	exit 1
endif

# gem cannot be called by itself with VOV, vov_gem should be called instead.
if ( $?VOV_HOST_NAME ) then
	echo $B "Tracing of gem is not possible at this time"
	echo $B "Use vov_gem instead if you want tracing."
	echo $B
	unsetenv VOV_HOST_NAME
endif

#default options for tools in GEM
set TECHDIR = "$OCTTOOLS/tech/scmos/gem"
set MISARG =  "-f $OCTTOOLS/lib/gem/gem.script -T oct"
set GENIEARG = "-w -netl 6 -tech $TECHDIR/genie.tech"
set GENIEARG_RERUN = "-w -tech $TECHDIR/genie.tech"
set ELECTRAARG = "-d 0 -s -c $TECHDIR"
set SPARCSARG = "-t -w PLACE -w NWEL"
#set JUMBOARG = "-vt -a 2 -N 40"
set PUPPYARG = "-t"
set PADPARG = "-P -i"
set MOSAICOARG = ""

#names of the programs invoked in this script
set MIS = "misII"
set GENIE = "genie"
set ELECTRA = "electra"
set SPARCS = "sparcs"
#set JUMBO = "~kring/src/jumbo/jumbo"
set OCTMVLIB = "octmvlib"
set PUPPY = "puppy"
set PADP = "padplace"
set MOSAICO = "mosaico"
set OCTPROP = "octprop"
set PARPUPPY1 = "$TECHDIR/parPuppy1"
set PARPUPPY2 = "$TECHDIR/parPuppy2"

set ROWS = -1
set COLS = -1
set BINSIZE = 100
set RUNMIS

@ argnum = 1
while ( $argnum < $#argv )
	if ( x$argv[$argnum] == "x-r" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set ROWS = `expr $argv[$argnum]`
	else if ( x$argv[$argnum] == "x-c" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set COLS = `expr $argv[$argnum]`
	else if ( x$argv[$argnum] == "x-b" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo $B -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set BINSIZE = `expr $argv[$argnum]`
	else if ( x$argv[$argnum] == "x-nomis" ) then
		unset RUNMIS
	else if ( x$argv[$argnum] == "x-oct" ) then
		set OCTINPUT
	else if ( x$argv[$argnum] == "x-p" ) then
		set PART
	else if ( x$argv[$argnum] == "x-f" ) then
		set JUMBOARG = "-f $JUMBOARG"
	else if ( x$argv[$argnum] == "x-a" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo $B -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set ASPECT_RATIO = `expr $argv[$argnum] "*" 2`
	else if ( x$argv[$argnum] == "x-g" ) then
		set GRAPHICS
	else if ( x$argv[$argnum] == "x-st" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo $B -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set START = $argv[$argnum]
	else goto usage
	endif
	echo $B -n "$argv[$argnum] "
	@ argnum = $argnum + 1
end
if( $?GRAPHICS == 0 ) set PUPPYARG = "-g $PUPPYARG"


set CELL = $argv[$#argv]

set WD = gemtmp.$$
mkdir $WD
cp $CELL.blif $WD
cd $WD


echo $B This is the GEM module generator. Version 2.0

if( $?START ) then
	echo $B restarting at $START
	goto $START
endif

#logic optimization
mis:
if( $?RUNMIS && $?OCTINPUT ) then
	echo $B input = $CELL\:optlogic output = $CELL\:gem
	time $MIS $MISARG -toct -o $CELL\:optlogic $CELL\:optlogic || exit
else if( $?RUNMIS ) then
	echo $B input = $CELL.blif output = $CELL\:gem
	time $MIS $MISARG -o $CELL\:optlogic $CELL.blif || exit
else
	echo $B input = $CELL\:optlogic output = $CELL\:gem
endif

#partitioning the cell
jumbo:
if( $?PART == 0 ) goto single_genie
    $JUMBO $JUMBOARG -r $ROWS -c $COLS -s $BINSIZE -p :optlogic -o :puppy $CELL\:optlogic
endif

#folding the cell
genie:
set SUBCELLS = `$OCTPROP -v JUMBO_PARTITIONS $CELL\:puppy`
if( "$SUBCELLS" == "no_partitions" )goto single_genie
foreach SUBCELL ( `eval echo $SUBCELLS` )
        echo $B running genie on $SUBCELL
	set CWD = $cwd
	cd $SUBCELL
	echo 1 10 0 | $GENIE $GENIEARG -i -octr $CWD/$SUBCELL ||\
	echo 1 5 0 | $GENIE $GENIEARG -i -octr $CWD/$SUBCELL ||\
	echo 1 10 0 | $GENIE $GENIEARG_RERUN -i -octr $CWD/$SUBCELL ||\
	echo 1 5 0 | $GENIE $GENIEARG_RERUN -i -octr $CWD/$SUBCELL ||\
	$GENIE $GENIEARG_RERUN -octr $CWD/$SUBCELL || exit
        cd $CWD
end
#running electra -F on the partitions
electra1:
set SUBCELLS = `$OCTPROP -v JUMBO_PARTITIONS $CELL\:puppy`
if( "$SUBCELLS" == "no_partitions" )goto single_electra
foreach SUBCELL ( `eval echo $SUBCELLS` )
        echo $B running electra floorplanning preparation on $SUBCELL
	$ELECTRA $ELECTRAARG -F -i $SUBCELL/final $SUBCELL\:optlogic || exit
end

puppy1:
#running puppy to constrain the subcells terminals
echo $B floorplanning the cell with puppy
cp $PARPUPPY1 parPuppy
$PUPPY $PUPPYARG $CELL\:puppy ||exit

electra2:
set SUBCELLS = `$OCTPROP -v JUMBO_PARTITIONS $CELL\:puppy`
foreach SUBCELL ( `eval echo $SUBCELLS` )
        echo creating $SUBCELL with electra and sparcs
	$ELECTRA $ELECTRAARG -i $SUBCELL/final -o :gem $SUBCELL\:optlogic || exit
	$SPARCS $SPARCSARG -o $SUBCELL\:gem $SUBCELL\:gem  || exit
end

#reinstanciating the subcells as the gem view
octmvlib:
$OCTMVLIB :optlogic :gem $CELL\:puppy || exit

puppy2:
#running puppy as a last optimization and to ensure that there is no overlap
cp $PARPUPPY2 parPuppy
time $PUPPY -t $PUPPYARG $CELL\:puppy || exit

padp:
#creating pads
time $PADP $PADPARG $CELL\:puppy || exit

mosaico:
#mosaico place and route
time $MOSAICO $MOSAICOARG -o gem $CELL\:puppy || exit
 
exit

usage:
echo $B usage\: "$0 [-nomis] [-oct] cell"
exit
single_genie:

	echo 1 10 0 | $GENIE $GENIEARG -i -octr $CELL ||\
	echo 1 5 0 | $GENIE $GENIEARG -i -octr $CELL ||\
	echo 1 10 0 | $GENIE $GENIEARG_RERUN -i -octr $CELL ||\
	echo 1 5 0 | $GENIE $GENIEARG_RERUN -i -octr $CELL ||\
	$GENIE $GENIEARG_RERUN -octr $CELL || exit
single_electra:
	$ELECTRA $ELECTRAARG -o :gem $CELL\:optlogic || exit
	$SPARCS $SPARCSARG -o $CELL\:gem $CELL\:gem || exit


cp -r $CELL/gem ../$CELL
cd ..
rm -rf $WD

echo $B "Output is now in ${CELL}:gem"
exit 0
