#include "copyright.h"
#include "genie.h"


/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/


inner_loop_criterion(num_states_per_stage,num_states_generated)
int num_states_per_stage,num_states_generated;
{

/*
*  if number of states generated is high enough or the cost function
*  is high quit
*/
    if (num_states_per_stage < num_states_generated)
       return(TRUE);
    else {
    /*   if (num_states_generated > 100  &&  starting_cost < init_cost/5 )
          return(TRUE);         
        else   */
          return(FALSE);
    }

} /* end of inner_loop_criterion */


stopping_criterion(cost1,cost2,cost3,cost4,stop_temp,temp)
int cost1,cost2,cost3,cost4;
double stop_temp,temp;
{


  if (stop_temp < temp )
     return(FALSE);  
  if ( temp < 1e-4 || (cost1 is cost2 && cost1 is cost3 && cost1 is cost4 ) )
    return(TRUE);
  else
    return(FALSE);

} /* end of stopping_criterion */


read_tech_file(tech,tFile)
int tech;
char *tFile;
{

  FILE *Tech;
  char line[MAXLINE];
  char *techFile;
  char stdName[ID_LENGTH];
  STDCELL std,findStd();
  struct plist **HAcreate();


   techFile = ALLOC(char,MAXLINE);
   if (tech is TRUE)
     (void) strcpy(techFile,tFile);
   else
     (void) sprintf(techFile, "%s/lib/genie.tech", util_file_expand( "$OCTTOOLS" ));
   if ((Tech = fopen(techFile,"r")) is NULL) {
	(void) fprintf( stderr,"Couldn't open tech file %s\n", techFile );
	exit(-1);
   }
   Stdhash = HAcreate( STDCELLHASH );
   NextLine = 0;
   while(fgets(line,MAXLINE,Tech) isnt NULL) {
     NextLine++;
     parse_line(line);
     if (targv[0] is NULL) continue;
     if (targv[0][0] is '#') continue;
     if (targc isnt 3) 
       error("in stdfile");
     (void) strcpy(stdName,targv[0]);
     std = findStd(stdName,TRUE);
     std->l = atoi(targv[1]);
     std->w = atoi(targv[2]);
   }
   FREE(techFile);
  return;

} /* end of read_tech_file */


STDCELL
findStd(stdName,install)
/*
 * check for presence of stdcell, else allocate one.
 *
 */
STRING stdName;
int install;
{
    STDCELL np;
    STRING loc, hname;
 
    if(HAfind(stdName, &loc, Stdhash, STDCELLHASH) != FALSE) {
        np = (struct stdcell *)loc;
        return (np);
    }
    if (install is FALSE)  {
      np = NIL(struct stdcell);
      return(np);
    }
    np = ALLOC(struct stdcell, 1);
    if(HAinstall(stdName,(char *)np,&hname,ELMEN,Stdhash,STDCELLHASH) != OK) {
        error("stdcell installation failed");
    }

    std_num++ ; 
    stds[std_num] = np;
    np->num = std_num;
    np->stdName = hname;
    return(np);
  

} /*  end of findStd  */


alloc_stdcells()
{

  STDCELL std,findStd();
  int i;
  CELL cell;
  CELL_PTR cptr;
  char stdName[ID_LENGTH];


   for(i=1;i<=gate_num;i++) {
     cell = cells[i];
     stdName[0] = cell->ch;
     stdName[1] = '\0';
     std = findStd(stdName,FALSE);
     if (std is NIL(struct stdcell)) {
       (void) fprintf(stderr,"Tile %s not defined in genie.tech technology file\n",stdName);
       error("couldn't find stdcell");
     }
     cell->stdType = std;
   }
   for(i=1;i<=net_num;i++) {
     nets[i]->size = 0;
     if (nets[i]->hor is TRUE) {
       for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                     cptr=cptr->nextPtr) {
         if (cptr->clink->type < PADLEFT)
           nets[i]->size = MAX(cptr->clink->stdType->l,nets[i]->size);
       }
     }
     else {
       for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                     cptr=cptr->nextPtr) 
         nets[i]->size = MAX(cptr->clink->stdType->w,nets[i]->size);
     }
   }
     
  return;

} /* end of alloc_stdcells */


#define FUN_LENGTH 500


compute_numcols(net,type)
NET net;
int type;
{
/* 
* routine which computes the number of columns to realize a net given
* the function the net is supposed to implement and the ordering of
* signals to the net.
*/
 int top[100],bot[100];
 SIG_PTR sptr;
 NET findNet(),signet;
 int i,numco,level,token_num;
 char funName[FUN_LENGTH];


  if (net->complex is FALSE) return(1);
  if (type is 0) {
    (void) strcpy(funName,net->fName);
    parse_line(funName);
    if (targc < 3) crash("not a complex column");
    signet = findNet(targv[0]);
    sptr = ALLOC(struct sig_ptr, 1);
    sptr->num = signet->num;
    sptr->level = 1;
    sptr->ch = 'o';
    net->sig_list = sptr;
    if (strcmp(targv[1],"(") isnt FALSE) 
      crash("funName corrupted");
    if (strcmp(targv[2],"s") is FALSE)
      type = SERIES;
    else if (strcmp(targv[2],"p") is FALSE)
      type = PARALLEL;
    else
      crash("type corrupted");
      
    token_num = 2; level = 0;
    compute_level(&token_num,&level,net,type);
    for(i=1;i<=level;i++) {
      top[i] = 0;
      bot[i] = INFINITY;
    }
    for(sptr=net->sig_list;sptr!=NIL(struct sig_ptr);sptr=sptr->nextPtr) {
      top[sptr->level] = MAX(nets[sptr->num]->row,top[sptr->level]);
      bot[sptr->level] = MIN(nets[sptr->num]->row,bot[sptr->level]);
    }
    net->level = numco = level;
    for(i=1;i<level;i++) {
      if (top[i] < bot[i+1]) {
        numco--;
        bot[i+1] = bot[i];
      }
      else if (bot[i] > top[i+1]) {
        numco--;
        top[i+1] = top[i];
      }
    }

  } else if (type is INTHORNET) {
    for(i=1;i<=net->level;i++) {
      top[i] = 0;
      bot[i] = INFINITY;
    }
    if (shift_net2 is NIL(struct net)) {
      for(sptr=net->sig_list;sptr!=NIL(struct sig_ptr);sptr=sptr->nextPtr) {
        if (sptr->num is shift_net1->num) {
          top[sptr->level] = MAX(new_row,top[sptr->level]);
          bot[sptr->level] = MIN(new_row,bot[sptr->level]);
        }
        else {
          top[sptr->level] = MAX(nets[sptr->num]->row,top[sptr->level]);
          bot[sptr->level] = MIN(nets[sptr->num]->row,bot[sptr->level]);
        }
      }
    }
    else {
      for(sptr=net->sig_list;sptr!=NIL(struct sig_ptr);sptr=sptr->nextPtr) {
        if (sptr->num is shift_net1->num) {
          top[sptr->level] = MAX(shift_net2->row,top[sptr->level]);
          bot[sptr->level] = MIN(shift_net2->row,bot[sptr->level]);
        }
        else if (sptr->num is shift_net2->num) {
          top[sptr->level] = MAX(shift_net1->row,top[sptr->level]);
          bot[sptr->level] = MIN(shift_net1->row,bot[sptr->level]);
        }
        else {
          top[sptr->level] = MAX(nets[sptr->num]->row,top[sptr->level]);
          bot[sptr->level] = MIN(nets[sptr->num]->row,bot[sptr->level]);
        }
      }
    }
    numco = level = net->level;
    for(i=1;i<level;i++) {
      if (top[i] < bot[i+1]) {
        numco--;
        bot[i+1] = bot[i];
      }
      else if (bot[i] > top[i+1]) {
        numco--;
        top[i+1] = top[i];
      }
    }
  }

  return(numco);

} /* end of compute_numcols */



compute_level(num,level,net,type)
int *num,*level;
NET net;
int type;
{
/*
* compute the level of every signal feeding into a complex column
* recursively. The level is the column which the transistor corresponding
* to the signal will reside in the worst case.
*/
  int tlevel,first;
  NET hornet,findNet();
  SIG_PTR sptr;

   
   first = TRUE;
   while( *num < targc-1 ) {
     (*num)++;
     if (strcmp(targv[(*num)],"(") is FALSE) {
       (*num)++;
       if ( *num >= targc-1 ) crash("exceeded count");
       if (strcmp(targv[(*num)],"p") is FALSE )
         compute_level(num,level,net,PARALLEL);
       else if (strcmp(targv[(*num)],"s") is FALSE)
         compute_level(num,level,net,SERIES);
       else 
         crash("unknown nature of gate");
     }
     else if (strcmp(targv[(*num)],")") is FALSE) {
       return;
     }
     else {
       hornet = findNet(targv[(*num)]);
       if (first is TRUE) {
         (*level)++; tlevel = *level;
         first = FALSE;
       }
       sptr = ALLOC(struct sig_ptr, 1);
       sptr->num = hornet->num;
       sptr->level = tlevel;
       if (type is SERIES) {
         sptr->ch = 's';
       }
       else if (type is PARALLEL) {
         sptr->ch = 'p';
       }
       else {
         crash("unknown type");
       }
       sptr->nextPtr = net->sig_list;
       net->sig_list = sptr;
     }
   }
 return;

} /* end of  compute_level */
       

#ifndef lint
print_numcols()
{

 int i;

  for(i=1;i<=net_num;i++)
    if (nets[i]->complex is TRUE)
      (void) fprintf(stdout,"%s %d\n",nets[i]->netName,nets[i]->num_cols);

 return;


} /* end of print_numcols */
#endif

calc_numcols()
{

 int i,comp_cols;

  comp_cols = 0;
  shift_net1 = shift_net2 = NIL(struct net);
  for(i=1;i<=net_num;i++)
    if (nets[i]->complex is TRUE)
      comp_cols += compute_numcols(nets[i],INTHORNET);

 return(comp_cols);

} /* end of calc_numcols */
