#include "copyright.h"
#include "genie.h"
#include <sys/times.h>


/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/


main(argc,argv)
int argc;
char **argv;
{
  int i,j;
  struct plist **HAcreate();
  char inpFile[ID_LENGTH], plaInFile[ID_LENGTH],tFile[ID_LENGTH];
  char octInFile[ID_LENGTH],octOutFile[ID_LENGTH];
  double atof();
  int tech;
  int nofold = FALSE;

  timer(stdout,"START");
  (void) strcpy(inpFile,"xxx");
/*
*  default gatetypes
*/
  gatetype[1] = AND;
  gatetype[2] = OR;
  debug = pla = expand = specify = varwidth = simplecol = FALSE;
  rowonly = colonly = statcmos = statnmos = domino = dogleg = FALSE;
  rowpreferred = nopads = TRUE;twolevel = FALSE; ordercon = FALSE;
  boundcon = FALSE;
  simplerow = FALSE;
  no_output = FALSE;
  
  /* default annealing parameters				*/
  STATES_BEGIN = 1.0;
  STATES_END = 10.0;
  ROW_PARAM = 5;
  COL_PARAM = 15;
  WID_LEN_PARAM = 0;
  COMPLEX_PARAM = 5;
  NET_OVERLAP_PARAM = 3.0;
  OVERLAP_OFFSET = 50;
  OVERLAP_PARAM = 50;

#ifdef THREED
   threed = TRUE;
#else
   threed = FALSE;
#endif

  octwrite = FALSE;octread = FALSE;gateread = FALSE;
  complex = TRUE; plaread = FALSE; mgread = FALSE;
  octmread = FALSE; tech = FALSE;
  NET_LENGTH_PARAM = 10.0;
  i = 1;
  while(i< argc) {
    if (strcmp(argv[i],"-octr") is 0) {
      octread = TRUE;
      (void) strcpy(octInFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-netl") is 0) {
      NET_LENGTH_PARAM = atof(argv[i+1]);
      i+=2;
    }
    else if( strcmp(argv[i], "-fT") == 0 ){
      STATES_END = atof(argv[i+1]);
      i +=2;
    }
    else if( strcmp(argv[i], "-sT") == 0 ){
      STATES_BEGIN = atof(argv[i+1]);
      i +=2;
    }
    else if (strcmp(argv[i],"-octmr") is 0) {
      octmread = TRUE; mgread = TRUE;
      (void) strcpy(octInFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-pla") is 0) {
      plaread = TRUE;
      (void) strcpy(plaInFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-no_output") is 0) {
      no_output = TRUE;
      i+=1;
    }
    else if (strcmp(argv[i],"-tech") is 0) {
      tech = TRUE;
      (void) strcpy(tFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-octw") is 0) {
      octwrite = TRUE;
      (void) strcpy(octOutFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-mread") is 0) {
      gateread = TRUE; mgread = TRUE;
      (void) strcpy(inpFile,argv[i+1]);
      i+=2;
    }
    else if (strcmp(argv[i],"-gread") is 0) {
      gateread = TRUE;
      (void) strcpy(inpFile,argv[i+1]);
      i+=2;
    }
    else if (argv[i][0] is '-') {
      j = 1;
      while (argv[i][j] isnt '\0') {
        if (argv[i][j] is 'g') {
          dogleg = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'l') {
          simplecol = TRUE;
          j++ ;
        }
	else if (argv[i][j] is 'n') {
	    nofold = TRUE;
	    j++;
	}
        else if (argv[i][j] is 't') {
          twolevel = TRUE;
          complex = FALSE;
          j++ ;
        }
        else if (argv[i][j] is 'd') {
          debug = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'r') {
          rowonly = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'c') {
          colonly = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'v') {
          varwidth = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'n') {
          nopads = FALSE;
          j++ ;
        }
        else if (argv[i][j] is 'p') {
          pla = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 's') {
          statcmos = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'w') {
          rowpreferred = FALSE;
          j++ ;
        }
        else if (argv[i][j] is 'm') {
          simplerow = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'o') {
          domino = TRUE;
          j++ ;
        }
        else if (argv[i][j] is 'i') {
          specify = TRUE;
          j++ ;
        }
        else
          crash("unrecognized option");
      }
      i++ ;
    }
    else if (strcmp(argv[i],"-g") is 0) {
      if (strcmp(argv[i+1],"NAND") is 0) 
         gatetype[1] = gatetype[2] = NAND;
      if (strcmp(argv[i+1],"NOR") is 0)
         gatetype[1] = gatetype[2] = NOR;
      i+=2;
    }
    else  {
      (void) strcpy(inpFile,argv[i]);
      i++ ;
    }
  }
  if (octread is FALSE || gateread is TRUE) {
    if (strcmp(inpFile,"xxx") is FALSE) {
       input = stdin;
    }
    else if ( (input = fopen(inpFile,"r")) is NULL) {
       (void) fprintf(stderr,"Couldn't open input file\n");
       exit(1) ;
    }
  }
  if (plaread is TRUE) {
    pla_to_genie(plaInFile);
    exit(0);
  }
  if (statcmos is TRUE) rowonly = TRUE;
  if (colonly is TRUE) rowpreferred = FALSE;
  net_num = cell_num = 0; std_num = 0;
  if( specify ) {
    (void) fprintf(stderr,"\nBeginning states per unit (%lf):", STATES_BEGIN);
    (void) fscanf(stdin,"%lf",&STATES_BEGIN);

    (void) fprintf(stderr,"\nEnding states per unit (%lf):", STATES_END);
    (void) fscanf(stdin,"%lf",&STATES_END); 

    (void) fprintf(stderr,"\nRow parameter(%d):",ROW_PARAM );
    (void) fscanf(stdin,"%d",&ROW_PARAM); 

    (void) fprintf(stderr,"\nCol parameter(%d):",COL_PARAM );
    (void) fscanf(stdin,"%d",&COL_PARAM); 

    (void) fprintf(stderr,"\nNet length parameter(%lf):", NET_LENGTH_PARAM);
    (void) fscanf(stdin,"%lf",&NET_LENGTH_PARAM); 

    (void) fprintf(stderr,"\nNet overlap parameter(%lf):", NET_OVERLAP_PARAM);
    (void) fscanf(stdin,"%lf",&NET_OVERLAP_PARAM); 

    (void) fprintf(stderr,"\nCell overlap parameter:(%d)", OVERLAP_PARAM );
    (void) fscanf(stdin,"%d",&OVERLAP_PARAM); 

    (void) fprintf(stderr,"\nCell overlap offset:(%d)", OVERLAP_OFFSET );
    (void) fscanf(stdin,"%d",&OVERLAP_OFFSET); 

/*
    (void) fscanf(stdin,"%d",&COMPLEX_PARAM); 
    (void) fprintf(stdout,"Complex gate parameter:%d\n",COMPLEX_PARAM);
    if (varwidth is TRUE) {
      (void) fprintf(stderr,"Length Width parameter:");
      (void) fscanf(stdin,"%lf",&WID_LEN_PARAM); 
      (void) fprintf(stdout,"Length Width parameter:%lf\n",WID_LEN_PARAM);
    }
*/
  }
  (void) fprintf(stderr,"\n %lf Beginning states per unit\n", STATES_BEGIN);
  (void) fprintf(stderr,"%lf ending states per unito\n", STATES_END);
  (void) fprintf(stderr,"%d row parameter\n",ROW_PARAM );
  (void) fprintf(stderr,"%d col parameter\n",COL_PARAM );
  (void) fprintf(stderr,"%lf net length parameter\n", NET_LENGTH_PARAM);
  (void) fprintf(stderr,"%lf net overlap parameter\n", NET_OVERLAP_PARAM);
  (void) fprintf(stderr,"%d cell overlap parameter\n", OVERLAP_PARAM );
  (void) fprintf(stderr,"%d cell overlap offset\n", OVERLAP_OFFSET );

  Nethash = HAcreate( NETHASH ) ;
  Cellhash = HAcreate( CELLHASH ) ;
/* 
* read in the input matrix and constraints
*/
  if (octread is TRUE || octmread is TRUE) {
#ifndef dist
    oct_read(octInFile);
#endif
    (void) strcpy(inpFile,"genie.log");
    if ( (input = fopen(inpFile,"r")) is NULL) {
       (void) fprintf(stderr,"Couldn't open input file\n");
       exit(1) ;
    }
    read_gates();
    place_nets_and_cells();
    timer(stdout,"for oct read");
  }
  else if (gateread is TRUE) {
    read_gates();
    place_nets_and_cells();
#ifndef dist
    timer(stdout,"for gate read");
#endif
  }
  else {
    read_input_file();
  /*
  * allocate storage for all cells/pads  required
  */
    make_cells();
    if (pla is TRUE) 
      add_connects();
    if (dogleg is TRUE) {
      dogleg_nets();
    }
    add_pads();
/*
    if (simplerow is TRUE) 
      simple_folding_pads();
*/
    place_pads();
#ifndef dist
    timer(stdout,"for matrix read");
#endif
  }
/*
*  get dimensions of the pads/cells
*/
  read_tech_file(tech,tFile);
  alloc_stdcells();
/*
* find initial cost and starting temperature.
*/
/*
  if (ordercon is TRUE)
    construct_acceptable_order(); 
*/
  ready_for_anneal();
  if (dogleg is FALSE) {
    init_cost = find_cost(TRUE,INTHORNET) ;
    find_starting_temp();
/*
    print_lists();
*/
  }
/*
* anneal.
*/
/*
  if (dogleg is TRUE) {
    print_array(strcat(inpFile,"_d"));
    exit(0);
  }
  else
    print_array("init");
  plot_cells("pli");
*/
  if (nofold == FALSE)
      simulated_anneal(); 
#ifndef dist
  timer(stdout,"for simulated annealing");
#endif

/*
* do the non trivial routing
*/ 
  route_annealed_array();
  print_array("final");

  for(i=1;i<=net_num;i++) {
    if (!nets[i]->cell_list) {
      (void) fprintf(stdout, "Warning: net %s has no connections!\n", nets[i]->netName);
      continue;
    }
    if (nets[i]->hor is TRUE)
      nets[i]->row = nets[i]->cell_list->clink->rect.y;
    else
      nets[i]->col = nets[i]->cell_list->clink->rect.x;
  }

#ifndef dist
  if ( (octread is TRUE || gateread is TRUE) && (complex is TRUE) ) {
    (void) fprintf(stdout,"%d columns to realize complex gates initially\n",
                                                                complex_cols);
    complex_cols = calc_numcols();
    (void) fprintf(stdout,"%d columns to realize complex gates finally\n",
                                                                complex_cols);
  }
  if (octwrite is TRUE)
    oct_write(octOutFile);
#endif
  free_storage();

  timer(stdout,"total");
  exit(0);

} /* end of main */

/*
*  parse input line into tokens, filling up targv and setting targc 
*/

parse_line(line)
  register char *line;
  {	register char **carg = targv;
	register char ch;

	targc = 0;
	while (ch = *line++) {
          if ((ch == ';')||(ch == '!' )) break ;
	  if ((ch <= ' ')||(ch == ',')) continue;
	  targc++;
	  *carg++ = line-1;
	  while ((ch = *line) && (ch > ' ' ) && (ch != ';')
                    &&(ch != ',') && (ch != '=')) line++;
	  if (ch) *line++ = '\0';
	}
	*carg = 0;

} /*  end of parse_line  */


/*
 * print error message and die
 */
error(message)
STRING message;
{
    (void) fprintf(stderr,"genie: %s around line %d\n",message,NextLine);
    exit(-1);
}

crash(message)
STRING message;
{
    (void) fprintf(stderr,"genie: %s \n",message);
    exit(-1);
}

read_input_file()
/*
* read in the matrix and store 
*/
{
 int i,j,nr,nc,start_line,buf_line,offset;
 NET nt,findNet();
 char *line;


  NextLine = 0;nr = 0;buf_line = TRUE;
  line = ALLOC (char,MAXLINE);
  while(fgets(line,MAXLINE,input) isnt NULL) {
    NextLine++ ;
    parse_line(line);
    if (targv[0] is NULL) continue;
    if (targv[0][0] is '#' || targv[0][0] is '*') continue;
    if (strcmp(targv[0],"new") is FALSE) {
      num_rows = atoi(targv[1]);
      num_cols = atoi(targv[2]);
      if (statcmos is TRUE || pla is TRUE) {
        if (targc isnt 4 )
          crash("row for P and N division not specified"); 
        else
          middle_row = atoi(targv[3]);
      }
      allocate_storage();
      offset = 2*MIN(MAX(num_rows,num_cols),10);
      continue;
    }
    if (strcmp(targv[0],"left") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->left = TRUE;
      }
      continue;
    }
    if (strcmp(targv[0],"right") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->right = TRUE;
      }
      continue;
    }
    if (strcmp(targv[0],"inter") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->inter = TRUE;
      }
      continue;
    }
    if (strcmp(targv[0],"bound") is FALSE) {
      bounding_constraints();
      continue;
    }
    if (strcmp(targv[0],"ordered") is FALSE) {
      order_constraints();
      continue;
    }
    if (domino is TRUE && buf_line is TRUE) {
      nc = 0;
      buf_line = FALSE ;
      do {
        for(i=0;i< targc;i++) {
          j = 0;
          while(targv[i][j] isnt '\0') {
            nc++;
            if (nc > num_cols)
              crash("more columns than specified");
            pbuffer[1][nc+offset] = targv[i][j];
            pbuffer[2][nc+offset] = ' ' ;
            j++;
          }
        }
      } while ( nc < num_cols) ;
      continue;
    }
    nt = findNet(targv[0]);
    nt->hor = TRUE;
    nr++;
    if (nr > num_rows)
      crash("more rows than specified");
    nt->row = nr;
    if (statcmos is TRUE || pla is TRUE) {
      if (nr <= middle_row)
        nt->type = PCHANNEL;
      else 
        nt->type = NCHANNEL;
    }
    nc = 0;
    start_line = 1;
    do {
      for(i=MAX(0,start_line);i<targc;i++) {
        j = 0;
        while(targv[i][j] isnt '\0') {
          nc++;
          if (nc > num_cols)
            crash("more columns than specified");
            inp_matrix[nc][nr] = targv[i][j];
          j++;
        }
      }
      if ( nc is num_cols) continue;
      if ( fgets(line,MAXLINE,input) is NULL)
        crash("premature end of input");
      else {
         parse_line(line);
         start_line = 0;
      }
    } while ( nc < num_cols) ;
    if (nc > num_cols)
      crash("error in number of columns specified");
  }
  if (num_rows <= 0 || num_cols <= 0) {
    crash("premature end of input");
  }
  hor_net_num = net_num;
 return;

} /* end of read_input_file */


add_net_cell(cell,net)
CELL cell;
NET net;
{

  CELL_PTR cptr;
  NET_PTR nptr;

  nptr = ALLOC(struct net_ptr, 1);
  nptr->nextPtr = cell->net_list;
  cell->net_list = nptr;
  nptr->nlink = net;
  cptr = ALLOC(struct cell_ptr, 1);
  cptr->nextPtr = net->cell_list;
  net->cell_list = cptr;
  cptr->clink = cell;

 return;

} /* end of add_net_cell */



delete_net_cell(cell,net)
CELL cell;
NET net;
/*
*  Deletes the cell from the net.
*/
{

 CELL_PTR cptr,prev_cptr;
 NET_PTR nptr,prev_nptr;

  prev_cptr = cptr = net->cell_list;
  while(cptr isnt NIL(struct cell_ptr)  && cptr->clink isnt cell ) {
    prev_cptr = cptr;
    cptr = cptr->nextPtr;
  }
  if (cptr is NIL(struct cell_ptr))
    error("in delete_net_cell");
  if (prev_cptr is cptr) 
    net->cell_list = net->cell_list->nextPtr;
  else
    prev_cptr->nextPtr = cptr->nextPtr;

  prev_nptr = nptr = cell->net_list;
  while(nptr isnt NIL(struct net_ptr)  && nptr->nlink isnt net ) {
    prev_nptr = nptr;
    nptr = nptr->nextPtr;
  }
  if (nptr is NIL(struct net_ptr))
    error("in deleting net in delete_net_cell");
  if (prev_nptr is nptr) 
    cell->net_list = cell->net_list->nextPtr;
  else
    prev_nptr->nextPtr = nptr->nextPtr;

 return;

} /* delete_net_cell */



NET
findNet(netName)
/*
 * check for presence of net, else allocate one.
 *
 */
STRING netName;
{
    NET np;
    STRING loc, hname;
 
    if(HAfind(netName, &loc, Nethash, NETHASH) != FALSE) {
        np = (struct net *)loc;
        return (np);
    }
    np = ALLOC(struct net, 1);
    if(HAinstall(netName,(char *)np,&hname,ELMEN,Nethash,NETHASH) != OK) {
        crash("net installation failed");
    }

    net_num++ ; 
    nets[net_num] = np;
    np->num = net_num;
    np->netName = hname;
    np->input = FALSE; np->output = FALSE;
    np->right = FALSE; np->left = FALSE; np->inter = FALSE;
    np->hor = FALSE; np->vert = FALSE;
    np->fName = NULL;
    np->cell_list = NIL(struct cell_ptr);
    np->complex = FALSE;
    np->col = np->row = 0;
    np->gate = np->tilda = np->weight = np->level = np->cnt = 0;
    np->proc = np->nnum_cols = np->num_cols = 0;
    np->sig_list = NIL(struct sig_ptr);
    np->stagger = np->size = np->straightened = 0;
    np->bound = FALSE;
    np->tb = 0; np->bb = INFINITY;
    np->ordernum = 0;
    np->type = 0;
    return(np);
  

} /*  end of findNet  */


CELL
findCell(cellName)
/*
 * check for presence of cell, else allocate one.
 *
 */
STRING cellName;
{
    CELL np;
    STRING loc, hname;
 
    if(HAfind(cellName, &loc, Cellhash, CELLHASH) != FALSE) {
        np = (struct cell *)loc;
        return (np);
    }
    np = ALLOC(struct cell, 1);
    if(HAinstall(cellName,(char *)np,&hname,ELMEN,Cellhash,CELLHASH) != OK) {
        crash("cell installation failed");
    }

    cell_num++ ; 
    cells[cell_num] = np;
    np->num = cell_num;
    np->cellName = hname;
    np->type = 0;
    np->net_list = NIL(struct net_ptr);
    np->special = 0;
    np->rect.x = np->rect.y = 0;
    np->xfix = np->yfix = FALSE;
    np->stdType = NIL(struct stdcell);

    return(np);
  

} /*  end of findCell  */


void
get_pad_name(Pname,one)
STRING Pname;
int one;
{

  (void) sprintf(Pname,"pad_%d",one);
 return;

} /* end of get_pad_name  */


void
get_cell_name(Cname,one,two)
STRING Cname;
int one,two;
{

  (void) sprintf(Cname,"cell_%d_%d",one,two);
 return;

} /* end of get_cell_name  */


void
get_net_name(Nname,one)
STRING Nname;
int one;
{

  (void) sprintf(Nname,"net_%d",one);
 return;

} /* end of get_net_name */


/*
 *    get the elapsed cputime
 *        (call it the first time with message "START")
 */ 
struct tms Buffer;
double StartTime, LastTime;

timer( file, message )
FILE *file;
char *message;
{
    if(strcmp(message,"START") == 0) {
        (void) times(&Buffer);
        StartTime = (double)(Buffer.tms_utime);
        LastTime = StartTime;
        return;
    }
    (void) times(&Buffer);
    (void) fprintf(file,"#genie: cputime %s: %.1fs %.1fs\n", message,
        (Buffer.tms_utime-LastTime)/60.0, (Buffer.tms_utime-StartTime)/60.0);
    LastTime = Buffer.tms_utime;
    return;
}
 
allocate_storage()
{

  int i,offset;
  int nnum_cols;

  
   offset = 2*MIN(MAX(num_rows,num_cols),10) ;
   M_SIZE = 2*MIN(MAX(num_rows,num_cols),10) + MAX(num_rows,num_cols) + 5;
   nnum_cols = num_cols;
   if (dogleg is TRUE) {
     offset += 50;
     M_SIZE += 50;
   }
   if (pla is TRUE) {
     M_SIZE += middle_row/2 + 2;
     nnum_cols += middle_row/2 + 2;
   }
   
   tilda_column = ALLOC (int32, M_SIZE + 1);
   Row =  ALLOC (int32, M_SIZE + 1);
   Col =  ALLOC (int32, M_SIZE + 1);

   matrix = ALLOC (char *, nnum_cols+offset + 4);
   for (i = 0; i < nnum_cols+offset + 4; i++)
     matrix[i] =  ALLOC (char, num_rows+offset + 4);

   pbuffer = ALLOC (char *, 3);
   for (i = 0; i < 3; i++) 
     pbuffer[i] = ALLOC (char, M_SIZE + 1);

#ifdef THREED
   net_matrix = ALLOC (struct horver **,3);
   net_matrix[1] = ALLOC(struct horver *,nnum_cols+offset+4);
   net_matrix[2] = ALLOC(struct horver *,nnum_cols+offset+4);
   for (i = 0; i < nnum_cols+offset + 4; i++) { 
     net_matrix[1][i] = ALLOC (struct horver,num_rows+offset + 4);
     net_matrix[2][i] = ALLOC (struct horver,num_rows+offset + 4);
   }

   t_matrix = ALLOC (char **,3);
   t_matrix[1] = ALLOC(char *,nnum_cols+offset+4);
   t_matrix[2] = ALLOC(char *,nnum_cols+offset+4);
   for (i = 0; i < nnum_cols+offset + 4; i++) { 
     t_matrix[1][i] = ALLOC (char ,num_rows+offset + 4);
     t_matrix[2][i] = ALLOC (char ,num_rows+offset + 4);
   }

#else
   net_matrix = ALLOC (struct horver *,nnum_cols+offset + 4);
   for (i = 0; i < nnum_cols+offset + 4; i++) 
     net_matrix[i] = ALLOC (struct horver,num_rows+offset + 4);
#endif

   inp_matrix = ALLOC (char *,nnum_cols+offset+4);
   for (i = 0; i < nnum_cols+offset + 4; i++) 
     inp_matrix[i] =  ALLOC (char,num_rows+offset+4);

 return;

} /* end of allocate_storage */

 
free_storage()
{

  int i;

   FREE(tilda_column);
   FREE(Row); FREE(Col);

   for (i = 0; i < num_cols + 1; i++)
     FREE(matrix[i]);
   FREE(matrix);

   for (i = 0; i < 3; i++) 
     FREE(pbuffer[i]);
   FREE(pbuffer);

#ifdef THREED
   for (i = 1; i < num_cols + 1; i++)  {
     FREE(net_matrix[1][i]);
     FREE(net_matrix[2][i]);
   }
   FREE(net_matrix[1]); FREE(net_matrix[2]);
   FREE(net_matrix);

   for (i = 1; i < num_cols + 1; i++)  {
     FREE(t_matrix[1][i]);
     FREE(t_matrix[2][i]);
   }
   FREE(t_matrix[1]); FREE(t_matrix[2]);
   FREE(t_matrix);
#else
   for (i = 0; i < num_cols + 1; i++) 
     FREE(net_matrix[i]);
   FREE(net_matrix);
#endif

   for (i = 0; i < num_cols + 1; i++) 
     FREE(inp_matrix[i]);
   FREE(inp_matrix);

 return;

} /* end of free_storage */

