  
#include "port.h"
#include "utility.h"


/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/

#define NETHASH 1000
#define CELLHASH 5000
#define STDCELLHASH 100
#define is ==
#define TRUE	1
#define FALSE	0

#define INFINITY 1000000
#define MAXLINE 1024
#define ID_LENGTH 1024
#define EITHER 0
/*
* type of net
*/
#define PCHANNEL 1
#define NCHANNEL 2
#define SCOMPLEX 3
#define PCOMPLEX 4
#define SLCOMPLEX 5
#define PLCOMPLEX 6
#define SFCOMPLEX 7
#define PFCOMPLEX 8
#define SMCOMPLEX 9
#define PMCOMPLEX 10
/* 
* type of cell
*/
#define SERIES 0
#define PARALLEL 1
#define LOAD 2
#define TRANSISTOR 3
#define CAPLOAD 4
#define CAPSER 5
#define CAPPAR 6

#define AND 1
#define OR 2
#define NOT 3
#define NAND 4
#define NOR 5
#define CONNECT 6
#define PADLEFT 7
#define PADRIGHT 8
#define PADLR 9
/*
* array sizes
*/
#define NUM_CELLS 10000
#define NUM_COLS 600
#define NUM_STD 100
#define NUM_NETS 5000
/*
* types of moves
*/
#define INTHORNET 1
#define INTVERTNET 2
#define INTPAD 3
#define DISPLAYER 4
/*
* layers
#define LPOLY 1
#define LMETA 2
/*
* hash table 
*/
#define OK 1
#define ELMEN 1
#define NOROOM -1
#define DUP 2
/* 
* constraints on nets
*/
#define INPUT 1
#define OUTPUT 2
#define INTER 3
/*
* convenient definitions
*/
#define END -1000
#define isnt !=

typedef int *INTPTR;
typedef char *STRING;

struct	plist {		/* basic hash table entry */
	char *name;	/* string to be hashed */
	char *loc; 	/* pointer to element `name' (cast to char) */
	int  type;	/* either ELEMENT or TABLE  (what loc points to) */
	struct plist *next;	/* next entry in chain */
};

struct plist **Nethash;
struct plist **Stdhash;
struct plist **Cellhash;
/*
 * hack added by Chuck Kring to print PLA complements
 */
typedef struct clist {
    struct clist *next;
    char *pos;
    char *neg;
} s_clist;
struct clist *CompList;
 
struct horver {
  char hor,ver;
};

typedef struct rectangle {
   int x,y;
} RECT;


typedef struct stdcell {
     STRING stdName;
     int num;
     int l,w;
} *STDCELL ;

typedef struct sig_ptr {
     int num;
     int level;
     char ch;
     struct sig_ptr *nextPtr;
} *SIG_PTR;

typedef struct cell {
     STRING cellName;
     int num;
     int type;
     RECT rect;
     int xfix,yfix;
     int special;
#ifndef dist
     int octObjectId;
#endif
     char ch;
     STDCELL stdType;
     struct net_ptr *net_list;
} *CELL ;



typedef struct cell_ptr {
     CELL clink;
     struct cell_ptr *nextPtr;
} *CELL_PTR ;


typedef struct net {
     STRING netName;
     STRING fName;
     int layer;
     int bound, tb, bb;
     int num,complex;
     int input,output;
     int left,right,inter;
     int hor,vert,type;
     int weight,tilda,gate,size;
     int stagger,row,col;
     int straightened,cnt;
     int ordernum;
     char buffer;
     int num_cols,nnum_cols;
     int level,proc;
     SIG_PTR sig_list;
     CELL_PTR cell_list;

} *NET ;  


typedef struct net_ptr {
     NET nlink;
     int type;
     struct net_ptr *nextPtr;
} *NET_PTR;  


FILE *input;
CELL cells[NUM_CELLS],shift_cell1,shift_cell2;
NET  nets[NUM_NETS];
STDCELL stds[NUM_STD];
char **inp_matrix,**pbuffer;
int32 rowonly,colonly,rowpreferred,nopads,debug,domino,pla,plaread;
int32 gateread,octwrite,octread,threed,ordercon,simplerow,mgread,octmread;
int32 boundcon;
int32 statcmos,statnmos,specify,varwidth,simplecol,dogleg,twolevel,complex;
int no_output;
int32 net_num,cell_num,std_num,gate_num,pad_num,*Row,*Col,*tilda_column;
int expand,targc,NextLine,pad_left_num,stag_cost,over_cost,new_stag;
int OFFSET,overlap_penalty,new_penalty,OVERLAP_OFFSET,pad_right_num;
int MHsize,MVsize,newloc_x,newloc_y,init_cost;
char **matrix;
int max_xcoor,max_ycoor;
double STATES_BEGIN,STATES_END;
int max_cost,OVERLAP_PARAM,num_rows,num_cols,p_rows,p_cols;
int M_SIZE,ROW_PARAM,COL_PARAM,COMPLEX_PARAM;

#ifdef THREED
struct horver ***net_matrix;
char ***t_matrix;
#else
struct horver **net_matrix;
#endif

int new_net_overlap,net_overlap,middle_row;
double  NET_OVERLAP_PARAM, NET_LENGTH_PARAM;
char *targv[100];
int horizontal_span,vertical_span,min_ycoor,min_xcoor,gatetype[3];
int hor_net_num,RATIO_CELL_DISP,RATIO_NET_CELL,RATIO_NET_DISP;
int new_layer,new_row,new_col;
int complex_cols;
int left_ord_num,right_ord_num,*left_ord,*right_ord;
NET shift_net1,shift_net2;
double ALPHA,START_TEMP,WID_LEN_PARAM;


/* function declarations */
extern void bounding_constraints();
extern void order_constraints();
extern void get_pad_name();
extern void get_cell_name();
extern void get_net_name();
extern void find_starting_temp();
extern int32 generate_new_state();
