#include "copyright.h"
#include "port.h"


/**********************************************************************/
/*        This program is a gate matrix layout package for static     */
/*        CMOS, static NMOS and DOMINO design styles. It provides     */
/*          for a variety of constraints on the terminals and         */
/*            non uniform transistor dimensions. It uses the          */
/*              combinatorial optimization technique called           */
/*                     simulated annealing.                           */
/*                 Copyright: SRINIVAS DEVADAS  1985                  */
/**********************************************************************/


/*
 * definitions
 */
#define	NEWLINE		'\n'
#define	BLANK		' '
#define	COMMENT		'/'
#define STAR		'*'
#define HASH		'#'
#define	DOT		'.'
#define SEP		':'
#define TERM		';'
#define PART_SEP        '%'
#define PARALLEL	'p'
#define OUTPUT		'o'
#define SERIES		's'
#define is		==
#define isnt		!=
#define and 		&&
#define or		||
#define FALSE		0
#define TRUE		1
#define AND             1
#define OR              2
#define NOT             3
#define NAND            4
#define NOR             5

extern int NextLine;

getcluster (line,inputf)
char *line;
FILE *inputf;
{
 char buffer[30], buf;
 int signal,level;
 int c, next, i;

  level = 0; signal = FALSE;
  *line = '\0';
  while ((c = getc(inputf)) isnt TERM) {
    switch (c) {
      case EOF:
        if (signal)
          return (TRUE);
        else
          return(FALSE);
	break;

      case NEWLINE:
        NextLine++ ;
        break;

      case BLANK:
      case '\t':
      case SEP:
        break;

      case TERM:
        if (signal)
          return(TRUE);
        else
          return(FALSE);
	break;

      case COMMENT:
	while (c = getc(inputf) isnt COMMENT and c isnt EOF);
	if (c is EOF) {
          error("unclosed comment");
	}
	break;

      case HASH:
      case STAR:
	while (c = getc(inputf) isnt NEWLINE and c isnt EOF);
	if (c is EOF) { 
          error("unclosed comment");
	}
	break;


      case ')' :
        level-- ;
        if (level >= 0)
          (void) strcat(line," )");
	break;

      case '(':
        if (signal is TRUE)
          level++;
	while ((next = getc(inputf)) is BLANK or (next is '\t')
	      or (next is NEWLINE));
	if (next is 'P') next = 'p';
	if (next is 'S') next = 's';
	if (next is 'O') next = 'o';
	if (next isnt SERIES and next isnt PARALLEL and next isnt OUTPUT) {
          error("operator missing");
	}
        if (next is SERIES) 
          (void) strcat(line," ( s");
        else if (next is PARALLEL) 
          (void) strcat(line," ( p");
        
        next = getc(inputf);
        if (next isnt '*')
          (void) ungetc(next,inputf);
	break;

      default:					/* signal name		*/
          (void) ungetc(c,inputf);
          i= 0;
          while ( 1 ) {
            if (fscanf(inputf,"%c", &buf) is 1) {
              if  (buf is ')' || buf <= ' ')
                break;
              buffer[i] = buf;
              i++;
            }
            else
              error("reading signal");
          }
          (void) ungetc(buf,inputf);
          buffer[i] = '\0';
          (void) strcat(line," ");
          (void) strcat(line,buffer);
          signal = TRUE;
       
	break;
    }
  }
  return(TRUE);
} /* end of getcluster  */
