#include "copyright.h"
#include "genie.h"
/*
 * HASH FUNCTIONS FOR GENERALISED ARRAY OPTIMIIZER - GENIE
 *
 */

static struct plist *lookup();



/*
 *    HASH FUNCTIONS FOR HASH TABLE GENERATION AND INDEXING
 *        HAcreate   : creates an empty table
 *        HAinstall  : installs an entry
 *        HAfind     : looks up an entry
 *        HAtprint   : dumps the hashtable structure
 *
 */
 
struct plist **
HAcreate( hashsize )
/*
 *    create an empty table of hash-entry pointers, of length
 *    hashsize. Return a pointer to the beginning of the table.
 *    This pointer becomes the hash table id "hashtab" in later calls.
 */
int hashsize;
{
    int i;
    struct plist **np;
 
    np = ALLOC(struct plist *, hashsize);
    if ( np == NIL(struct plist *) )
        (void) fprintf( stderr, "error: malloc failed in create");
    for (i=0; i<hashsize; i++)
        *(np+i) = NIL(struct plist);
    return( np );
}
 
HAinstall( name, loc, hname, type, hashtab, hashsize )
/*
 *    name    : string to be hashed as key
 *    loc    : pointer to keyed data (caste to char)
 *    hname    : pointer to stored string "name" (for direct name lookup)
 *    type    : ELEMENT or TABLE, for hierarchical hashing
 *    hashtab    : pointer to the hash table itself (returned by create)
 *    hashsize: size of the hash table
 *
 *    Put `name' and `loc' pointer in hash table `hashtab'
 *    of length `hashsize'.
 *    Returns: NOROOM - if malloc fails.
 *         DUP - if name already installed. No action taken.
 *         OK - if install succeeds.
 */
char *name, *loc, **hname;
struct plist *hashtab[];    /* pointer table */
int hashsize, type;
{
    struct plist *np, *lookup();
    unsigned hashval;
 
    if ((np = lookup(name,hashtab,hashsize)) == NIL(struct plist) ){
        np = ALLOC(struct plist, 1);
        if (np == NIL(struct plist)) return (NOROOM);
        if ((np->name = util_strsav(name)) == NIL(char)) return (NOROOM);
        *hname = np->name;
        hashval = hash(np->name, hashsize);
        np->next = hashtab[hashval];
        np->loc  = loc;
        np->type = type;
        hashtab[hashval] = np;
        return ( OK );
    }
    else {
        np->loc = loc;
        np->type = type;
        return ( DUP );
    }
}
 
HAfind( name, loc, hashtab, hashsize )
/*
 *    find string `name' in `hashtab' and return a pointer to its
 *    structure origin (cast as char) in loc.
 *    Returns: ELEMENT - if type is element
 *         TABLE   - if type is hash table pointer
 *         FALSE    - if entry no present
 */
char *name, **loc;
struct plist *hashtab[];
int hashsize;
{
    struct plist *np, *lookup();
 
    if ((np=lookup(name,hashtab,hashsize)) == NIL(struct plist) ) return(FALSE); 
    *loc = (char *) np->loc;
    return ( np->type );
}
 
hash( s, hashsize )
/*
 *    form a hash value for string `s'
 *        (only good for small strings)
 */
char *s;
int hashsize;
{
    unsigned hashval, scale;
 
    for (hashval = 0, scale = 1; *s != '\0'; scale *= 3 )
        hashval += (*s++)*scale;
    return (hashval % hashsize);
}
 
static struct plist *
lookup( s, hashtab, hashsize )
/*
 *    look for `s' in `hashtab' of length `hashsize'
 */
char *s;
struct plist *hashtab[];
int hashsize;
{
    struct plist *np;
    int hashindex;
 
    hashindex = hash(s,hashsize);
    for (np = hashtab[hashindex]; np != NIL(struct plist); np = np->next)
        if (strcmp(s, np->name) == 0)
            return (np);    /* found it! */
        return (NIL(struct plist));    /* didn't find it */
}
