#include "copyright.h"
#include "port.h"
#include "genie.h"

extern NET findNet();

/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/


#define FUN_LENGTH 500

/*ARGSUSED*/
read_nets(netName1,netName2,netType,cellType)
STRING netName1,netName2;
int netType,cellType;
{
/*
* read a vertical net from oct which corresponds to a signal.
* keep attaching the cells to the horizontal net which corresponds
* to the gate.
*/

 NET hornet,vernet,findNet();
 CELL cell,findCell();
 char cellName[ID_LENGTH];
 NET_PTR nptr;
 CELL_PTR cptr;

  hornet = findNet(netName1);
  hornet->hor = TRUE;
  hornet->vert = FALSE;
/*
  if (netType is INPUT)
    hornet->output = FALSE;
  else if (netType is OUTPUT)
    hornet->input = FALSE;
  else
    crash("netType bad");
*/

  vernet = findNet(netName2);
  vernet->vert = TRUE;
  vernet->hor = FALSE;
  
  if (twolevel is FALSE) {
    for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
      cell = cptr->clink;
      for(nptr=cell->net_list;nptr!=NIL(struct net_ptr);nptr=nptr->nextPtr) {
        if (nptr->nlink is vernet) return;
      }
    }
  }
         
  (void) sprintf(cellName,"cell_%d",cell_num+1);
  cell = findCell(cellName);
  add_net_cell(cell,hornet);
  add_net_cell(cell,vernet);
  if (cellType is SERIES)
    cell->ch = 's';
  else if (cellType is PARALLEL)
    cell->ch = 'p';
  else if (cellType is LOAD)
    cell->ch = 'o';
  else if (cellType is TRANSISTOR)
    cell->ch = 'A';
  else if (cellType is CAPLOAD)
    cell->ch = 'O';
  else if (cellType is CAPPAR)
    cell->ch = 'P';
  else if (cellType is CAPSER)
    cell->ch = 'S';

 return;

} /* end of read_nets */


place_nets_and_cells()
{
/*
* place the nets and cells and print out the matrix corresponding
* to them.
*/

 int i, j, rownum, colnum, old_net_num, attach;
 CELL_PTR cptr;
 NET net,signet,hornet,hornet1,hornet2,vernet;
 CELL cell,findCell(),make_or_find_cell();
 int cnt,old,token_num,level,type;
 char funName[FUN_LENGTH];
 char netName[ID_LENGTH],cellName[ID_LENGTH];
 SIG_PTR sptr;
  

  /* tag the vertical nets corresponding to complex gates */
  old_net_num = net_num;
  for(i=1;i<=old_net_num;i++) {
    net = nets[i];
    if (net->hor is FALSE) {
      if (net->fName == NIL(char)) {
	net->fName = ALLOC(char, strlen(funName) - 1);
      }
      (void) strcpy(funName, net->fName);
      parse_line(funName);
      cnt = 0;
      for(j=0;j<targc-1;j++) {
        if (strcmp(targv[j],"(") is FALSE) {
          if (strcmp(targv[j+1],"p") is FALSE ||
                                  strcmp(targv[j+1],"s") is FALSE) cnt++;
        }
        if (cnt > 1) break;
      }
      if (cnt > 1)
        net->complex = TRUE;
      /* stick this in for now */
      net->complex = FALSE;
    }
#ifndef dist
    if (twolevel is TRUE && net->complex is TRUE) {
      (void) strcpy(funName, net->fName);
      parse_line(funName);
      if (targc < 3) crash("not a complex column");
      signet = findNet(targv[0]);
      sptr = ALLOC(struct sig_ptr, 1);
      sptr->num = signet->num;
      sptr->level = 1;
      sptr->ch = 'o';
      net->sig_list = sptr;
      net->sig_list->nextPtr = NIL(struct sig_ptr);
      if (strcmp(targv[1],"(") isnt FALSE) 
        crash("funName corrupted");
      if (strcmp(targv[2],"s") is FALSE)
        type = SERIES;
      else if (strcmp(targv[2],"p") is FALSE)
        type = PARALLEL;
      else
        crash("type corrupted");
        
      hornet1 = signet;

      (void) sprintf(netName,"%s_one",signet->netName);
      hornet2 = findNet(netName);
      hornet2->hor = TRUE;

      (void) sprintf(cellName,"cell_%d",cell_num+1);
      cell = findCell(cellName);
      cell->ch = 'o';
      add_net_cell(cell,net);
      add_net_cell(cell,hornet2);

      if (type is PARALLEL)
        net->type = PCOMPLEX;
      else
        net->type = SCOMPLEX;

      token_num = 2; level = 0;
      compute_level(&token_num,&level,net,type);

      net->level = level;
      for(j=2;j<=level;j++) {

        (void) sprintf(netName,"%s_%d",net->netName,j);
        vernet = findNet(netName);
        vernet->vert = TRUE;

        (void) sprintf(cellName,"cell_%d",cell_num+1);
        cell = findCell(cellName);
        cell->ch = 'o';
        add_net_cell(cell,vernet);
        add_net_cell(cell,hornet1);

        (void) sprintf(cellName,"cell_%d",cell_num+1);
        cell = findCell(cellName);
        cell->ch = 'o';
        add_net_cell(cell,vernet);
        add_net_cell(cell,hornet2);

        for(sptr=net->sig_list;sptr!=NIL(struct sig_ptr);sptr=sptr->nextPtr) {
          if (sptr->level isnt j) continue;
          hornet = nets[sptr->num];
          
          cell = make_or_find_cell(net,hornet,&old);
          if  (type is PARALLEL)
            cell->ch = 's';
          else
            cell->ch = 'p';
          if (old is FALSE) crash("error couldnt find cell");

          delete_net_cell(cell,net);
          add_net_cell(cell,vernet);
        }
      }
      for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
        if (cptr->clink->ch is 'o') continue;
        if (type is PARALLEL)
          cptr->clink->ch = 's';
        else if (type is SERIES)
          cptr->clink->ch = 'p';
      }
    }
#endif
  }
  
  if (mgread is FALSE) {

    /* add the "t" cells for the primary inputs */
    old_net_num = net_num;
    for(i=1;i<=old_net_num;i++) {
      if (nets[i]->hor is TRUE) {
        attach = TRUE;
        for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                 cptr=cptr->nextPtr) {
          if (cptr->clink->ch is 's' || cptr->clink->ch is 'p' ||
                                                     cptr->clink->ch is 't') {
            attach = FALSE;
            break;
          }
        }
        /* stick this in for now */
        attach = FALSE;
        if (attach) {
  
          (void) sprintf(netName, "%s_po", nets[i]->netName);
          vernet = findNet(netName);
          vernet->vert = TRUE; vernet->tilda = FALSE;
  
          (void) sprintf(cellName, "cell_%d", cell_num+1);
          cell = findCell(cellName);
          cell->ch = 't';
  
          add_net_cell(cell, vernet);
          add_net_cell(cell, nets[i]);

        }
      }
    }
  }
  add_pads();
#ifndef dist
  (void) fprintf(stdout,"Number of nets is %d\n",net_num);
  (void) fprintf(stdout,"Number of cells is %d pads %d\n",cell_num,cell_num-gate_num);
#endif
        
  rownum = colnum = 0;
  
  for(i=1;i<=net_num;i++) {
    net = nets[i];
    if (net->hor is TRUE) {
      rownum++;
      net->row = rownum;
      for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                   cptr=cptr->nextPtr) {
        cell = cptr->clink;
        cell->rect.y = rownum;
      }
    }
    else if (net->hor is FALSE) {
      colnum++;
      net->col = colnum;
      for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                   cptr=cptr->nextPtr) {
        cell = cptr->clink;
        cell->rect.x = colnum;
      }
    }
    else
      crash("neither horizontal or vertical net");
  }

  num_rows = rownum;
  num_cols = colnum;
  allocate_storage();

 return;

} /* end of place_nets_and_cells */



attach_function(netName,funName)
char *netName,*funName;
{
/*
* for complex gates, need the funName to reconstruct.
*/
 char *fName;
 NET net,findNet();

  fName = ALLOC(char,FUN_LENGTH+1);
  (void) strcpy(fName,funName);
  net = findNet(netName);
  net->fName = fName;

 return;

} /* end of attach_function */
 



read_gates()
{
/*
* read a gate level description from the input file, in factored
* form, and produce the connectivity matrix of the transistors.
*/

 char *funName,*fName;
 int type,i,first,firstcell,simple;
 int stype,token_num,level;
 char netName1[ID_LENGTH],netName2[ID_LENGTH];
 char outName[ID_LENGTH];
 NET tildanet,nt,findNet();
 NET primarynet;
 char hornetName[ID_LENGTH],vernetName[ID_LENGTH];
 SIG_PTR sptr;
  

  funName = ALLOC(char,MAXLINE+1);
  fName = ALLOC(char,MAXLINE+1);

  while(getcluster(funName,input) isnt FALSE) {
    (void) strcpy(fName,funName);
    parse_line(funName);

    if (strcmp(targv[0],"left") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->left = TRUE;
      }
      continue;
    }
    else if (strcmp(targv[0],"output") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->output = TRUE;
      }
      continue;
    }
    else if (strcmp(targv[0],"input") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->input = TRUE;
      }
      continue;
    }
    else if (strcmp(targv[0],"right") is FALSE) {
      for(i=1;i<targc;i++) {
        nt = findNet(targv[i]);
        nt->right = TRUE;
      }
      continue;
    }
    else if (strcmp(targv[0],"ordered") is FALSE) {
      order_constraints();
    }

    i = 1; first = TRUE; simple = TRUE;
    while (i < targc ) {
      if (strcmp(targv[i],"(") is FALSE) {
        if (strcmp(targv[i+1],"s") is FALSE) {
          if (first is TRUE) {
            type = SERIES;
            first = FALSE;
          }
          else {
            simple = FALSE;
          }
          i += 2;
        }
        else if (strcmp(targv[i+1],"p") is FALSE) {
          if (first is TRUE) {
            type = PARALLEL;
            first = FALSE;
          }
          else {
            simple = FALSE;
          }
          i += 2;
        }
      }
      else {
        i += 1;
      }
    }
      
    if (simple is TRUE) {
      if (no_output is TRUE) {
        if (type is PARALLEL)
          (void) sprintf(netName1, "p_%s",targv[0]);
        else if (type is SERIES)
          (void) sprintf(netName1, "s_%s",targv[0]);
      }
      else
        (void) sprintf(netName1,"gate_%s",targv[0]);
      attach_function(netName1,fName);
/*
      if (mgread is FALSE) {
        (void) sprintf(netName2,"output_%s",targv[0]);

        (void) sprintf(outName,"%s_o",targv[0]);
        read_nets(targv[0],netName2,OUTPUT,LOAD);
        read_nets(outName,netName2,OUTPUT,LOAD);

        tildanet = findNet(netName2);
        tildanet->tilda = TRUE;
      }
      else
*/
        (void) sprintf(outName,"%s",targv[0]);

      first = TRUE; i = 1; firstcell = TRUE;
      while(i <= targc-1) {
        if (strcmp(targv[i],"(") is FALSE) {
          if (strcmp(targv[i+1],"s") is FALSE) {
            if (first is TRUE) {
              type = SERIES;
              first = FALSE;
            }
            i += 2;
          }
          else if (strcmp(targv[i+1],"p") is FALSE) {
            if (first is TRUE) {
              type = PARALLEL;
              first = FALSE;
            }
            i += 2;
          }
          else
            error("wierd operator after (");
        }
        else if (strcmp(targv[i],")") is FALSE) {
          i += 1;
        }
        else {
          if (firstcell is TRUE) {
            if ( mgread is FALSE && no_output is FALSE )
              read_nets(outName,netName1,OUTPUT,type);
            else if ( mgread is TRUE )
              read_nets(outName,netName1,OUTPUT,LOAD);
            firstcell = FALSE;
          }
          if (mgread is FALSE) 
            read_nets(targv[i],netName1,INPUT,TRANSISTOR);
          else 
            read_nets(targv[i],netName1,INPUT,type);
          i += 1;
        }
      }
    }
    else {
/*
* Complex gate.
*/
      /* make the main gate column, with s or p and O's and some A's*/
      (void) sprintf(netName1,"cgate_%s",targv[0]);
      attach_function(netName1,fName);

      /* tilda output column with the two o's */
      (void) sprintf(netName2,"Output_%s",targv[0]);

      /* jogged output signal with one o and a P or a S */
      (void) sprintf(outName,"%s_o",targv[0]);

      /* attach p or s to primary column and jogged output horiz. signal */
      if (type is SERIES)
        read_nets(outName,netName1,OUTPUT,SERIES);
      else if (type is PARALLEL)
        read_nets(outName,netName1,OUTPUT,PARALLEL);

      /* attach o to real output horiz. net and tilda output column */
      read_nets(targv[0],netName2,OUTPUT,LOAD);
    
      /* attach o to jogged output horiz. net and tilda output column */
      read_nets(outName,netName2,OUTPUT,LOAD);

      tildanet = findNet(netName2);
      tildanet->tilda = TRUE;
    
      primarynet = findNet(netName1);
      token_num = 2; level = 0;
      compute_level(&token_num,&level,primarynet,type);
      primarynet->level = level;

      if (type is SERIES)
        stype = CAPPAR;
      else if (type is PARALLEL)
        stype = CAPSER;

      for(i=2;i<=level;i++) {
        
        /* each level has a derivable horiz. net */
        (void) sprintf(hornetName,"%s_%d",targv[0],i);
        read_nets(hornetName,netName1,OUTPUT,CAPLOAD);

        /* each level has a derivable vertical net */
        (void) sprintf(vernetName,"gate_%s_%d",targv[0],i);
        read_nets(hornetName,vernetName,INPUT,stype);
      }
      
      for(sptr=primarynet->sig_list;sptr!=NIL(struct sig_ptr);
                                                      sptr=sptr->nextPtr) {
        if (sptr->level is 1)
          (void) strcpy(vernetName,primarynet->netName);
        else
          (void) sprintf(vernetName,"gate_%s_%d",targv[0],sptr->level);
        read_nets(nets[sptr->num]->netName,vernetName,INPUT,TRANSISTOR);
      }
    }
  }
 return;

} /* end of read_gates */


#ifndef lint
void
print_net_sig(net)
NET net;
{

 SIG_PTR sptr;


  for(sptr=net->sig_list;sptr!=NIL(struct sig_ptr);sptr=sptr->nextPtr)
    (void) fprintf(stdout,"sptr level %d number %d\n",sptr->level,sptr->num);

  return;
}
#endif


void
bounding_constraints()
{

 NET net1, net2;
 int i;
  
  boundcon = TRUE;
  net1 = findNet(targv[1]);
  net2 = findNet(targv[2]);

  net1->bound = TRUE;
  net2->bound = TRUE;
  for(i=1;i<=net_num;i++) {
    if (nets[i]->hor is TRUE && nets[i] isnt net1 && nets[i] isnt net2) {
      if (nets[i]->row > net1->row && nets[i]->row < net2->row ) {
        nets[i]->tb = net1->row;
        nets[i]->bb = net2->row;
      }
    }
  }

 return;

} /* end of bounding_constraints */


void
order_constraints()
{
/*
* from parsed line construct ordering constraints on nets.
*/
 int i,left;
 NET net, findNet();

  ordercon = TRUE;
  if (strcmp(targv[1],"left") is FALSE) {
    left_ord = ALLOC(int, targc + 1);
    left = TRUE;
    left_ord_num = 0;
  }
  else if (strcmp(targv[1],"right") is FALSE) {
    right_ord = ALLOC(int,targc + 1);
    left = FALSE;
    right_ord_num = 0;
  }
  else
   crash("wrong identifier in ordered statement");
  for(i=2;i< targc;i++) {
    net = findNet(targv[i]);
    net->ordernum = i - 1;
    if (left is TRUE) {
      left_ord_num++;
      left_ord[left_ord_num] = net->num;
    }
    else {
      right_ord_num++;
      right_ord[right_ord_num] = net->num;
    }
  }

 return;

} /* end of order_constraints */


int
is_output(name)
STRING name;
{

 int length;

  length = strlen(name);
  if (name[length-1] is 'o' && name[length-2] is '_') return(TRUE);
  else return(FALSE);

} /* end of is_output */
