#include "copyright.h"
#include "genie.h"


/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/


int32
range_limit_disp(type)
int type;
{
  int minx,miny,maxx,maxy;
  double RANDOM();

  if (type is INTHORNET) {
/*
* range in which net can be displaced vertically.
*/
    miny = shift_net1->row - vertical_span;
    maxy = shift_net1->row + vertical_span;
    if (statcmos is TRUE || pla is TRUE) {
      if (shift_net1->type is PCHANNEL) {
        miny = MAX(OFFSET+1,miny);
        maxy = MIN(maxy,middle_row+OFFSET);
      }
      else if (shift_net1->type is NCHANNEL) {
        miny = MAX(OFFSET+middle_row+1,miny);
        maxy = MIN(maxy,num_rows+OFFSET);
      }
      else if (shift_net1->type is EITHER && shift_net1->output is TRUE) {
        miny = MAX(OFFSET+1,miny);
        maxy = MIN(maxy,num_rows+OFFSET);
      }
      else if (shift_net1->type is EITHER && shift_net1->output isnt TRUE)
        error("static cmos matrix and horizontal net type undefined");
      if (pla is TRUE) {
        if (shift_net1->gate isnt 0) {
          miny = MAX(miny, nets[shift_net1->gate]->row-1);
          maxy = MIN(maxy, nets[shift_net1->gate]->row+1);
        }
      }
    }
    else {
      miny = MAX(OFFSET+1,miny);
      maxy = MIN(maxy,num_rows+OFFSET);
    }
    if (boundcon is TRUE) {
      miny = MAX(miny,shift_net1->tb);
      maxy = MIN(maxy,shift_net1->bb);
    }
    new_row = RANDOM(miny,maxy+1);
    if (new_row is shift_net1->row || (pla is TRUE && shift_net1->gate isnt 0
                               && new_row is nets[shift_net1->gate]->row) ) {
       new_row = RANDOM(miny,maxy+1);
       if (new_row is shift_net1->row || (pla is TRUE && shift_net1->gate isnt 0
                               && new_row is nets[shift_net1->gate]->row) )
         return(FALSE);
    }
    if (boundcon is TRUE) {
      if (check_disp_bound(shift_net1,new_row) is FALSE)
        return(FALSE);
    }
    if (ordercon is TRUE) {
      if (check_disp_order(shift_net1,new_row) is FALSE)
        return(FALSE);
    }
    return(TRUE);
  }
  else if (type is INTVERTNET) {
/*
* range in which net can be displaced horizontally.
*/
    minx = shift_net1->col - horizontal_span;
    maxx = shift_net1->col + horizontal_span;
    minx = MAX(OFFSET+1,minx);
    maxx = MIN(maxx,num_cols+OFFSET);
    new_col = RANDOM(minx,maxx+1);
    if (domino is TRUE) {
/*
*  check for columns which cannot be folded for TINKER compatibility.
*/
      if (tilda_column[new_col] is TRUE )
        return(FALSE);
    }
/*
* if an output net, dont fold, double precaution.
*/
    else {
      if (shift_net1->tilda is TRUE)
        return(FALSE);
    }

    if (new_col is shift_net1->col)
      return(FALSE);
    else
      return(TRUE);
  }
 return(TRUE);

} /* end of range_limit_disp */


int32
generate_new_state()
{

 CELL pad;
 double RANDOM(), f_num, s_num;
 int first_num, second_num, num_pad;
 int32 check_range;
 CELL_PTR cptr;
 
    
   new_layer = 0;
   shift_net1 = shift_net2 = NIL(struct net);
   shift_cell1 = NIL(struct cell);
   f_num = RANDOM(1,net_num+1);
   first_num = f_num;
   s_num = RANDOM(1,net_num*RATIO_NET_DISP + 1);
   second_num = s_num;
   if (nets[first_num]->cell_list is NIL(struct cell_ptr))
     return(FALSE);
   if (s_num <= net_num) {
/*
*  try interchanging nets.
*/
     if (nets[second_num]->cell_list is NIL(struct cell_ptr))
       return(FALSE);
     if (nets[first_num]->hor is nets[second_num]->hor) {
       if (domino is TRUE) {
/*
*  need to check for vertical nets which should not be folded for
*  output compatiblity with TINKER.
*/
         if (nets[first_num]->hor is FALSE && nets[second_num]->hor is FALSE) {
           if ( (nets[first_num]->tilda is TRUE && 
                 nets[second_num]->tilda is FALSE) || 
                (nets[first_num]->tilda is FALSE &&
                 nets[second_num]->tilda is TRUE) )
             return(FALSE);
           if (nets[first_num]->gate isnt nets[second_num]->gate)
             return(FALSE);
         }
       }
       if (pla is TRUE) {
         if (nets[first_num]->hor is TRUE && nets[second_num]->hor is TRUE) {
           if (nets[first_num]->gate isnt 0 ) {
             if (abs(nets[second_num]->row -
                          nets[nets[first_num]->gate]->row) > 1) {
               /* interchange first_num with complement */
               shift_net1 = nets[first_num];
               shift_net2 = nets[nets[first_num]->gate];
               if (shift_net1->row is shift_net2->row)
                 return(FALSE);
               else if (abs(shift_net1->row - shift_net2->row) > 1)
                 crash("input and complement far apart");
               else
                 second_num = shift_net2->num;
             }
           }
           if (nets[second_num]->gate isnt 0 ) {
             if (abs(nets[first_num]->row - 
                                   nets[nets[second_num]->gate]->row) > 1) {
               /* interchange second_num with complement */
               shift_net2 = nets[second_num];
               shift_net1 = nets[nets[second_num]->gate];
               if (shift_net1->row is shift_net2->row)
                 return(FALSE);
               else if (abs(shift_net1->row - shift_net2->row) > 1)
                 crash("input and complement far apart");
               else
                 first_num = shift_net1->num;
             }
           }
         }
       }

       shift_net1 = nets[first_num];
       shift_net2 = nets[second_num];
       if (debug)
         (void) fprintf(stdout,"interchanging nets\n");
       if (statcmos is TRUE || pla is TRUE) {
         if (shift_net1->hor is TRUE && shift_net2->hor is TRUE) {
           if ((shift_net1->type is PCHANNEL && shift_net2->type is NCHANNEL) ||
               (shift_net1->type is NCHANNEL && shift_net2->type is PCHANNEL))
             return(FALSE);
           if ( (shift_net1->type is PCHANNEL && shift_net2->row > middle_row
               + OFFSET) || (shift_net1->type is NCHANNEL && shift_net2->row
                                                     <= middle_row + OFFSET) )
             return(FALSE); 
           if ( (shift_net2->type is PCHANNEL && shift_net1->row > middle_row
               + OFFSET) || (shift_net2->type is NCHANNEL && shift_net1->row
                                                     <= middle_row + OFFSET) )
             return(FALSE); 
         }
       }

       if (ordercon is TRUE)
         if (shift_net1->hor is TRUE && shift_net2->hor is TRUE)
           if (check_inter_order(shift_net1,shift_net2) is FALSE)
             return(FALSE);

       if (boundcon is TRUE)
         if (shift_net1->hor is TRUE && shift_net2->hor is TRUE)
           if (check_inter_bound(shift_net1,shift_net2) is FALSE)
             return(FALSE);

       if (shift_net1->hor is TRUE) {
         if (shift_net1->row isnt shift_net2->row)  
           return(TRUE);
       }
       else {
         if (shift_net1->col isnt shift_net2->col) 
           return(TRUE);
       }
     }
     else if (nopads is FALSE) {
/*
*  try displacing pads.
*/
       if (pad_num isnt 0) {
         num_pad = RANDOM(1,pad_num+1);
         pad = cells[num_pad+gate_num];
         if (pad->type is PADLEFT || pad->type is PADRIGHT)
           return(FALSE);
         if (pad->rect.x < OFFSET)
           newloc_x = MHsize - 2;
         else
           newloc_x = 2;
         shift_cell1 = pad;
         return(TRUE);
       }
     }
   }
   else { 
/*
*  try displacing a net.
*/
     if ( threed is FALSE || (threed is TRUE && 
           s_num <= net_num*RATIO_NET_DISP/2 ) || 
                               nets[first_num]->hor is FALSE ) {
          
       shift_net1 = nets[first_num];
       shift_net2 = NIL(struct net);
       if (domino is TRUE) {
/*
*    check for columns which cannot be folded for TINKER compatibility.
*/
         if (tilda_column[shift_net1->col] is TRUE )
            return(FALSE);
       }
/*
* if an output net dont fold it
*/
       else {
         if (shift_net1->tilda is TRUE)
           return(FALSE);
       }

/*
*    check for row only or column only folding option.
*/
       if (shift_net1->hor is TRUE && colonly is TRUE)
         return(FALSE);

       if (shift_net1->hor is FALSE)
         if (rowonly is TRUE)
           return(FALSE);

       if (shift_net1->hor is TRUE) {
         if ( nopads is FALSE && s_num <= 2*net_num &&
                shift_net1->left is FALSE && shift_net1->right is FALSE) {
            for(cptr=shift_net1->cell_list;cptr!=NIL(struct cell_ptr);
                                                   cptr=cptr->nextPtr) {
              if (cptr->clink->type is PADLR) {
                 shift_cell1 = cptr->clink;
                 if (shift_cell1->rect.x < OFFSET)
                   newloc_x = MHsize - 2;
                 else
                   newloc_x = 2;
                 shift_net1 = NIL(struct net);
                 return(TRUE);
              }
            }
         }
         check_range = range_limit_disp(INTHORNET);
       }
       else
         check_range = range_limit_disp(INTVERTNET);
       if (debug)
         (void) fprintf(stdout,"displacing a net\n");
       return(check_range);
     }
     else {
/*
* going to try displacing a net off to another layer
*/
       shift_net1 = nets[first_num];
       shift_net2 = NIL(struct net);
       if (domino is TRUE) {
         if (tilda_column[shift_net1->col] is TRUE )
            return(FALSE);
       }
       if (colonly is TRUE)  crash("column folding only and three d");
       if (shift_net1->layer is 1 )
         new_layer = 2;
       else if (shift_net1->layer is 2)
         new_layer = 1;
       else
         crash("unknown layer number");
       return(TRUE);
     }
   }
     
  return(FALSE);

} /* end of generate_new_state */


int
check_inter_order(net1,net2)
NET net1,net2;
{
/*
* check if this interchange would violate the ordering constraint.
*/

 int ord1,ord2;
 int check_disp_order();

  if ( (ord1 = net1->ordernum) isnt 0 || (ord2 = net2->ordernum) isnt 0) {
    if (ord1 isnt 0 && ord2 isnt 0) {
      if (net1->left is TRUE && net2->left is TRUE)
        return(FALSE);
      if (net1->right is TRUE && net2->right is TRUE)
        return(FALSE);
    }
    if ( ord1 isnt 0 ) {
      if (check_disp_order(net1,net2->row) is FALSE)
        return(FALSE);
    }
    if ( ord2 isnt 0 ) {
      if (check_disp_order(net2,net1->row) is FALSE)
        return(FALSE);
    }
  }

 return(TRUE);

} /* end of check_inter_order */


int
check_inter_bound(net1,net2)
NET net1, net2;
{
  
   if (net1->bound is TRUE || net2->bound is TRUE)
     return(FALSE);
   if (net1->tb < net2->row && net1->bb > net2->row )
     if (net2->tb < net1->row && net2->bb > net1->row )
       return(TRUE);

 return(FALSE);

} /* end of check_inter_bound */


int
check_disp_bound(net,row)
NET net;
int row;
{

   if (net->bound is TRUE) return(FALSE);

   if (net->tb < row && net->bb > row)
     return(TRUE);
   else
     return(FALSE);

} /* end of check_disp_bound */


int
check_disp_order(net,row)
NET net;
int row;
{
/*
* check that the displacement of the given net to the given row
* does not violate any ordering constraints.
*/

 int i;

  if (net->left is TRUE) {
 
    /* work with left_ord */
    for(i=1;i< net->ordernum;i++) {
      if (nets[left_ord[i]]->row >= row)
        return(FALSE);
    }
    for(i=net->ordernum+1;i<=left_ord_num;i++) {
      if (nets[left_ord[i]]->row <= row)
        return(FALSE);
    }
  }
  else if (net->right is TRUE) {
 
    /* work with right_ord */
    for(i=1;i< net->ordernum;i++) {
      if (nets[right_ord[i]]->row >= row)
        return(FALSE);
    }
    for(i=net->ordernum+1;i<=right_ord_num;i++) {
      if (nets[right_ord[i]]->row <= row)
        return(FALSE);
    }
  }

 return(TRUE);

} /* end of check_disp_order */
