#include "copyright.h"
#include "genie.h"


/**************************************************************************/
/*    This program inputs a  multi-level personality matrix 	*/
/*         and finds the optimum placement of cells     	*/
/*      reducing the number of rows and columns using 		*/
/*                 simulated annealing.  	                */
/*           Copyright: SRINIVAS DEVADAS  1985   	        */
/**************************************************************************/



#define HOR 1
#define VER 2
#define EITHER 0

extern int *hor_net_order,*vert_net_order;
extern char **fixed_loc;
extern int hor_num,ver_num;
int *numnets_on_col;
extern int *numnets_on_row;


straighten_vertical_nets()
{
/*
*  all the vertical nets' stagger is removed.After this the array is 
*  routable. 
*/
   int i;
   NET net;

   numnets_on_col =  ALLOC(int,MHsize+10);
   for(i=0;i<ver_num;i++) {
     net = nets[vert_net_order[i]];
     if (net->cell_list isnt NIL(struct cell_ptr)) {
       if (rowpreferred is TRUE) 
         strgtn_vert_net(net);
       else
         strgtn_ver(net);
     }
   }
  return;

} /* end of straighten_vertical_nets */


strgtn_vert_net(net)
NET net;
{

  int miny,maxy,q;
  CELL_PTR cptr;

  q = 0;
  miny = INFINITY;
  maxy = 0;
  for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
    q++ ;
    miny = MIN(miny,cptr->clink->rect.y);
    maxy = MAX(maxy,cptr->clink->rect.y);
  }
  if (miny > maxy)
    error("error in strgn_vert_net");
  if (maxy - miny < q -1)
    maxy = miny + q - 1;
  add_col_rows_for_net(net,miny,maxy);

 return;

} /* end of strgtn_vert_net */



strgtn_hor(net)
NET net;
{

  int minx,maxx,q;
  CELL_PTR cptr;

  q = 0;
  minx = INFINITY;
  maxx = 0;
  for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
    q++ ;
    minx = MIN(minx,cptr->clink->rect.x);
    maxx = MAX(maxx,cptr->clink->rect.x);
  }
  if (minx > maxx)
    error("error in strgn_hor");
  if (maxx - minx < q -1)
    maxx = minx + q - 1;
  add_row_cols_for_net(net,minx,maxx);

 return;

} /* end of strgtn_hor */


add_col_rows_for_net(net,miny,maxy)
NET net;
int miny,maxy;
{

  int minx,maxx,lminx,lmaxx,col,*nrows,ncost;
  int column,cost,*yes,row,i,j,num_cells,same;
  int r,s,aldone,*poss_col,*cyes;
  int can_move ,c_mov,loc_x,ran_miny,ran_maxy;
  int range_miny,range_maxy,diff,row_exist,sel_row,crow,psize;
  int ccol,cminx,ccminx,*hor_loc,ccmaxx,cmaxx,*all_occ;
  CELL_PTR cptr,cptr2,ccptr;
  NET hornet,chornet,vertnet,cvertnet;
  CELL cell;
/*
*  add a column and/or row(s) and change the cell locations.
*  First try to find a column which maximises the number of cells
*  which can be put on it without adding rows and disrupting horizontal
*  nets.
*/

   lminx = INFINITY;
   lmaxx = 0;
   poss_col = ALLOC (int,num_cols *3 +20);
   nrows =  ALLOC (int,MAX(MHsize,MVsize) + 2);
   cyes = ALLOC (int,MAX(MHsize,MVsize) + 2);
   hor_loc = ALLOC (int,MAX(MHsize,MVsize) + 2);
   all_occ = ALLOC (int,MAX(MHsize,MVsize) + 2);
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     lminx = MIN(lminx,cptr->clink->rect.x);
     lmaxx = MAX(lmaxx,cptr->clink->rect.x);
   }
   if (lminx > lmaxx)
    error("lminx greater than lmaxx in add_col_rows_for_net");
/*
* check if any column can be used.
*/
   for(col=lminx-1;col<=lmaxx+1;col++) {
     for(i=1;i<=MHsize;i++)
       all_occ[i] = FALSE;
     poss_col[col] = TRUE;
     for(i=miny;i<=maxy;i++) {
       if (fixed_loc[col][i] is TRUE)
          poss_col[col] = FALSE;
     }
     if (simplecol is TRUE) {
       if (numnets_on_col[col] is 2) 
         poss_col[col] = FALSE;
       else if (numnets_on_col[col] > 2)
         error("simple column folding and more than two nets on a column");
     }
     for(i=1;i<=cell_num;i++) {
       if (cells[i]->rect.x isnt col) continue;
       if ( net->tilda is TRUE || rowonly is TRUE) {
         same = FALSE;
         for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
           if (cptr->clink is cells[i])
             same = TRUE;
         if (same is FALSE ) {
             poss_col[col] = FALSE;
         }
       }
       if (cells[i]->rect.y < miny || cells[i]->rect.y > maxy) continue;
       same = FALSE;
       for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
         if (cptr->clink is cells[i])
           same = TRUE;
       if (same is FALSE ) {
         if (cells[i]->xfix is TRUE)
           poss_col[col] = FALSE;
         else {
/*
*  check if cell can be moved to some other location on its horizontal net.
*/
           if (cells[i]->type >= PADLEFT) 
             hornet = cells[i]->net_list->nlink;
           else {
             if (cells[i]->net_list->nlink->hor is TRUE)
               hornet = cells[i]->net_list->nlink;
             else
               hornet = cells[i]->net_list->nextPtr->nlink;
           }
           minx = INFINITY;
           maxx = 0;
           for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);
                                                  cptr=cptr->nextPtr) {
             if (cptr->clink->type >= PADLEFT) {
               if (cptr->clink is cells[i])
/*
*  dont want to mess around with a pad.
*/
                 poss_col[col] = FALSE;
               continue;
             }
             minx = MIN(minx,cptr->clink->rect.x);
             maxx = MAX(maxx,cptr->clink->rect.x);
           }
           if (minx > maxx)
             error("minx greater than maxx");
           can_move = FALSE;
           for(s=minx;s<=maxx;s++) {
             c_mov = TRUE;
             if (fixed_loc[s][hornet->cell_list->clink->rect.y] is TRUE ||
                                                         all_occ[s] is TRUE)
               c_mov = FALSE;
             for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);
                                                    cptr=cptr->nextPtr) {
               if (cptr->clink->rect.x is s) {
                 c_mov = FALSE;
                 break;
               }
             }
             if (c_mov is TRUE) {
               can_move = TRUE;
               all_occ[s] = TRUE;
               break;
             }
           }
           if (can_move is FALSE) 
             poss_col[col] = FALSE;
         }
       }
     }
   }
  
   for(col=lminx-1;col<=lmaxx+1;col++) {
     nrows[col] = 0;
     for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
       cell = cptr->clink;
       if (cell->net_list->nlink->hor is TRUE)
         hornet = cell->net_list->nlink;
       else
         hornet = cell->net_list->nextPtr->nlink;
       minx = INFINITY;
       maxx = 0;
       num_cells = 0;
       for(cptr2=hornet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                   cptr2=cptr2->nextPtr) {
         minx = MIN(minx,cptr2->clink->rect.x);
         maxx = MAX(maxx,cptr2->clink->rect.x);
         num_cells++;
       }
       if (minx > maxx)
         error("minx greater than maxx in add_col_rows_for_net");
       if (col < minx || col > maxx && num_cells > 1) {
         if (varwidth is TRUE)
           nrows[col] += hornet->size;
         else
           nrows[col] += 1;
       }
     }
   }
/*
*  pick the column with least cost
*/
   column = 0;cost = INFINITY;
   for(col=lminx-1;col<=lmaxx+1;col++) {
     if (varwidth is TRUE)
       ncost = nrows[col];
     else
       ncost = nrows[col]*2;
     if (poss_col[col] is FALSE) {
/*
* cost of adding a column may be higher than that of adding a row.
* Area consideration.
*/
       if (varwidth is TRUE)
         ncost += net->size;
        else
         ncost += 2;
     }
     if (cost > ncost) {
       cost = ncost;
       column = col;
     }
   }
   if (column is 0)
     error("couldn't find a column with less than infinite cost");
/*
*  add the column if necessary.find the troublesome cells.
*/
   if (poss_col[column] is FALSE) 
     col = add_col(column,column+1);
   else 
     col = column;
   if (simplecol is TRUE)
     numnets_on_col[col] += 1;
/*
*  find the range of the net on the column.Cells may be placed anywhere 
*  on this range.
*/
   range_miny = 0;
   range_maxy = INFINITY;
   ran_miny = INFINITY;
   ran_maxy = 0;
   for(i=1;i<=cell_num;i++) {
     ran_miny = MIN(ran_miny,cells[i]->rect.y);
     ran_maxy = MAX(ran_maxy,cells[i]->rect.y);
     if (cells[i]->rect.x is column) {
       if (cells[i]->rect.y < miny)
         range_miny = MAX(range_miny,cells[i]->rect.y);
       if (cells[i]->rect.y > maxy) 
         range_maxy = MIN(range_maxy,cells[i]->rect.y);
     }
   }
   range_miny = MAX(range_miny,ran_miny);
   range_maxy = MIN(range_maxy,ran_maxy);
   if (poss_col[col] is TRUE) {
     for(i=1;i<=cell_num;i++) {
       if (cells[i]->rect.x isnt col) continue;
       if (cells[i]->rect.y < miny || cells[i]->rect.y > maxy) continue;
       same = FALSE;
       for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
         if (cptr->clink is cells[i])
           same = TRUE;
       if (same is FALSE) {
         if (cells[i]->xfix is TRUE)
           error("have to add column here");
         else {
/*
* cell should be able to move to some location in the range of its horiz. net.
*/
           if (cells[i]->type >= PADLEFT) 
             hornet = cells[i]->net_list->nlink;
           else {
             if (cells[i]->net_list->nlink->hor is TRUE)
               hornet = cells[i]->net_list->nlink;
             else
               hornet = cells[i]->net_list->nextPtr->nlink;
           }
           minx = INFINITY;
           maxx = 0;
           for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);
                                                  cptr=cptr->nextPtr) {
             if (cptr->clink->type >= PADLEFT) {
               if (cptr->clink is cells[i])
                 error("have to add a column here");
             }
             minx = MIN(minx,cptr->clink->rect.x);
             maxx = MAX(maxx,cptr->clink->rect.x);
           }
           if (minx > maxx)
             error("minx greater than maxx");
           can_move = FALSE;diff = INFINITY;
           for(s=minx;s<=maxx;s++) {
             c_mov = TRUE;
             if (fixed_loc[s][hornet->cell_list->clink->rect.y] is TRUE)
               c_mov = FALSE;
             for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);
                                                    cptr=cptr->nextPtr) {
               if (cptr->clink->rect.x is s) {
                 c_mov = FALSE;
                 break;
               }
             }
             if (c_mov is TRUE) {
               can_move = TRUE;
               if (diff > abs(col - s)) {
                 loc_x = s;
                 diff = abs(col - s);
               }
             }
           }
           if (can_move is FALSE) 
             error("have to add a column here");
           else 
/*
*  make the move.
*/
             cells[i]->rect.x = loc_x;
         }
       }
     }
   }
   yes = ALLOC (int,cell_num + 1);
   for(i=1;i<=cell_num;i++)
     yes[i] = FALSE;
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cell = cptr->clink;
     if (cell->net_list->nlink->hor is TRUE)
       hornet = cell->net_list->nlink;
     else
       hornet = cell->net_list->nextPtr->nlink;
     minx = INFINITY;
     maxx = 0;
     for(cptr2=hornet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr) {
       minx = MIN(minx,cptr2->clink->rect.x);
       maxx = MAX(maxx,cptr2->clink->rect.x);
     }
     for(i=MIN(minx,col);i<=MAX(maxx,col);i++)
       hor_loc[i] = FALSE;
     for(i=1;i<=cell_num;i++) {
       if (cells[i]->rect.y isnt cell->rect.y || cells[i] is cell) continue;
       if (cells[i]->net_list->nlink->hor is TRUE) 
         chornet = cells[i]->net_list->nlink;
       else 
         chornet = cells[i]->net_list->nextPtr->nlink;
       if (chornet is hornet) continue;
       cminx = INFINITY;
       cmaxx = 0;
       for(ccptr=chornet->cell_list;ccptr!=NIL(struct cell_ptr);
                                             ccptr=ccptr->nextPtr) {
         cminx = MIN(ccptr->clink->rect.x,cminx);
         cmaxx = MAX(ccptr->clink->rect.x,cmaxx);
       }
       for(r=cminx;r<=cmaxx;r++) 
         hor_loc[r] = TRUE ;
     }
     for(i=MIN(minx,col);i<=MAX(maxx,col);i++)
       if (hor_loc[i] is TRUE)
         yes[cell->num] = TRUE;
   }
/*
*  try to put the cells on it,adding rows when required.
*/
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cell = cptr->clink;
     if (yes[cell->num] is FALSE && fixed_loc[col][cell->rect.y] is FALSE) {
       cell->xfix = TRUE;
       cell->rect.x = col;
       fixed_loc[col][cell->rect.y] = TRUE;
     }
   }
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cell = cptr->clink;
     aldone = FALSE;
     if ((fixed_loc[col][cell->rect.y] is TRUE || aldone is FALSE )&& 
                                                   cell->xfix is FALSE) {

/*
*  all is not yet lost,look for rows in the range range_miny to range_maxy
*  and see if the cell's net can be placed on it.
*/
       if (cell->net_list->nlink->hor is TRUE)
         hornet = cell->net_list->nlink;
       else
         hornet = cell->net_list->nextPtr->nlink;
       cminx = INFINITY;
       cmaxx = 0;
       for(cptr2=hornet->cell_list;cptr2!=NIL(struct cell_ptr);
                                               cptr2=cptr2->nextPtr) {
          cminx = MIN(cptr2->clink->rect.x,cminx);
          cmaxx = MAX(cptr2->clink->rect.x,cmaxx);
       }
       cminx = MIN(cminx,col);
       cmaxx = MAX(cmaxx,col);
       if (cmaxx < cminx)
         error("cminx greater than cmaxx");
       row_exist = FALSE;

       for(crow=range_miny;crow<=range_maxy;crow++) {
         cyes[crow] = TRUE;
         if (crow is cell->rect.y ) { 
           cyes[crow] = FALSE;
           continue;
         }
         for(cptr2=hornet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr) {
           if (cptr2->clink is cell) continue;
           vertnet = NIL(struct net);
           if (cptr2->clink->net_list->nlink->hor is FALSE)
             vertnet = cptr2->clink->net_list->nlink;
           else if (cptr2->clink->type < PADLEFT)
             vertnet = cptr2->clink->net_list->nextPtr->nlink;
           if (vertnet is NIL(struct net)) continue;
           for(i=1;i<=cell_num;i++) {
             if (cells[i]->rect.x is cptr2->clink->rect.x &&
                                        cells[i]->xfix is TRUE ) {
               if ( (cells[i]->rect.y >= cptr2->clink->rect.y
                    && cells[i]->rect.y <= crow )||
                    (cells[i]->rect.y <= cptr2->clink->rect.y &&
                     cells[i]->rect.y >= crow) ) {
                 cvertnet = NIL(struct net);
                 if (cptr2->clink->net_list->nlink->hor is FALSE)
                   cvertnet = cptr2->clink->net_list->nlink;
                 else if (cptr2->clink->type < PADLEFT)
                   cvertnet = cptr2->clink->net_list->nextPtr->nlink;
                 if (cvertnet is NIL(struct net)) continue;
                 if (cvertnet isnt vertnet)
                   cyes[crow] = FALSE;
               }
             }
           }
         }
     
         for(ccol=cminx;ccol<=cmaxx;ccol++)
           hor_loc[ccol] = FALSE;
         for(i=1;i<=cell_num;i++) {
           if (cells[i]->rect.y isnt crow) continue;
           if (cells[i]->net_list->nlink->hor is TRUE) 
             chornet = cells[i]->net_list->nlink;
           else 
             chornet = cells[i]->net_list->nextPtr->nlink;
/*
*  need to check that moving the cells to the row will not mess up their
*  vertical nets.
*/
           ccminx = INFINITY;
           ccmaxx = 0;
           for(ccptr=chornet->cell_list;ccptr!=NIL(struct cell_ptr);
                                                 ccptr=ccptr->nextPtr) {
             ccminx = MIN(ccptr->clink->rect.x,ccminx);
             ccmaxx = MAX(ccptr->clink->rect.x,ccmaxx);
           }
           for(ccol=ccminx;ccol<=ccmaxx;ccol++) 
             hor_loc[ccol]  = TRUE ;
         }
         for(ccol=cminx;ccol<=cmaxx;ccol++) {
           if (hor_loc[ccol] is TRUE ) {
              cyes[crow] = FALSE;
              break;
           }
         }
         if (cyes[crow] is TRUE)
           row_exist = TRUE;
       }
       if (row_exist is TRUE) {
/*
*  find closest row or row with closest size.
*/
         diff = INFINITY;
         for(crow=range_miny;crow<=range_maxy;crow++) {
           if (cyes[crow] is TRUE) {
             if (varwidth is TRUE) {
               psize = 0;
               for(j=1;j<=net_num;j++) {
                 if (nets[j]->hor is FALSE) continue;
                 if (nets[j]->cell_list->clink->rect.y is crow &&
                                                     nets[j] isnt hornet)
                    psize = MAX(psize,nets[j]->size);
               }
               if (diff > abs(psize - hornet->size)) {
                 sel_row = crow;
                 diff = abs(psize - hornet->size);
               }
             }
             else {
               if (diff > abs(crow - cell->rect.y)) {
                 diff = abs(crow - cell->rect.y);
                 sel_row = crow;
               }
             }
           }
         }
/*
*  place cells on it.
*/
         for(ccptr=hornet->cell_list;ccptr!=NIL(struct cell_ptr);
                                                 ccptr=ccptr->nextPtr) 
           ccptr->clink->rect.y = sel_row;
         cell->xfix = TRUE;
         cell->rect.x = col;
         fixed_loc[col][sel_row] = TRUE;
       }
       else {
/*
*  have to add a row.
*/
         row = add_row(cell->rect.y,cell->rect.y + 1);
         for(cptr2=hornet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr)
           cptr2->clink->rect.y = row;
         cell->rect.x = col;
         cell->xfix = TRUE;
         fixed_loc[col][row] = TRUE;
       }
     }
   }
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cptr->clink->xfix = TRUE; 
   }
   for(i=miny;i<=maxy;i++)
     fixed_loc[col][i] = TRUE;
   net->straightened = TRUE;
   FREE(yes);
   FREE(cyes);
   FREE(poss_col);
   FREE(nrows);
   FREE(hor_loc);
   FREE(all_occ);
  return;

} /* end of add_col_rows_for_net */


add_row_cols_for_net(net,minx,maxx)
NET net;
int minx,maxx;
{

  int miny,maxy,lminy,lmaxy,row,*ncols,ncost;
  int rownum,cost,*yes,col,i,j,num_cells,same;
  int r,s,aldone,*poss_row,*cyes;
  int can_move ,c_mov,loc_y,ran_minx,ran_maxx;
  int range_minx,range_maxx,diff,col_exist,sel_col,ccol,psize;
  int crow,cminy,ccminy,*ver_loc,ccmaxy,cmaxy,*all_occ;
  CELL_PTR cptr,cptr2,ccptr;
  NET hornet,chornet,vertnet,cvertnet;
  CELL cell;
/*
*  add a row and/or column(s) and change the cell locations.
*  First try to find a row which maximises the number of cells
*  which can be put on it without adding columns and disrupting vertical
*  nets.
*/

   lminy = INFINITY;
   lmaxy = 0;
   poss_row =  ALLOC (int,num_rows *3 + 10);
   ncols =  ALLOC (int,MAX(MHsize,MVsize) + 2);
   cyes = ALLOC (int,MAX(MHsize,MVsize) + 2);
   ver_loc =  ALLOC (int,MAX(MHsize,MVsize) + 2);
   all_occ = ALLOC (int,MAX(MHsize,MVsize) + 2);
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     lminy = MIN(lminy,cptr->clink->rect.y);
     lmaxy = MAX(lmaxy,cptr->clink->rect.y);
   }
   if (lminy > lmaxy)
    error("lminy greater than lmaxy in add_row_cols_for_net");
/*
* check if any row can be used.
*/
   for(row=lminy-1;row<=lmaxy+1;row++) {
     for(i=1;i<=MVsize;i++)
       all_occ[i] = FALSE;
     poss_row[row] = TRUE;
     for(i=minx;i<=maxx;i++) {
       if (fixed_loc[i][row] is TRUE)
          poss_row[row] = FALSE;
     }
     for(i=1;i<=gate_num;i++) {
       if (cells[i]->rect.y isnt row) continue;
       if ( colonly is TRUE ) {
         same = FALSE;
         for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
           if (cptr->clink is cells[i])
             same = TRUE;
         if (same is FALSE ) {
             poss_row[row] = FALSE;
         }
       }
       if (cells[i]->rect.x < minx || cells[i]->rect.x > maxx) continue;
       same = FALSE;
       for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
         if (cptr->clink is cells[i])
           same = TRUE;
       if (same is FALSE ) {
         if (cells[i]->yfix is TRUE)
           poss_row[row] = FALSE;
         else {
/*
*  check if cell can be moved to some other location on its vertical net.
*/
           if (cells[i]->net_list->nlink->hor is FALSE)
             vertnet = cells[i]->net_list->nlink;
           else
             vertnet = cells[i]->net_list->nextPtr->nlink;
           miny = INFINITY;
           maxy = 0;
           for(cptr=vertnet->cell_list;cptr!=NIL(struct cell_ptr);
                                                  cptr=cptr->nextPtr) {
             miny = MIN(miny,cptr->clink->rect.y);
             maxy = MAX(maxy,cptr->clink->rect.y);
           }
           if (miny > maxy)
             error("miny greater than maxy");
           can_move = FALSE;
           for(s=miny;s<=maxy;s++) {
             c_mov = TRUE;
             if (fixed_loc[vertnet->cell_list->clink->rect.x][s] is TRUE ||
                                                         all_occ[s] is TRUE)
               c_mov = FALSE;
             for(cptr=vertnet->cell_list;cptr!=NIL(struct cell_ptr);
                                                    cptr=cptr->nextPtr) {
               if (cptr->clink->rect.y is s) {
                 c_mov = FALSE;
                 break;
               }
             }
             if (c_mov is TRUE) {
               can_move = TRUE;
               all_occ[s] = TRUE;
               break;
             }
           }
           if (can_move is FALSE) 
             poss_row[row] = FALSE;
         }
       }
     }
   }
  
   for(row=lminy-1;row<=lmaxy+1;row++) {
     ncols[row] = 0;
     for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
       if (cptr->clink->type >= PADLEFT) continue;
       cell = cptr->clink;
       if (cell->net_list->nlink->hor is FALSE)
         vertnet = cell->net_list->nlink;
       else
         vertnet = cell->net_list->nextPtr->nlink;
       miny = INFINITY;
       maxy = 0;
       num_cells = 0;
       for(cptr2=vertnet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                   cptr2=cptr2->nextPtr) {
         miny = MIN(miny,cptr2->clink->rect.y);
         maxy = MAX(maxy,cptr2->clink->rect.y);
         num_cells++;
       }
       if (miny > maxy)
         error("miny greater than maxy in add_row_cols_for_net");
       if (row < miny || row > maxy && num_cells > 1) {
         if (varwidth is TRUE)
           ncols[row] += vertnet->size;
         else
           ncols[row] += 1;
       }
     }
   }
/*
*  pick the row with least cost
*/
   row = 0;cost = INFINITY;
   for(row=lminy-1;row<=lmaxy+1;row++) {
     if (varwidth is TRUE)
       ncost = ncols[row];
     else
       ncost = ncols[row]*2;
     if (poss_row[row] is FALSE) {
/*
* cost of adding a row may be higher than that of adding a column.
* Area consideration.
*/
       if (varwidth is TRUE)
         ncost += net->size;
       else
         ncost += 2;
     }
     if (cost > ncost) {
       cost = ncost;
       rownum = row;
     }
   }
   if (rownum is 0)
     error("couldn't find a row with less than infinite cost");
/*
*  add the row if necessary.find the troublesome cells.
*/
   if (poss_row[rownum] is FALSE) 
     row = add_row(rownum,rownum+1);
   else 
     row = rownum;
/*
*  find the range of the net on the row.Cells may be placed anywhere 
*  on this range.
*/
   range_minx = 0;
   range_maxx = INFINITY;
   ran_minx = INFINITY;
   ran_maxx = 0;
   for(i=1;i<=cell_num;i++) {
     ran_minx = MIN(ran_minx,cells[i]->rect.x);
     ran_maxx = MAX(ran_maxx,cells[i]->rect.x);
     if (cells[i]->rect.y is row) {
       if (cells[i]->rect.x < minx)
         range_minx = MAX(range_minx,cells[i]->rect.x);
       if (cells[i]->rect.x > maxx) 
         range_maxx = MIN(range_maxx,cells[i]->rect.x);
     }
   }
   range_minx = MAX(range_minx,ran_minx);
   range_maxx = MIN(range_maxx,ran_maxx);
   if (poss_row[row] is TRUE) {
     for(i=1;i<=gate_num;i++) {
       if (cells[i]->rect.y isnt row) continue;
       if (cells[i]->rect.x < minx || cells[i]->rect.x > maxx) continue;
       same = FALSE;
       for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) 
         if (cptr->clink is cells[i])
           same = TRUE;
       if (same is FALSE) {
         if (cells[i]->yfix is TRUE)
           error("have to add row here");
         else {
/*
* cell should be able to move to some location in the range of its vertical net.
*/
           if (cells[i]->net_list->nlink->hor is FALSE)
             vertnet = cells[i]->net_list->nlink;
           else
             vertnet = cells[i]->net_list->nextPtr->nlink;
           miny = INFINITY;
           maxy = 0;
           for(cptr=vertnet->cell_list;cptr!=NIL(struct cell_ptr);
                                                  cptr=cptr->nextPtr) {
             miny = MIN(miny,cptr->clink->rect.y);
             maxy = MAX(maxy,cptr->clink->rect.y);
           }
           if (miny > maxy)
             error("miny greater than maxy");
           can_move = FALSE;diff = INFINITY;
           for(s=miny;s<=maxy;s++) {
             c_mov = TRUE;
             if (fixed_loc[vertnet->cell_list->clink->rect.x][s] is TRUE)
               c_mov = FALSE;
             for(cptr=vertnet->cell_list;cptr!=NIL(struct cell_ptr);
                                                    cptr=cptr->nextPtr) {
               if (cptr->clink->rect.y is s) {
                 c_mov = FALSE;
                 break;
               }
             }
             if (c_mov is TRUE) {
               can_move = TRUE;
               if (diff > abs(row - s)) {
                 loc_y = s;
                 diff = abs(row - s);
               }
             }
           }
           if (can_move is FALSE) 
             error("have to add a row here");
           else 
/*
*  make the move.
*/
             cells[i]->rect.y = loc_y;
         }
       }
     }
   }
   yes = ALLOC (int,cell_num + 1);
   for(i=1;i<=cell_num;i++)
     yes[i] = FALSE;
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     if (cptr->clink->type >= PADLEFT) continue;
     cell = cptr->clink;
     if (cell->net_list->nlink->hor is FALSE)
       vertnet = cell->net_list->nlink;
     else
       vertnet = cell->net_list->nextPtr->nlink;
     miny = INFINITY;
     maxy = 0;
     for(cptr2=vertnet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr) {
       miny = MIN(miny,cptr2->clink->rect.y);
       maxy = MAX(maxy,cptr2->clink->rect.y);
     }
     for(i=MIN(miny,row);i<=MAX(maxy,row);i++)
       ver_loc[i] = FALSE;
     for(i=1;i<=gate_num;i++) {
       if (cells[i]->rect.x isnt cell->rect.x || cells[i] is cell) continue;
       if (cells[i]->net_list->nlink->hor is FALSE) 
         cvertnet = cells[i]->net_list->nlink;
       else 
         cvertnet = cells[i]->net_list->nextPtr->nlink;
       if (cvertnet is vertnet) continue;
       cminy = INFINITY;
       cmaxy = 0;
       for(ccptr=cvertnet->cell_list;ccptr!=NIL(struct cell_ptr);
                                             ccptr=ccptr->nextPtr) {
         cminy = MIN(ccptr->clink->rect.y,cminy);
         cmaxy = MAX(ccptr->clink->rect.y,cmaxy);
       }
       for(r=cminy;r<=cmaxy;r++) 
         ver_loc[r] = TRUE ;
     }
     for(i=MIN(miny,row);i<=MAX(maxy,row);i++)
       if (ver_loc[i] is TRUE)
         yes[cell->num] = TRUE;
   }
/*
*  try to put the cells on it,adding columns when required.
*/
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cell = cptr->clink;
     if (yes[cell->num] is FALSE && fixed_loc[cell->rect.x][row] is FALSE) {
       cell->yfix = TRUE;
       cell->rect.y = row;
       fixed_loc[cell->rect.x][row] = TRUE;
     }
   }
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     if (cptr->clink->type >= PADLEFT) continue;
     cell = cptr->clink;
     aldone = FALSE;
     if ((fixed_loc[cell->rect.x][row] is TRUE || aldone is FALSE )&& 
                                                   cell->yfix is FALSE) {

/*
*  all is not yet lost,look for rows in the range range_minx to range_maxx
*  and see if the cell's net can be placed on it.
*/
       if (cell->net_list->nlink->hor is FALSE)
         vertnet = cell->net_list->nlink;
       else
         vertnet = cell->net_list->nextPtr->nlink;
       cminy = INFINITY;
       cmaxy = 0;
       for(cptr2=vertnet->cell_list;cptr2!=NIL(struct cell_ptr);
                                               cptr2=cptr2->nextPtr) {
          cminy = MIN(cptr2->clink->rect.y,cminy);
          cmaxy = MAX(cptr2->clink->rect.y,cmaxy);
       }
       cminy = MIN(cminy,row);
       cmaxy = MAX(cmaxy,row);
       if (cmaxy < cminy)
         error("cminy greater than cmaxy");
       col_exist = FALSE;

       for(ccol=range_minx;ccol<=range_maxx;ccol++) {
         cyes[ccol] = TRUE;
         if (ccol is cell->rect.x ) { 
           cyes[ccol] = FALSE;
           continue;
         }
         for(cptr2=vertnet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr) {
           if (cptr2->clink is cell) continue;
           hornet = NIL(struct net);
           if (cptr2->clink->net_list->nlink->hor is TRUE)
             hornet = cptr2->clink->net_list->nlink;
           else if (cptr2->clink->type < PADLEFT)
             hornet = cptr2->clink->net_list->nextPtr->nlink;
           if (hornet is NIL(struct net)) continue;
           for(i=1;i<=cell_num;i++) {
             if (cells[i]->rect.y is cptr2->clink->rect.y &&
                                        cells[i]->yfix is TRUE ) {
               if ( (cells[i]->rect.x >= cptr2->clink->rect.x
                    && cells[i]->rect.x <= ccol )||
                    (cells[i]->rect.x <= cptr2->clink->rect.x &&
                     cells[i]->rect.x >= ccol) ) {
                 chornet = NIL(struct net);
                 if (cptr2->clink->net_list->nlink->hor is TRUE)
                   chornet = cptr2->clink->net_list->nlink;
                 else if (cptr2->clink->type < PADLEFT)
                   chornet = cptr2->clink->net_list->nextPtr->nlink;
                 if (chornet is NIL(struct net)) continue;
                 if (chornet isnt hornet)
                   cyes[ccol] = FALSE;
               }
             }
           }
         }
     
         for(crow=cminy;crow<=cmaxy;crow++)
           ver_loc[crow] = FALSE;
         for(i=1;i<=cell_num;i++) {
           if (cells[i]->rect.x isnt ccol) continue;
           if (cells[i]->net_list->nlink->hor is FALSE) 
             cvertnet = cells[i]->net_list->nlink;
           else if (cells[i]->type < PADLEFT)
             cvertnet = cells[i]->net_list->nextPtr->nlink;
           else
             continue;
/*
*  need to check that moving the cells to the column will not mess up their
*  horizontal nets.
*/
           ccminy = INFINITY;
           ccmaxy = 0;
           for(ccptr=cvertnet->cell_list;ccptr!=NIL(struct cell_ptr);
                                                 ccptr=ccptr->nextPtr) {
             ccminy = MIN(ccptr->clink->rect.y,ccminy);
             ccmaxy = MAX(ccptr->clink->rect.y,ccmaxy);
           }
           for(crow=ccminy;crow<=ccmaxy;crow++) 
             ver_loc[crow]  = TRUE ;
         }
         for(crow=cminy;crow<=cmaxy;crow++) {
           if (ver_loc[crow] is TRUE ) {
              cyes[ccol] = FALSE;
              break;
           }
         }
         if (cyes[ccol] is TRUE)
           col_exist = TRUE;
       }
       if (col_exist is TRUE) {
/*
*  find closest column.
*/
         diff = INFINITY;
         for(ccol=range_minx;ccol<=range_maxx;ccol++) {
           if (cyes[ccol] is TRUE) {
             if (varwidth is TRUE) {
               psize = 0;
               for(j=1;j<=net_num;j++) {
                 if (nets[j]->hor is TRUE) continue;
                 if (nets[j]->cell_list->clink->rect.x is ccol &&
                                                        nets[j] isnt vertnet)
                    psize = MAX(psize,nets[j]->size);
               }
               if (diff > abs(psize - vertnet->size)) {
                 sel_col = ccol;
                 diff = abs(psize - vertnet->size);
               }
             }
             else {
               if (diff > abs(ccol - cell->rect.x)) {
                 diff = abs(ccol - cell->rect.x);
                 sel_col = ccol;
               }
             }
           }
         }
/*
*  place cells on it.
*/
         for(ccptr=vertnet->cell_list;ccptr!=NIL(struct cell_ptr);
                                                 ccptr=ccptr->nextPtr) 
           ccptr->clink->rect.x = sel_col;
         cell->yfix = TRUE;
         cell->rect.y = row;
         fixed_loc[sel_col][row] = TRUE;
       }
       else {
/*
*  have to add a column.
*/
         col = add_col(cell->rect.x,cell->rect.x + 1);
         for(cptr2=vertnet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                 cptr2=cptr2->nextPtr)
           cptr2->clink->rect.x = col;
         cell->rect.y = row;
         cell->yfix = TRUE;
         fixed_loc[col][row] = TRUE;
       }
     }
   }
   for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
     cptr->clink->yfix = TRUE; 
   }
   for(i=minx;i<=maxx;i++)
     fixed_loc[i][row] = TRUE;
   net->straightened = TRUE;
   FREE(yes);
   FREE(cyes);
   FREE(poss_row);
   FREE(ncols);
   FREE(ver_loc);
   FREE(all_occ);
  return;

} /* end of add_row_cols_for_net */


#ifndef lint
plot_cells(outFile)
STRING outFile;
{
 
 FILE *Output;
 int i;
 double l,w,max_l,max_w;
 CELL_PTR cptr;
  
  if ((Output = fopen(outFile,"w")) is NULL)
    error("couldn't open output file");
  
  max_l = max_w = 0;
  for(i=1;i<=gate_num;i++) {
    max_l = MAX(max_l,cells[i]->stdType->l);
    max_w = MAX(max_w,cells[i]->stdType->w);
  }
  for(i=1;i<=cell_num;i++) {
    if (cells[i]->type is CONNECT && pla is TRUE) continue;
    if (varwidth is FALSE || cells[i]->type >= PADLEFT) {
      (void) fprintf(Output,"%d %d \n",cells[i]->rect.x*3-1,cells[i]->rect.y*3-1);
      (void) fprintf(Output,"%d %d \n",cells[i]->rect.x*3-1,cells[i]->rect.y*3+1);
      (void) fprintf(Output,"%d %d \n",cells[i]->rect.x*3+1,cells[i]->rect.y*3+1);
      (void) fprintf(Output,"%d %d \n",cells[i]->rect.x*3+1,cells[i]->rect.y*3-1);
      if (cells[i]->special is TRUE) 
        (void) fprintf(Output,"%d %d \",\"\n",cells[i]->rect.x*3-1,
                                                cells[i]->rect.y*3-1);
      else
        (void) fprintf(Output,"%d %d \" \"\n",cells[i]->rect.x*3-1,
                                                cells[i]->rect.y*3-1);
    }
    else {
      l = (double) cells[i]->stdType->l/max_l;
      w = (double) cells[i]->stdType->w/max_w;
      (void) fprintf(Output,"%f %f \n",cells[i]->rect.x*3-w,cells[i]->rect.y*3-l);
      (void) fprintf(Output,"%f %f \n",cells[i]->rect.x*3-w,cells[i]->rect.y*3+l);
      (void) fprintf(Output,"%f %f \n",cells[i]->rect.x*3+w,cells[i]->rect.y*3+l);
      (void) fprintf(Output,"%f %f \n",cells[i]->rect.x*3+w,cells[i]->rect.y*3-l);
      if (cells[i]->special is TRUE) 
        (void) fprintf(Output,"%f %f \",\"\n",cells[i]->rect.x*3-w,
                                                cells[i]->rect.y*3-l);
      else
        (void) fprintf(Output,"%f %f \" \"\n",cells[i]->rect.x*3-w,
                                                cells[i]->rect.y*3-l);
    }
  }
  for(i=1;i<=net_num;i++) {
    if (nets[i]->type is CONNECT && pla is TRUE) continue;
    for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
      if (cptr->clink->type is CONNECT && pla is TRUE) continue;
      if (cptr->nextPtr isnt NIL(struct cell_ptr)) {
        (void) fprintf(Output,"%d %d \n",cptr->clink->rect.x*3,
                                         cptr->clink->rect.y *3); 
      }
      else {
        (void) fprintf(Output,"%d %d \" \"\n",cptr->clink->rect.x*3,
                                         cptr->clink->rect.y*3); 
      }
    }
  }

 (void) fclose(Output);
 return;

} /* end of plot_cells */
#endif
