#include "copyright.h"
#include "genie.h"
#include "oct.h"
#include "oh.h"

/*
 *  routines to interface to oct database.
 */

#define FUN_LENGTH 500
#define ROWH 3
#define COLV 4
#define HORI 0
#define VERT 1
#define NUM_FOLDS 30

#define ANDPLANE 1
#define ORPLANE 2

char TRAN1[] = "t1";
char TRAN2[] = "t2";
char TRAN3[] = "t3";
char TRAN4[] = "t4";

char ETLIB[1024];
int minx, miny;

oct_write(name)
char *name;
{
    octObject new_facet(), new_bag();
    octObject cell;
    octObject arrayBag, marrayBag, instBag;
    octObject masterColBag, masterRowBag;
    octObject andRowBag, orRowBag;
    octObject folRowBag, folColBag;
    octObject rowBag, colBag;
    int foldedrow, foldedcol;
    int i,j,plane;
    int prev_col,prev_row;
    int numcol, numrow, netcol, netrow;
    int **vnets, **hnets, *net_order;
    int tran_num;
    int hor_net_num, vert_net_num;
    NET net;
    char bagName[ID_LENGTH];
    CELL_PTR cptr;

    /* sanitize pla, get rid of connecting cells and nets, coordinates
       of cells and nets start from 1 */
    for(i=1;i<=net_num;i++) {
      net = nets[i];
      if (net->hor is TRUE) {
        net->col = INFINITY;
        for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr)
          net->col = MIN(net->col,cptr->clink->rect.x);
      }
    }
    if (pla is TRUE)
      sanitize_pla();
  
    (void) sprintf(ETLIB, "%s/%s", "$OCTTOOLS", "lib/technology/scmos/octopus");

    /* start the oct write */
    octBegin();

    /* create the cell */
    cell = new_facet(name, "unspaced");

    /* create the arrayStructure octBag  */
     arrayBag = new_bag(&cell,"GENIESTRUCTURE");
     marrayBag = new_bag(&cell,"MKARRAYSTRUCTURE");
     instBag = new_bag(&cell,"INSTANCES");

     /* master row bag which contains AND and OR planes */
     masterRowBag = new_bag(&cell,"ROW");
     OH_ASSERT(octAttach(&arrayBag,&masterRowBag));
     new_prop(&masterRowBag,"PURPOSE","master_row");

    /* rowBag which connects all column bags in the AND plane in order */
     andRowBag = new_bag(&cell,"ROW");
     OH_ASSERT(octAttach(&masterRowBag,&andRowBag));
     new_prop(&andRowBag,"PURPOSE","rowANDPlane");

    /* rowBag which connects all column bags in the OR plane in order */
     orRowBag = new_bag(&cell,"ROW");
     OH_ASSERT(octAttach(&masterRowBag,&orRowBag));
     new_prop(&orRowBag,"PURPOSE","rowORPlane");

    /* order the cells in horizontal nets */
    order_cells_within_net(2);

    net_order =  ALLOC(int,net_num+1);

    /* order the horizontal nets in increasing y */
    hnets = ALLOC(int *,net_num+1);
    for(i=0;i<net_num+1;i++)
      hnets[i] = ALLOC(int,NUM_FOLDS);

    j = 0;
    for(i=1;i<=net_num;i++) {
      if (nets[i]->hor is TRUE) {
        net_order[j] = nets[i]->num;
        j++;
      }
    }
    hor_net_num = j;
    sort_hor_or_vert_nets(net_order,hor_net_num,ROWH);

    /* put into 2-D array so we know what is in the same column (row) */
    prev_row = 0; numrow = 0; netrow = 1;
    for(i=0;i < hor_net_num;i++) {
      if (prev_row isnt nets[net_order[i]]->row) {
        hnets[numrow][netrow+1] = END;
        numrow++; netrow = 1;
        hnets[numrow][netrow] = net_order[i];
      }
      else {
        netrow++;
        hnets[numrow][netrow] = net_order[i];
      }
      prev_row = nets[net_order[i]]->row;
    } 
    hnets[numrow][netrow+1] = END;
   
    /* begin the write */
    plane = ANDPLANE; tran_num = 0;
    for(i=1;i<= numrow;i++) {

      j = 1;
      while ( hnets[i][j+1] isnt END) j++;
      if ( j > 1 ) foldedrow = TRUE;
      else foldedrow = FALSE;
      if (foldedrow is TRUE) {
         
        /* this is a folded row, allocate a foldedColumn */
        folColBag = new_bag(&cell,"COLUMN");
        new_prop(&folColBag,"PURPOSE","folded_col");

        net = nets[hnets[i][1]];
        if (net->type is PCHANNEL) plane = ANDPLANE;
        else if (net->type is NCHANNEL) plane = ORPLANE;
        else crash("horiz. net type not defined oct_write");

        if (plane is ANDPLANE)
          OH_ASSERT(octAttach(&andRowBag,&folColBag));
        else if (plane is ORPLANE)
          OH_ASSERT(octAttach(&orRowBag,&folColBag));
        j = 1;
        while ( hnets[i][j] isnt END ) {
          
          /* allocate column bag for this horizontal net */
          net = nets[hnets[i][j]];

          colBag = new_bag(&cell,"COLUMN");
          new_prop(&colBag,"PURPOSE",net->netName);
          OH_ASSERT(octAttach(&folColBag,&colBag));

          if (net->type is PCHANNEL) plane = ANDPLANE;
          else if (net->type is NCHANNEL) plane = ORPLANE;
          else crash("horiz. net type not defined oct_write");

          /* routine which takes a net and makes instances of the cells
             stores oct objectId in cells ; pads are thrown away */
          net_to_bag(&cell,&instBag,net,&colBag,&tran_num,TRUE);
            
          j++;
        }
      }
      else {
        net = nets[hnets[i][1]];
        if (net->type is PCHANNEL) plane = ANDPLANE;
        else if (net->type is NCHANNEL) plane = ORPLANE;
        else crash("horiz. net type not defined oct_write");

        colBag = new_bag(&cell,"COLUMN");
        new_prop(&colBag,"PURPOSE",net->netName);

        if (plane is ANDPLANE)
          OH_ASSERT(octAttach(&andRowBag,&colBag));
        else if (plane is ORPLANE)
          OH_ASSERT(octAttach(&orRowBag,&colBag));
        net_to_bag(&cell,&instBag,net,&colBag,&tran_num,TRUE);
      }
    }


    /* order the cells in the vertical nets */
    order_cells_within_net(1);

    /* order the vertical nets in increasing x */
    vnets = ALLOC(int *,net_num+1);
    for(i=0;i<net_num+1;i++)
      vnets[i] = ALLOC(int,NUM_FOLDS);

    j = 0;
    for(i=1;i<=net_num;i++) {
      
      if (nets[i]->hor is FALSE &&
                   (pla is FALSE || nets[i]->type isnt CONNECT) ) {
        net_order[j] = nets[i]->num;

        /* for sorting, we want the leftmost net on the same row first */
        nets[i]->row = nets[i]->cell_list->clink->rect.y;
        j++;
      }
    }
    vert_net_num = j;
    sort_hor_or_vert_nets(net_order,vert_net_num,COLV);

    /* put into 2-D array so we know what is in
                             the same row (column for me) */
    prev_col = 0; numcol = 0; netcol = 1;
    for(i=0;i < vert_net_num;i++) {
      if (prev_col isnt nets[net_order[i]]->col) {
        vnets[numcol][netcol+1] = END;
        numcol++; netcol = 1;
        vnets[numcol][netcol] = net_order[i];
      }
      else {
        netcol++;
        vnets[numcol][netcol] = net_order[i];
      }
      prev_col = nets[net_order[i]]->col;
    } 
    vnets[numcol][netcol+1] = END;
   
    /* begin the write */
    masterColBag = new_bag(&cell,"COLUMN");
    new_prop(&masterColBag,"PURPOSE","master_column");
    OH_ASSERT(octAttach(&arrayBag,&masterColBag));

    for(i=1;i<= numcol;i++) {

      j = 1;
      while ( vnets[i][j+1] isnt END) j++;
      if ( j > 1 ) foldedcol = TRUE;
      else foldedcol = FALSE;
      if (foldedcol is TRUE) {
         
        /* this is a folded column, allocate a foldedRow */
        folRowBag = new_bag(&cell,"ROW");
        OH_ASSERT(octAttach(&masterColBag,&folRowBag));
        new_prop(&folRowBag,"PURPOSE","folded_row");

        j = 1;
        while ( vnets[i][j] isnt END ) {
          net = nets[vnets[i][j]];

          /* allocate row bag for this vertical net */
          rowBag = new_bag(&cell,"ROW");
          OH_ASSERT(octAttach(&folRowBag,&rowBag));

          (void) sprintf(bagName,"row_%d",i);
          new_prop(&rowBag,"PURPOSE",bagName);

          /* routine which takes a net, grabs instances via objectId on
             cells connected to net and attaches to bag */
          net_to_bag(&cell,&instBag,net,&rowBag,&tran_num,FALSE);
            
          j++;
        }
      }
      else {
        net = nets[vnets[i][1]];

        rowBag = new_bag(&cell,"ROW");
        OH_ASSERT(octAttach(&masterColBag,&rowBag));

        (void) sprintf(bagName,"row_%d",i);
        new_prop(&rowBag,"PURPOSE",bagName);

        net_to_bag(&cell,&instBag,net,&rowBag,&tran_num,FALSE);
      }
    }


    FREE(vnets); FREE(hnets);
    FREE(net_order);

    /* close the facet */
    if (octCloseFacet(&cell) != OCT_OK) {
        octError("Closing Facet");
	exit(-1);
    }
    octEnd();

} /* end of oct_write */


CELL
make_or_find_cell(hornet,vernet,old)
NET hornet,vernet;
int *old;
{
/*
* find a cell common to hornet, and vernet.
*/
 CELL cell,findCell();
 CELL_PTR cptr2,cptr;
 char cellName[ID_LENGTH];


  for(cptr=hornet->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
    cell = cptr->clink;
    for(cptr2=vernet->cell_list;cptr2!=NIL(struct cell_ptr);
                                                   cptr2=cptr2->nextPtr) {
      if (cptr2->clink is cell) {
        *old = TRUE;
        return(cell);
      }
    }
  }
  /* couldn't find cell, make a new one */
  (void) sprintf(cellName,"cell_%d",cell_num+1);
  cell = findCell(cellName);
  *old = FALSE;
  return(cell);

} /* end of make_or_find_cell */


order_cells_within_net(type)
int type;
{

  NET_PTR nptr,nnptr;
  CELL_PTR cptr;
  int *net_cells;
  int i,j,numcells;

  
   /* Ordering of cells according to x, y coords within a net*/
   net_cells =  ALLOC(int,cell_num+10);
   for(i=1;i<=cell_num;i++) {
     if (cells[i]->type >= PADLEFT || 
                     cells[i]->net_list is NIL(struct net_ptr)) continue;
     nptr = cells[i]->net_list;
     nnptr = cells[i]->net_list->nextPtr;
     if (type is 1) {
       if (nptr->nlink->hor is TRUE) {
         nptr->nextPtr = NIL(struct net_ptr);
       }
       else {
         cells[i]->net_list = nnptr;
       }
     }
     else if (type is 2) {
       if (nptr->nlink->hor is FALSE) {
         nptr->nextPtr = NIL(struct net_ptr);
       }
       else {
         cells[i]->net_list = nnptr;
       }
     }
   }

   for(i=1;i<=net_num;i++) {
     if (type is 1 && nets[i]->hor is TRUE) continue;
     if (type is 2 && nets[i]->hor is FALSE) continue;
     numcells = 0;
     for(cptr=nets[i]->cell_list;cptr!=NIL(struct cell_ptr);
                                                   cptr=cptr->nextPtr) {
       if (cptr->clink->type < PADLEFT) {
         net_cells[numcells] = cptr->clink->num;
         numcells++;
       }
     }

     /* sort with increasing x, or increasing y */
     if (numcells > 1) {
       sort_cells(net_cells,numcells,type);
     }

     if (nets[i]->cell_list isnt NIL(struct cell_ptr))
       FREE(nets[i]->cell_list);
     nets[i]->cell_list = NIL(struct cell_ptr);
     for(j=numcells-1;j>=0;j--)
       add_net_cell(cells[net_cells[j]],nets[i]);
   }
   FREE(net_cells);

 return;

} /* end of order_cells_within_net */



net_to_bag(facet,instBag,net,bag,num,new)
octObject *facet,*instBag;
NET net;
octObject *bag;
int *num;
int new;
{

 octObject inst;
 CELL_PTR cptr;
 char TRAN[ID_LENGTH];


  if (new is TRUE) {
    for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
      if (cptr->clink->type is CONNECT) continue;
      if (net->type is PCHANNEL) {
        if ( (cptr->clink->rect.y - miny + 1) % 2 is 1)
          (void) strcpy(TRAN,TRAN1);
        else
          (void) strcpy(TRAN,TRAN2);
      }
      else if (net->type is NCHANNEL) {
        if ( (cptr->clink->rect.x - minx + 1) % 2 is 1)
          (void) strcpy(TRAN,TRAN3);
        else
          (void) strcpy(TRAN,TRAN4);
      }
      make_new_instance(facet,TRAN,num,&inst);
      cptr->clink->octObjectId = inst.objectId;
      OH_ASSERT(octAttach(bag,&inst));
      OH_ASSERT(octAttach(instBag,&inst));
    }
  }
  else {
    for(cptr=net->cell_list;cptr!=NIL(struct cell_ptr);cptr=cptr->nextPtr) {
      if (cptr->clink->type is CONNECT) continue;
      inst.objectId = cptr->clink->octObjectId;
      if (octGetById(&inst) != OCT_OK) {
        octError("couldn't find instance");
        exit(-1);
      }
      OH_ASSERT(octAttach(bag,&inst));
      OH_ASSERT(octAttach(instBag,&inst));
    }
  }
 return;

}  /* end of net_to_bag */


octObject new_facet(name, view)
char *name;
char *view;
{
    octObject *facet;

    facet = ALLOC(octObject, 1);
    facet->type = OCT_FACET;
    facet->contents.facet.cell = name;
    facet->contents.facet.view = view;
    facet->contents.facet.version = OCT_CURRENT_VERSION;
    facet->contents.facet.facet = "contents";
    facet->contents.facet.mode = "w";
    if (octOpenFacet(facet) < OCT_OK) {
	octError("Opening cell");
	exit(-1);
    }

    return *facet;
}


octObject new_bag(container, name)
octObject *container;
char *name;
{
    octObject *bag;

    bag = ALLOC(octObject, 1);
    bag->type = OCT_BAG;
    bag->contents.bag.name = name;
    if (octCreate(container, bag) != OCT_OK) {
	octError("Creating bag");
	exit(-1);
    }

    return *bag;
}



octObject new_generic_inst(container, name, master_name)
octObject *container;
char *name;
char *master_name;
{
    octObject *inst;
    char *str_cat();

    check_master(master_name);

    inst = ALLOC(octObject, 1);
    inst->type = OCT_INSTANCE;
    inst->contents.instance.name = name;
    inst->contents.instance.view = "physical";
    inst->contents.instance.facet = "contents";
    inst->contents.instance.version = OCT_CURRENT_VERSION;
    inst->contents.instance.master = str_cat(ETLIB,master_name);
    inst->contents.instance.transform.translation.x = 0;
    inst->contents.instance.transform.translation.y = 0;
    inst->contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    if (octCreate(container, inst) != OCT_OK) {
	octError("Creating instance");
	exit(-1);
    }

    return *inst;
}

check_master(cell_name)
char *cell_name;
{
    octObject master;
    char *str_cat();

    master.type = OCT_FACET;
    master.contents.facet.cell = str_cat(ETLIB,cell_name);
    master.contents.facet.view = "physical";
    master.contents.facet.version = OCT_CURRENT_VERSION;
    master.contents.facet.facet = "contents";
    master.contents.facet.mode = "r";
    if (octOpenFacet(&master) < OCT_OK) {
       octError("Opening the generic master");
       exit(-1);
    }
    OH_ASSERT(octCloseFacet(&master));

    return;
}


new_prop(container, name, type)
octObject *container;
char *name;
STRING type;
{
    octObject prop;

    prop.type = OCT_PROP;
    prop.objectId = oct_null_id;
    prop.contents.prop.name = name; 
    prop.contents.prop.type = OCT_STRING;    
    prop.contents.prop.value.string = type;    
    if (octCreate(container, &prop) != OCT_OK) {
	octError("Creating cellStatus property");
        exit(-1);
    }

    return;
}


make_new_instance(facet,master_name,num,instance)
octObject *facet;
STRING master_name;
int *num;
octObject *instance;
{

 char *instName;
 octObject new_generic_inst();

  (*num)++;
  instName = ALLOC(char,ID_LENGTH+1);
  (void) sprintf(instName,"%s_%d",master_name,*num);
  *instance = new_generic_inst(facet,instName,master_name);

 return;

} /* end of make_new_instance */


char *str_cat(str1, str2)
char *str1, *str2;
{
    char *result, *r ;

    result = r = ALLOC(char,strlen(str1) + strlen(str2) + 1);
    while (*result++ = *str1++)
	;
    *result--;
    while (*result++ = *str2++)
	;
    *result = '\0';

    return r;
}


sanitize_pla()
{

 int i;
 
  for(i=1;i<=gate_num;i++) 
    if (cells[i]->type is CONNECT)
      delete_net_cell(cells[i],cells[i]->net_list->nextPtr->nlink);

  minx = miny = INFINITY;
  for(i=1;i<=gate_num;i++) {
    minx = MIN(minx,cells[i]->rect.x);
    miny = MIN(miny,cells[i]->rect.y);
  }

  i = gate_num;
  while( i>=1 && cells[i]->type is CONNECT) i--;
  gate_num = i;

  i = net_num;
  while (i >= 1 && nets[i]->type is CONNECT) i--;
  net_num = i;

  for(i=0;i<=num_cols+10;i++) Col[i] = 0;

  for(i=1;i<=gate_num;i++) {
    Col[cells[i]->rect.x-OFFSET] += 1;
  }
  shrink_matrix();

 return;

} /* end of sanitize_pla */
