#include "port.h"
#include "utility.h"
#include "fang.h"
#include "fa.h"

/*
 * faPathDirection: Determines direction of path
 *    returns: 2 -- if path has zero length
 *             1 -- if path is vertical
 *             0 -- if path is horizontal
 *            -1 -- if there is an error
 */
faPathDirection(p1x, p1y, p2x, p2y)
fa_coord p1x, p1y, p2x, p2y;
{
    if ((p1x != p2x) && (p1y != p2y)) return -1;
    if ((p1x == p2x) && (p1y == p2y)) return 2;
    if (p1x == p2x) return 1;
    if (p1y == p2y) return 0;

    /* shouldn't happen */
    return -1;
}


/*
 * faIsBetween: Determines if point1 is between point2 and point3
 *   returns: 1 -- if true
 *            0 -- otherwise
 */
faIsBetween(pt1, pt2, pt3)
int pt1, pt2, pt3;
{
    if ((pt1 >= pt2) && (pt1 <= pt3)) {
	return 1;
    } else if ((pt1 >= pt3) && (pt1 <= pt2)) {
	return 1;
    } else {
	return 0;
    }
}


/*
 * faGenerateBox: Generates a newbox from box and update box
 */
int faGenerateBox(box, newbox)
fa_box **box;
fa_box *newbox;
{
    if (*box == ((fa_box *) 0)) {
	return 0;
    } else {
	*newbox = *(*box);
	newbox->next = ((fa_box *) 0);
	*box = (*box)->next;
	return 1;
    }
}


/*
 * faGenerateRegion: Generates a newregion from region and update region
 */
int faGenerateRegion(region, newregion)
fa_region **region;
fa_region *newregion;
{
    if (*region == ((fa_region *) 0)) {
	return 0;
    } else {
	*newregion = *(*region);
	newregion->next = ((fa_region *) 0);
	*region = (*region)->next;
	return 1;
    }
}


/*
 * faRegionGeo: Converts region to geo
 */
void
faRegionGeo(region, geo)
fa_region *region;
fa_geometry *geo;
{
    fa_init(geo);
    fa_add_regions(geo, region);
    return;
}


/*
 * faRegionEdges: Converts region in edgelists
 */
void
faRegionEdges(region, h_list, v_list)
fa_region *region;
fa_edge **h_list, **v_list;
{
    fa_geometry geo;
    
    fa_init(&geo);
    fa_add_regions(&geo, region);
    fa_to_edges(&geo, h_list, v_list);
    fa_free_geo(&geo);
    return;
}


/*
 * faBoxCenter: Finds center of box
 */
void
faBoxCenter(box, x, y)
fa_box *box;
fa_coord *x, *y;
{
    *x = (box->right - box->left)/2 + box->left;
    *y = (box->top - box->bottom)/2 + box->bottom;
    return;
}


/*
 * faAndNotInPlace: Does logical AND of geo1 and NOT geo2 in place
 */
void
faAndNotInPlace(geo, notgeo)
fa_geometry *geo;
fa_geometry *notgeo;
{
    fa_geometry tmp;

    fa_copy(geo, &tmp);
    fa_andnot(&tmp, notgeo, geo);
    fa_free_geo(&tmp);
    return;
}


/*
 * faCountBoxes: Counts the number of boxes in the current generator
 */
int faCountBoxes(box)
fa_box *box;
{
    fa_box *c;
    int i;

    i = 0;
    for (c = box; c != ((fa_box *) 0); c = c->next) {
	i++;
    }

    return i;
}


/*
 * faCoordInGeo - returns 1 if x,y is in the given geo
 */

int faCoordInGeo(geo, x, y)
    fa_geometry *geo;
    int x,y;
{
    fa_region *regions, region;
    
    fa_to_regions(geo, &regions);
    while(faGenerateRegion(&regions, &region) == 1)
	if(faCoordInRegion(&region, x, y) == 1) {
	    fa_free_regions(regions);
	    return 1;
	}
    fa_free_regions(regions);
    return 0;
}

/*
 * faCoordInRegion: Checks if given coordinates are contained in the region
 *    returns: 1 -- if contained
 *             0 -- otherwise
 */
int faCoordInRegion(region, x, y)
fa_region *region;
int x, y;
{
    fa_box *box_gen, *boxes, box;
    fa_coord l, b, r, t;

    fa_regions_to_boxes(region, &boxes); 

    box_gen = boxes;
    while (faGenerateBox(&box_gen, &box) == 1) {
	l = box.left; r = box.right; t = box.top; b = box.bottom;
	if  ( (x >= l) && (x <= r) && (y >= b) && (y <= t) ) {
	    return 1;
	}
    }

    fa_free_boxes(boxes);

    return 0;
}


/*
 * faCoordInBox: Checks if given coordinates are contained in the box
 *    returns: 1 -- if contained
 *             0 -- otherwise
 */
int faCoordInBox(box, x, y)
fa_box *box;
int x, y;
{
    fa_coord l, b, r, t;

    l = box->left; r = box->right; t = box->top; b = box->bottom;
    if  ( (x >= l) && (x <= r) && (y >= b) && (y <= t) ) {
	return 1;
    }
    return 0;
}



faGrowShrink( geo, amount )
    fa_geometry *geo;
    int amount;
{
    if ( amount ) {
	fa_geometry tmp;

	fa_grow( geo, amount, &tmp );
	fa_grow( &tmp, -amount, geo );
	
	fa_free_geo( &tmp );
    }
    return 0;
}


int faGeoRegionIntersect( g, r )
    fa_geometry* g;
    fa_region* r;
{
    int ret = 0;
    fa_geometry geo, tmp;
    fa_region reg;

    reg = *r;			/* Copy the region. */
    reg.next = 0;		/* Isolate this region from others in the list */
    reg.interior_regions =0;	/*  */

    fa_init( &geo );
    fa_add_regions( &geo, &reg );
    fa_grow( &geo, 1, &tmp );
    ret = fa_intersect_p( g, &tmp );

    fa_free_geo( &geo );
    fa_free_geo( &tmp );

    return ret;
}
