/*
 * formals.c: Routines for handling explicit formal terminals
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"

/*
 * handleExplicitFormals:
 *    for each explicit formal terminal {
 *        add to terminal list;
 *        make master if necessary;
 *        instantiate and make terminal;
 *    }
 */
/*ARGSUSED*/
handleExplicitFormals(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
    octObject *inFacetPtr, *outFacetPtr;
    harList *geoListPtr;
    khList *tsListPtr;
{
    octObject connectorBag, instance, net, fterm, aterm, layer, box;
    octGenerator genTerm;
    char *termname, *layername;
    int x, y;
    lsList termList;
    kh_term *termEntry;
    int width, length;
    char master[BUFFER_SIZE], view[BUFFER_SIZE], name[BUFFER_SIZE];
    octTransformType type;
    octObject bag, prop;
    int minimum;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%shandling explicit formal terminals ...\n", INDENT);
    }

    /* get connector bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &connectorBag, "CONNECTORS"));

    OH_ASSERT(octInitGenContents(inFacetPtr, OCT_TERM_MASK, &genTerm));
    while (octGenerate(&genTerm, &fterm) == OCT_OK) {
	OH_ASSERT(gnPhysicalFormalTerminalBoxAndLayer(&fterm, &box, &layer));
	layername = layer.contents.layer.name;
	OH_ASSERT(gnBoxCenter(&box, &x, &y));
	OH_ASSERT(gnBoxWidthAndLength(&box, &width, &length));

	/* grab minimum width for error checking */
	OH_ASSERT(ohGetByBagName(&globalFacet, &bag, "NETS"));
	if ( ohGetByLayerName(&bag, &layer, layername) != OCT_OK ) {
	    fprintf( stderr, "%s has illegal implementation in %s\n", ohFormatName(&fterm), ohFormatName(&layer));
	    continue;
	}
	OH_ASSERT(ohGetByPropName(&layer, &prop, "MINIMUM_WIDTH"));
	minimum = prop.contents.prop.value.integer;

	/* cannot make a connector smaller than minimum net width */
	if ((width < minimum) || (length < minimum)) {
	    (void) fprintf(stderr, "Ignored: cannot make a connector ");
	    (void) fprintf(stderr, "smaller than minimum net width\n");
	} else {
	    /* 
	     * make master if necessary,
	     * then instantiate it
	     */
	    makeConnectorMaster(layername, width, length, &type, master, view);
	    mkInstanceName(name);
	    /* keep count */
	    connectorCount++;
	    OH_ASSERT(ohCreateInstance(outFacetPtr, &instance, name, master, view, x, y, type));
	    OH_ASSERT(octAttach(&connectorBag, &instance));

	    /*
	     * make formal terminal 
	     */
	    termname = fterm.contents.term.name;
	    OH_ASSERT(ohCreateTerm(outFacetPtr, &fterm, termname));

	    /*
	     * make a net.
	     * get actual terminal of instance.
	     * attach fterm to net, and aterm to fterm
	     * and net
	     */
	    mkNetName(name);
	    OH_ASSERT(octGenFirstContent(&instance, OCT_TERM_MASK, &aterm));
	    OH_ASSERT(ohCreateNet(outFacetPtr, &net, name));
	    OH_ASSERT(octAttachOnce(&net, &fterm));
	    OH_ASSERT(octAttach(&net, &aterm));
	    OH_ASSERT(octAttachOnce(&fterm, &aterm));

	    /* make terminal entry */
	    termList = khGetNamedLs(layername, tsListPtr);
	    termEntry = termConstruct( x, y, aterm.objectId);
	    (void) lsNewBegin(termList, (Pointer) termEntry, LS_NH);
	}
    }
}


