/*
 * instances.c: Routines for instantiating disassembled masters
 *    and promoting actual terminals
 */
#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"

/*
 * handleInstances:
 *    for each module instance {
 *        instantiate the disassembled master;
 *        for each actual terminal {
 *            get bounding box of terminal;
 *            add to fang geo list;
 *            add to terminal table;
 *        }
 *    }
 */
handleInstances(inFacetPtr, outFacetPtr, geoListPtr, tsListPtr)
octObject *inFacetPtr, *outFacetPtr;
harList *geoListPtr;
khList *tsListPtr;
{
    char *master, *view;
    char name[BUFFER_SIZE];
    octObject layer, instance, newInstance, term, box;
    octGenerator genInst, genTerm;
    struct octTransform tr;
    char *layername;
    fa_geometry *geoPtr;
    kh_term *termEntry;
    int x, y;
    octObject instanceBag;
    lsList termList;

    /* print trace message */
    if (trace) {
	(void) fprintf(stderr, "%spromoting instances ...\n", INDENT);
    }

    /* get instance bag */
    OH_ASSERT(ohGetByBagName(outFacetPtr, &instanceBag, "INSTANCES"));

    /* handle the instances */
    OH_ASSERT(octInitGenContents(inFacetPtr, OCT_INSTANCE_MASK, &genInst));
    while (octGenerate(&genInst, &instance) == OCT_OK) {
	/* instantiate the instances */
	tr.translation.x = instance.contents.instance.transform.translation.x;
	tr.translation.y = instance.contents.instance.transform.translation.y;
	tr.transformType = instance.contents.instance.transform.transformType;
	master = instance.contents.instance.master;
	view =  outFacetPtr->contents.facet.view;
	mkInstanceName(name);
	/* keep count */
	subcellCount ++;
	OH_ASSERT(ohCreateInstance(outFacetPtr, &newInstance, name, master, 
	  view, tr.translation.x, tr.translation.y, tr.transformType));
	OH_ASSERT(octAttach(&instanceBag, &newInstance));

	/* handle actual terminals */
	OH_ASSERT(octInitGenContents(&newInstance, OCT_TERM_MASK, &genTerm));
	while (octGenerate(&genTerm, &term) == OCT_OK) {
	    /* update fang geo */
	    OH_ASSERT(gnSymbolicActualTerminalBoxAndLayer(&term,&box,&layer));
	    layername = layer.contents.layer.name;
	    geoPtr = harGetNamedGeo(layername, geoListPtr);
	    harAddObjectToFang(&box, geoPtr, &tr);

	    /* update terminal table */
	    termList = khGetNamedLs(layername, tsListPtr);
	    OH_ASSERT(gnBoxCenter(&box, &x, &y));
	    termEntry = termConstruct( x,y,term.objectId );

	    (void) lsNewBegin(termList, (Pointer) termEntry, LS_NH);
	}
    }
}


