/*
 * map.h
 *
 * Generic Virtual Coordinate Mapping package
 *
 * Bill Lin
 * University of California, Berkeley
 *
 * This file contains public type definitions for the Virtual 
 * Coordinate Mapping package implemented in map.c.  This is 
 * stand alone package.
 */

#define	MAP_OK		0
#define	MAP_ERROR	1

typedef int mapStatus;			/* return code */

typedef struct mapTable {		/* defining a map table */
    int size;				/* the size of the table */
    int *table;				/* the actual table */
} mapTable;

extern mapTable *mapCreate();		/* creates a map table */
extern mapStatus mapSortTable();	/* sorts the map table */
extern mapStatus mapVirtualCoord();	/* real to virtual coord */
extern mapStatus mapRealCoord();	/* virtual to real coord */
extern int mapCompare1();		/* compare func for increasing order */
extern int mapCompare2();		/* compare func for decreasing order */
extern void mapDumpTable();		/* prints map table in readable form */
extern void mapFree();			/* frees a map table */

#define MAP_ASSERT(fct)\
	if ((fct) != MAP_OK) {\
	    (void) fprintf(stderr, "MAP Assertion failed: file %s, line %d\n",\
		__FILE__, __LINE__);\
	    exit(-1);\
	}

