#include "port.h"
#include "utility.h"
#include "mz.h"
#include "kahlua.h"
#include "wormhole.h"

/* flags */
extern mzList*  initProblem();



extern int heuristic;
extern int trace;

int markTerminalsAsReached( grid, sx,  sy, list  )
    int** grid;
    int sx,sy;
    mzList **list;
{
    int count = 1;
    kh_term *term;
    kh_term *equiv;

    grid[sy][sx] = MZ_TRM;
    mzAddEntry( sx, sy, list );

    term = WHget( sx, sy );
    if ( term && term->equiv_head ) {
	for ( equiv = term->equiv_head; equiv != 0; equiv = equiv->equiv_next ) {
	    int x = equiv->gridPosition.x;
	    int y = equiv->gridPosition.y;
	    if ( x != -1 && grid[y][x] != MZ_TRM ) {
		grid[y][x] = MZ_TRM;
		count++;
		mzAddEntry( x, y, list );
	    }
	}
    }
    if ( count > 1 ) {
	fprintf( stderr, "\t%d equivalent pins reached\n", count );
    }
    return count;
}



/*
 * maze
 *
 * results
 *    0 if unsuccessful, 1 if successful
 *
 * side effects
 *    grid will be routed
 */
maze(grid, xsize, ysize)
int **grid;
int xsize, ysize;
{
    mzList *tree_list, *expand_list, *new_list, *path_list, *ptr;
    int tcount, x, y, nx, ny;
    int sx, sy;
    int reached, found;
    int wave;
    int cont;
    int direction;
    int connected_terms = 0;

    /* initialize problem */
    traceMsgs("initialize problem ...");
    tree_list = initProblem(grid, xsize, ysize, &connected_terms, &tcount);

    fprintf(stderr, "%d out of %d terminals connected to begin with\n", connected_terms, tcount );

    if (trace) debugGrid(grid, xsize, ysize, "stderr");

    /* for each unconnected terminal */
    while ( connected_terms < tcount ) {

	/* clear the grid from previous loop */
	traceMsgs("  clearing grid ...");
	clearGrid(grid, xsize, ysize);

	/* initialize variables */
	wave = 0;
	expand_list = tree_list;
	reached = 0;

	/* 
	 * continually propagate a wavefront outwards from 
	 * the current tree until another terminal is reached.
	 */
	traceMsgs("  propagating wavefront ...");
	while (!reached) {
	    /* expand the wavefront */
	    traceMsgs("    expanding next wave ...");
	    wave++;
            new_list = NIL(mzList);
	    ptr = expand_list;
	    /* error check */
	    if (ptr == NIL(mzList)) {
		(void) fprintf(stderr, "Panic: cannot expand any further\n");
		(void) fprintf(stderr, "still %d out of %d terminals unconnected\n",
			       tcount-connected_terms, tcount);
		(void) fprintf(stderr, "\n");
		clearGrid(grid, xsize, ysize);
		/* 
		 * just print out the grid
		 * instead of aborting entirely
		 */
		debugGrid(grid, xsize, ysize, "stderr");
		return 0;
	    }

	    while ((!reached) && (ptr != NIL(mzList))) {
		traceMsgs("      next ripple ...");
		/* current position */
		x = ptr->position.x;
		y = ptr->position.y;

		/* look left */
		if (!reached) {
		    nx = x - 1;
		    ny = y;
		    reached = expandWave(grid, nx, ny, wave, &new_list);
		}

		/* look down */
		if (!reached) {
		    nx = x;
		    ny = y - 1;
		    reached = expandWave(grid, nx, ny, wave, &new_list);
		}

		/* look right */
		if (!reached) {
		    nx = x + 1;
		    ny = y;
		    reached = expandWave(grid, nx, ny, wave, &new_list);
		}

		/* look up */
		if (!reached) {
		    nx = x;
		    ny = y + 1;
		    reached = expandWave(grid, nx, ny, wave, &new_list);
		}

		/* next item on expand_list */
		ptr = ptr->next;
	    }

	    traceMsgs("    update for next wave ...");
	    /* update expand_list */
	    if (expand_list != tree_list) {
		mzFree(expand_list);
	    }
	    expand_list = new_list;

	    /* free memory if no longer needed */
	    if (reached) {
		mzFree(new_list);
	    }
	}


	/*
	 * now backtrack to find a solution.  update tree 
	 * list also.
	 */
	sx = nx; sy = ny;	/* Save coordinates of reached terminal */
	connected_terms += markTerminalsAsReached( grid, sx, sy , &tree_list);

	traceMsgs("  begin backtracking whole wave ...");
	found = 0;
	cont = 1;
	direction = pickDirection(grid, xsize, ysize, sx, sy, heuristic);
        path_list = NIL(mzList);
	while ( !found ) {
	    traceMsgs("    try backtracking a step ...");
	    if (!cont) {
		direction = newDirection(direction);
		cont = 1;
	    }

	    switch(direction) {
		case LEFT:  x = nx - 1; y = ny; break;
		case DOWN:  x = nx; y = ny - 1; break;
		case RIGHT: x = nx + 1; y = ny; break;
		case UP:    x = nx; y = ny + 1; break;
	    }

	    {
		int dd = 0;
		if ( dd ) debugGrid(grid, xsize, ysize, "stderr");

	    }
	    wave--;			  /* Try to backtrack a step. */
	    if ((grid[y][x] == MZ_TRM) || (grid[y][x] == MZ_PTH)) {
		found = 1;		  /* Exit loop when a complete path is found */
	    } else if (grid[y][x] == wave) {
		/*
		 * continue in this direction, add entry to path list
		 * used to store the current soln; update new x and y
		 * locations.
		 */
		mzAddEntry(x, y, &path_list);
		nx = x;
		ny = y;
	    } else {
	       wave++;		/* Mark false. */
	       cont = 0;	/* Try to go for a new direction. */
	    }
	}

	/* 
	 * now go back and mark the path and
	 * add to tree list
	 */

	traceMsgs("  marking the path ...");
	ptr = path_list;
	while (ptr != NIL(mzList)) {
	    x = ptr->position.x;
	    y = ptr->position.y;
	    grid[y][x] = MZ_PTH;
	    mzAddEntry(x, y, &tree_list);
	    ptr = ptr->next;
	}

	/* free path list */
	mzFree(path_list);

	if ( trace ) debugGrid(grid, xsize, ysize, "stderr");

    }


    /* final clean up */
    traceMsgs("cleaning up ...");
    clearGrid(grid, xsize, ysize);
    mzFree(tree_list);
    return 1;
}
		    
