#include "port.h"
#include "utility.h"
#include "oct.h"
#include "fang.h"
#include "list.h"
#include "oh.h"

#include "map.h"
#include "kh.h"
#include "mz.h"
#include "kahlua.h"
#include "wormhole.h"

/* flags */
int heuristic = 2;
int trace_router = 0;

router(region, termList, pathList, xTable, yTable)
fa_region *region;
lsList termList;
lsList pathList;
mapTable *xTable;
mapTable *yTable;
{
    int **grid;
    int ret_val;

    /* run router */
    inGrid(&grid, region, termList, xTable, yTable);
    ret_val = maze(grid, xTable->size+2, yTable->size+2);
    outGrid(grid, pathList, xTable, yTable);
    freeGrid(&grid, yTable->size+2);
    WHclean();			/* Clean worm holes. */

    return ret_val;
}


traceMsgs(str)
char *str;
{
    if (trace_router) {
	(void) fprintf(stderr, "trace_router: %s\n", str);
    }
}


freeGrid(gridPtr, size)
int ***gridPtr;
int size;
{
    int **grid;
    int *row;
    int i;

    grid = *gridPtr;
    for (i=0; i<size; i++) {
	row = grid[i];
	FREE(row);
    }

    FREE(grid);
}


inGrid(gridPtr, region, termList, xTable, yTable)
int ***gridPtr;
fa_region *region;
lsList termList;
mapTable *xTable, *yTable;
{
    fa_point *cnrList;
    lsGen termGen;
    kh_term *termEntry;
    int lx, ly;
    int vx, vy;
    int ox = 0;
    int oy = 0;
    int **grid, *row;
    int xsize, ysize;
    int first = 1;
    int i, j;

    /*
     * first initialize grid: blocks at the periphery, dots everywhere else.
     */
    xsize = xTable->size + 2;
    ysize = yTable->size + 2;
    grid = ALLOC(int *, ysize);
    for (i=0; i<ysize; i++) {
        row = ALLOC(int, xsize);	
	for (j=0; j<xsize; j++) {
	    if (i==0 || i==ysize-1 || j==0 || j==xsize-1) {
		row[j] = MZ_BLK;
	    } else {
		row[j] = MZ_DOT;
	    }
	}
	grid[i] = row;
    }

    /*
     * write region boundaries
     */
    cnrList = region->head;

    while (cnrList != NIL(fa_point)) {
	/* translate into virtual grid mode */
	MAP_ASSERT(mapVirtualCoord(xTable, &(cnrList->x), &vx));
	MAP_ASSERT(mapVirtualCoord(yTable, &(cnrList->y), &vy));

	vx++; vy++;
	if (first) {
	    first = 0;
	    lx = vx; ly = vy;
	} else {
	    addEdge(grid, ox, oy, vx, vy);
	}

	ox = vx; oy = vy;
	cnrList = cnrList->next;
    }
    addEdge(grid, ox, oy, lx, ly);

    /*
      * Are there holes in the region? (AC May 10 1991)
      */
    if ( region->interior_regions ) {
	fa_region  *hole = region->interior_regions;
	cnrList = hole->head;
	first = 1;
	while( cnrList ) {
	    MAP_ASSERT(mapVirtualCoord(xTable, &(cnrList->x), &vx));
	    MAP_ASSERT(mapVirtualCoord(yTable, &(cnrList->y), &vy));

	    vx++; vy++;
	    if (first) {
		first = 0;
		lx = vx; ly = vy;
	    } else {
		addEdge(grid, ox, oy, vx, vy);
	    }

	    ox = vx; oy = vy;
	    cnrList = cnrList->next;
	}
	addEdge(grid, ox, oy, lx, ly);
    }

    /*
     * write term list
     */
    termGen = lsStart(termList);
    WHinit();
    while (lsNext(termGen, (Pointer *) &termEntry, LS_NH) == LS_OK) {
	int useCenter = 1;
	if ( useCenter ) {
	    /* translate into virtual grid mode */
	    MAP_ASSERT(mapVirtualCoord(xTable, &(termEntry->position.x), &vx));
	    MAP_ASSERT(mapVirtualCoord(yTable, &(termEntry->position.y), &vy));
	    vx++; vy++;
	    if ( grid[vy][vx] == MZ_CON ) {
		fprintf( stderr, "Overlapping terminals at %d,%d", vx,vy );
		termEntry->overlap = WHget( vx, vy );
	    } else {
		grid[vy][vx] = MZ_CON;
		WHadd( vx,vy, termEntry );
	    }
	    termEntry->gridPosition.x = vx;
	    termEntry->gridPosition.y = vy;

	} else {
	    /* 
	     * This code does not work, because the bound box has nothing to do with the
	     * virtual grid 
	     */
	    octObject term;
	    octBox    bb;
	    int llx,lly,urx,ury;
	    OH_ASSERT(ohGetById( &term, termEntry->termId ));
	    OH_ASSERT(ohTerminalBB( &term, &bb ));

	    MAP_ASSERT(mapVirtualCoord(xTable, &(termEntry->position.x), &vx));
	    MAP_ASSERT(mapVirtualCoord(yTable, &(termEntry->position.y), &vy));

	    llx = urx = vx;	/* THese are the defaults */
	    lly = ury = vy;

	    /* Some of these may fail, in which case, we use the defaults */
	    mapVirtualCoord(xTable, &(bb.lowerLeft.x), &llx );
	    mapVirtualCoord(yTable, &(bb.lowerLeft.y), &lly );
	    mapVirtualCoord(xTable, &(bb.upperRight.x), &urx );
	    mapVirtualCoord(yTable, &(bb.upperRight.y), &ury );
	    
	    {
		int i,j;
		int pseudoArea = 0;
		if ( llx > urx ) { int x = llx;llx=urx;urx=x;}
		if ( lly > ury ) { int y = lly;lly=ury;ury=y;}
		for ( i = llx ; i <= urx; i++ ) {
		    for ( j=lly; j <= ury; j++ ) {
			grid[j+1][i+1] = MZ_CON;
			pseudoArea++;
		    }
		}
		fprintf( stderr, "%s has area %d\n", ohFormatName(&term),pseudoArea );
	    }
	}
    }

    /* free generator */
    lsFinish(termGen);

    /* return */
    *gridPtr = grid;
}


addEdge(grid, ox, oy, vx, vy)
int **grid, ox, oy, vx, vy;
{
    int x1, y1, x2, y2;
    int i, j;

    x1 = MIN(ox, vx);
    y1 = MIN(oy, vy);
    x2 = MAX(ox, vx);
    y2 = MAX(oy, vy);

    for (i=y1; i<=y2; i++) {
	for (j=x1; j<=x2; j++) {
	    grid[i][j] = MZ_BLK;
	}
    }
}


outGrid(grid, pathList, xTable, yTable)
int **grid;
lsList pathList;
mapTable *xTable, *yTable;
{
    int x, y;

    while ( findStartingPoint(&x, &y, grid, xTable->size+2, yTable->size+2) ) {
	tracePath(grid, x, y, 0, 0, pathList, xTable, yTable);
    }
}


tracePath(grid, x, y, inc_x, inc_y, pathList, xTable, yTable)
int **grid;
int x, y;
int inc_x, inc_y;
lsList pathList;
mapTable *xTable, *yTable;
{
    int nx, ny;

    nx = x; ny = y;

    if (inc_x != 0 || inc_y != 0) {
	grid[ny][nx] = MZ_STR;
	nx = nx + inc_x;
	ny = ny + inc_y;
	while (checkNode(grid, nx, ny, inc_x, inc_y) != 0) {
	    grid[ny][nx] = MZ_STR;
	    nx = nx + inc_x;
	    ny = ny + inc_y;
	}
	addPath(x-1, y-1, nx-1, ny-1, pathList, xTable, yTable);
    }

    /* right */
    if (grid[ny][nx+1] == MZ_PTH || grid[ny][nx+1] == MZ_TRM)	
	tracePath(grid, nx, ny, 1, 0, pathList, xTable, yTable);	

    /* down */
    if (grid[ny-1][nx] == MZ_PTH || grid[ny-1][nx] == MZ_TRM)	
	tracePath(grid, nx, ny, 0,-1, pathList, xTable, yTable);	

    /* left */
    if (grid[ny][nx-1] == MZ_PTH || grid[ny][nx-1] == MZ_TRM)
	tracePath(grid, nx, ny,-1, 0, pathList, xTable, yTable);	

    /* up */
    if (grid[ny+1][nx] == MZ_PTH || grid[ny+1][nx] == MZ_TRM)	
	tracePath(grid, nx, ny, 0, 1, pathList, xTable, yTable);	

    {
	kh_term* term = WHget( x, y );
	kh_term* t ;
	if ( term ) {
	    grid[ny][nx] = MZ_STR; /* Eliminate this terminal ( useless ) */
	    if ( term->equiv_head ) {
		term = term->equiv_head;
	    
		for ( t =  term->equiv_head; t != 0; t = t->equiv_next ) {
		    int xx = t->gridPosition.x;
		    int yy = t->gridPosition.y;
		    if ( xx != -1 &&  grid[yy][xx] != MZ_STR ) {
			tracePath(grid, xx, yy, 0, 0, pathList, xTable, yTable);
		    }
		}
	    }
	}
    }
}


checkNode(grid, x, y, inc_x, inc_y)
int **grid;
int x, y;
int inc_x, inc_y;
{
    int c=0;

    /* if reached a terminal */
    if (grid[y][x] == MZ_TRM) return 0;

    /* if can't continue further */
    if (grid[y+inc_y][x+inc_x] != MZ_PTH && grid[y+inc_y][x+inc_x] != MZ_TRM) 
      return 0;

    /* if node branches into more than one direction */
    if (grid[y][x+1] == MZ_PTH || grid[y][x+1] == MZ_TRM) c++;
    if (grid[y-1][x] == MZ_PTH || grid[y-1][x] == MZ_TRM) c++;
    if (grid[y][x-1] == MZ_PTH || grid[y][x-1] == MZ_TRM) c++;
    if (grid[y+1][x] == MZ_PTH || grid[y+1][x] == MZ_TRM) c++;
    if (c > 1) return 0;

    return 1;
}


findStartingPoint(px, py, grid, xsize, ysize)
    int *px, *py;
    int **grid;
    int xsize, ysize;
    /*
     * Find any terminal in the grid.
     */
{
    int i, j;

    for (i=0; i<ysize; i++) {
	for (j=0; j<xsize; j++) {
	    if (grid[i][j] == MZ_TRM) {
		*py = i;
		*px = j;
		return 1;
	    }
	}
    }
    return 0;			/* No starting point. */
}


addPath(x1, y1, x2, y2, pathList, xTable, yTable)
    int x1, y1, x2, y2;
    lsList pathList;
    mapTable *xTable, *yTable;
{
    int rx1, ry1, rx2, ry2;
    kh_path *pathEntry;

    /* translate coordinates */
    MAP_ASSERT(mapRealCoord(xTable, &x1, &rx1));
    MAP_ASSERT(mapRealCoord(yTable, &y1, &ry1));
    MAP_ASSERT(mapRealCoord(xTable, &x2, &rx2));
    MAP_ASSERT(mapRealCoord(yTable, &y2, &ry2));

    /* add to path list */
    pathEntry = ALLOC(kh_path, 1);
    pathEntry->start.x = rx1;
    pathEntry->start.y = ry1;
    pathEntry->end.x = rx2;
    pathEntry->end.y = ry2;
    pathEntry->width = 0;	/* currently undefined */
    (void) lsNewBegin(pathList, (Pointer) pathEntry, LS_NH);
}


debugGrid(grid, xsize, ysize, fname)
int **grid;
int xsize;
int ysize;
char *fname;
{
    int i, j;
    int *row;
    FILE *stream;

    /* open file */
    if (!strcmp(fname, "stdout")) {
	stream = stdout;
    } else if (!strcmp(fname, "stderr")) {
	stream = stderr;
    } else if ((stream = fopen(util_file_expand(fname), "w")) == NULL) {
	(void) fprintf(stderr, "panic: couldn't open output file for writing\n");
	exit(-1);
    }

    /* now dump the grid */
    (void) fprintf(stream, "%d %d\n", ysize, xsize);
    for (i=0; i<ysize; i++) {
	row = grid[i];
	for (j=0; j<xsize;j++) {
	    switch(row[j]) {
		case MZ_BLK:
		    (void) fprintf(stream, "%c", SMZ_BLK);
		    break;
		case MZ_DOT:
		    (void) fprintf(stream, "%c", SMZ_DOT);
		    break;
		case MZ_CON:
		    (void) fprintf(stream, "%c", SMZ_CON);
		    break;
		case MZ_TRM:
		    (void) fprintf(stream, "%c", SMZ_TRM);
		    break;
		case MZ_PTH:
		    (void) fprintf(stream, "%c", SMZ_PTH);
		    break;
		case MZ_STR:
		    (void) fprintf(stream, "%c", SMZ_STR);
		    break;
		default:
		    (void) fprintf(stream, "%d", row[j]);
		    /* (void) fprintf(stderr, "cannot recognize number %d\n", row[j]); */
		    break;
	    }
	}
	(void) fprintf(stream, "\n");
    }

    /* close file */
    if ((stream != stdout) && (stream != stderr)) {
	(void) fclose(stream);
    }
}
