#include "port.h"
#include "utility.h"
#include "mz.h"

#define EXPAND(x) (x!=MZ_BLK)&&(x!=MZ_CON)&&(x!=MZ_TRM)&&(x!=MZ_PTH)

clearGrid(grid, xsize, ysize)
int **grid;
int xsize, ysize;
{
    int x, y;

    for (x=0; x<xsize; x++) {
	for(y=0; y<ysize; y++) {
	    switch (grid[y][x]) {
		case MZ_BLK:
		case MZ_DOT:
		case MZ_CON:
		case MZ_TRM:
		case MZ_PTH:
		    break;
		default:
		    grid[y][x] = MZ_DOT;
		    break;
	    }
	}
    }
}


mzList *initProblem(grid, xsize, ysize, connected, tc)
int **grid;
int xsize, ysize;
int *connected;
int *tc;
{
    int x, y;
    int done = 0;		/* ! when the first terminal has been found. */
    mzList *list = 0;

    /* initialize */
    *tc = 0;

    for (x=0; x<xsize; x++) {
	for(y=0; y<ysize; y++) {
	    switch (grid[y][x]) {
	    case MZ_BLK:
	    case MZ_DOT:
		break;
	    case MZ_CON:	/* A contact, turn the first into a terminal. */
		*tc += 1;
		if (!done) {
		    done = 1;
		    *connected = markTerminalsAsReached( grid, x, y, &list );
		    *tc += *connected - 1;
		}
		break;
	    case MZ_PTH:
		fprintf( stderr, "Adding path to list\n" );
		mzAddEntry( x, y, &list );
		break;
	    case MZ_TRM:
		fprintf( stderr, "Skip term already reached\n" );
		break;
	    default:
		grid[y][x] = MZ_DOT;
		break;
	    }
	}
    }
    return list;
}


expandWave(grid, x, y, wave, list)
int **grid;
int x, y;
int wave;
mzList **list;
/* 
  * Returns 1 if a contact is reached.
  */
{
    switch (grid[y][x]) {
	case MZ_DOT:
	    grid[y][x] = wave;
	    mzAddEntry(x, y, list);
	    return 0;
	case MZ_CON:
	    return 1;
	default:
	    return 0;
    }
    /*NOTREACHED*/
}


pickDirection(grid, xsize, ysize, x, y, heuristic)
int **grid;
int xsize, ysize;
int x, y;
int heuristic;
{
    int cx, cy;
    int h, v;
    int ml, md, mr, mu;

    if (heuristic == 1) {
	cx = xsize / 2;
	cy = ysize / 2;
	h = abs(x - cx);
	v = abs(y - cy);
	if (h >= v) {
            if (x >= cx) {
		return LEFT;
	    } else {
		return RIGHT;
	    }
	} else {
            if (y >= cy) {
		return DOWN;
	    } else {
		return UP;
	    }

	}
    } else if (heuristic == 2) {
	ml=1; md=1; mr=1; mu=1;
	while (EXPAND(grid[y][x-ml])) ml++;	/* left */
	while (EXPAND(grid[y-md][x])) md++;	/* down */
	while (EXPAND(grid[y][x+mr])) mr++;	/* right */
	while (EXPAND(grid[y+mu][x])) mu++;	/* up */
	if ((ml >= md) && (ml >= mr) && (ml >= mu)) return LEFT;
	if ((md >= ml) && (md >= mr) && (md >= mu)) return DOWN;
	if ((mr >= md) && (mr >= ml) && (mr >= mu)) return RIGHT;
	if ((mu >= md) && (mu >= mr) && (mu >= ml)) return UP;
    } else {
	return LEFT;
    }
    /*NOTREACHED*/
}


newDirection(dir)
int dir;
{
    switch (dir) {
	case LEFT:
	    return DOWN;
	case DOWN:
	    return RIGHT;
	case RIGHT:
	    return UP;
	case UP:
	    return LEFT;
    }
    /*NOTREACHED*/
}

