/* Multidigit.c: Operate with multidigit transistors */


#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"
#include "da.h"

extern char *defLibDir;

#define PASS_ASSERT(exp)        { int status = (exp); \
                                  if (status < 0) return(status); }

char* multiDigitTermNames[] = {"DRAIN", "SOURCE", "O", 0};


MDCreateDiffRegions( diffGeo, dlayer, facet )
    fa_geometry *diffGeo;
    octObject *dlayer;
    octObject *facet;
{
    int   count_regions = 0;
    char  name[64];
    fa_region *regionList;
    fa_region *regionGen, region;

    fa_to_regions(diffGeo, &regionList);

    regionGen = regionList;
    while(faGenerateRegion(&regionGen, &region) == 1) {
	if ( count_regions == 0 ) strcpy( name, "DRAIN" ); 
	else if ( count_regions == 1 ) strcpy( name, "SOURCE" );
	else sprintf( name, "REG_%d", count_regions-2 );

	MDmyCreateTerm(region, &dlayer, &facet, name );
    }
}





MDmakeMyMosfetMaster(gateBoxList, diffGeo,
		   mosfetRegion, name, playername, dlayername,
		   length, width, master, view)
    fa_box gateBoxList[];
    fa_geometry *diffGeo;
    fa_region *mosfetRegion;
    char *name;
    char *playername, *dlayername;
    int length, width;
    char *master,*view;
{
    char *root, *path;
    static int number=0;
    octObject facet, ifacet, player, dlayer, prop, bag, box, fterm, jtBag, jBag;
    fa_geometry geo;
    fa_box *boxList, fangBox;
    int count;
        
    number++;
    (void) (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = util_strsav(defLibDir);

    (void) (void) sprintf(master, "%s/%s.%dx%d.multi%d", path, root,
		   width, length, number);

    OH_ASSERT(ohGetOrCreateBag(&masBag, &bag, master));
    if (trace) {
	(void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
		INDENT, INDENT, master, view);
    }

    /* keep count */
    uniqueMosfetCount++;
	
    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    /* create layers */
    OH_ASSERT(ohGetOrCreateLayer(&facet, &player, playername));
    OH_ASSERT(ohGetOrCreateLayer(&facet, &dlayer, dlayername));

    /* create active area */
    faRegionGeo(mosfetRegion, &geo);
    fa_regions_to_boxes(mosfetRegion, &boxList);
    while(faGenerateBox(&boxList, &fangBox) == 1) {
	OH_ASSERT(ohCreateBox(&player, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
	OH_ASSERT(ohCreateBox(&dlayer, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
    }
    
    /* create source */
    MDCreateDiffRegions( diffGeo, mosfetGeo, &dlayer, &facet );


    /* create gates */
    OH_ASSERT(ohCreateBox(&player, &box, gateBoxList[0].left,
			 gateBoxList[0].bottom, gateBoxList[0].right,
			 gateBoxList[0].top));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&fterm, &box));

    OH_ASSERT(ohCreateBox(&player, &box, gateBoxList[1].left,
			 gateBoxList[1].bottom, gateBoxList[1].right,
			 gateBoxList[1].top));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE2"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* create properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));

    /* additional interface information */
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "MOSFET"));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "MOSFETTYPE", name));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETLENGTH", (double) width));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETWIDTH", (double) length));

    /* make join terms for gate terminals */
    OH_ASSERT(ohCreateBag(&ifacet, &jtBag, "JOINEDTERMS"));
    OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&jBag, &fterm));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE2"));
    OH_ASSERT(octAttach(&jBag, &fterm));

    /* make join terms for source terminals */
    if(ohGetByTermName(&ifacet, &fterm, "SOURCE1") > 0) {
	OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
	name = "SOURCE";
	count = 0;
	while(ohGetByTermName(&ifacet, &fterm, name) > 0) {
	    OH_ASSERT(octAttach(&jBag, &fterm));
	    count++;
	    (void) sprintf(name, "SOURCE%d", count);
	}
    }
    
       
    /* make join terms for source terminals */
    if(ohGetByTermName(&ifacet, &fterm, "DRAIN1") > 0) {
	OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
	name = "DRAIN";
	count = 0;
	while(ohGetByTermName(&ifacet, &fterm, name) > 0) {
	    OH_ASSERT(octAttach(&jBag, &fterm));
	    count++;
	    (void) sprintf(name, "DRAIN%d", count);
	}
    }
    
    OH_ASSERT(octFlushFacet(&ifacet));
    OH_ASSERT(octFlushFacet(&facet));
}


MDfindTranRegions(mosfetRegion, gateBoxList, sourceRegion, drainRegion,
	       polyGeo, diffGeo, poverlap)
    fa_region *mosfetRegion;
    fa_box    **gateBoxList;
    fa_region *sourceRegion, *drainRegion;
    fa_geometry *polyGeo, *diffGeo;
    int poverlap;
{
    fa_geometry geo;
    fa_vertex *faVertex;
    dary faVertexList = daAlloc( fa_vertex, 10 );
    int vcount = 0;
    int gateTerms = 0;

    faRegionGeo(mosfetRegion, &geo);
    for(faVertex = geo.head; faVertex != NULL; faVertex = faVertex->next) {
	switch(faVertex->type) {
	case UPPER_RIGHT_NEG:
	case UPPER_LEFT_NEG:
	case LOWER_RIGHT_NEG:
	case LOWER_LEFT_NEG:
	    break;
	case LEFT_DIAGONAL:
	case RIGHT_DIAGONAL:
	    (void) fprintf(stderr,"Error in find gate stuff\n");
	case UPPER_RIGHT_POS:
	case LOWER_LEFT_POS:
	    /* - + , + - */
	    if(faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y-1) == 1) {
		*daSet( fa_vertex, faVertexList, vcount ) =  *faVertex;
		vcount++;
	    }
	    break;
	case UPPER_LEFT_POS:
	case LOWER_RIGHT_POS:
	    /* + + , - - */
	    if(faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y-1) == 1) {
		*daSet( fa_vertex, faVertexList, vcount ) =  *faVertex;
		vcount++;
	    }
	    break;
	}
   

    }

    gateTerms = vcount / 2;		/* there are two terms per gateTerms */
    if ( gateTerms > 2 ) {
	printf( "Multidigit transistor found %d\n", gateTerms );
    }
    
    MDmakeGateRegions( faVertexList, vcount, poverlap, polyGeo, &gateBoxList);
    MDmakeDrainAndSourceRegions(drainRegion,sourceRegion,faVertexList,diffGeo);
}


void MDmakeGateRegions( vertexList, vcount, overlap, polyGeo, gateBoxList )
    dary vertexList;
    int vcount;
    int overlap;
    fa_geometry *polyGeo;
    fa_box **gateBoxList;
{
    fa_vertex tmp;
    fa_box *boxList, *boxGen, box;
    fa_box barray[vcount];

    int gateTerms = vcount / 2;

    *gateBoxList = ALLOC( fa_box, gateTerms );
    
    /* Find the boxes closest to where the gate terms need to be added */
    fa_to_boxes(polyGeo, &boxList);
    rno = 0;
    while(rno < vcount) {
	boxGen = boxList;
	while(faGenerateBox(&boxGen, &box) == 1) {
	    if(faCoordInBox(&box, vertexList[rno].x, vertexList[rno].y) == 1) {
		barray[rno++] = box;
		break;
	    }
	}
    }
    
    for ( i = 0; i < vcount; i++ ) {
	int j = ( i + 1 ) % vcount;
	fa_vertex *cv = daGet( fa_vertex, vertexList, i ); /* Get current vertex */
	fa_vertex *nv = daGet( fa_vertex, vertexList, j ); /* Get next vertex  */

	if ( !faCoordInBox( &barray[i], nv->x, nv->y ) ) {
	    continue;	/* If vertices do not share the same box, skip */
	}
	if( cv->x == nv->x) {
	    gateBoxList[rno].left   = leftOverlap( cv, polyGeo, overlap);
	    gateBoxList[rno].right  = rightOverlap( cv, polyGeo, overlap);
	    gateBoxList[rno].bottom = MIN( cv->y, nv->y);
	    gateBoxList[rno].top    = MAX( cv->y, nv->y);
	} else if(vertexList[2*rno].y == vertexList[(2*rno)+1].y) {
	    gateBoxList[rno].bottom = bottomOverlap( cv, polyGeo, overlap);
	    gateBoxList[rno].top    = topOverlap( cv, polyGeo, overlap);
	    gateBoxList[rno].right  = MAX(cv->x, nv->x);
	    gateBoxList[rno].left   = MIN(cv->x, nv->x);
	} else {
	    errRaise( "kahlua", 1, "unaligned boxes" );
	}
    }
}

MDmakeDrainAndSourceRegions(drainRegion, sourceRegion, 
			  vertexList, diffGeo)
    fa_region *drainRegion, *sourceRegion;
    dary  vertexList;
    fa_geometry *diffGeo;
    
{
    fa_region *regionList, *regionGen, region;
    fa_to_regions(diffGeo, &regionList);

    dary diffRegions;

    diffRegions = daAlloc( fa_region );

    regionGen = regionList;

    while(faGenerateRegion(&regionGen, &region) == 1) {
	int i = 0;
	for ( i = 0; i < daLen( vertexList ) ; i += 2 ) {
	    int j = i / 2;
	if(faCoordInRegion(&region, vertexList[i].x,vertexList[i].y) == 1) {
	    diffRegion = daSet( 
	    diffRegion 
	    *drainRegion = region;
	}
	if(faCoordInRegion(&region, vertexList[1].x, vertexList[1].y) == 1) {
	    *sourceRegion = region;
	}
    }
}


MDcalcLengthAndWidth(boxList, width, length)
    fa_box *boxList;
    int *length, *width;
{
    int count;

    *length = 0;
    *width = 0;
    count = faCountBoxes(boxList);
    while(boxList != NULL) {
	*length += MAX(abs(boxList->right - boxList->left),
		       abs(boxList->top - boxList->bottom));
	*width += MIN(abs(boxList->right - boxList->left),
		      abs(boxList->top - boxList->bottom));
	boxList = boxList->next;
    }
    *width /= (double)count;
}


MDmyCreateTerm(region, layer, facet, name)
    fa_region *region;
    octObject *layer, *facet;
    char *name;
{
    fa_box *boxList;
    int boxCount = 0;
    octObject box,term;
    char myname[BUFSIZ];
    
    fa_regions_to_boxes(region, &boxList);
    while(boxList != NULL) {
	if(boxCount == 0) 
	    (void) strcpy(myname,name);
	else
	    (void) sprintf(myname, "%s%d", name, boxCount);
	OH_ASSERT(ohCreateBox(layer, &box, boxList->left, boxList->bottom,
			     boxList->right, boxList->top));
	/* term.contents.term.name = myname; */
	OH_ASSERT(ohCreateTerm(facet, &term, myname));
	OH_ASSERT(octAttach(&term, &box));
	boxList = boxList->next;
	boxCount++;
    }
}














