#include "port.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "fa.h"
#include "gn.h"
#include "list.h"
#include "kh.h"
#include "kahlua.h"

extern char *defLibDir;

#define PASS_ASSERT(exp)        { int status = (exp); \
                                  if (status < 0) return(status); }

makeMyMosfetMaster(gateBoxList, sourceRegion, drainRegion,
		   mosfetRegion, name, playername, dlayername,
		   length, width, master, view)
    fa_box gateBoxList[];
    fa_region *sourceRegion, *drainRegion, *mosfetRegion;
    char *name;
    char *playername, *dlayername;
    int length, width;
    char *master,*view;
{
    char *root, *path;
    static int number=0;
    octObject facet, ifacet, player, dlayer, prop, bag, box, fterm, jtBag, jBag;
    fa_geometry geo;
    fa_box *boxList, fangBox;
    int count;
        
    number++;
    (void) (void) sprintf(view, "%s", leafView);
    root = util_strsav(name);
    lower(root);
    path = defLibDir;

    (void) (void) sprintf(master, "%s/%s.%dx%d.bent%d", path, root,
		   width, length, number);

    OH_ASSERT(ohGetOrCreateBag(&masBag, &bag, master));
    if (trace) {
	(void) fprintf(stderr, "%s%screating leaf cell [%s:%s] ...\n", 
		INDENT, INDENT, master, view);
    }

    /* keep count */
    uniqueMosfetCount++;
	
    /* open master */
    OH_ASSERT(ohOpenFacet(&facet, master, view, "contents", "w"));

    /* create layers */
    OH_ASSERT(ohGetOrCreateLayer(&facet, &player, playername));
    OH_ASSERT(ohGetOrCreateLayer(&facet, &dlayer, dlayername));

    /* create active area */
    faRegionGeo(mosfetRegion, &geo);
    fa_regions_to_boxes(mosfetRegion, &boxList);
    while(faGenerateBox(&boxList, &fangBox) == 1) {
	OH_ASSERT(ohCreateBox(&player, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
	OH_ASSERT(ohCreateBox(&dlayer, &box, fangBox.left, fangBox.bottom,
			     fangBox.right, fangBox.top));
    }
    
    /* create source */
    myCreateTerm(drainRegion, &dlayer, &facet, "SOURCE");
    
    /* create drain */
    myCreateTerm(sourceRegion, &dlayer, &facet, "DRAIN");

    /* create gates */
    OH_ASSERT(ohCreateBox(&player, &box, gateBoxList[0].left,
			 gateBoxList[0].bottom, gateBoxList[0].right,
			 gateBoxList[0].top));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&fterm, &box));

    OH_ASSERT(ohCreateBox(&player, &box, gateBoxList[1].left,
			 gateBoxList[1].bottom, gateBoxList[1].right,
			 gateBoxList[1].top));
    OH_ASSERT(ohCreateTerm(&facet, &fterm, "GATE1"));
    OH_ASSERT(octAttach(&fterm, &box));

    /* create properties */
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "TECHNOLOGY", technology));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "VIEWTYPE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(&facet, &prop, "CELLCLASS", "LEAF"));

    /* close facet and generate interface */
    OH_ASSERT(gnGenInterfaceFacet(&facet,&ifacet));

    /* additional interface information */
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "CELLTYPE", "MOSFET"));
    OH_ASSERT(ohCreatePropStr(&ifacet, &prop, "MOSFETTYPE", name));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETLENGTH", (double) width));
    OH_ASSERT(ohCreatePropReal(&ifacet, &prop, "MOSFETWIDTH", (double) length));

    /* make join terms for gate terminals */
    OH_ASSERT(ohCreateBag(&ifacet, &jtBag, "JOINEDTERMS"));
    OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE"));
    OH_ASSERT(octAttach(&jBag, &fterm));
    OH_ASSERT(ohGetByTermName(&ifacet, &fterm, "GATE1"));
    OH_ASSERT(octAttach(&jBag, &fterm));

    /* make join terms for source terminals */
    if(ohGetByTermName(&ifacet, &fterm, "SOURCE1") > 0) {
	OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
	name = "SOURCE";
	count = 0;
	while(ohGetByTermName(&ifacet, &fterm, name) > 0) {
	    OH_ASSERT(octAttach(&jBag, &fterm));
	    count++;
	    (void) sprintf(name, "SOURCE%d", count);
	}
    }
    
       
    /* make join terms for source terminals */
    if(ohGetByTermName(&ifacet, &fterm, "DRAIN1") > 0) {
	OH_ASSERT(ohCreateBag(&jtBag, &jBag, "JOINED"));
	name = "DRAIN";
	count = 0;
	while(ohGetByTermName(&ifacet, &fterm, name) > 0) {
	    OH_ASSERT(octAttach(&jBag, &fterm));
	    count++;
	    (void) sprintf(name, "DRAIN%d", count);
	}
    }
    
    OH_ASSERT(octFlushFacet(&ifacet));
    OH_ASSERT(octFlushFacet(&facet));
}


findTranRegions(mosfetRegion, gateBoxList, sourceRegion, drainRegion,
	       polyGeo, diffGeo, poverlap)
    fa_region *mosfetRegion;
    fa_box gateBoxList[];
    fa_region *sourceRegion, *drainRegion;
    fa_geometry *polyGeo, *diffGeo;
    int poverlap;
{
    fa_geometry geo;
    fa_vertex *faVertex, faVertexList[4];
    int vcount;
    
    faRegionGeo(mosfetRegion, &geo);
    for(faVertex = geo.head,vcount=0; faVertex != NULL && vcount < 4; faVertex = faVertex->next) {
	switch(faVertex->type) {
	case UPPER_RIGHT_NEG:
	case UPPER_LEFT_NEG:
	case LOWER_RIGHT_NEG:
	case LOWER_LEFT_NEG:
	    break;
	case LEFT_DIAGONAL:
	case RIGHT_DIAGONAL:
	    (void) fprintf(stderr,"Error in find gate stuff\n");
	case UPPER_RIGHT_POS:
	case LOWER_LEFT_POS:
	    /* - + , + - */
	    if(faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y-1) == 1)
		faVertexList[vcount++] = *faVertex;
	    break;
	case UPPER_LEFT_POS:
	case LOWER_RIGHT_POS:
	    /* + + , - - */
	    if(faCoordInGeo(polyGeo, faVertex->x+1, faVertex->y+1) == 1 ||
	       faCoordInGeo(polyGeo, faVertex->x-1, faVertex->y-1) == 1)
		faVertexList[vcount++] = *faVertex;
	    break;
	}
    }

    makeGateRegions(gateBoxList, faVertexList, poverlap, polyGeo);
    makeDrainAndSourceRegions(drainRegion,sourceRegion,faVertexList,diffGeo);
}


makeGateRegions(gateBoxList, vertexList, overlap, polyGeo)
    fa_box gateBoxList[];
    fa_vertex *vertexList;
    int overlap;
    fa_geometry *polyGeo;
{
    fa_vertex tmp;
    fa_box *boxList, *boxGen, barray[4], box;
    int rno;

    fa_to_boxes(polyGeo, &boxList);
    rno = 0;
    while(rno < 4) {
	boxGen = boxList;
	while(faGenerateBox(&boxGen, &box) == 1) {
	    if(faCoordInBox(&box, vertexList[rno].x, vertexList[rno].y) == 1) {
		barray[rno++] = box;
		break;
	    }
	}
    }
    
    for(rno=2;rno<4;rno++)
	if(faCoordInBox(&barray[0],vertexList[rno].x, vertexList[rno].y) == 1) {
	    tmp = vertexList[1];
	    vertexList[1] = vertexList[rno];
	    vertexList[rno] = tmp;
	    break;
	}
    
    for(rno=0;rno<2;rno++) {
	if(vertexList[2*rno].x == vertexList[(2*rno)+1].x) {
	    gateBoxList[rno].left   = leftOverlap(&vertexList[2*rno], polyGeo, overlap);
	    gateBoxList[rno].right  = rightOverlap(&vertexList[2*rno], polyGeo, overlap);
	    gateBoxList[rno].bottom = MIN(vertexList[2*rno].y,vertexList[(2*rno)+1].y);
	    gateBoxList[rno].top    = MAX(vertexList[2*rno].y,vertexList[(2*rno)+1].y);
	} else {
	    if(vertexList[2*rno].y == vertexList[(2*rno)+1].y) {
		gateBoxList[rno].bottom = bottomOverlap(&vertexList[2*rno], polyGeo, overlap);
		gateBoxList[rno].top = topOverlap(&vertexList[2*rno], polyGeo, overlap);
		gateBoxList[rno].right = MAX(vertexList[2*rno].x,vertexList[(2*rno)+1].x);
		gateBoxList[rno].left = MIN(vertexList[2*rno].x,vertexList[(2*rno)+1].x);
	    }
	}
    }
}

makeDrainAndSourceRegions(drainRegion, sourceRegion, 
			  vertexList, diffGeo)
    fa_region *drainRegion, *sourceRegion;
    fa_vertex vertexList[];
    fa_geometry *diffGeo;
{
    fa_region *regionList, *regionGen, region;
    
    fa_to_regions(diffGeo, &regionList);
    regionGen = regionList;
    while(faGenerateRegion(&regionGen, &region) == 1) {
	if(faCoordInRegion(&region, vertexList[0].x,vertexList[0].y) == 1)
	    *drainRegion = region;
	if(faCoordInRegion(&region, vertexList[1].x, vertexList[1].y) == 1)
	    *sourceRegion = region;
    }
}


topOverlap(vertex, geo, overlap)
    fa_vertex *vertex;
    fa_geometry *geo;
    int overlap;
{
    if(faCoordInGeo(geo, vertex->x, vertex->y + overlap) == 1)
	return(vertex->y + overlap);
    return(vertex->y);
}

bottomOverlap(vertex, geo, overlap)
    fa_vertex *vertex;
    fa_geometry *geo;
    int overlap;
{
    if(faCoordInGeo(geo, vertex->x, vertex->y - overlap) == 1)
	return(vertex->y - overlap);
    return(vertex->y);
}

leftOverlap(vertex, geo, overlap)
    fa_vertex *vertex;
    fa_geometry *geo;
    int overlap;
{
    if(faCoordInGeo(geo, vertex->x - overlap, vertex->y) == 1)
	return(vertex->x - overlap);
    return(vertex->x);
}

rightOverlap(vertex, geo, overlap)
    fa_vertex *vertex;
    fa_geometry *geo;
    int overlap;
{
    if(faCoordInGeo(geo, vertex->x + overlap, vertex->y) == 1)
	return(vertex->x + overlap);
    return(vertex->x);
}


calcLengthAndWidth(boxList, width, length)
    fa_box *boxList;
    int *length, *width;
{
    int count;

    *length = 0;
    *width = 0;
    count = faCountBoxes(boxList);
    while(boxList != NULL) {
	*length += MAX(abs(boxList->right - boxList->left),
		       abs(boxList->top - boxList->bottom));
	*width += MIN(abs(boxList->right - boxList->left),
		      abs(boxList->top - boxList->bottom));
	boxList = boxList->next;
    }
    *width /= (double)count;
}


myCreateTerm(region, layer, facet, name)
    fa_region *region;
    octObject *layer, *facet;
    char *name;
{
    fa_box *boxList;
    int boxCount = 0;
    octObject box,term;
    char myname[BUFSIZ];
    
    fa_regions_to_boxes(region, &boxList);
    while(boxList != NULL) {
	if(boxCount == 0) 
	    (void) strcpy(myname,name);
	else
	    (void) sprintf(myname, "%s%d", name, boxCount);
	OH_ASSERT(ohCreateBox(layer, &box, boxList->left, boxList->bottom,
			     boxList->right, boxList->top));
	/* term.contents.term.name = myname; */
	OH_ASSERT(ohCreateTerm(facet, &term, myname));
	OH_ASSERT(octAttach(&term, &box));
	boxList = boxList->next;
	boxCount++;
    }
}


#ifndef lint
print_all_nets(facet)
    octObject *facet;
{
    octGenerator net_gen, term_gen;
    octObject net, term;
    
    OH_ASSERT(octInitGenContents(facet, OCT_NET_MASK, &net_gen));
    
    (void) printf("printing nets:\n");
    while( octGenerate(&net_gen, &net) == OCT_OK ) {
	(void) printf("\t%s\n", ohFormatName(&net));

	OH_ASSERT(octInitGenContents(&net, OCT_TERM_MASK, &term_gen));
	
	while( octGenerate(&term_gen, &term) == OCT_OK) {
	    (void) printf("\t\t%s\n", ohFormatName(&term));
	}
    }
    (void) octFreeGenerator(&net_gen);
    (void) octFreeGenerator(&term_gen);
    (void) printf("\n");
}
#endif
