/*
 *  MAKESOFT
 */
#include "defs.h"
#include "options.h"
#include "vov.h"

octObject	FloorplanBag;
octObject	MinHeightOverWidth;
octObject	MaxHeightOverWidth;
octObject	FloatingEdge;
octObject	TermEdge;
octObject	TermRelPosition;
octObject	TermType;
octObject	NetType;
octObject	Criticality;
octObject	Technology;

char *edges = NULL;


int		verbose = FALSE;

octObject 	Chip;
octObject 	ChipInterface;

char 		buffer[256];
#define M(s)	(void) fprintf(stderr, s)

optionStruct optionList[] = {
    { "D",	0,		"delete all propertites of the soft cell" },
    { "F",	"LRTB",		"specify edges for floating terminals" },
    { "w",	"width",	"width of the cell (in lambda)" },
    { "h",	"height",	"height of the cell (in lambda)" },
    { "v",	0,		"verbose mode" },
    { "o",	"cell:view",	"output cell and view" },
    { OPT_RARG,	"cell:view",	"input cell and view" },
    { 0,	0,		0 }
};


extern void FatalError();
extern void Warning();

main(argc,argv)
    int argc;
    char *argv[];
{
    int option;
    extern int	optind;
    extern char	*optarg;
    char*	output_cell = 0;
    char*       output_view = 0;
    char  	*input_cell, *input_view;
    int		width = 0;
    int		height = 0;
    int		deleteAll = 0;
    octGenerator  genTerm,genBox;
    octObject	Term;
    octObject	Box;
    octObject	LayerPlace;
    char	edgesparsed[256];
    char	*edgename;
    octStatus	status;
    int		i;
    int 	dash_o = 0;	/* Is option -o used ? */

    extern char	*version();

    /* parse command line options*/
    output_cell = NIL(char);
    output_view = NIL(char);

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	case 'D':
	    deleteAll = 1;
	    break;

	case 'F':
	    edges = strsav(optarg);
	    break;
		
	case 'w':
	    width = atoi(optarg);
	    break;

	case 'h':
	    height = atoi(optarg);
	    break;

	case 'v':
	    verbose = 1;
	    break;

	case 'o':
	    dash_o = 1;
	    output_cell = strsav(optarg);
	    output_view = strchr(output_cell, ':');
	    if (output_view != NIL(char)) {
		*output_view++ = '\0';
	    }
	}
    }

    if (argc == optind+1) {
	input_cell = strsav(argv[optind]);
	input_view = strchr(input_cell, ':');
	if (input_view != NIL(char)) {
	    *input_view++ = '\0';
	} else {
	    input_view = "symbolic";
	}
    } else {
	optUsage();
    }

    VOVbegin( argc, argv );
    (void) fprintf(stderr, "makeSoft: %s\n", version());

    initPlacementObjects();

    if (output_cell != NIL(char)) {
        if (output_view == NIL(char)) output_view = "alba";
        (void) copyFacets(input_cell , input_view , output_cell , output_view);
    } else {
        output_view = input_view;
        output_cell = input_cell;
    }

    Chip.type = OCT_FACET;
    Chip.contents.facet.cell = output_cell;
    Chip.contents.facet.view = output_view;
    Chip.contents.facet.facet = "contents";
    Chip.contents.facet.version = OCT_CURRENT_VERSION;
    Chip.contents.facet.mode = "a";

    if ( ! dash_o ) {
	VOVinputFacet( &Chip );
	VOVoutputFacet( &Chip );
    }
    if (octOpenFacet(&Chip) < OCT_OK) FatalError("opening facet");

    ChipInterface.type = OCT_FACET;
    ChipInterface.contents.facet = Chip.contents.facet;
    ChipInterface.contents.facet.facet = "interface";
    ChipInterface.contents.facet.mode = "a";

    if ( ! dash_o ) {
	VOVinputFacet( &ChipInterface );
	VOVoutputFacet( &ChipInterface );
    }
    if (octOpenFacet(&ChipInterface) < OCT_OK) FatalError("Opening interface");

    edgesparsed[0] = '\0';
    if (edges!=NULL) {
	for ( i = 0 ; i < strlen(edges) ; i++) {
	    switch (*(edges+i)) {
	    case 'L':
		edgename = "LEFT ";
		break;
	    case 'R':
		edgename = "RIGHT ";
		break;
	    case 'T':
		edgename = "TOP ";
		break;
	    case 'B':
		edgename = "BOTTOM ";
		break;
	    default:
		(void) sprintf(buffer,"Unparsable letter %c for -F option", *(edges+i));
		Warning(buffer);
		edgename = " ";
		break;
	    }
	    (void) strcat(edgesparsed, edgename);
	}
    }
    if (verbose) (void) printf("Value of TERM_FLOATING_EDGES: \"%s\"\n", edgesparsed);
  
    OCT_ASSERT(octInitGenContents(&ChipInterface,OCT_TERM_MASK,&genTerm));
    while ((status = octGenerate(&genTerm,&Term))==OCT_OK) {
	OCT_ASSERT(octGetOrCreate(&Term,&FloatingEdge));
	if (deleteAll) {
	    OCT_ASSERT(octDelete(&FloatingEdge));
	} else {
	    if (verbose) (void) printf("Term %s\n", Term.contents.term.name);
	    FloatingEdge.contents.prop.value.string = edgesparsed;
	    OCT_ASSERT(octModify(&FloatingEdge));
	}
    }

    if (status != OCT_GEN_DONE) FatalError("Terminal generator failed");

    LayerPlace.type = OCT_LAYER;
    LayerPlace.contents.layer.name = "PLACE";

    OCT_ASSERT(octGetOrCreate(&ChipInterface,&LayerPlace));
  

    OCT_ASSERT(octGetOrCreate(&ChipInterface,&FloorplanBag));
    OCT_ASSERT(octInitGenContents(&FloorplanBag,OCT_BOX_MASK,&genBox));
    while (octGenerate(&genBox,&Box)==OCT_OK) 
      OCT_ASSERT(octDelete(&Box));

  
  
    if (deleteAll) {
	OCT_ASSERT(octDelete(&LayerPlace));
    } else {
	Box.type = OCT_BOX;
	if (width * height > 0) {
	    Box.contents.box.lowerLeft.x = 0;
	    Box.contents.box.lowerLeft.y = 0;
	    Box.contents.box.upperRight.x = width * OCT_LAMBDA;
	    Box.contents.box.upperRight.y = height * OCT_LAMBDA;
	} else {
	    /* Warning("You must specify width and height; Using BB"); */
	    OCT_ASSERT(octBB(&Chip,&Box.contents.box));
	}
	OCT_ASSERT(octCreate(&FloorplanBag,&Box));
	OCT_ASSERT(octAttach(&LayerPlace,&Box));
    }
    OCT_ASSERT(octFlushFacet(&Chip));
    OCT_ASSERT(octFlushFacet(&ChipInterface));

    VOVend( 0 ) ;
}



copyFacets ( cellName , oldView , newCell,  newView )
     char *cellName, *oldView , *newCell, *newView;
{
    octObject  		Input, Output;

    if (verbose) {
	(void)printf("Copying %s:%s into %s:%s\n",
		     cellName,oldView,cellName,newView);
    }
    Input.type = OCT_FACET;
    Input.contents.facet.cell = cellName;
    Input.contents.facet.view = oldView;
    Input.contents.facet.facet = "contents";
    Input.contents.facet.version = OCT_CURRENT_VERSION;

    Output.type = OCT_FACET;
    Output.contents.facet.cell = newCell;
    Output.contents.facet.view = newView;
    Output.contents.facet.facet = "contents";
    Output.contents.facet.version = OCT_CURRENT_VERSION;

    VOVinputFacet( &Input );
    VOVoutputFacet( &Output );
    if (octCopyFacet(&Output,&Input) != OCT_OK)  FatalError("Copying contents Facet");

    Input.contents.facet.facet = "interface";
    Output.contents.facet.facet = "interface";
    VOVinputFacet( &Input );
    VOVoutputFacet( &Output );
    if (octCopyFacet(&Output,&Input) != OCT_OK)  Warning("Copying interface Facet");

}




/*
 *   initPlacementObjects()
 *   Initializes some useful oct objects
 */
initPlacementObjects()
{

  TermType.type = OCT_PROP;
  TermType.contents.prop.name = "TERMTYPE";
  TermType.contents.prop.type = OCT_STRING;


  NetType.type = OCT_PROP;
  NetType.contents.prop.name = "NETTYPE";
  NetType.contents.prop.type = OCT_STRING;

  Criticality.type = OCT_PROP;
  Criticality.contents.prop.name = "CRITICALITY";
  Criticality.contents.prop.type = OCT_INTEGER;


  FloorplanBag.type = OCT_BAG;
  FloorplanBag.contents.bag.name = "FLOORPLAN";
    
  MinHeightOverWidth.type = OCT_PROP;
  MinHeightOverWidth.contents.prop.name = "MIN_Y_OVER_X";
  MinHeightOverWidth.contents.prop.type = OCT_REAL;

  MaxHeightOverWidth.type = OCT_PROP;
  MaxHeightOverWidth.contents.prop.name = "MAX_Y_OVER_X";
  MaxHeightOverWidth.contents.prop.type = OCT_REAL;

  FloatingEdge.type = OCT_PROP;
  FloatingEdge.contents.prop.name = "TERM_FLOATING_EDGES";
  FloatingEdge.contents.prop.type = OCT_STRING;

  TermEdge.type = OCT_PROP;
  TermEdge.contents.prop.name = "TERM_EDGE";
  TermEdge.contents.prop.type = OCT_STRING;

  TermRelPosition.type = OCT_PROP;
  TermRelPosition.contents.prop.name = "TERM_RELATIVE_POSITION";
  TermRelPosition.contents.prop.type = OCT_REAL;

  Technology.type = OCT_PROP;
  Technology.contents.prop.name = "TECHNOLOGY";
  Technology.contents.prop.type = OCT_STRING;

}

