 
/*
#ifndef lint
static char rcsid[] = "$Header: /octtools/release-3.0/src/mercury/FPgraph.c,v 1.1 90/02/07 17:02:23 octtools Exp $";
#endif
*/

/*
 *  FPgraph.c
 *  *********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines concerning floor plan graph
 *
 *  functions contained in this file:
 *	getFPgraph()
 *	getFPgraphDFS()
 *	addPin()
 *	removePin()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"




/*
 *  getFPgraph()
 *
 *  DESCRIPTION
 *  -----------
 *	Search the tile plane and Get a floor plan graph from it
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	GNodeP	<input>	: pointer to the first node in a tile plane
 *	adjP[]	<output>: pointers to adjacency list of a floor plan graph
 *
 *  Local variables:
 *  ---------------
 *	saveNode	: saves pointer to the first node in adjacency list
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Fill in GID2Reg[] with appropriate values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GID2Reg[i]	: pointer to a region with id i+1
 *
 */


getFPgraph(GRegP,adjP)


GRegionDef	*GRegP;
adjNodeType	*adjP[];

{


adjNodeType	*saveNode;
GRegionDef	*nextRegion;
int		i;
GRegionDef	*baseReg;
GRegionDef	*nextAdj;


for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    GID2Reg[nextRegion->id - 1] = nextRegion;

    if (nextRegion->type == G_INST) {
	++(statis.num_block);
    } else if (nextRegion->type == G_DOM_XSEC) {
	++(statis.num_GRgraph_node);
    } else if (nextRegion->type == G_HOR_BN) {
	++(statis.num_hor_bot);
    } else if (nextRegion->type == G_VER_BN) {
	++(statis.num_ver_bot);
    } else if (nextRegion->type == G_OVERCELL) {
	++(statis.num_overcell);
    }

    if (nextRegion->type != G_DOM_XSEC) {
	continue;
    }

    /* Get edge information if we are where two dominant tiles cross */
    /*---------------------------------------------------------------*/
    for (i = 0; i <= G_NDIR-1; ++i) {

	for (baseReg = nextRegion; (nextAdj = getNeighbor(baseReg,i)) != NULL;
		baseReg = NULL) {

	    if ((nextAdj->type == G_HOR_BN) || (nextAdj->type == G_VER_BN)) {
		break;
	    }
	}
	if (nextAdj == NULL) {
	    continue;
	}

	/* Add adjacent node info at the head of adjacency list */

	++(statis.num_GRgraph_edge);

	saveNode = adjP[nextRegion->id - 1];
	adjP[nextRegion->id - 1] = MALLOC(adjNodeType,1);
	adjP[nextRegion->id - 1]->id = nextAdj->adjReg[i]->id;

	switch (i) {
	case G_UP :
	    adjP[nextRegion->id - 1]->weight
		= nextAdj->adjReg[i]->centerLocY - nextRegion->centerLocY;
	    break;
	case G_LEFT :
	    adjP[nextRegion->id - 1]->weight
		= nextRegion->centerLocX - nextAdj->adjReg[i]->centerLocX;
	    break;
	case G_DOWN :
	    adjP[nextRegion->id - 1]->weight
		= nextRegion->centerLocY - nextAdj->adjReg[i]->centerLocY;
	    break;
	case G_RIGHT :
	    adjP[nextRegion->id - 1]->weight
		= nextAdj->adjReg[i]->centerLocX - nextRegion->centerLocX;
	    break;
	default:
	    break;
	}

	adjP[nextRegion->id - 1]->next = saveNode;
    }
}

/* We have counted the same edge twice from both endpoints */
statis.num_GRgraph_edge /= 2;


}






/*
 *  getFPgraphDFS()
 *
 *  DESCRIPTION
 *  -----------
 *	Depth first search the tile plane and Get a floor plan graph from it
 *	This is a DFS version of getFPgraph() and is OBSOLETE
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	GNodeP	<input>	: pointer to the first node in a tile plane
 *	adjP[]		: pointers to adjacency list of a floor plan graph
 *
 *  Local variables:
 *  ---------------
 *	saveNode	: saves pointer to the first node in adjacency list
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


getFPgraphDFS(GRegP,adjP)


GRegionDef	*GRegP;
adjNodeType	*adjP[];

{


adjNodeType	*saveNode;
GRegionDef	*nextRegion;
GRegionDef	*baseReg;
GRegionDef	*nextAdj;
int		i;


nextRegion = GRegP;

DFSflag[nextRegion->id - 1] = ON;
GID2Reg[nextRegion->id - 1] = nextRegion;

/* Get edge information if we are where two dominant tiles cross */
/*---------------------------------------------------------------*/
if (nextRegion->type == G_DOM_XSEC) {
    for (i = 0; i <= G_NDIR-1; ++i) {

	for (baseReg = nextRegion; (nextAdj = getNeighbor(baseReg,i)) != NULL;
		baseReg = NULL) {

	    if ((nextAdj->type == G_HOR_BN) || (nextAdj->type == G_VER_BN)) {
		break;
	    }
	}
	if (nextAdj == NULL) {
	    continue;
	}

	/* Add adjacent node info at the head of adjacency list */

	saveNode = adjP[nextRegion->id - 1];
	adjP[nextRegion->id - 1] = MALLOC(adjNodeType,1);
	adjP[nextRegion->id - 1]->id = nextAdj->adjReg[i]->id;

	switch (i) {
	case G_UP :
	    adjP[nextRegion->id - 1]->weight
		= nextAdj->adjReg[i]->centerLocY - nextRegion->centerLocY;
	    break;
	case G_LEFT :
	    adjP[nextRegion->id - 1]->weight
		= nextRegion->centerLocX - nextAdj->adjReg[i]->centerLocX;
	    break;
	case G_DOWN :
	    adjP[nextRegion->id - 1]->weight
		= nextRegion->centerLocY - nextAdj->adjReg[i]->centerLocY;
	    break;
	case G_RIGHT :
	    adjP[nextRegion->id - 1]->weight
		= nextAdj->adjReg[i]->centerLocX - nextRegion->centerLocX;
	    break;
	default:
	    break;
	}

	adjP[nextRegion->id - 1]->next = saveNode;
    }
}

for (i = 0; i <= G_NDIR-1; ++i) {
    if (nextRegion->adjReg[i] == NULL) {
        continue;
    }

    /* Skip if the neighbor has already been threaded */
    /*------------------------------------------------*/
    if (DFSflag[nextRegion->adjReg[i]->id - 1] == ON) {
        continue;
    }

    getFPgraphDFS(nextRegion->adjReg[i],adjP);
}


}





/*
 *  addPin()
 *
 *  DESCRIPTION
 *  -----------
 *	Add to the floor plan graph those nodes corresponding to the pins
 *		of the given net group
 *
 *  RETURNS
 *  -------
 *
 *  Arguments:
 *  ---------
 *	netGrP	<input>	: pointer to a net group whose pins should be added
 *
 *  Local variables:
 *  ---------------
 *	node1		: pointer to one node adjacent to G_DOM_XSEC region
 *	node2		: pointer to other node adjacent to G_DOM_XSEC region
 *	distance1	: distance between node1 and G_DOM_XSEC region
 *	distance2	: distance between node2 and G_DOM_XSEC region
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Add pins to graph
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	adjList[]	: pointers to adjacency list of graph
 *
 */


addPin(netGrP)

GNetGrDef	*netGrP;

{


GRegionDef	*node1, *node2;
adjNodeType	*nextAdj;
int		distance1, distance2;
GNodeDef	*nextNode;
int		orient;


/* For each region insert a corresponding extra graph node */
/*---------------------------------------------------------*/
for (nextNode = netGrP->node; nextNode != NULL; nextNode = nextNode->next) {

    if (nextNode->region->type == G_DOM_XSEC) {
	continue;
    }
    /* Get two adjacent G_DOM_XSEC regions to this bottleneck region */
    /*---------------------------------------------------------------*/
    orient = nextNode->region->orient;
    if (orient == G_HOR) {
	node1 = nextNode->region->adjReg[G_DOWN];
	node2 = nextNode->region->adjReg[G_UP];
    } else {
	node1 = nextNode->region->adjReg[G_LEFT];
	node2 = nextNode->region->adjReg[G_RIGHT];
    }

    /* Update adjacency info for node1 */
    nextAdj = adjList[node1->id - 1];
    while (nextAdj->id != node2->id) {
	nextAdj = nextAdj->next;
    }
    nextAdj->id = nextNode->region->id;
    distance1 = nextAdj->weight = 
		calcDistance(node1, nextNode->region, orient, netGrP);

    /* Do the same for node2 */
    nextAdj = adjList[node2->id - 1];
    while (nextAdj->id != node1->id) {
	nextAdj = nextAdj->next;
    }
    nextAdj->id = nextNode->region->id;
    distance2 = nextAdj->weight =
		calcDistance(nextNode->region, node2, orient, netGrP);

    /* Set up adjacency list for a new node */
    /*--------------------------------------*/
    nextAdj = adjList[nextNode->region->id - 1]
		= MALLOC(adjNodeType,1);
    nextAdj->id = node1->id;
    nextAdj->weight = distance1;
    nextAdj->next = MALLOC(adjNodeType,1);
    nextAdj->next->id = node2->id;
    nextAdj->next->weight = distance2;
    nextAdj->next->next = NULL;
}


}




/*
 *  calcDistance()
 *
 *  DESCRIPTION
 *  -----------
 *	Calculate the distance between newly inserted node and adjacent node.
 *
 *  RETURNS
 *  -------
 *	distance	: distance
 *
 *  Arguments:
 *  ---------
 *	reg1,2  <input>	: calculate distance from reg1 to reg2
 *	orient  <input>	: orientation of bottleneck corresponding to new node
 *	netGrP	<input>	: pointer to a net group whose pins should be added
 *
 *  Local variables:
 *  ---------------
 *	botNeck		: points to bottleneck region (either reg1 or reg2)
 *	coord_start	: smaller coordinate of the two nodes
 *	coord_end	: larger coordinate of the two nodes
 *	pin_count	: counter for #pins belonging to the bottleneck
 *	offset_sum	: sum of the offset values for the pins above
 *	ave_offset	: offset_sum / pin_count
 *	disp_start	: displacement of center from bottom/left for _start
 *	disp_end	: displacement of center from bottom/left for _end
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	adjList[]	: pointers to adjacency list of graph
 *
 */

calcDistance(reg1, reg2, orient, netGrP)

GRegionDef	*reg1, *reg2;
GNetGrDef	*netGrP;
int		orient;

{

    GRegionDef	*botNeck;
    GPinDef	*nextPin;
    int		coord_start, coord_end;
    int		pin_count;
    int		offset_sum;
    int		ave_offset;
    int		disp_start;
    int		disp_end;


    /* First, just take the center location values */
    if (orient == G_HOR) {
	coord_start = reg1->centerLocY;
	coord_end = reg2->centerLocY;
	disp_start = reg1->dimension[G_VER] / 2;
	disp_end = reg2->dimension[G_VER] / 2;
    } else {
	coord_start = reg1->centerLocX;
	coord_end = reg2->centerLocX;
	disp_start = reg1->dimension[G_HOR] / 2;
	disp_end = reg2->dimension[G_HOR] / 2;
    }

    /* Check which one is a bottlenck tile */
    if (orient == G_HOR) {
	if (reg1->type == G_HOR_BN) {
	    botNeck = reg1;
	} else {
	    botNeck = reg2;
	}
    } else {
	if (reg1->type == G_VER_BN) {
	    botNeck = reg1;
	} else {
	    botNeck = reg2;
	}
    }

    /* If net groups are NOT formed, or the net group contains a critical net,
     * get an exact location of terminal
     *----------------------------------------------------------------------*/
    if (parm.form_netgroup == OFF || netGrP->net->critical == YES) {

	pin_count = offset_sum = 0;
	for (nextPin = netGrP->net->pin; nextPin; nextPin = nextPin->next) {
	    if (nextPin->region == botNeck) {
		++pin_count;
		offset_sum += nextPin->offset;
	    }
	}
	ave_offset = offset_sum / pin_count;

	if (botNeck == reg1) {
	    coord_start += ave_offset - disp_start ;
	} else {
	    coord_end += ave_offset - disp_end ;
	}
    }

    return(coord_end - coord_start);

}




/*
 *  removePin()
 *
 *  DESCRIPTION
 *  -----------
 *	Remove from floor plan graph those nodes corresponding to pins
 *		of the given net group
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netGrP	<input>	: pointer to a net group whose pins should be added
 *
 *  Local variables:
 *  ---------------
 *	node1		: pointer to one node adjacent to G_DOM_XSEC region
 *	node2		: pointer to other node adjacent to G_DOM_XSEC region
 *	distance	: distance between the centers of node1 and node2
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Remove pins from graph
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	adjList[]	: pointers to adjacency list of graph
 *
 */


removePin(netGrP)

GNetGrDef	*netGrP;

{


GRegionDef	*node1, *node2;
adjNodeType	*nextAdj;
int		distance;
GNodeDef	*nextNode;
int		orient;


/* Do for each region removing an extra graph node */
/*-------------------------------------------------*/
for (nextNode = netGrP->node; nextNode != NULL; nextNode = nextNode->next) {

    if (nextNode->region->type == G_DOM_XSEC) {
	continue;
    }
    /* Get two adjacent G_DOM_XSEC regions to this bottleneck region */
    /*---------------------------------------------------------------*/
    orient = nextNode->region->orient;
    if (orient == G_HOR) {
	node1 = nextNode->region->adjReg[G_DOWN];
	node2 = nextNode->region->adjReg[G_UP];
    } else {
	node1 = nextNode->region->adjReg[G_LEFT];
	node2 = nextNode->region->adjReg[G_RIGHT];
    }

    /* Update adjacency info for node1 */
    nextAdj = adjList[node1->id - 1];
    while (nextAdj->id != nextNode->region->id) {
	nextAdj = nextAdj->next;
    }
    nextAdj->id = node2->id;
    if (orient == G_HOR) {
	distance = nextAdj->weight = node2->centerLocY - node1->centerLocY;
    } else {
	distance = nextAdj->weight = node2->centerLocX - node1->centerLocX;
    }

    /* Do the same for node2 */
    nextAdj = adjList[node2->id - 1];
    while (nextAdj->id != nextNode->region->id) {
	nextAdj = nextAdj->next;
    }
    nextAdj->id = node1->id;
    nextAdj->weight = distance;

    /* Remove adjacency list of the pin node */
    /*---------------------------------------*/
    nextAdj = adjList[nextNode->region->id - 1];
    FREE(nextAdj->next);
    FREE(nextAdj);
    adjList[nextNode->region->id - 1] = NULL;
}



}




