 
/*
#ifndef lint
static char rcsid[] = "$Header: /octtools/release-3.0/src/mercury/density.c,v 1.1 90/02/07 17:02:26 octtools Exp $";
#endif
*/

/*
 *  density.c
 *  *********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for obtaining a density of a routing region
 *
 *  functions contained in this file:
 *	getInnateDens()
 *	getMaxDens()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"





/*
 *  getInnateDens()
 *
 *  DESCRIPTION
 *  -----------
 *	Scan bottleneck regions and get innate density for each
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	GRegP	<input>	: pointer to the first tile plane node
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Fills in FIX matrix with values of innate channel density
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

getInnateDens(GRegP)


GRegionDef	*GRegP;

{

int		density;
GRegionDef	*nextRegion;



for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    if ((nextRegion->type != G_HOR_BN) && (nextRegion->type != G_VER_BN)) {
        continue;
    }

    if ((density = getMaxDens(nextRegion)) != 0) {
	printf("Innate density for region %d = %d\n", nextRegion->id, density);
    }
    FIX[nextRegion->type][nextRegion->id - 1] += density * G_UNITPITCH;
    nextRegion->estUsage += density * G_UNITPITCH;
}


}




/*
 *  getMaxDens()
 *
 *  DESCRIPTION
 *  -----------
 *	Get max column density for a given region
 *
 *  RETURNS
 *  -------
 *	maxDensity	max column density in this region
 *
 *  Arguments:
 *  ---------
 *	regionP	<input>	: pointer to a region for which to obtain density
 *
 *  Local variables:
 *  ---------------
 *	density		current column density while scanning columns
 *	maxDensity	max column density in this region
 *
 *  Possible errors:
 *  ---------------
 *	Works correctly only if the pins are sorted in nondecreasing order
 *	by their offset
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


getMaxDens(regionP)


GRegionDef	*regionP;

{

int	density,maxDensity;
GPinDef	*nextPin;


density = maxDensity = 0;

for (nextPin = regionP->pin; nextPin != NULL; nextPin = nextPin->next) {
    if (nextPin->netStat == ON) {
	++density;
    } else {
	--density;
    }
    if (density > maxDensity) {
	maxDensity = density;
    }
}
return(maxDensity);


}





