/* graph.h - Data structure for graph operation used in Mercury.
		The representation here is abstract compared to tiles and so on.
 
   Copyright (C) 1988 by Yoshihito Nishizaki.  All rights reserved.

 */


typedef struct net {
    int critical;
    int maxLength;
    int trivial;
    struct net *next;
    struct netNode *node;
    }
    netType;
typedef struct netNode {
    int id;
    struct netNode *next;
    struct nodeGroup *equiv;
    }
    netNodeType;

#ifdef OLD
typedef struct nodeGroup {
    int id;
    struct nodeGroup *next;
    }
    nodeGroupType;
typedef struct path {
    int length;
    struct nodeGroup *node;
    struct path *next;
    }
    pathType;
#endif


/* Graph representation by adjacency list */
/*----------------------------------------*/
typedef struct adjNode {
    int id;
    int weight;
    short blocked;
    struct adjNode *next;
    }
    adjNodeType;

/* Pointer to the first adjacent node */
/*------------------------------------*/
adjNodeType *adjList[G_MAXNODE];
/* Flags. If YES, this node is blocked and no edge from this node is usable */
/*--------------------------------------------------------------------------*/
short nodeBlock[G_MAXNODE];


/* Stuff for dijkstra shortest path algorithm */
/*--------------------------------------------*/
typedef struct dijkstraNode {
    struct dijkstraNode *v2link;
    int d;
    int parent;
    int vset;
    struct adjNode *adjlist;
    }
    dijkstraNodeType;
dijkstraNodeType *dijkstraPointer[G_MAXNODE];
dijkstraNodeType *v2link0;

/* Modified graph representaion for lawler forced inclusion and exclusion */
/*------------------------------------------------------------------------*/
adjNodeType *modAdjList[G_MAXNODE];

/* Stuff for m shortest paths algorithm */
typedef struct lawlerPath {
    int constrainedNode;		/* forced to trace nodes up to here */
    int constrainedLength;		/* length of forced path up to here */
    int length;				/* total length */
    struct lawlerNode *node;		/* chain of nodes */
    }
    lawlerPathType;
typedef struct lawlerNode {
    int id;				/* node id */
    struct lawlerNode *next;		/* next node */
    struct nodeGroup *closedNode;	/* nodes closed from this node */
    }
    lawlerNodeType;


/* Pointer to m shortest paths for each of lawler execution */
/*----------------------------------------------------------*/
lawlerPathType *pathPointer[G_MAXPATH];

/* Connected and unconnected nodes in multiLawler route construction */
/*-------------------------------------------------------------------*/
netNodeType *C,*U;

/* 
   List of paths obtained by Lawler's algorithm
   --------------------------------------------
	1st suffix : #connected pin group (level of multiLawler recursion)
	2nd suffix : suffix for m shortest paths in each level
*/
lawlerPathType *pathList[G_MAXNODE][G_MAXPATH];


/* Which path of the m shortest paths is used now for route construction */
/*-----------------------------------------------------------------------*/
/*	index : # of iteration in multiLawler algorithm */
int pathNo[G_MAXNODE];


routeType	*routePointer[G_MAXROUTE];
netType		*netPointer[G_MAXNET];


