
/*
#ifndef lint
static char rcsid[] = "$Header: /octtools/release-3.0/src/mercury/intProg.c,v 1.1 90/02/07 17:02:36 octtools Exp $";
#endif
*/

/*
 *  intProg.c
 *  *********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for formulating and solving a global router
 *	problem by interger programming
 *
 *  functions contained in this file:
 *	intProg()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  intProg()
 *
 *  DESCRIPTION
 *  -----------
 *	Formulate and Solve a global routing problem as integer programming
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	netNo		: net ID (used to specify critical nets in a file)
 *	M		: # of shortest routes as candidate per net group
 *	Mcon		: # of paths selected per pin connection 
 *			  when obtaining M candidates for a net group
 *	buf		: string buffer for system command
 *	netGrCount	: counter for net group
 *	pass_times[i]	: # of times the given net passes region i
 *				pass-thru	: 1
 *				termination	: 0.5
 *				T cross		: 1.5
 *				+ cross		: 2.0
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	parm		: parameter struct
 *	statis		: statistics struct
 *
 *
 */

intProg()


{


GRegionDef	*nextRegion,*getNextRegion();
GNetGrDef	*nextNetGr;
GNetDef		*nextNet;
GNodeDef	*nextGNode;
critPathType	*critPathP;
netType		*graphNet, *makeNet();

int		netNo;
int		i,j,k;

int		netGrCount;
int		M;
int		Mcon;



/**********************/
/* Formulation begins */
/**********************/


/* Some initialization */
/*---------------------*/
M = parm.M;
for (i = 0; i <= NO_REGIONTYPE-1; ++i) {
    for (j = 0; j <= G_MAXREGION - 1; ++j) {
	FIX[i][j] = 0;
    }
}
for (i = 0; i <= NO_REGIONTYPE-2; ++i) {
    for (j = 0; j <= G_MAXPOLAR - 1; ++j) {
	for (k = 0; k <= G_MAXREGION - 1; ++k) {
	    INC[i][j][k] = 0;
	}
    }
}

statis.num_net = 0;
statis.num_crit_net = 0;
statis.num_trivial_net = 0;
statis.num_netgroup = 0;
statis.num_nontriv_netgroup = 0;
statis.ave_net_per_netgroup = 0.0;

statis.num_block = 0;

statis.num_hor_bot = 0;
statis.num_ver_bot = 0;
statis.num_overcell = 0;

statis.num_GRgraph_node = 0;
statis.num_GRgraph_edge = 0;

statis.init_area = 0;
statis.init_netlength = 0;
statis.final_area = 0;
statis.area_ratio = 0.0;

statis.num_variable = 0;
statis.num_constraint = 0;

for (i = 0; i < G_NHV; ++i) {
    statis.num_polarpath[i] = 0;
    statis.critical_dim[i] = 0;
    statis.LP_init_dim[i] = 0;
    statis.LP_improv_dim[i] = 0;
    statis.final_dim[i] = 0;
    statis.dim_ratio[i] = 0.0;
}

for (i = 0; i < G_MAXROUTE; ++i) {
    statis.num_net_cand_i[i] = 0;
    statis.percent_net_cand_i[i] = 0.0;
}


/* Get a floorplan graph from tile plane */
/*---------------------------------------*/
/* OBSOLETE -- Next stuff is for DFS version of getFPgraph()
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    DFSflag[i] = OFF;
}
getFPgraphDFS(GRegList,adjList);
 */

getFPgraph(GRegList,adjList);
printf("Printing the floor plan graph .....\n");
printAdjMatrix(adjList);
printCPUtime("Floor plan graph obtained\n", NO);


/* Get a horizontal and vertical polar graph from tile plane */
/*-----------------------------------------------------------*/
getPolarGraph(GRegList);
printCPUtime("Polar graph obtained\n", NO);


/* Select and mark the critical nets to be routed first or manually */
/*------------------------------------------------------------------*/
for (nextNet = GNet; nextNet != NULL; nextNet = nextNet->next) {
    ++(statis.num_net);
    if (askCritical(nextNet) == YES) {
	printf("Net %s is critical\n", nextNet->name);
	nextNet->critical = YES;
	++statis.num_crit_net;
    }
}
printf("\n");

/* Divide nets into groups based on net topology
 *	and initialize WGT using real # of net groups rather than MAX # */
/*----------------------------------------------------------------------*/
/* if not specified by command, determine if we form net groups or not */
if (parm.form_netgroup < 0) {
    for (i = 0, nextNet = GNet; nextNet; ++i, nextNet = nextNet->next);
    if (i >= NETGROUP_THRESHOLD) {
	parm.form_netgroup = ON;
    } else {
	parm.form_netgroup = OFF;
    }
}
statis.num_netgroup = makeNetGroup();
sprintf(buf, "%d net groups formed\n", statis.num_netgroup);
printCPUtime(buf, NO);

for (j = 0; j <= G_MAXREGION - 1; ++j) {
    WGT[j] = (float *)malloc(M * statis.num_netgroup * sizeof(float));
    for (k = 0; k <= M * statis.num_netgroup - 1; ++k) {
	WGT[j][k] = 0.0;
    }
}
printCPUtime("WGT matrix initialized\n", NO);


/* Get a density demand innate to each bottleneck region
 * It is independent of global routing results
 * The result is stored in FIX matrix
 *----------------------------------------------------*/
getInnateDens(GRegList);
printCPUtime("Innate density obtained\n", NO);


/* Obtain M shortest routes and calculate the related matrices */
/*-------------------------------------------------------------*/
for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

    if (askTrivialRoute(nextNetGr) == YES) {
	statis.num_trivial_net += nextNetGr->NoNet;
	printf("%dth net group (%d nets) is trivial\t: region = %d\n\n",
		nextNetGr->id, nextNetGr->NoNet, nextNetGr->node->region->id);
	continue;
    }

    ++(statis.num_nontriv_netgroup);

    /* Print the net group information */
    printf("Routing %dth net group (%d nets)\t: region =",
		nextNetGr->id,nextNetGr->NoNet);
    for (nextGNode = nextNetGr->node; nextGNode != NULL;
		nextGNode = nextGNode->next) {
	printf(" %d",nextGNode->region->id);
    }
    printf("\n\n");

    /* Determine Mcon */
    /*----------------*/
    Mcon = parm.m;
    if (nextNetGr->NoRegion >= parm.netSizeLimit) {
	Mcon = 1;
    }
    if (nextNetGr->NoRegion == 2) {
	Mcon = M;
    }

    /* Obtain M shortest routes */
    /*--------------------------*/
    addPin(nextNetGr);

#ifdef MULTI_LAWLER_DEBUG
    printf("\nPrinting the new global routing graph\n");
    printAdjMatrix(adjList);
#endif

    graphNet = makeNet(nextNetGr);
    multiLawler(graphNet, nextNetGr->NoRegion, parm.M, Mcon);
    copyRoute(nextNetGr);
    deleteNet(graphNet);
    removePin(nextNetGr);

    /* Calculate the FIX and WGT matrices */
    /*------------------------------------*/
    calcMatrix(nextNetGr);
}

printCPUtime("M shortest routes obtained\n", NO);


/* Check the validity of the critical net constraints */

for (nextRegion = GRegList; nextRegion != NULL; nextRegion = nextRegion->next) {
    if (nextRegion->type != G_OVERCELL) {
	continue;
    }
    if (nextRegion->capacity[0] < FIX[G_OVERCELL][nextRegion->id - 1]) {
	overCellError(nextRegion);
	break;
    }
}


/* Get a critical path on the H/V polar graph */
/*--------------------------------------------*/
critPathP = MALLOC(critPathType,1);
for (i = 0; i < G_NHV; ++i) {
    critPathP->maxLength = 0;
    critPathP->pathNo = 1;
    for (j = 0; j <= G_MAXREGION - 1; ++j) {
	critPathP->regionID[j] = 0;
    }

    getPolarPath(i, GPGraph[i], &INC[i][0][0], FIX, critPathP, 0);

    statis.critical_dim[i] = 
	RINT((float)(critPathP->maxLength) / (float)OCT_PER_MICRON);
    statis.num_polarpath[i] = --(critPathP->pathNo);

    /* Expand the chip so that it can accommodate all the nets */
    statis.LP_init_dim[i] = 
		RINT((float)(critPathP->maxLength) * parm.expandFactor);
    printf("%d %s polar paths found : critical dimension = %d u\n",
	statis.num_polarpath[i], i == G_HOR ? "HOR" : "VER", 
	statis.critical_dim[i]);
    printf("Set initial %s dimension for LP = %d oct_units\n",
	i == G_HOR ? "HOR" : "VER", statis.LP_init_dim[i]);
}
printf("\n");

printCPUtime("Critical polar path obtained\n", NO);


/* Print Matrices */
/*----------------*/
/*
printFIX();
printINC();
printWGT();
 */


/* M = 1 && placement refinement mode = 0 is a special case */
if (parm.M == 1 && parm.placeRefMode == 0) {

    for (nextNetGr = GNetGr; nextNetGr; nextNetGr = nextNetGr->next) {
	if ((askTrivialRoute(nextNetGr) == YES)
		|| (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
	    continue;
	}
	nextNetGr->IPresult[0] = nextNetGr->NoNet - nextNetGr->NoCritNet;
    }
    statis.num_net_cand_i[0] = statis.num_net;

} else {

    /* Formulate the global routing as an integer programming problem */
    /*----------------------------------------------------------------*/
    formulateLindo(statis.LP_init_dim);
    printCPUtime("Lindo input file created\n", NO);

    /* Solve the global routing as an integer programming problem */
    /*------------------------------------------------------------*/
    (void) sprintf(buf, "rm lindo.out");
    system(buf);
    (void) sprintf(buf, "lindo < lindo.in");
    if (system(buf) != 0) {
	printf("\nYOU DO NOT HAVE LINDO LINEAR PROGRAMMING PACKAGE\n");
	printf("EXITING.....\n");
	printf("PLEASE OBTAIN LINDO FROM\n\n");
	printf("\tThe Scientific Press, 540 University Ave.\n");
	printf("\tPalo Alto, CA 94301, USA\n");
	printf("\tTEL : (415) 322-5221\n\n");
	printf("OR RUN MERCURY WITH -M 1 -p 0 option\n\n");
	exit(5);
    }

    printCPUtime("Lindo executed\n", NO);

    /* Read Lindo result and assign a route to each net as appropriate */
    /*-----------------------------------------------------------------*/
    getLindoResult(M);
}

for (j = 0; j <= G_MAXREGION - 1; ++j) {
    free(WGT[j]);
}

assignRouteToNet();
printCPUtime("Assigned a route for each net\n", NO);


}



