 
/*
#ifndef lint
static char rcsid[] = "$Header: /octtools/release-3.0/src/mercury/polarGraph.c,v 1.1 90/02/07 17:02:56 octtools Exp $";
#endif
*/

/*
 *  polarGraph.c
 *  ************
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines concerning polar graph
 *
 *  functions contained in this file:
 *	getPolarGraph()
 *	getPolarPath()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"





/*
 *  getPolarGraph()
 *
 *  DESCRIPTION
 *  -----------
 *	Get a polar graph from tile plane
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	GRegP	<input>	: pointer to the first region in a tile plane
 *
 *  Local variables:
 *  ---------------
 *	id2Polar[i]	: stores pointer to the polar node representing region i
 *	baseReg		: base region for which to get neighbors
 *	direction	: direction of the neighbor from the base region
 *	reg_s1, reg_s2	: two regions closer to the source on the + junction
 *	reg_t1, reg_t2	: two regions closer to the sink on the + junction
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Constructs polar grphs and fill in values in GPGraph[]
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GID2Reg[i]	: stores pointer to a region i
 *
 */


getPolarGraph(GRegP)


GRegionDef	*GRegP;

{


GPolarDef	*id2Polar[G_MAXREGION];
GPolarDef	*nextPolar;
GPolarDef	*savePolar;
GRegionDef	*nextRegion;
GRegionDef	*nextAdjReg;
int		i;
GRegionDef	*baseReg;
int		direction;
GRegionDef	*reg_s1, *reg_s2, *reg_t1, *reg_t2;


/* Construct horizontal polar graph first */
/*----------------------------------------*/

/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    id2Polar[i] = NULL;
}

/* Allocate GPolarDef for each polar graph node */

for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    if ((nextRegion->type == G_DOM_XSEC) || (nextRegion->type == G_VER_BN)) {
	continue;
    }
    nextPolar = id2Polar[nextRegion->id - 1] = MALLOC(GPolarDef,1);
    nextPolar->region = GID2Reg[nextRegion->id - 1];
    nextPolar->pGraph = NULL;
    nextPolar->nextPGraph = NULL;
    nextPolar->length = 0;
}

/* Obtain horizontal polar information */

for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    if ((nextRegion->type == G_DOM_XSEC) || (nextRegion->type == G_VER_BN)) {
	continue;
    }

/* If the region has no left neighbor, it is adjacent to the source */
/*------------------------------------------------------------------*/
    if (nextRegion->adjReg[G_LEFT] == NULL) {
	savePolar = GPGraph[G_HOR];
	nextPolar = GPGraph[G_HOR] = id2Polar[nextRegion->id - 1];
	nextPolar->nextPGraph = savePolar;
    }
/* Make a polar connection for each of its right neighbor */
/*--------------------------------------------------------*/
    for (baseReg = nextRegion, direction = G_RIGHT;
	(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; baseReg = NULL) {

	if ((nextAdjReg->type != G_DOM_XSEC)
		&& (nextAdjReg->type != G_VER_BN)) {

	    if (id2Polar[nextRegion->id - 1]->pGraph == NULL) {
		nextPolar = id2Polar[nextRegion->id - 1]->pGraph
			= id2Polar[nextAdjReg->id - 1];
	    } else {
		nextPolar->nextPGraph = id2Polar[nextAdjReg->id - 1];
		nextPolar = nextPolar->nextPGraph;
	    }
	}
    }
}

/* If model diagonal effect, extend the adjancency in the diagonal direction */
/*---------------------------------------------------------------------------*/
if (parm.modelDiagonal == ON) {
    for (nextRegion = GRegP; nextRegion != NULL; 
		nextRegion = nextRegion->next) {

	if (nextRegion->type != G_DOM_XSEC) {
	    continue;
	} else if (askPlusJunction(nextRegion) == NO) {
	    continue;
	}

	/* Find the four blocks surrounding the + junction */
	for (baseReg = nextRegion, direction = G_LEFT;
		(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; 
		baseReg = NULL) {
	    if (nextAdjReg->type == G_VER_BN) {
		reg_s1 = nextAdjReg->adjReg[G_UP];
		reg_s2 = nextAdjReg->adjReg[G_DOWN];
		break;
	    }
	}
	for (baseReg = nextRegion, direction = G_RIGHT;
		(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; 
		baseReg = NULL) {
	    if (nextAdjReg->type == G_VER_BN) {
		reg_t1 = nextAdjReg->adjReg[G_UP];
		reg_t2 = nextAdjReg->adjReg[G_DOWN];
		break;
	    }
	}
	/* Make a polar connection in diagonal direction */
	for (nextPolar = (id2Polar[reg_s1->id - 1])->pGraph; 
		nextPolar->nextPGraph != NULL;
		nextPolar = nextPolar->nextPGraph) { }
	nextPolar->nextPGraph = MALLOC(GPolarDef,1);
	nextPolar = nextPolar->nextPGraph;
	nextPolar->region = GID2Reg[nextRegion->id - 1];
	nextPolar->pGraph = id2Polar[reg_t2->id - 1];
	nextPolar->nextPGraph = NULL;
	nextPolar->length = 0;

	for (nextPolar = (id2Polar[reg_s2->id - 1])->pGraph; 
		nextPolar->nextPGraph != NULL;
		nextPolar = nextPolar->nextPGraph) { }
	nextPolar->nextPGraph = MALLOC(GPolarDef,1);
	nextPolar = nextPolar->nextPGraph;
	nextPolar->region = GID2Reg[nextRegion->id - 1];
	nextPolar->pGraph = id2Polar[reg_t1->id - 1];
	nextPolar->nextPGraph = NULL;
	nextPolar->length = 0;
    }
}

/* Print polar graph info */
printf("Printing horizontal polar graph\n");
printf("First HOR polar node = %d\n",GPGraph[G_HOR]->region->id);
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    if (id2Polar[i] == NULL) {
	continue;
    }
    printf("region %d : pGraph = %d, nextPGraph = %d\n", i+1,
	(id2Polar[i]->pGraph == NULL) ? 0 : id2Polar[i]->pGraph->region->id,
	(id2Polar[i]->nextPGraph == NULL) ?
		0 : id2Polar[i]->nextPGraph->region->id);

    /* There may be two extra polar nodes representing the + junction */
    /*----------------------------------------------------------------*/
    if (id2Polar[i]->nextPGraph != NULL && 
		id2Polar[i]->nextPGraph->region->type == G_DOM_XSEC) {

	nextPolar = id2Polar[i]->nextPGraph;
	printf("region %d : pGraph = %d, nextPGraph = %d\n",
		nextPolar->region->id,
		(nextPolar->pGraph == NULL) ? 
			0 : nextPolar->pGraph->region->id,
		(nextPolar->nextPGraph == NULL) ?
			0 : nextPolar->nextPGraph->region->id);

	if (nextPolar->nextPGraph != NULL && 
		nextPolar->nextPGraph->region->type == G_DOM_XSEC) {

	    nextPolar = nextPolar->nextPGraph;
	    printf("region %d : pGraph = %d, nextPGraph = %d\n",
		nextPolar->region->id,
		(nextPolar->pGraph == NULL) ? 
			0 : nextPolar->pGraph->region->id,
		(nextPolar->nextPGraph == NULL) ?
			0 : nextPolar->nextPGraph->region->id);
	}
    }
}
printf("\n");


/* Construct vertical polar graph next */
/*-------------------------------------*/

/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    id2Polar[i] = NULL;
}

/* Allocate GPolarDef for each polar graph node */

for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    if ((nextRegion->type == G_DOM_XSEC) || (nextRegion->type == G_HOR_BN)) {
	continue;
    }
    nextPolar = id2Polar[nextRegion->id - 1]
		= MALLOC(GPolarDef,1);
    nextPolar->region = GID2Reg[nextRegion->id - 1];
    nextPolar->pGraph = NULL;
    nextPolar->nextPGraph = NULL;
    nextPolar->length = 0;
}

/* Obtain vertical polar information */

for (nextRegion = GRegP; nextRegion != NULL; nextRegion = nextRegion->next) {

    if ((nextRegion->type == G_DOM_XSEC) || (nextRegion->type == G_HOR_BN)) {
	continue;
    }

/* If the region has no down neighbor, it is adjacent to the source */
/*------------------------------------------------------------------*/
    if (nextRegion->adjReg[G_DOWN] == NULL) {
	savePolar = GPGraph[G_VER];
	nextPolar = GPGraph[G_VER] = id2Polar[nextRegion->id - 1];
	nextPolar->nextPGraph = savePolar;
    }
/* Make a polar connection for each of its up neighbor */
/*-----------------------------------------------------*/
    for (baseReg = nextRegion, direction = G_UP;
	(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; baseReg = NULL) {

	if ((nextAdjReg->type != G_DOM_XSEC)
		&& (nextAdjReg->type != G_HOR_BN)) {

	    if (id2Polar[nextRegion->id - 1]->pGraph == NULL) {
		nextPolar = id2Polar[nextRegion->id - 1]->pGraph
			= id2Polar[nextAdjReg->id - 1];
	    } else {
		nextPolar->nextPGraph = id2Polar[nextAdjReg->id - 1];
		nextPolar = nextPolar->nextPGraph;
	    }
	}
    }
}

/* If model diagonal effect, extend the adjancency in the diagonal direction */
/*---------------------------------------------------------------------------*/
if (parm.modelDiagonal == ON) {
    for (nextRegion = GRegP; nextRegion != NULL; 
		nextRegion = nextRegion->next) {

	if (nextRegion->type != G_DOM_XSEC) {
	    continue;
	} else if (askPlusJunction(nextRegion) == NO) {
	    continue;
	}

	/* Find the four blocks surrounding the + junction */
	for (baseReg = nextRegion, direction = G_DOWN;
		(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; 
		baseReg = NULL) {
	    if (nextAdjReg->type == G_HOR_BN) {
		reg_s1 = nextAdjReg->adjReg[G_LEFT];
		reg_s2 = nextAdjReg->adjReg[G_RIGHT];
		break;
	    }
	}
	for (baseReg = nextRegion, direction = G_UP;
		(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; 
		baseReg = NULL) {
	    if (nextAdjReg->type == G_HOR_BN) {
		reg_t1 = nextAdjReg->adjReg[G_LEFT];
		reg_t2 = nextAdjReg->adjReg[G_RIGHT];
		break;
	    }
	}
	/* Make a polar connection in diagonal direction */
	for (nextPolar = (id2Polar[reg_s1->id - 1])->pGraph; 
		nextPolar->nextPGraph != NULL;
		nextPolar = nextPolar->nextPGraph) { }
	nextPolar->nextPGraph = MALLOC(GPolarDef,1);
	nextPolar = nextPolar->nextPGraph;
	nextPolar->region = GID2Reg[nextRegion->id - 1];
	nextPolar->pGraph = id2Polar[reg_t2->id - 1];
	nextPolar->nextPGraph = NULL;
	nextPolar->length = 0;

	for (nextPolar = (id2Polar[reg_s2->id - 1])->pGraph; 
		nextPolar->nextPGraph != NULL;
		nextPolar = nextPolar->nextPGraph) { }
	nextPolar->nextPGraph = MALLOC(GPolarDef,1);
	nextPolar = nextPolar->nextPGraph;
	nextPolar->region = GID2Reg[nextRegion->id - 1];
	nextPolar->pGraph = id2Polar[reg_t1->id - 1];
	nextPolar->nextPGraph = NULL;
	nextPolar->length = 0;
    }
}

/* Print polar graph info */
printf("Printing vertical polar graph\n");
printf("First VER polar node = %d\n",GPGraph[G_VER]->region->id);
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    if (id2Polar[i] == NULL) {
	continue;
    }
    printf("region %d : pGraph = %d, nextPGraph = %d\n", i+1,
	(id2Polar[i]->pGraph == NULL) ? 0 : id2Polar[i]->pGraph->region->id,
	(id2Polar[i]->nextPGraph == NULL) ?
		0 : id2Polar[i]->nextPGraph->region->id);

    /* There may be two extra polar nodes representing the + junction */
    /*----------------------------------------------------------------*/
    if (id2Polar[i]->nextPGraph != NULL && 
		id2Polar[i]->nextPGraph->region->type == G_DOM_XSEC) {

	nextPolar = id2Polar[i]->nextPGraph;
	printf("region %d : pGraph = %d, nextPGraph = %d\n",
		nextPolar->region->id,
		(nextPolar->pGraph == NULL) ? 
			0 : nextPolar->pGraph->region->id,
		(nextPolar->nextPGraph == NULL) ?
			0 : nextPolar->nextPGraph->region->id);

	if (nextPolar->nextPGraph != NULL && 
		nextPolar->nextPGraph->region->type == G_DOM_XSEC) {

	    nextPolar = nextPolar->nextPGraph;
	    printf("region %d : pGraph = %d, nextPGraph = %d\n",
		nextPolar->region->id,
		(nextPolar->pGraph == NULL) ? 
			0 : nextPolar->pGraph->region->id,
		(nextPolar->nextPGraph == NULL) ?
			0 : nextPolar->nextPGraph->region->id);
	}
    }
}
printf("\n");


}






/*
 *  getPolarPath()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain a critical polar path
 *
 *  RETURNS
 *  -------
 *	node		when a shortest path is found
 *
 *  Arguments:
 *  ---------
 *	direction	<input>	: direction of polar path (G_HOR or G_VER)
 *	source		<input>	: current source node for the search
 *	incMatrix	<in/out>: incidence matrix
 *				  represents which region belongs to which path
 *	fixMatrix	<in/out>: fixed usage matrix
 *				  represents fixed usage of a region
 *	critPathP	<input>	: pointer to critical path struct
 *	pathLength	<input>	: path length for current polar path
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


getPolarPath(direction,source,incMatrix,fixMatrix,critPathP,pathLength)


int		direction;
GPolarDef	*source;
int		incMatrix[][G_MAXREGION];
int		fixMatrix[][G_MAXREGION];
critPathType	*critPathP;
int		pathLength;

{


int		saveLength;
int		i,j,idx;


/* Update some  critical path data */
/*---------------------------------*/
saveLength = pathLength;
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    if (critPathP->regionID[i] != 0) {
	continue;
    } else {
	critPathP->regionID[i] = source->region->id;
	idx = i;
	break;
    }
}
if (source->region->type == G_INST) {
    pathLength += source->region->dimension[direction];
    fixMatrix[direction][source->region->id - 1]
		= source->region->dimension[direction];
} else {
    pathLength += (source->region->estUsage + source->length);
}


/* Proceed to the sink */
/*---------------------*/
if (source->pGraph != NULL) {
    getPolarPath(direction,source->pGraph,incMatrix,
		fixMatrix,critPathP,pathLength);

/* If the next region is the sink, you have found a path */
/*-------------------------------------------------------*/
} else {
    printf("Found a polar path %d (length = %d) : ",
		critPathP->pathNo, pathLength);
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	if (critPathP->regionID[i] == 0) {
	    printf("\n");
	    break;
	}
	printf(" %d",critPathP->regionID[i]);
	for (j = 0; j <= G_MAXREGION - 1; ++j) {
	    if (incMatrix[critPathP->pathNo - 1][j] == 0) {
		incMatrix[critPathP->pathNo - 1][j] = critPathP->regionID[i];
		break;
	    } else {
		continue;
	    }
	}
    }

/* Update critical polar path information */
/*----------------------------------------*/
    if (pathLength > critPathP->maxLength) {
	critPathP->maxLength = pathLength;
	critPathP->critPathNo = critPathP->pathNo;
    }
    ++(critPathP->pathNo);
}

/* If the preceding polar node has a branch path, search the sibling */
/*-------------------------------------------------------------------*/
if (source->nextPGraph != NULL) {
    critPathP->regionID[idx] = 0;
    getPolarPath(direction,source->nextPGraph,incMatrix,
		fixMatrix,critPathP,saveLength);
    return;
} else {
    critPathP->regionID[idx] = 0;
    return;
}


}



