#include "mighty.h"

void S_printstats()
/*
 *  print statistics - number of vias, total line length,
 *                     number of rows and columns.
 */
{
    LINKPTR tlink;
    LINKLISTPTR list;
    int i;  /* counter */
    int x,
	y;
    int hlink,
        vlink; 		/* To mark the existence of the links */
    int hlength = 0;
    int vlength = 0;
    int nvia = 0;

    /*
     *  find the wire length
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tlink = net_array[i].path; tlink != NULL;
	     tlink = tlink->netnext )
	{
	    if( tlink->layer == HORIZ )
		hlength += tlink->x2 - tlink->x1 + tlink->y2 - tlink->y1;
	    else
		vlength += tlink->x2 - tlink->x1 + tlink->y2 - tlink->y1;
	}
    }

    /*
     *  count the vias
     */
    for( y = 1; y <= num_rows; y++ )
    {
	for( x = 1; x <= num_cols; x++ )
	{
	    if( hlayer[y][x] == vlayer[y][x] && hlayer[y][x] != 0 &&
		hlayer[y][x] <= num_nets )
	    {
		hlink = vlink = NO;
		for( list = linklayer[y][x]; list != NULL; list = list->next )
		{
		    tlink = list->link;
		    if( tlink->layer == HORIZ )
			hlink = YES;
		    else
			vlink = YES;

		}

		/*
		 *  To include vias on the boundary
		 */
		if( playerh[y][x] > 0 )
		    hlink = YES;
		if( playerv[y][x] > 0 )
		    vlink = YES;

		if( hlink == YES && vlink == YES )
		{
		    nvia++;
		}
	    }
	}
    }
    printf(" There are %d vias\n", nvia );
    printf(" The total net_length is %d\n", hlength + vlength );
    printf(" ( hlength = %d,  vlength = %d )\n", hlength, vlength );
}

S_dumppins( net )
int net;
{
    PINPTR pin1;

    printf("*** pins of net %d ***\n", net );
    for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next )
    {
	printf("(%d %d) %s %s\n", pin1->x, pin1->y,
	    (pin1->layer == HORIZ ? "H":"V"),
	    (pin1->type == TEMP ? "T":"P") );
    }
}

S_dumppath( net )
int net;
/*
 *  dump all path of the net
 */
{
    LINKPTR tlink;
    printf(" path of net %d\n", net );
    for( tlink = net_array[net].path; tlink != NULL;
	tlink = tlink->netnext )
    {
	printf("%s (%d, %d) to (%d, %d)\n",
	  (tlink->layer == HORIZ ? "H" : (tlink->layer == VERT ? "V" : "N" )),
	  tlink->x1, tlink->y1, tlink->x2, tlink->y2 );
    }
    printf(" tmp of net %d\n", net );
    for( tlink = net_array[net].tmplist; tlink != NULL;
	tlink = tlink->netnext )
    {
	printf("%s (%d, %d) to (%d, %d)\n",
	  (tlink->layer == HORIZ ? "H" : (tlink->layer == VERT ? "V" : "N" )),
	  tlink->x1, tlink->y1, tlink->x2, tlink->y2 );
    }
    printf(" oldpath of net %d\n", net );
    for( tlink = net_array[net].oldpath; tlink != NULL;
	tlink = tlink->netnext )
    {
	printf("%s (%d, %d) to (%d, %d)\n",
	  (tlink->layer == HORIZ ? "H" : (tlink->layer == VERT ? "V" : "N" )),
	  tlink->x1, tlink->y1, tlink->x2, tlink->y2 );
    }
    printf(" oldtmp of net %d\n", net );
    for( tlink = net_array[net].oldtmplist; tlink != NULL;
	tlink = tlink->netnext )
    {
	printf("%s (%d, %d) to (%d, %d)\n",
	  (tlink->layer == HORIZ ? "H" : (tlink->layer == VERT ? "V" : "N" )),
	  tlink->x1, tlink->y1, tlink->x2, tlink->y2 );
    }
}

S_markborder()
/*
 *  Mark the boundary of the routing area and
 *  obstacles
 */
{
    LINKPTR blink;
    int **flayer;
    int border;
    int xout,
	yout,
	xpos,
	ypos;

    border = num_nets + 1;

    /*
     *  Mark bounding box
     */
    for( blink = bounding_box; blink != NULL; blink = blink->netnext )
    {
/*
print_debug ("bb", "horiz", channel->horiz_layer, num_rows, num_cols, num_nets);
print_debug ("bb", "vert", channel->vert_layer, num_rows, num_cols, num_nets);
printf("bb (%d, %d) (%d, %d)\n", blink->x1, blink->y1, blink->x2, blink->y2 );
*/
	if( blink->x1 < blink->x2 )
	{
	    ypos = blink->y1;
	    yout = ypos - 1;
	    for( xpos = blink->x1; xpos <= blink->x2; xpos++ )
	    {
		hlayer[ypos][xpos] = vlayer[ypos][xpos] = border;
		hlayer[yout][xpos] = vlayer[yout][xpos] = border;
	    }
	}
	else if( blink->x2 < blink->x1 )
	{
	    ypos = blink->y1;
	    yout = ypos + 1;
	    for( xpos = blink->x2; xpos <= blink->x1; xpos++ )
	    {
		hlayer[ypos][xpos] = vlayer[ypos][xpos] = border;
		hlayer[yout][xpos] = vlayer[yout][xpos] = border;
	    }
	}
	else if( blink->y1 < blink->y2 )
	{
	    xpos = blink->x1;
	    xout = xpos + 1;
	    for( ypos = blink->y1; ypos <= blink->y2; ypos++ )
	    {
		hlayer[ypos][xpos] = vlayer[ypos][xpos] = border;
		hlayer[ypos][xout] = vlayer[ypos][xout] = border;
	    }
	}
	else if( blink->y2 < blink->y1 )
	{
	    xpos = blink->x1;
	    xout = xpos - 1;
	    for( ypos = blink->y2; ypos <= blink->y1; ypos++ )
	    {
		hlayer[ypos][xpos] = vlayer[ypos][xpos] = border;
		hlayer[ypos][xout] = vlayer[ypos][xout] = border;
	    }
	}
    }

    /*
     *  Mark obstacles
     */
    for( blink = obstacle; blink != NULL; blink = blink->netnext )
    {
/*
print_debug ("bb", "horiz", channel->horiz_layer, num_rows, num_cols, num_nets);
print_debug ("bb", "vert", channel->vert_layer, num_rows, num_cols, num_nets);
printf("ob %d, ( %d, %d) (%d, %d)\n", blink->layer, blink->x1, blink->y1,
blink->x2, blink->y2 );
*/
	if( blink->layer == HORIZ )
	    flayer = hlayer;
	else
	    flayer = vlayer;

	if( blink->x1 < blink->x2 )
	{
	    ypos = blink->y1;
	    for( xpos = blink->x1; xpos <= blink->x2; xpos++ )
	    {
		flayer[ypos][xpos] = border;
	    }
	}
	else if( blink->x2 < blink->x1 )
	{
	    ypos = blink->y1;
	    for( xpos = blink->x2; xpos <= blink->x1; xpos++ )
	    {
		flayer[ypos][xpos] = border;
	    }
	}
	else if( blink->y1 < blink->y2 )
	{
	    xpos = blink->x1;
	    for( ypos = blink->y1; ypos <= blink->y2; ypos++ )
	    {
		flayer[ypos][xpos] = border;
	    }
	}
	else if( blink->y2 < blink->y1 )
	{
	    xpos = blink->x1;
	    for( ypos = blink->y2; ypos <= blink->y1; ypos++ )
	    {
		flayer[ypos][xpos] = border;
	    }
	}
	else
	{
	    flayer[blink->y1][blink->x1] = border;
	}
    }
}

S_equivpins( net )
int net;
/*
 *  Mark set of equivalent pins which are on one grid-point
 */
{
    PINPTR pin1,
	   pin2;
    int numequiv = 0;

    for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next )
    {
	pin1->equiv = 0;
    }

    if( net_array[net].pin != NULL )
    {
      for( pin1 = net_array[net].pin; pin1->next != NULL; pin1 = pin1->next )
      {
	for( pin2 = pin1->next; pin2 != NULL; pin2 = pin2->next )
	{
	    if( pin1->x == pin2->x && pin1->y == pin2->y )
	    {
		numequiv++;
		pin1->equiv = numequiv;
		pin2->equiv = numequiv;
	    }
	}
      }
    }
    net_array[net].numequiv = numequiv;
}

int S_compensateequiv( net )
int net;
/*
 *  equivalent pins and links must have the same group number
 */
{
    PINPTR pin1;
    LINKPTR link;
    int i;
    int mingp = INFINITY;

    for( i = 1; i <= net_array[net].numequiv; i++ )
    {
	/*
	 *  find minimum group number of class i
	 */
	for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next)
	{
	    if( pin1->equiv == i )
	        mingp = MIN( mingp, pin1->group );
	}
	/*
	 *  set group number
	 */
	for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next)
	{
	    if( pin1->equiv == i && pin1->group != mingp )
	    {
		for( link = net_array[net].path; link != NULL; link = link->netnext )
		{
		    if( link->group == pin1->group )
			link->group = mingp;
		}
		pin1->group = mingp;
	    }
	}
    }

    mingp = INFINITY;
    for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next)
    {
	/* If mutiple components, return 2 */
	if( pin1->group > 1 )
	    mingp = MIN( mingp, pin1->group );
    }
    if( mingp == INFINITY )
	return( 1 );
    else if( mingp == 2 )
	return( 2 );
    else
    {
	/*
	 *  Shift the mingp to 2, so that group 1 and 2 exist
	 *  S_improve and S_powerimprove use groups 1 and 2
	 */
	for( pin1 = net_array[net].pin; pin1 != NULL; pin1 = pin1->next)
	{
	    if( pin1->group == mingp )
		pin1->group = 2;
	}
	for( link = net_array[net].path; link != NULL; link = link->netnext )
	{
	    if( link->group == mingp )
		link->group = 2;
	}
    }
    return( 3 ); /* NOT EXACT */
}


void S_connectboundary()
/*
 *  connect boundary - connect the pins on the boundary( tmppin )
 */
{
    PINPTR cpin;
    int i;  /* counter */
    int x2,
	y2;

    /*
     *  add link to connect tmppins
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( cpin = net_array[i].tmppin; cpin != NULL;
	     cpin = cpin->next )
	{
	    x2 = cpin->x;
	    y2 = cpin->y;

	    switch( cpin->group )
	    {
	    	case DOWN :
		   y2++;
		   break;
	    	case UP   :
		   y2--;
		   break;
		case LEFT :
		   x2--;
		   break;
		case RIGHT :
		   x2++;
		   break;
	    }
	    S_addpathlink( cpin->layer, cpin->x, cpin->y, x2, y2, i );
	}
    }
}

