#include "mighty.h"
extern int PREASSIGN;
int *cap;
int **cost = 0;
int *row;
int *col;
int *nexts;
int *h1;
int *h2;
int *h3;
int *h4;
int *h5;
int *h6;

void S_initassign();
void S_shortestpath();
void S_augmentation();
void S_transformation();

/*
 *  Pseudo pins are made.
 *  This is to give guide in routing long channels
 *  or highly irregular shaped channels.
 */


int **
S_structure(m, n)
int m, /* number of rows( crossing nets ) */
    n; /* number of cols( tracks ) */
{
    int j;
    int *S_allocatevector();
    int ** S_allocatematrix();

    cap = S_allocatevector( n );
    for( j = 1; j <= n; j++ )
	cap[j] = 1;
    cost = S_allocatematrix( m, n );
    row = S_allocatevector( n );
    col = S_allocatevector( m );
    nexts = S_allocatevector( m );
    h1 = S_allocatevector( m );
    h2 = S_allocatevector( n );
    h3 = S_allocatevector( m );
    h4 = S_allocatevector( n );
    h5 = S_allocatevector( n );
    h6 = S_allocatevector( n );
    return( cost );
}


void S_resetstructure(m, n)
int m, /* number of rows( crossing nets ) */
    n; /* number of cols( tracks ) */
/*
 *  reset to 0 all the entries, except capacities which are 1
 */
{
int i,
    j;
    for( j = 1; j <= n; j++ )
	cap[j] = 1;
    for( i = 1; i <= m; i++ )
        for( j = 1; j <= n; j++ )
	    cost[i][j] = 0;
    for( j = 1; j <= n; j++ )
	row[j] = 0;
    for( i = 1; i <= m; i++ )
	col[i] = 0;
    for( i = 1; i <= m; i++ )
	nexts[i] = 0;
    for( i = 1; i <= m; i++ )
	h1[i] = 0;
    for( j = 1; j <= n; j++ )
	h2[j] = 0;
    for( i = 1; i <= m; i++ )
	h3[i] = 0;
    for( j = 1; j <= n; j++ )
	h4[j] = 0;
    for( j = 1; j <= n; j++ )
	h5[j] = 0;
    for( j = 1; j <= n; j++ )
	h6[j] = 0;
}


/*ARGSUSED*/
int *
S_assign( cost, m, n )
int **cost;
int m,
    n;
/*
 *  For bipatite graph with m source nodes ( i : rows )
 *  and n sink nodes ( j : cols ) connected by edges with cost[i][j] >=0
 *  an assignment of rows i to columns j is calculated which minimoizes
 *  the cost
 *
 *        z = sum cost[i][j] * x[i][j]
	      i,j
 *  subject to
 *
 *   sum x[i][j] <= cap[j];  sum x[i][j] = 1;  x[i][j] >= 0
 *    i                       j
 *
 *  INPUT :  The capacity vector cap[1..n]
 *	     cost matrix         cost[1..m][1..n]
 *     	     problem size m, n.  m <= sum cap[j]
 *				       i
 *  OUTPUT :  Two vectors row[1..n], col[1..m] describing rows and
 *	     columns of the selected matrix entries.
 * 	     row[j] points to one of the rows assigned to col j.
 *	     Others may be accessed by nexts[row[j]] ...
 *  scratch : h1[1..m], h2[1..n], h3[1..m], h4[1..n], h5[1..n], h6[1..n]
 *
 *  method :  First an optimal assignment of a subset of rows and
 *	     columns is constructed heuristically.  Then shortest
 *	     augmenting paths are generated starting from unassigned
 *	     source nodes to sink nodes with free capacity, to determine
 *	     the solution of the complete problem.
 *	     REF : Lauther's class note; Burkhard & Derigs, "assignment
 *	     and matching problems, springer 1980
 *
 *	     Problems with m < n are handled as an n by n problem with
 *	     c[i][j] = INFINITY for i > m.
 *	     Sink nodes j with cap[j] > 1 are handled like a bunch of
 *	     cap[j] nodes with capacity 1 each.
 */
{
    int i,
	j,
	k;

    /*
     *  Initialize candidate list for shortest path
     */
    S_initassign(h1, h2, m, n);

    for( j = 2; j <= n; j++ )
	h6[j] = j - 1;
    h6[1] = n;

    for( i = 1; i <= m; i++ )
    {
	if( col[i] == 0 )
	{
	    S_shortestpath( h1, h2, h3, h4, h5, h6, i, &j, &k, m, n );
	    S_augmentation( h5, i, j );
	    S_transformation( h1, h2, h3, h4, k, m, n );
	}
    }
    return( col );
}


static void
S_initassign(ys, yt, m, n)
int ys[],
    yt[];
int m, n;
/*
 *  Initializes, checks feasibility and constructs an optimal assignment
 *  for a subset of rows and columns :
 *  First we calculate ys[i] = min cost[i][j];  yt[j] = min (cost[i][j] - ys[i])
 *				j			 i
 *
 *  Then admissible assignments (i,j) with c[i][j] = ys[i] + yt[j]
 *  are made.
 */
{
    int i,
	j,
	j0,
	ui,
	cij;
    int neff = 0;  /* total capacity */

    /*  initial assignment */
    for( j = 1; j <= n; j++ )
    {
	row[j] = 0;
	neff += cap[j];
    }
    if( neff < m )
    {
	printf("ERROR : total capacity too small\n");
	fprintf(output_file, "failed\n");
	exit( 03 );
    }

    /*
     *  calculate ys and make trvial assignments
     */
    for( i = 1; i <= m; i++ )
    {
	col[i] = 0;
	ui = INFINITY;
	for( j = 1; j <= n; j++ )
	{
	    if( cost[i][j] < ui )
	    {
		ui = cost[i][j];
		j0 = j;
	    }
	}
	ys[i] = ui;
	if( cap[j0] > 0 )
	{
	    cap[j0]--;
	    nexts[i] = row[j0];
	    row[j0] = i;
	    col[i] = j0;
	    yt[j0] = 0;
	}
    }
    /*
     *  calculate yt
     */
    if( m < neff )
    {
	for( j = 1; j <= n; j++ )
	    yt[j] = 0;
    }
    else
    {
	for( j = 1; j <= n; j++ )
	{
	    if( row[j] == 0 )
	    {
		ui = INFINITY;
		for( i = 1; i <= m; i++ )
		{
		    if( (cij = cost[i][j] - ys[i]) < ui )
			ui = cij;
		}
		yt[j] = ui;
	    }
	}
    }
    /*
     *  make admissible assignments
     */
    for( i = 1; i <=m; i++ )
    {
	if( col[i] == 0 )
	{
	    ui = ys[i];
	    for( j = 1; j <=n; j++ )
	    {
		if( cap[j] > 0 )
		{
		    if( cost[i][j] == ui + yt[j] )
		    {
			cap[j]--;
			nexts[i] = row[j];
			row[j] = i;
			col[i] = j;
			goto DONE;
		    }
		}
	    }
	}
DONE :
	continue;
    }
}

     
static void
S_shortestpath( ys, yt, dplus, dminus, back, next, i, indexp, dp, m, n )
int ys[],
    yt[],
    dplus[],
    dminus[],
    back[],
    next[];
int *indexp;
int *dp;
int i,
    m,
    n;
/*
 *  Finds a shortest path from starting node i to sink node index
 *  using dijkstra's algorithm and modified costs.
 *  next is a cyclic list of all sink nodes.
 *  next[headcand..lastcand] are the active candidate nodes.
 *  back points back in the path.
 */
{
int lastcand,
    headcand,
    v,
    vgl,
    j,
    dalt,
    ysi;

    for( j = 1; j <= n; j++ )
	dminus[j] = INFINITY;
    for( j = 1; j <= m; j++ )
	dplus[j] = INFINITY;

    lastcand = headcand = n;
    *dp = 0;

    /* till sink node with free capacity is reached */
    for( ;; )
    {
	dplus[i] = dalt = *dp;
	*dp = INFINITY;
	v = headcand;
	ysi = ys[i];

	/* for all active and feasible sink nodes */
	for( ;; )
	{
	    j = next[v];
	    if( cap[j] > 0 || row[j] > 0 )
	    {
		/*  update distance  */
		vgl = dalt + cost[i][j] - ysi - yt[j];
		if( vgl < dminus[j] )
		{
		    dminus[j] = vgl;
		    back[j] = i;
		}
		if( dminus[j] < *dp )
		{
		    *dp = dminus[j];
		    *indexp = v;
		}
	    }
	    v = j;
	    if( v == lastcand ) break;
	}

	/*
	 *  remove next[v] from candidate set
	 */
	v = *indexp;
	*indexp = next[v];
	if( *indexp == lastcand )
	    lastcand = v;
	else
	{
	    next[v] = next[*indexp];
	    next[*indexp] = next[headcand];
	    next[headcand] = *indexp;
	    headcand = *indexp;
	}
	i = row[*indexp];
        if( cap[*indexp] > 0 ) break;
    }
}


static void
S_augmentation( back, u, ind )
int back[];
int u,
    ind;
/*
 *  tracing back the augmenting path from index back to u,
 *  assignments are updated accordingly.
 */
{
int oldind;
int w = 0;

    (cap[ind])--;

    for( ; w != u; )
    {
	oldind = ind;

	w = back[ind];
	if( w != u )
	{
	    /* remove w, col[w] from list associated with col[w]  */
	    ind = col[w];
	    row[ind] = nexts[w];
	}

	/*  add ede w, index to list associated with index  */
	nexts[w] = row[oldind];
	row[oldind] = w;
	col[w] = oldind;
    }
}


static void
S_transformation( ys, yt, dplus, dminus, d, m, n )
int ys[],
    yt[],
    dplus[],
    dminus[];
int d;
/*
 *  update ys and yt
 */
{
int i;
    
    for( i = 1; i <= m; i++ )
    {
	if( dplus[i] != INFINITY )
	    ys[i] = ys[i] + d - dplus[i];
    }
    for( i = 1; i <= n; i++ )
    {
	if( dminus[i] < d )
	    yt[i] = yt[i] - d + dminus[i];
    }
}


S_printresults( m, n )
int m,
    n;
{
int i,
    j,
    sum = 0,
    max = 0,
    cij;

    for( i = 1; i <= m; i++ )
    {
	if( (cij = cost[i][col[i]]) > max )
	    max = cij;
	sum += cij;
    }
    /* if too high, exit */
    if( max >= PREASSIGN )
        printf("ERROR in pin-assignment\n");
    printf("sum = %d, max = %d\n", sum, max );
    for( j = 1; j <= n; j++ )
	printf("%d  %d\n", j, cap[j]);
    for( i = 1; i<= m; i++ )
    {
	for( j = 1; j <=n; j++ )
	{
	    if( col[i] == j )
		printf("[%d]", cost[i][j]);
	    else
		printf(" %d ", cost[i][j]);
	}
	printf("\n");
    }
}
