#include "mighty.h"

extern SCHED_PTR schedhead;

void S_makejump();

/*
 *  move a segment of a net by more than unit distance
 *  possibly over other nets
 */
int S_jumppush( dir, x, y, net )
int dir;
int x,
    y;
int net;
{
    LINKLISTPTR linklist;
    int jumpto;
    int x1,
	y1,
	x2,
	y2;
    int newx,
	newy;

    /*
     *  find interval to jump
     */
    x1 = x2 = x;
    y1 = y2 = y;
    for( linklist = linklayer[y][x]; linklist != 0;
	linklist = linklist->next )
    {
	if( linklist->link->net == net )
	{
	    x1 = MIN( x1, linklist->link->x1 );
	    x2 = MAX( x2, linklist->link->x2 );
	    y1 = MIN( y1, linklist->link->y1 );
	    y2 = MAX( y2, linklist->link->y2 );
	}
    }
    /*
     *  If wrongdir, there may be no link on linklayer[y][x]
     *  Solution :
     *     if( x1 == x2 && y1 == y2 ) return NO
     *     if( x1 == x2 ) , NO jump UP/DOWN or sidejump LEFT/RIGHT
     *     if( y1 == y2 ) , NO sidejump UP/DOWN or jump LEFT/RIGHT
     */
    if( x1 == x2 && y1 == y2 ) return ( NO );

    if( dir == UP || dir == DOWN )
    {
	if( x1 == x2 ) goto SIDEJP1;
    	/*  No pins on the link  */
	for( newx = x1; newx <= x2; newx++ )
	{
	    /* Check playerh only, since we do not jump if wrongdir */
	    if( playerh[y][newx] == net )
	   	goto SIDEJP1;
	}

	if( S_wrongdir( x, y, net ) || S_wrongdir( x1, y, net ) ||
	    S_wrongdir( x2, y, net ) )
	    return( NO );
        if( (jumpto = S_canjump( dir, x, y, net, x1, x2 )) > 0 )
        {
            S_makejump( dir, x, y, net, x1, x2, jumpto );
	    return( YES );
        }
SIDEJP1:
	if( y1 == y2 ) return( NO );
	/*  No pins on the link  */
	for( newy = y1; newy <= y2; newy++ )
	{
	    /* Check playerv only, since we do not jump if wrongdir */
	    if( playerv[newy][x] == net )
        return( NO );
	}

	if( S_wrongdir( x, y1, net ) || S_wrongdir( x, y2, net ) )
	    return( NO );
        if( (jumpto = S_cansidejump( dir, x, y, net, y1, y2 )) > 0 )
	{
	    if( jumpto > x )
                S_makejump( RIGHT, x, y, net, y1, y2, jumpto );
	    else
                S_makejump( LEFT, x, y, net, y1, y2, jumpto );
	    return( YES );
	}
        return( NO );
    }
    else /* LEFT/RIGHT */
    {
	if( y1 == y2 ) goto SIDEJP2;
	/*  No pins on the link  */
	for( newy = y1; newy <= y2; newy++ )
	{
	    /* Check playerv only, since we do not jump if wrongdir */
	    if( playerv[newy][x] == net )
		goto SIDEJP2;
	}

	if( S_wrongdir( x, y, net ) || S_wrongdir( x, y1, net ) ||
	    S_wrongdir( x, y2, net ) )
	    return( NO );
        if( (jumpto = S_canjump( dir, x, y, net, y1, y2 )) > 0 )
        {
            S_makejump( dir, x, y, net, y1, y2, jumpto );
	    return( YES );
        }
SIDEJP2:
	if( x1 == x2 ) return( NO );
	/*  No pins on the link  */
	for( newx = x1; newx <= x2; newx++ )
	{
	    /* Check playerh only, since we do not jump if wrongdir */
	    if( playerh[y][newx] == net )
        return( NO );
	}

	if( S_wrongdir( x1, y, net ) || S_wrongdir( x2, y, net ) )
	    return( NO );
        if( (jumpto = S_cansidejump( dir, x, y, net, x1, x2 )) > 0 )
	{
	    if( jumpto > y )
                S_makejump( DOWN, x, y, net, x1, x2, jumpto );
	    else
                S_makejump( UP, x, y, net, x1, x2, jumpto );
	    return( YES );
	}
        return( NO );
    }
}

/*
 *  Jump a segment, if there is a space
 *
 *        |     |			|     |
 *        |     |			----------
 *        |     |			         |
 *        ----------			      ^
 *		^  |			      |
 *		|			      |
 */
S_canjump( dir, x, y, net, low, high )
int dir;
int x,
    y;
int net;
int low,
    high; /* interval to jump */
{
    int tmpx,
	tmpy,
	jumpto;
    int pass; 		/* skip a row/column or not */
    int fitness,
        oldfitness;
    int S_linkgroup();

    if( dir == UP || dir == DOWN )
    {
	if( dir == DOWN )
	{
	    /*
	     *  see if there is a available row
	     */
	    oldfitness = 0;
	    jumpto = 0;
	    for( tmpy = y+1; tmpy <= num_rows; tmpy++ )
	    {
		if( vlayer[tmpy][low] != 0 && vlayer[tmpy][low] != net )
		    break; /* zzz recursion later */
		else if( vlayer[tmpy][high] != 0 && vlayer[tmpy][high] != net )
		    break;
		else if( vlayer[tmpy][x] != 0 && vlayer[tmpy][x] != net )
		    break;

		/*
		 *  don't block other pins
		 */
		pass = NO;
		for( tmpx = 1; tmpx <= num_cols; tmpx++ )
		{
		    if( playerh[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerh[tmpy][tmpx] ) == 0)
		    {
			pass = YES;
			break;
		    }
		}
		if( pass == YES )
	    continue;

		fitness = 0;
	        for( tmpx = low; tmpx <= high; tmpx++ )
		{
		    if( hlayer[tmpy][tmpx] == net )
			fitness++;
		    else if( hlayer[tmpy][tmpx] != 0 )
			break;
		}
		/* found - if same fitness go farther */
		if( tmpx > high && fitness >= oldfitness )
		{
		    jumpto = tmpy;
		    oldfitness = fitness;
		}
	    }
	    return( jumpto );
	}
	else  /* dir == UP */
	{
	    /*
	     *  see if there is a available row
	     */
	    oldfitness = 0;
	    jumpto = 0;
	    for( tmpy = y-1; tmpy >= 1; tmpy-- )
	    {
		if( vlayer[tmpy][low] != 0 && vlayer[tmpy][low] != net )
		    break; /* zzz recursion later */
		else if( vlayer[tmpy][high] != 0 && vlayer[tmpy][high] != net )
		    break;
		else if( vlayer[tmpy][x] != 0 && vlayer[tmpy][x] != net )
		    break;
		/*
		 *  don't block other pins
		 */
		pass = NO;
		for( tmpx = 1; tmpx <= num_cols; tmpx++ )
		{
		    if( playerh[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerh[tmpy][tmpx] ) == 0)
		    {
			pass = YES;
			break;
		    }
		}
		if( pass == YES )
	    continue;

		fitness = 0;
	        for( tmpx = low; tmpx <= high; tmpx++ )
		{
		    if( hlayer[tmpy][tmpx] == net )
			fitness++;
		    else if( hlayer[tmpy][tmpx] != 0 )
			break;
		}
		/* found - if same fitness go farther */
		if( tmpx > high && fitness >= oldfitness )
		{
		    jumpto = tmpy;
		    oldfitness = fitness;
		}
	    }
/*
if( jumpto > 0 )
{
print_debug ("canjump", "hlayer", hlayer, num_rows, num_cols, num_nets);
print_debug ("canjump", "vlayer", vlayer, num_rows, num_cols, num_nets);
}
*/
	    return( jumpto );
	}
    }
    else  /* LEFT or RIGHT */
    {
	if( dir == RIGHT )
	{
	    /*
	     *  see if there is a available row
	     */
	    oldfitness = 0;
	    jumpto = 0;
	    for( tmpx = x+1; tmpx <= num_cols; tmpx++ )
	    {
		if( hlayer[low][tmpx] != 0 && hlayer[low][tmpx] != net )
		    break; /* zzz recursion later */
		else if( hlayer[high][tmpx] != 0 && hlayer[high][tmpx] != net )
		    break;
		else if( hlayer[y][tmpx] != 0 && hlayer[y][tmpx] != net )
		    break;

		/*
		 *  don't block other pins
		 */
		pass = NO;
		for( tmpy = 1; tmpy <= num_rows; tmpy++ )
		{
		    if( playerv[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerv[tmpy][tmpx] ) == 0)
		    {
			pass = YES;
			break;
		    }
		}
		if( pass == YES )
	    continue;

		fitness = 0;
	        for( tmpy = low; tmpy <= high; tmpy++ )
		{
		    if( vlayer[tmpy][tmpx] == net )
			fitness++;
		    else if( vlayer[tmpy][tmpx] != 0 )
			break;
		}
		/* found - if same fitness go farther */
		if( tmpy > high && fitness >= oldfitness )
		{
		    jumpto = tmpx;
		    oldfitness = fitness;
		}
	    }
	    return( jumpto );
	}
	else  /* dir == LEFT */
	{
	    /*
	     *  see if there is a available col
	     */
	    oldfitness = 0;
	    jumpto = 0;
	    for( tmpx = x-1; tmpx >= 1; tmpx-- )
	    {
		if( hlayer[low][tmpx] != 0 && hlayer[low][tmpx] != net )
		    break; /* zzz recursion later */
		else if( hlayer[high][tmpx] != 0 && hlayer[high][tmpx] != net )
		    break;
		else if( hlayer[y][tmpx] != 0 && hlayer[y][tmpx] != net )
		    break;
		/*
		 *  don't block other pins
		 */
		pass = NO;
		for( tmpy = 1; tmpy <= num_rows; tmpy++ )
		{
		    if( playerv[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerv[tmpy][tmpx] ) == 0)
		    {
			pass = YES;
			break;
		    }
		}
		if( pass == YES )
	    continue;

		fitness = 0;
	        for( tmpy = low; tmpy <= high; tmpy++ )
		{
		    if( vlayer[tmpy][tmpx] == net )
			fitness++;
		    else if( vlayer[tmpy][tmpx] != 0 )
			break;
		}
		/* found - if same fitness go farther */
		if( tmpy > high && fitness >= oldfitness )
		{
		    jumpto = tmpx;
		    oldfitness = fitness;
		}
	    }
	    return( jumpto );
	}
    }
}

/*
 *  make a real jump
 */
static void
S_makejump( dir, x, y, net, low, high, jumpto )
int dir,
    x,
    y,
    net;
int low,
    high,
    jumpto;
{
    LINKPTR link;
    int tmpx,
        tmpy;

    while( (link = net_array[net].tmplist) != 0 )
    {
        net_array[net].tmplist = link->netnext;
        S_retlink( link );
    }

    if( dir == DOWN || dir == UP )
    {
	S_jumplinks( dir, x, y, jumpto, net, low, high );
	if( x != low )
	    S_jumplinks( dir, low, y, jumpto, net, low, high );
	if( x != high )
	    S_jumplinks( dir, high, y, jumpto, net, low, high );
	/*
	 *  clear the channel
	 */
	for( tmpx = low; tmpx <= high; tmpx++ )
	{
	    if( hlayer[y][tmpx] == net )
	        hlayer[y][tmpx] = 0;
	    if( vlayer[y][tmpx] == net )
	        vlayer[y][tmpx] = 0;
	}
	for( tmpy = MIN( y, jumpto ); tmpy <= MAX( y, jumpto ); tmpy++ )
	{
	    vlayer[tmpy][x] = 0;
	    vlayer[tmpy][low] = 0;
	    vlayer[tmpy][high] = 0;
	}
    }
    else /* dir == LEFT || dir == RIGHT */
    {
	S_jumplinks( dir, x, y, jumpto, net, low, high );
	if( y != low )
	    S_jumplinks( dir, x, low, jumpto, net, low, high );
	if( y != high )
	    S_jumplinks( dir, x, high, jumpto, net, low, high );
	/*
	 *  clear the channel
	 */
	for( tmpy = low; tmpy <= high; tmpy++ )
	{
	    if( hlayer[tmpy][x] == net )
	        hlayer[tmpy][x] = 0;
	    if( vlayer[tmpy][x] == net )
	        vlayer[tmpy][x] = 0;
	}
	for( tmpx = MIN( x, jumpto ); tmpx <= MAX( x, jumpto ); tmpx++ )
	{
	    hlayer[y][tmpx] = 0;
	    hlayer[low][tmpx] = 0;
	    hlayer[high][tmpx] = 0;
	}
    }
    for( link = net_array[net].path; link != 0; link = link->netnext )
	S_writelink( link, net );

    if( S_trytmppath( net ) == NO )
    {
	S_dumppath( net );
	exit( 6 );
    }

    /*  clean up links  */
    S_adjustlink( net );

    /*
     *  Set marked
     */
    net_array[net].marked = YES;
    /*
     *  reschedule is done in S_improve
     *  LATER check any severe cost increase
     */
    if (S_findcomponents( net ) > 1 )
    {
        if( S_findpath( net, YES ) == NO )
	{
	    printf("findpath3 not possible - must reroute\n");
	}
    }
}

/*
 *   not complete, links may overlap
 */
S_jumplinks( dir, x, y, jumpto, net, low, high )
int dir,
    x,
    y,
    jumpto,
    net;
int low,
    high;
{
    LINKLISTPTR linklist;
    LINKPTR link;
    int mark = NO;

    if( dir == UP || dir == DOWN )
    {
        linklist = linklayer[y][x];
        while( linklist != 0 )
        {
	    if( linklist->link->net != net )
	    {
	        linklist = linklist->next;
    	continue;
	    }
	    else
	        link = linklist->link;
	    if( link->layer == HORIZ )  /* horiz */
	    {
		if( low <= link->x1 && link->x2 <= high )
	            S_addlink( link->layer, link->x1, jumpto,
				 link->x2, jumpto, net );
		else
		{
	            S_addlink( link->layer, link->x1, link->y1,
				 link->x2, link->y2, net );
	            mark = YES;
		}
	    }
	    else if( link->y1 == y && link->y2 != jumpto )
	        S_addlink( link->layer, link->x1, jumpto, link->x2, link->y2, net );
	    else if( link->y2 == y && link->y1 != jumpto )
	        S_addlink( link->layer, link->x1, link->y1, link->x2, jumpto, net );
	    linklist = linklist->next;
	    S_dellink( link );
        }
	if( mark == YES || playerv[y][x] == net )
	    S_addlink( VERT, x, y, x, jumpto, net );
    }
    else /* dir == LEFT || dir == RIGHT */
    {
        linklist = linklayer[y][x];
        while( linklist != 0 )
        {
	    if( linklist->link->net != net )
	    {
	        linklist = linklist->next;
    	continue;
	    }
	    else
	        link = linklist->link;
	    if( link->layer == VERT )  /* VERT */
	    {
		if( low <= link->y1 && link->y2 <= high )
	            S_addlink( link->layer, jumpto, link->y1, 
				 jumpto, link->y2, net );
		else
		{
	            S_addlink( link->layer, link->x1, link->y1,
				 link->x2, link->y2, net );
	            mark = YES;
		}
	    }
	    else if( link->x1 == x && link->x2 != jumpto )
	        S_addlink( link->layer, jumpto, link->y1, link->x2, link->y2, net );
	    else if( link->x2 == x && link->x1 != jumpto )
	        S_addlink( link->layer, link->x1, link->y1, jumpto, link->y2, net );
	    linklist = linklist->next;
	    S_dellink( link );
        }
	if( mark == YES || playerh[y][x] == net )
	    S_addlink( HORIZ, x, y, jumpto, y, net );
    }
}


/*
 *      ------                    -------------
 *           |                                |
 *           ------------                     ------
 *           ^                         ^
 *           |			       |
 */
S_cansidejump( dir, x, y, net, low, high )
int dir;
int x,
    y;
int net;
{
    int tmpx,
	tmpy,
	jumpto;
    int pass;
    int fitness,
        oldfitness;
    int S_linkgroup();

    if( dir == LEFT || dir == RIGHT )
    {
	/*
	 *  see if there is a available row
	 */
	oldfitness = - INFINITY;
	jumpto = 0;
	for( tmpy = y+1; tmpy <= num_rows; tmpy++ )
	{
	    if( vlayer[tmpy][low] != 0 && vlayer[tmpy][low] != net )
		break; /* zzz recursion later */
	    else if( vlayer[tmpy][high] != 0 && vlayer[tmpy][high] != net )
		break;

	    /*
	     *  don't block other pins
	     */
	    pass = NO;
	    for( tmpx = 1; tmpx <= num_cols; tmpx++ )
	    {
		if( playerh[tmpy][tmpx] > 0 &&
	          S_linkgroup( tmpx, tmpy, playerh[tmpy][tmpx] ) == 0)
		{
	  	    pass = YES;
		    break;
		}
	    }
	    if( pass == YES )
	continue;

	    fitness = y - tmpy;  /* to go less */
	    for( tmpx = low; tmpx <= high; tmpx++ )
	    {
		if( hlayer[tmpy][tmpx] == net )
		    fitness += 10;
		else if( hlayer[tmpy][tmpx] != 0 )
		    break;
	    }
	    /* found - if same fitness go less */
	    if( tmpx > high && fitness > oldfitness )
	    {
		jumpto = tmpy;
		oldfitness = fitness;
	    }
	}
	for( tmpy = y-1; tmpy >= 1; tmpy-- )
	{
	    if( vlayer[tmpy][low] != 0 && vlayer[tmpy][low] != net )
		break; /* zzz recursion later */
	    else if( vlayer[tmpy][high] != 0 && vlayer[tmpy][high] != net )
		break;

	    /*
	     *  don't block other pins
	     */
	    pass = NO;
	    for( tmpx = 1; tmpx <= num_cols; tmpx++ )
	    {
		if( playerh[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerh[tmpy][tmpx] ) == 0)
		{
	  	    pass = YES;
		    break;
		}
	    }
	    if( pass == YES )
	continue;

	    fitness = tmpy - y;  /* to go less */
	    for( tmpx = low; tmpx <= high; tmpx++ )
	    {
		if( hlayer[tmpy][tmpx] == net )
		    fitness += 10;
		else if( hlayer[tmpy][tmpx] != 0 )
		    break;
	    }
	    /* found - if same fitness go less */
	    if( tmpx > high && fitness > oldfitness )
	    {
		jumpto = tmpy;
		oldfitness = fitness;
	    }
	}
	return( jumpto );
    }
    else /* dir == UP || dir == DOWN */
    {
	/*
	 *  see if there is a available row
	 */
	oldfitness = - INFINITY;
	jumpto = 0;
	for( tmpx = x+1; tmpx <= num_cols; tmpx++ )
	{
	    if( hlayer[low][tmpx] != 0 && hlayer[low][tmpx] != net )
		break; /* zzz recursion later */
	    else if( hlayer[high][tmpx] != 0 && hlayer[high][tmpx] != net )
		break;

	    /*
	     *  don't block other pins
	     */
	    pass = NO;
	    for( tmpy = 1; tmpy <= num_rows; tmpy++ )
	    {
		if( playerv[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerv[tmpy][tmpx] ) == 0)
		{
	  	    pass = YES;
		    break;
		}
	    }
	    if( pass == YES )
	continue;

	    fitness = x - tmpx;  /* to go less */
	    for( tmpy = low; tmpy <= high; tmpy++ )
	    {
		if( vlayer[tmpy][tmpx] == net )
		    fitness += 10;
		else if( vlayer[tmpy][tmpx] != 0 )
		    break;
	    }
	    /* found - if same fitness go less */
	    if( tmpy > high && fitness > oldfitness )
	    {
		jumpto = tmpx;
		oldfitness = fitness;
	    }
	}
	for( tmpx = x-1; tmpx >= 1; tmpx-- )
	{
	    if( hlayer[low][tmpx] != 0 && hlayer[low][tmpx] != net )
		break; /* zzz recursion later */
	    else if( hlayer[high][tmpx] != 0 && hlayer[high][tmpx] != net )
		break;

	    /*
	     *  don't block other pins
	     */
	    pass = NO;
	    for( tmpy = 1; tmpy <= num_rows; tmpy++ )
	    {
		if( playerv[tmpy][tmpx] > 0 &&
	         S_linkgroup( tmpx, tmpy, playerv[tmpy][tmpx] ) == 0)
		{
	  	    pass = YES;
		    break;
		}
	    }
	    if( pass == YES )
	continue;

	    fitness = tmpx - x;  /* to go less */
	    for( tmpy = low; tmpy <= high; tmpy++ )
	    {
		if( vlayer[tmpy][tmpx] == net )
		    fitness += 10;
		else if( vlayer[tmpy][tmpx] != 0 )
		    break;
	    }
	    /* found - if same fitness go less */
	    if( tmpy > high && fitness > oldfitness )
	    {
		jumpto = tmpx;
		oldfitness = fitness;
	    }
	}
	return( jumpto );
    }
}

/*
 *  return YES, running vetically on horiz_layer or vice versa
 */
int S_wrongdir( x, y, net )
int x,
    y,
    net;
{
    LINKLISTPTR linklist;
    LINKPTR tmplink;

    for( linklist = linklayer[y][x]; linklist != 0;
	linklist = linklist->next )
    {
	if( linklist->link->net == net )
	{
	    tmplink = linklist->link;
	    if( tmplink->layer == HORIZ && tmplink->y1 != tmplink->y2 )
		return( YES );
	    if( tmplink->layer == VERT && tmplink->x1 != tmplink->x2 )
		return( YES );
        }
    }
    return( NO );
}
