#include "mighty.h"
/*
int Insidepin = FALSE;
*/
int xshift;
int yshift;

/*******************************************************************
 *
 *  This routine reads the input for swrt, and returns an array of
 *  NETs.  The array contains all of the net data structures.
 *  xmin is shifted to x = 1, ymin is shifted to y = 1.
 *  The input is returned to
 *  the main routine in the form of arrays of pointers to NET structures.
 *
 *  Input should be in the following format:????????????????????????
 *   channeldef
 *   number_of_nets #
 *   rectagoncorners #
 *   number_of_pins #
 *   obstacles #
 *   left_list #
 *   right_list #
 *   sensitive_nets #
 *
 *
 *  Last modified: 8-6-86  for input error checking in S_onboundary
 *
 ********************************************************************/

NETPTR S_input ( num_nets, num_box, num_rows, num_cols, num_obstacles, num_pins )

int *num_nets,		/* total number of nets */
    *num_box,           /* number of corners of box */
    *num_rows,		/* size of bounding box in y */
    *num_cols,		/* size of bounding box in x */
    *num_obstacles,	/* total number of obstacles */
    *num_pins;		/* total number of pins */

{
    NETPTR net_array;		/* used to allocate space for net data
				   structure */
    LINKPTR link,
	    slink;
    PINPTR pinptr;
    PINPTR newpin;
    int netname;
    int netnum;			/* used to translate input net name to a
				   pointer to the net */
    int i;			/* loop counter */
    char input_string[15];	/* used to parse input */
    int layer;
    int dir;
    int x1,
	y1,
	x2,
	y2;
    int xmin,
	ymin,
	xmax,
	ymax;
    int leftcol,
	rightcol;
    int snet;		/* sensitive net */
    int sen_nets = 0;	/* number of sensitive nets */
    int *S_allocatevector();
    LINKPTR S_getlink();
    PINPTR S_getpin();
    int S_onboundary();
    int S_onsmpboundary();
    NETPTR allocate_nets ();

    /*
     *  initialize variables in case some are not read in
     */

    *num_nets = 0;
    *num_box = 0;
    *num_obstacles = 0;
    *num_pins = 0;
    obstacle = 0;
    bounding_box = 0;
    num_left = num_right = 0;
    left = right = 0;

    /*
     *  read input
     */
    {
	while (fscanf (input_file, "%s", input_string) != EOF)
	{
	    if (strcmp (input_string, "number_of_nets") == 0)
	    {
		if (*num_nets == 0)
		{
		    (void) fscanf (input_file, "%d", num_nets);
		    net_array = allocate_nets (*num_nets + 2);
		    init_name_translation (*num_nets);
		}
		else
		    fprintf (output_file,
		  "number_of_nets already given, first number used.\n");
	    }
	    else if (strcmp (input_string, "sensitive_nets") == 0)
	    {
		if (sen_nets == 0)
		{
		    (void) fscanf (input_file, "%d", &sen_nets);
		    for( i = 1; i <= sen_nets; i++ )
		    {
		        (void) fscanf (input_file, "%d", &netname);
			snet = name_to_number( netname );
			net_array[snet].sens = YES;
		    }
		}
		else
		    fprintf (output_file,
		  "sensitive nets already given, first list used.\n");
	    }
	    else if (strcmp (input_string, "left_list") == 0)
	    {
		if (num_left == 0)
		{
		    (void) fscanf (input_file, "%d", &num_left);
		    left = S_allocatevector(num_left);
		    for( i = 1; i <= num_left; i++ )
		    {
		        (void) fscanf (input_file, "%d", &netname);
			left[i] = name_to_number( netname );
			net_array[left[i]].left = YES;
		    }
		}
		else
		    fprintf (output_file,
		  "left_list already given, first list used.\n");
	    }
	    else if (strcmp (input_string, "right_list") == 0)
	    {
		if (num_right == 0)
		{
		    (void) fscanf (input_file, "%d", &num_right);
		    right = S_allocatevector(num_right);
		    for( i = 1; i <= num_right; i++ )
		    {
		        (void) fscanf (input_file, "%d", &netname);
			right[i] = name_to_number( netname );
			net_array[right[i]].right = YES;
		    }
		}
		else
		    fprintf (output_file,
		  "right_list already given, first list used.\n");
	    }
	    else if (strcmp (input_string, "rectagoncorners") == 0)
	    {
		if (*num_box == 0)
		{
		    (void) fscanf (input_file, "%d", num_box);
		    xmin = INFINITY;
		    ymin = INFINITY;
		    xmax = - INFINITY;
		    ymax = - INFINITY;
		    /*
		     *  save boundary of the routing area as linls
		     */
		    for( i = 1; i <= *num_box; i++ )
		    {
		        (void) fscanf (input_file, "%d", &x2);
		        (void) fscanf (input_file, "%d", &y2);
			xmin = MIN( xmin, x2 );
			ymin = MIN( ymin, y2 );
			xmax = MAX( xmax, x2 );
			ymax = MAX( ymax, y2 );
			link = S_getlink();
			if( bounding_box != 0 )
			{
			    link->x1 = bounding_box->x2;
			    link->y1 = bounding_box->y2;
			}
			else
			    slink = link;

			link->netnext = bounding_box;
		        bounding_box = link;
			link->x2 = x2;
			link->y2 = y2;
		    }
		    slink->x1 = bounding_box->x2;
		    slink->y1 = bounding_box->y2;

		    /*
		     *  To shift (xmin, ymin) to (1, 1)
		     */
		    xshift = xmin - 1;
		    yshift = ymin - 1;
		    *num_cols = xmax - xshift;
		    *num_rows = ymax - yshift;
		}
		else
		    fprintf (output_file,
		  "bounding_box already specified, first list used.\n");
	    }
	    else if (strcmp (input_string, "number_of_pins") == 0)
	    {
		if (*num_pins == 0)
		{
		    (void) fscanf (input_file, "%d", num_pins);
		    for( i = 1; i <= *num_pins; i++ )
		    {
		        (void) fscanf (input_file, "%d", &netname);
		        (void) fscanf (input_file, "%d", &x1);
		        (void) fscanf (input_file, "%d", &y1);
		        (void) fscanf (input_file, "%d", &layer);
			netnum = name_to_number( netname );
			pinptr = S_getpin();
			pinptr->type = PERM;
			pinptr->next = net_array[netnum].pin;
			net_array[netnum].pin = pinptr;
			pinptr->layer = layer;
			pinptr->x = x1;
			pinptr->y = y1;
		    }
		}
		else
		    fprintf (output_file,
			 "pins already specified, first list used.\n");
	    }
	    else if (strcmp (input_string, "obstacles") == 0)
	    {
		if (*num_obstacles == 0)
		{
		    obstacle = 0;
		    (void) fscanf (input_file, "%d", num_obstacles);
		    for( i = 1; i <= *num_obstacles; i++ )
		    {
			link = S_getlink();
			link->netnext = obstacle;
		        obstacle = link;
		        (void) fscanf (input_file, "%d", &x1);
		        (void) fscanf (input_file, "%d", &y1);
		        (void) fscanf (input_file, "%d", &x2);
		        (void) fscanf (input_file, "%d", &y2);
		        (void) fscanf (input_file, "%d", &layer);
			if( x1 != x2 && y1 != y2 )
			{
			    printf("ERROR in obstacles\n");
			    fprintf(output_file, "failed\n");
			    exit( 3 );
			}
			link->layer = layer;
			link->x1 = MIN( x1, x2 );
			link->y1 = MIN( y1, y2 );
			link->x2 = MAX( x1, x2 );
			link->y2 = MAX( y1, y2 );
		    }
		}
		else
		    fprintf (output_file,
			 "obstacles already specified, first list used.\n");
	    }
	}
    }

    /*
     *  Shift xmin to 1 and ymin to 1.
     */
    for( i = 1; i <= *num_nets; i++ )
    {
	for( pinptr = net_array[i].pin; pinptr != 0; pinptr = pinptr->next )
	{
	    pinptr->x -= xshift;
	    pinptr->y -= yshift;
	}
    }

    for( link = bounding_box; link != 0; link = link->netnext )
    {
	link->x1 -= xshift;
	link->x2 -= xshift;
	link->y1 -= yshift;
	link->y2 -= yshift;
    }

    for( link = obstacle; link != 0; link = link->netnext )
    {
	link->x1 -= xshift;
	link->x2 -= xshift;
	link->y1 -= yshift;
	link->y2 -= yshift;
    }

    /*
     *  if pin is on the boundary, move the pin
     *  one unit inside the routing area,
     *  and mark the direction of the movement
     */
    for( i = 1; i <= *num_nets; i++ )
    {
	/* To find bounds */
	if( net_array[i].left == YES )
	    leftcol = 1;
	else
	    leftcol = *num_cols;
	if( net_array[i].right == YES )
	    rightcol = *num_cols;
	else
	    rightcol = 1;

	for( pinptr = net_array[i].pin; pinptr != 0; pinptr = pinptr->next )
	{
	    x1 = pinptr->x;
	    y1 = pinptr->y;

	    if( (dir = S_onboundary( x1, y1 )) != 0 )
	    {
		switch( dir )
		{
		case DOWN : (pinptr->y)++;
		   break;
		case UP   : (pinptr->y)--;
		   break;
		case LEFT : (pinptr->x)--;
		   break;
		case RIGHT : (pinptr->x)++;
		   break;
		}

		newpin = S_getpin();
		newpin->type = PERM;
		newpin->next = net_array[i].tmppin;
		net_array[i].tmppin = newpin;
		newpin->layer = pinptr->layer;
	        newpin->x = x1;
		newpin->y = y1;
		newpin->group = dir;
	    }

	    leftcol = MIN( pinptr->x, leftcol );
	    rightcol = MAX( pinptr->x, rightcol );
	}
	if( leftcol <= rightcol )
	{
	    net_array[i].colleft = leftcol;
	    net_array[i].colright = rightcol;
	}
	else
	{
	    net_array[i].colleft = 0;
	    net_array[i].colright = 0;
	}
    }
    return (net_array);
}

int S_onboundary( px, py )
int px,
    py;
{
    LINKPTR link;
    int indir = 0;
    int outdir = 0;

    for( link = bounding_box; link != 0; link = link->netnext )
    {
	if( link->y1 == py && link->x1 < px && px < link->x2 )
	    return( DOWN );
	if( link->y1 == py && link->x2 < px && px < link->x1 )
	    return( UP );
	if( link->x1 == px && link->y1 < py && py < link->y2 )
	    return( LEFT );
	if( link->x1 == px && link->y2 < py && py < link->y1 )
	    return( RIGHT );
    }

    /*  Now check if there is a pin on a convex corner, for which
	routing is impossible  */
    for( link = bounding_box; link != 0; link = link->netnext )
    {
        if( link->y1 == py && link->x1 == px && px < link->x2 )
	    outdir = RIGHT;
        else if( link->y1 == py && link->x1 == px && py < link->y2 )
	    outdir = DOWN;
        else if( link->y1 == py && link->x1 == px && px > link->x2 )
    	    outdir = LEFT;
        else if( link->y1 == py && link->x1 == px && py > link->y2 )
	    outdir = UP;

        if( link->y2 == py && link->x2 == px && px > link->x1 )
	    indir = RIGHT;
        else if( link->y2 == py && link->x2 == px && py > link->y1 )
	    indir = DOWN;
        else if( link->y2 == py && link->x2 == px && px < link->x1 )
	    indir = LEFT;
        else if( link->y2 == py && link->x2 == px && py < link->y1 )
	    indir = UP;
    }

    if( indir == UP && outdir == RIGHT ||
	indir == RIGHT && outdir == DOWN ||
	indir == DOWN && outdir == LEFT ||
	indir == LEFT && outdir == UP )
    {
	/*
	 *  Report the erroneous pin and exit -
	 *  Note that the direction of links are reversed
	 */
	printf("ERROR: Isolated pin on (%d, %d)\n", px+xshift, py+yshift);
	fprintf(output_file, "failed\n");
	exit( 4 );
    }
    return( 0 );
}
