#include "mighty.h"
int Densityonly;

/******************************************************************
 *
 *  command_line processes all the information on the command line.
 *  Global flags are set from this, and input and output files are opened.
 *  If no input and output files are specified, stdin and stdout are
 *  assumed.
 *
 *
 *****************************************************************/

void command_line (argc, argv)
int argc;			/* argc and argv are used to get information
				   from the command line */
char *argv[];
{
    int i;			/* loop counter */

    /* set defaults */
    DEBUG = NO;
    debug = NO;
    input_file = stdin;
    result_file = stdin;
    output_file = stdout;
    ioform = HUMAN;
    outform = HUMAN;
    stats_only = NO;
    maximize_metal = YES;
    Densityonly = NO;

    for (i = 1; i < argc; i++)
    {
	if (strcmp (argv[i], "-H") == 0)
	    ioform = MACHINE;
	else if (strcmp (argv[i], "-O") == 0)
	{
	    outform = MACHINE;
	    if (i < argc-1) result_file = fopen(argv[++i], "w");
	    else outform = HUMAN;
	}
	else if (strcmp (argv[i], "-d") == 0)
	    debug = YES;
	else if (strcmp (argv[i], "-s") == 0)
	    stats_only = YES;
	else if (strcmp (argv[i], "-m") == 0)
	    maximize_metal = NO;
	else if (strcmp (argv[i], "-e") == 0)
	    Densityonly = YES;
/*
	else if (strcmp (argv[i], "-z") == 0)
	{
	    i++;
	    if (i < argc)
		sscanf (argv[i], "%d", &cut_column);
	}
	else if (strcmp (argv[i], "-c") == 0)
	{
	    i++;
	    if (i < argc)
		sscanf (argv[i], "%d", &initial_column);
	}
*/
	else
	{
	    if (ioform == HUMAN)
	    {
		input_file = fopen (argv[i], "r");
		if (i + 1 < argc)
		    output_file = fopen (argv[i + 1], "w");
		fprintf (output_file, "\nInput file: %s\n\n", argv[i]);
		return;
	    }
	}
    }
}


/*****************************************************
 *
 *  name_to_number and number_to_name handle the translation from the
 *  external (input) to the internal numbers used for the various nets
 *  being routed.  init_name_translation is called to set up the static
 *  variables declared below that are used in the translation.
 *  Since this name_to_number is called only once per pin on the channel
 *  it is not done in a very efficient way.  number_to_name is called
 *  (at least) once for every piece of metal or poly in the final routed
 *  channel, so it is very efficient.
 *  
 ******************************************************/

static int max_net_number;	/* largest net number used so far */

void init_name_translation (num_nets)
int num_nets;		/* number of nets that will be in the channel
				   */
{
    int i;			/* loop counter */
    int *S_allocatevector();

    /* allocate the name_array */
    name_array = S_allocatevector( num_nets );

    /* empty the array */
    for (i = 0; i <= num_nets; name_array[i++] = 0);

    /* set max_net_number */
    max_net_number = 0;
}

/* number_to_name has been changed to a macro.  It (and name_array) would
 * have to be moved to "mighty.h"
 */
/* int number_to_name (number)
int number;			* the net number whose name is desired *
{
    return (name_array[number]);
} */

int correct_num_nets ()
{
    return (max_net_number);
}

int name_to_number (name)
int name;			/* the external name of a net */
{
    int i;			/* loop counter */

/*     name_array[name] = name;
    return(name);
 */

    /* if "name" has already been assigned a number, return that number */
    for (i = 0; i <= max_net_number; i++)
	if (name_array[i] == name)
	    return (i);

    /* assign max_net_number+1 to be name */
    max_net_number++;
    name_array[max_net_number] = name;
    return (max_net_number);
}

