#include "mighty.h"

extern int xshift, yshift;

void initformat();
void writeline();

/*****************************************************************
 *
 *  print_channel prints out the contents of the horizontal and vertical
 *  route layers of the channel.  One column is included for each (left
 *  and right) edge.
 *
 *
 *  Date: 3-30-86
 *
 ****************************************************************/

void print_channel (channel, num_rows, num_cols, num_nets)
CHANNELPTR channel;		/* the channel to be printed */
int num_rows;		/* number of rows in the channel */
int num_cols;		/* number of columns in the channel */
int num_nets;		/* number of nets in the channel */
{
    register int i, j;		/* loop counters */
    register int num, l;	/* used to convert integers to characters */
    char temp[20];		/* the caracter representation of a number */

    {
	fprintf (output_file, "\nThe horizontal layer (turned sideways):\n");
	fprintf (output_file, "     ");
	fprintf (output_file, "\n");
	for (j = 1; j <= num_cols; j++)
	{
	    putc(' ', output_file);
	    for (i = num_rows; i >= 1; i--)
	    {
		num = channel->horiz_layer[i][j];

		if (num > num_nets)
		{
		    putc(' ',output_file);
		    putc('X',output_file);
		    putc(' ',output_file);
		}
		else
		{
		    num = number_to_name(num);

		    /* Generate digits (reverse order) */
		    l = 0; do temp[l++] = num % 10 + '0'; while ((num /= 10) > 0);

		    /* right justify the number in a two digit field */
		    if (l == 1)
		        putc(' ', output_file);
    
		    /* Copy them back in correct order */
		    do putc(temp[--l], output_file); while (l > 0);

		    /* insert a space to separate the numbers */
		    putc(' ', output_file);
	        }
	    }
	    fprintf (output_file, " col = %3d\n", j );
	}
	fprintf(output_file, "     ");
	fprintf(output_file, "\n");

	fprintf (output_file, "\nThe vertical layer (turned sideways):\n");
	fprintf (output_file, "     ");
	fprintf (output_file, "\n");
	for (j = 1; j <= num_cols; j++)
	{
	    putc(' ', output_file);
	    for (i = num_rows; i >= 1; i--)
	    {
		num = channel->vert_layer[i][j];
		if (num > num_nets)
		{
		    putc(' ',output_file);
		    putc('X',output_file);
		    putc(' ',output_file);
		}
		else
		{
		    num = number_to_name(num);

		    /* Generate digits (reverse order) */
		    l = 0; do temp[l++] = num % 10 + '0'; while ((num /= 10) > 0);

		    /* right justify the number in a 2 digit field */
		    if (l == 1)
		        putc(' ', output_file);

		    /* Copy them back in correct order */
		    do putc(temp[--l], output_file); while (l > 0);

		    /* insert a space to separate the numbers */
		    putc(' ', output_file);
	        }
	    }
	    fprintf (output_file, "  col = %3d\n", j );
	}
	fprintf(output_file, "     ");
	fprintf(output_file, "\n");
    }
}



/*****************************************************************
 *
 *  print_channel prints out the contents of the horizontal and vertical
 *  route layers of the channel.  One column is included for each (left
 *  and right) edge.
 *
 ****************************************************************/

void print_debug (msg, name, layer, num_rows, num_cols, num_nets)
char *msg;
char *name;
int **layer;		/* the layer to be printed */
int num_rows;		/* number of rows in the channel */
int num_cols;		/* number of columns in the channel */
int num_nets;		/* number of nets in the channel */
{
    register int i, j;		/* loop counters */
    register int num, l;	/* used to convert integers to characters */
    char temp[20];		/* the caracter representation of a number */

    {
	fprintf (output_file, "\nThe %s layer (%s):\n", name, msg);
	fprintf (output_file, "     ");
	fprintf (output_file, "\n");
	for (j = 0; j <= num_cols+1; j++)
	{
	    putc(' ', output_file);
	    for (i = num_rows; i >= 0; i--)
	    {
		num = layer[i][j];
		if (num < 0)
		{
		    num = -num;
		    if( num < 99 )
		        putc('-', output_file);
		    else
		    {
			num = num % 100;
		        putc('*', output_file);
		    }
		}
		else
		{
		    /* insert a space to separate the numbers */
		    putc(' ', output_file);
		}
		if( layer[i][j] > num_nets )
		{
		    putc(' ', output_file);
		    putc('X', output_file);
		}
		else
		{
		    /* Generate digits (reverse order) */
		    l = 0; do temp[l++] = num % 10 + '0'; while ((num /= 10) > 0);

		    /* right justify the number in a 2 digit field */
		    if (l == 1)
		        putc(' ', output_file);

		    /* Copy them back in correct order */
		    do putc(temp[--l], output_file); while (l > 0);

	        }
	    }
	    fprintf (output_file, "  col = %2d\n", j);
	}
	fprintf(output_file, "     ");
	fprintf(output_file, "\n");
    }
}



int pitch = 7*20;
int npitch = - 7*20;
/*
int npitch = 7*20;
*/
int instnum = 0;

void S_printsq()
/*
 *  print statistics - number of vias, total line length,
 *                     number of rows and columns.
 *  Mirror the whole layout upside down when print out
 */
{
    LINKPTR tlink;
    LINKLISTPTR list;
    int i;  /* counter */
    int x,
	y;
    int hlink,
        vlink; 		/* To mark the existence of the links */
    int nvia = 0;
    int numlink = 0;
    int width;


    /*
     *  Initial SQUID format
     */
    initformat();

    /*  instnum must be bigger than netnum  */
    instnum += num_nets;

    /*
     *  find the wire length
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tlink = net_array[i].path; tlink != NULL;
	     tlink = tlink->netnext )
	{
	    numlink++;
	}
    }
    /*
    fprintf(output_file, "SUBNET %d\n", numlink );
    */
    /*
     *  print the wires
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tlink = net_array[i].path; tlink != NULL;
	     tlink = tlink->netnext )
	{
	    /*
	    fprintf(output_file, "%d %d %d %d %d %d\n",
	     number_to_name(i), tlink->layer,
	     tlink->x1, tlink->y1, tlink->x2, tlink->y2 );
	     */
	if( tlink->layer == HORIZ )
	{
	    width = 60;
	    if( tlink->x1 == tlink->x2 )
	    {
		fprintf(output_file, "MK INST %d INST_NAME \"60by60mterm\" MASTER \"60by60mterm\" \"layout\" MAT 1 0 0 1 %d %d\n",
		++instnum, pitch*tlink->x1, npitch*tlink->y1 );
		/*
		fprintf(output_file, "MK TERM \"i\" NET_ID %d INST_ID %d\n",
		hlayer[tlink->y1][tlink->x1], instnum );
		*/

		fprintf(output_file, "MK INST %d INST_NAME \"60by60mterm\" MASTER \"60by60mterm\" \"layout\" MAT 1 0 0 1 %d %d\n",
		++instnum, pitch*tlink->x2, npitch*tlink->y2 );
		/*
		fprintf(output_file, "MK TERM \"i\" NET_ID %d INST_ID %d\n",
		hlayer[tlink->y2][tlink->x2], instnum );
		*/
	    }
	}
	else
	{
	    width = 40;
	    if( tlink->y1 == tlink->y2 )
	    {
		fprintf(output_file, "MK INST %d INST_NAME \"40by40pterm\" MASTER \"40by40pterm\" \"layout\" MAT 1 0 0 1 %d %d\n",
		++instnum, pitch*tlink->x1, npitch*tlink->y1 );
		/*
		fprintf(output_file, "MK TERM \"i\" NET_ID %d INST_ID %d\n",
		vlayer[tlink->y1][tlink->x1], instnum );
		*/

		fprintf(output_file, "MK INST %d INST_NAME \"40by40pterm\" MASTER \"40by40pterm\" \"layout\" MAT 1 0 0 1 %d %d\n",
		++instnum, pitch*tlink->x2, npitch*tlink->y2 );
		/*
		fprintf(output_file, "MK TERM \"i\" NET_ID %d INST_ID %d\n",
		vlayer[tlink->y2][tlink->x2], instnum );
		*/
	    }
	}
	writeline(pitch*tlink->x1, npitch*tlink->y1, pitch*tlink->x2,
		  npitch*tlink->y2, width, i);

	}
    }

    /*
     *  count the vias
     */
    for( y = 1; y <= num_rows; y++ )
    {
	for( x = 1; x <= num_cols; x++ )
	{
	    if( hlayer[y][x] == vlayer[y][x] && hlayer[y][x] != 0 &&
		hlayer[y][x] <= num_nets )
	    {
		hlink = vlink = NO;
		for( list = linklayer[y][x]; list != NULL; list = list->next )
		{
		    tlink = list->link;
		    if( tlink->layer == HORIZ )
			hlink = YES;
		    else
			vlink = YES;

		}

		/*
		 *  To include vias on the boundary
		 */
		if( playerh[y][x] > 0 )
		    hlink = YES;
		if( playerv[y][x] > 0 )
		    vlink = YES;

		if( hlink == YES && vlink == YES )
		{
		    nvia++;
		}
	    }
	}
    }
    /*
    fprintf(output_file, "VIA %d\n", nvia );
    */
    /*
     *  print the vias
     */
    for( y = 1; y <= num_rows; y++ )
    {
	for( x = 1; x <= num_cols; x++ )
	{
	    if( hlayer[y][x] == vlayer[y][x] && hlayer[y][x] != 0 &&
		hlayer[y][x] <= num_nets )
	    {
		hlink = vlink = NO;
		for( list = linklayer[y][x]; list != NULL; list = list->next )
		{
		    tlink = list->link;
		    if( tlink->layer == HORIZ )
			hlink = YES;
		    else
			vlink = YES;

		}

		/*
		 *  To include vias on the boundary
		 */
		if( playerh[y][x] > 0 )
		    hlink = YES;
		if( playerv[y][x] > 0 )
		    vlink = YES;

		if( hlink == YES && vlink == YES )
		{
		    nvia++;
		    /*
		    fprintf(output_file, "%d %d %d %d\n", hlayer[y][x],
					     1, x, y );
		    */
	      fprintf(output_file, "MK INST %d INST_NAME \"80by80pmc\" MASTER \"80by80pmc\" \"layout\" MAT 1 0 0 1 %d %d\n",
	      ++instnum, pitch*x, npitch*y);
	      /*
	      fprintf(output_file, "MK TERM \"i\" NET_ID %d INST_ID %d\n",
	      hlayer[y][x], instnum );
	      */
		}
	    }
	}
    }
    /*
    printf(" There are %d vias\n", nvia );
    */
}


static void
initformat()
/*
 *  write the initialization part in SQUID format
 */
{
    int i;

    fprintf(output_file, "SQUID\n");
    fprintf(output_file, "PUT VIEW \"%dby%dchannel\" \"layout\" \"w\" \"squidNextObjectID\" INT 777\n",num_rows, num_cols);
    for( i = num_nets; i >= 1; i-- )
    {
	fprintf(output_file, "MK NET %d NET_NAME \"\"\n", i );
    }
}



int numline = 0;

static void
writeline(x1, y1, x2, y2, width, net)
int x1, y1, x2, y2;
int width, net;
/*
 *  write lines in SQUID format
 */
{
    numline++;
    fprintf(output_file, "MK LINE %ld NET %d LAYER \"%s\" FILL WIDTH %d PATH %d %d %d %d\n",
        INFINITY - numline, net, (width == 40 ? "NP" : "NM"),
	width, x1, y1, x2, y2);
    fprintf(output_file, "PUT LINE %ld \"ishorizontal\" BOOL %s\n", INFINITY - numline,
	(y1 == y2 ? "TRUE" : "FALSE"));
}



void S_printstandard()
/*
 *  print statistics - number of vias, total line length,
 *                     number of rows and columns.
 *  Mirror the whole layout upside down when print out
 */
{
    LINKPTR tlink;
    LINKLISTPTR list;
    int i;  /* counter */
    int x,
	y;
    int hlink,
        vlink; 		/* To mark the existence of the links */
    int nvia = 0;
    int numlink = 0;


    /*
     *  Initial format
     */
    fprintf(output_file, "channelwiring\n" );

    /*
     *  count the vias
     */
    for( y = 1; y <= num_rows; y++ )
    {
	for( x = 1; x <= num_cols; x++ )
	{
	    if( hlayer[y][x] == vlayer[y][x] && hlayer[y][x] != 0 &&
		hlayer[y][x] <= num_nets )
	    {
		hlink = vlink = NO;
		for( list = linklayer[y][x]; list != NULL; list = list->next )
		{
		    tlink = list->link;
		    if( tlink->layer == HORIZ )
			hlink = YES;
		    else
			vlink = YES;

		}

		/*
		 *  To include vias on the boundary
		 */
		if( playerh[y][x] > 0 )
		    hlink = YES;
		if( playerv[y][x] > 0 )
		    vlink = YES;

		if( hlink == YES && vlink == YES )
		{
		    nvia++;
		}
	    }
	}
    }
    fprintf(output_file, "vias %d\n", nvia );
    /*
     *  print the vias
     */
    for( y = 1; y <= num_rows; y++ )
    {
	for( x = 1; x <= num_cols; x++ )
	{
	    if( hlayer[y][x] == vlayer[y][x] && hlayer[y][x] != 0 &&
		hlayer[y][x] <= num_nets )
	    {
		hlink = vlink = NO;
		for( list = linklayer[y][x]; list != NULL; list = list->next )
		{
		    tlink = list->link;
		    if( tlink->layer == HORIZ )
			hlink = YES;
		    else
			vlink = YES;

		}

		/*
		 *  To include vias on the boundary
		 */
		if( playerh[y][x] > 0 )
		    hlink = YES;
		if( playerv[y][x] > 0 )
		    vlink = YES;

		if( hlink == YES && vlink == YES )
		{
		    fprintf(output_file, "%d %d %d %d %d\n",
			number_to_name(hlayer[y][x]),
			x + xshift, y + yshift, 1, 2 );
		}
	    }
	}
    }

    /*
     *  count number of links
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tlink = net_array[i].path; tlink != NULL;
	     tlink = tlink->netnext )
	{
	    numlink++;
	}
    }
    fprintf(output_file, "wires %d\n", numlink );

    /*
     *  print the wires
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tlink = net_array[i].path; tlink != NULL;
	     tlink = tlink->netnext )
	{
	    fprintf(output_file, "%d %d %d %d %d %d\n",
	     number_to_name(i),
	     tlink->x1 + xshift, tlink->y1 + yshift,
	     tlink->x2 + xshift, tlink->y2 + yshift, tlink->layer );
	}
    }
    fprintf(output_file, ".end\n");
}

