#include "mighty.h"

extern int HCOST1;
extern int VCOST1;

extern int HCOST2;
extern int VCOST2;
extern int SWCOST;

extern int LIMCOST;
extern int MODCOST;

SEARCHPTR *searchq;

/*
 *  find min_cost path of net i
 *  return YES or NO
 */
int S_findpath( net, cleanup )
int net;
int cleanup;
{
    int	**hgp,     /* horiz_conn_components */
	**vgp;     /* vert_conn_components */
    SEARCHPTR tmpq;
    PINPTR tpin;
    LINKPTR tmplink;
    int i;         /* counter */
    int keycost,   /* cost to access searchq ( = cost % MODCOST) */
	cost;
    int group,
	layer;
    int x,
	y;

    hgp = channel->hcomp;
    vgp = channel->vcomp;

    /*
     * clear tmppath
     */
    while( (tmplink = net_array[net].tmplist) != 0 )
    {
	net_array[net].tmplist = tmplink->netnext;
	S_retlink( tmplink );
    }

    /*
     *  mark all pins
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tpin = net_array[i].pin; tpin != 0; tpin = tpin->next )
	{
	    if( tpin->layer == HORIZ )
		hlayer[tpin->y][tpin->x] = i;
	    if( tpin->layer == VERT )
		vlayer[tpin->y][tpin->x] = i;
	}
    }

/*
print_channel (channel, num_rows, num_cols, num_nets);
*/
    /*
     *  push current locations of routing
     */
    cost = 1;
    for( tpin = net_array[net].pin; tpin != 0; tpin = tpin->next )
    {
	if( tpin->layer == VERT )
	    S_searchnext(vlayer, hlayer, HCOST2, VCOST1, SWCOST,
			 vgp, hgp, tpin->x, tpin->y, tpin->group, cost);

	if( tpin->layer == HORIZ )
	    S_searchnext(hlayer, vlayer, HCOST1, VCOST2, SWCOST,
			 hgp, vgp, tpin->x, tpin->y, tpin->group, cost);
    }
    for( tmplink = net_array[net].path; tmplink != 0;
	 tmplink = tmplink->netnext )
    {
	if( tmplink->layer == HORIZ )
	{
	    if( tmplink->x1 == tmplink->x2 )
	        for( y = tmplink->y1; y <= tmplink->y2; y++ )
	            S_searchnext(hlayer, vlayer, HCOST1, VCOST2, SWCOST,
			hgp, vgp, tmplink->x1, y, tmplink->group, cost);
	    else
	        for( x = tmplink->x1; x <= tmplink->x2; x++ )
	            S_searchnext(hlayer, vlayer, HCOST1, VCOST2, SWCOST,
			hgp, vgp, x, tmplink->y1, tmplink->group, cost);
	}
	else
	{
	    if( tmplink->x1 == tmplink->x2 )
	        for( y = tmplink->y1; y <= tmplink->y2; y++ )
	            S_searchnext(vlayer, hlayer, HCOST2, VCOST1, SWCOST,
			vgp, hgp, tmplink->x1, y, tmplink->group, cost);
	    else
	        for( x = tmplink->x1; x <= tmplink->x2; x++ )
	            S_searchnext(vlayer, hlayer, HCOST2, VCOST1, SWCOST,
			vgp, hgp, x, tmplink->y1, tmplink->group, cost);
	}
    }

    /*
     *  find the shortest link which can reduce the number of
     *  connected components
     */
    while( cost <= LIMCOST )
    {
	    keycost = cost % MODCOST;
	    if (searchq[keycost] == 0)
	    {
		cost++;
	continue;
	    }
	    /*
	     *  pop the first job from searchq
	     */
	    tmpq = searchq[keycost];
	    x = tmpq->x;
	    y = tmpq->y;
	    group = tmpq->group;
	    layer = tmpq->layer;
	    searchq[keycost] = tmpq->next;
	    S_retsearchq( tmpq );
	    if( layer == HORIZ &&
	        S_searchnext(hlayer, vlayer, HCOST1, VCOST2, SWCOST,
		hgp, vgp, x, y, group, cost) == YES)
	break;
	    if( layer == VERT &&
	        S_searchnext(vlayer, hlayer, HCOST2, VCOST1, SWCOST,
		vgp, hgp, x, y, group, cost) == YES)
	break;
    }
    if( cost > LIMCOST )
    {
/*
if( DEBUG )
{
printf("path search failed at cost %d\n", cost );
print_debug ("search failed", "horiz", channel->horiz_layer, num_rows, num_cols, num_nets);
print_debug ("search failed", "vert", channel->vert_layer, num_rows, num_cols, num_nets);
print_debug ("search failed", "hcomp", channel->hcomp, num_rows, num_cols, num_nets);
print_debug ("search failed", "vcomp", channel->vcomp, num_rows, num_cols, num_nets);
}
*/
	net_array[net].cost = INFINITY;
	S_cleansearchq();
	if( cleanup == YES )
	    S_cleantrace();
	return( NO );
    }
    else
    {
	net_array[net].cost = cost;  /* zzz tmp */
	S_cleansearchq();
	S_backtrack( layer, x, y, group, net, cost );
	S_cleantrace();
        return( YES );  /* path found */
    }
}

/*
 *  mark current point and search the next point.
 */
S_searchnext(flayer, slayer, hcost, vcost, swcost, fgp, sgp, x, y, group, cost)
int **flayer,   /* primary layer */
    **slayer;   /* secondary layer */
int hcost,
    vcost,
    swcost;
int **fgp,      /* group of primary layer */
    **sgp;      /* group of secondary layer */
int x,
    y,
    group,
    cost;
{
    int newx,
        newy;
    int touchcost;

    if( flayer[y][x] != 0 && fgp[y][x] != 0 )
    {
	if( fgp[y][x] == group )
	{
	    if( flayer[y][x] >= - cost )
	        return( NO );   /* already searched */
	    /*
	    else
	        flayer[y][x] = - cost;
	    */
	}
	else
	    return( YES );  /* found */
    }
    /*
     *  still flayer[y][x] == 0
     *  mark point [y][x] and check neighbors
     */
    if( flayer[y][x] <= 0 )
        flayer[y][x] = - cost;
    fgp[y][x] = group;
    /*
     *  check UP, DOWN, LEFT, RIGHT, and OTHER LAYER
     */
    newy = y + 1;
    {
	if(flayer[newy][x] == 0 )
	    S_pushsearchq( cost + vcost, x, newy, group, flayer[0][0]);
	else if(fgp[newy][x] > 0 && fgp[newy][x] != group)
	{
	    touchcost = (cost + vcost + MAX( 1, - flayer[newy][x]))/2;
	    S_pushsearchq( touchcost, x, newy, group, flayer[0][0]);
	}
    }
    newy = y - 1;
    {
	if( flayer[newy][x] == 0 )
	    S_pushsearchq( cost + vcost, x, newy, group, flayer[0][0] );
	else if (fgp[newy][x] > 0 && fgp[newy][x] != group)
	{
	    touchcost = (cost + vcost + MAX( 1, - flayer[newy][x]))/2;
	    S_pushsearchq( touchcost, x, newy, group, flayer[0][0]);
	}
    }
    newx = x + 1;
    {
	if(flayer[y][newx] == 0 )
	    S_pushsearchq( cost + hcost, newx, y, group, flayer[0][0] );
	else if (fgp[y][newx] > 0 && fgp[y][newx] != group)
	{
	    touchcost = (cost + hcost + MAX( 1, - flayer[y][newx]))/2;
	    S_pushsearchq( touchcost, newx, y, group, flayer[0][0]);
	}
    }
    newx = x - 1;
    {
	if(flayer[y][newx] == 0 )
	    S_pushsearchq( cost + hcost, newx, y, group, flayer[0][0] );
	else if(fgp[y][newx] > 0 && fgp[y][newx] != group)
	{
	    touchcost = (cost + hcost + MAX( 1, - flayer[y][newx]))/2;
	    S_pushsearchq( touchcost, newx, y, group, flayer[0][0]);
	}
    }
    if( slayer[y][x] == 0 )
	S_pushsearchq( cost + swcost, x, y, group, slayer[0][0] );
    else if (sgp[y][x] > 0 && sgp[y][x] != group)
    {
	touchcost = (cost + swcost + MAX( 1, - slayer[y][x]))/2;
	S_pushsearchq( touchcost, x, y, group, slayer[0][0]);
    }
    return( NO );
}

/*
 *  if tmppath of net i has not been blocked, route the path
 *  if routing is successful, append tmplist to list, set pointers
 *                  on linklayer and return YES
 *		    Also clear the history.
 *  else remove every thing done here and free the tmplist, return NO
 */
int S_trytmppath( i )
int i;
{
    int **tmplayer,
	**gplayer;
    int blocked = FALSE;
    LINKPTR link;
    int x,
        y;

/*
printf("trytmppath of net %d\n", i);
*/
    /*
     *  mark the path
     */
    for( link = net_array[i].tmplist; link != 0;
	 link = link->netnext)
    {
/*
printf("try link %s net = %d (%d, %d), (%d, %d)\n",
( link->layer == HORIZ ? "H" : "V" ),
link->net, link->x1, link->y1, link->x2, link->y2);
*/
	if( link->layer == 0 )
    continue;
	if( link->layer == HORIZ )
	{
	    tmplayer = channel->horiz_layer;
	    gplayer = channel->hcomp;
	}
	else
	{
	    tmplayer = channel->vert_layer;
	    gplayer = channel->vcomp;
	}
	if( link->x1 == link->x2 )
	{
	    for( y = link->y1; y <= link->y2; y++ )
	    {
		if( tmplayer[y][link->x1] == 0 )
		{
		    tmplayer[y][link->x1] = i;
		    gplayer[y][link->x1] = -1;
		}
		else if( tmplayer[y][link->x1] != i )
		{
		    blocked = TRUE;
		    break;
		}
	    }
	}
	else
	{
	    for( x = link->x1; x <= link->x2; x++ )
	    {
		if( tmplayer[link->y1][x] == 0 )
		{
		    tmplayer[link->y1][x] = i;
		    gplayer[link->y1][x] = -1;
		}
		else if( tmplayer[link->y1][x] != i )
		{
		    blocked = TRUE;
		    break;
		}
	    }
	}
    }
    if ( blocked == FALSE )
    {
	/*
	 *  clean up component layer
	 */
        for( link = net_array[i].tmplist; link != 0;
	     link = link->netnext)
        {
	    if( link->layer == 0 )
	continue;
	    if( link->layer == HORIZ )
	        gplayer = channel->hcomp;
	    else
	        gplayer = channel->vcomp;
	    if( link->x1 == link->x2 )
	        for( y = link->y1; y <= link->y2; y++ )
		    gplayer[y][link->x1] = 0;
	    else
	        for( x = link->x1; x <= link->x2; x++ )
		    gplayer[link->y1][x] = 0;
	}
	/*
	 *  move tmppath to path
	 */
	while( (link = net_array[i].tmplist) != 0 )
	{
            net_array[i].tmplist = link->netnext;
	    if( link->layer == 0 )
	    {
		S_divlink( link->x1, link->y1, i );
		S_retlink( link );
	    }
	    else
	    {
   	        S_addlklayer(link, link->x1, link->y1);
	        S_addlklayer(link, link->x2, link->y2);
                link->netnext = net_array[i].path;
                net_array[i].path = link;
	    }
	}
        return( YES );
    }
    else
    {
/*
print_debug ("tmppath failed", "horiz", channel->horiz_layer, num_rows, num_cols, num_nets);
print_debug ("tmppath failed", "vert", channel->vert_layer, num_rows, num_cols, num_nets);
print_debug ("tmppath failed", "hcomp", channel->hcomp, num_rows, num_cols, num_nets);
print_debug ("tmppath failed", "vcomp", channel->vcomp, num_rows, num_cols, num_nets);
*/
	/*
	 *  remove all the marks made here
	 */
        for( link = net_array[i].tmplist; link != 0;
	     link = link->netnext)
        {
	    if( link->layer == HORIZ )
	    {
	        tmplayer = channel->horiz_layer;
	        gplayer = channel->hcomp;
	    }
	    else
	    {
	        tmplayer = channel->vert_layer;
	        gplayer = channel->vcomp;
	    }
	    if( link->x1 == link->x2 )
	    {
	        for( y = link->y1; y <= link->y2; y++ )
	        {
		    if( gplayer[y][link->x1] == -1 )
		    {
			gplayer[y][link->x1] = 0;
			tmplayer[y][link->x1] = 0;
		    }
		}
	    }
	    else
	    {
	        for( x = link->x1; x <= link->x2; x++ )
	        {
		    if( gplayer[link->y1][x] == -1 )
		    {
			gplayer[link->y1][x] = 0;
			tmplayer[link->y1][x] = 0;
		    }
		}
	    }
	}
	/*
	 *  return tmplist
	 */
	/*  delete at the beginning of findpath
	while( (link = net_array[i].tmplist) != 0 )
	{
	    net_array[i].tmplist = net_array[i].tmplist->netnext;
	    S_retlink( link );
	}
	*/
    }
    return( NO );
}

S_pushsearchq( totalcost, x, y, group, layer)
int totalcost,  /* cost to reach (x, y) */
    x,
    y,          /* coordinates */
    group,
    layer;
{
    int keycost;
    SEARCHPTR newq;
    SEARCHPTR S_getsearchq();

    /*  push in the queue  */
    keycost = totalcost % MODCOST;
    newq = S_getsearchq();
    newq->x = x;
    newq->y = y;
    newq->group = group;
    newq->layer = layer;
    newq->next = searchq[keycost];
    searchq[keycost] = newq;
}

/*
 *   clean the trace
 */
S_cleantrace()
{
    PINPTR tpin;
    int	**hcomp,
	**vcomp;
    int i;
    int x,
	y;

    hcomp = channel->hcomp;
    vcomp = channel->vcomp;
    for( x = 1; x <= num_cols; x++ )
    {
	for( y = 1; y <= num_rows; y++ )
	{
	    if( hlayer[y][x] < 0 /* && playerh[y][x] == 0 */)
	    {
		hlayer[y][x] = 0;
	    }
	    hcomp[y][x] = 0;
	    if( vlayer[y][x] < 0 /* && playerv[y][x] == 0 */)
	    {
		vlayer[y][x] = 0;
	    }
	    vcomp[y][x] = 0;
	}
    }

    /*
     *  mark all pins
     */
    for( i = 1; i <= num_nets; i++ )
    {
	for( tpin = net_array[i].pin; tpin != 0; tpin = tpin->next )
	{
	    if( tpin->layer == HORIZ )
		hlayer[tpin->y][tpin->x] = i;
	    if( tpin->layer == VERT )
		vlayer[tpin->y][tpin->x] = i;
	}
    }
}
	    

/*
 *  backtrack the path one step
 */
S_backnext( xp, yp, group, costp, layer, net )
int *xp,
    *yp;
int group;
int *costp;
int layer,
    net;
{
    int **flayer,
	**slayer,
	**fgp,
	**sgp;
    int hcost,
	vcost;
    int newx,
	newy;

/*
printf("back : x= %d, y=%d, net= %d group=%d %s cost=%d\n", *xp, *yp,
net, group, (layer == HORIZ ? "H" : "V"), *costp );
*/
    if( layer == HORIZ )
    /*
     *  set environment
     */
    {
	flayer = hlayer;
	slayer = vlayer;
	fgp = channel->hcomp;
	sgp = channel->vcomp;
	hcost = HCOST1;
	vcost = VCOST2;
    }
    else
    {
	slayer = hlayer;
	flayer = vlayer;
	sgp = channel->hcomp;
	fgp = channel->vcomp;
	hcost = HCOST2;
	vcost = VCOST1;
    }
    if( fgp[*yp][*xp] == group && flayer[*yp][*xp] == net )
    {
	return( NO );  /* finished at a link: no more path */
    }
#ifdef WDEBUG
    else if( fgp[*yp][*xp] == group && flayer[*yp][*xp] == -1 &&
	(playerh[*yp][*xp] == net || playerv[*yp][*xp] == net) )
    {
	return( NO );  /* finished at a pin: no more path */
    }
#endif
    /*
     *  On HORIZ layer check HORIZ direction first
     *  Else VERT first
     */
    if( layer == HORIZ )
    {
        if( fgp[*yp][(newx = *xp + 1)] == group &&
	      flayer[*yp][newx] == net )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[*yp][(newx = *xp - 1)] == group &&
	      flayer[*yp][newx] == net )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[(newy = *yp + 1)][*xp] == group &&
	      flayer[newy][*xp] == net )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[(newy = *yp - 1)][*xp] == group &&
	      flayer[newy][*xp] == net )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( sgp[*yp][*xp] == group &&
	      slayer[*yp][*xp] == net )
        {
	    (*costp) -= SWCOST;
	    return( SWITCH );
        }

        if( fgp[*yp][(newx = *xp + 1)] == group &&
	      flayer[*yp][newx] == - *costp + hcost )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[*yp][(newx = *xp - 1)] == group &&
	      flayer[*yp][newx] == - *costp + hcost )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[(newy = *yp + 1)][*xp] == group &&
	      flayer[newy][*xp] == - *costp + vcost )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[(newy = *yp - 1)][*xp] == group &&
	      flayer[newy][*xp] == - *costp + vcost )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
    }
    else /* VERT first */
    {
        if( fgp[(newy = *yp + 1)][*xp] == group &&
	      flayer[newy][*xp] == net )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[(newy = *yp - 1)][*xp] == group &&
	      flayer[newy][*xp] == net )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[*yp][(newx = *xp + 1)] == group &&
	      flayer[*yp][newx] == net )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[*yp][(newx = *xp - 1)] == group &&
	      flayer[*yp][newx] == net )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( sgp[*yp][*xp] == group &&
	      slayer[*yp][*xp] == net )
        {
	    (*costp) -= SWCOST;
	    return( SWITCH );
        }

        if( fgp[(newy = *yp + 1)][*xp] == group &&
	      flayer[newy][*xp] == - *costp + vcost )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[(newy = *yp - 1)][*xp] == group &&
	      flayer[newy][*xp] == - *costp + vcost )
        {
	    *yp = newy;
	    (*costp) -= vcost;
	    return( VERT );
        }
        else if( fgp[*yp][(newx = *xp + 1)] == group &&
	      flayer[*yp][newx] == - *costp + hcost )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
        else if( fgp[*yp][(newx = *xp - 1)] == group &&
	      flayer[*yp][newx] == - *costp + hcost )
        {
	    *xp = newx;
	    (*costp) -= hcost;
	    return( HORIZ );
        }
    }
    if( sgp[*yp][*xp] == group &&
	slayer[*yp][*xp] == - *costp + SWCOST )
    {
	(*costp) -= SWCOST;
	return( SWITCH );
    }
    else
    {
print_debug ("backnext failed", "horiz", channel->horiz_layer, num_rows, num_cols, num_nets);
print_debug ("backnext failed", "vert", channel->vert_layer, num_rows, num_cols, num_nets);
print_debug ("backnext failed", "hcomp", channel->hcomp, num_rows, num_cols, num_nets);
print_debug ("backnext failed", "vcomp", channel->vcomp, num_rows, num_cols, num_nets);
	S_dumppath( net );
	exit( 136 );
    }
    /*NOTREACHED*/
}


/*
 *  clean up seachq
 */
S_cleansearchq()
{
    int count;
    SEARCHPTR tmpq;

    for( count = 0; count < MODCOST; count++ )
    {
	while( (tmpq = searchq[count]) != 0 )
	{
	    searchq[count] = tmpq->next;
	    S_retsearchq( tmpq );
	}
    }
}
