#include "copyright.h"
#include "port.h"

#include "oct.h"
#include "st.h"
#include "header.h"
#include "region.h"

extern st_table *PATH_HASH;

struct octBox *get_term_box2();
octObject *getMaster();

/*******************************************************************/

make_clusters()
{
octGenerator path_gen,bag_gen;
octObject path1,bag,layer;
octId last_path_id;
pcellist met1,met2,cl,cl2;
pcell cell1;
  
#ifndef QUIET
    printf("making clusters...");
    fflush(stdout);
#endif

OCTCHKOK(octInitGenContents(&network,OCT_BAG_MASK,&bag_gen));
while(octGenerate(&bag_gen,&bag) == OCT_OK)
  {
    if(strncmp(bag.contents.bag.name,"CROSSING",8) == 0)
      {
	cl = met1 = met2 = NULL;
	OCTCHKOK(octInitGenContents(&bag,OCT_PATH_MASK,&path_gen));
	while(octGenerate(&path_gen,&path1) == OCT_OK)
	  {
	    cl = new_cellist();
	    if(st_lookup(PATH_HASH,(char *) path1.objectId,(char **)&(cl->cell)) == 0)
	      {
		printf("(MIZER): serious internal ERROR\n");
		quit();
	      }
	    OCTCHKOK(octGenFirstContainer(&path1,OCT_LAYER_MASK,&layer));
	    if(octIdsEqual(layer.objectId, layers[LAYER1].objectId))
	      {
		cl->next = met1;
		met1 = cl;
		last_path_id = path1.objectId;
	      }
	    else
	      {
		cl->next = met2;
		met2 = cl;
	      }
	  }
	if((met1 != NULL) && (met2 != NULL))
	  {
	    for(cl = met1;cl != NULL;cl = cl->next)
	      {
		make_hard_link(cl->cell,met2->cell);
	      }
	    if(st_lookup(PATH_HASH,(char *)last_path_id,(char **)&(cell1)) == 0)
	      {
		printf("(MIZER): serious internal ERROR\n");
		quit();
	      }
	    for(cl = met2->next;cl != NULL;cl = cl->next)
	      {
		make_hard_link(cl->cell,cell1);
	      }
	  }
	for(cl = met1;cl != NULL;)
	  {
	    cl2 = cl;
	    cl = cl->next;
	    kill_cellist(cl2);
	  }
	for(cl = met2;cl != NULL;)
	  {
	    cl2 = cl;
	    cl = cl->next;
	    kill_cellist(cl2);
	  }
	OCTCHKOK(octDelete(&bag));
	OCTCHKOK(octFreeGenerator(&path_gen));
      }
  }
#ifndef QUIET
    printf("time %.2f\n",tot_time());
    fflush(stdout);
#endif
}

/***********************************************************/

make_hard_link(cell1,cell2)
pcell cell1,cell2;
{
pcell cell3,cell4;

if(cell1 == cell2)
  return;

if((cell1->hard_link == NULL) && (cell2->hard_link == NULL))
  {
    cell1->hard_link = cell2;
    cell2->hard_link = cell1;
    if(cell1->fixed == TRUE)
      {
	if(cell2->fixed == TRUE)
	  {
	    if(cell2->side == cell1->side)
	      {
		printf("ERROR fixed cells on opposite side\n");
		quit();
	      }
	  }
	else
	  {
	    cell2->fixed = TRUE;
	    cell2->side = (cell1->side == 1)?2:1;
	  }
      }
    else
      {
	if(cell2->fixed == TRUE)
	  {
	    cell1->fixed = TRUE;
	    cell1->side = (cell2->side == 1)?2:1;
	  }
      }
  }    
else
  if((cell1->hard_link != NULL) && (cell2->hard_link == NULL))
    {
      merge_hard_links(cell1->hard_link,cell2);
    }
  else
    if((cell1->hard_link == NULL) && (cell2->hard_link != NULL))
      {
	merge_hard_links(cell2->hard_link,cell1);
      }
    else
      {
	cell3 = cell1->hard_link;
	cell4 = cell2->hard_link;
	merge_hard_links(cell2,cell3);
	merge_hard_links(cell1,cell4);
	cell1->hard_link = cell2;
	cell2->hard_link = cell1;
      }
}

/*******************************************************************/

merge_hard_links(cell3,cell2)
pcell cell3,cell2;
{
ppathList pl;

if(cell3 == cell2)
  return;
if(cell2->fixed == TRUE)
  {
    if(cell3->fixed == FALSE)
      {
	cell3->fixed = TRUE;
	cell3->side = cell2->side;
	if(cell3->hard_link != NULL)
	  {
	    if(cell3->hard_link->fixed == TRUE)
	      {
		if(cell3->hard_link->side == cell2->side)
		  {
		    printf("ERROR fixed cells on wrong side\n");
		    quit();
		  }
	      }
	    else
	      {
		cell3->hard_link->fixed = TRUE;
		cell3->hard_link->side = (cell3->side == 1)?2:1;
	      }
	  }
      }
    else
      {
	if(cell3->side != cell2->side)
	  {
	    printf("ERROR in merging two opposite fixed cells\n");
	    quit();
	  }
      }
  }
else
  {
    if(cell3->fixed == TRUE)
      {
	cell2->fixed = TRUE;
	cell2->side = cell3->side;
	if(cell2->hard_link != NULL)
	  {
	    if(cell2->hard_link->fixed == TRUE)
	      {
		if(cell2->hard_link->side == cell3->side)
		  {
		    printf("ERROR fixed cells on wrong side\n");
		    quit();
		  }
	      }
	    else
	      {
		cell2->hard_link->fixed = TRUE;
		cell2->hard_link->side = (cell2->side == 1)?2:1;
	      }
	  }
      }
  }
merge_two_nodes(cell3,cell2);

for(pl = cell3->paths;pl != NULL;pl = pl->next)
  {
    if(st_insert(PATH_HASH, (char *)pl->id, (char *)cell3) != 1)
      {
	printf("ERROR: hash insert problems\n");
	quit();
      }
  }
}

/*******************************************************************/

remove_from_cell_list(cell,cell_list)
pcell cell,*cell_list;
{

if((*cell_list) == cell)
  {
    (*cell_list) = (*cell_list)->next;
    if((*cell_list) != NULL)
      (*cell_list)->prev = NULL;
  }
else
  {
    if(cell->prev != NULL)
      cell->prev->next = cell->next;
    if(cell->next != NULL)
      cell->next->prev = cell->prev;
  }
FREE_MEM(1,cell,struct cell);
}


