#include "copyright.h"
#include "port.h"

#include "oct.h"
#include "st.h"
#include "header.h"
pcell FIXED_MET1 = NULL,FIXED_MET2 = NULL;
octObject *getMaster();
ppathList new_pathList();
char *oct_find_str_prop();

extern int CONT_WIDTH;
extern int FAST;
extern int SLOW;

st_table *PATH_HASH;
pcell CELLS = NULL;
pnet NETS = NULL;
int ESSENTIAL;
int OTHER_LAYERS_EXIST = FALSE;
pnet DEAD_NETS = NULL;

make_graph()
{
int found;
/***
int found2;
***/

PATH_HASH = st_init_table(st_numcmp,st_numhash);

make_paths();
make_terminals();
make_clusters();
printf("(MIZER %.2f sec): graph construction complete...\n",tot_time());
fflush(stdout);

set_up_fixed_cells();
#ifndef QUIET
     printf("count %d\n",count_cells());
#endif
ESSENTIAL = 0;

/***
do 
  {
    found2 = FALSE;
***/
    do
      {
	found = FALSE;
	found = remove_essential_vias();
	found |= remove_3_cycles();
	found |= remove_terminal_overconstraints();
	found |= remove_terminal_conflicts();
	while(check_for_essential(FIXED_MET1) == TRUE)
	  {
	    ESSENTIAL++;
	    found = TRUE;
	  }
	while((FAST == FALSE) && (simplify_cells() == TRUE))
	  {
	    found = TRUE;
	  }
	/***
	if(found == TRUE)
	  found2 = TRUE;
	***/
      }
    while((CELLS != NULL) && (found == TRUE));
/***
  }
while((SLOW == TRUE) && 
      (CELLS != NULL) && 
      ((run1() != 0) || (found2 == TRUE)));
***/

printf("(MIZER %.2f sec): graph simplification complete...\n",tot_time());
exhaustive();
/*
print_graph();
show_clusters();
*/

if(count_cells() == 0)
  printf("(MIZER %.2f sec): optimal solution obtained\n",tot_time());

printf("(MIZER %.2f sec): %d essential vias identified\n",tot_time(),ESSENTIAL);
fflush(stdout);

#ifndef QUIET
check_cells_for_dups();
       printf("time %.2f\n",tot_time());
       printf("end graph\n");
       fflush(stdout);
#endif
st_free_table(PATH_HASH);
}

make_paths()
{
octGenerator path_gen,inst_gen, net_gen;
octObject term,path,bag,inst,oct_net,layer;
pcell cell;
pnet net;
pcellist cl,cellList;
pnetlist nl;
struct octBox box;  
int max_dim;
int num;
int cnt;
int fixed;
int num_of_paths;

#ifndef QUIET
  printf("start making paths at %.2f ...",tot_time());
#endif
bag.type = OCT_BAG;
bag.contents.bag.name = "CONNECTORS";
OCTCHKOK(octGetByName(&network,&bag));

OCTCHKOK(octInitGenContents(&bag,OCT_INSTANCE_MASK,&inst_gen));
while(octGenerate(&inst_gen,&inst) == OCT_OK)
  {
    OCTCHKOK(octGenFirstContent(&inst,OCT_TERM_MASK,&term));
    if(term_is_formal(&term) == TRUE)
      continue;
    OCTCHKOK(octBB(&inst,&box));
    if(box.upperRight.x - box.lowerLeft.x > box.upperRight.y - box.lowerLeft.y)
      max_dim = box.upperRight.x - box.lowerLeft.x;
    else
      max_dim = box.upperRight.y - box.lowerLeft.y;
    if(max_dim > CONT_WIDTH)
      {
	num = (max_dim + (CONT_WIDTH - 1)) / CONT_WIDTH;
/*
	printf("weighting contact at %d %d times %d\n",
	       inst.contents.instance.transform.translation.x,
	       inst.contents.instance.transform.translation.y,num);
*/
      }
    else
      {
	num = 1;
      }
    num_of_paths = 0;
    fixed = FALSE;
    OCTCHKOK(octInitGenContainers(&term,OCT_PATH_MASK,&path_gen));
    while(octGenerate(&path_gen,&path) == OCT_OK)
      {
	OCTCHKOK(octGenFirstContainer(&path,OCT_LAYER_MASK,&layer));
	if(which_layer(&layer) == OTHER_LAYER)
	  {
	    fixed = TRUE;	    
	  }
	else
	  num_of_paths++;
      }
    for(;num > 0;num--)
      {
	if((num_of_paths > 1) && (fixed == FALSE))
	  {
	    net = new_net();
	    net->next = NETS;
	    if(net->next != NULL)
	      net->next->prev = net;
	    net->name = (char *)inst.objectId;
	    net->OCTid = inst.objectId;
	    NETS = net;
	  }

	cellList = NULL;
	OCTCHKOK(octInitGenContainers(&term,OCT_PATH_MASK,&path_gen));
	cnt = 0;
	while(octGenerate(&path_gen,&path) == OCT_OK)
	  {
	    OCTCHKOK(octGenFirstContainer(&path,OCT_LAYER_MASK,&layer));
	    if(which_layer(&layer) != OTHER_LAYER)
	      {
		if(st_lookup(PATH_HASH, (char *)path.objectId, (char **)&cell) != 1)
		  {		
		    cell = new_cell();
		    cell->next = CELLS;
		    if(CELLS != NULL)
		      CELLS->prev = cell;
		    CELLS = cell;
		    cell->paths = new_pathList();
		    cell->paths->id = path.objectId;
		    st_insert(PATH_HASH, (char *)path.objectId, (char *)cell);
		  }
		if((num_of_paths > 1) && (fixed == FALSE))
		  {
		    cl = new_cellist();
		    cnt++;
		    cl->next = cellList;
		    cellList = cl;
		    cl->cell = cell;
		    nl = new_netlist();
		    nl->net = net;
		    nl->next = cell->nets;
		    cell->nets = nl;
		  }
		if(fixed == TRUE)
		  {
/*
		    if(which_layer(&layer) != MET1)
		      {
			struct octPoint pp[2];
			
			get_path2(&path,pp);
			printf("(MIZER):ERROR path at %d %d %d %d has multi layer contacts\n",pp[0].x,pp[0].y,pp[1].x,pp[1].y);
			quit();
		      }
*/
		    cell->fixed = TRUE;
		    cell->side = LAYER1;
		  }
	      }
	  }
	if((num_of_paths > 1) && (fixed == FALSE))
	  {
	    net->cells = cellList;
	    if(cnt == 2)
	      net->two_conn = TRUE;
	  }

	OCTCHKOK(octFreeGenerator(&path_gen));
      }
  }
OCTCHKOK(octInitGenContents(&network, OCT_NET_MASK, &net_gen))
while(octGenerate(&net_gen,&oct_net) == OCT_OK)
  {
    OCTCHKOK(octInitGenContents(&oct_net,OCT_PATH_MASK,&path_gen));
    while(octGenerate(&path_gen,&path) == OCT_OK)
      {
	OCTCHKOK(octGenFirstContainer(&path,OCT_LAYER_MASK,&layer));
	if(which_layer(&layer) != OTHER_LAYER)
	  {
	    if(st_lookup(PATH_HASH, (char *)path.objectId, (char **)&cell) != 1)
	      {
/*
		printf("found isolated path\n"); 
*/
		cell = new_cell();
		cell->next = CELLS;
		if(CELLS != NULL)
		  CELLS->prev = cell;
		CELLS = cell;
		cell->paths = new_pathList();
		cell->paths->id = path.objectId;
		st_insert(PATH_HASH, (char *)path.objectId, (char *)cell);
	      }
	  }
      }
  }
#ifndef QUIET
    printf("end at %.2f\n",tot_time());
    MEM_SIZE("after making paths");
#endif
}

make_terminals()
{
octGenerator net_gen,term_gen,path_gen;
octObject net,term,path,term2,*pterm,net2,layer;
int type;
pcell cell;
#ifndef QUIET
int num_met1,num_met2,num_either;
#endif

#ifndef QUIET
    printf("start making terminals at %.2f ...\n",tot_time());
    fflush(stdout);
#endif

#ifndef QUIET
num_met1 = num_met2 = num_either = 0;
#endif
init_masters();  
OCTCHKOK(octInitGenContents(&network,OCT_NET_MASK,&net_gen));
while(octGenerate(&net_gen,&net) == OCT_OK)
  {
    OCTCHKOK(octInitGenContents(&net,OCT_TERM_MASK,&term_gen));
    while(octGenerate(&term_gen,&term) == OCT_OK)
      {
	if (octIdIsNull(term.contents.term.instanceId))
	   {
	     if(octGenFirstContent(&term,OCT_TERM_MASK,&term2) != OCT_OK)
	       {
		 printf("(MIZER): WARNING -- net %s formal terminal %s has no implementation\n",
			net.contents.net.name,
			term.contents.term.name);
		 continue;
	       }
	     if(octGenFirstContainer(&term2,OCT_NET_MASK,&net2) == OCT_OK)
	       continue;
	     pterm = &term2;
	   }
	else
	  pterm = &term;
	OCTCHKOK(octInitGenContainers(pterm,OCT_PATH_MASK,&path_gen));
	while(octGenerate(&path_gen,&path) == OCT_OK)
	  {
	    OCTCHKOK(octGenFirstContainer(&path,OCT_LAYER_MASK,&layer));
	    if((which_layer(&layer) == OTHER_LAYER))
	      {
		continue;
	      }
	    else
	      {
		if(st_lookup(PATH_HASH, (char *)path.objectId, (char **)&cell) == 0)
		  {
		    printf("(MIZER): internal ERROR - hash lookup\n");
		    quit();
		  }
		type = get_term_type(pterm);
	      }
	    if(cell->fixed == TRUE)
	      continue;

	    if(type != 3)
	      {
		cell->side = type;
		cell->fixed = TRUE;
#ifndef QUIET
		if(type == 1)
		  num_met1++;
		else
		  num_met2++;
#endif
	      }
#ifndef QUIET
	    else
	      {
		num_either++;
	      }
#endif
	  }
      }
    OCTCHKOK(octFreeGenerator(&term_gen));
  }
free_masters();
#ifndef QUIET
if(num_met1 != 0)
  printf("%d fixed met1 terminals\n",num_met1);
if(num_met2 != 0)
  printf("%d fixed met2 terminals\n",num_met2);
if(num_either != 0)
  printf("%d floating terminals\n",num_either);
  printf("end at %.2f\n",tot_time());
#endif
}

/*******************************************************************/

term_is_on_poly(term)
octObject *term;
{
octObject inst,*master,term2,box2,layer;
octGenerator box_gen;

if(OTHER_LAYERS_EXIST == FALSE)
  return(FALSE);
OCTCHKOK(octGenFirstContainer(term,OCT_INSTANCE_MASK,&inst));

master = getMaster(inst.contents.instance.master,inst.contents.instance.view);

term2.type = OCT_TERM;
term2.contents.term.name = term->contents.term.name;
OCTCHKOK(octGetByName(master,&term2));

OCTCHKOK(octInitGenContents(&term2,OCT_BOX_MASK,&box_gen));
while(octGenerate(&box_gen,&box2) == OCT_OK)
  {
    OCTCHKOK(octGenFirstContainer(&box2,OCT_LAYER_MASK,&layer));
    if(which_layer(&layer) == OTHER_LAYER)
      {
	return(TRUE);
      }
  }
return(FALSE);
}

get_term_type(term)
octObject *term;
{
octObject inst,*master,term2,box2,layer;
octGenerator box_gen,inst_gen;
struct octBox met1,met2,temp;
int found_met1,found_met2;

OCTCHKOK(octInitGenContainers(term,OCT_INSTANCE_MASK,&inst_gen));
OCTCHKOK(octGenerate(&inst_gen,&inst));
OCTCHKOK(octFreeGenerator(&inst_gen));

master = getMaster(inst.contents.instance.master,inst.contents.instance.view);

/*
printf("of master %s ",master->contents.facet.cell);
*/

term2.type = OCT_TERM;
term2.contents.term.name = term->contents.term.name;
OCTCHKOK(octGetByName(master,&term2));

found_met1 = found_met2 = FALSE;
OCTCHKOK(octInitGenContents(&term2,OCT_BOX_MASK,&box_gen));
while(octGenerate(&box_gen,&box2) == OCT_OK)
  {
    OCTCHKOK(octGenFirstContainer(&box2,OCT_LAYER_MASK,&layer));
    if(strcmp(layer.contents.layer.name,layer_nameTable[LAYER1]) == 0)
      {
	if(found_met1 == FALSE)
	  {
	    OCTCHKOK(octBB(&box2,&met1));
	    found_met1 = TRUE;
	  }
	else
	  {
	    OCTCHKOK(octBB(&box2,&temp));
	    if(box_is_bigger(&temp,&met1))
	      {
		OCTCHKOK(octBB(&box2,&met1));
	      }	    
	  }
      }
  }
OCTCHKOK(octInitGenContents(&term2,OCT_BOX_MASK,&box_gen));
while(octGenerate(&box_gen,&box2) == OCT_OK)
  {
    OCTCHKOK(octGenFirstContainer(&box2,OCT_LAYER_MASK,&layer));
    if(strcmp(layer.contents.layer.name,layer_nameTable[LAYER2]) == 0)
      {
	if(found_met2 == FALSE)
	  {
	    OCTCHKOK(octBB(&box2,&met2));
	    found_met2 = TRUE;
	  }
	else
	  {
	    OCTCHKOK(octBB(&box2,&temp));
	    if(box_is_bigger(&temp,&met2))
	      {
		OCTCHKOK(octBB(&box2,&met2));
	      }	    
	  }
      }
  }
if((found_met1 == TRUE) && (found_met2 == TRUE))
  {
    if(box_is_bigger(&met1,&met2))
      return(LAYER1);
    if(box_is_bigger(&met2,&met1))
      return(LAYER2);
    return(3);
  }
else
  {
    if(found_met1 == TRUE)
      return(LAYER1);
    if(found_met2 == TRUE)
      return(LAYER2);
  }
OCTCHKOK(octBB(&inst,&met1));
printf("(MIZER): ERROR -- terminal %s at %d %d %d %d has no implementation on MET1 or MET2\n",
       term->contents.term.name,       
       met1.lowerLeft.x,
       met1.lowerLeft.y,
       met1.upperRight.x,
       met1.upperRight.y);
quit();
/*NOTREACHED*/
}


remove_essential_vias()
{
pcell cell;
int cnt;

cnt = 0;
for(cell = CELLS;cell != NULL;cell = cell->next)
  {
    if((cell->hard_link != NULL) && 
       (cell->paths->id <= cell->hard_link->paths->id))
      {
	while(check_for_essential(cell) == TRUE)
	  {
	    cnt++;
	  }
      }
  }
if(cnt != 0)
  {
#ifndef QUIET
      printf("%d essential vias found\n",cnt);
#endif
    ESSENTIAL += cnt;
    return(TRUE);
  }
return(FALSE);
}

check_for_essential(cell)
pcell cell;
{
pnetlist nl;
pcellist cl;

for(nl = cell->nets;nl != NULL;nl = nl->next)
  {
    for(cl = nl->net->cells;cl != NULL;cl = cl->next)
      {
	if(cl->cell == cell->hard_link)
	  {
/*
	    printf("essential net %d\n",nl->net->id);
*/
	    kill_net(nl->net);
	    return(TRUE);
	  }
      }
  }
return(FALSE);
}

kill_net(net)
pnet net;
{
pnetlist nl, nl2;
pcellist cl, cl2;

/*
printf("killing net %d\n",net->id);
*/
if(net == NETS)
  {
    NETS = NETS->next;
    if(NETS != NULL)
      NETS->prev = NULL;
  }
else
  {
    if(net->prev != NULL)
      net->prev->next = net->next;
    if(net->next != NULL)
      net->next->prev = net->prev;
  }
for(cl = net->cells;cl != NULL;)
  {
    nl = cl->cell->nets;
    if(nl->net == net)
      {
	nl2 = nl;
	cl->cell->nets = nl->next;
      }
    else
      {
	for(;(nl != NULL) && 
	    (nl->next != NULL) && 
	    (nl->next->net != net);nl = nl->next)
	  {}
	if((nl == NULL) || (nl->next == NULL))
	  {
	    printf("ERROR in deleting net connection\n");
	    quit();
	  }
	nl2 = nl->next;
	nl->next = nl->next->next;
      }
    cl2 = cl;
    cl = cl->next;
    FREE_MEM(1, cl2, struct cellist);
    FREE_MEM(1, nl2, struct netlist);
  }
net->next = DEAD_NETS;
DEAD_NETS = net;
}

set_up_fixed_cells()
{
    pcell cell,c;

    while( CELLS != NULL && CELLS->fixed == TRUE)
      {
	  cell = CELLS;
	  CELLS = CELLS->next;
	  if(CELLS != NULL)
	    CELLS->prev = NULL;
	  affix_cell(cell);
      }
    cell = CELLS;
    while((cell != NULL) && (cell->next != NULL))
      {
	  if(cell->next->fixed == TRUE)
	    {
		c = cell->next->next;
		affix_cell(cell->next);
		cell->next = c;
		if(c != NULL)
		  c->prev = cell;
	    }
	  else
	    cell = cell->next;
      }
    if(FIXED_MET1 == NULL)
      {
	  FIXED_MET1 = new_cell();
	  FIXED_MET1->fixed = TRUE;
	  FIXED_MET1->side = LAYER1;
      }
    if(FIXED_MET2 == NULL)
      {
	  FIXED_MET2 = new_cell();
	  FIXED_MET2->fixed = TRUE;
	  FIXED_MET2->side = LAYER2;
      }

    FIXED_MET1->hard_link = FIXED_MET2;
    FIXED_MET2->hard_link = FIXED_MET1;
}

affix_cell(cell)
pcell cell;
{

if((cell->hard_link != NULL) && (cell->hard_link->fixed == FALSE))
  {
    printf("ERROR in affix_cell\n");
    quit();
  }
cell->next = cell->prev = NULL;
if(cell->side == 1)
  {
    if(FIXED_MET1 == NULL)
      FIXED_MET1 = cell;
    else
      {
	merge_two_nodes(FIXED_MET1,cell);
      }
  }
else
  {
    if(FIXED_MET2 == NULL)
      FIXED_MET2 = cell;
    else
      {
	merge_two_nodes(FIXED_MET2,cell);
      }
  }
}


num_of_nets(cell)
pcell cell;
{
int i;
pnetlist nl;

for(i = 0,nl = cell->nets;nl != NULL;nl = nl->next,i++)
  {}
return(i);
}

num_of_cells(net)
pnet net;
{
int i;
pcellist cl;

for(i = 0,cl = net->cells;cl != NULL;cl = cl->next,i++)
  {}
return(i);
}

which_layer(layer)
octObject *layer;
{

if(octIdsEqual(layer->objectId, layers[LAYER1].objectId))
  return(LAYER1);
if(octIdsEqual(layer->objectId, layers[LAYER2].objectId))
  return(LAYER2);
return(OTHER_LAYER);
}

other_layers_exist()
{
octObject path,lay;
octGenerator path_gen;

OTHER_LAYERS_EXIST = FALSE;
OCTCHKOK(octInitGenContents(&network,OCT_PATH_MASK,&path_gen));
while(octGenerate(&path_gen,&path) == OCT_OK)
  {
    if(path_layer(&path) == OTHER_LAYER)
      {
	OTHER_LAYERS_EXIST = TRUE;
	return;
      }
  }    
}

box_is_bigger(b1,b2)
struct octBox *b1,*b2;
{
if(b1->upperRight.x - b1->lowerLeft.x > b2->upperRight.x - b2->lowerLeft.x)
  return(TRUE);
if(b1->upperRight.y - b1->lowerLeft.y > b2->upperRight.y - b2->lowerLeft.y)
  return(TRUE);
return(FALSE);
}

simplify_cells()
{
int found;

found = FALSE;
while(merge_simple_nodes(CELLS) == TRUE)
  {
    found = TRUE;
  }

found |= remove_terminal_overconstraints();
found |= remove_terminal_conflicts();

while(merge_linked_nodes(CELLS) == TRUE)
  {
    found = TRUE;
  }

found |= remove_terminal_overconstraints();
found |= remove_terminal_conflicts();
clean_up_unconnected();
return(found);
}

clean_up_unconnected()
{
pcell c,next;
pcell c2;

for(c = CELLS;c != NULL;)
  {
    next = c->next;
    if((c->nets == NULL) && 
       ((c->hard_link == NULL) || (c->hard_link->nets == NULL)))
      {
	if(c->hard_link != NULL)
	  {
	    if(next == c->hard_link)
	      {
		next = next->next;
	      }
	    merge_two_nodes(FIXED_MET2,c->hard_link);
	  }
	merge_two_nodes(FIXED_MET1,c);
      }
    c = next;
  }
}

do_free_cell(c)
pcell c;
{
ppathList pl,pl2;
pnetlist nl,nl2;

for(pl = c->paths;pl != NULL;)
  {
    pl2 = pl;
    pl = pl->next;
    FREE_MEM(1, pl2, ppathList);
  }
nl = c->nets;
while(nl != NULL)
  {
    nl2 = nl;
    nl = nl->next;
    FREE_MEM(1,nl2,struct netlist);
  }
FREE_MEM(1, c, struct cellist);
}

