#include "port.h"
#include "oct.h"
#include "memchk.h" 

#define OCTCHKOK(octCall) \
{ \
if (octCall!=OCT_OK) \
    octError("Failed: \"octCall==OCT_OK\""); \
}
#define MIN(a,b) (((a) < (b))?(a):(b))
#define MAX(a,b) (((a) > (b))?(a):(b))
#define two_pt_connection(net) (net->two_conn)

typedef struct cell *pcell;
typedef struct cellist *pcellist;
typedef struct net *pnet;
typedef struct netlist *pnetlist;
typedef int *pgain;
typedef struct bucket *pbucket;
typedef struct hash *phash;
typedef struct boxList *pboxList;
typedef struct pathList *ppathList;

struct bucket 
{
  int global_max;
  pcell *cells;
  int num;
};

struct cell
{
  pcell next;
  pcell prev;
  pcell bucket_next;
  pcell bucket_prev;
  pnetlist nets;
  pbucket bucket;
  char *name;
  pcell hard_link;
  pcell hl1;
  pcell hl2;
  ppathList paths;
  int HLparent;
  int gain;
  int part;
  int side;
  int id;
  int order;
  int area;
  int width;
  int x,y;
  int fixed;
};

struct boxList
{
  struct octBox *box;
  octId bagId;
  pboxList next;
};

struct pathList
{
octId id;
ppathList next;
};

struct cellist
{
  pcell cell;
  pcellist next;
};

struct netlist
{
  pnet net;
  pnetlist next;
};

struct net
{
  pnet next;
  pnet prev;
  pcellist cells;
  char *name;
  octId OCTid;
  int id;
  int lock;
  int free[2],locked[2];
  int two_conn;
};

typedef struct hashObj *phashObj;
struct hashObj
{
  char *obj;
  int key;
  phashObj next;
};

struct hash
{
  phashObj *start;
  int size;
  int contents;
  int curr;
};

/*********************************************************************/

extern int BALANCE;
extern int P;
extern pcell CELLS;
extern pnet NETS;
extern int LOCKED_NETS;
extern pbucket FREE[2];
extern pbucket LOCKED[2];
extern pbucket DOUB_FREE[2][2];
extern pbucket DOUB_LOCKED[2][2];
extern int NUM_OF_CELLS;
extern int NUM_OF_NETS;
extern int total_cuts;
extern int X_SCALE;
extern int Y_SCALE;
extern char *FLOOD;
extern int CELL_WIDTH;
extern int X_OFFSET;
extern int TYPE;
extern int PROP;
extern int PAD;
extern pcell PADS[4];
extern int STD_CELL;
extern int CUT_DIR;
extern int TOT_AREA;
extern int GROUP;
extern double DENSITY;
extern int ROUND;

/*********************************************************************/

#define OTHER_LAYER 0
#define LAYER1 1
#define LAYER2 2
extern char *layer_nameTable[];
extern int layer_width[];
extern int layer_spacing[];
extern octObject layers[];
extern octObject network;

/*********************************************************************/

char *CALLOC_CHK();
pnetlist new_netlist();
phash new_hash();
pnet new_net();
pcell new_cell();
pcellist new_cellist();
pbucket new_bucket();
pcell pick_move_cell();
double tot_time();

/*********************************************************************/

#define top right
#define bot left
#define TRUE 1
#define FALSE 0
#define PART_A 0
#define PART_B 1
#define SIDE_1 0
#define SIDE_2 1
#define EXH_THRESH 16
#define VER 99
#define HOR 100
#define RIGHT 32
#define LEFT 34
#define TOP 12
#define BOT 22
#define SINGLE 11
#define DOUBLE 14
#define INF 0x7fffffff

#define remove_from_bucket(cell) delete_from_bucket(cell)
#define put_in_bucket(cell,buc) insert_in_bucket(buc,cell)

#define add_net_gain(from_free,from_locked,to_free,to_locked,cell) set_net_gain(from_free,from_locked,to_free,to_locked,cell,1)

#define subt_net_gain(from_free,from_locked,to_free,to_locked,cell) set_net_gain(from_free,from_locked,to_free,to_locked,cell,-1)

#define get_max_cell(bucket) ((bucket->num == 0)? NULL : (bucket->cells[bucket->global_max]))

#define check_balance(from_part,to_part) (FREE[from_part]->num + LOCKED[from_part]->num - (get_max_cell(FREE[from_part]))->area + BALANCE > FREE[to_part]->num + LOCKED[to_part]->num + (get_max_cell(FREE[from_part]))->area)

#define compare_gains(g1,g2) ((g1 > g2)?g1:((g2 > g1)?g2:((random()%2 == 1)?g1:g2)))
