#include "copyright.h"
#include "port.h"

#include "oct.h"
#include "region.h"
#include "header.h"
#define POLY 0
extern octObject layers[3];
#ifdef OLD
static int *REGION_LIST = NULL;
#else
static ppathList REGION_LIST = NULL;
#endif

static int kd_tos;


#define DIMENSIONALITY 4
#define BIGGEST 100000000
#define SMALLEST -100000000
#define TRUE 1
#define FALSE 0
#define HISON_VAL 0
#define LOSON_VAL 1
#define LOSON(p)  ((p)->kd_sons[LOSON_VAL])
#define HISON(p)  ((p)->kd_sons[HISON_VAL])
#define LOMINBOUND(p) ((p)->kd_lo_min_bound)
#define HIMAXBOUND(p) ((p)->kd_hi_max_bound)
#define OTHERBOUND(p) ((p)->kd_other_bound)
#define NEXTDISC(i)   (((i) + 1)%DIMENSIONALITY)
#define K(i,p)        ((p)->kd_keys[i])

typedef int kdkey_t;


struct kdtree {
    kdkey_t	kd_keys[DIMENSIONALITY];
    octId	        id;
    kdkey_t	kd_lo_min_bound,
    kd_hi_max_bound,
    kd_other_bound;
    struct kdtree	*kd_sons[2];
};

typedef struct kdtree kdtree_t;


/* ****** Forward declaration of static functions ****** */
static kdtree_t *_kd_optimize();
static void _kd_region_search();
static int _kd_touches_region();




static kdtree_t *KDTREE_ROOT = NULL;
struct {
    kdkey_t kdrecdef[DIMENSIONALITY];
    kdtree_t *kdroot;
} kdglobal;

kdtree_t *new_kdtree_node()
{
    kdtree_t *kdt;

    kdt = CALLOC(1,kdtree_t);
    kdt->kd_keys[0] = 0;
    kdt->kd_keys[1] = 0;
    kdt->kd_keys[2] = 0;
    kdt->kd_keys[3] = 0;
    kdt->kd_lo_min_bound = 0;
    kdt->kd_hi_max_bound = 0;
    kdt->kd_other_bound = 0;
    kdt->kd_sons[0] = NULL;
    kdt->kd_sons[1] = NULL;
    return(kdt);
}

ppathList new_pathList()
{
    ppathList pl;

    pl = CALLOC(1,struct pathList);
    pl->next = NULL;
    pl->id = 0;
    return(pl);
}

/***********************************************************************/

kdtree_t *kdtree_optimize(brect,llist)
    kdkey_t brect[];
    kdtree_t *llist;
{
    register i;
    kdtree_t *_kd_optimize();

    for(i=0;i<DIMENSIONALITY;i++)
      kdglobal.kdrecdef[i] = brect[i];

    kdglobal.kdroot = _kd_optimize(llist,0);
    return(kdglobal.kdroot);
}

/***********************************************************************/

static kdtree_t *_kd_optimize(A,j)
    register kdtree_t *A;
    int j;
{
    register i,m;
    register kdtree_t *next;
    kdtree_t *Al,*Ah,*nroot;
    int  tmp,looking = 1;
    int lomin,lomax,himin,himax;

    if(A == NULL)
      return(NULL);

    i = j&01;
    m = (kdglobal.kdrecdef[i+2]+kdglobal.kdrecdef[i])/2;

    Al = Ah = NULL;
    nroot = A;
    lomin = himin = BIGGEST;
    lomax = himax = SMALLEST;
    next = HISON(A);
    while(A = next)  {    
	next = HISON(A);
	if(looking) {
	    if(K(j,A) == m) {
		if(K(j,nroot) < m)  {
		    HISON(nroot) = Al;
		    Al = nroot;
		    lomin = MIN(lomin,K(i,Al));
		    lomax = MAX(lomax,K(i + 2,Al));
		} else {
		    HISON(nroot) = Ah;
		    Ah = nroot;
		    himin = MIN(himin,K(i,Ah));
		    himax = MAX(himax,K(i + 2,Ah));
		}

		nroot = A;
		looking = 0;
		continue;
	    } else if ((K(j,A) < m) && (K(j,A) > K(j,nroot))) {
		HISON(nroot) = Al;
		Al = nroot;
		lomin = MIN(lomin,K(i,Al));
		lomax = MAX(lomax,K(i+2,Al));
		nroot = A;
		continue;
      
	    } else if ((K(j,A) > m) && (K(j,A) < K(j,nroot))) {
		HISON(nroot) = Ah;
		Ah = nroot;
		himin = MIN(himin,K(i,Ah));
		himax = MAX(himax,K(i + 2,Ah));
		nroot = A;
		continue;

	    }
	}
	if(K(j,A) < m) {
	    HISON(A) = Al;
	    Al = A;
	    lomin = MIN(lomin,K(i,Al));
	    lomax = MAX(lomax,K(i+2,Al));
	} else {
	    HISON(A) = Ah;
	    Ah = A;
	    himin = MIN(himin,K(i,Ah));
	    himax = MAX(himax,K(i + 2,Ah));
	}
    }

    LOMINBOUND(nroot) = MIN(lomin,K(i,nroot));
    HIMAXBOUND(nroot) = MAX(himax,K(i+2,nroot));

    if(j & 2) 
      OTHERBOUND(nroot) = MIN(himin,K(i,nroot));
    else
      OTHERBOUND(nroot) = MAX(lomax,K(i+2,nroot));

    j = NEXTDISC(j);

    tmp = kdglobal.kdrecdef[i+2];
    kdglobal.kdrecdef[i+2] = m;
    LOSON(nroot) = _kd_optimize(Al,j);

    kdglobal.kdrecdef[i+2] = tmp;
    tmp = kdglobal.kdrecdef[i];
    kdglobal.kdrecdef[i] = m;
    HISON(nroot) = _kd_optimize(Ah,j);
    kdglobal.kdrecdef[i] = tmp;

    return(nroot);
}

/***********************************************************************/

kdtree_region_search(region,P)
    kdkey_t region[];
    kdtree_t *P;
{
    register i;

    for(i = 0;i < DIMENSIONALITY;i++)
      kdglobal.kdrecdef[i] = region[i];

#ifdef OLD
    kd_tos = 0;
#else
    REGION_LIST = NULL;
#endif

    _kd_region_search(P,0);
}

/***********************************************************************/

static void _kd_region_search(p,j)
    register kdtree_t *p;
    register j;
{
    register i;

    if(kd_touches_region(p)) kd_found(p);
    i = j&01;

    if((LOSON(p) != NULL) && 
       ((((j&2) == 0) &&
	 ((kdglobal.kdrecdef[i] < OTHERBOUND(p)) &&
	  (kdglobal.kdrecdef[i+2] > LOMINBOUND(p))))
	||
	((j&2) &&
	 ((kdglobal.kdrecdef[i] < K(j,p)) && 
	  (kdglobal.kdrecdef[i+2] > LOMINBOUND(p))))))
      _kd_region_search(LOSON(p),NEXTDISC(j));

    if((HISON(p) != NULL) && 
       ((((j&2) == 0) &&
	 ((kdglobal.kdrecdef[i] < HIMAXBOUND(p)) &&
	  (kdglobal.kdrecdef[i+2] > K(j,p))))
	||
	((j&2) &&
	 ((kdglobal.kdrecdef[i] < HIMAXBOUND(p)) &&
	  (kdglobal.kdrecdef[i+2] > OTHERBOUND(p))))))
      _kd_region_search(HISON(p),NEXTDISC(j));
}

/***********************************************************************/

static kd_touches_region(p)
    register kdtree_t *p;
{
    if(K(0,p) >= kdglobal.kdrecdef[2] ||    
       K(1,p) >= kdglobal.kdrecdef[3] ||
       K(2,p) <= kdglobal.kdrecdef[0] ||
       K(3,p) <= kdglobal.kdrecdef[1])
      return(FALSE);
    return(TRUE);
}

/***********************************************************************/

kd_found(p)
register kdtree_t *p;
{
#ifdef OLD
    REGION_LIST[kd_tos++] = p->id;
#else
    register ppathList pl2;

    pl2 = new_pathList();
    pl2->id = p->id;
    pl2->next = REGION_LIST;
    REGION_LIST = pl2;

#endif
}

/***********************************************************************/
/***********************************************************************/

octObject network;

/***********************************************************************/

kdget_pathBB(path,rect)
    octObject *path;
    kdkey_t rect[];
{
    octGenerator term_gen;
    octObject term1,term2;
    struct octBox *box1,*box2,*get_term_box3();
    struct octPoint p[2];

    OCTCHKOK(octInitGenContents(path,OCT_TERM_MASK,&term_gen));
    OCTCHKOK(octGenerate(&term_gen,&term1));
    OCTCHKOK(octGenerate(&term_gen,&term2));

    OCTCHKOK(octFreeGenerator(&term_gen));

    box1 = get_term_box3(&term1,path);
    box2 = get_term_box3(&term2,path);

    get_path2(path,p);
    if(direction(p) == HOR)
      {
	  rect[0] = MIN(box1->lowerLeft.x,box2->lowerLeft.x);
	  rect[1] = p[0].y -MAX((path->contents.path.width/2),40);
	  rect[2] = MAX(box1->upperRight.x,box2->upperRight.x);
	  rect[3] = p[0].y+MAX((path->contents.path.width/2),40);
      }
    else
      {
	  rect[0] = p[0].x -MAX((path->contents.path.width/2),40);
	  rect[1] = MIN(box1->lowerLeft.y,box2->lowerLeft.y);
	  rect[2] = p[0].x+MAX((path->contents.path.width/2),40);
	  rect[3] = MAX(box1->upperRight.y,box2->upperRight.y);
      }
    FREE_MEM(1,box1,struct octBox);
    FREE_MEM(1,box2,struct octBox);
}

/***********************************************************************/

kdRegObjStart(box)
struct octBox *box;
{
    kdkey_t rect[4];

    if(KDTREE_ROOT == NULL)
      {
	  make_search_tree();
#ifdef OLD
	  REGION_LIST = CALLOC(10000,int);
	  kd_tos = 0;
#else
	  REGION_LIST = NULL;
#endif
      }

    rect[0] = box->lowerLeft.x;
    rect[1] = box->lowerLeft.y;
    rect[2] = box->upperRight.x;
    rect[3] = box->upperRight.y;

    kdtree_region_search(rect,KDTREE_ROOT);

#ifdef OLD
    if(kd_tos == 0)
      return(REG_NOMORE);
#else
    if(REGION_LIST == NULL)
      return(REG_NOMORE);
#endif

    return(REG_OK);
}

/***********************************************************************/

kdRegObjNext(obj)
octObject *obj;
{
    ppathList pl;

#ifdef OLD
    if(kd_tos == 0)
      return(REG_NOMORE);
#else
    if(REGION_LIST == NULL)
      return(REG_NOMORE);
#endif

#ifdef OLD
    obj->objectId = REGION_LIST[--kd_tos];
#else
    pl = REGION_LIST;
    REGION_LIST = REGION_LIST->next;
    obj->objectId = pl->id;
    FREE_MEM(1,pl,struct pathList);
#endif
    OCTCHKOK(octGetById(obj));

    return(REG_OK);
}

/***********************************************************************/

make_search_tree()
{
    octGenerator path_gen,net_gen;
    octObject path,net,layer;
    kdtree_t *list,*last,*kdt;
    kdkey_t brect[4];
    struct octBox box;

#ifndef QUIET
    printf("\n%.2f make search tree\n",tot_time());
#endif
    last = list = NULL;
    OCTCHKOK(octInitGenContents(&network,OCT_NET_MASK,&net_gen));
    while(octGenerate(&net_gen,&net) == OCT_OK)
      {
	  OCTCHKOK(octInitGenContents(&net,OCT_PATH_MASK,&path_gen));
	  while(octGenerate(&path_gen,&path) == OCT_OK)
	    {
		OCTCHKOK(octGenFirstContainer(&path,OCT_LAYER_MASK,&layer));
		if(which_layer(&layer) > POLY)
		  {
		      kdt = new_kdtree_node();
		      kdget_pathBB(&path,kdt->kd_keys);
		      kdt->id = path.objectId;
		      if(list == NULL)
			list = kdt;
		      else
			{
			    HISON(last) = kdt;
			}
		      last = kdt;
		  }
	    }
	  OCTCHKOK(octFreeGenerator(&path_gen));
      }

    OCTCHKOK(octBB(&network,&box));
    brect[0] = box.lowerLeft.x;
    brect[1] = box.lowerLeft.y;
    brect[2] = box.upperRight.x;
    brect[3] = box.upperRight.y;
    KDTREE_ROOT = kdtree_optimize(brect,list);

#ifndef QUIET
    printf("%.2f tree made\nchecking tree ...",tot_time());
    fflush(stdout);
#endif

    kd_ckeck_tree(KDTREE_ROOT,0);
#ifndef QUIET
    printf("done %.2f\n",tot_time());
#endif
}

/***********************************************************************/

kd_ckeck_tree(kdt,j)
    kdtree_t *kdt;
    int j;
{
    if(kdt == NULL)
      return;
    kd_check_node_hi(HISON(kdt),j,K(j,kdt));
    kd_check_node_lo(LOSON(kdt),j,K(j,kdt));
    kd_ckeck_tree(HISON(kdt),NEXTDISC(j));
    kd_ckeck_tree(LOSON(kdt),NEXTDISC(j));
}

/*******************************************************************/

kd_check_node_hi(kdt,j,val)
    kdtree_t *kdt;
    int j,val;
{
    if(kdt == NULL)
      return;

    if(K(j,kdt) < val)
      printf("hi tree err\n");
    kd_check_node_hi(HISON(kdt),j,val);
    kd_check_node_hi(LOSON(kdt),j,val);
}

/*******************************************************************/

kd_check_node_lo(kdt,j,val)
    kdtree_t *kdt;
    int j,val;
{
    if(kdt == NULL)
      return;

    if(K(j,kdt) > val)
      printf("lo tree err\n");
    kd_check_node_lo(HISON(kdt),j,val);
    kd_check_node_lo(LOSON(kdt),j,val);
}

/*******************************************************************/

kd_kill_tree()
{
    _kd_kill_tree(KDTREE_ROOT);
}


/*******************************************************************/

_kd_kill_tree(kdt)
kdtree_t *kdt;
{
    if(kdt == NULL)
      return;
    _kd_kill_tree(HISON(kdt));
    _kd_kill_tree(LOSON(kdt));
    FREE_MEM(1,kdt,kdtree_t);
}

new_box(kdt)
    kdtree_t *kdt;
{
    octObject obj,layer;

    printf("box %d %d %d %d\n",
	   kdt->kd_keys[0],
	   kdt->kd_keys[1],
	   kdt->kd_keys[2],
	   kdt->kd_keys[3]);

    obj.type = OCT_BOX;
    obj.contents.box.lowerLeft.x = kdt->kd_keys[0];
    obj.contents.box.lowerLeft.y = kdt->kd_keys[1];
    obj.contents.box.upperRight.x = kdt->kd_keys[2];
    obj.contents.box.upperRight.y = kdt->kd_keys[3];
    OCTCHKOK(octCreate(&network,&obj));
    layer.type = OCT_LAYER;
    layer.contents.layer.name = "NDIFF";
    OCTCHKOK(octGetOrCreate(&network,&layer));
    OCTCHKOK(octAttach(&layer,&obj));
}

struct octBox *get_term_box3(term,path)
    octObject *term,*path;
{
    octObject net,inst,*master,term2,box3,*getMaster();
    struct octBox *box;
    struct octPoint p[2];

    box = CALLOC(1,struct octBox);
    if(octGenFirstContainer(term,OCT_NET_MASK,&net) == OCT_OK)
      {
    
	  OCTCHKOK(octGenFirstContainer(term,OCT_INSTANCE_MASK,&inst));

	  master = getMaster(inst.contents.instance.master,inst.contents.instance.view);
    
	  term2.type = OCT_TERM;
	  term2.contents.term.name = term->contents.term.name;
	  OCTCHKOK(octGetByName(master,&term2));

	  get_term_BB(&term2,&box3);
	  octTransformGeo(&box3,&inst.contents.instance.transform);

	  box->lowerLeft.x = box3.contents.box.lowerLeft.x;
	  box->lowerLeft.y = box3.contents.box.lowerLeft.y;
	  box->upperRight.x = box3.contents.box.upperRight.x;
	  box->upperRight.y = box3.contents.box.upperRight.y;
	  get_path2(path,p);
	  if(direction(p) == HOR)
	    {
		if(box->lowerLeft.x > MIN(p[0].x,p[1].x))
		  {
		      box->upperRight.x = MAX(p[0].x,p[1].x);
		  }
		else
		  {
		      box->lowerLeft.x = MIN(p[0].x,p[1].x);
		  }
	    }
	  else
	    {
		if(box->lowerLeft.y > MIN(p[0].y,p[1].y))
		  {
		      box->upperRight.y = MAX(p[0].y,p[1].y);
		  }
		else
		  {
		      box->lowerLeft.y = MIN(p[0].y,p[1].y);
		  }
	    }
      }
    else
      {
	  OCTCHKOK(octGenFirstContainer(term,OCT_INSTANCE_MASK,&inst));
	  OCTCHKOK(octBB(&inst,box));
      }

    return(box);
}
