#include "copyright.h"
#include "port.h"

#include "oct.h"

#define OCTCHKOK(octCall) \
{ \
if (octCall!=OCT_OK) \
    octError("Failed: \"octCall==OCT_OK\""); \
}

/*******************************************************************/

openFacet(octFacet, cell, view, name, mode)
octObject *octFacet;
char *cell;
char *view;
char *name;
char *mode;
{

    octFacet->type = OCT_FACET;
    octFacet->contents.facet.cell    = cell;
    octFacet->contents.facet.view    = view;
    octFacet->contents.facet.facet   = name;
    octFacet->contents.facet.mode    = mode;
    octFacet->contents.facet.version = OCT_CURRENT_VERSION;
    if(octOpenFacet(octFacet) < OCT_OK)
      {
	  octError("Failed: octOpenFacet(octFacet)");
	  exit(9);
      }
}

/*******************************************************************/

openOldFacet(octFacet, cell, view, name, mode)
octObject *octFacet;
char *cell;
char *view;
char *name;
char *mode;
{
    int stat;

    octFacet->type = OCT_FACET;
    octFacet->contents.facet.cell    = cell;
    octFacet->contents.facet.view    = view;
    octFacet->contents.facet.facet   = name;
    octFacet->contents.facet.mode    = mode;
    octFacet->contents.facet.version = OCT_CURRENT_VERSION;
    if((stat = octOpenFacet(octFacet)) < OCT_OK)
      {
	  printf("ERROR %s:%s not found\n",cell,view);
	  exit(1);
      }
    if(stat == OCT_NEW_FACET)
      {
	  printf("ERROR %s:%s not found\n",cell,view);
	  exit(1);
      }
}

/*******************************************************************/

octId newInstance( Facet, name, master, view, version, transform )
octObject *Facet;
char *name;
char *master;
char *view;
char *version;
enum octTransformType transform;
{
    octObject instbag;
    octObject inst;

    inst.type = OCT_INSTANCE;
    inst.contents.instance.name = name;
    inst.contents.instance.master = master;
    inst.contents.instance.view = view;
    inst.contents.instance.version = version;
    inst.contents.instance.transform.translation.x = 0;
    inst.contents.instance.transform.translation.y = 0;
    inst.contents.instance.transform.transformType = transform;

    OCTCHKOK( octCreate( Facet, &inst ));

    instbag.type = OCT_BAG;
    instbag.contents.bag.name = "INSTANCES";
    OCTCHKOK( octGetByName(Facet,&instbag));
    OCTCHKOK( octAttach( &instbag, &inst ));

return(inst.objectId);
}

/*******************************************************************/

octId newNet( facet, name)
octObject *facet;
char *name;
{
octObject net;

net.type = OCT_NET;
net.contents.net.name = name;
OCTCHKOK(octCreate(facet,&net));
return(net.objectId);
}

/*******************************************************************/

set_inst_loc(Id,x,y,trans)
octId Id;
octCoord x,y;
octTransformType trans;
{
octObject inst;

inst.type = OCT_INSTANCE;
inst.objectId = Id;
OCTCHKOK(octGetById(&inst));
/*
printf("%s oct name %d %d \n",inst.contents.instance.name,x,y);
*/
inst.contents.instance.transform.translation.x = x;
inst.contents.instance.transform.translation.y = y;
inst.contents.instance.transform.transformType = trans;
OCTCHKOK(octModify(&inst));
}

/*******************************************************************/

octId newBox( facet, layer, x, y, w, h)
octObject *facet;
octObject *layer;
int x,y,h,w;
{
octObject box;

box.type = OCT_BOX;
box.contents.box.lowerLeft.x = x;
box.contents.box.lowerLeft.y = y;
box.contents.box.upperRight.x = x + w;
box.contents.box.upperRight.y = y + h;
OCTCHKOK(octCreate(facet,&box));
OCTCHKOK(octAttach(layer,&box));
return(box.objectId);
}

/*******************************************************************/

newLayer( facet, layer, name)
octObject *facet;
char *name;
octObject *layer;
{

layer->type = OCT_LAYER;
layer->contents.layer.name = name;
OCTCHKOK(octCreate(facet,layer));
}

/*******************************************************************/

octId oldLayer( facet, name)
octObject *facet;
char *name;
{
octObject layer;

layer.type = OCT_LAYER;
layer.contents.layer.name = name;
OCTCHKOK(octGetByName(facet,&layer));
return(layer.objectId);
}

/*******************************************************************/

octId newBag( obj, name)
octObject *obj;
char *name;
{
octObject bag;

bag.type = OCT_BAG;
bag.contents.bag.name = name;
OCTCHKOK(octCreate(obj,&bag));
return(bag.objectId);
}

/*******************************************************************/

split_oct_name(in,cell,view)
char *in,**cell,**view;
{
int i,j;

for(i = 0;(in[i] != '\0') && (in[i] != ':');i++)
  {}
if((in[i] == '\0') || (i == 0))
  {
    return(0);
  }
else
  {
    in[i] = '\0';
    (*cell) = (char*)malloc(sizeof(char) * (i + 1));
    (void) strcpy((*cell),in);
  }

for(j=1;(in[i + j] != '\0');j++)
  {}
if(j == 1)
  {
    return(0);
  }
else 
  {
    (*view) = (char*)malloc(sizeof(char) * j);
    (void) strcpy((*view),&(in[i+1]));
  }
/*
printf("%s:%s\n",(*cell),(*view));
*/
return(1);
}

/*******************************************************************/

char *oct_find_str_prop(obj,prop_name)
octObject *obj;
char *prop_name;
{
octGenerator p_gen;
octObject prop;
octStatus stat;

OCTCHKOK(octInitGenContents(obj,OCT_PROP_MASK,&p_gen));
while(((stat = octGenerate(&p_gen,&prop)) == OCT_OK) && (strcmp(prop.contents.prop.name,prop_name) != 0))
  {}
if(stat != OCT_OK)
  {
    char str[50];

    (void) sprintf(str,"could not find oct property %s",prop_name);
    octError(str);
    return("");
  }
return(prop.contents.prop.value.string);
}

/*******************************************************************/

oct_find_bag_by_type(obj,bag,name)
octObject *obj,*bag;
char *name;
{
octGenerator gen;
octStatus stat;

OCTCHKOK(octInitGenContents(obj,OCT_BAG_MASK,&gen));
while(((stat = octGenerate(&gen,bag)) == OCT_OK) && (strcmp(oct_find_str_prop(bag,"TYPE"),name) != 0))
  {}
if(stat != OCT_OK)
  {
    char str[50];

    (void) sprintf(str,"could not find bag %s",name);
    octError(str);
    return(0);
  }
return(1);
}

/*******************************************************************/

oct_find_bag_by_name(obj,bag,name)
octObject *obj,*bag;
char *name;
{
octGenerator gen;
octStatus stat;

OCTCHKOK(octInitGenContents(obj,OCT_BAG_MASK,&gen));
while(((stat = octGenerate(&gen,bag)) == OCT_OK) && (strncmp(bag->contents.bag.name,name,strlen(name)) != 0))
  {}
return(stat == OCT_OK);
}

/*******************************************************************/

make_bag(facet,bag,name)
octObject *bag;
octObject *facet;
char *name;
{
  bag->type = OCT_BAG;
  bag->objectId = newBag(facet,name);
  OCTCHKOK(octGetById(bag));
}

/*******************************************************************/

get_bag(bag,id)
octObject *bag;
octId id;
{
  bag->type = OCT_BAG;
  bag->objectId = id;
  OCTCHKOK(octGetById(bag));
}

/*******************************************************************/

octId new_prop(obj,name,string)
octObject *obj;
char *string;
char *name;
{
octObject prop;

prop.type = OCT_PROP;
prop.contents.prop.name = name;
prop.contents.prop.type = OCT_STRING;
prop.contents.prop.value.string = string;
OCTCHKOK(octCreate(obj,&prop));
return(prop.objectId);
}

/*******************************************************************/

newBox2( facet, box, x, y, w, h)
octObject *facet;
octObject *box;
int x,y,h,w;
{

box->type = OCT_BOX;
box->contents.box.lowerLeft.x = x;
box->contents.box.lowerLeft.y = y;
box->contents.box.upperRight.x = x + w;
box->contents.box.upperRight.y = y + h;
OCTCHKOK(octCreate(facet,box));
}
