#include "copyright.h"
#include "port.h"

#include "header.h"
#include "options.h"
#include "utility.h"
#include "vov.h"

     optionStruct optionList[] = {
       {"o",   "output_cell:view", "input oct cell:view"},
       {"f",   "","fast mode (less optimization done)"},
       {"s",   "","check for shorts only"},
       {"L",   "layer_name", "signal layer name"},
       {OPT_RARG, "input_cell:view", "input oct cell:view"},
       {0  ,   0,        0}
     };
char *out_cell_name;
char *out_cell_view;
char *tmp_out_cell_name;
int FAST = FALSE;
int SLOW = FALSE;
int SHORTS = FALSE;

process_command_line(argc,argv)
int argc;
char **argv;
{
    int output,option;
    char *cell;
    char *view;
    int layer_num;

    output = FALSE;
    layer_nameTable[LAYER1] = util_strsav("MET1");
    layer_nameTable[LAYER2] = util_strsav("MET2");
    layer_num = LAYER1;
    while ((option = optGetOption(argc, argv)) != EOF) 
      {
	  switch (option) 
	    {
	    case 'o':
		split_oct_name(util_strsav(optarg),&out_cell_name,&out_cell_view);      
		output = TRUE;
		break;
	    case 'f':
		FAST = TRUE;
		break;
	    case 's':
		SHORTS = TRUE;
		break;
	    case 'L':
		layer_nameTable[layer_num++] = util_strsav(optarg);
		break;
	    default:
		err("command line error\n");
	    }
      }
    if (optind >= argc)
      {
	  optUsage();
      }
    else
      {
	  split_oct_name(util_strsav(argv[optind]),&cell,&view);      
      }

    octBegin();
    VOVbegin( argc, argv );

    if(output == TRUE)
      {
	  octObject old;

	  printf("(MIZER %.2f sec): copying %s:%s to %s:%s\n",
		 tot_time(),
		 cell,view,
		 out_cell_name,out_cell_view);

	  old.type = OCT_FACET;
	  old.contents.facet.cell = cell;
	  old.contents.facet.view = view;
	  old.contents.facet.version = OCT_CURRENT_VERSION;
	  old.contents.facet.facet = "interface";

	  network.type = OCT_FACET;
	  network.contents.facet.cell = out_cell_name;
	  network.contents.facet.view = out_cell_view;
	  network.contents.facet.facet = "interface";
	  network.contents.facet.version = OCT_CURRENT_VERSION;
	  network.contents.facet.mode = "a";
	  if(octCopyFacet(&network,&old) != OCT_OK)
	    printf("(MIZER):WARNING no interface facet exists %s:%s\n",cell,view);

	  network.contents.facet.facet = "contents";
	  old.contents.facet.facet = "contents"; 

	  VOVinputFacet( &old );
	  VOVoutputFacet( &network );
	  
	  if(octCopyFacet(&network,&old) != OCT_OK)
	    {
		printf("(MIZER):ERROR in copying file %s:%s to %s:%s\n",
		       cell, view, out_cell_name, out_cell_view);
		exit(9);	
	    }

	  printf("(MIZER %.2f sec): opening %s:%s\n",tot_time(),
		 out_cell_name,out_cell_view);
	  network.contents.facet.mode = "a";
	  if(octOpenFacet(&network) < OCT_OK)	
	    {
		printf("ERROR in opening new file %s : %s\n",out_cell_name,out_cell_view);
		exit(9);
	    }
      }
    else
      {
	  printf("(MIZER %.2f sec): opening %s:%s\n",tot_time(),cell,view);
	  openFacet(&network, cell, view, "contents", "a");

	  VOVinputFacet( &network );
	  VOVoutputFacet( &network );
	  out_cell_name = cell;
	  out_cell_view = view;
      }
    printf("(MIZER %.2f sec): done opening files\n",tot_time());
}


void
mm_set(a, b)
int a, b;
{
    return;
}
