#include "copyright.h"
#include "port.h"

#include "oct.h"
#include "st.h"
#include "header.h"

extern pcell FIXED_MET1,FIXED_MET2;
extern int ESSENTIAL;
extern int NUM_OF_NETS;
pcell over_connected(),linked_over_connected();
pcell fixed_over_connected();

#define MET1 1
#define MET2 2
extern int CONT_WIDTH;
extern pcell CELLS;
extern pnet NETS;
octObject *getMaster();
ppathList new_pathList();


/**********************************************************************/

remove_3_cycles()
{
pcell cell;
int cnt;

cnt = 0;
for(cell = CELLS;cell != NULL;cell = cell->next)
  {
    if((cell->hard_link != NULL) && (cell->id < cell->hard_link->id))
      {
	while(common_connection(cell,cell->hard_link) == TRUE)
	  {
	    cnt++;
	  }
      }
  }
if(cnt != 0)
  {
#ifndef QUIET
      printf("%d three cycles\n",cnt);
#endif
    ESSENTIAL += cnt;
    return(TRUE);
  }
return(FALSE);
}

/**********************************************************************/

common_connection(cell1,cell2)
pcell cell1,cell2;
{
pnetlist nl,nl2;
pcellist cl,cl2;
static int *touched_array = NULL;
static counter = 0;
int found,i;

counter++;
if(touched_array == NULL)
  {
    touched_array = CALLOC(NUM_OF_CELLS + 1,int);
    for(i = 0;i <= NUM_OF_CELLS;i++)
      {
	touched_array[i] = 0;
      }
  }

for(nl = cell1->nets;nl != NULL;nl = nl->next)
  {
    if(two_pt_connection(nl->net) == TRUE)
      {
	for(cl = nl->net->cells;cl != NULL;cl = cl->next)
	  {
	    touched_array[cl->cell->id] = counter;
	  }
      }
  }
touched_array[cell1->id] = 0;
touched_array[cell2->id] = 0;

found = FALSE;
for(nl2 = cell2->nets;nl2 != NULL;nl2 = nl2->next)
  {
    if(two_pt_connection(nl2->net) == TRUE)
      {
	for(cl2 = nl2->net->cells;cl2 != NULL;cl2 = cl2->next)
	  {
	    if(touched_array[cl2->cell->id] == counter)
	      {
		found = TRUE;
		break;
	      }	
	  }
      }
  }
if(found == TRUE)
  {
    common_connection2(cell1,cell2);
  }
return(found);
}

/**********************************************************************/

common_connection2(cell1,cell2)
pcell cell1,cell2;
{
pnetlist nl,nl2;
pcellist cl,cl2;

for(nl = cell1->nets;nl != NULL;nl = nl->next)
  {
    if(two_pt_connection(nl->net) == TRUE)
      {
	for(nl2 = cell2->nets;nl2 != NULL;nl2 = nl2->next)
	  {
	    if(two_pt_connection(nl2->net) == TRUE)
	      {
		for(cl = nl->net->cells;cl != NULL;cl = cl->next)
		  {
		    for(cl2 = nl2->net->cells;cl2 != NULL;cl2 = cl2->next)
		      {
			if(cl->cell == cl2->cell)
			  {
			    kill_net(nl->net);
			    kill_net(nl2->net);
			    return;
			  }
		      }
		  }
	      }
	  }
      }
  }
}

/**********************************************************************/

remove_terminal_conflicts()
{
pcell cell;
int cnt;

cnt = 0;
for(cell = CELLS;cell != NULL;cell = cell->next)
  {
    if((cell->hard_link != NULL) && (cell->id < cell->hard_link->id))
      {
	cnt += kill_terminal_conflict(cell);
      }
  }
if(cnt != 0)
  {
#ifndef QUIET
      printf("%d terminal conflicts found\n",cnt);
#endif
    ESSENTIAL += cnt;
    return(TRUE);
  }
return(FALSE);
}

/**********************************************************************/

kill_terminal_conflict(cell)
pcell cell;
{
int layer1,layer2;
pnet net1,net2;

layer1 = is_constrained_to_layer(cell,&net1);
layer2 = is_constrained_to_layer(cell->hard_link,&net2);
if((layer1 != 0) && (layer2 != 0))
  {
    if(layer1 == layer2)
      {
	kill_net(net1);
	kill_net(net2);
	return(1);
      }
  }
return(0);
}

/**********************************************************************/

remove_terminal_overconstraints()
{
pcell cell;
int cnt;

cnt = 0;
for(cell = CELLS;cell != NULL;cell = cell->next)
  {
    cnt += kill_terminal_overconstraint(cell);
  }
if(cnt != 0)
  {
#ifndef QUIET
      printf("%d terminal over constriaints found\n",cnt);
#endif
    ESSENTIAL += cnt;
    return(TRUE);
  }
return(FALSE);
}

/**********************************************************************/

kill_terminal_overconstraint(cell)
pcell cell;
{
pnet net1,net2;

if((is_constrained_to_a_particular_layer(cell,&net1,MET1) == TRUE) && 
   (is_constrained_to_a_particular_layer(cell,&net2,MET2) == TRUE))
  {
    kill_net(net1);
    kill_net(net2);
    return(1);
  }
return(0);
}

/**********************************************************************/

is_constrained_to_a_particular_layer(cell,net,layer)
pcell cell;
pnet *net;
int layer;
{
pnetlist nl;
pcellist cl;

for(nl = cell->nets;nl != NULL;nl = nl->next)
  {
    for(cl = nl->net->cells;cl != NULL;cl = cl->next)
      {
	if((cl->cell->fixed == TRUE) && 
	   (cl->cell->side == layer) && 
	   (two_pt_connection(nl->net) == TRUE))
	  {
	    (*net) = nl->net;
	    return(TRUE);
	  }
      }
  }
return(FALSE);
}

/**********************************************************************/

is_constrained_to_layer(cell,net)
pcell cell;
pnet *net;
{pnetlist nl;
pcellist cl;

for(nl = cell->nets;nl != NULL;nl = nl->next)
  {
    for(cl = nl->net->cells;cl != NULL;cl = cl->next)
      {
	if((cl->cell->fixed == TRUE) && (two_pt_connection(nl->net) == TRUE))
	  {
	    (*net) = nl->net;
	    return(cl->cell->side);
	  }
      }
  }
return(0);
}

/**********************************************************************/

merge_simple_nodes(list)
pcell list;
{
pcell cell, c;
pcell best,dummy;
int found;

found = FALSE;
for(cell = list;cell != NULL;)
  {
    c = cell->next;
    if((cell->hard_link == NULL) && 
       ((best = over_connected(cell,&dummy)) != NULL))
      {
	merge_two_nodes(best, cell);
	clean_up(best, NULL);
	found = TRUE;
      }
    cell = c;
  }
return(found);
}

/**********************************************************************/

merge_two_nodes(cell3,cell2)
pcell cell3,cell2;
{
pnetlist nl,nl2,nl3,nl4;
pcellist cl,cl2;
ppathList pl;

/*
printf("merging %d and %d\n",cell3->id,cell2->id);
*/

for(nl = cell2->nets;nl != NULL;nl = nl->next)
  {
    for(cl = nl->net->cells;(cl != NULL) && (cl->cell != cell2);cl = cl->next)
      {}
    if(cl == NULL)
      {
	printf("ERROR - net %d only connected to cell %d\n",nl->net->id,cell2->id);
	quit();
      }
    cl->cell = cell3;
  }

for(nl = NULL,nl2 = cell2->nets;nl2 != NULL;)
  {
    for(nl3 = cell3->nets;
	(nl3 != NULL) && 
	(nl2->net != nl3->net);nl3 = nl3->next)
      {}
    nl4 = nl2;
    nl2 = nl2->next;
    if(nl3 == NULL)
      {
	nl4->next = nl;
	nl = nl4;
      }
    else
      {
	if(nl4->net->cells->cell == cell3)
	  {
	    cl = nl4->net->cells;
	    nl4->net->cells = nl4->net->cells->next;
	  }
	else
	  {
	    for(cl2 = nl4->net->cells;
		(cl2 != NULL) && 
		(cl2->next != NULL) &&
		(cl2->next->cell != cell3);cl2 = cl2->next)
	      {}
	    if((cl2 == NULL) ||	(cl2->next == NULL))
	      {
		printf("ERROR in deleting net connection\n");
		quit();
	      }
	    cl = cl2->next;
	    cl2->next = cl2->next->next;
	  }
	if((nl4->net->cells == NULL) ||
	   (nl4->net->cells->next == NULL))
	  kill_net(nl4->net);
	else
	  {
	    if(nl4->net->cells->next->next == NULL)
	      nl4->net->two_conn = TRUE;
	  }
	FREE_MEM(1, cl, struct cellist);
	FREE_MEM(1, nl4, struct netlist);
      }
  }
if(nl != NULL)
  {
    for(nl2 = nl;(nl2 != NULL) && (nl2->next != NULL);nl2 = nl2->next)
      {}
    nl2->next = cell3->nets;
    cell3->nets = nl;
  }

if(cell3->paths == NULL)
  cell3->paths = cell2->paths;
else
  {
    for(pl = cell3->paths;(pl != NULL) && (pl->next != NULL);pl = pl->next)
      {}
    pl->next = cell2->paths;
  }

remove_from_cell_list(cell2,&CELLS);
}

/**********************************************************************/

merge_linked_nodes(list)
pcell list;
{
pcell cell,best,c,next;
int cnt;

cnt = 0;
for(cell = list;cell != NULL;)
  {
    if((cell->hard_link != NULL) && (cell->id > cell->hard_link->id) && 
       ((best = over_connected(cell,&c)) != NULL))
      {
	next = cell->next;
	if(best->hard_link == NULL)
	  {
	    if(next == best)
	      next = next->next;
	    merge_two_nodes(c,best);
	    clean_up(c, NULL);
	  }
	else
	  {
	    if(next == c)
	      {
		next = next->next;
		if(next == c->hard_link)
		  next = next->next;
	      }
	    else
	      if(next == c->hard_link)
		{
		  next = next->next;
		  if(next == c)
		    next = next->next;
		}
	    merge_two_nodes(best->hard_link,c->hard_link);	    
	    merge_two_nodes(best,c);	    
	    clean_up(best, best->hard_link);
	  }
	cnt++;
      }
    else
      next = cell->next;
    cell = next;
  }
if(cnt != 0)
  {
    return(TRUE);
  }
return(FALSE);
}

/*************************************************************************/

num_of_two_point_conn(cell1,cell2)
pcell cell1,cell2;
{
pnetlist nl;
int cnt;

cnt = 0;
for(nl = cell1->nets;nl != NULL;nl = nl->next)
  {
    if(two_pt_connection(nl->net) == TRUE)
      {
	if((nl->net->cells->cell == cell2) || 
	   (nl->net->cells->next->cell == cell2))
	  cnt++;
      }
  }
return(cnt);
}

/**********************************************************************/

num_of_non_two_point_conn(cell1,cell2)
pcell cell1,cell2;
{
pnetlist nl;
int cnt;
pcellist cl;

cnt = 0;
for(nl = cell1->nets;nl != NULL;nl = nl->next)
  {
    if(two_pt_connection(nl->net) == FALSE)
      {
	for(cl = nl->net->cells;cl != NULL;cl = cl->next)
	  if(cl->cell == cell2)
	    {
	      cnt++;
	      break;
	    }
      }
  }
return(cnt);
}

/**********************************************************************/

pcell over_connected(cell, returnCell)
pcell cell;
pcell *returnCell;
{
int i,id;
int tot,max;
pcell best,c;
pnetlist nl;
pcellist cl;
int n, r, p, s, m, q, u, v, w;
static pcell *cell_array = NULL;
static int *touched_array = NULL;
static counter = 0;
int tos;
int diff;

if((cell->nets == NULL) && 
   ((cell->hard_link == NULL) || (cell->hard_link->nets == NULL)))
  {
    return(NULL);
  }

counter++;

if(cell_array == NULL)
  {
    cell_array = CALLOC(NUM_OF_CELLS + 1,pcell);
    touched_array = CALLOC(NUM_OF_CELLS + 1,int);
    for(i = 0;i <= NUM_OF_CELLS;i++)
      {
	cell_array[i] = NULL;
	touched_array[i] = 0;
      }
  }

tot = 0;
tos = 0;
for(nl = cell->nets;nl != NULL;nl = nl->next)
  {
    tot++;
    for(cl = nl->net->cells;cl != NULL;cl = cl->next)
      {
	if((cl->cell != cell) && (cl->cell != cell->hard_link))
	  {
	    id = cl->cell->id;
	    if((cell->hard_link != NULL) && (cl->cell->hard_link != NULL))
	      {
		id = MAX(cl->cell->id,cl->cell->hard_link->id);
	      }
	    if(touched_array[id] != counter)
	      {
		touched_array[id] = counter;
		cell_array[tos++] = cl->cell;
	      }
	  }
      }
  }
if(cell->hard_link != NULL)
  for(nl = cell->hard_link->nets;nl != NULL;nl = nl->next)
    {
      tot++;
      for(cl = nl->net->cells;cl != NULL;cl = cl->next)
	{
	  if((cl->cell != cell) && (cl->cell != cell->hard_link))
	    {
	      if(cl->cell->hard_link == NULL)
		id = cl->cell->id;
	      else
		{
		  id = MAX(cl->cell->id,cl->cell->hard_link->id);
		}
	      if(touched_array[id] != counter)
		{
		  touched_array[id] = counter;
		  cell_array[tos++] = cl->cell;
		}
	    }
	}
    }

max = 0;
best = NULL;
for(i = 0;i < tos;i++)
  {
    c = cell_array[i];

    v = w = n = r = p = s = m = q = u = 0;
    n = num_of_two_point_conn(cell,c);
    q = num_of_non_two_point_conn(cell,c);
    if(c->hard_link != NULL)
      {
	w = num_of_two_point_conn(cell,c->hard_link);
	s = num_of_non_two_point_conn(cell,c->hard_link);
      }
    if(cell->hard_link != NULL)
      {
	v = num_of_two_point_conn(cell->hard_link,c);
	p = num_of_non_two_point_conn(cell->hard_link,c);
	if(c->hard_link != NULL)
	  {
	    r = num_of_two_point_conn(cell->hard_link,c->hard_link);
	    u = num_of_non_two_point_conn(cell->hard_link,c->hard_link);
	  }
      }
    m = tot - (n + r + p + q + s + u + v + w);
    diff = (n + r ) - (m + p + s + v + w);
    if(diff >= 0)
      {
	if((best == NULL) || 
	   (diff > max) || 
	   ((diff == max) && ((c == FIXED_MET1) || (c == FIXED_MET2))))
	  {
	    (*returnCell) = cell;
	    best = c;
	    max = diff;
	  }
      }
    
    if(cell->hard_link != NULL)
      {
	diff = (v + w) - (m + n + r + q + u);
	if(diff >= 0)
	  {
	    if((best == NULL) || 
	       (diff > max) || 
	       ((diff == max) && ((c == FIXED_MET1) || (c == FIXED_MET2))))
	      {
		(*returnCell) = cell->hard_link;
		best = c;
		max = diff;
	      }
	  }
      }
  }
return(best);
}

/**********************************************************************/

clean_up_node(cell)
pcell cell;
{

while(check_for_essential(cell) == TRUE)
  {
    ESSENTIAL++;
  }
if(cell->hard_link != NULL)
  {
    while(common_connection(cell,cell->hard_link) == TRUE)
      {
	ESSENTIAL++;
      }
  }
}

/**********************************************************************/

clean_up(cell,cell2)
pcell cell,cell2;
{
pnetlist nl;
pcellist cl;
static pcell *cell_array = NULL;
static int *touched_array = NULL;
static counter = 0;
int tos,i,id;

clean_up_node(cell);
if(cell->hard_link != NULL)
  clean_up_node(cell->hard_link);

counter++;

if(cell_array == NULL)
  {
    cell_array = CALLOC(NUM_OF_CELLS + 1,pcell);
    touched_array = CALLOC(NUM_OF_CELLS + 1,int);
    for(i = 0;i <= NUM_OF_CELLS;i++)
      {
	cell_array[i] = NULL;
	touched_array[i] = 0;
      }
  }

tos = 0;
touched_array[cell->id] = counter;
for(nl = cell->nets;nl != NULL;nl = nl->next)
  {
    for(cl = nl->net->cells;cl != NULL;cl = cl->next)
      {
	id = cl->cell->id;
	if(touched_array[id] != counter)
	  {
	    touched_array[id] = counter;
	    cell_array[tos++] = cl->cell;
	  }
      }
  }
if(cell2 != NULL)
  {
    touched_array[cell2->id] = counter;
    for(nl = cell2->nets;nl != NULL;nl = nl->next)
      {
	for(cl = nl->net->cells;cl != NULL;cl = cl->next)
	  {
	    id = cl->cell->id;
	    if(touched_array[id] != counter)
	      {
		touched_array[id] = counter;
		cell_array[tos++] = cl->cell;
	      }
	  }
      }
  }

for(i = 0;i < tos;i++)
  {
   clean_up_node(cell_array[i]);
  }
}

/**********************************************************************/

swap(a,b)
register int *a,*b;
{
register int tmp;

tmp = *a;
(*a) = (*b);
(*b) = tmp;
}
