#include "copyright.h"
#include "port.h"

#include "header.h"
#include "tap.h"

#define FATAL 1

extern int  INPUT_SCALE;
extern int  OUTPUT_SCALE;

static char errBuf[80];
char *oct_find_str_prop();
char    *TECH_DIR;

char *get_tech_name(){
char *techName;

if(oct_find_str_prop(&network,"TECHNOLOGY") == NULL)
  logError("Could not find TECHNOLOGY property\n",0);
TECH_DIR = tapGetDirectory(&network);
printf("(MIZER %.2f sec): technology %s\n",tot_time(),TECH_DIR);

for(techName = TECH_DIR + strlen(TECH_DIR) - 1;(techName > TECH_DIR) && (*techName != '/');techName--)
  {}
techName++;
return(techName);
}

set_up_width_and_spacing(layers)
octObject layers[];
{
  layer_width[LAYER1] = tapGetMinWidth(&(layers[LAYER1]));
  layer_width[LAYER2] = tapGetMinWidth(&(layers[LAYER2]));
  
  layer_spacing[LAYER1] = tapGetMinSpacing(&(layers[LAYER1]),&(layers[LAYER1]));
  layer_spacing[LAYER2] = tapGetMinSpacing(&(layers[LAYER2]),&(layers[LAYER2]));
  if((layer_width[LAYER1] == -1) || 
     (layer_width[LAYER2] == -1) || 
     (layer_spacing[LAYER1] == -1) || 
     (layer_spacing[LAYER2] == -1))
    logError("(tap_stuff): error in getting layer widths and spacing\n",FATAL);
}

get_connector_BB(box,layerOne,layerTwo,layerOne_dir,layerTwo_dir)
struct octBox *box;
int layerOne,layerTwo,layerOne_dir,layerTwo_dir;
{
    octObject mast,conn;
    octStatus stat;

    get_connector_info(&mast,layerOne,layerTwo,layerOne_dir,layerTwo_dir);
    /*
      printf("master %s view %s\n",mast.contents.instance.master,mast.contents.instance.view);
      */

    if((stat = openFacet2(&conn,mast.contents.instance.master,mast.contents.instance.view,"contents",OCT_CURRENT_VERSION,"r")) < OCT_OK)
      {
	  (void) sprintf(errBuf,"(tap_stuff): ERROR in opening via master -- status %d\n",stat);
	  logError(errBuf,FATAL);
      }

    if(octBB(&conn,box) != OCT_OK)
      {
	  (void) sprintf(errBuf,"(tap_stuff): ERROR in getting bound box for layer %d to %d connector\n",layerOne,layerTwo);
	  logError(errBuf,FATAL);
      }
    /*
      printf("%d %d %d %d del %d %d\n",
      box->lowerLeft.x,
      box->lowerLeft.y,
      box->upperRight.x,
      box->upperRight.y,
      box->upperRight.x - box->lowerLeft.x,
      box->upperRight.y - box->lowerLeft.y);
      */
}

set_up_layers(layers)
octObject layers[];
{
  layers[LAYER1].type = OCT_LAYER;
  layers[LAYER1].contents.layer.name = layer_nameTable[LAYER1];
  OCTCHKOK(octGetOrCreate(&network,&(layers[LAYER1])));
  layers[LAYER2].type = OCT_LAYER;
  layers[LAYER2].contents.layer.name = layer_nameTable[LAYER2];
  OCTCHKOK(octGetOrCreate(&network,&(layers[LAYER2])));
}

get_connector_info(master,layerOne,layerTwo,layerOne_dir,layerTwo_dir)
int layerOne,layerTwo,layerOne_dir,layerTwo_dir;
octObject *master;
{
    tapLayerListElement layerlist[2];
    octStatus stat;

    layerlist[0].layer.type = OCT_LAYER;
    layerlist[0].layer.contents.layer.name = layer_nameTable[layerOne];
    OCTCHKOK(octGetOrCreate(&network,&(layerlist[0].layer)));
    layerlist[1].layer.type = OCT_LAYER;
    layerlist[1].layer.contents.layer.name = layer_nameTable[layerTwo];
    OCTCHKOK(octGetOrCreate(&network,&(layerlist[1].layer)));

    layerlist[0].width = layer_width[layerOne];
    layerlist[1].width = layer_width[layerTwo];
    layerlist[0].direction = layerOne_dir;
    layerlist[1].direction = layerTwo_dir;

    /*
      printf("layer %s width %d dir %d\n",
      layerlist[0].layer.contents.layer.name,
      layerlist[0].width,
      layerlist[0].direction);
      printf("layer %s width %d dir %d\n",
      layerlist[1].layer.contents.layer.name,
      layerlist[1].width,
      layerlist[1].direction);
      */
    if((stat = tapGetConnector(2,layerlist,master)) != OCT_OK)      {
	(void) sprintf(errBuf,"(tap_stuff): ERROR in getting connector-- status %d\n",stat);
	logError(errBuf,FATAL);
    }
}

logError(str,stat)
char *str;
int stat;
{
    printf("%s\n",str);
    if(stat == FATAL)
      exit(5);
}


octStatus openFacet2(obj,cell,view,facet,version,mode)
char *cell,*view,*facet,*mode,*version;
octObject *obj;
{
    obj->type = OCT_FACET;
    obj->contents.facet.cell = cell;
    obj->contents.facet.view = view;
    obj->contents.facet.facet = facet;
    obj->contents.facet.version = version;
    obj->contents.facet.mode = mode;

    return(octOpenFacet(obj));
}
