#include "cpmInternal.h"
/*
 *	Level-oriented scheduler for nodes
 */

/*
 *	Global variables
 */
static struct cpmNode **LFirstNode, **LLastNode;
static struct cpmNode *LschedEOL;
static int L_sched_count=0, SchedMaxLevels=0, ActiveNodes=0;

void
Lschedule ( node )
struct cpmNode *node;
{
	if(node->level > SchedMaxLevels || node->level < 0) {
		fprintf(stderr,"Lschedule: level %d at %s gt %d\n",
			node->level, node->name, SchedMaxLevels);
		exit(0);
	}

        if(node->sched != NULL) return; /* already scheduled */
	if(LFirstNode[node->level] == LschedEOL)
		LFirstNode[node->level] = node;
	else	LLastNode[node->level]->sched = node;
	
	LLastNode[node->level] = node;
	node->sched = LschedEOL;
	if( (L_sched_count++ & 0xFFFF) == 0) {
	    printf( "." );
	    fflush(stdout);
	}
	ActiveNodes++;
        return;
}

/* processing routine for scheduler */
struct cpmNode *
LnextNode( direction )
int direction;
{
    struct cpmNode *node;
    int level;

    if(ActiveNodes <= 0) return( NULL ); /* none left */

    if( direction != REVERSE ) {
	for(level=0;level<SchedMaxLevels;level++) {
	    if(LFirstNode[level] != LschedEOL) {
		node = LFirstNode[level];
		LFirstNode[level] = node->sched;
		node->sched = NULL;
		ActiveNodes--;
		if(LFirstNode[level] == LschedEOL) LLastNode[level] = LschedEOL;

		return( node );
	    }
	}
    }
    else {
	for(level=SchedMaxLevels-1;level>=0;level--) {
	    if(LFirstNode[level] != LschedEOL) {
		node = LFirstNode[level];
		LFirstNode[level] = node->sched;
		node->sched = NULL;
		ActiveNodes--;
		if(LFirstNode[level] == LschedEOL) LLastNode[level] = LschedEOL;

		return( node );
	    }
	}
    }
    fprintf(stderr,"nextNodeLevel: lost a node somehow\n");
    fprintf(stderr,"nextNodeLevel: active=%d schedCount=%d\n",ActiveNodes,L_sched_count);
    exit(-1);
}

/*
 *  initialize the scheduler
 */
void
LschedInit( maxLevels )
int maxLevels;
{
    int level;

    L_sched_count=0;
    ActiveNodes=0;
    LschedEOL = ALLOC( struct cpmNode, 1 );
    LFirstNode = ALLOC( struct cpmNode *, maxLevels + 1 );
    LLastNode = ALLOC( struct cpmNode *, maxLevels + 1 );
    SchedMaxLevels = maxLevels;
    for(level=0;level<=SchedMaxLevels;level++) {
	    LFirstNode[level] = LschedEOL;
	    LLastNode[level] = LschedEOL;
    }
}

/*
 *  return the number of schedules
 */
double
LschedCount()
{
	return ( (double)(L_sched_count) );
}
