/*LINTLIBRARY*/
#include "mkInternal.h"
/*
 * addArrayArea
 *
 * Copyright (c) 1987, Charles J Kring. All rights reserved.
 */
void
mkAddArrayArea( arrayId, area )
mkId arrayId;
octBox *area;
{
    int numAdded;	/* number of rows or columns added to the array	*/
    int numMoved;	/* number of rows or columns moved in the array	*/
    int numTotal;	/* numbe of rows or columns after addition	*/
    int row, col;	/* indices into the array			*/
    s_array *array;	/* the array be added to 			*/
    s_arraylist *sublist;	/* list of subarrays of array		*/
    int min, max;
    int *map;
    s_elemlist ***rowptr;
    s_elemlist **colptr;
    s_cpmNode **grid;
    s_cpmBranch **branch;
    char nodeName[16];
    static int callNum = 0;

    MK_START("mkAddArrayArea");
    array = mk_get_array( arrayId );

    /* no mapped area since this routine does not support indirection	*/

    if( array->type == SUBARRAY ){

	/* adding rows to the subarray					*/
	
	min = MIN( area->lowerLeft.y, area->upperRight.y );
	max = MAX( area->lowerLeft.y, area->upperRight.y );

	if ( min >= array->numRows )min = array->numRows;

	numAdded = max - min + 1;
	numMoved = array->numRows - min;
	numTotal = array->numRows + numAdded;

	if( array->numRows == 0 ){
	    array->contents.subarray.rowmap = ALLOC( int, numTotal);
	} else {
	    array->contents.subarray.rowmap = 
		REALLOC( int, array->contents.subarray.rowmap, numTotal);
	}
	map = &array->contents.subarray.rowmap[min];
	for( row = numMoved; row > 0; row-- ){
	    map[numAdded] = *map;
	    map++;
	}

	map = &array->contents.subarray.rowmap[min];
	for( row = numAdded; row > 0; row--) *map++ = MK_UNMAPPED_COORD;

	array->numRows = numTotal;

	/* adding cols to the subarray					*/

	min = MIN( area->lowerLeft.x, area->upperRight.x );
	max = MAX( area->lowerLeft.x, area->upperRight.x );

	if ( min >= array->numCols )min = array->numCols;

	numAdded = max - min + 1;
	numMoved = array->numCols - min;
	numTotal = array->numCols + numAdded;

	if( array->numCols == 0 ){
	    array->contents.subarray.colmap = ALLOC( int, numTotal);
	} else {
	    array->contents.subarray.colmap = 
		REALLOC( int, array->contents.subarray.colmap, numTotal);
	}

	map = &array->contents.subarray.colmap[min];
	for( col = numMoved; col > 0; col-- ){
	    map[numAdded] = *map;
	    map++;
	}
	map = &array->contents.subarray.colmap[min];
	for( col = numAdded; col > 0; col--) *map++ = MK_UNMAPPED_COORD;

	array->numCols = numTotal;

	MK_END();
	return;
    }

    /* otherwise it is a regular array					*/

    if(( area->lowerLeft.x != MK_NULL_COORD )&&
       ( area->upperRight.x != MK_NULL_COORD )){

	min = MIN( area->lowerLeft.x, area->upperRight.x );
	max = MAX( area->lowerLeft.x, area->upperRight.x );

	if ( min >= array->numCols )min = array->numCols;

	numAdded = max - min + 1;
	numMoved = array->numCols - min;
	numTotal = array->numCols + numAdded;

	/* updating the existing rows in the element array		*/

	rowptr = array->contents.array.elemlist;
	for( row = 0; row < array->numRows; row++ ){

	    rowptr[row] = REALLOC(s_elemlist *, rowptr[ row ], numTotal);

	    colptr = &rowptr[row][ min ];
	    for( col = numMoved - 1; col >= 0; col-- ){
		colptr[ col + numAdded ] = colptr[ col ];
	    }
	    for( col = 0; col < numAdded; col++ ){
		colptr[ col ] = NIL( s_elemlist );
	    }
	}

	/* the column virtual grid nodes				*/

	if( array->numCols == 0 ){
	    array->contents.array.colGrid = ALLOC( s_cpmNode *, numTotal + 1);
	} else {
	    array->contents.array.colGrid = REALLOC(s_cpmNode *, 
			array->contents.array.colGrid, numTotal + 1);
	    grid = &array->contents.array.colGrid[ min ];
	    for( col = numMoved; col >= 0; col-- ){
		grid[ col + numAdded ] = grid[ col ];
	    }
	}
	grid = &array->contents.array.colGrid[ min ];
	if( array->numCols == 0 ){
	    *grid = ALLOC( struct cpmNode, 1 );
	    cpmInitNode( *grid, "col_0" );
	    cpmInstallNode( array->contents.array.Xgraph, *grid );
	    grid++;
	}
	for( col = 0; col < numAdded; col++ ){
	    grid[ col ] = ALLOC( struct cpmNode, 1 );
	    if( callNum = 0 ) (void) sprintf( nodeName, "col_%d", col + 1 );
	    else (void) sprintf( nodeName, "col_%d_%d", callNum, col + 1 );
	    cpmInitNode( grid[ col ], nodeName );
	    cpmInstallNode( array->contents.array.Xgraph, grid[ col ]);
	}

	/* the column branches						*/

	if( array->numCols == 0 ){
	    array->contents.array.colBranch = ALLOC( s_cpmBranch *, numTotal );
	} else {
	    array->contents.array.colBranch = REALLOC( s_cpmBranch *, 
			 array->contents.array.colBranch, numTotal);
	    branch = &array->contents.array.colBranch[ min ];
	    for( col = numMoved - 1; col >= 0; col-- ){
		branch[ col + numAdded ] = branch[ col ];
	    }
	}
	if(( array->numCols != 0 )&&( min >= array->numCols )){
	    branch = &array->contents.array.colBranch[ array->numCols - 1 ];
	    grid = &array->contents.array.colGrid[ array->numCols ];
	} else {
	    branch = &array->contents.array.colBranch[ min ];
	    grid = &array->contents.array.colGrid[ min ];
	}
	for( col = 0; col < numAdded; col++ ){
	    branch[ col ] = ALLOC( struct cpmBranch, 1 );
	    cpmInstallBranch( array->contents.array.Xgraph, grid[ col ], 
			      grid[ col + 1 ], branch[ col ], 
			      array->contents.array.colPitch, CPM_INFINITY, 0 );
	}
	/* resetting the branch fanning into the first new grid		*/
	if(( min > 0 )&&( min < array->numCols )){

	    branch = &array->contents.array.colBranch[ min - 1 ];
	    grid = &array->contents.array.colGrid[ min - 1 ];

	    cpmRemoveBranch( array->contents.array.Xgraph, *branch );
	    cpmInstallBranch( array->contents.array.Xgraph, grid[ 0 ], 
			      grid[ 1 ], *branch,
			      array->contents.array.colPitch, CPM_INFINITY, 0 );
	}
	/* the subarray's columns					*/

	if( numAdded > 0 ){
	    for( sublist = array->contents.array.sublist; 
		 sublist != NIL( s_arraylist ); sublist = sublist->next ){

		    map = sublist->array->contents.subarray.colmap;
		for( col = sublist->array->numCols; col > 0; col-- ){
		    if( *map > min ) *map++ += numAdded;
		}
	    }
	}
	array->numCols = numTotal;
    }

    /* now the rows in the array					*/

    if(( area->lowerLeft.y != MK_NULL_COORD )&&
       ( area->upperRight.y != MK_NULL_COORD )){

	min = MIN( area->lowerLeft.y, area->upperRight.y );
	max = MAX( area->lowerLeft.y, area->upperRight.y );

	if ( min >= array->numRows )min = array->numRows;

	numAdded = max - min + 1;
	numMoved = array->numRows - min;
	numTotal = array->numRows + numAdded;

	/* adding rows in the element array				*/

	if( array->numRows == 0 ){
	    array->contents.array.elemlist = ALLOC( s_elemlist **, numTotal );
	} else {
	    array->contents.array.elemlist =
	    REALLOC( s_elemlist **, array->contents.array.elemlist, numTotal );

	    rowptr = &array->contents.array.elemlist[ min ];
	    for( row = numMoved - 1; row >= 0; row-- ){
		rowptr[ row + numAdded] = rowptr[ row ];
	    }
	}
	rowptr = &array->contents.array.elemlist[ min ];
	for( row = 0; row < numAdded; row++ ){
	    rowptr[ row ] = ALLOC( s_elemlist *, array->numCols );
	    colptr = rowptr[ row ];
	    for( col = 0; col < array->numCols; col ++ ){
		colptr[col] = NIL(s_elemlist);
	    }
	}

	/* the row virtual grid nodes					*/

	if( array->numRows == 0 ){
	    array->contents.array.rowGrid = ALLOC( s_cpmNode *, numTotal + 1);
	} else {
	    array->contents.array.rowGrid = REALLOC( s_cpmNode *, 
			     array->contents.array.rowGrid, numTotal + 1);
	    grid = &array->contents.array.rowGrid[  min ];
	    for( row = numMoved; row >= 0; row-- ){
		grid[ row + numAdded ] = grid[ row ];
	    }
	}
	grid = &array->contents.array.rowGrid[ min ];
	if( array->numRows == 0 ){
	    *grid = ALLOC( struct cpmNode, 1 );
	    cpmInitNode( *grid, "row_0" );
	    cpmInstallNode( array->contents.array.Ygraph, *grid );
	    grid++;
	}
	for( row = 0; row < numAdded; row++ ){
	    grid[ row ] = ALLOC( struct cpmNode, 1 );
	    if( callNum = 0 ) (void) sprintf( nodeName, "col_%d", col + 1 );
	    else (void) sprintf( nodeName, "col_%d_%d", callNum, col + 1 );
	    cpmInitNode( grid[ row ], nodeName );
	    cpmInstallNode( array->contents.array.Ygraph, grid[ row ] );
	}

	/* the row branches						*/

	if( array->numRows == 0 ){
	    array->contents.array.rowBranch = ALLOC( s_cpmBranch *, numTotal );
	} else {
	    array->contents.array.rowBranch = REALLOC( s_cpmBranch *, 
			 array->contents.array.rowBranch, numTotal);
	    branch = &array->contents.array.rowBranch[ min ];
	    for( row = numMoved - 1; row >= 0; row-- ){
		branch[ row + numAdded ] = branch[ row ];
	    }
	}
	if(( array->numRows != 0 )&&( min >= array->numRows )){
	    branch = &array->contents.array.rowBranch[ min - 1 ];
	    grid = &array->contents.array.rowGrid[ min - 1 ];
	} else {
	    branch = &array->contents.array.rowBranch[ min ];
	    grid = &array->contents.array.rowGrid[ min ];
	}
	for( row = 0; row < numAdded; row++ ){
	    branch[ row ] = ALLOC( struct cpmBranch, 1 );
	    cpmInstallBranch( array->contents.array.Ygraph, grid[ row ], 
			      grid[ row + 1 ], branch[ row ], 
			      array->contents.array.rowPitch, CPM_INFINITY, 0 );
	}
	/* reseting the branch fanning into the first new grid          */
	 if(( min > 0 )&&( min < array->numRows )){
	    branch = &array->contents.array.rowBranch[ min - 1 ];
	    grid = &array->contents.array.rowGrid[ min - 1 ];

	    cpmRemoveBranch( array->contents.array.Ygraph, *branch );
	    cpmInstallBranch( array->contents.array.Ygraph, grid[ 0 ],
			      grid[ 1 ], *branch,
			      array->contents.array.rowPitch, CPM_INFINITY, 0 );
	}

	/* the subarray's rows						*/

	if( numAdded > 0 ){
	    for( sublist = array->contents.array.sublist; 
		 sublist != NIL( s_arraylist ); sublist = sublist->next ){

		map = sublist->array->contents.subarray.rowmap;
		for( row = sublist->array->numRows; row > 0; row-- ){
		    if( *map > min ) *map++ += numAdded;
		}
	    }
	}
	array->numRows = numTotal;
    }
    callNum++;
    MK_END();
    return;
}
