/*LINTLIBRARY*/
#include "mkInternal.h"
/*
 * areaMap.c
 * Copyright (c) 1987, Charles J Kring. All rights reserved.
 *
 * routines which map points, rows, and columns into areas
 */
static octBox AreaArray[MK_AREA_ARRAY_SIZE];
static int AreaIndex = 0;
#define NEXT_AREA &AreaArray[ (AreaIndex++)%MK_AREA_ARRAY_SIZE ]
/*
 * mkPoint
 * 
 * map a point into an area
 */
octBox *
mkPoint( row, col )
int row, col;		/* of the point				*/

{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.y = area->upperRight.y = row;
    area->lowerLeft.x = area->upperRight.x = col;
    return( area );
}
/*
 * mkArea
 * 
 * map an area bounded by lower, upper, left and right into an area
 */
octBox *
mkArea( lower, upper, left, right )
int lower, upper;	/* coordinate of the lower and upper rows	*/
int left, right;	/* coordinate of the left and right columns	*/
{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.y = lower;
    area->upperRight.y = upper;
    area->lowerLeft.x = left;
    area->upperRight.x = right;
    return( area );
}

/*
 * mkRow
 * 
 * map a row into an area
 */
octBox *
mkRow( row )
int row;		/* the row				*/

{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.x = area->upperRight.x = MK_NULL_COORD;
    area->lowerLeft.y = area->upperRight.y = row;
    return( area );
}
/*
 * mkRows
 * 
 * map a range of rows into an area
 */
octBox *
mkRows( lower, upper )
int lower, upper;	/* coordinate of the lower and upper rows	*/
{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.x = area->upperRight.x = MK_NULL_COORD;
    area->lowerLeft.y = lower;
    area->upperRight.y = upper;
    return( area );
}
/*
 * mkCol
 * 
 * map a column into an area
 */
octBox *
mkCol( column )
int column;		/* the column				*/
{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.x = area->upperRight.x = column;
    area->lowerLeft.y = area->upperRight.y = MK_NULL_COORD;
    return( area );
}
/*
 * mkCols
 * 
 * map a range of columns into an area
 */
octBox *
mkCols( left, right )
int left, right;	/* coordinate of the left and right columns	*/
{
    octBox *area;
    area = NEXT_AREA;
    area->lowerLeft.x = left;
    area->upperRight.x = right;
    area->lowerLeft.y = area->upperRight.y = MK_NULL_COORD;
    return( area );
}

/*
 * mkIntersection
 *
 * returns the intersection of two areas
 */
octBox *
mkIntersection( area1, area2 )
octBox *area1;
octBox *area2;
{
    octBox *area;
    octBox a1, a2;
    area = NEXT_AREA;

    MK_START( "mkIntersection" );
    if(( area1->lowerLeft.x != MK_NULL_COORD )&&
       ( area1->upperRight.x != MK_NULL_COORD )&&
       ( area1->upperRight.x < area1->lowerLeft.x )){
       a1.lowerLeft.x = area1->upperRight.x;
       a1.upperRight.x = area1->lowerLeft.x;
    } else {
       a1.lowerLeft.x = area1->lowerLeft.x;
       a1.upperRight.x = area1->upperRight.x;
    }
    if(( area1->lowerLeft.y != MK_NULL_COORD )&&
       ( area1->upperRight.y != MK_NULL_COORD )&&
       ( area1->upperRight.y < area1->lowerLeft.y )){
       a1.lowerLeft.y = area1->upperRight.y;
       a1.upperRight.y = area1->lowerLeft.y;
    } else {
       a1.lowerLeft.y = area1->lowerLeft.y;
       a1.upperRight.y = area1->upperRight.y;
    }
    if(( area2->lowerLeft.x != MK_NULL_COORD )&&
       ( area2->upperRight.x != MK_NULL_COORD )&&
       ( area2->upperRight.x < area2->lowerLeft.x )){
       a2.lowerLeft.x = area2->upperRight.x;
       a2.upperRight.x = area2->lowerLeft.x;
    } else {
       a2.lowerLeft.x = area2->lowerLeft.x;
       a2.upperRight.x = area2->upperRight.x;
    }
    if(( area2->lowerLeft.y != MK_NULL_COORD )&&
       ( area2->upperRight.y != MK_NULL_COORD )&&
       ( area2->upperRight.y < area2->lowerLeft.y )){
       a2.lowerLeft.y = area2->upperRight.y;
       a2.upperRight.y = area2->lowerLeft.y;
    } else {
       a2.lowerLeft.y = area2->lowerLeft.y;
       a2.upperRight.y = area2->upperRight.y;
    }

    if( a1.lowerLeft.x == MK_NULL_COORD )
	area->lowerLeft.x = a2.lowerLeft.x;
    else if( a2.lowerLeft.x == MK_NULL_COORD )
	area->lowerLeft.x = a1.lowerLeft.x;
    else
	area->lowerLeft.x = MAX( a1.lowerLeft.x, a2.lowerLeft.x );

    if( a1.lowerLeft.y == MK_NULL_COORD )
	area->lowerLeft.y = a2.lowerLeft.y;
    else if( a2.lowerLeft.y == MK_NULL_COORD ) 
	area->lowerLeft.y = a1.lowerLeft.y;
    else
	area->lowerLeft.y = MAX( a1.lowerLeft.y, a2.lowerLeft.y );

    if( a1.upperRight.x == MK_NULL_COORD )
	area->upperRight.x = a2.upperRight.x;
    else if( a2.lowerLeft.x == MK_NULL_COORD )
	area->upperRight.x = a1.upperRight.x;
    else
	area->upperRight.x = MIN( a1.upperRight.x, a2.upperRight.x );

    if( a1.upperRight.y == MK_NULL_COORD )
	area->upperRight.y = a2.upperRight.y;
    else if( a2.upperRight.y == MK_NULL_COORD )
	area->upperRight.y = a1.lowerLeft.y;
    else
	area->upperRight.y = MIN( a1.upperRight.y, a2.upperRight.y );

    MK_END();
    return( area );
}

/*
 * mkUnionBB
 *
 * returns the bounding box of the union of two areas
 */
octBox *
mkUnionBB( area1, area2 )
octBox *area1;
octBox *area2;
{
    octBox *area;
    octBox a1, a2;
    area = NEXT_AREA;

    MK_START( "mkUnionBB" );
    if(( area1->lowerLeft.x != MK_NULL_COORD )&&
       ( area1->upperRight.x != MK_NULL_COORD )&&
       ( area1->upperRight.x < area1->lowerLeft.x )){
       a1.lowerLeft.x = area1->upperRight.x;
       a1.upperRight.x = area1->lowerLeft.x;
    } else {
       a1.lowerLeft.x = area1->lowerLeft.x;
       a1.upperRight.x = area1->upperRight.x;
    }
    if(( area1->lowerLeft.y != MK_NULL_COORD )&&
       ( area1->upperRight.y != MK_NULL_COORD )&&
       ( area1->upperRight.y < area1->lowerLeft.y )){
       a1.lowerLeft.y = area1->upperRight.y;
       a1.upperRight.y = area1->lowerLeft.y;
    } else {
       a1.lowerLeft.y = area1->lowerLeft.y;
       a1.upperRight.y = area1->upperRight.y;
    }
    if(( area2->lowerLeft.x != MK_NULL_COORD )&&
       ( area2->upperRight.x != MK_NULL_COORD )&&
       ( area2->upperRight.x < area2->lowerLeft.x )){
       a2.lowerLeft.x = area2->upperRight.x;
       a2.upperRight.x = area2->lowerLeft.x;
    } else {
       a2.lowerLeft.x = area2->lowerLeft.x;
       a2.upperRight.x = area2->upperRight.x;
    }
    if(( area2->lowerLeft.y != MK_NULL_COORD )&&
       ( area2->upperRight.y != MK_NULL_COORD )&&
       ( area2->upperRight.y < area2->lowerLeft.y )){
       a2.lowerLeft.y = area2->upperRight.y;
       a2.upperRight.y = area2->lowerLeft.y;
    } else {
       a2.lowerLeft.y = area2->lowerLeft.y;
       a2.upperRight.y = area2->upperRight.y;
    }
    if( a1.lowerLeft.x == MK_NULL_COORD )
	area->lowerLeft.x = a1.lowerLeft.x;
    else if( a2.lowerLeft.x == MK_NULL_COORD )
	area->lowerLeft.x = a2.lowerLeft.x;
    else
	area->lowerLeft.x = MIN( a1.lowerLeft.x, a2.lowerLeft.x );

    if( a1.lowerLeft.y == MK_NULL_COORD )
	area->lowerLeft.y = a1.lowerLeft.y;
    else if( a2.lowerLeft.y == MK_NULL_COORD )
	area->lowerLeft.y = a2.lowerLeft.y;
    else
	area->lowerLeft.y = MIN( a1.lowerLeft.y, a2.lowerLeft.y );

    if( a1.upperRight.x == MK_NULL_COORD )
	area->upperRight.x = a1.upperRight.x;
    else if( a2.lowerLeft.x == MK_NULL_COORD )
	area->upperRight.x = a2.upperRight.x;
    else
	area->upperRight.x = MAX( a1.upperRight.x, a2.upperRight.x );

    if( a1.upperRight.y == MK_NULL_COORD )
	area->upperRight.y = a1.upperRight.y;
    else if( a2.upperRight.y == MK_NULL_COORD )
	area->upperRight.y = a1.lowerLeft.y;
    else
	area->upperRight.y = MAX( a1.upperRight.y, a2.upperRight.y );

    MK_END();
    return( area );
}

/*
 * mkArrayBB
 *
 * returns the bounding box of array
 */
octBox *
mkArrayBB( arrayId )
mkId arrayId;
{
    s_array *array;
    octBox *area;

    MK_START( "mkArrayBB" );
    area = NEXT_AREA;
    array = mk_get_array( arrayId );
    area->lowerLeft.x = area->lowerLeft.y = 0;
    area->upperRight.x = array->numCols - 1;
    area->upperRight.y = array->numRows - 1;
    MK_END();
    return( area );
}
/*
 * mkObjectBB
 *
 * set area equal to the bounding box of object in array 
 */
octBox *
mkObjectBB( arrayId, object )
mkId arrayId;
octObject *object;
{
    s_element *element;
    s_array *array;
    octBox *area;

    MK_START( "mkObjectBB" );
    area = NEXT_AREA;
    array = mk_get_array( arrayId );
    element = mk_get_element( array, object->objectId );
    *area = element->bbox;
    MK_END();
    return( area );
}
/*
 * mkTranslate
 *
 */
octBox *
mkTranslate( subArrayId, subArea )
mkId subArrayId;
octBox *subArea;
{
    s_array *subArray;
    octBox *area;

    MK_START( "mkTranslate" );
    area = NEXT_AREA;

    subArray = mk_get_array( subArrayId );

    if( subArray->type == ARRAY ){
	*area = *subArea;

    } else {

	if( subArea->lowerLeft.x == MK_NULL_COORD )
	    area->lowerLeft.x = MK_NULL_COORD;
	else
	    area->lowerLeft.x = 
		subArray->contents.subarray.colmap[subArea->lowerLeft.x];

	if( subArea->lowerLeft.y == MK_NULL_COORD )
	    area->lowerLeft.y = MK_NULL_COORD;
	else
	    area->lowerLeft.y = 
		subArray->contents.subarray.rowmap[subArea->lowerLeft.y];

	if( subArea->upperRight.x == MK_NULL_COORD )
	    area->upperRight.x = MK_NULL_COORD;
	else
	    area->upperRight.x = 
		subArray->contents.subarray.colmap[subArea->upperRight.x];

	if( subArea->upperRight.y == MK_NULL_COORD )
	    area->upperRight.y = MK_NULL_COORD;
	else
	    area->upperRight.y = 
		subArray->contents.subarray.rowmap[subArea->upperRight.y];

	if(( area->lowerLeft.x == MK_UNMAPPED_COORD )||
	   ( area->lowerLeft.y == MK_UNMAPPED_COORD )||
	   ( area->upperRight.x == MK_UNMAPPED_COORD )||
	   ( area->upperRight.y == MK_UNMAPPED_COORD )){
	    errRaise( MK_PKG_NAME, 0,
	              "The subarray was not mapped to the array");
	}
    }
    MK_END();
    return( area );
}
