/*LINTLIBRARY*/
#include "mkInternal.h"
/*
 * mkAttach
 * Copyright (c) 1987, Charles J Kring. All rights reserved.
 *
 */
void
mkAttach( arrayId, area, object )
mkId arrayId;
octBox *area;
octObject *object;
{
    s_array *array;
    octBox *mappedArea;
    s_element *element;

    MK_START("mkAttach");

    array = mk_get_array( arrayId );
    mk_sanitize_area( array, area );

    if( array->type == SUBARRAY ) {
	mappedArea = mkTranslate( arrayId, area );
	array = array->contents.subarray.array;
    } else {
	mappedArea = area;
    }

    if(( object->type&LEGAL_ELEM_MASK ) == (unsigned)0 ){
	errRaise( MK_PKG_NAME, 0, 
		  "%s type objects cannot be attached to the array",
		  ohTypeName( object ));
    }

    if(( element = mk_get_element( array, object->objectId)) == NIL(s_element)){
	element = mk_malloc_element();
	element->objectId = object->objectId;
	mk_hash_element( array, element );
	element->type = object->type;
	if( element->type == OCT_INSTANCE ){
	    mk_read_instance( array, element );
	    mk_attach_elem( array, mappedArea, element, 0 );
	} else if( element->type == OCT_TERM ){
	    mk_read_term( array, element );
	    mk_attach_elem( array, mappedArea, element, 0 );
	    mk_update_instBB( array, element );
	}
    } else {
	mk_attach_elem( array, mappedArea, element, 0 );
	if ( element->type == OCT_TERM ){
	    mk_update_instBB( array, element );
	} else if ( element->type == OCT_INSTANCE ){
	    mk_update_instBB( array, element );
	}
    }
    MK_END();
}

void
mk_attach_elem( array, area, element, bbOnly )
s_array *array;
octBox *area;
s_element *element;
int bbOnly;
{
    s_elemlist ***rowptr;
    s_elemlist **colptr;
    s_elemlist *elemlist;
    s_cpmNode **rowGrid, **colGrid;
    int row, col;

    rowGrid = array->contents.array.rowGrid;
    colGrid = array->contents.array.colGrid;

    if(( element->bbox.lowerLeft.x == MK_NULL_COORD )||
       ( element->bbox.lowerLeft.x > area->lowerLeft.x )){

	if( element->bbox.lowerLeft.x != MK_NULL_COORD )
	    cpmRemoveBranch( array->contents.array.Xgraph, &element->lbranch );

	cpmInstallBranch( array->contents.array.Xgraph,
			  colGrid[ area->lowerLeft.x ],
			  &element->lnode, &element->lbranch,
			  0, CPM_INFINITY, 0 );
	element->bbox.lowerLeft.x = area->lowerLeft.x;
    }

    if(( element->bbox.upperRight.x == MK_NULL_COORD )||
       ( element->bbox.upperRight.x < area->upperRight.x )){

	if( element->bbox.upperRight.x != MK_NULL_COORD )
	    cpmRemoveBranch( array->contents.array.Xgraph, &element->rbranch );

	cpmInstallBranch( array->contents.array.Xgraph,
			  &element->rnode,
			  colGrid[ area->upperRight.x + 1 ], 
			  &element->rbranch,
			  0, CPM_INFINITY, 0 );
	element->bbox.upperRight.x = area->upperRight.x;
    }

    if(( element->bbox.lowerLeft.y == MK_NULL_COORD )||
       ( element->bbox.lowerLeft.y > area->lowerLeft.y )){

	if( element->bbox.lowerLeft.y != MK_NULL_COORD )
	    cpmRemoveBranch( array->contents.array.Ygraph, &element->bbranch );

	cpmInstallBranch( array->contents.array.Ygraph,
			  rowGrid[ area->lowerLeft.y ],
			  &element->bnode, &element->bbranch,
			  0, CPM_INFINITY, 0 );
	element->bbox.lowerLeft.y = area->lowerLeft.y;
    }

    if(( element->bbox.upperRight.y == MK_NULL_COORD )||
       ( element->bbox.upperRight.y < area->upperRight.y )){

	if( element->bbox.upperRight.y != MK_NULL_COORD )
	    cpmRemoveBranch( array->contents.array.Ygraph, &element->tbranch );

	cpmInstallBranch( array->contents.array.Xgraph,
			  &element->tnode,
			  rowGrid[ area->upperRight.y + 1 ], 
			  &element->tbranch,
			  0, CPM_INFINITY, 0 );
	element->bbox.upperRight.y = area->upperRight.y;
    }

    if( bbOnly )return;
    rowptr = &array->contents.array.elemlist[area->lowerLeft.y];
    for( row = area->upperRight.y - area->lowerLeft.y + 1; 
	 row > 0; row--){
	colptr = &(*rowptr)[area->lowerLeft.x];
	for( col = area->upperRight.x - area->lowerLeft.x + 1; 
	     col > 0; col-- ){

	    /* no sense attaching something twice			*/
	    for( elemlist = *colptr; elemlist != NIL( s_elemlist); 
		 elemlist = elemlist->next ){
		if( elemlist->element->objectId == element->objectId ) continue;
	    }

	    /* attaching object to array[row][col]			*/
	    elemlist = ALLOC( s_elemlist, 1 );
	    elemlist->element = element;
	    elemlist->next = *colptr;
	    *colptr = elemlist;
	    colptr++;
	}
    rowptr++;
    }
}

/*
 * mk_update_instBB
 *
 * resets the instance bounding box if an actual terminal has been attached
 * outside of the instance's bounding box
 */
void
mk_update_instBB( array, element )
s_array *array;
s_element *element;
{
    octBox bbox;
    s_aterm *term;
    s_element *instance;
    int first = TRUE;

    if ( element->type == OCT_TERM ){
	instance = element->contents.term->instance;
    } else if( element->type == OCT_INSTANCE ){
	instance = element;
    } else {
	MK_INT_ERROR( "corrupt element type");
    }

    if(( instance->bbox.lowerLeft.x == MK_NULL_COORD )||
       ( instance->bbox.lowerLeft.y == MK_NULL_COORD )||
       ( instance->bbox.upperRight.x == MK_NULL_COORD )||
       ( instance->bbox.upperRight.y == MK_NULL_COORD )) return;

    if ( element->type == OCT_TERM ){
	bbox = element->bbox;
    } else {
	for( term = element->contents.instance->term; term != NIL( s_aterm );
	     term = term->next ){
	    if( term->element == NIL( s_element ))continue;
	    if( first ){
		first = FALSE;
		bbox = term->element->bbox;
	    } else {
		bbox.lowerLeft.x = MIN( bbox.lowerLeft.x, 
					term->element->bbox.lowerLeft.x );
		bbox.lowerLeft.y = MIN( bbox.lowerLeft.y, 
					term->element->bbox.lowerLeft.y );
		bbox.upperRight.x = MAX( bbox.upperRight.x, 
					 term->element->bbox.upperRight.x );
		bbox.upperRight.y = MAX( bbox.upperRight.y, 
					 term->element->bbox.upperRight.y );
	    }
	}
    }
    if(( bbox.lowerLeft.x < instance->bbox.lowerLeft.x )||
       ( bbox.lowerLeft.y < instance->bbox.lowerLeft.y )||
       ( bbox.upperRight.x > instance->bbox.upperRight.x )||
       ( bbox.upperRight.y > instance->bbox.upperRight.y ))
	mk_attach_elem( array, &bbox, instance, 1 );
}
