#include "mkInternal.h" 
s_ftermlist *FreeTlist = NIL( s_ftermlist );
/* 
 * buildRules
 *
 * these routines build mkArray connection rules
 */
/*
 * mkGetOrCreateRule
 *
 * gets or creates a rulegroup named ruleName
 * if the rule exists this routine returns MK_ALREADY_OPEN, otherwise
 * it returns MK_OK.  If ruleName is "" or NIL( char ), this routine 
 * prints a message in MkErrBuffer, and returns MK_ERROR.
 */
void
mkCreateRule( ruleName )
char *ruleName;
{

    s_rulegroup *rulegroup;
    s_rulegroup *mk_get_rule();

    MK_START( "mkCreateRule" );

    if(( ruleName == NIL( char ))||( strcmp( ruleName, "" ) == 0 )){
	errRaise( MK_PKG_NAME, 0, "The rule name was NIL or \"\"" );
    }
    if(( rulegroup = mk_get_rule( ruleName )) != NIL( s_rulegroup )){
	errRaise( MK_PKG_NAME, 0, 
		  "A rule named %s has already been created", ruleName );
    }
    rulegroup = mk_malloc_rulegroup();
    rulegroup->name = util_strsav( ruleName );
    rulegroup->enabled = FALSE;
    mk_hash_rule( rulegroup );
    MK_END();
}
void
mkAddConnectToRule( arrayId, ruleName, oTo, oFrom, 
		    layerName, widthSpec, orientation )
mkId arrayId;
char *ruleName;
octObject *oTo;
octObject *oFrom;
char *layerName;
int widthSpec;
int orientation;
{
    s_array *array;
    s_ftermlist *toList;
    s_ftermlist *fromList;
    s_rulegroup *rulegroup;
    s_layer *layer;

    MK_START( "mkCreateRule" );
    array = mk_get_array( arrayId );
    if(( rulegroup = mk_get_rule( ruleName )) == NIL( s_rulegroup )){
	errRaise( MK_PKG_NAME, 0, 
		  "A rule named %s has never been created", ruleName );
    }
    mk_build_tlist( array, oTo, &toList );
    if( oFrom == NIL( octObject )){
	oFrom = oTo;
    } else {
	mk_build_tlist( array, oFrom, &fromList );
    }
    if(( layerName == NIL( char ))||( strcmp( layerName, "" ) == 0 )){
	layer = NIL( s_layer );
    } else {
	mk_get_layer( array, layerName, &layer );
    }
    mk_match_tlists( rulegroup, toList, fromList, 
		     layer, widthSpec, orientation );
    MK_END();
}
/*
 * mk_match_tlists
 *
 * matches toList and fromList to create rules in rulegroup
 * puts the toList and fromList back in the FreeTlist
 */
void
mk_match_tlists( rulegroup, toList, fromList, layer, widthSpec, orientation )
s_rulegroup *rulegroup;
s_ftermlist *toList;
s_ftermlist *fromList;
s_layer *layer;
int widthSpec;
int orientation;
{
    int onelist;
    s_rule *rule;
    s_ftermlist *toPtr, *fromPtr, *next;
    
    onelist = (( fromList == NIL( s_ftermlist ))||( fromList == toList ));
    /* matching the tlists to derive the rules			*/
    for( toPtr = toList; toPtr != NIL( s_ftermlist ); toPtr = toPtr->next ){
	for(( onelist ? ( fromPtr = toList ) : ( fromPtr = fromList ));
	      fromPtr != NIL( s_ftermlist ); fromPtr = fromPtr->next ){

	    rule = mk_malloc_rule();
	    rule->widthSpec = widthSpec;
	    rule->orientation = orientation;
	    rule->group = rulegroup;
	    rule->fterm = fromPtr->fterm;
	    rule->next = toPtr->fterm->rule;
	    toPtr->fterm->rule = rule;
	    rule->layer = layer;
	}
    }
    /* putting the tlists back on the FreeTlist			*/
    for( toPtr = toList; toPtr != NIL( s_ftermlist ); ){
	next = toList->next;
	toPtr->next = FreeTlist;
	FreeTlist = toPtr;
	toPtr = next;
    }
    if( !onelist ){
	for( fromPtr = fromList; fromPtr != NIL( s_ftermlist ); ){
	    next = fromList->next;
	    fromPtr->next = FreeTlist;
	    FreeTlist = fromPtr;
	    fromPtr = next;
	}
    }
}

/*
 * mk_build_tlist
 *
 * build a list of formal terms 
 */
void
mk_build_tlist( array, oObj, tlist )
s_array *array;
octObject *oObj;
s_ftermlist **tlist;
{
    s_ftermlist *listptr;
    s_fterm *fterm;
    octObject oTerm;
    octGenerator gTerm;

    (*tlist) = NIL( s_ftermlist );
    if( oObj->type == OCT_TERM ){
	mk_get_fterm( array, oObj, &fterm );
	if( FreeTlist == NIL( s_ftermlist )){
	    listptr = mk_malloc_ftermlist();
	} else { 
	    listptr = FreeTlist; 
	    FreeTlist = FreeTlist->next; 
	}
	listptr->fterm = fterm;
	listptr->next = *tlist;
	*tlist = listptr;
    } else if( oObj->type == OCT_BAG ){
	octInitGenContents( oObj, OCT_TERM_MASK, &gTerm );
	while( octGenerate( &gTerm, &oTerm ) == OCT_OK ){
	    mk_get_fterm( array, &oTerm, &fterm ); 
	    if( FreeTlist == NIL( s_ftermlist )){
		listptr = mk_malloc_ftermlist();
	    } else { 
		listptr = FreeTlist; 
		FreeTlist = FreeTlist->next; 
	    }
	    listptr->fterm = fterm;
	    listptr->next = *tlist;
	    *tlist = listptr;
	}
    } else {
	MK_INT_ERROR( "corrupt object type" );
    }
}
