
#include "port.h"

#define CPM_PKG_NAME "cpm"
#define CPM_INFINITY (int)100000000
#define CPM_OK 1
#define CPM_SANITIZE_ERR -1
#define CPM_SOLVE_ERR -2
#define CPM_OVERCON_ERR -3
#define CPM_NONCRIT_ERR -4
#define CPM_SANITIZE (unsigned)(1<<0)
#define CPM_SOLVE    (unsigned)(1<<1)
#define CPM_OVERCON  (unsigned)(1<<2)
#define CPM_NONCRIT  (unsigned)(1<<3)
#define CPM_VERBOSE  (unsigned)(1<<4)
#define CPM_DEBUG  (unsigned)(1<<5)
 
typedef struct cpmNode {
        char *name;     /* node name */
        int lower;      /* lower bound on loc */
        int loc;        /* present location of node */
        int upper;      /* upper bound on loc */

        int level;      /* "level" of node from source (level0) */
        int count;      /* counter for number of visits to node */
	int upperInFlg;	/* true if fanin branch with upper bound */
	int lowerInFlg;	/* true if fanin branch with lower bound */
	int upperOutFlg;	/* true fanout branch with upper bound */
	int lowerOutFlg;	/* true fanout branch with lower bound */
        int overcon;    /* true if node part of overconstraint loop */
        struct cpmNode *list;      /* all nodes are linked in order read */
        struct cpmNode *sched;     /* schedule pointer for event-driven */
        struct cpmPath *path;      /* path list for overconstraint check */
        int numfanout;          /* number of fanouts */
        int numfanin;           /* number of fanins */
        struct cpmBranch *fanout;  /* linked list of fanout branches */
        struct cpmBranch *fanin;   /* linked list of fanin branches */
} s_cpmNode;

typedef struct cpmBranch {
        int upper;      /* value of upper bound (INFINITY is a large no.) */
        int lower;      /* value of lower bound */
        int proc;       /* "processed" counter - mainly used for leveling */
        int force;      /* attractive force RIGHT, LEFT, DONTCARE */
        struct cpmBranch *list;    /* all branches are linked */   
        struct cpmNode *fromNode;  /* node from which branch points  */
        struct cpmNode *toNode;    /* node to which branch points */
        struct cpmBranch *nextfo;  /* next fanout branch from this node */
        struct cpmBranch *nextfi;  /* next fanin branch from this node */
} s_cpmBranch;

typedef struct cpmGraph {
	struct cpmNode *nodeList;
	struct cpmBranch *branchList;
	struct cpmNode sourceNode;
	struct cpmNode sinkNode;
	int numNodes, numBranches;
	int maxLevels, numUpperBounds, numLowerBounds;
	unsigned status;
	FILE *graphfile;
	char *graphName;
} s_cpmGraph;
