/*LINTLIBRARY*/
#include "mkInternal.h"
/*
 * mkFreeArray
 *
 * frees all storage associated with array.  If array is a regular
 * array, also frees the arrays subarrays.  All subsequent operations
 * on the array will fail.
 */
void
mkFreeArray( arrayId )
mkId arrayId;
{
    s_array *array;
    s_arraylist *arraylist;
    extern enum st_retval mk_free_element();
    int row, col;
    s_net *net, *nextnet;
    s_segment *segment, *nextseg;
    s_elemlist **elemptr, *elemlist, *nextelem;

    MK_START( "mkFreeArray");
    array = mk_get_array( arrayId );
    if( array->type == ARRAY ){

	/* free the subarrays						*/
	for( arraylist = array->contents.array.sublist;
	     arraylist != NULL; arraylist = arraylist->next ){
	    mkFreeArray( arraylist->array->arrayId );
	}

	/* free the elemlist array					*/
	for( row = array->numRows - 1; row >= 0; row-- ){
	    elemptr = array->contents.array.elemlist[row];
	    for( col = array->numCols - 1; col >= 0; col-- ){
		for( elemlist = *elemptr; elemlist != NIL( s_elemlist );
		     elemlist = nextelem ){
		    nextelem = elemlist->next;
		    FREE( elemlist );
		}
		elemptr++;
	    }
	    FREE( array->contents.array.elemlist[row]);
	}
	FREE( array->contents.array.elemlist );

	/* free the constraint graphs					*/
	/*
	cpmFreeGraph( array->contents.array.Xgraph );
	cpmFreeGraph( array->contents.array.Ygraph );
	*/

	/* free the nets and segments					*/
	for( net = array->contents.array.netlist; net != NIL( s_net );
	     net = nextnet ){
	    nextnet = net->next;
	    for( segment = net->segment; segment != NIL( s_segment );
		 segment = nextseg ){
		nextseg = segment->netNext;
		FREE( segment );
	    }
	    FREE( net );
	}
	/* free the elements						*/
	st_foreach( array->contents.array.elemTbl, mk_free_element, NIL(char));
	if( array->contents.array.elemTbl != NIL( st_table )){
	    st_free_table( array->contents.array.elemTbl );
	}
		    
    } else if ( array->type == SUBARRAY ){

	/* free the rowmap and colmap					*/
	FREE( array->contents.subarray.rowmap );
	FREE( array->contents.subarray.colmap );
    }

    /* free the array							*/
    FREE( array );
    MK_END();
    return;
}
/*
 * mk_free_element
 *
 * free all storage associated with element
 */
/*ARGSUSED*/
enum st_retval
mk_free_element( key, value, arg )
char *key;		/* not used					*/
char *value;		/* the element					*/
char *arg;		/* not used					*/
{
    s_element *element;

    element = (s_element *)value;

    /*
    cpmFreeNode( element->nodes->bnode );
    cpmFreeNode( element->nodes->tnode );
    cpmFreeNode( element->nodes->lnode );
    cpmFreeNode( element->nodes->rnode );
    */
    FREE( element );
    return( ST_CONTINUE );
}
