#include "mkInternal.h"
/*
 * mk_get_fterm
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 *
 * Return a pointer to the s_fterm of the formal terminal
 * corresponding the the actual oct terminal termobj;
 * Returns NULL if the master is not in the hash table.
 */
void
mk_get_fterm(array, oTerm, fterm )
s_array *array;
octObject *oTerm;		/* the oct object of the actual terminal */
s_fterm **fterm;
{
    s_master *master;		/* the master of the formal terminal */
    octObject oMaster;		/* the oct master */
    octObject oInst;		/* the oct instance */
    s_master *mk_get_master();	/* return a master by objectId */
    s_instance instance;
    /*
     * get the oMaster from the oInst which contains oTerm
     */
    if( oTerm->type != OCT_TERM ){
	errRaise( MK_PKG_NAME, 0,
		  "%s line %s: object type is %s, expected an OCT_TERM",
		  __FILE__, __LINE__, ohTypeName( oTerm ));
    }
    oInst.objectId = oTerm->contents.term.instanceId;
    octGetById( &oInst );
    oMaster.type = OCT_FACET;
    oMaster.contents.facet.facet = "interface";
    oMaster.contents.facet.mode = "r";
    oMaster.contents.facet.version = OCT_CURRENT_VERSION;
    if( octOpenMaster( &oInst, &oMaster ) < OCT_OK ){
	errRaise( MK_PKG_NAME, 0,
		  "%s line %s: Could not open %s of %s",
		  __FILE__, __LINE__, 
		  ohFormatName( &oMaster ), ohFormatName( &oInst ));
    }

    /* if mk_get_master returns NIL, read the master			*/
    if ((master = mk_get_master( oMaster.objectId )) == NIL(s_master)){
	mk_read_master( array, &instance, &oInst );
	master = instance.master;
    }

    /* find the formal terminal on the master 				*/
    for( *fterm = master->term; *fterm != NIL(s_fterm); 
	 *fterm = (*fterm)->next ){
	if((*fterm)->xId == oTerm->contents.term.formalExternalId )return;
    }
    errRaise( MK_PKG_NAME, 0,
	      "%s line %s: %s of %s doesn't match a terminal in %s",
	      __FILE__, __LINE__, ohFormatName( oTerm ),
	      ohFormatName( &oInst ), ohFormatName( &oMaster ));
}
