#include "mkInternal.h"
#include "st.h"
st_table *mk_array_tbl = NIL( st_table);
/*
 * mk_hash_array
 *
 * Insert an array into mk_array_tbl by arrayId.  Initialize mk_array_tbl if necessary.
 */
void
mk_hash_array( array )
s_array *array;
{
    if ( mk_array_tbl == NIL( st_table )) 
	mk_array_tbl = st_init_table( st_numcmp, st_numhash );

    if( st_insert( mk_array_tbl, (char *)array->arrayId, (char *)array))
	errRaise( MK_PKG_NAME, 0, 
		  "%s line %s: array %d was already in the hash table",
		  array->arrayId );
}
/*
 * mk_get_array
 *
 * return an array in mk_array_tbl with the correct arrayId
 */
s_array *
mk_get_array( arrayId )
mkId arrayId;
{
    s_array *array = NIL( s_array );
    if (( mk_array_tbl == NULL )||
        ( !st_lookup( mk_array_tbl,(char *)arrayId, (char **)(&array)))){
	return( NIL( s_array ));
    }
    return( array );
}

/*
 * mk_hash_element
 *
 */
void
mk_hash_element( array, element )
s_array *array;
s_element *element;
{
    octObject oObj;
    /*
     * initialize the mk_array_tbl if it is NULL
     */
    if( array->type == SUBARRAY )array = array->contents.subarray.array;

    if ( array->contents.array.elemTbl == NULL ) {
	array->contents.array.elemTbl = st_init_table( st_numcmp, st_numhash );
    }
    if( st_insert( array->contents.array.elemTbl, 
		   (char *)element->objectId, (char *)element)){
	oObj.objectId = element->objectId;
	octGetById( &oObj );
	errRaise( MK_PKG_NAME, 0, 
		  "%s line %s: %d was already in the array %d's hash table",
		  __FILE__, __LINE__, ohFormatName( &oObj ), array->arrayId );
    }
}
/*
 * mk_get_element
 *
 * return an enement from array's hash table
 */
s_element *
mk_get_element( array, objectId )
s_array *array;
octId objectId;
{
    s_element *element = NIL(s_element);

    if( array->type == SUBARRAY )array = array->contents.subarray.array;

    if (( array->contents.array.elemTbl == NIL( st_table ))||
        ( !st_lookup( array->contents.array.elemTbl, 
		     (char *)objectId, (char **)(&element)))){
	return( NIL( s_element ));
    }
    return( element );
}

st_table *mk_master_tbl = NIL(st_table);	/* hashtable for masters */
/*
 * mk_hash_master
 *
 * Insert a master into mk_master_tbl by objectId.  
 * Initialize mk_master_tbl if necessary.
 */
void
mk_hash_master( master )
s_master *master;
{
    octObject oObj;
    if ( mk_master_tbl == NIL( st_table ))
	mk_master_tbl = st_init_table( st_numcmp, st_numhash );
    
    if( st_insert( mk_master_tbl, (char *)master->objectId, (char *)master)){
	oObj.objectId = master->objectId;
	octGetById( &oObj );
	errRaise( MK_PKG_NAME, 0, "%s line %s: %s is already in the hash table",
		  __FILE__, __LINE__, ohFormatName( &oObj ));
    }
}
/*
 * mk_get_master
 *
 * return an element in mk_master_tbl with the correct objectId
 */
s_master *
mk_get_master( objectId )
octId objectId;
{
    s_master *master;

    if(( mk_master_tbl == NIL( st_table ))||
       ( !st_lookup( mk_master_tbl, (char *)objectId, (char **)(&master))))
	return ( NIL( s_master ));
    return ( master );
}

st_table *mk_rule_tbl = NIL(st_table);	/* hashtable for rulegroups */
/*
 * mk_hash_rule
 *
 * Insert a rulegroup into mk_rule_tbl by name.  
 * Initialize mk_rule_tbl if necessary.
 */
void
mk_hash_rule( rulegroup )
s_rulegroup *rulegroup;
{
    if( mk_rule_tbl == NIL(st_table)) 
	mk_rule_tbl = st_init_table( strcmp, st_strhash );
    if( st_insert( mk_rule_tbl, (char *)rulegroup->name, (char *)rulegroup))
	errRaise( MK_PKG_NAME, 0, 
		  "%s line %s: rule %s is already in the hash table",
		  __FILE__, __LINE__, rulegroup->name );
}
/*
 * mk_get_rule
 *
 * return an element in mk_rule_tbl with the correct name
 */
s_rulegroup *
mk_get_rule( name )
char *name;
{
    s_rulegroup *rulegroup = NIL(s_rulegroup);
    if(( mk_rule_tbl == NIL( st_table ))||
       ( !st_lookup( mk_rule_tbl, name, (char **)(&rulegroup)))) 
	return ( NIL( s_rulegroup ));
    return ( rulegroup );
}

st_table *mk_net_tbl = NIL(st_table);	/* hashtable for nets */
/*
 * mk_hash_net
 *
 * Insert a net into mk_net_tbl by name.  Initialize mk_net_tbl if necessary.
 */
void
mk_hash_net( net )
s_net *net;
{
    octObject oObj;

    if ( mk_net_tbl == NIL(st_table) )
	mk_net_tbl = st_init_table( st_numcmp, st_numhash );

    if( st_insert( mk_net_tbl, ( char *)net->objectId, (char *)net )){
	oObj.objectId = net->objectId;
	octGetById( &oObj );
	errRaise( MK_PKG_NAME, 0, "%s line %s: %s is already in the hash table",
		  __FILE__, __LINE__, ohFormatName( &oObj ));
    }
}
/*
 * mk_get_net
 *
 * return an element in mk_net_tbl with the correct name
 */
s_net *
mk_get_net( objectId )
octId objectId;
{
    s_net *net = NIL(s_net);

    if(( mk_net_tbl == NIL(st_table))||
       ( !st_lookup( mk_net_tbl, (char *)objectId, (char **)(&net))))
	return (NIL ( s_net ));
    return ( net );
}
