/*LINTLIBRARY*/
#include "mkInternal.h"
/* 
 * mkLinkSubarray
 *
 * links subArea in arrayId to area in arraId's array
 */
void
mkLinkSubarray( arrayId, subArea, area )
mkId arrayId;
octBox *subArea;
octBox *area;
{
    s_array *subArray;
    s_array *array;
    int index;
    int offset;
    int size, subSize;

    MK_START( "mkLinkSubArray" );

    subArray = mk_get_array( arrayId );

    if( subArray->type == ARRAY )
	errRaise( MK_PKG_NAME, 0,
		  "You cannot link a regular array to another array");

    array = subArray->contents.subarray.array;

    if(( subArea->upperRight.x != MK_NULL_COORD )||
       ( subArea->lowerLeft.x != MK_NULL_COORD )){

	if( subArea->upperRight.x == MK_NULL_COORD )
	    subArea->upperRight.x = subArray->numCols;
	if( subArea->lowerLeft.x == MK_NULL_COORD )
	    subArea->lowerLeft.x = 0;

	if(( subArea->lowerLeft.x < 0 )||
	   ( subArea->upperRight.x >= subArray->numCols )){
	    errRaise( MK_PKG_NAME, 0,
		      "The subarea %s overstepped the the subarray %s",
		      mk_print_area( subArea ), 
		      mk_print_array_bounds( subArray ));
	 }

	if( area->upperRight.x == MK_NULL_COORD )
	    area->upperRight.x = array->numCols;
	if( area->lowerLeft.x == MK_NULL_COORD )
	    area->lowerLeft.x = 0;

	if(( area->lowerLeft.x < 0 )||
	   ( area->upperRight.x >= array->numCols )){
	    errRaise( MK_PKG_NAME, 0,
		      "The area %s overstepped the the array %s",
		      mk_print_area( area ), 
		      mk_print_array_bounds( array ));
	 }

	subSize = subArea->upperRight.x - subArea->lowerLeft.x;
	size = area->upperRight.x - area->lowerLeft.x;

	if(( size > 0 ? size : -size ) != ( subSize > 0 ? subSize : -subSize )){
	    errRaise( MK_PKG_NAME, 0,
		      "The subarea %s and area %s are not the same size",
		      mk_print_area( subArea ), mk_print_area( area ));
	}
	if( subSize >= 0 ){
	    if( size >= 0 ) offset = area->lowerLeft.x - subArea->lowerLeft.x;
	    else            offset = area->upperRight.x - subArea->lowerLeft.x;

	    for( index = subArea->lowerLeft.x; 
		 index <= subArea->upperRight.x; index++ ){
		subArray->contents.subarray.colmap[index] = index + offset;
	    }
	} else {
	    if( size >= 0 ) offset = area->lowerLeft.x - subArea->upperRight.x;
	    else            offset = area->upperRight.x - subArea->upperRight.x;

	    for( index = subArea->upperRight.x; 
		 index <= subArea->lowerLeft.x; index++ ){
		subArray->contents.subarray.colmap[index] = index + offset;
	    }
	}
    }
    if(( subArea->upperRight.y != MK_NULL_COORD )||
        ( subArea->lowerLeft.y != MK_NULL_COORD )){

	if( subArea->upperRight.y == MK_NULL_COORD )
	    subArea->upperRight.y = subArray->numRows;
	if( subArea->lowerLeft.y == MK_NULL_COORD )
	    subArea->lowerLeft.y = 0;

	if(( subArea->lowerLeft.y < 0 )||
	   ( subArea->upperRight.y >= subArray->numRows )){
	    errRaise( MK_PKG_NAME, 0,
		      "The subarea %s overstepped the the subarray %s",
		      mk_print_area( subArea), 
		      mk_print_array_bounds( subArray ));
	 }

	if( area->upperRight.y == MK_NULL_COORD )
	    area->upperRight.y = array->numRows;
	if( area->lowerLeft.y == MK_NULL_COORD )
	    area->lowerLeft.y = 0;

	if(( area->lowerLeft.y < 0 )||
	   ( area->upperRight.y >= array->numRows )){
	    errRaise( MK_PKG_NAME, 0,
		      "The area %s overstepped the the array %s",
		      mk_print_area( area), mk_print_array_bounds( array ));
	 }

	subSize = subArea->upperRight.y - subArea->lowerLeft.y;
	size = area->upperRight.y - area->lowerLeft.y;

	if(( size > 0 ? size : -size ) != ( subSize > 0 ? subSize : -subSize )){
	    errRaise( MK_PKG_NAME, 0,
		      "The subarea %s and area %s are not the same size",
		      mk_print_area( area ), mk_print_array_bounds( array ));
	}
	if( subSize >= 0 ){
	    if( size >= 0 ) offset = area->lowerLeft.y - subArea->lowerLeft.y;
	    else            offset = area->upperRight.y - subArea->lowerLeft.y;
	    for( index = subArea->lowerLeft.y; 
		 index <= subArea->upperRight.y; index++ ){
		subArray->contents.subarray.rowmap[index] = index + offset;
	    }
	} else {
	    if( size > 0 ) offset = area->lowerLeft.y - subArea->upperRight.y;
	    else           offset = area->upperRight.y - subArea->upperRight.y;
	    for( index = subArea->upperRight.y; 
		 index <= subArea->lowerLeft.y; index++ ){
		subArray->contents.subarray.rowmap[index] = index + offset;
	    }
	}
    }
    MK_END();
}
