/*
 * mkInternal.h
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 *
 * This is the header file for mkutils, a set of utilities to read and write
 * array based structures into the mkarray format.
 *
 * Users of mkutils don't have to include this file while compiling.  Instead,
 * the file mk.h contains all of the information necessary.  This file contains
 * the data structures internal to mkutils.
 */
#include "port.h"	/* for portability 				*/
#include "utility.h"	/* standard utilities				*/
#include "errtrap.h"
#include "oct.h"	/* oct data base				*/
#include "oh.h"		/* oct utilities				*/
#include "cpm.h"	/* cpm ( constraint package ) data s 		*/
#include "st.h"		/* hash table package				*/
#include "tap.h"	/* technology access package ( design rules )	*/
#include "mkarray.h"	/* names of bags and properties read by mkarray */
 
/* general stuff 							*/
#define TRUE 1
#define FALSE 0
#define UNDEF -1
#define UNDEF_WIDTH -2
#define INFINITY CPM_INFINITY

/* this should be in OCT						*/
/*
typedef struct octPoint octPoint;
typedef struct octBox octBox;
typedef struct octTransform octTransform;
*/

/* malloc functions */
extern struct array *mk_malloc_array();
extern struct arraylist *mk_malloc_arraylist();
extern struct aterm *mk_malloc_aterm();
extern struct element *mk_malloc_element();
extern struct fterm *mk_malloc_fterm();
extern struct ftermlist *mk_malloc_ftermlist();
extern struct instance *mk_malloc_instance();
extern struct layer *mk_malloc_layer();
extern struct master *mk_malloc_master();
extern struct net *mk_malloc_net();
extern struct rule *mk_malloc_rule();
extern struct rulegroup *mk_malloc_rulegroup();
extern struct segment *mk_malloc_segment();
extern struct tframe *mk_malloc_tframe();

/* hash table functions							*/
extern struct array *mk_get_array();
extern struct element *mk_get_element();
extern struct master *mk_get_master();
extern struct rulegroup *mk_get_rule();
extern struct net *mk_get_net();
extern void mk_hash_array();
extern void mk_hash_element();
extern void mk_hash_master();
extern void mk_hash_rule();
extern void mk_hash_net();

/* internal void mkarray functions					*/
extern void mk_attach_elem();
extern void mk_build_tlist();
extern void mk_get_fterm();
extern void mk_get_layer();
extern void mk_get_seg_width();
extern void mk_init_array();
extern void mk_init_subarray();
extern void mk_match_tlists();
extern void mk_merge_net();
extern void mk_new_segment();
extern void mk_read_instance();
extern void mk_read_layers();
extern void mk_read_master();
extern void mk_read_term();
extern void mk_relative_tCenter();
extern void mk_relax_segments();
extern void mk_sanitize_area();
extern void mk_set_seg_tframes();
extern void mk_transform_tframe();
extern void mk_update_instBB();
extern void mk_wire();
extern void mk_wire_term();

/* errTrap macros							*/
extern void mkErrHandler();
extern int mkDepth;
extern char *mkRoutine;

extern char *mk_print_area();
extern char *mk_print_array_bounds();

#define MK_START( name ) \
	if( mkDepth++ == 0 ){ \
	    mkRoutine = name; \
	    errPushHandler( mkErrHandler ); \
	}
									    
#define MK_END() \
	if( --mkDepth == 0 ) errPopHandler();

#define MK_INT_ERROR( str )\
	errRaise( MK_PKG_NAME, 0, \
		  "An internal error occured: file %s line %s: %s",\
		  __FILE__, __LINE__, str )
	 

/* values of array->type						*/
#define ARRAY (unsigned)1	/* a regular array			*/
#define SUBARRAY (unsigned)2	/* a subarray				*/

/* values of array->status						*/
#define FIRST_ERROR (unsigned)1<<0
#define ARRAY_BUILD (unsigned)1<<1

/* values of element->status						*/
#define SET (unsigned)1<<0	/* a status bit in elements		*/
#define CLR ~SET
#define SIMPLE (unsigned)1<<1	/* elements covering a single box 	*/
#define COMPLEX ~SIMPLE		/* elements covering a complex area 	*/ 

/* values of generator->type						*/
#define AREAGEN (unsigned)0 	/* initialized with mkInitGenArea	*/
#define OBJECTGEN (unsigned)1 	/* initialized with mkInitGenObject	*/

/* values of term->status						*/
#define WIRED (unsigned)1<<0

/* values of rule->type							*/
#define MAX_RULE (unsigned)1<<0

/* An error buffer which is accessed with mkErrorString()		*/
extern unsigned MkStatusBuffer;

/* typedefault 								*/
#define DEFAULT_TECH "scmos"
#define DIDITBAG "MKARRAYDIDIT"

/* the orientation of a segment	*/
#define HORIZONTAL MK_HOR_SEG
#define VERTICAL MK_VER_SEG

/* values of instance->type						*/
#define INSTANCE 1
#define CONNECTOR 2

/* every element of the array is hashed into a table of ELEMTBL_SIZE */
/* every master is hashed into a table of MASTERTBL_SIZE */
#define ELEMTBL_SIZE 500  
#define MASTERTBL_SIZE 25  
#define RULETBL_SIZE 10  

/* a row, column, or contains an instance */
typedef struct element {  
    octId objectId;
    unsigned type;		/* OCT object type of the element	*/
    octBox bbox;		/* bounding box of element in the array	*/
    union {
	struct instance *instance;	/* element is an instance 	*/
	struct aterm *term;		/* element is an actual term 	*/
    } contents;
    struct cpmNode bnode;		/* bottom edge of the element 	*/
    struct cpmNode tnode;		/* top edge of the element 	*/
    struct cpmNode lnode;		/* left edge of the element 	*/
    struct cpmNode rnode;		/* right edge of the element 	*/
    struct cpmBranch bbranch;		/* branch to bnode		*/
    struct cpmBranch tbranch;		/* branch from tnode		*/
    struct cpmBranch lbranch;		/* branch to lnode		*/
    struct cpmBranch rbranch;		/* branch from rnode		*/
} s_element;

/* links elements into lists */
typedef struct elemlist {  
    struct elemlist *next;
    struct element *element;
} s_elemlist;

/* an instance */
typedef struct instance {  
    struct master *master;	/* the master of the instance		*/
    struct aterm *term;		/* actual terminals on the instance	*/
    struct octPoint offset;  	/* places boundingbox at 0,0 */
    int type;
} s_instance;

/* a master */
typedef struct master {  
    octId objectId;		/* objectId of the OCT facet		*/
    struct fterm *term;		/* list of formal terms on the master 	*/
} s_master;

/* a net */
typedef struct net { 
    octId objectId;		/* objectId of the OCT net		*/
    char *name;			/* name of the net			*/
    int temporary;		/* TRUE if the net is termporary	*/
    struct segment *segment;	/* segments attached to the net		*/
    struct net *next;		/* next net in the netlist		*/
} s_net;

/* a net segment */
typedef struct segment { 
    octId objectId;		/* objectId of the OCT path		*/
    struct cpmNode node;	/* node in either the constraint graph	*/
    struct aterm *termIn;	/* the upper or left aterm		*/
    struct tframe *frameIn;	/* implementation of termIn		*/
    struct aterm *termOut;	/* the lower or right aterm		*/
    struct tframe *frameOut;	/* implementation of termOut		*/
    int orientation;		/* HORIZONTAL, VERTICAL, or UNDEF	*/
    struct segment *netNext;	/* next segment from the net		*/
    struct segment *inNext;	/* next segment from termIn		*/
    struct segment *outNext;	/* next segment from termOut		*/
} s_segment;

/* an actual terminal - an instance's terminal */
typedef struct aterm {  
    octId objectId;		/* objectId of the OCT actual term	*/
    struct net *net;		/* the net the term is connected to 	*/
    struct element *element;	/* the element the term is in		*/
    struct element *instance;	/* the terminals instance		*/
    struct aterm *next;		/* the next aterm in its instance	*/
    struct fterm *fterm;	/* the corresponding formal term	*/
    struct segment *segOut;	/* segments going out of the term	*/
    struct segment *segIn;	/* segments coming in to the term	*/
} s_aterm;

/* a formal terminal - a master's terminal */
typedef struct fterm { 
    octId xId;		      	/* externalId of the OCT formal term	*/
    struct tframe *frame;	/* implementations of this term		*/
    struct fterm *next;		/* the next formal term on the master	*/
    struct rule *rule;		/* rules connecting "from" this term	*/
    int joinedGroup;		/* used to group JOINED terms		*/
} s_fterm;

/* a box/layer pair implemention of a terminal */
typedef struct tframe { 
    octBox box;		
    struct layer *layer;
    struct tframe *next; 	/* other implementations of the term	*/
} s_tframe;

/* a layer */
typedef struct layer { 
    octId objectId;		/* objectId of the OCT layer		*/
    char *name;			/* the name of the layer		*/
    struct layer *next;		/* the next layer in the list of layers	*/
    int minWidth;		/* minimum feature size of the layer	*/
    int widthSpec;		/* specified width for the layer	*/
    int hpriority;		/* horizontal priority of the layer	*/
    int vpriority;		/* vertical priority of the layer	*/
} s_layer;

/* a group of rules created with a common name */
typedef struct rulegroup {
    char *name;			/* the name given to the rules		*/
    int enabled;		/* if TRUE, this group's rules can fire	*/
}  s_rulegroup;

/* a connection rule between two fterms */
/* connects aterms of the "from" and "to" fterms when group is enabled	*/
/* the "from" fterm is the head of the list of rules - fterm->next...	*/
typedef struct rule {
    struct rulegroup *group;	/* the group this rule belongs to	*/
    struct layer *layer;	/* the layer of connection (if not NIL) */
    int widthSpec;		/* user specified width			*/
    int orientation;		/* user specified orientation		*/
    struct fterm *fterm;	/* the "to" fterm 			*/
    struct rule *next;		/* the next rule from the "from" fterm	*/
} s_rule;

/* to make lists of fterms during rule construction			*/
typedef struct ftermlist {
    struct fterm *fterm;
    struct ftermlist *next;
} s_ftermlist;

typedef struct array {
    mkId arrayId;
    unsigned status;
    unsigned type;
    int numRows;
    int numCols;
    union {
	struct {
	    struct array *array;
	    int *rowmap;
	    int *colmap;
	} subarray;
	struct {
	    octId objectId;
	    octId facetId;
	    octId diditBagId;
	    st_table *elemTbl;
	    struct cpmGraph *Xgraph;
	    struct cpmGraph *Ygraph;
	    struct cpmNode **rowGrid;
	    struct cpmNode **colGrid;
	    struct cpmBranch **rowBranch;
	    struct cpmBranch **colBranch;
	    struct arraylist *sublist;
	    struct elemlist ***elemlist;
	    s_net *netlist;
	    int debug;
	    int printGraph;
	    int wireIntraInst;
	    int mergeRules;
	    int rowPitch;
	    int colPitch;
	    double expansion;
	    int createNoGeo;
	} array;
    } contents;
} s_array;

typedef struct arraylist {
    struct array *array;
    struct arraylist *next;
} s_arraylist;
