#include "mkInternal.h"
/*
 * mk_malloc_element
 */
s_element *
mk_malloc_element()
{
    s_element *element;

    element = ALLOC( s_element, 1 );
    element->objectId = oct_null_id;
    element->type = UNDEF;
    element->bbox.lowerLeft.x = element->bbox.lowerLeft.y = MK_NULL_COORD;
    element->bbox.upperRight.x = element->bbox.upperRight.y = MK_NULL_COORD;
    element->contents.instance = NIL( s_instance );
    cpmInitNode( &element->bnode, NIL( char ));
    cpmInitNode( &element->tnode, NIL( char ));
    cpmInitNode( &element->lnode, NIL( char ));
    cpmInitNode( &element->rnode, NIL( char ));
    return( element );
}
/*
 * mk_malloc_instance
 */
s_instance *
mk_malloc_instance()
{
    s_instance *instance;

    instance = ALLOC( s_instance, 1 );
    instance->master = NIL( s_master );
    instance->term = NIL( s_aterm );
    instance->offset.x = instance->offset.y = 0;
    return( instance );
} 
/*
 * mk_malloc_master
 */
s_master *
mk_malloc_master()
{
    s_master *master;

    master = ALLOC( s_master, 1 );
    master->objectId = oct_null_id;
    master->term = NIL( s_fterm );
    return( master );
}
/*
 * mk_malloc_net
 */
s_net *
mk_malloc_net( array, oNet )
s_array *array;			/* the array the net is in		*/
octObject *oNet;		/* the nets object			*/
{
    s_net *net;

    net = ALLOC( s_net, 1 );
    if( oNet == NIL( octObject )){
	net->objectId = oct_null_id;
	net->name = NIL( char );
	net->temporary = TRUE;
    } else {
	net->objectId = oNet->objectId;
	net->name = oNet->contents.net.name;
	net->temporary = FALSE;
	mk_hash_net( net );
    }
    net->next = array->contents.array.netlist;
    array->contents.array.netlist = net;
    net->segment = NIL( s_segment );
    return( net );
} 
/*
 * mk_malloc_segment
 */
s_segment *
mk_malloc_segment()
{
    s_segment *segment;

    segment = ALLOC( s_segment, 1 );
    segment->objectId = oct_null_id;
    cpmInitNode( &segment->node, NIL( char ));
    segment->termIn = NIL( s_aterm );
    segment->frameIn = NIL( s_tframe );
    segment->termOut = NIL( s_aterm );
    segment->frameOut = NIL( s_tframe );
    segment->orientation = UNDEF;
    segment->netNext = NIL( s_segment );
    segment->inNext = NIL( s_segment );
    segment->outNext = NIL( s_segment );
    return( segment );
} 
/*
 * mk_malloc_aterm
 */
s_aterm *
mk_malloc_aterm( element, objectId )
s_element *element;		/* the aterm's element			*/
octId objectId;			/* objectId of the actual term		*/
{
    s_aterm *aterm;

    aterm = ALLOC( s_aterm, 1 );
    aterm->objectId = objectId;
    aterm->net = NIL( s_net );
    aterm->element = element;
    aterm->instance = NIL( s_element );
    if( element->type == OCT_INSTANCE ){
	aterm->next = element->contents.instance->term;
	element->contents.instance->term = aterm;
    } else {
	aterm->next = NIL( s_aterm );
    }
    aterm->fterm = NIL( s_fterm );
    aterm->segIn = NIL( s_segment );
    aterm->segOut = NIL( s_segment );
    return( aterm );
}
/*
 * mk_malloc_fterm
 */
s_fterm *
mk_malloc_fterm( xId )
octId xId;
{
    s_fterm *fterm;

    fterm = ALLOC( s_fterm, 1 );
    fterm->xId = xId;
    fterm->frame = NIL( s_tframe );
    fterm->next = NIL( s_fterm );
    fterm->rule = NIL( s_rule );
    fterm->joinedGroup = 0;

    return( fterm );
}
/*
 * mk_malloc_tframe
 */
s_tframe *
mk_malloc_tframe()
{
    s_tframe *tframe;
    tframe = ALLOC( s_tframe, 1 );
    tframe->layer = NIL( s_layer );
    tframe->box.lowerLeft.x = tframe->box.lowerLeft.y = MK_NULL_COORD;
    tframe->box.upperRight.x = tframe->box.upperRight.y = MK_NULL_COORD;
    tframe->next = NIL( s_tframe );
    return( tframe );
}
/*
 * mk_malloc_layer
 */
s_layer *
mk_malloc_layer()
{
    s_layer *layer;

    layer = ALLOC( s_layer, 1 );
    layer->objectId = oct_null_id;
    layer->name = NIL( char );
    layer->next = NIL( s_layer );
    layer->minWidth = UNDEF;
    layer->widthSpec = MK_MIN_WIDTH;
    layer->hpriority = UNDEF;
    layer->vpriority = UNDEF;
    return( layer );
}
/*
 * mk_malloc_rulegroup
 */
s_rulegroup *
mk_malloc_rulegroup()
{
    s_rulegroup *rulegroup;

    rulegroup = ALLOC( s_rulegroup, 1 );
    rulegroup->name = NIL( char );
    rulegroup->enabled = FALSE;
    return( rulegroup );
}
/*
 * mk_malloc_rule
 */
s_rule *
mk_malloc_rule()
{
    s_rule *rule;

    rule = ALLOC( s_rule, 1 );
    rule->layer = NIL( s_layer );
    rule->group = NIL( s_rulegroup );
    rule->widthSpec = UNDEF_WIDTH;
    rule->fterm = NIL( s_fterm );
    rule->next = NIL( s_rule );
    return( rule );
}
/*
 * mk_malloc_ftermlist
 */
s_ftermlist *
mk_malloc_ftermlist()
{
    s_ftermlist *ftermlist;

    ftermlist = ALLOC( s_ftermlist, 1 );
    ftermlist->fterm = NIL( s_fterm );
    ftermlist->next = NIL( s_ftermlist );
    return( ftermlist );
}
/*
 * mk_malloc_array
 */
s_array *
mk_malloc_array()
{
    s_array *array;
    static int arrayId = 0;

    array = ALLOC( s_array, 1 );
    array->arrayId = ++arrayId;
    array->status = (unsigned)0;
    array->type = UNDEF;
    array->numRows = 0;
    array->numCols = 0;
    /* array contents is a union and is initialized in 			*/
    /* mk_init_array and mk_init_subarray 				*/
    return( array );
}
/*
 * mk_init_array
 */
void
mk_init_array( array )
s_array *array;
{
    array->type = ARRAY;
    array->contents.array.objectId = oct_null_id;
    array->contents.array.facetId = oct_null_id;
    array->contents.array.diditBagId = oct_null_id;
    array->contents.array.debug = 0;
    array->contents.array.printGraph = FALSE;
    array->contents.array.mergeRules = FALSE;
    array->contents.array.wireIntraInst = FALSE;
    array->contents.array.expansion = 1.0;
    array->contents.array.createNoGeo = 0;
    array->contents.array.elemTbl = NIL( st_table );
    array->contents.array.Xgraph = NIL( s_cpmGraph );
    array->contents.array.Ygraph = NIL( s_cpmGraph );
    array->contents.array.rowGrid = NIL( s_cpmNode * );
    array->contents.array.colGrid = NIL( s_cpmNode * );
    array->contents.array.rowBranch = NIL( s_cpmBranch * );
    array->contents.array.colBranch = NIL( s_cpmBranch * );
    array->contents.array.sublist = NIL( s_arraylist );
    array->contents.array.elemlist = NIL( s_elemlist ** );
    array->contents.array.netlist = NIL( s_net );
    array->contents.array.rowPitch = 0;
    array->contents.array.colPitch = 0;
}
/* 
 * mk_init_subarray
 */
void
mk_init_subarray( subarray )
s_array *subarray;
{
    subarray->type = SUBARRAY;
    subarray->contents.subarray.array = NIL( s_array );
    subarray->contents.subarray.rowmap = NIL( int );
    subarray->contents.subarray.colmap = NIL( int );
}
/*
 * mk_malloc_arraylist
 */
s_arraylist *
mk_malloc_arraylist()
{
    s_arraylist *arraylist;

    arraylist = ALLOC( s_arraylist, 1 );
    arraylist->array = NIL( s_array );
    arraylist->next = NIL( s_arraylist );
    return( arraylist );
}
