/*LINTLIBRARY*/
#include "mkInternal.h"
/*
 * mkOptions
 *
 * This file contains routines to set some of the array options.
 *
 * mkSetPrintGraph - sets array->contents.array.printGraph which 
 *                   causes the constraint graphs to be printed.
 * mkSetExpansion - sets array->contents.array.expansion which
 *                  causes the effective bounding boxes of 
 *                  instances to expand which separates the 
 *                  instances in the final array.
 * mkSetCreateNoGeo - sets array->contents.array.createNoGeo which
 *                    disables the creation of OCT_PATHS.
 * mkWireIntraInst - sets array->contents.array.wireIntraInst
 *                   which causes mkarray to try to wire terminals
 *                   which belong to the same instance.
 * mkMergeRules - sets array->contents.array.mergeRules
 *                which causes mkarray to merge wiring rules
 *                if they have the same name - otherwise and error occurs
 */
/*
 * mkSetDebug
 */
void 
mkSetDebug( arrayId, value )
mkId arrayId;
int value;
{
    s_array *array;

    MK_START( "mkSetDebug" );

    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.debug = value;
    MK_END();
}
/*
 * mkSetPrintGraph
 */
void 
mkSetPrintGraph( arrayId, value )
mkId arrayId;
int value;
{
    s_array *array;

    MK_START( "mkSetPrintGraph" );

    if(( value != TRUE )&&( value != FALSE )){
	errRaise( MK_PKG_NAME, 0,
	          "The value of PrintGraph must be TRUE or FALSE");
    }
    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.printGraph = value;
    array->contents.array.Xgraph->graphfile = fopen( "mkarray.x", "w" );
    array->contents.array.Ygraph->graphfile = fopen( "mkarray.y", "w" );
    if( array->contents.array.debug > 0 ){
	setbuf( array->contents.array.Xgraph->graphfile, NIL(char) );
	setbuf( array->contents.array.Ygraph->graphfile, NIL(char) );
    }
    MK_END();
}
/*
 * mkSetCreateNoGeo
 */
void 
mkSetCreateNoGeo( arrayId, value )
mkId arrayId;
int value;
{
    s_array *array;

    MK_START( "mkSetCreateNoGeo" );

    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.createNoGeo = value;
    MK_END();
}
/*
 * mkSetExpansion
 */
void 
mkSetExpansion( arrayId, value )
mkId arrayId;
double value;
{
    s_array *array;

    MK_START( "mkSetExpansion" );

    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.expansion = value;
    MK_END();
}
/*
 * mkWireIntraInst
 */
void 
mkWireIntraInstance( arrayId, value )
mkId arrayId;
int value;
{
    s_array *array;

    MK_START( "mkWireIntraInstance" );

    if(( value != TRUE )&&( value != FALSE )){
	errRaise( MK_PKG_NAME, 0,
	          "The value of WireIntraInstance must be TRUE or FALSE");
    }
    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.wireIntraInst = value;
    MK_END();
}
/*
 * mkMergeRules
 */
void 
mkMergeRules( arrayId, value )
mkId arrayId;
double value;
{
    s_array *array;

    MK_START( "mkMergeRules" );

    if(( value != TRUE )&&( value != FALSE )){
	errRaise( MK_PKG_NAME, 0,
	          "The value of MergeRules must be TRUE or FALSE");
    }
    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;
    array->contents.array.mergeRules = value;
    MK_END();
}
/*
 * mkSetArrayPitch
 *
 * sets the minimum horizontal and vertical pitch in the array
 */
void 
mkSetArrayPitch( arrayId, area, rowPitch, colPitch )
mkId arrayId;
octBox *area;
int rowPitch;
int colPitch;
{
    s_array *array;
    struct cpmBranch **branch;
    int index;

    MK_START( "mkSetArrayPitch" );

    array = mk_get_array( arrayId );
    mk_sanitize_area( array, area );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;

    array->contents.array.rowPitch = rowPitch ;
    array->contents.array.colPitch = colPitch ;

    /* setting the existing branches in the array			*/
    branch = array->contents.array.rowBranch;
    for( index = area->lowerLeft.y; index <= area->upperRight.y; index++ ){
	cpmSetBranch( array->contents.array.Ygraph, branch[ index ], 
		      rowPitch, CPM_INFINITY, 0 );
    }
    branch = array->contents.array.colBranch;
    for( index = area->lowerLeft.x; index <= area->upperRight.x; index++ ){
	cpmSetBranch( array->contents.array.Xgraph, branch[ index ], 
		      colPitch, CPM_INFINITY, 0 );
    }
	
    MK_END();
}
/*
 * mkSetArrayOrigin
 *
 * sets the lower left hand corner of the arrayt
 */
void 
mkSetArrayOrigin( arrayId, x, y, relative )
mkId arrayId;
int x;
int y;
int relative;
{
    s_array *array;

    MK_START( "mkSetArrayOrigin" );

    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY )array = array->contents.subarray.array;

    cpmSetGraphOrigin( array->contents.array.Xgraph, x, relative );
    cpmSetGraphOrigin( array->contents.array.Ygraph, y, relative );
    MK_END();
}
