#include "mkInternal.h"
/*
 * mkNewSegment
 *
 * create a segment between oTerm1 and oTerm2
 */
void
mkNewSegment( arrayId, oT1, oT2, layerName, widthSpec, orientation )
mkId arrayId;
octObject *oT1;
octObject *oT2;
char *layerName;
int widthSpec;
int orientation;
{
    s_array *array;
    s_element *elem1;
    s_element *elem2;
    s_aterm *term1;
    s_aterm *term2;
    s_layer *layer;
    octObject oTerm1;
    octObject oTerm2;
    octObject oInst;
    octGenerator gTerm;
    double x1Percent, x2Percent, y1Percent, y2Percent;
    int vDist;
    int hDist;

    MK_START( "mkNewSegment" );
    array = mk_get_array( arrayId );

    if(( layerName == NIL( char ))||( strcmp( layerName, "" ) == 0 )){
	layer = NIL( s_layer );
    } else {
	mk_get_layer( array, layerName, &layer );
    }

    if( oT1->type != OCT_TERM ){
	oInst = *oT1;
	octInitGenContents( &oInst, OCT_TERM_MASK, &gTerm );
	octGenerate( &gTerm, &oTerm1 );
	if( octGenerate( &gTerm, &oInst ) != OCT_GEN_DONE ){
	    octFreeGenerator( &gTerm );
	    errRaise( MK_PKG_NAME, 0, "%s %s",
		      "The first object must be an OCT_TERM",
		      "or an OCT_INSTANCE with one terminal");
	}
    } else {
	oTerm1 = *oT1;
    }
    if( oT2->type != OCT_TERM ){
	oInst = *oT2;
	octInitGenContents( &oInst, OCT_TERM_MASK, &gTerm );
	octGenerate( &gTerm, &oTerm2 );
	if( octGenerate( &gTerm, &oInst ) != OCT_GEN_DONE ){
	    octFreeGenerator( &gTerm );
	    errRaise( MK_PKG_NAME, 0, "%s %s",
		      "The second object must be an OCT_TERM",
		      "or an OCT_INSTANCE with one terminal");
	}
    } else {
	oTerm2 = *oT2;
    }
    /*  finding elem1 and term1						*/
    if(( elem1 = mk_get_element( array, oTerm1.objectId)) == NIL( s_element )){
	if(( elem1 = mk_get_element( array, oTerm1.contents.term.instanceId ))
	    == NIL( s_element )){
	    errRaise( MK_PKG_NAME, 0, "The first object isn't in the array");
	} else {
	    term1 = elem1->contents.instance->term;
	}
    } else {
	term1 = elem1->contents.term->instance->contents.instance->term;
    }
    while (!octIdsEqual(term1->objectId, oTerm1.objectId)) {
	term1 = term1->next;
	if( term1 == NIL( s_aterm )) MK_INT_ERROR( "lost an actual terminal");
    }

    /*  finding elem2 and term2						*/
    if(( elem2 = mk_get_element( array, oTerm2.objectId)) == NIL( s_element )){
	if(( elem2 = mk_get_element( array, oTerm2.contents.term.instanceId ))
	    == NIL( s_element )){
	    errRaise( MK_PKG_NAME, 0, "The second object isn't in the array");
	} else {
	    term2 = elem2->contents.instance->term;
	}
    } else {
	term2 = elem2->contents.term->instance->contents.instance->term;
    }
    while (!octIdsEqual(term2->objectId, oTerm2.objectId)) {
	term2 = term2->next;
	if( term2 == NIL( s_aterm )) MK_INT_ERROR( "lost an actual terminal");
    }
    if( orientation == MK_DC_SEG ){
	if(( elem1->bbox.upperRight.x < elem2->bbox.lowerLeft.x )||
	   ( elem2->bbox.upperRight.x < elem1->bbox.lowerLeft.x )){
	    if(( elem1->bbox.upperRight.y < elem2->bbox.lowerLeft.y )||
	       ( elem2->bbox.upperRight.y < elem1->bbox.lowerLeft.y )){
		errRaise( MK_PKG_NAME, 0,
			 "%s: %s (%d,%d)(%d,%d), %s (%d,%d)(%d,%d): %s",
			 "mkNewSegment: The segment is not manhattan",
			 "the first element's bbox is",
			 elem1->bbox.lowerLeft.x, elem1->bbox.lowerLeft.y,
			 elem1->bbox.upperRight.x, elem1->bbox.upperRight.y,
			 "the second element's bbox is",
			 elem2->bbox.lowerLeft.x, elem2->bbox.lowerLeft.y,
			 elem2->bbox.upperRight.x, elem2->bbox.upperRight.y,
			 "the bboxes must overlap in the x or y dimension");
	    } else {
		orientation = MK_HOR_SEG;
	    }
	} else if(( elem1->bbox.upperRight.y < elem2->bbox.lowerLeft.y )||
		  ( elem2->bbox.upperRight.y < elem1->bbox.lowerLeft.y )){
	    orientation = MK_VER_SEG;
	} else {
	    /* the elements overlap.  In this case, a heuristic 	*/
	    /* distance between the centers of the two terminals is 	*/
	    /* found in both dimensions and the orientation is set to 	*/
	    /* the dimension with the longest distance.	 The center	*/
	    /* of a element is weighted by the center of the term	*/
	    /* in the transformed instance.  See the code for details.	*/

	    /* finding the transformed bounding box of the terms	*/
	    mk_relative_tCenter( term1, &x1Percent, &y1Percent );
	    mk_relative_tCenter( term2, &x2Percent, &y2Percent );

	    hDist = elem1->bbox.lowerLeft.x - elem2->bbox.lowerLeft.x +
	            (( elem1->bbox.upperRight.x - elem1->bbox.lowerLeft.x ) *
		     x1Percent ) -
	            (( elem2->bbox.upperRight.x - elem2->bbox.lowerLeft.x ) *
		     x2Percent );
	    if( hDist < 0 )hDist = -hDist;

	    vDist = elem1->bbox.lowerLeft.y - elem2->bbox.lowerLeft.y +
	            (( elem1->bbox.upperRight.y - elem1->bbox.lowerLeft.y ) *
		     y1Percent ) -
	            (( elem2->bbox.upperRight.y - elem2->bbox.lowerLeft.y ) *
		     y2Percent );
	    if( vDist < 0 )vDist = -vDist;

	    if( vDist >= hDist ) orientation = MK_VER_SEG;
	    else orientation = MK_HOR_SEG;
	}
    } else if ( orientation == MK_NONMAN_SEG ){
	orientation = UNDEF;
    } else if (( orientation != MK_HOR_SEG )&&( orientation != MK_VER_SEG )){
	errRaise( MK_PKG_NAME, 0,
		  "The orientation field must be %s, %s, %s or %s",
		  "MK_DC_SEG", "MK_HOR_SEG", 
		  "MK_VER_SEG", "MK_NONMAN_SEG");
    }

    if( widthSpec >= MK_MAX_WIDTH ){
	mk_new_segment( array, term1, term2, orientation, 
			layer, widthSpec, TRUE );
    } else {
	errRaise( MK_PKG_NAME, 0,
	         "The minimum field must be %s, %s, or %s",
		 "MK_MIN_WIDTH", "MK_MAX_WIDTH", "> 0" );
    }
    MK_END();
}
