/*LINTLIBRARY*/
#include "cpmInternal.h"
/*
 *	place non-critical nodes in correct locations.
 *	modified for general branch weight: 1/9/85
 */
cpmNoncrit( graph )
struct cpmGraph *graph;
{
    int debug = FALSE;
    struct cpmNode *node, *FnextNode();
    struct cpmBranch *br;
    int newub, newlb, newloc, force, forceIn, forceOut;
    int verbose = FALSE;
        
    if(((graph->status&CPM_VERBOSE) == CPM_VERBOSE)||
       ((graph->status&CPM_DEBUG) == CPM_DEBUG)){
       verbose = TRUE;
    }
    if((graph->status&CPM_DEBUG) == CPM_DEBUG)debug = TRUE;
    if(verbose)fprintf(stderr,"cpmNoncrit: ");
	
    FschedInit();
    for(node=graph->nodeList; node != NULL; node=node->list) {
	if( node->upper == node->lower ){
	    node->loc=node->lower;
	} else {
	    node->loc = (node->upper+node->lower)/2;
	    Fschedule( node );
	}
    }

    while( (node=FnextNode()) != NULL ) {

	newlb = -CPM_INFINITY;
	newub = CPM_INFINITY;	

	for(forceIn=0, br=node->fanin; br!= NULL; br=br->nextfi) {
	    newlb= MAX(newlb, ((br->fromNode)->loc+br->lower));
	    newub= MIN(newub, ((br->fromNode)->loc+br->upper));
	    forceIn += br->force;
	}
	for(forceOut=0, br=node->fanout; br != NULL; br=br->nextfo) {
	    newlb= MAX(newlb, ((br->toNode)->loc-br->upper));
	    newub= MIN(newub, ((br->toNode)->loc-br->lower));
	    forceOut += br->force;
	}

	force = forceIn + forceOut;
		
	if(forceIn >= CPM_INFINITY) {
	    if(forceOut < CPM_INFINITY) newloc=newlb;
	    else newloc = (newlb+newub)/2;
	} else if(forceOut >= CPM_INFINITY) newloc=newub;
	else if(force == 0) newloc = (newlb+newub)/2;
	else newloc = newlb+(newub-newlb)*forceOut/force;
	if(debug)fprintf(stderr,"noncrit: %s : nlb=%d nub=%d (%d %d->%d %d)",
	    node->name,newlb,newub,node->lower,node->loc,newloc,node->upper);

	if(( newloc != node->loc )||( newloc < node->lower )||
	   ( newloc > node->upper)){
	    node->loc = newloc;
	    if(debug)fprintf(stderr," - changed\n");
	    for(br=node->fanin; br != NULL; br=br->nextfi){
		if((br->fromNode)->lower != (br->fromNode)->upper)
		    Fschedule((br->fromNode));
	    }
	    for(br=node->fanout; br != NULL; br=br->nextfo){
		if((br->toNode)->lower != (br->toNode)->upper)
		    Fschedule((br->toNode));
	    }
	} else if(debug) fprintf(stderr," - no change\n");
    }
    if(verbose)fprintf(stderr,"\n");
    return( OK );
}
