/*LINTLIBRARY*/
#include "cpmInternal.h"

int ActiveBubbles, BubbleCount;

/*
 * check for overconstraints using the event-driven
 * algorithm.
 *
 */
cpmOvercon(graph )
struct cpmGraph *graph;
{

    int verbose = FALSE;
    int debug = FALSE;
    struct cpmBranch *br;
    struct cpmNode *node;
    struct cpmPath *bubble, *bubPath;
    double maxSched;
    extern double LschedCount();
    int numOvers, moved, newLow, newHigh;
    extern struct cpmNode *LnextNode();

    if(( graph->status&CPM_DEBUG ) == CPM_DEBUG )debug = TRUE;
    if(((graph->status&CPM_VERBOSE) == CPM_VERBOSE)||(debug == TRUE))
	verbose = TRUE;
    maxSched = (double)graph->numNodes*(double)graph->numBranches;
    ActiveBubbles = 0;
    BubbleCount = 0;
    numOvers = 0;
    if(verbose)fprintf(stderr,"overcon: solving for overconstraints; %d levels\n",
	graph->maxLevels);

    LschedInit( graph->maxLevels );
/*
 *     Create the origin bubble at each upper-bound node
 */
    for(node=graph->nodeList; node != NULL; node=node->list) {
        if(node->upperOutFlg && node->numfanout > 1) {
            bubble = ALLOC( struct cpmPath, 1 );
            bubble->branchPoint = node;
            bubble->lower = 0;
            bubble->upper = 0;    /* 0 to anchor it */
            bubble->active = TRUE;
            bubble->next = NULL;
            node->path = bubble;
            node->count = 1;
            ActiveBubbles++;
            BubbleCount++;
/*      schedule nodes with upper bound output constraints to start */
            Lschedule( node );
            if(debug) fprintf(stderr,"scheduled %s\n",node->name);
        }
    }


/*
 *    Now loop through all the nodes until no more activity
 */
    while((node=LnextNode( FORWARD )) != NULL && LschedCount()<maxSched) {
    if(debug) fprintf(stderr,"PROC. NODE %s:\n",node->name);

/*
 *    ... for each bubble at this node
 */
    for(bubble=node->path; bubble != NULL; bubble=bubble->next) {

/*
 *    For all active bubbles, but not the last one...
 */
    if(bubble->active == TRUE && ((bubble->branchPoint)->count > 2 ||
	bubble->branchPoint==node)) {
/*
 * process fanout branches
 */
for(br=node->fanout; br != NULL; br=br->nextfo) {
    if(debug) fprintf(stderr,"NODE %s:\n",node->name);
    if(br->toNode != bubble->branchPoint && br->lower > -CPM_INFINITY){
        moved = FALSE;
        for(bubPath=(br->toNode)->path;; bubPath = bubPath->next) {
            if(bubPath == NULL) {
/*
 *  not there, so add it
 */
                bubPath = ALLOC( struct cpmPath, 1 );
                bubPath->branchPoint = bubble->branchPoint;
                bubPath->lower= MAX((bubble->lower)+br->lower,-CPM_INFINITY);
                bubPath->upper= MIN((bubble->upper)+br->upper,CPM_INFINITY);
	        bubPath->lowerConstraint = node;
	        bubPath->upperConstraint = node;
                bubPath->next = (br->toNode)->path;
                (br->toNode)->path = bubPath;
                BubbleCount++;
	        ActiveBubbles++;
	        bubPath->active = TRUE;
	        (bubble->branchPoint)->count++;
	        moved = TRUE;
		if(debug)fprintf(stderr,"created ");
                break;
           }
/*
 *  found it
 */
           if(bubPath->branchPoint == bubble->branchPoint) {
	       if(debug)fprintf(stderr,"found ");
	       break;
	    }
        }
/*
 *  update constraints & schedule if necessary
 */
        newLow = MAX((bubble->lower)+br->lower,-CPM_INFINITY);
        if(newLow > bubPath->lower) {
            bubPath->lower = newLow;
            bubPath->lowerConstraint = node;
            moved = TRUE;
        }
        newHigh = MIN((bubble->upper)+br->upper,CPM_INFINITY);
        if(newHigh < bubPath->upper) {
          bubPath->upper = newHigh;
          bubPath->upperConstraint = node;
          moved = TRUE;
        }
	if(debug) {
	    fprintf(stderr,"fanout: from %s at %s (%d %d), count: %d",
	       (bubPath->branchPoint)->name, (br->toNode)->name,
	       bubPath->lower, bubPath->upper, (bubPath->branchPoint)->count);
	    if(moved) fprintf(stderr," moved\n");
	    else fprintf(stderr,"\n");
	}
        if(moved) {
	    Lschedule( br->toNode );
	    if( bubPath->lower > bubPath->upper) {
		numOvers++;
		(br->toNode)->overcon = TRUE;
		(bubble->branchPoint)->overcon = TRUE;
	    }
            if( bubPath->active == FALSE ) {
		   bubPath->active = TRUE;
	   	   ActiveBubbles++;
	           (bubble->branchPoint)->count++;
	    }
	}
    }
}

/*
 * process fanin branches
 */
    for(br=node->fanin; br != NULL; br=br->nextfi ) {
	if(br->upper<CPM_INFINITY && br->fromNode != bubble->branchPoint){
            moved = FALSE;
            for(bubPath=(br->fromNode)->path;; bubPath = bubPath->next) {
                if(bubPath == NULL) {
/*
 *  not there, so add it
 */
                    bubPath = ALLOC( struct cpmPath, 1 );
                    bubPath->branchPoint = bubble->branchPoint;
                    bubPath->lower= MIN((bubble->lower)-br->upper,CPM_INFINITY);
		    if(bubble->upper >= CPM_INFINITY) bubPath->upper = CPM_INFINITY;
                  else bubPath->upper=MIN((bubble->upper)-br->lower,CPM_INFINITY);
	            bubPath->lowerConstraint = node;
	            bubPath->upperConstraint = node;
                    bubPath->next = (br->fromNode)->path;
                    (br->fromNode)->path = bubPath;
                    BubbleCount++;
	            ActiveBubbles++;
	            bubPath->active = TRUE;
	            (bubble->branchPoint)->count++;
	            moved = TRUE;
	            if(debug)fprintf(stderr,"created ");
                    break;
               }
/*
 *  found it
 */
               if(bubPath->branchPoint == bubble->branchPoint){
	            if(debug)fprintf(stderr,"found ");
		    break;
                }
	    }
/*
 *  update constraints & schedule if necessary
 */
            newLow = MAX((bubble->lower)-br->upper,-CPM_INFINITY);
            if(newLow > bubPath->lower) {
                bubPath->lower = newLow;
                bubPath->lowerConstraint = node;
                moved = TRUE;
            }
	    if(bubble->upper >= CPM_INFINITY) newHigh = CPM_INFINITY;
            else newHigh = MIN((bubble->upper)-br->lower,CPM_INFINITY);
            if(newHigh < bubPath->upper) {
              bubPath->upper = newHigh;
              bubPath->upperConstraint = node;
              moved = TRUE;
            }
	    if(debug) {
	        fprintf(stderr,"fanin: from %s at %s (%d %d), count: %d",
	          (bubPath->branchPoint)->name, (br->fromNode)->name,
	          bubPath->lower,bubPath->upper,(bubPath->branchPoint)->count);
	        if(moved) fprintf(stderr," moved\n");
	        else fprintf(stderr,"\n");
	    }
            if(moved) {
		Lschedule( br->fromNode );
	        if( bubPath->lower > bubPath->upper) {
		    numOvers++;
		    (br->fromNode)->overcon = TRUE;
		    (bubble->branchPoint)->overcon = TRUE;
	        }
   	        if( bubPath->active == FALSE ) {
			bubPath->active = TRUE;
			ActiveBubbles++;
	                (bubble->branchPoint)->count++;
	        }
	    }
	}
    }
/*
 *  deactivate this bubble
 */
    if( bubble->branchPoint != node ) {
        bubble->active = FALSE;
        (bubble->branchPoint)->count--;
        ActiveBubbles--;
    }
   }
  }
}
fprintf(stderr,"overcon: overconstraints detected = %d\n", numOvers);
fprintf(stdout,"overcon: overconstraints detected = %d\n", numOvers);
fprintf(stdout,"overcon: the nodes involved are:\n");
/*
for(node=graph->nodeList;node != NULL; node = node->list) {
    if(debug) {
        for(bubble=node->path;bubble != NULL; bubble=bubble->next)
            if(bubble->active == TRUE)
	fprintf(stderr, "source: %s, terminal: %s lower: %d upper: %d\n",
	bubble->branchPoint->name, node->name,bubble->lower,bubble->upper);
    }
    if(node->overcon == TRUE ) {
        for(bubble=node->path;bubble != NULL; bubble=bubble->next)
        if(bubble->active == TRUE && bubble->lower > bubble->upper) {
          fprintf(stdout, "source: %s, terminal: %s\n", bubble->branchPoint->name,
                node->name);
	    fprintf(stdout, "upper path: ");
            traceUpper(node,bubble->branchPoint);
            fprintf(stdout, "lower path: ");
            traceLower(node,bubble->branchPoint);
        }
    }
}
*/
fprintf(stderr,"overcon: total schedules = %g\n",LschedCount());
fprintf(stderr,"overcon: total bubbles created = %d\n",BubbleCount);
fprintf(stderr,"overcon: total active bubbles remaining = %d\n",ActiveBubbles);

if(numOvers != 0) return(FALSE);
else return(TRUE);
}

/*
 *    Trace the upper-limiting path to the source node
 */
traceUpper(fromNode, toNode)
struct cpmNode *fromNode, *toNode;
{
    struct cpmPath *bubble;
    fprintf(stdout,"%s ",fromNode->name);
    if(fromNode == toNode) {
    fprintf(stdout,"\n");
    return;
    }
    for(bubble=fromNode->path;bubble != NULL; bubble=bubble->next) {
    if(bubble->branchPoint == toNode)
    if(bubble->upperConstraint != NULL) {
       traceUpper(bubble->upperConstraint,toNode);
       return;
    }
    else {
       fprintf(stdout, "... lost it!\n");
       return;
    }
    }
    fprintf(stdout, "... lost it!\n");
    return;
}

/*
 *    Trace the lower-limiting path to the source node
 */
traceLower(fromNode, toNode)
struct cpmNode *fromNode, *toNode;
{
    struct cpmPath *bubble;
    fprintf(stdout,"%s ",fromNode->name);
    if(fromNode == toNode) {
    fprintf(stdout,"\n");
    return;
    }
    for(bubble=fromNode->path;bubble != NULL; bubble=bubble->next) {
    if(bubble->branchPoint == toNode)
    if(bubble->lowerConstraint != NULL) {
       traceLower(bubble->lowerConstraint,toNode);
       return;
    }
    else {
       fprintf(stdout, "... lost it!\n");
       return;
    }
    }
    fprintf(stdout, "... lost it!\n");
    return;
}
