/*LINTLIBRARY*/
/* Copyright (c) 1987, Charles J Kring. All rights reserved.		*/
#include "mkInternal.h" 
/*
 * mkPlaceInstances()
 *
 * Recover the coordinates of the instances from the constraint graphs. 
 */
void
mkPlaceInstances( array )
s_array *array;
{
    extern enum st_retval mk_place_instance();

    MK_START( "mkPlaceInstances" );
    st_foreach( array->contents.array.elemTbl, mk_place_instance,
		(char *)array);
    MK_END();
}

/*
 * mk_place_instance
 *
 * If element is an instance, its corresponding oct instance is translated
 * so that the lowerLeft corner of its bounding box is at the point 
 * element->lnode.loc, element->bnode.loc.
 *
 * Element->offset compensates for the fact that the master may not have
 * its lowerLeft corner at the origin.
 */
/*ARGSUSED*/
enum st_retval
mk_place_instance( key, value, arg )
char *key;			/* unused				*/
char *value;			/* element passed in fron st_foreach	*/
char *arg;			/* unused				*/
{
    s_element *element;		/* the element to be placed		*/
    octObject oInst;		/* OCT instance to be translated	*/

    element = (s_element *)value;
    if( element->type != OCT_INSTANCE )return( ST_CONTINUE );

    oInst.objectId = element->objectId;
    if( octGetById( &oInst ) < OCT_OK ){
	return( ST_STOP );
    }
    oInst.contents.instance.transform.translation.x = 
	element->lnode.loc + element->contents.instance->offset.x;
    oInst.contents.instance.transform.translation.y = 
	element->bnode.loc + element->contents.instance->offset.y;

    if( octModify( &oInst ) < OCT_OK ){
	return( ST_STOP );
    }
    return( ST_CONTINUE );
}


/*
 * mkPlaceNets
 *
 * The endpoints of every segment in every net are placed.  If a
 * segment is manhattan, the horizontal or vertical location of 
 * the segment is derived from the constraint graph and the endpoints
 * are derived from the terminal frames.  Otherwise, a message is printed
 * and the endpoints are totally derived from the terminal frames.
 */
void
mkPlaceNets( array ) 
s_array *array;
{
    octObject oFacet;	
#ifdef DIDIT
    octObject oDiditBag;
#endif
    octObject oNet;	
    octObject oSegment;	
    octObject oTerm;	
    octObject oFterm;	
    octPoint points[2];		/* end points of each oSegment		*/
    s_net *net;			/* nets in the netlist			*/
    s_segment *segment;		/* the segments				*/
    octBox frameIn;		/* segment->termIn terminal frame	*/
    octBox frameOut;		/* segment->termOut terminal frame	*/
    s_element *instIn;		/* segment->termIn->instance		*/
    s_element *instOut;		/* segment->termOut->instance		*/
    int newNet;

    MK_START( "mkPlaceNets" );

    oFacet.objectId = array->contents.array.facetId;
    octGetById( &oFacet );
#ifdef DIDIT
    oDiditBag.objectId = array->contents.array.diditBagId;
    octGetById( &oDiditBag );
#endif

    for( net = array->contents.array.netlist; net != NIL(s_net); 
	 net = net->next ){

	if( net->segment == NIL( s_segment ))continue;

	/* create a net if necessary					*/
	if (octIdIsNull(net->objectId)) {

	    oNet.type = OCT_NET;
	    oNet.contents.net.name = "";
	    oNet.contents.net.width = 1;
	    octCreate( &oFacet, &oNet );
#ifdef DIDIT
	    octAttach( &oDiditBag, &oNet );
#endif
	    net->objectId = oNet.objectId;
	    newNet = TRUE;
	} else {
	    newNet = FALSE;
	}
	for( segment = net->segment; segment != NIL( s_segment); 
	     segment = segment->netNext){

	    frameIn = segment->frameIn->box;
	    instIn = segment->termIn->instance;
	    mk_transform_tframe( instIn, &frameIn );

	    frameOut = segment->frameOut->box;
	    instOut = segment->termOut->instance;
	    mk_transform_tframe( instOut, &frameOut );

	    if ( segment->orientation == MK_HOR_SEG ) {

		points[0].x = instIn->lnode.loc +
			      (frameIn.upperRight.x + frameIn.lowerLeft.x)/2;
		points[1].x = instOut->lnode.loc +
			      (frameOut.upperRight.x + frameOut.lowerLeft.x)/2;
		if( points[0].x == points[1].x )points[1].x++;
		points[0].y = points[1].y = segment->node.loc;

	    } else if ( segment->orientation == MK_VER_SEG ) { 

		points[0].x = points[1].x = segment->node.loc;
		points[0].y = instIn->bnode.loc +
			      (frameIn.upperRight.y + frameIn.lowerLeft.y)/2;
		points[1].y = instOut->bnode.loc +
			      (frameOut.upperRight.y + frameOut.lowerLeft.y)/2;
		if( points[0].y == points[1].y )points[1].y++;

	    } else {

		points[0].x = instIn->lnode.loc +
			      (frameIn.upperRight.x + frameIn.lowerLeft.x)/2;
		points[1].x = instOut->lnode.loc +
			      (frameOut.upperRight.x + frameOut.lowerLeft.x)/2;

		points[0].y = instIn->bnode.loc +
			      (frameIn.upperRight.y + frameIn.lowerLeft.y)/2;
		points[1].y = instOut->bnode.loc +
			      (frameOut.upperRight.y + frameOut.lowerLeft.y)/2;
		if(( points[0].x == points[1].x )&&
		   ( points[0].y == points[1].y )){
		    /* 0 length nonmanhattanpath, making it horizontal	*/
		    points[1].y++;
		} else {
		    fprintf(stderr, "%s: endpoints are (%d, %d),(%d, %d))\n",
			    "WARNING: Mkarray produced a non manhattan segment",
			    points[ 0 ].x, points[ 0 ].y,
			    points[ 1 ].x, points[ 1 ].y );
		}
	    }
	    oSegment.objectId = segment->objectId;
	    /* if array.contents.array.createNoGeo ... */
	    if (!octIdIsNull(oSegment.objectId)) {
		octGetById( &oSegment );
	        octPutPoints( &oSegment , 2 , points );
            }
	    if( newNet ){
		/* attaching the OCT segment and terms to the net	*/
		/* attaching any promoted terminals to the net		*/

		if (!octIdIsNull(oSegment.objectId))
		    octAttachOnce( &oNet, &oSegment );

		oTerm.objectId = segment->termIn->objectId;
		octGetById( &oTerm );
		if (!octIdIsNull(oSegment.objectId))
		    octAttachOnce( &oSegment, &oTerm );

		if( octGenFirstContainer( &oTerm, OCT_TERM_MASK, 
					  &oFterm ) == OCT_OK )
		    octAttachOnce( &oNet, &oFterm );

		if( instIn->contents.instance->type == INSTANCE ){
		    octAttachOnce( &oNet, &oTerm );
		}

		oTerm.objectId = segment->termOut->objectId;
		octGetById( &oTerm );
		if (!octIdIsNull(oSegment.objectId))
		    octAttachOnce( &oSegment, &oTerm );

		if( octGenFirstContainer( &oTerm, OCT_TERM_MASK, 
					  &oFterm ) == OCT_OK )
		    octAttachOnce( &oNet, &oFterm );

		if( instOut->contents.instance->type == INSTANCE ){
		    octAttachOnce( &oNet, &oTerm );
		}
	    }
	}
    }
    MK_END();
}
