#include "mkInternal.h"
/*
 * mk_read_instance
 * Copyright (c) 1987, Charles J Kring. All rights reserved.
 *
 * The master and terminal information needed for each instance is read
 * in this routine.  If the master has not been seen before, mk_read_master
 * is called to read the master and hash it.  Then the actual terminals
 * in the instance are matched up to formal terminal in the master.
 *
 */
void
mk_read_instance( array, element )
s_array *array;
s_element *element;
{
    octObject oInst;	
    octObject oMaster;	
    octObject oTerm;		/* actual terms attached to oInst 	*/
    octObject oFterm;		/* formal terms containing oTerm 	*/
    octObject oNet;		/* optional net containing an oTerm	*/
    octObject oBag;		/* the INSTANCES and CONNECTORS bags	*/
    octGenerator gTerm;		/* generates oTerms			*/
    s_instance *inst;		/* the instance 			*/
    s_fterm *fterm;		/* a formal term on the master		*/
    s_aterm *term;		/* an actual terminal on the instance	*/
    s_net *net;			/* net containing the actual term	*/
    octBox bbox;		/* bounding box of the instance		*/
    octCoord size;		/* size of the instance 		*/
    char name[128];		/* a buffer for constraint node names	*/

    inst = element->contents.instance = mk_malloc_instance();

    oInst.objectId = element->objectId;
    octGetById( &oInst );

    mk_read_master(array, inst, &oInst );

    /* check to see if this is an INSTANCE of CONNECTOR			*/
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = INSTANCESBAG;
    if( octGetContainerByName( &oInst, &oBag ) >= OCT_OK ){
	inst->type = INSTANCE;
    } else {
	oBag.contents.bag.name = CONNECTORSBAG;
	if( octGetContainerByName( &oInst, &oBag ) >= OCT_OK ){
	    inst->type = CONNECTOR;
	} else {
	    fprintf( stderr,
		     "WARNING: %s %s: %s",
		     ohFormatName( &oInst ),
		     "isn't in either the INSTANCES or the CONNECTORS bag",
		     "assuming its an INSTANCE" );
	    inst->type = INSTANCE;
	}
    }

    /* normalizing the instance transform and getting the inst bbox	*/
    oInst.contents.instance.transform.translation.x = 0;
    oInst.contents.instance.transform.translation.y = 0;
    octModify( &oInst );
    octBB( &oInst, &bbox );

    inst->offset.x = - MIN( bbox.lowerLeft.x, bbox.upperRight.x );
    inst->offset.y = - MIN( bbox.lowerLeft.y, bbox.upperRight.y );

    /* generating the actual terminals on the instance and match them 	*/
    /* to the corresponding formal terminals on the master 		*/

    octInitGenContents( &oInst, OCT_TERM_MASK, &gTerm );
    while( octGenerate( &gTerm , &oTerm ) == OCT_OK ){
	for( fterm = inst->master->term; fterm != NIL( s_fterm ); 
	     fterm = fterm->next ){
	    if( fterm->xId == oTerm.contents.term.formalExternalId ){
		term = mk_malloc_aterm( element, oTerm.objectId );
		term->fterm = fterm;
		term->instance = element;
		break;
	    }
	}

	/* if the terms don't match up, the interface facet is bad	*/
	if( fterm == NIL( s_fterm )){
	    oMaster.objectId = inst->master->objectId;
	    octGetById( &oMaster );
	    errRaise( MK_PKG_NAME, 0,
		    "%s doesn't match a formal terminal in %s",
		    ohFormatName( &oTerm ),
		    ohFormatName( &oMaster ));
	}

	/* if the term is attached to a net, this net is propogated 	*/
	/* in the wiring phase						*/
	if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet ) == OCT_OK ){
	    if(( net = mk_get_net( oNet.objectId )) == NIL(s_net )){
		net = mk_malloc_net( array, &oNet );
	    }
	    term->net = net;
	    /* fixing this  ( sometimes useful )symbolic policy violation*/
	    if( inst->type == CONNECTOR )octDetach( &oNet, &oTerm );

	} else if( octGenFirstContainer( &oTerm, OCT_TERM_MASK, 
					 &oFterm ) == OCT_OK ){
	    if( octGenFirstContainer(&oFterm, OCT_NET_MASK, &oNet) == OCT_OK ){
		if(( net = mk_get_net( oNet.objectId )) == NIL(s_net )){
		    net = mk_malloc_net( array, &oNet );
		}
		term->net = net;
	    }
	} else {
	    term->net = NIL( s_net );
	}
    }

    /* adding the instance to the constraint graph			*/

    if( array->contents.array.printGraph ){
	(void) sprintf(name,"%s.l", oInst.contents.instance.name);
	cpmResetNodeName( &element->lnode, name );

	(void) sprintf(name,"%s.r", oInst.contents.instance.name);
	cpmResetNodeName( &element->rnode, name );

	(void) sprintf(name,"%s.t", oInst.contents.instance.name);
	cpmResetNodeName( &element->tnode, name );

	(void) sprintf(name,"%s.b", oInst.contents.instance.name);
	cpmResetNodeName( &element->bnode, name );
    }
    cpmInstallNode( array->contents.array.Xgraph, &element->lnode );
    cpmInstallNode( array->contents.array.Xgraph, &element->rnode );
    cpmInstallNode( array->contents.array.Ygraph, &element->tnode );
    cpmInstallNode( array->contents.array.Ygraph, &element->bnode );

    size = ( bbox.upperRight.x - bbox.lowerLeft.x) * 
	     array->contents.array.expansion;

    cpmInstallBranch( array->contents.array.Xgraph, 
		      &element->lnode , &element->rnode , NIL( s_cpmBranch ),
		      size, size, 0 );

    size = ( bbox.upperRight.y - bbox.lowerLeft.y) * 
	     array->contents.array.expansion;

    cpmInstallBranch( array->contents.array.Ygraph, 
		      &element->bnode, &element->tnode, NIL( s_cpmBranch ),
		      size, size, 0);

}
