#include "mkInternal.h"
/*
 * mk_read_master
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 */
void
mk_read_master(array, instance, oInstance ) 
s_array *array;
s_instance *instance;
octObject *oInstance;		/* Need to find the master of this instance */
{
    octObject oMaster;		/* oct master */
    octObject oTerm;		/* oct formal terminals */
    octObject oBox;		/* oct boxes implementing the formal terms */
    octObject oLayer;		/* oct layers the boxes are attached to */
    octGenerator gTerm;		/* oct terminal gernerator */
    octGenerator gBox;		/* oct box generator */
    s_master *master;		/* the master 				*/
    s_fterm *term;		/* formal terminals on the master	*/
    s_layer *layer;		/* layers generated off terminals	*/
    s_tframe *frame;	
    s_master *mk_get_master();	/* return a master from the hash table  */
    int joinedGroup = 0;
    octObject oJTerms;
    octObject oJoined;
    octGenerator gJoined;
    /*
     * open the master
     */
    oMaster.type = OCT_FACET;
    oMaster.contents.facet.cell = oInstance->contents.instance.master;
    oMaster.contents.facet.view = oInstance->contents.instance.view;
    oMaster.contents.facet.facet = "interface";
    oMaster.contents.facet.version = OCT_CURRENT_VERSION;
    oMaster.contents.facet.mode = "r";
    if ( octOpenFacet( &oMaster ) < OCT_OK ) {
	errRaise( MK_PKG_NAME, 0, 
		  "Could not open %s", ohFormatName( &oMaster ));
    } 

    /*
     * If the master has already been opened, get it and return.
     * Otherwise, it needs to be created
     */

    instance->master = mk_get_master( oMaster.objectId );
    if( instance->master != NIL( s_master )) return;
    /*
     * Create and hash the master
     */
    master = mk_malloc_master();
    master->objectId = oMaster.objectId;
    mk_hash_master( master );
    instance->master = master;
    /*
     * generate the terminals on the master
     */
    octInitGenContents( &oMaster, OCT_TERM_MASK, &gTerm );
    while( octGenerate( &gTerm , &oTerm ) == OCT_OK ){

	term = mk_malloc_fterm( oTerm.contents.term.formalExternalId );
	term->next = master->term;
	master->term = term;

	/* generate the implementation of the formal term		*/
	octInitGenContents( &oTerm, OCT_BOX_MASK, &gBox );
	while( octGenerate( &gBox, &oBox ) == OCT_OK ){

	    /* generate the layer the box is attached to		*/
	    octGenFirstContainer( &oBox, OCT_LAYER_MASK, &oLayer );
	    mk_get_layer( array, oLayer.contents.layer.name, &layer );

	    /*** check for multiple implementations on one layer? */
	    frame = mk_malloc_tframe();
	    frame->layer = layer;
	    frame->box = oBox.contents.box;
	    frame->next = term->frame;
	    term->frame = frame;
	}
    }

    /* reading the JOINEDTERMS and JOINED bags the determine 	*/
    /* the electrically equivalent terms				*/

    oJTerms.type = OCT_BAG;
    oJTerms.contents.bag.name = JOINEDTERMSBAG;
    if( octGetByName( &oMaster, &oJTerms ) >= OCT_OK ){
	octInitGenContents( &oJTerms, OCT_BAG_MASK, &gJoined );
	while( octGenerate( &gJoined, &oJoined ) == OCT_OK ){
	    if( strcmp( oJoined.contents.bag.name, JOINEDBAG ) != 0 ) continue;
	    joinedGroup++;
	    octInitGenContents( &oJoined, OCT_TERM_MASK, &gTerm );
	    while( octGenerate( &gTerm, &oTerm ) == OCT_OK ){
		for( term = master->term; term != NIL( s_fterm );
		     term = term->next ){
		    if( term->xId == oTerm.contents.term.formalExternalId )
			break;
		}
		if( term == NIL( s_fterm )){
		    errRaise( MK_PKG_NAME, 0,
			     "Couldn't match a term to JOINED %s in %s",
			     JOINEDBAG,
			     ohFormatName( &oTerm ),
			     ohFormatName( &oMaster ));
		} else {
		    term->joinedGroup = joinedGroup;
		}
	    }
	}
    }
}
