/*LINTLIBRARY*/
#include "mkInternal.h"
/* list of free ftermlists instead on mallocing an freeing all the time */
static s_ftermlist *freeTlist = NIL(s_ftermlist);
/*
 * mkReadRules
 * Copyright (c) 1987, Charles J Kring. All rights reserved.
 *
 * Open the Rulefacet and read the connection rules it contains.  The 
 * rules are hashed according by name in Ruletbl and linked to the row
 * and column structure in readRow and readCol. 
 */
/*
 * for every bag name RULEBAG attached to the oFacet
 *    for every bag name CONNECTBAG attached to the rule bag
 *       generate the first bag named TERMGROUPBAG or terminal
 *       generate the second bag named TERMGROUPBAG or terminal
 *           if a bag was found, generate the m terminals is contains
 *      if only one terminal was found, make a self rule.
 *      if only one bag was found, make a rule between every pair
 *        of terminals it constains, including self rules ( n*n rules )
 *	    if one bag and one terminal was found, make a fule between 
 *        the terminal and every termainal in the bag ( n or m rules )
 *      otherwise two bags were found, and make a rule between every
 *        terminal in the first bag and every terminal in the second bag
 *        (mxn rules)
 */
void
mkReadRules( arrayId, oFacet )
mkId arrayId;
octObject *oFacet;
{
    octObject oRule;		/* RULEBAG attached to Rfacet */
    octObject oConnect;		/* CONNECTBAG attached to RULEBAG */
    octObject oObj;		/* objects connected to CONNECTBAG */
    octGenerator gRule;		/* oRule generator */
    octGenerator gConnect;	/* oConnect generator */
    octGenerator gObj;		/* oObj generator */
    s_ftermlist *tlist1;
    s_ftermlist *tlist2;
    s_ftermlist *ptr1;
    s_ftermlist *ptr2;
    s_rule *rule;		/* a connection rule */
    s_rulegroup *rulegroup;	/* a set of connection rules */
    s_rulegroup *tmpgroup;	/* tmp rulegroup		*/
    s_rulegroup *mk_get_rule();
    char *mk_group_name();		/* return the name of a rulegroup */
    s_array *array;
    int oneList;
    int widthSpec;
    int orientation;
    s_layer *layer;
    void mk_read_layers();

    MK_START( "mkReadRules" );

    array = mk_get_array( arrayId );

    mk_read_layers( array, oFacet );

    /* generating RULEBAG bags which become rulegroups			*/
    octInitGenContents( oFacet, OCT_BAG_MASK, &gRule );
    while( octGenerate( &gRule, &oRule ) == OCT_OK ){

	if( strcmp( oRule.contents.bag.name, RULEBAG ) != 0 )continue;

	rulegroup = mk_malloc_rulegroup();
	if(( rulegroup->name = mk_group_name( &oRule )) == NIL(char) ){
	    fprintf( stderr,"WARNING: ignoring an unnamed rulegroup\n" );
	    FREE( rulegroup );
	    continue;
	}
	if(( tmpgroup = mk_get_rule( rulegroup->name )) != NIL( s_rulegroup )){
	    if( array->contents.array.mergeRules ){
		fprintf( stderr, 
			 "WARNING: merging multiple rules named %s\n",
			 rulegroup->name );
	    } else {
		errRaise( MK_PKG_NAME, 0,
			  "Found two rules named %s: %s",
			  rulegroup->name,
			  "they can be merged with mkMergeRules()");
	    }
	    FREE( rulegroup );
	    rulegroup = tmpgroup;
	} else {
	    mk_hash_rule( rulegroup );
	}

	octInitGenContents( &oRule, OCT_BAG_MASK, &gConnect );
	while( octGenerate( &gConnect, &oConnect ) == OCT_OK ){

	    if( strcmp( oConnect.contents.bag.name, CONNECTBAG ) != 0 )continue;
	    /* checking for a layer the connection should be made in	*/
	    octInitGenContents(&oConnect, OCT_LAYER_MASK, &gObj );
	    if( octGenerate( &gObj, &oObj ) == OCT_OK ){
		mk_get_layer( array, oObj.contents.layer.name, &layer );
	    } else {
		layer = NIL( s_layer );
	    }

	    /* reading the MK_SEG_WIDTH prop to set the rule specWidth	*/
	    /* defaults to a UNDEF_WIDTH 				*/
	    oObj.type = OCT_PROP;
	    oObj.contents.prop.name = MK_WIDTH_PROP;
	    if( octGetByName( &oConnect, &oObj ) == OCT_OK ){
		if( oObj.contents.prop.type == OCT_INTEGER ){
		    widthSpec = oObj.contents.prop.value.integer;
		} else {
		    fprintf( stderr,
			     "Ignoring a non integer valued %s property on a %s bag\n",
			     MK_WIDTH_PROP, CONNECTBAG );
		}
	    } else {
		widthSpec = UNDEF_WIDTH;
	    }
	    /* reading the MK_ORIENTATION_PROP to set the rule orientation */
	    /* defaults to a MK_DC_SEG 					*/
	    oObj.type = OCT_PROP;
	    oObj.contents.prop.name = MK_ORIENTATION_PROP;
	    if( octGetByName( &oConnect, &oObj ) == OCT_OK ){
		if( oObj.contents.prop.type == OCT_INTEGER ){
		    orientation = oObj.contents.prop.value.integer;
		} else {
		    fprintf( stderr,
			     "Ignoring a non integer valued %s property on a %s bag\n",
			     MK_WIDTH_PROP, CONNECTBAG );
		}
	    } else {
		orientation = MK_DC_SEG;
	    }
	    octInitGenContents( &oConnect, OCT_BAG_MASK|OCT_TERM_MASK, &gObj );
	    if( octGenerate( &gObj, &oObj ) == OCT_OK ){
		mk_build_tlist( array, &oObj, &tlist1 );
	    } else {
		continue;
	    }
	    if( octGenerate( &gObj, &oObj ) == OCT_OK ){
		mk_build_tlist( array, &oObj, &tlist2 );
		oneList = FALSE;
		octFreeGenerator( &gObj );
	    } else {
		oneList = TRUE;
	    }

	    /* matching the tlists to derive the rules			*/
	    for( ptr1 = tlist1; ptr1 != NIL( s_ftermlist ); ptr1 = ptr1->next ){
		for(( oneList ? ( ptr2 = tlist1 ) : ( ptr2 = tlist2 ));
		    ptr2 != NIL( s_ftermlist ); ptr2 = ptr2->next ){

		    rule = mk_malloc_rule();
		    rule->widthSpec = widthSpec;
		    rule->orientation = orientation;
		    rule->group = rulegroup;
		    rule->fterm = ptr2->fterm;
		    rule->next = ptr1->fterm->rule;
		    rule->layer = layer;
		    ptr1->fterm->rule = rule;
		}
	    }
	    /* putting the tlists back on the freeTlist			*/
	    if( tlist1 != NIL( s_ftermlist )){
		for( ptr1 = tlist1; ptr1->next != NIL( s_ftermlist ); 
		     ptr1 = ptr1->next );
		ptr1->next = freeTlist;
		freeTlist = ptr1;
	    }
	    if(( !oneList )&&( tlist2 != NIL( s_ftermlist ))){
		for( ptr2 = tlist2; ptr2->next != NIL( s_ftermlist ); 
		     ptr2 = ptr2->next );
		ptr2->next = freeTlist;
		freeTlist = ptr2;
	    }
	}
    }
    /* freeing the freeTlist						*/
    while( freeTlist != NIL( s_ftermlist )){
	ptr1 = freeTlist;
	freeTlist = freeTlist->next;
	FREE( ptr1 );
    }
    MK_END();
}

/*
 * mk_group_name()
 *
 * Return a string which contains the value of the property named RULENAMEPROP.
 */
char *
mk_group_name( object )
octObject *object;
{
    octObject prop;
    prop.type = OCT_PROP;
    prop.contents.prop.name = RULENAMEPROP;
    if( octGetByName( object, &prop ) == OCT_OK ){
	if( prop.contents.prop.type == OCT_STRING ){
	    return ( util_strsav( prop.contents.prop.value.string ));
	}
    }
    return ( NIL(char));
}

void
mk_read_layers( array, oFacet )
s_array *array;
octObject *oFacet;
{
    octObject oBag;
    octObject oProp;
    octObject oLayer;
    octGenerator gLayer;
    s_layer *layer;
    int priority;

    oProp.type = OCT_PROP;
    oBag.type = OCT_BAG;
    priority = 100;
    oBag.contents.bag.name = HORIZONTAL_LAYERS;
    if( octGetByName( oFacet, &oBag ) == OCT_OK ){
	octInitGenContents( &oBag, OCT_LAYER_MASK, &gLayer );
	while( octGenerate( &gLayer, &oLayer ) == OCT_OK ){
	    mk_get_layer( array, oLayer.contents.layer.name, &layer );
	    layer->hpriority = priority--;
	}
    }
    priority = 100;
    oBag.contents.bag.name = VERTICAL_LAYERS;
    if( octGetByName( oFacet, &oBag ) == OCT_OK ){
	octInitGenContents( &oBag, OCT_LAYER_MASK, &gLayer );
	while( octGenerate( &gLayer, &oLayer ) == OCT_OK ){
	    mk_get_layer( array, oLayer.contents.layer.name, &layer );
	    layer->vpriority = priority--;
	}
    }
    octInitGenContents( oFacet, OCT_LAYER_MASK, &gLayer );
    while( octGenerate( &gLayer, &oLayer ) == OCT_OK ){
	mk_get_layer( array, oLayer.contents.layer.name, &layer );
	oProp.contents.prop.name = MK_WIDTH_PROP;
	if( octGetByName( &oLayer, &oProp ) == OCT_OK ){
	    if( oProp.contents.prop.type == OCT_INTEGER ){
		layer->widthSpec = oProp.contents.prop.value.integer;
	    } else {
		fprintf( stderr,
			"Ignoring a non string valued %s prop on layer %s\n",
		     MK_WIDTH_PROP, oLayer.contents.layer.name );
	    }
	}
    }
}
