/*LINTLIBRARY*/
#include "cpmInternal.h"
/*
 *  Sanity check for the graph. Determine source & sink.
 */
cpmSanitize(graph)
struct cpmGraph *graph;
{
    int verbose = FALSE;
    int debug = FALSE;
    struct cpmNode *node, *foNode;
    struct cpmBranch *br, *fanin, *fanout;
    int allDone, faninLevel;
    extern struct cpmNode *FnextNode();

    if(( graph->status&CPM_DEBUG ) == CPM_DEBUG )debug = TRUE;
    if(((graph->status&CPM_VERBOSE) == CPM_VERBOSE)||(debug == TRUE))verbose = TRUE;

    for(node=graph->nodeList;node != NULL;node = node->list) {
	if (debug) {
	    fprintf(stderr,"%s :       fanins:  ",node->name);
			    for(br=node->fanin;br != NULL;br=br->nextfi)
	    fprintf(stderr,"(%s %d %d) ",(br->fromNode)->name,br->lower,br->upper);
	    fprintf(stderr,"\n fanouts: ");
			    for(br=node->fanout;br != NULL;br=br->nextfo)
	    fprintf(stderr,"(%s %d %d) ",(br->toNode)->name,br->lower,br->upper);
	    fprintf(stderr,"\n");
	}
	if (node->numfanin == 0) {
	    cpmInstallBranch( graph, &graph->sourceNode, node, 
			      NIL( s_cpmBranch ), 0,CPM_INFINITY,0);
	}
	if (node->numfanout == 0) {
	    cpmInstallBranch( graph, node, &graph->sinkNode,
			      NIL( s_cpmBranch ), 0, CPM_INFINITY, 0);
	}
    }
/*
 *      Level the graph; note that loops are stderrors in this case.
 */
    FschedInit( );
    graph->sourceNode.level = 0;
    Fschedule(&graph->sourceNode);
    if(debug) fprintf(stderr,"sanitize: entering levelling code\n");

    while((node=FnextNode()) != NULL) {
/*
 *      Update level of current node
 */
    for(fanin=node->fanin; fanin != NULL; fanin=fanin->nextfi) {
	    faninLevel = (fanin->fromNode)->level;
	node->level = MAX( node->level, faninLevel+1 );
    }
    graph->maxLevels = MAX( node->level, graph->maxLevels );
    if(debug) fprintf(stderr,"sanitize: at %s level is %d\n",
		    node->name, node->level);
/*
 *      Schedule fanout nodes if all their inputs have been processed
 */
    for(fanout=node->fanout;fanout != NULL;fanout=fanout->nextfo) {
	    foNode = fanout->toNode;
	if(foNode->sched == NULL) {
	    if(debug)fprintf(stderr,"sanitize: proc %s,",
				    foNode->name);
	    allDone = TRUE;
	    for(fanin=foNode->fanin;fanin!=NULL; fanin=fanin->nextfi) {
		if((fanin->fromNode)->level == UNDEF) allDone = FALSE;
	    }
	    if(allDone == TRUE) {
		Fschedule(foNode);
		if(debug)fprintf(stderr, " scheduled\n");
	    }
	    else if(debug)fprintf(stderr," not scheduled\n");
	    }
	}
    }
    if(debug)fprintf(stderr,"sanitize: finished levelling\n");
    graph->maxLevels++; /* add 1 for level 0 */
/*
 *      Walk node list & look for unprocessed (loop) nodes
 */
    allDone = TRUE;
    for(node=graph->nodeList; node != NULL; node=node->list)
	if(node->level == UNDEF) {
	    if(debug) fprintf(stderr,"sanitize: node %s is part of loop\n",
		              node->name);
	    allDone = FALSE;
	}
    if(allDone == FALSE) return(ERROR);
    if(verbose)fprintf(stderr,"cpmSanitize: %d schedules\n",FschedCount());
    graph->status &= ~CPM_SANITIZE;
    return(OK);
}
