#include "mkInternal.h"
/*
 * mk_set_seg_width
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 * 
 */
void
mk_get_seg_width( segment, widthSpec, width )
s_segment *segment;
int widthSpec;
int *width;
{
    octCoord size1 = INFINITY;
    octCoord size2 = INFINITY;
    s_layer *layer;
    octBox frameIn;
    octBox frameOut;

    layer = segment->frameIn->layer;
    if( widthSpec == MK_MIN_WIDTH ){
	*width = layer->minWidth;
    } else if ( widthSpec > MK_MIN_WIDTH ){
	*width = widthSpec;
    } else if(( widthSpec == MK_MAX_WIDTH )||
	      ( layer->widthSpec == MK_MAX_WIDTH )){

	frameIn = segment->frameIn->box;
	mk_transform_tframe( segment->termIn->instance, &frameIn );
	frameOut = segment->frameOut->box;
	mk_transform_tframe( segment->termOut->instance, &frameOut );
	if( segment->orientation == MK_HOR_SEG ){

	    size1 = frameIn.upperRight.y - frameIn.lowerLeft.y;
	    size2 = frameOut.upperRight.y - frameOut.lowerLeft.y;

	} else if( segment->orientation == MK_VER_SEG ){

	    size1 = frameIn.upperRight.x - frameIn.lowerLeft.x;
	    size2 = frameOut.upperRight.x - frameOut.lowerLeft.x;

	} else {

	    size1 = MIN(( frameIn.upperRight.x - frameIn.lowerLeft.x ),
			( frameIn.upperRight.y - frameIn.lowerLeft.y ));
	    size2 = MIN(( frameOut.upperRight.x - frameOut.lowerLeft.x ),
			( frameOut.upperRight.y - frameOut.lowerLeft.y ));

	}
	*width = MIN( size1, size2); 
    } else if ( layer->widthSpec > 0 ){
	*width = layer->widthSpec;
    } else {
	*width = layer->minWidth;
    }
}

/*
 * mk_set_seg_layer
 *
 * if layer is not NIL, find an implemetation for termIn and termOut.
 * if it is, find a common layer implementation for termIn and termOut
 */
void
mk_set_seg_tframes( segment, layer )
s_segment *segment;
s_layer *layer;		/* optional layer the segment will be made in	*/
{
    s_tframe *frameIn, *frameOut; 
    s_tframe *bestIn = NIL( s_tframe );
    s_tframe *bestOut = NIL( s_tframe );
    int priority = UNDEF - 1;
    octObject oTermIn, oTermOut, oMasterIn, oMasterOut;

    if( layer != NIL( s_layer )){
	for( bestIn = segment->termIn->fterm->frame; 
	     bestIn != NIL( s_tframe ); bestIn = bestIn->next ){
	    if( bestIn->layer == layer )break;
	}
	for( bestOut = segment->termOut->fterm->frame; 
	     bestOut != NIL( s_tframe ); bestOut = bestOut->next ){
	    if( bestOut->layer == layer )break;
	}
    } else {
	/* find a common layer between the two terminals 		*/
	if( segment->orientation == MK_VER_SEG ){
	    for( frameIn = segment->termIn->fterm->frame; 
		 frameIn != NIL( s_tframe ); frameIn = frameIn->next ){
		for( frameOut = segment->termOut->fterm->frame; 
		     frameOut != NIL(s_tframe ); frameOut = frameOut->next ){
		    if(( frameIn->layer == frameOut->layer )&&
		       ( frameIn->layer->vpriority > priority )){
			priority = frameIn->layer->vpriority;
			bestIn = frameIn;
			bestOut = frameOut;
		    }
		}
	    }
	} else if ( segment->orientation == MK_HOR_SEG ){
	    for( frameIn = segment->termIn->fterm->frame; 
		 frameIn != NIL( s_tframe ); frameIn = frameIn->next ){
		for( frameOut = segment->termOut->fterm->frame; 
		     frameOut != NIL( s_tframe ); frameOut = frameOut->next ){
		    if(( frameIn->layer == frameOut->layer )&&
		       ( frameIn->layer->hpriority > priority )){
			priority = frameIn->layer->hpriority;
			bestIn = frameIn;
			bestOut = frameOut;
		    }
		}
	    }
	} else {
	    for( frameIn = segment->termIn->fterm->frame; 
		 frameIn != NIL( s_tframe ); frameIn = frameIn->next ){
		layer = frameIn->layer;
		for( frameOut = segment->termOut->fterm->frame; 
		     frameOut != NIL( s_tframe ); frameOut = frameOut->next ){
		    if( frameIn->layer == frameOut->layer ){
			if ( frameIn->layer->vpriority > priority ){
			    priority = frameIn->layer->vpriority;
			    bestIn = frameIn;
			    bestOut = frameOut;
			}
			if ( frameIn->layer->hpriority > priority ){
			    priority = frameIn->layer->hpriority;
			    bestIn = frameIn;
			    bestOut = frameOut;
			}
		    }
		}
	    }
	}
    }
    segment->frameIn = bestIn;
    segment->frameOut = bestOut;
    if( layer == NIL( s_layer )){
	if(( bestIn == NIL( s_tframe ))||( bestOut == NIL( s_tframe ))){
	    oTermIn.objectId = segment->termIn->objectId;
	    oMasterIn.objectId = 
		segment->termIn->instance->contents.instance->master->objectId;
	    oTermOut.objectId = segment->termOut->objectId;
	    oMasterOut.objectId = 
		segment->termOut->instance->contents.instance->master->objectId;
	    octGetById( &oTermIn );
	    octGetById( &oMasterIn );
	    octGetById( &oTermOut );
	    octGetById( &oMasterOut );
	    errRaise( MK_PKG_NAME, 0,
		     "Couldn't find a common layer in %s of %s and %s of %s",
		     ohFormatName( &oTermIn ), ohFormatName( &oMasterIn ),
		     ohFormatName( &oTermOut ), ohFormatName( &oMasterOut ));
	}
    } 
    if( bestIn == NIL( s_tframe )){
	oTermIn.objectId = segment->termIn->objectId;
	oMasterIn.objectId = 
	    segment->termIn->instance->contents.instance->master->objectId;
	octGetById( &oTermIn );
	octGetById( &oMasterIn );
	errRaise( MK_PKG_NAME, 0,
		  "%s of %s does not have an implementation in layer %s",
		  ohFormatName( &oTermIn ), ohFormatName( &oMasterIn ),
		  layer->name );
    }
    if( bestOut == NIL( s_tframe )){
	oTermIn.objectId = segment->termOut->objectId;
	oMasterIn.objectId = 
	    segment->termOut->instance->contents.instance->master->objectId;
	octGetById( &oTermOut );
	octGetById( &oMasterOut );
	errRaise( MK_PKG_NAME, 0,
		  "%s of %s does not have an implementation in layer %s",
		  ohFormatName( &oTermOut ), ohFormatName( &oMasterOut ),
		  layer->name );
    }
}
