#include "mkInternal.h"
/*
 * segments.c
 *
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 *
 * routines to create new segments and add them to the mkarray data structure
 */
/*
 * mk_new_segment
 *
 * create a new segment if there was not one between term1 and term2 already
 */
void
mk_new_segment( array, term1, term2, orientation, 
		layer, widthSpec, redundantSegs )
s_array *array;
s_aterm *term1;		/* the first terminal 				*/
s_aterm *term2;		/* the second terminal 				*/
int orientation;	/* the segment is horizontal or vertical or NULL */
s_layer *layer;		/* optional specification of layer 		*/
int widthSpec;		/* the width specification of the rule		*/
int redundantSegs;	/* if TRUE, allows redundant segs		*/
{
    octObject oFacet;		/* OCT facet containing the array	*/
    octObject oSegment;		/* OCT path implementing the segment	*/
    octObject oNet;		/* OCT net containing oSegment		*/
    octObject oTerm;		/* OCT terminals containing oSegment	*/
    octObject oFterm;		/* OCT formal terminals			*/
    octObject oLayer;		/* OCT layer containing oSegment	*/
#ifdef DIDIT
    octObject oDidit;		/* OCT bag containing new objects	*/
#endif
    s_segment *segment;		/* the segment which is created 	*/
    octBox box1;		/* terminal frames and bboxes		*/
    octBox box2;		/* terminal frames and bboxes		*/
    s_aterm *termIn;		/* the upperLeft of term1 and term2	*/
    s_aterm *termOut;		/* the lowerRight of term1 and term2	*/
    static int segId = 0; 	/* unique segment id if printGraph	*/
    char nameBuffer[10];	/* buffer for segment node names	*/
    int width;			/* width of the segment			*/
    int joinedGroup;
    s_aterm *jterm;
    void mk_merge_net();

    /* return MK_OK if a segment already exists between term1 and term2	*/
    for( segment = term1->segIn; segment != NIL( s_segment );
	 segment = segment->inNext ){
	if( segment->termOut == term2 ) return;
    }

    for( segment = term1->segOut; segment != NIL( s_segment );
	 segment = segment->outNext ){
	if( segment->termIn == term2 ) return;
    }
    oFacet.objectId = array->contents.array.facetId;
    octGetById( &oFacet );

    /* order term1 and term2 into termIn and termOut			*/

    box1 = term1->element->bbox;
    box2 = term2->element->bbox;
    if( orientation == MK_VER_SEG ){

	if(( box2.lowerLeft.y < box1.lowerLeft.y )||
	   (( box2.lowerLeft.y == box1.lowerLeft.y )&&
            ( box2.upperRight.y < box1.upperRight.y ))){
	    termIn = term1;
	    termOut = term2;
	} else {
	    termIn = term2;
	    termOut = term1;
	}

    } else if ( orientation == MK_HOR_SEG ){

	if(( box2.lowerLeft.x < box1.lowerLeft.x )||
	   (( box2.lowerLeft.x == box1.lowerLeft.x )&&
            ( box2.upperRight.x < box1.upperRight.x ))){
	    termIn = term2;
	    termOut = term1;
	} else {
	    termIn = term1;
	    termOut = term2;
	}

    } else {

	if(( box2.lowerLeft.x < box1.lowerLeft.x )||
	   (( box2.lowerLeft.x == box1.lowerLeft.x )&&
            ( box2.upperRight.x < box1.upperRight.x ))){
	    termIn = term2;
	    termOut = term1;
	} else {
	    termIn = term1;
	    termOut = term2;
	}

    }

    segment = mk_malloc_segment();
    segment->termIn = termIn;
    segment->termOut = termOut;
    segment->orientation = orientation;
    mk_set_seg_tframes( segment, layer );
    mk_get_seg_width( segment, widthSpec, &width );

    /* return if a segment will overlap another segment 		*/
    /* coming out of termIn or into termOut 				*/
    if(( redundantSegs == FALSE )&&( mk_redundant_segment( segment ))){
        FREE( segment );
	return;
    }

    /* merging termIn->net and termOut->net				*/
    if(( termIn->net == NIL( s_net ))&&(termOut->net != NIL( s_net ))){
	termIn->net = termOut->net;
    } else if(( termOut->net == NIL( s_net ))&&(termIn->net != NIL( s_net ))){
	termOut->net = termIn->net;
    } else if(( termIn->net == NIL( s_net ))&&(termOut->net == NIL( s_net ))){
	termIn->net = termOut->net = mk_malloc_net( array, NIL( octObject ));
    } else if( termIn->net != termOut->net ){
	mk_merge_net( termIn->net, termOut->net );
    }
    /* merging termIn's JOINED terminals				*/
    if(( joinedGroup = termIn->fterm->joinedGroup ) != 0 ){
	for( jterm = termIn->instance->contents.instance->term;
	    jterm != NIL( s_aterm ); jterm = jterm->next ){
	    if((jterm->fterm->joinedGroup == joinedGroup)&&(jterm != termIn)){
		if( jterm->net == NIL( s_net ))
		    jterm->net = termIn->net;
		else if( jterm->net != termIn->net )
		    mk_merge_net( jterm->net, termIn->net );
	    }
	}
    }
    /* merging termOut's JOINED terminals				*/
    if(( joinedGroup = termOut->fterm->joinedGroup ) != 0 ){
	for( jterm = termOut->instance->contents.instance->term;
	    jterm != NIL( s_aterm ); jterm = jterm->next ){
	    if((jterm->fterm->joinedGroup == joinedGroup)&&(jterm != termOut)){
		if( jterm->net == NIL( s_net ))
		    jterm->net = termOut->net;
		else if( jterm->net != termOut->net )
		    mk_merge_net( jterm->net, termOut->net );
	    }
	}
    }

    /* creating the OCT segment	and necessary attachments		*/

    layer = segment->frameIn->layer;
    oLayer.objectId = layer->objectId;
    octGetById( &oLayer );

    if( array->contents.array.createNoGeo == 0 ){
	oSegment.type = OCT_PATH;
	oSegment.contents.path.width = width;
	octCreate( &oLayer, &oSegment );
	segment->objectId = oSegment.objectId;
#ifdef DIDIT
	oDidit.objectId = array->contents.array.diditBagId;
	octGetById( &oDidit );
	octAttach( &oDidit, &oSegment );
#endif
    }

    if (!octIdIsNull(oNet.objectId = termOut->net->objectId)) {
	octGetById( &oNet );
	if( array->contents.array.createNoGeo == 0 )
	    octAttach( &oNet, &oSegment );

	oTerm.objectId = segment->termIn->objectId;
	octGetById( &oTerm );
	if( array->contents.array.createNoGeo == 0 )
	    octAttach( &oSegment, &oTerm );

	if(octGenFirstContainer( &oTerm, OCT_TERM_MASK, &oFterm ) == OCT_OK)
	    octAttachOnce( &oNet, &oFterm );

	if( segment->termIn->instance->contents.instance->type == INSTANCE){
	   octAttachOnce( &oNet, &oTerm );
	}

	oTerm.objectId = segment->termOut->objectId;
	octGetById( &oTerm );
	if( array->contents.array.createNoGeo == 0 )
	    octAttach( &oSegment, &oTerm );

	if(octGenFirstContainer( &oTerm, OCT_TERM_MASK, &oFterm ) == OCT_OK)
	    octAttachOnce( &oNet, &oFterm );

	if(segment->termOut->instance->contents.instance->type == INSTANCE){
	   octAttachOnce( &oNet, &oTerm );
	}
    }


    /* insert the segment into the data structure			*/

    segment->netNext = term1->net->segment;
    term1->net->segment = segment;

    segment->inNext = segment->termIn->segIn;
    segment->termIn->segIn = segment;

    segment->outNext = segment->termOut->segOut;
    segment->termOut->segOut = segment;

    /* add the segment to the constraint graph				*/

    box1 = segment->frameIn->box;
    mk_transform_tframe( termIn->instance, &box1 );
    box2 = segment->frameOut->box;
    mk_transform_tframe( termOut->instance, &box2 );
    if( orientation == MK_VER_SEG ){

	if( array->contents.array.printGraph ){
	    (void) sprintf(nameBuffer,"%d.v", ++segId );
	    cpmResetNodeName( &segment->node, nameBuffer );
	}
	cpmInstallNode( array->contents.array.Xgraph, &segment->node );
	cpmInstallBranch(array->contents.array.Xgraph, 
			 &segment->termIn->instance->lnode, &segment->node,
			 NIL( s_cpmBranch ),
			 box1.lowerLeft.x + width/2,
			 box1.upperRight.x - width/2, 0);
	cpmInstallBranch(array->contents.array.Xgraph, 
			 &segment->termOut->instance->lnode, &segment->node,
			 NIL( s_cpmBranch ),
			 box2.lowerLeft.x + width/2,
			 box2.upperRight.x - width/2, 0);

    } else if ( orientation == MK_HOR_SEG ){

	if( array->contents.array.printGraph ){
	    (void) sprintf(nameBuffer,"%d.h", ++segId );
	    cpmResetNodeName( &segment->node, nameBuffer );
	}
	cpmInstallNode( array->contents.array.Ygraph, &segment->node );
	cpmInstallBranch(array->contents.array.Ygraph, 
			 &segment->termIn->instance->bnode, &segment->node,
			 NIL( s_cpmBranch ),
			 box1.lowerLeft.y + width/2,
			 box1.upperRight.y - width/2, 0);
	cpmInstallBranch(array->contents.array.Ygraph, 
			 &segment->termOut->instance->bnode, &segment->node,
			 NIL( s_cpmBranch ),
			 box2.lowerLeft.y + width/2,
			 box2.upperRight.y - width/2, 0);
    }
    return;
}

/*
 * mk_relax_segments
 *
 * unconstrain any segments which cause the constraint graph to be
 * overconstrained or to contain loops
 */
void
mk_relax_segments( array, orientation )
s_array *array;
int orientation;
{
    s_net *net;
    s_segment *segment;
    struct cpmGraph *graph;
    if( orientation == MK_HOR_SEG ){
	graph = array->contents.array.Ygraph;
    } else if( orientation == MK_VER_SEG ){
	graph = array->contents.array.Xgraph;
    } else {
	MK_INT_ERROR("bad orientation");
    }
    for( net = array->contents.array.netlist; net != NIL( s_net );
	 net = net->next ){
	for( segment = net->segment; segment != NIL( s_segment );
	     segment = segment->netNext ){
	    if(( segment->orientation == orientation )&&
	       (( segment->node.level == UNDEF )||
		( segment->node.overcon == TRUE ))){
		segment->orientation = UNDEF;
		graph->status |= CPM_SANITIZE;
		cpmSetBranch( graph, segment->node.fanin,
			      0, CPM_INFINITY, 0 );
		cpmSetBranch( graph, segment->node.fanin->nextfi,
			      0, CPM_INFINITY, 0 );
	       
	    }
	}
    }
    return;
}

/* 
 * mk_merge_net
 *
 * merges net1 and net2
 *
 * delete the OCT implementation of net1 if necessary
 */
void
mk_merge_net( net1, net2 )
s_net *net1;
s_net *net2;
{
    s_segment *segment;
    octObject oNet;
    octObject oSegment;
    octObject oTerm;
    octObject oFterm;
    s_net *merged, *used;

    if( net1 == net2 ) return;

    if( net1->temporary && net2->temporary ){
	if (octIdIsNull(net1->objectId)) {
	    merged = net1;
	    used = net2;
	} else {
	    used = net1;
	    merged = net2;
	}
    } else if( net1->temporary ){
	merged = net1;
	used = net2;
    } else if( net2->temporary ){
	merged = net2;
	used = net1;
    } else {
	merged = net1;
	used = net2;
	fprintf( stderr, "%s %s %s %s and %s: merging %s into %s\n",
		 "Initial netlist and connection rules",
		 "produce an illegal final netlist where",
		 "a segment connects between two nets -",
		 merged->name, used->name,
		 merged->name, used->name );
    }

    /* attach the merged net's contents to the used net 		*/
    /* delete the merged net						*/
    if (!octIdIsNull(merged->objectId)) {
	oNet.objectId = merged->objectId;
	octGetById( &oNet );
	octDelete( &oNet );
	merged->objectId = oct_null_id;
    }
    if (!octIdIsNull(used->objectId)) {
	oNet.objectId = used->objectId;
	octGetById( &oNet );
	for( segment = merged->segment; segment != NIL( s_segment );
	     segment = segment->netNext ){
	    if (octIdIsNull(segment->objectId)) continue;

	    oSegment.objectId = segment->objectId;
	    octGetById( &oSegment );
	    octAttachOnce( &oNet, &oSegment );

	    oTerm.objectId = segment->termIn->objectId;
	    octGetById( &oTerm );
	    octAttachOnce( &oSegment, &oTerm );

	    if(octGenFirstContainer( &oTerm, OCT_TERM_MASK, &oFterm ) == OCT_OK)
		octAttachOnce( &oNet, &oFterm );

	    if( segment->termIn->instance->contents.instance->type == INSTANCE )
		octAttachOnce( &oNet, &oTerm );

	    oTerm.objectId = segment->termOut->objectId;
	    octGetById( &oTerm );
	    octAttachOnce( &oSegment, &oTerm );

	    if(octGenFirstContainer( &oTerm, OCT_TERM_MASK, &oFterm ) == OCT_OK)
		octAttachOnce( &oNet, &oFterm );

	    if( segment->termOut->instance->contents.instance->type == INSTANCE)
		octAttachOnce( &oNet, &oTerm );
	}
    }

    /* fixing the data structure					*/
    for( segment = merged->segment; segment != NIL( s_segment );
	 segment = segment->netNext ){
	segment->termIn->net = used;
	segment->termOut->net = used;
    }
    if( used->segment == NIL( s_segment )){
	used->segment = merged->segment;
    } else {
	segment = used->segment; 
	while( segment->netNext != NIL( s_segment )) segment = segment->netNext;
	segment->netNext = merged->segment;
    }
    merged->segment = NIL( s_segment );

}

/*
 * mk_redundant_segment
 *
 * return TRUE is the segment is redundant
 */
int
mk_redundant_segment( segment )
s_segment *segment;
{
    s_segment *seg;
    octBox *box1;
    octBox *box2;

    if( segment->orientation == MK_HOR_SEG ){
	box1 = &segment->termOut->element->bbox;
	for( seg = segment->termIn->segIn; seg != NIL( s_segment );
	     seg = seg->inNext ){
	    if( seg->orientation != MK_HOR_SEG ) continue;
	    box2 = &seg->termOut->element->bbox;
	    if(((( box1->lowerLeft.y >= box2->lowerLeft.y )&&
	         ( box1->lowerLeft.y <= box2->upperRight.y ))||
	        (( box2->lowerLeft.y >= box1->lowerLeft.y )&&
	         ( box2->lowerLeft.y <= box1->upperRight.y )))&&
	       ( segment->frameOut->layer == seg->frameOut->layer ))
		return( TRUE );
	}
	box1 = &segment->termIn->element->bbox;
	for( seg = segment->termOut->segOut; seg != NIL( s_segment );
	     seg = seg->outNext ){
	    if( seg->orientation != MK_HOR_SEG ) continue;
	    box2 = &seg->termIn->element->bbox;
	    if(((( box1->lowerLeft.y >= box2->lowerLeft.y )&&
	         ( box1->lowerLeft.y <= box2->upperRight.y ))||
	        (( box2->lowerLeft.y >= box1->lowerLeft.y )&&
	         ( box2->lowerLeft.y <= box1->upperRight.y )))&&
	       ( segment->frameIn->layer == seg->frameIn->layer ))
		return( TRUE );
	}
    } else if( segment->orientation == MK_VER_SEG ){
	box1 = &segment->termOut->element->bbox;
	for( seg = segment->termIn->segIn; seg != NIL( s_segment );
	     seg = seg->inNext ){
	    if( seg->orientation != MK_VER_SEG ) continue;
	    box2 = &seg->termOut->element->bbox;
	    if(((( box1->lowerLeft.x >= box2->lowerLeft.x )&&
	         ( box1->lowerLeft.x <= box2->upperRight.x ))||
	        (( box2->lowerLeft.x >= box1->lowerLeft.x )&&
	         ( box2->lowerLeft.x <= box1->upperRight.x )))&&
	       ( segment->frameOut->layer == seg->frameOut->layer ))
		return( TRUE );
	}
	box1 = &segment->termIn->element->bbox;
	for( seg = segment->termOut->segOut; seg != NIL( s_segment );
	     seg = seg->outNext ){
	    if( seg->orientation != MK_VER_SEG ) continue;
	    box2 = &seg->termIn->element->bbox;
	    if(((( box1->lowerLeft.x >= box2->lowerLeft.x )&&
	         ( box1->lowerLeft.x <= box2->upperRight.x ))||
	        (( box2->lowerLeft.x >= box1->lowerLeft.x )&&
	         ( box2->lowerLeft.x <= box1->upperRight.x )))&&
	       ( segment->frameIn->layer == seg->frameIn->layer ))
		return( TRUE );
	}
    }
    return( FALSE );
}
