/*LINTLIBRARY*/
#include "cpmInternal.h"

cpmInstallBranch( graph, fp, tp, bp, lower, upper, weight )
struct cpmGraph *graph;
struct cpmNode *fp, *tp;
struct cpmBranch *bp;
int lower, upper, weight;
{
    CPM_START( "cpmInstallBranch" );
    if(( fp == tp )||(fp == NIL( s_cpmNode ))||(tp == NIL( s_cpmNode ))){
	return;
    }
    if( bp == NIL( s_cpmBranch )) bp = ALLOC( s_cpmBranch, 1 );
	
    if ( graph->graphfile != NIL( FILE )){
	fprintf( graph->graphfile, "%s %s %d %d %d\n", 
	         fp->name, tp->name, lower, upper, weight );
    }
    cpmInitBranch( bp );
    cpmSetBranch( graph, bp, lower, upper, weight );

    bp->list  = graph->branchList;
    graph->branchList = bp;

    bp->fromNode = fp;
    bp->toNode = tp;

    bp->nextfo = fp->fanout;
    fp->fanout = bp;
    fp->numfanout++;

    bp->nextfi = tp->fanin;
    tp->fanin = bp;
    tp->numfanin++;
	
    if( upper < CPM_INFINITY ) {
	fp->upperOutFlg = TRUE;
	tp->upperInFlg = TRUE;
	graph->numUpperBounds++;
    }
    if( lower > -CPM_INFINITY ) {
	fp->lowerOutFlg = TRUE;
	tp->lowerInFlg = TRUE;
	graph->numLowerBounds++;
    }
    graph->numBranches++;
    /*
     * check to see if this branch upsets the level of the graph. 
     * if so set the graph status so that cpmSanitize is run before cpmSolve
     */
    if( tp->level < fp->level + 1 )graph->status |= CPM_SANITIZE;

    CPM_END();
}
/*
 * Change the weight of a branch in the node.  Set the solve bit if
 * the change could affect the solution of the graph
 */
cpmSetBranch( graph, bp, lower, upper, weight )
struct cpmGraph *graph;
struct cpmBranch *bp;
int lower, upper, weight;
{
    
    CPM_START( "cpmSetBranch" );
    /* if the branch was or may now be on the critical path,   		*/
    /* mark the graph so it will be solved again			*/
    /* this is not checked if the graph already need solving		*/
    if(( graph->status&CPM_SOLVE) != CPM_SOLVE ){
	if((( bp->lower != lower )&&
	    ( bp->fromNode->lower + bp->lower == bp->toNode->lower ))||
	   (( bp->upper != upper )&&
	    ( bp->fromNode->upper + bp->upper == bp->toNode->upper ))){
	    graph->status |= CPM_SOLVE;
	}
    }
    /* reset the branch							*/
    bp->lower = lower;
    bp->upper = upper;
    bp->force = weight;

    CPM_END();
}
/*
 * cpmRemoveBranch
 *
 * Remove a branch from the graph
 */
cpmRemoveBranch( graph, branch )
s_cpmGraph *graph;
s_cpmBranch *branch;
{
    s_cpmBranch *bp;
    int sanitize = TRUE;

    CPM_START( "cpmRemoveBranch" );

    /* in case the branch was never in the graph			*/
    if(( branch->fromNode == NIL( s_cpmNode ))||
       ( branch->toNode == NIL( s_cpmNode )))return;


    if(( bp = graph->branchList ) == branch ){
	graph->branchList = bp->list;
    } else {
	while( bp->list != NIL( s_cpmBranch )){
	    if( bp->list ==  branch ){
		bp->list = branch->list;
		break;
	    }
	    bp = bp->list;
	}
    }

    if(( bp = branch->fromNode->fanout ) == branch ){
	branch->fromNode->fanout = bp->nextfo;
    } else {
	while( bp->nextfo != NIL( s_cpmBranch )){
	    if( bp->nextfo ==  branch ){
		bp->nextfo = branch->nextfo;
		break;
	    }
	    bp = bp->nextfo;
	}
    }
    if(( bp = branch->toNode->fanin ) == branch ){
	branch->toNode->fanin = bp->nextfi;
    } else {
	while( bp->nextfi != NIL( s_cpmBranch )){
	    if( bp->nextfi ==  branch ){
		bp->nextfi = branch->nextfi;
		break;
	    }
	    bp = bp->nextfi;
	}
    }
    /*
     * check to see if this branch upsets the level of the graph. 
     * if so set the graph status so that cpmSanitize is run before cpmSolve
     */
    for( bp = branch->toNode->fanin; bp != NIL( s_cpmBranch );
	 bp = bp->nextfi ){
	if( bp->fromNode->level == bp->toNode->level - 1 ){
	    sanitize = FALSE;
	    break;
	}
    }
    if( sanitize ) graph->status |= CPM_SANITIZE;

    branch->fromNode->numfanout--;
    branch->fromNode = NIL( s_cpmNode );
    branch->toNode->numfanin--;
    branch->toNode = NIL( s_cpmNode );
    branch->nextfo = NIL( s_cpmBranch );
    branch->nextfi = NIL( s_cpmBranch );
    graph->numBranches--;
    branch->list = NIL( s_cpmBranch );

    CPM_END();
}

/*
   Install a node in the graph
*/
cpmInstallNode ( graph , node )
struct cpmGraph *graph;
struct cpmNode *node;
{
    struct cpmNode *ptr;

    CPM_START( "cpmInstallNode" );

    for( ptr = graph->nodeList; ptr != NIL( s_cpmNode ); ptr = ptr->list ){
	if( ptr == node )return;
    }
    graph->status |= CPM_SOLVE|CPM_SANITIZE;
    graph->numNodes++;
    node->list = graph->nodeList;
    graph->nodeList = node;

    CPM_END();

}
/*
 * Initialize a branch
 */
cpmInitBranch( branch )
struct cpmBranch *branch;
{

    CPM_START( "cpmInitBranch" );

    branch->upper = 0;
    branch->lower = 0;
    branch->proc = 0;
    branch->force = 0;
    branch->list = NULL;
    branch->fromNode = NULL;
    branch->toNode = NULL;
    branch->nextfo = NULL;
    branch->nextfi = NULL;

    CPM_END();
}

/*
 * Initialize a graph
 */
cpmInitGraph( graph )
struct cpmGraph *graph;
{

    CPM_START( "cpmInitGraph" );
    graph->nodeList = NIL( struct cpmNode );
    graph->branchList = NIL( struct cpmBranch );

    graph->numNodes = 0;
    graph->numBranches = 0;
    graph->maxLevels = 0;
    graph->numUpperBounds = 0;
    graph->numLowerBounds = 0;
    graph->status = CPM_SANITIZE|CPM_SOLVE;
    graph->graphfile = NIL( FILE );
    graph->graphName = NIL( char );

    cpmInitNode( &graph->sourceNode, "source" );
    cpmInitNode( &graph->sinkNode, "sink" );
    cpmInstallNode( graph, &graph->sourceNode );
    graph->sourceNode.lower = 0;
    cpmInstallNode( graph, &graph->sinkNode );

    cpmResetGraph( graph );

    CPM_END();
}

/*
 * cpmResetGraph
 *
 * reset a graph so CPM can run again
 */
cpmResetGraph( graph )
struct cpmGraph *graph;
{
    struct cpmNode *node;
    int origin;

    CPM_START( "cpmResetGraph" );

    if(( graph->status&CPM_SANITIZE) == CPM_SANITIZE )graph->maxLevels = 0;

    /* save the origin of the graph					*/
    origin = graph->sourceNode.lower;

    /* reset every node in the graph 					*/
    for( node = graph->nodeList; node != NULL; node = node->list ){
	cpmResetNode( graph, node );
    }
    /* reset the origin of the graph					*/
    graph->sourceNode.lower = graph->sourceNode.upper = origin;

    CPM_END();
}
cpmSetGraphOrigin( graph, origin, relative )
struct cpmGraph *graph;
int origin;
int relative;
{
    CPM_START( "cpmSetGraphOrigin" );
    if( relative ){
	graph->sourceNode.lower += origin;
	graph->status |= CPM_SOLVE;
    } else if( graph->sourceNode.lower != origin ){
	graph->sourceNode.lower = origin;
	graph->status |= CPM_SOLVE;
    }

    CPM_END();
}
    
/*
 * reset a node
 */
cpmResetNode( graph, node )
struct cpmGraph *graph;
struct cpmNode *node;
{
    CPM_START( "cpmResetNode" );
    node->loc = 0;
    node->lower = -CPM_INFINITY;
    node->upper = CPM_INFINITY;
    node->sched = NULL;
    node->path = NULL;
    node->overcon = FALSE;
    node->count = 0;
    if(( graph->status&CPM_SANITIZE) == CPM_SANITIZE ){
	node->level = UNDEF;
    }

    CPM_END();
}
/*
 * malloc a new node
 */
struct cpmNode *
cpmNewNode( )
{
    struct cpmNode *node;

    CPM_START( "cpmNewNode" );
    node = ALLOC( struct cpmNode, 1 );
    cpmInitNode( node, NIL( char ));
    CPM_END();
    return ( node );
}
/*
 * initialize an existing node
 */
cpmInitNode( node, name )
struct cpmNode *node;
char *name;
{
    CPM_START( "cpmInitNode" );
    if( name != NIL( char ))node->name = util_strsav( name );
    else node->name = NIL( char );
    node->lower = -CPM_INFINITY;
    node->loc = 0;
    node->upper = CPM_INFINITY;
    node->level = UNDEF;
    node->count = 0;
    node->lowerInFlg = FALSE;
    node->lowerOutFlg = FALSE;
    node->upperInFlg = FALSE;
    node->upperOutFlg = FALSE;
    node->overcon = FALSE;
    node->list = NULL;
    node->sched = NULL;
    node->path = NULL;
    node->numfanout = 0;
    node->numfanin  = 0;
    node->fanout = NULL;
    node->fanin = NULL;
    CPM_END();
}
/*
 * cpmResetNodeName
 *
 * reset a node name
 */
cpmResetNodeName( node, name )
struct cpmNode *node;
char *name;
{
    CPM_START( "cpmResetNodeName" );
    if( node->name != NIL( char ))FREE( node->name );
    node->name = util_strsav( name );
    CPM_END();
}
