/*LINTLIBRARY*/
#include "cpmInternal.h"
/*
 * solve the graph
 *
 */

cpmSolve(graph )
struct cpmGraph *graph;
{
    int verbose = FALSE;
    struct cpmBranch *br;
    struct cpmNode *node, *nextNode();
    extern struct cpmNode *LnextNode();
    extern double LschedCount();
    int newlb, newub, newloc;
    double maxSched;
        
    if(((graph->status&CPM_VERBOSE) == CPM_VERBOSE)||
       ((graph->status&CPM_DEBUG) == CPM_DEBUG)){
       verbose = TRUE;
    }
    maxSched = (double)graph->numNodes*(double)graph->numBranches;
    if(verbose)fprintf(stderr,"cpmSolve: ");

/*
 * determine absolute lower bounds
 */
    LschedInit( graph->maxLevels );
    for(br=graph->sourceNode.fanout;br != NULL; br=br->nextfo)
	Lschedule(br->toNode);

    while((node=LnextNode( FORWARD )) != NULL && LschedCount()<maxSched) {
	newlb = -CPM_INFINITY;
	for(br=node->fanin;br != NULL; br = br->nextfi){
	    newlb = MAX ( newlb, ( (br->fromNode)->lower+br->lower));
	}
	newub = -CPM_INFINITY;
	for(br=node->fanout;br != NULL; br=br->nextfo){
	    newub = MAX ( newub, ( (br->toNode)->lower-br->upper ));
	}
	newloc = MAX ( node->lower, MAX ( newlb, newub ) );
	if ( newloc != node->lower ) {
	    node->lower = newloc;
	    for(br=node->fanin;br != NULL; br=br->nextfi){
		if(br->upper < CPM_INFINITY) Lschedule(br->fromNode);
	    }
	    for(br=node->fanout;br != NULL;br=br->nextfo){
		if(br->lower < CPM_INFINITY) Lschedule(br->toNode);
	    }
	}
    }
    if(verbose)fprintf(stderr,"%g lower bound schedules ",LschedCount());
    if(LschedCount() >= maxSched) {
	if( verbose )fprintf(stderr,"The graph could not be solved\n");
	return(FALSE);
    }
        
/*
 * determine absolute upper bounds
 */
    graph->sinkNode.upper = graph->sinkNode.lower;
    LschedInit( graph->maxLevels );
    for(br=graph->sinkNode.fanin;br != NULL; br=br->nextfi)
	Lschedule(br->fromNode);
        
    while((node=LnextNode( REVERSE )) != NULL && LschedCount()<maxSched) {
	newlb = CPM_INFINITY;
	for(br=node->fanin;br != NULL; br = br->nextfi)
	    newlb = MIN ( newlb, ( (br->fromNode)->upper+br->upper));
	newub = CPM_INFINITY;
	for(br=node->fanout;br != NULL; br=br->nextfo)
	    newub = MIN ( newub, ( (br->toNode)->upper-br->lower ));
	newloc = MIN ( node->upper, MIN ( newlb, newub ) );
	if ( newloc != node->upper ) {
	    node->upper = newloc;
	    for(br=node->fanin;br != NULL; br=br->nextfi)
		if(br->lower < CPM_INFINITY) Lschedule(br->fromNode);
	    for(br=node->fanout;br != NULL;br=br->nextfo)
		if(br->upper < CPM_INFINITY) Lschedule(br->toNode);
	}
    }
    if(verbose)fprintf(stderr,"%g upper bound schedules\n",LschedCount());
    graph->status &= ~CPM_SOLVE;
    return(OK);
}
