/*LINTLIBRARY*/
#include "cpmInternal.h" 
int cpmDepth;
char *cpmRoutine;
/* 
 * cpmSolveGraph
 * Copyright (c) 1986, Charles J Kring. All rights reserved.
 * 
 * The heart of the constraint solution.  The graph is levelized and
 * checked for forward loops in sanitize() and solved in solve().  If the
 * solve return code != OK and the graph may be overconstrained, overcon
 * is called to check for overconstraints.
 * 
 */
cpmSolveGraph( graph, opflag )
struct cpmGraph *graph;		/* graph to be solved */
unsigned opflag;		/* which operations to perform on graph */
{ 
    int status;
    int verbose = FALSE;

    if((opflag&CPM_DEBUG) == CPM_DEBUG){
	graph->status |= CPM_DEBUG;
    }
    if((opflag&CPM_VERBOSE) == CPM_VERBOSE){
	graph->status |= CPM_VERBOSE;
	verbose = TRUE;
	if( graph->graphName == NULL ){
	    fprintf(stderr,"Entering CPM\n");
	} else {
	    fprintf(stderr,"Entering CPM - %s graph\n", graph->graphName);
	}
    }

    if( graph->graphfile != NIL( FILE )) (void) fflush( graph->graphfile );
    if ((opflag&CPM_SOLVE) == CPM_SOLVE ){
	if ((graph->status&CPM_SOLVE) == CPM_SOLVE ) cpmResetGraph( graph );
	if ((graph->status&CPM_SANITIZE) == CPM_SANITIZE ){
	    status = cpmSanitize( graph );
	    if( graph->graphfile != NIL( FILE )) (void) fflush( graph->graphfile );
	    if ( status != CPM_OK ){
		return( CPM_SANITIZE_ERR );
	    }
	}
	if (((opflag&CPM_OVERCON) == CPM_OVERCON)&&
	    ( graph->numUpperBounds > 0 )){
	    status = cpmOvercon( graph );
	    if ( status != CPM_OK ){
		return ( CPM_OVERCON_ERR );
	    }
	}
	if ((graph->status&CPM_SOLVE) == CPM_SOLVE ){
	    status = cpmSolve( graph );
	    if (status != CPM_OK ){
		if ((opflag&CPM_OVERCON) != CPM_OVERCON){

		    if ( cpmSanitize( graph ) != CPM_OK ){
			return( CPM_SANITIZE_ERR );
		    }
		    cpmResetGraph( graph );
		    status = cpmOvercon( graph );
		    if ( status != CPM_OK ) {
			return ( CPM_OVERCON_ERR );
		    }
		}
		return ( CPM_SOLVE_ERR );
	    }
	}
    }
    if ((opflag&CPM_NONCRIT) == CPM_NONCRIT ){
	status = cpmNoncrit( graph );
	if ( status != CPM_OK ) {
	    return ( CPM_NONCRIT_ERR );
	}
    }
    if( verbose)fprintf(stderr,"Leaving CPM\n");
    return ( status );
}

/*
 * cpmErrHandler
 *
 */
/*ARGSUSED*/
void
cpmErrHandler( pkgName, code, message )
char *pkgName;
int code;
char *message;
{
    cpmDepth = 0;
    errPopHandler();
    if( strcmp( pkgName, CPM_PKG_NAME ) == 0 ) pkgName = "internal";
    errPass("(function %s) unexpected %s error: %s:",
	    cpmRoutine, pkgName, message );
}
