#include "mkInternal.h"
#include "tr.h"
/*
 * mk_transform_tframe
 *
 * tranforms a tframe according to the instances transform.  The 
 * origin of the tframes coordinate system is adjusted so that the
 * lower left hand corner of the instance is a 0,0.
 */
void
mk_transform_tframe( element, frame )
s_element *element;
octBox *frame;
{
    octObject oInst;
    octCoord tmp;

    oInst.objectId = element->objectId;
    octGetById( &oInst );
    tr_oct_itransform( &oInst.contents.instance.transform,
		      &frame->lowerLeft.x,
		      &frame->lowerLeft.y );
    tr_oct_itransform( &oInst.contents.instance.transform,
		      &frame->upperRight.x,
		      &frame->upperRight.y );

    if( frame->lowerLeft.x > frame->upperRight.x ){
	tmp = frame->lowerLeft.x;
	frame->lowerLeft.x = frame->upperRight.x;
	frame->upperRight.x = tmp;
    }
    if( frame->lowerLeft.y > frame->upperRight.y ){
	tmp = frame->lowerLeft.y;
	frame->lowerLeft.y = frame->upperRight.y;
	frame->upperRight.y = tmp;
    }
    /* translating frame so that it is relative to the lowerLeft 	*/
    /* corner of the oInst bbox						*/
    frame->lowerLeft.x += element->contents.instance->offset.x;
    frame->lowerLeft.y += element->contents.instance->offset.y;
    frame->upperRight.x += element->contents.instance->offset.x;
    frame->upperRight.y += element->contents.instance->offset.y;
}
void
mk_relative_tCenter( term, xPercent, yPercent )
s_aterm *term;
double *xPercent;
double *yPercent;
{
    octObject oInst;
    octBox tbb;
    octBox ibb;
    int tmp;
    s_tframe *tframe;

    tbb.lowerLeft.x = tbb.lowerLeft.y = INFINITY;
    tbb.upperRight.x = tbb.upperRight.y = 0;
    for( tframe = term->fterm->frame; tframe != NIL( s_tframe );
	 tframe = tframe->next ){
	tbb.lowerLeft.x = MIN( tbb.lowerLeft.x, tframe->box.lowerLeft.x);
	tbb.lowerLeft.y = MIN( tbb.lowerLeft.y, tframe->box.lowerLeft.y);
	tbb.upperRight.x = MAX( tbb.upperRight.x, tframe->box.upperRight.x);
	tbb.upperRight.y = MAX( tbb.upperRight.y,tframe->box.upperRight.y);
    }
    mk_transform_tframe( term->instance, &tbb );
    oInst.objectId = term->instance->objectId;
    octGetById( &oInst );
    octBB( &oInst, &ibb );
    if( ibb.lowerLeft.x > ibb.upperRight.x ){
	tmp = ibb.lowerLeft.x;
	ibb.lowerLeft.x = ibb.upperRight.x;
	ibb.upperRight.x = tmp;
    }
    if( ibb.lowerLeft.y > ibb.upperRight.y ){
	tmp = ibb.lowerLeft.y;
	ibb.lowerLeft.y = ibb.upperRight.y;
	ibb.upperRight.y = tmp;
    }
    /* translating frame so that it is relative to the lowerLeft 	*/
    /* corner of the oInst bbox						*/
    ibb.lowerLeft.x += term->instance->contents.instance->offset.x;
    ibb.lowerLeft.y += term->instance->contents.instance->offset.y;
    ibb.upperRight.x += term->instance->contents.instance->offset.x;
    ibb.upperRight.y += term->instance->contents.instance->offset.y;

    *xPercent = ( tbb.upperRight.x + tbb.lowerLeft.x ) /
		( ibb.upperRight.x + ibb.lowerLeft.x );

    *yPercent = ( tbb.upperRight.y + tbb.lowerLeft.y ) /
		( ibb.upperRight.y + ibb.lowerLeft.y );
}
