/* Each pair of sequential instances in the elements elemlist are	*/
/* Copyright (c) 1987, Charles J Kring. All rights reserved.		*/
#include "mkInternal.h"
/*
 * mkWireArea
 *
 * Each pair of sequential instances in the elements elemlist are
 * connectable. Every pair of terminals in the connectable instances
 * which belong to the same net are connected with a segment if 
 * a segment is not already there.
 */
void
mkWireArea( arrayId, area, ruleName )
mkId arrayId;
octBox *area;			/* area to wire 			*/
char *ruleName;			/* name of the wiring rule		*/
{

    s_array *array;
    s_rulegroup *rule;
    int lr = TRUE; 		/* TRUE is scanning from left to right	*/
    int tb = TRUE;		/* TRUE is scanning from top to bottom	*/
    static int colBufSize = 0;	/* current size of the colBuffer	*/
    static s_elemlist **colBuffer;	
				/* array of last elements found in col	*/
    s_elemlist **colbuf;	/* points to colBuffer[col]		*/
    s_elemlist *rowbuf;		/* last element found in row		*/

    octBox *mappedArea;		/* wiring area mapped to root array 	*/
    int row, col;		/* indices in mappedArea		*/
    int numRows;		/* number of rows in mappedArea		*/
    int numCols;		/* number of cols in mappedArea		*/
    s_elemlist ***rowptr; 	/* points to array->elemlist[row]	*/
    s_elemlist **colptr;	/* points to rowptr[col]		*/
    s_rulegroup *mk_get_rule();
    void mk_wire();
    void mk_wire_term();

    MK_START( "mkWireArea" );

    if( area->lowerLeft.x > area->upperRight.x )lr = FALSE;
    if( area->lowerLeft.y > area->upperRight.y )tb = FALSE;

    array = mk_get_array( arrayId );
    if( array->type == SUBARRAY ) {
	mappedArea = mkTranslate( arrayId, area );
	array = array->contents.subarray.array;
    } else {
	mappedArea = area;
    }
    mk_sanitize_area( array, mappedArea );

    if( ruleName != NIL( char )){
	if(( rule = mk_get_rule( ruleName )) == NIL( s_rulegroup )){
	    errRaise( MK_PKG_NAME, 0,
	              "Could not find rule named %s", ruleName );
	}
	rule->enabled = TRUE;
    }

    numRows = mappedArea->upperRight.y - mappedArea->lowerLeft.y + 1;
    numCols = mappedArea->upperRight.x - mappedArea->lowerLeft.x + 1;

    /* allocating the colBuffer						*/
    if( numCols > colBufSize ){
	if( colBufSize == 0 ){
	    colBuffer = ALLOC( s_elemlist *, numCols );
	} else {
	    colBuffer = REALLOC( s_elemlist *, colBuffer, numCols );
	}
	colBufSize = numCols;
    }

    /* initializing the colBuffer					*/
    colbuf = colBuffer;
    for( col = numCols; col > 0; col-- ){
	*colbuf = NIL( s_elemlist );
	colbuf++;
    }

    /* stepping through the mappedArea in the proper direction looking	*/
    /* for pairs of non NIL elemlists - they are passed to mk_wire to 	*/
    /* see if anything in the pair should be wired			*/

    if( tb )rowptr = &array->contents.array.elemlist[mappedArea->upperRight.y];
    else    rowptr = &array->contents.array.elemlist[mappedArea->lowerLeft.y];
    for( row = numRows; row > 0; row-- ){

	rowbuf = NIL( s_elemlist );

	if( lr )colptr = &(*rowptr)[mappedArea->lowerLeft.x];
	else    colptr = &(*rowptr)[mappedArea->upperRight.x];
	colbuf = colBuffer;

	for( col = numCols; col > 0; col-- ){

	    if( *colptr != NIL( s_elemlist )){

		/* this elemlist and the one to the left or right	*/
		if( rowbuf != NIL( s_elemlist )){
		    mk_wire( array, rowbuf, *colptr, MK_HOR_SEG);
		}
		/* this elemlist and the one to the top or bottom	*/
		if( *colbuf != NIL( s_elemlist )){
		    mk_wire( array, *colbuf, *colptr, MK_VER_SEG );
		}
		rowbuf = *colptr;
		*colbuf = *colptr;
	    }

	    if ( lr ) colptr++;
	    else colptr--;
	    colbuf++;
	}
	if( tb )rowptr--;
	else    rowptr++;
	/* check to see if rowbuf should be wired to itself		*/
	if(( rowbuf != NIL( s_elemlist ))&&
	   ( rowbuf->next != NIL( s_elemlist ))){
	    mk_wire( array, rowbuf, NIL( s_elemlist ), MK_DC_SEG );
	}
    }
    /* check to see if the colbufs should be wired to themself		*/
    colbuf = colBuffer;
    for( col = numCols; col > 0; col-- ){
	if(( (*colbuf) != NIL( s_elemlist ))&&
	   ( (*colbuf)->next != NIL( s_elemlist ))){
	    mk_wire( array, (*colbuf), NIL( s_elemlist ), MK_DC_SEG );
	}
	colbuf++;
    }
    if( ruleName != NIL( char )) rule->enabled = FALSE;
    MK_END();
}

/* 
 * mk_wire
 *
 * finds every term in elemlist one and calls mk_wire_term on it
 * and elemlist2.  
 */
void
mk_wire( array, elemlist1, elemlist2, orientation )
s_array *array;
s_elemlist *elemlist1;
s_elemlist *elemlist2;
int orientation;
{
    s_elemlist *elemlist;
    s_aterm *term;

    for( elemlist = elemlist1; elemlist != NIL( s_elemlist );
	 elemlist = elemlist->next ){
	if( elemlist->element->type == OCT_INSTANCE ){
	    for( term = elemlist->element->contents.instance->term; 
		 term != NIL( s_aterm ); term = term->next ){
		if( elemlist->next != NIL( s_elemlist )){
		    mk_wire_term( array, term, elemlist->next, orientation );
		}
		mk_wire_term( array, term, elemlist2, orientation );
	    }
	} else if ( elemlist->element->type == OCT_TERM ){
	    if( elemlist->next != NIL( s_elemlist )){
		mk_wire_term( array, elemlist->element->contents.term, 
			      elemlist->next, orientation );
	    }
	    mk_wire_term( array, elemlist->element->contents.term,
			  elemlist2, orientation );
	}
    }
}

/*
 * mk_wire_term
 *
 * checks term with every term2 in elemlist2. If a term / term2 pair 
 * should be connected, calls mk_new_segment to create the segment.
 * A pair should be connected if a term->rule exists which is enabled
 * and points to term2.
 *
 */
void
mk_wire_term( array, term, elemlist2, orientation )
s_array *array;
s_aterm *term;
s_elemlist *elemlist2;
int orientation;	
{
    s_rule *rule;		/* rules on term1->fterm		*/
    s_aterm *term2;		/* aterms in elemlist			*/
    s_elemlist *elemlist;	/* points to elemlist			*/
    int realOrientation;

    for( rule = term->fterm->rule; rule != NIL( s_rule ); 
	 rule = rule->next){

	if( rule->group->enabled == FALSE ) continue;
	for( elemlist = elemlist2; elemlist != NIL( s_elemlist );
	     elemlist = elemlist->next ){
	    if( elemlist->element->type == OCT_INSTANCE ){
		if(( elemlist->element == term->instance )&&
		   ( array->contents.array.wireIntraInst == FALSE )) continue;
		for(term2 = elemlist->element->contents.instance->term;
		    term2 != NIL( s_aterm ); term2 = term2->next ){
		    if( rule->fterm == term2->fterm ){
			if( orientation == MK_DC_SEG )
			    realOrientation = rule->orientation;
			else if(( orientation == MK_HOR_SEG && 
				  rule->orientation == MK_VER_SEG )||
			        ( orientation == MK_VER_SEG && 
				  rule->orientation == MK_HOR_SEG ))continue;
			else
			    realOrientation = orientation;
			mk_new_segment( array, term, term2, realOrientation, 
				        rule->layer, rule->widthSpec, FALSE );
		    }
		}
	    } else if( elemlist->element->type == OCT_TERM ){
		term2 = elemlist->element->contents.term;
		if(( term2->instance == term->instance )&&
		   ( array->contents.array.wireIntraInst == FALSE )) continue;
		if( rule->fterm == term2->fterm ){
		    if( orientation == MK_DC_SEG )
			realOrientation = rule->orientation;
		    else if(( orientation == MK_HOR_SEG && 
			      rule->orientation == MK_VER_SEG )||
			    ( orientation == MK_VER_SEG && 
			      rule->orientation == MK_HOR_SEG ))continue;
		    else
			realOrientation = orientation;
		    mk_new_segment( array, term, term2, realOrientation, 
				    rule->layer, rule->widthSpec, FALSE  );
		}
	    }
	}
    }
}
