/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPInc.h

Globals for interface of channel preparation for channel router from OCT
database.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CPInc.h,v $
 * Revision 1.2  90/02/07  16:14:04  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/26  14:20:20  octtools
 * Initial revision
 * 
 * Revision 1.7  88/04/25  15:16:24  mitch
 * Fixed bug in corner remapping of z-shaped channels.
 * 
 * Revision 1.4  88/04/08  16:31:20  mitch
 * Put in pin direction as an option to mosaicoDR.
 * 
 * Revision 1.2  88/04/05  15:56:03  mitch
 * Added pseudo pin direction field in CPPinDef data structure.
 * 
 * Revision 1.1  88/03/31  11:23:53  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
typedef	struct	CPPinDef	CPPinDef;
typedef	struct	CPNetDef	CPNetDef;
typedef	struct	CPInstDef	CPInstDef;
typedef	struct	CPLocDef	CPLocDef;
typedef	struct	CPCornDef	CPCornDef;
typedef	struct	CPWhereDef	CPWhereDef;
typedef	struct	CPParmDef	CPParmDef;

#define CP_MAX_NUM_CORN		20
#define CP_OCT_PREROUTE_VIEW	"preroute"
#define CP_OCT_POSTROUTE_VIEW	"symbolic"
#define CP_EXT_INST_NAME	"CPExtInst"
#define CP_NULL_CHANNEL		MAXSHORT
#define CP_NULL_NET		-1
#define CP_PTERM_INSTNUM	-1
#define CP_CW			 1
#define CP_CCW			-1
#define CP_V			 0
#define CP_H			 1

struct CPLocDef {
    int		x, y;
};

struct CPPinDef {
    String	name;			/* Name of pin */
    String	newName;		/* New name of pin (name if same) */
    octId	oldTermId;		/* Old terminal octId */
    octId	newTermId;		/* New terminal octId */
    int		width;			/* Width of pin */
    int		widthForSpider;		/* Width of pin for spider */
    double	peakCurrent;		/* Peak power current */
    Boolean	wasAPTermFlg;		/* Was a pseudo terminal ? */
    Boolean	isAFTermFlg;		/* Is a real formal terminal ? */
    char	dir;			/* Pin direction: */
					/*   R=right, L=left, T=top, */
					/*   B=bottom, C=center/don't care */
    char	exitDir;		/* Other pins on the same net dir */
					/*   H=high, L=lo, X=other */
    short	instNum;		/* Index into CPInst[] */
					/*   CP_PTERM_INSTNUM if pTerm */
    short	chanNum;		/* Belongs to this channel */
					/*   adjacent chan if pTerm */
    short	subNetNum;		/* Pin belongs to this net */
					/*   CP_NULL_NET if no net attachment */
    short	layer;			/* Pin layer number */
    CPLocDef	loc;			/* Center location of pin */
    int		locHi;			/* Range of pin (center location) */
    int		locLo;

    CPPinDef	*next;
};

struct CPNetDef {
    octId	subNetId;		/* Subnet octId of net */
    octId	netId;			/* Net octId of net */
    short	netNum;			/* Net num of subNet */
    Boolean	dupFlg;			/* Multiple subnets under net ? */
    String	name;			/* Name of subnet */
};

struct CPInstDef {
    String	name;			/* Name of instance */
    octId	oldInstId;		/* Old instance octId */
    octId	newInstId;		/* New instance octId */
    char	type;			/* t=top, b=bottom */
    CPLocDef	loc;			/* Location of instance */
    CPLocDef	newLoc;
    short	numBdy;			/* Number of boundary elements */
    short	bdyOrient;		/* Original boundary orientation: */
					/*   Clockwise or Counterclockwise ? */
    short	numCorn;		/* Number of corner elements */
    CPLocDef	*bdy;			/* Boundary polygon - always goes */
					/*   counterclockwise around inst */
    CPCornDef	*corn;			/* Corners description array */
    short	startBdyNum;		/* Start/End bdy indicies of instance */
    short	endBdyNum;		/*   channel boundary */
    octObject	octInstMaster;		/* Instance master interface facet */
    Boolean	skipCornFlg;		/* Skip next corn after current corner
					 *   in merged new master? */
};

struct CPWhereDef {
    short	instOnTop;		/* Upper or right inst adj to channel */
    short	instOnBot;		/* Lower or left inst adj to channel */
    short	chanOnRight;		/* Next channel going clockwise from */
					/*   top to bottom inst */
    short	chanOnLeft;		/* Prev channel ... */
    short	chanDir;		/* Direction of channel */
};

struct CPCornDef {
    short	bdyNum;			/* Bdy num of corner */
    short	chanNum;		/* Channel number of next */
};

struct CPParmDef {
    String	cellName;
    String	inputView;
    String	outputView;
    String	routerName;
    String	compactChanFile;
    String	routeDirName;
    Boolean	partialMode;
    Boolean	interactiveMode;
    Boolean	pinDirMode;
    Boolean	createPurgedFacet;
    Boolean	skipChExtern;
    Boolean	compactChanMode;
    Boolean	viaMinMode;
    short	numCompactChan;
    short	*compactChanNum;
};

CPParmDef	CPParm;
int		CPNumChan;
int		CPCurChan;
int		CPNumNet;
int		CPNumSubNet;
octId		*CPChanOctId;
octObject	CPOctMaster;
octObject	CPOctSubMaster;
octObject	CPOctSubMasterInst;
octObject	CPOctTech;
String		CPSubMasterName;
CPPinDef	*CPPin;
CPNetDef	*CPNet;
CPInstDef	CPInst[2];
CPWhereDef	CPWhere;
short		CPExtInstNum;
char		CPTempBuf[1024];

octObject	CPOctChRouted;
octObject	CPOctChInterface;

#define CP_TERM_TO_CHANNUM(chanNum99, term99) \
{ \
octObject	subNet99; \
octObject	chan99; \
\
*(chanNum99)  = CP_NULL_CHANNEL; \
subNet99.type = OCT_NET; \
chan99  .type = OCT_BAG; \
subNet99.contents.net.name = "CDSubnet"; \
chan99  .contents.bag.name = "CDChanDef"; \
if (MIOctGetByNameContainers(term99, &subNet99)==OCT_OK) { \
    OH_ASSERT(MIOctGetByNameContainers(&subNet99, &chan99)); \
    CP_CHAN_TO_CHANNUM(chanNum99, &chan99); \
} \
}

#define CP_CHAN_TO_CHANNUM(chanNum99, chan99) \
{ \
octObject	prop99; \
\
*(chanNum99) = CP_NULL_CHANNEL; \
if (ohGetByPropName(chan99, &prop99, "CDChanOrder")==OCT_OK) \
    *(chanNum99) = (short)prop99.contents.prop.value.integer; \
}
