/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPMain.c

Main routine for channel preparation for channel routing & routing.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CPMain.c,v $
 * Revision 1.4  89/02/12  13:29:44  octtools
 * turned off interactive mode (till I put in the routeChan replacement)
 * 
 * Revision 1.3  89/02/07  13:20:55  octtools
 * got rid of routeChan
 * ./
 * 
 * Revision 1.2  89/01/27  23:05:10  octtools
 * fixed lint error
 * 
 * Revision 1.1  89/01/26  14:20:11  octtools
 * Initial revision
 * 
 * Revision 1.10  88/06/15  14:04:09  mitch
 * Consider also OCT_BAGs when copying the contents of aTerms
 * 
 * Revision 1.9  88/04/27  16:54:50  mitch
 * Fatal error in mosaico when clean mode is all ok.
 * 
 * Revision 1.8  88/04/27  13:03:21  mitch
 * Updated message about non POLY, MET1 or MET2 terminal contacts.
 * 
 * Revision 1.7  88/04/25  15:17:05  mitch
 * Fixed bug in corner remapping of z-shaped channels.
 * 
 * Revision 1.6  88/04/24  16:42:03  mitch
 * Put in package prefix.
 * Fixed subnet aTerm pointer.
 * 
 * Revision 1.5  88/04/11  12:14:58  mitch
 * Changed CPCompLayerNum to return CP_UNKNOWN_LAYER_NUM for everything
 * except POLY, MET1 and MET2.
 * 
 * Revision 1.4  88/04/08  16:31:58  mitch
 * Put in pin direction as an option to cprep.
 * Deleted no subnet found for aterm warning message.
 * Fixed strange bug in pin direction initialization.
 * 
 * Revision 1.3  88/04/08  12:17:58  mitch
 * Put in octBegin and octEnd into main procedure.
 * 
 * Revision 1.2  88/04/05  15:57:23  mitch
 * Got rid of nested admLogInTool/admLogOutTool since adm cannot handle it.
 * Added pin direction stuff into router interface file.
 * 
 * Revision 1.1  88/03/31  14:24:01  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#ifndef lint
static char RCSId[] = "@(#) $Header: CPMain.c,v 1.4 89/02/12 13:29:44 octtools Exp $";
#endif  lint
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "vulcan.h"
#include "options.h"
#include "miutil.h"
#include "CPInc.h"
#include "vov.h"


extern	void	octMakeSureItsFree();
int closeFacetsFlag = 0;	/* If facets are not closed, it should run faster */

/* Main procedure for channel preparation
----------------------------------------------------------------------------*/
main(argc, argv)
    int		argc;
    String	argv[];
{
    short	k;
    octObject	octChan;
    String	chanNameO = ".ch";
    String	chanNameR = ".ti";
    int		rcClean;
    String	primaryCellName;
    String	strrchr();
    int		c;
    octObject	octExtChanProp;
    octObject	chanBag;
    int		lastExtChan;
    int		spiderRc;
    octObject	Routed;
    char	*routerBin;
    char	buffer[2048];
    Boolean	compactChanFlag;
    int		printMemUsage = 1; /*  -1 means print it out */
				/* >=0 means don't print */
    MIPrefix = "mosaicoDR";
    errProgramName("mosaicoDR");
    CPEvalParm(argc, argv);

#ifdef hpux
    setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
    setvbuf(stderr, NULL, _IOLBF, BUFSIZ);
#else
    setlinebuf(stdout);
    setlinebuf(stderr);
#endif

    {
	char	month[10];
	int		date, year;

	printf("%s: Routing driver and channel preparation (Rev 2.5", MIPrefix);
#ifdef CUR_DATE
	printf(" compiled %s)\n", CUR_DATE);
#else				/* CUR_DATE */
	printf(")\n");
#endif				/* CUR_DATE */
    }

    MIModStatStart("MosaicoDR", "debugCp", TRUE);
    MSG(1, ("Cell name <%s>, inView <%s> and outView <%s>\n",
	    CPParm.cellName, CPParm.inputView, CPParm.outputView));
    MI_STACK_ALLOC(1000);
    if (strrchr(CPParm.cellName,'/'))
      primaryCellName = strrchr(CPParm.cellName,'/')+1;
    else
      primaryCellName = CPParm.cellName;

    octBegin();
    CPOctInput(CPParm.cellName, CPParm.inputView, CPParm.outputView);
    /* Count number of external channels */
    for (c=0; c<CPNumChan; c++) {
	octChan.objectId = CPChanOctId[c];
	if (octChan.objectId==oct_null_id) continue;
	OH_ASSERT(octGetById(&octChan));
	if (ohGetByPropName(&octChan, &octExtChanProp, "CDOuterChannel")==OCT_OK)
	  continue;
	lastExtChan = c;
    }

    if (CPParm.compactChanMode && !CPParm.compactChanFile) {
	CPParm.numCompactChan = CPNumChan;
	CPParm.compactChanNum = ALLOC_ARRAY(CPParm.numCompactChan, short);
	for (c=0; c<CPParm.numCompactChan; c++)
	  CPParm.compactChanNum[c] = c;
    }

    MSG(-1, ("There are %d channels in cell %s %s\n",
	     CPParm.skipChExtern ? lastExtChan+1 : CPNumChan,
	     CPParm.cellName, CPParm.inputView));
    rcClean = CPOctCleanSubNets();
    if (CPParm.createPurgedFacet)
      goto CPMainEnd;

    for (CPCurChan=0; CPCurChan<CPNumChan; CPCurChan++) {
	if (CPChanOctId[CPCurChan]==oct_null_id) continue;

	OH_ASSERT(ohGetById(&octChan, CPChanOctId[CPCurChan]));
	if (CPParm.skipChExtern && CPCurChan>lastExtChan) break;
	/* Open subMaster */
	(void) sprintf(MIBuf,"%s/%s%s%d",
		       CPParm.routeDirName, primaryCellName, chanNameO, CPCurChan);
	CPSubMasterName = STR_COPY(MIBuf);
	printf("%s: Working on %20s (%3d more to go)\n",
	       MIPrefix,
	       CPSubMasterName,
	       (CPParm.skipChExtern ? lastExtChan+1 : CPNumChan)-CPCurChan-1);
	OH_ASSERT(ohOpenFacet(&CPOctSubMaster, CPSubMasterName,
			      CP_OCT_PREROUTE_VIEW, "contents", "w"));
	VOVoutputFacet( &CPOctSubMaster );
	/* Init submaster stuff */
	CPInitSubMaster(&CPOctSubMaster, &octChan);
	CPWhosOnTop();
	CPAdjustInstLoc();
	(void) sprintf(CPTempBuf,"%s/%s%s%d",
		       CPParm.routeDirName, primaryCellName, chanNameR, CPCurChan);
	CPRouteInter(CPTempBuf);

	(void) sprintf(CPTempBuf,"%s/%s%s%d",
		       CPParm.routeDirName, primaryCellName, chanNameO, CPCurChan);
	CPOctChRouted.contents.facet.cell = STR_COPY(CPTempBuf);
	/* Close subMaster */
	for (k=0; k<2; k++) {
	    if (k==CPExtInstNum) continue;
	    OH_ASSERT(octCloseFacet(&CPInst[k].octInstMaster));
	}
	OH_ASSERT(octFlushFacet(&CPOctSubMaster));
	if ( closeFacetsFlag ) {
	    octMakeSureItsFree(&CPOctSubMaster);
	}

	/*
	 * replacing routeChan:
	 *
	 *  - copy facet prerouted to symbolic
	 *  - system("spider")
	 *  - unlink temp files
	 */
	if (CPParm.interactiveMode) {
	    printf("Interactive mode turned off for the time being\n");
	    exit(-1);

	    printf("Check preroute facet of chan%d ? ", CPCurChan);
	    (void) scanf("%s", MIBuf);
	    (void) system(CPTempBuf);
	    printf("Check routed facet of chan%d ? ", CPCurChan);
	    (void) scanf("%s", MIBuf);
	    if (*MIBuf=='y') {
		printf("Do you want to quit right now ? ");
		(void) scanf("%s", MIBuf);
		if (*MIBuf=='y') goto CPMainEnd;
	    }
	}
	else {
	    Routed = CPOctSubMaster;
	    Routed.contents.facet.view = "symbolic";
	    (void) octCopyFacet(&Routed, &CPOctSubMaster);

	    if (CPParm.routerName == (char *) 0 || CPParm.routerName[0] == '\0') {
		routerBin = "yacr";
	    } else if (strcmp(CPParm.routerName, "YACR") == 0) {
		routerBin = "yacr";
	    } else {
		routerBin = "mighty";
	    }

	    for (c=0; c<CPParm.numCompactChan; c++)
	      if (CPParm.compactChanNum[c]==CPCurChan) break;
	    compactChanFlag = (c<CPParm.numCompactChan);

	    (void) sprintf(buffer, "spider %s/%s %d %s %s %s %s",
			   CPParm.routeDirName, primaryCellName,
			   CPCurChan, CPParm.routerName, routerBin,
			   compactChanFlag ? "COMPACT" : "",
			   compactChanFlag && CPParm.viaMinMode ? "MIZER" : "");
	    spiderRc = system(buffer)/256;
	    switch(spiderRc) {
	    case 0:
		printf("%s: Channel %c is degenerate (has no routing)\n", 
		       MIPrefix, CPCurChan);
		break;
	    case 1:
		printf("%s: Channel router crapped out in channel %d ... terminating\n",
		       MIPrefix, CPCurChan);
		exit(1);
	    case -10:
		printf("%s: Channel router had oct problems in channel %d ... terminating\n",
		       MIPrefix, CPCurChan);
		exit(1);
	    case 2:
	    case 3:
	    case 4:
		break;
	    default:
		printf("%s: Unknown return code %d from spider in channel %d ... terminating\n",
		       MIPrefix, spiderRc, CPCurChan);
		exit(1);
	    }

	    /* remove summary/error files */
	    (void) sprintf(buffer, "%s/%s.summary",
			   CPParm.routeDirName, primaryCellName);
	    (void) unlink(buffer);
	    (void) sprintf(buffer, "%s/%s.err%d",
			   CPParm.routeDirName, primaryCellName, CPCurChan);
	    (void) unlink(buffer);

	    /* remove router input and output files */
	    (void) sprintf(buffer, "%s.in%d", routerBin, CPCurChan);
	    (void) unlink(buffer);
	    (void) sprintf(buffer, "%s.out%d", routerBin, CPCurChan);
	    (void) unlink(buffer);

	}

	OH_ASSERT(ohOpenFacet(&CPOctChRouted, CPOctChRouted.contents.facet.cell,
			      "symbolic", "contents", "r"));
	OH_ASSERT(ohOpenFacet(&CPOctChInterface, CPOctChRouted.contents.facet.cell,
			      "symbolic", "interface", "w"));
	VOVoutputFacet( &CPOctChRouted );
	VOVoutputFacet( &CPOctChInterface );

	MAXMEM(printMemUsage, "Before vulcan");
	(void) vulcanParameterized(&CPOctChRouted, &CPOctChInterface,
				   VULCAN_DFLT_FLAGS, VULCAN_DFLT_MIN_FEATURE, VULCAN_NO_DEBUG);
	MAXMEM(printMemUsage, "After vulcan");
	{
	    octObject	octTerm, octProp, octInstTerm;
	    octGenerator	genTerm, genProp;

	    OH_ASSERT(octInitGenContents(&CPOctChRouted, OCT_TERM_MASK, &genTerm));
	    while (octGenerate(&genTerm, &octTerm)==OCT_OK) {
		octInstTerm = octTerm;
		OH_ASSERT(octGetByName(&CPOctChInterface, &octInstTerm));
		OH_ASSERT(octInitGenContents(&octTerm, OCT_PROP_MASK, &genProp));
		while (octGenerate(&genProp, &octProp)==OCT_OK) {
		    OH_ASSERT(octCreateOrModify(&octInstTerm, &octProp));
		}
	    }
	}
	CPRedoVulcan(&CPOctChRouted, "save");
	if ( closeFacetsFlag ) {
	    OH_ASSERT(octCloseFacet(&CPOctChRouted));
	    octMakeSureItsFree(&CPOctChRouted);
	}
    
	CPRedoVulcan(&CPOctChInterface,"replace");
	if ( closeFacetsFlag ) {
	    OH_ASSERT(octCloseFacet(&CPOctChInterface));
	    octMakeSureItsFree(&CPOctChInterface);
	}
	CPRebind(&octChan);
	CPCleanChan(&octChan);

	if (CPParm.interactiveMode) {
	    printf("Do you want to quit right now ? ");
	    (void) scanf("%s", MIBuf);
	    if (*MIBuf=='y') goto CPMainEnd;
	}
	if (CPParm.partialMode) {
	    printf("Do you want to partial save right now ? ");
	    (void) scanf("%s", MIBuf);
	    if (*MIBuf=='y') {
		MAXMEM(printMemUsage, "Before flush facet");
		OH_ASSERT(octFlushFacet(&CPOctMaster));
	    }
	}
	MAXMEM(printMemUsage, "After rebinding and cleaning facet");
	MAXMEM(-1, "After routing channel");
    }

 CPMainEnd:
    CPCleanDegenNets(&CPOctMaster);
    if (!CPParm.createPurgedFacet) {
	if (ohGetByBagName(&CPOctMaster,&chanBag,"CDChanDef")==OCT_OK) {
	    ohRecursiveDelete( &chanBag, OCT_POINT_MASK|OCT_BAG_MASK|OCT_PROP_MASK);
	}
    }

    OH_ASSERT(octCloseFacet(&CPOctMaster));
    /* octMakeSureItsFree(&CPOctMaster); */
    octEnd();

    MI_STACK_FREE();

    MIModStatEnd();
    {
	char command[1024];
	sprintf( command, "octgc -d %s:%s", CPOctMaster.contents.facet.cell,
		CPOctMaster.contents.facet.view );
	if ( util_csystem( command ) == 127 ) {
	    printf( "Failed command: %s", command );
	}

    }
    VOVend(0);
}				/* ............................. End of main procedure */


/* Remove degenerate nets from master
----------------------------------------------------------------------------*/
CPCleanDegenNets(octMaster)
    octObject	*octMaster;
{
    octObject	octNet, octTerm;
    octGenerator	genNet, genTerm;

    MI_MOD_START("CPCleanDegenNets");

    OH_ASSERT(octInitGenContents(octMaster, OCT_NET_MASK, &genNet));
    while (octGenerate(&genNet, &octNet)==OCT_OK) {
	if (octGenFirstContent(&octNet, OCT_TERM_MASK,  &octTerm)==OCT_OK) {
	    break;
	}
	OH_ASSERT(octDelete(&octNet)); /* No terminals in net - purging */
    }

    MI_MOD_END();
}				/* ............................. End of procedure CPCleanDegenNets */


/* Remove degenerate subnets
----------------------------------------------------------------------------*/
int
CPOctCleanSubNets()
{
#define		CP_MAX_PTERM_PER_NET 256
#define		CP_MAX_SUBNET_PER_PTERM 256
#define		CP_FIX_FABIO

    octObject	octNet, octNet2;
    octObject	octSubNet, octSubNet2;
    octObject	octPTerm, octATerm, octFTerm;
    octObject	octSubNetChan;
    octObject	octSubStuff, octSubStuff2;
    octObject	octInst;
    octGenerator	genNet, genNet2;
    octGenerator	genSubNet;
    octGenerator	genTerm;
    octGenerator	genPTerm;
    octGenerator	genSubStuff, genSubStuff2;
    short		chanNum, chanNum1, chanNum2;
    short		numSubNet;
    short		p, psn;
    Boolean		netFoundFlg;
    int		rc = 0;

    short		numPTerm;
    int		pTermOctId[CP_MAX_PTERM_PER_NET];

    int		pTermSubNetOctId  [CP_MAX_SUBNET_PER_PTERM];
    int		pTermSubNetChanNum[CP_MAX_SUBNET_PER_PTERM];
    int		pTermSubNetChan1  [CP_MAX_SUBNET_PER_PTERM];
    int		pTermSubNetChan2  [CP_MAX_SUBNET_PER_PTERM];
    short		numPTermSubNetChan1, numPTermSubNetChan2;

    MI_MOD_START("CPOctCleanSubNets");
    /* Get rid of subnet->fterm of fterm */
    /*   attachment (if the fTerm has a */
    /*   aterm with a subnet) */
    OH_ASSERT(octInitGenContents(&CPOctMaster, OCT_TERM_MASK, &genTerm));
    while (octGenerate(&genTerm, &octFTerm)==OCT_OK) {
	OH_ASSERT(octInitGenContainers(&octFTerm, OCT_NET_MASK, &genNet));
	while (octGenerate(&genNet, &octNet)==OCT_OK) {
	    if (strcmp(octNet.contents.net.name, "CDSubnet")) continue;

	    /* Check if unassigned fTerm */
	    if (octGenFirstContent(&octFTerm, OCT_TERM_MASK, &octATerm)==OCT_OK) {
		/* Check if aTerm under fTerm has */
		/*   a subnet */
		OH_ASSERT(octInitGenContainers(&octATerm, OCT_NET_MASK, &genNet));
		for (netFoundFlg=FALSE;
		     !netFoundFlg && octGenerate(&genNet, &octNet2)==OCT_OK; )
		  netFoundFlg = !strcmp(octNet2.contents.net.name, "CDSubnet");
		if (!netFoundFlg)
		  OH_ASSERT(octAttach(&octNet, &octATerm));
		OH_ASSERT(octDetach(&octNet, &octFTerm));
	    }
	}
    }

    octSubNetChan.type = OCT_BAG;
    octSubNetChan.contents.bag.name = "CDChanDef";

    OH_ASSERT(octInitGenContents(&CPOctMaster, OCT_NET_MASK, &genNet));
    /* Scan each net for degen pTerms */
    while (octGenerate(&genNet, &octNet)==OCT_OK) {

	/* Make sure all aTerms connect to only one subnet */

	OH_ASSERT(octInitGenContents(&octNet, OCT_TERM_MASK, &genTerm));
	while (octGenerate(&genTerm, &octATerm)==OCT_OK) {
	    /* Check if fTerm */
	    if (octATerm.contents.term.instanceId==(octId)0) {
		octFTerm = octATerm;
		if (octGenFirstContent(&octFTerm, OCT_TERM_MASK, &octATerm)==OCT_OK) {
		    /* Check if aTerm is attached to net */
		    OH_ASSERT(octInitGenContainers(&octATerm, OCT_NET_MASK, &genNet2));
		    netFoundFlg = FALSE;
		    while (octGenerate(&genNet2, &octNet2)==OCT_OK) {
			if (octNet2.objectId==octNet.objectId) {
			    netFoundFlg = TRUE;
			    break;
			}
		    }
		    octFreeGenerator( &genNet2 );

		    if (netFoundFlg) continue;
		}
		else continue;
	    }
	    /* Go through the subnets of aTerm */
	    OH_ASSERT(octInitGenContainers(&octATerm, OCT_NET_MASK, &genSubNet));
	    chanNum1 = CP_NULL_CHANNEL;
	    while (octGenerate(&genSubNet, &octSubNet)==OCT_OK) {
		if (strcmp(octSubNet.contents.net.name, "CDSubnet")) continue;

		OH_ASSERT(MIOctGetByNameContainers(&octSubNet, &octSubNetChan));
		CP_CHAN_TO_CHANNUM(&chanNum, &octSubNetChan);
		if (chanNum==CP_NULL_CHANNEL) {
		    MSG(0, ("aTerm subnet not attached to any channel\n"));
		}
		else if (chanNum1==CP_NULL_CHANNEL) {
		    chanNum1   = chanNum;
		    octSubNet2 = octSubNet;
		}
		else {
		    if (octATerm.contents.term.instanceId!=(octId)0) {
			OH_ASSERT(octGenFirstContainer(&octATerm,
						       OCT_INSTANCE_MASK, &octInst));
			MSG(0, ("aTerm %s of inst %s has more than one subnet\n",
				octATerm.contents.term.name,
				octInst.contents.instance.name));
		    }
		    else {
			MSG(0, ("fTerm %s has more than one subnet\n",
				octATerm.contents.term.name));
		    }

		    /* Move stuff from 2nd subnet to 1st subnet of aterm */

		    OH_ASSERT(octInitGenContents(&octSubNet, OCT_ALL_MASK, &genSubStuff));
		    while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
			OH_ASSERT(octInitGenContents(&octSubNet2, OCT_ALL_MASK, &genSubStuff2));
			while (octGenerate(&genSubStuff2, &octSubStuff2)==OCT_OK) {
			    if (octSubStuff.objectId==octSubStuff2.objectId)  {
				break;
			    }
			}
			octFreeGenerator( &genSubStuff2 );

			if (octSubStuff.objectId!=octSubStuff2.objectId)
			  OH_ASSERT(octAttach(&octSubNet2, &octSubStuff));
		    }

		    MSG(2, ("Deleting subnet (octId=%d)\n", octSubNet.objectId));
		    OH_ASSERT(octDelete(&octSubNet));
		    if (chanNum1!=chanNum) {
			MSG(0, ("aTerm subnets attached to more than 1 channel\n"));
		    }
		    rc = 8;
		}
	    }
	}

	/* Init pTermOctId[] (unique list of all pTerms in net) */

	OH_ASSERT(octInitGenContents(&octNet, OCT_NET_MASK, &genSubNet));
	MSG(2, ("Cleaning subnets of net %s (octId=%d)\n",
		octNet.contents.net.name, octNet.objectId));
	numPTerm = 0;
	while (octGenerate(&genSubNet, &octSubNet)==OCT_OK) {
	    if (strcmp(octSubNet.contents.net.name, "CDSubnet")) continue;
	    OH_ASSERT(octInitGenContents(&octSubNet, OCT_BAG_MASK, &genPTerm));
	    /* Generate all pTerms of all nets */
	    while (octGenerate(&genPTerm, &octPTerm)==OCT_OK) {
		if (strcmp(octPTerm.contents.bag.name, "CDPseudoTerm")) continue;

		for (p=0; p<numPTerm; p++)
		  if (pTermOctId[p]==octPTerm.objectId) break;
		if (p==numPTerm) {
		    DURANGE(numPTerm, CP_MAX_PTERM_PER_NET);
		    pTermOctId[numPTerm++] = octPTerm.objectId;
		}
	    }
	    octFreeGenerator( &genPTerm );
	}

	/* Check all pTerms under this net */

	for (p=0; p<numPTerm; p++) {
	    octPTerm.objectId = pTermOctId[p];
	    if (octGetById(&octPTerm)!=OCT_OK) continue;
	    MSG(3, ("On pTerm (octId=%d)\n", octPTerm.objectId));

	    /* Look at parent subnets of pTerm */

	    OH_ASSERT(octInitGenContainers(&octPTerm, OCT_NET_MASK, &genSubNet));
	    numSubNet = 0;
	    chanNum1 = chanNum2 = CP_NULL_CHANNEL;
	    while (octGenerate(&genSubNet, &octSubNet)==OCT_OK) {
		DASSERT(octSubNet.objectId!=octNet.objectId);
		OH_ASSERT(MIOctGetByNameContainers(&octSubNet, &octSubNetChan));
		CP_CHAN_TO_CHANNUM(&chanNum, &octSubNetChan);
		if (chanNum==CP_NULL_CHANNEL) {
		    MSG(0, ("pTerm subnet of net %s not attached to any channel\n",
			    octNet.contents.net.name));
		    continue;
		}
		else if (chanNum1==CP_NULL_CHANNEL)
		  chanNum1 = chanNum;
		else if (chanNum2==CP_NULL_CHANNEL) {
		    if (chanNum1!=chanNum)
		      chanNum2 = chanNum;
		    else {
			MSG(0, ("pTerm (%d) connected to at least 2 subnets of same channel %d\n",
				octPTerm.objectId, chanNum));
		    }
		}
		else if (chanNum==chanNum1 || chanNum==chanNum2) {
		    MSG(0, ("pTerm (%d) connected to at least 3 subnets with 2 of the same channel %d\n",
			    octPTerm.objectId, chanNum));
		}
		else {
		    MSG(0, ("pTerm (%d) connected to more than two channels\n",
			    octPTerm.objectId));
		    rc = 8;
#ifdef CP_FIX_FABIO
		    {
			octObject	oldSubNet, newSubNet;
			octObject	newPTerm;
			octObject	midChan;
			short		midChanNum;
			short		k;

			/* Don't ask me what the fuck this shit does */

			MSG(-1, ("Attempting to fix it\n"));
			midChan.type = OCT_BAG;
			midChan.contents.bag.name = "CDChanDef";
			midChanNum = MAX(chanNum, MAX(chanNum1, chanNum2));
			if (midChanNum==chanNum) {
			    OH_ASSERT(MIOctGetByNameContainers(&octSubNet, &midChan));
			    OH_ASSERT(ohGetById(&oldSubNet, pTermSubNetOctId[numSubNet-1]));
			}
			else {
			    for (k=0; k<numSubNet; k++)
			      if (pTermSubNetChanNum[k]==midChanNum) break;
			    OH_ASSERT(ohGetById(&oldSubNet, pTermSubNetOctId[k]));
			    OH_ASSERT(MIOctGetByNameContainers(&oldSubNet, &midChan));
			    oldSubNet = octSubNet;
			}
			OH_ASSERT(octDetach(&oldSubNet, &octPTerm));
			OH_ASSERT(ohCreateBag(&oldSubNet, &newPTerm, "CDPseudoTerm"));
			pTermOctId[numPTerm++] = newPTerm.objectId;
			OH_ASSERT(ohCreateNet(&octNet, &newSubNet, "CDSubnet"));
			OH_ASSERT(octAttach(&newSubNet, &octPTerm));
			OH_ASSERT(octAttach(&newSubNet, &newPTerm));
			OH_ASSERT(octAttach(&midChan, &newSubNet));
			if (midChanNum==chanNum) {
			    pTermSubNetOctId  [numSubNet-1] = newSubNet.objectId;
			    pTermSubNetChanNum[numSubNet-1] = midChanNum;
			}
			else {
			    octSubNet = newSubNet;
			    chanNum   = midChanNum;
			}
		    }
#endif CP_FIX_FABIO
		}

		DURANGE(numSubNet, CP_MAX_SUBNET_PER_PTERM);
		pTermSubNetOctId  [numSubNet] = octSubNet.objectId;
		pTermSubNetChanNum[numSubNet] = chanNum;
		numSubNet++;
	    }
	    octFreeGenerator( &genSubNet );

	    if (numSubNet!=2) {
		MSG(0, ("in net %s (octId=%d)\n",
			octNet.contents.net.name, octNet.objectId));
		MSG(0, ("%d subnets connected to pTerm (octId=%d) should be 2\n",
			numSubNet, octPTerm.objectId));
	    }

	    /* Tabulate subnets of pTerm for channel 1 and 2 */

	    numPTermSubNetChan1 = 0;
	    numPTermSubNetChan2 = 0;
	    for (psn=0; psn<numSubNet; psn++) {
		if (pTermSubNetChanNum[psn]==chanNum1)
		  pTermSubNetChan1[numPTermSubNetChan1++] = pTermSubNetOctId[psn];
		if (pTermSubNetChanNum[psn]==chanNum2 &&
		    chanNum2!=CP_NULL_CHANNEL)
		  pTermSubNetChan2[numPTermSubNetChan2++] = pTermSubNetOctId[psn];
	    }

	    /* If pTerm has multiple subnets of 1st channel then move other pterms of
	       other subnets to a single subnet */

	    if (numPTermSubNetChan1>1) {
		OH_ASSERT(ohGetById(&octSubNet, pTermSubNetChan1[0]));
		rc = 8;
	    }
		
	    for (psn=1; psn<numPTermSubNetChan1; psn++) {
		OH_ASSERT(ohGetById(&octSubNet2, pTermSubNetChan1[psn]));
		OH_ASSERT(octInitGenContents(&octSubNet2, OCT_ALL_MASK, &genSubStuff));
		while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK)
		  if (octSubStuff.objectId!=octPTerm.objectId)
		    OH_ASSERT(octAttach(&octSubNet, &octSubStuff));
		MSG(0, ("Deleting subNet (octId=%d) of net %s\n",
			octSubNet2.objectId, octNet.contents.net.name));
		OH_ASSERT(octDelete(&octSubNet2));
	    }

	    /* If pTerm comes from subnets of only one channel then delete it */

	    if (chanNum2==CP_NULL_CHANNEL) {
		MSG(0, ("PTerm(%d) of net %s chan %d is degenerate\n",
			octPTerm.objectId, octNet.contents.net.name, chanNum1));
		MSG(0, ("Deleting pTerm (octId=%d) of net %s\n",
			octPTerm.objectId, octNet.contents.net.name));
		OH_ASSERT(octDelete(&octPTerm));
		rc = 8;
	    }

	    /* If pTerm has multiple subnets of 2nd channel then move other pterms of
	       other subnets to a single subnet */

	    if (numPTermSubNetChan2>1) {
		OH_ASSERT(ohGetById(&octSubNet, pTermSubNetChan2[0]));
		rc = 8;
	    }
		
	    for (psn=1; psn<numPTermSubNetChan2; psn++) {
		OH_ASSERT(ohGetById(&octSubNet2, pTermSubNetChan2[psn]));
		OH_ASSERT(octInitGenContents(&octSubNet2, OCT_ALL_MASK, &genSubStuff));
		while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK)
		  if (octSubStuff.objectId!=octPTerm.objectId)
		    OH_ASSERT(octAttach(&octSubNet, &octSubStuff));
		MSG(0, ("INFO Deleting subNet (octId=%d) of net %s\n",
			octSubNet2.objectId, octNet.contents.net.name));
		OH_ASSERT(octDelete(&octSubNet2));
	    }
	}
    }

    MI_MOD_END();
    return(rc);
}				/* ............................. End of procedure CPOctCleanSubNets */

optionStruct optionList[] = {
    { "i",	"","interactive channel by channel mode"},
    { "p",	"","prompting for partial save"},
    { "c",	"","create \"cleaned\" facet for mosaicoDR"},
    { "d",	"","use pin direction property mode"},
    { "C",	"","compact all channels"},
    { "F",	"chan_compact_file",
		    "file with the numbers of the channels to be compacted"},
    { "m",	"","via minimization (on channels to be compacted)"},
    { "R",	"route_directory_name",
		    "directory name for routing stuff (default is RTstuff)"},
    { "r",	"router_name","router name (YACR or MIGHTY, default is YACR)"},
    { "o",	"output_view","output view name (default is input_view)"},
    { OPT_RARG,	"cell:view","the cell to be routed"},
    { 0,	0, 0}
};
/* Evaluate input parameters
----------------------------------------------------------------------------*/
CPEvalParm(argc, argv)
int		argc;
String		argv[];
{
    int		option;
    int		c;
    String		strchr();
    FILE		*f;

    CPParm.partialMode       = FALSE;
    CPParm.interactiveMode   = FALSE;
    CPParm.pinDirMode        = FALSE;
    CPParm.createPurgedFacet = FALSE;
    CPParm.skipChExtern      = TRUE;
    CPParm.compactChanMode   = FALSE;
    CPParm.viaMinMode        = FALSE;
    CPParm.compactChanFile   = NULL;
    CPParm.routerName        = "YACR";
    CPParm.outputView        = NULL;
    CPParm.routeDirName      = "RTstuff";

    CPParm.numCompactChan    = 0;
    CPParm.compactChanNum    = NULL;

    while ((option=optGetOption(argc, argv))!=EOF) {
	switch(option) {
	case 'i':	CPParm.interactiveMode = TRUE;
	    break;
	case 'p':	CPParm.partialMode = TRUE;
	    break;
	case 'c':	CPParm.createPurgedFacet = TRUE;
	    break;
	case 'd':	CPParm.pinDirMode = TRUE;
	    break;
	case 'C':	CPParm.compactChanMode = TRUE;
	    break;
	case 'm':	CPParm.viaMinMode = TRUE;
	    break;
	case 'R':	CPParm.routeDirName = STR_COPY(optarg);
	    if ( access( CPParm.routeDirName, 0 ) == -1 ) {
		printf( "mosaicoDR: creating %s\n", CPParm.routeDirName );
		mkdir( CPParm.routeDirName, 0777 );
	    }
	    break;
	case 'F':	CPParm.compactChanMode = TRUE;
	    CPParm.compactChanFile = STR_COPY(optarg);
	    break;
	case 'r':	CPParm.routerName = STR_COPY(optarg);
	    break;
	case 'o':	CPParm.outputView = STR_COPY(optarg);
	    break;
	default:	optUsage();
	}
    }

    if (optind+1!=argc) optUsage();


    CPParm.cellName = STR_COPY(argv[optind]);
    CPParm.inputView = strchr(CPParm.cellName,':');
    if (CPParm.inputView) {
	*CPParm.inputView = '\0';
	CPParm.inputView++;
	if (!CPParm.outputView)
	  CPParm.outputView = CPParm.inputView;
    }
    if (!CPParm.inputView || !(*CPParm.inputView))
      optUsage();

    VOVbegin( argc, argv );

    if (CPParm.compactChanFile) {
	f = MIfopen(CPParm.compactChanFile,"r","channels to be compacted");
	if (!f) optUsage();
	for ( ; fscanf(f, "%d", &c)!=EOF; CPParm.numCompactChan++);
	fclose(f);
	CPParm.compactChanNum = ALLOC_ARRAY(CPParm.numCompactChan, short);
	f = MIfopen(CPParm.compactChanFile,"r","channels to be compacted");
	for (c=0 ; fscanf(f, "%hd", &CPParm.compactChanNum[c])!=EOF; c++);
	fclose(f);
    }

}				/* ............................. End of procedure CPEvalParm */
void	octMakeSureItsFree(octFacet)
octObject	*octFacet;
{
    int		i = 0;

    if ( closeFacetsFlag ) {
	while (octGetById(octFacet)==OCT_OK && octFreeFacet(octFacet)==OCT_OK) i++;
    }
    if (i>0) {
	printf("%s: Facet %s:%s had to be freed %d times\n", MIPrefix,
	       octFacet->contents.facet.cell, octFacet->contents.facet.view, i);
    }
}
