/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPRedoVulcan.c

Create new geometry for layer "PLACE" which surrounds interface facet.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CPRedoVulcan.c,v $
 * Revision 1.2  89/02/06  16:44:27  octtools
 * removed 'OCT_ASSERT' from octGetPoints when
 * used for finding out the number of points
 * 
 * Revision 1.1  89/01/26  14:20:14  octtools
 * Initial revision
 * 
 * Revision 1.10  88/06/15  14:06:55  mitch
 * Include connector polygons when regenerating the PLACE layer
 * 
 * Revision 1.6  88/04/24  16:52:14  mitch
 * Use octOpenMaster instead of MIOctOpenMasterOfInst.
 * 
 * Revision 1.1  88/03/31  11:23:10  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"
#include "miutil.h"
#include "CPInc.h"
#include "vov.h"

extern	void	octMakeSureItsFree();
/* Get oct data
----------------------------------------------------------------------------*/
CPRedoVulcan(octFacet, mode)

octObject	*octFacet;
String		mode;
{
octObject	placeLayer;
octObject	octGeom;

harList		harPerimList;
octCoord	growth;
octCoord	minFeatureSize = 20 * 10;
int		k;
Boolean		foundFlg = FALSE;
static Boolean	savedPlacePtsFlag = FALSE;
static struct octPoint	*octPtArray, *harPtArray;
static int32	numPoints;

int		x, y;
octObject	instBag, connBag;
octObject	octInst;
octObject	masterPlaceLayer;
octObject	instMaster;
octGenerator	genInst;
octGenerator	genGeom;
Boolean		conInstFlag = FALSE;

MI_MOD_START("CPRedoVulcan");

if (!strcmp(mode, "save")) {
    savedPlacePtsFlag = FALSE;

    /* Must add place layers of instance facets */
    OH_ASSERT(ohGetOrCreateLayer(octFacet, &placeLayer, "PLACE"));

/*    OH_ASSERT(ohGetByBagName(octFacet, &instBag, "INSTANCES")); */
/*    OH_ASSERT(octInitGenContents(&instBag, OCT_INSTANCE_MASK, &genInst)); */
    /* Andrea said to try all instances */
    OH_ASSERT(octInitGenContents(octFacet, OCT_INSTANCE_MASK, &genInst));

    while (octGenerate(&genInst, &octInst)==OCT_OK) {
CPRedoVulcanAddInstPoly:
	OH_ASSERT(ohOpenMaster(&instMaster, &octInst, "interface", "r"));
	VOVinputFacet( &instMaster );
	if (ohGetByLayerName(&instMaster, &masterPlaceLayer,"PLACE")!=OCT_OK) {
	    octGeom.type = OCT_BOX;
	    OH_ASSERT(octBB(&octInst, &octGeom.contents.box));
	    OH_ASSERT(octCreate(&placeLayer, &octGeom));
	    if (*debugFlag>=2) {
		octBox b;
		OH_ASSERT(octBB(&octGeom, &b));
		MSG(-1, ("%s: bBox (%d %d) (%d %d)\n",
		    ohFormatName(&octGeom),
		    b.lowerLeft.x, b.lowerLeft.y,
		    b.upperRight.x, b.upperRight.y));
	    }
	    continue;
	}
        OH_ASSERT(octInitGenContents(&masterPlaceLayer, OCT_GEO_MASK, &genGeom));
	while (octGenerate(&genGeom, &octGeom)==OCT_OK) {
	    switch (octGeom.type) {
	    case OCT_BOX:
		tr_oct_transform(&octInst.contents.instance.transform,
		    &octGeom.contents.box.lowerLeft.x, &octGeom.contents.box.lowerLeft.y);
		tr_oct_transform(&octInst.contents.instance.transform,
		    &octGeom.contents.box.upperRight.x, &octGeom.contents.box.upperRight.y);
		if (octGeom.contents.box.lowerLeft.x>octGeom.contents.box.upperRight.x) {
		    x = octGeom.contents.box.lowerLeft.x,
		    octGeom.contents.box.lowerLeft.x = octGeom.contents.box.upperRight.x;
		    octGeom.contents.box.upperRight.x = x;
		}
		if (octGeom.contents.box.lowerLeft.y>octGeom.contents.box.upperRight.y) {
		    y = octGeom.contents.box.lowerLeft.y,
		    octGeom.contents.box.lowerLeft.y = octGeom.contents.box.upperRight.x;
		    octGeom.contents.box.upperRight.y = y;
		}
		OH_ASSERT(octCreate(&placeLayer, &octGeom));
		if (*debugFlag>=2) {
		    octBox b;
		    OH_ASSERT(octBB(&octGeom, &b));
		    MSG(-1, ("%s: bBox (%d %d) (%d %d)\n",
			ohFormatName(&octGeom),
			b.lowerLeft.x, b.lowerLeft.y,
			b.upperRight.x, b.upperRight.y));
		}
		break;
	    case OCT_POLYGON:
		numPoints = 0;
		(void) octGetPoints(&octGeom, &numPoints, octPtArray);
		octPtArray = ALLOC_ARRAY(numPoints, struct octPoint);
		OH_ASSERT(octGetPoints(&octGeom, &numPoints, octPtArray));
		for (k=0; k<numPoints; k++)
		    tr_oct_transform(&octInst.contents.instance.transform,
			&octPtArray[k].x, &octPtArray[k].y);
		OH_ASSERT(octCreate(&placeLayer, &octGeom));
		OH_ASSERT(octPutPoints(&octGeom, numPoints, octPtArray));
		if (*debugFlag>=2) {
		    octBox b;
		    OH_ASSERT(octBB(&octGeom, &b));
		    MSG(-1, ("%s: bBox (%d %d) (%d %d)\n",
			ohFormatName(&octGeom),
			b.lowerLeft.x, b.lowerLeft.y,
			b.upperRight.x, b.upperRight.y));
		}
	    default: break;
	    }
	}
	OH_ASSERT(octCloseFacet(&instMaster));
	octMakeSureItsFree(&instMaster);

	if (conInstFlag) break;
    }
    octFreeGenerator( &genInst );

    if (!conInstFlag) {
	k = 0;
	conInstFlag = TRUE;
    }
    for ( ; k<2; k++)
	if (k!=CPExtInstNum) {
	    OH_ASSERT(ohGetByInstName(octFacet, &octInst, CPInst[k].name));
	    connBag.type = OCT_BAG;
	    connBag.contents.bag.name = "CONNECTORS";
	    if (MIOctGetByNameContainers(&octInst, &connBag)==OCT_OK) {
		k++;
		goto CPRedoVulcanAddInstPoly;
	    }
	}

    for (k=1; k<=20 && !foundFlg; k++) {
	harPerimList = HAR_EMPTY_LIST;
	growth = k*minFeatureSize;
	
 	harFacetToPerimList(octFacet, growth, 0, &harPerimList);

	if (harRestOfList(harPerimList)!=HAR_EMPTY_LIST) {
	    MSG(2, ("More than one geometry on harPerimList\n"));
	    harFreeList(&harPerimList);
	    continue;
	}
	foundFlg = TRUE;
    }
    if (foundFlg) {
        MSG(1, ("growth factor = %d\n", growth));

	savedPlacePtsFlag = TRUE;
        numPoints = harFirstPerimCount(harPerimList);
	octPtArray = ALLOC_ARRAY(numPoints, struct octPoint);
	harPtArray = harFirstPerimArray(harPerimList);
	MSG(2, ("numPoints in polygon = %d\n", numPoints));
	for (k=0; k<numPoints; k++) {
	    octPtArray[k] = harPtArray[k];
	    MSG(3, ("polygon pt%d = (%d %d)\n",
	      k, octPtArray[k].x, octPtArray[k].y));
	}
	harFreeList(&harPerimList);

	ohDeleteContents(&placeLayer, OCT_GEO_MASK);

	/* Check to see if polygon is rectilinear */
	for (k=0; k<numPoints; k++) {
	    int kn = (k+1)%numPoints;
	    if ((octPtArray[k].x!=octPtArray[kn].x) &&
	        (octPtArray[k].y!=octPtArray[kn].y)) {
		MIerror("New interface PLACE polygon is not rectilinear at location%d = (%d %d)\n",
		    k, octPtArray[k].x, octPtArray[k].y);
	    }
	}
    }
    else
        MIerror("WARNING Could not make place layer into one geometry\n");
}

if (!strcmp(mode, "replace") && savedPlacePtsFlag) {
    MSG(1, ("replacing place layer\n"));
    OH_ASSERT(ohGetByLayerName(octFacet, &placeLayer, "PLACE"));
    ohDeleteContents(&placeLayer, OCT_ALL_MASK);
    octGeom.type = OCT_POLYGON;
    OH_ASSERT(octCreate(&placeLayer, &octGeom));
    OH_ASSERT(octPutPoints(&octGeom, numPoints, octPtArray));
    if (*debugFlag>=2) {
	octBox b;
	OH_ASSERT(octBB(&octGeom, &b));
	MSG(-1, ("%s: bBox (%d %d) (%d %d)\n",
	    ohFormatName(&octGeom),
	    b.lowerLeft.x, b.lowerLeft.y,
	    b.upperRight.x, b.upperRight.y));
    }
    FREE_ARRAY(numPoints, octPtArray, struct octPoint);
}

MI_MOD_END();
} /* ............................. End of procedure CPRedoVulcan */
