/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPRemapCorn.c

Remap corner points of new channel instance.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CPRemapCorn.c,v $
 * Revision 1.2  90/02/07  16:13:48  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/26  14:20:15  octtools
 * Initial revision
 * 
 * Revision 1.7  88/04/25  15:17:58  mitch
 * Fixed bug in corner remapping of z-shaped channels.
 * 
 * Revision 1.1  88/03/31  11:23:18  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CPInc.h"
#include "vov.h"

extern	void	octMakeSureItsFree();
/* Remap corners of new merged instance
----------------------------------------------------------------------------*/
CPRemapCorners()
{
    short		numCorn = 0;
    CPCornDef	corn[CP_MAX_NUM_CORN];
    short		k;
    short		c, cStart;
    int		delta, minDelta;
    short		b, minBdyNum;
    short		chanNum;
    CPLocDef	cornLoc;

#define		MAX_OCT_PTS 1000
    int32		numOctPts;
    octObject	octInstMaster;
    octObject	octInstMasterCont;
    octObject	octOldInst;
    struct octPoint	octPts[MAX_OCT_PTS];
    Boolean		orientFlg;

    int		minExtLoc, maxExtLoc;
    short		extBdyNum;

    short		bdyIdx;
    octObject	octCEdgeChan;
    octObject	octCorner;
    octObject	octCEdge;
    octObject	octCProp;

    MI_MOD_START("CPRemapCorners");

    OH_ASSERT(ohOpenMaster(&octInstMaster, &CPOctSubMasterInst, "interface", "r"));
    VOVinputFacet( &octInstMaster );
    numOctPts = MAX_OCT_PTS;
    MIOctGetInstBdy(&octInstMaster, &CPOctSubMasterInst, octPts,
		    &numOctPts, &orientFlg);
    OH_ASSERT(octCloseFacet(&octInstMaster));
    octMakeSureItsFree(&octInstMaster);

    OH_ASSERT(ohOpenMaster(&octInstMasterCont, &CPOctSubMasterInst,
			   "contents", "r"));
    for (k=0; k<2; k++) {
	if (k==CPExtInstNum) continue;

	MSG(1, ("Remapping corners of inst %s:\n", CPInst[k].name));

	OH_ASSERT(ohGetByInstName(&octInstMasterCont, &octOldInst, CPInst[k].name));
	CPInst[k].newLoc.x = octOldInst.contents.instance.transform.translation.x;
	CPInst[k].newLoc.y = octOldInst.contents.instance.transform.translation.y;
	MSG(2, ("Wired version of inst translated by (%d %d)\n",
		CPInst[k].newLoc.x-CPInst[k].loc.x,
		CPInst[k].newLoc.y-CPInst[k].loc.y));

	/* Find starting corner */
	for (c=0; c<CPInst[k].numCorn; c++)
	  if (CPInst[k].corn[c].chanNum==CPCurChan) break;
	if (c==CPInst[k].numCorn)
	  MIerror("Corners of inst %s bad! Nothing bordering channel\n",
		  CPInst[k].name);
	cStart = c + (CPInst[k].skipCornFlg ? 2 : 1 );
	cStart = cStart%CPInst[k].numCorn;
	MSG(2, ("CornStart=%d numCorn=%d\n", cStart, CPInst[k].numCorn));

	for (c=cStart; CPInst[k].corn[c].chanNum!=CPCurChan;
	     c=(c+1)%CPInst[k].numCorn) {
	    cornLoc.x  = CPInst[k].bdy[CPInst[k].corn[c].bdyNum].x;
	    cornLoc.y  = CPInst[k].bdy[CPInst[k].corn[c].bdyNum].y;
	    cornLoc.x += CPInst[k].newLoc.x-CPInst[k].loc.x;
	    cornLoc.y += CPInst[k].newLoc.y-CPInst[k].loc.y;
	    chanNum    = CPInst[k].corn[c].chanNum;

	    MSG(2, ("Trying to find corner %d: (%d %d) chan=%d\n",
		    c, cornLoc, chanNum));

	    minDelta = MAXINT;
	    for (b=0; b<numOctPts; b++) {
		delta = ABS(octPts[b].x-cornLoc.x) + ABS(octPts[b].y-cornLoc.y);
		if (delta<minDelta) {
		    minDelta  = delta;
		    minBdyNum = b;
		}
	    }
	    if (minDelta>0) {
		MIerror("Could not remap channel corners\n");
		MSG(-1, ("Caused by bad boundary of new instance master %s\n",
			 CPOctSubMasterInst.contents.instance.master));
		MSG(-1, ("Corner at (%d %d) of inst %s mapped to (%d %d)\n",
			 cornLoc, CPInst[k].name, octPts[minBdyNum]));
	    }
	    MSG(2, ("Corner remapped to b=%d dx=%d\n", minBdyNum, minDelta));

	    URANGE(numCorn, CP_MAX_NUM_CORN);
	    corn[numCorn].bdyNum  = minBdyNum;
	    corn[numCorn].chanNum = chanNum;
	    numCorn++;
	}
    }
    OH_ASSERT(octCloseFacet(&octInstMasterCont));
    octMakeSureItsFree(&octInstMasterCont);

    if (CPExtInstNum==1) {	/* An external channel: */
	MSG(1, ("Initialize external channel corners\n"));

	cStart  = (cStart-1+CPInst[0].numCorn)%CPInst[0].numCorn;
	chanNum = CPInst[0].corn[cStart].chanNum;

	minExtLoc =   MAXINT;
	maxExtLoc = -(MAXINT);
	if (CPWhere.chanDir==CP_V) {
	    if (CPExtInstNum==CPWhere.instOnBot) { /* West side */
		MSG(2, ("Channel on West side\n"));
		for (b=0; b<numOctPts; b++)
		  if (octPts[b].x<minExtLoc) {
		      minExtLoc = octPts[b].x;
		      extBdyNum = b;
		  }
		if (extBdyNum==0 && octPts[1].x>minExtLoc)
		  extBdyNum = numOctPts-1;
	    }
	    else {		/* East side */
		MSG(2, ("Channel on East side\n"));
		for (b=0; b<numOctPts; b++)
		  if (octPts[b].x>maxExtLoc) {
		      maxExtLoc = octPts[b].x;
		      extBdyNum = b;
		  }
		if (extBdyNum==0 && octPts[1].x<maxExtLoc)
		  extBdyNum = numOctPts-1;
	    }
	}
	else {
	    if (CPExtInstNum==CPWhere.instOnBot) { /* South side */
		MSG(2, ("Channel on South side\n"));
		for (b=0; b<numOctPts; b++)
		  if (octPts[b].y<minExtLoc) {
		      minExtLoc = octPts[b].y;
		      extBdyNum = b;
		  }
		if (extBdyNum==0 && octPts[1].y>minExtLoc)
		  extBdyNum = numOctPts-1;
	    }
	    else {		/* North side */
		MSG(2, ("Channel on North side\n"));
		for (b=0; b<numOctPts; b++)
		  if (octPts[b].y>maxExtLoc) {
		      maxExtLoc = octPts[b].y;
		      extBdyNum = b;
		  }
		if (extBdyNum==0 && octPts[1].y<maxExtLoc)
		  extBdyNum = numOctPts-1;
	    }
	}

	corn[numCorn].bdyNum  = extBdyNum;
	corn[numCorn].chanNum = CP_NULL_CHANNEL;
	numCorn++;
	corn[numCorn].bdyNum  = (extBdyNum+1)%numOctPts;
	corn[numCorn].chanNum = chanNum;
	numCorn++;
    }
    /* Put corners into oct */
    MSG(2, ("Putting corners into oct\n"));
    OH_ASSERT(ohCreateBag(&CPOctSubMasterInst, &octCorner, "CDCorners"));
    for (c=0; c<numCorn; c++) {	/* Each corner: */
	chanNum = corn[c].chanNum;
	if (orientFlg==CP_CW)
	  then bdyIdx = numOctPts-corn[c].bdyNum-1;
	else bdyIdx = corn[c].bdyNum;
	MSG(3, ("On corner %d: bdyIdx=%d\n", c, bdyIdx));

	OH_ASSERT(ohCreatePropInt((c==0 ? &octCorner : &octCEdge),
				  &octCProp, "CDCorner", bdyIdx));
	/* Attach to channel */
	MSG(3, ("Corner edge in channel %d\n", chanNum));
	if (chanNum==CP_NULL_CHANNEL)
	  then octCEdgeChan.objectId = octCorner.objectId;
	else octCEdgeChan.objectId = CPChanOctId[chanNum];

	OH_ASSERT(octGetById(&octCEdgeChan));
	OH_ASSERT(ohCreateBag(&octCEdgeChan, &octCEdge, "CDCornerEdge"));
	OH_ASSERT(octAttach(&octCEdge, &octCProp));
	if (chanNum==CP_NULL_CHANNEL)
	  OH_ASSERT(octDetach(&octCEdgeChan, &octCEdge));
    }
    MI_MOD_END();
}				/* ............................. End of procedure CPRemapCorners */
