/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

Create shadow of the channel boundary.

CPShadowChBdy		Put shadow of channel boundary.

Author(s):		Mitsuru Igusa
Revision history:	$Log:	CPShadowChBdy.c,v $
 * Revision 1.3  90/02/07  16:13:49  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.2  89/02/07  13:21:14  octtools
 * got rid of routeChan
 * 
 * Revision 1.1  89/01/26  14:20:16  octtools
 * Initial revision
 * 
 * Revision 1.1  88/03/31  11:23:29  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"

typedef	struct	CPLocDef	CPLocDef;
struct CPLocDef {
    int         x, y;
};

Boolean		CPShadowVertMode;
Boolean		CPShadowHiToLoMode;
CPLocDef	*CPShadowEdge[2];

CPShadowChBdy(bdy, sBdyNum, eBdyNum, mod, vertMode, hiToLoMode,
    f, layerNum)

CPLocDef	*bdy;
int		sBdyNum, eBdyNum, mod;
Boolean		vertMode, hiToLoMode;
FILE		*f;
short		layerNum;
{
int		numEdge, e, e2, es, ej;
int		numBdy, numSBdy, b;
int		*edgeOrder;
CPLocDef	*edge[2], *edgeCopy[2];
CPLocDef	*sBdy;
CPLocDef	loc;
int		CPQsortOrderChEdge();
Boolean		foundFlg;
int		minC, maxC;

MI_MOD_START("CPShadowChBdy");

CPShadowVertMode   = vertMode;
CPShadowHiToLoMode = hiToLoMode;

numBdy = eBdyNum-sBdyNum+1;
if (sBdyNum>eBdyNum) numBdy += mod;

DASSERT(!(numBdy%2));
numEdge = numBdy/2;
edge[0] = ALLOC_ARRAY(numEdge, CPLocDef);
edge[1] = ALLOC_ARRAY(numEdge, CPLocDef);
if (hiToLoMode)
    for (b=sBdyNum, e=0; e<numEdge; b=(b+2)%mod, e++) {
	edge[0][e] = bdy[b];
	edge[1][e] = bdy[(b+1)%mod];
    }
else
    for (b=eBdyNum, e=0; e<numEdge; b=(b-2+mod)%mod, e++) {
	edge[0][e] = bdy[b];
	edge[1][e] = bdy[(b-1+mod)%mod];
    }

MSG(1, ("chanDir=%c hiToLo=%d\n", vertMode ? 'v' : 'h', hiToLoMode));
for (e=0; e<numEdge; e++) {
    MSG(1, ( "Edge %d: (%d %d) - (%d %d)\n", e, edge[0][e], edge[1][e]));
}

edgeOrder = ALLOC_ARRAY(numEdge, int);
for (e=0; e<numEdge; e++)
    edgeOrder[e] = e;
CPShadowEdge[0] = edge[0];
CPShadowEdge[1] = edge[1];
qsort((char *)edgeOrder, numEdge, sizeof(int), CPQsortOrderChEdge);

edgeCopy[0] = ALLOC_ARRAY(numEdge, CPLocDef);
edgeCopy[1] = ALLOC_ARRAY(numEdge, CPLocDef);
for (e=0; e<numEdge; e++) {
    e2 = edgeOrder[e];
    edgeCopy[0][e] = edge[0][e2];
    edgeCopy[1][e] = edge[1][e2];
}
FREE_ARRAY(numEdge, edge[0], CPLocDef);
FREE_ARRAY(numEdge, edge[1], CPLocDef);
FREE_ARRAY(numEdge, edgeOrder, int);
edge[0] = edgeCopy[0];
edge[1] = edgeCopy[1];
					/* Check qsort */
for (e=0; e<numEdge; e++) {
    MSG(1, ( "Edge %d: (%d %d) - (%d %d)\n", e, edge[0][e], edge[1][e]));
    if (vertMode)
	DASSERT(edge[0][e].x==edge[1][e].x)
    else
	DASSERT(edge[0][e].y==edge[1][e].y)
    if (e==numEdge-1) break;
    if (vertMode) {
	if (edge[0][e].y!=edge[0][e+1].y)
	    DASSERT(edge[0][e].y>edge[0][e+1].y)
	else if (hiToLoMode)
	    DASSERT(edge[0][e].x<edge[0][e+1].x)
	else
	    DASSERT(edge[0][e].x>edge[0][e+1].x)
    }
    else {
	if (edge[0][e].x!=edge[0][e+1].x)
	    DASSERT(edge[0][e].x>edge[0][e+1].x)
	else if (hiToLoMode)
	    DASSERT(edge[0][e].y<edge[0][e+1].y)
	else
	    DASSERT(edge[0][e].y>edge[0][e+1].y)
    }
}
					/* Init shadow boundary */
sBdy = ALLOC_ARRAY(numEdge+1, CPLocDef);
numSBdy = 0;
sBdy[numSBdy] = edge[0][0];
MSG(1, ("Bdy %d: (%d %d)\n", numSBdy, sBdy[numSBdy]));
numSBdy++;
for (e=ej=0; e<numEdge-1; ) {
    loc = edge[1][e];
    foundFlg = FALSE;
    for (e2=ej+1; e2<numEdge; e2++) {
	if (vertMode) {
	    if (edge[0][e2].y<edge[1][e2].y) continue;
	    if (edge[1][e2].y>=edge[0][e].y) {
		ej = e2;
		continue;
	    }
	    if ((!hiToLoMode && edge[1][e2].x>loc.x) ||
		( hiToLoMode && edge[1][e2].x<loc.x)) {
		sBdy[numSBdy] = edge[0][e2];
		foundFlg = TRUE;
		break;
	    }
	    if (edge[0][e2].y<=loc.y) break;
	}
	else {
	    if (edge[0][e2].x<edge[1][e2].x) continue;
	    if (edge[1][e2].x>=edge[0][e].x) {
		ej = e2;
		continue;
	    }
	    if (( hiToLoMode && edge[1][e2].y>loc.y) ||
		(!hiToLoMode && edge[1][e2].y<loc.y)) {
		sBdy[numSBdy] = edge[0][e2];
		foundFlg = TRUE;
		break;
	    }
	    if (edge[0][e2].x<=loc.x) break;
	}
    }
    if (foundFlg) {
	e = ej = e2;
	MSG(1, ("Bdy %d: (%d %d)\n", numSBdy, sBdy[numSBdy]));
	numSBdy++;
	continue;
    }
    sBdy[numSBdy] = loc;
    minC =  MAXINT;
    maxC = -MAXINT;
    es = 0;
    for (e2=1; e2<numEdge; e2++) {
	if (( vertMode && edge[0][e2].y<loc.y) ||
	    (!vertMode && edge[0][e2].x<loc.x)) break;
	if (vertMode && edge[1][e2].y<loc.y) {
	    if (hiToLoMode && edge[1][e2].x<minC) {
		sBdy[numSBdy].x = minC = edge[1][e2].x;
		es = e2;
	    }
	    else if (!hiToLoMode && edge[1][e2].x>maxC) {
		sBdy[numSBdy].x = maxC = edge[1][e2].x;
		es = e2;
	    }
	}
	if (!vertMode && edge[1][e2].x<loc.x) {
	    if (!hiToLoMode && edge[1][e2].y<minC) {
		sBdy[numSBdy].y = minC = edge[1][e2].y;
		es = e2;
	    }
	    else if (hiToLoMode && edge[1][e2].y>maxC) {
		sBdy[numSBdy].y = maxC = edge[1][e2].y;
		es = e2;
	    }
	}
    }
    if (es>0) {
	e = es;
	MSG(1, ("Bdy %d: (%d %d)\n", numSBdy, sBdy[numSBdy]));
	numSBdy++;
    }
    else break;
}

fprintf(f, "%d %d\n", numSBdy*2, layerNum);

if (hiToLoMode)				/* Print bdy */
    for (e=0; e<numSBdy; e++) {
	fprintf(f, "%d %d ", sBdy[e].x, sBdy[e].y); /* Changed for SUN4 */
	if (vertMode)
	     fprintf(f, "%d %d ",
		 sBdy[e].x, (e<numSBdy-1 ? sBdy[e+1].y : bdy[eBdyNum].y));
	else fprintf(f, "%d %d ",
		 (e<numSBdy-1 ? sBdy[e+1].x : bdy[eBdyNum].x), sBdy[e].y);
    }
else
    for (e=numSBdy-1; e>=0; e--) {
	if (vertMode)
	     fprintf(f, "%d %d ",
		 sBdy[e].x, (e<numSBdy-1 ? sBdy[e+1].y : bdy[sBdyNum].y));
	else fprintf(f, "%d %d ",
		 (e<numSBdy-1 ? sBdy[e+1].x : bdy[sBdyNum].x), sBdy[e].y);
	fprintf(f, "%d %d ", sBdy[e].x, sBdy[e].y);
    }
fprintf(f, "\n");

FREE_ARRAY(numEdge, edge[0], CPLocDef);
FREE_ARRAY(numEdge, edge[1], CPLocDef);
FREE_ARRAY(numEdge+1, sBdy, CPLocDef);

MI_MOD_END();
} /* ............................. End of procedure CPShadowChBdy */
/* Channel edge ordering program.
----------------------------------------------------------------------------*/
int
CPQsortOrderChEdge(e1, e2)

int		*e1, *e2;
{
int		d;

if (CPShadowVertMode) {
    d = CPShadowEdge[0][*e2].y - CPShadowEdge[0][*e1].y;
    if (d) return(d);			/* Higher edge first */
    else {
	d = CPShadowEdge[0][*e1].x - CPShadowEdge[0][*e2].x;
	if (CPShadowHiToLoMode)
	then return(d);			/* Low x first */
	else return(-d);		/* Hi x first */
    }
}
else {
    d = CPShadowEdge[0][*e2].x - CPShadowEdge[0][*e1].x;
    if (d) return(d);			/* Higher edge first */
    else {
	d = CPShadowEdge[0][*e1].y - CPShadowEdge[0][*e2].y;
	if (CPShadowHiToLoMode)
	then return(d);			/* Low y first */
	else return(-d);		/* Hi y first */
    }
}
} /* ............................. End of procedure CPQsortOrderChEdge */
