/*

Taken from spiderPower.c:

 	SPquantizeWidth()
 	SPquantizeWidthDOWN()
 	SPcurrentDensity()

To compute the width of a terminal do the following:

    width = peakCurrentDensityOfPin/SPcurrentDensity(pinLayer);
    width = SPquantizeWidth(width);


Taken from tap_stuff.c:

	set_up_width_and_spacing()

*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "tap.h"
#include "miutil.h"
/* #include <sys/resource.h> */

#define OCT_LAMBDA 20
#define DEFAULT_MAX_CURRENT_DENSITY 0.1e-4 /* in amperes per lambda */
#define POLY 0
#define MET1 1
#define MET2 2

int	layer_width[3];
int	layer_spacing[3];
char	*layer_nameTable[3] = {"POLY", "MET1", "MET2"};
char	*TECH_DIR;

int
SPquantizeWidth(w)
    int	w;			/* original width */
{
  int	m;			/* w MOD OCT_LAMBDA*/
  int	nw;			/* new width in lambda */

  /* The width is in "OCT" units: It must be rounded to the nearest LAMBDA */
  
  m = w % OCT_LAMBDA;
  if (m != 0) {
    w += OCT_LAMBDA - m;	/* round up to next OCT_LAMBDA */
  }

  /*
   * Quantize: Quantization sequence is (in lambda) :
   * 1,2,3,4,5,6,8,10,12,14,17,20,25,30,35,.... (up by 5)
   */
  nw = w / OCT_LAMBDA;		/* Work in lambda  */
  if (nw > 6) {
    if ( nw <= 14 ) {
      if (nw % 2 != 0)  nw = ((nw / 2) + 1) * 2;
    } else if ( nw <= 20) { 
      if ((nw-14) % 3 != 0)   nw = 14 + (((nw-14) / 3) + 1) * 3;
    } else if (nw < 30) {
      if (nw % 5 != 0)   nw = ((nw / 5) + 1) * 5;
    } else {
      nw = nw + 5 - nw%5;
    }
  }
  return(nw * OCT_LAMBDA);	/* return OCT units */
}


int
SPquantizeWidthDOWN(w)
    int	w;			/* original width */
{
  int	m;			/* w MOD OCT_LAMBDA*/
  int	nw;			/* new width in lambda */

  /* The width is in "OCT" units: It must be rounded to the nearest LAMBDA */
  
  m = w % OCT_LAMBDA;
  if (m != 0) {
    w -= m;	/* round down to next OCT_LAMBDA */
  }

  /*
   * Quantize: Quantization sequence is (in lambda) :
   * 1,2,3,4,5,6,8,10,12,14,17,20,25,30,35,.... (up by 5)
   */
  nw = w / OCT_LAMBDA;		/* Work in lambda  */
  if (nw > 6) {
    if ( nw <= 14 ) {
      if (nw % 2 != 0)  nw = (nw / 2) * 2;
    } else if ( nw <= 20) { 
      if ((nw-14) % 3 != 0)   nw = 14 + (((nw-14) / 3)) * 3;
    } else if (nw < 30) {
      if (nw % 5 != 0)   nw = (nw / 5) * 5;
    } else {
      nw = 30;			/* MAX allowed size (temporary) */
      MSG(-1, ("WARNING Maximum width is 30\n"));
    }
  }
  return(nw * OCT_LAMBDA);	/* return OCT units */
}










set_up_width_and_spacing(layers)
octObject layers[];
{
octObject facet;

  OH_ASSERT(octGenFirstContainer(layers, OCT_FACET_MASK, &facet));
  TECH_DIR = tapGetDirectory(&facet);

  layer_width[POLY] = tapGetMinWidth(&(layers[POLY]));
  layer_width[MET1] = tapGetMinWidth(&(layers[MET1]));
  layer_width[MET2] = tapGetMinWidth(&(layers[MET2]));
  
  layer_spacing[POLY] = tapGetMinSpacing(&(layers[POLY]),&(layers[POLY]));
  layer_spacing[MET1] = tapGetMinSpacing(&(layers[MET1]),&(layers[MET1]));
  layer_spacing[MET2] = tapGetMinSpacing(&(layers[MET2]),&(layers[MET2]));

  MSG(1, ("POLY min width = %d, min spacing = %d\n",
      layer_width[POLY], layer_spacing[POLY]));
  MSG(1, ("MET1 min width = %d, min spacing = %d\n",
      layer_width[MET1], layer_spacing[MET1]));
  MSG(1, ("MET2 min width = %d, min spacing = %d\n",
      layer_width[MET2], layer_spacing[MET2]));

  if((layer_width[POLY] == -1) || (layer_width[MET1] == -1) || (layer_width[MET2] == -1) || (layer_spacing[POLY] == -1) || (layer_spacing[MET1] == -1) || (layer_spacing[MET2] == -1))
    MIerror("in getting layer widths and spacing\n");
}




static double curr_denTable[3] = {0.0, 0.0, 0.0};

int SPinit( facet )
    octObject* facet;
{
    static int init = 0;

    if ( !init && facet != 0 ) {
	int i;
	octObject layer;
	double coordSize = 1.e-6;
	tapLayerListElement layerList[1];

	if( curr_denTable[MET1] == 0.0 )  {
	    if ( tapGetProp( facet, TAP_COORD_SIZE ) ) {
		coordSize = TAP_COORD_SIZE->value.real;
	    }
	    for ( i = 0; i < 3 ; i++ ) {
		layer.type = OCT_LAYER;
		layer.contents.layer.name = layer_nameTable[i];
		octGetOrCreate( facet, &layer );
		layerList[0].layer = layer;
		if ( tapGetLyrProp( 1, layerList, TAP_MAX_CURRENT_DENSITY ) ) {
		    curr_denTable[i] = TAP_MAX_CURRENT_DENSITY->value.real * coordSize;
		} else {
		    curr_denTable[i] = 0.0015;
		}
		/* printf("%s current density %lf (amps/oct unit)\n", layer_nameTable[i], curr_denTable[i] ); */
	    }
	}
	init = 1;
    }
    return init;
}

/*
 * returns the maximum current density 
 * for a particular layer. If no layer is specified, then
 * some default is returned. 
 */
double SPcurrentDensity(layerNum)
    int layerNum;
{
    if ( ! SPinit( 0 ) ) {
	printf( "spider power not initialized" );
    }

    if(((layerNum != MET1) && (layerNum != MET2) &&  (layerNum != POLY)) ||
       (curr_denTable[layerNum] == 0.0)) {
	return(DEFAULT_MAX_CURRENT_DENSITY);
    }
    return curr_denTable[layerNum];
}
