#include "port.h"
#include "defs.h"
#include "oct.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"

#define 	MAX_NUM_POINTS		400
#define		MAX_NUM_EDGES		400
#define 	MAX_NUM_CHANNELS	2000
#define		MAX_NUM_NODES		2000
#define		MAX_NUM_PINS		500
#define		MAX_NUM_CHAR		500
#define		NUM_LAYERS		3
#define		BUFFER_DIM		500
#define		POLYGON

#define		VERTICAL		0
#define		HORIZONTAL  		1
#define		NORTH			0
#define		WEST			1
#define		SOUTH			2
#define		EAST			3
#define		ANY_DIRECTION		-1


struct point{
	octId			id;
	int			index;
	struct octPoint		*ptoPoint;
	octId			firstChannel;
	octId			secondChannel;
};

struct instInfo{
	octId			instanceId;
	octTransformType	transform;
	char			*name;
	int 			bottom;
	int			top;
	int			right;
	int			left;
};

struct instList{
	struct instInfo		*instInfo;
	struct instList		*nextInList;
};

struct edge{
	octId			id;
	int			channelNumber;
	int			outerEdge;
	int			width;
};

struct channelInfo{
	octId			id;
	int			channelNumber; /* CDchannel number */
	int			width;
	int			type;
	int			centerLine;
	int			length;
	int			capacity;
	int			outerChannel;
	int			numberOfTerms;
	int			lengthFactor;
	int			used;
};

struct node{
	int			number;
	int			channelNumber; /* CDchannel number */
	int			x;
	int			y;
	int			width;
	struct channelList 	*ptoFirstStartingChannel;
	struct channelList 	*ptoFirstEndingChannel;
};

struct netList{
	octId			id;
	char			*name;
	struct netList 		*ptoNextNetInList;
};

struct layerList{
	char			*layerName;
	struct layerList 	*ptoNextInList;
};

struct pseudoTerm{
	octId			pseudoTermBagId;
	int			nodeNumber;
	int			numberSubnets;
	struct subnet		*subnet[5];

};

typedef struct channel{
	int			number;
	struct node		*firstNode;
	struct node   		*secondNode;
	struct channelInfo 	*info;
	struct subnetList	*ptoFirstSubnet;
	struct termList		*ptoFirstTerm;
} *PTOCHANNEL;

struct channelList{
	struct channel	  	*ptoChannel;
	struct channelList 	*ptoNextChannel;
};

struct dir{
	int			distance;
	struct layerList	*layer[3];
};

struct terminal{
	octId			actualTermId;
	octId			jointBagId;
	char			*name;
	char			*termName;
	struct instInfo		*instInfo;
	int			number;
	struct terminal		*equivTerm;
	int			routeEquivTerm; /* pseudo-equivTerminal
						   added to account for
						   terminals that can be
						   projected on more than
						   one channel */
	int			distance;       /* computed from firstNode */
	int			channelNumber;
	int			implNumber;
	struct node		*firstNode;
	struct node   		*secondNode;
	struct netList		*firstNetInList; /* net connected to the term */
						/* implementation of the terminal */
	struct octBox		box;		/* on the master */
	struct octPoint		*center;	/* on the instance */
	struct octPoint		*lowerLeft;	/* on the instance */
	struct octPoint		*upperRight;	/* on the instance */
	struct layerList	*layerIds;
	int			numberOfImplementations;
	struct dir		direction[4];   /* direction of the projection */
	int			normalProj;     /* flag that is normally set
						   to 1 to signify that the terminal
						   has been projected perpendicularly
						   on the channel. Its value being 0
						   means that the projection is 
						   not perpendicular. This happens
						   when the terminal is to be projected onto
						   a Z-shape channel */
};


struct termList{
	struct terminal		*ptoTerm;
	struct termList 	*ptoNextTerm;
};


struct net{
	octId			id;
	char			*name;
	int 			number;
	int			length;
	struct net		*ptoNextNet;
	struct termList 	*firstTermOnNet;
	struct subnetList	*ptoFirstSubnet;
};

struct subnet{
	int			startingPoint;
	int			endingPoint;
	int			firstNodeIndex;
	int			secondNodeIndex;
	struct channel		*ptoChannel;
	struct termList 	*ptoFirstTermInList;
	int			tricky;
	octId			id;
};

struct subnetList{
	struct subnet		*ptoSubnet;
	struct subnetList	*ptoNextSubnet;
};
extern struct subnetList* newSubnetList();

struct gphEdge{
	int			firstNodeIndex;
	int			secondNodeIndex;
	int			startingPoint;
	int			endingPoint;
	struct channel		*ptoChannel;
	struct subnet		*ptoSubnet;
};

