#include "cds.h"
#include "oh.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern FILE		*dump ;
extern int		numChannels ;
extern int 		numNodes ;
extern struct channel	*arrayOfChannels[MAX_NUM_CHANNELS];
extern struct node	*arrayOfNodes[MAX_NUM_NODES];
extern struct pseudoTerm       arrayOfPseudoTerms[MAX_NUM_NODES];
extern int		outerChan;
struct gbcObj{
	octId		id;
	struct gbcObj	*next;
};
struct gbcObj		*firstGbcObj = NIL(struct gbcObj);



sanitizeChannels()
{
    int 		i;
    int			starting;
    int			counter;
    int			channelUsed = numChannels;
    struct channelList	*pointer;

    if( outerChan == NO ){
	for( i = 0 ; i < numChannels ; i++ ){
	    if( arrayOfChannels[i]->info->outerChannel == TRUE ){
		if( arrayOfChannels[i]->info->numberOfTerms != 0 ){
		    arrayOfChannels[i]->info->used = YES;
		    findPath( arrayOfChannels[i] );
		}
	    }
	}
	for( i = 0 ; i < numChannels ; i++ ){
	    if( arrayOfChannels[i]->info->used <= UNUSED ){
		removeChannel( arrayOfChannels[i] );
		channelUsed--;
	    }
	}
    }
    /* Now remove channels like b in fig.

			- 0 - - a - - 0 - - b - - 0
			  |           |
			  |           |
			  c           d
			  |           |
			  |           |
			- 0 - - e - - 0 -
			  |           |

       with no terminals attached to it */

    for( i = 0 ; i < numNodes ; i++ ){
	counter = 0 ;
	for( pointer = arrayOfNodes[i]->ptoFirstStartingChannel;
	     pointer != NIL(struct channelList);
	     pointer = pointer->ptoNextChannel ){
	    counter++;
	    starting = TRUE;
	}
	for( pointer = arrayOfNodes[i]->ptoFirstEndingChannel;
	     pointer != NIL(struct channelList);
	     pointer = pointer->ptoNextChannel ){
	    counter++;
	    starting = FALSE;
	}
	if( counter > 1 || counter == 0 )
	    continue;
	else if( starting == FALSE ){
	    if( arrayOfNodes[i]->ptoFirstEndingChannel->ptoChannel->info->numberOfTerms == 0 ){
		removeChannel( arrayOfNodes[i]->ptoFirstEndingChannel->ptoChannel );
		channelUsed--;
	    }
	}
	else if( starting == TRUE ){
	    if( arrayOfNodes[i]->ptoFirstStartingChannel->ptoChannel->info->numberOfTerms == 0 ){
		removeChannel( arrayOfNodes[i]->ptoFirstStartingChannel->ptoChannel );
		channelUsed--;
	    }
	}
    }
    return(channelUsed);
}	




removeChannel( ptoChannel )
struct channel	*ptoChannel;
{
    int			index;
    int			counter = 0;
    struct channelList	*pointer;
    struct channelList	*prevPtr;

    index = ptoChannel->firstNode->number - 1;
    prevPtr = arrayOfNodes[index]->ptoFirstStartingChannel;
    for( pointer = arrayOfNodes[index]->ptoFirstStartingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	counter++;
	if( pointer->ptoChannel == ptoChannel ){
	    if( counter == 1 )
		arrayOfNodes[index]->ptoFirstStartingChannel = pointer->ptoNextChannel;
	    else
		prevPtr->ptoNextChannel = pointer->ptoNextChannel;
	    break;
	} else {
	    prevPtr = pointer;
	} 
    }
    index = ptoChannel->secondNode->number - 1;
    prevPtr = arrayOfNodes[index]->ptoFirstEndingChannel;
    for( pointer = arrayOfNodes[index]->ptoFirstEndingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	counter++;
	if( pointer->ptoChannel == ptoChannel ){
	    if( counter == 1 )
		arrayOfNodes[index]->ptoFirstEndingChannel = pointer->ptoNextChannel;
	    else
		prevPtr->ptoNextChannel = pointer->ptoNextChannel;
	    break;
	} else {
	    prevPtr = pointer;
	} 
    }
    index = ptoChannel->number - 1;
    arrayOfChannels[index]->firstNode = NIL(struct node);
    arrayOfChannels[index]->secondNode = NIL(struct node);
    arrayOfChannels[index]->info = NIL(struct channelInfo);
    arrayOfChannels[index]->ptoFirstSubnet = NIL(struct subnetList);
    arrayOfChannels[index]->ptoFirstTerm = NIL(struct termList);
}




findPath(ptoChannel)
struct channel	*ptoChannel;
{

    /*

	This routine does not account for cases like the one represented in
	the following picture:


	0 - - e - - 0 - - e - - 0 - - e - - 0
	| +---x---+ | +-------+ | +---x---+ |
	| |       | | |       | | |       | |     x = terminal to be connected 
	e | cell1 | i x cell2 | i | cell3 | e         by the same net
	| |       | | |       | | |       | |     i = internal Channel
	| +---x---+ | +-------+ | +---x---+ |     e = external Channel
	0 - - e - - 0 - - e - - 0 - - e - - 0

	After the elimination of unnecessary channels the situation becomes:

	0 - - e - - 0           0 - - e - - 0
	| +---x---+ | +-------+ | +---x---+ |
	| |       | | |       | | |       | |     x = terminal to be connected 
	e | cell1 | i x cell2 | i | cell3 | e         by the same net
	| |       | | |       | | |       | |     i = internal Channel
	| +---x---+ | +-------+ | +---x---+ |     e = external Channel
	0 - - e - - 0           0 - - e - - 0

	To avoid a situation like this, it is necessary to determine for each 
	terminal, assigned to an external channel, a path that connects it to 
	the other terminals on the same net.
	Note however, that in a layout  like the one showed above, the outer 
	channels are really part of the routing areas...

    */


    int			index, i;
    int			counter = 0;
    int			pathFound = FALSE;
    struct channelList	*pointer;
    struct channel	*strChan[4], *endChan[4];

    for( i = 0 ; i < 4 ; i++ ){
	strChan[i] = NIL(struct channel);
	endChan[i] = NIL(struct channel);
    }
    index = ptoChannel->firstNode->number - 1;
    for( pointer = arrayOfNodes[index]->ptoFirstStartingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	if( pointer->ptoChannel == ptoChannel ) continue;
	if( pointer->ptoChannel->info->outerChannel == NO ) {
	    for( i = 0 ; i < counter ; i++ ){
		strChan[i]->info->used--;
	    }
	    pathFound = TRUE;
	    counter = 0;
	    break;
	}
	strChan[counter] = pointer->ptoChannel;
	strChan[counter++]->info->used++;
    }
    for( pointer = arrayOfNodes[index]->ptoFirstEndingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	if( pointer->ptoChannel->info->outerChannel == NO ) {
	    for( i = 0 ; i < counter ; i++ ){
		endChan[i]->info->used--;
	    }
	    pathFound = TRUE;
	    break;
	}
	endChan[counter] = pointer->ptoChannel;
	endChan[counter++]->info->used++;
    }
    for( i = 0 ; i < 4 ; i++ ){
	strChan[i] = NIL(struct channel);
	endChan[i] = NIL(struct channel);
    }
    counter = 0;
    index = ptoChannel->secondNode->number - 1;
    for( pointer = arrayOfNodes[index]->ptoFirstEndingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	if( pointer->ptoChannel == ptoChannel ) continue;
	if( pointer->ptoChannel->info->outerChannel == NO ) {
	    for( i = 0 ; i < counter ; i++ ){
		endChan[i]->info->used--;
	    }
	    counter = 0;
	    pathFound = TRUE;
	    break;
	}
	endChan[counter] = pointer->ptoChannel;
	endChan[counter++]->info->used++;
    }
    for( pointer = arrayOfNodes[index]->ptoFirstStartingChannel;
	 pointer != NIL(struct channelList);
	 pointer = pointer->ptoNextChannel){
	if( pointer->ptoChannel->info->outerChannel == NO ) {
	    for( i = 0 ; i < counter ; i++ ){
		strChan[i]->info->used--;
	    }
	    pathFound = TRUE;
	    break;
	}
	strChan[counter] = pointer->ptoChannel;
	strChan[counter++]->info->used++;
    }
    if(pathFound) return;
    /*
    fprintf(stderr,"MOSAICO_GLOBAL_ROUTE_ERROR:\n");
    fprintf(stderr,"The channel between nodes %d <%d,%d> and %d <%d,%d>\n",
		    ptoChannel->firstNode->number,
		    ptoChannel->firstNode->x,
		    ptoChannel->firstNode->y,
		    ptoChannel->secondNode->number,
		    ptoChannel->secondNode->x,
		    ptoChannel->secondNode->y );
    fprintf(stderr,"may be disconnected\n");
    */
}



cleanSubnet( pTerm, numPseudoTerms, index )
struct pseudoTerm	*pTerm;
int			numPseudoTerms, index;
{
    /* This routine is to eliminate un-necessary subnets.
       A situation like this:



		     |        |
		     |        |
		     |        |
    -------*---------+--------+--------------       *  terminal
           |					    o  pseudoTerminal (junction between
	   +----1----o---+2---o---3---                 two subnets)
                         |	     |
    ---------------------*-----------*-------       #  subnets


       Becomes:

		     |        |
		     |        |
		     |        |
    -------*---------+--------+--------------
           |                         
	   +-------1-----+------------               
                         |	     |
    ---------------------*-----------*-------        

      Number of subnets     3 --> 1
      Number of pseudoterms 2 --> 0

    */


    /* clean up the database */

    octObject	PTerm, FirstSubnet, SecondSubnet, Obj;
    octGenerator objGen;
    int		i, j;

    PTerm.objectId = pTerm->pseudoTermBagId;

    OH_ASSERT( octGetById(&PTerm) );
    OH_ASSERT( octGetById( &FirstSubnet ) );
    OH_ASSERT( octGetById( &SecondSubnet ) );

    addToGbcList( &PTerm );
    
    OH_ASSERT( octInitGenContents( &SecondSubnet, OCT_ALL_MASK, &objGen ) );

    while( octGenerate( &objGen, &Obj ) != OCT_GEN_DONE ){
	OH_ASSERT( octDetach( &SecondSubnet, &Obj ) );
	OH_ASSERT( octAttach( &FirstSubnet, &Obj ) );
    }

    OH_ASSERT( octDelete( &SecondSubnet ));

    /* now update the internal datastructure  
       this is the minimum necessary to proceed in the cleaning with no
       problems. 
       If the checks are runned after this point, some inconsistencies may
       arise...*/

    for( i = 0 ; i < numPseudoTerms ; i++ ){
	if( i == index ) continue;
	for(j = 0 ; j < arrayOfPseudoTerms[i].numberSubnets ; j++ ){
	    /*
	    if (octIdsEqual(arrayOfPseudoTerms[i].subnetId[j], FirstSubnet.objectId)) {
		firstPterm = i;
		firstSubnet = j;
		break;
	    }
	    */
	    if (octIdsEqual(arrayOfPseudoTerms[i].subnet[j]->id, SecondSubnet.objectId)) {
		arrayOfPseudoTerms[i].subnet[j]->id = FirstSubnet.objectId;
		return;
	    }
	}
    }
}



addToGbcList( Obj )
octObject 	*Obj;
{
    struct gbcObj	newObj;

    newObj.id = Obj->objectId;
    newObj.next = firstGbcObj;
    firstGbcObj = &newObj;
}

cleanGbcList()
{
    struct gbcObj	*obj;
    octObject		Obj;
    int			counter = 0;

    for( obj = firstGbcObj; obj != NIL(struct gbcObj); obj = obj->next ){
	Obj.objectId = obj->id;
	OH_ASSERT( octGetById( &Obj ) );
	counter += octDeleteR(&Obj);
    }
    if( debug ){
	fprintf( dump, "Total number of objects deleted from the database %d\n",
		    counter);
    }
}
