#include "cds.h"
#include "oh.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern int		verbose;
extern FILE		*dump ;

checkPseudoTerms( PseudoTerm )
octObject	*PseudoTerm;
{
    octObject		PsTerm;
    octObject		Net, Subnet;
    octGenerator	PTgen, subnetGen, genNet;
    octStatus		status, status1, status2, status3;
    int 		counter;

    OH_ASSERT( octInitGenContents( PseudoTerm, OCT_BAG_MASK, &PTgen) );

    while( (status = octGenerate( &PTgen, &PsTerm)) == OCT_OK )  {
	OH_ASSERT( octInitGenContainers( &PsTerm, OCT_NET_MASK, &subnetGen) );
	counter = 0;
	while( (status1 = octGenerate( &subnetGen, &Subnet)) == OCT_OK )  {
	    if (strCmp( Subnet.contents.net.name , "CDSubnet") == 0)  
		counter++;
	}
	if (status1 != OCT_GEN_DONE)  {
	    octError("Bad exit from subnet generator");
	    myExit( -1, __FILE__, __LINE__);
	}
	switch( counter ) {
	    case 0 :
		if(verbose){
		    fprintf( dump, "PseudoTerm %d  with no subnet connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
		    while( (status2 = octGenerate( &genNet, &Net )) == OCT_OK )
			fprintf(dump, "Subnet %d belongs to net %s \n",
				Subnet.objectId,
				Net.contents.net.name);
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from net generator");
		    }
		}
		break;

	    case 1 :
		if(verbose){
		    fprintf( dump, "PseudoTerm %d  with only one subnet connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
		    while( ( status2 = octGenerate( &genNet, &Net ) ) == OCT_OK )
			fprintf(dump, "Subnet %d  belongs to net %s \n",
				Subnet.objectId,
				Net.contents.net.name);
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from net generator");
		    }
		}
		break;

	    case 2 :
		if(debug){
		    fprintf( dump, "PseudoTerm %d  with two subnets connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &PsTerm, OCT_NET_MASK, &subnetGen) );
		    while( ( status2 = octGenerate( &subnetGen, &Subnet )) == OCT_OK ){
			if( strCmp( Subnet.contents.net.name , "CDSubnet") == 0 ) {
			    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
			    while( ( status3 = octGenerate( &genNet, &Net ) ) == OCT_OK )
				fprintf(dump, "Subnet %d  belongs to net %s \n",
					Subnet.objectId,
					Net.contents.net.name);
			    if( status3 != OCT_GEN_DONE ) {
				octError("Bad exit from net generator");
			    }
			}
		    }
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from subnet generator");
		    }
		}
		break;
	    case 3 :
		if(debug){
		    fprintf( dump, "PseudoTerm %d  with three subnets connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &PsTerm, OCT_NET_MASK, &subnetGen) );
		    while( ( status2 = octGenerate( &subnetGen, &Subnet )) == OCT_OK ){
			if( strCmp( Subnet.contents.net.name , "CDSubnet") == 0 ) {
			    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
			    while( ( status3 = octGenerate( &genNet, &Net ) ) == OCT_OK )
				fprintf(dump, "Subnet %d  belongs to net %s \n",
					Subnet.objectId,
					Net.contents.net.name);
			    if( status3 != OCT_GEN_DONE ) {
				octError("Bad exit from net generator");
			    }
			}
		    }
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from subnet generator");
		    }
		}
		break;
	    case 4 :
		if(debug){
		    fprintf( dump, "PseudoTerm %d  with four subnets connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &PsTerm, OCT_NET_MASK, &subnetGen) );
		    while( ( status2 = octGenerate( &subnetGen, &Subnet )) == OCT_OK ){
			if( strCmp( Subnet.contents.net.name , "CDSubnet") == 0 ) {
			    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
			    while( ( status3 = octGenerate( &genNet, &Net ) ) == OCT_OK )
				fprintf(dump, "Subnet %d  belongs to net %s \n",
					Subnet.objectId,
					Net.contents.net.name);
			    if( status3 != OCT_GEN_DONE ) {
				octError("Bad exit from net generator");
			    }
			}
		    }
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from subnet generator");
		    }
		}
		break;
	    default :
		if(debug){
		    fprintf( dump, "PseudoTerm %d  with more than four subnets connected to it!!\n",
			    PsTerm.objectId);
		    OH_ASSERT( octInitGenContainers( &PsTerm, OCT_NET_MASK, &subnetGen) );
		    while( ( status2 = octGenerate( &subnetGen, &Subnet ) ) == OCT_OK ){
			if( strCmp( Subnet.contents.net.name , "CDSubnet") == 0 ) {
			    OH_ASSERT( octInitGenContainers( &Subnet, OCT_NET_MASK, &genNet ) );
			    while( ( status3 = octGenerate( &genNet, &Net ) ) == OCT_OK )
				fprintf(dump, "Subnet %d  belongs to net %s \n",
					Subnet.objectId,
					Net.contents.net.name);
			    if( status3 != OCT_GEN_DONE ) {
				octError("Bad exit from net generator");
			    }
		    }
		    }
		    if( status2 != OCT_GEN_DONE ) {
			octError("Bad exit from subnet generator");
		    }
		}
		break;
	}
    }
    if (status != OCT_GEN_DONE)  {
	octError("Bad exit from p-term generator");
    }
}


channelInfo( channel )
struct channel	*channel;
{
    fprintf( dump, "channel number %d < mitch number %d >  id %d \n",
	channel->number,
	channel->info->channelNumber,
	channel->info->id);
    fprintf( dump, "first node %d  x  %d   y   %d \n",
	channel->firstNode->number,
	channel->firstNode->x,
	channel->firstNode->y );
    fprintf( dump, "second node %d  x  %d   y   %d \n",
	channel->secondNode->number,
	channel->secondNode->x,
	channel->secondNode->y );
}

checkSubnets( Chip )
octObject *Chip;
{
    octObject		Net;
    octObject		Subnet;
    octObject		Term;
    octObject		PseudoTerm;
    octGenerator	genNet;
    octGenerator	genSubnet;
    octGenerator	genTerm;
    octGenerator	genPseudoTerm;

    OH_ASSERT( octInitGenContents( Chip, OCT_NET_MASK, &genNet )  );
    while( octGenerate( &genNet, &Net ) == OCT_OK ){
	fprintf( dump, "Net name %s \n", Net.contents.net.name );
	OH_ASSERT( octInitGenContents( &Net, OCT_NET_MASK, &genSubnet )  );
	while( octGenerate( &genSubnet, &Subnet ) == OCT_OK ){
	    fprintf( dump, "Subnet name %s \n", Subnet.contents.net.name );
	    OH_ASSERT( octInitGenContents( &Subnet, OCT_TERM_MASK, &genTerm )  );
	    while( octGenerate( &genTerm, &Term ) == OCT_OK ){
		fprintf( dump, "Term name %s \n", Term.contents.term.name );
	    }
	    OH_ASSERT( octInitGenContents( &Subnet, OCT_BAG_MASK, &genPseudoTerm )  );
	    while( octGenerate( &genPseudoTerm, &PseudoTerm ) == OCT_OK ){
		if( strCmp( PseudoTerm.contents.bag.name , "CDPseudoTerm") == 0 ){
		    fprintf( dump, "PseudoTerm name %s <octId %d> \n", 
			    PseudoTerm.contents.bag.name,
			    PseudoTerm.objectId);
		}
	    }
	}
    }
}
