#include "port.h"
#include "utility.h"
/*
 *   MACRO DEFINITIONS
 *
 *   Macros are distinguished by using solely capital letters in their
 *   identifiers.  This contrasts with C defined identifiers which are strickly 
 *   lower case, and program variable and procedure names which use both upper 
 *   and lower case.
 */

/* Begin macros. */

/* Boolean data type */
#define  BOOLEAN	int
#define  FALSE		0
#define  TRUE		1
#define  NOT		!
#define  AND		&&
#define  OR		||


#define YES  	1
#define NO	0
#define MAYBE	-1
#define OK   1
#define KO   0
#define LEFT 0
#define RIGHT 1
#define TOP 2
#define BOTTOM 3
#define FIRST  1	/* used in owner link/unlink routine */
#define LAST   9	/* used in owner link/unlink routine */
#define NEW	1
#define CURRENT	0
#define UNUSED 0
#define UNDEFINED 0
#define CWISE 1
#define CCWISE -1



/* Macro commands */


/* Macro function that returns the absolute value of a floating point number. */
#define  SIGN(a)            ((a) < 0.0 ? -1. : ((a) > 0.0 ? 1. : 0.0 ))

/* Macro procedure that swaps two entities. */
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}



/* Macro for random number generation */
#define RAND(a) ( a = ( a * 1103515245 + 12345 ) & MAX_INT ))
#define NORM(n) ( (double) n / (double) MAX_INT )
#define NRAND(a) ( (double) ( a = ( a * 1103515245 + 12345 ) & MAX_INT ) / (double) MAX_INT )


#define ACCEPT(d,t) ( (d) < 0. 			? 1. : \
		    ( (t) == 0. 		? 0. : \
		    (((double) (d) / (t)) > 25. ? 0. : exp( -(double)(d) / (t) ))))

#define EXP(d,t) ( (t) == 0. ? 0. : exp( -(double) (d) / (t) ) )
#define MY_EXP(x) ( (x) > 88.03 ? MAX_DOUBLE : ( (x) < - 84.12 ? MIN_DOUBLE : exp(x)))
#define MY_POW(x,p) ( ( (p) == 0. && (x) != 0. )	? 1. : \
		    ( (p) == 1. 		        ? x  : pow((x),(p))))

#define LOG_B(base,x) ( log(x)/log(base) )

#define RINT(a) ( (int) floor( (double) a + 0.5 ) )

/*
 *  This macro finds the intersection between the intervals (min1, max1) and
 *  (min2, max2).  The intersection is returned in (resultMin, resultMax).
 *  If the intersection is empty, resultMin-resultMax gives the distance
 *  between (min1, max1) and (min2, max2).
 */
#define INTERVAL_INTERSECT(resultMin, resultMax, min1, max1, min2, max2) \
    resultMin = ((min1) > (min2) ? (min1) : (min2)); \
    resultMax = ((max1) < (max2) ? (max1) : (max2))

