#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"
extern int		debug ; 
extern FILE		*dump ;
extern int 		numPoints ;
extern int		numPolyPoints ;
extern int 		numMagicPoints ;
extern struct point	magicPoints[MAX_NUM_POINTS];
extern struct octPoint  polyPoints[MAX_NUM_POINTS];
extern int		magicPointsOrientation;
int			lowerCoordinate, upperCoordinate;


/*
*  	getPolygon(ptrBag, points, edges )
*
*	Routine to reconstruct the polygon from bags and points
*	
*/
getPolygon(ptrBag, points, edges)
octObject	*ptrBag;
struct octPoint	points[];
struct edge	edges[];
{
    octId		oldBag;
    octGenerator	genPoint;
    octObject		Point, NewPoint;
    struct edge		*presentEdge;
    int			searchDone = FALSE;

    oldBag = ptrBag->objectId;
    if(octInitGenContents(ptrBag, OCT_POINT_MASK, &genPoint) != OCT_OK){
	octError("generating the generator for the points in the polygon");
	myExit( -1, __FILE__, __LINE__);
    }
    octGenerate(&genPoint, &Point);
    presentEdge = ALLOC( struct edge, 1);
    numPoints = 0;
    while(searchDone == FALSE){
	points[numPoints].x = Point.contents.point.x;
	points[numPoints].y = Point.contents.point.y;
	if(debug) fprintf( dump, " %d    x %d   y %d\n" ,numPoints, points[numPoints].x,points[numPoints].y);
        searchDone = getNewPoint(&Point, &NewPoint, presentEdge, &oldBag );
	edges[numPoints].channelNumber = presentEdge->channelNumber;	
	edges[numPoints].id = presentEdge->id;	
	edges[numPoints].width = presentEdge->width;	
	edges[numPoints].outerEdge = presentEdge->outerEdge;	
	(numPoints)++;
	Point = NewPoint;
	if( numPoints >= MAX_NUM_POINTS ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "The number of poly points found %d exceeds the maximum number allowed %d\n", numPoints, MAX_NUM_POINTS );
	    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_POINTS in file cd2tw.h\n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( numPoints >= MAX_NUM_EDGES ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "The number of poly edges found %d exceeds the maximum number allowed %d\n", numPoints, MAX_NUM_EDGES );
	    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_EDGES in file cd2tw.h\n");
	    myExit( -1, __FILE__, __LINE__);
	}
    }
    return(numPoints);
}



getNewPoint(point, newPoint, presentEdge, oldBag)
octObject	*point, *newPoint;
struct edge	*presentEdge;
octId		*oldBag;
{
    octGenerator	genBag, genParentBag;
    octGenerator	genProp;
    octGenerator	genPoint;
    octObject		Bag, parentBag;
    octObject		Prop;
    octObject		Prop1;
    octObject		Prop2;
    int 		propertyFound;
    int			status, bagStatus;
    int 		parentBagFound;
    int			bagFound;

    Prop1.type = OCT_PROP;
    Prop1.contents.prop.name = "CDChanOrder";
    Prop2.type = OCT_PROP;
    Prop2.contents.prop.name = "CDOuterChannel";
    if(octInitGenContainers(point, OCT_BAG_MASK, &genBag) != OCT_OK){
	octError("generating the generator for the bag representing the subchannel");
	myExit( -1, __FILE__, __LINE__);
    }
    bagFound = FALSE;
    while( ( bagStatus = octGenerate(&genBag, &Bag) ) == OCT_OK){
	if(strCmp(Bag.contents.bag.name, "CDGraphEdge") == 0 && !octIdsEqual(Bag.objectId, *oldBag)) {
	    *oldBag = Bag.objectId;
	    bagFound = TRUE;
	    if(octInitGenContents(&Bag, OCT_PROP_MASK, &genProp) != OCT_OK){
		octError("generating the generator for the property representing the subchannel");
		myExit( -1, __FILE__, __LINE__);
	    }
	    propertyFound = FALSE;
	    while(octGenerate(&genProp,&Prop) == OCT_OK){
		if(strCmp(Prop.contents.prop.name, "CDEdgeWidth") == 0){
		    presentEdge->width = Prop.contents.prop.value.integer;
		    propertyFound = TRUE ;
		    break;
		}
	    }
	    if(propertyFound == FALSE){
		octError("no property present");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if(octInitGenContainers(&Bag, OCT_BAG_MASK, &genParentBag) != OCT_OK){
		octError("generating the generator for the parent bag representing the main channel");
		myExit( -1, __FILE__, __LINE__);
	    }
	    parentBagFound = FALSE;
	    while(octGenerate(&genParentBag, &parentBag) == OCT_OK){
		if(strCmp(parentBag.contents.bag.name, "CDChanDef") == 0){
		    if( octGetByName(&parentBag, &Prop1) != OCT_NOT_FOUND){
			presentEdge->channelNumber = Prop1.contents.prop.value.integer;
			presentEdge->id = parentBag.objectId;
			parentBagFound = TRUE;
			if( (status = octGetByName(&parentBag, &Prop2)) == OCT_NOT_FOUND){
			    presentEdge->outerEdge = FALSE;
			} else if( status == OCT_OK ){
			    presentEdge->outerEdge = TRUE;
			}
			break;
		    }
		} 
	    }
	    if(parentBagFound == FALSE){
		octError("the main channel bag was not found");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if(octInitGenContents(&Bag, OCT_POINT_MASK,&genPoint) != OCT_OK){
		octError("generating the generator for the bag representing the new point");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if(debug){
		fprintf( dump, " newchannel found !!  width %d  id %d \n", 
				    presentEdge->width, presentEdge->id);
	    }
	    while((status = octGenerate(&genPoint, newPoint)) == OCT_OK){
		if (!octIdsEqual(point->objectId, newPoint->objectId)) {
		    return(FALSE);
		}
	    }
	    if(status == OCT_GEN_DONE) {
		return(TRUE);
	    }
	} 
    }
    
    if( bagStatus == OCT_GEN_DONE && bagFound == FALSE ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Bag CDGraphEdge not found\n");
	myExit( -1, __FILE__, __LINE__);
	/*NOTREACHED*/
    }
    /*NOTREACHED*/
}

/*
*	numMagicPoints = getMagicPoints( &Instance );
*
*	routine that extract from OCT the magic points of the channel
*	the magic points determine for each side of the instance the 
*	subchannel to which project the terminals
*
*/
getMagicPoints( Instance )
octObject	*Instance;
{
    octId		oldBag;
    octStatus		status;
    octObject		CornerBag;
    octObject		CornerProperty;
    octObject		EdgeBag;
    octObject		Channel;
    octGenerator	genEdgeBag;
    octGenerator	genCornerProperty;
    octGenerator	genChannels;
    int			numMagicPoints = 0;
    int			presentProperty;
    int 		newPropertyFound = FALSE;
    int			i, index, nextIndex;
    int			presentLine, lowestLine;
    int			presentDiff, nextDiff;
    int			channelFound;

    if(debug) fprintf( dump, "starting generating magic points\n");
    CornerBag.type = OCT_BAG;
    CornerBag.contents.bag.name = "CDCorners";
    CornerProperty.type = OCT_PROP;
    CornerProperty.contents.prop.name = "CDCorner";
    EdgeBag.type = OCT_BAG;
    EdgeBag.contents.bag.name = "CDCornerEdge";
    Channel.type = OCT_BAG;
    lowestLine = MAX_INT;
    magicPointsOrientation = UNDEFINED;
     
    if( octGetByName( Instance, &CornerBag ) != OCT_OK ){
	octError("CornerBag not found");
	myExit( -1, __FILE__, __LINE__);
    }
    oldBag = CornerBag.objectId;
    if( octGetByName( &CornerBag, &CornerProperty ) != OCT_OK ){
	octError("CornerProperty not found");
	myExit( -1, __FILE__, __LINE__);
    }
    for( i = 0 ; i < numPolyPoints ; i++ ){
	if( octInitGenContainers( &CornerProperty, OCT_BAG_MASK, &genEdgeBag ) != OCT_OK ){
	    octError("Initializing EdgeBag generator");
	    myExit( -1, __FILE__, __LINE__);
	}
	while( ( status = octGenerate( &genEdgeBag, &EdgeBag ) ) != OCT_GEN_DONE){
	    if (octIdsEqual(EdgeBag.objectId, oldBag)) {
		continue;
	    } else if( strCmp(EdgeBag.contents.bag.name, "CDCornerEdge") == 0 ) {
		oldBag = EdgeBag.objectId;
		break;
	    } else {
		continue;
	    }
	}
	if( status == OCT_GEN_DONE ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "NewEdge not found\n");
	    myExit( -1, __FILE__, __LINE__);
	}
	presentProperty = CornerProperty.contents.prop.value.integer;
	magicPoints[numMagicPoints].index = presentProperty;
	magicPoints[numMagicPoints].ptoPoint = &polyPoints[presentProperty];
	magicPoints[numMagicPoints].id = CornerProperty.objectId;
	presentLine = polyPoints[presentProperty].x + polyPoints[presentProperty].y;
	if( lowestLine > presentLine ){
	    lowestLine = presentLine;
	    index = numMagicPoints;
	}
	if( octInitGenContainers( &EdgeBag, OCT_BAG_MASK, &genChannels ) < OCT_OK ){
	    octError("Generating the generator for channels \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	channelFound = FALSE;
	while( octGenerate( &genChannels, &Channel ) != OCT_GEN_DONE ){
	    if( strCmp( Channel.contents.bag.name, "CDChanDef" ) == 0 ){
		channelFound = TRUE;
		magicPoints[numMagicPoints].firstChannel = Channel.objectId ;
		if( numMagicPoints > 0 ) {
		    magicPoints[numMagicPoints].secondChannel = 
				magicPoints[numMagicPoints-1].firstChannel;
		}
		break;
	    }
	}
	if( channelFound == FALSE ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "No channel was found attached to edge <%s>\n",
			 octFormatId(EdgeBag.objectId));
	    myExit( -1, __FILE__, __LINE__);
	}
	numMagicPoints++;
	if( numMagicPoints >= MAX_NUM_POINTS ){
	    fprintf( stderr, "MOSAICO_GR_ERROR: \n");
	    fprintf( stderr, "The number of poly points found %d exceeds the maximum number allowed %d\n", numPoints, MAX_NUM_POINTS );
	    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_POINTS in file cd2tw.h\n");
	    myExit( -1, __FILE__, __LINE__);
	}
	newPropertyFound = FALSE;

	if(octInitGenContents(&EdgeBag, OCT_PROP_MASK,&genCornerProperty) != OCT_OK){
	    octError("generating the generator for properties attached to the edge bag");
	    myExit( -1, __FILE__, __LINE__);
	}
	while((status = octGenerate(&genCornerProperty, &CornerProperty)) == OCT_OK){
	    if( CornerProperty.contents.prop.value.integer == presentProperty ){
		continue;
	    } else {
		newPropertyFound = TRUE;
		break;
	    }
	}
	if(newPropertyFound == FALSE){
	    if(debug){
		fprintf( dump, "found %d magic points\n", numMagicPoints);
	        fprintf( dump, "Leaving find magicPoints\n");
	    }
	    magicPoints[0].secondChannel = magicPoints[numMagicPoints-1].firstChannel;
	    if( debug )
		for( i = 0 ; i < numMagicPoints ; i ++ )
		    fprintf( dump, "channel %d magic Point %d  channel %d \n",
			magicPoints[i].secondChannel, i ,
			magicPoints[i].firstChannel);

	    /* determine the orientation of magicPoints  */
	    nextIndex = (index + 1) % numMagicPoints ;
	    presentDiff = magicPoints[index].ptoPoint->y -
			  magicPoints[index].ptoPoint->x;
	    nextDiff = magicPoints[nextIndex].ptoPoint->y -
		       magicPoints[nextIndex].ptoPoint->x;
	    if( nextDiff > presentDiff ) {
		magicPointsOrientation = CWISE;
	    } else if( nextDiff < presentDiff ) {
		magicPointsOrientation = CCWISE;
	    } 
	    return(numMagicPoints);
	}
    }
    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
    fprintf( stderr, "More than %d magic points: possible error, I quit !!\n", numPolyPoints);
    myExit( -1, __FILE__, __LINE__);
    /*NOTREACHED*/
}


getSegment( subnet )
struct subnet	*subnet;
{
    struct node		*first, *second;
    struct channel	*channel;

    channel = subnet->ptoChannel;
    first = channel->firstNode;
    second = channel->secondNode;
    if( channel->info->type == HORIZONTAL ){
	if( ( subnet->startingPoint == 0 ) && ( subnet->endingPoint == 0 ) ){
	    lowerCoordinate = MIN( first->x, second->x );
	    upperCoordinate = MAX( first->x, second->x );
	    return;
	}			       
	if( ( subnet->startingPoint != 0 ) && ( subnet->endingPoint != 0 ) ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->x < second->x ){
		    lowerCoordinate = first->x + subnet->startingPoint;
		    upperCoordinate = second->x - subnet->endingPoint;
		    return;
		} else {
		    upperCoordinate = first->x - subnet->startingPoint;
		    lowerCoordinate = second->x + subnet->endingPoint;
		    return;
		}
	    } else {
		if( first->x < second->x ){
		    lowerCoordinate = first->x + subnet->endingPoint;
		    upperCoordinate = second->x - subnet->startingPoint;
		    return;
		} else {
		    upperCoordinate = first->x - subnet->endingPoint;
		    lowerCoordinate = second->x + subnet->startingPoint;
		    return;
		}
	    }
	}
	if( subnet->endingPoint == 0 ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->x < second->x ){
		    lowerCoordinate = first->x + subnet->startingPoint;
		    upperCoordinate = second->x;
		    return;
		} else {
		    upperCoordinate = first->x - subnet->startingPoint;
		    lowerCoordinate = second->x;
		    return;
		}
	    } else {	
		if( first->x < second->x ){
		    lowerCoordinate = first->x;
		    upperCoordinate = second->x - subnet->startingPoint;
		    return;
		} else {
		    lowerCoordinate = second->x + subnet->startingPoint;
		    upperCoordinate = first->x;
		    return;
		}
	    }
	}
	if( subnet->startingPoint == 0 ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->x < second->x ){
		    lowerCoordinate = first->x;
		    upperCoordinate = second->x - subnet->endingPoint;
		    return;
		} else {
		    upperCoordinate = first->x;
		    lowerCoordinate = second->x + subnet->endingPoint;
		    return;
		}
	    } else {	
		if( first->x < second->x ){
		    lowerCoordinate = first->x + subnet->endingPoint;
		    upperCoordinate = second->x;
		    return;
		} else {
		    lowerCoordinate = second->x;
		    upperCoordinate = first->x - subnet->endingPoint;
		    return;
		}
	    }
	}
    } else {
	if( ( subnet->startingPoint == 0 ) && ( subnet->endingPoint == 0 ) ){
	    lowerCoordinate = MIN( first->y, second->y );
	    upperCoordinate = MAX( first->y, second->y );
	    return;
	}			       
	if( ( subnet->startingPoint != 0 ) && ( subnet->endingPoint != 0 ) ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->y < second->y ){
		    lowerCoordinate = first->y + subnet->startingPoint;
		    upperCoordinate = second->y - subnet->endingPoint;
		    return;
		} else {
		    upperCoordinate = first->y - subnet->startingPoint;
		    lowerCoordinate = second->y + subnet->endingPoint;
		    return;
		}
	    } else {
		if( first->y < second->y ){
		    lowerCoordinate = first->y + subnet->endingPoint;
		    upperCoordinate = second->y - subnet->startingPoint;
		    return;
		} else {
		    upperCoordinate = first->y - subnet->endingPoint;
		    lowerCoordinate = second->y + subnet->startingPoint;
		    return;
		}
	    }
	}
	if( subnet->endingPoint == 0 ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->y < second->y ){
		    lowerCoordinate = first->y + subnet->startingPoint;
		    upperCoordinate = second->y;
		    return;
		} else {
		    upperCoordinate = first->y - subnet->startingPoint;
		    lowerCoordinate = second->y;
		    return;
		}
	    } else {	
		if( first->y < second->y ){
		    lowerCoordinate = first->y;
		    upperCoordinate = second->y - subnet->startingPoint;
		    return;
		} else {
		    lowerCoordinate = second->y + subnet->startingPoint;
		    upperCoordinate = first->y;
		    return;
		}
	    }
	}
	if( subnet->startingPoint == 0 ){
	    if( first->number == subnet->firstNodeIndex ){
		if( first->y < second->y ){
		    lowerCoordinate = first->y;
		    upperCoordinate = second->y - subnet->endingPoint;
		    return;
		} else {
		    upperCoordinate = first->y;
		    lowerCoordinate = second->y + subnet->endingPoint;
		    return;
		}
	    } else {	
		if( first->y < second->y ){
		    lowerCoordinate = first->y + subnet->endingPoint;
		    upperCoordinate = second->y;
		    return;
		} else {
		    lowerCoordinate = second->y;
		    upperCoordinate = first->y - subnet->endingPoint;
		    return;
		}
	    }
	}
    }
}
