#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"

extern int		debug ; 
extern FILE		*dump ;
extern int		numPolyPoints ;
extern int 		numMagicPoints ;
extern struct point	magicPoints[MAX_NUM_POINTS];
extern struct octPoint  polyPoints[MAX_NUM_POINTS];
extern int		orientation;




int findChannel(presentTerm)
struct terminal		*presentTerm;
{
    struct channel 	*closestChannel[5];
    struct channel	*tempChan, *getGRChannel();
    octId		id;
    octId		getCDchannelId();
    int 		i, I;
    int			closestMagicPoint;
    int			closestPolyPoint;
    int			counter;
    int			distance, distance1;
    int			indexSeparation, smallestIndexSeparation;
    int			smallestDistance;
    int			xDistanceL, yDistanceL;
    int			xDistanceU, yDistanceU;
    int			direction;


    if(debug) fprintf( dump, "entering findChannel\n");

    /* routine to find the channel for a terminal that is not coincident
       with a magic point   */

    smallestDistance = MAX_INT;
    counter = 0;
    for( i = 0 ; i < 4 ; i++ )
	presentTerm->direction[i].distance = -1;
    for( i = 0 ; i < numPolyPoints ; i++ ){
	if( polyPoints[i].x == polyPoints[i+1].x ){
	    /* check if the terminal can be projected on the side */

	    if( ( presentTerm->center->y >= MIN( polyPoints[i].y, polyPoints[i+1].y) ) &&
		( presentTerm->center->y <= MAX( polyPoints[i].y, polyPoints[i+1].y) ) ){
		xDistanceL = presentTerm->lowerLeft->x - polyPoints[i].x; 
		xDistanceU = presentTerm->upperRight->x - polyPoints[i].x;
		distance = MIN( ABS( xDistanceL ) , ABS( xDistanceU ));
		smallestDistance = MIN( smallestDistance, distance );
		if( xDistanceL >= 0 ){
		    direction = WEST;
		} else {
		    direction = EAST;
		}
		if( checkProjection( presentTerm, distance, direction ) != OK ){
		    continue;
		}
	    } else {
		continue;
	    }
	} else {
	    /* check if the terminal can be projected on the side */

	    if( ( presentTerm->center->x >= MIN( polyPoints[i].x, polyPoints[i+1].x) ) &&
		( presentTerm->center->x <= MAX( polyPoints[i].x, polyPoints[i+1].x) ) ){

		yDistanceL = presentTerm->lowerLeft->y - polyPoints[i].y; 
		yDistanceU = presentTerm->upperRight->y - polyPoints[i].y;

		distance = MIN( ABS( yDistanceL ) , ABS( yDistanceU ));
		smallestDistance = MIN( smallestDistance, distance );
		if( yDistanceL >= 0 ){
		    direction = SOUTH;
		} else {
		    direction = NORTH;
		}
		if( checkProjection( presentTerm, distance, direction ) != OK ){
		    continue;
		}
	    } else {
		continue;
	    }
	}
	closestPolyPoint = i;
	closestMagicPoint = -1;
	smallestIndexSeparation = MAX_INT;
	for( I = 0 ; I < numMagicPoints ; I++ ){
	    indexSeparation = closestPolyPoint - magicPoints[I].index;
	    if( indexSeparation  < smallestIndexSeparation && indexSeparation >= 0 ){
		smallestIndexSeparation = indexSeparation;
		closestMagicPoint = I;
	    }
	}
	if( closestMagicPoint < 0 ){
	    closestPolyPoint += numPolyPoints - 1 ;
	    for( I = 0 ; I < numMagicPoints ; I++ ){
		indexSeparation = closestPolyPoint - magicPoints[I].index;
		if( indexSeparation  < smallestIndexSeparation && indexSeparation >= 0 ){
		    smallestIndexSeparation = indexSeparation;
		    closestMagicPoint = I;
		}
	    }
	}
	id = getCDchannelId(closestMagicPoint);
	closestChannel[counter] = getGRChannel(id, presentTerm, direction);
	if( closestChannel[counter] == NIL(struct channel) ) continue;
	/* if the distance is 0 and the channel is not an external one we're done.
	    otherwise store the channel and keep going...  */
	if( distance == 0 && closestChannel[counter]->info->outerChannel == FALSE ){
	    if( attachTerm(presentTerm, closestChannel[counter]) == OK ){
		return(OK); 
	    }
	} else {
	    if( counter++ == 4 ){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "Terminal %s < %d, %d > of instance %s\n" , 
			    presentTerm->termName, 
			    presentTerm->center->x, presentTerm->center->y, 
			    presentTerm->instInfo->name);
		fprintf( stderr, "More than 4 channel found for the projection \n");
		myExit( -1, __FILE__, __LINE__);
	    }
	}
    }
    if( counter == 0 ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Terminal %s < %d, %d > of instance %s\n" , 
		    presentTerm->termName, 
		    presentTerm->center->x, presentTerm->center->y, 
		    presentTerm->instInfo->name);
	fprintf( stderr, "Two cases are possible:\n");
	fprintf( stderr, "1) The terminal is inside the instance and there are no rectilinear paths\n");
	fprintf( stderr, "   from its location to any of the sides of the instance.\n");
	fprintf( stderr, "2) The terminal is on the boundary of the cell but there is not enought\n");
	fprintf( stderr, "   room between it and another terminal assigned to the same channel\n"); 
	fprintf( stderr, "To convince yourself, take a look at the interface of the master of cell %s\n", 
		presentTerm->instInfo->name);
	myExit( -1, __FILE__, __LINE__ );
    }
    if( closestChannel[0] == NIL(struct channel) ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "No channel found for the terminal %s < %d, %d > of instance %s\n" , 
		    presentTerm->termName, 
		    presentTerm->center->x, presentTerm->center->y, 
		    presentTerm->instInfo->name);
	myExit( -1, __FILE__, __LINE__);
    }
		    
    /*
    *     we have found the candidate channels    
    *     now we complete the information of term
    */

    /* 	Sort the channels to find the most interesting one     */


    do{
	tempChan = closestChannel[0];
	for (  I = 1 ; I < counter ; I++ ){
	    if(( distance = dist( presentTerm, closestChannel[I-1] ) ) > 
	       ( distance1 = dist( presentTerm, closestChannel[I] ) ) ){
		tempChan = closestChannel[I-1];
		closestChannel[I-1] = closestChannel[I];
		closestChannel[I] = tempChan;
	    } else if ( distance1 == distance ){
		if( ( closestChannel[I-1]->info->outerChannel == TRUE )&&
		    ( closestChannel[I]->info->outerChannel == FALSE ) ){
		    tempChan = closestChannel[I-1];
		    closestChannel[I-1] = closestChannel[I];
		    closestChannel[I] = tempChan;
		}
	    }
	}
    } while( tempChan != closestChannel[0] );
    /* first look for a channel which is suitable and not external */
    for( I = 0 ; I < counter ; I++ ){
	if( closestChannel[I]->info->outerChannel == TRUE ) continue;
	if( attachTerm(presentTerm, closestChannel[I]) == KO ) continue;
	if( smallestDistance > 0 ) updateProtFrame(closestChannel[I], presentTerm);
	return(OK);
    }
    /* if fails try again with all the other channels */
    for( I = 0 ; I < counter ; I++ ){
	if( closestChannel[I]->info->outerChannel == FALSE ) continue;
	if( attachTerm(presentTerm, closestChannel[I]) == KO ) continue;
	if( smallestDistance > 0 ) updateProtFrame(closestChannel[I], presentTerm);
	return(OK);
    }
    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
    fprintf( stderr, "Terminal %s < %d, %d > of instance %s\n" , 
		presentTerm->termName, 
		presentTerm->center->x, presentTerm->center->y, 
		presentTerm->instInfo->name);
    fprintf( stderr, "The terminal cannot be projected on any of the sides of the instance \n");
    fprintf( stderr, "To convince yourself, take a look at the interface of the master of cell %s\n", 
	    presentTerm->instInfo->name);
    myExit( -1, __FILE__, __LINE__ );
    /*NOTREACHED*/
}




octId getCDchannelId(closestMagicPoint)
int	closestMagicPoint;
{
    int			channelFound = FALSE;
    struct point 	firstMagicPoint, secondMagicPoint;
    octId		channelId;

    firstMagicPoint = magicPoints[closestMagicPoint];

    if( closestMagicPoint == 0 ){
	if( orientation == 1 ){
	    secondMagicPoint = magicPoints[1];
	} else {
	    secondMagicPoint = magicPoints[numMagicPoints-1];
	}
    } else if( closestMagicPoint == numMagicPoints - 1){
	if( orientation == 1 ){
	    secondMagicPoint= magicPoints[0];
	} else {
	    secondMagicPoint = magicPoints[closestMagicPoint-1];
	}
    } else {
	if( orientation == 1 ){
	    secondMagicPoint = magicPoints[closestMagicPoint+1];
	} else {
	    secondMagicPoint = magicPoints[closestMagicPoint-1];
	}
    }
    channelFound = FALSE;
    if( firstMagicPoint.firstChannel == secondMagicPoint.secondChannel ){
	channelFound = TRUE;
	channelId = firstMagicPoint.firstChannel;
    } else if( firstMagicPoint.secondChannel == secondMagicPoint.firstChannel ){
	channelFound = TRUE;
	channelId = firstMagicPoint.secondChannel;
    }
    if( channelFound == FALSE ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "No channel found \n");
	myExit( -1, __FILE__, __LINE__);
    } 
    return(channelId);
}
