#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern FILE		*dump ;
extern int		numNets ;
extern int		numConnectedTerms ;
extern octObject	Chip;
int			numNetsAdded;



struct net *attachTermToNet(ptoTerm, firstNet, netFound)
struct terminal *ptoTerm;
struct net	*firstNet;
int		*netFound;
{
    struct net 		*getTheNet();
    struct terminal	*presentTerm;
    octObject		ActualTerm;
    octObject		FormalTerm;
    octGenerator	genFterm;

    *netFound = FALSE;
    for( presentTerm = ptoTerm; presentTerm != NIL(struct terminal);
	 presentTerm = presentTerm->equivTerm){
	ActualTerm.objectId = presentTerm->actualTermId;
	if( octGetById(&ActualTerm) != OCT_OK ){
	    if(debug) fprintf( dump, "ActualTerm %s \n", presentTerm->termName);
	    octError("ActualTerm Id is corrupted");
	    myExit( -1, __FILE__, __LINE__);
	}
	firstNet = getTheNet( ptoTerm, firstNet, &ActualTerm, netFound );
    }

	/*  No net found attached to the terminal. 
	    Check whether the terminal is a formal terminal for
	    the chip
	*/
    if( *netFound == FALSE ){
	for( presentTerm = ptoTerm; presentTerm != NIL(struct terminal);
	     presentTerm = presentTerm->equivTerm){
	    ActualTerm.objectId = presentTerm->actualTermId;
	    if( octGetById(&ActualTerm) != OCT_OK ){
		if(debug) fprintf( dump, "ActualTerm %s \n", presentTerm->termName);
		octError("ActualTerm Id is corrupted");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if( octInitGenContainers( &ActualTerm, OCT_TERM_MASK, &genFterm ) != OCT_OK ){
		octError("generating formalTerminal attached to terminal");
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "Terminal   %s \n", presentTerm->name );
		myExit( -1, __FILE__, __LINE__);
	    }
	    while( octGenerate( &genFterm, &FormalTerm ) == OCT_OK ){
		if( octGetByName( &Chip, &FormalTerm )== OCT_OK ){
		    /* Check if there is a net attached to the formal terminal
		       of the chip
		    */
		}
		firstNet = getTheNet( ptoTerm, firstNet, &FormalTerm, netFound );
	    }
	}
    }
    if( *netFound == TRUE ){
	numConnectedTerms++;
	if(debug) fprintf( dump, "New connected terminal %s %d\n" , ptoTerm->name, numConnectedTerms);
    } else {
	if(debug) fprintf( dump, "The terminal %s has no nets connected to it \n",
				 ptoTerm->name);
    }
    return(firstNet) ;
}



struct net *getTheNet( ptoTerm, firstNet, Term, netFound)
struct octObject *Term;
struct terminal  *ptoTerm;
struct net	 *firstNet;
int		 *netFound;
{
    int			netAlreadyInList;
    int			termFound;
    struct net		*presentNet;
    struct net		*newNet;
    struct termList	*newTermOnNet, *termOnNet;
    struct netList	*newNetOnList;
    octObject		Net;
    octGenerator	genNet;
    char		buffer[512];

    if( octInitGenContainers( Term, OCT_NET_MASK, &genNet ) != OCT_OK ){
	octError("generating nets attached to terminal");
	fprintf( stderr, "MOSAICO_GR_ERROR:|n");
	fprintf( stderr, "Terminal   %s \n", Term->contents.term.name );
	myExit( -1, __FILE__, __LINE__);
    }
    while( octGenerate( &genNet, &Net ) == OCT_OK ){
	*netFound = TRUE;
	netAlreadyInList = FALSE;
	if( strCmp( Net.contents.net.name, NIL(char) ) != 0 ){
	    for ( presentNet = firstNet ;
		  presentNet != NIL(struct net) ; 
		  presentNet = presentNet->ptoNextNet ){
		if ( strCmp( Net.contents.net.name , presentNet->name ) == 0 ){
		    netAlreadyInList = TRUE ;
		    break;
		}
	    }
	} else {
	    (void) sprintf( buffer,"mosaicoNet_%d_%s\n", numNets + 1,
			octFormatId(Net.objectId));
	    Net.contents.net.name = util_strsav(buffer);
	    (void) octModify(&Net);
	    numNetsAdded++;
	}
	/*
	*	Attach the the terminal to the net
	*/
	if( netAlreadyInList == TRUE ){
	    /*
	    *	Check whether the term was already connected to the net
	    */
	    termFound = FALSE;
	    for( termOnNet = presentNet->firstTermOnNet;
		 termOnNet != NIL(struct termList);
		 termOnNet = termOnNet->ptoNextTerm ){
		if( termOnNet->ptoTerm == ptoTerm ){
		    termFound = TRUE;
		    break;
		}
	    }
	    if( termFound == TRUE ) continue;
	    newTermOnNet = ALLOC( struct termList , 1);
	    newTermOnNet->ptoTerm = ptoTerm;
	    newTermOnNet->ptoNextTerm = presentNet->firstTermOnNet;
	    presentNet->firstTermOnNet = newTermOnNet;
	    newNetOnList = ALLOC( struct netList , 1 );
	    newNetOnList->name = util_strsav(Net.contents.net.name);
	    newNetOnList->id = Net.objectId;
	    newNetOnList->ptoNextNetInList = ptoTerm->firstNetInList;
	    ptoTerm->firstNetInList = newNetOnList;
	} else {
	    newTermOnNet = ALLOC( struct termList , 1);
	    newTermOnNet->ptoTerm = ptoTerm;
	    newTermOnNet->ptoNextTerm = NIL(struct termList);
	    newNet = ALLOC( struct net , 1 );
	    newNet->name = util_strsav(Net.contents.net.name);
	    newNet->id = Net.objectId;
	    newNet->number = ++numNets ;
	    newNet->firstTermOnNet = newTermOnNet;
	    newNet->ptoNextNet = firstNet;
	    firstNet = newNet ;
	    newNetOnList = ALLOC( struct netList , 1 );
	    newNetOnList->name = util_strsav(Net.contents.net.name);
	    newNetOnList->id = Net.objectId;
	    newNetOnList->ptoNextNetInList = ptoTerm->firstNetInList;
	    ptoTerm->firstNetInList = newNetOnList;
	}
    }
    return(firstNet);
}
