#include "cds.h"
#include "utility.h"
#include "vov.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"
extern int		debug ; 
extern FILE		*dump ;
extern int 		levelCounter;
extern int 		numPoints ;
int32			numPolyPoints;
int			numMagicPoints;
struct octPoint		polyPoints[MAX_NUM_POINTS];
struct point		magicPoints[MAX_NUM_POINTS];
struct instList		*instanceList;
int			magicPointsOrientation, polyPointsOrientation, orientation;
fa_geometry		termGeo, geoOnPoly, geoOnMet1, geoOnMet2;
octCoord		maxSpacing, spacing_M1, spacing_M2, spacing_PO;



/*
*     generateCell(ptrBag)
*
*     Routine to generate the informations of the cell contained in the leaf bag
*     of the channel definition Graph
*
*/

generateCell(ptrBag)
octObject *ptrBag;
{
    octGenerator	genInstance;
    octGenerator	genBag;
    octGenerator 	genPolygon;
    octGenerator 	genBox;
    octObject		MasterInterface;

    octObject		Instance;
    octObject		Bag;
    octObject		Box;
    octObject		joinedBag;
    octObject		Polygon;
    octObject		Layer;

    octStatus		status, layerStatus;

    struct octBox	instanceBB;
    struct octTransform	cellTransform;
    struct instList	*tmpInstList;

    tr_stack		*stack;
    char		*cellName, *tapGetDirectory();
    int			instanceFound;
    int			polygonFound;
    int			t_x, t_y, type;
    int			bagFlag;
    int			firstBoxFound;
    int			i, j, k;
    int 		nextIndex, index;
    int 		lowestLine, presentLine;
    /*
    *   Now we look for the Instance of the cell that is contained in the
    *   polygon defining the channel
    */

    if(debug) fprintf( dump, "Entering generateCell\n");
    polyPointsOrientation = UNDEFINED ;
    MasterInterface.type = OCT_FACET;
    MasterInterface.contents.facet.facet = "interface";
    MasterInterface.contents.facet.mode = "r";
    stack = tr_create();
    if(octInitGenContents(ptrBag, OCT_INSTANCE_MASK, &genInstance) != OCT_OK){
	octError("generating the Instance generator \n");
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Occurred at level %d \n", levelCounter);
	myExit( -1, __FILE__, __LINE__);
    }
    Layer.type = OCT_LAYER;
    Layer.contents.layer.name = "PLACE";

    instanceFound = FALSE;
    while((status = octGenerate(&genInstance,&Instance)) == OCT_OK){
	if(instanceFound == TRUE){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Occurred at level %d \n", levelCounter);
	    fprintf( stderr, "More than one instance found in a leaf bag!!!\n");
	    fprintf( stderr, "I do not know what to do therefore I quit\n");
	    myExit( -1, __FILE__, __LINE__);
	}
	instanceFound = TRUE;
	MasterInterface.contents.facet.cell = Instance.contents.instance.master;
	MasterInterface.contents.facet.view = Instance.contents.instance.view;
	MasterInterface.contents.facet.version = Instance.contents.instance.version;
	cellName = util_strsav(Instance.contents.instance.name);
	cellTransform = Instance.contents.instance.transform;
	if(debug) fprintf( dump, "cellName %s\n" , cellName);

	VOVinputFacet( &MasterInterface );
	if(octOpenFacet(&MasterInterface) < OCT_OK){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("opening master \n");
	    myExit( -1, __FILE__, __LINE__);
	}

	/*
	****  set up the stack for the transformation of terminals  ****
	*/
	tr_identity(stack);

	type = (int) cellTransform.transformType;
	t_x = cellTransform.translation.x;
	t_y = cellTransform.translation.y;

	tr_translate(stack, (int32) t_x, (int32) t_y, 0);

	if(type != (int) OCT_FULL_TRANSFORM){
	    tr_do_op(stack, type, 0);
	} else {
	    double (*matrix)[2] = cellTransform.generalTransform;
	    tr_rotate_matrix(stack, matrix, 0);
	}
	(void) octBB( &Instance, &instanceBB);
	Layer.contents.layer.name = "MET1";
	fa_init(&geoOnMet1);
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Error getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( layerStatus == OCT_OK ){
	    harAddContentsToFang(&Layer, &geoOnMet1, HAR_NO_TRANSFORM);
	    /* use tap to get the design rule correct spacing */
	    spacing_M1 = tapGetMinSpacing( &Layer, &Layer );	
	    maxSpacing = spacing_M1;
	}
	Layer.contents.layer.name = "MET2";
	fa_init(&geoOnMet2);
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( layerStatus == OCT_OK ){
	    harAddContentsToFang(&Layer, &geoOnMet2, HAR_NO_TRANSFORM);
	    spacing_M2 = tapGetMinSpacing( &Layer, &Layer );	
	    maxSpacing = MAX( maxSpacing, spacing_M2 );
	}
	Layer.contents.layer.name = "POLY";
	fa_init(&geoOnPoly);
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( layerStatus == OCT_OK ){
	    harAddContentsToFang(&Layer, &geoOnPoly, HAR_NO_TRANSFORM);
	    spacing_PO = tapGetMinSpacing( &Layer, &Layer );	
	    maxSpacing = MAX( maxSpacing, spacing_PO );
	}
	/*
	    Add geometry on PDIF and NDIF to avoid creating transistors while making
	    connections.
	*/
	Layer.contents.layer.name = "NDIF";
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( layerStatus == OCT_OK ){
	    harAddContentsToFang(&Layer, &geoOnPoly, HAR_NO_TRANSFORM);
	}
	Layer.contents.layer.name = "PDIF";
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Error getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( layerStatus == OCT_OK ){
	    harAddContentsToFang(&Layer, &geoOnPoly, HAR_NO_TRANSFORM);
	}
	Layer.contents.layer.name = "PLACE";
	if( (layerStatus = octGetByName(&MasterInterface, &Layer) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Error getting the layer \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( debug ){
	    fprintf( dump, "The technology used is %s\n", tapGetDirectory(&Layer) );
	    fprintf( dump,"Spacing on MET1 %d\n", spacing_M1 );
	    fprintf( dump,"Spacing on MET2 %d\n", spacing_M2 );
	    fprintf( dump,"Spacing on POLY %d\n", spacing_PO );
	}
	/*
	*
	*	find the polygon that bounds the master
	*
	*/
	if( layerStatus == OCT_NOT_FOUND ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    fprintf( stderr, "Layer PLACE not found \n");
	}
	if( layerStatus == OCT_OK ){
	    if(octInitGenContents(&Layer, OCT_POLYGON_MASK, &genPolygon) != OCT_OK){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "In cell %s \n", cellName);
		octError("generating the generator for polygons");
		myExit( -1, __FILE__, __LINE__);
	    }
	    polygonFound = FALSE;
	     
	    /*
	    *   We look first for the polygon that contains all the informations
	    *   related to the protection frame of the master
	    */
	    while((status = octGenerate(&genPolygon,&Polygon)) == OCT_OK){
		if(polygonFound == TRUE){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "In cell %s \n", cellName);
		    fprintf( stderr, "More than one polygon found in the master!!!\n");
		    fprintf( stderr, "I do not know what to do therefore I quit\n");
		    myExit( -1, __FILE__, __LINE__);
		}
		numPolyPoints = 0 ;
		polygonFound = TRUE;
		(void) octGetPoints(&Polygon, &numPolyPoints, (struct octPoint *) 0);
		if( numPolyPoints >= MAX_NUM_POINTS ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "The number of poly points found %d exceeds the maximum number allowed %d\n",
			numPolyPoints, MAX_NUM_POINTS);
		    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_POINTS in file cd2tw.h\n");
		    myExit( -1, __FILE__, __LINE__);
		}
		(void) octGetPoints(&Polygon, &numPolyPoints, polyPoints);
	    }
	    /*
	    *   Then we consider the bounding box as a polygon with four corners 
	    */
	    if( status == OCT_GEN_DONE && polygonFound == FALSE ){
		if(octInitGenContents(&Layer, OCT_BOX_MASK, &genBox) != OCT_OK){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "In cell %s \n", cellName);
		    octError("generating the generator for bounding box");
		    myExit( -1, __FILE__, __LINE__);
		}
		firstBoxFound = FALSE;
		while( ( status = octGenerate(&genBox, &Box)) == OCT_OK  && firstBoxFound == FALSE ){
		    firstBoxFound = TRUE;
		    numPolyPoints = 4;
		    polyPoints[0].x = Box.contents.box.lowerLeft.x;
		    polyPoints[0].y = Box.contents.box.lowerLeft.y;
		    polyPoints[1].x = Box.contents.box.upperRight.x;
		    polyPoints[1].y = Box.contents.box.lowerLeft.y;
		    polyPoints[2].x = Box.contents.box.upperRight.x;
		    polyPoints[2].y = Box.contents.box.upperRight.y;
		    polyPoints[3].x = Box.contents.box.lowerLeft.x;
		    polyPoints[3].y = Box.contents.box.upperRight.y;
		}
	    }
	} else {
	    numPolyPoints = 4;
	    polyPoints[0].x = instanceBB.lowerLeft.x;
	    polyPoints[0].y = instanceBB.lowerLeft.y;
	    polyPoints[1].x = instanceBB.upperRight.x;
	    polyPoints[1].y = instanceBB.lowerLeft.y;
	    polyPoints[2].x = instanceBB.upperRight.x;
	    polyPoints[2].y = instanceBB.upperRight.y;
	    polyPoints[3].x = instanceBB.lowerLeft.x;
	    polyPoints[3].y = instanceBB.upperRight.y;
	}
	polyPoints[numPolyPoints] = polyPoints[0];
	if( layerStatus == OCT_OK ){
	    for( i = 0 ; i <= numPolyPoints ; i++ ){
		tr_transform(stack , &polyPoints[i].x , &polyPoints[i].y );
	    }
	}
	    /* determine the orientation of polyPoints  */
	lowestLine = MAX_INT;
	for( k = 0 ; k < numPolyPoints ; k++ ) {
	    presentLine = polyPoints[k].x + polyPoints[k].y;
	    if( lowestLine > presentLine ){
		lowestLine = presentLine;
		index = k;
	    }
	}
	nextIndex = (index + 1) % numPolyPoints ;
	if( polyPoints[index].x == polyPoints[nextIndex].x ){
	    polyPointsOrientation = CWISE;
	} else if( polyPoints[index].y == polyPoints[nextIndex].y){
	    polyPointsOrientation = CCWISE;
	}
	
	/*
	*	find channel corners in the instance 
	*/
	numMagicPoints = 0;
	numMagicPoints = getMagicPoints( &Instance );
	orientation = polyPointsOrientation * magicPointsOrientation ;


	if(debug){
	    fprintf( dump, "Points of the polygon bounding the master\n");
	    if( polyPointsOrientation == CWISE )
		fprintf( dump, "polyPointsOrientation = CWISE \n");
	    else if( polyPointsOrientation == CCWISE )
		fprintf( dump, "polyPointsOrientation = CCWISE \n");
	    else 
		fprintf( dump, "polyPointsOrientation = UNDEFINED \n");
	    for( i = 0 ; i < numPolyPoints ; i++ ){
		fprintf( dump, "%d point, x = %d , y = %d \n", 
			i,
			polyPoints[i].x,
			polyPoints[i].y);
	    }
	    fprintf( dump, "Magic points \n");
	    if( magicPointsOrientation == CWISE )
		fprintf( dump, "magicPointsOrientation = CWISE \n");
	    else if( magicPointsOrientation == CCWISE )
		fprintf( dump, "magicPointsOrientation = CCWISE \n");
	    else 
		fprintf( dump, "magicPointsOrientation = UNDEFINED \n");

	    for( j = 0 ; j < numMagicPoints ; j++ ){
		fprintf( dump, "%d magicPoint (%d), x = %d , y = %d \n", 
			j, magicPoints[j].index,
			magicPoints[j].ptoPoint->x,
			magicPoints[j].ptoPoint->y);
	    }
	}
	if( instanceList == NIL(struct instList) ){
	    instanceList = ALLOC( struct instList, 1 );
	    instanceList->nextInList = NIL(struct instList);
	} else {
	    tmpInstList = ALLOC( struct instList, 1 );
	    tmpInstList->nextInList = instanceList;
	    instanceList = tmpInstList;
	}
	instanceList->instInfo = ALLOC( struct instInfo, 1);
	instanceList->instInfo->instanceId = Instance.objectId;
	instanceList->instInfo->name = util_strsav(cellName);
	instanceList->instInfo->transform = Instance.contents.instance.transform.transformType;
	instanceList->instInfo->bottom = instanceBB.lowerLeft.y; 
	instanceList->instInfo->top = instanceBB.upperRight.y; 
	instanceList->instInfo->right = instanceBB.upperRight.x; 
	instanceList->instInfo->left = instanceBB.lowerLeft.x; 
	/*
	 **** find all the terminals in Master ****
	 */

	/**** First generate terminals with equivalent pins on the cell ****/

	if(debug) fprintf( dump, "Start to generate the terminals \n");
	joinedBag.type = OCT_BAG;
	joinedBag.contents.bag.name = "JOINEDTERMS";
	if(octGetByName(&MasterInterface , &joinedBag) == OCT_NOT_FOUND){
	    if(debug) fprintf( dump, "No joined terminals on this cell\n");
	} else {
	    bagFlag = TRUE;
	    if(debug) fprintf( dump, "Start with joined terminals \n");
	    if(octInitGenContents(&joinedBag,OCT_BAG_MASK,&genBag) != OCT_OK){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "In cell %s \n", cellName);
		octError("Generating the generator for joined terms");
		myExit( -1, __FILE__, __LINE__);
	    } else {
		while( octGenerate(&genBag,&Bag) == OCT_OK ){
		    generateTerms(&Bag, stack, &bagFlag , &Instance, instanceList->instInfo);
		}
	    }
	}
	/**** repeat almost the same procedure for terminals without equivalents ****/
	bagFlag = FALSE;
	if(debug) fprintf( dump, "Then with single terminals \n");
	generateTerms(&MasterInterface, stack, &bagFlag , &Instance, instanceList->instInfo );
    }

    if((status == OCT_GEN_DONE) && (instanceFound == FALSE)){
	struct octPoint points[MAX_NUM_POINTS];
	struct edge     edges[MAX_NUM_EDGES];
	Bag.type = OCT_BAG;
	Bag.contents.bag.name = "CDGraph";
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "No instance was found inside the polygon:\n");
	if (octGetByName(ptrBag, &Bag) != OCT_OK) {
	    octError("can not find the CDGraph bag");
	    exit(-1);
	}
	numPoints = getPolygon(&Bag, points, edges);
	for( i = 0 ; i < numPoints ; i ++ ){
	    fprintf( stderr, "%d point, x %d , y %d \n",
		    i, points[i].x , points[i].y );
	}
	myExit( -1, __FILE__, __LINE__);
    }
    if(debug) fprintf( dump, "Leaving createCell \n");
    fa_free_geo(&geoOnMet1);
    fa_free_geo(&geoOnMet2);
    fa_free_geo(&geoOnPoly);
    fa_free_geo(&termGeo);
}

