#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int	debug ; 
extern FILE	*dump ;
extern int 	levelCounter;
extern int	numChannels ;
extern int 	numNodes ;
extern int 	numPoints ;
int		channelIndex[MAX_NUM_POINTS];
extern int	outerChan;
struct channel	*arrayOfChannels[MAX_NUM_CHANNELS];
struct node	*arrayOfNodes[MAX_NUM_NODES];


/*     createChannels(ptrBag)
*
*      Routine to generate channel informations fron the leaf bag of channel definition
*      graph
*
*/

createChannels(ptrBag)
octObject	*ptrBag;
{
    octStatus		status;
    octObject		Bag;
    octGenerator 	genBag;
    struct octPoint	points[MAX_NUM_POINTS];
    struct edge		edges[MAX_NUM_EDGES];
    struct channelList	*newChannelL;

    int 		polygonFound;	
    int			channelFound;
    int			nodeFound;
    int			i ,j;
    int			nodeIndex[MAX_NUM_POINTS];

    if(debug) fprintf( dump, "Entering createChannels \n");
    if(octInitGenContents(ptrBag, OCT_BAG_MASK, &genBag) != OCT_OK){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Level %d of the Channel Def. tree \n", levelCounter);
	octError("generating the generator for Bag containing the polygon \n");
	myExit( -1, __FILE__, __LINE__);
    }
    polygonFound = FALSE;
     
    /*
    *   We look first for the polygon that contains all the informations
    *   related to the structure of the channel
    */
    while((status = octGenerate(&genBag,&Bag)) == OCT_OK){
	if(polygonFound == TRUE){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Level %d of the Channel Def. tree \n", levelCounter);
	    fprintf( stderr, "More than one polygon found in a leaf bag!!!\n");
	    fprintf( stderr, "I do not know what to do therefore I quit\n");
	    myExit( -1, __FILE__, __LINE__);
	}
	polygonFound = TRUE;
	numPoints = getPolygon(&Bag, points , edges);
	if( numPoints >= MAX_NUM_POINTS ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "The number of poly points found %d exceeds the maximum number allowed %d\n", numPoints, MAX_NUM_POINTS );
	    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_POINTS in file cd2tw.h\n");
	    myExit( -1, __FILE__, __LINE__);
	}

	if(debug){
	    fprintf( dump, "polygon describing the channels\n");
	    for( i = 0 ; i < numPoints ; i++ ){
		fprintf( dump, "%d point , x %d , y %d \n",
			i,
			points[i].x,
			points[i].y);
	    }
	}

	for( i = 0 ; i < numPoints ; i++ ){
	    nodeFound = FALSE ;
	    for( j = 0 ; j < numNodes ; j++){
		if((arrayOfNodes[j]->x == points[i].x) && 
		    (arrayOfNodes[j]->y == points[i].y)){
		    /* the node was found */
		    nodeIndex[i] = j;
		    nodeFound = TRUE;
		    break;
		}
	    }
	    /* the node just found is new */
	    if(nodeFound == FALSE){
		arrayOfNodes[numNodes] = ALLOC( struct node, 1 );
		arrayOfNodes[numNodes]->number = numNodes + 1;
		arrayOfNodes[numNodes]->x = points[i].x;
		arrayOfNodes[numNodes]->y = points[i].y;
		arrayOfNodes[numNodes]->channelNumber = edges[i].channelNumber;
		arrayOfNodes[numNodes]->width = edges[i].width;
		arrayOfNodes[numNodes]->ptoFirstStartingChannel = NIL(struct channelList);
		arrayOfNodes[numNodes]->ptoFirstEndingChannel = NIL(struct channelList);
		nodeIndex[i] = numNodes;
		if( ++numNodes >= MAX_NUM_NODES ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "The number of nodes %d exceeds the maximum number allowed %d\n", numNodes, MAX_NUM_NODES );
		    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_NODES in file cd2tw.h\n");
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	}
	if(debug){ 
	    fprintf( dump, "Got the points( %d ), now build channel\n",
			    numPoints);
	    fprintf( dump, "%d nodes in the graph determined so far\n",
			    numNodes);
	}
	nodeIndex[numPoints] = nodeIndex[0];
	for( i = 0 ; i < numPoints ; i++ ){
	    arrayOfChannels[numChannels]->number = numChannels + 1;
	    arrayOfChannels[numChannels]->firstNode = arrayOfNodes[nodeIndex[i]];
	    arrayOfChannels[numChannels]->secondNode = arrayOfNodes[nodeIndex[i+1]];

	    /* Now we check wheather the channel has already been found
	       This is done comparing the coordinates of the present
	       pair of pins with those of the channels found so far */
		
	    channelFound = FALSE;
	    for( j = 0 ; j < numChannels && channelFound == FALSE ; j++ ){
		if((arrayOfChannels[j]->firstNode->x ==
		    arrayOfChannels[numChannels]->secondNode->x) &&
		    (arrayOfChannels[j]->firstNode->y ==
		    arrayOfChannels[numChannels]->secondNode->y) &&
		    (arrayOfChannels[j]->secondNode->x ==
		    arrayOfChannels[numChannels]->firstNode->x) &&
		    (arrayOfChannels[j]->secondNode->y ==
		    arrayOfChannels[numChannels]->firstNode->y)){
		    channelFound = TRUE; 
		    channelIndex[i] = j;
		    if( debug ){
			fprintf( dump, "old channel\n");
			channelInfo( arrayOfChannels[j] );
		    }
		}
	    }
	    if(channelFound == FALSE){
		arrayOfChannels[numChannels]->info = ALLOC(struct channelInfo, 1);
		arrayOfChannels[numChannels]->info->channelNumber = edges[i].channelNumber ;
		arrayOfChannels[numChannels]->info->id = edges[i].id ;
		arrayOfChannels[numChannels]->info->capacity = UNDEFINED ;
		arrayOfChannels[numChannels]->info->width = edges[i].width ;
		if( edges[i].outerEdge == TRUE ){
		    arrayOfChannels[numChannels]->info->outerChannel = TRUE;
		    arrayOfChannels[numChannels]->info->used = NO ;
		} else {
		    arrayOfChannels[numChannels]->info->outerChannel = FALSE;
		    arrayOfChannels[numChannels]->info->used = YES ;
		}
		arrayOfChannels[numChannels]->info->numberOfTerms = 0;
	   	if( arrayOfChannels[numChannels]->firstNode->x ==
		    arrayOfChannels[numChannels]->secondNode->x ){
		    arrayOfChannels[numChannels]->info->type = VERTICAL ;
		    arrayOfChannels[numChannels]->info->centerLine = 
			arrayOfChannels[numChannels]->firstNode->x;
		    arrayOfChannels[numChannels]->info->length =
		    ABS( arrayOfChannels[numChannels]->firstNode->y -
			arrayOfChannels[numChannels]->secondNode->y );
	   	}
	   	if( arrayOfChannels[numChannels]->firstNode->y ==
		    arrayOfChannels[numChannels]->secondNode->y ){
		    arrayOfChannels[numChannels]->info->type = HORIZONTAL ;
		    arrayOfChannels[numChannels]->info->centerLine = 
			arrayOfChannels[numChannels]->firstNode->y;
		    arrayOfChannels[numChannels]->info->length =
		    ABS( arrayOfChannels[numChannels]->firstNode->x -
			arrayOfChannels[numChannels]->secondNode->x );
		}
		if( edges[i].outerEdge == TRUE && outerChan == NO ){
		    arrayOfChannels[numChannels]->info->lengthFactor = 10 ;
		} else
		    arrayOfChannels[numChannels]->info->lengthFactor = 1;
		channelIndex[i] = numChannels;
		arrayOfChannels[numChannels]->ptoFirstSubnet = NIL(struct subnetList);
		arrayOfChannels[numChannels]->ptoFirstTerm = NIL(struct termList);
		newChannelL = ALLOC(struct channelList, 1);
		newChannelL->ptoChannel = arrayOfChannels[numChannels];
		newChannelL->ptoNextChannel =
			    arrayOfNodes[nodeIndex[i]]->ptoFirstStartingChannel;
		arrayOfNodes[nodeIndex[i]]->ptoFirstStartingChannel = newChannelL;
		newChannelL = ALLOC(struct channelList, 1);
		newChannelL->ptoChannel = arrayOfChannels[numChannels];
		newChannelL->ptoNextChannel =
			    arrayOfNodes[nodeIndex[i+1]]->ptoFirstEndingChannel;
		arrayOfNodes[nodeIndex[i+1]]->ptoFirstEndingChannel = newChannelL;
		if( debug ){
		    fprintf( dump, "new channel\n");
		    channelInfo( arrayOfChannels[numChannels] );
		}
		if( numChannels >= MAX_NUM_CHANNELS ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "The number of channels found %d exceeds the maximum number allowed %d\n", numChannels, MAX_NUM_CHANNELS );
		    fprintf( stderr, "Remedy: Increase the value of MAX_NUM_CHANNELS in file cd2tw.h\n");
		    myExit( -1, __FILE__, __LINE__);
		}
		arrayOfChannels[++numChannels] = ALLOC( struct channel, 1 );
		if(debug) fprintf( dump, "Got the channel\n");
	    }
	}
	if(debug) fprintf( dump, "Got all channels\n");
    }
	
    if((status == OCT_GEN_DONE) && (polygonFound == FALSE)){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Level %d of the Channel Def. tree \n", levelCounter);
	fprintf( stderr, "The polygon was not found !!!");
	myExit( -1, __FILE__, __LINE__);
    }
    if(debug) fprintf( dump, "Leaving createChannels \n");
}
