#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern FILE		*dump ;
extern int		numTerms;
extern struct net	*firstNet;
extern fa_geometry	geoOnPoly, geoOnMet1, geoOnMet2;




/*
*     generateTerms(container, stack, bagFlag, Instance, instInfo)
*     
*     Routine to generate the terminals given their container:
*
*     1)   Bag 		if we are dealing with joined terminals;
*     2)   Master 	if we are dealing with simple terminals;
*
*
*/
generateTerms(container, stack, bagFlag , Instance, instInfo )
octObject	*container;
octObject	*Instance;
struct instInfo	*instInfo;
tr_stack	*stack;
int		*bagFlag;
{
    octGenerator	genTerm;
    octGenerator	genBag;
    octObject		Term;
    octObject		actualTerm;
    octObject		Bag;
    octId		jointBagId;

    struct terminal	*term, *ptoFirstTerm;
    struct net		*attachTermToNet();
    int			netFound;
    int			joinedTerms;

    actualTerm.type = OCT_TERM; 
    ptoFirstTerm = NIL(struct terminal);
    if(debug) fprintf( dump, "Entering generateTerms \n");
    if(octInitGenContents(container,OCT_TERM_MASK,&genTerm) != OCT_OK){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "In cell %s \n", instInfo->name);
	octError("Generating the generator for terms");
	myExit( -1, __FILE__, __LINE__);
    } 
    if( container->type == OCT_BAG ){
	jointBagId = container->objectId;
    } else {
	jointBagId = oct_null_id;
    }
    while(octGenerate(&genTerm,&Term) == OCT_OK){
    /*
    *	in the case we are looking at single terminals:
    *	for each terminal generated chech whether it is not already contained
    *	in a bag of type joined. If it is do not generated the terminal
    *	again.
    *
    */
	if(*bagFlag == FALSE){
	    if(octInitGenContainers(&Term, OCT_BAG_MASK, &genBag) != OCT_OK){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "In cell %s \n", instInfo->name);
		octError("Initializing the generator for containers of a term");
		myExit( -1, __FILE__, __LINE__);
	    }
	    joinedTerms = FALSE;
	    while(octGenerate(&genBag, &Bag) == OCT_OK){
		if(strCmp(Bag.contents.bag.name, "JOINED") == 0){
		    if(debug)
			fprintf( dump, "Term already found when searching for joined nets \n");
		    joinedTerms = TRUE;
		    break;
		}
	    }
	    if( joinedTerms == TRUE ) continue;
	}

	/***** find the geometry that represents term *****/


	term = ALLOC(struct terminal, 1);
	term->termName = util_strsav(Term.contents.term.name);
	term->name = ALLOC(char,(strLen(instInfo->name)+strLen(Term.contents.term.name)+4));
	sprintf(term->name,"%sFIR%s", instInfo->name, Term.contents.term.name);
	term->number = ++numTerms;
	term->firstNetInList = NIL(struct netList);
	term->routeEquivTerm = FALSE;
	term->firstNode = NIL(struct node);
	term->secondNode = NIL(struct node);
	term->jointBagId = jointBagId;
	term->instInfo = instInfo;

	actualTerm.contents.term.name = Term.contents.term.name;
	if(octGetByName( Instance , &actualTerm ) != OCT_OK){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Instance %s \n", Instance->contents.instance.name );
	    fprintf( stderr, "terminal %s\n", Term.contents.term.name );
	    octError("actualTerm not found") ;
	    myExit( -1, __FILE__, __LINE__);
	}
	term->actualTermId = actualTerm.objectId;
	if( findImplementation( instInfo->name, &Term , term) == NO ) continue;
	findAbsCoor( stack, term );


	/*
	*	We have to check whether there are nets connected to this terminal.
	*   Two cases have to be considered:
	*
	*   1) the case in which we're dealing with joned terminals;
	*   2) the case in which we're dealing with single terminals.
	*
	*/
	if(*bagFlag == TRUE){					/* joined terminals */
	    term->equivTerm = ptoFirstTerm;
	    ptoFirstTerm = term;
	} else {						/* single terminals */
	    term->equivTerm = NIL(struct terminal);
	    firstNet = attachTermToNet(term, firstNet, &netFound);
	    if( netFound == TRUE ) assignTermToChannel(term);
	}
    }
    /*
    *	Now, in the case we are dealing with joined terminals we still have
    *	to find the nets that are attached to the terminal
    */
    if(*bagFlag == TRUE){
	firstNet = attachTermToNet(ptoFirstTerm, firstNet, &netFound);
	if( netFound == TRUE ) assignTermToChannel(ptoFirstTerm);
    }
    if(debug) fprintf( dump, "Leaving generateTerms \n");
}
