#include "cds.h"
#include "oh.h"
#include "tap.h"
#include "utility.h"
#include "errtrap.h"
#include "vov.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern int		silent ; 
extern char		*version();
octObject		Chip;
int 			TWgr, Ggr;
int			verbose;
int			outerChan;
FILE			*dump;
extern int		numNetsAdded;
char			*techDir;

#define TO 0
#define FROM 1
#define RUN 2


main(argc,argv)
int argc;
char *argv[];

{
    char	*openMode;
    char	*twdir;
    char	*fileName;
    char	*chipName;
    char	*inCell, *outCell;
    char	command[512];
    int		argIndex;
    int		globalRouter;
    extern	void readFromOct();
    extern char	*optarg;
    extern int	*optind;




    if( argc > 10 OR argc < 2 ){
	usage();
	exit(-1);
    }
    VOVbegin( argc, argv );
    /*
     *  default values
     */
    debug = NO;
    verbose = NO;
    globalRouter = RUN;
    silent = NO;
    TWgr = YES;
    Ggr = NO;
    outerChan = YES;
    numNetsAdded = 0;
    twdir = NIL(char);
    outCell = NIL(char);
    techDir = NIL(char);
    inCell = argv[1];
    if( inCell == NIL(char) ) {
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "The input cellname must be present \n");
	usage();
	exit(-1);
    }

    ohUnpackDefaults(&Chip, "r", "::contents");
    (void) ohUnpackFacetName(&Chip, inCell);
    chipName = util_strsav(Chip.contents.facet.cell) ;
    if( (chipName = strrchr( chipName, '/' )) != NIL(char) ) 
	chipName++;
    else
	chipName = Chip.contents.facet.cell ;



    /* Now parse the options */
    argIndex = 2;
    while( argIndex < argc ){ 
	if( strCmp(argv[argIndex] , "-d") == 0 ){
	    debug = YES;
	    verbose = YES;
	    if( argv[argIndex+1][0] == '-' || argv[argIndex+1][0] == '\0' ){
		fileName = ALLOC( char , strLen(chipName) + 10 );
		sprintf( fileName, "%s_cds.deb", chipName);
	    } else {
		fileName = argv[++argIndex];
	    }
	} else if( strCmp(argv[argIndex] , "-s") == 0 ){
	    silent = YES;
	} else if( strCmp(argv[argIndex] , "-t") == 0 ){
	    globalRouter = TO;
	} else if( strCmp(argv[argIndex] , "-f") == 0 ){
	    globalRouter = FROM;
	} else if( strCmp(argv[argIndex] , "-TW") == 0 ){
	    TWgr = YES;
	    Ggr = NO;
	/*
	} else if( strCmp(argv[argIndex] , "-G") == 0 ){
	    TWgr = NO;
	    Ggr = YES;
	    */
	} else if( strCmp(argv[argIndex] , "-v") == 0 ){
	    verbose = YES;
	    if( argv[argIndex+1][0] == '-' || argv[argIndex+1][0] == '\0' ){
		fileName = ALLOC( char , strLen(chipName) + 10 );
		sprintf( fileName, "%s_cds.deb", chipName);
	    } else {
		fileName = argv[++argIndex];
	    }
	} else if( strCmp(argv[argIndex] , "-e") == 0 ){
	    outerChan = NO;
	} else if( strCmp(argv[argIndex] , "-r") == 0 ){
	    globalRouter = RUN;
	} else if( strCmp(argv[argIndex] , "-od") == 0 ){
	    if( argv[argIndex+1][0] == '-' || argv[argIndex+1][0] == '\0' ){
		fprintf( stderr, "MOSAICO_GR WARNING \n" );
		fprintf( stderr, "the option -od requires an argument namely the output directory \n" );
		fprintf( stderr, "option ignored\n" );
	    } else {
		twdir = argv[++argIndex];
	    }
	} else if( strCmp(argv[argIndex] , "-ov") == 0 ){
	    if( argv[argIndex+1][0] == '-' || argv[argIndex+1][0] == '\0' ){
		fprintf( stderr, "MOSAICO_GR WARNING \n" );
		fprintf( stderr, "the option -ov requires an argument namely the output view \n" );
		fprintf( stderr, "option ignored\n" );
	    } else {
		if( ( outCell = strchr( argv[++argIndex], ':' ) ) == NIL(char) ){
		    outCell = ALLOC( char, strlen(argv[argIndex]) + 2); /* length + colon + \0 */
		    sprintf( outCell, ":%s", argv[argIndex] );
		} else {
		    outCell =  argv[argIndex];
		}
	    }
	} else if( strCmp(argv[argIndex] , "-T") == 0 ){
	    techDir =  argv[++argIndex];
	    (void) tapRootDirectory( techDir );
	} else {
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Unparsable keyword --%s-- found in the input command\n",
			argv[argIndex] );
	    usage();
	    if( dump != NIL(FILE) ) (void) fclose(dump);
	    exit(-1);
	}
	argIndex++;
    }
    if( silent == NO ){
	printf("\nMOSAICO_GR %s \n", version() );
    }
    if( debug || verbose ) {
	if( (dump = fopen( util_file_expand(fileName), "w") ) == NIL(FILE) ){
	    errRaise( "MOSAICO_GR", 1, "Error opening the file %s", fileName );
	}
    }


    if( globalRouter == TO ){
	if( silent == NO ) {
	    if( TWgr == YES )
		printf("\n\nMOSAICO_GR environment -> Global Router environment(TW)\n\n\n");
	    else
		printf("\n\nMOSAICO_GR environment -> Global Router environment(NL)\n\n\n");
	}
	openMode = "r";
	readFromOct( inCell, outCell, openMode );
	printForTW(chipName, twdir);
	/*
	if( TWgr == YES ) {
	    printForTW(chipName, twdir);
	} else {
	    printForGopal(chipName, twdir);
	}
	*/
    } else if( globalRouter == FROM ){
	if( silent == NO ) {
	    if( TWgr == YES )
		printf("\n\nMOSAICO_GR environment <- Global Router environment(TW)\n\n\n");
	    else
		printf("\n\nMOSAICO_GR environment <- Global Router environment(NL)\n\n\n");
	}
	openMode = "a";
	readFromOct( inCell, outCell, openMode );
	readFromTw(chipName, twdir);
	storeInOct();
    } else if( globalRouter == RUN ){
	if( silent == NO ) {
	    if( TWgr == YES )
		printf("\n\nMOSAICO_GR environment <- Global Router environment(TW)\n\n\n");
	    else
		printf("\n\nMOSAICO_GR environment <- Global Router environment(NL)\n\n\n");
	}
	openMode = "a";
	readFromOct( inCell, outCell, openMode );
	printForTW(chipName, twdir);
	/*
	if( TWgr == YES ) {
	    printForTW(chipName, twdir);
	} else {
	    printForGopal(chipName, twdir);
	}
	*/
	if( twdir == NIL(char) )
	    sprintf( command, "TimberWolfMC TW%s", chipName);
	else {
	    char pathName[512];
	    extern char *getwd();

	    (void) getwd( pathName );
	    sprintf( command, "cd %s ; TimberWolfMC TW%s ; cd %s ", twdir, chipName,
			pathName);
	    }
	if( silent == NO )
	    printf("Starting the global router with the following command \n---> %s \n", command );
	(void) system(command);
	if( silent == NO ) {
	    printf("Global router completed \n");
	    if( TWgr == YES )
		printf("\n\nMOSAICO_GR environment <- Global Router environment(TW)\n\n\n");
	    else
		printf("\n\nMOSAICO_GR environment <- Global Router environment(NL)\n\n\n");
	}
	readFromTw(chipName, twdir);
	storeInOct();
    } 
    if( dump != NIL(FILE) ) (void) fclose(dump);
    if( numNetsAdded > 0 || globalRouter == FROM 
			 || globalRouter == RUN ) (void) octCloseFacet(&Chip);
    octEnd();
    {
	char command[1024];
	sprintf( command, "mosaicoDR -c %s:%s > /dev/null", 
		Chip.contents.facet.cell, Chip.contents.facet.view );
	/* printf( "mosaicoGR: final cleanup: %s\n" , command ); */
	if ( util_csystem( command ) == 127 ) {
	    fprintf(stderr, "Error in invoking %s\n", command );
	}
	sprintf( command, "PGcurrent %s:%s",
		Chip.contents.facet.cell, Chip.contents.facet.view );
	if ( util_csystem( command ) == 127 ) {
	    fprintf(stderr, "Error in invoking %s\n", command );
	}
	
    }
    VOVend(0);
}
usage()
{
    fprintf( stderr, "\nMOSAICO_GR %s \n", version() );
    fprintf( stderr, "Error: usage  MOSAICO_GR cellName[:cellView] [options] \n");
    fprintf( stderr, "Available options:\n");
    fprintf( stderr, "[-od dirName] 	directory where the created files are stored \n");
    fprintf( stderr, "[-f] (*)	the results of global routing are stored in Oct\n");
    fprintf( stderr, "[-t] (*)	the data for global routing are prepared\n");
    fprintf( stderr, "[-r]		1) data for global routing are prepared\n"); 
    fprintf( stderr, "		2) global router is executed,\n"); 
    fprintf( stderr, "		3) results of global routing are stored in Oct\n");
    /*
    fprintf( stderr, "[-TW] 		files prepared to use TimberWolfMC global router\n");
    fprintf( stderr, "[-G]		files prepared to use NLAGR global router\n");
    */
    fprintf( stderr, "[-e]		external channels are not used by the global router\n");
    fprintf( stderr, "[-ov view]	specify output view \n");
    fprintf( stderr, "[-T  dir ]	specify technology directory \n");

    fprintf( stderr, "[-s]		no printouts are generated \n"); 
    fprintf( stderr, "(*)		Note that if the option -e was used to generate\n");
    fprintf( stderr, "		the input data for the global router,\n");
    fprintf( stderr, "		the same option must be used when the\n");
    fprintf( stderr, "		results are stored in the oct.\n");
}
