#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"

extern int		debug ; 
extern FILE		*dump ;
int			lowerCoordinate, upperCoordinate;



/*
*     findAbsCoor(stack,ptoTerm)
*
*     Routine to generate the absolute coordinates of all pins.
*     It makes use of trans routine of Peter Moore.
*
*/
findAbsCoor(stack, ptoTerm)
tr_stack		*stack;	
struct terminal		*ptoTerm;
{
    int32			x1, x2, y1, y2;
    /****  find the absolute coordinates of the terminal implementation ****/

    if(debug) fprintf( dump, "Entering findAbsCoor \n");
    x1 = ptoTerm->box.lowerLeft.x;
    y1 = ptoTerm->box.lowerLeft.y;
    x2 = ptoTerm->box.upperRight.x;
    y2 = ptoTerm->box.upperRight.y;

    if(debug){
	fprintf( dump, "Before the change of coordinates \n");
	fprintf( dump, "the box for terminal %s is (%d,%d) -> (%d,%d)\n",
	   (ptoTerm->name != NIL(char) ?
	    ptoTerm->name : "<NO_NAME>"),
	    x1 , y1 , x2 , y2 );
    }

    tr_transform(stack, &x1 , &y1 );
    tr_transform(stack, &x2 , &y2 );

    ptoTerm->center = ALLOC( struct octPoint, 1);
    ptoTerm->lowerLeft = ALLOC( struct octPoint, 1);
    ptoTerm->upperRight = ALLOC( struct octPoint, 1);
    ptoTerm->center->x = (x1+x2)/2;
    ptoTerm->center->y = (y1+y2)/2;
    ptoTerm->lowerLeft->x = MIN(x1,x2);
    ptoTerm->lowerLeft->y = MIN(y1,y2);
    ptoTerm->upperRight->x = MAX(x1,x2);
    ptoTerm->upperRight->y = MAX(y1,y2);
    if(debug){
	fprintf( dump, "After the change of coordinates \n");
	fprintf( dump, "the box for terminal %s is (%d,%d) -> (%d,%d)\n",
	   (ptoTerm->name != NIL(char) ?
	    ptoTerm->name : "<NO_NAME>"),
	    ptoTerm->lowerLeft->x , ptoTerm->lowerLeft->y , 
	    ptoTerm->upperRight->x , ptoTerm->upperRight->y );
    }

    if(debug) fprintf( dump, "Leaving findAbsCoor \n");
}


insideWindow( subnet, term )
struct terminal		*term;
struct subnet		*subnet;
{
    if( subnet->ptoChannel->info->type == HORIZONTAL ){
	if( term->center->x >= lowerCoordinate &&
	    term->center->x <= upperCoordinate )
	    return(TRUE);
    } else {
	if( term->center->y >= lowerCoordinate &&
	    term->center->y <= upperCoordinate )
	    return(TRUE);
    }
    if( term->normalProj == FALSE ){  /* to account for Z-shaped channels */
	return(TRUE);
    }
    return(FALSE);
}



dist( term, channel )
struct terminal 	*term;
struct channel 		*channel;
{
    if( channel->info->type == HORIZONTAL ) {
	if( term->center->y - channel->info->centerLine >= 0 ){
	    if( term->direction[SOUTH].distance >= 0 ){
		return(term->direction[SOUTH].distance);
	    } else {
		return(MAX_INT);
	    }
	} else {
	    if( term->direction[NORTH].distance >= 0 ){
		return(term->direction[NORTH].distance);
	    } else {
		return(MAX_INT);
	    }
	}
    } else {
	if( term->center->x - channel->info->centerLine >= 0 ){
	    if( term->direction[WEST].distance >= 0 ){
		return(term->direction[WEST].distance);
	    } else {
		return(MAX_INT);
	    }
	} else {
	    if( term->direction[EAST].distance >= 0 ){
		return(term->direction[EAST].distance);
	    } else {
		return(MAX_INT);
	    }
	}
    }
}



dir( term, channel )
struct terminal 	*term;
struct channel 		*channel;
{
    if( channel->info->type == HORIZONTAL ) {
	if( term->center->y - channel->info->centerLine >= 0 ){
	    return(SOUTH);
	} else {
	    return(NORTH);
	}
    } else {
	if( term->center->x - channel->info->centerLine >= 0 ){
	    return(WEST);
	} else {
	    return(EAST);
	}
    }
}



octDeleteR(Root)
octObject	*Root;
{
    octObject	Obj;
    octGenerator genObj;
    int	counter = 0;

    if( octInitGenContents( Root, OCT_ALL_MASK, &genObj ) != OCT_OK ){
	octError("Creating the generator");
	myExit( -1, __FILE__, __LINE__);
    }
    while( octGenerate( &genObj, &Obj ) != OCT_GEN_DONE ){
	counter += octDeleteR( &Obj );
    }
    if( octDelete( Root ) != OCT_OK ){
	octError("Deleteing Root");
	myExit( -1, __FILE__, __LINE__);
    }
    counter++;
    return(counter);
}



myExit( code, file, line )
int	code, line;
char	*file;
{
    if( dump != NIL(FILE) ) (void) fclose(dump);
    fprintf( stderr,"Exit routine called in file %s at line %d\n", file, line );
    exit(code);
}
