#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

int		debug ; 
int		silent;
FILE		*dump ;



mustJoined( Chip )
octObject	*Chip;
{
    octGenerator 	instGen, bagGen, termGen;
    octObject		Instance, MasterInterface, Term, *ATerm, FTerm;
    octObject		MustJoin, MustJoinTerms, MustJoinNets;
    octObject		NewNet, *getInstTerm(), InstancesBag;
    octObject		NetType, Criticality, Object;
    octStatus		status;
    char		*netName, *findProp();
    int			errorFlag, netNumber = 0;

    NewNet.type = OCT_NET;
    NetType.type = OCT_PROP;
    NetType.contents.prop.name = "NETTYPE";
    NetType.contents.prop.type = OCT_STRING;
    NetType.contents.prop.value.string = "SIGNAL";
    Criticality.type = OCT_PROP;
    Criticality.contents.prop.name = "CRITICALITY";
    Criticality.contents.prop.type = OCT_INTEGER;
    Criticality.contents.prop.value.integer = 0;
    InstancesBag.type = OCT_BAG;
    MustJoinTerms.type = OCT_BAG;
    MustJoinNets.type = OCT_BAG;
    MustJoin.type = OCT_BAG;
    InstancesBag.contents.bag.name = "INSTANCES";
    MustJoinTerms.contents.bag.name = "MUSTJOINTERMS";
    MustJoinNets.contents.bag.name = "MUSTJOINNETS";
    MustJoinNets.objectId = -1;
    MustJoin.contents.bag.name = "MUSTJOIN";
    MasterInterface.type = OCT_FACET;
    MasterInterface.contents.facet.facet = "interface";
    MasterInterface.contents.facet.mode = "r";

    if( octGetOrCreate( Chip, &MustJoinNets ) != OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Getting or creating the MustJoinNets Bag\n");
	octError(NIL(char));
	myExit( -1, __FILE__, __LINE__);
    }

    if( octGetByName( Chip, &InstancesBag ) != OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Getting by name the instance Bag\n");
	octError(NIL(char));
	myExit( -1, __FILE__, __LINE__);
    }
	
    if( octInitGenContents( &InstancesBag, OCT_INSTANCE_MASK, &instGen ) != OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Generating the instance generator\n");
	octError(NIL(char));
	myExit( -1, __FILE__, __LINE__);
    }

    while( octGenerate( &instGen, &Instance ) == OCT_OK ){
	MasterInterface.contents.facet.cell = Instance.contents.instance.master;
	MasterInterface.contents.facet.view = Instance.contents.instance.view;
	MasterInterface.contents.facet.version = Instance.contents.instance.version;

	if(octOpenFacet(&MasterInterface) < OCT_OK){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", Instance.contents.instance.name);
	    octError("opening masterInterface \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( ( status = octGetByName( &MasterInterface, &MustJoinTerms ) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Getting by name the MustJoinTerms Bag\n");
	    octError(NIL(char));
	    myExit( -1, __FILE__, __LINE__);
	}
	if( status == OCT_NOT_FOUND ) continue;
	if( octInitGenContents( &MustJoinTerms, OCT_BAG_MASK, &bagGen) != OCT_OK ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Generating the MustJoind bag generator\n");
	    octError(NIL(char));
	    myExit( -1, __FILE__, __LINE__);
	}
	while( octGenerate( &bagGen, &MustJoin ) == OCT_OK ){
	    if( strCmp( MustJoin.contents.bag.name, "MUSTJOIN" ) != 0 ){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "The bag %s is contained in the MUSTJOINTERM bag\n",
		    MustJoin.contents.bag.name);
		continue;
	    }
	    if( octInitGenContents( &MustJoin, OCT_TERM_MASK, &termGen) != OCT_OK ){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "Generating the term generator\n");
		octError(NIL(char));
		myExit( -1, __FILE__, __LINE__);
	    }
	    netName = NIL(char);
	    while( octGenerate( &termGen, &Term ) == OCT_OK ){
		if( ( errorFlag = findTheNet( &Term, &Instance, netName ) ) != OK ){
		    break;
		}
	    }
	    if( errorFlag == KO) break; /* an error in the terminals has been detected */
	    if( netName == NIL(char) ){
		/* the net must be created and the proerties attached to it */
		netName = ALLOC( char, 20 + strLen(Instance.contents.instance.name));
		sprintf( netName, "MustJoin_%s_%d", 
			Instance.contents.instance.name,
			netNumber++);
		NewNet.contents.net.name = util_strsav(netName);
		if( octCreate( Chip, &NewNet ) < OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Creating the net to implement the must joint connection \n");
		    octError(NIL(char));
		    myExit( -1, __FILE__, __LINE__);
		}
		if( octAttachOnce( &MustJoinNets, &NewNet ) < OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Attaching the net to bag mustJoinNets \n");
		    octError("attaching the net");
		    myExit( -1, __FILE__, __LINE__);
		}
		if( octGetOrCreate( &NewNet, &Criticality ) < OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Attaching criticality property to net %s\n",
				NewNet.contents.net.name);
		    octError(NIL(char));
		    myExit( -1, __FILE__, __LINE__);
		}
		if( ( NetType.contents.prop.value.string = 
			    util_strsav(findProp( &MustJoin ) ) ) != NIL(char ) ){
		    if( octGetOrCreate( &NewNet, &NetType ) < OCT_OK ){
			fprintf( stderr, "MOSAICO_GR_ERROR:\n");
			fprintf( stderr, "Attaching nettype property to net %s\n",
				    NewNet.contents.net.name);
			octError(NIL(char));
			myExit( -1, __FILE__, __LINE__);
		    }
		}
	    } else {
		NewNet.contents.net.name = util_strsav(netName);
		if( octGetByName( Chip, &NewNet ) < OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Getting the net by name \n");
		    octError(NIL(char));
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	    if( octInitGenContents( &MustJoin, OCT_TERM_MASK, &termGen) != OCT_OK ){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "Generating the term generator\n");
		octError(NIL(char));
		myExit( -1, __FILE__, __LINE__);
	    }
	    while( octGenerate( &termGen, &Term ) == OCT_OK ){
		ATerm = getInstTerm( &Term, &Instance );
		(void) octAttachOnce( &NewNet, ATerm );
		if( octGenFirstContainer( ATerm, OCT_TERM_MASK, &FTerm ) == OCT_OK ){
		    (void) octAttachOnce( &NewNet, &FTerm );
		}
	    }
	}
    }
    if( octGenFirstContent( &MustJoinNets, OCT_ALL_MASK, &Object ) == OCT_GEN_DONE ){
	(void) octDelete( &MustJoinNets );
    }
}

    

octObject *getInstTerm( Term, Instance )
octObject *Term;
octObject *Instance;
{
    octObject ATerm;

    ATerm.type = OCT_TERM;
    ATerm.contents.term.name = Term->contents.term.name;

    if( octGetByName( Instance, &ATerm ) != OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	octError("Getting the ATerm");
	myExit( -1, __FILE__, __LINE__);
    }
    return(&ATerm);
}



findTheNet( Term, Instance, netName )
octObject *Term;
octObject *Instance;
char *netName;

{
    octObject  	*getInstTerm(), Net;
    octStatus   status;


    if( ( status = 
	  octGenFirstContainer( getInstTerm(Term,Instance), OCT_NET_MASK, &Net ) ) == OCT_OK ){
	if( netName == NIL(char) ){
	    netName = util_strsav(Net.contents.net.name);
	    return(OK);
	}
	if( strCmp( netName, Net.contents.net.name) == 0 ){
	    return(OK);
	} else {
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Two terminals in the same MUSTJOIN bag connected to different nets\n" );
	    fprintf( stderr, "The terminals will not bbe connected\n" );
	    return(KO);
	}
    } else if( status == OCT_GEN_DONE ){
	return(OK);
    } else if( status < OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Getting the net for terminal %s \n",
		Term->contents.term.name );
	myExit( -1, __FILE__, __LINE__);
    }
    /*NOTREACHED*/
}

char *findProp( Bag )
octObject *Bag;
{

    octGenerator 	termGen;
    octObject		Term, TermType;
    octStatus		status;
    char		*type;

    TermType.type = OCT_PROP;
    TermType.contents.prop.name = "TERMTYPE";
    type = NIL(char);

    if( octInitGenContents( Bag, OCT_TERM_MASK, &termGen) != OCT_OK ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Generating the term generator\n");
	octError(NIL(char));
	myExit( -1, __FILE__, __LINE__);
    }
    while( octGenerate( &termGen, &Term ) == OCT_OK ){
	if( ( status = octGetByName( &Term, &TermType ) ) == OCT_ERROR ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Getting by name the TERMTYPE Property on terminal %s\n",
			    Term.contents.term.name);
	    octError(NIL(char));
	    myExit( -1, __FILE__, __LINE__);
	}
	if( status == OCT_NOT_FOUND )
	    TermType.contents.prop.value.string = util_strsav("SIGNAL");
	if( type == NIL(char) ){
	    type = util_strsav(TermType.contents.prop.value.string);
	} else {
	    if( strCmp( type, TermType.contents.prop.value.string) == 0 ){
		continue;
	    } else {
		/* inconstencies found in the termtype properties for terminals 
		   that have to be connected together */
		if( octInitGenContents( Bag, OCT_TERM_MASK, &termGen) != OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Generating the term generator\n");
		    octError(NIL(char));
		    myExit( -1, __FILE__, __LINE__);
		}
		fprintf( stderr, "MOSAICO_GR_WARNING:\n");
		fprintf( stderr, "The following terminals are in a MUSTJOINT bag but do not have\n");
		fprintf( stderr, "the same value of the TERMTYPE property\n");
		while( octGenerate( &termGen, &Term ) == OCT_OK ){
		    if( octGetByName( &Term, &TermType ) != OCT_OK ){
			fprintf( stderr, "MOSAICO_GR_ERROR:\n");
			fprintf( stderr, "Getting by name the termtype Property \n");
			octError(NIL(char));
			myExit( -1, __FILE__, __LINE__);
		    }
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr,"Terminal %s Value of TERMTYPE prop. %s \n",
			    Term.contents.term.name, 
			    TermType.contents.prop.value.string );
		}
		type = NIL(char);
	    }
	}
    }
    return(type);
}
