#include "cds.h"
#include "utility.h"
#include "vov.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern int		silent ; 
extern FILE		*dump ;
extern int		numChannels ;
extern int 		numNodes ;
extern struct net	*firstNet;
extern struct channel	*arrayOfChannels[MAX_NUM_CHANNELS];


printForTW(chipName,twdir)
char *chipName;
char *twdir;

{
    FILE		*fopen(), *fp1, *fp2, *fp3;
    char		*fileName1;
    char		*fileName2;
    char		*fileName3;
    struct net		*presentNet;
    struct termList	*prsnTermInList;
    struct terminal	*presentTerm;
    int 		i;
    int			firstTermFound;


    if( silent == NO ) printf("Printing input files for TimberWolf\n");
    if ( twdir == NIL(char) ) {
	fileName1 = ALLOC( char , strLen(chipName) + 10) ;
	fileName2 = ALLOC( char , strLen(chipName) + 10) ;
	fileName3 = ALLOC( char , strLen(chipName) + 10) ;
	(void) sprintf( fileName1,"TW%s.twf", chipName );
	(void) sprintf( fileName2,"TW%s.gph", chipName );
	(void) sprintf( fileName3,"TW%s.par", chipName );
    } else {
	fileName1 = ALLOC( char , strLen(chipName) +  strLen(twdir) + 10) ;
	fileName2 = ALLOC( char , strLen(chipName) +  strLen(twdir) + 10) ;
	fileName3 = ALLOC( char , strLen(chipName) +  strLen(twdir) + 10) ;
	(void) sprintf( fileName1,"%s/TW%s.twf", twdir, chipName );
	(void) sprintf( fileName2,"%s/TW%s.gph", twdir, chipName );
	(void) sprintf( fileName3,"%s/TW%s.par", twdir, chipName );
    }
    fp1 = VOVfopen( util_file_expand(fileName1), "w" );
    if( fp1 == NIL(FILE) ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Opening file %s \n", fileName1);
	myExit( -1, __FILE__, __LINE__);
    }
    fp2 = VOVfopen( util_file_expand(fileName2), "w" );
    if( fp2 == NIL(FILE) ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Opening file %s \n", fileName2);
	myExit( -1, __FILE__, __LINE__);
    }
    fp3 = fopen( util_file_expand(fileName3), "r" );
    if( fp3 == NIL(FILE) ){
	fp3 = fopen( util_file_expand(fileName3), "w" );
	if( fp3 == NIL(FILE) ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Opening file %s \n", fileName3);
	    myExit( -1, __FILE__, __LINE__);
	}
	fprintf( fp3, "router.max.paths 15\n");
	fprintf( fp3, "router.extra.sources 3\n");
	fprintf( fp3, "do.global.route");
	(void) fclose(fp3);
    }
    VOVoutput( VOV_UNIX_FILE, util_file_expand(fileName3) );

    for( i = 0 ; i < numChannels ; i++ ){
	if( arrayOfChannels[i]->firstNode != NIL(struct node) )
	    fprintf( fp2 ,"edge   %3d   %3d   length   %6d   capacity   %3d \n",
			arrayOfChannels[i]->firstNode->number,
			arrayOfChannels[i]->secondNode->number,
			arrayOfChannels[i]->info->length * arrayOfChannels[i]->info->lengthFactor,
			arrayOfChannels[i]->info->width );
    }

    for( presentNet = firstNet ; presentNet != NIL(struct net) ;
			presentNet = presentNet->ptoNextNet ){
	    /* 	If there is only one terminal attached to the net
	    	there is no need to perform global router for this
		net */ 
	if( presentNet->firstTermOnNet->ptoNextTerm == NIL(struct termList ) ) continue;
	fprintf( fp1 , "\n");
	fprintf( fp1 , "net   %s_%d\n" , presentNet->name, presentNet->number  );
	for( prsnTermInList = presentNet->firstTermOnNet; prsnTermInList != NIL(struct termList);
			     prsnTermInList = prsnTermInList->ptoNextTerm ){
	    firstTermFound = FALSE;
	    for( presentTerm = prsnTermInList->ptoTerm; presentTerm != NIL(struct terminal);
				    presentTerm = presentTerm->equivTerm ){
		if( ( presentTerm->firstNode == NIL(struct node) ) || 
		    ( presentTerm->secondNode == NIL(struct node) ) ) continue;
		if ( firstTermFound == FALSE){
		    firstTermFound = TRUE ;
		    fprintf( fp1, "pin   %s   nodes   %3d   %3d   at   %3d\n",
				    presentTerm->name,
				    presentTerm->firstNode->number,
				    presentTerm->secondNode->number,
				    presentTerm->distance );
		} else {
		    fprintf( fp1, "equiv %s   nodes   %3d   %3d   at   %3d\n",
				    presentTerm->name,
				    presentTerm->firstNode->number,
				    presentTerm->secondNode->number,
				    presentTerm->distance );
		}
	    }
	}
    }
    (void) fclose(fp1);
    (void) fclose(fp2);
}
