#include "cds.h"
#include "oh.h"
#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 


extern FILE		*dump;
extern fa_geometry	termGeo, geoOnMet1, geoOnMet2, geoOnPoly;
extern octCoord		spacing_M1, spacing_M2, spacing_PO, maxSpacing;
extern char		*techDir;




int findImplementation( cellName, Term, term )
octObject	*Term;
char		*cellName;
struct terminal *term;
{
    octStatus		status;
    octGenerator	geoGem, layerGen;
    octObject		Geo, Layer, ATerm, Net;
    struct octBox	box, largestBox;
    int			counter = 0, result, numGeos = 0;
    struct layerList	*layerIds, *layerId;

    /* an equivalent term is created for each of the implementations of the term */

    Net.type = OCT_NET;


    if(octInitGenContents(Term,OCT_GEO_MASK,&geoGem) != OCT_OK){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "In cell %s \n", cellName);
	octError("Initializing the geo generator");
	myExit( -1, __FILE__, __LINE__);
    }
    while( octGenerate(&geoGem,&Geo) == OCT_OK ){
	numGeos++;
	(void) octBB( &Geo, &box );
	if( octInitGenContainers( &Geo, OCT_LAYER_MASK, &layerGen ) != OCT_OK ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Initializing the layer generator");
	    myExit( -1, __FILE__, __LINE__);
	}
	while( octGenerate(&layerGen,&Layer) == OCT_OK ){
	    if( ( strCmp( Layer.contents.layer.name, "POLY" ) != 0 ) &&
	        ( strCmp( Layer.contents.layer.name, "MET1" ) != 0 ) &&
		( strCmp( Layer.contents.layer.name, "MET2" ) != 0 ) ){
		continue;
	    } else {
		layerId = ALLOC( struct layerList, 1);
		layerId->layerName = util_strsav(Layer.contents.layer.name);
		layerId->ptoNextInList = NIL(struct layerList);
		if( counter++ == 0 ){
		    largestBox = box;
		    layerIds = layerId;
		} else {
		    if( ( result = boxContain( &box, &largestBox ) ) == 1 ){
			largestBox = box;
			layerIds = layerId;
		    } else if( result == 3 ){
			layerId->ptoNextInList = layerIds;
			layerIds = layerId;
		    }
		}
		break;
	    }
	}
    }
    if( numGeos == 0 ){
	octObject facet;
	octGetFacet( Term, &facet );
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, ohFormatName( &facet ) );
	fprintf( stderr, "The %s does not have any geometry\n",
		ohFormatName( Term ) );
	myExit( -1, __FILE__, __LINE__);
    }
    if( counter > 0 ){
	term->box = largestBox;
	term->numberOfImplementations = counter;
	term->layerIds = layerIds;
	return(YES);
    } else {
	/* check to see if the terminal has a net attached to it.
	   if it doesn't, forget about it.
	   If it does, two cases are possible:
	   1) the terminal is in a joined bag. in this case it can be disregarded
	   2) the terminal is simple. Then issue an error and stop.
	*/
	ATerm.objectId = term->actualTermId;
	if( octGetById( &ATerm ) != OCT_OK ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Getting the actual term by Id");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( ( status = octGenFirstContainer( &ATerm, OCT_NET_MASK, &Net ) ) < OCT_OK ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "In cell %s \n", cellName);
	    octError("Getting the net for the actual term ");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( status == OCT_GEN_DONE ) return(NO);
	if (octIdIsNull(term->jointBagId)) {
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Cell %s \n", cellName);
	    fprintf( stderr, 
		    "The terminal %s on net %s does not have geometry on any of\n",
		    Term->contents.term.name, Net.contents.net.name);
	    fprintf( stderr, "the legal layers (POLY, MET1, MET2)\n");
	    fprintf( stderr, "The last layer found was: %s \n",
		    Layer.contents.layer.name);
	    myExit( -1, __FILE__, __LINE__);
	}
    }
}




int checkProjection( term, distance, instDirection )
struct terminal *term;
int		instDirection;
int		distance;
{

    int			masterDirection, i, counter = 0;
    struct octBox	region, *makeRegion();
    octCoord		spacing;
    struct layerList	*layerId;

    masterDirection = findDirection( instDirection, term->instInfo->transform );
    term->direction[instDirection].distance = distance;
    if( distance == 0 ){
	for( layerId = term->layerIds; layerId != NIL(struct layerList) ;
	     layerId = layerId->ptoNextInList){
	    term->direction[instDirection].layer[counter++] = layerId;
	}
	for( i = counter; i < NUM_LAYERS ; i++ ){
	    term->direction[instDirection].layer[i] = NIL(struct layerList);
	}
	return(OK);
    }
    for( i = 0; i < NUM_LAYERS ; i++ ){
	term->direction[instDirection].layer[i] = NIL(struct layerList);
    }
    for( layerId = term->layerIds; layerId != NIL(struct layerList) ;
	 layerId = layerId->ptoNextInList){

	if( strCmp(layerId->layerName, "MET1") == 0 ){
	    spacing = spacing_M1;
	} else if( strCmp(layerId->layerName, "MET2") == 0 ){
	    spacing = spacing_M2;
	} else if( strCmp(layerId->layerName, "POLY") == 0 ){
	    spacing = spacing_PO;
	}

	/* determine the region */

	/* first make the check with the layer */
	region = *makeRegion( &(term->box), spacing, distance, masterDirection, 0 );
	if( regionIntersection( &region, layerId->layerName ) == NO ){
	    term->direction[instDirection].layer[counter++] = layerId;
	}
    }
    for( i = 0; i < NUM_LAYERS ; i++ ){
	if( term->direction[instDirection].layer[i] != NIL(struct layerList) )
	    return(OK);
    }
    return(KO);
}



struct octBox *makeRegion( box, spacing, distance, newDirection, flag )
struct octBox *box;
int 		spacing, distance, newDirection, flag;
{
    struct octBox	region;

    /* if flag = 1 then the region will include box
		   otherwise the region won't		*/

	
    if( newDirection == NORTH ){
	region.lowerLeft.x = box->lowerLeft.x - spacing;
	if( flag == 0 )
	    region.lowerLeft.y = box->upperRight.y;
	else
	    region.lowerLeft.y = box->lowerLeft.y;
	region.upperRight.x = box->upperRight.x + spacing;
	region.upperRight.y = box->upperRight.y + distance + 1;
    } else if( newDirection == SOUTH ){
	region.lowerLeft.x = box->lowerLeft.x - spacing;
	region.lowerLeft.y = box->lowerLeft.y - distance - 1;
	region.upperRight.x = box->upperRight.x + spacing;
	if( flag == 0 )
	    region.upperRight.y = box->lowerLeft.y;
	else 
	    region.upperRight.y = box->upperRight.y;
    } else if( newDirection == EAST ){
	if( flag == 0 )
	    region.lowerLeft.x = box->upperRight.x;
	else
	    region.lowerLeft.x = box->lowerLeft.x;
	region.lowerLeft.y = box->lowerLeft.y - spacing;
	region.upperRight.x = box->upperRight.x + distance + 1;
	region.upperRight.y = box->upperRight.y + spacing;
    } else if( newDirection == WEST ){
	region.lowerLeft.x = box->lowerLeft.x - distance - 1;
	region.lowerLeft.y = box->lowerLeft.y - spacing;
	if( flag == 0 )
	    region.upperRight.x = box->lowerLeft.x;
	else
	    region.upperRight.x = box->upperRight.x;
	region.upperRight.y = box->upperRight.y + spacing;
    }
    return(&region);
}




int findDirection( direction, transform)
int			direction;
octTransformType	transform;
{
    switch(transform) {
	case OCT_NO_TRANSFORM:
	    return(direction);
	case OCT_MIRROR_X:
	    if( direction == WEST ) return(EAST);
	    if( direction == EAST ) return(WEST);
	    return(direction);
	case OCT_MIRROR_Y:
	    if( direction == NORTH ) return(SOUTH);
	    if( direction == SOUTH ) return(NORTH);
	    return(direction);
	case OCT_ROT180:
	    if( direction == WEST ) return(EAST);
	    if( direction == EAST ) return(WEST);
	    if( direction == NORTH ) return(SOUTH);
	    if( direction == SOUTH ) return(NORTH);
	case OCT_ROT90:
	    if( direction == WEST ) return(NORTH);
	    if( direction == EAST ) return(SOUTH);
	    if( direction == NORTH ) return(EAST);
	    if( direction == SOUTH ) return(WEST);
	case OCT_ROT270:
	    if( direction == WEST ) return(SOUTH);
	    if( direction == EAST ) return(NORTH);
	    if( direction == NORTH ) return(WEST);
	    if( direction == SOUTH ) return(EAST);
	case OCT_MX_ROT90:
	    if( direction == WEST ) return(NORTH);
	    if( direction == EAST ) return(SOUTH);
	    if( direction == NORTH ) return(WEST);
	    if( direction == SOUTH ) return(EAST);
	case OCT_MY_ROT90:
	    if( direction == WEST ) return(SOUTH);
	    if( direction == EAST ) return(NORTH);
	    if( direction == NORTH ) return(EAST);
	    if( direction == SOUTH ) return(WEST);
	default:
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Type of transformation not known! \n");
	    myExit( -1, __FILE__, __LINE__);
    }
}


regionIntersection( region, layerName )
char	*layerName;
struct octBox	*region;
{
    fa_geometry geo;
    fa_coord  left, right, top, bottom;
    int status;

    left = region->lowerLeft.x;
    right = region->upperRight.x;
    top = region->upperRight.y;
    bottom = region->lowerLeft.y;
    fa_init(&geo);
    fa_add_box( &geo, left, bottom, right, top );

    if( strCmp( layerName, "MET1" ) == 0 ){
	status = fa_intersect_p( &geo, &geoOnMet1 );
    } else if( strCmp( layerName, "MET2" ) == 0 ){
	status = fa_intersect_p( &geo, &geoOnMet2 );
    } else if( strCmp( layerName, "POLY" ) == 0 ){
	status = fa_intersect_p( &geo, &geoOnPoly );
    } else if( layerName == NIL(char) ){
	status = fa_intersect_p( &geo, &termGeo );
    } else {
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Routine regionIntersection: Layer %s not allowed\n",
	    layerName );
	myExit( -1, __FILE__, __LINE__);
    }
    fa_free_geo(&geo);
    if (status) {
	return(YES);
    } else {
	return(NO);
    }
}



int boxIntersect( box1, box2 )
struct octBox 	*box1, *box2;
{
    int min, max;

    INTERVAL_INTERSECT( min, max, box1->lowerLeft.x, box1->upperRight.x,
		    	         box2->lowerLeft.x, box2->upperRight.x);
    if( (max - min) <= 0 ) return(NO);   
    INTERVAL_INTERSECT( min, max, box1->lowerLeft.y, box1->upperRight.y,
				 box2->lowerLeft.y, box2->upperRight.y);
    if( (max - min) <= 0 ) return(NO);   
    return(YES);
}

int boxContain( box1, box2 )
struct octBox 	*box1, *box2;
{
    if( box2->upperRight.x == box1->upperRight.x &&
        box2->upperRight.y == box1->upperRight.y &&
        box2->lowerLeft.x == box1->lowerLeft.x &&
        box2->lowerLeft.y == box1->lowerLeft.y ){
	return(3);
    } else if( box1->upperRight.x >= box2->upperRight.x &&
        box1->upperRight.y >= box2->upperRight.y &&
        box1->lowerLeft.x <= box2->lowerLeft.x &&
        box1->lowerLeft.y <= box2->lowerLeft.y ){
	return(1);
    } else if( box2->upperRight.x >= box1->upperRight.x &&
        box2->upperRight.y >= box1->upperRight.y &&
        box2->lowerLeft.x <= box1->lowerLeft.x &&
        box2->lowerLeft.y <= box1->lowerLeft.y ){
	return(2);
    } else {
	return(NO);
    }
}

printPoly( Polygon )
octObject *Polygon;
{
    int32 numPoints;
    int i;
    struct octPoint *array;

    (void) octGetPoints(Polygon, &numPoints, (struct octPoint *) 0);
    array = ALLOC( struct octPoint, numPoints );
    (void) octGetPoints(Polygon, &numPoints, array); 
    for( i = 0; 1 < numPoints; i++ ){
	printf("%d  x = %d  y = %d \n", i, array[i].x, array[i].y );
    }
}

updateProtFrame( channel, term )
struct channel	*channel;
struct terminal *term;
{
    int		direction, newDirection, i;
    struct octBox	region, *makeRegion();
    fa_coord  left, right, top, bottom;
    struct layerList	*layerLst;


    if( channel->info->type == HORIZONTAL ) {
	if( term->center->y - channel->info->centerLine >= 0 ){
	    direction = SOUTH;
	} else {
	    direction = NORTH;
	}
    } else {
	if( term->center->x - channel->info->centerLine >= 0 ){
	    direction = WEST;
	} else {
	    direction = EAST;
	}
    }
    newDirection = findDirection( direction, term->instInfo->transform);
    if( term->direction[newDirection].distance == 0 ) return;
    region = *makeRegion( &(term->box), 0, term->direction[newDirection].distance, 
			    newDirection, 0);

    left = region.lowerLeft.x;
    right = region.upperRight.x;
    top = region.upperRight.y;
    bottom = region.lowerLeft.y;
    
    for( i = 0 ; i < NUM_LAYERS ; i++ ){
	if( ( layerLst = term->direction[newDirection].layer[i] ) == NIL(struct layerList) )
	    continue;
	if( strCmp( layerLst->layerName, "MET1" ) == 0 ){
	    fa_add_box( &geoOnMet1, left, bottom, right, top );
	} else if( strCmp( layerLst->layerName, "MET2" ) == 0 ){
	    fa_add_box( &geoOnMet2, left, bottom, right, top );
	} else if( strCmp( layerLst->layerName, "POLY" ) == 0 ){
	    fa_add_box( &geoOnPoly, left, bottom, right, top );
	} 
    }
}
