#include "cds.h"
#include "utility.h"
#include "vov.h"
#include "errtrap.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

extern int		debug ; 
extern int		verbose ; 
extern int		silent ; 
extern FILE		*dump ;
extern int 		numNodes;
extern struct net	*firstNet;
extern struct node	*arrayOfNodes[MAX_NUM_NODES];
extern int		TWgr, Ggr;
struct gphEdge		edges[MAX_NUM_NODES];
int			numTrickyTerms;

/*
*	readFromTw(chipName,twdir)
*
*	Routine to read back  the informations after the global
*	router has been performed.
*
*/
readFromTw(chipName,twdir)
char *chipName;
char *twdir;
{
    char		*arrayOfPins[MAX_NUM_PINS];
    char		*termName;
    struct net		*presentNet;
    struct subnet	*subnet;
    struct subnetList	*presentSubnetL;
    struct channelList	*ptoChannelL;
    struct termList	*presentTermL;
    struct termList	*ptoTermInSubnet;
    FILE		*fopen(), *fp;
    char		*fileName;
    char		*buffer, *buffer1;
    char		*keyword;
    char		*netName;
    int			i, j, length;
    int			numPins;
    int			numSubnets;
    int			target1, target2;
    int			channelFound;
    struct subnet	*getTrickySubnet(), *trickySubnet;


    numTrickyTerms = numNodes;
    if( silent == NO ) printf("Reading output files from TimberWolf\n");
    buffer = ALLOC( char , BUFFER_DIM );
    buffer1 = ALLOC( char , BUFFER_DIM );
    keyword = ALLOC( char , BUFFER_DIM );
    netName = ALLOC( char , BUFFER_DIM );

    {
	char* dir = ( twdir==NIL(char) )? "" : twdir;
	char* sep = ( twdir==NIL(char) )? "" : "/";
	char* prefix = ( TWgr == YES ) ? "TW" : "NL";
	char  file[1024];

	(void) sprintf( file,"%s%s%s%s.pyt",dir,sep,prefix,chipName );
	fileName = util_file_expand( file );
    }
    fp = fopen( fileName, "r" );
    /* It is an output!! Created by TW */
    VOVoutput( VOV_UNIX_FILE, fileName ); 

    if ( fp == NIL(FILE) ){
	errRaise( "MOSAICO_GR", 1, "Unable to open %s\n", fileName);
    }
    
    while( fscanf( fp , "%s", keyword ) != EOF ){
	if( strCmp( keyword , "net") == 0 ){
	    (void) fscanf( fp, "%s %s %d", buffer, buffer1, &length );
	    if(debug) fprintf( dump, "net  %s\n", buffer);
	    (void) sprintf(netName,"%s", buffer);
	    for( presentNet = firstNet; presentNet != NIL(struct net) ;
		 presentNet = presentNet->ptoNextNet ){
		(void) sprintf(buffer1,"%s_%d", 
			presentNet->name, presentNet->number);
		if( strCmp(buffer1, buffer) == 0 ){
		    presentNet->length = length;
		    break;
		}
	    }
	} else if( strCmp( keyword , "pins" ) == 0 ){
	    (void) fscanf( fp, "%d" , &numPins );
	    if( numPins > MAX_NUM_PINS ){
		errRaise( "MOSAICO_GR", 1, 
			 "Too many pins in this net.Increase MAX_NUM_PINS to at least %d", numPins);
	    } 
	    if(debug) fprintf( dump, "numPins  %d\n", numPins);
	    for( i = 0 ; i < numPins ; i++ ){ 
		(void) fscanf( fp , "%s", buffer ); 
		arrayOfPins[i] = util_strsav(buffer);
	    } 
	} else if( strCmp( keyword , "channels" ) == 0 ){
	    (void) fscanf( fp, "%d" , &numSubnets ); 
	    if(debug) fprintf( dump, "numSubnets  %d\n", numSubnets);
	    if( numSubnets > MAX_NUM_NODES - 2 ) {
		errRaise( "MOSAICO_GR", 1, 
			 "Number of edges %d larger than max %d\nIncrease MAX_NUM_NODES in file cds2tw.h",
			 numSubnets, MAX_NUM_NODES - 2 );
	    }
	    for( i = 1 ; i <= numSubnets ; i++ ){ 
		(void) fscanf( fp, "%d %d %d %d", 
			&(edges[i].firstNodeIndex), &(edges[i].secondNodeIndex),
			&(edges[i].startingPoint), &(edges[i].endingPoint) );
	    }
	    for( i = 1 ; i <= numSubnets ; i++ ){ 
		if( i == 1 ){
		    presentNet->ptoFirstSubnet = newSubnetList();
		    presentSubnetL = presentNet->ptoFirstSubnet;
		} else {
		    presentSubnetL->ptoNextSubnet = newSubnetList();
		    presentSubnetL = presentSubnetL->ptoNextSubnet;
		}
		presentSubnetL->ptoSubnet = ALLOC( struct subnet, 1);
		subnet = presentSubnetL->ptoSubnet;
		subnet->ptoFirstTermInList = NIL(struct termList);
		subnet->firstNodeIndex = edges[i].firstNodeIndex;
		subnet->secondNodeIndex = edges[i].secondNodeIndex;
		subnet->startingPoint = edges[i].startingPoint;
		subnet->endingPoint = edges[i].endingPoint;
		subnet->tricky = FALSE ;
		target1 = MIN( subnet->firstNodeIndex , subnet->secondNodeIndex );
		target2 = MAX( subnet->firstNodeIndex , subnet->secondNodeIndex );
		channelFound = FALSE;
		for( j = 0 ; j < numNodes && channelFound == FALSE ; j++ ){   
		    if( arrayOfNodes[j]->number == target1 ){
			/*
			*	first look for channels with forward orientation
			*/
			for( ptoChannelL = arrayOfNodes[j]->ptoFirstStartingChannel ;
			     ptoChannelL != NIL(struct channelList) ;
			     ptoChannelL = ptoChannelL->ptoNextChannel ){
			    if( ptoChannelL->ptoChannel->secondNode->number == target2 ){
				subnet->ptoChannel = ptoChannelL->ptoChannel;
				edges[i].ptoChannel = ptoChannelL->ptoChannel;
				if( subnet->ptoChannel->info->outerChannel == TRUE ){
				    subnet->startingPoint /= subnet->ptoChannel->info->lengthFactor;
				    subnet->endingPoint /= subnet->ptoChannel->info->lengthFactor;
				}
				channelFound = TRUE;
				break;
			    }
			}
			/*
			*	If not found
			*	look for channels with backward orientation
			*/
			if( channelFound == FALSE ){
			    for( ptoChannelL = arrayOfNodes[j]->ptoFirstEndingChannel ;
				 ptoChannelL != NIL(struct channelList) ;
				 ptoChannelL = ptoChannelL->ptoNextChannel ){
				if( ptoChannelL->ptoChannel->firstNode->number == target2 ){
				    subnet->ptoChannel = ptoChannelL->ptoChannel;
				    edges[i].ptoChannel = ptoChannelL->ptoChannel;
				    if( subnet->ptoChannel->info->outerChannel == TRUE ){
					subnet->startingPoint /= subnet->ptoChannel->info->lengthFactor;
					subnet->endingPoint /= subnet->ptoChannel->info->lengthFactor;
				    }
				    channelFound = TRUE;
				    break;
				}
			    }
			}
			if( channelFound == FALSE ){
			    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
			    fprintf( stderr, "Cannot find the channel between nodes %d %d \n",
				    subnet->firstNodeIndex,
				    subnet->secondNodeIndex);
			    myExit( -1, __FILE__, __LINE__);
			}
		    }
		}
		getSegment( subnet );
		for( presentTermL = subnet->ptoChannel->ptoFirstTerm;
		     presentTermL != NIL(struct termList);
		     presentTermL = presentTermL->ptoNextTerm ){
		    termName = presentTermL->ptoTerm->name;
		    for( j = 0 ; j < numPins ; j++ ){
			if( strCmp( termName, arrayOfPins[j] ) == 0 ){
			    if( insideWindow( subnet, presentTermL->ptoTerm ) == TRUE ){
				ptoTermInSubnet = ALLOC(struct termList, 1);
				ptoTermInSubnet->ptoNextTerm = subnet->ptoFirstTermInList;
				ptoTermInSubnet->ptoTerm = presentTermL->ptoTerm;
				subnet->ptoFirstTermInList = ptoTermInSubnet;
			    }
			}
		    }
		}
		edges[i].ptoSubnet = subnet;
		if( edges[i].endingPoint == 0 ){
		    if( ( i < numSubnets ) && 
			( edges[i].secondNodeIndex == edges[i+1].firstNodeIndex ) ){
			trickySubnet = NIL(struct subnet);
			trickySubnet = getTrickySubnet( i , i+1 , netName );
			if( trickySubnet != NIL(struct subnet) ){
			    presentSubnetL->ptoNextSubnet = newSubnetList();
			    presentSubnetL = presentSubnetL->ptoNextSubnet;
			    presentSubnetL->ptoSubnet = trickySubnet;
			}
		    }
		    /*
		    if( ( ( i < numSubnets ) && 
			(edges[i].secondNodeIndex != edges[i+1].firstNodeIndex ) ) ||
			( edges[numSubnets].endingPoint == 0 ) ){

			for( j = 1 ; j < i ; j++ ){
			    if( edges[i].secondNodeIndex == edges[j].secondNodeIndex ){
				trickySubnet = NIL(struct subnet);
				trickySubnet = getTrickySubnet( i , j , netName ); 
				if( trickySubnet != NIL(struct subnet) ){
				    presentSubnetL->ptoNextSubnet = newSubnetList();
				    presentSubnetL = presentSubnetL->ptoNextSubnet;
				    presentSubnetL->ptoSubnet = trickySubnet;
				}
				break;
			    }
			}
		    }
		    */
		}
	    }
	} else {
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Unexpected keyword found %s\n", keyword);
	    myExit( -1, __FILE__, __LINE__);
	}
    }
    if(debug) fprintf( dump, "Files successfully opened and read\n");
    (void) fclose(fp);
}



struct subnet *getTrickySubnet(index1, index2 , netName)
int	index1, index2;
char	*netName;
{
    struct channelList  *presentChannelL;
    struct channel 	*channel[4];
    struct subnet	*subnet;
    int			i = 2, J, j = -1;
    int 		flag;

    channel[0] = edges[index1].ptoChannel;
    channel[1] = NIL(struct channel);
    channel[2] = NIL(struct channel);
    channel[3] = NIL(struct channel);
    if( index2 == index1+1 )
	flag = TRUE;
    else
	flag = FALSE;
    for( J = 0 ; J < numNodes ; J++ ){
	if( arrayOfNodes[J]->number == edges[index1].secondNodeIndex ){
	    j = J ;
	    break ;
	}
    }
    if( j < 0 ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Node not found\n");
	myExit( -1, __FILE__, __LINE__);
    }
    if( debug ){
	fprintf( dump, "node %d startingChannels\n", j+1);
	printList( arrayOfNodes[j]->ptoFirstStartingChannel );
    }
    for( presentChannelL = arrayOfNodes[j]->ptoFirstStartingChannel;
	 presentChannelL != NIL(struct channelList);
	 presentChannelL = presentChannelL->ptoNextChannel ){
	if( flag == TRUE ){
	    if( presentChannelL->ptoChannel->secondNode->number == 
		edges[index2].secondNodeIndex ){
		channel[1] = presentChannelL->ptoChannel;    
	    } else if( presentChannelL->ptoChannel->secondNode->number == 
		edges[index1].firstNodeIndex ){
		    continue;
	    } else {
		channel[i++] = presentChannelL->ptoChannel;
		if( i > 4) {
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Net %s\n", netName );
		    fprintf( stderr, "more than 4 channels connected to node %d\n",
			edges[index1].secondNodeIndex);
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	} else {
	    if( presentChannelL->ptoChannel->secondNode->number == 
		edges[index2].firstNodeIndex ){
		channel[1] = presentChannelL->ptoChannel;    
	    } else if( presentChannelL->ptoChannel->secondNode->number == 
		edges[index1].firstNodeIndex ){
		    continue;
	    } else {
		channel[i++] = presentChannelL->ptoChannel;
		if( i > 4) {
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Net %s\n", netName );
		    fprintf( stderr, "more than 4 channels connected to node %d\n",
			edges[index1].secondNodeIndex);
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	}
    }
    if( debug ){
	fprintf( dump, "node %d endingChannels\n", j+1);
	printList( arrayOfNodes[j]->ptoFirstEndingChannel );
    }
    for( presentChannelL = arrayOfNodes[j]->ptoFirstEndingChannel;
	 presentChannelL != NIL(struct channelList);
	 presentChannelL = presentChannelL->ptoNextChannel ){
	if( flag == TRUE ){
	    if( presentChannelL->ptoChannel->firstNode->number == 
		edges[index2].secondNodeIndex ){
		channel[1] = presentChannelL->ptoChannel;    
	    } else if( presentChannelL->ptoChannel->firstNode->number == 
		edges[index1].firstNodeIndex ){
		    continue;
	    } else {
		channel[i++] = presentChannelL->ptoChannel;
		if( i > 4) {
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Net %s\n", netName );
		    fprintf( stderr, "more than 4 channels connected to node %d\n",
			edges[index1].secondNodeIndex);
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	} else {
	    if( presentChannelL->ptoChannel->firstNode->number == 
		edges[index2].firstNodeIndex ){
		channel[1] = presentChannelL->ptoChannel;    
	    } else if( presentChannelL->ptoChannel->firstNode->number == 
		edges[index1].firstNodeIndex ){
		    continue;
	    } else {
		channel[i++] = presentChannelL->ptoChannel;
		if( i > 4) {
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Net %s\n", netName );
		    fprintf( stderr, "more than 4 channels connected to node %d\n",
			edges[index1].secondNodeIndex);
		    myExit( -1, __FILE__, __LINE__);
		}
	    }
	}
    }
    if( channel[2] == NIL(struct channel) || channel[3] == NIL(struct channel) )
	return(NIL(struct subnet));
    if( channel[2]->info->channelNumber == channel[3]->info->channelNumber ){
	subnet = ALLOC( struct subnet, 1);
	subnet->firstNodeIndex = ++numTrickyTerms;
	subnet->secondNodeIndex = edges[index1].secondNodeIndex;
	subnet->startingPoint = 0;
	subnet->endingPoint = 0;
	subnet->ptoFirstTermInList = NIL(struct termList);
	subnet->ptoChannel = channel[2];
	subnet->tricky = TRUE;
	/* the secondNodeIndex of the previous subnet is changed to its 
	   reciprocal to take care of the tricky net
	   before the insertion of the trickynet

	   1st subnet  x y
	   2nd subnet  y z

	   after

	   1st subnet  x 		numTrickyTerms
	   tricky      numTrickyTerms	y
	   2nd subnet  y 		z

	*/
	edges[index1].ptoSubnet->secondNodeIndex = numTrickyTerms;
	if( debug ){
	    fprintf( dump, "TrickySubnet created <net %s>\n", netName);
	    fprintf( dump, "starting node %d\n", subnet->firstNodeIndex );
	    fprintf( dump, "ending node %d\n", subnet->secondNodeIndex );
	    fprintf( dump, "starting point %d   ending point %d \n",
		subnet->startingPoint, subnet->endingPoint);
	}

	return(subnet);
    }
    return(NIL(struct subnet));
}


printList( pointer )
struct channelList *pointer;
{

    struct channel		*channel;
    struct channelList		*presentChannel;

    for( presentChannel = pointer;
	 presentChannel != NIL(struct channelList);
	 presentChannel = presentChannel->ptoNextChannel ){
	channel = presentChannel->ptoChannel;
	fprintf(dump, "channel number %d\n", channel->number );
	fprintf(dump, "mitch channel  %d\n", channel->info->channelNumber);
	fprintf(dump, "first node     %d <%d %d>\n", 
		channel->firstNode->number,
		channel->firstNode->x,
		channel->firstNode->y);
	fprintf(dump, "second node    %d <%d %d>\n", 
		channel->secondNode->number,
		channel->secondNode->x,
		channel->secondNode->y);
    }
}




markTrickySubnet( subnetId )
octId	subnetId;
{
    octObject		Subnet, MustAlign, PTerm;
    octGenerator	ptermGen;
    int			counter = 0;

    Subnet.type = OCT_NET;
    Subnet.objectId = subnetId;
    MustAlign.type = OCT_BAG;
    MustAlign.contents.bag.name = "GRMustAlign";
    PTerm.type = OCT_BAG;
    PTerm.contents.bag.name = "PseudoTerm";

    if( octGetById( &Subnet ) != OCT_OK ){
	octError( "getting the subnet by id \n");
	myExit( -1, __FILE__, __LINE__);
    }

    if( octInitGenContents( &Subnet, OCT_BAG_MASK, &ptermGen ) != OCT_OK ){
	octError( "Generating the generator \n");
	myExit( -1, __FILE__, __LINE__);
    }
    while( octGenerate( &ptermGen, &PTerm ) != OCT_OK ) {
	if( strCmp( PTerm.contents.bag.name , "PseudoTerm" ) != 0 ) continue;
	if( ++counter == 1 ){
	    if( octCreate( &Subnet, &MustAlign ) != OCT_OK ){
		octError( "Creating the MustAlign bag \n");
		myExit( -1, __FILE__, __LINE__);
	    }
	}
	if( octAttach( &MustAlign, &PTerm ) != OCT_OK ){
	    octError( "Attaching Pterm to the MustAlign bag \n");
	    myExit( -1, __FILE__, __LINE__);
	}
    }
}
