#include "cds.h"
#include "vov.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "oct.h"
#include "oh.h"
#include "tap.h"

int		debug ; 
int		silent;
FILE		*dump ;
int 		levelCounter;
int		numChannels = 0 ;
int 		numNodes = 0 ;
int		numNets = 0 ;
int		numTerms = 0 ;
int		numConnectedTerms = 0 ;
int 		numPoints ;
int		maxLength;
extern int	outerChan;
struct net	*firstNet = NIL(struct net);
extern struct channel	*arrayOfChannels[MAX_NUM_CHANNELS];
extern octObject 	Chip;
extern int	numNetsAdded;
struct octPoint chipCenter;


/*
*	readFromOct(inCell, outCell, openMode)
*
*	Routine to read the information out from oct and stro them in a 
*	local data structure to prepare data for the TimberWolf global
*	router phase.
*
*/
void
readFromOct(inCell, outCell, openMode )
char *inCell;
char *outCell;
char *openMode;
{
    octObject*	        NewChipPtr = &Chip;
    octObject 		OldChip;
    octObject		Bag;
    octStatus 		status, facetStatus;
    struct octBox	chipBB;
    int			numChannelUsed;  /* number of channels available to the
					   global router */
    extern void		findLeaf();


    octBegin();

    OldChip.type = OCT_FACET;
    OldChip.contents.facet.cell = NIL(char);
    OldChip.contents.facet.view = NIL(char);
    OldChip.contents.facet.version = OCT_CURRENT_VERSION;
    NewChipPtr->contents.facet.cell = NIL(char);
    NewChipPtr->contents.facet.view = NIL(char);
    NewChipPtr->contents.facet.version = OCT_CURRENT_VERSION;
    Bag.type = OCT_BAG;
     
    if( strCmp( openMode, "r" ) == 0){
	ohUnpackDefaults(NewChipPtr, "a", ":cd:contents");
	(void) ohUnpackFacetName(NewChipPtr, inCell);
	if( silent == NO )
	    printf("Reading oct input %s \n\n\n", ohFormatName(NewChipPtr) );
    } else {
	/* first copy the facet */
	if( strCmp( inCell, outCell ) != 0 ) {
	    ohUnpackDefaults(&OldChip, "a", "::interface");
	    (void) ohUnpackFacetName(&OldChip, inCell);
	    if( OldChip.contents.facet.view == NIL(char) ) 
		OldChip.contents.facet.view = "cd";
	    if( outCell == NIL(char) ) {
		NewChipPtr->contents.facet.view = "cd_after_gr";
	    } else {
		(void) ohUnpackFacetName(NewChipPtr, outCell);
	    }
	    if( NewChipPtr->contents.facet.cell == NIL(char) )
		NewChipPtr->contents.facet.cell = OldChip.contents.facet.cell;
	    NewChipPtr->contents.facet.facet = "interface";
	    NewChipPtr->contents.facet.mode = "a";

	    VOVinputFacet( &OldChip );
	    VOVoutputFacet( NewChipPtr );
	    if(octCopyFacet(NewChipPtr, &OldChip) != OCT_OK ){
		octError("error in copy interface facet");
		myExit( -1, __FILE__, __LINE__);
	    }
	    OldChip.contents.facet.facet = "contents";
	    NewChipPtr->contents.facet.facet = "contents";

	    VOVinputFacet( &OldChip );
	    VOVoutputFacet( NewChipPtr );
	    if(octCopyFacet(NewChipPtr, &OldChip) != OCT_OK ){
		octError("error in copy contents facet");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if( silent == NO )
		printf("Reading oct input %s \n\n\n", ohFormatName(&OldChip) );
	} else {
	    ohUnpackDefaults(NewChipPtr, "a", ":cd:contents");
	    (void) ohUnpackFacetName(NewChipPtr, outCell);
	    if( silent == NO )
		printf("Reading oct input %s \n\n\n", ohFormatName(NewChipPtr) );
	}


	/* open the facet for reading or for appending */
	    
    }
    
    VOVinputFacet( NewChipPtr );
    VOVoutputFacet( NewChipPtr );
    if(( facetStatus = octOpenFacet(NewChipPtr)) == OCT_ERROR){
	octError("Couldn't open contents facet");
	octError( ohFormatName( NewChipPtr ) );
	myExit( -1, __FILE__, __LINE__);
    }
    if( facetStatus == OCT_NO_EXIST || facetStatus == OCT_NEW_FACET ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n\tThe facet %s does not exist\n", ohFormatName( NewChipPtr ));
	myExit( -1, __FILE__, __LINE__);
    }
    if( facetStatus == OCT_NO_PERM ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n\tThe facet %s does not have the right permission\n",
		ohFormatName( NewChipPtr ));
	myExit( -1, __FILE__, __LINE__);
    }
    arrayOfChannels[0] = ALLOC( struct channel, 1 );
    mustJoined(NewChipPtr);
    (void) octBB( NewChipPtr, &chipBB );
    maxLength = chipBB.upperRight.x + chipBB.upperRight.y -
    		chipBB.lowerLeft.x + chipBB.lowerLeft.y;
    chipCenter.x = ( chipBB.upperRight.x + chipBB.lowerLeft.x ) / 2;
    chipCenter.y = ( chipBB.upperRight.y + chipBB.lowerLeft.y ) / 2;

    /* initialize oct to generate all the bags in the facet */
      
    Bag.contents.bag.name = "CDChanDef";
    if( (status = octGetByName(NewChipPtr,&Bag)) == OCT_OK ){
	if(debug){
	    levelCounter = 0 ;
	    fprintf( dump, "The technology used is %s\n", tapGetDirectory(&Bag) );
	}
	findLeaf(&Bag);
    } else if( status == OCT_NOT_FOUND ){ 
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "The bag CDChanDef was not found !!!");
	myExit( -1, __FILE__, __LINE__);
    }
    numChannelUsed = sanitizeChannels();
    if( silent == NO ){
	if( strCmp( openMode , "r" ) == 0 )
	    printf("****************  %s  ******************\n", NewChipPtr->contents.facet.cell);
	else
	    printf("****************  %s  ******************\n", OldChip.contents.facet.cell);
	printf("Number of channels            = %d\n",numChannels);
	if( outerChan == NO )
	    printf("Number of channels used       = %d\n",numChannelUsed);
	printf("Number of terms               = %d\n",numTerms);
	printf("Number of connected terms     = %d\n",numConnectedTerms);
	printf("Number of nets                = %d\n",numNets);
	if( numNetsAdded > 0 )
	    printf("Number of nets without a name = %d\n",numNetsAdded);
	printf("\n\n");
    }
    return;
} 




/*
*
*	findLeaf(ptrBag);
*
*	This routine recursively searchs the channel definition tree for
*	leafs where are stored the elementary channels together with the 
*	cells that are on the channel
*
*/
void
findLeaf(ptrBag)
octObject *ptrBag;
{
    octStatus		status;
    octObject 		Bag;
    octGenerator 	genBag;

    int			bagFound = FALSE;

    if(debug){
	levelCounter++;
	fprintf( dump, " entering level %d , everything ok so far \n",
				    levelCounter ) ;
    }
    if(octInitGenContents(ptrBag,OCT_BAG_MASK, &genBag) != OCT_OK){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr, "Level %d of the Channel Def. tree \n", levelCounter);
	octError("generating bags");
	myExit( -1, __FILE__, __LINE__);
    }

    /* generate all the subbags contained in the bag */
      
    while((status = octGenerate(&genBag,&Bag)) == OCT_OK){

	if(debug) fprintf( dump, "MASTER %s\n",Bag.contents.bag.name);

	/*
	*    Check if the present bag is the one we are loocking for
	*    i.e. the bag that contains the information related to 
	*    channel definition
	*/

	if(strCmp(Bag.contents.bag.name,"CDChanDef") == 0){
	    /* We found the right bag  but it is not a leaf 
	    *  therefore we continue the search */
	    findLeaf(&Bag);
	    bagFound = TRUE ;
	}
    }
    if(status == OCT_GEN_DONE){
	if(bagFound == FALSE){
	    /* 
	    *  It is likely that we have found the leaf of the tree and we 
	    *  start popping the informations 
	    */
	    if(debug) fprintf( dump, "Looking for channels and polygons\n");
	    createChannels(ptrBag);
	    generateCell(ptrBag);
	}
    }
    if(debug){ 
	fprintf( dump, " leaving level %d , everything ok so far \n",
				levelCounter ) ;
	levelCounter-- ;
    }
    return;
}
