#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"
extern int		debug ; 
extern int		verbose ; 
extern int		silent ; 
extern FILE		*dump ;
extern struct net	*firstNet;
extern octObject	Chip;
extern int		TWgr, Ggr;
struct pseudoTerm       arrayOfPseudoTerms[MAX_NUM_NODES];



/*
*
*	storeInOct()
*
*	Routine that restores the information in oct after the global router
*	step
*/
storeInOct()
{
    
    octObject		Net;
    octObject		Bag;
    octObject		InstanceBag;
    octObject		Subnet;
    octObject		Term;
    octObject		FormalTerm;
    octObject		Channel;
    octObject		Prop;
    octObject		Instance;
    octObject		PseudoTerm;

    octStatus 		status;
    octGenerator 	genTerm, genFterm;
    int			subnetCounter;
    int			numPseudoTerms;
    int			i, index, counter;
    struct net		*presentNet;
    struct subnet	*subnet;
    struct subnetList	*presentSubnetL;
    struct terminal	*presentTerm;
    struct termList 	*presentTermL;
    struct pseudoTerm	*addSubnet();
    struct pseudoTerm	*pTerm;

    /* open the facet for reading */
	
    /*
    debug = 1 ;
    */
    if( silent == NO ) printf("Storing information back to OCT\n");
    Net.type = OCT_NET;
    Bag.type = OCT_BAG;
    InstanceBag.type = OCT_BAG;
    PseudoTerm.type = OCT_BAG;
    Bag.contents.bag.name = "CDPseudoTerm";
    InstanceBag.contents.bag.name = "INSTANCES";
    PseudoTerm.contents.bag.name = "PseudoTerm";
    Channel.type = OCT_BAG;
    Channel.contents.bag.name = "CDChanDef";
    Subnet.type = OCT_NET;
    Subnet.contents.net.name = "CDSubnet";
    Prop.type = OCT_PROP;
    Prop.contents.prop.name = "CDOuterChannel";
    Instance.type = OCT_INSTANCE;
    Term.type = OCT_TERM;

    for( i = 0 ; i < MAX_NUM_NODES ; i++ ){
	arrayOfPseudoTerms[i].numberSubnets = 0;
    }

    if( (status = octCreate(&Chip, &PseudoTerm)) != OCT_OK){
	octError("Error I cannot create the Pseudo term bag\n");
	myExit( -1, __FILE__, __LINE__);
    }
    for( presentNet = firstNet ; presentNet != NIL(struct net) ;
			presentNet = presentNet->ptoNextNet ){
	numPseudoTerms = 0;
	Net.objectId= presentNet->id;
	if( (status = octGetById(&Net)) != OCT_OK){
	    octError("Oct is corrupted: I cannot find the net");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( octInitGenContents( &Net, OCT_TERM_MASK, &genTerm ) != OCT_OK ){
	    octError("error generating the generator for terms \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	while( octGenerate( &genTerm, &Term ) == OCT_OK ){
	    if (octIdIsNull(Term.contents.term.instanceId)) continue;
	    if( ( status = octDetach( &Net, &Term ) ) != OCT_OK ){
		fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		fprintf( stderr, "Detaching the terminal %s from net %s \n",
			Term.contents.term.name,
			Net.contents.net.name);
		fprintf( stderr, "Oct returns %d \n", status);
		octError("Error detaching a terminal \n");
		myExit( -1, __FILE__, __LINE__);
	    }
	}
	if( debug )
	    fprintf( dump, "Net name %s \n", Net.contents.net.name);
	subnetCounter = 0 ;
	/*
	*	for each subnet attached to present net first generate the 
	*	octObject subnet 
	*/
	for( presentSubnetL = presentNet->ptoFirstSubnet;
	     presentSubnetL != NIL(struct subnetList);
	     presentSubnetL = presentSubnetL->ptoNextSubnet ){
	    subnetCounter++;
	    subnet = presentSubnetL->ptoSubnet;
	    if( (status = octCreate(&Net,&Subnet)) != OCT_OK ){
		octError("error in the creation of the subnet attached to net \n");
		myExit( -1, __FILE__, __LINE__);
	    }
	    subnet->id = Subnet.objectId;
	    if( debug ){
		fprintf( dump, "Subnet created %d <octId %d> %d -> %d \n", 
			    subnetCounter, Subnet.objectId, 
			    subnet->firstNodeIndex, subnet->secondNodeIndex);
		if( subnet->tricky == TRUE )
		    fprintf( dump, "this is a tricky one \n");
	    }
	    for( presentTermL = subnet->ptoFirstTermInList;
		 presentTermL != NIL(struct termList);
		 presentTermL = presentTermL->ptoNextTerm ){
		presentTerm = presentTermL->ptoTerm;

		Instance.contents.instance.name = presentTerm->instInfo->name;
		if(octGetByName(&Chip, &Instance) != OCT_OK){
		    octError("error getting the instance");
		    myExit( -1, __FILE__, __LINE__);
		}
		Term.contents.term.name = presentTerm->termName;
		if(octGetByName(&Instance, &Term) != OCT_OK){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Terminal %s\n", presentTerm->termName );
		    octError("Oct is corrupted: I cannot find the term");
		    myExit( -1, __FILE__, __LINE__);
		}
		/* 
		*	Now check if the terminal is a fTerm
		*/
		if( octInitGenContainers( &Term, OCT_TERM_MASK, &genFterm ) != OCT_OK ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Terminal   %s \n", presentTerm->name );
		    octError("generating formalTerminal attached to terminal");
		    myExit( -1, __FILE__, __LINE__);
		}
		counter = 0;
		if( octGenerate( &genFterm, &FormalTerm ) == OCT_OK ){
		    counter++;
		    if( octGetByName( &Chip, &FormalTerm )== OCT_OK ){
			if( (status = octAttach(&Subnet,&FormalTerm)) != OCT_OK){
			    octError("Oct is corrupted: I cannot attach the term to subnet");
			    myExit( -1, __FILE__, __LINE__);
			}
			if( (status = octAttachOnce(&Net,&FormalTerm)) == OCT_ERROR){
			    octError("Oct is corrupted: I cannot attach the term to net");
			    myExit( -1, __FILE__, __LINE__);
			}
			/*
			*	Check if the instance is a connector instance
			*/
			if( ( status = octGetContainerByName( &Instance, &InstanceBag ))== OCT_NOT_FOUND ){
			    continue;
			} else if( status == OCT_OK ){
			    if( (status = octAttach(&Subnet,&Term)) != OCT_OK){
				octError("Oct is corrupted: I cannot attach the term to subnet");
				myExit( -1, __FILE__, __LINE__);
			    }
			    if( (status = octAttachOnce(&Net,&Term)) == OCT_ERROR){
				octError("Oct is corrupted: I cannot attach the term to net");
				myExit( -1, __FILE__, __LINE__);
			    }
			} else {
			    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
			    fprintf( stderr,"Instance   %s \n", presentTerm->instInfo->name );
			    octError("looking for the instance Bag ");
			    myExit( -1, __FILE__, __LINE__);
			}
		    }
		}
		if( counter == 0 ){
		    /*
		    *	if the counter == 0 it means that no fTerms were found 
		    *	attached to Term
		    */
		    if( (status = octAttach(&Subnet,&Term)) != OCT_OK){
			octError("Oct is corrupted: I cannot attach the term to subnet");
			myExit( -1, __FILE__, __LINE__);
		    }
		    if( (status = octAttachOnce(&Net,&Term)) == OCT_ERROR){
			octError("Oct is corrupted: I cannot attach the term to net");
			myExit( -1, __FILE__, __LINE__);
		    }
		}

		if( debug )
		    fprintf( dump, "subnet %d attached to term %s\n",
			Subnet.objectId, Term.contents.term.name );
	    }
	    if( subnet->startingPoint == 0 ){
		Bag.objectId = -1 ;
		for( i = 0 ; i < numPseudoTerms ; i++ ){
		    if( subnet->firstNodeIndex == arrayOfPseudoTerms[i].nodeNumber ){
			pTerm = &arrayOfPseudoTerms[i];
			pTerm->subnet[pTerm->numberSubnets] = subnet;
			Bag.objectId = pTerm->pseudoTermBagId;
			/*  special case in which there are more than 3 subnets attached
			    to the same pseudo term. This means that three channels
			    intersect in a cross pattern. In this case an extra subnet 
			    must be introduced */
			if( pTerm->numberSubnets == 2 ){
			    if( ( pTerm->subnet[0]->ptoChannel->info->id != 
				  pTerm->subnet[1]->ptoChannel->info->id ) && 
			        ( pTerm->subnet[0]->ptoChannel->info->id != 
				  pTerm->subnet[2]->ptoChannel->info->id ) && 
			        ( pTerm->subnet[1]->ptoChannel->info->id != 
				  pTerm->subnet[2]->ptoChannel->info->id ) ){
				pTerm = addSubnet( i, &Bag.objectId, &Net, &PseudoTerm,
						    &numPseudoTerms, subnet);
				Bag.objectId = pTerm->pseudoTermBagId;
				index = numPseudoTerms;
			    } else {
				index = i+1;
			    }
			} else if( pTerm->numberSubnets > 2 ){
			    pTerm = addSubnet( i, &Bag.objectId, &Net, &PseudoTerm,
						&numPseudoTerms, subnet);
			    Bag.objectId = pTerm->pseudoTermBagId;
			    index = numPseudoTerms;
			} else {
			    index = i+1;
			}
			break;
		    }
		}
		if( Bag.objectId > 0 ){
		    if( octGetById(&Bag) < OCT_OK ){
			octError("retriving information on pseudoTerms \n");
			myExit( -1, __FILE__, __LINE__);
		    }
		    if( octAttach(&Subnet, &Bag) < OCT_OK ){
			octError("attaching pseudoTerms to subnets\n");
			myExit( -1, __FILE__, __LINE__);
		    }
		    pTerm->numberSubnets++;
		} else {
		    if( (status = octCreate(&Subnet, &Bag)) != OCT_OK ){
			octError("error in the creation of the object");
			myExit( -1, __FILE__, __LINE__);
		    }
		    arrayOfPseudoTerms[numPseudoTerms].pseudoTermBagId = Bag.objectId;
		    arrayOfPseudoTerms[numPseudoTerms].numberSubnets = 1;
		    arrayOfPseudoTerms[numPseudoTerms].subnet[0] = subnet;
		    arrayOfPseudoTerms[numPseudoTerms++].nodeNumber =
				    subnet->firstNodeIndex;
		    index = numPseudoTerms;
		    if( octAttach( &PseudoTerm , &Bag ) < OCT_OK ){
			octError("error attaching pseudo terms \n");
			myExit( -1, __FILE__, __LINE__);
		    }
		}
		if( debug )
		    fprintf( dump, "subnet %d attached to psuedoterm %d <octId %d>\n",
			Subnet.objectId, index, Bag.objectId );
	    } 
	    if( subnet->endingPoint == 0 ){
		Bag.objectId = -1;
		for( i = 0 ; i < numPseudoTerms ; i++ ){
		    if( subnet->secondNodeIndex == arrayOfPseudoTerms[i].nodeNumber ){
			pTerm = &arrayOfPseudoTerms[i];
			pTerm->subnet[pTerm->numberSubnets] = subnet;
			Bag.objectId = pTerm->pseudoTermBagId;
			/*  special case in which there are more than 3 subnets attached
			    to the same pseudo term. This means that three channels
			    intersect in a cross pattern. In this case an extra subnet 
			    must be introduced */
			if( pTerm->numberSubnets == 2 ){
			    if( ( pTerm->subnet[0]->ptoChannel->info->id != 
				  pTerm->subnet[1]->ptoChannel->info->id ) && 
			        ( pTerm->subnet[0]->ptoChannel->info->id != 
				  pTerm->subnet[2]->ptoChannel->info->id ) && 
			        ( pTerm->subnet[1]->ptoChannel->info->id != 
				  pTerm->subnet[2]->ptoChannel->info->id ) ){
				pTerm = addSubnet( i, &Bag.objectId, &Net, &PseudoTerm,
						    &numPseudoTerms, subnet);
				Bag.objectId = pTerm->pseudoTermBagId;
				index = numPseudoTerms;
			    } else {
				index = i+1;
			    }
			} else if( pTerm->numberSubnets > 2 ){
			    pTerm = addSubnet( i, &Bag.objectId, &Net, &PseudoTerm,
						&numPseudoTerms, subnet);
			    Bag.objectId = pTerm->pseudoTermBagId;
			    index = numPseudoTerms;
			} else {
			    index = i+1;
			}
			break;
		    }
		}
		if( Bag.objectId > 0 ){
		    if( octGetById(&Bag) < OCT_OK ){
			octError("retriving information on pseudoTerms \n");
			myExit( -1, __FILE__, __LINE__);
		    }
		    if( octAttach(&Subnet, &Bag) < OCT_OK ){
			octError("attaching pseudoTerms to subnets\n");
			myExit( -1, __FILE__, __LINE__);
		    }
		    pTerm->numberSubnets++;
		} else {
		    if( (status = octCreate(&Subnet, &Bag)) != OCT_OK ){
			octError("error in the creation of the object");
			myExit( -1, __FILE__, __LINE__);
		    }
		    arrayOfPseudoTerms[numPseudoTerms].pseudoTermBagId = Bag.objectId;
		    arrayOfPseudoTerms[numPseudoTerms].numberSubnets = 1;
		    arrayOfPseudoTerms[numPseudoTerms].subnet[0] = subnet;
		    arrayOfPseudoTerms[numPseudoTerms++].nodeNumber =
				    subnet->secondNodeIndex;
		    index = numPseudoTerms;
		    if( octAttach( &PseudoTerm , &Bag ) < OCT_OK ){
			octError("error attaching pseudo terms \n");
			myExit( -1, __FILE__, __LINE__);
		    }
		}
		if( debug )
		    fprintf( dump, "subnet %d attached to psuedoterm %d <octId %d>\n",
			Subnet.objectId, index, Bag.objectId );
	    }
	    Channel.objectId = subnet->ptoChannel->info->id;
	    if(octGetById(&Channel) != OCT_OK){
		octError("error finding the channel");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if( subnet->ptoChannel->info->outerChannel == YES ){
		/* modify the property to notify cprep that this channel has
		    to be used  */
	        subnet->ptoChannel->info->outerChannel = NO;
		if( octGetByName( &Channel, &Prop ) == OCT_OK ){
		    (void) octDelete( &Prop );
		} 
	    }

	    if( (status = octAttach(&Channel,&Subnet)) != OCT_OK){
		octError("Oct is corrupted: I cannot attach the subnet to the channel");
		myExit( -1, __FILE__, __LINE__);
	    }
	    if( debug )
		fprintf( dump, "subnet %d attached to channel %d <octId %d>\n",
		    Subnet.objectId, subnet->ptoChannel->info->channelNumber, 
		    Channel.objectId );
	}
	if( ( subnetCounter == 0 ) &&
	    ( presentNet->firstTermOnNet->ptoNextTerm != NIL(struct termList ) ) ){ 
	    fprintf( stderr, "MOSAICO_GR_WARNING:\n");
	    fprintf( stderr, "Net %s has no subnets: It is probably missing in file *.pyt\n",
		presentNet->name );
	}
	orderPseudoTerms(numPseudoTerms);
	for( i = 0 ; i < numPseudoTerms ; i++ ){
	    arrayOfPseudoTerms[i].numberSubnets = 0;
	}
    }
    if( verbose ) {
	if( silent == NO ) printf("running the checks\n");
	checkPseudoTerms( &PseudoTerm );
	checkSubnets( &Chip );
    }
    if( silent == NO ) printf("Flushing data back to OCT\n");
}



struct pseudoTerm *addSubnet( index , bagId , Net , PseudoTerm , numPT, subnet)
octObject		*Net;
octObject		*PseudoTerm;
octId			*bagId;
int			index;
int			*numPT;
struct subnet		*subnet;
{
    octObject		Bag;
    octObject		Subnet;
    octObject		Channel;
    int			chanIndex;
    struct pseudoTerm 	*newPTerm;
    struct subnet	*newSubnet;
     
    Bag.type = OCT_BAG;
    Bag.contents.bag.name = "CDPseudoTerm";
    Channel.type = OCT_BAG;
    Channel.contents.bag.name = "CDChanDef";
    Subnet.type = OCT_NET;
    Subnet.contents.net.name = "CDSubnet";


    if( octCreate(Net,&Subnet) != OCT_OK ){
	octError("error in the creation of the subnet attached to net \n");
	myExit( -1, __FILE__, __LINE__);
    }
    Bag.objectId = *bagId;
    if( octGetById(&Bag) < OCT_OK ){
	octError("retriving information on pseudoTerms \n");
	myExit( -1, __FILE__, __LINE__);
    }
    if( octAttach(&Subnet, &Bag) < OCT_OK ){
	octError("attaching pseudoTerms to subnets\n");
	myExit( -1, __FILE__, __LINE__);
    }
    if( octCreate(&Subnet, &Bag) != OCT_OK ){
	octError("error in the creation of the object");
	myExit( -1, __FILE__, __LINE__);
    }
    arrayOfPseudoTerms[index].numberSubnets++;
    chanIndex = getChannel( &arrayOfPseudoTerms[index] );
    newSubnet = ALLOC( struct subnet, 1);
    newSubnet->startingPoint = 0;
    newSubnet->endingPoint = 0;
    newSubnet->ptoChannel = arrayOfPseudoTerms[index].subnet[chanIndex]->ptoChannel;
    newSubnet->ptoFirstTermInList = NIL( struct termList );
    newSubnet->firstNodeIndex = arrayOfPseudoTerms[index].nodeNumber;
    newSubnet->secondNodeIndex = arrayOfPseudoTerms[index].nodeNumber;
    newSubnet->tricky = FALSE;
    newPTerm = &arrayOfPseudoTerms[((*numPT)++)];
    newPTerm->nodeNumber = - arrayOfPseudoTerms[index].nodeNumber;
    newPTerm->numberSubnets = 1;  /* it will be incremented outside the routine */
    newPTerm->subnet[0] = subnet;
    newPTerm->subnet[1] = newSubnet;
    newPTerm->pseudoTermBagId = Bag.objectId;
    if( octAttach( PseudoTerm , &Bag ) < OCT_OK ){
	octError("error attaching pseudo terms \n");
	myExit( -1, __FILE__, __LINE__);
    }
    Channel.objectId = arrayOfPseudoTerms[index].subnet[chanIndex]->ptoChannel->info->id;
    if(octGetById(&Channel) != OCT_OK){
	octError("error finding the channel 1 ");
	myExit( -1, __FILE__, __LINE__);
    }
    if( octAttach(&Channel,&Subnet) != OCT_OK){
	octError("Oct is corrupted: I cannot attach the subnet to the channel");
	myExit( -1, __FILE__, __LINE__);
    }
    if( debug ){
	fprintf( dump, "fanta-subnet %d attached to channel <octId %d> \n",
	    Subnet.objectId, Channel.objectId );
	fprintf( dump, "fanta-subnet %d between pseudoTerms <octId %d> and <octId %d> \n",
	    Subnet.objectId, *bagId, Bag.objectId);
    }
    return( newPTerm );
}

int  getChannel( pTerm )
struct pseudoTerm 	*pTerm;
{
    octObject		Prop;
    int			i = 0, j;
    int			max = -1, maxIndex = -1;

    Prop.type = OCT_PROP;
    Prop.contents.prop.name = "CDChanOrder";

    for( i = 0 ; i < pTerm->numberSubnets ; i++ ){
	for( j = i+1 ; j < pTerm->numberSubnets ; j++ ){
	    if( pTerm->subnet[i]->ptoChannel->info->id == 
		pTerm->subnet[j]->ptoChannel->info->id ) return(i);
	}
    }
    /*    three different channels attached to a three subnet net  */
    for( i = 0 ; i < 3 ; i++ ){
	if( max < pTerm->subnet[i]->ptoChannel->info->channelNumber ){
	    max = pTerm->subnet[i]->ptoChannel->info->channelNumber;
	    maxIndex = i;
	}
    }
    if( maxIndex < 0 ){
	fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	fprintf( stderr,"No channel for the fantaSubnet found\n");
	myExit( -1, __FILE__, __LINE__);
    }
    return(maxIndex);
}



orderPseudoTerms( numPseudoTerms )
int	numPseudoTerms;
{
    int i, j;
    octObject	PTerm, OrderProp;
    struct pseudoTerm *pTerm;
    int	max, min, child, parent;

    PTerm.type = OCT_BAG;
    OrderProp.type = OCT_PROP;
    OrderProp.contents.prop.type = OCT_STRING;
    OrderProp.contents.prop.name = "GRSideTermOrder";

    for( i = 0 ; i < numPseudoTerms ; i++ ){
	pTerm = &arrayOfPseudoTerms[i];
	PTerm.objectId = pTerm->pseudoTermBagId;
	if( pTerm->numberSubnets < 2 ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr,"The pseudo term has less than two subnets attached to it\n");
	    myExit( -1, __FILE__, __LINE__);
	} else if( pTerm->numberSubnets > 2 ){
	    OrderProp.contents.prop.value.string = "CENTER";
	} else if( pTerm->subnet[0]->tricky == TRUE ){
	    OrderProp.contents.prop.value.string = "CENTER";
	    markTrickySubnet( pTerm->subnet[0]->id );
	} else if( pTerm->subnet[1]->tricky == TRUE ){
	    OrderProp.contents.prop.value.string = "CENTER";
	    markTrickySubnet( pTerm->subnet[1]->id );
	} else {
	    max = 0;
	    min = MAX_INT;
	    for( j = 0 ; j < pTerm->numberSubnets ; j++ ){
		if( max <= pTerm->subnet[j]->ptoChannel->info->channelNumber ){
		    max = pTerm->subnet[j]->ptoChannel->info->channelNumber;
		    parent = j;
		}
		if( min >= pTerm->subnet[j]->ptoChannel->info->channelNumber ){
		    min = pTerm->subnet[j]->ptoChannel->info->channelNumber;
		    child = j;
		}
	    }
	    if( parent == child ){
		/* not used for now 
		    those are part of the subnets that are not presently
		    used by mitch
		cleanSubnet( pTerm, numPseudoTerms, i );  
		continue;
		*/
		OrderProp.contents.prop.value.string = "CENTER";
	    } else {
		if( pTerm->subnet[child]->ptoChannel->info->type == HORIZONTAL ){
		    max = MAX( pTerm->subnet[parent]->ptoChannel->firstNode->y ,
			       pTerm->subnet[parent]->ptoChannel->secondNode->y );
		    min = MIN( pTerm->subnet[parent]->ptoChannel->firstNode->y ,
			       pTerm->subnet[parent]->ptoChannel->secondNode->y );
		    if( pTerm->subnet[child]->ptoChannel->firstNode->y == min ){
			OrderProp.contents.prop.value.string = "UP";
		    } else {
			OrderProp.contents.prop.value.string = "DOWN";
		    }
		} else {
		    max = MAX( pTerm->subnet[parent]->ptoChannel->firstNode->x ,
			       pTerm->subnet[parent]->ptoChannel->secondNode->x );
		    min = MIN( pTerm->subnet[parent]->ptoChannel->firstNode->x ,
			       pTerm->subnet[parent]->ptoChannel->secondNode->x );
		    if( pTerm->subnet[child]->ptoChannel->firstNode->x == min ){
			OrderProp.contents.prop.value.string = "RIGHT";
		    } else {
			OrderProp.contents.prop.value.string = "LEFT";
		    }
		}
	    }	
	}
	if( octGetById( &PTerm ) != OCT_OK ){
	    octError( "Getting the PseudoTerm Bag \n");
	    myExit( -1, __FILE__, __LINE__);
	}
	if( octCreate( &PTerm, &OrderProp ) != OCT_OK ){
	    octError( "Creating the OrderProp property \n");
	    myExit( -1, __FILE__, __LINE__);
	}
    }
}

