#include "cds.h"

#ifndef OCTTOOLS_COPYRIGHT_H
#define OCTTOOLS_COPYRIGHT_H
/*
 * Oct Tools Distribution 3.0
 *
 * Copyright (c) 1988, 1989, Regents of the University of California.
 * All rights reserved.
 *
 * Use and copying of this software and preparation of derivative works
 * based upon this software are permitted.  However, any distribution of
 * this software or derivative works must include the above copyright
 * notice.
 *
 * This software is made available AS IS, and neither the Electronics
 * Research Laboratory or the University of California make any
 * warranty about the software, its performance or its conformity to
 * any specification.
 *
 * Suggestions, comments, or improvements are welcome and should be
 * addressed to:
 *
 *   octtools@eros.berkeley.edu
 *   ..!ucbvax!eros!octtools
 */

#if !defined(lint) && !defined(SABER)
static char copyright[] = "Copyright (c) 1988, Regents of the University of California.  All rights reserved.";
#endif
#endif 

#include "Numbers.h"

extern int		debug ; 
extern FILE		*dump ;
extern int 		numPoints ;
extern int 		numMagicPoints ;
extern int              channelIndex[MAX_NUM_POINTS] ;
extern struct channel	*arrayOfChannels[MAX_NUM_CHANNELS];
extern struct point	magicPoints[MAX_NUM_POINTS];
extern octCoord		maxSpacing, spacing_M1, spacing_M2, spacing_PO;
extern struct octPoint 	chipCenter;
int			mulProjection = FALSE;  /*  inserted because Mitch is not ready
						    to handle this */





assignTermToChannel(ptoTerm)
struct terminal		*ptoTerm;
{
    struct channel 	*closestChannel[5], *getGRChannel();
    struct terminal	*presentTerm, *term, *createGRTerm();
    struct point	*point[5];
    octId		getCDchannelId();
    int 		i, j, J, I, l;
    int			counter;
    int			channelFound;
    int 		findChannel();
    int			direction;


    if(debug) fprintf( dump, "entering assignTermToChannel\n");

    for(presentTerm = ptoTerm ; presentTerm != NIL(struct terminal) ;
	presentTerm = presentTerm->equivTerm){	
	if( presentTerm->routeEquivTerm == TRUE ) continue;


	/* 	If terminals are inside the cell this  projection 
		may cause mistakes      */


	/* 	there are two cases: 
		1) the terminal has at least one corner coincident
		   with a magic point; in this case the terminal
		   has to be assigned to two (only one magic point)
		   or to three (two magic points).
		2) the terminal has no magic points.
		   check to see if it is on the border of the cell
		   or inside it. if it is inside find out where to
		   project it on the perifery.
		   then find the channel with the standard procedure

	*/

	/*     	check if any ot the magic points coincides with the 
		corners of the terminal 

	*/
	counter = 0;
	for( i = 0; i < numMagicPoints ; i++ ){
	    if( (magicPoints[i].ptoPoint->x == presentTerm->lowerLeft->x ) &&
		(magicPoints[i].ptoPoint->y == presentTerm->lowerLeft->y ) ){
		point[counter++] = &(magicPoints[i]);
		continue;
	    }
	    if( (magicPoints[i].ptoPoint->x == presentTerm->upperRight->x ) &&
		(magicPoints[i].ptoPoint->y == presentTerm->upperRight->y ) ){
		point[counter++] = &(magicPoints[i]);
		continue;
	    }
	    if( (magicPoints[i].ptoPoint->x == presentTerm->upperRight->x ) &&
		(magicPoints[i].ptoPoint->y == presentTerm->lowerLeft->y ) ){
		point[counter++] = &(magicPoints[i]);
		continue;
	    }
	    if( (magicPoints[i].ptoPoint->x == presentTerm->lowerLeft->x ) &&
		(magicPoints[i].ptoPoint->y == presentTerm->upperRight->y ) ){
		point[counter++] = &(magicPoints[i]);
		continue;
	    }
	}
	 
	channelFound = FALSE;
	I = 0;
	if( counter > 0 ){
	    if( mulProjection == TRUE ){
		closestChannel[I] = getGRChannel(point[0]->firstChannel, presentTerm, ANY_DIRECTION);
		channelFound = attachTerm(presentTerm, closestChannel[I++]);
		term = createGRTerm( presentTerm, 0 );
		closestChannel[I] = getGRChannel(point[0]->secondChannel, term, ANY_DIRECTION);
		channelFound = attachTerm(presentTerm, closestChannel[I++]);
		for( i = 1 ; i < counter && i < 3 ; i++ ){
		/*	first add the route terminal to the list of terminals */
		/* then find out the surrounding channels */
		    if( ( point[i]->firstChannel != point[i-1]->firstChannel ) &&
			( point[i]->firstChannel != point[i-1]->secondChannel ) ){
			term = createGRTerm( presentTerm, i );
			closestChannel[I] = getGRChannel(point[i]->firstChannel, term, ANY_DIRECTION);
			channelFound = attachTerm(presentTerm, closestChannel[I++]);
		    }
		    if( ( point[i]->secondChannel != point[i-1]->firstChannel ) &&
			( point[i]->secondChannel != point[i-1]->secondChannel ) ){
			term = createGRTerm( presentTerm, i );
			closestChannel[I] = getGRChannel(point[i]->secondChannel, term, ANY_DIRECTION);
			channelFound = attachTerm(presentTerm, closestChannel[I++]);
		    }
		}
	    } else {
		/* This code has been added to correctly project terminals that have at
		   least one of their corners coincident with a magic point.
		   The terminal will be projected on the side where the terminal is
		   tangent to the bounding box of the instance. This is to be removed
		   when cprep is capable of handling multiple projections.*/

		for( l = 0 ; l < counter; l++ ){
		    for( J = 0 ; J < numPoints ; J++ ){
			j = channelIndex[J];
			if( arrayOfChannels[j]->info->id == point[l]->firstChannel ){
			    closestChannel[I] = getGRChannel(point[l]->firstChannel, presentTerm, ANY_DIRECTION);
			} else if( arrayOfChannels[j]->info->id == point[l]->secondChannel ){
			    closestChannel[I] = getGRChannel(point[l]->secondChannel, presentTerm, ANY_DIRECTION);
			} else {
			    continue;
			}
			channelFound = FALSE;
			for( i = 0 ; i < I ; i++ ){
			    if( closestChannel[i] == closestChannel[I] ){
				channelFound = TRUE;
				break;
			    }
			}
			if( channelFound == FALSE ) I++;
		    }
		}
		if( I == 0 ){
		    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
		    fprintf( stderr, "Terminal %s < %d, %d > of instance %s\n" , 
				presentTerm->termName, 
				presentTerm->center->x, presentTerm->center->y, 
				presentTerm->instInfo->name);
		    fprintf( stderr, "The terminal is at the corner and cannot be projected to any of the sides of the instance\n");
		    fprintf( stderr, "To convince yourself, take a look at the interface of the master of cell %s\n", 
			    presentTerm->instInfo->name);
		    myExit( -1, __FILE__, __LINE__ );
		}

		direction = -1;
		channelFound = -1;
		for( i = 0 ; i < counter ; i++ ){
		    /* Find the side of the terminal which is tangent to the BB of
		       the instance and then find out which one is the proper channel.  */
		     
		    if( point[i]->ptoPoint->x == presentTerm->instInfo->left ){
			if( checkProjection( presentTerm, 0, WEST ) == OK ) {
			    direction = VERTICAL;
			    for( j = 0 ; j < I ; j++ ){
				if( ( closestChannel[j]->info->id == point[i]->firstChannel ||
				      closestChannel[j]->info->id == point[i]->secondChannel ) &&
				    closestChannel[j]->info->type == direction &&
				    closestChannel[j]->info->outerChannel == FALSE ){
				    if( channelFound == -1 ){
					channelFound = j;
				    } else if( closestChannel[channelFound]->info->channelNumber >
					closestChannel[j]->info->channelNumber ) {
					channelFound = j;
				    }
				    break;
				}
			    }
			}
		    } else if( point[i]->ptoPoint->x == presentTerm->instInfo->right ){
			if( checkProjection( presentTerm, 0, EAST ) == OK ) {
			    direction = VERTICAL;
			    for( j = 0 ; j < I ; j++ ){
				if( ( closestChannel[j]->info->id == point[i]->firstChannel ||
				      closestChannel[j]->info->id == point[i]->secondChannel ) &&
				    closestChannel[j]->info->type == direction &&
				    closestChannel[j]->info->outerChannel == FALSE ){
				    if( channelFound == -1 ){
					channelFound = j;
				    } else if( closestChannel[channelFound]->info->channelNumber >
				    	closestChannel[channelFound]->info->channelNumber >
					closestChannel[j]->info->channelNumber ) {
					channelFound = j;
				    }
				    break;
				}
			    }
			}
		    } 
		    if( point[i]->ptoPoint->y == presentTerm->instInfo->top ){
			if( checkProjection( presentTerm, 0, NORTH ) == OK ) {
			    direction = HORIZONTAL; 
			    for( j = 0 ; j < I ; j++ ){
				if( ( closestChannel[j]->info->id == point[i]->firstChannel ||
				      closestChannel[j]->info->id == point[i]->secondChannel ) &&
				    closestChannel[j]->info->type == direction &&
				    closestChannel[j]->info->outerChannel == FALSE ){
				    if( channelFound == -1 ){
					channelFound = j;
				    } else if( closestChannel[channelFound]->info->channelNumber >
					closestChannel[j]->info->channelNumber ) {
					channelFound = j;
				    }
				    break;
				}
			    }
			}
		    } else if( point[i]->ptoPoint->y == presentTerm->instInfo->bottom ){
			if( checkProjection( presentTerm, 0, SOUTH ) == OK ) {
			    direction = HORIZONTAL; 
			    for( j = 0 ; j < I ; j++ ){
				if( ( closestChannel[j]->info->id == point[i]->firstChannel ||
				      closestChannel[j]->info->id == point[i]->secondChannel ) &&
				    closestChannel[j]->info->type == direction &&
				    closestChannel[j]->info->outerChannel == FALSE ){
				    if( channelFound == -1 ){
					channelFound = j;
				    } else if( closestChannel[channelFound]->info->channelNumber >
					closestChannel[j]->info->channelNumber ) {
					channelFound = j;
				    }
				    break;
				}
			    }
			}
		    } else {
			continue;
		    }
		}
		if( channelFound == -1 ) channelFound = 0;
		channelFound = attachTerm(presentTerm, closestChannel[channelFound]);
	    }
	}
	if( counter == 0 || channelFound == FALSE ){
	    channelFound = findChannel(presentTerm);
	}
	if( channelFound == FALSE ){
	    fprintf( stderr, "MOSAICO_GR_ERROR:\n");
	    fprintf( stderr, "Terminal %s < %d, %d > of instance %s\n" , 
			presentTerm->termName, 
			presentTerm->center->x, presentTerm->center->y, 
			presentTerm->instInfo->name);
	    fprintf( stderr, "The terminal cannot be projected to any of the sides of the instance\n");
	    fprintf( stderr, "To convince yourself, take a look at the interface of the master of cell %s\n", 
		    presentTerm->instInfo->name);
	    myExit( -1, __FILE__, __LINE__ );
	}
    }
    if(debug) fprintf( dump, "leaving assign\n");
}



struct terminal *createGRTerm( presentTerm, i)
struct terminal *presentTerm;
int i;
{
    struct terminal *term;
    term = ALLOC(struct terminal, 1);
    *term = *presentTerm;
    term->name = ALLOC(char, strLen(presentTerm->name) + 4 );
    sprintf( term->name,"%s_%d", presentTerm->name, i );
    term->termName = util_strsav(presentTerm->termName);
    term->firstNetInList = NIL(struct netList);
    term->equivTerm = presentTerm->equivTerm;
    term->routeEquivTerm = TRUE;
    presentTerm->equivTerm = term;
    return(term);
}


attachTerm(presentTerm, closestChannel)
struct terminal *presentTerm;
struct channel *closestChannel;
{
    struct termList 	*newTerm = NIL(struct termList), *prsnTerm;
    int			findDistance();
    struct octBox	box, box1, box2, *makeRegion();
    octCoord		spacing;
    int			direction, i, numAttach = 0;
    struct layerList    *layerLst;
    octObject		DirProp, LayerProp, Bag, Obj, ATerm;
    octGenerator	objGen;
    octStatus		status;	/* added by Andrea */
    /*	check if the terminal projected on this channel interferes with any of the
	other terminals already projected */

    
    DirProp.type = OCT_PROP;
    DirProp.contents.prop.name = "Direction";
    DirProp.contents.prop.type = OCT_INTEGER;
    LayerProp.type = OCT_PROP;
    LayerProp.contents.prop.name = "LayerName";
    LayerProp.contents.prop.type = OCT_STRING;
    Bag.type = OCT_BAG;
    Bag.contents.bag.name = "MosaicoTermInfo";
    ATerm.type = OCT_TERM;

    direction = dir( presentTerm, closestChannel );
    DirProp.contents.prop.value.integer = direction;
    for( i = 0 ; i < NUM_LAYERS ; i++ ){
	if( ( layerLst = presentTerm->direction[direction].layer[i] ) == NIL(struct layerList)){
	    continue;
	}
	if( strCmp( layerLst->layerName, "MET1" ) == 0 ){
	    spacing = spacing_M1;
	} else if( strCmp( layerLst->layerName, "MET2") == 0){
	    spacing = spacing_M2;
	} else if( strCmp( layerLst->layerName, "POLY") == 0){
	    spacing = spacing_PO;
	}
	box.lowerLeft.x = presentTerm->lowerLeft->x;
	box.lowerLeft.y = presentTerm->lowerLeft->y;
	box.upperRight.x = presentTerm->upperRight->x;
	box.upperRight.y = presentTerm->upperRight->y;

	box1 = *makeRegion( &(box), spacing, presentTerm->direction[direction].distance, 
				direction, 1);
	for( prsnTerm = closestChannel->ptoFirstTerm; prsnTerm != NIL(struct termList);
	     prsnTerm = prsnTerm->ptoNextTerm){
	    if( prsnTerm->ptoTerm->actualTermId == presentTerm->actualTermId ){
		continue;
	    }
	    box2.lowerLeft.x = prsnTerm->ptoTerm->lowerLeft->x;
	    box2.lowerLeft.y = prsnTerm->ptoTerm->lowerLeft->y;
	    box2.upperRight.x = prsnTerm->ptoTerm->upperRight->x;
	    box2.upperRight.y = prsnTerm->ptoTerm->upperRight->y;
	    if( boxIntersect( &box1, &box2 ) == YES ){
		/* now check if the terminals that intersect are joined:
		    if they are, no problems return OK.
		    if they are not return KO  */
		if( presentTerm->jointBagId != prsnTerm->ptoTerm->jointBagId ){
		    presentTerm->direction[direction].layer[i] = NIL(struct layerList);
		    break;
		} 
	    }
	}
	if( presentTerm->direction[direction].layer[i] == NIL(struct layerList) ) continue;
	if( newTerm == NIL(struct termList) ){
	    newTerm = ALLOC(struct termList, 1);
	    newTerm->ptoTerm = presentTerm;
	    newTerm->ptoNextTerm = closestChannel->ptoFirstTerm;
	    closestChannel->ptoFirstTerm = newTerm;
	    closestChannel->info->numberOfTerms++;
	    if ( closestChannel->info->type == VERTICAL ){
		presentTerm->firstNode = closestChannel->firstNode;
		presentTerm->secondNode = closestChannel->secondNode;
		presentTerm->distance = closestChannel->info->lengthFactor * 
					findDistance( presentTerm, VERTICAL );
		presentTerm->channelNumber = closestChannel->number;
		if( debug ){
		    fprintf( dump, "Terminal %s assigned to the VERTICAL channel\n",
			presentTerm->termName );
		    channelInfo(closestChannel);
		}
	    }
	    if ( closestChannel->info->type == HORIZONTAL ){
		presentTerm->firstNode = closestChannel->firstNode;
		presentTerm->secondNode = closestChannel->secondNode;
		presentTerm->distance = closestChannel->info->lengthFactor * 
					findDistance( presentTerm, HORIZONTAL );
		presentTerm->channelNumber = closestChannel->number;
		if( debug ){
		    fprintf( dump, "Terminal %s assigned to the HORIZONTAL channel \n",
			presentTerm->termName );
		    channelInfo(closestChannel);
		}
	    }
	}
    }
    ATerm.objectId = presentTerm->actualTermId;
    if( octGetById( &ATerm ) != OCT_OK ){
	octError("Getting the Aterm");
	myExit( -1, __FILE__, __LINE__);
    }
    if( octGetOrCreate( &ATerm, &Bag ) != OCT_OK ){
	octError("Getting the MosInfBag");
	myExit( -1, __FILE__, __LINE__);
    }

    if( octInitGenContents( &Bag, OCT_ALL_MASK, &objGen ) != OCT_OK ){
	octError("Generating the contents of the MosInfBag");
	myExit( -1, __FILE__, __LINE__);
    }
    while( (status = octGenerate( &objGen, &Obj )) == OCT_OK ){
	if( Obj.type == OCT_PROP ) {
	    if( octDelete( &Obj ) == OCT_ERROR ){
		octError("Deleting the contents of the MosInfBag");
		myExit( -1, __FILE__, __LINE__);
	    }
	} else {
	    if( octDetach( &Bag, &Obj ) == OCT_ERROR ){
		octError("Detaching the contents of the MosInfBag");
		myExit( -1, __FILE__, __LINE__);
	    }
	}
    }
    if (status != OCT_GEN_DONE) {
	octError("Generator failed");
	myExit( -1, __FILE__, __LINE__);
    }
    if( octCreate( &Bag, &DirProp ) != OCT_OK ){
	octError("Creating the contents (direction) of the MosInfBag");
	myExit( -1, __FILE__, __LINE__);
    }
    for( i = 0 ; i < NUM_LAYERS ; i++ ){
	if( ( layerLst = presentTerm->direction[direction].layer[i] ) == NIL(struct layerList)) continue;
	LayerProp.contents.prop.value.string = util_strsav( layerLst->layerName );
	if( octCreate( &Bag, &LayerProp ) != OCT_OK ){
	    octError("Creating the contents (layer) of the MosInfBag");
	    myExit( -1, __FILE__, __LINE__);
	}
	numAttach++;
    }
    if( numAttach > 0 )
	return(OK);
    else
	return(KO);
}



int findDistance( pTerm, direction )
struct terminal *pTerm;
int		direction;
{
    int distance;

    if( direction == HORIZONTAL ){
	/* first check if the terminal is inside the segment */
	if( pTerm->firstNode->x < pTerm->secondNode->x ){
	    if( ( pTerm->center->x >= pTerm->firstNode->x ) &&
		( pTerm->center->x <= pTerm->secondNode->x ) ) {
		distance = pTerm->center->x - pTerm->firstNode->x;
		pTerm->normalProj = TRUE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->x >= pTerm->secondNode->x ) {
		distance = pTerm->secondNode->x - pTerm->firstNode->x - 1;
		pTerm->normalProj = FALSE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->x <= pTerm->firstNode->x ) {
		distance = 1;
		pTerm->normalProj = FALSE;
	    }
	} else {
	    if( ( pTerm->center->x <= pTerm->firstNode->x ) &&
		( pTerm->center->x >= pTerm->secondNode->x ) ) {
		distance = pTerm->firstNode->x - pTerm->center->x;
		pTerm->normalProj = TRUE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->x >= pTerm->firstNode->x ) {
		distance = 1;
		pTerm->normalProj = FALSE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->x <= pTerm->secondNode->x ) {
		distance = pTerm->firstNode->x - pTerm->secondNode->x - 1;
		pTerm->normalProj = FALSE;
	    }
	}
    } else {
	if( pTerm->firstNode->y < pTerm->secondNode->y ){
	    if( ( pTerm->center->y >= pTerm->firstNode->y ) &&
		( pTerm->center->y <= pTerm->secondNode->y ) ) {
		distance = pTerm->center->y - pTerm->firstNode->y;
		pTerm->normalProj = TRUE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->y >= pTerm->secondNode->y ) {
		distance = pTerm->secondNode->y - pTerm->firstNode->y - 1;
		pTerm->normalProj = FALSE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->y <= pTerm->firstNode->y ) {
		distance = 1;
		pTerm->normalProj = FALSE;
	    }
	} else {
	    if( ( pTerm->center->y <= pTerm->firstNode->y ) &&
		( pTerm->center->y >= pTerm->secondNode->y ) ) {
		distance = pTerm->firstNode->y - pTerm->center->y;
		pTerm->normalProj = TRUE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->y >= pTerm->firstNode->y ) {
		distance = 1;
		pTerm->normalProj = FALSE;
	    } /* to account for Z-shaped channels */
	    if( pTerm->center->y <= pTerm->secondNode->y ) {
		distance = pTerm->firstNode->y - pTerm->secondNode->y - 1;
		pTerm->normalProj = FALSE;
	    }
	}
    }
    return(distance);
}



struct channel *getGRChannel(id, presentTerm, direction)
octId id;
struct terminal *presentTerm;
int direction;   /* added to account for z-shaped channels */
{
    struct channel *closestChannel;
    int j, J, distance, maxDistance = MAX_INT;

    closestChannel = NIL(struct channel);
    for( J = 0 ; J < numPoints ; J++ ){
	j = channelIndex[J];
	if( arrayOfChannels[j]->info->id == id ){
	    if( ( direction == NORTH || direction == SOUTH ) && 
		arrayOfChannels[j]->info->type == VERTICAL  ) continue;
	    if( ( direction == EAST || direction == WEST ) && 
		arrayOfChannels[j]->info->type == HORIZONTAL  ) continue;
	    if( arrayOfChannels[j]->info->type == HORIZONTAL ){
		if( (presentTerm->center->x >= arrayOfChannels[j]->firstNode->x  &&
		     presentTerm->center->x <= arrayOfChannels[j]->secondNode->x ) ||
		    (presentTerm->center->x >= arrayOfChannels[j]->secondNode->x  &&
		     presentTerm->center->x <= arrayOfChannels[j]->firstNode->x )){
		    closestChannel = arrayOfChannels[j];
		    break;
		} else { /* to account for Z-shaped channels */
		    distance = ABS( arrayOfChannels[j]->firstNode->x - presentTerm->center->x );
		    if( maxDistance > distance ){
			maxDistance = distance;
			closestChannel = arrayOfChannels[j];
		    }
		    distance = ABS( arrayOfChannels[j]->secondNode->x - presentTerm->center->x );
		    if( maxDistance > distance ){
			maxDistance = distance;
			closestChannel = arrayOfChannels[j];
		    }
		}
	    }
	    if( arrayOfChannels[j]->info->type == VERTICAL ){
		if( (presentTerm->center->y >= arrayOfChannels[j]->firstNode->y  &&
		     presentTerm->center->y <= arrayOfChannels[j]->secondNode->y ) ||
		    (presentTerm->center->y >= arrayOfChannels[j]->secondNode->y  &&
		     presentTerm->center->y <= arrayOfChannels[j]->firstNode->y )){
		    closestChannel = arrayOfChannels[j];
		    break;
		} else { /* to account for Z-shaped channels */
		    distance = ABS( arrayOfChannels[j]->firstNode->y - presentTerm->center->y );
		    if( maxDistance > distance ){
			maxDistance = distance;
			closestChannel = arrayOfChannels[j];
		    }
		    distance = ABS( arrayOfChannels[j]->secondNode->y - presentTerm->center->y );
		    if( maxDistance > distance ){
			maxDistance = distance;
			closestChannel = arrayOfChannels[j];
		    }
		}
	    }
	}
    }
    return(closestChannel);
}
