#include <stdio.h>
#include "copyright.h"
#include "port.h"

#define XId
#include "rpc.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#define STR_COPY(s) strcpy(malloc(strlen(s)+1),s)

char	errmsg[256];

/*
static IVWindow* ivwin = 0;
static int start = 0;
*/
/* Global variables to tune behavior of the monitor */
static int	maxTries = 200;
static int	sleepTime = 2;

#define	DEMOV_CELLNAME_DIM 1000
Display* disp;			/* The X connection */
int       stopFlag;		/* The flag to stop the monitor */

int	demovRpcResume(), demovRpcStart(), demovRpcExit();
int	demovRpcNext();
int	demovDebug();
int     demovRpcDiskConsistency();
char	demovFileName[256];
char*	demovCellName[DEMOV_CELLNAME_DIM];
char*	demovViewName[DEMOV_CELLNAME_DIM];
char*	demovCwd;
int	demovNumCell;
int	demovCellIdx;
int 	debug = 0;

RPCFunction CommandArray[] = {
    /* {demovRpcStart,	"mosaicoMonitor", "start-monitoring"}, */
    {demovRpcNext,	"mosaicoMonitor", "next" , "N", 0},
    {demovRpcResume,	"mosaicoMonitor", "resume-monitoring"},
    {demovRpcDiskConsistency,	"mosaicoMonitor", "auto-re-read"},
    {demovDebug,	"mosaicoMonitor", "debug"},
    {demovRpcExit,	"mosaicoMonitor", "exit"}
};


time_t getTimestamp( file )
    char* file;
{
    struct stat buf;
    if (debug ) fprintf( stderr, "checking %s", file );
    if ( stat( file, &buf ) == -1 ) {
	if (debug) perror("XX"); 
	return 0;
    } else {
	if (debug) fprintf( stderr, "\t(TS:%d)\n", buf.st_mtime );
	return buf.st_mtime;
    }
}


/*ARGSUSED*/
long UserMain(display, spot, cmdList, uow, array)
    char		*display;
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
    RPCFunction		**array;
{
    demovNumCell = -1;

    VEMMSG("mosaicoMonitor is now ready.\n");
    vemPrompt();
    *array = CommandArray;
    return(sizeof(CommandArray) / sizeof(RPCFunction));

}

/*ARGSUSED*/
demovRpcExit(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    RPCExit( RPC_OK );
}

int demovDebug(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    debug = !debug;
    VEMMSG("Debugging is ");
    VEMMSG( ( debug ) ? "ON\n" : "OFF\n" );
    vemPrompt();
    return RPC_OK;
}


char* demovGetWorkingDirectory( FacetP )
    octObject *FacetP;
{
  octObject Prop;
  char* cwd = 0;

  Prop.type = OCT_PROP;
  Prop.contents.prop.name = "MOSAICO_DEMO_CWD";
  Prop.contents.prop.type = OCT_STRING;
  if (octGetByName (FacetP, &Prop) != OCT_OK) {
    VEMMSG("MOSAICO_DEMO_CWD property was not found\n");
    return 0;
  } else {
    cwd = STR_COPY(Prop.contents.prop.value.string); 
    return cwd;
  }
}



int
readSequenceFile( file )
    char* file;
{
    FILE* 	f;
    char	cell[250], view[250];
    char*	viewName = "symbolic";
    char 	input_buffer[256];

    if (debug) fprintf(stderr,"file name = %s\n", file);
    f = fopen( file, "r");
    if (!f) {
	return 0;
    }

    demovNumCell = 0;
    demovCellIdx = 0;

    while (fscanf(f,"%s", input_buffer)!=EOF) {
	if (demovNumCell+1>=DEMOV_CELLNAME_DIM) {
	    return 0;
	}
	if (index(input_buffer,':')) {
	    sscanf(input_buffer,"%[^:]:%s", cell, view);
	    demovCellName[demovNumCell] = STR_COPY(cell);
	    demovViewName[demovNumCell] = STR_COPY(view);
	} else {
	    demovCellName[demovNumCell] = STR_COPY(input_buffer);
	    demovViewName[demovNumCell] = STR_COPY(viewName);
	}
	demovNumCell++;
    }
    return demovNumCell;
}

/*ARGSUSED*/
int
getSequenceFile( spot, cmdList, uow )
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    static int  sequenceRead = 0;
    RPCArg	*arg;
    char  	buf[256];

    if ( sequenceRead ) return RPC_OK;

    if (lsLength(cmdList) > 1) {
	VEMMSG("ERROR: [\"file\"] : sequence\nmosaicoMonitor> ");
	return 0;
    }
    if (lsLength(cmdList) == 1) {
  	lsFirstItem(cmdList, (Pointer *) &arg, (lsHandle *) 0);
	if (arg->argType != VEM_TEXT_ARG) {
	    VEMMSG("ERROR: [\"string\"] sequence\nmosaicoMonitor> ");
	    return 0;
	}
	sprintf(demovFileName,"%s",arg->argData.string);
    } else {
	octObject Facet;
	Facet.objectId = spot->facet;
	octGetById( &Facet );
	demovCwd = demovGetWorkingDirectory( &Facet );
	sprintf(demovFileName,"%s/mosaicoDemoRpc.seq", demovCwd) ;
    }

    demovNumCell = readSequenceFile ( demovFileName );
    sprintf(buf,"Read in sequence file %s\n (number of cells = %d)\nmosaicoMonitor> ",
	    demovFileName, demovNumCell);
    VEMMSG(buf);
    sequenceRead = 1;
    return 1 ;
}


void waitStableFile( file )
    char *file;
{
    int	timestamp = getTimestamp( file );
    int changingTime;
    do {
	changingTime = timestamp;
	sleep( 2 );		/* Allow time to stop writing */
	timestamp = getTimestamp( file );
    } while (changingTime != timestamp );
    return;
}



void reread( spot )
    RPCSpot *spot;
{
    octObject Facet;
    lsList	vCmdList;

    Facet.objectId = spot->facet;
    if ( octGetById( &Facet ) != OCT_OK ) {
	octError("in reread: getbyid");
	RPCExit( RPC_ERROR );
    }

    if (debug) fprintf(stderr, "Rereading %s:%s \n" , Facet.contents.facet.cell, Facet.contents.facet.view );

    vCmdList = lsCreate();
    (void)vemCommand("rr-nc", spot, vCmdList, 0);
    (void)vemWnFlush();
    
    Facet.contents.facet.mode = "r";
    if (octOpenFacet( &Facet) < OCT_OK) {
	VEMMSG("MosaicoMonitor:Cannot get new facet after reread");
	octError("mosaicomonitor:during reread");
	RPCExit( RPC_ERROR );
    }
    /* Gotta get the new Id for the facet */
    spot->facet = Facet.objectId;
    octFreeFacet( &Facet );	/* Do not use close to avoid write out to disk */
}

void switchfacet( spot , cell, view)
    RPCSpot *spot;
    char *cell;
    char *view;
{
    char cell_view[256];
    lsList	vCmdList;
    RPCArg	arg;
    octObject Facet;

    if (debug) fprintf(stderr, "Switching to %s:%s\n" , cell, view );

    sprintf(cell_view,"%s:%s",	  cell, view );
    vCmdList = lsCreate();
    arg.argType = VEM_TEXT_ARG;
    arg.argData.string = cell_view;
    lsNewBegin(vCmdList, (Pointer) &arg, (lsHandle *) 0);
    (void)vemCommand("switch-facet", spot, vCmdList, 0);
    (void)vemWnFlush();

    Facet.type = OCT_FACET;
    Facet.contents.facet.cell = cell;
    Facet.contents.facet.view = view;
    Facet.contents.facet.facet = "contents";
    Facet.contents.facet.version = OCT_CURRENT_VERSION;
    Facet.contents.facet.mode = "r";
    if ( octOpenFacet( &Facet) < OCT_OK ) {
	VEMMSG("MosaicoMonitor:Cannot get new facet after switch");
	RPCExit( RPC_ERROR );
    } ;	/* Gotta get the new Id for the facet */
    spot->facet = Facet.objectId;
    octFreeFacet( &Facet ); /* Do not use close to avoid write out to disk */
}

int demovMonitor( spot )
    RPCSpot *spot;
{
    time_t 	timestamp;
    time_t      curfacetTS;	/* Actual timestamp for the current facet */
    time_t	virtualTS;	/* VirtualTS for the current facet */
    int		k;
    char	curfName[1024];	/* For the current facet */
    char	newfName[1024];	/* For the next facet to be displayed */

    octObject Facet;

    /* Initialize: Get current facet and its timestamp */
    Facet.objectId = spot->facet;
    if (  octGetById( &Facet ) != OCT_OK ) {
	octError("Cannot get facet in monitor");
    }
    sprintf( curfName, "%s/%s/%s/contents;", demovCwd, Facet.contents.facet.cell, Facet.contents.facet.view );
    curfacetTS = getTimestamp( curfName );
    virtualTS = 0;		/*  */

    stopFlag = 0;

    for ( ; demovCellIdx<demovNumCell; demovCellIdx++) {
	sprintf(newfName,"%s/%s/%s/contents;", demovCwd, demovCellName[demovCellIdx], demovViewName[demovCellIdx]);
	if (debug) fprintf(stderr,"Trying to open cell %s\n", newfName); 

	for ( k = 0; k<maxTries; k++ ) {
	    if ( curfacetTS != getTimestamp( curfName ) ) {
		waitStableFile ( curfName );
		curfacetTS =  getTimestamp( curfName );
		reread( spot );
		k = 0;		/* restart the loop */
		sleep( (unsigned)sleepTime );
		continue;
	    }
	    timestamp = getTimestamp( newfName );
	    if (timestamp >= virtualTS ) {
		waitStableFile( newfName );
		sprintf(curfName, "%s/%s/%s/contents;", demovCwd, demovCellName[demovCellIdx], demovViewName[demovCellIdx]);
		virtualTS = getTimestamp( curfName );
		switchfacet( spot, demovCellName[demovCellIdx], demovViewName[demovCellIdx]);
		break;
	    }
	    sleep( (unsigned)sleepTime );
	    if ( stopFlag == 1 ) {
		break;
	    }
	}
	if ( k == maxTries ) {
	    if ( timestamp == 0) {
		sprintf(errmsg, "Could not open file %s (resume scan again)\nmosaicoMonitor> ", newfName);
		VEMMSG(errmsg);
		break;
	    } else {
		sprintf(errmsg, "Opening %s even if it is not new", newfName);
		VEMMSG( errmsg );
		sprintf(curfName, "%s/%s/%s/contents;", demovCwd, demovCellName[demovCellIdx], demovViewName[demovCellIdx]);
		virtualTS = getTimestamp( curfName );
		switchfacet( spot, demovCellName[demovCellIdx], demovViewName[demovCellIdx]);
	    }
	}
	sleep( (unsigned)sleepTime );
	curfacetTS = getTimestamp( curfName );
    }
    return 1;
}

/*ARGSUSED*/
demovRpcStart(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    if ( getSequenceFile( spot, cmdList, uow) ) {
	demovCellIdx = 0;
	demovMonitor( spot );
    }
    VEMMSG("Monitoring completed\n");
    return RPC_OK;
}

demovRpcNext(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    static index = 0;
    char	newfName[1024];	/* For the next facet to be displayed */
    char* cell;
    char* view;
    char cell_view[128];
    lsList	vCmdList;
    RPCArg	arg;

    if ( getSequenceFile( spot, cmdList, uow) ) {
	demovCellIdx = 0;
    }
    cell = demovCellName[index];
    view = demovViewName[index];
    sprintf(cell_view,"%s:%s",	  cell, view );

    sprintf(errmsg, "Opening %s\n", cell_view);
    VEMMSG( errmsg );

    vCmdList = lsCreate();
    arg.argType = VEM_TEXT_ARG;
    arg.argData.string = cell_view;
    lsNewBegin(vCmdList, (Pointer) &arg, (lsHandle *) 0);
    (void)vemCommand("switch-facet", spot, vCmdList, 0);
    (void)vemWnFlush();

    index++;
    cell = demovCellName[index];
    view = demovViewName[index];
    sprintf( errmsg, "NEXT facet will be N%s:%s\n", cell, view );
    VEMMSG( errmsg );
    vemPrompt();
    return RPC_OK;
}

demovRpcResume(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    if ( getSequenceFile( spot, cmdList, uow) ) {
	demovMonitor( spot );
    }
    VEMMSG("Monitoring completed\n");
    return RPC_OK;
}


int
demovRpcDiskConsistency(spot, cmdList, uow)
    RPCSpot		*spot;
    lsList		cmdList;
    long		uow;
{
    time_t      curfacetTS;	/* Actual timestamp for the current facet */
    char	curfName[1024];	/* For the current facet */

    octObject Facet;
    char *fullName;
    char msg[256];

    /* Initialize: Get current facet and its timestamp */
    Facet.objectId = spot->facet;
    if (  octGetById( &Facet ) != OCT_OK ) {
	octError("Cannot get facet in monitor");
    }
    octFullName( &Facet, &fullName );
    sprintf( curfName, "%s/%s/contents;", fullName, Facet.contents.facet.view );
    curfacetTS = getTimestamp( curfName );

    stopFlag = 0;

    while( 1 ) {
	if ( curfacetTS != getTimestamp( curfName ) ) {
	    waitStableFile ( curfName );
	    sprintf( msg, "Auto Rereading %s...\n", fullName );
	    VEMMSG( msg );
	    reread( spot );
	    curfacetTS =  getTimestamp( curfName );
	}
	sleep( (unsigned)sleepTime );
	if ( stopFlag == 1 ) {
	    break;
	}
    }
    return RPC_OK;
}

