#include "copyright.h"
#include "port.h"
#include "checks.h"
#include "oh.h"

#define			MAX_NUM_GEOS 	10
extern int		debug ; 
extern int		numErrors, maxNumErrors;
extern octObject	Facet;
extern octObject	MarkBag;
extern char		*reportFileName;
int			levelCounter = 0;
char			msgs[1000];
extern int		silent;



getNewTerm( Net, Path, oldTermId )
octObject 	*Net, *Path;
octId 		oldTermId;
{
    octGenerator	termGen, genPath, termGenJoin, ftermGen;
    octGenerator	newNetGen;
    octObject		MasterBag, NewTerm, NewPath, JoinTerm, Fterm;
    octObject		MasterNewTerm, MasterJoinTerm, Instance;
    octObject		NewNet;

    int 		numPaths;
    int			numTerms;
    int 		status, counter;
    int			loopFound = FALSE;

    MasterBag.type = OCT_BAG;
    MasterBag.contents.bag.name = "JOINED";
    JoinTerm.type = OCT_TERM;


    /*   first get the new terminal attached to the path   */

    if (debug) FPRINTF( stderr, "Routine getNewTerm, level  %d \n",
				++levelCounter);
    OCT_INIT_GEN_CONTENTS( Path, OCT_TERM_MASK, &termGen, "Generator to get new terminal");
    numTerms = 0 ;
    while( octGenerate( &termGen, &NewTerm ) != OCT_GEN_DONE ){
	numTerms++;
	if (octIdsEqual(NewTerm.objectId, oldTermId)) continue;

	if (debug) FPRINTF(stderr, "Terminal %s (%s)\n",
			NewTerm.contents.term.name,
			octFormatId(NewTerm.objectId));
			
	checkConnection( &NewTerm, Path );
	if( markTerm( &NewTerm ) == FOUND ){
	    /* two cases have to be considered:
		1) a terminal is attached to more than one net;
		2) there is a loop in the net implementation;

	    */
	    OCT_INIT_GEN_CONTAINERS( &NewTerm, OCT_NET_MASK, &newNetGen,
				"Generator for the net");
	    counter = 0;
	    while( octGenerate( &newNetGen, &NewNet ) != OCT_GEN_DONE ){
		if( counter++ == 0 ){
		    reportError( &NewTerm, "Terminal connected to more than one net", YES );
		    FPRINTF( stderr, "Terminal %s is attached to the following nets: \n",
				NewTerm.contents.term.name);
		}
		FPRINTF( stderr, "Net %s\n", NewNet.contents.net.name);
		SPRINTF( msgs, "Errors on net %s", NewNet.contents.net.name );
		reportError(&NewNet, msgs, NO );
	    }
	    if( counter <= 1 ){
		SPRINTF( msgs, "Net %s\n Terminal is part of a loop", 
				Net->contents.net.name );
		reportError( &NewTerm, msgs, YES );
		loopFound = TRUE;
	    } else {
		loopFound = FALSE;
	    }
	    break;
	}

	/*   We have to check whether the terminal is contained by a joined bag.
	     If it is contained by the joined bag then the path generation 
	     must be continued from all the joined terms not yet touched.  */

	/*	 check if the terminal is in a joined bag */

	Instance.objectId = NewTerm.contents.term.instanceId;
	OCT_GET_BY_ID( &Instance, "Getting the instance by ID" );
	getMasterTerm( &NewTerm, &MasterNewTerm, &Instance );
	OCT_GET_CONTAINER_BY_NAME( &MasterNewTerm, &MasterBag , "Getting Joined Bag" );
	status = Mstatus;
	if( status == OCT_OK ){
	    OCT_INIT_GEN_CONTENTS( &MasterBag, OCT_TERM_MASK, &termGenJoin, "Generating joined terms" );
	    while( octGenerate( &termGenJoin, &MasterJoinTerm ) != OCT_GEN_DONE ){
		JoinTerm.contents.term.name = MasterJoinTerm.contents.term.name;
		OCT_GET_BY_NAME( &Instance, &JoinTerm, "Getting joinTerm by name");
		if (debug) FPRINTF(stderr, "Terminal %s (%s)\n",
				JoinTerm.contents.term.name,
				octFormatId(JoinTerm.objectId));
		if (!octIdsEqual(JoinTerm.objectId, NewTerm.objectId)) {
		    markTerm( &JoinTerm );
		}
		OCT_INIT_GEN_CONTAINERS( &JoinTerm, OCT_PATH_MASK, &genPath, "Generator to get paths");
		numPaths = 0;
		while( octGenerate( &genPath, &NewPath ) != OCT_GEN_DONE ){
		    numPaths++;
		    if (octIdsEqual(NewPath.objectId, Path->objectId)) continue;
		    OCT_GET_CONTAINER_BY_NAME( &NewPath, Net, "Getting the net for the path" );
		    if( Mstatus == OCT_NOT_FOUND ){
			
			SPRINTF( msgs, "Path not attached to net %s",
				Net->contents.net.name);
			reportError(&NewPath, msgs, YES);
		    }
		    checkConnection( &JoinTerm, &NewPath );
		    getNewTerm( Net, &NewPath, JoinTerm.objectId );
		}
	    }
	} else if( status == OCT_NOT_FOUND ){
	    OCT_INIT_GEN_CONTAINERS( &NewTerm, OCT_PATH_MASK, &genPath, "Generator to get paths");
	    numPaths = 0;
	    while( octGenerate( &genPath, &NewPath ) != OCT_GEN_DONE ){
		numPaths++;
		if (octIdsEqual(NewPath.objectId, Path->objectId)) continue;
		OCT_GET_CONTAINER_BY_NAME( &NewPath, Net, "Getting the net for the path" );
		if( Mstatus == OCT_NOT_FOUND ){
			
		    SPRINTF( msgs, "Path not attached to net %s",
			    Net->contents.net.name);
		    reportError(&NewPath, msgs, YES);
		}
		checkConnection( &NewTerm, &NewPath );
		getNewTerm( Net, &NewPath, NewTerm.objectId );
	    }
	    if ((numPaths < 2 ) && (instanceTerm( &NewTerm ) != YES)){
		/*	The terminal we have found has no joined terms and
			and there is only one path hanging off it.
			Three cases are possible:
			1) the terminal belongs to an instance
			2) the terminal is a connector but is used as a formal
			   terminal of the Facet
			3) the terminal is a connector but not a formal terminal
			   of the facet. Only in this case we have an error.  */
		/*	Only the third case is then checked   */

		OCT_INIT_GEN_CONTAINERS( &NewTerm, OCT_TERM_MASK, &ftermGen,
				"Creating the generator for Fterms");
		while( octGenerate( &ftermGen, &Fterm ) != OCT_GEN_DONE ){
		    if( octIsAttached( &Facet, &Fterm ) == OCT_NOT_FOUND ){
			SPRINTF(msgs, "The connector terminal %s has only one path attached",
					NewTerm.contents.term.name);
			reportError(&NewTerm, msgs, YES );
		    }
		}
	    }
	}
    }
    if( numTerms != 2 && loopFound == FALSE ){
	SPRINTF( msgs, "%d terminal(s) connected to the path", numTerms );
	reportError(Path, msgs, NO );
	reportError(&NewTerm, msgs, YES );
	if( debug ){
	    FPRINTF(stderr, "Net %s \n", Net->contents.net.name );
	    FPRINTF(stderr, "%d terminal(s) connected to a path\n",
		    numTerms);
	}
    }
    if (debug) FPRINTF( stderr, "Leaving getNewTerm, level %d\n",
				--levelCounter);
}



checkForFormal( Term , Net)
octObject *Term, *Net;
{
    octObject		ATerm, FTerm, InstanceBag, Instance;
    int 		status;

    InstanceBag.type = OCT_BAG;
    InstanceBag.contents.bag.name = "INSTANCES";
    Instance.type = OCT_INSTANCE;

    if (octIdIsNull(Term->contents.term.instanceId)) {
	/* Yes the term is a formal term for the cell. Find now the aterm attached to it */
	if( ( status = octGenFirstContent( Term, OCT_TERM_MASK, &ATerm) ) == OCT_NOT_FOUND ){
	    SPRINTF( msgs, "Term %s is a formTerm without ATerm\n",
			    Term->contents.term.name );
	    incrErrorCount(msgs, Term);
	    return;
	} else if( status == OCT_OK ){
	    Instance.objectId = ATerm.contents.term.instanceId;
	    (void) octGetById(&Instance);
	    if( octGetContainerByName( &Instance, &InstanceBag ) == OCT_OK ){
		if( ( status = octGetByName( Net, &ATerm ) ) == OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) but it is not attached to the Net\n",
				octFormatId(ATerm.objectId),
				octFormatId(Term->objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
	    } else if( status == OCT_NOT_FOUND ){
		if( ( status = octGetByName( Net, &ATerm ) ) != OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) but it is also attached to the net\n",
				octFormatId(ATerm.objectId),
				octFormatId(Term->objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
	    }
	    Term = &ATerm;
	    return;
	}
    } else {
	if( ( status = octGenFirstContainer( Term, OCT_TERM_MASK, &FTerm) ) == OCT_GEN_DONE ){
	    if( instanceTerm(Term) == YES ){
		OCT_GET_CONTAINER_BY_NAME( Term, Net, "Getting the net for the terminal" );
		if( Mstatus != OCT_OK ){
		    SPRINTF( msgs, "Instance terminal %s not connected to net %s",
				Term->contents.term.name,
				Net->contents.net.name );
		    reportError( Term, msgs, YES );
		}
	    } else {
		OCT_GET_CONTAINER_BY_NAME( Term, Net, "Getting the net for the terminal" );
		if( Mstatus == OCT_OK ){
		    SPRINTF( msgs, "Connector terminal %s connected to net %s",
				Term->contents.term.name,
				Net->contents.net.name );
		    reportError( Term, msgs, YES );
		}
	    }

	    return;
	} else if( status == OCT_OK ){
	    Instance.objectId = Term->contents.term.instanceId;
	    OCT_GET_BY_ID( &Instance, "Getting the instance");
	    octGenFirstContainer( Term, OCT_INSTANCE_MASK, &Instance);
	    if( ( status = octGenFirstContainer( &Instance, OCT_BAG_MASK, &InstanceBag)) == OCT_GEN_DONE ){
		if( octGetByName( Net, Term ) != OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) and it is attached to the net\n",
				octFormatId(Term->objectId),
				octFormatId(FTerm.objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
		if( octGetByName( Net, &FTerm ) == OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) and Fterm is not attached to the net\n",
				octFormatId(Term->objectId),
				octFormatId(FTerm.objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
	    } else {
		if( octGetByName( Net, Term ) == OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) and it is not attached to the net\n",
				octFormatId(Term->objectId),
				octFormatId(FTerm.objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
		if( octGetByName( Net, &FTerm ) == OCT_NOT_FOUND ) {
		    SPRINTF( msgs, "Symbolic policy violation: For details see file %s",
			    reportFileName);
		    reportError( Term, msgs, YES);
		    FPRINTF( stderr, "Aterm (%s) is attached to Fterm (%s) and Fterm is not attached to the net\n",
				octFormatId(Term->objectId),
				octFormatId(FTerm.objectId));
		    FPRINTF( stderr, "This is a violation of the Oct Symbolic Policy\n");
		}
	    }
	    return;
	}
    }
}


checkConnection( Term, Path )
octObject	*Term, *Path;
{
    
    struct octPoint		*pathPoints;
    int				i;
    int32 numPathPoints;
    octObject			Master, Instance, Box, Fterm;
    octObject			Layer, MasterLayer;
    octGenerator		boxGen, instGen, layerGen;
    char 			*instanceName;
    octStatus			status;
    int				geoFound = 0;
    octId			arrayOfGeo[MAX_NUM_GEOS];

    

    Fterm.type = OCT_TERM;
    Layer.type = OCT_LAYER;
    MasterLayer.type = OCT_LAYER;

    /* 	 Get the layer for the path   */

    OCT_INIT_GEN_CONTAINERS( Path, OCT_LAYER_MASK, &layerGen, "Generator for path layer");
    OCT_GENERATE( &layerGen, &Layer, "Getting the layer for the path");
    MasterLayer.contents.layer.name = Layer.contents.layer.name;

    /*   Now get the points in the path   */

    ohGetPoints( Path, &numPathPoints, &pathPoints);
    if( numPathPoints > 2 ){
	FPRINTF(stderr, "Error: The number of path points found %d exceeds the maximum number allowed 2\n", numPathPoints);
	return;
    }

    /* check for 0 length paths */
    if( ( ( pathPoints[0].x - pathPoints[1].x ) == 0 ) &&
        ( ( pathPoints[0].y - pathPoints[1].y ) == 0 ) ){

	if( objectInList( Path ) == NO ) {
	    FPRINTF( stderr, "Path on layer has zero length %s\n", 
				MasterLayer.contents.layer.name);
	    for( i = 0 ; i < numPathPoints ; i++ ){
		FPRINTF( stderr, "point %d <%d,%d>\n", i, pathPoints[i].x, pathPoints[i].y);
	    }
	    addToErrorBag(Path, "Zero length path " );
	}
    }


    OCT_INIT_GEN_CONTAINERS( Term, OCT_INSTANCE_MASK, &instGen, "Generating the instance for terminal" );
    OCT_GENERATE( &instGen, &Instance , "Getting the instance for the terminal");
    if( debug ) FPRINTF( stderr, "Term %s belongs to instance %s \n",
				Term->contents.term.name,
				Instance.contents.instance.name);


    OH_ASSERT( ohOpenMaster( &Master, &Instance, "interface", "r" ) );

    instanceName = SAVE(Instance.contents.instance.name);



    Fterm.contents.term.name = Term->contents.term.name;
    OCT_GET_BY_NAME( &Master, &Fterm, "Getting the Fterm" );
    OCT_GET_BY_NAME( &Master, &MasterLayer, "Getting the MasterLayer" );

    if(debug) FPRINTF( stderr, "Looking for the box for the terminal \n");
    OH_ASSERT( octInitGenContents(&Fterm,OCT_BOX_MASK,&boxGen) );

    while( (status = octGenerate(&boxGen,&Box) ) == OCT_OK ){
	if( octIsAttached( &MasterLayer, &Box ) != OCT_OK ) {
	    continue;
	}
	if( geoFound < MAX_NUM_GEOS ) {
	    arrayOfGeo[geoFound++] = Box.objectId;
	}
	octTransformGeo( &Box, &(Instance.contents.instance.transform) );
	for( i = 0 ; i < numPathPoints ; i++ ) {
	    if( POINT_IN_BOX( (&pathPoints[i]), (&Box.contents.box) ) ) {
		return;
	    }
	}
    }

    if( geoFound == 0){
	FPRINTF(stderr, "No box found attached to the layer %s", ohFormatName(&MasterLayer));
	return;
    }

    if( (objectInList( Term ) == NO ) || (objectInList( Path ) == NO ) ){
	FPRINTF( stderr, "Terminal %s (id %s) and Path (id %s) are not connected \n",
		Term->contents.term.name,
		octFormatId(Term->objectId),
		octFormatId(Path->objectId));
	FPRINTF( stderr, "Their coordinate are: \n" );
	for( i = 0 ; i < geoFound ; i++ ){
	    Box.objectId = arrayOfGeo[i];
	    OCT_GET_BY_ID( &Box, "Getting the box by id");
	    FPRINTF( stderr, "box on layer %s  <%d,%d> ---> <%d,%d> \n",
		    Layer.contents.layer.name,
		    Box.contents.box.lowerLeft.x, Box.contents.box.lowerLeft.y,
		    Box.contents.box.upperRight.x, Box.contents.box.upperRight.y);
	}
	FPRINTF( stderr, "path on layer %s\n", MasterLayer.contents.layer.name);
	for( i = 0 ; i < numPathPoints ; i++ ){
	    FPRINTF( stderr, "Point %d <%d,%d>\n", i, pathPoints[i].x, pathPoints[i].y);
	}
    }
    addToErrorBag(Path, "Path and Terminal not Connected" );
    addToErrorBag(Term, "Path and Terminal not Connected" );

    incrErrorCount(NIL(char), NIL(octObject));
}

int boxContain( box1, box2 )
octObject *box1, *box2;
{
    if( box1->contents.box.upperRight.x >= box2->contents.box.upperRight.x &&
        box1->contents.box.upperRight.y >= box2->contents.box.upperRight.y &&
        box1->contents.box.lowerLeft.x <= box2->contents.box.lowerLeft.x &&
        box1->contents.box.lowerLeft.y <= box2->contents.box.lowerLeft.y ){
	return(1);
    } else if( box2->contents.box.upperRight.x >= box1->contents.box.upperRight.x &&
        box2->contents.box.upperRight.y >= box1->contents.box.upperRight.y &&
        box2->contents.box.lowerLeft.x <= box1->contents.box.lowerLeft.x &&
        box2->contents.box.lowerLeft.y <= box1->contents.box.lowerLeft.y ){
	return(2);
    } else {
	return(NO);
    }
}

