#include "copyright.h"
#include "port.h"
#include "checks.h"
#include "oh.h"

extern int		debug ; 
extern int		numErrors, maxNumErrors;
extern octObject	Facet;
extern octObject	MarkBag;
extern char		*reportFileName;
char			msgs[1000];
extern int		silent;




checkConnectivity()
{
    octGenerator	netGen, termGen, genPath, termGenJoin;
    octObject		Net, MasterBag, Path,  Term, JoinedTerm, ATerm, Instance;
    octObject		MasterTerm, MasterJoinedTerm;
    int			numberOfBranches = 0;
    int			numberOfNets = 0;
    int			numPaths;
    int			goToNextNet;
    int			status;

    if( silent == NO ) Message( "Checking connectivity.");
    MasterBag.type = OCT_BAG;
    MasterBag.contents.bag.name = "JOINED";
    JoinedTerm.type = OCT_TERM;

    if (debug) FPRINTF( stderr ,"Routine checkConnectivity \n");
    /* for each net attached to the facet */
    OCT_INIT_GEN_CONTENTS( &Facet, OCT_NET_MASK, &netGen, "net generator" );
    while ( octGenerate( &netGen, &Net ) != OCT_GEN_DONE ){
	goToNextNet = FALSE;
	++numberOfNets;
	if( silent == NO ){
	    PRINTF( ".");
	    if ( numberOfNets % 50 == 0 ) PRINTF( "  %d\n", numberOfNets );
	}
	if (debug) { 
	    FPRINTF( stderr, "Checking net %s\n", Net.contents.net.name) ;
	}
	/* for each terminal attached to the net */
	OCT_INIT_GEN_CONTENTS( &Net, OCT_TERM_MASK, &termGen, "Term generator" );
	numberOfBranches = 0;
	while (octGenerate(&termGen, &Term) != OCT_GEN_DONE && goToNextNet == FALSE ) {
	    if (debug){
		FPRINTF( stderr, "Checking term %s (%s) on net %s\n", 
			    Term.contents.term.name,
			    octFormatId(Term.objectId),
			    Net.contents.net.name);

	    }
	    if (octIdIsNull(Term.contents.term.instanceId)){
		if( instanceTerm( &Term ) == YES ){
		    continue;
		}
		OCT_GEN_FIRST_CONTENT( &Term, OCT_TERM_MASK, &ATerm, "Getting the Aterm");
		if( Mstatus == OCT_GEN_DONE ){
		    SPRINTF( msgs, "No implementation found for the Fterminal %s\n",
			    Term.contents.term.name );
		    incrErrorCount(msgs, &Term);
		    continue;
		} else if( Mstatus == OCT_OK ){
		    Term = ATerm;
		}
	    }

		
	    /* first: check to see if the terminal has already been touched
		      before */

	    if( markTerm( &Term ) == FOUND ) continue;  /* terminal already found before */

	    /* second: start following the path */

	    /*	check if the teminal is in a joined bag */

	    numberOfBranches++;
	    Instance.objectId = Term.contents.term.instanceId;
	    OCT_GET_BY_ID( &Instance, "Getting Instance by Id");
	    getMasterTerm( &Term, &MasterTerm, &Instance );
	    OCT_GET_CONTAINER_BY_NAME( &MasterTerm, &MasterBag , "Getting Joined Bag" );
	    status = Mstatus;
	    if( status == OCT_NOT_FOUND ){
		OCT_INIT_GEN_CONTAINERS( &Term, OCT_PATH_MASK, &genPath, "Generator to get paths");
		numPaths = 0;
		while ( octGenerate( &genPath, &Path ) != OCT_GEN_DONE ){
		    numPaths++;
		    OCT_GET_CONTAINER_BY_NAME( &Path, &Net, "Getting the net for the path" );
		    if( Mstatus == OCT_NOT_FOUND ){
			
			SPRINTF( msgs , "Path not attached to net %s",
				Net.contents.net.name );
			reportError(&Path, msgs, YES );
		    }
		    if (debug) {
			FPRINTF( stderr, "Pat %s\n", octFormatId(Path.objectId));
		    }
		    getNewTerm( &Net, &Path, Term.objectId );
		}
		if( numPaths == 0 ){
			SPRINTF( msgs , "No path attached to %s",
				ohFormatName( &Term ) );
			reportError(&Term, msgs, YES );
		}
	    } else if( status == OCT_OK ){
		OCT_INIT_GEN_CONTENTS( &MasterBag, OCT_TERM_MASK, &termGenJoin, "Generating joined terms" );
		while( octGenerate( &termGenJoin, &MasterJoinedTerm ) != OCT_GEN_DONE ){
		    JoinedTerm.contents.term.name = MasterJoinedTerm.contents.term.name;
		    OCT_GET_BY_NAME( &Instance, &JoinedTerm, "getting joinedTerm by name ");
		    if (!octIdsEqual(JoinedTerm.objectId, Term.objectId)) markTerm( &JoinedTerm );
		    OCT_INIT_GEN_CONTAINERS( &JoinedTerm, OCT_PATH_MASK, &genPath, "Generator to get paths");
		    while( octGenerate( &genPath, &Path ) != OCT_GEN_DONE ){
			if (debug) {
			    FPRINTF( stderr, "Pat %s\n", octFormatId(Path.objectId) );
			}
			OCT_GET_CONTAINER_BY_NAME( &Path, &Net, "Getting the net for the path" );
			getNewTerm( &Net, &Path, JoinedTerm.objectId );
		    }
		}
	    }
	}
	if( numberOfBranches > 1 ){
	    SPRINTF( msgs, "More than one branch (%d) connected to %s",
		    numberOfBranches,
		    ohFormatName( &Net ) );
	    reportError( &Net, msgs, YES );
	}
    }
    if( silent == NO ){
	PRINTF("\n");
	PRINTF("Total number of nets checked %d\n" , numberOfNets );
	PRINTF( "\n\n");
    }
}


checkPaths()
{
    octGenerator	layerGen, termGen, pathGen, termGenJoin;
    octObject		Layer, Term, Path, Instance, MasterTerm, MasterBag; 
    octObject		JoinedTerm, MasterJoinedTerm;
    int			numTerms, numberOfPaths = 0;
    octStatus		status;

    MasterBag.type = OCT_BAG;
    MasterBag.contents.bag.name = "JOINED";
    JoinedTerm.type = OCT_TERM;

    if( silent == NO ) PRINTF( "Checking for disconnected paths\n");
    /*	for each net attached to the facet  */

    if (debug ) FPRINTF( stderr, "entering checkPaths\n" );
    OCT_INIT_GEN_CONTENTS( &Facet, OCT_LAYER_MASK, &layerGen, "Layer generator" );
    while( octGenerate( &layerGen, &Layer ) != OCT_GEN_DONE ){
	/* for each layer */ 
	if( ( strcmp( Layer.contents.layer.name, "MET1") != 0 ) &&
	    ( strcmp( Layer.contents.layer.name, "MET2") != 0 ) &&
	    ( strcmp( Layer.contents.layer.name, "POLY") != 0 ) ){
	    continue;
	}
	OCT_INIT_GEN_CONTENTS( &Layer, OCT_PATH_MASK, &pathGen, "Path generator" );
	while( octGenerate( &pathGen, &Path ) != OCT_GEN_DONE ){
	    ++numberOfPaths;
	    if( silent == NO ){
		PRINTF( ".");
		if ( numberOfPaths % 50 == 0 ) PRINTF( "  %d\n", numberOfPaths );
	    }
	    if( debug ) FPRINTF( stderr, "checking path %s\n", octFormatId(Path.objectId)) ;
	    OCT_INIT_GEN_CONTENTS( &Path, OCT_TERM_MASK, &termGen, "Term generator" );
	    numTerms = 0;
	    while( octGenerate( &termGen, &Term ) != OCT_GEN_DONE ){
		if( debug ) FPRINTF( stderr, 
			"checking  connection of path %s with term %s\n", 
				octFormatId(Path.objectId),
				Term.contents.term.name) ;
		checkConnection( &Term, &Path );
		if( unmarkTerm( &Term ) != NOT_FOUND ){
		    Instance.objectId = Term.contents.term.instanceId;
		    OCT_GET_BY_ID( &Instance , "Getting the instance by id");
		    getMasterTerm( &Term, &MasterTerm, &Instance );
		    OCT_GET_CONTAINER_BY_NAME( &MasterTerm, &MasterBag , "Getting Joined Bag" );
		    status = Mstatus;
		    if( status == OCT_OK ){
			OCT_INIT_GEN_CONTENTS( &MasterBag, OCT_TERM_MASK, &termGenJoin, "Generating joined terms" );
			while( octGenerate( &termGenJoin, &MasterJoinedTerm ) != OCT_GEN_DONE ){
			    JoinedTerm.contents.term.name = MasterJoinedTerm.contents.term.name;
			    OCT_GET_BY_NAME( &Instance, &JoinedTerm, "getting joinedTerm by name ");
			    unmarkTerm( &JoinedTerm );
			}
		    }
		}
		numTerms++;
	    } 
	    if( numTerms != 2 ){
		SPRINTF( msgs, "%d terminal(s) connected to the path", numTerms );
		reportError(&Path, msgs, YES);
		if( debug ){
		    FPRINTF(stderr, "%d terminal(s) connected to a path\n",
			    numTerms);
		}
	    }
	}
    }
    if( silent == NO ){
	PRINTF("\n");
	PRINTF("Total number of paths checked %d\n" , numberOfPaths );
	PRINTF( "\n\n");
    }
}



finalCheck()
{

    octGenerator	propGen;
    octObject		Prop, Term;
    int			numResProp = 0;

    if( silent == NO ) PRINTF( "Performing the final check\n");
    OCT_GET_BY_NAME( &Facet, &MarkBag, "Getting the marks bag");
    OCT_INIT_GEN_CONTENTS( &MarkBag, OCT_PROP_MASK, &propGen, 
			     "Initialiting the generator for the mark properties" );
    while( octGenerate( &propGen, &Prop ) != OCT_GEN_DONE ) {
	OCT_GEN_FIRST_CONTAINER( &Prop, OCT_TERM_MASK, &Term, 
		    "Getting the terminal that contains the property");
	SPRINTF( msgs, "%s not correctly unmarked", ohFormatName( &Term ) );
	reportError(&Term, msgs, YES);
	numResProp++;
    }
    if( numResProp == 0 ) OCT_DELETE_M( &MarkBag , "Deleting MarkBag");
    return(numResProp);
}

checkConsistency(view)
char	*view;
{
    
    octObject		Master, CompareMaster;
    octObject		CompareFacet, MasterTerm, CompareMasterTerm;
    octObject		Term, CompareTerm, Net, CompareNet, CompJoinedTerm;
    octObject		JoinedBag, CompareJoined, Instance, InstanceBag, CompareInstance;

    octGenerator	termGen, instGen, compTermGen;

    int		netFound;
    int		numberOfInstances = 0, numberOfFterms= 0, i;
    int status;


    JoinedBag.type = OCT_BAG;
    JoinedBag.contents.bag.name = "JOINED";
    Instance.type = OCT_INSTANCE;
    InstanceBag.type = OCT_BAG;
    InstanceBag.contents.bag.name = "INSTANCES";
    Master.type = OCT_FACET;
    Master.contents.facet.facet = "interface";
    Master.contents.facet.mode = "r";
    MasterTerm.type = OCT_TERM;
    CompareFacet = Facet;
    CompareFacet.contents.facet.view = view;
    CompareFacet.contents.facet.mode = "r";
    CompareJoined.type = OCT_BAG;
    CompareJoined.contents.bag.name = "JOINED";
    CompareTerm.type = OCT_TERM;
    CompJoinedTerm.type = OCT_TERM;
    CompareNet.type = OCT_NET;
    CompareInstance.type = OCT_INSTANCE;
    CompareMaster.type = OCT_FACET;
    CompareMaster.contents.facet.facet = "interface";
    CompareMaster.contents.facet.mode = "r";

    if( octOpenFacet( &CompareFacet ) < 0 ){
	FPRINTF( stderr, "The view %s is not present\n" , view );
	FPRINTF( stderr, "No consistency check performed\n");
	return;
    }
    VOVinputFacet( &CompareFacet );
    if( silent == NO )
	PRINTF( "Checking net-list consistency of view %s with respect to view %s\n",
		Facet.contents.facet.view, CompareFacet.contents.facet.view);
    for( i = 0 ; i < 2 ; i++ ){
	numberOfInstances = 0;
	numberOfFterms = 0;
	if ( i == 1 ) SWAP( octObject, Facet, CompareFacet );
	OCT_GET_BY_NAME( &Facet, &InstanceBag, "Getting the instance bag");
	OCT_INIT_GEN_CONTENTS( &InstanceBag, OCT_INSTANCE_MASK, &instGen, 
			"Generator for instances in Facet");
	while( octGenerate( &instGen, &Instance ) != OCT_GEN_DONE ){
	    ++numberOfInstances;
	    if( silent == NO ){
		PRINTF( ".");
		if ( numberOfInstances % 50 == 0 ) PRINTF( "  %d\n", numberOfInstances );
	    }
	    CompareInstance.contents.instance.name = Instance.contents.instance.name;
	    OCT_GET_BY_NAME( &CompareFacet, &CompareInstance , "Getting the instance");
	    if (Mstatus == OCT_NOT_FOUND){
		SPRINTF( msgs, "Instance %s on facet %s:%s does not have a correspondent one on facet %s:%s\n",
			Instance.contents.instance.name,
			Facet.contents.facet.cell,
			Facet.contents.facet.view,
			CompareFacet.contents.facet.cell,
			CompareFacet.contents.facet.view);
		incrErrorCount( msgs, &Instance );
		continue;
	    }
	    Master.contents.facet.cell = Instance.contents.instance.master;
	    Master.contents.facet.view = Instance.contents.instance.view;
	    Master.contents.facet.version = Instance.contents.instance.version;
	    OCT_OPEN_FACET( &Master, "opening the master" );
	    CompareMaster.contents.facet.cell = CompareInstance.contents.instance.master;
	    CompareMaster.contents.facet.view = CompareInstance.contents.instance.view;
	    CompareMaster.contents.facet.version = CompareInstance.contents.instance.version;
	    OCT_OPEN_FACET( &CompareMaster, "opening the Comparemaster" );
	    OCT_INIT_GEN_CONTENTS( &Instance, OCT_TERM_MASK, &termGen, 
				"Generator for Terms in Facet");

	    while( octGenerate( &termGen, &Term ) != OCT_GEN_DONE ){
		OCT_GEN_FIRST_CONTAINER( &Term, OCT_NET_MASK, &Net, "Getting the net");
		if( Mstatus == OCT_GEN_DONE ) continue;
		if( numTermsOnNet( &Net, TRUE ) < 2 ) continue;
		CompareTerm.contents.term.name = Term.contents.term.name;
		OCT_GET_BY_NAME( &CompareInstance, &CompareTerm , "Getting the terminal");
		if( Mstatus == OCT_NOT_FOUND ){
		    SPRINTF( msgs, "Terminal %s on instance %s in facet %s:%s does not have a correspondent one in facet %s:%s\n",
			Term.contents.term.name,
			Instance.contents.instance.name,
			Facet.contents.facet.cell,
			Facet.contents.facet.view,
			CompareFacet.contents.facet.cell,
			CompareFacet.contents.facet.view);
		    incrErrorCount( msgs, &Term );
		    continue;
		}
		/*	Check if the terminal is joined with others	*/
		MasterTerm.contents.term.name = Term.contents.term.name;
		OCT_GET_BY_NAME(&Master, &MasterTerm, "Getting terminal on master");
		OCT_GET_CONTAINER_BY_NAME( &MasterTerm, &JoinedBag, "Getting the JoinedBag");
		status = Mstatus;
		if( status == OCT_NOT_FOUND ){
		    CompareNet.contents.net.name =  Net.contents.net.name ;
		    OCT_GET_CONTAINER_BY_NAME( &CompareTerm, &CompareNet,
				    "Getting the net for the CompareTerm");
		    if( Mstatus == OCT_NOT_FOUND ){
			OCT_GEN_FIRST_CONTAINER( &CompareTerm, OCT_NET_MASK, &CompareNet, 
				    "Getting the second Net" );
			if( Mstatus == OCT_GEN_DONE ){
			    SPRINTF( msgs, "The terminal %s of instance %s in facet %s:%s is not attached to any net \nThe terminal %s of instance %s in facet %s:%s is attached to net %s\n\n",
				CompareTerm.contents.term.name,
				CompareInstance.contents.instance.name,
				CompareFacet.contents.facet.cell,
				CompareFacet.contents.facet.view,
				CompareTerm.contents.term.name,
				Instance.contents.instance.name,
				Facet.contents.facet.cell,
				Facet.contents.facet.view,
				Net.contents.net.name);
			    } else {
				SPRINTF( msgs, "The terminal %s of instance %s in facet %s:%s is attached to net %s\nThe terminal %s of instance %s in facet %s:%s is attached to net %s\n\n",
				CompareTerm.contents.term.name,
				CompareInstance.contents.instance.name,
				CompareFacet.contents.facet.cell,
				CompareFacet.contents.facet.view,
				CompareNet.contents.net.name,
				CompareTerm.contents.term.name,
				Instance.contents.instance.name,
				Facet.contents.facet.cell,
				Facet.contents.facet.view,
				Net.contents.net.name);
			    }
			incrErrorCount( msgs, &CompareTerm );
			continue;
		    }
		} else {
		    CompareMasterTerm.contents.term.name =
			    CompareTerm.contents.term.name;
		    OCT_GET_BY_NAME( &CompareMaster, &CompareTerm, 
			    "Getting the terminal on the compare Master m");
		    OCT_GET_CONTAINER_BY_NAME( &CompareTerm, &CompareJoined,
			    "Getting the joined bag on the master term");
		    if( Mstatus == OCT_NOT_FOUND ){
			    SPRINTF( msgs, "The terminal %s in facet %s:%s is not contained in a joined bag while the terminal %s in facet %s:%s is \n\n",
				CompareTerm.contents.term.name,
				CompareFacet.contents.facet.cell,
				CompareFacet.contents.facet.view,
				Term.contents.term.name,
				Facet.contents.facet.cell,
				Facet.contents.facet.view);
		    }
		    OCT_INIT_GEN_CONTENTS( &CompareJoined, OCT_TERM_MASK, &compTermGen,
			    "Generating the joined term on the compare Master");
		    netFound = FALSE;
		    CompareNet.contents.net.name = Net.contents.net.name;
		    while( octGenerate( &compTermGen, &CompJoinedTerm ) != OCT_GEN_DONE ){
			CompareTerm.contents.term.name = CompJoinedTerm.contents.term.name;
			OCT_GET_BY_NAME( &CompareInstance, &CompareTerm, 
				"Getting the joined term on the instance");
			OCT_GET_CONTAINER_BY_NAME( &CompareTerm, &CompareNet,
					"Getting the CompareNet");
			if( Mstatus == OCT_OK ){
			    netFound = TRUE ;
			    break;
			}
		    }
		    if( netFound == FALSE ){
			OCT_GEN_FIRST_CONTAINER( &CompareTerm, OCT_NET_MASK, &CompareNet, "Getting the second Net" );
			if( Mstatus == OCT_GEN_DONE ){
			    SPRINTF( msgs, "The terminal %s of instance %s in facet %s:%s is not attached to any net \nThe terminal %s of instance %s in facet %s:%s is attached to net %s\n\n",
				CompareTerm.contents.term.name,
				CompareInstance.contents.instance.name,
				CompareFacet.contents.facet.cell,
				CompareFacet.contents.facet.view,
				Term.contents.term.name,
				Instance.contents.instance.name,
				Facet.contents.facet.cell,
				Facet.contents.facet.view,
				Net.contents.net.name);
			} else {
			    SPRINTF( msgs, "The terminal %s of instance %s in facet %s:%s is attached to net %s\nThe terminal %s of instance %s in facet %s:%s is attached to net %s\n\n",
				CompareTerm.contents.term.name,
				CompareInstance.contents.instance.name,
				CompareFacet.contents.facet.cell,
				CompareFacet.contents.facet.view,
				CompareNet.contents.net.name,
				Term.contents.term.name,
				Instance.contents.instance.name,
				Facet.contents.facet.cell,
				Facet.contents.facet.view,
				Net.contents.net.name);
			}
			incrErrorCount( msgs, &CompareTerm );
		    }
		}
	    }
	}
	if( silent == NO ) {
	    PRINTF("\n");
	    PRINTF("Total number of instances checked %d\n" , numberOfInstances );
	    PRINTF("\n");
	}
	/* now check the formal terminals  */
	OCT_INIT_GEN_CONTENTS( &Facet, OCT_TERM_MASK, &termGen, 
			"Getting Fterms on Facet");
	while( octGenerate( &termGen, &Term ) != OCT_GEN_DONE ){
	    ++numberOfFterms;
	    if( silent == NO ){
		PRINTF( ".");
		if ( numberOfFterms % 50 == 0 ) PRINTF( "  %d\n", numberOfFterms );
	    }
	    CompareTerm.contents.term.name = Term.contents.term.name;
	    OCT_GET_BY_NAME( &CompareFacet, &CompareTerm, "Getting the fTerm on Cfacet");
	    if( Mstatus == OCT_NOT_FOUND ){
		SPRINTF( msgs, "The terminal %s is not present on facet %s:%s\n",
			CompareTerm.contents.term.name,
			CompareFacet.contents.facet.cell,
			CompareFacet.contents.facet.view);
		incrErrorCount( msgs, &CompareTerm );
		continue;
	    }
	    OCT_GEN_FIRST_CONTAINER( &Term, OCT_NET_MASK, &Net, 
				    "Getting the first Net" );
	    if( Mstatus == OCT_GEN_DONE ) continue;
	    if( numTermsOnNet( &Net, TRUE ) < 2 ) continue;
	    OCT_GEN_FIRST_CONTAINER( &CompareTerm, OCT_NET_MASK, &CompareNet, 
				    "Getting the second Net" );
	    if( Mstatus == OCT_GEN_DONE ){
		CompareNet.contents.net.name = "UNKNOWN";
	    }
	    if( strcmp( Net.contents.net.name, CompareNet.contents.net.name ) != 0 ){
		SPRINTF( msgs, "The terminal %s in facet %s:%s is attached to net %s\nThe terminal %s in facet %s:%s is attached to net %s\n\n",
			CompareTerm.contents.term.name,
			CompareFacet.contents.facet.cell,
			CompareFacet.contents.facet.view,
			CompareNet.contents.net.name,
			Term.contents.term.name,
			Facet.contents.facet.cell,
			Facet.contents.facet.view,
			Net.contents.net.name);
		incrErrorCount( msgs, &CompareTerm );
		continue;
	    }
	}
	if( silent == NO ) {
	    PRINTF("\n");
	    PRINTF("Total number of Fterms checked %d\n" , numberOfFterms );
	    PRINTF("\n");
	}
    }
    SWAP( octObject, CompareFacet, Facet );
    if( silent == NO ) PRINTF( "\n\n" );
}


checkTerms()
{
    octGenerator	netGen, termGen;
    octObject		Net, Term;
    int			numberOfTerms = 0; 

    if( silent == NO ) PRINTF( "Checking Terminals\n");

    if (debug) FPRINTF( stderr ,"Routine checkTerms \n");
    /* for each net attached to the facet */
    OCT_INIT_GEN_CONTENTS( &Facet, OCT_NET_MASK, &netGen, "net generator" );
    while ( octGenerate( &netGen, &Net ) != OCT_GEN_DONE ){
	if (debug) { 
	    FPRINTF( stderr, "Checking net %s\n", Net.contents.net.name) ;
	}
	/* for each terminal attached to the net */
	OCT_INIT_GEN_CONTENTS( &Net, OCT_TERM_MASK, &termGen, "Term generator" );
	while (octGenerate(&termGen, &Term) != OCT_GEN_DONE ) {
	    numberOfTerms++ ;
	    if( silent == NO ){
		PRINTF( ".");
		if ( numberOfTerms % 50 == 0 ) PRINTF( "  %d\n", numberOfTerms );
	    }
	    if (debug){
		FPRINTF( stderr, "Checking term %s (%s) on net %s\n", 
			    Term.contents.term.name,
			    octFormatId(Term.objectId),
			    Net.contents.net.name);

	    }
	    checkForFormal( &Term, &Net );
	}
    }
    if( silent == NO ){
	PRINTF("\n");
	PRINTF("Total number of terminals checked %d\n" , numberOfTerms );
	PRINTF( "\n\n");
    }
}



checkConnectors()
{
    octGenerator	instGen, termGen, pathGen;
    octObject		Instance, Net, Term, Path, ConnBag, FTerm; 
    int			numPaths, numberOfTerms = 0;

    ConnBag.type = OCT_BAG;
    ConnBag.contents.bag.name = "CONNECTORS";

    if( silent == NO )
	PRINTF( "Checking for disconnected connectors\n");

    OCT_GET_BY_NAME( &Facet, &ConnBag, "Getting the connector Bag");
    if (debug ) FPRINTF( stderr, "entering checkConnectors\n" );
    OCT_INIT_GEN_CONTENTS( &ConnBag, OCT_INSTANCE_MASK, &instGen, "Instance generator" );
    while( octGenerate( &instGen, &Instance ) != OCT_GEN_DONE ){
	/* for each instance */
	OCT_INIT_GEN_CONTENTS( &Instance, OCT_TERM_MASK, &termGen, "Terminal generator" );
	while( octGenerate( &termGen, &Term ) != OCT_GEN_DONE ){
	    ++numberOfTerms;
	    if( silent == NO ){
		PRINTF( ".");
		if ( numberOfTerms % 50 == 0 ) PRINTF( "  %d\n", numberOfTerms );
	    }
	    OCT_INIT_GEN_CONTAINERS( &Term, OCT_PATH_MASK, &pathGen, "Path generator" );
	    numPaths = 0;
	    while( octGenerate( &pathGen, &Path ) != OCT_GEN_DONE ){
		numPaths++;
		checkConnection( &Term, &Path );
	    } 
	    if( numPaths < 2 ){
		
		OCT_GEN_FIRST_CONTAINER( &Term, OCT_TERM_MASK, &FTerm, "Getting the term")
		if( Mstatus == OCT_NOT_FOUND ) {
		    OCT_GEN_FIRST_CONTAINER( &Path, OCT_NET_MASK, &Net, "Getting the net")
		    SPRINTF( msgs, "Errors on net %s", Net.contents.net.name );
		    reportError(&Term, msgs, YES );
		    FPRINTF( stderr, "Connector  terminal connected to less than two paths\n");
		}
	    }
	}
    }
    if( silent == NO ){
	PRINTF("\n");
	PRINTF("Total number of connectors checked %d\n" , numberOfTerms );
	PRINTF( "\n\n");
    }
}
getMasterTerm( Term, MasterTerm, Instance )
octObject *Term, *MasterTerm, *Instance;
{
    octObject Master;

    Master.type = OCT_FACET;
    Master.contents.facet.facet = "interface";
    Master.contents.facet.mode = "r";
    Master.contents.facet.cell = Instance->contents.instance.master;
    Master.contents.facet.view = Instance->contents.instance.view;
    Master.contents.facet.version = Instance->contents.instance.version;
    MasterTerm->type = OCT_TERM;
    MasterTerm->contents.term.name = Term->contents.term.name;

    OCT_OPEN_FACET( &Master, "opening the master" );
    OCT_GET_BY_NAME( &Master, MasterTerm, "getting master term");
}
