#include "copyright.h"
#include "port.h"
#include "checks.h"
#include "oh.h"
#include "options.h"
#include "vov.h"

int		debug = NO; 
int		numErrors = 0 ;
int 		maxNumErrors = 10;
char		*reportFileName = NIL(char);
FILE*           file = NIL(FILE);
octObject	Facet;
octObject	MarkBag;
octObject	ErrorBag;

int 		deltax, deltay;
int 	silent = YES;
extern char		*version();


optionStruct optionList[] = {
    { "f", "fileName",       "file where the errors are printed (stderr by default)"},
    { "c", "view", "compares the contents of facet with view"},
    { "m", "maxNumErrors",       "maximum number of errors tolerated"},
    { "F", 0,			"flush the facet to show the errors" },
    { "d", 0,			"debug mode" },
    { "v", 0,     		    "verbose mode"},
    { OPT_RARG, "cell:view",     "input facet"},
    { 0, 0 , 0}
};


main(argc,argv)
    int argc;
    char *argv[];
{
    char	*fileName;
    char	*facetName;
    char	*inFacet;
    char	*compareView = NIL(char);
    int		argIndex;
    int		doFlush = 0;
    struct octBox	Box;
    octStatus	facetStatus;
    extern char* optarg;
    extern int   optind;
    int  c;

    /*
     *  default values
     */
    /* Now parse the options */
    while ( (c = optGetOption( argc, argv )) != EOF )  {
	switch ( c ) {
	case 'F':
	    doFlush = 1;
	    break;
	case 'd':
	    debug = 1;
	    break;
	case 'm':
	    maxNumErrors = atoi( optarg );
	    break;
	case 'f':
	    reportFileName = optarg;
	    break;
	case 'c':
	    compareView = optarg;
	    break;
	case 'v':
	    silent = NO;
	    break;
	default:
	    optUsage();
	}
    }
    
    if ( argc == 1 || optind != argc-1 ) optUsage();


    VOVbegin( argc, argv );

    octBegin();

    inFacet = argv[optind];
    Facet.type = OCT_FACET;
    ohUnpackDefaults(&Facet, "a", "::contents");
    OH_ASSERT( ohUnpackFacetName(&Facet, inFacet) );
    facetName = Facet.contents.facet.cell ;
    OH_ASSERT( octOpenFacet( &Facet ));
    VOVinputFacet( &Facet );


    if( silent == NO ){
	PRINTF( "mosaicoRC %s\n\n", version() );
    }




    octBB(&Facet, &Box);
    deltax = (Box.upperRight.x - Box.lowerLeft.x ) / 100.;
    deltay = (Box.upperRight.y - Box.lowerLeft.y ) / 100.;



    if( reportFileName == NIL(char) ){
	file = stderr;
    }  else {
	if( (file = VOVfopen( reportFileName, "w") ) == NIL(FILE) ){
	    errRaise( "mosaicoRC", 1, "Error opening the file %s \n", reportFileName );
	}
    }



    createBags();

    if( compareView != NIL(char) ){
	checkConsistency( compareView );
    }
    checkConnectivity();
    checkTerms();
    checkConnectors();
    checkPaths();

    if( finalCheck() > 0 || numErrors > 0 ){
	printInstr( doFlush );
	if ( doFlush ) {
	    VOVoutputFacet( &Facet );
	    OH_ASSERT( octCloseFacet(&Facet) );
	}
	VOVend( 1 );
    } else {
	if ( silent == NO ) Message( "Routing is OK." );
    }

    deleteBag();
    octEnd();
    VOVend( 0 );
}


printInstr( doFlush )
    int doFlush;
{
    if ( doFlush ) {
	FPRINTF( stderr, "\n\n" );
	FPRINTF( stderr, "\tThe error(s) can be displayed with Vem as follows:\n");
	FPRINTF( stderr, "\t1) Run Vem;\n" );
	FPRINTF( stderr, "\t2) Open %s:%s:%s\n",
		Facet.contents.facet.cell,
		Facet.contents.facet.view,
		Facet.contents.facet.facet );
	FPRINTF( stderr, "\t3) Select the bag ROUTING_ERRORS\n");
	FPRINTF( stderr, "\tAt this point the location where errors occurred will be\n");
	FPRINTF( stderr, "\tdisplayed.\n");
	FPRINTF( stderr, "\tEach of the object that is highlighted will have a property\n");
	FPRINTF( stderr, "\tattached describing the type of the error detected.\n");
	FPRINTF( stderr, "\n\n" );
    }

    Error( "Some routing inconsistencies detected.");
}


void Message( s )
    char* s;
{
    (void)fprintf( stderr, "mRC: Message: %s\n", s );
}

void Error( s )
    char* s;
{
    (void)fprintf( stderr, "mRC: Error: %s\n", s );
}

