#include "copyright.h"
#include "port.h"
#include "errtrap.h"
#include "checks.h"
#include "oh.h"


extern int		debug ; 
extern int		numErrors, maxNumErrors;
extern octObject	Facet;
extern octObject	MarkBag;
extern octObject	ErrorBag;
extern int		netSignaled;
extern int		deltax, deltay;
extern char		*reportFileName;



struct objectList	*stack = NIL(struct objectList);



incrErrorCount(msgs, Obj)
char *msgs;
octObject *Obj;
{
    ++numErrors;
    if( msgs != NIL(char) ){
	FPRINTF( stderr, msgs );
    }
    if( Obj != NIL( octObject ) ){
	if( objectInList( Obj ) == NO ){
	    pushIntoList( Obj );
	}
    }
    if( numErrors > maxNumErrors ){
	errRaise( "mosaicoRC", 1, "Too many errors" );
    }
}


markTerm( Term )
octObject 	*Term;
{
    octObject   Mark;
    octObject	ATerm, FTerm;

    Mark.type = OCT_PROP;
    Mark.contents.prop.name = "MARK";
    Mark.contents.prop.type = OCT_STRING;
    Mark.contents.prop.value.string = "";
    if( debug ){
	FPRINTF( stderr, "Routine markTerm\n");
	FPRINTF( stderr, "Marking terminal %s (%d)\n",
		    Term->contents.term.name,
		    Term->objectId);
    }
    /* I have to mark also the Formal Terminal */


    if( octGetByName( Term, &Mark ) == OCT_NOT_FOUND ){
        OCT_CREATE_M( Term, &Mark, "Creating the property Mark" );
	OCT_ATTACH( &MarkBag, &Mark, "Attaching the property Mark" );
    } else {   
	return(FOUND);
    }
    if (octIdIsNull(Term->contents.term.instanceId)) {
	octGenFirstContent( Term, OCT_TERM_MASK, &ATerm );
	if( octGetByName( &ATerm, &Mark ) == OCT_OK ){
	    OCT_CREATE_M( &ATerm, &Mark, "Creating the property Mark" );
	    OCT_ATTACH( &MarkBag, &Mark, "Attaching the property Mark" );
	} 
    } else {
	if( octGenFirstContainer( Term, OCT_TERM_MASK, &FTerm ) == OCT_OK ){
	    if( octGetByName( &FTerm, &Mark ) == OCT_OK ){
		OCT_CREATE_M( &FTerm, &Mark, "Creating the property Mark" );
		OCT_ATTACH( &MarkBag, &Mark, "Attaching the property Mark" );
	    } 
	}
    }
    return(CREATED);
}


unmarkTerm( Term )
octObject 	*Term;
{
    octObject   Mark;

    Mark.type = OCT_PROP;
    Mark.contents.prop.name = "MARK";
    if( debug ){
	FPRINTF( stderr, "Routine unmarkTerm\n");
	FPRINTF( stderr, "Marking terminal %s (%d)\n",
		    Term->contents.term.name,
		    Term->objectId);

    }

    if( octGetByName( Term, &Mark ) == OCT_OK ){
	OCT_DELETE_M( &Mark, "Deleting the property Mark" );
	return(DELETED);
    } else {
	return(NOT_FOUND);
    }
}

int instanceTerm( Term ) 
octObject	*Term;
{

    octObject 	Instance, Bag;

    Bag.type = OCT_BAG;
    Bag.contents.bag.name = "INSTANCES";
    Instance.type = OCT_INSTANCE;


    if (octIdIsNull(Term->contents.term.instanceId)) return(-1);
    Instance.objectId = Term->contents.term.instanceId;
    OCT_GET_BY_ID(&Instance, "Getting the instance for the terminal");
    OCT_GET_CONTAINER_BY_NAME( &Instance, &Bag , "Getting the instances bag");
    if( Mstatus == OCT_OK ){
	if( debug ){
	    FPRINTF( stderr, "Routine instanceTerm\n");
	    FPRINTF( stderr, "terminal %s (%d) belong to instance %s (%d)\n",
			Term->contents.term.name,
			Term->objectId,
			Instance.contents.instance.name,
			Instance.objectId);
	}
	return(YES);
    } else {
	return(NO);
    }
}

addToErrorBag(Obj, msgs)
octObject	*Obj;
char 		*msgs;
{
    struct octPoint		pathPoints[2];
    octObject 			Instance, Master, Fterm, MasterLayer;
    octObject			Box, TermBox, Term;
    octGenerator		termGen;
    int32 			numPathPoints;
    char			*instanceName;

    Box.type = OCT_BOX;
    Fterm.type = OCT_TERM;
    MasterLayer.type =OCT_LAYER;

    if( Obj->type == OCT_PATH ){
	if( objectInList(Obj) ){
	    addProp(Obj, msgs );
	    return;
	}
	octGetPoints( Obj, &numPathPoints, (struct octPoint *) 0);
	octGetPoints( Obj, &numPathPoints, pathPoints);
	if( pathPoints[0].x == pathPoints[1].x ){
	    Box.contents.box.lowerLeft.x = MIN( pathPoints[0].x ,pathPoints[1].x ) - deltax;
	    Box.contents.box.upperRight.x = MAX( pathPoints[0].x ,pathPoints[1].x ) + deltax;
	    Box.contents.box.lowerLeft.y = MIN( pathPoints[0].y ,pathPoints[1].y );
	    Box.contents.box.upperRight.y = MAX( pathPoints[0].y ,pathPoints[1].y );
	} else {
	    Box.contents.box.lowerLeft.y = MIN( pathPoints[0].y ,pathPoints[1].y ) - deltay;
	    Box.contents.box.upperRight.y = MAX( pathPoints[0].y ,pathPoints[1].y ) + deltay;
	    Box.contents.box.lowerLeft.x = MIN( pathPoints[0].x ,pathPoints[1].x );
	    Box.contents.box.upperRight.x = MAX( pathPoints[0].x ,pathPoints[1].x );
	}
	octCreate( Obj, &Box );
	octAttach( &ErrorBag, &Box );
	addProp(Obj, msgs );
	pushIntoList( Obj );
    } else if( Obj->type == OCT_TERM ){
	
	if (octIdIsNull(Obj->contents.term.instanceId)) {
	    OCT_GEN_FIRST_CONTENT( Obj, OCT_TERM_MASK, &Term, "Getting the terminal");
	    if( Mstatus == OCT_GEN_DONE ){
		FPRINTF( stderr, "No implementation found for terminal %s\n",
			Obj->contents.term.name );
		return;
	    }
	    Instance.objectId = Term.contents.term.instanceId;
	} else {
	    Instance.objectId = Obj->contents.term.instanceId;
	}
	if( objectInList(Obj) ){
	    addProp(Obj, msgs );
	    return;
	}
	OCT_GET_BY_ID( &Instance , "Getting the instance");
	OH_ASSERT( ohOpenMaster( &Master, &Instance, "interface", "r" ) );

	if (octIdIsNull(Obj->contents.term.instanceId)) {
	    Fterm.contents.term.name = Term.contents.term.name;
	} else {
	    Fterm.contents.term.name = Obj->contents.term.name;
	}
	OCT_GET_BY_NAME( &Master, &Fterm, "Getting the Fterm" );
	OCT_INIT_GEN_CONTENTS(&Fterm,OCT_BOX_MASK,&termGen, "Generator for term imple");
	octGenerate( &termGen, &TermBox );
	while( octGenerate( &termGen, &Box ) != OCT_GEN_DONE ){
	    if( boxContain( &Box, &TermBox ) == 1){
		TermBox = Box;
	    }
	}
	octTransformGeo( &TermBox, &(Instance.contents.instance.transform) );
	Box.contents.box.lowerLeft.x 
		= MIN( TermBox.contents.box.lowerLeft.x ,TermBox.contents.box.upperRight.x ) - deltax ;
	Box.contents.box.upperRight.x 
		= MAX( TermBox.contents.box.lowerLeft.x ,TermBox.contents.box.upperRight.x ) + deltax ;
	Box.contents.box.lowerLeft.y 
		= MIN( TermBox.contents.box.lowerLeft.y ,TermBox.contents.box.upperRight.y ) - deltay ;
	Box.contents.box.upperRight.y 
		= MAX( TermBox.contents.box.lowerLeft.y ,TermBox.contents.box.upperRight.y ) + deltay ;
	octCreate( Obj, &Box );
	octAttach( &ErrorBag, &Box );
	addProp(Obj, msgs );
	pushIntoList( Obj);
    } else if( Obj->type == OCT_NET ){
	if( octGetByName( &ErrorBag, Obj ) == OCT_NOT_FOUND ) {
	    octAttach( &ErrorBag, Obj );
	}
    }
}

int objectInList( Obj )
octObject *Obj;
{
    struct objectList	*prstElmt;

    for( prstElmt = stack ; prstElmt != NIL(struct objectList ) ;
			    prstElmt = prstElmt->nextInList ){
	if (octIdsEqual(prstElmt->id, Obj->objectId)) return(YES);
    }
    return(NO);
}

pushIntoList( Obj )
octObject	*Obj;
{  			
    struct objectList	*element;
		    
    if( objectInList( Obj ) ) return;
    element = CALLOC( struct objectList, 1 );
    element->id = Obj->objectId;		
    element->type = Obj->type;
    element->nextInList = stack; 
    stack = element; 	
    return;
}

addProp( obj, msg )
octObject	*obj;
char *msg;
{
    octObject Prop;

    Prop.type = OCT_PROP;
    Prop.contents.prop.name = "ROUTING_ERROR";
    Prop.contents.prop.type = OCT_STRING;
    Prop.contents.prop.value.string = SAVE(msg);

    octCreate( obj, &Prop );
    octAttach( &ErrorBag, &Prop );
}

createBags()
{
    octGenerator netGen, boxGen, propGen;
    octObject	Net, Box, Prop;

    ErrorBag.type = OCT_BAG;
    ErrorBag.contents.bag.name = "ROUTING_ERRORS";
    MarkBag.type = OCT_BAG;
    MarkBag.contents.bag.name = "MARK_BAG";

    if( octGetByName( &Facet, &ErrorBag ) == OCT_OK ){
	int mask = OCT_NET_MASK | OCT_BOX_MASK | OCT_PROP_MASK;
	octGenerator gen;
	octObject obj;
	(void) octInitGenContents( &ErrorBag, mask  , &gen );
	while( octGenerate( &gen, &obj ) != OCT_GEN_DONE ){
	    octDetach( &ErrorBag, &obj );
	}
    } else {
	octCreate( &Facet, &ErrorBag );
    }

    if( octGetByName( &Facet, &MarkBag ) != OCT_NOT_FOUND ){
	(void) octInitGenContents( &MarkBag, OCT_PROP_MASK , &propGen );
	while( octGenerate( &propGen, &Prop ) != OCT_GEN_DONE ){
	    octDelete( &Prop );
	}
    } else {
	octCreate( &Facet, &MarkBag );
    }
}

deleteBag()
{
    octGenerator 	gen;
    octObject		Obj;

    ErrorBag.type = OCT_BAG;
    ErrorBag.contents.bag.name = "ROUTING_ERRORS";

    if( octGetByName( &Facet, &ErrorBag ) != OCT_NOT_FOUND ){
	(void) octInitGenContents( &ErrorBag, OCT_ALL_MASK , &gen );
	while( octGenerate( &gen, &Obj ) != OCT_GEN_DONE ){
	    octDelete( &Obj );	/* Added by Andrea */
	}
	octDelete( &ErrorBag );
    }
    return;
}

reportError( obj, msgs, flag )
octObject	*obj;
char		*msgs;
int		flag;
{
    Error( msgs );
    addToErrorBag( obj, msgs );
    if( flag ) incrErrorCount(NIL(char), NIL(octObject));
}



int numTermsOnNet( Net , flag)
octObject	*Net;
int		flag;
{

    octGenerator 	termGen;
    octObject		Term;
    int			numTerms = 0;

    OCT_INIT_GEN_CONTENTS( Net, OCT_TERM_MASK, &termGen, "Creating the generator");
    while( octGenerate( &termGen, &Term ) != OCT_GEN_DONE ){
	if( flag && ( ++numTerms > 1 ) ){
	    return(numTerms);
	}
    }
    return(numTerms);
}
