#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "options.h"
#include "vov.h"
#include "utility.h"
#include "pep.h"
/*
 *   >>>  GLOBAL VARIABLES DECLARATION
 */


int verbose = 0;

optionStruct optionList[] = {
    {"f", "tech file name",    "File with tchnological info" },
    {"C", "number",            "Min parasitic capacitance (in pF)" },
    {"v", "",                  "Verbose mode"},
    { 0, 0 , 0}
};



main(argc,argv)
    int argc;
    char *argv[];
{
    char  *inputName = 0;
    char  *techFileName = 0;
    int   c;
    double minCap = 0.020e-12;	/* 0.020 pF */

    /* parse command line options*/
    while ((c = optGetOption(argc,argv)) != EOF) {
	switch(c) {
	case 'v':
	    verbose = 1;
	    VOVredirect( 0 );
	    break;
	case 'C':
	    minCap = atof( optarg ) * 1e-12;
	    break;
	case 'f':
	    techFileName = util_strsav( optarg );
	    break;
	default:
	    optUsage();
	    break;
	}
    }

    /* the remaining arguments are argv[optind ... argc-1] */
    if (optind >= argc) {
	optUsage();
    } else {
	inputName  = util_strsav(argv[optind]);
    }

    octBegin();
    VOVbegin( argc, argv );

    if ( techFileName ) {
	thReadFile( techFileName );
    }

    {
	octObject Chip, net;
	octGenerator genNet;

	ohInputFacet( &Chip, inputName );

	octInitGenContents( &Chip, OCT_NET_MASK, &genNet );
	while ( octGenerate( &genNet, &net ) == OCT_OK ) {
	    double cap = pepCapacitance( &net );
	    octObject prop;
	    if ( ohGetByPropName( &net, &prop, "NETTYPE" ) != OCT_OK ) {
		prop.contents.prop.value.string = "SIGNAL";
	    }

	    printf( "%18s : %8s : cap %.3gpF %s\n", ohFormatName( &net ), 
		   prop.contents.prop.value.string, 
		   cap *1e12,
		   cap < minCap ? "small": "");
	}

	{
	    octObject schemFacet;
	    schemFacet = Chip;
	    schemFacet.contents.facet.view = "schematic";
	    schemFacet.contents.facet.mode = "w";
	
	    VOVoutputFacet( &schemFacet );
	    OH_ASSERT(octOpenFacet( &schemFacet ));

	    extToSchematic( &Chip, &schemFacet, minCap );
	    
	    OH_ASSERT(octFlushFacet( &schemFacet ));
	}
    }
    octEnd();
    VOVend(0);			/* Successful. */
}



