#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "pep.h"
#include "th.h"

#define PROPEQ( p, s ) (!strcmp( (p)->contents.prop.value.string, (s)))


static octId groundNetId = oct_null_id;
static octId supplyNetId = oct_null_id;


void GetOrCreate( schemFacet, sym_obj, schem_obj )
    octObject* schemFacet;
    octObject *sym_obj;
    octObject *schem_obj;
{
    *schem_obj =  *sym_obj;
    if ( octGetByName( schemFacet, schem_obj ) != OCT_OK ) {
	octGenerator gen;
	octObject obj;
	octCreate( schemFacet, schem_obj );
	octInitGenContents( sym_obj, OCT_PROP_MASK, &gen );
	while (octGenerate( &gen,&obj ) == OCT_OK ) {
	    octCreate( schem_obj, &obj );
	}
    }
}


void ohNetBB( net, bb )
    octObject *net;
    struct octBox *bb;
{
    octGenerator gen;
    octObject obj;
    struct octBox tmpbb;
    int count = 0;
    
    octInitGenContents( net, OCT_GEO_MASK | OCT_TERM_MASK, &gen );

    bb->lowerLeft.x = 0;
    bb->lowerLeft.y = 0;
    
    bb->upperRight.x = 0;
    bb->upperRight.y = 0;
    while ( octGenerate( &gen, &obj ) == OCT_OK ) {
	ohBB( &obj, &tmpbb );
	if ( count++ == 0 ) {
	    *bb = tmpbb;
	} else {
	    bb->lowerLeft.x = MIN(bb->lowerLeft.x, tmpbb.lowerLeft.x );
	    bb->lowerLeft.y = MIN(bb->lowerLeft.y, tmpbb.lowerLeft.y );

	    bb->upperRight.x = MAX(bb->upperRight.x, tmpbb.upperRight.x );
	    bb->upperRight.y = MAX(bb->upperRight.y, tmpbb.lowerLeft.y );
	}
    }
}

scaleInstance( inst )
    octObject *inst;
{
    inst->contents.instance.transform.translation.x *= 1.5;
    inst->contents.instance.transform.translation.y *= 1.5;
}

static doFormals( symFacet, schemFacet )
    octObject *symFacet;
    octObject *schemFacet;
{
    octGenerator gen;
    octObject  sym_term, schem_term, cterm, sym_net, schem_net, contact, cbag;
    struct octBox bb;
    struct octPoint center;

    sym_net.type = OCT_NET;

    ohGetOrCreateBag( schemFacet, &cbag, "CONNECTORS" );
    
    octInitGenContents( symFacet, OCT_TERM_MASK, &gen );
    while ( octGenerate( &gen, &sym_term ) == OCT_OK ) {
	ohTerminalBB( &sym_term, &bb );
	ohBoxCenter( &bb, &center );
	schem_term = sym_term;
	GetOrCreate( schemFacet, &sym_term, &schem_term );

	OH_ASSERT(ohCreateInstance( schemFacet, &contact, schem_term.contents.term.name, 
				   "$OCTTOOLS/tech/schematic/symbols/con2",
				   "physical", center.x,center.y, OCT_NO_TRANSFORM ));
	scaleInstance( &contact );
	OH_ASSERT(octModify( &contact ));
	octAttach( &cbag, &contact );
	OH_ASSERT( ohGetByTermName( &contact, &cterm, "t" ));
	octAttach( &schem_term, &cterm );

	if ( ohTerminalNet( &sym_term, &sym_net ) == OCT_OK ) {
	    schem_net = sym_net;
	    GetOrCreate( schemFacet, &sym_net, &schem_net );
	    octAttach( &schem_net, &schem_term );
	    octAttach( &schem_net, &cterm );
	}

    }
}


int isMosfet( inst )
    octObject *inst;
{
    octObject master, prop;
    OH_ASSERT( ohOpenMaster( &master, inst, "interface", "r" ) );
    if ( ohGetByPropName( &master, &prop, "CELLTYPE" ) == OCT_OK ) {
	return PROPEQ( &prop, "MOSFET" );
    }
    return 0;
}

struct correspondence {
    char *sym_name;
    char *schem_name;
    int  bulkFlag;
} mosfetCorr[] = {
    { "DRAIN", "DRAIN"  , 0},
    { "SOURCE", "SOURCE", 0},
    { "GATE", "GATE"    , 0},
    { "GATE1", "GATE"   , 0},
    { "BULK",  "BULK"   , 1},
    { 0, 0 , 0}
};



doConnectMosfet( symFacet, schemFacet, inst, new_inst )
    octObject *symFacet;
    octObject *schemFacet;
    octObject *inst, *new_inst;
{
    octObject sym_term, schem_term, sym_net, schem_net;
    int i = 0;
    while ( mosfetCorr[i].sym_name ) {
	if ( ohGetByTermName( inst, &sym_term, mosfetCorr[i].sym_name ) == OCT_OK) {
	    OH_ASSERT( ohGetByTermName( new_inst, &schem_term, mosfetCorr[i].schem_name ));
	    if ( ohTerminalNet( &sym_term, &sym_net ) == OCT_OK ) {
		GetOrCreate( schemFacet, &sym_net, &schem_net );
		octAttachOnce( &schem_net, &schem_term );
	    }
	}

	if ( mosfetCorr[i].bulkFlag ) {
	    octObject master,prop;
	    char *type;
	    
	    OH_ASSERT( ohGetByTermName( new_inst, &schem_term, mosfetCorr[i].schem_name ));
	    OH_ASSERT( ohOpenMaster( &master, new_inst, "interface", "r" ));
	    OH_ASSERT(ohGetByPropName( &master, &prop, "MOSFETTYPE" ));
	    
	    type = prop.contents.prop.value.string;
	    if ( type[0] == 'N' || type[0] == 'n' ) {
		ohGetById( &schem_net, groundNetId );
	    } else {
		ohGetById( &schem_net, supplyNetId );
	    }
	    octAttachOnce( &schem_net, &schem_term );
	}

	i++;
    }

}



static doInstances( symFacet, schemFacet )
    octObject *symFacet;
    octObject *schemFacet;
{
    octGenerator gen, genTerm;
    octObject instance, new_inst, term, sym_net, schem_net, bag, ibag;
   
    ohGetByBagName( symFacet, &bag, "INSTANCES" );
    octInitGenContents( &bag, OCT_INSTANCE_MASK, &gen );
    while ( octGenerate( &gen, &instance ) == OCT_OK ) {
	new_inst = instance;
	new_inst.contents.instance.view = "schematic";
	new_inst.contents.instance.facet = "contents";
	scaleInstance( &new_inst );
	octCreate( schemFacet, &new_inst );
	ohGetOrCreateBag( schemFacet, &ibag, "INSTANCES" );
	octAttach( &ibag, &new_inst );

	if ( isMosfet( &instance ) ) {
	    doConnectMosfet( symFacet, schemFacet, &instance, &new_inst );
	} else {		/* Do generic instance */
	    octInitGenContents( &instance, OCT_TERM_MASK, &genTerm );
	    while ( octGenerate( &genTerm, &term ) == OCT_OK ) {
		if ( octGenFirstContainer( &term, OCT_NET_MASK, &sym_net ) == OCT_OK) {
		    OH_ASSERT(octGetByName( &new_inst, &term ));
		    GetOrCreate( schemFacet, &sym_net, &schem_net );
		    octAttachOnce( &schem_net, &term );
		}
	    }
	}
    }
}


doParasitics( symFacet, schemFacet, minCap )
    octObject *symFacet;
    octObject *schemFacet;
    double minCap;
{
    octGenerator genNet;
    octObject net, prop, capacitor, groundNet, term, ibag;
    struct octBox net_bb;
    struct octPoint center;
    double cap;
    char capName[128];

    ohGetById( &groundNet, groundNetId );

    ohGetOrCreateBag( schemFacet, &ibag, "INSTANCES" );

    octInitGenContents( symFacet, OCT_NET_MASK, &genNet );
    while ( octGenerate( &genNet, &net ) == OCT_OK ) {
	
	if ( ohGetByPropName( &net, &prop, "NETTYPE" ) == OCT_OK && 
	    ( PROPEQ( &prop, "SUPPLY" )  ||PROPEQ( &prop, "GROUND" ) )) continue;
	/* thTechnology( &net ); */	/* Make sure we use the right technology. */
	/* thVerbose( 1 ); */
	cap = pepCapacitance( &net );
	
	if ( cap < minCap ) continue; /* Ignore small parasitic. */
	
	printf( "Cap %e >= %e\n", cap, minCap );


	ohNetBB( &net, &net_bb );
	ohBoxCenter( &net_bb, &center );
	
	sprintf( capName, "C%s", net.contents.net.name );
	capacitor.type = OCT_INSTANCE;
	capacitor.contents.instance.name = capName;
	capacitor.contents.instance.master = "$OCTTOOLS/tech/schematic/symbols/smallcapgrd";
	capacitor.contents.instance.view = "schematic";
	capacitor.contents.instance.facet = "contents";
	capacitor.contents.instance.version = OCT_CURRENT_VERSION;
	capacitor.contents.instance.transform.transformType  = OCT_NO_TRANSFORM;
	capacitor.contents.instance.transform.translation =  center;
	capacitor.contents.instance.transform.translation.x -= 1.5 * thLambda();
	
	scaleInstance( &capacitor );
	OH_ASSERT(octCreate( schemFacet, &capacitor ));
	octAttach( &ibag, &capacitor );

	octGetOrCreate( schemFacet, &net );
	OH_ASSERT(ohGetByTermName( &capacitor, &term, "TOP" ));
	octAttach( &net, &term );

	ohCreatePropReal( &capacitor, &prop, "c", cap );
	
    }
}    

findPowerAndGround( symFacet, schemFacet )
    octObject *symFacet;
    octObject *schemFacet;
{
    octGenerator genNet;
    octObject sym_net, schem_net, prop;

    octInitGenContents( symFacet, OCT_NET_MASK, &genNet );
    while ( octGenerate( &genNet, &sym_net ) == OCT_OK ) {
	
	if ( ohGetByPropName( &sym_net, &prop, "NETTYPE" ) == OCT_OK ) { 
	    if ( PROPEQ( &prop, "SUPPLY" ) ){
		GetOrCreate( schemFacet, &sym_net, &schem_net );
		supplyNetId = schem_net.objectId;
	    }

	    if( PROPEQ( &prop, "GROUND" ) ) {
		GetOrCreate( schemFacet, &sym_net, &schem_net );
		groundNetId = schem_net.objectId;
	    }
	}
    }
    if ( octIdIsNull( groundNetId ) ) {
	ohCreateNet( schemFacet, &schem_net, "GROUND_NET" );
	ohCreatePropStr( &schem_net, &prop, "NETTYPE", "GROUND" );
	groundNetId = schem_net.objectId;
    }

    if ( octIdIsNull( supplyNetId ) ) {
	ohCreateNet( schemFacet, &schem_net, "SYPPLY_NET" );
	ohCreatePropStr( &schem_net, &prop, "NETTYPE", "SUPPLY" );
	supplyNetId = schem_net.objectId;
    }
}

void    doPowerAndGround( schemFacet )
    octObject *schemFacet;
{
    octGenerator genNet;
    octObject    inst, net, prop, term, bag;
    octBox  facetBB, netBB;
    octPoint netBBcenter;

    octBB( schemFacet, &facetBB );
    octInitGenContents( schemFacet, OCT_NET_MASK, &genNet );
    while ( octGenerate(&genNet, &net ) == OCT_OK ) {
	if ( ohGetByPropName( &net, &prop, "NETTYPE" ) == OCT_OK ) {
	    if ( ! PROPEQ( &prop, "SUPPLY" ) && ! PROPEQ( &prop, "GROUND" )) continue;
	    

	    ohNetBB( &net, &netBB );
	    ohBoxCenter( &netBB, &netBBcenter );
	    
	    if ( PROPEQ( &prop, "SUPPLY" ) ) {
		OH_ASSERT( ohCreateInstance( schemFacet, &inst, "supply", 
				 "$OCTTOOLS/tech/schematic/symbols/vsupply", "schematic", 
				 netBBcenter.x , facetBB.upperRight.y + 100, OCT_NO_TRANSFORM ));
	    } else if ( PROPEQ( &prop, "GROUND" )) {
		OH_ASSERT(ohCreateInstance( schemFacet, &inst, "ground", 
				 "$OCTTOOLS/tech/schematic/symbols/ground", "schematic", 
				 netBBcenter.x , facetBB.lowerLeft.y - 100 , OCT_NO_TRANSFORM ));
	    }
	    ohGetOrCreateBag( schemFacet, &bag, "INSTANCES" );
	    octAttach( &bag, &inst );
	    OH_ASSERT(octGenFirstContent( &inst, OCT_TERM_MASK, &term ));
	    OH_ASSERT(octAttach( &net, &term ));
	}
    }
}



extToSchematic( symFacet, schemFacet, minCap )
    octObject *symFacet;
    octObject *schemFacet;
    double minCap;
{
    octObject prop;
    
    ohCreateOrModifyPropStr( schemFacet, &prop, "TECHNOLOGY", "schematic" );
    ohCreateOrModifyPropStr( schemFacet, &prop, "VIEWTYPE", "SCHEMATIC" );
    ohCreateOrModifyPropStr( schemFacet, &prop, "EDITSTYLE", "SCHEMATIC" );

    findPowerAndGround( symFacet, schemFacet );

    doInstances( symFacet, schemFacet );

    doFormals( symFacet, schemFacet );

    doParasitics( symFacet, schemFacet, minCap );

    doPowerAndGround( schemFacet );
    implementNets( schemFacet );
}
