#!/bin/csh -f
#
set B = "octdrc:"

if ($?OCTTOOLS) then
	set octtools = $OCTTOOLS
else
	set octtools = _CADRUNDIR_
endif

if ( $?VOV_HOST_NAME ) then
	echo $B "I must disable tracing. Use vov_octdrc if you want tracing" 
	unsetenv VOV_HOST_NAME
endif


set SCRIPTS = scmos/octmm-scripts

set nwell = 1
set pwell = 0
set twinwell = 0
set fullusage = 0
set usage = 0
set flatten = 1
set outViewName = drc

while ($#argv)
	switch($1)
	case -n:
		set nwell = 1
		set pwell = 0
		set twinwell = 0
		breaksw
	case -p:
		set pwell = 1
		set nwell = 0
		set twinwell = 0
		breaksw
	case -t:
		set twinwell = 1
		set pwell = 0
		set nwell = 0
		breaksw
	case -o:
		set outViewName = $2
		shift
		breaksw
	case -P:
		set flatten = 0
		breaksw
	case -help:
		set fullusage = 1
		goto usage
		breaksw
	default:
        set option = `echo $1 | sed 's/^-/this is an option/'`
        if ( $#option > 1 ) then
            echo "octdrc: unknown option" $1
	goto usage
        else 
            break
        endif
 	endsw
	shift
end

if ($#argv < 1) then
   goto usage
endif

set cell = `echo $1 | sed 's/:.*//'`
set flatView = `echo $1 | sed 's/.*://'`

#//////////////////////////////////////////////////////
if (($nwell && $pwell) || ($nwell && $twinwell) || ($pwell && $twinwell)) then
   echo $B "Illegal combination of options."
   goto usage
endif

if ($flatten) then
   echo $B "Flattening $1."
   octflatten -t PHYSICAL -o :flat-drc $1 >& /dev/null
   set flatView = flat-drc
endif

if ($nwell) then
   echo $B "Running DRC, using nwell script" 
   octmm -f $SCRIPTS/nwell.drc -o :$outViewName $cell\:$flatView
   if ($flatten) then 
      goto cleanup
   endif
endif

if ($pwell) then 
   echo $B "Running DRC, using pwell script" 
   octmm -f $SCRIPTS/pwell.drc -o :$outViewName $cell\:$flatView
   if ($flatten) then 
      goto cleanup
   endif
endif

if ($twinwell) then 
  echo $B "Running DRC, using twinwell script" 
  octmm -f $SCRIPTS/twinwell.drc -o :$outViewName $cell\:$flatView
  if ($flatten) then 
     goto cleanup
  endif
endif
exit 0

#//////////////////////////////////////////////////////
usage:
echo "Usage is: octdrc [options] cell:view"
if ($fullusage) then
echo "Options are"
echo "     -n         Use Nwell rule set. (Default)"
echo "     -p         Use Pwell rule set."
echo "     -t         Use twinwell rule set."
echo "     -o :view   Specify the output view name. (:drc is the default)"
echo "     -P         Don't flatten the input cell. This should"
echo "                be used ONLY if the input cell is allready a"
echo "                flat PHYSICAL view."
echo "     -help      Print this message."
else 
echo "Use: octdrc -help for more information."
endif
exit 0


#//////////////////////////////////////////////////////
cleanup:
echo $B "Cleaning up intermediate views."
rm -rf $cell\/flat-drc
exit 0
