#include "octopus.h"

extern octObject *getInst();

extern s_arrays Arrays;
extern s_cells Cells;
extern int AndCols, OrCols, Rows;
extern s_input **Inputs;
extern s_output **Outputs;
extern s_ruleNames RuleNames;
extern s_options Options;

/* file globals to keep track of the state of folding of the PLA	*/
int 	TopInputs = FALSE;
int 	BottomInputs = FALSE;
int 	TopOutputs = FALSE;
int 	BottomOutputs = FALSE;


/* 
 * createIOBuffers
 *
 * instanciate and attach the input and output buffers
 */
createIOBuffers()
{
    if( Options.clocking & CLOCKED_OUTPUT ){
	createSingleOutputBuffers();
    } else {
	createOutputBuffers();
    }
    createInputBuffers();
    doPower();
}

/*
 * createSingleOutputBuffers
 *
 * instanciate and attach only single output buffers
 */
createSingleOutputBuffers()
{
    octObject *getInst();
    octObject *oInst, oTerm;
    char *output1Name;
    int col;

    if( Options.clocking & CLOCKED_OUTPUT ){
	output1Name = Cells.output1_clk.name;
    } else {
	output1Name = Cells.output1.name;
    }

    for( col = 0; col < OrCols; col ++ ){

	if( Outputs[ col ] == NIL(s_output)) continue;

	TopOutputs = TRUE;

	oInst = getInst( output1Name, OCT_NO_TRANSFORM );
	promoteAterm( oInst, "OUT", Outputs[ col ]->name, "SIGNAL", "OUTPUT" );
	mkAttach( Arrays.output, mkArea( 6, 11, col, col ), oInst );
	OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	mkAttach( Arrays.outputTerms, mkPoint( 1, col ), &oTerm);
	mkAttach( Arrays.output, mkPoint( 11, col ), &oTerm);

	mkAttach( Arrays.orPullup, mkPoint( 1, col ),
			     getInst( Cells.orPullup.name, OCT_NO_TRANSFORM ));

	if( Outputs[ col ]->next == NIL(s_output)) continue;

	BottomOutputs = TRUE;

	oInst = getInst( output1Name, OCT_MIRROR_Y );
	promoteAterm( oInst, "OUT", Outputs[ col ]->next->name, "SIGNAL", "OUTPUT" );
	mkAttach( Arrays.output, mkArea( 0, 5, col, col ), oInst );
	OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	mkAttach( Arrays.outputTerms, mkPoint( 0, col ), &oTerm);
	mkAttach( Arrays.output, mkPoint( 0, col ), &oTerm);

	mkAttach( Arrays.orPullup, mkPoint( 0, col ),
			     getInst( Cells.orPullup.name, OCT_MIRROR_Y ));
    }
}

/* 
 * createOutputBuffers
 *
 * instanciate and attach the output buffers
 */
createOutputBuffers()
{
    octObject *getInst();
    octObject *oInst, oTerm;
    char *output1Name, *output2Name;
    int col;
    int thisTop, nextTop, thisBot, nextBot;
    s_output *outThisTop, *outNextTop;

    if( Options.clocking & CLOCKED_OUTPUT ){
	/* this routine should not be called with CLOCKED_OUTPUT */
	fprintf( stderr, "Warning: Clocked outputs in createOutputBuffers\n");
	output1Name = Cells.output1_clk.name;
	output2Name = Cells.output2_clk.name;
    } else {
	output1Name = Cells.output1.name;
	output2Name = Cells.output2.name;
    }

    for( col = 0; col < OrCols; col += 2 ){

	thisTop = nextTop = thisBot = nextBot = FALSE;

	if( Outputs[ col ] != NIL(s_output)) {
	    if( Outputs[ col ]->bottom != 0) {
		thisTop = TRUE;
		outThisTop = Outputs[ col ];
	    } else if( Outputs[ col ]->top != Rows - 1)
		thisBot = TRUE;
	    else { /*could be top or bottom (check user constraints...)	*/ 
		thisTop = TRUE;
		outThisTop = Outputs[ col ];
	    }
	    if((Outputs[ col ]->next != NIL(s_output))) {
		if( thisTop ) {
		    errRaise( optProgName, -1,
			"two top outputs, col %d.", col);
		}
		thisTop = TRUE;
		outThisTop = Outputs[ col ]->next;
	    }
	}
	if(( col + 1 < OrCols) && ( Outputs[ col + 1 ] != NIL(s_output))) {
	    if( Outputs[ col + 1 ]->bottom != 0) {
		nextTop = TRUE;
		outNextTop = Outputs[ col + 1 ];
	    } else if( Outputs[ col + 1 ]->top != Rows - 1)
		nextBot = TRUE;
	    else { /*could be top or bottom (check user constraints...)	*/ 
		nextTop = TRUE;
		outNextTop = Outputs[ col + 1 ];
	    }
	    if((Outputs[ col + 1 ]->next != NIL(s_output))) {
		if( nextTop ) {
		    errRaise( optProgName, -1,
			"two top outputs, col %d.", col + 1);
		}
		nextTop = TRUE;
		outNextTop = Outputs[ col + 1 ]->next;
	    }
	}
	if( thisTop || nextTop ) TopOutputs = TRUE;
	if( thisBot || nextBot ) BottomOutputs = TRUE;

	if( thisTop && nextTop ){
	    oInst = getInst( output2Name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "OUT", outThisTop->name, "SIGNAL", "OUTPUT" );
	    promoteAterm( oInst, "OUT1", outNextTop->name, "SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 6, 11, col, col + 1), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 1, col ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 11, col ), &oTerm);
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN1" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 1, col + 1 ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 11, col + 1 ), &oTerm);
	} else if ( thisTop ){
	    oInst = getInst( output1Name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "OUT", outThisTop->name, "SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 6, 11, col, col ), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 1, col ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 11, col ), &oTerm);
	} else if ( nextTop ){
	    oInst = getInst( output1Name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "OUT1", outNextTop->name, "SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 6, 11, col + 1, col + 1 ), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 1, col + 1 ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 11, col + 1 ), &oTerm);
	}
	if( thisBot && nextBot ){
	    oInst = getInst( output2Name, OCT_MIRROR_Y );
	    promoteAterm( oInst, "OUT", Outputs[ col ]->name,
			"SIGNAL", "OUTPUT" );
	    promoteAterm( oInst, "OUT1", Outputs[ col + 1 ]->name,
			"SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 0, 5, col, col + 1), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 0, col ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 0, col ), &oTerm);
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN1" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 0, col + 1 ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 0, col + 1 ), &oTerm);
	} else if ( thisBot ){
	    oInst = getInst( output1Name, OCT_MIRROR_Y );
	    promoteAterm( oInst, "OUT", Outputs[ col ]->name,
			"SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 0, 5, col, col ), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 0, col ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 0, col ), &oTerm);
	} else if ( nextBot ){
	    oInst = getInst( output1Name, OCT_ROT180 );
	    promoteAterm( oInst, "OUT", Outputs[ col + 1 ]->name,
			"SIGNAL", "OUTPUT" );
	    mkAttach( Arrays.output, mkArea( 0, 5, col + 1, col + 1 ), oInst );
	    OH_ASSERT( ohGetByTermName( oInst, &oTerm, "IN" ));
	    mkAttach( Arrays.outputTerms, mkPoint( 0, col + 1 ), &oTerm);
	    mkAttach( Arrays.output, mkPoint( 0, col + 1 ), &oTerm);

	}

	if( thisTop )
	    mkAttach( Arrays.orPullup, mkPoint( 1, col ),
				 getInst( Cells.orPullup.name,
						OCT_NO_TRANSFORM ));
	if( nextTop )
	    mkAttach( Arrays.orPullup, mkPoint( 1, col + 1 ),
				 getInst( Cells.orPullup.name,
						OCT_NO_TRANSFORM ));
	if( thisBot )
	    mkAttach( Arrays.orPullup, mkPoint( 0, col ),
				 getInst( Cells.orPullup.name, OCT_MIRROR_Y ));
	if( nextBot )
	    mkAttach( Arrays.orPullup, mkPoint( 0, col + 1 ),
				 getInst( Cells.orPullup.name, OCT_MIRROR_Y ));
    }
}

s_inputCol	*InCols;
/*
 * createInputBuffers
 *
 * Instantiate and wire the input buffers, being careful to place them
 * to try to minimize core stretching.
 *
 * Algorithm:
 *  (1) make an array representation of the inputs
 *	(a) one entry (with a top and bottom) per column
 *	(b) classify inputs as single, empty, or (default) double.
 *	(c) if only one input in column, store its data in the bottom field
 *	    of the structure and mark it as unplaced (ie it could be placed
 *	    on either the top or bottom of the core).
 * (2) place the unplaced inputs
 *	(a) partition the unplaced inputs into sets of adjacent unplaced
 *	    inputs, bounded by either columns with placed inputs or
 *	    edge of the core.
 *	(b) for each continuous partition (segment) of unplaced inputs...
 *	    (i) look at the boundry columns and determine which boundry
 *		will stretch the core the most if it is not placed
 *		optimally against.
 *	   (ii) place the input column adjacent to that boundry optimally.
 *		Then, to place the rest of the inputs in that segment,
 *		just alternate sides of the core.
 */
createInputBuffers()
{
    s_inputCol	*newInputColArray();
    s_input	*input;
    int		i, j, folded;

    InCols = newInputColArray( AndCols );
    for( i = 0; i < AndCols; ++i) {
	InCols[i].bottom = InCols[i].top = IN_EMPTY;
    }

    /* classify the input and store the data in an array		*/
    for( i = 0; i < AndCols; ++i) {
	for( input = Inputs[i]; input != NIL(s_input); input = input->next) {
	    /* if its a single, place it				*/
	    if(( input->comp == NIL(s_input)) || (input->comp->col < 0)) {
		if( input->bottom == 0) {
		    if( input->top == Rows - 1) {
			InCols[i].bottom = IN_UNPLACED | IN_SINGLE;
		    } else {
			InCols[i].bottom = IN_SINGLE;
		    }
		} else {
		    InCols[i].top = IN_SINGLE;
		}
	    } else { /* else its part of comlement pair			*/
		/* if its the left side of pair, deal with it		*/
		if( input->col < input->comp->col) {
		    /* if there both alone in there rows...		*/
		    if(( input->bottom == 0) && (input->top == Rows - 1)
			&& (input->comp->bottom == 0)
			&& (input->comp->top == Rows - 1)) {
			InCols[i].bottom = IN_UNPLACED | IN_LEFT_SIDE;
			InCols[input->comp->col].bottom = IN_UNPLACED;
		    } else { /* its constrained somewhere		*/
			if(( input->bottom != 0) 
			    || ( input->comp->bottom != 0)) {
			    InCols[i].top = IN_LEFT_SIDE;
			    InCols[input->comp->col].top = 0;
			} else {
			    /* must be constriained to the bottom	*/
			    InCols[i].bottom = IN_LEFT_SIDE;
			    InCols[input->comp->col].bottom = 0;
			}
		    }
		} /* else its the right side, so its been handlesd	*/
	    }
	}
    }

    /* see if inputs are folded (ie if there are any cols with multiple	*/
    /* inputs). If not, place all inputs on top.			*/
    folded = FALSE;
    for( i = 0; (i < AndCols) && !(folded); ++i) {
	if( !( InCols[i].bottom & (IN_UNPLACED | IN_EMPTY)) ) {
	    folded = TRUE;
	}
    }

    if( !folded ) {
	/* all inputs can go on one side, so put them all on top	*/
	for( i = 0; i < AndCols; ++i ) {
	    InCols[i].top = InCols[i].bottom & ~(IN_UNPLACED);
	    InCols[i].bottom = IN_EMPTY;
	}
    } else {
	/* Inputs are folded, so handle the general case.		*/
	/* Assign locations to single-boundry continuous sets		*/
	/* of unplaced inputs.						*/
	i = j = 0;
	while((i < AndCols) && (j < AndCols)) { 
	    while(( i < AndCols) && (InCols[i].bottom & IN_UNPLACED)) {
		++i;
	    }
	    if( i != j ) {
		placeInputSegment( i - 1, j);
		j = i;
	    }
	    while(( j < AndCols) && !(InCols[j].bottom & IN_UNPLACED)) {
		++j;
	    }
	    i = j;
	}

#ifdef DEBUG_INPUTS
	for( i = 0; i < AndCols; ++i) {
	    if( InCols[i].bottom & IN_SINGLE) {
		printf("S ");
	    } else if( InCols[i].bottom & IN_EMPTY) {
		printf("E ");
	    } else {
		printf("D ");
	    }
	    if( InCols[i].top & IN_SINGLE) {
		printf("S\n");
	    } else if( InCols[i].top & IN_EMPTY) {
		printf("E\n");
	    } else {
		printf("D\n");
	    }
	}
#endif

    }

    attachInputBuffers();

    FREE( InCols );
}

/*
 * placeInputSegment
 *
 * Place the inputs within the segment.
 * Assume:
 *  (1) all inputs within the segment can be placed on either side of the core.
 *  (2) empty cols are bounries, so they will not be within a segment.
 *  (3) double phase inputs are in adjacent columns
 * Placement strategy is simple:
 *  (1) figure out which boundry interaction is most important to satisfy
 *  (2) satisfy that requirement, then alternate placement sides
 *      for the rest of the inputs.
 */
placeInputSegment( col1, col2)
int	col1, col2;
{
    int		start, end, groups, i;
    int		rPriority, lPriority;
    coreSide	rLargerSide, lLargerSide, placeSide;

    /* sort input columns						*/
    if( col1 < col2) {
	start = col1;
	end = col2;
    } else {
	start = col2;
	end = col1;
    }

    /* classify boundry columns						*/
    classifyBoundryCol( start - 1, &lPriority, &lLargerSide);
    classifyBoundryCol( end + 1, &rPriority, &rLargerSide);

    /* count the number of groups to place (double phase buffers are 1)	*/
    groups = 0;
    for( i = start; i <= end; ++i) {
	if( InCols[i].bottom & IN_SINGLE)
	    ++groups;
	else if( InCols[i].bottom & IN_EMPTY) {
	    errRaise( optProgName, -1,
		"Bad call to placeInputSegment(%d, %d), %s %d.",
		col1, col2, "segment contains empty col at", i);
	} else { /* double phase, takes 2 cols but its 1 group		*/
	    ++i;
	    ++groups;
	}
    }

    /* place members of the segment (ie, decide position of first one)	*/
    /* if one priority dominates, use that				*/
    if( rPriority != lPriority ) {
	if( lPriority > rPriority ) {
	    if( lLargerSide == TOP ) {
		placeSide = BOTTOM;
	    } else {
		placeSide = TOP;
	    }
	} else {
	    if( rLargerSide == TOP ) {
		placeSide = (groups % 2) ? BOTTOM : TOP;
	    } else {
		placeSide = (groups % 2) ? TOP : BOTTOM;
	    }
	}
    } else { /* else look at boundry-{start,end} col interaction	*/
	/*** get fancy later						*/
	placeSide = TOP;
    }
    for( i = start; i <= end; ++i) {
	if( placeSide == TOP) {
	    if( !( InCols[i].bottom & IN_LEFT_SIDE)) {
		placeSide = BOTTOM;
	    }
	    InCols[i].top = InCols[i].bottom & ~(IN_UNPLACED);
	    InCols[i].bottom = IN_EMPTY;
	} else {
	    if( !( InCols[i].bottom & IN_LEFT_SIDE)) {
		placeSide = TOP;
	    }
	    InCols[i].bottom &= ~(IN_UNPLACED);
	}
    }
}

/*
 * classifyBoundryCol
 *
 * Assign a priority level to the boundry input column passed in.
 * The priority gives an ordering to the benefit recieved if the neighboring
 * unplaced input column is placed minimally against this boundry column.
 * The higher number idicates higher priority.
 * Also return which side is "larger" (ie which side (TOP or BOTTOM) infuences
 * the placement the most).
 */
classifyBoundryCol( col, pri, lSide)
int		col, *pri; /* col number and where to return priority	*/
coreSide	*lSide; /* which side of column is bigger		*/
{
    if(( col < 0) || ( col >= AndCols)
	|| ( InCols[col].top == InCols[col].bottom)){
	*pri = 0;
	return;
    }

    /* else 3 possible configurations: (E,S), (E,D), and (D,S)		*/
    if( InCols[col].bottom == IN_EMPTY) {
	if( InCols[col].top & IN_SINGLE) {
	    *pri = 3;
	    *lSide = TOP;
	} else { /* must be part of a double				*/
	    *pri = 2;
	    *lSide = TOP;
	}
    } else if( InCols[col].bottom & IN_SINGLE) {
	if( InCols[col].top & IN_EMPTY) {
	    *pri = 3;
	    *lSide = BOTTOM;
	} else { /* must be part of a double				*/
	    *pri = 1;
	    *lSide = BOTTOM;
	}
    } else { /* bottom is double					*/
	if( InCols[col].top & IN_EMPTY) {
	    *pri = 2;
	    *lSide = BOTTOM;
	} else { /* must be part of a single				*/
	    *pri = 1;
	    *lSide = TOP;
	}
    }
}


/*
 * attachInputBuffers
 *
 * attach the input buffers based on the InCols array.
 */
attachInputBuffers()
{
    int		col;
    octObject	*oInst, *getInst(), oTerm;
    s_input	*input;
    struct octBox *instArea;
    char *inputName;

    if( Options.clocking & CLOCKED_INPUT ){
	inputName = Cells.input_clk.name;
    } else {
	inputName = Cells.input.name;
    }

    col = 0;
    TopInputs = BottomInputs = FALSE;
    /* do bottom inputs						*/
    while( col < AndCols) {
	if( InCols[col].bottom & IN_EMPTY) {
	    ++col;
	    continue;
	} else if( InCols[col].bottom & IN_SINGLE) {
	    input = Inputs[col];
	    if(( col == 0) || ((InCols[col-1].bottom & IN_EMPTY)
			&& !(InCols[col-1].bottom & IN_OVERHANG))) {
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_ROT180);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_MIRROR_Y);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		if( col == 0) instArea = mkArea( 0, 5, col, col );
		else instArea = mkArea( 0, 5, col, col-1 );
	    } else if(( col == AndCols-1) || (InCols[col+1].bottom & IN_EMPTY)){
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_MIRROR_Y);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_ROT180);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		if ( col + 1 < AndCols ) 
		    InCols[col+1].bottom |= IN_OVERHANG;
		if( col == AndCols - 1) instArea = mkArea( 0, 5, col, col );
		else instArea = mkArea( 0, 5, col, col+1 );
	    } else {
		/* overhang it left (could go either way)		*/
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_ROT180);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_MIRROR_Y);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		instArea = mkArea( 0, 5, col, col);
	    }
	    if( input->phase == POS) {
		promoteAterm( oInst, "IN", input->name, "SIGNAL", "INPUT" );
	    } else {
		promoteAterm( oInst, "IN", input->comp->name, "SIGNAL", "INPUT" );
	    }
	    mkAttach( Arrays.inputTerms, mkPoint( 0, col), &oTerm);
	    mkAttach( Arrays.input, instArea, oInst);
	    BottomInputs = TRUE;
	    ++col;
	} else {
	    /* its a double input					*/
	    input = Inputs[col];
	    if( input->phase == POS) {
		oInst = getInst( inputName, OCT_MIRROR_Y);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		mkAttach( Arrays.inputTerms, mkPoint( 0, col), &oTerm);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		mkAttach( Arrays.inputTerms, mkPoint( 0, col+1), &oTerm);
	    } else {
		oInst = getInst( inputName, OCT_ROT180);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		mkAttach( Arrays.inputTerms, mkPoint( 0, col), &oTerm);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		mkAttach( Arrays.inputTerms, mkPoint( 0, col+1), &oTerm);
	    }
	    if( input->phase == POS) {
		promoteAterm( oInst, "IN", input->name, "SIGNAL", "INPUT" );
	    } else {
		promoteAterm( oInst, "IN", input->comp->name,
				"SIGNAL", "INPUT" );
	    }
	    mkAttach( Arrays.input, mkArea( 0, 5, col, col + 1), oInst);
	    BottomInputs = TRUE;
	    col += 2;
	}
    }

    col = 0;
    /* do top inputs						*/
    while( col < AndCols) {
	if( InCols[col].top & IN_EMPTY) {
	    ++col;
	    continue;
	} else if( InCols[col].top & IN_SINGLE) {
	    if( Inputs[col]->next == NIL(s_input)) {
		input = Inputs[col];
	    } else {
		input = Inputs[col]->next;
	    }
	    if(( col == 0) || ((InCols[col-1].top & IN_EMPTY)
			&& !(InCols[col-1].top & IN_OVERHANG))) {
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_MIRROR_X);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_NO_TRANSFORM);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		if( col == 0) instArea = mkArea( 6, 11, col, col );
		else instArea = mkArea( 6, 11, col, col-1 );
	    } else if(( col == AndCols-1) || (InCols[col+1].top & IN_EMPTY)){
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_NO_TRANSFORM);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_MIRROR_X);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		if ( col + 1 < AndCols ) 
		    InCols[col+1].top |= IN_OVERHANG;
		if( col == AndCols - 1) instArea = mkArea( 6, 11, col, col );
		else instArea = mkArea( 6, 11, col, col+1 );
	    } else {
		/* overhang it left (could go either way)		*/
		if( input->phase == POS) {
		    oInst = getInst( inputName, OCT_MIRROR_X);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		} else {
		    oInst = getInst( inputName, OCT_NO_TRANSFORM);
		    OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		}
		instArea = mkArea( 6, 11, col, col);
	    }
	    if( input->phase == POS) {
		promoteAterm( oInst, "IN", input->name, "SIGNAL", "INPUT" );
	    } else {
		promoteAterm( oInst, "IN", input->comp->name, "SIGNAL","INPUT");
	    }
	    mkAttach( Arrays.inputTerms, mkPoint( 1, col), &oTerm);
	    mkAttach( Arrays.input, instArea, oInst);
	    TopInputs = TRUE;
	    ++col;
	} else {
	    /* its a double input					*/
	    if( Inputs[col]->next == NIL(s_input)) {
		input = Inputs[col];
	    } else {
		input = Inputs[col]->next;
	    }
	    if( input->phase == POS) {
		oInst = getInst( inputName, OCT_NO_TRANSFORM);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		mkAttach( Arrays.inputTerms, mkPoint( 1, col), &oTerm);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		mkAttach( Arrays.inputTerms, mkPoint( 1, col+1), &oTerm);
	    } else {
		oInst = getInst( inputName, OCT_MIRROR_X);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "NEG"));
		mkAttach( Arrays.inputTerms, mkPoint( 1, col), &oTerm);
		OH_ASSERT( ohGetByTermName(oInst, &oTerm, "POS"));
		mkAttach( Arrays.inputTerms, mkPoint( 1, col+1), &oTerm);
	    }
	    if( input->phase == POS) {
		promoteAterm( oInst, "IN", input->name, "SIGNAL", "INPUT" );
	    } else {
		promoteAterm( oInst, "IN", input->comp->name, "SIGNAL", "INPUT" );
	    }
	    mkAttach( Arrays.input, mkArea( 6, 11, col, col + 1), oInst);
	    TopInputs = TRUE;
	    col += 2;
	}
    }
}

/*
 * doBufferVddVss
 *
 * adds the lVss, lVdd and cVss instances to the array
 */
doPower()
{
    octObject *oInst;
    struct octBox ac, ar, ab, ap;

    /* the wires connecting to the input buffer GND terminals		*/
    if( TopInputs ){
	/* the pom1 and vdd_vias to power the andPullups		*/
	oInst = getInst( Cells.vdd_via.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 1, 0 ), oInst );

	oInst = getInst( Cells.m1m1_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 3, 1 ), oInst );
	oInst = getInst( Cells.m1m1_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 3, 0 ), oInst );

	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 2, 1 ), oInst );
	oInst = getInst( Cells.pom1.name, OCT_ROT90 );
	mkAttach( Arrays.vss, mkPoint( 2, 0 ), oInst );

	
	if( Options.clocking & CLOCKED_INPUT ){

	    oInst = getInst( Cells.pom1_1.name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "t", "inPhi", "CLOCK", "INPUT" );
	    mkAttach( Arrays.inPhi, mkPoint( 1, 0 ), oInst );

	    oInst = getInst( Cells.pom1_2.name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "t", "inPhiBar", "CLOCK", "INPUT" );
	    mkAttach( Arrays.inPhiBar, mkPoint( 1, 0 ), oInst );
	}
    }
    if( BottomInputs ){
	/* the pom1 and vdd_vias to power the andPullups		*/
	oInst = getInst( Cells.vdd_via.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 0, 0 ), oInst );

	oInst = getInst( Cells.m1m1_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 0, 1 ), oInst );
	oInst = getInst( Cells.m1m1_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 0, 0 ), oInst );

	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 1, 1 ), oInst );
	oInst = getInst( Cells.pom1.name, OCT_ROT270 );
	mkAttach( Arrays.vss, mkPoint( 1, 0 ), oInst );

	if( Options.clocking & CLOCKED_INPUT ){

	    oInst = getInst( Cells.pom1_1.name, OCT_NO_TRANSFORM );
	    if( TopOutputs && Options.extraTerms ){
		promoteAterm( oInst, "t", "inPhi1", "CLOCK", "INPUT" );
	    } else {
		promoteAterm( oInst, "t", "inPhi", "CLOCK", "INPUT" );
	    }
	    mkAttach( Arrays.inPhi, mkPoint( 0, 0 ), oInst );

	    oInst = getInst( Cells.pom1_2.name, OCT_NO_TRANSFORM );
	    if( TopOutputs && Options.extraTerms ){
		promoteAterm( oInst, "t", "inPhiBar1", "CLOCK", "INPUT" );
	    } else {
		promoteAterm( oInst, "t", "inPhiBar", "CLOCK", "INPUT" );
	    }
	    mkAttach( Arrays.inPhiBar, mkPoint( 0, 0 ), oInst );
	}
    }
    /* the vdd lines between input and output buffers			*/
    if( TopInputs && TopOutputs ){
	oInst = getInst( Cells.m2m2.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 1, 1 ), oInst );
	oInst = getInst( Cells.m2m2_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 1, 1 ), oInst );
    }
    if( TopOutputs ){
	oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 3, 1 ), oInst );
    }
    if( BottomInputs && BottomOutputs ){
	oInst = getInst( Cells.m2m2.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 0, 1 ), oInst );
	oInst = getInst( Cells.m2m2_1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vdd, mkPoint( 0, 1 ), oInst );
    }
    if( BottomOutputs ){
	oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.vss, mkPoint( 0, 1 ), oInst );
    }

    /* the VSS and VDD contacts and formal terms			*/
    if( TopOutputs ){
	oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
	promoteAterm( oInst, "t", "GND", "GROUND", "INPUT" );
	mkAttach( Arrays.vss, mkPoint( 3, 2 ), oInst );

	oInst = getInst( Cells.vdd_via.name, OCT_NO_TRANSFORM );
	promoteAterm( oInst, "t", "Vdd", "SUPPLY", "INPUT" );
	mkAttach( Arrays.vdd, mkPoint( 1, 2 ), oInst );

	if( Options.clocking & CLOCKED_OUTPUT ){
	    oInst = getInst( Cells.pom1_1.name, OCT_NO_TRANSFORM );
	    promoteAterm( oInst, "t", "outPhi", "CLOCK", "INPUT" );
	    mkAttach( Arrays.outPhi, mkPoint( 1, 0 ), oInst );
	}
    }

    if( BottomOutputs ){
	oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
	if( TopOutputs && Options.extraTerms ){
	    promoteAterm( oInst, "t", "GND", "GROUND", "INPUT" );
	} else {
	    promoteAterm( oInst, "t", "GND", "GROUND", "INPUT" );
	}
	mkAttach( Arrays.vss, mkPoint( 0, 2 ), oInst );

	oInst = getInst( Cells.vdd_via.name, OCT_NO_TRANSFORM );
	if( TopOutputs && Options.extraTerms ){
	    promoteAterm( oInst, "t", "Vdd", "SUPPLY", "INPUT" );
	} else {
	    promoteAterm( oInst, "t", "Vdd", "SUPPLY", "INPUT" );
	}
	mkAttach( Arrays.vdd, mkPoint( 0, 2 ), oInst );

	if( Options.clocking & CLOCKED_OUTPUT ){
	    oInst = getInst( Cells.pom1_1.name, OCT_NO_TRANSFORM );
	    if( TopOutputs && Options.extraTerms ){
		promoteAterm( oInst, "t", "outPhi1", "CLOCK", "INPUT" );
	    } else {
		promoteAterm( oInst, "t", "outPhi", "CLOCK", "INPUT" );
	    }
	    mkAttach( Arrays.outPhi, mkPoint( 0, 0 ), oInst );
	}
    }

    /* the contacts to the or plane pullups				*/

    ar = *mkTranslate( Arrays.vdd, mkCol( 2 ));
    ac = *mkTranslate( Arrays.vss, mkCol( 1 ));
    ap = *mkTranslate( Arrays.andOrInterface, mkCol( 0 ));
    ab = *mkTranslate( Arrays.orPullup, mkRow( 0 ));

    if( BottomOutputs ){
	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ar, &ab ), oInst);
	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ac, &ab ), oInst);
	oInst = getInst( Cells.pom1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ap, &ab ), oInst);
    }

    ab = *mkTranslate( Arrays.orPullup, mkRow( 1 ));

    if( TopOutputs ){
	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ar, &ab ), oInst );
	oInst = getInst( Cells.m1m1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ac, &ab ), oInst);
	oInst = getInst( Cells.pom1.name, OCT_NO_TRANSFORM );
	mkAttach( Arrays.full, mkIntersection( &ap, &ab ), oInst);
    }

    oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
    ab = *mkTranslate( Arrays.orPower, mkRow( 0 ));
    mkAttach( Arrays.full, mkIntersection( &ac, &ab ), oInst );

    oInst = getInst( Cells.vss_via.name, OCT_NO_TRANSFORM );
    ab = *mkTranslate( Arrays.orPower, mkRow( 1 ));
    mkAttach( Arrays.full, mkIntersection( &ac, &ab ), oInst );
}
