#include "octopus.h"

extern int	Rows, AndCols, OrCols, OddEvenBreak;
extern char	**Core;
extern s_arrays	Arrays;
extern s_cells	Cells;
extern s_output	**Outputs;
extern s_input	**Inputs;
extern s_ruleNames RuleNames;
extern s_options Options;

octObject	*getInst();

/*
 * createPlaCore
 *
 * Instanciate and wire together all transistors, signal connectors,
 * and ground connectors in the core.
 */
createPlaCore()
{
    doOrPlaneOutputs();
    doProductTerms();
    doLiterals();
    doGround();
}

/*
 * doOrPlaneOutputs
 *
 * attach the OR plane contacts and wire the outputs.
 */
doOrPlaneOutputs()
{
    int		row, col;
    s_output	*output;
    struct octBox t_area, b_area;

    for ( row = 0; row < Rows; row += 2) {
	for ( col = OddEvenBreak; col < OrCols; ++col) {
	    if ((IS_PRODUCT( row, col ))||
		(( row > 0 ) && (IS_PRODUCT( row - 1, col )))){
		mkAttach( Arrays.orEvenCont, mkPoint( row, col ),
			getInst( Cells.orCont.name, OCT_NO_TRANSFORM ));
	    }
	}
	for ( col = OddEvenBreak - 1; col >= 0; --col) {
	    if (( IS_PRODUCT( row, col )) ||
		(( row < Rows - 1 ) && ( IS_PRODUCT( row + 1, col )))){
		mkAttach( Arrays.orOddCont, mkPoint( row, col ),
			getInst( Cells.orCont.name, OCT_NO_TRANSFORM ));
	    }
	}
    }
    /* handle top row signal contacts on Even side when Rows is even	*/
    if( Rows % 2 == 0) {
	row = Rows - 1;
	for ( col = OddEvenBreak; col < OrCols; ++col) {
	    if ((IS_PRODUCT( row, col ))||
		(( row > 0 ) && (IS_PRODUCT( row - 1, col )))){
		mkAttach( Arrays.orEvenCont, mkPoint( row, col ),
			getInst( Cells.orCont.name, OCT_NO_TRANSFORM ));
	    }
	}
    }

    if( Options.wire ){
	for( col = 0; col < OrCols; col++ ){
	    for( output = Outputs[ col ]; output != NIL( s_output );
		 output = output->next ){
		/* wire the OR output column 				*/
		if(( output->bottom < 0) || ( output->bottom >= Rows) ||
		   ( output->top < 0) || ( output->top >= Rows)) {
		    continue; /* unconnected				*/
		}
		if( output->top == Rows - 1 ){
		    t_area = *mkTranslate( Arrays.outputTerms,
					mkPoint( 1, col));
		} else {
		    if( col < OddEvenBreak ) {
			t_area = *mkTranslate( Arrays.orOddCont, 
					    mkPoint( output->top, col));
		    } else {
			t_area = *mkTranslate( Arrays.orEvenCont, 
					    mkPoint( output->top, col));
		    }
		}
		if( output->bottom == 0 ){
		    b_area = *mkTranslate( Arrays.outputTerms, mkPoint(0, col));
		} else {
		    if( col < OddEvenBreak ) {
			b_area = *mkTranslate( Arrays.orOddCont, 
					    mkPoint( output->bottom, col));
		    } else {
			b_area = *mkTranslate( Arrays.orEvenCont, 
					    mkPoint( output->bottom, col));
		    }
		}
		mkWireArea( Arrays.full, 
				       mkUnionBB( &t_area, &b_area ),
				       RuleNames.orOutput );
	    }
	}
    }
}

/*
 * doProductTerms
 *
 * Instantiate and wire together the literal rows.  This includes (from left
 * to right) the literal pullups, the AND plane signal contacts, the
 * and-plane/or-plane interface contacts, and the OR plane transistors
 * and signal contacts.
 */
doProductTerms()
{
    int		row, col;

    for ( row = 0; row < Rows; ++row) {

	/* and plane pullups and andOrInterface contacts		*/

	mkAttach( Arrays.andPullup, mkRow( row ),
		getInst( Cells.andPullup.name, OCT_NO_TRANSFORM ));

	mkAttach( Arrays.andOrInterface, mkPoint( row, 0 ),
		getInst( Cells.andOrInterface.name, OCT_NO_TRANSFORM ));

	/* or plane transistors						*/
	for ( col = 0; col < OrCols; ++col){
	    if ( IS_PRODUCT( row, col )){
		if((( col >= OddEvenBreak )&&( row%2 ))||
		   (( col < OddEvenBreak )&&( row%2 == 0 ))){
		    mkAttach( Arrays.orTran, mkPoint( row, col ),
			    getInst( Cells.orTran.name, OCT_NO_TRANSFORM ));
		} else {
		    mkAttach( Arrays.orTran, mkPoint( row, col ),
			    getInst( Cells.orTran.name, OCT_ROT180 ));
		}

	    }
	}
	/* and plane contacts						*/
	for ( col = 0; col < AndCols; col += 2) {
	    if (( IS_LITERAL( row, col )) ||
		(( col < AndCols - 1 ) && (IS_LITERAL( row, col + 1 )))){
		mkAttach( Arrays.andCont, mkPoint( row, col ),
			getInst( Cells.andCont.name, OCT_NO_TRANSFORM ));
	    }
	}
	/* wire the row in the core by pattern   			*/
	if( Options.wire )
	    mkWireArea( Arrays.andPullup, mkRow( row ),
				   RuleNames.productTerm );
    }
    if( Options.wire )
	mkWireArea( Arrays.andPullup, mkCols( 0, 1 ),
			       RuleNames.andPullup );
}

/*
 * doLiterals
 *
 * attach the and plane transistors and wire their gates together by pattern.
 * In the wiring, be careful to wire top inputs to top input buffers and
 * bottom inputs to bottom input buffers.
 */
doLiterals()
{
    int row, col;
    s_input *input;
    struct octBox t_area, b_area;

    for ( row = 0; row < Rows; ++row) {
	for ( col = 0; col < AndCols; ++col){
	    if ( IS_LITERAL( row, col )){
		if( col%2 ){
		    mkAttach( Arrays.andTran, mkPoint( row, col ),
				    getInst( Cells.andTran.name, OCT_ROT90 ));
		} else {
		    mkAttach( Arrays.andTran, mkPoint( row, col ),
				    getInst( Cells.andTran.name, OCT_ROT270 ));
		}
	    }
	}
    }
    if( Options.wire ){
	for ( col = 0; col < AndCols; ++col){
	    for( input = Inputs[ col ]; input != NIL( s_input );
		 input = input->next ){
		if(( input->col == - 1) ||
		   ( input->bottom < 0) || ( input->bottom >= Rows) ||
		   ( input->top < 0) || ( input->top >= Rows)) {
		    continue; /* unconnected				*/
		}
		if( input->top == Rows - 1 ){
		    t_area = *mkTranslate( Arrays.inputTerms,
					    mkPoint( 1, col));
		} else {
		    t_area = *mkTranslate( Arrays.andTran,
					    mkPoint( input->top, col));
		}
		if( input->bottom == 0 ){
		    b_area = *mkTranslate( Arrays.inputTerms,
					    mkPoint( 0, col));
		} else {
		    b_area = *mkTranslate( Arrays.andTran,
					    mkPoint( input->bottom, col));
		}
		mkWireArea ( Arrays.full,
					    mkUnionBB( &t_area, &b_area ),
					    RuleNames.literal );
	    }
	}
    }
}
