#include "octopus.h"
#ifndef VERSION
#define VERSION "2.0"
#endif
#ifndef CUR_DATE
#define CUR_DATE "undated"
#endif
/*
 * octopus
 *
 * an OCT based PLA generator
 */
 
/* global variables */

s_cells		Cells;		/* contains cell names			*/
s_ruleNames	RuleNames;	/* names of the connection rules	*/
char		**Core;	    	/* sanitized personality matrix		*/
char		**GivenCore;    /* given personality matrix		*/
s_options	Options;	/* command line options			*/
s_arrays	Arrays;		/* the various mkarray mkId's		*/
octObject 	oFacet;		/* the output PLA facet			*/
octObject 	oPalette;	/* the palette of instances		*/

s_input		**Inputs;	/* parameters of sanitized matrix	*/
s_output	**Outputs;
int 		Rows;		/* number of rows in the PLA		*/
int		AndCols;	/* number of columns in the AND plane	*/
int		OrCols;		/* number of columns in the OR plane	*/
int		OddEvenBreak;	/* split in OR plane between fold types	*/

s_input		**GivenInputs; /* parameters of given matrix		*/
s_output	**GivenOutputs;
int 		GivenRows;
int		GivenAndCols;
int		GivenOrCols;

main(argc, argv)
int argc;
char *argv[];
{
    /* fprintf(stderr,"%s version %s\nmade %s\n", argv[0], VERSION, CUR_DATE ); */

    /* set Options defaults						*/
    optionsDefaults();

    /* set other defaults						*/
    setDefaults();
    setCellAndRuleNameDefaults();

    /* parse command line */
    parseComLine( argc, argv );

    /* read the input personality matrix 				*/
    readMatrix( Options.inFile);

    /* open the facet (This could be done later, but I'm using openFacet
     * to set the fields in oFacet so the call to printPLA will work.
     */
    octBegin();
    openFacet(Options.plaFacet); 

    /*
    if( Options.debug > 1 )printIn();
    */
    /* dump PLA(5) output */
    if(( Options.debug > 1) && (Options.plaVerifyFile != NIL(char))){
	printPLA( GivenRows, GivenOrCols, GivenCore, GivenOutputs);
    }

    /* fix any problems with the personality matrix			*/
    sanitizeMatrix();
    if( Options.debug > 0 )printTrue();

    /* dump PLA(5) output */
    if( Options.plaVerifyFile != NIL(char)){
	printPLA( Rows, OrCols, Core, Outputs);
    }

    /* open the palette							*/
    if( Options.localPaletteName != NIL( char )){
	openLocalPalette( Options.localPaletteName );
    } else {
	openPalette( Options.paletteName );
    }

    /* create the array and subArrays					*/
    createArrays();

    /* read the wiring rules from oPalette				*/
    mkReadRules( Arrays.full, &oPalette );

    /* instanciate and wire the buffers 				*/
    createIOBuffers();

    /* instanciate and wire the core of the PLA				*/
    createPlaCore();

    /* wiring the perimeter of the array				*/
    wirePerimeter();

    /* solve the array							*/
    mkPlaceArray( Arrays.full );

    /* close oFacet and exit						*/ 
    moveInterfaceProps();
    OH_ASSERT( octCloseFacet( &oFacet ));
    octEnd();
    exit(0);
}
