/*
 * octopus.h
 *
 * This is the header file for the program octopus.
 */

#include "port.h"
#include "oct.h"
#include "oh.h"
#include "mkarray.h"
#include "utility.h"
#include "options.h"
#include "errtrap.h"

/* general constants */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define STRLEN 200
#define NAMESIZE 80


#ifndef MIN
#define MIN(a,b) (a>b?b:a)
#endif
#ifndef MAX
#define MAX(a,b) (a>b?a:b)
#endif

/* assertion macros							*/
#define ASSERT(fct)\
	if ((fct) < 0) {\
	    fprintf( stderr, "An assertion failed: %s line %d\n", \
		    __FILE__, __LINE__ );\
	    exit(1);\
	}

/* instance types 							*/
typedef enum instanceType { INSTANCE, CONNECTOR } instanceType;

/* clocking constants for command line parsing */
#define CLOCKED_INPUT	0x1
#define CLOCKED_OUTPUT	0x2
#define CLOCKED_PRE_OUT	0x4	/* clocked prechaged output buffer */

/* bits for classifying inputs for placing them				*/
#define IN_UNPLACED	0x1	/* else placed				*/
#define IN_SINGLE	0x2	/* else both phases present		*/
#define IN_LEFT_SIDE	0x4	/*else right side of double phase input	*/
#define IN_OVERHANG	0x8	/* overhang of single input		*/
#define IN_EMPTY	0x10	/* nothing to connect at this location	*/

typedef enum coreSide { TOP, BOTTOM, LEFT, RIGHT } coreSide;

/* types of charactors in the matrix */
#define DOT '.'
#define SUM 's'
#define PRODUCT 'p'
#define CONNECTION 'c'

/* Core testing macros							*/
#define IS_LITERAL(row,col)\
	( Core[row][col] == SUM )

#define IS_PRODUCT(row,col)\
	( Core[row][col + AndCols] == PRODUCT )
	
/* data types */

typedef struct options {
    /* definitely using */
    char	*technology;	/* name of technology			*/
    char	*paletteName;	/* name of palette in tech dir		*/
    char	*localPaletteName;	/* name of local palette	*/
    int		debug;		/* debugging level			*/
    char	*inFile;	/* name of input personality matrix	*/
    char	*plaVerifyFile;	/* name of PLA(5) format verify file	*/
    int		wire;		/* bool. Add wiingres or not		*/
    char	*plaFacet;	/* name of output PLA facet		*/
    /* questionable */
    int		printGraph;	/* bool. Print mkarray constraint graph	*/
    int		andGndLineSep;	/* holds separation of met 1 gnd lines	*/
    int		orGndLineSep;	/* holds separation of met 1 gnd lines	*/
    int		clocking;	/* bitwise or-ing of clocking types	*/
    int		extraTerms;	/* produces extra formal terms if true	*/
} s_options;

typedef struct input {
    int		top, bottom, col;  /* column and range of input line	*/
    char	*name;		   /* signal name			*/
    enum phaseType {POS, NEG} phase; /* phase of input			*/
    struct input *comp;		   /* complement of this input		*/
    struct input *next;		   /* next in list at this collum	*/
} s_input;

typedef struct output {
    int		top, bottom, col;	/* column and range of output	*/
    char	*name;			/* signal name			*/
    enum outputType {ODD, EVEN, DONT_CARE} type; /* output type		*/
    struct output *next;		/* next in list			*/
} s_output;

typedef struct arrays {
    mkId full;			/* the full array			*/
    mkId andPullup;		/* AND plane pullups			*/
    mkId andPlaneGnd;		/* AND plane ground tracks		*/
    mkId andTran;		/* AND plane transistor sites		*/
    mkId andCont;		/* AND plane product term contacts 	*/
    mkId andGnd;		/* AND plane ground connectors		*/
    mkId andOrInterface;	/* connectors between AND and OR planes	*/
    mkId orPullup;		/* OR plane pullups			*/
    mkId orPower;		/* OR plane ground supply		*/
    mkId orOddPlaneGnd;		/* odd phase OR plane ground tracks	*/
    mkId orEvenPlaneGnd;	/* even phase OR plane ground tracks	*/
    mkId orTran;		/* OR plane transistor sites		*/
    mkId orEvenGnd;		/* even phase OR plane ground connectors*/
    mkId orOddGnd;		/* odd phase OR plane ground connectors	*/
    mkId orEvenCont;		/* even phase OR plane output contacts  */
    mkId orOddCont;		/* odd phase OR plane output contacts  	*/
    mkId inputTerms;		/* input buffer terminals 		*/
    mkId outputTerms;		/* output buffer terminals 		*/
    mkId input;			/* input buffers			*/
    mkId output;		/* output buffers			*/
    mkId vss;			/* left side VSS			*/
    mkId vdd;			/* left side VDD			*/
    mkId inPhi; 		/* input clock				*/
    mkId inPhiBar; 		/* input clock bar			*/
    mkId outPhi; 		/* output clock				*/
} s_arrays;

typedef struct cellStruct {
    char *name;
    char *master;
    char *view;
} s_cellStruct;

typedef struct cells {
    struct cellStruct andPullup;
    struct cellStruct andPlaneGnd;
    struct cellStruct andTran;
    struct cellStruct andGnd;
    struct cellStruct andCont;
    struct cellStruct andOrInterface;
    struct cellStruct orPullup;
    struct cellStruct orPlaneGnd;
    struct cellStruct orTran;
    struct cellStruct orGnd;
    struct cellStruct orCont;
    struct cellStruct input;
    struct cellStruct input_clk;
    struct cellStruct output1;
    struct cellStruct output1_clk;
    struct cellStruct output2;
    struct cellStruct output2_clk;
    struct cellStruct pom1;
    struct cellStruct pom1_1;
    struct cellStruct pom1_2;
    struct cellStruct m1m1;
    struct cellStruct m1m1_1;
    struct cellStruct m2m2;
    struct cellStruct m2m2_1;
    struct cellStruct vdd_via;
    struct cellStruct vss_via;
} s_cells;

typedef struct ruleNames {
    char *orOutput;
    char *productTerm;
    char *andGndDiffusion;
    char *andTranDiffusion;
    char *orGndDiffusion;
    char *orTranDiffusion;
    char *orGround;
    char *andGround;
    char *literal;
    char *andPullup;
    char *orPullup;
    char *buffer;
    char *cPower;
} s_ruleNames;

typedef struct inputCol {
    unsigned	top, bottom;
} s_inputCol;
