/*
 * parse commmand line and set values in options structure.
 */


#include "octopus.h"

optionStruct optionList[] = {
    { "i",     "input filename", "name of the input personality matrix."},
    { "T",     "technology", "name of the technology."},
    { "p",     "palette-name", "palette name known to TAP(3CAD)."},
    { "P",     "user-palette[:view[:facet[:version]]]",
				"name of user palette. Overrides the '-p' option."},
    { "C",     "buffer","specify clocking for buffers."},
    { OPT_CONT, 0,"\n    legal types are 'in', for clocked input buffer."},
    { OPT_CONT, 0,"\n                    'out', for clocked output buffer."},
    { OPT_CONT, 0,"\n                    'pre', for clocked precharged output buffer."},
    { "v",	"PLA(5CAD) filename", "output a PLA(5CAD) format file for verification/comparison with the given PLA(5CAD) file."},
    { "g",      "",           "prints the mkarray constraint graphs."},
    { "s",      "separation", "give ground line separation in AND plane."},
    { "S",      "separation", "give ground line separation in OR plane."},
    { "w",      "",           "suppress wiring."},
    { "d",     "level",	      "sets the level of debugging."},
    { OPT_CONT, 0,"\n    0   : run silently and quickly ( default )."},
    { OPT_CONT, 0,"\n    1-n : do various special things depending on value."},
    { OPT_RARG, "cell[:view[:facet[:version]]]",
		"name, view, facet, and version of the output facet."},
    { OPT_DESC, 0,"Octopus is a programmable logic array (PLA) module generator."},
    { 0,        0,               0}
};

/* global variables set by the command line				*/

extern s_options Options;

/*
 * parseComLine
 *
 * parse the command line 
 */
parseComLine( argc, argv )
int argc;
char **argv;
{

    extern char	*optarg;
    int		option;
    char	*mapToLower();
    char 	*util_strsav();

    while(( option = optGetOption( argc,argv)) != EOF ){
	switch ( option ) {
	    case 'T':
		Options.technology = util_strsav( optarg );
		break;
	    case 'p':
		Options.paletteName = util_strsav( optarg );
		break;
	    case 'P':
		Options.localPaletteName = util_strsav( optarg );
		break;
	    case 'C':
		if (!strcmp (optarg, "in")) {
		    Options.clocking |= CLOCKED_INPUT;
		} else if (!strcmp (optarg, "out")) {
		    Options.clocking |= CLOCKED_OUTPUT;
		/*
		} else if (!strcmp (optarg, "pre")) {
		    Options.clocking |= CLOCKED_PRE_OUT;
		*/
		} else {
		    optUsage();
		}
		break;
	    case 'd':
		Options.debug = atoi( optarg );
		break;
	    case 'g':
		Options.printGraph = TRUE;
		break;
	    case 'i':
		Options.inFile = util_strsav( optarg );
		break;
	    case 'v':
		Options.plaVerifyFile = util_strsav( optarg );
		break;
	    case 's':
		Options.andGndLineSep = atoi (optarg );
		break;
	    case 'S':
		Options.orGndLineSep = atoi (optarg );
		break;
	    case 'w':
		Options.wire = FALSE;
		break;
	    default:
		optUsage();
		break;
	}
    }
    if(( argc == 1 )||( optind == argc )){
	optUsage();
	exit(-1);
    }
    Options.plaFacet = util_strsav (argv[ optind ] );
}
