/*
 * Read the genie persionality matrix.
 * Calculate various global values and set up the array core
 * description structure and input/output arrays.
 */

#include "octopus.h"

extern int	GivenRows;	/* dimentions given in personality matrix */
extern int	GivenAndCols, GivenOrCols;
s_input		**GivenInputs;
s_output	**GivenOutputs;
extern char	**GivenCore;


/*
 * readMatrix
 *
 * Read a personality matrix of the form written by Genie.  This just
 * blindly copies in what it finds in the input file.  The matrix is
 * is constructed in the current Genie orientation (upside down) and
 * empty columns and useless CONNECTION cells.  Processing is done
 * later on on this marix (GivenXXX) to construct a better version
 * for actually creating the PLA.
 *
 * Note: The convention that genie uses for orientation of row and column
 * 	 exactly opposite to that used by octopus.  So row in the input file
 *	 is column in the data structure of octopus, and the complementary
 * 	 rule applies to columns.
 */
readMatrix(fileName)
char	*fileName;
{
    FILE	*fopen(), *ifd;
    int		allCols, i, j, c, isOrCol;
    char	matrixType[ NAMESIZE ];


    /* open input file */
    if ( fileName == NIL(char)) {
	ifd = stdin;
	fprintf(stderr,"Input file is stdin\n");
    } else {
	if (( ifd = fopen( fileName, "r" )) == NIL(FILE)) {
	    errRaise( optProgName, -1, "Could not open %s.", fileName);
	}
    } 
   
    /* read the first line which consists of a string and 2 numbers which
     * define the size of the array.  Genie considers these to by Row x Col,
     * but here in octopus we use them as Col x Row.
     */
    if ( fscanf(ifd, "%s %d %d", matrixType, &allCols, &GivenRows) != 3 ){
	errRaise( optProgName, -1, "%s %s",
		"The first line of personality matrix should",
		"have a name and two integers.");
    }

    /*** here we could check the matrixType				*/

    /* allocate memory for GivenCore matrix				*/ 
    allocGivenCore(GivenRows, allCols);
   
    /* read characters in personality matrix				*/
    isOrCol = FALSE;
    for ( i = 0; i < allCols; ++i) {
	/* get rid of white space */
	while ( ((c = getc(ifd)) == ' ') || (c == '\t') || (c == '\n') ) {
	     ;
	}
	if ( c == '(' ) {
	    errRaise( optProgName, -1, "%s %s",
			"Not enough lines (octopus columns,",
			" genie rows) in personality matrix.");
	}
	(void)ungetc(c, ifd);
	/* read a line of the character array				*/
	for ( j=0; j < GivenRows; ++j) {
	    if (((c = getc(ifd)) == DOT) || (c == SUM) || (c == PRODUCT) ||
						(c == CONNECTION)) {
		GivenCore[j][i] = c;
		if (c == PRODUCT) {
		    isOrCol = TRUE;
		}
	    } else if ( c == '\n') {
		errRaise( optProgName, -1,
			"personality matrix not wide enough.");
	    } else if ( c == EOF) {
		errRaise( optProgName, -1,
			"unexpected EOF while reading personality matrix.");
	    } else {
		errRaise( optProgName, -1,
			"Unrecognizable character '%c' in personality matrix.",
			(char) c);
	    }
	}	/* finish reading a line */

	if (isOrCol) {
	    ++GivenOrCols;			/* just read an or column */
	} else {
	    ++GivenAndCols;			/* just read an and column */
 	}
	/* read to the end of the line */
	while ((c = getc(ifd)) != '\n') {
	    switch (c) {
		case EOF: 
		    errRaise( optProgName, -1,
			"Unexpected EOF while reading personality matrix.");
		    break;
		case ' ':
		case '\t': break;
		default: 
		    errRaise( optProgName, -1,
		      "Unrecognizable char '%c' in personality matrix[%d,%d].",
				c, j, i);
	    }
	}
    } /* end reading core matrix					*/

    /* remove white space between matrix and row definitions		*/
    while ((( c = getc(ifd) ) == ' ')||(c == '\t')||(c == '\n')) ;

    if ( c != '(' ) {
	errRaise( optProgName, -1, "Too many lines in matrix core.");
    } else {
	(void)ungetc(c, ifd);
    }

    /* read the signal names and locations				*/
    readSignals(ifd);
}


/* readSignals
 *
 * Read the row and column signal definitions
 * ifd is assumed to be at the begining of the first (row ...) definition.
 */
readSignals( ifd )
FILE	*ifd; /* input file descriptor					*/
{
    char	keyword[ NAMESIZE ], sigName[ STRLEN ];
    int		index, start;
    int		c;
    s_input	*getInputByName(), *sig1, *sig2; /* for assigning complements*/

    /* allocate memory for GivenInputs and GivenOutputs arrays		*/
    GivenInputs  = ALLOC( s_input *, GivenAndCols);
    GivenOutputs = ALLOC( s_output *, GivenOrCols);

    (void) memset((char *) GivenInputs, 0, sizeof(s_input *) * GivenAndCols);
    (void) memset((char *) GivenOutputs, 0, sizeof(s_output *) * GivenOrCols);

    /* read row definitions						*/
    while ( fscanf(ifd, " ( %s", keyword) == 1) {
	if (strcmp( keyword, "row") == 0) {
	    /* get the row number					*/
	    fscanf(ifd, " %d", &index);
	    /* for each signal name in the row definition...		*/
	    fscanf(ifd, " %s", sigName);
	    while ( strcmp(")", sigName) != 0) {
		/* get the associated starting positions		*/
		if ( fscanf(ifd, " %d", &start) != 1) {
		    fprintf(stderr,"expected int after signal %s in row %d\n",
				sigName, index);
		}
		/* append to either inputs or outputs list at index	*/
		if(index < GivenAndCols) {
		    /* process this row as inputs			*/
		    appendInput(index, sigName, start);
		} else {
		    /* process this row as outputs			*/
		    appendOutput(index, sigName, start);
		}
		fscanf(ifd, " %s", sigName);
	    }
	} else if (strcmp (keyword, "column") == 0) {
	    /* read to end of line.
	     * The column definitions are not being looked at because
	     * octopus is not currently handling column folding or
	     * multi-plane PLA's.  If these features were to be added,
	     * then the column definitions would be necessary.
	     */
	    while( ((c = getc(ifd)) != '\n') && (c != EOF)) ;
	} else if( strcmp(keyword, "complement") == 0) {
	    if ((fscanf( ifd, "%s", sigName) == 1) && (strcmp(")", sigName))) {
		sig1 = getInputByName(sigName);
		if ((fscanf( ifd,"%s", sigName) == 1) && (strcmp(")", sigName))) {
		    /* have two signals					*/
		    sig1->phase = POS;
		    sig2 = getInputByName(sigName);
		    sig2->phase = NEG;
		    sig1->comp = sig2;
		    sig2->comp = sig1;
		} else {
		    /* only one signal					*/
		    sig1->phase = NEG;
		}
	    }
	    /* read to begining of next line, or possibly EOF		*/
	    while( ((c = getc(ifd)) != '\n') && (c != EOF)) ;
	} else {
	    /* unknown keyword						*/
	    errRaise( optProgName, -1,
			"Unknown keyword \"%s\" in personality matrix file.",
			keyword);
	}
    }
    
}

/* appendInput
 * Add a new input on to the list of inputs at GivenInputs[column].
 * Since it is assumed there will be only one (row ...) line in the 
 * personality matrix file, all new inputs on at column will simply be
 * appended on the the list of inputs at GivenInputs[column].
 */
appendInput( column, sigName, start)
int	column;
char	*sigName;
int	start;
{
    s_input	*getInputByName(), *input;
    s_input	*tempIn;

    if ((input = getInputByName(sigName)) == NIL(s_input)) {
	errRaise( optProgName, -1,
		"Could not get input siginal \"%s\".", sigName);
    }

    input->col = column;
    input->top = start;

    /* append it to list						*/
    tempIn = GivenInputs[column];
    if ( tempIn == NIL(s_input)) {
	/* list is empty, so make this the first element		*/
	input->bottom = GivenRows - 1;
	GivenInputs[column] = input;
    } else {
	/* list has stuff alread, add this as the last element		*/
	for( ; tempIn->next != NIL(s_input); tempIn = tempIn->next);
	input->bottom = tempIn->bottom;
	tempIn->bottom = start - 1;
	tempIn->next = input;
    }
}

/* appendOutput
 */
appendOutput( column, sigName, start)
int	column;
char	*sigName;
int	start;
{
    s_output	*newOutput(), *output, *tempOut;

    output = newOutput ( sigName );

    output->col = column;
    output->top = start;

    /* append it to list						*/
    tempOut = GivenOutputs[column - GivenAndCols];
    if ( tempOut == NIL(s_output)) {
	/* list is empty, so make this the first element		*/
	output->bottom = GivenRows - 1;
	GivenOutputs[column - GivenAndCols] = output;
    } else {
	/* list has stuff alread, add this as the last element		*/
	for( ; tempOut->next != NIL(s_output); tempOut = tempOut->next);
	output->bottom = tempOut->bottom;
	tempOut->bottom = start - 1;
	tempOut->next = output;
    }
}
