#!/bin/csh -f
if ($#argv < 1) goto usage

# Prepare a banner.
set B = "octpla: "

#names of the programs invoked in this script
if ($?OCTTOOLS) then
	set octtools = $OCTTOOLS
else
	set octtools = _CADRUNDIR_
endif

# octpla cannot be called by itself with VOV, vov_octpla should be called instead.
if ( $?VOV_HOST_NAME ) then
	echo $B "Tracing of octpla is not possible at this time: use vov_octpla instead."
	echo 
	unsetenv VOV_HOST_NAME
endif

set CADBIN = $octtools/bin

set MIS = "$CADBIN/misII"
set ESPRESSO = "$CADBIN/espresso"
set GENIE = "$CADBIN/genie"
set OCTOPUS = "$CADBIN/octopus"
set SPARCS = "$CADBIN/sparcs"
set VULCAN = "$CADBIN/vulcan"

#default options for tools in octpla

set MISARG =  "-t blif -T pla"
set ESPRESSOARG =  ""
set GENIEARG = "-pm"
set OCTOPUSARG = ""
set SPARCSARG = "-w PLACE -w NWEL -w PWEL"
set VULCANARG = ""

set CELL = $argv[$#argv]
if ( ${CELL} =~ -* ) then
	goto usage
endif

set VIEW = "pla"

@ argnum = 1
echo -n "$0 "
while ( $argnum < $#argv )
	if ( x$argv[$argnum] == "x-st" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set START = $argv[$argnum]
	else if ( x$argv[$argnum] == "x-o" ) then
		if ( $argnum == $#argv - 1 ) goto usage
		echo -n "$argv[$argnum] "
		@ argnum = $argnum + 1
		set VIEW = $argv[$argnum]
	else if ( x$argv[$argnum] == "x-clock" ) then
		set OCTOPUSARG = "-C in -C out $OCTOPUSARG"
	else if ( x$argv[$argnum] == "x-nofold" ) then
		set GENIEARG = "-n $GENIEARG"
	else if ( x$argv[$argnum] == "x-cin" ) then
		set OCTOPUSARG = "-C in $OCTOPUSARG"
	else if ( x$argv[$argnum] == "x-cout" ) then
		set OCTOPUSARG = "-C out $OCTOPUSARG"
	else if ( x$argv[$argnum] == "x-help" ) then
		goto usage
	else goto usage
	endif
	echo -n "$argv[$argnum] "
	@ argnum = $argnum + 1
end
echo $CELL


echo $B This is the octopus pla generator 

if( $?START ) then
	echo $B restarting at $START
	goto $START
endif

if(-e $CELL.blif == 0 ) then
	echo $B Could not find $CELL.blif
	exit
endif


## Since some tools always use the same output (final , genie.out ... )
## Create a special temporary directory to run this particular job.
set WD = octpla.tmp.$$

echo $B "Creating working directory $WD"
mkdir $WD

# Move needed file in tmp directory
cp $CELL.blif $WD


cd $WD

mis:
echo $B Running $MIS
set LASTTOOL = $MIS
$MIS $MISARG -o ${CELL}.pla ${CELL}.blif || goto abort

espresso:
echo $B Running $ESPRESSO
set LASTTOOL = $ESPRESSO
( $ESPRESSO $ESPRESSOARG < $CELL.pla > $CELL.opt ) || goto abort

genie:
echo $B Running $GENIE
set LASTTOOL = $GENIE
( $GENIE -pla  ${CELL}.opt | ( $GENIE $GENIEARG >& genie.log )) || goto abort

octopus:
echo $B Running $OCTOPUS
set LASTTOOL = $OCTOPUS
$OCTOPUS $OCTOPUSARG -i final ${CELL}:${VIEW} || goto abort

sparcs:
echo $B Running $SPARCS
set LASTTOOL = $SPARCS
$SPARCS $SPARCSARG -o ${CELL}:${VIEW} ${CELL}:${VIEW} || goto abort

vulcan:
echo $B Running $VULCAN
set LASTTOOL = $VULCAN
$VULCAN $VULCANARG ${CELL}:${VIEW} || goto abort

echo $B Cleaning up $WD
cp -r $CELL ..
cd ..
rm -rf $WD

echo $B Done: the output is in ${CELL}:${VIEW}
exit 0

abort:
	echo $B "////////////////////////////////////////////////"
	echo $B $LASTTOOL failed.
	echo $B Temporary files left in $WD
	echo $B "////////////////////////////////////////////////"
	exit 10
usage:
	echo $B usage: $0 \[-clock\] \[-nofold\] \[-cin\] \[-cout\] \[-o view\] \[-st \<toolname\>\] cell
	echo $B ""
	echo $B expects cell.blif as input and will create cell:pla or
	echo $B cell:view if the -o option is specified
	echo $B ""
	echo $B -nofold will produce an unfolded pla \( hopefully \)
	echo $B -cin uses latched inputs, clock terms are inPhi and inPhiBar
	echo $B -cout uses latched output, clock term is outPhi
	echo $B -clock is equivalent to -cin -cout
	echo $B -st \<toolname\> starts execution with specified tool
	echo $B ""
	echo $B the ground terminal is GND
	echo $B the power terminal is Vdd
	echo $B ""
	echo $B inPhiBar and outPhi should be connected to the same clock phase
	echo $B inPhi should be the complementary clock phase
	echo $B the pla inputs are held when inPhi is low 
	echo $B the pla outputs are held when outPhi is low 
	exit 10
