#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "math.h"

#define LARGEINTEGER ( 1<<( sizeof(int)*8-2))
             
/*
 *  PADPgetArea()
 *  Determine the size of available Area
 */
void PADPgetArea( chippads, useFlpArea, padLimited, fixed_pads )
    ChipPads* chippads;
    int useFlpArea;
    int padLimited;
    int fixed_pads;
{
    int	 	cellsInCore = 0;
    int 	totalArea = 0; 	/*  */
    int 	instanceArea = 0.0; 	/* excluding PADS */
    struct octBox bb;		/* BB of each instance. */
    struct octBox core;		/* BB of the chip core. */
    struct octBox chip;		/* BB of the chip.      */
    octGenerator  genInst; 	/*  */
    octObject	  Instance;		/*  */
    int  padHeight = chippads->maxPadHeight;
    int  corePadsSpace = LAMBDA * 50; /* space between core and pads */
    double	  areaExpansion = 1.5;

    if (verbose) Message("Computing placement area");
    if (useFlpArea==TRUE) {
	(void)getAvailableAreaFromOct( &chip );
    } else {
	/* find area of the core (all the cells that are not pads) */
	core.lowerLeft.x  = LARGEINTEGER;
	core.lowerLeft.y  = LARGEINTEGER;
	core.upperRight.x = -LARGEINTEGER; 
	core.upperRight.y = -LARGEINTEGER;
	chip.lowerLeft.x  = LARGEINTEGER;
	chip.lowerLeft.y  = LARGEINTEGER;
	chip.upperRight.x = -LARGEINTEGER; 
	chip.upperRight.y = -LARGEINTEGER;

	OH_ASSERT(octInitGenContents(&Chip, OCT_INSTANCE_MASK , &genInst));
	while (octGenerate(&genInst, &Instance) == OCT_OK) {
	    OH_ASSERT(octBB(&Instance,&bb) );
	    totalArea +=  box_area( &bb, LAMBDA ); /* In LAMBDA square */

	    if ( !isPad(&Instance)) {
		cellsInCore++;
		core.lowerLeft.x  = MIN(core.lowerLeft.x,bb.lowerLeft.x);
		core.lowerLeft.y  = MIN(core.lowerLeft.y,bb.lowerLeft.y);
		core.upperRight.x = MAX(core.upperRight.x,bb.upperRight.x);
		core.upperRight.y = MAX(core.upperRight.y,bb.upperRight.y);
		instanceArea += box_area( &bb, LAMBDA );
	    } 
	    chip.lowerLeft.x  = MIN(chip.lowerLeft.x,bb.lowerLeft.x);
	    chip.lowerLeft.y  = MIN(chip.lowerLeft.y,bb.lowerLeft.y);
	    chip.upperRight.x = MAX(chip.upperRight.x,bb.upperRight.x);
	    chip.upperRight.y = MAX(chip.upperRight.y,bb.upperRight.y);
	}

	if (cellsInCore == 0) {
	    Warning("No cells to define the core");
	}

	/* If pads are not fixed, adjust chip area. */
	if ( ! fixed_pads && ! padLimited) {
	    if ( box_area( &core, LAMBDA ) > instanceArea * 0.9 ) {
		
		chip.lowerLeft.x = core.lowerLeft.x - padHeight - corePadsSpace;
		chip.lowerLeft.y = core.lowerLeft.y - padHeight - corePadsSpace;
		chip.upperRight.x = core.upperRight.x + padHeight +corePadsSpace;
		chip.upperRight.y = core.upperRight.y + padHeight +corePadsSpace;
	    } else {
		if (verbose) (void) printf(" The cells are badly placed\n");

		/* probably the cells have not be placed at all */
		chip.lowerLeft.x = 0;
		chip.lowerLeft.y = 0;
		chip.upperRight.x = sqrt((double)totalArea)*LAMBDA*areaExpansion;
		chip.upperRight.y = sqrt((double)totalArea)*LAMBDA*areaExpansion;
	    }
	}
    }


    if (verbose) {
	(void) printf("Core area %d,\t\tBBox %6d,%6d - %6d,%6d\n", box_area( &core, LAMBDA ),
		      core.lowerLeft.x, core.lowerLeft.y,
		      core.upperRight.x, core.upperRight.y);
	(void) printf("Chip area %d,\t\tBBox %6d,%6d - %6d,%6d\n", box_area( &chip, LAMBDA ),
		      chip.lowerLeft.x, chip.lowerLeft.y,
		      chip.upperRight.x, chip.upperRight.y);
	(void) printf("max Height of pads = %d\n", chippads->maxPadHeight);
	(void) printf("space between core and pads %d\n" , corePadsSpace);
	fflush( stdout );
    }

    chippads->coreArea = core;
    chippads->chipArea = chip;
    chippads->width  =  box_width( &chip );
    chippads->height =  box_height( &chip );
}

int isPad(inst)
    octObject *inst;
{
    octObject Master;

    if (octGetByName(inst,&PlacementClass)==OCT_OK && propEqual(&PlacementClass,"PAD")) {
	return 1;
    }
    OH_ASSERT( ohOpenMaster( &Master, inst, "interface", "r" ) );
    if (octGetByName(&Master,&PlacementClass)==OCT_OK &&propEqual(&PlacementClass,"PAD")) {
	OH_ASSERT( octCreate(inst,&PlacementClass) );
	return 1;
    }
    return 0;
}


getAvailableAreaFromOct(boxP)
    struct octBox	*boxP;
{
  octObject    FlpAreaBox;	/*  */


  if (verbose ) {
    (void) printf("Getting available area from OCT\n");
    (void) printf("I'm looking for the bag %s in %s:%s:%s\n",
	   FloorplanBag.contents.bag.name,
	   ChipInterface.contents.facet.cell,
	   ChipInterface.contents.facet.view,
	   ChipInterface.contents.facet.facet);
    (void) printf("Inside the bag I expect a box which may (or may not) be attached to layer PLACE\n");
  }

  if (interfaceOpen == FALSE) PADPFatalError("Could not open the interface facet");
  if (octGetByName(&ChipInterface,&FloorplanBag)!=OCT_OK) PADPFatalError("No FLOORPLAN bag");
  if (octGenFirstContent(&FloorplanBag,OCT_BOX_MASK,&FlpAreaBox)!=OCT_OK) 
    PADPFatalError("No box in FLOORPLAN bag");
  (*boxP = FlpAreaBox.contents.box);
  return OCT_OK;
}
