/*
 *
 *  GLOBAL DEFINITIONS
 *
 *  Advising professor:
 *      Alberto Sangiovanni-Vincentelli
 *
 *	Andrea Casotto
 */


#include "copyright.h"
#include "port.h"
#include "errno.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"


/*
 *   MACRO DEFINITIONS
 *
 *   Macros are distinguished by using solely capital letters in their
 *   identifiers.  This contrasts with C defined identifiers which are strickly 
 *   lower case, and program variable and procedure names which use both upper 
 *   and lower case.
 */

/* Begin macros. */

/* Boolean data type */
#define  BOOLEAN	int
#define  FALSE		0
#define  TRUE		1
#define  NOT		!
#define  AND		&&
#define  OR		||
#define nobreak 


#define YES  	1
#define NO	0
#ifndef NULL
#define NULL	0
#endif
#define OK   1


/* Macro commands */

/* Macro procedure that swaps two entities. */
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}


/* Memory allocation and dealocation macros. */
extern char *CheckPtr();

#define DEF_ROW_SPACING 4000
#define DEF_BLOCK_SPACING 4000
#define DEF_BLOCK_SIZE 6
#define DEF_MAX_ROWS 3
#define CUR_DEFAULT 0.01
#define DIR_DEFAULT "special_contacts"
#define MIN_CONNECTOR_SIZE 4 /* lambda */

typedef enum {
    SCHEME_UNDEFINED,
    SCHEME_DONT_MOVE,
    SCHEME_MULTI_ROW,
    SCHEME_CLOCKWISE,
    SCHEME_FLP_CONSTRAINTS,
    SCHEME_MIN_NET_LENGTH 
  } schemeType;



#define NO_SIDE -1
#define LEFT     0
#define TOP      1
#define RIGHT    2
#define BOTTOM   3

/*
 *  A linked list is defined for all the pads
 */
typedef struct padListEl {
  char			*name;
  octId			FTermId;
  octId	   		instanceId;
  octId			implementationId;
  int			connectorImplementation;
  char			*bondingTermName;
  char			*internalTermName;
  char			*MasterCell;
  char			*MasterView;
  char			*termType;
  char			*direction;
  int                   masterOnly;
  double		powerRequired;
  double		peakCurrent; 		/* only used in SUPPLY or GROUND terms */
  octId			netId;
  char			*netName;
  int			row;
  int              	side;			/* side of the pad */
  int 		   	positionInSide;		/*  */
  struct octPoint  	targetLowerLeft;
  struct octBox	   	masterBBox;
  int		   	width,height;
  double             	relPos;
  octTransformType 	transform;
  int	           	internalEdgeOfMaster;	/* which edge should be facing in */
  char			*floatingEdges;
  char			*termEdge;
  int	       		flp_side;		/* side required by the floorplanner */
  double		flp_relPos;		/*  */
  struct padListEl	*nextInEdge;
  struct padListEl	*precInEdge;
  struct padListEl 	*next;
  struct padListEl 	*prec;
} Pad;


struct padFamily {
  char* 	name;
  octObject 	facet;
  char          *technology;
  int		connectorType;
  int		numberOfPads;
  int		minimumSeparation;
  Pad		*padArray;
};


typedef struct _ChipPads {
    Pad		*FirstPad;
    Pad		*LastPad;
    Pad		*FirstPadInEdge[DEF_MAX_ROWS][4];
    Pad		*LastPadInEdge[DEF_MAX_ROWS][4];
    int         numberOfPads;
    int 	availableSpaceSide[4];
    int		padsInEdge[DEF_MAX_ROWS][4];
    int		totalWidth[DEF_MAX_ROWS][4];
    int         width;
    int         height;
    struct octBox coreArea;
    struct octBox chipArea;
    int maxPadHeight,maxPadWidth;
} ChipPads;

typedef  struct padFamily PadFamily;


extern int verbose;
extern char buf[];

extern int createPads;

extern octObject 	Chip;
extern octObject 	ChipInterface;
extern int		interfaceOpen;

#define LAMBDA 20




