/*  Last modified: April 26, 1993      */
/*  Enrico Malavasi, Edoardo Charbon (the famous EM and EC) */

#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"
#include "padp.h"
#include "utility.h"
#include "tap.h"
#include "th.h"
#include "oct.h"
#include "oh.h"
#include "prim.h"
#include <sys/types.h>
#include <sys/dir.h>


int noBondingRings = FALSE;
static int cDens_init = FALSE;

void PADPreadPadFamily( chippads, family )
    ChipPads* chippads;
    PadFamily* family;
{
    char 		*cell;
    char 		*view;
    octObject 	*Family;
    octObject	Instance;
    octObject	Term;
    octObject   bag;
    octObject	Net;
    octGenerator	genInst,genTerm, genNet;
    Pad		*padp;
    char		fullPathname[256];
  

    cell = family->name;
    view = strchr(cell, ':');
    if (view != NIL(char)) {
	*view++ = '\0';
    } else {
	view = "symbolic";
    }

    if (octGetByName(&Chip,&Technology) != OCT_OK ) {
	PADPFatalError("The TECHNOLOGY property is missing");
    }
    
    (void) sprintf(fullPathname,"%s/tech/%s/pads/%s", 
		   "$OCTTOOLS", Technology.contents.prop.value.string, cell);

    Family = &family->facet;
    family->technology = util_strsav(Technology.contents.prop.value.string);

    if ( ohOpenFacet( Family, cell, view, "contents", "r" ) < OCT_OK ) {
	if ( ohOpenFacet( Family, fullPathname, view, "contents", "r" ) < OCT_OK ) {
	    PADPFatalError( "Opening pad family facet" );
	}
    }
    VOVinputFacet( Family );
  
    if (verbose) {
	(void) printf("Using pad family %s\n", ohFormatName( Family ) );
    }

    OH_ASSERT(octInitGenContents(Family,OCT_INSTANCE_MASK,&genInst));
    family->numberOfPads = 0;
    while (octGenerate(&genInst,&Instance)==OCT_OK) family->numberOfPads++;
  
    if (verbose) (void) printf("%d pads in this family\n", family->numberOfPads);

    family->padArray = ALLOC(Pad, family->numberOfPads);

    family->connectorType = FALSE;
    if (ohGetByBagName(Family,&bag,"CONNECTORS")==OCT_OK && 
	ohCountContents(&bag,OCT_INSTANCE_MASK ) > 0 ) {
	family->connectorType = TRUE;
	if (verbose) Message("Family is of connectors" );
    }
  
    if (octGetByName(Family,&PadMinSeparation) != OCT_OK) {
	PADPFatalError("Can not get the minimum pad separation from the family description");
    }
    family->minimumSeparation = PadMinSeparation.contents.prop.value.real;
  
    if (octGetByName(Family,&PadBondingPinsNet) != OCT_OK) {
	PADPFatalError("Can not find the the net containings the bonding pad terminals");
    }
    if (octGetByName(Family,&PadInternalPinsNet) != OCT_OK) {
	PADPFatalError("Can not find the net containings the internal terminals");
    }

    if (family->connectorType == FALSE) {
	if (octGetByName(Family,&PadSupplyRing) != OCT_OK) {
	    PADPFatalError("Can not get the power suppy ring from the family description");
	}
	if (octGetByName(Family,&PadGroundRing) != OCT_OK) {
	    PADPFatalError("Can not get the power suppy ring from the family description");
	}
    }

    OH_ASSERT(octInitGenContents(Family,OCT_INSTANCE_MASK,&genInst));
    family->numberOfPads = 0;
    while (octGenerate(&genInst,&Instance)==OCT_OK) {
	padp = &family->padArray[family->numberOfPads++];
	padp->instanceId = Instance.objectId;

	OH_ASSERT(octGetByName(&Instance,&Direction));
	padp->direction = util_strsav(Direction.contents.prop.value.string) ;
	OH_ASSERT(octGetByName(&Instance,&TermType));
	padp->termType = util_strsav(TermType.contents.prop.value.string) ;
	OH_ASSERT(octGetByName(&Instance,&PadPowerRequired));
	padp->powerRequired = PadPowerRequired.contents.prop.value.real;
	padp->masterOnly = FALSE;

	readPadMaster( chippads, padp );
    

	OH_ASSERT(octInitGenContents(&Instance,OCT_TERM_MASK,&genTerm));
	while (octGenerate(&genTerm,&Term) == OCT_OK) {
	    OH_ASSERT(octInitGenContainers(&Term,OCT_NET_MASK,&genNet));
	    while (octGenerate(&genNet,&Net)==OCT_OK) {
		if (octIdsEqual(Net.objectId, PadBondingPinsNet.objectId)) {
		    padp->bondingTermName = util_strsav(Term.contents.term.name) ;
		} else if (octIdsEqual(Net.objectId, PadInternalPinsNet.objectId)) {
		    padp->internalTermName =  util_strsav(Term.contents.term.name) ;
		}
	    }
	}
    }
}


/* given the peak-current info, find a quantize widht */
int
getImplWidth(chip, termLayer, peakCurrent) 
    octObject *chip;
    char *termLayer;
    double peakCurrent;
{
    int width;			/* in meters */
    octObject curLayer;
 
    if (!cDens_init) {
	tap_cDensityInit( chip );
	cDens_init = TRUE;
    } 

    curLayer.contents.layer.name = util_strsav(termLayer);
    width = tap_getWidthFromPeakCurrent( peakCurrent, &curLayer );

    return width;		/* in lambda */

}


/* quick and dirty check for cell existence */

int
cellExist(cell, view, facet)
    char *cell, *view, *facet;
{
    DIR *director;
    struct direct *dp;
    char buffer[2000];
    int len;

    if (!cell || !view || !facet) return OCT_ERROR;

    if (!(director = opendir(cell))) return OCT_ERROR;
    closedir(director);
    sprintf (buffer, "%s/%s", cell, view);
    if (!(director = opendir(buffer))) return OCT_ERROR; 
    len = strlen(facet);
    for (dp = readdir(director); dp != NULL; dp = readdir(director)) {
	if (dp->d_namlen == len+1 && !strncmp(dp->d_name, facet, len)) {
	    closedir(director);
	    return OCT_OK;
	}
    }

    closedir(director);
    return OCT_ERROR;

}


Pad *createPad( family, padP, width)
    PadFamily* family;
    Pad *padP;
    int width;			/* in lambda */
{
    int	i;
    Pad	*pad;
    char *name, *connName;
    octObject Facet, termInst, prop;
    char *conLayer, *firstLayer, *secondLayer;
    DIR *specContDir;

    if ((specContDir = opendir(DIR_DEFAULT)) == NULL) {
	/* directory not existing. Creating one */
	sprintf(buf, "mkdir %s", DIR_DEFAULT);
	if (system(buf)) {
	    sprintf (buf, "Opening directory \"%s\"", DIR_DEFAULT);
	    PADPFatalError(buf);
	}
    }

    if (width < MIN_CONNECTOR_SIZE) width = MIN_CONNECTOR_SIZE;
    sprintf(buf, "%s/MET1-MET2.%dx%d", DIR_DEFAULT, width, width );
    name = util_strsav(buf);
    
    if (cellExist(name, "physical", "contents") == OCT_ERROR) {
	sprintf(buf,"Creating a %d lambda wide pad! for %s", 
		width, padP->name);
	Warning(buf);
	ohUnpackDefaults(&Facet, "w", ":physical:contents");
	ohUnpackFacetName(&Facet, name);
	symOpenFacet(&Facet, "w", family->technology);
	ohCreateOrModifyPropStr(&Facet, &prop, "CELLTYPE", "MOS");
	ohCreateOrModifyPropStr(&Facet, &prop, "CELLCLASS", "LEAF");

	width *= LAMBDA;
	conLayer = util_strsav("COM2");
	firstLayer = util_strsav("MET1");
	secondLayer = util_strsav("MET2");
	thTechnology( &Facet );	/* hash all info about the technology */
	primThreeLayerContact(&Facet, conLayer, firstLayer, secondLayer, 
			      width, width);
	octCloseFacet(&Facet);    
    } else {
	sprintf(buf,"Using a %d lambda wide pad! for %s", width, padP->name);
	Warning(buf);
    }
    pad = ALLOC(Pad, 1);
    pad->termType = util_strsav(padP->termType);
    pad->direction = util_strsav(padP->direction);
    pad->powerRequired = FALSE;
    pad->bondingTermName = util_strsav("NONE");
    pad->internalTermName = util_strsav("t");
    pad->instanceId = 1;
    pad->masterOnly = TRUE;
    pad->MasterCell = name;
    pad->MasterView = util_strsav("physical");
    noBondingRings = TRUE;

    (family->numberOfPads)++;
    family->padArray = REALLOC (Pad, family->padArray, family->numberOfPads+1);
    family->padArray[family->numberOfPads-1] = *pad;
    
    return(pad);
}


Pad *PF_getPad( family, direction, termType, peakCurrent)
    PadFamily* family;
    char *direction;
    char *termType;
    double peakCurrent;
{
    int	i;
    Pad	*pad;

    if (strcmp(termType,"SUPPLY")==0 || strcmp(termType,"GROUND")==0) {
	for (i = 0 ; i < family->numberOfPads ; i++) {
	    pad = &family->padArray[i];
	    if (strcmp(pad->termType,termType) == 0) {
		if (!(peakCurrent) || pad->peakCurrent == CUR_DEFAULT ||
		      pad->peakCurrent == peakCurrent)
		  return pad;
	    }
	}
    } else {
	/* linear search */
	for (i = 0 ; i < family->numberOfPads ; i++) {
	    pad = &family->padArray[i];
	    if ((strcmp(pad->direction,direction) == 0) &&
		(strcmp(pad->termType,termType) == 0)) {
		return(pad);
	    }
	}
    }
    return(NULL);
}


static int j = 0;

/* 
 * If the pad lack an implementation , give one to it using pads form the pad family 
 */ 
void PADPimplementFormals( chippads, family )
    ChipPads* chippads;
    PadFamily* family;
{
  Pad		*padP;
  Pad		*implPad;
  octObject	FTerm, BondingTerm, InternalTerm, Net;
  octObject	NewInstance;
  octObject     netProp;
  octObject     Master;
  char		uniqueName[256];
  int           width; 
  char          *termLayer;

  termLayer = util_strsav("MET1");
  if (verbose) (void) printf("Implementing all the formal terminals\n");

  OH_ASSERT(octGetOrCreate(&Chip,&InstanceBag));
  OH_ASSERT(octGetOrCreate(&Chip,&ConnectorBag));
  
  for (padP = chippads->FirstPad ; padP != NULL ; padP = padP->next) {
      if (padP->implementationId != 0)  continue;
      if (!createPads) {
	  implPad = PF_getPad( family, padP->direction, padP->termType, 
			      padP->peakCurrent);
      } else {
	  width = getImplWidth(&Chip, termLayer, padP->peakCurrent); 
	  implPad = createPad(family, padP, width);
      }	
      if (implPad == NULL) {
	  (void)sprintf(buf,"No pad for %s (%s,%s) in family",
			padP->name, padP->termType, padP->direction);
	  Warning(buf);
	  continue;
      }

      if (verbose) {
	  (void)printf("%s (%s,%s)\n", padP->name, padP->termType, padP->direction);
      }
      
      Net.objectId = padP->netId;
      OH_ASSERT(ohGetById(&FTerm, padP->FTermId));
      /* special connector */
      if (implPad->masterOnly) {	
	  (void) sprintf(uniqueName, "auto_gen_%s", padP->name);
	  ohCreateInst(&Chip, &NewInstance, util_strsav(uniqueName), 
		       implPad->MasterCell, implPad->MasterView);
	  
      /* regular connector */
      } else {
	  OH_ASSERT(ohGetById(&NewInstance, implPad->instanceId));
	  (void) sprintf(uniqueName,"%s_%s", NewInstance.contents.instance.name, 
			 padP->name);
	  NewInstance.contents.instance.name = util_strsav(uniqueName);
	  ohBestName( &Chip, &NewInstance, 0 );
	  OH_ASSERT(octCreate(&Chip,&NewInstance));
      }
      
      PlacementClass.contents.prop.value.string = "PAD";
      OH_ASSERT(octCreate(&NewInstance,&PlacementClass));
       
      if (implPad->bondingTermName && strcmp(implPad->bondingTermName, "NONE"))
	OH_ASSERT(ohGetByTermName(&NewInstance,&BondingTerm, 
				  implPad->bondingTermName));
      OH_ASSERT(ohGetByTermName(&NewInstance,&InternalTerm, 
				implPad->internalTermName));
      
      if (family->connectorType) {
	  padP->connectorImplementation = TRUE;
	  OH_ASSERT(octAttach(&ConnectorBag,&NewInstance));
	  OH_ASSERT(octAttach(&FTerm,&InternalTerm));
      } else {
	  padP->connectorImplementation = FALSE;
	  OH_ASSERT(octAttach(&InstanceBag,&NewInstance));
	  OH_ASSERT(octAttach(&FTerm,&BondingTerm));
	  if (strcmp(padP->direction,"INOUT")==0) {
	      (void)sprintf(buf,"Term %s is bidirectional. I do not know its enable signal",
			     padP->name);
	      Warning(buf);
	      Warning("This is probably an error.Fix it with attache or with bdnet.");
	  }
      }
       
      if (!octIdIsNull(Net.objectId)) {
	  OH_ASSERT(octGetById(&Net));
	  if (padP->connectorImplementation) {
	      OH_ASSERT(octAttachOnce(&Net,&FTerm));
	  } else {
	      OH_ASSERT(octAttach(&Net,&InternalTerm));
	      if (octIsAttached(&Net,&FTerm)==OCT_OK) {
		  OH_ASSERT(octDetach(&Net,&FTerm));
	      }
	      /* Change the name of the net so that it does not conflict with the
	       * name of the formal terminal. This is needed for bdsim
	       */
	      (void) sprintf(uniqueName,"%s-internal", Net.contents.net.name);
	      Net.contents.net.name = uniqueName;
	      OH_ASSERT(octModify(&Net));
	      /*  Add the terminal property to the net */
	      OH_ASSERT( ohCreateSprop( &Net, &netProp, "NETTYPE", padP->termType) );
	  }
      }

      padP->instanceId = NewInstance.objectId;
      padP->implementationId = BondingTerm.objectId;
      padP->connectorImplementation = implPad->connectorImplementation;
      padP->bondingTermName = implPad->bondingTermName;
      padP->internalTermName = implPad->internalTermName;
      padP->MasterCell = implPad->MasterCell;
      padP->MasterView = implPad->MasterView;
      padP->powerRequired = implPad->powerRequired;
       
      padP->masterBBox = implPad->masterBBox;
      padP->width = implPad->width;
      padP->height = implPad->height;
      padP->internalEdgeOfMaster = implPad->internalEdgeOfMaster;
       
  }
}  
       
       
       
       
       
