#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"

#ifndef lint

static st_table *st_fterm;	/* Hash table to get formal terminals from Instance ids */

/*
 *  FUNCTION: ListAllPads(FacetPtr)
 *  There are two ways to find pads: 
 * 1
 *  Generates all the formal terminals from FacetPtr and assumes that
 *  the instance attached to the terminal, or the instance containing the
 *  actual terminal attached to the fterm, is a pad
 * 2
 *  Generate all instances and check if they have the property INSTANCE CLASS = PAD
 */
int
ListAllPads( chippads, useFlpConstraints, implementFlag, promoteFTermsToPads )
    ChipPads* chippads;
    int useFlpConstraints;
    int implementFlag;
    int promoteFTermsToPads;
{
  Pad			*padP=NULL;
  Pad			*lastPad=NULL; /* auxiliary pointer */
  octGenerator 	gTerm;
  octGenerator 	gInst;
  octStatus 		sTerm;
  octStatus 		sInst;
  extern octObject    ConnectorBag;
  extern octObject    Technology;
  octObject		oMaster;
  octObject		oTerm;
  octObject		oInst;
  octObject		oATerm;
  char 		cellBuffer[256];
  char 		*stringSide();
  extern int octIdCmp(), octIdHash();


  st_fterm = st_init_table(octIdCmp, octIdHash);

  OH_ASSERT(octGetOrCreate( &Chip, &ConnectorBag ));
  OH_ASSERT(octGetByName( &Chip, &Technology ));

  if( implementFlag ) {
    (void) sprintf( cellBuffer,"%s/tech/%s/primitives/com2.4x4",
	    "$OCTTOOLS", Technology.contents.prop.value.string);
  }

  if (promoteFTermsToPads) {
    /* generate all formal terminals, find the pads and flag them */
    OH_ASSERT( octInitGenContents(&Chip, OCT_TERM_MASK, &gTerm));
    while(( sTerm = octGenerate( &gTerm, &oTerm )) == OCT_OK){

      if (!octIdIsNull(oTerm.contents.term.instanceId)) continue;

      /* find the implementation of the pad			*/
      sInst = octGenFirstContent( &oTerm, OCT_INSTANCE_MASK|OCT_TERM_MASK, &oInst );
      OH_ASSERT( sInst );
      if( sInst == OCT_GEN_DONE ){
	if( implementFlag ){

	  /* create an implementation of the formal term	*/
	  oInst.type = OCT_INSTANCE;
	  oInst.contents.instance.name = oTerm.contents.term.name;
	  oInst.contents.instance.master = cellBuffer;
	  oInst.contents.instance.view = "physical";
	  oInst.contents.instance.facet = "contents";
	  oInst.contents.instance.version = OCT_CURRENT_VERSION;
	  oInst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
	  oInst.contents.instance.transform.translation.x = 0;
	  oInst.contents.instance.transform.translation.y = 0;
	    
	  OH_ASSERT( octCreate( &Chip, &oInst ));
	  OH_ASSERT( octAttach( &ConnectorBag, &oInst ));

	  OH_ASSERT( octGenFirstContent( &oInst, OCT_TERM_MASK, &oATerm));
	  OH_ASSERT( octAttach(&oTerm,&oATerm));

	} else {

	  (void) sprintf( buf, "Formal terminal %s has no implementation", oTerm.contents.term.name);
	  Warning( buf );
	  oInst.objectId = oct_null_id;

	}
      } else if ( oInst.type == OCT_TERM ) {
	oInst.objectId = oInst.contents.term.instanceId;
	OH_ASSERT( octGetById( &oInst ));
      }
      if (!octIdIsNull(oInst.objectId)) {
	octId *id1, *id2;
	*id1 = oInst.objectId;
	*id2 = oTerm.objectId;
	st_insert(st_fterm, (char *) id1, (char *) id2);
	checkPadProperties( &oInst );
      } 
    }
    OH_ASSERT( sTerm );
  }

  /* generate all instances, check prop PLACEMENT_CLASS on Master */
  /* this might be kind of slow */
  OH_ASSERT( octInitGenContents(&Chip, OCT_INSTANCE_MASK, &gInst));
  while(( sInst = octGenerate(&gInst, &oInst)) == OCT_OK ) {

    /* If we already know this is a pad, just go on */
    if(( octGetByName( &oInst, &PlacementClass ) == OCT_OK )&& 
       ( strcmp(PlacementClass.contents.prop.value.string,"PAD") ==0 ))
      continue;

    /* else check whether the master says it is a pad */
    oMaster.type = OCT_FACET;
    oMaster.contents.facet.facet="interface";
    oMaster.contents.facet.mode="r";
    if( octOpenMaster( &oInst, &oMaster) < OCT_OK ) 
      PADPFatalError("opening master");
    if(( octGetByName( &oMaster, &PlacementClass) == OCT_OK )&&
       ( strcmp( PlacementClass.contents.prop.value.string, "PAD" ) == 0 )){
       checkPadProperties( &oInst );
     }
  }
  OH_ASSERT( sInst );

  /* generate all instances, if they are pads add them to the list */
  OH_ASSERT( octInitGenContents( &Chip, OCT_INSTANCE_MASK, &gInst ));
  while(( sInst = octGenerate(&gInst, &oInst)) == OCT_OK ) {

    if( octGetByName( &oInst, &PlacementClass ) == OCT_OK &&
       strcmp(PlacementClass.contents.prop.value.string,"PAD" )==0)  {
	chippads->numberOfPads++;
       padP = ALLOC( Pad, 1 );
       /* Link the pad to the list of all pads */
	if ( chippads->numberOfPads == 1) {
	    chippads->FirstPad = lastPad = padP;
	    padP->prec = NULL;
	} else {
	    padP->prec = lastPad;
	    lastPad->next =  padP;
	    lastPad = padP;
	}
	padP->next= NULL;

       /* Fill in the structure */
       padP->instanceId = oInst.objectId;
       padP->transform = OCT_NO_TRANSFORM; /* default */
       if (octGetByName(&oInst,&PadEdge) == OCT_OK) {
	 padP->side = parseEdgeName(PadEdge.contents.prop.value.string);
       } else {
	 padP->side = NO_SIDE;	/* default */
       }
       padP->positionInSide = 0;
       padP->nextInEdge = NULL;
       padP->precInEdge = NULL;

       if (useFlpConstraints ) getFlpConstraints(padP);
       (void) readPadMaster(padP);
       if (verbose) (void) printf("pad N.%d: name %10s, BBox %d,%d - %d,%d (%6s,%4.2f)\n",
				  chippads->numberOfPads,
			   oInst.contents.instance.name,
			   padP->masterBBox.lowerLeft.x,
			   padP->masterBBox.lowerLeft.y,
			   padP->masterBBox.upperRight.x,
			   padP->masterBBox.upperRight.y,
			   stringSide(padP->flp_side) , padP->flp_relPos);
     
     }
  }
  st_free_table(st_fterm);
  return(OK);
}

getFlpConstraints(padP)
    struct padListEl *padP;
{
    octObject Term;
    char *ptr;

    /* set up the deafult values */
    padP->flp_side = NO_SIDE;
    padP->flp_relPos = 0.0;
    if (interfaceOpen == FALSE) PADPFatalError("Interface must be open to get flp constraints");

    /* Find the formal terminal attached to a terminal of this instance */
    if (st_lookup(st_fterm, (char *) &padP->instanceId, &ptr) == 0 ){
	Warning("Entry not found");
	return;
    }
    Term.objectId = *(octId *) ptr;
    OH_ASSERT( octGetById( &Term ));
    OH_ASSERT(octGetByName(&ChipInterface,&Term));
    if (octGetByName(&Term,&TermEdge)==OCT_OK) {
	padP->flp_side = parseEdgeName(TermEdge.contents.prop.value.string);
	if (padP->flp_side == NO_SIDE) {
	    (void) sprintf(buf,"Illegal edge specification `%s' for term %s",
			   TermEdge.contents.prop.value.string, 
			   Term.contents.term.name);
	    Warning(buf);
	}
    }
    if (octGetByName(&Term,&TermRelPosition)==OCT_OK) {
	padP->flp_relPos = TermRelPosition.contents.prop.value.real;
    }
}

checkPadProperties(oInst)
     octObject *oInst;		/* pointer to an instance */
{
    extern octObject TranCons;

    PlacementClass.contents.prop.value.string = "PAD";
    OH_ASSERT( octCreateOrModify( oInst, &PlacementClass ));

    TranCons.contents.prop.value.integer =
	~(1<<(int)oInst->contents.instance.transform.transformType);
    OH_ASSERT( octCreateOrModify( oInst, &TranCons ));

}
#endif
