#include "copyright.h"
#include "port.h"
#include "options.h"
#include "oct.h"
#include "vov.h"
#include "utility.h"
#include "oh.h"
#include "symlib.h"
/*
 *  PADP: Pad placement program
 *  Places all the formal terminals along the periphery of the chip.
 */
#include "defs.h"
#include "padp.h"

char    	*progName;
char		buf[256];			/* buffer for messages and errors */
int		verbose = FALSE;
int		moduleGenerator = FALSE;
int   		useFlpArea = FALSE;
int		useFlpConstraints = TRUE;
int		promoteFTermsToPads = FALSE;
int   		updateMobilityOfPads = TRUE;	/* Flag, to fix the pads after the placement */
int		implementFlag = FALSE;
int		sparcsConstraints = FALSE;
int		listOnlyFlag = FALSE;
int             createPads = FALSE;


octObject 	Chip;
octObject 	ChipInterface;
int		interfaceOpen = FALSE;



optionStruct optionList[] = {
    {"a",   "",               "Area specified in FLOORPLAN bag"},
    {"c",   "",               "Place pads evenly clockwise around the chip"},
    {"n",   "",               "Place pads to minimize netlength"},
    {"e",   "",		      "Exact placement of pads as specified (do not space equally)" },
    {"P",   "",               "The chip is pad-limited. Squeeze pads close together."},
    {"p",   "",               "Create new pads and put them in the dir. SPECIAL_CONTACTS."},
    {"f",   "",               "Do not move the pads, keep them fixed" },
    {"g",   "",               "Use padplace as a module generator"},
    {"S",   "",               "Add constraints for sparcs"},
    {"R",   "",               "Route the rings" },
    {"l",   "",               "List all formal terminals"},
    {"D", "fileName",         "Read formal terminal description from filename"},
    {"u", "familyName",       "Use a pad family to implement the formal terminals"},
    {"o", "cell:view",        "Specify output cell and view"},
    {"v",   "",               "Verbose mode"},
    {"V",   0,                "Print version number"},
    {OPT_RARG,    "",         "cell:view"},
    { 0, 0 , 0}
};

main(argc,argv)
    int argc;
    char *argv[];
{
    char* inputFacetName = NIL(char);
    char* outputFacetName = NIL(char);

    int 	c;
    schemeType	scheme = SCHEME_FLP_CONSTRAINTS;
    int 	fixed_pads = FALSE;
    int		exactPosition = FALSE;

    FILE	*padDescFile = 0;
    char	*descFile = NULL;
    octStatus	status;
    octObject 	Technology;
    int 	saveProps = TRUE;
    extern char	*version();
    ChipPads      chippads;
    int routeRings = FALSE;
    int	padLimited = FALSE;
    PadFamily family;
    int	PF_use = FALSE;
    char	*PF_name;	/* Name of the pad family */

    progName = argv[0];
    initPlacementObjects();
    /* Initialize Chippads */
    {
	int side;
	int row = 0;
	chippads.numberOfPads = 0;
	chippads.width = 0;
	chippads.height = 0;
	chippads.FirstPad = 0;
	chippads.LastPad = 0;
  
	for ( side  = LEFT; side <= BOTTOM; side++) {
	    chippads.availableSpaceSide[side] = 0;
	    for ( row = 0; row < DEF_MAX_ROWS; row++ ) {
		chippads.FirstPadInEdge[row][side] = NULL;
		chippads.LastPadInEdge[row][side] = NULL;
		chippads.padsInEdge[row][side] = 0;
		chippads.totalWidth[row][side] = 0;
	    }
	}
    }
    /* parse command line options*/
    while ((c = optGetOption(argc,argv)) != EOF) {
	switch(c) {
	case 'a':
	    useFlpArea = TRUE;
	    break;
	case 'u':
	    PF_use = TRUE;
	    PF_name = util_strsav(optarg);
	    break;
	case 'e':
	    exactPosition = TRUE;
	    break;
	case 'c':
	    scheme = SCHEME_CLOCKWISE;
	    break;
	case 'D':
	    descFile = util_strsav(optarg);
	    break;
	case 'f':
	    fixed_pads = TRUE;
	    break;
	case 'g':
	    PF_use = TRUE;
	    PF_name = "connFamily";
	    moduleGenerator = TRUE;
	    /* scheme = SCHEME_FLP_CONSTRAINTS; */
	    useFlpArea = TRUE;
	    useFlpConstraints = TRUE;
	    implementFlag = TRUE;
	    promoteFTermsToPads = TRUE;
	    break;
	case 'n':
	    scheme = SCHEME_MIN_NET_LENGTH;
	    break;
	case 'P':
	    padLimited = TRUE;
	    break;
	case 'p':
	    createPads = TRUE;
	    break;
	case 'R':
	    scheme = SCHEME_DONT_MOVE;
	    fixed_pads = TRUE;
	    routeRings = TRUE;
	    break;
	case 'S':
	    scheme = SCHEME_DONT_MOVE;
	    fixed_pads = TRUE;
	    sparcsConstraints = TRUE;
	    break;
	case 'l':
	    listOnlyFlag = TRUE;
	    break;
	case 'v':
	    verbose = 1;
	    break;
	case 'V':
	    Message( version() );
	    break;
	case 'o':
	    outputFacetName = util_strsav( optarg );
	    break;
	default:
	    optUsage();
	    break;
	}
    }

    /* the remaining arguments are argv[optind ... argc-1] */
    if (optind != argc -1) {
	optUsage();
    }

    inputFacetName = util_strsav( argv[optind] );

    VOVbegin( argc, argv );

    if ( descFile != 0 ) {
	VOVinput( VOV_UNIX_FILE, descFile );
	if (!(padDescFile = fopen( descFile, "r"))) {
	    (void) sprintf(buf,"Cannot read file: %s", descFile);
	    PADPFatalError(buf);
	}
    }


    if (moduleGenerator) {
	createModule( inputFacetName, outputFacetName, &Chip, &ChipInterface );
    } else {
	ohWorkFacet( &Chip, inputFacetName, outputFacetName );
    }


    
    
    ChipInterface = Chip;
    ChipInterface.contents.facet.facet = "interface";
    ChipInterface.contents.facet.mode = listOnlyFlag ? "r": "a";
    if (  octOpenFacet(&ChipInterface) < OCT_OK) {
	(void) sprintf(buf,"Problems opening %s",ohFormatName(&ChipInterface));
	PADPFatalError(buf);
    }

    if ( !listOnlyFlag ) {
	VOVinputFacet( &ChipInterface );
	VOVoutputFacet( &ChipInterface );
    }
    
    interfaceOpen = TRUE;

    if (ohGetByPropName( &Chip, &Technology, "TECHNOLOGY" ) != OCT_OK) {
	PADPFatalError("The TECHNOLOGY property is missing");
    }


    if (listOnlyFlag) {
	PADPgetArea( &chippads, 0, 0, 1 );
	PADPgetAllPads( &chippads, listOnlyFlag );
	PADPprintAllPads( chippads.FirstPad, stdout );
	VOVend( 0 );
    } 

    (void) sprintf(buf,"%s\nPlacement of pads for %s\n", version(), 
		   ohFormatName( &Chip ));
    Message(buf);


    PADPgetArea( &chippads, useFlpArea, padLimited, fixed_pads ); /* Get initial area of chip. */


    if ( PF_use ) {
	family.name = PF_name;
	PADPreadPadFamily(&chippads, &family );
    }

    PADPgetAllPads( &chippads );


    if ( padDescFile != NULL) {
	(void)scanAllFTerms( chippads.FirstPad, padDescFile );
    }

    /* Implement Formals and recompute the area of the chip now. */
    if ( PF_use )  {
	PADPimplementFormals( &chippads, &family );
	PADPgetArea( &chippads, useFlpArea, padLimited, fixed_pads );
    }

    PADPplace( &chippads, scheme, padLimited );

    if ( ! fixed_pads ) {
	PADPspace( &chippads, padLimited, exactPosition );
    }

    if ( verbose ) Message("Cleaning up all constraints" );
    symCleanAllConstraints( &Chip );

    if ( routeRings ) {
	PADPdoRings( &Chip, &chippads );
	VOVend( 0 );
    } 

    PADPsaveResult(&chippads,saveProps,sparcsConstraints,fixed_pads,updateMobilityOfPads);
  
    if ( moduleGenerator ) {
	createModuleInterface();
    }

    VOVend( 0 );
}



copyFacets ( cellName , oldView , newCell,  newView )
     char *cellName, *oldView , *newCell, *newView;
{
  octObject  		Input, Output;

  if (verbose) (void) printf("Copying %s:%s into %s:%s\n",cellName,oldView,cellName,newView);
  Input.type = OCT_FACET;
  Input.contents.facet.cell = cellName;
  Input.contents.facet.view = oldView;
  Input.contents.facet.facet = "contents";
  Input.contents.facet.version = OCT_CURRENT_VERSION;

  Output.type = OCT_FACET;
  Output.contents.facet.cell = newCell;
  Output.contents.facet.view = newView;
  Output.contents.facet.facet = "contents";
  Output.contents.facet.version = OCT_CURRENT_VERSION;

  VOVinputFacet( &Input );
  VOVoutputFacet( &Output );
  if (octCopyFacet(&Output,&Input) != OCT_OK)  PADPFatalError("Copying contents Facet");

  Input.contents.facet.facet = "interface";
  Output.contents.facet.facet = "interface";
  VOVinputFacet( &Input );
  VOVoutputFacet( &Output );
  if (octCopyFacet(&Output,&Input) != OCT_OK)  Warning("Copying interface Facet");

}


CheckType(prop, type)
    octObject *prop;
    octPropType type;
{
    if (prop->contents.prop.type != type) {
	(void) sprintf(buf,"Prop %s has wrong type. Trying to fix it",
				 prop->contents.prop.name);
	Message(buf);
	prop->contents.prop.type = type;
	switch (type) {
	case OCT_STRING:
	    prop->contents.prop.value.string = "";
	    break;
	case OCT_REAL:
	    prop->contents.prop.value.real = 0.0;
	    break;
	case OCT_INTEGER:
	    prop->contents.prop.value.integer = 0;
	    break;
	default:
	    break;
	}
    }
}
