#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "vulcan.h"
#include "vov.h"

static   octObject  		Input, Output;
static   octObject		InputInt, OutputInt;


createModule ( inputFacetName, outputFacetName, outChip, outChipInt )
    char *inputFacetName;
    char *outputFacetName;
    octObject* outChip;
    octObject* outChipInt;
{
    octGenerator	genTerm;
    octGenerator	genProp;
    octObject		IITerm,OITerm;
    octObject		Term,Prop,Geo;
    octObject		FloorplanBag;
    octObject		FloatingEdge;
    octObject		BigF, BigFLayer;
    octCoord		cx, cy, lx,ly;
    octObject		ViewType, EditStyle, Technology;

    ohUnpackDefaults( &Input, "r", "::contents" );
    ohUnpackFacetName( &Input, inputFacetName );
    InputInt = Input;
    InputInt.contents.facet.facet = "interface";

    Output = Input;
    if ( outputFacetName ) {
	ohUnpackFacetName( &Output, outputFacetName );
    } else {
	ohUnpackFacetName( &Output, ":padpMG" );
    }
    Output.contents.facet.mode = "w";

    OutputInt = Output;
    OutputInt.contents.facet.facet = "interface";

    if (verbose) (void) printf("Opening all facets");
    VOVinputFacet( &Input );
    VOVinputFacet( &InputInt );
    VOVoutputFacet( &Output );
    if (octOpenFacet(&Input)<OCT_OK) PADPFatalError("opening input facet");
    if (octOpenFacet(&InputInt)<OCT_OK) PADPFatalError("opening input facet");
    if (octOpenFacet(&Output)<OCT_OK) PADPFatalError("opening output facet");

    if (verbose) {
	(void) printf("Creating %s\n", ohFormatName(&Output) );
    }
    OH_ASSERT(octInitGenContents(&Input,OCT_TERM_MASK,&genTerm));
    while(octGenerate(&genTerm,&Term)==OCT_OK) {
	OH_ASSERT(octCreate(&Output,&Term));
    }
  

    FloorplanBag.type = OCT_BAG;
    FloorplanBag.contents.bag.name = "FLOORPLAN";
    if (octGetByName(&InputInt,&FloorplanBag)!= OCT_OK) 
      PADPFatalError("No floorplan bag");
    if (octGenFirstContent(&FloorplanBag,OCT_BOX_MASK,&Geo)!=OCT_OK)
      PADPFatalError("No box in floorplan bag");

    /* Create the cell on a few good layers */
    {
	octObject layer;
	ohGetOrCreateLayer(&Output,&layer,"MET1" );
	octCreate(&layer,&Geo);
	ohGetOrCreateLayer(&Output,&layer,"MET2" );
	octCreate(&layer,&Geo);
    }


    /* Put a big F inside the cell so that we can tell the orientation of the cell */
    cx = (Geo.contents.box.upperRight.x + Geo.contents.box.lowerLeft.x ) / 2;
    cy = (Geo.contents.box.upperRight.y + Geo.contents.box.lowerLeft.y ) / 2;
    lx = (Geo.contents.box.upperRight.x - Geo.contents.box.lowerLeft.x ) * 0.1;
    ly = (Geo.contents.box.upperRight.y - Geo.contents.box.lowerLeft.y ) * 0.1;


    BigFLayer.type = OCT_LAYER;
    BigFLayer.contents.layer.name = "POLY";
    OH_ASSERT(octCreate(&Output,&BigFLayer));

    BigF.type = OCT_BOX;
    BigF.contents.box.lowerLeft.x = cx - 2 * lx;
    BigF.contents.box.lowerLeft.y = cy - 3 * ly;
    BigF.contents.box.upperRight.x = cx - 1 * lx;
    BigF.contents.box.upperRight.y = cy + 3 * ly;
    OH_ASSERT(octCreate(&BigFLayer,&BigF));

    BigF.contents.box.lowerLeft.x = cx - 1* lx;
    BigF.contents.box.lowerLeft.y = cy + 2 * ly;
    BigF.contents.box.upperRight.x = cx + 2 * lx;
    BigF.contents.box.upperRight.y = cy + 3 * ly;
    OH_ASSERT(octCreate(&BigFLayer,&BigF));

    BigF.contents.box.lowerLeft.x = cx - 1 * lx;
    BigF.contents.box.lowerLeft.y = cy - 1 * ly;
    BigF.contents.box.upperRight.x = cx - 0 * lx;
    BigF.contents.box.upperRight.y = cy + 0 * ly;
    OH_ASSERT(octCreate(&BigFLayer,&BigF));


    VOVoutputFacet( &OutputInt );
    if (octOpenFacet(&OutputInt)<OCT_OK) PADPFatalError("opening output facet");

    /* Copy formal terminal properties */
    OH_ASSERT(octInitGenContents(&InputInt,OCT_TERM_MASK,&genTerm));
    while(octGenerate(&genTerm,&IITerm)==OCT_OK) {
	OITerm.type = OCT_TERM;
	OITerm.contents.term.name = IITerm.contents.term.name;
	OH_ASSERT(octGetByName(&OutputInt,&OITerm));
	OH_ASSERT(octInitGenContents(&IITerm,OCT_PROP_MASK,&genProp));
	while (octGenerate(&genProp,&Prop)==OCT_OK) {
	    OH_ASSERT(octCreate(&OITerm,&Prop));
	}

	/* Remove the TERM_FLOATING_EDGES property if existing */

	if (ohGetByPropName(&OITerm,&FloatingEdge,"TERM_FLOATING_EDGES" )==OCT_OK)
	  OH_ASSERT(octDelete(&FloatingEdge));
    }

    OH_ASSERT(octCreate(&OutputInt,&FloorplanBag));
    OH_ASSERT(octCreate(&FloorplanBag,&Geo));


    Technology.type = OCT_PROP;
    Technology.contents.prop.type = OCT_STRING;
    Technology.contents.prop.name = "TECHNOLOGY";
    OH_ASSERT(octGetByName(&Input,&Technology));
    OH_ASSERT(octCreate(&Output,&Technology));
    OH_ASSERT(octCreate(&OutputInt,&Technology));

    ViewType.type = OCT_PROP;
    ViewType.contents.prop.name = "VIEWTYPE";
    ViewType.contents.prop.type = OCT_STRING;
    ViewType.contents.prop.value.string = "PHYSICAL";
    OH_ASSERT(octCreate(&Output,&ViewType));
    OH_ASSERT(octCreate(&OutputInt,&ViewType));

    EditStyle.type = OCT_PROP;
    EditStyle.contents.prop.name = "EDITSTYLE";
    EditStyle.contents.prop.type = OCT_STRING;
    EditStyle.contents.prop.value.string = "PHYSICAL";
    OH_ASSERT(octCreate(&Output,&EditStyle));
    OH_ASSERT(octCreate(&OutputInt,&EditStyle));


    *outChip = Output;
    *outChipInt = OutputInt;


}


createModuleInterface()
{
    if (verbose) (void) printf("Calling vulcan on %s:%s\n",
			       Output.contents.facet.cell, Output.contents.facet.view);
    switch (vulcan(&Output,&OutputInt)) {
    case VULCAN_PARTIAL:
    case VULCAN_DEBUGGING:
    case VULCAN_INFORMATION:
	break;
    case VULCAN_WARNING:
	break;
    case VULCAN_SEVERE:
    case VULCAN_FATAL:
	PADPFatalError("vulcan failed");
	break;
    default:
	break;
    }

    if (verbose) (void) printf("Flushing new cell and its interface\n");

    OH_ASSERT(octFlushFacet(&Output));
    OH_ASSERT(octFlushFacet(&OutputInt));


}


