#include "copyright.h"
#include "port.h"
/*
 *  FILE: openFacet.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 */
#include "defs.h"
#include "octPlacement.h"
#include "oh.h"





void PADPsaveResult( chippads, savePadProps, sparcs, fixed, mobility )
    ChipPads* chippads;
    int savePadProps;		/* save TERM_EDGE and TERM_RELATIVE_POSITION */
    int sparcs;			/* put sparcs constraints */
    int fixed;			/* keep pads fixed */
    int mobility;		/* update MOBILITY prop  */
{
    struct padListEl	*padPtr;
    octObject 	oInstance;
    octObject	prop;		/* generic property */
    extern char	*stringSide();
  

    if (verbose) Message("Saving results");
    Chip.contents.facet.mode = "a";
    OH_ASSERT( octOpenFacet(&Chip) );
  
    for (padPtr = chippads->FirstPad ; padPtr != NULL ; padPtr = padPtr->next) {
	if ( savePadProps && !octIdIsNull(padPtr->FTermId)) {
	    octObject term;
	    OH_ASSERT( ohGetById( &term, padPtr->FTermId ) );
	    OH_ASSERT( octGetByName( &ChipInterface, &term) );
	    ohCreateOrModifyPropStr( &term, &prop, "TERM_EDGE", padPtr->termEdge );
	    ohCreateOrModifyPropReal( &term, &prop, "TERM_RELATIVE_POSITION", padPtr->relPos );
	}
	/* If the cell has no implementation skip this terminal. */
	if (octIdIsNull(padPtr->instanceId)) continue;

	OH_ASSERT( ohGetById( &oInstance, padPtr->instanceId ) );
	if ( ! fixed  ) {
	    oInstance.contents.instance.transform.transformType = padPtr->transform;
	    computeTranslation(padPtr, &oInstance );    
	    OH_ASSERT( octModify( &oInstance ));
	    if (verbose) {
		(void) printf("Pad %20s: TRANSFORM_CODE %d TRANSLATION %d,%d\n",
			      oInstance.contents.instance.name,
			      oInstance.contents.instance.transform.transformType,
			      oInstance.contents.instance.transform.translation.x,
			      oInstance.contents.instance.transform.translation.y);
	    }
	    PadEdge.contents.prop.value.string = stringSide(padPtr->side);
	    OH_ASSERT(octCreateOrModify(&oInstance,&PadEdge));
	}
	if ( sparcs )   putSparcsConstraints( &oInstance, padPtr->side);
	if ( mobility ) ohCreateOrModifyPropStr( &oInstance, &prop, "MOBILITY", "FIXED");
    }
    OH_ASSERT( octFlushFacet(&Chip) );
    OH_ASSERT( octFlushFacet(&ChipInterface));
}




pointTransform(point,transform)
    struct octPoint *point;
    octTransformType transform;
{
    octCoord tempx;
    /*
     *  apply the correct transforms to the  point
     */
    switch (transform) {
    case OCT_NO_TRANSFORM:
	break;
    case OCT_MIRROR_X:
	point->x = -point->x;
	break;
    case OCT_MIRROR_Y:
	point->y = -point->y;
	break;
    case OCT_ROT180:
	point->x = -point->x;
	point->y = -point->y;
	break;
    case OCT_ROT90:
	tempx = point->x;
	point->x = -point->y;
	point->y =  tempx;
	break;
    case OCT_ROT270:
	tempx = point->x;
	point->x =  point->y;
	point->y = -tempx;
	break;
    case OCT_MX_ROT90:
	tempx = point->x;
	point->x = -point->y;
	point->y = -tempx;
	break;
    case OCT_MY_ROT90:
	tempx = point->x;
	point->x =  point->y;
	point->y =  tempx;
	break;
    default:
	break;
    }
}

/*
 *  computeTranslation(firstVertex)
 *    given the instance, with the transformation already set
 *    and given the first vertex found in the XXX.dat file
 *    for this instance, compute the translation vector
 */
computeTranslation(padPtr, oInstance)
    struct padListEl *padPtr;    
    octObject *oInstance;
{
    struct octPoint lowerLeft;	 

    switch (padPtr->transform) {
    case OCT_NO_TRANSFORM:
    case OCT_MY_ROT90:
	lowerLeft.x = padPtr->masterBBox.lowerLeft.x;
	lowerLeft.y = padPtr->masterBBox.lowerLeft.y;
	break;
    case OCT_MIRROR_X:
    case OCT_ROT270:
	lowerLeft.x = padPtr->masterBBox.upperRight.x;
	lowerLeft.y = padPtr->masterBBox.lowerLeft.y;
	break;
    case OCT_MIRROR_Y:
    case OCT_ROT90:
	lowerLeft.x = padPtr->masterBBox.lowerLeft.x;
	lowerLeft.y = padPtr->masterBBox.upperRight.y;
	break;
    case OCT_ROT180:
    case OCT_MX_ROT90:
	lowerLeft.x = padPtr->masterBBox.upperRight.x;
	lowerLeft.y = padPtr->masterBBox.upperRight.y;
	break;
    default:
	break;
    }
    pointTransform(&lowerLeft,padPtr->transform);
    /*
     *  Finally, compute and store the translation vector.
     */
    oInstance->contents.instance.transform.translation.x = 
      padPtr->targetLowerLeft.x - lowerLeft.x;
    oInstance->contents.instance.transform.translation.y = 
      padPtr->targetLowerLeft.y - lowerLeft.y;
}
