#include "copyright.h"
#include "port.h"
#include "defs.h"
char ErrorBuffer[256];
/*
 * getBestPadPlacement
 *
 * Generates all actual terminals on all nets which pads terminals, finds the 
 * actual terminal closest to an edge, and sets the pad
 * location to the projection of that terminal on the edge.
 * Returns YES if a location was found, NO if not.
 */
int
getBestPadPlacement(chippads, padP )
    ChipPads* chippads;
    Pad *padP;
{
    struct octBox* area = &chippads->chipArea;

    octObject oInst, oPterm, oNetOrTerm, oNet, oTerm;
    octGenerator gPterm, gTerm;
    octStatus sPterm, sTerm;
    struct octPoint termloc;	/* location of each actual terminal	*/
    int bestloc;		/* location of the best projection	*/
    int bestside;		/* the best side			*/
    int bestD = 1000000000;	/* best distance found so far		*/
    int found1 = NO;		/* YES if an actual terminal was found	*/
    int	i;
    int	distance[4];


    oInst.objectId = padP->instanceId;
    if (octIdIsNull(oInst.objectId)) {
	(void) sprintf(buf,"Term %s has no implementation: cannot find best placement",
		       padP->name);
	Warning(buf);
	return(NO);
    }
    OH_ASSERT( octGetById( &oInst ));
    OH_ASSERT( octInitGenContents( &oInst, OCT_TERM_MASK, &gPterm ));
    while(( sPterm = octGenerate( &gPterm, &oPterm )) == OCT_OK ) {
	if( octGenFirstContainer( &oPterm, OCT_NET_MASK|OCT_TERM_MASK, &oNetOrTerm ) == OCT_OK ) {
	    /* case where oPterm is a promoted INSTANCE term 		*/
	    if( oNetOrTerm.type == OCT_NET ){
		oNet.objectId = oNetOrTerm.objectId;
		oNet.contents.net.name = oNetOrTerm.contents.net.name;

		/* case where oPterm is a promoted CONNECTOR term 		*/
	    } else if( octGenFirstContainer( &oNetOrTerm, OCT_NET_MASK, &oNet ) != OCT_OK ){
		continue;
	    }
	    OH_ASSERT( octInitGenContents( &oNet, OCT_TERM_MASK, &gTerm ));
	    while(( sTerm = octGenerate( &gTerm, &oTerm )) == OCT_OK ){
		if (octIdIsNull(oTerm.contents.term.instanceId) || 
		    octIdsEqual(oTerm.contents.term.instanceId,  oInst.objectId))
		  continue;
		if( getCenterOfTerm( &oTerm, &termloc )) {

		    found1 = 1;

		    distance[LEFT] = termloc.x - area->lowerLeft.x;
		    distance[BOTTOM] = termloc.y - area->lowerLeft.y;
		    distance[TOP] = area->upperRight.y - termloc.y;
		    distance[RIGHT] = area->upperRight.x - termloc.x;

		    for (i = LEFT ; i <= BOTTOM ; i++) {
			if (distance[i] < bestD) {
			    bestD = distance[i];
			    bestloc = (i == LEFT || i == RIGHT) ? distance[BOTTOM] : distance[LEFT];
			    bestside = i;
			}
		    }
		}
		OH_ASSERT( sTerm );
	    }
	}
    }
    OH_ASSERT( sPterm );
    if( found1 ){
	padP->flp_side = bestside;
	if(( bestside == LEFT )||( bestside == RIGHT )){
	    padP->flp_relPos = (double)bestloc/(double)chippads->height;
	} else {
	    padP->flp_relPos = (double)bestloc/(double)chippads->width;
	}
    }
    return found1;
}

/* 
 * getCenterOfTerm
 *
 * Sets termloc to the center of oTerm's bounding box.
 * To do this, oTerm's instance's master's interface facet is opened,
 * the formal term is gotton by oTerm's name, the bounding box of the
 * term is derived, the bounding box is transformed by oTerm's instance's
 * transform, and termloc is set to the center of the transformed bbox.
 *
 * returns YES if everything worked correctly, NO otherwise
 */
getCenterOfTerm( oTerm, termloc )
    octObject *oTerm;		/* must be a valid octObject		*/
    struct octPoint *termloc;	/* must be allocated			*/
{
    octObject oInst, oMaster, oFterm, oBox;
    octGenerator gBox;
    octStatus sBox;
    struct octBox bbox;

    oInst.objectId = oTerm->contents.term.instanceId;
    OH_ASSERT( octGetById( &oInst ));

    oMaster.type = OCT_FACET;
    oMaster.contents.facet.cell = oInst.contents.instance.master;
    oMaster.contents.facet.view = oInst.contents.instance.view;
    oMaster.contents.facet.facet = "interface";
    oMaster.contents.facet.version = OCT_CURRENT_VERSION;
    oMaster.contents.facet.mode = "r";
    if( octOpenFacet( &oMaster ) < OCT_OK ){
	(void) sprintf(ErrorBuffer, 
		"Could not open master %s:%s:%s of instance %s of term %s", 
		oMaster.contents.facet.cell,
		oMaster.contents.facet.view, 
		oMaster.contents.facet.facet,
		oInst.contents.instance.name,
		oTerm->contents.term.name );
	octError( ErrorBuffer );
	return(NO);
    }
    oFterm = *oTerm;
    if( octGetByName( &oMaster, &oFterm ) < OCT_OK ){
	(void) sprintf(ErrorBuffer, 
		"Could not match term %s to a formal term on master %s:%s:%s",
		oTerm->contents.term.name,
		oMaster.contents.facet.cell,
		oMaster.contents.facet.view, 
		oMaster.contents.facet.facet );
	octError( ErrorBuffer );
	return(NO);
    }
    OH_ASSERT( octInitGenContents( &oFterm, OCT_BOX_MASK, &gBox ));
    /* setting bbox to the first generated implementation		*/
    OH_ASSERT( sBox = octGenerate( &gBox, &oBox ));
    if( sBox > OCT_OK ){
	(void) sprintf(ErrorBuffer, 
		"Formal term %s on master %s:%s:%s has no implementation",
		oTerm->contents.term.name,
		oMaster.contents.facet.cell,
		oMaster.contents.facet.view, 
		oMaster.contents.facet.facet );
	octError( ErrorBuffer );
	return(NO);
    } else {
	bbox = oBox.contents.box;
    }
    /* setting bbox to the max( bbox, other implementations )		*/
    while(( sBox = octGenerate( &gBox, &oBox )) == OCT_OK ){
	bbox.lowerLeft.x = MIN( bbox.lowerLeft.x, oBox.contents.box.lowerLeft.x );
	bbox.lowerLeft.y = MIN( bbox.lowerLeft.y, oBox.contents.box.lowerLeft.y );
	bbox.upperRight.x = MAX( bbox.upperRight.x, oBox.contents.box.upperRight.x );
	bbox.upperRight.y = MAX( bbox.upperRight.y, oBox.contents.box.upperRight.y );
    }
    OH_ASSERT( sBox );
    /* transforming bbox 						*/
    oBox.contents.box = bbox;
    octTransformGeo( &oBox, &oInst.contents.instance.transform );
    termloc->x = ( oBox.contents.box.lowerLeft.x + oBox.contents.box.upperRight.x )/2;
    termloc->y = ( oBox.contents.box.lowerLeft.y + oBox.contents.box.upperRight.y )/2;
    return( YES );
}
