#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"

void 
readPadMaster(chippads, padP)
    ChipPads* chippads;
    Pad* padP;
{
    octObject oMaster;
    octObject oInst;

    /* 
     *   get the Master of this instance
     *   Open the interface facet to find the bounding box
     */

    oMaster.type = OCT_FACET;
    oMaster.contents.facet.facet="interface";
    oMaster.contents.facet.mode="r";
    oInst.objectId = padP->instanceId;
    OH_ASSERT( octGetById( &oInst ));

    padP->MasterCell = oInst.contents.instance.master;
    padP->MasterView = oInst.contents.instance.view;
    padP->transform  =  oInst.contents.instance.transform.transformType;
    OH_ASSERT( octOpenMaster( &oInst, &oMaster) );
    OH_ASSERT( octBB( &oMaster, &padP->masterBBox) );

    padP->internalEdgeOfMaster = findInternalEdge( &oMaster );

    padP->width  = padP->masterBBox.upperRight.x -  padP->masterBBox.lowerLeft.x;
    padP->height = padP->masterBBox.upperRight.y -  padP->masterBBox.lowerLeft.y;
    if (( padP->internalEdgeOfMaster == LEFT )|| ( padP->internalEdgeOfMaster == RIGHT )){
	SWAP(int, padP->width, padP->height);
    } 
    chippads->maxPadHeight = MAX (chippads->maxPadHeight, padP->height);
    chippads->maxPadWidth  = MAX (chippads->maxPadWidth , padP->width);
}

int
findInternalEdge(Master)
     octObject *Master;		/* must be already open */
{
    static int PF_internalEdge = NO_SIDE;
    char *stringSide();
    octObject Term;
    octGenerator gTerm;
    struct octBox masterBBox, termBBox;
    int	termsInSide[4];
    int   shortestDistanceSide[4];
    octCoord   distance[4];	/* of a terminal from all the edges */
    int   shortestDistanceIndex;
    int	maxSideLength;
    int	sideWithMostTerms;
    int	probableInnerSide;
    int   I;			/* a counter */
    octObject InternalEdge;

    if ( PF_internalEdge != NO_SIDE ) return PF_internalEdge;
    
    InternalEdge.type = OCT_PROP;
    InternalEdge.contents.prop.name = "PAD_INTERNAL_EDGE";
    if (octGetByName(Master,&InternalEdge) == OCT_OK) {
	probableInnerSide = parseEdgeName(InternalEdge.contents.prop.value.string);
	if (probableInnerSide != NO_SIDE) {
	    return(probableInnerSide);
	} else {
	    Warning("Illegal value for PAD_INTERNAL_EDGE prop");
	}
    }

    {
	static int once = 0;
	if ( ! once++ ) {
	    Warning("Missing prop PAD_INTERNAL_EDGE: trying to guess it");
	    Warning("This message is given only once");
	}
    }
    OH_ASSERT(octBB(Master, &masterBBox));

    /* Initialize some arrays */
    maxSideLength = MAX (masterBBox.upperRight.x - masterBBox.lowerLeft.x,
			 masterBBox.upperRight.y - masterBBox.lowerLeft.y);
    for (I=LEFT; I<=BOTTOM; I++) {
	termsInSide[I] = 0;
	shortestDistanceSide[I] = maxSideLength;
    }

    OH_ASSERT(octInitGenContents(Master, OCT_TERM_MASK , &gTerm));
    while (octGenerate(&gTerm,&Term) == OCT_OK) {
	if (getTermBBox(&Term,&termBBox) == FALSE) {
	    (void) (void) printf("WARNING:: Terminal %s of %s has no implementation\n",
				 Term.contents.term.name, Master->contents.facet.cell);
	} else {
	    distance[LEFT] = ABS(masterBBox.lowerLeft.x - termBBox.lowerLeft.x);
	    distance[BOTTOM] = ABS(masterBBox.lowerLeft.y - termBBox.lowerLeft.y);
	    distance[RIGHT] = ABS(masterBBox.upperRight.x - termBBox.upperRight.x);
	    distance[TOP] = ABS(masterBBox.upperRight.y - termBBox.upperRight.y);
	    shortestDistanceIndex = LEFT;
	    for (I=TOP; I<=BOTTOM; I++) {
		if (distance[I] < distance[shortestDistanceIndex]) {
		    shortestDistanceIndex = I;
		}
	    }
	    termsInSide[shortestDistanceIndex]++;
	    shortestDistanceSide[shortestDistanceIndex] =
	      MIN(shortestDistanceSide[shortestDistanceIndex] , distance[shortestDistanceIndex] );
	}
    }
    probableInnerSide = LEFT;	/*  */
    sideWithMostTerms = LEFT;	/* synonymous of probableInner side */
    for (I=TOP; I<=BOTTOM; I++) {
	if (termsInSide[I] >= termsInSide[sideWithMostTerms] && 
	    shortestDistanceSide[I] < shortestDistanceSide[probableInnerSide]) {
	    probableInnerSide = I;
	    sideWithMostTerms = I;
	}
    }
    /*  check if there are terms also on other edges */
    /*  this might be a bad sign */
    for (I=LEFT; I<=BOTTOM; I++) {
	if (I!=sideWithMostTerms AND termsInSide[I] == termsInSide[sideWithMostTerms]) {
	    (void) printf("Warning: %s has %d terms in %s side and %d in %s side\n",
			  Master->contents.facet.cell,
			  termsInSide[sideWithMostTerms],
			  stringSide(sideWithMostTerms),
			  termsInSide[I],
			  stringSide(I));
	}
    }
    return(sideWithMostTerms);
}


getTermBBox(Term,termBBox)
     octObject *Term;
     struct octBox *termBBox;
{
  octGenerator genGeo;
  octObject    Geo;
  
  OH_ASSERT(octInitGenContents(Term, OCT_GEO_MASK | OCT_INSTANCE_MASK, &genGeo));
  if (octGenerate(&genGeo,&Geo) == OCT_OK) {
    OH_ASSERT(octBB(&Geo,termBBox));
    return(TRUE);
  } else {
     return(FALSE);
  }
}


void PADPplaceDontMove( chippads )
    ChipPads* chippads;
{
    octObject instance;
    struct octBox bb;
    Pad* padP;
    int side;
    double pos;

    Message( "PADS are not moved" );
    
    for (padP=chippads->FirstPad; padP!=NULL; padP=padP->next) {
	OH_ASSERT( ohGetById( &instance, padP->instanceId ) );
	OH_ASSERT( octBB( &instance, &bb ) );
	side= padP->side;
	if ( side == LEFT || side == RIGHT ) {
	    pos = bb.lowerLeft.y;
	} else {
	    pos = bb.lowerLeft.x;
	}

	addPadToEdge(chippads, padP, side, pos, padP->row);
    } 
}

/* place the pads clockwise, starting from lower Left corner */
void PADPplace( chippads, scheme, padLimited )
    ChipPads* chippads;
    schemeType		scheme;
    int padLimited;
{
    extern octTransformType putPinsInside();
    int 	averagePadPerEdge;
    int 	side;
    int 	tWidthInThisSide = 0;
    int 	tPadsInThisSide = 0 ;
    int         maxPadsUpToHere = 0;
    int		positionInSide;
    int 	ntPadsSide[4];
    int 	tWidthPadsSide[4];
    long 	random();
    long	time();
    Pad		*padP;
    char 	*stringSide();
    int		row;
    double 	pos;		/* position in edge of a pad */
    int		bestPlacementFound;
    int         deadCorner = getDeadCorner( chippads );

    if (verbose) Message("Placing the pads");

    /* Place pads using the constraints given by the floorplanner */
    chippads->availableSpaceSide[LEFT]  = chippads->height;
    chippads->availableSpaceSide[TOP]   = chippads->width;
    chippads->availableSpaceSide[RIGHT] = chippads->height;
    chippads->availableSpaceSide[BOTTOM]= chippads->width;

    if ((chippads->numberOfPads % 4) == 0) {
	averagePadPerEdge = (chippads->numberOfPads / 4) ;
    } else {
	averagePadPerEdge = (chippads->numberOfPads / 4) + 1;
    }
  
    switch (scheme) {
    case SCHEME_CLOCKWISE:
	if (verbose) {
	    (void)printf("Placing pads clockwise\n");
	    if ( padLimited) (void) printf("Placing pads in assuming chip is PAD LIMITED\n");
	}
	/* Place pads clockwise starting from left edge */

	/* Assign the pads to the sides */

	side = LEFT;
	tWidthInThisSide = 2 * deadCorner;	/* reset */
	tPadsInThisSide = 0;	/* reset */
	maxPadsUpToHere = averagePadPerEdge;
	for (padP= chippads->FirstPad; padP!=NULL; padP=padP->next) {
	    if (side == LEFT || side == TOP) {
		positionInSide = tWidthInThisSide;
	    } else {
		positionInSide = chippads->availableSpaceSide[side] - tWidthInThisSide;
	    }
	    row = 0;
	    if ( padLimited ) {
		pos = (double)tPadsInThisSide / (double)averagePadPerEdge;
	    } else {
		pos = (double)positionInSide / (double)chippads->availableSpaceSide[side];
	    }
	    addPadToEdge(chippads, padP , side , pos, row );

	    if ( padP->width ) {
		tWidthInThisSide += padP->width;
	    } else {
		/* This takes care of chips with pads not yet implemented (e.g. wolfe cell) */
		tWidthInThisSide += chippads->availableSpaceSide[side]/averagePadPerEdge;
	    }

	    tPadsInThisSide++;	/* one more pad on this side */
	    ntPadsSide[side]= tPadsInThisSide;
	    tWidthPadsSide[side]= tWidthInThisSide;
	    {
		int nextSideFlag = 0;
		if ( padLimited ) {
		    if ( tPadsInThisSide >= maxPadsUpToHere ) {
			chippads->availableSpaceSide[side] = tWidthInThisSide;
			nextSideFlag = 1;
		    }
		} else {
		    if (tPadsInThisSide>=maxPadsUpToHere||tWidthInThisSide>chippads->availableSpaceSide[side]) {
			nextSideFlag = 1;
		    }
		}
		if ( nextSideFlag ) {
		    /* Allow pads that did not fit this edge to be put in the next one */
		    maxPadsUpToHere = averagePadPerEdge + (maxPadsUpToHere - tPadsInThisSide);

		    side++;	/* Go on to next side */
		    tWidthInThisSide = 2 * deadCorner; /* reset */
		    tPadsInThisSide = 0; /* reset */
		    if ( side > BOTTOM && padP->next!=NULL) {
			PADPFatalError("Insufficient space around core");
		    }
		}
	    }
	}
	if ( padLimited ) {
	    chippads->availableSpaceSide[side] = tWidthInThisSide;
	}


    
	break;

    case SCHEME_FLP_CONSTRAINTS:
    case SCHEME_MIN_NET_LENGTH:

	for (padP=chippads->FirstPad; padP!=NULL; padP=padP->next) {

	    if ( scheme == SCHEME_MIN_NET_LENGTH ) {
		bestPlacementFound = getBestPadPlacement( chippads, padP );
	    }

	    if (scheme == SCHEME_FLP_CONSTRAINTS && padP->flp_side == NO_SIDE ||
		scheme == SCHEME_MIN_NET_LENGTH && ! bestPlacementFound ) {
		padP->flp_side   = (int)    (random() &0xF) % 4;
		padP->flp_relPos = (double) (random()&0xFFFF) / (double) 0xFFFF ;
		(void) printf("Random placement: %20s (%s,%.2f)\n",
			      padP->name, stringSide(padP->flp_side), padP->flp_relPos);
	    }

	    addPadToEdge(chippads, padP, padP->flp_side, padP->flp_relPos, padP->row);

	} 

	break;

    case SCHEME_DONT_MOVE:
	PADPplaceDontMove( chippads );
	break;
    case SCHEME_UNDEFINED:
    default:
	PADPFatalError("Unknown scheme");
	break;
    }
}

