#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"
#include "padp.h"

extern char	*stringSide();
extern Pad* newPad();

void determinePadSide( chippads, pad, Instance )
    ChipPads* chippads;
    Pad* pad;
    octObject *Instance;
{
    struct octBox instBB;
    struct octBox chipBB;
    octObject PadEdge;
    int distance[4];
    int actualSide = NO_SIDE;
    int propSide = NO_SIDE;
    int side;

    chipBB = chippads->chipArea;
    OH_ASSERT( octBB( Instance, &instBB ) );
    
    distance[LEFT]   = ABS( chipBB.lowerLeft.x - instBB.lowerLeft.x );
    distance[RIGHT]  = ABS( instBB.upperRight.x - chipBB.upperRight.x );
    distance[BOTTOM] = ABS( chipBB.lowerLeft.y - instBB.lowerLeft.y );
    distance[TOP]    = ABS( instBB.upperRight.y - chipBB.upperRight.y );
    
    actualSide = LEFT;
    for ( side = LEFT ; side <= BOTTOM; side++ ) {
	if ( distance[side] < distance[actualSide] ) {
	    actualSide = side;
	}
    }

    if ( verbose ) {
	sprintf( buf, "%s assigned to %s side (L%d,T%d,R%d,B%d)",
		ohFormatName( Instance ),
		stringSide(actualSide),
		distance[LEFT],distance[TOP],distance[RIGHT],distance[BOTTOM] );
	Message( buf );
    }

    pad->side = actualSide;
    if (ohGetByPropName( Instance, &PadEdge, "PAD_EDGE" ) == OCT_OK) {
	propSide =  parseEdgeName(PadEdge.contents.prop.value.string);
	if ( propSide == NO_SIDE ) {
	    sprintf( buf, "Cannot parse PAD_EDGE in %s", ohFormatName( Instance ) );
	    Warning( buf );
	} else {
	  if ( actualSide != NO_SIDE && propSide != actualSide ) {
	      sprintf(buf, "%s is actually on %s side, but PAD_EDGE says %s side",
		      ohFormatName( Instance ), 
		      stringSide( actualSide ),
		      stringSide( propSide ) );
	      Warning( buf );
	  }
	  pad->side = propSide;
	}
    } else {
	OH_ASSERT(ohCreatePropStr(Instance,&PadEdge,"PAD_EDGE",stringSide(actualSide)));
    }

}

void getFormalTermProps( padP, term )
    Pad* padP;
    octObject *term;
{
    octObject oTermInterface, prop;
    
    OH_ASSERT( ohGetByTermName(&ChipInterface,&oTermInterface,term->contents.term.name));

    if (octGetByName(&oTermInterface,&Direction)==OCT_OK) {
	padP->direction = util_strsav(Direction.contents.prop.value.string);
    }
    if (octGetByName(&oTermInterface,&TermType)==OCT_OK) {
	padP->termType = util_strsav(TermType.contents.prop.value.string);
    }
    if (octGetByName(&oTermInterface,&FloatingEdge)==OCT_OK) {
	CheckType(&FloatingEdge,OCT_STRING);
	padP->floatingEdges = util_strsav(FloatingEdge.contents.prop.value.string);
    }
    if (ohGetByPropName(&oTermInterface,&prop,"TERM_ROW")==OCT_OK) {
	CheckType(&prop,OCT_INTEGER);
	padP->row = prop.contents.prop.value.integer;
    }
    if (octGetByName(&oTermInterface,&TermEdge)==OCT_OK) {
	padP->termEdge = util_strsav(TermEdge.contents.prop.value.string);
	padP->flp_side = parseEdgeName(TermEdge.contents.prop.value.string);
	if (padP->flp_side == NO_SIDE) {
	    (void) sprintf(buf,"Illegal edge specification `%s' for term %s",
			   TermEdge.contents.prop.value.string, 
			   oTermInterface.contents.term.name);
	    Warning(buf);
	}
    }
    if (octGetByName(&oTermInterface,&TermRelPosition)==OCT_OK) {
	CheckType(&TermRelPosition,OCT_REAL);
	padP->flp_relPos = TermRelPosition.contents.prop.value.real;
    }
    
    if (octGetByName(&oTermInterface,&PeakCurrent)==OCT_OK) {
	CheckType(&PeakCurrent,OCT_REAL);
	padP->peakCurrent = PeakCurrent.contents.prop.value.real;
    }

}

void getFormalTermPads( chippads, padBag )
    ChipPads* chippads;
    octObject *padBag;
{
    Pad			*padP=NULL;
    octGenerator 	gTerm;
    extern octObject   	ConnectorBag;
    octObject		Instance, Net;
    octObject		oTerm;
    octObject		oImplTerm;


    OH_ASSERT(octGetOrCreate( &Chip, &ConnectorBag ));

    /* generate all formal terminals, find the pads and flag them */
    OH_ASSERT( octInitGenContents(&Chip, OCT_TERM_MASK, &gTerm));
    while ( octGenerate( &gTerm, &oTerm ) == OCT_OK) {
	if (!octIdIsNull(oTerm.contents.term.instanceId))  continue;

	padP = newPad( chippads );
	padP->FTermId = oTerm.objectId;
	padP->name = util_strsav(oTerm.contents.term.name);

	/* 
	 * Find the implementation of the formal terminal.
	 */
	if (octGenFirstContent( &oTerm, OCT_TERM_MASK, &oImplTerm )==OCT_OK) {
	    padP->implementationId = oImplTerm.objectId;
	    padP->bondingTermName = util_strsav(oImplTerm.contents.term.name);
	    padP->instanceId = oImplTerm.contents.term.instanceId;
	    OH_ASSERT(ohGetById(&Instance, padP->instanceId ) ); 
	    OH_ASSERT( octAttach(padBag, &Instance ));

	    if (octIsAttached(&ConnectorBag,&Instance)==OCT_OK) {
		padP->connectorImplementation = TRUE;
	    }
	    readPadMaster( chippads, padP );
      
	    determinePadSide( chippads, padP, &Instance );
	    if (octGetByName(&Instance,&PadEdge) == OCT_OK) {
		padP->side = parseEdgeName(PadEdge.contents.prop.value.string);
	    }
	} else {		/* No implementation. */
	    if (octGenFirstContainer(&oTerm,OCT_NET_MASK,&Net)!=OCT_OK)  {
		(void)sprintf(buf,"%s has no implementation and is not connected.",oTerm.contents.term.name);
		Warning(buf);
	    } else {
		padP->netId = Net.objectId;
		padP->netName = util_strsav(Net.contents.net.name);
	    }
	}
	getFormalTermProps( padP, &oTerm );
    }
}


getInstancePads( chippads, padBag ) 
    ChipPads* chippads;
    octObject *padBag;
{
    Pad			*padP=NULL;
    octObject oInst;
    octGenerator gInst;
    /* 
     * Now look at all the instances.
     */
    
    OH_ASSERT( octInitGenContents(&Chip, OCT_INSTANCE_MASK, &gInst));
    while( octGenerate(&gInst, &oInst) == OCT_OK ) {
	if ( !isPad( &oInst ) ) continue;
	if ( octIsAttached( padBag, &oInst ) == OCT_OK ) continue;
	
	padP = newPad( chippads );

	OH_ASSERT( octAttach(padBag, &oInst ));
	padP->instanceId = oInst.objectId;
	padP->FTermId = 0;
	padP->name = util_strsav(oInst.contents.instance.name);

	if (octIsAttached(&ConnectorBag,&oInst)==OCT_OK) {
	    padP->connectorImplementation = TRUE;
	}
	readPadMaster( chippads, padP );
      
	determinePadSide( chippads, padP, &oInst );

	if (octGetByName(&oInst,&PadEdge) == OCT_OK) {
	    padP->side = parseEdgeName(PadEdge.contents.prop.value.string);
	} 
    }
}

void 
PADPgetAllPads( chippads )
    ChipPads* chippads;
{
    octObject padBag;

    if (verbose) Message("Reading all the formal terminals");

    OH_ASSERT( ohCreateBag( &Chip, &padBag, "TEMPORARY PAD BAG" ) );
    octMarkTemporary( &padBag );
    getFormalTermPads( chippads, &padBag );
    getInstancePads( chippads, &padBag );
}


void initPad( pad )
    Pad* pad;
{
    pad->FTermId = 0;
    pad->name = "<unspecified>";
    pad->connectorImplementation = FALSE;
    pad->implementationId = 0;
    pad->bondingTermName = "<unspecified>"; 		/*  */
    pad->transform = OCT_NO_TRANSFORM;	
    pad->MasterCell = "";
    pad->MasterView = "";
    pad->instanceId = 0;
    pad->termType = "<unspecified>";
    pad->direction = "<unspecified>";
    pad->flp_side   = NO_SIDE;
    pad->termEdge = "<unspecified>";
    pad->flp_relPos = 0.0;
    pad->floatingEdges = "";				/*  */
    pad->positionInSide = 0;
    pad->netId = 0;
    pad->netName = "<unspecified>";
    pad->row = 0;					
    pad->peakCurrent = 0.0;

    pad->nextInEdge = NULL;
    pad->precInEdge = NULL;

}


void printString( fp, n, c, l )
    FILE* fp;
    char* n;			/* New value for this pad. */
    char** c;			/* Current value. */
    char*  l;			/* Label */
{
    if ( strcmp( n, *c ) != 0 ) {
	if (  n[0] != '\0' ) {
	    (void) fprintf(fp, "\t%s %s\n", l, n );
	}
	*c = n;
    }
}


void printInt( fp, n, c, l )
    FILE* fp;
    int n;			/* New value for this pad. */
    int* c;			/* Current value. */
    char*  l;			/* Label */
{
    if (  n != *c )  {
	(void) fprintf(fp, "\t%s %d\n", l, n );
	*c = n;
    }
}

void printDouble( fp, n, c, l )
    FILE* fp;
    double n;			/* New value for this pad. */
    double* c;			/* Current value. */
    char*  l;			/* Label */
{
    if (  n != *c )  {
	(void) fprintf(fp, "\t%s %.2f\n", l, n );
	*c = n;
    }
}




void
PADPprintAllPads( padListHandle, fp )
    Pad* padListHandle;
    FILE *fp;
{
    Pad		*padP;
    Pad		curpad;

    initPad( &curpad );

    (void) fprintf(fp,"#\n# Formal Terminal description for %s:%s\n",
		   Chip.contents.facet.cell,
		   Chip.contents.facet.view);
    (void) fprintf(fp,"# padp recognizes the following keywords:\n");
    (void) fprintf(fp,"# TERMTYPE (SUPPLY | GROUND | SIGNAL | CLOCK) : DEFAULT = SIGNAL\n");
    (void) fprintf(fp,"# DIRECTION (INPUT | OUTPUT | INOUT) : DEFAULT = INOUT\n");
    (void) fprintf(fp,"# PEAKCURRENT ( x (in Amperes)) : DEFAULT = 0.0\n");
    (void) fprintf(fp,"#   NOTE: PEAKCURRENT applies only to SUPPLY and GROUND terminals\n");
    (void) fprintf(fp,"# TERM_FLOATING_EDGES (ALL | {LEFT|RIGHT|TOP|BOTTOM})\n");
    (void) fprintf(fp,"# TERM_EDGE (LEFT | RIGHT | TOP | BOTTOM) : DEFAULT = none\n");
    (void) fprintf(fp,"# TERM_ROW ( 0 | 1 | 2 | ... | n) : DEFAULT = 0\n");
    (void) fprintf(fp,"# TERM_RELATIVE_POSITION ( 0.0 <= x <= 1.0) : DEFAULT = none\n");
    (void) fprintf(fp,"# TERM_RELATIVE_POSITION_STEP ( 0.0 <= x <= 1.0) : DEFAULT = 0.0\n");
    (void) fprintf(fp,"\n");
    (void) fprintf(fp,"\
# This file uses a 'REVERSE' logic: \n\
# First you specify the properties\n\
# then the terminal for which the props apply\n\
");

    for (padP = padListHandle ; padP != NULL ; padP = padP->next) {
	printString(fp, padP->termType, &curpad.termType, "TERMTYPE" );
	printString(fp, padP->direction, &curpad.direction, "DIRECTION" );
	printString(fp, padP->termEdge, &curpad.termEdge, "TERM_EDGE" );
	printString(fp, padP->floatingEdges, &curpad.floatingEdges, "TERM_FLOATING_EDGES");
	printInt(fp, padP->row, &curpad.row, "TERM_ROW" );
	printDouble(fp, padP->flp_relPos, &curpad.flp_relPos, "TERM_RELATIVE_POSITION" );
	if (strcmp(padP->termType,"SUPPLY")==0 || strcmp(padP->termType,"GROUND")==0) {
	    printDouble(fp, padP->peakCurrent, &curpad.peakCurrent, "PEAKCURRENT");
	}
	if ( octIdIsNull( padP->FTermId ) ) {
	    (void)fprintf(fp, "PAD_INSTANCE %s\n" , padP->name);
	} else {
	    (void)fprintf(fp, "FORMAL_TERMINAL %s\n" , padP->name);
	}
    }
}



int
scanAllFTerms( listHandle, fp)
    Pad* listHandle;
    FILE *fp;
{
    Pad			padStatus;
    char 		input[256]; 	/*  */
    char		string[256];
    int 		test;		/* Return value of the fscanf() function */
    int			integer; 	/*  */
    int			i;		/*  */
    double 		real;		/*  */
    double		relPosStep = 0.0;

    if (verbose) Message("Parsing the description file");
    initPad( &padStatus );

    while (fscanf(fp , " %s " , input) ==1) {
	/* The "#" sign at the beginnig of a line indicates a comment in the parPuppy file */
	if  (strcmp( input , "#" ) == 0) { 
	    while ( getc(fp) != '\n') ;	/*   COMMENT: go on to the next line */
	    continue;
	}
     
	if (strcmp(input,"FORMAL_TERMINAL") == 0) {
	    test = fscanf( fp, " %s " , string);
	    if (test != 1) {
		Warning("illegal syntax");
		continue;
	    }
	    padStatus.name = string;
	    checkPad( listHandle, &padStatus, 1);
	    padStatus.flp_relPos += relPosStep;
	} else if (strcmp(input,"PAD_INSTANCE") == 0) {
	    test = fscanf( fp, " %s " , string);
	    if (test != 1) {
		Warning("illegal syntax");
		continue;
	    }
	    padStatus.name = string;
	    checkPad( listHandle, &padStatus, 0 );
	    padStatus.flp_relPos += relPosStep;
	} else if (strcmp(input,"TERMTYPE") == 0) {
	    test = fscanf( fp, " %s " , string);
	    padStatus.termType = util_strsav(string);
	} else if (strcmp(input,"DIRECTION") == 0) {
	    test = fscanf( fp, " %s " , string);
	    padStatus.direction = util_strsav(string);
	} else if (strcmp(input,"TERM_EDGE") == 0) {
	    test = fscanf( fp, " %s " , string);
	    padStatus.flp_side = parseEdgeName(string);
	    padStatus.termEdge = util_strsav(string);
	} else if (strcmp(input,"PEAKCURRENT") == 0) {
	    test = fscanf( fp, " %lf " , &real);
	    padStatus.peakCurrent = real;
	} else if (strcmp(input,"TERM_ROW") == 0) {
	    test = fscanf( fp, " %d " , &integer);
	    padStatus.row = integer;
	} else if (strcmp(input,"TERM_RELATIVE_POSITION") == 0) {
	    test = fscanf( fp, " %lf " , &real);
	    padStatus.flp_relPos = real;
	} else if (strcmp(input,"TERM_RELATIVE_POSITION_STEP") == 0) {
	    test = fscanf( fp, " %lf " , &real);
	    relPosStep = real;
	} else if (strcmp(input,"NETNAME") == 0) {
	    test = fscanf( fp, " %s " , string);
	    padStatus.netName = util_strsav(string);
	} else if (strcmp(input,"TERM_FLOATING_EDGES") == 0) {
	    /* Go on to the end of the line */
	    i = 0;
	    while ((string[i] = getc(fp)) != '\n') i++;
	    string[i] = '\0';
	    padStatus.floatingEdges = util_strsav(string);
	} else {
	    (void) sprintf(buf,"Unparsable word >>> %s found\n",input );
	    Warning(buf);
	}
    }
    return(OK);
}


checkPad( listHandle, padP, ftermFlag)
    Pad* listHandle;
    Pad *padP;
    int ftermFlag;
{
    octObject	Term;
    octObject	TermInterface;
    octObject   prop;
    Pad		*curpad;
    int		found = FALSE;

    /* Find current information about this pad */
    /* Linear search. this will take a while */
    for (curpad = listHandle; curpad !=NULL ; curpad = curpad->next) {
	if (strcmp(curpad->name,padP->name)==0) {
	    if ( ftermFlag ) {
		if ( curpad->FTermId != 0 )  {
		    found = TRUE;
		    break;
		}
	    } else {
		found = TRUE;
		break;
	    }
	}
    }
    if (!found) {
	(void) sprintf(buf,"Pad %s not existing", padP->name);
	PADPFatalError(buf);
    }

    if ( ! ftermFlag ) {
	/* It is a PAD_INSTANCE */
	curpad->floatingEdges = util_strsav(padP->floatingEdges);
	curpad->termEdge = util_strsav(padP->termEdge);
	curpad->flp_side = padP->flp_side;
	curpad->flp_relPos = padP->flp_relPos;
	curpad->row = padP->row;
    } else {
	Term.type = OCT_TERM;
	TermInterface.type = OCT_TERM;

	Term.contents.term.name = TermInterface.contents.term.name = padP->name;
	OH_ASSERT(octGetByName(&Chip,&Term));
	OH_ASSERT(octGetByName(&ChipInterface,&TermInterface));
  
	Direction.contents.prop.value.string = util_strsav(padP->direction);
	OH_ASSERT(octCreateOrModify(&TermInterface,&Direction));
	curpad->direction = util_strsav(padP->direction);

	TermType.contents.prop.value.string = util_strsav(padP->termType);
	OH_ASSERT(octCreateOrModify(&TermInterface,&TermType));
	curpad->termType = util_strsav(padP->termType);

	if (strcmp(padP->termType,"SUPPLY")==0 || strcmp(padP->termType,"GROUND")==0 ) {
	    PeakCurrent.contents.prop.value.real = padP->peakCurrent;
	    OH_ASSERT(octCreateOrModify(&TermInterface,&PeakCurrent));
	    curpad->peakCurrent = padP->peakCurrent;
	}


	if (strcmp(padP->floatingEdges,"")!=0) {
	    FloatingEdge.contents.prop.value.string = util_strsav(padP->floatingEdges);
	    OH_ASSERT(octCreateOrModify(&TermInterface,&FloatingEdge));
	}
	curpad->floatingEdges = util_strsav(padP->floatingEdges);
 
	if (padP->termEdge[0] != '\0' && strcmp(padP->termEdge, curpad->termEdge)!=0) {
	    TermEdge.contents.prop.value.string = util_strsav(padP->termEdge);
	    OH_ASSERT(octCreateOrModify(&TermInterface,&TermEdge));
	    curpad->termEdge = util_strsav(padP->termEdge);
	    curpad->flp_side = padP->flp_side;
	}
	if (padP->flp_relPos != curpad->flp_relPos) {
	    TermRelPosition.contents.prop.value.real = padP->flp_relPos;
	    OH_ASSERT(octCreateOrModify(&TermInterface,&TermRelPosition));
	    curpad->flp_relPos = padP->flp_relPos;
	}
	if (padP->row != curpad->row) {
	    ohCreateOrModifyPropInt(&TermInterface,&prop, "TERM_ROW", padP->row);
	    curpad->row = padP->row;
	}
    }
}

