#include "copyright.h"
#include "port.h"
#include "defs.h"

extern char  *stringSide();

static char ringbuffer[256];

typedef struct _ring {
    char* name;
    octObject net;
    octObject layer;
    int offset;			/* Apparent offset from edge of chip */
    int width;
    int coords[4];		/* Centerline coordinates. */
} Ring;    
      




octStatus
ohTerminalBBOnLayer(term, bb, layer)
octObject *term;
struct octBox *bb;
octObject *layer;
{
    if (octIdIsNull(term->contents.term.instanceId)) {
	return ohFormalTerminalBBOnLayer(term, bb,layer);
    } else {
	return ohActualTerminalBBOnLayer(term, bb, layer);
    }
}

octStatus
ohActualTerminalBBOnLayer(aterm, bb, layer)
octObject *aterm;
struct octBox *bb;
octObject *layer;
{
    octObject instance, fterm, box;
    octStatus status;

    /* Get the instance from the terminal */
    status = ohGetById(&instance, aterm->contents.term.instanceId);
    if (status != OCT_OK) return status;

    /* Get the corresponding formal terminal */
    status = ohFindFormal(&fterm, aterm);
    if (status < OCT_OK) return status;

    /* Get the formal terminal's BB */
    status = ohFormalTerminalBBOnLayer(&fterm, &box.contents.box, layer);
    if (status < OCT_OK) return status;

    /* transform the BB */
    box.type = OCT_BOX;
    octTransformGeo(&box, &instance.contents.instance.transform);
    *bb = box.contents.box;
    return OCT_OK;
}

#define MIN_ASSIGN(dst, src)	if ((src) < (dst)) (dst) = (src)
#define MAX_ASSIGN(dst, src)	if ((src) > (dst)) (dst) = (src)

octStatus
ohFormalTerminalBBOnLayer(fterm, bb, layer)
octObject *fterm;
struct octBox *bb;
octObject* layer;
{
    octStatus status;
    octGenerator implGen;
    octObject locallayer;
    octObject facet;
    octObject implementation;
    int count = 0;
    struct octBox tempBB;
    
    octGetFacet( fterm, &facet );

    if ( ohGetByLayerName( &facet, &locallayer, layer->contents.layer.name )!=OCT_OK ) {
	return OCT_NO_BB;
    }
    /* Run through all the implementing terminals and geometries */
    status = octInitGenContents(fterm, OCT_TERM_MASK|OCT_GEO_MASK, &implGen);
    
    if (status != OCT_OK) return status;
    while ((status = octGenerate(&implGen, &implementation)) == OCT_OK) {
	/* Find the BB of this implementation */
	if (implementation.type == OCT_TERM) {
	    status = ohTerminalBBOnLayer(&implementation, &tempBB, layer);
	} else {
	    if ( octIsAttached( &locallayer, &implementation ) != OCT_OK ) continue;
	    status = octBB(&implementation, &tempBB);
	}
	if (status == OCT_NO_BB) continue;
	if (status != OCT_OK) return status;

	/* If this is the first implementation, use its BB; else add it in */
	if (count++ == 0) {
	    *bb = tempBB;
	} else {
	    MIN_ASSIGN(bb->lowerLeft.x, tempBB.lowerLeft.x);
	    MIN_ASSIGN(bb->lowerLeft.y, tempBB.lowerLeft.y);
	    MAX_ASSIGN(bb->upperRight.x, tempBB.upperRight.x);
	    MAX_ASSIGN(bb->upperRight.y, tempBB.upperRight.y);
	}
    }
    if (status < OCT_OK) return status;

    if (count == 0) {
	return OCT_NO_BB;
    } else {
	return OCT_OK;
    }
}




/*
* getRingTerm
* Get the term in pad which is to be used for the routing of this ring.
* If there is more than one term, get the one which is closest
* to ref (which is also a terminal)
*/
int getRingTerm( ring, pad, ref, term ) 
    Ring* ring;
    Pad* pad;
    octObject* ref;		/* Reference term. Can be 0*/
    octObject* term;
{
    octObject instance;
    octObject tmpTerm;		/*  */
    octObject formal;
    octObject ringProp;
    octGenerator genTerm;
    int found = 0;
    int distance = 1000000;	/* Large number. */

    OH_ASSERT( ohGetById( &instance, pad->instanceId ) );
    
    OH_ASSERT( octInitGenContents( &instance, OCT_TERM_MASK, &genTerm ));
    while ( octGenerate( &genTerm, &tmpTerm ) == OCT_OK ) {
	OH_ASSERT( ohFindFormal( &formal, &tmpTerm ) );
	if ( ohGetByPropName( &formal, &ringProp, "RINGNAME") != OCT_OK ) continue;
	if ( propEqual( &ringProp , ring->name ) ) {
	    int d;
	    if ( ref == 0 ||  (d = termDistance( &tmpTerm, ref ) ) < distance  ) {
		distance = d;
		found++;
		ohGetById( term, tmpTerm.objectId );
		if ( verbose ) {
		    sprintf( ringbuffer, "%s",
			    ohFormatName( term ));
		    Message( ringbuffer  );
		}
	    }
	    if ( ref == 0 ) return 1;
	}
    }
    if ( found ) {
	return 1;
    } else {
	sprintf( ringbuffer, "%s has no terminal in ring %s",
		ohFormatName( &instance) , ring->name );
	Warning( ringbuffer );
	return 0;
    }
}

void setRingCoord( ring, side , p )
    Ring* ring;
    int side;
    struct octPoint* p;
{
    int c = ( side == LEFT || side == RIGHT )  ?  p->x : p->y ;
    if ( ring->coords[side] != c ) {
	ring->offset = ABS( c - ring->coords[side] );
	if (verbose) printf("Ring offset %s is %d\n", stringSide(side), ring->offset );
    }
    ring->coords[side] = c;
}

int connectRingTerminals( ring, side, term, isConnector )
    Ring* ring;
    int side ;
    octObject term[2];
    int isConnector[2];
{
    extern void boxCenter( /* box, point */ );
    int i;
    octObject path;
    struct octBox bb[2];
    struct octPoint p[2];
    
    if ( verbose ) {
	sprintf( ringbuffer, "Connecting %s to %s",
		ohFormatName( &term[0] ),
		ohFormatName( &term[1] ));
	Message( ringbuffer );
    }
    for ( i = 0; i < 2 ; i++ ) {
	if ( ohTerminalBBOnLayer( &term[i], &bb[i], &ring->layer ) != OCT_OK ) {
	    sprintf(ringbuffer,"%s has no BB on %s",
		    ohFormatName( &term[i]) , ohFormatName( &ring->layer ) );
	    Warning( ringbuffer );
	    return 0;
	}
	boxCenter( &bb[i], &p[i] );
	if ( ! isConnector[i] ) {
	    octAttach( &ring->net, &term[i] );
	}
    }
    
    setRingCoord( ring, side, &p[0] );
    
    ohCreateSegment( &ring->layer, &path, p[0].x, p[0].y, p[1].x, p[1].y, ring->width );
    octAttach( &ring->net, &path );
    octAttach( &path, &term[0] );
    octAttach( &path, &term[1] );
    return 1;
}

int connectRing( ring, side, pads )
    Ring* ring;
    int side ;
    Pad* pads[];
{
    octObject term[2];
    int isConnector[2];

    isConnector[0] = 0;
    isConnector[1] = 0;

    /* Get any term in pads[0] */
    getRingTerm( ring, pads[0], (octObject*)0, &term[0] );
    /* Get term in pads[1] closest to term[0] */
    getRingTerm( ring, pads[1], &term[0], &term[1] );
    /* Get term in pads[1] closest to term[0] */
    getRingTerm( ring, pads[0], &term[1], &term[0] );
    
    connectRingTerminals( ring, side , term, isConnector );
}


char* getRingCornerName( ring )
    Ring* ring;
{
    static char cornerName[256];
    int width = ring->width / LAMBDA;
    sprintf( cornerName, "ring%s.%dx%d",
	    ring->layer.contents.layer.name, width, width );
    return cornerName;
}

void  createRingCorner( ring )
    Ring* ring;
{
    octObject Corner;
    octObject CornerInterface;
    octStatus s;
    octObject layer;
    octObject box;
    octObject term;
    octObject prop;
    octCoord h = ring->width / 2;

    char* cornerName = getRingCornerName( ring );
    
    s = ohOpenFacet( &Corner, cornerName, "physical", "interface", "r" );
    if ( s >= OCT_OK )  {
	/* VOVinputFacet( &Corner ); */
	return;
    }
    VOVoutputFacet( &Corner );
    s = ohOpenFacet( &Corner, cornerName, "physical", "contents", "w" );
    if ( s != OCT_NEW_FACET ) PADPFatalError( "Cannot create corner" );
    
    if ( verbose ) Message( "Creating ring corner" );
    OH_ASSERT( ohCreatePropStr( &Corner, &prop, "VIEWTYPE", "PHYSICAL" ) );
    OH_ASSERT( ohCreatePropStr( &Corner, &prop, "TECHNOLOGY", "scmos" ) );
    OH_ASSERT( ohCreatePropStr( &Corner, &prop, "EDITSTYLE", "PHYSICAL" ) );
    
    OH_ASSERT( ohCreateLayer( &Corner, &layer, ring->layer.contents.layer.name ) );
    OH_ASSERT( ohCreateBox( &layer, &box, -h, -h , h, h ) );
    OH_ASSERT( ohCreateTerm( &Corner, &term , "t" ) );
    OH_ASSERT( octAttach( &term, &box ) );
    OH_ASSERT( octFlushFacet( &Corner ) );

    CornerInterface.type = OCT_FACET;
    CornerInterface.contents.facet = Corner.contents.facet;
    CornerInterface.contents.facet.facet = "interface";
    OH_ASSERT( octCopyFacet( &CornerInterface, &Corner ) );
    
    if ( verbose ) Message( "Corner created" );
}

int doRingCorners ( chip, chippads, ring )
    octObject* chip;
    ChipPads* chippads;
    Ring* ring;
{
    octObject corners[4];
    octObject cornerTerms[4];
    octObject connectors;
    int side, ix, iy;
    int row = 0;
    int padsep = 200;		/* Min separation of pads form core. */

    /* Message( "Putting ring corners" ); */
    
    createRingCorner( ring );
    OH_ASSERT( ohGetByBagName( chip, &connectors, "CONNECTORS" ) );
    
    /* First create the corners */
    for ( side = LEFT ; side <= BOTTOM ; side++ ) {
	if ( chippads->padsInEdge[row][side] == 0 ) {
	    if ( side == LEFT || side == BOTTOM ) {
		ring->coords[side] += ring->offset - chippads->maxPadHeight - padsep;
	    } else {
		ring->coords[side] -= ring->offset - chippads->maxPadHeight - padsep;
	    }
	}
    }
    for ( side = LEFT ; side <= BOTTOM ; side++ ) {
	corners[side].type = OCT_INSTANCE;
	
	corners[side].contents.instance.master = getRingCornerName( ring );
	corners[side].contents.instance.view = "physical";
	corners[side].contents.instance.facet = "interface";
	corners[side].contents.instance.name = "corner";
	corners[side].contents.instance.version = OCT_CURRENT_VERSION;
    
	corners[side].contents.instance.transform.transformType = OCT_NO_TRANSFORM;
	ix = ( side == LEFT || side == TOP ) ? LEFT : RIGHT;
	iy = ( side == LEFT || side == BOTTOM ) ? BOTTOM : TOP;
	corners[side].contents.instance.transform.translation.x = ring->coords[ ix ];
	corners[side].contents.instance.transform.translation.y = ring->coords[ iy ];
    
	OH_ASSERT( octCreate( chip, &corners[side] ) );
	OH_ASSERT( octAttach( &connectors, &corners[side] ) );
	OH_ASSERT( octGenFirstContent( &corners[side], OCT_TERM_MASK, &cornerTerms[side] ) );
    }

    /* Now connect them. */

    for ( side = LEFT ; side <= BOTTOM ; side++ ) {
	octObject terms[2];
	int isConnector[2];
	Pad* pad1;
	Pad* pad2;
	if ( chippads->padsInEdge[row][side] == 0 ) {
	    terms[0] = cornerTerms[side];
	    terms[1] = cornerTerms[(side+1)%4];
	    isConnector[0] = 1;
	    isConnector[1] = 1;
	    connectRingTerminals( ring, side, terms, isConnector );
	} else {
	    if ( side == LEFT || side == TOP ) {
		pad1 = chippads->FirstPadInEdge[row][side];
		pad2 = chippads->LastPadInEdge[row][side];
	    } else {
		pad1 = chippads->LastPadInEdge[row][side];
		pad2 = chippads->FirstPadInEdge[row][side];
	    }
	    terms[0] = cornerTerms[side];
	    if ( getRingTerm( ring, pad1, &terms[0], &terms[1] ) ) {
		isConnector[0] = 1;
		isConnector[1] = 0;
		connectRingTerminals( ring, side, terms, isConnector );
	    }
	    terms[0] = cornerTerms[(side+1)%4];
	    if ( getRingTerm( ring, pad2, &terms[0], &terms[1] ) ) {
		isConnector[0] = 1;
		isConnector[1] = 0;
		connectRingTerminals( ring, side, terms, isConnector );
	    }
	}
    }    
}

int doRing( chip, chippads, ring )
    octObject* chip;
    ChipPads* chippads;
    Ring* ring;
{
    int side;
    Pad* pad;
    int row = 0;
    int count;

    if ( verbose ) {
	(void)sprintf( ringbuffer, "Routing ring %s on %s width %d lambdas", 
		  ring->net.contents.net.name, 
		  ring->layer.contents.layer.name, 
		  ring->width / LAMBDA );
	Message( ringbuffer );
    }
    
    OH_ASSERT( octGetOrCreate( chip, &ring->net ) );
    
    
    ring->coords[LEFT]  = chippads->chipArea.lowerLeft.x;
    ring->coords[RIGHT] = chippads->chipArea.upperRight.x;
    ring->coords[BOTTOM]= chippads->chipArea.lowerLeft.y;
    ring->coords[TOP]   = chippads->chipArea.upperRight.y;
    
    if ( verbose ) {
	for ( side = LEFT ; side <= BOTTOM; side++ ) {
	    sprintf( ringbuffer, "Initial ring coords: %s %d", 
		    stringSide( side ) , ring->coords[side] );
	    Message( ringbuffer );
	}
    }

    /* Route the pads along each side. */
    for ( side = LEFT ; side <= BOTTOM ; side++ ) {
	count = 0;
	for ( pad = chippads->FirstPadInEdge[row][side] ; pad != 0 ; pad = pad->nextInEdge ) {
	    Pad* pads[2];

	    count++;
	    pads[0] = pad;
	    pads[1] = pad->nextInEdge;
	    if ( pads[1] != 0 ) {
		connectRing( ring, side, pads );
	    } else if ( count == 1 ) {
		octObject term;
		struct octBox bb;
		struct octPoint p;
		getRingTerm( ring, pads[0], (octObject*)0, &term );
		ohTerminalBBOnLayer( &term, &bb, &ring->layer );
		boxCenter( &bb, &p );
		setRingCoord( ring, side, &p );
	    }
	}
	if (verbose) {
	    sprintf( ringbuffer, "%d pads on %s side\n", count, stringSide( side ) );
	    Message( ringbuffer );
	}
    }
    
    /* now do corners */
    doRingCorners ( chip, chippads, ring );
    return 1;
}

void PADPdoRings( chip, chippads )
    octObject* chip;
    ChipPads* chippads;
{
    octGenerator genTerm;
    octGenerator genBox;
    octObject box;
    octObject instance;
    octObject master;
    octObject term;
    octObject ringProp;
    Pad* pad;
    Ring ring;
    

    pad = chippads->FirstPad;
    OH_ASSERT( ohGetById( &instance, pad->instanceId ));
    OH_ASSERT( ohOpenMaster( &master, &instance, "interface", "r" ));
    octInitGenContents( &master, OCT_TERM_MASK, &genTerm );
    while ( octGenerate( &genTerm, &term ) == OCT_OK ) {
	if ( ohGetByPropName( &term, &ringProp, "RINGNAME") == OCT_OK ) {
	    ring.name =  ringProp.contents.prop.value.string;
	    ring.net.type = OCT_NET;
	    ring.net.contents.net.name = ring.name;
	    if ( octGetByName( chip, &ring.net ) == OCT_OK ) {
		sprintf( ringbuffer, "Net %s exists already. Cannot route it as a ring.",
			ring.name );
		Warning( ringbuffer );
	    } else {
		OH_ASSERT( octCreate( chip, &ring.net ) );
		OH_ASSERT( octInitGenContents( &term, OCT_BOX_MASK, &genBox ));
		while ( octGenerate( &genBox, &box ) == OCT_OK ) {
		    octGenFirstContainer( &box, OCT_LAYER_MASK, &ring.layer );
		    ring.width = box_height( &box.contents.box );
		    OH_ASSERT( octGetOrCreate( chip, &ring.layer ) );
		    doRing( chip, chippads, &ring );
		}
	    }
	}
    }
    OH_ASSERT( octFlushFacet( chip ) );
}
