#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"
#include "padp.h"

extern char  *stringSide();
extern octTransformType putPinsInside();

void padPartitions( chippads, instId, side, row, block)
    ChipPads*   chippads;
    octId	instId;
    int		side;
    int		row;
    int		block;
{
    octObject		Instance;
    static octObject	PartitionBag;
    static octObject	NameProp;
    static octObject	PartitionType;
    static int		initFlag = FALSE;
    static octId	*bagArray;
    octObject		Object;
    char		name[256];
    int			index;
    static int		arraySize;

    Instance.objectId = instId;
    if (octGetById(&Instance) != OCT_OK) {
	Warning("Wrong argument instId in padPartitions. Should never happen");
	return;
    }

    if (!initFlag) {
	initFlag = TRUE;

	PartitionBag.type = OCT_BAG;
	PartitionBag.contents.bag.name = "PARTITION_PADP";
	NameProp.type = OCT_PROP;
	NameProp.contents.prop.name = "PARTITION_NAME";
	NameProp.contents.prop.type = OCT_STRING;
	PartitionType.type = OCT_PROP;
	PartitionType.contents.prop.name = "PARTITION_TYPE";
	PartitionType.contents.prop.type = OCT_STRING;
    
	arraySize =  4 * DEF_MAX_ROWS * chippads->numberOfPads ;
	bagArray = ALLOC(octId , arraySize);
    
	for (index = 0 ; index < arraySize ; index++)
	  bagArray[index] = oct_null_id;

	while (octGetByName(&Chip,&PartitionBag)==OCT_OK) {
	    OH_ASSERT(octDelete(&PartitionBag));
	}
    }


    index = block * 4 * DEF_MAX_ROWS + row * 4 + side;
    if (index >= arraySize)
      PADPFatalError("Array size is too small");

    /* If this is the first instance, store its id */
    /* else , create a bag */
    if (octIdIsNull(bagArray[index])) {
	bagArray[index] = instId;
    } else {
	Object.objectId = bagArray[index];
	OH_ASSERT(octGetById(&Object));
	if (Object.type == OCT_BAG) {
	    OH_ASSERT(octAttach(&Object,&Instance));
	} else if (Object.type == OCT_INSTANCE) {
	    PartitionType.contents.prop.value.string = "HARD";
	    (void) sprintf(name, "PAD_BLOCKS/PADS_%s.%d.%d", stringSide(side), row, block);
	    NameProp.contents.prop.value.string = name;
	    OH_ASSERT(octCreate(&Chip, &PartitionBag));
	    OH_ASSERT(octCreate(&PartitionBag,&NameProp));
	    OH_ASSERT(octCreate(&PartitionBag,&PartitionType));
	    OH_ASSERT(octAttach(&PartitionBag,&Object));
	    OH_ASSERT(octAttach(&PartitionBag,&Instance));
	    bagArray[index] = PartitionBag.objectId;
	} else {
	    PADPFatalError("padPartition: unknown object in array");
	}
    }
}

int getDeadCorner( chippads )
    ChipPads* chippads;
{
    int padHeight = chippads->maxPadHeight;
    return MAX( padHeight * 1.2 , padHeight + 100) ;
}

void PADPspace( chippads, padLimited, exactPosition )
    ChipPads* chippads;
    int padLimited;
    int exactPosition;
{
    int	side;			/* To do one side of the chip at a time. */
    int	positionInSide;		/*  */
    Pad	*padP;
    int sidelength[4];
    int	row;
    int	rowOffset;
    int blockOffset;
    int	width;

    struct octBox area;
    int i;

    int	deadCorner = getDeadCorner( chippads );

    if (verbose) Message("Spacing pads");

    area = chippads->chipArea;
    for (side = LEFT; side <= BOTTOM ; side++) {
	sidelength[side] = chippads->availableSpaceSide[side]; /* Initialize */
    }


    /* If padlimited determine chip size from pads. */
    if ( padLimited ) {
	for (side = LEFT; side <= BOTTOM ; side++) {
	    sidelength[side]  = 0;
	    for (row = 0 ; row < DEF_MAX_ROWS ; row++) {
		/* Width required by this row */
		if ( chippads->padsInEdge[row][side] > 0 ) {
		    width = chippads->totalWidth[row][side] + row * 2 * (deadCorner + DEF_ROW_SPACING) ;
		    if (row >= 1) width += chippads->padsInEdge[row][side] / DEF_BLOCK_SIZE * DEF_BLOCK_SPACING;
		    sidelength[side]  = MAX(sidelength[side] , width);
		}
	    }
	}
	chippads->width = MAX(sidelength[TOP],sidelength[BOTTOM]) + 2 * deadCorner ;
	chippads->height = MAX(sidelength[LEFT],sidelength[RIGHT]) + 2 * deadCorner ;

	sidelength[LEFT]  = chippads->height;
	sidelength[RIGHT] = chippads->height;
	sidelength[TOP]   = chippads->width;
	sidelength[BOTTOM]= chippads->width;

	area.upperRight.x = area.lowerLeft.x + chippads->width;
	area.upperRight.y = area.lowerLeft.y + chippads->height;
	chippads->chipArea = area;

	if (verbose) {
	    printf("PAD LIMITED CHIP: width = %d, height = %d\n", chippads->width, chippads->height );
	}
    }

    for (side=LEFT ; side <=BOTTOM ; side++) {
	for (row = 0 ; row < DEF_MAX_ROWS ; row++) {
	    int blockCounter = 0;
	    int padsInBlock = 0;
	    int spacing = 0;
	    /*
	     * Determine the spacing between pads 
	     */
	    if ( !exactPosition && !padLimited && chippads->padsInEdge[row][side] > 1) {
		width = sidelength[side] - 2 * deadCorner -
		  2 * row * (deadCorner + DEF_ROW_SPACING)
		    - chippads->totalWidth[row][side];
		if (row >= 1) {
		    width -= chippads->padsInEdge[row][side] / DEF_BLOCK_SIZE * DEF_BLOCK_SPACING ;
		}
		spacing = width / ( chippads->padsInEdge[row][side] - 1);
		if (spacing < 0) {
		    (void) sprintf(buf,"Not enough space for row %d in side %s", row, stringSide(side));
		    Warning(buf);
		    spacing = 0;
		}
	    }

	    if ( chippads->padsInEdge[row][side] == 1) {
		positionInSide = ( sidelength[side] - chippads->totalWidth[row][side] )/2;
	    } else {
		positionInSide = deadCorner;
	    }
      
	    rowOffset = row * (DEF_ROW_SPACING + chippads->maxPadHeight);
	    blockOffset = (row > 0) ? (deadCorner + DEF_ROW_SPACING) * row : 0;
	    for (padP=chippads->FirstPadInEdge[row][side] ; padP!= NULL ; padP=padP->nextInEdge) {

		if ( exactPosition ) {
		    padP->positionInSide = sidelength[side] * padP->flp_relPos;
		} else {
		    padP->positionInSide = positionInSide + blockOffset;
		    positionInSide += spacing + padP->width;
		}

		switch (side) {
		case LEFT:
		    padP->targetLowerLeft.x = area.lowerLeft.x + rowOffset;
		    padP->targetLowerLeft.y = area.lowerLeft.y + padP->positionInSide;
		    break;
		case TOP:
		    padP->targetLowerLeft.x = area.lowerLeft.x + padP->positionInSide;
		    padP->targetLowerLeft.y = area.upperRight.y - rowOffset - padP->height;
		    break;
		case RIGHT:
		    padP->targetLowerLeft.x = area.upperRight.x - rowOffset - padP->height;
		    padP->targetLowerLeft.y = area.lowerLeft.y +  padP->positionInSide;
		    break;
		case BOTTOM:
		    padP->targetLowerLeft.x = area.lowerLeft.x + padP->positionInSide;
		    padP->targetLowerLeft.y = area.lowerLeft.y + rowOffset;
		    break;
		default:
		    (void) sprintf(buf,"Wrong side (%d) for pad %s",side,padP->name);
		    PADPFatalError(buf);
		}

		if ( !octIdIsNull( padP->instanceId ) ) {
		    padPartitions( chippads, padP->instanceId, side, row, blockCounter);
		}
		padP->transform = putPinsInside(padP->internalEdgeOfMaster,padP->side);

		if (++padsInBlock >= DEF_BLOCK_SIZE) {
		    padsInBlock = 0;
		    if (row > 0) {
			blockCounter++;
			blockOffset += DEF_BLOCK_SPACING;
		    }
		}
	    }
	}
    }
}



