#include "copyright.h"
#include "port.h"
#include "defs.h"
#include "octPlacement.h"
#include "st.h"
/*
 * parseEdgeName
 *
 * given an edge name, return the edge
 */
int
parseEdgeName(name)
    char *name;
{
  if (strcmp(name,"LEFT")==0) {
    return(LEFT);
  } else   if (strcmp(name,"TOP")==0) {
    return(TOP);
  } else   if (strcmp(name,"RIGHT")==0) {
    return(RIGHT);
  } else   if (strcmp(name,"BOTTOM")==0) {
    return(BOTTOM);
  } else {
    return(NO_SIDE);
  }
}
/*
 * stringSide
 *
 * given a side return the edge name
 */
char *
stringSide(side)
     int side;
{
  switch (side) {
  case LEFT:
    return("LEFT");
  case TOP:
    return("TOP");
  case RIGHT:
    return("RIGHT");
  case BOTTOM:
    return("BOTTOM");
  default:
    return("*****");
  }
}
/*
 * putPinsInside
 *
 * return a transformType such that the internalEdgeOfMaster 
 * is on the correct side of the chip
 */
octTransformType
putPinsInside(internalEdgeOfMaster, side)
     int internalEdgeOfMaster, side;
{
  switch (internalEdgeOfMaster) {
  case LEFT:
    switch (side) {
    case LEFT:
      return(OCT_ROT180);
    case TOP:
      return(OCT_ROT90);
    case RIGHT:
      return(OCT_NO_TRANSFORM);
    case BOTTOM:
      return(OCT_ROT270);
    default:
      PADPFatalError("in putpinsinside");
    }
  case TOP:
    switch (side) {
    case LEFT:
      return(OCT_ROT270);
    case TOP:
      return(OCT_ROT180);
    case RIGHT:
      return(OCT_ROT90);
    case BOTTOM:
      return(OCT_NO_TRANSFORM);
    default:
      PADPFatalError("in putpinsinside");
    }
  case RIGHT:
    switch (side) {
    case LEFT:
      return(OCT_NO_TRANSFORM);
    case TOP:
      return(OCT_ROT270);
    case RIGHT:
      return(OCT_ROT180);
    case BOTTOM:
      return(OCT_ROT90);
    default:
      PADPFatalError("Wrong edge");
    }
  case BOTTOM:
    switch (side) {
    case LEFT:
      return(OCT_ROT90);
    case TOP:
      return(OCT_NO_TRANSFORM);
    case RIGHT:
      return(OCT_ROT270);
    case BOTTOM:
      return(OCT_ROT180);
    default:
      PADPFatalError("Wrong edge");
    }
  default:
    PADPFatalError("Wrong side");
  }
  return(OCT_NO_TRANSFORM);	/* make lint happy, but this will never happen */
}


octCoord 
box_width( b )
    struct octBox *b;
{
    return b->upperRight.x - b->lowerLeft.x;
}

octCoord 
box_height( b )
    struct octBox *b;
{
    return b->upperRight.y - b->lowerLeft.y;
}

int box_area( b, scale )
    struct octBox *b;
    int scale;
{
    return ( box_width( b ) / scale)  * ( box_height( b ) / scale );
}

void boxCenter( b, c )
    struct octBox* b;
    struct octPoint* c;
{
    c->x = ( b->lowerLeft.x + b->upperRight.x ) >> 1;
    c->y = ( b->lowerLeft.y + b->upperRight.y ) >> 1;
}

int termDistance( t1, t2 )
    octObject* t1;
    octObject* t2;
{
    struct octBox bb1;
    struct octBox bb2;
    struct octPoint p1;
    struct octPoint p2;

    OH_ASSERT( ohTerminalBB( t1, &bb1 ) );
    OH_ASSERT( ohTerminalBB( t2, &bb2 ) );
    boxCenter( &bb1, &p1 );
    boxCenter( &bb2, &p2 );
    
    return  ABS( p1.x - p2.x ) + ABS( p1.y - p2.y );
}

int propEqual( p, s )
    octObject* p;
    char* s;
{
    if ( p->type != OCT_PROP ) PADPFatalError("not a prop");
    if ( p->contents.prop.type != OCT_STRING ) PADPFatalError("not a string prop");
    return ( strcmp( p->contents.prop.value.string, s ) == 0);
}
