/*
 *  FILE: analogCost.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Oct.30.1990          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  int         curMatchingCost(costP)
 *  int         curSymmetryCost(costP) 
 *  int         MoveMatchingCost(costP)
 *  int         MoveSymmetryCost(costP)
 *  int         SymmetryToAxis(center1,center2, flag1, flag2)
 *  int         computeTotalSymmetryToAxes(SymCell,costP)  
 *      
 *  IMPORTED    FUNCTIONS:
 *  
 *    
 */


/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */

#include "defs.h"
#include "sa.h"
#include "global.h"
#include "analogVariables.h"

/* IMPORTED VARIABLES */
extern int acceptedFlag;  /* <--- sa.c */
extern int inconsistentError;  /* <--- move.c */

/* GLOBAL VARIABLES OF THIS MODULE ("analogCost.c") */
double matchDistance;
double matchOrienta;

/* DECLARED ROUTINES USED IN THIS FILE THAT DO NOT RETURN INTEGERS */

double
computeMoveMatching();

double
checkOrientation();

double
computeTotalMatchingCost();

double
computeTotalSymmetryToAxis();

double
symmetryToAxis();

double
checkAstigmatism();

MacroCell
*computeMoveSymmetryToAxis();

MacroCell
*searchOtherSymmetricCell();

/* GLOBAL CONSTANTS FOR "analogCost.c" */
int astigmatism = 500;            /* PENALTY FOR NON MIRRORED CELLS */
int singleCellDisplacement = 100; /* PENALTY FOR SINGLE CELLS FAR FROM
				   * SYMMETRY-AXIS */
int alignment = 40;               /* PENALTY FOR SYMMETRIC CELLS WHICH
				   * ARE MISALIGNED */
int positionInWrongWell = 500;    /* PENALTY FOR A CELL WHICH IS IN THE
				   * WRONG WELL */
static int       curFlag = FALSE;
static int       newFlag = TRUE;

/* IMPORTED FUNCTIONS */
extern int distance();   /* <-- "costWL.c */ 


/* FUNCTIONS */

/* procedure that scans the database gMatchCell and computes the current
 * cost. Notice the recursion inside it */

int
CurMatchingCost(costP)
     Cost *costP;

{
  octId cellToMove = 0;
  octId cell2ToMove = 0;

  if (!considerMatching || SM_matchWeight <= 0.0)  {
    costP->matching = 0.0;
    return OK;
  }
  costP->matching = computeTotalMatchingCost(gMatchCell->matchFirstCluster, cellToMove,
                    cell2ToMove, curFlag);  
}


double
computeTotalMatchingCost(MatchCell, cellToMoveP, cell2ToMoveP , flagP) 
    struct MatchingCluster *MatchCell;
    octId                 cellToMoveP;
    octId                cell2ToMoveP;
    int                         flagP;

{
  struct MatchingCluster *Cell;
  int clusterFirstIndex = 0;

  matchDistance = 0;
  matchOrienta = 0;
  Cell = MatchCell;
  computeInnerLoop(Cell, clusterFirstIndex, cellToMoveP, cell2ToMoveP, flagP);
  return matchDistance + matchOrienta;
}

 
int
computeInnerLoop(cluster, clusterIndex, MoveCell, MoveCell2, flag)
   int                clusterIndex;
   struct MatchingCluster *cluster;
   octId                  MoveCell;
   int                        flag;     /* this flag sets the computation of the cost to
                                         * current (FALSE) or new (TRUE) parameters for
                                         * the cell MoveCell */

{
  int  cellIndex;                         /* index */
  struct MatchingCluster          *child; /* visited branches */ 
  MacroCell                *cell, *cell2;
  MacroCell      *otherCell, *otherCell2;
  octPoint               center, center2;
  octPoint     otherCenter, otherCenter2;
  octTransformType transform, transform2; 
  AnalogStatus           status, status2; 
  octCoord                  axisPosition;
  SymmetryType                  axisType;
  int         singleCellInCouple = FALSE;
  int        singleCellInCouple2 = FALSE;
        

  while (cluster != NULL)  {
    cellIndex = 0;
    cell = cluster->MatchingFirstCell; 
    if (cell != NULL) 
      clusterIndex++;
    while (cell != NULL)  {
      cell2 = cell->matchNextCell;
      while (cell2 != NULL)  {
        if (!flag) {          /* current parameters */
          matchDistance += (double)RECT(distance(cell->matchLink->curCenter.x,
                           cell->matchLink->curCenter.y, cell2->matchLink->curCenter.x,
                           cell2->matchLink->curCenter.y) - cluster->ClusterLimitDistance);

	  if (cell->matchLink->analogStatus == MATCHED ||
	      cell->matchLink->analogStatus == MULTIPLE_MATCHED ||
	      cell2->matchLink->analogStatus == MATCHED ||
	      cell2->matchLink->analogStatus == MULTIPLE_MATCHED) {
            matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                             cell2->matchLink->curTransform);
	  } else {
	    /* (cell->matchLink->analogStatus == BOTH ||
	     *  cell->matchLink->analogStatus == BOTH_MULTIPLE) &&
	     * (cell2->matchLink->analogStatus == BOTH ||
	     *  cell2->matchLink->analogStatus == BOTH_MULTIPLE) */
 
	    /* analyze the situation when both cells are symmetric wrt a vertical axis */
	    if (cell->matchLink->ySymPointer != NULL && cell2->matchLink->ySymPointer != NULL) {	 
	      axisType = VERTICAL;
	      otherCell = searchOtherSymmetricCell (cell->matchLink, &axisPosition, 
						    &axisType, &singleCellInCouple);
	      otherCell2 = searchOtherSymmetricCell (cell2->matchLink, &axisPosition, 
						     &axisType, &singleCellInCouple2);
	  
	      if (singleCellInCouple && singleCellInCouple2) {
	        matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                                 cell2->matchLink->curTransform);
	     
	      } else {
	      
		if (!singleCellInCouple && !singleCellInCouple2) {
	  	  if (  (cell->matchLink->curCenter.x < otherCell->curCenter.x &&
		        cell2->matchLink->curCenter.x < otherCell2->curCenter.x) ||
		        (cell->matchLink->curCenter.x > otherCell->curCenter.x &&
		        cell2->matchLink->curCenter.x > otherCell2->curCenter.x)  ) {
	          /* both cells are at the same side of their own axes */
	          matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                                   cell2->matchLink->curTransform);
	          }
	        }
	      }
	    } else {
	      /* analyze the situation when both cells are symmetric wrt a horizontal axis */
	      if (cell->matchLink->xSymPointer != NULL && cell2->matchLink->xSymPointer != NULL) {	 
	        axisType = HORIZONTAL;
	        otherCell = searchOtherSymmetricCell (cell->matchLink, &axisPosition, 
						      &axisType, &singleCellInCouple);
	        otherCell2 = searchOtherSymmetricCell (cell2->matchLink, &axisPosition, 
						       &axisType, &singleCellInCouple2);

	        if (singleCellInCouple && singleCellInCouple2) {
	          matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                                   cell2->matchLink->curTransform);

	        } else {

	          if (!singleCellInCouple && !singleCellInCouple2) {
	  	    if (  (cell->matchLink->curCenter.y > otherCell->curCenter.y &&
		          cell2->matchLink->curCenter.y > otherCell2->curCenter.y) ||
		          (cell->matchLink->curCenter.y < otherCell->curCenter.y &&
		          cell2->matchLink->curCenter.y < otherCell2->curCenter.y)  ) {
	            /* both cells are at the same side of their own axes */
	            matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                                     cell2->matchLink->curTransform);
	            }
	          }
	        }
	      } else {
	        /* the cells have different types of symmetry */
	        matchOrienta += checkOrientation(cell->matchLink->curTransform,
                                                 cell2->matchLink->curTransform);
	      }	
	    }
	  }  
	} else {
	  /* the flag is TRUE and therefore the same calculations are repeated assuming
	   * that the cells are to be moved */ 
          center = cell->matchLink->curCenter;
          center2 = cell2->matchLink->curCenter;
          transform = cell->matchLink->curTransform;
          transform2 = cell2->matchLink->curTransform;
	  status = cell->matchLink->analogStatus;
	  status2 = cell2->matchLink->analogStatus;

          if (MoveCell != 0)  {
            if (MoveCell == cell->matchLink->cellId)  {
              center = cell->matchLink->newCenter;
              transform = cell->matchLink->newTransform;
            } else {
	      if (MoveCell == cell2->matchLink->cellId)  {
                center2 = cell2->matchLink->newCenter;
                transform2 = cell2->matchLink->newTransform;
	      }	
	    }
          }
          if (MoveCell2 !=0)  {
            if (MoveCell2 == cell->matchLink->cellId)  {
              center = cell->matchLink->newCenter;
              transform = cell->matchLink->newTransform;
            } else {
              if (MoveCell2 == cell2->matchLink->cellId)  { 
                center2 = cell2->matchLink->newCenter;
                transform2 = cell2->matchLink->newTransform;
	      }
	    }
          }          

#ifdef OBSOLETE_TO_BE_DELETED
sprintf(errmsg,"Cluster %d, cell %d (X,Y) %d,%d, curtran %d, cell2 %d (X,Y) %d,%d, curtran %d, cluster limit distance %d ", clusterIndex, cell->matchLink->cellId, center.x, center.y, transform, cell2->matchLink->cellId, center2.x, center2.y, transform2, cluster->ClusterLimitDistance);
Message(errmsg);
#endif
         
          matchDistance += (double)RECT(distance(center.x, center.y, center2.x,
                           center2.y) - cluster->ClusterLimitDistance);

	  if (status == MATCHED || status == MULTIPLE_MATCHED ||
	      status2 == MATCHED || status2 == MULTIPLE_MATCHED) {
            matchOrienta += checkOrientation(transform,
                                             transform2);
	  } else {
	    /* (status == BOTH || status == BOTH_MULTIPLE) &&
	     * (status2 == BOTH || status2 == BOTH_MULTIPLE) */
 
	    /* analyze the situation when both cells are symmetric wrt a vertical axis */
	    if (cell->matchLink->ySymPointer != NULL && cell2->matchLink->ySymPointer != NULL) {	 
	      axisType = VERTICAL;
	      otherCell = searchOtherSymmetricCell (cell->matchLink, &axisPosition, 
						    &axisType, &singleCellInCouple);
	  
	      otherCell2 = searchOtherSymmetricCell (cell2->matchLink, &axisPosition, 
						     &axisType, &singleCellInCouple2);
	     
	      if (singleCellInCouple && singleCellInCouple2) {
	        matchOrienta += checkOrientation(transform,
                                                 transform2);
	      
	      } else {
	      
		if (!singleCellInCouple && !singleCellInCouple2) {
		  /* check if the selected cell is one of the other two couple cells */  
		  if (MoveCell == otherCell->cellId || MoveCell2 == otherCell->cellId) {
		    otherCenter = otherCell->newCenter;
		  } else {
		    otherCenter = otherCell->curCenter;
		  }
		  if (MoveCell == otherCell2->cellId || MoveCell2 == otherCell2->cellId) {
		    otherCenter2 = otherCell2->newCenter;
		  } else {
		    otherCenter2 = otherCell2->curCenter;
		  }
		  /* end check */
	  	  if (  (center.x < otherCenter.x &&
		        center2.x < otherCenter2.x) ||
		        (center.x > otherCenter.x &&
		        center2.x > otherCenter2.x)  ) {
	          /* both cell and cell2 are at the same side of their own axes */
	          matchOrienta += checkOrientation(transform,
                                                   transform2);
	          }
	        }
	      }
	    } else {
	      /* analyze the situation when both cells are symmetric wrt a horizontal axis */
	      if (cell->matchLink->xSymPointer != NULL && cell2->matchLink->xSymPointer != NULL) {	 
	        axisType = HORIZONTAL;
	        otherCell = searchOtherSymmetricCell (cell->matchLink, &axisPosition, 
						      &axisType, &singleCellInCouple);
	      
		otherCell2 = searchOtherSymmetricCell (cell2->matchLink, &axisPosition, 
						       &axisType, &singleCellInCouple2);

	        if (singleCellInCouple && singleCellInCouple2) {
	          matchOrienta += checkOrientation(transform,
                                                   transform2);

	        } else {

	          if (!singleCellInCouple && !singleCellInCouple2) {

		    /* check if the selected cell is one of the other two couple cells */  
		    if (MoveCell == otherCell->cellId || MoveCell2 == otherCell->cellId) {
		      otherCenter = otherCell->newCenter;
		    } else {
		      otherCenter = otherCell->curCenter;
		    }
		    if (MoveCell == otherCell2->cellId || MoveCell2 == otherCell2->cellId) {
		      otherCenter2 = otherCell2->newCenter;
		    } else {
		      otherCenter2 = otherCell2->curCenter;
		    }
		    /* end check */
	
	  	    if (  (center.y > otherCenter.y && 
		          center2.y > otherCenter2.y) ||
		          (center.y < otherCenter.y &&
		          center2.y < otherCenter2.y)  ) {
	            /* both cell and cell2 are at the same side of their own axes */
	            matchOrienta += checkOrientation(transform,
                                                     transform2);
	            }
	          }
	        }
	      } else {
	        /* the cells have different types of symmetry */
	        matchOrienta += checkOrientation(transform,
                                                 transform2);
	      }	
	    }
	  }  
	}
        cell2 = cell2->matchNextCell;
      }
      cell = cell->matchNextCell;
    }
    child = cluster->Cluster;
    clusterIndex = computeInnerLoop(child, clusterIndex, MoveCell, MoveCell2, flag);
    cluster = cluster->matchNextCluster;
  }
  return clusterIndex;
}


/* This procedure calculates the current cost wrt the symmetry constraints  */

int
CurSymmetryCost(costP)
     Cost *costP;

{
  double                totalDistanceFromY;
  double                totalDistanceFromX;

  if (!considerSymmetry || SM_symWeight <= 0.0)  {
    costP->symmetry = 0.0;
    return OK;
  }
  costP->symmetry = 0.0;
  totalDistanceFromY = computeTotalSymmetryToAxes(gySymCell);
  totalDistanceFromX = computeTotalSymmetryToAxes(gxSymCell);
  costP->symmetry = totalDistanceFromY + totalDistanceFromX;
  return OK;
}



computeTotalSymmetryToAxes(SymCell)
  SymmetricCell  *SymCell;      /* data-base that represents symmetric cells  */  

{
  double                 totalDistance = 0.0; 
  double                 distance = 0.0;                /* temporary distance */   
  double                 astigma = 0.0;             /* temporary astigmatism */
  int                    m,l;                           /* indices */   
  int                    numberOfAxes;
  SymmetryType           typeOfAxis;
  int                    numberOfCouples;
  octCoord               positionOfAxis;
  struct SymmetryAxis    element; 
  octPoint               centerCell1,centerCell2;
  octTransformType       transform1, transform2;
  int                    singleCellInCouple = FALSE;    
  
  numberOfAxes = SymCell->numberOfAxis;
  typeOfAxis = SymCell->axisType;
  for (m = 0; m < numberOfAxes; m++)  {
    element = SymCell->Axis[m];
    numberOfCouples = element.numberOfCouple;
    positionOfAxis = element.axisPosition;
    for (l = 0; l < numberOfCouples; l++)  {
      singleCellInCouple = FALSE;
      centerCell1 = element.couple[l].cell1->curCenter;
      transform1 = element.couple[l].cell1->curTransform;
      if (element.couple[l].cell2 == NULL)  {
        singleCellInCouple = TRUE;
      } else {
        centerCell2 = element.couple[l].cell2->curCenter;
        transform2 = element.couple[l].cell2->curTransform;
      }
      distance = symmetryToAxis (centerCell1, centerCell2, positionOfAxis, typeOfAxis, singleCellInCouple);
      astigma = checkAstigmatism (transform1, transform2, typeOfAxis, singleCellInCouple);
      totalDistance += (distance + astigma);

#ifdef OBSOLETE_TO_BE_DELETED
sprintf(errmsg,"C1 %d, curtran %d, C2 %d, curtran %d, astigma %f",SymCell->Axis[m].couple[l].cell1->cellId, SymCell->Axis[m].couple[l].cell1->curTransform, SymCell->Axis[m].couple[l].cell2->cellId, SymCell->Axis[m].couple[l].cell2->curTransform, astigma);
Message(errmsg);
#endif
   
    }
  }
  return totalDistance;
}


/* This routine calculates the cost of placement of a couple.
 * It is assumed that the cells in the couple are IDENTICAL. */  

double
symmetryToAxis (center1, center2, axis, type, singleFlag)
  octPoint  center1;
  octPoint  center2;
  octCoord     axis;
  SymmetryType type;
  int        singleFlag;

{
  double       verticalMis;            /* vertical distance-mismatch of cells */
  double       horizontalMis;          /* horizontal distance-mismatch of cells */
 
  if (singleFlag) {
    if (type == VERTICAL) {
      verticalMis = ABS(center1.x - axis) * singleCellDisplacement;
      horizontalMis = 0;  
    } else {
      horizontalMis = ABS(center1.y - axis) * singleCellDisplacement;
      verticalMis = 0;   
    }
  } else {    
    if (type == VERTICAL) {
      horizontalMis = ABS(center1.x + center2.x - 2 * axis);
      verticalMis = ABS(center1.y - center2.y) * alignment;
    } else {
      verticalMis = ABS(center1.y + center2.y - 2 * axis);
      horizontalMis = ABS(center1.x - center2.x) * alignment;
    }
  } 
  return verticalMis + horizontalMis;
}  


/* This procedure checks the orientation of cells */


double
checkOrientation(transform1, transform2)
     octTransformType  transform1;
     octTransformType  transform2;
   
{
  /* the parameter "astigmatism" is used here as well as in checkAstigmatism 
   * i. e. the cost of being not parallel is the same of that of not being
   *  mirrored astigmatism is defined in "analogCost.c" (Top) */

  if (transform1 == OCT_NO_TRANSFORM && transform2 == OCT_NO_TRANSFORM)  {
    return 0.0;
  }
  if (transform1 == OCT_MIRROR_X && transform2 == OCT_MIRROR_X)  {
    return 0.0;
  }
  if (transform1 == OCT_MIRROR_Y && transform2 == OCT_MIRROR_Y)  {
    return 0.0;
  }
  if (transform1 == OCT_ROT90 && transform2 == OCT_ROT90)  {
    return 0.0;
  }
  if (transform1 == OCT_ROT180 && transform2 == OCT_ROT180)  {
    return 0.0;
  }
  if (transform1 == OCT_ROT270 && transform2 == OCT_ROT270)  {
    return 0.0;
  }
  if (transform1 == OCT_MX_ROT90 && transform2 == OCT_MX_ROT90)  {
    return 0.0;
  }
  if (transform1 == OCT_MY_ROT90 && transform2 == OCT_MY_ROT90)  {
    return 0.0;
  }
  return (double) astigmatism;
}


   
/* This procedure checks if the cells are mirrored in the direction of the symmetry axis */


double
checkAstigmatism(transform1, transform2, type, singleFlag)
     octTransformType  transform1;
     octTransformType  transform2;
     SymmetryType            type;
     int                   singleFlag;

{
  /* astigmatism is defined in "analogCost.c" (Top) */
  if (singleFlag == FALSE)  {
    if (type == VERTICAL)  {
      if ((transform1 == OCT_MIRROR_X && transform2 == OCT_NO_TRANSFORM) || (transform1 == OCT_NO_TRANSFORM && transform2 == OCT_MIRROR_X))  {
        return 0.0;
      }
      if ((transform1 == OCT_ROT270 && transform2 == OCT_MX_ROT90) || (transform1 == OCT_MX_ROT90 && transform2 == OCT_ROT270))  {
        return 0.0;
      }
      if ((transform1 == OCT_ROT180 && transform2 == OCT_MIRROR_Y) || (transform1 == OCT_MIRROR_Y && transform2 == OCT_ROT180))  {
        return 0.0;
      }
      if ((transform1 == OCT_MY_ROT90 && transform2 == OCT_ROT90) || (transform1 == OCT_ROT90 && transform2 == OCT_MY_ROT90))  {
        return 0.0;
      }
      return (double) astigmatism;

    } else {

      if ((transform1 == OCT_ROT90 && transform2 == OCT_MY_ROT90) || (transform1 == OCT_MY_ROT90 && transform2 == OCT_ROT90))  {
        return 0.0;
      }
      if ((transform1 == OCT_MIRROR_X && transform2 == OCT_ROT180) || (transform1 == OCT_ROT180 && transform2 == OCT_MIRROR_X))  {
        return 0.0;
      }
      if ((transform1 == OCT_NO_TRANSFORM && transform2 == OCT_MIRROR_Y) || (transform1 == OCT_MIRROR_Y && transform2 == OCT_NO_TRANSFORM))  {
        return 0.0;
      }
      if ((transform1 == OCT_ROT270 && transform2 == OCT_MX_ROT90) || (transform1 == OCT_MX_ROT90 && transform2 == OCT_ROT270))  {
        return 0.0;
      }
      return (double) astigmatism;
    }
  } else {
    return 0.0;
  }
}


int
CurWellCost(costP)
     Cost *costP;

{
  MacroCell *cell;
  double misPosition = 0.0;

  if (!considerWell || SM_wellWeight <= 0.0)  {
    costP->well = 0.0;
    return OK;
  }

  cell = gFirstCell;
  while (cell != NULL) {
    if ((cell->biasStatus == VDD && cell->curCenter.y < wellBoundary) ||
        (cell->biasStatus == VSS && cell->curCenter.y > wellBoundary)) {
      misPosition += positionInWrongWell;
    } else {
      if (cell->biasStatus == SELF) {
        misPosition += ABS(cell->curCenter.y - wellBoundary); 
      }
    }
    cell = cell->globalNext;
  }
  costP->well = misPosition; 
}


/* This procedure calculate the cost of a move wrt the matching constraints */


int
MoveMatchingCost(moveP)
     Move *moveP;

{
  struct MatchingCluster  *MatchCell;   
  MacroCell                    *cell;  /* first Cell involved in a transformation */
  MacroCell                   *cell2;  /* eventual second Cell involved in a transformation */
  double             deltaCost = 0.0;
  double            deltaCost2 = 0.0;
  octId                   cellToMove; 
  octId                  cell2ToMove; 
  int        computationFlag = FALSE;

  if (!considerMatching || SM_matchWeight <= 0.0)  {
    moveP->cost.matching = 0.0;
    return OK;
  }
  MatchCell = gMatchCell->matchFirstCluster;
  cell = moveP->cell; 
  switch (moveP->complexity)  {
  case 0 :
    return NOT_OK;
    break;
  case 1 :
    cell2 = NULL;
    break;
  case 2 :
    cell2 = moveP->cell2;
    break;
  default :  
    FatalError ("Number of Cells to be moved is illegal"); 
    break;
  }

  if (cell != NULL && (cell->analogStatus == MULTIPLE_MATCHED ||
      cell->analogStatus == BOTH_MULTIPLE || cell->analogStatus == MATCHED ||
      cell->analogStatus == BOTH))  {
    cellToMove = cell->cellId;
    computationFlag = TRUE;
  } else {
    cellToMove = 0;
  }  

  if (cell2 != NULL && (cell2->analogStatus == MULTIPLE_MATCHED ||
      cell2->analogStatus == BOTH_MULTIPLE || cell2->analogStatus == MATCHED ||
      cell2->analogStatus == BOTH))  {
    cell2ToMove = cell2->cellId;
    computationFlag = TRUE;
  } else {
    cell2ToMove = 0;
  }

  if (computationFlag)  { 
    deltaCost = computeTotalMatchingCost(MatchCell, cellToMove, cell2ToMove, newFlag) -
                computeTotalMatchingCost(MatchCell, cellToMove, cell2ToMove, curFlag);
  } 

  moveP->cost.matching = deltaCost;
  return OK;
}




/* This procedure calculate the cost of a move wrt the symmetry constraints */

int
MoveSymmetryCost(moveP)
     Move *moveP;

{
  MacroCell      *cell;      /* first Cell involved in a transformation */
  MacroCell      *cell2;     /* eventual second Cell involved in a transformation */
  MacroCell      *yOtherCell; /* cell symmetric to cell or cell2 in vert. symm. */
  MacroCell      *xOtherCell; /* cell symmetric to cell or cell2 in horiz. symm. */
  int            yOtherCellId, xOtherCellId;
  int            foundSymmetry = FALSE;
  int            singleCellInCouple = FALSE; /* True if in a couple there exist one cell only */
  SymmetryType   symmetryType;
  octCoord       symmetryAxis = 0;
  double         deltaCost = 0.0, xDeltaCost = 0.0, yDeltaCost = 0.0;

  if (!considerSymmetry || SM_symWeight <= 0.0)  {
    moveP->cost.symmetry = 0.0;
    return OK;
  }
  cell = moveP->cell; 
  if (cell == NULL) return OK;

  switch (moveP->complexity)  {
  case 0 :
    return NOT_OK;
    break;
  case 1 :
    cell2 = NULL;
    break;
  case 2 :
    cell2 = moveP->cell2;
    break;
  default :  
    FatalError ("Number of Cells to be moved is illegal"); 
    break;
  }
 
  if (cell->analogStatus == SYMMETRIC ||
      cell->analogStatus == BOTH || cell->analogStatus == BOTH_MULTIPLE)  {
    if (cell->ySymPointer != NULL)  { 
      /* Computing cost of move for the first selected cell in vert. symmetry */
      symmetryType = VERTICAL;
      yOtherCell = computeMoveSymmetryToAxis(cell, cell2, &yDeltaCost, &symmetryType);
      if (yOtherCell == NULL || symmetryType == NO_THIS_SYMMETRY) {
	yOtherCellId = 0;
      } else {
	yOtherCellId = yOtherCell->cellId;
      } 
    }
    
    if (cell->xSymPointer != NULL)  {
    /* Computing cost of move for the first selected cell in horiz. symmetry */
      symmetryType = HORIZONTAL;
      xOtherCell = computeMoveSymmetryToAxis(cell, cell2, &xDeltaCost, &symmetryType);
      if (xOtherCell == NULL || symmetryType != NO_THIS_SYMMETRY) {
	xOtherCellId = 0;
      } else {
	xOtherCellId = xOtherCell->cellId;
      }
    }

    deltaCost = xDeltaCost + yDeltaCost;
  }

  if (cell2 != NULL && (cell2->analogStatus == SYMMETRIC || cell2->analogStatus == BOTH ||
			cell2->analogStatus == BOTH_MULTIPLE ))  {
    xDeltaCost = yDeltaCost = 0.0;
    
    if (cell2->ySymPointer != NULL && cell2->cellId != yOtherCellId)  {
      /* Computing cost of move for the second selected cell in vert. symmetry */
      symmetryType = VERTICAL;
      yOtherCell = computeMoveSymmetryToAxis(cell2, cell2, &yDeltaCost, &symmetryType);
    }
    
    if (cell2->xSymPointer != NULL && cell2->cellId != xOtherCellId)  {
      /* Computing cost of move for the second selected cell in horiz. symmetry */
      symmetryType = HORIZONTAL;
      xOtherCell = computeMoveSymmetryToAxis(cell2, cell2, &xDeltaCost, &symmetryType);
    }
   
    deltaCost += xDeltaCost + yDeltaCost;
  }
  moveP->cost.symmetry = deltaCost;

}    


MacroCell
*computeMoveSymmetryToAxis(cellP, cell2P, deltaCost, symType)
  MacroCell     *cellP;    /* input cell */
  MacroCell     *cell2P;    /* input cell */
  double        *deltaCost;
  SymmetryType  *symType;
  
{  
  int   singleCellInCouple = FALSE;
  octCoord            symmetryAxis;
  double          curCost, newCost;
  SymmetryType            symmetry;
  double                  distance;
  double                   astigma;
  MacroCell             *otherCell;
  int                  otherCellId;

  otherCell = NULL;
  symmetry = *symType;
  *deltaCost = 0.0;

  otherCell = searchOtherSymmetricCell (cellP, &symmetryAxis, &symmetry, 
					&singleCellInCouple);
  if (otherCell == NULL) {
    if (!singleCellInCouple) *symType = NO_THIS_SYMMETRY;
    otherCellId = 0;
  } else {
    otherCellId = otherCell->cellId;
  }

  if (cell2P != NULL && cell2P->cellId == otherCellId)  {
    if (otherCell != NULL) {
      distance = symmetryToAxis (cellP->newCenter, otherCell->newCenter, symmetryAxis, 
				 symmetry, singleCellInCouple);
      astigma = checkAstigmatism (cellP->newTransform, otherCell->newTransform, symmetry,
				  singleCellInCouple);
    } else {
      distance = symmetryToAxis (cellP->newCenter, cellP->newCenter, symmetryAxis, symmetry, 
				 singleCellInCouple);
      astigma = 0.0;
    }
    newCost = distance + astigma;
  } else {
    if (otherCell != NULL) {
      distance = symmetryToAxis (cellP->newCenter, otherCell->curCenter, symmetryAxis, 
				 symmetry, singleCellInCouple);
      astigma = checkAstigmatism (cellP->newTransform, otherCell->curTransform, 
				  symmetry, singleCellInCouple);
    } else {
      distance = symmetryToAxis (cellP->newCenter, cellP->newCenter, symmetryAxis, symmetry, 
				 singleCellInCouple);
      astigma = 0.0;
    }
    newCost = distance + astigma;
  }    
  if (otherCell != NULL) {
    distance = symmetryToAxis (cellP->curCenter, otherCell->curCenter, symmetryAxis, 
			       symmetry, singleCellInCouple);
    astigma = checkAstigmatism (cellP->curTransform, otherCell->curTransform, symmetry, 
				singleCellInCouple);
  } else {
    distance = symmetryToAxis (cellP->curCenter, cellP->curCenter, symmetryAxis, symmetry, 
			       singleCellInCouple);
    astigma = 0.0;
  }
  curCost = distance + astigma;
  *deltaCost = newCost - curCost;

  return otherCell;

} 



MacroCell
*searchOtherSymmetricCell (cellG, axis, type, singleFlag)
  MacroCell   *cellG;       /* cell whose partner has to be found */
  octCoord    *axis;        /* axis location */
  SymmetryType *type;       /* (VERTICAL, HORIZONTAL) */
  int         *singleFlag;      /* 1 (TRUE) or 2 cells in a couple ? */

{
  struct SymmetricCouple   *ptrCell;
  octId                      CellId;                       
  SymmetryType       typeOfSymmetry;
  MacroCell              *otherCell;   /* partner cell */

  *singleFlag = FALSE;
  typeOfSymmetry = *type;

  if (cellG == NULL) {
    otherCell = NULL;
    *type = NO_SYMMETRY;
    return otherCell;
  }
  if (typeOfSymmetry == VERTICAL)  {
    ptrCell = cellG->ySymPointer;
    if (ptrCell == NULL) {
      otherCell = NULL;
      return otherCell;
    }
    /* no symmetry of this type */
  } else {
    ptrCell = cellG->xSymPointer;
    if (ptrCell == NULL) {
      otherCell = NULL;
      return otherCell;
    }
    /* no symmetry of this type */
  }
  if (ptrCell->cell1 != NULL && ptrCell->cell2 != NULL)  {
    if (cellG == ptrCell->cell1)  {
      otherCell = ptrCell->cell2;
      CellId = otherCell->cellId;
    } else { 
      otherCell = ptrCell->cell1;
      CellId = otherCell->cellId;
    }
  } else {
    if (ptrCell->cell1 == NULL && ptrCell->cell2 == NULL)  {
      FatalError("Missing Link detected"); 
    } else {
   /* i. e.  if (only ptrCell->cell2 == NULL)  */
      otherCell = NULL;
      CellId = 0; 
      *singleFlag = TRUE;
    }
  } 
    
  *axis = ptrCell->axisPosition;
  *type = ptrCell->axisType;

  return otherCell;

}   


/* This procedure calculate the cost of a move wrt the well constraints */

int
MoveWellCost(moveP)
     Move *moveP;
{
  MacroCell                    *cell;  /* first Cell involved in a transformation */
  MacroCell                   *cell2;  /* eventual second Cell involved in a transformation */
  double             deltaCost = 0.0;
  double            deltaCost2 = 0.0;
  octId                   cellToMove; 
  octId                  cell2ToMove; 

  if (!considerWell || SM_wellWeight <= 0.0)  {
    moveP->cost.well = 0.0;
    return OK;
  }

  cell = moveP->cell; 
  switch (moveP->complexity)  {
  case 0 :
    return NOT_OK;
    break;
  case 1 :
    cell2 = NULL;
    break;
  case 2 :
    cell2 = moveP->cell2;
    break;
  default :  
    FatalError ("Number of Cells to be moved is illegal"); 
    break;
  }

  if (cell != NULL && cell->biasStatus == VDD) {
    if (cell->newCenter.y < wellBoundary) {  
      deltaCost = positionInWrongWell;
    }
    if (cell->curCenter.y < wellBoundary) {  
      deltaCost -= positionInWrongWell;
    }
  } else {
    if (cell != NULL && cell->biasStatus == VSS) {
      if (cell->newCenter.y > wellBoundary) {  
        deltaCost = positionInWrongWell;
      }
      if (cell->curCenter.y > wellBoundary) {  
        deltaCost -= positionInWrongWell;
      }
    } else {
      if (cell != NULL && cell->biasStatus == SELF) {
        deltaCost = ABS(cell->newCenter.y - wellBoundary) -
                    ABS(cell->curCenter.y - wellBoundary);  
      }
    }
  }

  if (cell2 != NULL && cell2->biasStatus == VDD) {
    if (cell2->newCenter.y < wellBoundary) {  
      deltaCost2 = positionInWrongWell;
    }
    if (cell2->curCenter.y < wellBoundary) {  
      deltaCost2 -= positionInWrongWell;
    }
  } else {
    if (cell2 != NULL && cell2->biasStatus == VSS) {
      if (cell2->newCenter.y > wellBoundary) {  
        deltaCost2 = positionInWrongWell;
      }
      if (cell2->curCenter.y > wellBoundary) {  
        deltaCost2 -= positionInWrongWell;
      }
    } else {
      if (cell2 != NULL && cell2->biasStatus == SELF) {
        deltaCost2 = ABS(cell2->newCenter.y - wellBoundary) -
                     ABS(cell2->curCenter.y - wellBoundary);  
      }
    }
  }
  moveP->cost.well = deltaCost + deltaCost2;
  return OK;
}


































