/*
 *  FILE: analogLink.c
 *
 *  DESCRIPTION: link of the simmetric and matched cells to the pre-
 *               existing data structure Cell.
 *
 *
 *  CREATION DATE: Oct.21.1987          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  int    linkSymmetricCellsToCell()      
 *  int    linkMatchedCellsToCell()             
 *             
 */
 

#include "defs.h"
#include "parall.h"
#include "global.h"
#include "sa.h"
#include <X11/Xlib.h>
#include "analogVariables.h"

/* IMPORTED GLOBAL ANALOG VARIABLES AND STRUCTURES */ 


/* This procedure searches cells in the two symmetric data-bases and compare them
 * with those found in the global data-base, if it finds a couple of identical
 * cells, it links the cell in the symmetric data-base with that in the general
 * one and frees the memory occupied by the former. A second link is made 
 * from the cell in the general data-base to the "couple" structure in the symmetric
 * data-base to which it belongs */ 

int
linkSymmetricCellsToCell()
{
  MacroCell                *cell;
  SymmetryType    typeOfSymmetry;
  
  Message("linking symmetry constraints data-base with the global data-base");
  cell = gFirstCell;
  while (cell != NULL)  {
#ifdef OBSOLETE_TO_BE_DELETED 
    sprintf(errmsg,"Instance %d found in global data-base, analogStatus? %d ",cell->cellId, cell->analogStatus);
    Message(errmsg);
#endif
    if (cell->cellId !=0)  {
      typeOfSymmetry = VERTICAL;
      searchAndLinkSymmetricCell(gySymCell, cell, &typeOfSymmetry); 
 
      typeOfSymmetry = HORIZONTAL;  
      searchAndLinkSymmetricCell(gxSymCell, cell, &typeOfSymmetry);  

#ifdef OBSOLETE_TO_BE_DELETED  
      sprintf(errmsg,"Instance %d, cell->analogStatus = %d",cell->cellId, cell->analogStatus);
      Message(errmsg);
#endif

    }  
    cell = cell->globalNext; 
  }
  return OK;
}

int
searchAndLinkSymmetricCell(SymCell, globalCell, sym)
  SymmetricCell         *SymCell;      /* symmetric data-base */
  MacroCell             *globalCell;   /* data-base that represents all cells */
  SymmetryType          *sym;          /* vertical / horizontal symmetry */
 
{
  int                    m,l;			/* indices */
  int                    numberOfAxes;
  int                    numberOfCouples;
  SymmetryType           symmetry;
 
  symmetry = *sym;  
  numberOfAxes = SymCell->numberOfAxis;
  for (m = 0; m < numberOfAxes; m++)  {
    numberOfCouples = SymCell->Axis[m].numberOfCouple;
    for (l = 0; l < numberOfCouples; l++)  {
#ifdef OBSOLETE_TO_BE_DELETED
     sprintf(errmsg,"Axis id = %d, couple id = %d ,Cell1 = %d, Cell2 = %d ,id data-base generale = %d",m,l, SymCell->Axis[m].couple[l].cell1->cellId, SymCell->Axis[m].couple[l].cell2->cellId, globalCell->cellId);
     Message(errmsg);  
#endif
      if (SymCell->Axis[m].couple[l].cell1 != NULL)  {
        if (SymCell->Axis[m].couple[l].cell1->cellId == globalCell->cellId)  {
          SymCell->Axis[m].couple[l].cell1 = globalCell;
          if (symmetry == VERTICAL)  {
            globalCell->ySymPointer = &(SymCell->Axis[m].couple[l]);
	  } else {
            globalCell->xSymPointer = &(SymCell->Axis[m].couple[l]);
	  }
          globalCell->analogStatus = SYMMETRIC;
	  if (debug)  {
            sprintf(errmsg,"Instance1 %d of symmetry data-base linked to general data-base, Symmetric? %d", SymCell->Axis[m].couple[l].cell1->cellId, globalCell->analogStatus);
            Message(errmsg);
          }
	  return OK;
        }
      }
      if (SymCell->Axis[m].couple[l].cell2 != NULL)  {
        if (SymCell->Axis[m].couple[l].cell2->cellId == globalCell->cellId)  {  
          SymCell->Axis[m].couple[l].cell2 = globalCell;
          if (symmetry == VERTICAL)  {
            globalCell->ySymPointer = &(SymCell->Axis[m].couple[l]);
          } else {
            globalCell->xSymPointer = &(SymCell->Axis[m].couple[l]);
          }
          globalCell->analogStatus = SYMMETRIC;
          if (debug)  {
            sprintf(errmsg,"Instance2 %d of symmetry data-base linked to general data-base, Symmetric? %d", SymCell->Axis[m].couple[l].cell2->cellId, globalCell->analogStatus);
            Message(errmsg);
          }
          return OK;
        }
      }
    }
  }   
  return NOT_OK;     
}



/* This procedure searches cells in the matching data-base and compare them
 * with those found in the global data-base, if it finds a couple of identical
 * cells, it links the cell in the matching data-base with that in the general
 * one and frees the memory occupied by the former */ 

int
linkMatchedCellsToCell()

{
  MacroCell                   *cell;
  int         clusterFirstIndex = 0;
  struct MatchingCluster *MatchCell;   

  Message("linking matching constraints data-base with the global data-base");
 
#ifdef OBSOLETE_TO_BE_DELETED 
  sprintf(errmsg,"Instance %d found in global data-base, analogStatus? %d ",cell->cellId, cell->analogStatus);
  Message(errmsg);
#endif

  cell = gFirstCell;
  MatchCell = gMatchCell->matchFirstCluster;
  while (cell != NULL)  {
    clusterFirstIndex = 0;
    searchAndLinkMatchedCell(MatchCell, cell, clusterFirstIndex);
    cell = cell->globalNext;
  }
#ifdef OBSOLETE_TO_BE_DELETED
  scanMatching(gMatchCell->matchFirstCluster);
#endif 
  return OK;
}


MacroCell
*scanMacroCell(globalCellP, localCellP)
   MacroCell *globalCellP;
   MacroCell *localCellP;

{
  MacroCell *cell;

  cell = gFirstCell;
  while (cell != NULL)  { 
    if ((cell->cellId != 0) && (localCellP->cellId == cell->cellId))  
      return cell;
    cell = cell->globalNext; 
  }
}  



int
searchAndLinkMatchedCell(inputCluster, globalCell, clusterIndex)  
   struct MatchingCluster *inputCluster;
   MacroCell                *globalCell;
   int                     clusterIndex;

{
  int  cellIndex;                                           /* index */
  struct MatchingCluster *cluster, *child;                  /* visited branches */ 
  MacroCell                  *temp, *cell, *previousCell;   /* visited cells */ 

  cluster = inputCluster;
  while (cluster != NULL)  {
    cellIndex = 0; 
    cell = cluster->MatchingFirstCell;
    if (cell != NULL)  { 
      clusterIndex++;
      while (cell != NULL)  {
        if (cell->cellId == globalCell->cellId)  {
          assignStatus(globalCell); 
          globalCell->matchPointer = cluster;
          cell->matchLink = globalCell;
          if (debug)  { 
            sprintf(errmsg,"Cluster %d ( distance = %d), index %d, matched cell %d (X) %d linked to general data-base, analogStatus %d", clusterIndex, cluster->ClusterLimitDistance, cellIndex, cell->matchLink->cellId, cell->matchLink->curCenter.x, cell->matchLink->analogStatus);
            Message(errmsg);
          }   
        } 
        cellIndex++;
        cell = cell->matchNextCell;
      }   
    }
    child = cluster->Cluster;
    clusterIndex = searchAndLinkMatchedCell(child, globalCell, clusterIndex);
    cluster = cluster->matchNextCluster;
  }
  return clusterIndex; 
}


assignStatus(globalCell)
   MacroCell *globalCell;

{    
  switch (globalCell->analogStatus)  {
  case NONE:
    globalCell->analogStatus = MATCHED;
    break; 
  case SYMMETRIC:
    globalCell->analogStatus = BOTH;
    break;
  case MATCHED:
    globalCell->analogStatus = MULTIPLE_MATCHED;
    break;
  case BOTH:
    globalCell->analogStatus = BOTH_MULTIPLE;
    break; 
  default:
    break;
  }
}

















