/*
 *  FILE: analogVariables.h
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Oct.15.1990          AUTHOR: Edoardo Charbon
 *
 *  STRUCTURES USED IN ANALOG MODULES:
 *  
 */


/* DEFINITION OF GLOBAL CONSTANTS FOR ANALOG DATA-BASE */

#define MAX_NUMBER_OF_SYMMETRIES 16 
#define MAX_NUMBER_OF_COUPLES 100


/* IMPORTED STRUCTURES */

#include "sa.h" 

/*
 * SymmetricCouple is declared in sa.h
 *
 * struct SymmetricCouple  {
 *   octCoord        axisPosition;
 *   enum SymmetryType  axisTyope;
 *   MacroCell             *cell1;
 *   MacroCell             *cell2;
 * };
 *
 *
 * MatchingCluster is declared in sa.h
 *
 * struct MatchingCluster  {
 *   int                 ClusterLimitDistance;
 *   double                  ClusterInnerCost;    cost of placement of cluster 
 *   int              numberOfMatchingCluster;
 *   struct MatchingCluster *matchNextCluster;
 *   int                 numberOfMatchingCell;
 *   MacroCell             *MatchingFirstCell;
 *   struct MatchingCluster          *Cluster;  
 * };
 *
 */ 


/* STRUCTURES */

typedef struct symmetric_cell      SymmetricCell;
typedef struct matched_cell        MatchedCell;
typedef enum DiffusionType { NEITHER_TYPE, PDIF_TYPE, NDIF_TYPE } DiffusionType;

struct SymmetryAxis  {
    int                 numberOfCouple;
    octCoord              axisPosition;
    struct SymmetricCouple couple[MAX_NUMBER_OF_COUPLES];
  };

struct symmetric_cell  {
    enum SymmetryType   axisType; 
    int             numberOfAxis;   
    int          numberOfSymCell;
    struct SymmetryAxis Axis[MAX_NUMBER_OF_SYMMETRIES];
  };

struct matched_cell  {
    int                         numberOfMatchCell;
    struct MatchingCluster     *matchFirstCluster;
  };  


/* DEFINITION OF GLOBAL VARIABLES USED FOR THE ANALOG DATA_BASE */

SHARED extern SymmetricCell      *gxSymCell; /* structure containing all symmetric cells with respect to x-axis */

SHARED extern SymmetricCell      *gySymCell; /* structure containing all symmetric cells with respect to y-axis */

SHARED extern MatchedCell *gMatchCell; /* structure containing all matched cells */


SHARED extern int considerSymmetry;

SHARED extern int considerMatching;

SHARED extern int considerWell; 

SHARED extern int considerVirtual;

SHARED extern double SM_matchWeight;

SHARED extern double SM_symWeight;

SHARED extern double SM_wellWeight;

SHARED extern int SM_virtualSteps;
 
extern int        analogMode;       /* flag to enable analog placement */

extern octCoord   wellBoundary;     /* boundary between p and n wells */ 





