/*
 *  FILE: analogVirtualAxis.c
 *
 *  DESCRIPTION: computation and update of virtual axes
 *
 *
 *  CREATION DATE: Nov.13.1990          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  int         computeVirtualAxis()
 *  int         computeAxis(SymCell)
 *  int         virtualAxis()  
 *      
 *  TODAY IS THE BIRTHDAY OF MY SISTER ! GREETINGS !
 *
 *  IMPORTED    FUNCTIONS:
 *  
 *    
 */

#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "oh.h"


/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */


/* IMPORTED VARIABLES */


/* DECLARED ROUTINES USED IN THIS FILE THAT DO NOT RETURN INTEGERS */

double
virtualAxis ();
void
computeAxis ();


/* FUNCTIONS */

/* This routine calculates the position of the virtual axis of a cluster of
 * Cells. The cluster is determined by the initial bag "SYM" containig the 
 * cells which have to be symmetric to a certain axis.
 * The position calculated here updates the current position of the axis. in
 * tha data structure gySymCell , gxSymCell.
 */ 


updateVirtualAxis()
 
{
  computeAxis(gySymCell);
  computeAxis(gxSymCell);
  return OK;
}


void
computeAxis(SymCell)
  SymmetricCell  *SymCell;      /* data-base that represents symmetric cells  */  

{
  
  int                    m,l;                         /* indices */   
  int                    numberOfAxes;
  SymmetryType           typeOfAxis;
  int                    numberOfCouples;
  octCoord               curPositionOfAxis;           /* current Axis position */
  octCoord               newPositionOfAxis;           /* new Axis position */
  double                 accumulator;
  struct SymmetryAxis    element; 
  octPoint               centerCell1,centerCell2;
  int                    singleCellInCouple = FALSE;    
  
  if (SymCell->numberOfSymCell <= 0)  {
    return;          /* there are no cells around the axis */
  } else {
    numberOfAxes = SymCell->numberOfAxis;
    typeOfAxis = SymCell->axisType;
    for (m = 0; m < numberOfAxes; m++)  {
      element = SymCell->Axis[m];
      numberOfCouples = element.numberOfCouple;
      curPositionOfAxis = element.axisPosition;
      accumulator = 0;
      for (l = 0; l < numberOfCouples; l++)  {
        singleCellInCouple = FALSE;
        centerCell1 = element.couple[l].cell1->curCenter;
        if (element.couple[l].cell2 == NULL)  {
          singleCellInCouple = TRUE;
        } else {
          centerCell2 = element.couple[l].cell2->curCenter;
        }
        accumulator += virtualAxis (centerCell1, centerCell2, typeOfAxis, singleCellInCouple);
      }
    newPositionOfAxis = (int)(accumulator/numberOfCouples);
    SymCell->Axis[m].axisPosition = newPositionOfAxis; 
    }
  }
}


/* compute the position of the virtual symmetry axis of a couple */

double
virtualAxis (center1, center2, type, singleFlag)
  octPoint  center1;
  octPoint  center2;
  SymmetryType type;
  int        singleFlag;
  
{
  if (singleFlag) {
    if (type == VERTICAL)  {
      return center1.x;
    } else { 
      return center1.y;
    }
  } else {    
    if (type == VERTICAL) {
      return (center1.x + center2.x)/2;
    } else {
      return (center1.y + center2.y)/2;
    }
  } 
}

/* update axis position on OCT */

int
saveAxis2oct(chip, symCell)
     octObject *chip;
     SymmetricCell *symCell;

{
  octObject           SymmetryBag, SymBag;
  octObject           newProp, prop;
  octGenerator        genSym, genBag;
  octStatus           symStatus;
  octCoord            integer;
  int                 m;
  int                 numberOfAxis;
  SymmetryType        axisType;
  char                symmetry[1];

  numberOfAxis = symCell->numberOfAxis;
  axisType = symCell->axisType;
  
  SymmetryBag.type = OCT_BAG;
  switch (axisType) {
  case VERTICAL:
    SymmetryBag.contents.bag.name = "YSYMMETRIES";
    strcpy(symmetry,"Y");
    break;
  case HORIZONTAL:
    SymmetryBag.contents.bag.name = "XSYMMETRIES";
    strcpy(symmetry,"X");
    break;
  }

  if (symCell != NULL && symCell->numberOfAxis != 0) {
    if (octGetByName(chip, &SymmetryBag) == OCT_NOT_FOUND)  {
      sprintf(errmsg, "%s bag not found. Axes in this direction not saved",
	      SymmetryBag.contents.bag.name);
      Warning(errmsg);
      return NOT_OK;
    }
  } else {
    return NOT_OK;
  }
  
  OH_ASSERT( octInitGenContents(&SymmetryBag,OCT_BAG_MASK,&genSym) );
  /* search bags "SYM" inside the SYMMETRIES bag */
  m = 0;
  while ((symStatus = octGenerate(&genSym,&SymBag)) == OCT_OK)  {
  
    /* search the position of the axis which is contained in a prop */
    symStatus = ohGetByPropName(&SymBag, &prop, "POSITION");
    if (symStatus == OCT_NOT_FOUND) {   
      sprintf(errmsg, "Axis prop not found. %s Axes not updated", symmetry);
      Warning(errmsg);
      return NOT_OK;
    }

    integer = symCell->Axis[m].axisPosition;
    prop.contents.prop.value.integer = integer;
    octModify(&prop);
    m++;
  }

  octFreeGenerator(&genSym);  
  octFlushFacet(chip);

  if(numberOfAxis != 0 && m > numberOfAxis) {
    sprintf(errmsg, "The number number of axes in %s direction is corrupted",
	    symmetry);
    if (axisType == VERTICAL) {
      Warning(errmsg);
    } else {
      Message(errmsg);
    } 
    return NOT_OK;
  }
  
  sprintf(errmsg, "%d Axes updated in %s direction", 
	  symCell->numberOfAxis, symmetry);
  if (axisType == VERTICAL) {
    Message(errmsg);
  } else {
    Warning(errmsg);
  } 
  return OK;
}

