/*
 *   makebins.c
 *	stolen fron Carl Sechen's Timberwolf
 *	Improved variable names and documentation
 */

#include "defs.h"
#include "global.h"

#ifdef PARALLEL
shared slock_t	binsLock;
#endif

typedef int 	*p2int;
typedef p2int 	*p2p2int;

makebins()
{
  double 	bins;		/* number of bins (to be found) */
  double 	c;		/* double version of numberOfCells */
  double	error;

  int		b1;			/* auxiliary counter for number of bins */
  int		a;			/* auxiliary counter */
  int		xbins , ybins ;
  int 		saveHx , saveHy , saveLx , saveLy , totalH , totalL ;
  int 		saveL2x , saveL2y , totalL2 ;
  int		i , j , x , y , diff ;

  /* Terrible way to find the number of bins */
  c = (double) numberOfCells ;
  for( b1 = 1 ; b1 <= 19 ; b1++ ) {
    for( a = 0 ; a <= 99 ; a++ ) {
      bins = b1 + (double) a / 100.0 ;
      error = pow((double)bins, 4.0) - 
	(2 * sqrt(c) + 2.0) * pow((double)bins,3.0) +
	  (2 * sqrt(c) + 1.0) * pow((double)bins,2.0) + c ;
      if( error < 0 ) {
	b1 = 20 ;
	break ;
      }
    }
  }

  /*  Provisionally */
  xbins = (int) bins + 1 ;
  ybins = (int) bins + 1 ;
  /*   find best values for numBinX and numBinY */
  totalH = xbins + ybins ;
  totalL = totalH - 1 ;
  totalL2 = totalL - 1 ;
  saveHx = 100 ;		/* some large value */
  diff = 32000 ;		/* some large value */
  for( x = 1 ; x <= xbins ; x++ ) {
    y = totalH - x ;
    if( x * y < bins * bins - 0.5 ) {
      continue ;
    }
    if( x * y - bins * bins < diff ) {
      diff = (int) (x * y - bins * bins) ;
      saveHx = x ;
      saveHy = y ;
    }
  }
  saveLx = 100 ;		/* some large value */
  diff = 32000 ;		/* some large value */
  for( x = 1 ; x <= xbins ; x++ ) {
    y = totalL - x ;
    if( x * y < bins * bins - 0.5 ) {
      continue ;
    }
    if( x * y - bins * bins < diff ) {
      diff = (int) (x * y - bins * bins) ;
      saveLx = x ;
      saveLy = y ;
    }
  }
  saveL2x = 100 ;		/* some large value */
  diff = 32000 ;		/* some large value */
  for( x = 1 ; x <= xbins ; x++ ) {
    y = totalL2 - x ;
    if( x * y < bins * bins - 0.5 ) {
      continue ;
    }
    if( x * y - bins * bins < diff ) {
      diff = (int) (x * y - bins * bins) ;
      saveL2x = x ;
      saveL2y = y ;
    }
  }
  if( saveLx < 100 AND saveHx < 100 ) {
    if( saveLx * saveLy <= saveHx * saveHy ) {
      xbins = saveLx ;
      ybins = saveLy ;
    } else {
      xbins = saveHx ;
      ybins = saveHy ;
    }
  } else if( saveHx < 100 ) {
    xbins = saveHx ;
    ybins = saveHy ;
  } else if( saveLx < 100 ) {
    xbins = saveLx ;
    ybins = saveLy ;
  }
  if( saveLx < 100 OR saveHx < 100 ) {
    if( saveL2x < 100 ) {
      if( xbins * ybins >= saveL2x * saveL2y ) {
	xbins = saveL2x ;
	ybins = saveL2y ;
      }
    }
  } else {
    xbins = saveL2x ;
    ybins = saveL2y ;
  }
  numBinX = xbins ;
  numBinY = ybins ;
    
/*   printf("numBinX automatically set to:%d\n", numBinX ); */
/*   printf("numBinY automatically set to:%d\n", numBinY ); */
    
  /* allocate the blockArray in shared memory */
    
  blockArray = SH_MALLOC(p2p2int, (1 + numBinX) ) ; 
  for( i = 0 ; i <= numBinX ; i++ ) {
    blockArray[i] = SH_MALLOC(p2int,(1 + numBinY));
    for( j = 0 ; j <= numBinY ; j++ ) {
      blockArray[i][j] = SH_MALLOC( int, (numberOfCells + 1));
    }
  }
  bucket = blockArray[0][0] ;


#ifdef PARALLEL
  s_init_lock(&binsLock);
#endif
  return(OK);
}
    
    
/*
 *    loadbins()
 *	Initialization routine for the bins
 *	Always executed by the parent process
 *	No use of binsLock
 */
int
loadbins()
{
  struct 	MacroCell *cell ;
  int		I;			/* counter for the cells */
  int 		i , j , K ;
  int 		startx , endx , starty , endy , *occuptr ;
  int		lowBinX , highBinX , lowBinY , highBinY ;
    

  binWidthX = binWidthY = 1;	/* default value */
  /* find actual size of bins */
  if (chipBbox != NULL) {
    
    /* Make sure the bins are computed in the most convenient direction */
    if ((numBinX - numBinY) * ((chipBbox->upperRight.x - chipBbox->lowerLeft.x)
			       - (chipBbox->upperRight.y - chipBbox->lowerLeft.y)) < 0 ) {

      /* Reallocate the space for the bins */

      for( i = 0 ; i <= numBinX ; i++ ) {
	for( j = 0 ; j <= numBinY ; j++ ) {
	  SH_FREE(blockArray[i][j]);
	}
	SH_FREE(blockArray[i]);
      }
      SH_FREE (blockArray);

      SWAP(int,numBinX,numBinY);

      blockArray = SH_MALLOC(p2p2int, (1 + numBinX) ) ; 
      for( i = 0 ; i <= numBinX ; i++ ) {
	blockArray[i] = SH_MALLOC(p2int,(1 + numBinY));
	for( j = 0 ; j <= numBinY ; j++ ) {
	  blockArray[i][j] = SH_MALLOC( int, (numberOfCells + 1));
	}
      }
      bucket = blockArray[0][0] ;

    }
    binOffsetX = chipBbox->lowerLeft.x;
    binOffsetY = chipBbox->lowerLeft.y;
    binWidthX  = (chipBbox->upperRight.x - chipBbox->lowerLeft.x)/ numBinX;
    binWidthY  = (chipBbox->upperRight.y - chipBbox->lowerLeft.y)/ numBinY;
    if (binWidthX <= 0 || binWidthY <= 0) {
      /* 
       * Something is wrong, maybe the chipbbox has not been
       * initialized. Do not worry, just keep going, things will
       * eventually be fixed
       */
      Warning("Number of bins is not positive. Set to 1");
    }
  }

  /* Always zeroes everything */
  for( i = 0 ; i <= numBinX ; i++ ) {
    for( j = 0 ; j <= numBinY ; j++ ) {
      for( K = 0 ; K <= numberOfCells ; K++ ) {
	blockArray[i][j][K] = 0 ;
      }
    }
  }

  /* put all the cells in the correct bin */
    
  for( I = 1 ; I <= numberOfCells ; I++ ) {
    cell = allCellsArray[I] ;
    
    startx = cell->curBbox.lowerLeft.x;
    endx   = cell->curBbox.upperRight.x;
    starty = cell->curBbox.lowerLeft.y;
    endy   = cell->curBbox.upperRight.y;
    
    lowBinX = (startx - binOffsetX) / binWidthX ;
    lowBinX = MAX(lowBinX , 1);
    lowBinX = MIN(lowBinX , numBinX);

    highBinX = (endx - binOffsetX) / binWidthX ;
    highBinX = MAX(highBinX , 1);
    highBinX = MIN(highBinX , numBinX);

    lowBinY = (starty - binOffsetY) / binWidthY ;
    lowBinY = MAX(lowBinY , 1);
    lowBinY = MIN(lowBinY , numBinY);
      
    highBinY = (endy - binOffsetY) / binWidthY ;
    highBinY = MAX(highBinY , 1);
    highBinY = MIN(highBinY , numBinY);

    if( lowBinX == highBinX  && lowBinY == highBinY ) {
      occuptr = blockArray[lowBinX][lowBinY] ;
    } else {
      occuptr = bucket ;
    }
    occuptr[ ++occuptr[0] ] = cell->number;
    cell->curBin = occuptr;
  }
  return(OK) ;
}
    
    
    
int    
TWoverlap( cell, curOrNewFlag )
    MacroCell *cell;
    int 	curOrNewFlag;
{
  MacroCell 	*cellptr2;	/* pointer to another cell that might overlap with cell */
  octCoord	left , right;	/* edges of the cell */
  octCoord	bottom , top ;	/* edges of the cell */
  register int	i , j , K;	/* counters */
  int		*occuptr;	/* pointer to bins */
  int		lowBinX;	/* index to lowest bin in X dimension */
  int		highBinX;	/* index of highest bin in X dimension */
  int		lowBinY;	/* index of lowest bin in Y dimension */
  int		highBinY;	/* index of highest bin in Y dimension */
  int		overlap;	/* the returned value */

  if (curOrNewFlag == NEW ) {
    left   = cell->newBbox.lowerLeft.x ;
    right  = cell->newBbox.upperRight.x ;
    bottom = cell->newBbox.lowerLeft.y ;
    top    = cell->newBbox.upperRight.y ;
  } else {
    left   = cell->curBbox.lowerLeft.x ;
    right  = cell->curBbox.upperRight.x ;
    bottom = cell->curBbox.lowerLeft.y ;
    top    = cell->curBbox.upperRight.y ;
  }
  /*
   *   Hash these coordinates to find the possible overlaps
   *   Bins to check are: bucket + (lowBinX->highBinX, lowBinY->highBinY)
   */
  lowBinX = (left - binOffsetX) / binWidthX ;
  if (lowBinX < 1) lowBinX = 1;
  if (lowBinX > numBinX) lowBinX = numBinX;

  highBinX = (right - binOffsetX) / binWidthX ;
  if (highBinX < 1) highBinX = 1;
  if (highBinX > numBinX) highBinX = numBinX;

  lowBinY = (bottom - binOffsetY) / binWidthY ;
  if (lowBinY < 1) lowBinY = 1;
  if (lowBinY > numBinY) lowBinY = numBinY;

  highBinY = (top - binOffsetY) / binWidthY ;
  if (highBinY < 1) highBinY = 1;
  if (highBinY > numBinY) highBinY = numBinY;


  overlap = 0;
  for( i = lowBinX ; i <= highBinX ; i++ ) {
    for( j = lowBinY ; j <= highBinY ; j++ ) {
      occuptr = blockArray[i][j] ;
      if (occuptr[0] == 0) {
	/* no cells in this bin */
	continue;		/* j loop */
      }
      for( K = 1 ; K <= occuptr[0] ; K++ ) {
	cellptr2 = allCellsArray[occuptr[K] ];
	if (cellptr2 != cell) {
	  if (curOrNewFlag == NEW) {
	    overlap += newOverlap(cell, cellptr2);
	  } else {
	    overlap += curOverlap(cell, cellptr2);
	  }
	}
      }				/* next K */
    }				/* next j */
  }				/* next i */
  /* check the bucket */
  if (bucket[0] != 0) {
    for( K = 1 ; K <= bucket[0] ; K++ ) {
      cellptr2 = allCellsArray[ bucket[K] ];
      if (cellptr2 != cell) {
	if (curOrNewFlag == NEW) {
	  overlap += newOverlap(cell, cellptr2);
	} else {
	  overlap += curOverlap(cell, cellptr2);
	}
      }
    }				/* next K */
  }
  return (overlap);
}


/*
 *	binReplace(cell)
 *	This routine can be executed in parallel
 *	It's necessary to use the lock
 */
binReplace(cell)
MacroCell *cell;
{
  octCoord	left, right, top, bottom;
  int		*occuptr, *oldBin;
  int		lowBinX, lowBinY, highBinX, highBinY;
  int		K;		/* index for the bins arrays */

  left   = cell->curBbox.lowerLeft.x;
  right  = cell->curBbox.upperRight.x;
  bottom = cell->curBbox.lowerLeft.y;
  top    = cell->curBbox.upperRight.y;
    
  lowBinX = (left - binOffsetX) / binWidthX ;
  lowBinX = MAX (1 , lowBinX);
  lowBinX = MIN (numBinX , lowBinX);

  highBinX = (right - binOffsetX) / binWidthX ;
  highBinX = MAX (1 , highBinX);
  highBinX = MIN (numBinX , highBinX);

  if( lowBinX == highBinX ) {
    lowBinY = (bottom - binOffsetY) / binWidthY ;
    lowBinY = MAX (1 , lowBinY);
    lowBinY = MIN (numBinY , lowBinY);

    highBinY = (top - binOffsetY) / binWidthY ;
    highBinY = MAX (1 , highBinY);
    highBinY = MIN (numBinY , highBinY);

    if( lowBinY == highBinY ) {
      occuptr = blockArray[lowBinX][lowBinY] ;
    } else {
      occuptr = bucket ;
    }
  } else {
    occuptr = bucket ;
  }
  if (occuptr != cell->curBin) {
#ifdef PARALLEL
    s_lock(&binsLock);
#endif
    /*	add cell to new bin */
    occuptr[ ++occuptr[0] ] = cell->number;
    /*  find cell number in the old bin and remove it */
    oldBin = cell->curBin;
    K = 1;
    while (oldBin[K] != cell->number) K++;     /* the cell MUST be in the oldbin !! */
    /*
     *  The next couple of instructions can cause some
     *	extra errors in the evaluation of the cost of the
     *  move, if another process at this time is running
     *  through this bin to compute ovelaps
     *	The sequence is such that the error, if any, 
     *	will be in excess
     */
    oldBin[K] = oldBin[ oldBin[0] ];
    oldBin[0]-- ;
    cell->curBin = occuptr;
#ifdef PARALLEL
    s_unlock(&binsLock);
#endif
  }
}
