/*
 *  FILE: cellTransform.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Aug.21.1987		AUTHOR: Andrea Casotto
 *
 *
 *  FUNCTIONS:
 *	cellTransform.c
 *	Computes the bounding box for the cell after
 *	a transformation
 */


#include "defs.h"
#include "global.h"
#include "parall.h"

/* DEFINITIONS */
#define TR_MASK1 (int)(1<<OCT_NO_TRANSFORM | 1<<OCT_MIRROR_X | 1<<OCT_MIRROR_Y | 1<<OCT_ROT180)

/* FUNCTIONS */

cellTransform(cell,transform)
    MacroCell *cell;
    octTransformType transform;
{
  cell->newTransform= transform;
  cell->newCenter.x = cell->curCenter.x;
  cell->newCenter.y = cell->curCenter.y;
  
  if ( (1 << (int) transform) & TR_MASK1) {
    cell->newBbox.lowerLeft.x= (octCoord) cell->curCenter.x - cell->rHalfW;
    cell->newBbox.lowerLeft.y= (octCoord) cell->curCenter.y - cell->rHalfH;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rW;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rH;
  } else {
    cell->newBbox.lowerLeft.x= (octCoord) cell->curCenter.x - cell->rHalfH ;
    cell->newBbox.lowerLeft.y= (octCoord) cell->curCenter.y - cell->rHalfW ;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rH;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rW;
  }
  newTermPosition(cell);
  newPerimeter(cell);
}


/* 
 *	cellPlace()
 *
 */
int
cellPlace(cell,x,y)
     MacroCell *cell;
     octCoord x,y;
{
  cell->newCenter.x = x;
  cell->newCenter.y = y;
  cell->newTransform = cell->curTransform;
  /*  find new coordinates of the bounding box */
  /*	
   *	Computes the bounding box for the cell in the new position 
   */
  if ( 1 << (int) cell->newTransform & TR_MASK1) {
    cell->newBbox.lowerLeft.x= (octCoord) cell->newCenter.x - cell->rHalfW ;
    cell->newBbox.lowerLeft.y= (octCoord) cell->newCenter.y - cell->rHalfH ;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rW;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rH;
  } else {
    cell->newBbox.lowerLeft.x= (octCoord) cell->newCenter.x - cell->rHalfH ;
    cell->newBbox.lowerLeft.y= (octCoord) cell->newCenter.y - cell->rHalfW ;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rH;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rW;
  }
  newTermPosition(cell);
  newPerimeter(cell);
}


/*	cellPlaceAndTransform()
 *	Computes the bounding box for the cell after
 *	a transformation
 */
int
cellPlaceAndTransform(cell,cx,cy,transform)
    MacroCell 	*cell;
    octCoord	cx,cy;		/* new center coordinates */
    octTransformType transform;
{
  cell->newTransform = transform;
  cell->newCenter.x = cx;
  cell->newCenter.y = cy;

  if ( 1<< (int) transform & TR_MASK1) {
    cell->newBbox.lowerLeft.x= (octCoord) cx - cell->rHalfW ;
    cell->newBbox.lowerLeft.y= (octCoord) cy - cell->rHalfH ;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rW;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rH;
  } else {
    cell->newBbox.lowerLeft.x= (octCoord) cx - cell->rHalfH ;
    cell->newBbox.lowerLeft.y= (octCoord) cy - cell->rHalfW ;
    cell->newBbox.upperRight.x = cell->newBbox.lowerLeft.x + cell->rH;
    cell->newBbox.upperRight.y = cell->newBbox.lowerLeft.y + cell->rW;
  }
  newTermPosition(cell);
  newPerimeter(cell);
}


/* 
 *	cellTranslate()
 *
 */
int
cellTranslate(cell,tx,ty)
     MacroCell *cell;
     octCoord tx,ty;
{
  cell->newBbox.lowerLeft.x = cell->curBbox.lowerLeft.x + tx;
  cell->newBbox.lowerLeft.y = cell->curBbox.lowerLeft.y + ty;
  cell->newBbox.upperRight.x = cell->curBbox.upperRight.x + tx;
  cell->newBbox.upperRight.y = cell->curBbox.upperRight.y + ty;
  cell->newCenter.x = cell->curCenter.x + tx;
  cell->newCenter.y = cell->curCenter.y + ty;
  cell->newTransform = cell->curTransform;
  newTermPosition(cell);		/* can be improved */	
  newPerimeter(cell);
  return(OK);
}




initializeCells()
{
  MacroCell *cell;

  for (cell = gFirstCell->globalNext ; cell != NULL ; cell = cell->globalNext) {
    cell->moveStatArray = SH_MALLOC(struct moveStat  , (int)NUMBER_OF_MOVE_TYPES);
    cell->moveSelArray  = SH_MALLOC(struct moveSelect, (int)NUMBER_OF_MOVE_TYPES);
    /* initMoveStatArray(cell->moveStatArray); OBSOLETE */
    /* computeQualityFactor(cell->moveStatArray);  */
    /*    reorderMoveSelArray(cell->moveStatArray,cell->moveSelArray); */
    /*     resetMoveStatArray(cell->moveStatArray); */

    expandCell(cell);		/* Compute the expanded BBox of the cell */
    totalExpandedArea += cell->area;

    if (cell->floatingTerms>0) {
      FloorplanMode = TRUE;	/* turn on the floorplanning mode */
      computeFloatingSites(cell);
      assignFloatingSites(cell);
    }

    /* 
     * Final initialization step:
     * Recompute the position of all terminals and sites 
     * and make those positions current
     */
    cellPlace(cell, cell->curCenter.x, cell->curCenter.y); 
    ExecuteCellMovement(cell);
    lyoReplace(cell);
    binReplace(cell);
  }
}
