/*
 *
 * >>>>  ROUTINES
 *   setChipBoundary()
 */

#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "octPlacement.h"
#include "iv.h"

ivWhichItem *ivChipBoundary ;

char   *ChipBoundaryTypeName[] = {
  "AUTOMATIC" ,		
  "INITIAL_POSITION" ,	
  "FIXED_CELLS" ,		
  "SPECIFIED_BY_OCT",		
  "NUMBER_OF_CB_TYPES"	
  };

octBox	CBbox[ (int) CB_NUMBER_OF_CB_TYPES];

#define MIN_CHIP_BOX_X 100
#define MIN_CHIP_BOX_Y 100

void buildIvChipBoundary()
{
    int i;
     ivChipBoundary = CALLOC ( ivWhichItem, (int)CB_NUMBER_OF_CB_TYPES );
     for ( i = 0; i < (int)CB_NUMBER_OF_CB_TYPES; i++) {
	 ivChipBoundary[i].value = ChipBoundaryTypeName[i];
	 ivChipBoundary[i].select = ( i == (int)chipBoundaryType );
     }
}

/*
 *    setChipBoundary()
 *    sets the box within which the cells will be allowed to move, 
 *    the box will have the exact aspect ratio
 *    and an area equal to the total area of all the cells
 *    times some factor >1
 *  >>>> AFFECTS
 *   gRangeLimiterX   <SET> to initial value
 */
int
initChipBoundary()
{
  octCoord	x,y;		/* coordinates, various uses */
  double  	expansionFactor;
  MacroCell	*cell;		/* pointer used in for loop */
  octBox	box;		/* generic */
  int		fixedCellsCounter;
  int		i;		/* in the for loop */

  chipBbox    = SH_MALLOC(octBox,1);

  for ( i = (int) CB_AUTOMATIC ; i < (int) CB_NUMBER_OF_CB_TYPES ; i++) {
    switch ((enum ChipBoundaryType) i) {
  case CB_AUTOMATIC:
      if (automaticChipArea == 0.0) {
	expansionFactor = 1.4;	/* default */
	automaticChipArea = totalExpandedArea * expansionFactor * expansionFactor ;
      }
      y = (int) sqrt (automaticChipArea / desAspectRatio);
      x = (int) automaticChipArea  / y;
      
      
      /*  store this bounding box  */
      box.lowerLeft.x = 0;
      box.lowerLeft.y = 0;
      box.upperRight.x = x;
      box.upperRight.y = y;
      break;
  case CB_INITIAL_POSITION:
      /* determine the bounding box of all the cell centers */
      box.lowerLeft.x  =gFirstCell->globalNext->curBbox.lowerLeft.x;
      box.lowerLeft.y  =gFirstCell->globalNext->curBbox.lowerLeft.y;
      box.upperRight.x =gFirstCell->globalNext->curBbox.upperRight.x;
      box.upperRight.y =gFirstCell->globalNext->curBbox.upperRight.y;
    
      for (cell=gFirstCell->globalNext;cell!=NULL;cell=cell->globalNext) {
	box.lowerLeft.x = MIN (box.lowerLeft.x , cell->curBbox.lowerLeft.x);
	box.lowerLeft.y = MIN (box.lowerLeft.y , cell->curBbox.lowerLeft.y);
	box.upperRight.x = MAX (box.upperRight.x , cell->curBbox.upperRight.x);
	box.upperRight.y = MAX (box.upperRight.y , cell->curBbox.upperRight.y);
      }
      break;
  case CB_FIXED_CELLS:
      fixedCellsCounter = 0;
    
      for (cell=gFirstCell->globalNext;cell!=NULL;cell=cell->globalNext) {
	if (cell->mobility == FIXED ) {
	  if (++fixedCellsCounter == 1) {
	    box.lowerLeft.x =  cell->curBbox.lowerLeft.x;
	    box.lowerLeft.y =  cell->curBbox.lowerLeft.y;
	    box.upperRight.x = cell->curBbox.upperRight.x;
	    box.upperRight.y = cell->curBbox.upperRight.y;
	  } else {
	    box.lowerLeft.x = MIN (box.lowerLeft.x , cell->curBbox.lowerLeft.x);
	    box.lowerLeft.y = MIN (box.lowerLeft.y , cell->curBbox.lowerLeft.y);
	    box.upperRight.x = MAX (box.upperRight.x , cell->curBbox.upperRight.x);
	    box.upperRight.y = MAX (box.upperRight.y , cell->curBbox.upperRight.y);
	  }
	}
      }
      if (fixedCellsCounter < 1) { /* not enough fixed cells to define a good box */
	box = CBbox[(int) CB_AUTOMATIC]; /* fix it up some how */ 
      }
      break;
  case CB_SPECIFIED_BY_OCT:	/* should be treated independently */
      if (getAvailableAreaFromOct(&Chip,&box)==OCT_OK) {
	box.lowerLeft.x /= OCT_LAMBDA;
	box.lowerLeft.y /= OCT_LAMBDA;
	box.upperRight.x /= OCT_LAMBDA;
	box.upperRight.y /= OCT_LAMBDA;
      }
      break;
  default:
      break;
    }
    if (box.upperRight.x - box.lowerLeft.x < MIN_CHIP_BOX_X) {
      (void) sprintf(errmsg,"Width for chip %s set to %d", ChipBoundaryTypeName[i], MIN_CHIP_BOX_X);
      Warning(errmsg);
      box.upperRight.x = box.lowerLeft.x + MIN_CHIP_BOX_X;
    }
    if (box.upperRight.y - box.lowerLeft.y < MIN_CHIP_BOX_Y) {
      (void) sprintf(errmsg,"Height for chip %s set to %d",ChipBoundaryTypeName[i], MIN_CHIP_BOX_Y);
      Warning(errmsg);
      box.upperRight.y = box.lowerLeft.y + MIN_CHIP_BOX_Y;
    }

    {CBbox[ i ] = box;}
  }
  setChipBoundary(chipBoundaryType);
  return(OK);
}


int iv_changedAspectRatio()
{
  octCoord	x,y;		/* coordinates, various uses */
  octBox	box;		/* generic */
  int		i;		/* in the for loop */
  extern void   plotCells();

  /* bound aspect ratio within reasonable limits */

  if (desAspectRatio > 10.0 ) desAspectRatio = 10.0;
  if (desAspectRatio < 0.10 ) desAspectRatio = 0.10;


  /* Not a trivial change. Must maintain consistency of bin structure */
  /* Stop all processes and wait that all go into limbo */
  AllChildrenInLimbo(); 

  i = (int) CB_AUTOMATIC ;
  y = (int) sqrt (automaticChipArea / desAspectRatio);
  x = (int) automaticChipArea / y;
      
  box.lowerLeft.x = 0;   box.lowerLeft.y = 0;
  box.upperRight.x = x;  box.upperRight.y = y;

  if (box.upperRight.x - box.lowerLeft.x < MIN_CHIP_BOX_X) {
    (void) sprintf(errmsg,"Width for chip %s set to %d", ChipBoundaryTypeName[i], MIN_CHIP_BOX_X);
    Warning(errmsg);
    box.upperRight.x = box.lowerLeft.x + MIN_CHIP_BOX_X;
  }
  if (box.upperRight.y - box.lowerLeft.y < MIN_CHIP_BOX_Y) {
    (void) sprintf(errmsg,"Height for chip %s set to %d",ChipBoundaryTypeName[i], MIN_CHIP_BOX_Y);
    Warning(errmsg);
    box.upperRight.y = box.lowerLeft.y + MIN_CHIP_BOX_Y;
  }
  {CBbox[ i ] = box;}
  chipBoundaryType = CB_AUTOMATIC;
  setChipBoundary(chipBoundaryType);

  refreshConsoleWindow();
  plotCells();
  return(IV_OK);
}

setChipBoundary( type )
    enum ChipBoundaryType type;
{
  {*chipBbox = CBbox[(int) type];}
  loadbins();
}

initRangeLimiters()
{
  if (useCellBasedMoveStats == FALSE ) {	
    gRangeLimiterX = (chipBbox->upperRight.x - chipBbox->lowerLeft.x) / 8;
    gRangeLimiterY = (chipBbox->upperRight.y - chipBbox->lowerLeft.y) / 8;
  } else {
    /* Temporary instructions (just like the 'then' case) */
    gRangeLimiterX = (chipBbox->upperRight.x - chipBbox->lowerLeft.x) / 8;
    gRangeLimiterY = (chipBbox->upperRight.y - chipBbox->lowerLeft.y) / 8;
  }
}

PM_changeCBtype()
{
  /* Not a trivial change. Must maintain consistency of bin structure */
  /* Stop all processes and wait that all go into limbo */
  AllChildrenInLimbo(); 

  chipBoundaryType = (enum ChipBoundaryType)
   ( ((int) chipBoundaryType + 1) % (int) CB_NUMBER_OF_CB_TYPES);
  setChipBoundary(chipBoundaryType);
  refreshConsoleWindow();
  return(OK);
}

iv_changeCBtype()
{
    extern IVWindow *ivWin;
    extern ivWhichItem *ivChipBoundary;
    int	 flag;
    /* Not a trivial change. Must maintain consistency of bin structure */
    /* Stop all processes and wait that all go into limbo */
    
    AllChildrenInLimbo(); 

    chipBoundaryType =   (enum ChipBoundaryType)iv_WhichSelect( ivWin, ivChipBoundary );
    flag = ( chipBoundaryType == CB_AUTOMATIC );
    iv_SetEnableFlag( ivWin, &desAspectRatio, flag );
    iv_SetEnableFlag( ivWin, &automaticChipArea, flag );
    setChipBoundary( chipBoundaryType );
    refreshConsoleWindow();
    return IV_OK;
}



explainChipBoundaryType(fp)
    FILE *fp;
{
  int i;

  fprintf(fp,"# ------- Codes for `chipBoundaryType' -------\n");
  for (i = 0 ; i < (int) CB_NUMBER_OF_CB_TYPES ; i++) {
    fprintf(fp, "\t# %d ==> %s\n", i , ChipBoundaryTypeName[i]);
  }
}


readPlacementConstraints()
{

}


getAvailableAreaFromOct(FacetPtr,boxP)
    octObject *FacetPtr;
    struct octBox	*boxP;
{
  octObject    AvailAreaBox;	/*  */
  octObject    Interface;

  /* Get the available area from the interface facet */
  Interface.type = OCT_FACET;
  (Interface.contents.facet = FacetPtr->contents.facet);
  Interface.contents.facet.facet = "interface";
  Interface.contents.facet.mode = "r";

  if (chipBoundaryType == CB_SPECIFIED_BY_OCT) {
    if (debug) {
      printf("Getting available area from OCT\n");
      printf("I'm looking for the bag %s in %s:%s:%s\n",
	     FloorplanBag.contents.bag.name,
	     Interface.contents.facet.cell,
	     Interface.contents.facet.view,
	     Interface.contents.facet.facet);
      printf("Inside the bag I expect a box which may (or may not) be attached to layer PLACE\n");
    }
  }
  VOVinputFacet( &Interface );
  if (octOpenFacet(&Interface)<OCT_OK) {
    (void) sprintf(errmsg,"Problems opening %s:%s:%s;%s",
		    Interface.contents.facet.cell,
		    Interface.contents.facet.view,
		    Interface.contents.facet.facet,
		    Interface.contents.facet.version);
    Warning(errmsg);
    return(OCT_ERROR);
  }
  if (octGetByName(&Interface,&FloorplanBag)!=OCT_OK) {
    if (chipBoundaryType == CB_SPECIFIED_BY_OCT) Warning("No FLOORPLAN bag");
    return(OCT_ERROR);
  }
  if (octGenFirstContent(&FloorplanBag,OCT_BOX_MASK,&AvailAreaBox)!=OCT_OK) {
    Warning("No box in FLOORPLAN bag");
    return(OCT_ERROR);
  }
  (*boxP = AvailAreaBox.contents.box);
  return(OCT_OK);
}


/* translate all the cells in order to have the lower left corner of the gFirstCell in 0,0 */
placementPostProcessing()
{
/*
  *** THIS DOES NOT WORK FOR ALL THE CELLS. PADS ARE NOT MOVED ****
  *** That's why it is commented out  ****

  octCoord  left;
  octCoord  bottom;
  MacroCell *cell;

 

  if (chipBoundaryType == CB_AUTOMATIC) {
    left  = gFirstCell->lyoNext[PUPPY_L]->curBbox.lowerLeft.x;
    bottom= gFirstCell->lyoNext[PUPPY_B]->curBbox.lowerLeft.y;
    for (cell = gFirstCell->globalNext ; cell != NULL; cell = cell->globalNext) {
      cellTranslate(cell, -left, -bottom);
      ExecuteGraphicMovement(cell);
      lyoReplace(cell);
      binReplace(cell);
    }
  }
*/
}



