/*
 *   cluster.c
 *>>>> ROUTINES 
 */

#include "defs.h"
#include "sa.h"
#include "analogVariables.h"
#include "global.h"
#include "parall.h"


/*
 *   newMoment(moveP)
 *   this routine updates the fields related to the computation
 *   of the cluster cost, depending on the move
 *
 */
newMoment(moveP)
    Move *moveP;
{
  octCoord		ncx,ncy,ccx,ccy;
  double		area;
  struct ProcessorFrame *Proc;
  struct cluster_t 	*ncl;	/* new cluster in myProc */
  struct cluster_t 	*ccl;	/* cur cluster in myProc */

  ncl = &myProc->newCluster;
  ccl = &myProc->curCluster;
  switch (moveP->type) {
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
  case TRANSLATE_AND_TRANSFORM:
    myProc->newNumberOfCells = myProc->curNumberOfCells;

    area = moveP->cell->area;
    ncx = moveP->cell->newCenter.x;
    ncy = moveP->cell->newCenter.y;
    ccx = moveP->cell->curCenter.x;
    ccy = moveP->cell->curCenter.y;
    ncl->totalArea = ccl->totalArea;
    ncl->sumXArea = ccl->sumXArea + area * (ncx - ccx);
    ncl->sumYArea = ccl->sumYArea + area * (ncy - ccy);
    ncl->sumX2Area = ccl->sumX2Area + (ncx * ncx - ccx * ccx) *  area;
    ncl->sumY2Area = ccl->sumY2Area + (ncy * ncy - ccy * ccy) *  area;
    computeBaricenterAndMomentum(ncl);
    break;
  case TRANSFORMATION:		/* no change */
    myProc->newNumberOfCells = myProc->curNumberOfCells;
    myProc->newCluster = myProc->curCluster; /* no need to lock newCluster */
    break;
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
    myProc->newNumberOfCells = myProc->curNumberOfCells;
    ncl->totalArea = ccl->totalArea;
    /* first use cell and then cell2 */
    area = moveP->cell->area;
    ncx = moveP->cell->newCenter.x;
    ncy = moveP->cell->newCenter.y;
    ccx = moveP->cell->curCenter.x;
    ccy = moveP->cell->curCenter.y;
    ncl->sumXArea = ccl->sumXArea + area * (ncx - ccx);
    ncl->sumYArea = ccl->sumYArea + area * (ncy - ccy);
    ncl->sumX2Area = ccl->sumX2Area + area * (double) (ncx * ncx - ccx * ccx);
    ncl->sumY2Area = ccl->sumY2Area + area * (double) (ncy * ncy - ccy * ccy);
    if (moveP->cell2->owner == myself ) {
      area = moveP->cell2->area;
      ncx = moveP->cell2->newCenter.x;
      ncy = moveP->cell2->newCenter.y;
      ccx = moveP->cell2->curCenter.x;
      ccy = moveP->cell2->curCenter.y;
      ncl->sumXArea += area * (ncx - ccx);
      ncl->sumYArea += area * (ncy - ccy);
      ncl->sumX2Area += area * (double)(ncx * ncx - ccx * ccx);
      ncl->sumY2Area += area * (double)(ncy * ncy - ccy * ccy);
    } 
    /* compute baricenter and momentum */
    computeBaricenterAndMomentum(ncl);
    break;
  case TERMINAL_MOVE:
  case SUBNET_MOVE:
    myProc->newNumberOfCells = myProc->curNumberOfCells;
    myProc->newCluster = myProc->curCluster; /* no need to lock newCluster */
    break;
  case CHANGE_OWNERSHIP:
    Proc = &process[moveP->procNumber];

    myProc->newNumberOfCells = myProc->curNumberOfCells - 1 ;
    Proc->newNumberOfCells = Proc->curNumberOfCells + 1 ;

    myProc->newCluster = myProc->curCluster;
    Proc->newCluster = Proc->curCluster;
    SubtractCellToCluster(&myProc->newCluster, moveP->cell);
    AddCellToCluster(&Proc->newCluster, moveP->cell);

    computeBaricenterAndMomentum(&myProc->newCluster);
    computeBaricenterAndMomentum(&Proc->newCluster);
	    
    break;
  default:
    break;
  }
}


computeBaricenterAndMomentum(cl)
    struct cluster_t *cl;
{
  if (cl->totalArea < 0.0) {
    FatalError("Total area of cells is not positive");
  } else if (cl->totalArea == 0.0) {
    cl->momentum = 0.0;
  } else {
    cl->barX = cl->sumXArea / cl->totalArea;
    cl->barY = cl->sumYArea / cl->totalArea;
    cl->momentum = cl->sumX2Area + cl->sumY2Area -
      (cl->barX * cl->barX + cl->barY * cl->barY) * cl->totalArea;
  }
}

zeroesCluster(cl)
    struct cluster_t *cl;
{
  cl->totalArea = 0.0;
  cl->sumXArea = 0.0;
  cl->sumYArea = 0.0;
  cl->sumX2Area = 0.0;
  cl->sumY2Area = 0.0;
}

AddCellToCluster(cl,cell)
    struct cluster_t *cl;
    MacroCell	*cell;
{
  double	area;
  octCoord	cx,cy;		/* center coordinates */

  area = cell->area;
  cx = cell->curCenter.x;
  cy = cell->curCenter.y;
#ifdef PARALLEL
  s_lock(&cl->lock);		/* probably useless */
#endif
  cl->totalArea += area;
  cl->sumXArea += cx * area;
  cl->sumYArea += cy * area;
  cl->sumX2Area += cx * cx * area;
  cl->sumY2Area += cy * cy * area;
#ifdef PARALLEL
  s_unlock(&cl->lock);
#endif
}

SubtractCellToCluster(cl,cell)
    struct cluster_t *cl;
    MacroCell	*cell;
{
  double	area;
  octCoord	cx,cy;		/* center coordinates */

  area = cell->area;
  cx = cell->curCenter.x;
  cy = cell->curCenter.y;
#ifdef PARALLEL
  s_lock(&cl->lock);
#endif
  cl->totalArea -= area;
  if (cl->totalArea < 0.0){
    FatalError("In SubtractCellToCluster");
  }
  cl->sumXArea -= cx * area;
  cl->sumYArea -= cy * area;
  cl->sumX2Area -= cx * cx * area;
  cl->sumY2Area -= cy * cy * area;
#ifdef PARALLEL
  s_unlock(&cl->lock);
#endif
}



checkClusters()
{
  int 		n;
  AnnealVar	*annealVar;
  MacroCell 	*cell;
  double 	area;
  int	 	ncells;

  for (n = 0 ; n <P ; n++) {
    ncells = 0;
    area = 0.0;
    for (annealVar=process[n].firstAnnealVar; annealVar!=NULL; annealVar=annealVar->ownerNext) {
      if (annealVar->type == AV_CELL) {
	cell = annealVar->pointer.cell;
	ncells++;
	if (cell->status == BORROWED) {
	  (void) sprintf(errmsg,"Cell %s belongs to proc %d, but is BORROWED", cell->name, cell->owner);
	  Warning(errmsg);
	} else  if (cell->status == OWNED) {
	  area += cell->area;
	}
      }
    }
    if (ncells != process[n].curNumberOfCells) {
      (void) sprintf(errmsg,"Inconsistent cell count for proc %d:  %d vs %d",
	      n, ncells, process[n].curNumberOfCells);
      Warning(errmsg);
    }
    if (area != process[n].curCluster.totalArea) {
      (void) sprintf(errmsg,"Inconsistent area count for proc %d:  %g vs %g",
	      n, area, process[n].curCluster.totalArea);
      Warning(errmsg);
    }
  }
}
