/*
 *  FILE: constraintDrivenVariables.h
 *
 *  DESCRIPTION:
 * 
 *
 *  CREATION DATE: 3/22/91		AUTHOR: Edoardo Charbon
 * 
 *
 *  STRUCTURES:
 *
 */


/* IMPORTED STRUCTURES */

# include "sa.h"

/*
 * Sensitivity is declared in sa.h
 *
 * struct Sensitivity {
 *   char                 *performanceName;
 *   int                  performanceId;
 *   double               plus;
 *   double               minus;
 * };
 *
 */


/* STRUCTURES */

typedef struct ConstraintDriven ConstraintDriven;
typedef struct Stringa Stringa;
typedef struct NetConstraint NetConstraint;
typedef struct MatchedNets MatchedNets;
typedef struct Sensitivity Sensitivity;
typedef enum MetalType {FIRST_MET, SECOND_MET} MetalType;

struct Stringa {
  char                 str[40];
};

struct ConstraintDriven {
  int                  numberOfPerformance;
  int                  numberOfNet; 
  int                  numberOfMatchedPair;
  Sensitivity          *deltaWmax;
  NetConstraint        *netConstraint;
  MatchedNets          *matchedNets;
};

struct NetConstraint {
  char                 netName[40];
  Net                  *net;
  char                 groundNetName[40]; 
  Sensitivity          *S;
};

struct MatchedNets {
  char                 *matchingType;
  char                 net1Name[40];
  char                 net2Name[40];
};


/* DEFINITION OF GLOBAL CONSTANTS FOR ANALOG DATA-BASE */
#define PERFORMANCE_NUM 1


/* DEFINITION OF GLOBAL CONSTANTS FOR CONSTRAINT-DRIVEN DATA-BASE */

#define DATA_CONST 2
#define COUPLE_CONST 2
#define FIRST_CONST 1 
#define SECOND_CONST 2
#define NET_CONST 1
#define GROUND_CONST 0
#define SKIP_CONST 1


/* DEFINITION OF GLOBAL VARIABLES FOR CONSTRAINT-DRIVEN DATA-BASE */

SHARED extern int showConstrainedNets;     /* show the nodes on which there
				     	    * are constraints */ 

SHARED extern int considerConstraintDriven;     /* consider constraints */

SHARED extern double SM_constraintWeight; /* constraints weight */

SHARED extern ConstraintDriven *gConstraintDriven; /* structure containing all 
						    * critical nets */
SHARED extern double SM_topCapMod0;   /* top layer of metal capacitance model: 1st const */
SHARED extern double SM_topCapMod1;   /* top layer of metal capacitance model: 2nd const */
SHARED extern double SM_bottomCapMod0;/* bottom layer of metal capacitance model: 1st const */
SHARED extern double SM_bottomCapMod1;/* bottom layer of metal capacitance model: 2nd const */
SHARED extern double SlopeRatio;      /* ratio between the slopes of the curves used in the
				       * evaluation of the cost function for critical nets */
SHARED extern float  SM_minNetWidth;  /* min net width in lambda (analog mode only) */






















