#include "defs.h"
#include "global.h"
#include "octPlacement.h"
#include "vov.h"

static int totNetProps;
static int totCellProps;
static int getLine();
static int checkCell();
static int checkNet();

/*
 * Unfortunately unions cannot be initialized. Thus I have to resort
 * to this sort of kludgy mess
*/
typedef struct {
  char *name;
  octPropType type;
  char   *lastStringValue;
  char   *defaultStringValue;
  int	 lastIntegerValue;
  int    defaultIntegerValue;
  double lastRealValue;
  double defaultRealValue;
} auxPropType;

auxPropType  cellPropsAux[] = {
  { "MOBILITY" , OCT_STRING , "" , "MOVABLE" , 0 , 0 , 0.0, 0.0} ,  
  { "LEGAL_TRANSFORMATIONS" , OCT_STRING , "" , "ALL" , 0 , 0 , 0.0 , 0.0 },
  { "END_OF_LIST", (octPropType) 0}
}; 

auxPropType netPropsAux[] = {
  { "NET_CRITICALITY" , OCT_INTEGER ,"", "",  -99999 , 0 , 0.0, 0.0},
  { "NET_MAX_LENGTH"  , OCT_INTEGER ,"", "",  -99999 , 0 , 0.0, 0.0},
  { "END_OF_LIST", (octPropType) 0 }
};

octObject *CellProps;
octObject *NetProps;

initConstraints()
{
  static int initialized = FALSE;
  int i;

  if (initialized) return;

  i = 0; 
  while (strcmp(cellPropsAux[i].name, "END_OF_LIST") != 0) i++;
  totCellProps = i;
  i = 0; 
  while (strcmp(netPropsAux[i].name, "END_OF_LIST") != 0) i++;
  totNetProps = i;

  CellProps = SH_MALLOC(octObject, totCellProps);
  for (i = 0 ; i < totCellProps ; i++ ) {
    CellProps[i].type = OCT_PROP;
    CellProps[i].contents.prop.name = cellPropsAux[i].name;
    CellProps[i].contents.prop.type = cellPropsAux[i].type;
  }

  NetProps = SH_MALLOC(octObject, totNetProps);
  for (i = 0 ; i < totNetProps ; i++ ) {
    NetProps[i].type = OCT_PROP;
    NetProps[i].contents.prop.name = netPropsAux[i].name;
    NetProps[i].contents.prop.type = netPropsAux[i].type;
  }
  initialized = TRUE;
}

int
printConstraintsFile( doCells, doNets)
    int		doCells;
    int		doNets;
{
  extern char 	*progName;
  MacroCell 	*cell;
  Net		*net;
  int		i;
  octObject	Instance;
  octObject	ChipNet;
  char			constraintsFileName[128];
  FILE 			*fp;

  
  (void) sprintf(constraintsFileName,"%s.constraintsPuppy", Chip.contents.facet.cell);
  VOVoutput( VOV_UNIX_FILE, constraintsFileName );
  if (( fp=fopen(constraintsFileName , "w")) == (FILE *) NULL) {
    (void) sprintf(errmsg,"Could not open %s", constraintsFileName);
    Message(errmsg);
  } 
  (void) sprintf(errmsg,"Using %s file",constraintsFileName);
  Message(errmsg);
	  
  initConstraints();

  fprintf(fp,"#\n# Formal Terminal description for %s:%s\n",
	  Chip.contents.facet.cell, Chip.contents.facet.view);
  fprintf(fp,"# %s recognizes the following keywords:\n", progName);
  for (i = 0 ; i < totNetProps ; i++ ) {
    fprintf(fp, "# %s\n" , NetProps[i].contents.prop.name);
  }
  for (i = 0 ; i < totCellProps ; i++ ) {
    fprintf(fp, "# %s\n" , CellProps[i].contents.prop.name);
  }
  
  fprintf(fp,"# This file uses a 'REVERSE' logic: \n");
  fprintf(fp,"# First you specify the properties\n# then the objects for which the props apply\n");

  if (doCells) {
    for (cell = gFirstCell->globalNext ; cell != NULL ; cell = cell->globalNext) {
    
      Instance.objectId = cell->cellId;
      octGetById(&Instance);
      for (i = 0 ; i < totCellProps; i++ ) {
      
	switch (CellProps[i].contents.prop.type ) {
	case OCT_STRING:
	  if (octGetByName(&Instance,&CellProps[i]) != OCT_OK) {
	    CellProps[i].contents.prop.value.string = cellPropsAux[i].defaultStringValue;
	  }
	  if (strcmp(cellPropsAux[i].lastStringValue,CellProps[i].contents.prop.value.string)!=0) {
	    fprintf(fp, "\t%s %s\n" , cellPropsAux[i].name,CellProps[i].contents.prop.value.string);
	    cellPropsAux[i].lastStringValue = strsav(CellProps[i].contents.prop.value.string);
	  }
	  break;
	case OCT_INTEGER:
	  if (octGetByName(&Instance,&CellProps[i]) != OCT_OK) {
	    CellProps[i].contents.prop.value.integer = cellPropsAux[i].defaultIntegerValue;
	  }
	  if (cellPropsAux[i].lastIntegerValue != CellProps[i].contents.prop.value.integer) {
	    fprintf(fp, "\t%s %d\n" , cellPropsAux[i].name,CellProps[i].contents.prop.value.integer);
	    cellPropsAux[i].lastIntegerValue = CellProps[i].contents.prop.value.integer;
	  }
	  break;
	case OCT_REAL:
	  if (octGetByName(&Instance,&CellProps[i]) != OCT_OK) {
	    CellProps[i].contents.prop.value.real = cellPropsAux[i].defaultRealValue;
	  }
	  if (cellPropsAux[i].lastRealValue != CellProps[i].contents.prop.value.real) {
	    fprintf(fp, "\t%s %lf\n" , cellPropsAux[i].name,CellProps[i].contents.prop.value.real);
	    cellPropsAux[i].lastRealValue = CellProps[i].contents.prop.value.real;
	  }
	  break;
	default:
	  FatalError("Wrong cell property type in printConstraintsFile()");
	  break;
	}
      }
      fprintf(fp,"INSTANCE %s\n", Instance.contents.instance.name);
    }
  }


  if (doNets) {
    for (net = netSet->globalNext ; net != NULL ; net = net->globalNext) {
    
      ChipNet.type = OCT_NET;
      ChipNet.contents.net.name = net->name;
      if (octGetByName(&Chip,&ChipNet) != OCT_OK) FatalOctError("Internal inconsistency in printConstraintsFile");
      for (i = 0 ; i < totNetProps; i++ ) {
      
	switch (NetProps[i].contents.prop.type ) {
	case OCT_STRING:
	  if (octGetByName(&ChipNet,&NetProps[i]) != OCT_OK) {
	    NetProps[i].contents.prop.value.string = netPropsAux[i].defaultStringValue;
	  }
	  if (strcmp(netPropsAux[i].lastStringValue,NetProps[i].contents.prop.value.string)!=0) {
	    fprintf(fp, "\t%s %s\n" , netPropsAux[i].name,NetProps[i].contents.prop.value.string);
	    netPropsAux[i].lastStringValue = strsav(NetProps[i].contents.prop.value.string);
	  }
	  break;
	case OCT_INTEGER:
	  if (octGetByName(&ChipNet,&NetProps[i]) != OCT_OK) {
	    NetProps[i].contents.prop.value.integer = netPropsAux[i].defaultIntegerValue;
	  }
	  if (netPropsAux[i].lastIntegerValue != NetProps[i].contents.prop.value.integer) {
	    fprintf(fp, "\t%s %d\n" , netPropsAux[i].name,NetProps[i].contents.prop.value.integer);
	    netPropsAux[i].lastIntegerValue = NetProps[i].contents.prop.value.integer;
	  }
	  break;
	case OCT_REAL:
	  if (octGetByName(&ChipNet,&NetProps[i]) != OCT_OK) {
	    NetProps[i].contents.prop.value.real = netPropsAux[i].defaultRealValue;
	  }
	  if (netPropsAux[i].lastRealValue != NetProps[i].contents.prop.value.real) {
	    fprintf(fp, "\t%s %lf\n" , netPropsAux[i].name,NetProps[i].contents.prop.value.real);
	    netPropsAux[i].lastRealValue = NetProps[i].contents.prop.value.real;
	  }
	  break;
	default:
	  FatalError("Wrong Net property type in printConstraintsFile()");
	  break;
	}
      }
      fprintf(fp,"NET %20s  # (with %d terms)\n", ChipNet.contents.net.name,net->size);
    }
  }
  (void) sprintf(errmsg,"%s has been created (or updated)",constraintsFileName);
  Message(errmsg);
  return(OK);
}



/*
 * Must be able to read this file before puppy has read all the data from oct
 */
int
scanConstraintsFile()
{
  char 			input[256]; /*  */
  char			string[256];
  char			name[256];
  int 			test;	/* Return value of the fscanf() function */
  int			integer; /*  */
  int			i;
  int			found;
  double 		real;	/*  */
  char			constraintsFileName[128];
  FILE 			*fp;

  
  (void) sprintf(constraintsFileName,"%s.constraintsPuppy", Chip.contents.facet.cell);
  if (( fp=fopen(constraintsFileName , "r")) == (FILE *) NULL) {
    (void) sprintf(errmsg,"Could not find %s", constraintsFileName);
    Message(errmsg);
    (void) strcpy(constraintsFileName,"constraintsPuppy");
    if (( fp=fopen(constraintsFileName , "r")) == (FILE *) NULL) {
      (void) sprintf(errmsg,"Cannot even find %s",constraintsFileName);
      Message(errmsg);
      return(NOT_OK);
    } 
  } 
  VOVinput( VOV_UNIX_FILE, constraintsFileName );
  (void) sprintf(errmsg,"Using %s file",constraintsFileName);
  Message(errmsg);

  initConstraints();

  while (fscanf(fp , " %s " , input) ==1) {
    /* The "#" sign at the beginning of a line indicates a comment in the constraintsFileName file */
    if  (strcmp( input , "#" ) == 0) { 
      while ( getc(fp) != '\n') ; /*   COMMENT: go on to the next line */
      continue;
    }
     
    if (strcmp(input,"INSTANCE") == 0) {
      test = fscanf( fp, " %s " , name);
      if (test != 1) {
	(void) sprintf(errmsg,"Expected a name  after %s","INSTANCE");
	FatalError(errmsg);
      }
      checkCell(name);
    } else    if (strcmp(input,"NET") == 0) {
      test = fscanf( fp, " %s " , name);
      if (test != 1) {
	(void) sprintf(errmsg,"Expected a name  after %s","NET");
	FatalError(errmsg);
      }
      checkNet(name);
    } else {
      /* Do a linear search by name on the property lists */
      found = FALSE;
      for (i = 0 ; i < totCellProps ; i++) {
	if (strcmp(input,cellPropsAux[i].name) == 0) {
	  switch (cellPropsAux[i].type) {
	  case OCT_STRING:
	    getLine(fp,string);
	    CellProps[i].contents.prop.value.string = strsav(string);
	    break;
	  case OCT_INTEGER:
	    test = fscanf( fp, " %d " , &integer);
	    if (test!=1) {
	      (void) sprintf(errmsg,"Expected an integer after %s",cellPropsAux[i].name);
	      FatalError(errmsg);
	    }
	    CellProps[i].contents.prop.value.integer = integer;
	    break;
	  case OCT_REAL:
	    test = fscanf( fp, " %lf " , &real);
	    if (test!=1) {
	      (void) sprintf(errmsg,"Expected a real after %s",cellPropsAux[i].name);
	      FatalError(errmsg);
	    }
	    CellProps[i].contents.prop.value.real = real;
	    break;
	  default:
	    break;
	  }
	  found = TRUE;
	  break;
	}
      }
      for (i = 0 ; i < totNetProps ; i++) {
	if (strcmp(input,netPropsAux[i].name) == 0) {
	  switch (netPropsAux[i].type) {
	  case OCT_STRING:
	    getLine(fp,string);
	    NetProps[i].contents.prop.value.string = strsav(string);
	    break;
	  case OCT_INTEGER:
	    test = fscanf( fp, " %d " , &integer);
	    if (test!=1) {
	      (void) sprintf(errmsg,"Expected an integer after %s",netPropsAux[i].name);
	      FatalError(errmsg);
	    }
	    NetProps[i].contents.prop.value.integer = integer;
	    break;
	  case OCT_REAL:
	    test = fscanf( fp, " %lf " , &real);
	    if (test!=1) {
	      (void) sprintf(errmsg,"Expected a real after %s",netPropsAux[i].name);
	      FatalError(errmsg);
	    }
	    NetProps[i].contents.prop.value.real = real;
	    break;
	  default:
	    FatalError("Should never happen in scanConstraintsFile");
	    break;
	  }
	  found = TRUE;
	  break;
	}
      }

      if (!found) {
	(void) sprintf(errmsg,"Unparsable word >>> %s found\n",input );
	Warning(errmsg);
      }
    }

  }
  return(OK);
}

static int
getLine(fp,string)
    FILE *fp;
    char *string;
{
  int i = 0;
    /* Go on to the end of the line */
  i = 0;
  while ((string[i] = getc(fp)) != '\n') i++;
  string[i] = '\0';
    
}

static int 
checkCell(name)
  char *name;
{
  octObject Instance;
  int	i;

  Instance.type = OCT_INSTANCE;
  Instance.contents.instance.name = name;
  if (octGetByName(&Chip,&Instance)!= OCT_OK) {
    (void) sprintf(errmsg,"No instance found with name %s", name);
    FatalError(errmsg);
  }
  
  for (i = 0 ; i < totCellProps ; i++) {
    if (octCreateOrModify(&Instance,&CellProps[i])!=OCT_OK) {
      FatalOctError("in checkCell()");
    }
  }
}

static int
checkNet(name)
  char *name;
{
  octObject ChipNet;
  int	i;


  ChipNet.type = OCT_NET;
  ChipNet.contents.net.name = name;
  if (octGetByName(&Chip,&ChipNet)!= OCT_OK) {
    (void) sprintf(errmsg,"No net found with name %s", name);
    FatalError(errmsg);
  }
  
  for (i = 0 ; i < totNetProps ; i++) {
    if (octCreateOrModify(&ChipNet,&NetProps[i])!=OCT_OK) {
      FatalOctError("in checkNet()");
    }
  }
}
