#include "defs.h"
#include "global.h"
#include "parall.h"


/* Two ways to compute the area cost. One is absolute,
 * The other uses the ratio between the minimum possible area and
 * the actual area.
 * The method can be selected at compile time defining
 * the switch USE_ACTUAL_AREA
 */

#define USE_ACTUAL_AREA

#ifndef USE_ACTUAL_AREA
double areaCost(areaRatio)
    double areaRatio;
{
  if (areaRatio > 20.0) {
    return(exp(20.0));
  } else {
    return(exp(areaRatio));
  }
}
#endif

CurAreaCost( costP )
     Cost *costP;			/* RETURN */
{
  double 	curChipArea();
  double 	areaCost();
#ifndef USE_ACTUAL_AREA
  double	areaRatio;
#endif

  /* Cost due to chip area */
#ifdef USE_ACTUAL_AREA
  costP->area = curChipArea();
#else
  areaRatio = curChipArea() / totalExpandedArea;
  costP->area = areaCost (areaRatio);
#endif
}



MoveAreaCost(moveP)
    Move *moveP;
{
  double 	curAreaRatio;
  double 	newAreaRatio;
  double 	curChipArea(), newChipArea();
  double	areaCost();

  if (SM_areaWeight == 0.0) {
    moveP->cost.area = 0.0;
    return;
  }
  /* cost due to the area of the chip */
  switch (moveP->type) {
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
  case TRANSLATE_AND_TRANSFORM:
  case TRANSFORMATION:
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
#ifdef USE_ACTUAL_AREA
    curAreaRatio = curChipArea();
    newAreaRatio = newChipArea(moveP);
    moveP->cost.area = newAreaRatio - curAreaRatio;
#else
    curAreaRatio = curChipArea() / totalExpandedArea;
    newAreaRatio = newChipArea(moveP) / totalExpandedArea;
    moveP->cost.area = areaCost(newAreaRatio) - areaCost(curAreaRatio);
#endif
    break;
  case TERMINAL_MOVE:
  case SUBNET_MOVE:
  case CHANGE_OWNERSHIP:
  default:
    moveP->cost.area = 0.0;
    break;
  }
}

