/*
 *  FILE: costWL.c
 *
 *  DESCRIPTION: computation of cost due to wire length
 *
 *
 *  CREATION DATE: Aug.21.1987          AUTHOR: Andrea Casotto
 *
 *  LAST MODIFICATION DATE: Jan.24.1991 AUTHOR: Edoardo Charbon
 *  
 *  FUNCTIONS:
 *
 * 
 *
 */


#include "defs.h"
#include "global.h"
#include "parall.h"

/* If a subnet connects two floating terminals, its cost should
 * be reduced so that weight to use for those nets including some  floating pins */

double floatingNetWeight = 0.01;
double powerNetWeight = 0.0;

/*
 * Total cost due the wires (nets) 
 */

CurWireLengthCost( costP )
     Cost *costP;		/* return */
{
    struct net	 	*NetP;
    int			estimatedLength;
    int			subnetLength;
    octBox		curBBox;
    Term		*term1,*term2;
    int			i;	/* index for subnets */

    /* primaryCost due to estimated wire length */
    costP->wireLength = 0.0;
    costP->conViol = 0;
    if (SM_netWeight <=  0.0) return;

    if (enableSubnets == TRUE) {
	NetP= netSet;
	while ((NetP=NetP->globalNext) != NULL) {
	    if (NetP->weight == 0) continue;
	    estimatedLength = 0;
	    for (i=0 ; i< NetP->subnets ; i++) {
		term1 = NetP->subnetArray[i].curTerm[0];
		term2 = NetP->subnetArray[i].curTerm[1];
		subnetLength = distance(term1->curAbsCoordP->x, 
					term1->curAbsCoordP->y,
					term2->curAbsCoordP->x, 
					term2->curAbsCoordP->y);
	
		if (NetP->maxLength && subnetLength > NetP->maxLength) {
		    costP->conViol +=  subnetLength - NetP->maxLength ;
		}
		if (term1->floating && term2->floating) {
		    estimatedLength += subnetLength * floatingNetWeight;
		} else {
		    estimatedLength += subnetLength;
		}
	    }
	    costP->wireLength += estimatedLength * NetP->weight  ;
	}
    } else {
	NetP= netSet;
	while ((NetP=NetP->globalNext) != NULL) {
	    if (NetP->weight == 0) continue;
	    curNetBBox(NetP,&curBBox);
	    estimatedLength = halfPerimeter(&curBBox);
	    if (NetP->maxLength && estimatedLength > NetP->maxLength) {
		costP->conViol +=  estimatedLength - NetP->maxLength;
	    }
	    costP->wireLength += estimatedLength * NetP->weight ;
	}
    }
    
    if (!considerPower) return;
    if (powerNetWeight == 0.0) return;
    
    NetP= powerNetSet;
    while (NetP != NULL) {
      if (NetP->weight == 0) continue;
      curNetBBox(NetP,&curBBox);
      estimatedLength = halfPerimeter(&curBBox);
      
#ifdef TO_BE_ADDED
      if (NetP->maxLength && estimatedLength > NetP->maxLength) {
	costP->conViol +=  estimatedLength - NetP->maxLength;
      }
#endif
      
      costP->wireLength += 
	estimatedLength * NetP->weight * NetP->normPeakCurrent;
      NetP = NetP->globalNext;
    }
    return;
    
}


MoveWireLengthCost(moveP)
    Move *moveP;
{

    if (enableSubnets == TRUE) {
	MoveSubnetLengthCost(moveP);
    } else {
	MoveNetBBoxCost(moveP);
    }
    movePowerLengthCost(moveP);  /* power buses are represented as 2 
				    terminal nets, therefore the same 
				    computation is required for both cases */

}


MoveNetBBoxCost(moveP)
    Move *moveP;
{
  int	 	curWireL = 0;
  int		newWireL = 0;
  int		conViol = 0;		/* constraints violations */
  int		cel ;			/* current estimated length */
  int		nel ;			/* new estimated length */
  octBox	curBBox;		/* Current bounding box	 */
  octBox	newBBox;		/* New bounding box */
  struct netList	*n1,*n2;	/* for the case of moveComplexity <= 2 */
  Net		*net;
  Term		*termP;			/* moving terminal */
  octCoord	left,right,top,bottom;
  struct termList *termPtr;
  octPoint	*newAbsCoordP;	/* of a floating terminal */


  /* new version: two pointers, one for cell, the other for cell2 */
  /* we run through the list of nets, processing the nets avoiding repetitions */
  
  if (moveP->type == SUBNET_MOVE) {
    Warning("Subnet move does not affect BBox cost ");
    curWireL = newWireL = 0;
  } else if (moveP->type == TERMINAL_MOVE) {

    /* Find BBox of net without the moving terminal */
    termP = moveP->term;
    net = termP->net;
    if (net->size > 1) {
      termPtr = net->term0.next;
      /* Skip the first if this is the moving terminal */
      if (termPtr->term == termP) {
	termPtr = termPtr->next;
	if (termPtr == NULL) {
	  sprintf(errmsg,
		  "Inconsistent data for net %s:\n 1 term found instaed of 2",
		  net->name);
	  FatalError(errmsg);
	}
      }
      left   =   termPtr->term->curAbsCoordP->x;
      bottom =   termPtr->term->curAbsCoordP->y;
      right  =   termPtr->term->curAbsCoordP->x;
      top    =   termPtr->term->curAbsCoordP->y;
      while ((termPtr=termPtr->next) != NULL) {
	if (termPtr->term != termP ) {
	  left   = MIN(left,   termPtr->term->curAbsCoordP->x);
	  bottom = MIN(bottom, termPtr->term->curAbsCoordP->y);
	  right  = MAX(right,  termPtr->term->curAbsCoordP->x);
	  top    = MAX(top,    termPtr->term->curAbsCoordP->y);
	}
      }

      newAbsCoordP =  &termP->cell->floatingSites[termP->newSite].curAbsCoord;
 
      curBBox.lowerLeft.x =  MIN(left,   termP->curAbsCoordP->x);
      curBBox.lowerLeft.y =  MIN(bottom, termP->curAbsCoordP->y);
      curBBox.upperRight.x = MAX(right,  termP->curAbsCoordP->x);
      curBBox.upperRight.y = MAX(top,    termP->curAbsCoordP->y);
      newBBox.lowerLeft.x =  MIN(left,   newAbsCoordP->x);
      newBBox.lowerLeft.y =  MIN(bottom, newAbsCoordP->y);
      newBBox.upperRight.x = MAX(right,  newAbsCoordP->x);
      newBBox.upperRight.y = MAX(top,    newAbsCoordP->y);

      cel = halfPerimeter(&curBBox);
      nel = halfPerimeter(&newBBox);
      if (net->maxLength) {
	if ( cel > net->maxLength ) conViol -= cel - net->maxLength;
	if ( nel > net->maxLength ) conViol += nel - net->maxLength;
      }
      curWireL += cel * net->weight;
      newWireL += nel * net->weight;
    }
  } else if (moveP->complexity <= 2) {
    n1 = moveP->cell->net0.next;
    if (moveP->cell2) {
      n2 = moveP->cell2->net0.next;
    } else {
      n2 = NULL;
    }

    while (n1 != NULL || n2 != NULL) {
      if (n1 == NULL) {		/* list 1 is finished */
	net = n2->net; n2 = n2->next; /* take net in list 2 */
      } else if (n2 == NULL) {	/* list 2 is finished */
	net = n1->net; n1 = n1->next; /* take net in list 1 */
      } else if (n1->net->number < n2->net->number) { /* net in list 1 less than net in list 2 */
	net = n1->net; n1 = n1->next; /* take net in list 1 */
      } else if (n2->net->number < n1->net->number) { /* net in list 1 less than net in list 2 */
	net = n2->net; n2 = n2->next; /* take net in list 2 */
      } else {			/* Same net. Advance both pointers */
	net = n1->net; n1 = n1->next; n2 = n2->next;
      }
      curNetBBox(net,&curBBox);
      newNetBBox(net,&newBBox);
      cel = halfPerimeter(&curBBox);
      nel = halfPerimeter(&newBBox);
      if (net->maxLength) {
	if ( cel > net->maxLength ) conViol -= cel - net->maxLength;
	if ( nel > net->maxLength ) conViol += nel - net->maxLength;
      }
      curWireL += cel * net->weight;
      newWireL += nel * net->weight;
    }

  } else {
    /* move complexity is > 2, must use a more complex algorithm to 
       pik the nets to process */
    Warning("Move complexity > 2: code not yet implemented");
  }
  moveP->cost.conViol = conViol;
  moveP->cost.wireLength = (newWireL - curWireL);
}



/* TIME CRITICAL ROUTINE !! */

MoveSubnetLengthCost(moveP)
    Move *moveP;
{
  int	 	curWireL = 0;
  int		newWireL = 0;
  int		conViol = 0;		/* constraints violations */
  int		curSubnetLength;
  int		newSubnetLength;
  struct net	*net;
  struct termList *termPtr;
  SubNet	*subnetP;
  Term		*termP;
  Term		*term2P;
  Term		*oldTermP, *newTermP;
  int		movingTermIndex;
  MacroCell	*cellP;
  ListElt	*subnetLE;	/* subnet-list element */
  int		floatingDone;
  octPoint	*newAbsCoordP;	/* of a floating terminal */
  
  if (moveP->type == TERMINAL_MOVE) {
    /* One terminal is moving, recompute the subnets attached to it */
    termP = moveP->term;
    net = termP->net;
    newAbsCoordP =  
      &termP->cell->floatingSites[termP->newSite].curAbsCoord; 
    for (subnetLE = termP->firstSubnet; subnetLE != NULL ; subnetLE=subnetLE->next) {
      subnetP = subnetLE->pointer.subnet;
      term2P = 
	(subnetP->curTerm[0] == termP)? subnetP->curTerm[1] : subnetP->curTerm[0];
      if ( termP == term2P ) {
	sprintf( errmsg,
		"BUG:Subnet attached twice to the same terminal (address %d)",
		(int)termP);
	Warning( errmsg );
	Warning( "This is a known bug.Continuing anyway");
	continue;
      }
      curSubnetLength = distance(termP->curAbsCoordP->x, 
				 termP->curAbsCoordP->y,
				 term2P->curAbsCoordP->x, 
				 term2P->curAbsCoordP->y);

      newSubnetLength = distance(newAbsCoordP->x, 
				 newAbsCoordP->y,
				 term2P->curAbsCoordP->x , 
				 term2P->curAbsCoordP->y);
      
      if (net->maxLength) {
	if ( curSubnetLength > net->maxLength ) 
	  conViol -=  curSubnetLength - net->maxLength;
	if ( newSubnetLength > net->maxLength ) 
	  conViol +=  newSubnetLength - net->maxLength;
      }
      /* We know that termP is floating, check term2P */
      if (term2P->floating) {
	curWireL += curSubnetLength * floatingNetWeight * net->weight;
	newWireL += newSubnetLength * floatingNetWeight * net->weight;
      } else {
	curWireL += curSubnetLength * net->weight;
	newWireL += newSubnetLength * net->weight;
      }
    }
  } else if (moveP->type == SUBNET_MOVE) {
    subnetP = moveP->subnet;
    net = subnetP->curTerm[0]->net;
    movingTermIndex = moveP->subnetIndex;
    
    termP = subnetP->curTerm[1-movingTermIndex];
    oldTermP = subnetP->curTerm[movingTermIndex];
    newTermP = subnetP->newTerm[movingTermIndex];
    
    curSubnetLength = distance(termP->curAbsCoordP->x, 
			       termP->curAbsCoordP->y,
			       oldTermP->curAbsCoordP->x, 
			       oldTermP->curAbsCoordP->y);
 
    newSubnetLength = distance(termP->curAbsCoordP->x, 
			       termP->curAbsCoordP->y,
			       newTermP->curAbsCoordP->x, 
			       newTermP->curAbsCoordP->y);
 
    if (net->maxLength) {
      if ( curSubnetLength > net->maxLength ) 
	conViol -= curSubnetLength - net->maxLength;
      if ( newSubnetLength > net->maxLength ) 
	conViol += newSubnetLength - net->maxLength;
    }
    if (termP->floating && oldTermP->floating) {
      curWireL = curSubnetLength * floatingNetWeight * net->weight;
    } else {
      curWireL = curSubnetLength * net->weight;
    }
    if (termP->floating && newTermP->floating) {
      newWireL = newSubnetLength * floatingNetWeight *net->weight;
    } else {
      newWireL = newSubnetLength * net->weight;
    }
  } else {
    /* 
     ** For every moving cell and for every terminal in the cell and for 
     ** every subnet attached to the terminal ... 
     */
    for (cellP = moveP->cell; cellP != NULL ; cellP = cellP->moveNext) {
      /* Do the fixed terminal first, then the floating terminals */
      termPtr = cellP->term0.next;
      floatingDone = FALSE;
      while (TRUE) {	/* exit controlled by break */
	if (termPtr == NULL ) {
	  if (floatingDone == TRUE || cellP->floatingTerms == 0 ) break;
	  termPtr = cellP->floatTerm0.next; /* go through all floating terms*/
	  floatingDone = TRUE;
	}
	termP = termPtr->term;
	if (termP->status == USED) {
	  net = termP->net;
	  for (subnetLE = termP->firstSubnet; subnetLE != NULL ; subnetLE=subnetLE->next) {
	    subnetP = subnetLE->pointer.subnet;
	    term2P = (subnetP->curTerm[0] == termP)? subnetP->curTerm[1] : subnetP->curTerm[0];
	    /* Some subnets are encountered twice. Is this one of those ? */
	    if (term2P->cell->moveId == myMoveId && term2P < termP) {
	      /*
	       ** This net is encountered twice in this loop, but its 
	       ** contribution must be added only once. Therefore, skip 
	       ** the subnet if the address term2P is less than the address 
	       ** termP.
	       */
	      continue;
	    }
	    curSubnetLength = distance(termP->curAbsCoordP->x, 
				       termP->curAbsCoordP->y,
				       term2P->curAbsCoordP->x, 
				       term2P->curAbsCoordP->y);
	    
	    if (term2P->cell->moveId != myMoveId) {
	      newSubnetLength = distance(termP->newAbsCoordP->x, 
					 termP->newAbsCoordP->y,
					 term2P->curAbsCoordP->x , 
					 term2P->curAbsCoordP->y);
	    } else {
	      /* Both terminals are moving in the same move */
	      newSubnetLength = distance(termP->newAbsCoordP->x, 
					 termP->newAbsCoordP->y,
					 term2P->newAbsCoordP->x, 
					 term2P->newAbsCoordP->y);
	    }
	    
	    if (net->maxLength) {
	      if ( curSubnetLength > net->maxLength ) 
		conViol -= curSubnetLength - net->maxLength;
	      if ( newSubnetLength > net->maxLength ) 
		conViol += newSubnetLength - net->maxLength;
	    }
	    if (termP->floating && term2P->floating) {
	      curWireL += curSubnetLength * floatingNetWeight * net->weight;
	      newWireL += newSubnetLength * floatingNetWeight * net->weight;
	    } else {
	      curWireL += curSubnetLength * net->weight;
	      newWireL += newSubnetLength * net->weight;
	    }
	  }
	}
	termPtr = termPtr->next;
      }
    }
  }
  moveP->cost.conViol = conViol;
  moveP->cost.wireLength = (newWireL - curWireL);
}


/* Return the manhattan distance between two points */
/* TIME CRITICAL ROUTINE !! */

int distance( x1, y1, x2, y2)
    register int x1,y1,x2,y2;
{
  register int lx = x1 - x2;
  register int ly = y1 - y2;

  if (lx < 0) lx = -lx;
  if (ly < 0) ly = -ly;
  return lx + ly;
}


movePowerLengthCost(moveP)
  Move *moveP;
{

  Term *termP, *term2P;
  MacroCell *cellP;
  struct termList *termPtr;
  double curNetLength = 0.0;
  double newNetLength = 0.0;
  double curWL = 0.0;
  double newWL = 0.0;
  double conViol = 0.0;
  Net *net;
  
  if (!considerPower) return;
  if (powerNetWeight == 0.0) return;

  termP = moveP->term;
  net = moveP->net; 

  if (moveP->type == TERMINAL_MOVE) {
    if (termP->type == SUPPLY) return;
    term2P = net->padTerm;
    if (termP == term2P) return;
    
    curNetLength = distance(termP->curAbsCoordP->x, 
			    termP->curAbsCoordP->y,
			    term2P->curAbsCoordP->x, 
			    term2P->curAbsCoordP->y);
    
    newNetLength = distance(termP->newAbsCoordP->x, 
			    termP->newAbsCoordP->y,
			    term2P->curAbsCoordP->x , 
			    term2P->curAbsCoordP->y);
    
    newWL = newNetLength * net->weight * net->normPeakCurrent;
    curWL = curNetLength * net->weight * net->normPeakCurrent;

#ifdef TO_BE_ADDED  
    if (net->maxLength) {
      if ( curNetLength > net->maxLength ) 
	conViol -=  curSubnetLength - net->maxLength;
      if ( newNetLength > net->maxLength ) 
	conViol +=  newSubnetLength - net->maxLength;
    }
#endif

  } else if (moveP->type == SUBNET_MOVE) {
    return;
  } else {
    for (cellP = moveP->cell; cellP != NULL ; cellP = cellP->moveNext) {
      termPtr = cellP->term0.next;
      while (termPtr != NULL) {
	termP = termPtr->term;
	net = termP->net;
    	if (termP->type == SUPPLY) {
	  term2P = net->padTerm;
	  curNetLength = distance(termP->curAbsCoordP->x, 
				  termP->curAbsCoordP->y,
				  term2P->curAbsCoordP->x, 
				  term2P->curAbsCoordP->y);
	  
	  newNetLength = distance(termP->newAbsCoordP->x, 
				  termP->newAbsCoordP->y,
				  term2P->curAbsCoordP->x , 
				  term2P->curAbsCoordP->y);
	  
	  newWL += newNetLength * net->weight * termP->normPeakCurrent;
	  curWL += curNetLength * net->weight * termP->normPeakCurrent;

#ifdef TO_BE_ADDED
	  if (net->maxLength) {
	    if ( curNetLength > net->maxLength ) 
	      conViol -= curNetLength - net->maxLength;
	    if ( newNetLength > net->maxLength ) 
	      conViol += newNetLength - net->maxLength;
	  }
#endif
	}
	termPtr = termPtr->next;
      }
    }
  }

  moveP->cost.conViol += conViol;
  moveP->cost.wireLength += (newWL - curWL);
  return;

}
