#include "defs.h"
#include "global.h"
#include "fang.h"
#include "harpoon.h"

int
readCellGeometry(cell)
    MacroCell *cell;
{
    octPoint		lowerLeft;
    octPoint		upperRight;
    octBox		BBox;	/* bounding box of the Master */
    int 		I;    /* counter */
    int			llc;	/* temporary for lowestLeftCorner */
    harList		har_perimList; /* to hold the perimeter of a Master */
    octPoint		*tmpPerim; /* temporary perimeter of a cell */
    octObject 		Instance;
    octObject 		Master;

    Instance.objectId = cell->cellId;
    Master.objectId = cell->masterId;

    if (octGetById(&Instance)!=OCT_OK) FatalOctError("Corrupted cell instance id");
    if (octGetById(&Master)!=OCT_OK) FatalOctError("Corrupted cell master id");


    /* This function is void */
    harFacetToPerimList(&Master, HAR_INFINITY, OCT_LAMBDA*100, &har_perimList);
    
    if ((harList *) harRestOfList(har_perimList) != NIL(harList)) {
	(void) sprintf(errmsg,"More than 1 geometry to cover %s:%s", 
		       Master.contents.facet.cell,Master.contents.facet.view);
	Warning(errmsg);
    }
    /* get the information from HARPOON,copy it in shared memory and then free storage */
    tmpPerim	     = harFirstPerimArray(har_perimList);
    cell->edgesInPerim = harFirstPerimCount(har_perimList);
    
    cell->perimeter    = SH_MALLOC(octPoint, cell->edgesInPerim);
    cell->curPerimeter = SH_MALLOC(octPoint, cell->edgesInPerim);
    cell->newPerimeter = SH_MALLOC(octPoint, cell->edgesInPerim);
    cell->termsInEdge  = SH_MALLOC(int ,     cell->edgesInPerim);
    cell->widthOfEdge  = SH_MALLOC(int ,     cell->edgesInPerim);

    for (I=0;I<cell->edgesInPerim;I++) {
	cell->termsInEdge[I] = 0; /* Reset the array */
	cell->widthOfEdge[I] = 0; /* Reset the array */
	cell->perimeter[I].x = tmpPerim[I].x;
	cell->perimeter[I].y = tmpPerim[I].y;
    }
    harFreeList(&har_perimList); /* free all storage used by HARPOON */
    maxEdgesInPerim=MAX(maxEdgesInPerim,cell->edgesInPerim);

    /*
     * Scale the perimeter down by OCT_LAMBDA and determine the lowest left corner
     */
    cell->perimeter[0].x /= OCT_LAMBDA;
    cell->perimeter[0].y /= OCT_LAMBDA;
    BBox.lowerLeft.x = cell->perimeter[0].x;
    BBox.lowerLeft.y = cell->perimeter[0].y;
    BBox.upperRight.x = cell->perimeter[0].x;
    BBox.upperRight.y = cell->perimeter[0].y;
    llc = 0;			/* lowestLeftCorner */
    for (I = 1 ; I< cell->edgesInPerim ; I++) {
	cell->perimeter[I].x /= OCT_LAMBDA;
	cell->perimeter[I].y /= OCT_LAMBDA;
	if (cell->perimeter[I].y < cell->perimeter[llc].y ||
	    cell->perimeter[I].y == cell->perimeter[llc].y &&
	    cell->perimeter[I].y < cell->perimeter[llc].y ) {
	    llc = I;
	}
	BBox.lowerLeft.x = MIN( BBox.lowerLeft.x , cell->perimeter[I].x);
	BBox.lowerLeft.y = MIN( BBox.lowerLeft.y , cell->perimeter[I].y);
	BBox.upperRight.x = MAX( BBox.upperRight.x , cell->perimeter[I].x);
	BBox.upperRight.y = MAX( BBox.upperRight.y , cell->perimeter[I].y);
    }
    cell->lowestLeftCorner = llc;
    /* save the coordinates of this corner to simplify the "saving to OCT" process */
    cell->masterLLC.x = cell->perimeter[llc].x;
    cell->masterLLC.y = cell->perimeter[llc].y;

    lowerLeft.x = BBox.lowerLeft.x;
    lowerLeft.y = BBox.lowerLeft.y;
    upperRight.x = BBox.upperRight.x;
    upperRight.y = BBox.upperRight.y;
    cell->W =  upperRight.x - lowerLeft.x;
    cell->H =  upperRight.y - lowerLeft.y;

    cell->rW =  upperRight.x - lowerLeft.x;
    cell->rH =  upperRight.y - lowerLeft.y;
    cell->rHalfW = cell->rW >> 1;
    cell->rHalfH = cell->rH >> 1;
    cell->curCenter.x = (cell->W >> 1) + BBox.lowerLeft.x;
    cell->curCenter.y = (cell->H >> 1) + BBox.lowerLeft.y;
    /* 
     * Init the variables to take into account the fact that 
     * due to asymmetric expansion of the cell, the center of the cell
     * might be offset
     */
    cell->centerOffset.x = 0;
    cell->centerOffset.y = 0;
    cell->curBbox.lowerLeft.x = cell->curCenter.x - cell->rHalfW;
    cell->curBbox.lowerLeft.y = cell->curCenter.y - cell->rHalfH;
    cell->curBbox.upperRight.x = cell->curBbox.lowerLeft.x + cell->rW;
    cell->curBbox.upperRight.y = cell->curBbox.lowerLeft.y + cell->rH;


    cell->area = (double) cell->W * (double) cell->H;
    if (cell->area == 0.0) FatalError("Area of one cell is zero");

    /* convert the perimeter so that it refers to the center of the cell */
    for (I = 0 ; I< cell->edgesInPerim ; I++) {
	cell->perimeter[I].x -= cell->curCenter.x;
	cell->perimeter[I].y -= cell->curCenter.y;
    }
    return OK;
}
