/*
 *  FILE: global.h
 *
 *  DESCRIPTION: declaration of all global variables
 *
 *
 *  CREATION DATE: Jan 1987		AUTHOR: Andrea Casotto
 *  LAST MODIFICATION: Oct 1990         BY: Edoardo Charbon
 *
 */


#include "parall.h"
#include <X11/Xlib.h>
#include "sa.h"

/*  GLOBAL VARIABLE DECLARATION */
extern  char            usingParameterFile[128];/* parameter file used */
extern  long 		mypid;			/* system process identification */
extern  int		myself;			/* internal identification */
extern	int		myMoveId;		/* identifier for moves attempted by myself */
extern  struct ProcessorFrame  *myProc;		/*  */
#ifdef PARALLEL
shared extern  slock_t	lyoLock;		/*  */
shared extern  slock_t  waitLock;		/*  */
shared extern  slock_t	subnetMoveLock;
#endif
extern  int  		P;			/* number of processes used */

SHARED extern int 	debug;			/* debug flag */
SHARED extern int	saveInterval;
SHARED extern int	useQualityFactor;	/* To determine the frequency of the moves */
SHARED extern int	useCellBasedMoveStats;
SHARED extern int 	enableTranslations; 	/* To enable the translations */
SHARED extern int 	enableTransformations; 	/* To enable the transformations */
SHARED extern int	enableCenterInterchange; /*  */
SHARED extern int	enableSubnets;		/*  */
SHARED extern int       considerPower;          /* do power optimization */
SHARED extern octCoord  lambda;                 /* oct units / lambda */
SHARED extern double    maxCurrDensity;         /*  */
SHARED extern octCoord  minWireWidth;           /*  */
SHARED extern int	FloorplanMode;		/*  */
SHARED extern int	floatingTermsOnly;	/*  */
SHARED extern int	costOnly;		/*  */
extern char             *progName;              /* global program name */ 
extern char             *output_cell;           /* global cell name */

extern int  		graphics;		/* flag activating graphing output */
extern int		XDisplayActive;		/*  */
extern int		showBB;			/* if TRUE, show the BBox of the cell */
extern int		showAxes;		/* if TRUE, show the symmetry axes */
extern int		showWell;		/* if TRUE, show the well boundary */
extern int		showConstraintNodes;	/* if TRUE, show critical nodes */
extern int		showBias;		/* if TRUE, show the bias of the cell */
extern int		showNets;		/* if TRUE, show all the nets */
extern int		showSubNets;		/* if TRUE, show all the sub-nets */
extern int		showBins;		/* if TRUE, show the bins used for overlap */
extern int		showOwner;		/* if TRUE, show the owner of the cells */
extern int		showAge;		/* if TRUE, show the age of the cells */
extern char		errmsg[];		/*  */
extern long		randVar;		/*  */

extern octObject	Chip;			/*  */
SHARED extern ListElt	*ListOfAnnealVars; 	/*  */
SHARED extern MacroCell *gFirstCell;		/*  */
SHARED extern Net *netSet;		        /* list of all the nets	*/
SHARED extern Net *powerNetSet;		        /* list of all power nets */

extern int 		maxNets;		/* max # of nets in a cell */
extern int 		maxEdgeX;		/* used in overlap() */
extern int 		maxEdgeY;		/*  */
extern int		maxEdgesInPerim;	/* max number of edges in perimeter of a cell */

SHARED extern int 	numberOfCells;		/*  */
SHARED extern int 	numberOfMovingCells; 	/*  */
SHARED extern int 	numberOfNets;		/*  */
SHARED extern int	numberOfFloatingPins;   /*  */
SHARED extern int	numberOfSubnets;	/*  */
SHARED extern int	numberOfAnnealVars; 	/*  */

extern double 		maxArea;		/*  */
extern double 		totalArea;		/*  */
extern double		totalExpandedArea;
SHARED extern MacroCellPointer *allCellsArray;	/*  */
SHARED extern MacroCellPointer *varCellsArray;	/*  */
 

extern double 		desAspectRatio;		/*  */
extern double		automaticChipArea; 	/*  */
extern double		cellExpansion;		/*  */

SHARED extern enum ChipBoundaryType chipBoundaryType; /*  */
SHARED extern octBox	*chipBbox;		/*  */
extern int 		minNumberOfCellsPerProc; /*  */
extern int 		maxNumberOfCellsPerProc; /*  */
extern long 		moveAttempted;		/* used in Sim. Ann.		*/
extern long 		moveAccepted;		/* used in Sim. Ann.		*/
extern double 		rngLimT;		/* used in Sim. Ann.		*/
extern long 		totalNumberOfMoves;	/*  */
extern long 		totalNumberOfAcceptedMoves; /*  */

/* Global arrays for move statistics and selection and their lock*/
SHARED extern struct moveSelect gMoveSelArray[];
SHARED extern struct moveStat   gMoveStatArray[];
SHARED extern MoveSet gCellMoveSet;
SHARED extern MoveSet gNetMoveSet;
SHARED extern MoveSet gFloatTermMoveSet;
#ifdef PARALLEL
SHARED extern slock_t moveStatArrayLock;	/* to control the access to moveStatArray */
#endif

/* External variables for the bins ruotines */
SHARED extern int	 	numBinX,numBinY;	/*  */
SHARED extern octCoord		binWidthX,binWidthY;	/*  */
SHARED extern octCoord		binOffsetX,binOffsetY;	/*  */
SHARED extern int	***blockArray;		/*  */
SHARED extern int	*bucket;		/*  */

/* Shared variables for SA */
SHARED extern   double 	SM_T;			/*  */
SHARED extern	double	SM_startT;		/*  */
SHARED extern	double	SM_minimumTemperature;	/*  */
SHARED extern	double	SM_alpha;		/*  */
SHARED extern   double	SM_rngLimT;		/*  */
SHARED extern   octCoord gRangeLimiter;		/*  */
SHARED extern   octCoord gRangeLimiterX;	/*  */
SHARED extern   octCoord gRangeLimiterY;	/*  */
SHARED extern	int	keepWholeCellInsideBoundary; /*  */
SHARED extern   int	SM_shakeStrength;	/*  */
SHARED extern   int	SM_shakeX;		/*  */
SHARED extern   int	SM_shakeY;		/*  */
SHARED extern   int	movesPerAnnVar;		/*  */
SHARED extern   double	SM_totalCost;		/*  */
SHARED extern   double	SM_bestCost;		/*  */
SHARED extern   int 	SM_somebodyIsWaiting;	/* flag */
SHARED extern   double	SM_ovWeight;		/*  */
SHARED extern   double	SM_netWeight;		/*  */
SHARED extern   double	SM_areaWeight;		/*  */
SHARED extern   double	SM_conViolWeight;	/*  */
SHARED extern   double	SM_clusterWeight;	/*  */
SHARED extern   int 	SM_minNumberOfCellsPerProc; /*  */
SHARED extern   int 	SM_maxNumberOfCellsPerProc; /* not yet used */
SHARED extern   int	SM_minNumberOfAnnealVarsPerProc; /*  */
SHARED extern   int	SM_maxNumberOfAnnealVarsPerProc; /*  */


SHARED extern struct PFrame 	PFrame;		/*  */
SHARED extern struct ProcessorFrame *process;	/*  */

/* Variables for output to X display */
extern char    		*displayName;		/*  */
extern Display 		*display;		/*  */
extern Window 		consoleWin;		/* where the current status of program is shown */
extern Window		placementWin;		/* where the cells are displayed */
extern Window		graphWin;		/* for graphics */

extern SHARED JobQueue	*outerLoopJobQueue;	/*  */



