/*
 *    'grafici.c'
 *    This file contains all the routines that deal with the 
 *	X window manager.
 *
 *    ROUTINES CONTAINED IN THIS MODULE:
 *	setColors()
 *	plotCells()
 * 	createMonWindow()
 *	showWaiting()
 *	showWorking()
 */

#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "constraintDrivenVariables.h"

extern void plotCells();


octCoord	localOriginX;
octCoord	localOriginY;
double		scale;
int		globalOffsetX;
int		globalOffsetY;

#define PLOTX(a) (((a) - localOriginX) * scale + globalOffsetX)
#define PLOTY(a) ((localOriginY - (a)) * scale + globalOffsetY)


extern Window	placementWin;	/* plotting window */

/* We need at least 30 colors, because that is the max number of 
   processors in current models of Sequent computers */
char *procColorName[] = {
    "SeaGreen",			/* 1 */
    "MediumBlue",
    "salmon",
    "aquamarine",
    "tan",
    "NavyBlue",
    "wheat",
    "coral",
    "NavyBlue",
    "firebrick",		/* 10 */
    "brown",
    "CadetBlue",
    "magenta",
    "sandybrown",
    "forest green",
    "lime green",
    "medium forest green",
    "pale green",
    "SeaGreen",
    "yellow green",		/* 20 */
    "YellowGreen",
    "khaki",
    "medium orchid",
    "plum",
    "MidnightBlue",
    "indian red",
    "MediumVioletRed",
    "salmon",
    "sienna",
    "tan",			/* 30 */
    0
  };

int	ageStep = 10;		/* used to discriminate various age ranges */
int 	*myPixel;		/* array of pixel */

int numOfColors;


SHARED octBox  coreBbox ;


int setColors()
{
    int 	i = 0;		/* counter */
    extern int	bgPlacementPixel;

    i = numOfColors = 0;
    while ( procColorName[i] != (char*)0 ) {
	i = ++numOfColors;
    }
    numOfColors--;

    myPixel = CALLOC( int, numOfColors );

    i = 0;
    while ( procColorName[i] != (char*)0 ) {
	myPixel[i] = puppyGetColor( procColorName[i] ) ^ bgPlacementPixel;
	i++;
    }


    return OK;
}



void setCoreBbox( box )
    octBox *box;
{
    coreBbox.lowerLeft.x =  box->lowerLeft.x;
    coreBbox.lowerLeft.y =  box->lowerLeft.y;
    coreBbox.upperRight.x =  box->upperRight.x;
    coreBbox.upperRight.y =  box->upperRight.y;
}

int PM_resize()
{
    setCoreBbox( chipBbox );
    plotCells();
    refreshConsoleWindow();
    return OK;
}


/* This procedure scans all the axes in the global data structure SymCell,
 * and it plotted the symmetry axes present in it */
 
void readAndDrawAxisPosition(SymCell)
    SymmetricCell *SymCell;
 
{
    int                    m,l;                           /* indices */   
    int                    numberOfAxes;
    SymmetryType           typeOfAxis;
    octCoord               positionOfAxis;
    static int xAxisPixel;     /* color of the x-axis */  
    static int yAxisPixel;     /* color of the y-axis */
    static GC  axisGC = 0;
    short      pixel; 
    octCoord  x1, x2, y1, y2;   

    if (axisGC == 0) {
        xAxisPixel = puppyGetColor( "yellow" );
        yAxisPixel = puppyGetColor( "red" );
        axisGC = XCreateGC( display, placementWin, 0, 0 );
        XSetFunction( display, axisGC, GXcopy );
        XSetPlaneMask( display, axisGC, AllPlanes );
        XSetLineAttributes( display, axisGC, 2, LineOnOffDash, CapNotLast, JoinMiter );
    }    
    if (SymCell->numberOfSymCell == 0)  {
        return;
    }
    numberOfAxes = SymCell->numberOfAxis;
    typeOfAxis = SymCell->axisType;
    for (m = 0; m < numberOfAxes; m++)  {
        positionOfAxis = SymCell->Axis[m].axisPosition;
        if (typeOfAxis == VERTICAL) {
            pixel = yAxisPixel;
            y1 = PLOTY( chipBbox->lowerLeft.y );
            y2 = PLOTY( chipBbox->upperRight.y );
            x1 = PLOTX(positionOfAxis);
            x2 = PLOTX(positionOfAxis);
        } else {
            pixel = xAxisPixel;
            x1 = PLOTX( chipBbox->lowerLeft.x );
            x2 = PLOTX( chipBbox->upperRight.x );
            y1 = PLOTY(positionOfAxis); 
            y2 = PLOTY(positionOfAxis);
        }
        XSetForeground( display, axisGC, pixel );
        XDrawLine( display, placementWin, axisGC, x1, y1, x2, y2 );
    }
    XFlush( display );

}


void drawSymmetryAxis()
 
{
    readAndDrawAxisPosition(gxSymCell);
    readAndDrawAxisPosition(gySymCell);
}


void drawWellAxis()
   
{
    static GC  axisGC = 0;
    short      pixel; 
    octCoord  x1, x2, y1, y2;   
    static int wellAxisPixel;     /* color of the well boundary */
 
    if (axisGC == 0) {
        wellAxisPixel = puppyGetColor( "brown" );
        axisGC = XCreateGC( display, placementWin, 0, 0 );
        XSetFunction( display, axisGC, GXcopy );
        XSetPlaneMask( display, axisGC, AllPlanes );
        XSetLineAttributes( display, axisGC, 2, LineOnOffDash, CapNotLast, JoinMiter );
    }    
    pixel = wellAxisPixel;
    x1 = PLOTX( chipBbox->lowerLeft.x );
    x2 = PLOTX( chipBbox->upperRight.x );
    y1 = PLOTY( wellBoundary ); 
    y2 = PLOTY( wellBoundary );
    XSetForeground( display, axisGC, pixel );
    XDrawLine( display, placementWin, axisGC, x1, y1, x2, y2 );
    XFlush( display );  

}


void drawCells( frameGC)
    GC frameGC;
{
    static int	padPixel;	/* color for the pads */
    static int	cellPixel;	/* color of the cells */
    static int	softPixel;	/* color of softcells */
    static int  symmetricPixel; /* color of the symmetric cells */
    static int  matchedPixel;   /* color of matched cells */
    static int  bothPixel;      /* color of matched & symmetric cells */
    static GC   cellGC = 0;
    static XPoint *Xpoints = 0;
    int		numPoints;
    short 	pixel;
    int 	I;		/* counter */
    MacroCell	*cell;
    octCoord	left,right,top,bottom;
    int		x,y,width,height; /* of a rectangle */
    octPoint    biasLabelPosition;
    int 	lastPixel = -1;	/* init to an invalid pixel value */

    if ( !Xpoints ) {
	Xpoints = CALLOC( XPoint , maxEdgesInPerim + 1);
	cellPixel = myPixel[7];
	padPixel  = myPixel[4];
	softPixel = myPixel[6];
        symmetricPixel = myPixel[1];
        matchedPixel = myPixel[2];
        bothPixel = myPixel[3];
	cellGC = XCreateGC( display, placementWin, 0, 0 );
	XSetFunction( display, cellGC, GXxor );
	XSetPlaneMask( display, cellGC, AllPlanes );
	XSetLineAttributes( display, cellGC, 1, LineSolid, CapNotLast, JoinMiter );

    }

    for ( cell = gFirstCell->globalNext ; cell != 0 ; cell = cell->globalNext ) {
      	if (showOwner) {
	    pixel = myPixel[cell->owner];
	} else if (showAge) {
	    pixel = myPixel[MIN(cell->age / ageStep, 11)];
	} else {
	    if (cell->isPad) {
		pixel = padPixel;
	    } else  if ( cell->floatingTerms > 0 ) {
		pixel = softPixel;
	    } else  if ( cell->analogStatus == SYMMETRIC ) {
                pixel = symmetricPixel;  
	    } else  if ((cell->analogStatus == MATCHED) || (cell->analogStatus == MULTIPLE_MATCHED)) {
                pixel = matchedPixel;
	    } else  if ((cell->analogStatus == BOTH) || (cell->analogStatus == BOTH_MULTIPLE)) {
                pixel = bothPixel;
	    } else  if ( cell->analogStatus == NONE ) {
	        pixel = cellPixel;
	    } else {
                pixel = cellPixel; 
	      }
	}
	if (pixel != lastPixel) {
	    lastPixel = pixel;
	    XSetForeground( display, cellGC, pixel );
	}

	x =  PLOTX(cell->curBbox.lowerLeft.x);
	y =  PLOTY(cell->curBbox.upperRight.y);
	
	if (showBB) {
	    width = PLOTX( cell->curBbox.upperRight.x) - x;
	    height = PLOTY(cell->curBbox.lowerLeft.y) - y;
	    XFillRectangle( display, placementWin, cellGC, x, y, width, height );
	    XDrawRectangle( display, placementWin, frameGC, x, y, width, height );
	    biasLabelPosition.x =  PLOTX(cell->curBbox.lowerLeft.x);
	    biasLabelPosition.y =  PLOTY(cell->curBbox.lowerLeft.y);

	} else {		/* show actual perimeter */
	    numPoints = cell->edgesInPerim + 1 ;
	    for (I=0; I < cell->edgesInPerim; I++) {
		Xpoints[I].x = PLOTX(cell->curPerimeter[I].x );
		Xpoints[I].y = PLOTY(cell->curPerimeter[I].y );
	    }
	    Xpoints[numPoints-1] = Xpoints[0];
	    XDrawLines( display, placementWin, frameGC, Xpoints, numPoints , CoordModeOrigin );
	    XFillPolygon( display, placementWin, cellGC, Xpoints, numPoints, Nonconvex, CoordModeOrigin );
	    biasLabelPosition.x =  PLOTX(cell->curPerimeter[0].x);
	    biasLabelPosition.y =  PLOTY(cell->curPerimeter[0].y);
	}

        if (showBias) {
            switch (cell->biasStatus) {
            case VDD:
              pwprint("p",1,biasLabelPosition.x,biasLabelPosition.y);
              break;
            case VSS:
              pwprint("n",1,biasLabelPosition.x,biasLabelPosition.y);
              break;
            case INDEPENDENT:
              pwprint("i",1,biasLabelPosition.x,biasLabelPosition.y);
              break;
            case SELF:
              pwprint("s",1,biasLabelPosition.x,biasLabelPosition.y);
              break;
            default:
              FatalError("Wrong biasStatus");
              break;
            }
        }
    }
    XFlush( display );
    /* VOID function */
}

/* Plot multiple terminal nets only if subnets are enabled */
void drawNets()
{
    static int	net2Pixel;		/* color for 2 terminals nets */
    static int	net3Pixel;		/* color for nets with 3 or more terminals */
    static int	netCritPixel;		/* color for critical nets */
    static GC   netGC = 0;

    short 	pixel;
    struct net	*netPtr;
    Term		*termP;
    struct termList *termPtr;
    int		x1,y1,x2,y2;	/* used to draw 2 terminals nets */

    if ( ! showNets ) return ;

    if (netGC == 0) {
	net2Pixel = puppyGetColor( "Blue" );
	net3Pixel = puppyGetColor( "cyan" );
	netCritPixel = puppyGetColor( "red" );
	netGC = XCreateGC( display, placementWin, 0, 0 );
	XSetFunction( display, netGC, GXcopy );
	XSetPlaneMask( display, netGC, AllPlanes );
	XSetLineAttributes( display, netGC, 1, LineSolid, CapNotLast, JoinMiter );
    }

    for ( netPtr= netSet;  netPtr != 0; netPtr = netPtr->globalNext ) {
	if ( netPtr->weight == 0 ) continue; /* Skip irrelevant nets. */
	if (netPtr->size == 2) {
	    struct termList *termPtr = netPtr->term0.next;
	    Term* t1 = termPtr->term;
	    Term* t2 = termPtr->next->term;

	    pixel =  (netPtr->weight > 1 || netPtr->maxLength || 
		      netPtr->constraintStatus == CONSTRAINED) ? netCritPixel : net2Pixel;
	    if (!showConstrainedNets || showConstrainedNets && netPtr->specStatus == NOT_MET) {
	      XSetForeground( display, netGC, pixel );
	      x1 = PLOTX(t1->curAbsCoordP->x);
	      y1 = PLOTY(t1->curAbsCoordP->y);
	      x2 = PLOTX(t2->curAbsCoordP->x);
	      y2 = PLOTY(t2->curAbsCoordP->y);
	      XDrawLine( display, placementWin, netGC, x1,y1, x2,y2 );
	    }
	} else if (enableSubnets) {
	    int I;		/* counter */

	    pixel =  (netPtr->weight > 1 || netPtr->maxLength || 
		      netPtr->constraintStatus == CONSTRAINED)  ? netCritPixel : net3Pixel;
	    if (!showConstrainedNets || showConstrainedNets && netPtr->specStatus == NOT_MET) {
	        XSetForeground( display, netGC, pixel );
	        for (I = 0 ; I < netPtr->subnets; I++) {
		    Term* st1 = netPtr->subnetArray[I].curTerm[0];
		    Term* st2 = netPtr->subnetArray[I].curTerm[1];
		
		    x1 = PLOTX(st1->curAbsCoordP->x);
		    y1 = PLOTY(st1->curAbsCoordP->y);
		    x2 = PLOTX(st2->curAbsCoordP->x);
		    y2 = PLOTY(st2->curAbsCoordP->y);

		    XDrawLine( display, placementWin, netGC,  x1, y1, x2, y2 );
	        }
	    }
	}
    }
}

void drawBins( gc )
    GC gc;
{
    if (showBins) {
	int 	I;
	int	x1,y1,x2,y2;	/* used to draw 2 terminals nets */

	y1 = PLOTY(binOffsetY);
	y2 = PLOTY(binOffsetY + chipBbox->upperRight.y - chipBbox->lowerLeft.y);
	for (I = 0 ; I <= numBinX ; I++) {
	    x2 = x1 = PLOTX(binOffsetX + I * binWidthX);
	    XDrawLine( display, placementWin, gc, x1, y1, x2, y2 );
	}
	x1 = PLOTX(binOffsetX);
	x2 = PLOTX(binOffsetX + chipBbox->upperRight.x - chipBbox->lowerLeft.x);
	for (I = 0 ; I <= numBinY ; I++) {
	    y2 = y1 = PLOTY(binOffsetY + I * binWidthY);
	    XDrawLine( display, placementWin, gc, x1, y1, x2, y2 );
	}
    }
    /* VOID function */
}

void plotCells()
{
    static GC gc;
    XWindowAttributes    info;
    octCoord	left,right,top,bottom;
    octCoord	coreLeft,coreRight,coreTop,coreBottom; /* chip boundaries */
    struct net	*netPtr;
    static int firstPlot = 1;
    int x, y, width, height;
  
    if ( NOT XDisplayActive OR NOT graphics) return;

    if ( firstPlot ) {
	gc = DefaultGC( display, 0  );
	setCoreBbox( chipBbox );
	firstPlot = 0;
    }

    XClearWindow( display, placementWin );
  
    XGetWindowAttributes( display, placementWin , &info);
    /*  draw chip bounding box */
    coreLeft   = coreBbox.lowerLeft.x;
    coreBottom = coreBbox.lowerLeft.y;
    coreRight  = coreBbox.upperRight.x;
    coreTop    = coreBbox.upperRight.y;
    localOriginX = coreLeft;
    localOriginY = coreBottom;
    scale = MIN(info.width,info.height)*0.8 /(double) (MAX(coreRight-coreLeft,coreTop-coreBottom)) ;
    globalOffsetX = (int) info.width  * 0.1;
    globalOffsetY = (int) info.height * 0.9;

    x =  PLOTX( chipBbox->lowerLeft.x );
    y =  PLOTY( chipBbox->upperRight.y );
    width = PLOTX(chipBbox->upperRight.x) - x;
    height = PLOTY(chipBbox->lowerLeft.y) - y;
    XSetLineAttributes( display, gc, 2, LineOnOffDash, CapNotLast, JoinMiter );
    XDrawRectangle( display, placementWin, gc, x, y , width, height );

    /*  current dimensions of the chip */
    left   = gFirstCell->lyoNext[PUPPY_L]->curBbox.lowerLeft.x;
    bottom = gFirstCell->lyoNext[PUPPY_B]->curBbox.lowerLeft.y;
    right  = gFirstCell->lyoNext[PUPPY_R]->curBbox.upperRight.x;
    top    = gFirstCell->lyoNext[PUPPY_T]->curBbox.upperRight.y;

    x =  PLOTX( left  );
    y =  PLOTY( top );
    width = PLOTX(right) - x;
    height = PLOTY(bottom) - y ;
    XSetLineAttributes( display, gc, 2, LineSolid, CapNotLast, JoinMiter );
    XDrawRectangle( display, placementWin, gc, x, y , width, height );

    drawCells( gc );
    drawBins( gc );
    drawNets( );
    if (analogMode && considerSymmetry && showAxes)
      drawSymmetryAxis( gc );
    if (analogMode && considerWell && showWell)
      drawWellAxis( gc );
    XFlush( display );
}
























