/*
 *  FILE: main.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Aug.21.1987		AUTHOR: Andrea Casotto
 *
 *  LAST MODIFICATION DATE: Jan.24.1991 AUTHOR: Edoardo Charbon
 *
 *
 *  FUNCTIONS:
 *  		main(argc,argv)
 *  double	puppyTime()
 *  int		readParameters()
 *  int		copyFacets ( cellName , oldView , newView )
 */

#include "defs.h" 
#include "sa.h"
#include "analogVariables.h"
#include "constraintDrivenVariables.h"
#include "parall.h"
#include <X11/Xlib.h>
#include "vov.h"
#include "oct.h" 
#include "options.h"
#include "port.h"
#include "copyright.h"
#include <stdio.h>


/*
 *   >>>  GLOBAL VARIABLES DECLARATION
 *   >>>  Attributes:	SM	in shared memory
 *			K	constant for all the processes
 *			I	independent for each process
 *			M	used only by main process
 */

char			*progName;		/* The name of the program */
char 		        *output_cell;           /* output facet and view */
char 		        *input_cell;            /* input facet and view */
char                    usingParameterFile[128];/* parameter file used */
long			mypid;			/* PID of the process */
int			myself;			/* local id for a process. 0 for parent */
int			myMoveId = 1;		/* identifier for moves attempted by myself */
struct ProcessorFrame   *myProc;		/*  */

/* Global flags */
int			XDisplayActive = FALSE;	/*  */
int			showBB = TRUE;		/*  show the bounding box of the cells */
int                     showBias = FALSE;        /*  show the biasing of cells */ 
int                     showAxes = FALSE;        /*  show the symmetry axes */
int                     showWell = FALSE;        /*  show the well boundary */
int                     showConstrainedNets = FALSE; /* show the nodes on which
						      	there are constraints*/
int			showNets = FALSE;	/*  */
int			showBins = FALSE;
int			showOwner = FALSE;	/*  */
int			showAge = FALSE;	/*  */
int	   		graphics = TRUE;	/*  */
SHARED int		FloorplanMode = FALSE;	/*  */
SHARED int		floatingTermsOnly = FALSE;
SHARED int		costOnly = FALSE;	/* compute the cost only */
SHARED int   		enableTranslations = TRUE; /*  */
SHARED int   		enableTransformations = TRUE; /*  */
SHARED int		enableCenterInterchange = TRUE;	/* OBSOLETE  */
SHARED int		enableSubnets = TRUE;	/*  */
SHARED int              considerSymmetry = TRUE; /* consider symmetry constraints in the circuit */
SHARED int              considerMatching = TRUE; /* consider matching constraints in the circuit */
SHARED int              considerWell = TRUE;     /* separate the devices according
						  * to their biasing. */
SHARED int              considerConstraintDriven = FALSE; /* consider node constraints */
SHARED int              considerVirtual = FALSE; /* do not consider virtual axes */
SHARED int              considerPower = FALSE;   /* do power optimization */
SHARED int		useQualityFactor = TRUE; /* To determine the frequency of the moves */
SHARED int		useCellBasedMoveStats = FALSE;
SHARED int		saveInterval = 50; 	/* # of steps between savings */
SHARED int		debug = FALSE;		/*  */
SHARED int		newDataFile = FALSE;
SHARED int              analogMode = FALSE;     /* take into account the constraints due to analog layout */
SHARED octCoord         wellBoundary = 0;       /* boundary between p and n wells */ 
SHARED octCoord         lambda;                 /* oct units / lambda */
SHARED double           maxCurrDensity;         /* max wiring current density */
SHARED octCoord         minWireWidth;           /* min wiring width */
char			errmsg[256];		/* string for error messages */
long 			randVar; 		/* used to generate random numbers */

octObject		Chip;			/* The facet to be placed  */
SHARED ListElt		*ListOfAnnealVars = NULL; 	/*  */
SHARED MacroCell	*gFirstCell;		/* Handle for the list of all cells */
SHARED SymmetricCell    *gxSymCell;             /* structure containing all symmetric cells with respect to xaxis */
SHARED SymmetricCell    *gySymCell;             /* structure containing all symmetric cells with respect to yaxis */ 
SHARED MatchedCell      *gMatchCell;            /* structure containing all matched cells */
SHARED ConstraintDriven *gConstraintDriven;     /* structure containing all critical nets */
SHARED Net      	*netSet; 		/* Pointer to the first net of all */
SHARED Net      	*powerNetSet; 		/* Pointer to the first power net of all */
SHARED int 		numberOfCells = 0;	/* Total number of cells */
SHARED int		numberOfMovingCells = 0;
SHARED int		numberOfFloatingPins = 0; /*  */
SHARED int		numberOfSubnets = 0;	/*  */
SHARED int              numberOPerformances = 0; /* number of performances */
SHARED int	 	numberOfNets = 0;	 /* Total number of nets */
SHARED int		numberOfAnnealVars = 0;

int			maxEdgesInPerim;	/* Max number of edges in perimeter of a cell */
double 			maxArea = 0.0; 		/* Area of largest cell */
double			totalArea = 0.0; 	/* Area of all the cells */
double			totalExpandedArea = 0.0;
SHARED MacroCellPointer *allCellsArray;		/* Array for quick reference of the cells */
SHARED MacroCellPointer *varCellsArray; 	/* Array for quick reference of movingCells */

int 			maxNets = 0;		/*  */
int 			maxEdgeX = 0;		/*  */
int 			maxEdgeY = 0;		/*  */

SHARED    double 	SM_T = 0 ;		/* current temperature in the annealing */
SHARED    double	SM_startT = INFINITE_TEMPERATURE; /* Initial temperature for annealing */
SHARED	  double	SM_minimumTemperature = 100.0; /*  */
SHARED	  double	SM_alpha = 0.9;		/* .. of the cooling schedule */
SHARED    octCoord	gRangeLimiterX = 10;	/* OBSOLESCENT */
SHARED    octCoord	gRangeLimiterY = 10;	/* OBSOLESCENT */
SHARED    int		SM_shakeStrength = 4;	/*  */
SHARED    int		SM_shakeX = 2;		/*  */
SHARED    int		SM_shakeY = 2 ;		/*  */
SHARED    int		movesPerAnnVar = 20; 	/* Attempts per cell */
SHARED    double	SM_ovWeight = 4;	/*  */
SHARED    double	SM_netWeight = 1;	/*  */
SHARED    double	SM_areaWeight = 1;	/*  */
SHARED    double        SM_matchWeight = 1;     /* mismatch penalty */
SHARED    double        SM_symWeight = 1;       /* asymmetry penalty */
SHARED    double        SM_wellWeight = 1;      /* well discontinuity penalty */
SHARED    double        SM_constraintWeight = 1; /* constraints weight */
SHARED    int           SM_virtualSteps = 1;    /* number of steps btw. virtual axes updates */
SHARED    double	SM_conViolWeight = 1;	/*  */
SHARED    double	SM_clusterWeight = 1E-8;/*  */
SHARED    int 		SM_minNumberOfCellsPerProc = 1;	/*  */
SHARED    int 		SM_maxNumberOfCellsPerProc = 1000000; /*  */
SHARED    int 		SM_minNumberOfAnnealVarsPerProc = 1; /*  */
SHARED    int 		SM_maxNumberOfAnnealVarsPerProc = 1000000; /*  */
SHARED    double        SM_topCapMod0 = .062;   /* top layer of metal capacitance model: 1st const */
SHARED    double        SM_topCapMod1 = .013;   /* top layer of metal capacitance model: 2nd const */
SHARED    double        SM_bottomCapMod0 = .081;/* bottom layer of metal capacitance model: 1st const */
SHARED    double        SM_bottomCapMod1 = .027;/* bottom layer of metal capacitance model: 2nd const */
SHARED    float         SM_minNetWidth = 6;     /* min net width in lambda (analog mode only) */
double			desAspectRatio = 1.0;	/*  */
double			automaticChipArea = 0.0; /*  */
double			cellExpansion = 2.0;	/*  */
SHARED	  int		chipBoundaryOverlapFactor = 10;
SHARED enum ChipBoundaryType chipBoundaryType = CB_AUTOMATIC; /*  */
SHARED octBox 		*chipBbox = NULL;	/* Bounding box restricting movement of cells */
int 			P = 4;			/* number of processes, defaults to 4 */
SHARED struct ProcessorFrame *process;		/* Array, info about process activities */
SHARED struct PFrame  	PFrame;			/*  */
#ifdef PARALLEL
shared	slock_t		lyoLock; 		/* lock for lyo data structure */
shared  slock_t		waitLock; 		/* lock for processor intercomunication */
shared  slock_t		subnetMoveLock;
#endif
long			totalNumberOfMoves;	/*  */
long			totalNumberOfAcceptedMoves; /*  */

SHARED MoveSet gCellMoveSet;
SHARED MoveSet gNetMoveSet;
SHARED MoveSet gFloatTermMoveSet;
#ifdef PARALLEL
shared slock_t moveStatArrayLock;	/* to control the access to moveStatArray */
#endif
SHARED int		numBinX , numBinY ;	/*  */
SHARED octCoord		binWidthX , binWidthY ;	/*  */
SHARED octCoord		binOffsetX , binOffsetY ; /*  */
SHARED int		***blockArray;		/*  */
SHARED int		*bucket;		/*  */
char			*displayName;		/*  */
Display 		*display;		/*  */
SHARED JobQueue		*outerLoopJobQueue; 	/*  */
SHARED ListElt		*firstParameter = NULL;	/* Handle for list of parameters */
 

/*
 *   GLOBAL DECLARATIONS OF TIMING VARIABLES
 *   These variables are used throughout this file to time various routines.
 */
double  		StartTime;		/*  */
double			TotalTime;		/*  */


/*
 *   GLOBAL DECLARATION OF VERSION
 */
extern char *PUPPYversion();


/*
 *   OPTION LIST DECLARATION
 */
optionStruct optionList[] = {

  { "O",      "ov_weight",    "Overlap Weight in SA cost function"                 },
  /* SM_ovWeight */
  { "W",      "net_weight",   "Wire Length Weight in SA cost function"             },
  /* SM_netWeight */
  { "A",      "area_weight",  "Weight in SA cost function"                         },
  /* SM_areaWeight */
  { "C",      "con_viol",     "ConstrViolations Weight in SA cost function"        },
  /* SM_conViolWeight */
  { "M",      "match_weight", "Mismatch Weight in SA cost function"                },
  /* SM_matchWeight */
  { "Y",      "sym_weight",   "asYmmetry Weight in SA cost function"               },
  /* SM_symWeight */
  { "D",      "well_weight",  "well Discontinuity Weight in SA cost function"      },
  /* SM_wellWeight */
  { "N",      "crit_node_weight","critical Node Weight in SA cost function"        },
  /* SM_constraintWeight */
  { "V",      "virtual_steps","Steps between updates of virtual axes (analog mode only)"  },
  /* SM_virtualSteps */
  { "T",      "starting_temp","Starting Temperature for SA [automatic]"            },
  { "R",      "",             "Seed for the random number generator"               },
  /*   fprintf(stderr,"  -S number       Cooling Schedule (0,1,2)"); */
  { "a",      "anneal_moves", "Moves per annealing variable"                       },
  /* movesPerAnnVar */
  { "x",      "cell_expansion","Set expansion factor for the cells"                },
  /* cellExpansion */
  { "c",      "",             "Compute the cost of current placement and stop"     },
  { "f",      "",             "Place only floating terminals"                      },
  { "p",      "",             "Do power optimization"                              },
  { "q",      "",             "Disable the translations"                           },
  { "t",      "",             "Disable the transformations"                        },
  { "g",      "",             "Turn OFF graphic output"                            },
  { "r",      "",             "Enable analog placement mode"                       },
  { "y",      "",             "Do not consider symmetry constraints (analog mode only)"},
  { "m",      "",             "Do not consider matching constraints (analog mode only)"},
  { "w",      "",             "Do not consider well constraints (analog mode only)"},
  { "n",      "",             "Do consider critical node constraints (analog mode only)"},
  { "v",      "",             "Do consider virtual axes (analog mode only)"        },
  { "l",      "CNB",          "List C_ells or N_ets or B_oth. Also writes parPuppy file"},
#ifdef PARALLEL
  { "P",      "number_proc",  "Processors to be used"                              },
#endif
  { "d",      "",             "Enables debug"                                      },
  { "i",      "input_par_file" , "Specify input parameter file [cell.parPuppy]"          },
  { "o",      "output_cell:view",    "Specify output cell and view"                },
  { OPT_RARG, "cell:view",           "Input cell and view"                         },
  { OPT_DESC, 0,              "\n"                                                 }, 

#ifdef PARALLEL
  { OPT_DESC, 0, "THIS IS THE PARALLEL VERSION for the Sequent Balance\n"          }, 
#else
  { OPT_DESC, 0, "THIS IS THE SEQUENTIAL VERSION\n"                                },
#endif

  { OPT_DESC, 0,  "(Command line options override the values in the parPuppy file)\n"},
  { 0,        0,              0                                                    }  

};


/*
 *   MAIN PROGRAM
 */
main(argc,argv)
    int argc;
    char **argv;
{
    Cost		curCost; /* Cost of current configuration*/
    octStatus		status;
    int			holdFinalPlacement = 0;
    int			listOnly = FALSE;
    char 		*listCommand = "";
    extern int    	coolingSchedule;
    char                *cellName, *template;
    int 		c;	/* used to get the options */
    extern char 	*optarg;
    extern int 		optind;
    int                 rtnBadFlag = 0;
    char                usingCDFile[128];
    extern double 	puppyTime();
    extern void		plotCells();
    extern char         *readConstraintDrivenFile();


    char *readParameters();
   
    if ((progName = strrchr(argv[0],'/')) == 0) {
	progName = argv[0];
    } else {
	progName++;		/* get rid of the leading '/' */
    }
    randVar = (long) time((time_t*) 0);	/* system time */


    if (argc > 1) {
	VOVbegin( argc, argv ); 
        initParameterList();
	output_cell = strsav(argv[argc-1]);
	if (strchr(output_cell,':') == NIL(char)) optUsage();
	cellName = output_cell;
 
	/* read parameter file */
	strcpy(usingParameterFile, readParameters(cellName, argv, argc));
    }
 
    /* parse command line options*/
    output_cell = NIL(char);
    while ((c = optGetOption(argc, argv)) != EOF) {

 	switch(c) {
	case 'i':
	  /* the par file is read and base parameters stored */
	  /* following options overwrite these base parameters */
	  break;
	case 'c': costOnly = TRUE; break;
	case 'd': debug = 1; break;
	case 'P': P = atoi( optarg ); break;
	case 'e': newDataFile = TRUE; break;
	case 'p': considerPower = TRUE; break;
	case 'q': enableTranslations = FALSE; break;
	case 't': enableTransformations = FALSE; break;
	case 'f': floatingTermsOnly = TRUE; break;
	case 'g': graphics = FALSE; break;
	case 'l': listOnly = TRUE; listCommand = optarg; break;
        case 'r': analogMode = TRUE; break;
        case 'y': considerSymmetry = FALSE; break;
        case 'm': considerMatching = FALSE; break;
        case 'w': considerWell = FALSE; break;
        case 'n': considerConstraintDriven = TRUE; break;
        case 'v': considerVirtual = TRUE; break;
	case 'R': randVar = (long) atoi(optarg); break;
	case 'S': 
	    coolingSchedule = atoi(optarg);
	    if (coolingSchedule < 0 || coolingSchedule > 2) 
	      FatalError("Arg of -S should be 0, 1 or 2");
	    break;
	case 'x': 
	    cellExpansion = atof(optarg);
	    if (cellExpansion < 0.0) {
		Warning("Expansion factor cannot be negative");
		cellExpansion = 0.0;
	    }
	    if (cellExpansion > 10.0) {
		Warning("I believe your expansion factor is too big. Continuing anyway ");
	    }
	    break;
	case 'O': SM_ovWeight= atof(optarg); break;
	case 'W': SM_netWeight = atof(optarg); break;
	case 'A': SM_areaWeight = atof(optarg); break;
	case 'C': SM_conViolWeight = atof(optarg); break;
        case 'M': SM_matchWeight = atof(optarg); break; 
        case 'Y': SM_symWeight = atof(optarg); break;
        case 'D': SM_wellWeight = atof(optarg); break;
	case 'N': SM_constraintWeight = atof(optarg); break;
        case 'V': SM_virtualSteps = atoi(optarg); break;
	case 'T': SM_startT = atof(optarg); break;
	case 'a': movesPerAnnVar = atoi(optarg); break;
	case 'o': output_cell = strsav(optarg); break;
	
	default: optUsage(); break;
	}
    }
 
    /* the remaining arguments are argv[optind ... argc-1] */
    if (optind >= argc) {
	optUsage();
    } else {
	input_cell = strsav(argv[optind]);
	optind++;
    }

    sprintf (errmsg,"Using %s parameter file", usingParameterFile);
    Message (errmsg);

    /* check for extraneous arguments */
    if (optind < argc)    optUsage();
  
    initializeSharedVariables(); /*  */
    initAnalogParameterList();
    allocateShMvariables2();	/*  */

    /* open the input facet for reading and copy it to the output facet */
    ohWorkFacet( &Chip, input_cell, output_cell );

    initializePlacementObjects();	/* Init objects defined by OCT symbolic policy */

    if (analogMode)  {
        Message ("Placement of analog circuits");
    } else {
        Message ("Placement of digital circuits: analog parameters ignored");
    }

    scanConstraintsFile();
 
    if ( ! (costOnly || listOnly) ) {
      startX();		/* Open all X windows */
    } 
 
    DFdescribe();		/* Describe the format of the  data file */

    DFopen();			/* Open the data file */

    if (considerPower) 
      readTechnologyDataFromOct();
    
    readDataFromOct();

    if (debug) scanNets();

    if (analogMode && considerConstraintDriven) {
      /* read CDfile and construct data structure */
      strcpy(usingCDFile, readConstraintDrivenFile(cellName));
      sprintf (errmsg,"Using %s constraint-driven file", usingCDFile);
      Message (errmsg);
      assignConstraintsToNets();
    }
    
    /* Read instances and net list */
    DFwriteHeader();		/* 	 */
    checkData();		/* Check consistency of input data */

    if (listOnly) {
	printConstraintsFile(listCommand[0] == 'B' || listCommand[0] == 'C',
			     listCommand[0] == 'B' || listCommand[0] == 'N');

	saveParFile();

	VOVend(0);
    }

    /* CAREFUL!! The initialization steps must be done in the correct sequence */
    lyosort();			/* set up lyo data structure */
    makebins();			/* Data structure for overlap computation*/
    loadbins();
    initializeCells();		/* expand , setup for display, etc */
    initChipBoundary();		/*  */
    initGlobalMoveSets();

    if (XDisplayActive) {
	holdFinalPlacement = 10;
	plotCells();
	refreshConsoleWindow();
    }

    if (costOnly == TRUE) {
	configurationCost(&curCost); /* Compute cost of initial placement */
	printCost(&curCost, "Cost of current placement");
	if (analogMode && considerConstraintDriven) DFwriteCriticalNetsCap();
	printf("Bounding box of the chip is: ");
	printChipBBox();
	VOVend(0);
    } 
#ifdef PARALLEL
    P = 1 + manageChildren(P-1); /* Fork P-1 children*/
#else
    if (P!=1) {
	Warning("Only 1 processor can be used with the sequential version");
    }
    P = 1;
#endif
    if (numberOfAnnealVars > 0) {
	distributeAnnealingVars();
    } else {
	FatalError("There are no variables (cells, nets or floating terms) to anneal. ");
    }
    initRangeLimiters();	/* Initialize the range limiters */
    doubleCheck();		/* Other checks on the data structures */
  
    /* SIMULATED ANNEALING */
    StartTime = puppyTime();
    simulatedAnnealing();
    TotalTime = puppyTime() - StartTime;
  
    (void) sprintf(errmsg,"Total annealing time: %.2f seconds", TotalTime);
    DFwriteComment(errmsg);
    Message(errmsg);
    if (analogMode && considerConstraintDriven) DFwriteCriticalNetsCap();
  
#ifdef PARALLEL
    manageChildren(0);		/* Kill all children */
#endif

    saveResultsToOct(TRUE);	/* Save final result */
    saveParFile();

    configurationCost(&curCost); /* Compute cost of final placement */
    printCost(&curCost, "Cost of final placement");
    
    if (enableSubnets==TRUE) {
	/* Always report the BBox metric */
	enableSubnets = FALSE;
	configurationCost(&curCost); /* Compute cost of final placement */
	printCost(&curCost, "Cost of final placement");
    }
    printf("Bounding box of the chip is: ");
    printChipBBox();
    DFclose();

    if (holdFinalPlacement > 0) {
	(void) sprintf(errmsg,"Holding placement for %d seconds",
				       holdFinalPlacement);
	Message(errmsg);
	StartTime = puppyTime();
	while (puppyTime() - StartTime < (double) holdFinalPlacement) {
	    innerLoopXserver();
	}
    }
    endX();
    VOVend(0);			/* Success */
}


int
copyFacets ( cellName , oldView , newCell, newView )
    char *cellName, *oldView , *newCell, *newView;
{
  octObject  		Input, Output;

  (void) sprintf(errmsg,"Copying %s:%s into %s:%s",
				 cellName, oldView , newCell, newView);
  Message(errmsg);

  Input.type = OCT_FACET;
  Input.contents.facet.cell = cellName;
  Input.contents.facet.view = oldView;
  Input.contents.facet.facet = "contents";
  Input.contents.facet.version = OCT_CURRENT_VERSION;
  Input.contents.facet.mode = "r";
  if (octOpenFacet(&Input) < OCT_OK) {
    (void) sprintf(errmsg,"Facet %s:%s:%s has some problem",
	    Input.contents.facet.cell,
	    Input.contents.facet.view,
	    Input.contents.facet.facet);
    FatalOctError(errmsg);
  }
  Output.type = OCT_FACET;
  Output.contents.facet.cell = newCell;
  Output.contents.facet.view = newView;
  Output.contents.facet.facet = "contents";
  Output.contents.facet.version = OCT_CURRENT_VERSION;

  VOVinputFacet( &Input );
  VOVoutputFacet( &Output );
  if (octCopyFacet(&Output,&Input) != OCT_OK) FatalOctError("Copying contents");

  Input.contents.facet.facet = "interface";
  Output.contents.facet.facet = "interface";
  if (octOpenFacet(&Input) < OCT_OK) {
    (void) sprintf(errmsg,"Facet %s:%s:%s has some problem",
	    Input.contents.facet.cell,
	    Input.contents.facet.view,
	    Input.contents.facet.facet);
    Warning(errmsg);
  }
  VOVinputFacet( &Input );
  VOVoutputFacet( &Output );
  if (octCopyFacet(&Output,&Input) != OCT_OK) Warning("Copying interface");
  return(OK);
}



#define RUSAGE_SELF 0
#define RUSAGE_CHILDREN -1
/* Routine that queries the system to find the process time. */
/* a better routine for the timing */
double
puppyTime()
{
  struct timeval t;

  if (gettimeofday(&t,NIL(struct timezone))==-1) Warning("Problems with gettimeofday");
  return((double) t.tv_sec + t.tv_usec * 0.000001 );
}




/**** PARAMETERS: read / set analog dfault parameters ******/

initAnalogParameterList()
{

  if (!analogMode) return;
  
  if (considerConstraintDriven) showConstrainedNets = TRUE;
  showWell = TRUE;
  showAxes = TRUE;
  showBias = TRUE;

}


/**** PARAMETERS: read / write parameter file ******/

enum parType {PT_INTEGER, PT_REAL, PT_BOOLEAN, PT_STRING};
struct parameter {
  char *name;
  enum parType  type;
  union {
    int *integerP;
    double *realP;
  } value;
  char *comment;
};

addParameterToList(name, type, address, comment)
    char 		*name;
    enum parType  	type;
    double 		*address;
    char 		*comment;
{
  struct parameter 	*newP;
  ListElt		*LE;

  newP = SH_MALLOC(struct parameter, 1);
  LE = SH_MALLOC(ListElt, 1);
  LE->pointer.generic = (char *) newP;
  newP->name = strsav(name);
  newP->type = type;
  newP->comment = strsav(comment);
  switch (type) {
 case PT_INTEGER:
 case PT_BOOLEAN:
    newP->value.integerP = (int *) address;
    break;
 case PT_REAL:
    newP->value.realP = (double *) address;
    break;
 default:
    FatalError("Wrong parameter type");
    break;
  }
  linkEltToListHandle(&firstParameter,LE);
}


initParameterList()
{

  /* arbitrary parameters for annealing */ 
  extern int terminalSpacing;
  extern int maxSiteCapacity;
  extern int astigmatism;                /* defined in "analogCost.c" */
  extern int alignment;                  /* defined in "analogCost.c" */
  extern int singleCellDisplacement;     /* defined in "analogCost.c" */
  extern double slopeRatio;              /* defined in "constraintDriven.c" */
  
  /* 
   * Add parameters in the inverse order that you want them appear in the 
   * parameter file autaomatically generated (Pure esthetics)
   */
  addParameterToList("slopeRatio", PT_REAL, &slopeRatio,
		     "Curve shaping parameters in constraint-driven approach (analog mode only)");
  addParameterToList("MinNetWidth", PT_REAL, &SM_minNetWidth,
		     "Minimum net width in um (analog mode only)");
  addParameterToList("CoTop", PT_REAL, &SM_topCapMod0,
		     "1st const. of top cap. model (analog mode only)");
  addParameterToList("C1Top", PT_REAL, &SM_topCapMod1,
		     "2nd const. of top cap. model (analog mode only)");
  addParameterToList("CoBottom", PT_REAL, &SM_bottomCapMod0,
		     "1st const. of bottom cap. model (analog mode only)");
  addParameterToList("C1Bottom", PT_REAL, &SM_bottomCapMod1,
		     "2nd const. of bottom cap. model (analog mode only)");
  addParameterToList("singleCellDisplacement", PT_INTEGER, (double *)  &singleCellDisplacement,
		     "Relative cost of misaligned single cells");
  addParameterToList("alignment", PT_INTEGER, (double *) &alignment,
		     "Relative cost of misaligned couples of cells");
  addParameterToList("astigmatism", PT_INTEGER, (double *) &astigmatism ,
		     "Relative cost astigmatic cells");
  addParameterToList("chipBoundaryType", PT_INTEGER, (double *)  &chipBoundaryType,
		     "Area in which the cells can move");
  addParameterToList("boundaryOVFactor",PT_INTEGER, (double *) &chipBoundaryOverlapFactor,
		     "Relative cost of placement outside boundary");
  addParameterToList("shakingStrength", PT_INTEGER, (double *) &SM_shakeStrength ,
		     "Length of the shaking vector");
  addParameterToList("terminalSpacing", PT_INTEGER, (double *) &terminalSpacing,
		     "Spacing between floating terminals");
  addParameterToList("maxSiteCapacity", PT_INTEGER, (double *) &maxSiteCapacity,
		     "Max number of floating terms in a site");
  addParameterToList("cellExpansion", PT_REAL, (double *)&cellExpansion,
		     "Used to change the amount of routing area");
  addParameterToList("desAspectRatio", PT_REAL, (double *) &desAspectRatio,
		     "desired aspect ratio (width over height) (X/Y)");
  addParameterToList("automaticChipArea", PT_REAL, (double *) &automaticChipArea,
		     "Used together with desAspectRatio");
  addParameterToList("enableSubnets", PT_BOOLEAN, (double *) &enableSubnets,
		     "Wire length estimated using subnets");
  addParameterToList("useQualityFactor", PT_BOOLEAN,(double *)  &useQualityFactor,
		     "To determine selection of moves");
  addParameterToList("clusterWeight", PT_REAL, &SM_clusterWeight,
		     "weight for clustering");
  addParameterToList("conViolWeight",PT_REAL,&SM_conViolWeight,
		     "weight for constraints violations");
  addParameterToList("symWeight", PT_REAL, &SM_symWeight,
		     "weight for symmetry (analog mode only)");
  addParameterToList("wellWeight", PT_REAL, &SM_wellWeight,
		     "weight for well discontinuity (analog mode only)");
  addParameterToList("constraintWeight", PT_REAL, &SM_constraintWeight,
		     "weight of critical nodes (analog mode only)");
  addParameterToList("matchWeight", PT_REAL, &SM_matchWeight,
		     "weight for matching (analog mode only)");
  addParameterToList("areaWeight", PT_REAL, &SM_areaWeight,
		     "weight for area");
  addParameterToList("netWeight", PT_REAL, &SM_netWeight,
		     "weight for nets");
  addParameterToList("ovWeight", PT_REAL, &SM_ovWeight,
		     "weight for overlap");
  addParameterToList("processors", PT_INTEGER, (double *) &P,
		     "number of processors to be used");
  addParameterToList("virtualSteps", PT_INTEGER, (double *) &SM_virtualSteps,
		     "Number of steps betw. updates of virtual axes (analog mode only)");
  addParameterToList("movesPerVariable", PT_INTEGER, (double *) &movesPerAnnVar,
		     "Moves attempted per annealing variable");
  addParameterToList("startT", PT_REAL, &SM_startT,
		     "starting temperature");
  addParameterToList("alpha", PT_REAL, &SM_alpha,
		     "alpha (simulated annealing)");
}
  
  
/*
 *  Reads the parameters in the file parPuppy in the current directory
 *  Any error in the file will be echoed with a warning. The routine returns
 *  OK if everything went ok, NOT_OK otherwise
 */
char
*readParameters(cellName, argv, argc)
    char        *cellName;
    char        **argv;
    int         argc;

{
    FILE 		*fp;           /*  */
    FILE		*fopen();      /*  */
    char                parFileName[128];
    char                parPuppy[128]; /* name of the parPuppy file */
    char 		input[40];     /*  */
    int 		test;      /* Return value of the fscanf() function */
    int			status = OK;   /* Returned status */
    int			found;	       /* Flag */
    int			integer;       /*  */
    double 		real;	       /*  */
    ListElt 		*parLE;	/* Used to run through the list of all parameters */
    struct parameter 	*parP;	       /*  */
    int                 parFileSpecified = FALSE;
    int                 i;             /* option index */


    for (i = 0; i < argc; i++) {
      if (strcmp(argv[i], "-i") == 0) {
	parFileSpecified = TRUE;
	strncpy(parFileName, argv[i + 1], 128);
      }
    }

    if (parFileSpecified) { 
      strncpy(parPuppy, parFileName, 128);
      fp=fopen(parPuppy , "r");
    } else {
      strcpy(parPuppy, "any specified parameter files");
      fp = (FILE *) NULL;
    }
    
    if (fp == (FILE *) NULL) {
      (void) sprintf(errmsg,"Cannot find %s",parPuppy);
      Message(errmsg);
      strncpy(parPuppy, cellName, 128 - strlen(" .parPuppy "));
      *strchr(parPuppy, ':') = '\0';
      strcat(parPuppy, ".parPuppy");
      if (( fp=fopen(parPuppy , "r")) == (FILE *) NULL) {
	(void) sprintf(errmsg,"Cannot find %s",parPuppy);
	Message(errmsg);
	strcpy(parPuppy,"parPuppy");
	if (( fp=fopen(parPuppy , "r")) == (FILE *) NULL) {
	  (void) sprintf(errmsg,"Cannot even find %s",parPuppy);
	  Message(errmsg);
	  Message("You can create or modify the parPuppy file using the menu");
	  strcpy(parPuppy, "no");
	  return parPuppy;
	}
      }
    }
      
    VOVinput( VOV_UNIX_FILE, parPuppy );

    while (fscanf(fp , " %s " , input) ==1) {
	/* The "#" sign at the beginning of a line indicates a comment in the parPuppy file */
	if  (strcmp( input , "#" ) == 0) { 
	    while ( getc(fp) != '\n') ;	/*   COMMENT: go on to the next line */
	}  else  {
	    found = FALSE;
	    for (parLE = firstParameter; parLE!=NULL; parLE = parLE->next) {
		parP = (struct parameter *) parLE->pointer.generic;
		if (strcmp( input, parP->name)==0) {
		    found = TRUE;
		    switch (parP->type) {
		    case PT_INTEGER:
		    case PT_BOOLEAN:
			test = fscanf( fp, " %d " , &integer);
			if (test!=1) {
			    (void) sprintf(errmsg," Value of %s not properly entered\n", parP->name);
			    Warning(errmsg);
			    status = NOT_OK;
			}
			*parP->value.integerP = integer;
			break;
		    case PT_REAL:
			test = fscanf( fp, " %lf " , &real);
			if (test!=1) {
			    (void) sprintf(errmsg," Value of %s not properly entered\n", parP->name);
			    Warning(errmsg);
			    status = NOT_OK;
			}
			*parP->value.realP = real;
			break;
		    default:
			FatalError("Wrong type for a parameter");
		    }
		}
	    }
	    if (found == FALSE) {
		(void) sprintf(errmsg,"Unparsable word (%s) found in %s",input,parPuppy );
		Warning(errmsg);
		while ( getc(fp) != '\n') ; /*  go on to the next line */
		status = NOT_OK;
	    }
	}
    }
    (void) fclose(fp);
    return parPuppy;
}



/*
 *  OK if everything went ok, NOT_OK otherwise
 */

int
saveParFile()
{
  int status1, status2;

  status2 = saveUsedFile(usingParameterFile, FALSE);
  status1 = saveUsedFile(usingParameterFile, TRUE);
  if (status1 != OK && status2 != OK) return NOT_OK;
  return OK;

}


int
saveUsedFile(parFileName, mode)
     char    *parFileName;
     int     mode;
{
  FILE 	         *fp, *fopen();
  int  	           status = OK;	    /* returned status */
  ListElt 		*parLE;	    /* Used to run through all parameters */
  struct parameter 	 *parP;	    /*  */
  char            outCell[128];

  if ((mode == FALSE) && (strcmp(parFileName, "no") == 0)) {
    strncpy(outCell, output_cell, 128);
    *strchr(outCell, ':') = '\0';
    sprintf(parFileName,"%s.parPuppy", outCell);
  } else {
    sprintf(parFileName,"%s.par", output_cell);
  }

  /* read parPuppy file	*/
  if (( fp=fopen(parFileName , "w")) == (FILE *) NULL) {
    (void) sprintf(errmsg,"Cannot write %s",parFileName);
    Warning(errmsg); 
    return NOT_OK;
  }
  VOVoutput( VOV_UNIX_FILE, parFileName );
  sprintf(errmsg,"Saving parameter file in``%s''", parFileName);
  if (mode) {
    Message(errmsg);
  } else {
    Warning(errmsg);
  } 
  fprintf(fp,"# Automatically generated parameter file\n");
  fprintf(fp,"# The # sign (followed by a space) indicates the beginning of a comment\n\n");
  
  for (parLE = firstParameter; parLE!=NULL; parLE = parLE->next) {
    parP = (struct parameter *) parLE->pointer.generic;
    switch (parP->type) {
  case PT_INTEGER:
  case PT_BOOLEAN:
      fprintf(fp, "%-18s %12d # %s\n", parP->name, *parP->value.integerP, parP->comment);
      break;
  case PT_REAL:
      fprintf(fp, "%-18s %12.4g # %s\n", parP->name, *parP->value.realP, parP->comment);
      break;
  default:
      FatalError("Wrong TYPE for a parameter");
    }
  }

  explainChipBoundaryType(fp);
  
  (void) fclose(fp);
  sprintf(errmsg,"%s saved",parFileName);

  if (mode) {
    Message(errmsg);
  } else {
    Warning(errmsg);
  }

  return status;
}



int unpackFacetName(name, facet)
    char *name;
    octObject *facet;
{
  char *s;

  /* Check for silly input */
  if (name == NIL(char) || strcmp(name, "") == 0) return;
	
  /* Make a copy of the input string (because we will tear it apart) */ 
  s = strsav(name);

  if (s[0] != ':') facet->contents.facet.cell = s;
  if ((s = strchr(s, ':')) != NIL(char)) {
    *s++ = '\0';
    if (s[0] != ':') facet->contents.facet.view = s;
    if ((s = strchr(s, ':')) != NIL(char)) {
      *s++ = '\0';
      if (s[0] != ':') facet->contents.facet.facet = s;
      if ((s = strchr(s, ':')) != NIL(char)) {
	*s++ = '\0';
	if (s[0] != ':') facet->contents.facet.version = s;
      }
    }
  }
}


























