#include "defs.h"
#include "global.h"
#include "sa.h"
#include "st.h"

copyMasterInfo(la, cell)
    MacroCell *la;		/* lookAlikeCell */
    MacroCell *cell;
{
  int		I;			/* generic counter */
  int		llc;
  octObject	Instance;
  octObject	InstTerm;
  struct termList *termptr, *latermPtr;
  Term		*term , *laterm;
  extern	st_table *st_terminals;

  cell->masterId = la->masterId;

  llc = cell->lowestLeftCorner = la->lowestLeftCorner; 
  cell->masterLLC = la->masterLLC; 
  cell->W = la->W; 
  cell->H = la->H; 
  cell->rW = la->rW; 
  cell->rH = la->rH; 
  cell->rHalfW = la->rHalfW; 
  cell->rHalfH = la->rHalfH; 
  cell->area = la->area; 

  /* Make sure the cell is not translated or rotated */
  cell->edgesInPerim = la->edgesInPerim; 
  cell->perimeter    = SH_MALLOC(octPoint, cell->edgesInPerim);
  cell->curPerimeter = SH_MALLOC(octPoint, cell->edgesInPerim);
  cell->newPerimeter = SH_MALLOC(octPoint, cell->edgesInPerim);
  cell->termsInEdge = SH_MALLOC(int, cell->edgesInPerim);
  cell->widthOfEdge = SH_MALLOC(int, cell->edgesInPerim);

  for (I=0;I<cell->edgesInPerim;I++) {
    cell->perimeter[I] = la->perimeter[I];
    cell->curPerimeter[I].x = la->perimeter[I].x + cell->masterLLC.x + (cell->W >> 1);
    cell->curPerimeter[I].y = la->perimeter[I].y + cell->masterLLC.y + (cell->H >> 1);
    cell->newPerimeter[I] = cell->curPerimeter[I]; /* initial position */
    cell->termsInEdge[I] = 0;
    cell->widthOfEdge[I] = 0;
  }

  cell->curCenter.x = (cell->W >> 1) + cell->curPerimeter[llc].x;
  cell->curCenter.y = (cell->H >> 1) + cell->curPerimeter[llc].y;
  cell->curBbox.lowerLeft.x = cell->curCenter.x - cell->rHalfW;
  cell->curBbox.lowerLeft.y = cell->curCenter.x - cell->rHalfH;
  cell->curBbox.upperRight.x = cell->curBbox.lowerLeft.x + cell->rW;
  cell->curBbox.upperRight.y = cell->curBbox.lowerLeft.x + cell->rH;

  cell->numberOfSites = la->numberOfSites; 
  if (cell->numberOfSites > 0) {
    cell->floatingSites = SH_MALLOC(struct s_site, cell->numberOfSites );
    for (I = 0; I < cell->numberOfSites ; I++) {
      cell->floatingSites[I].cellCoord = la->floatingSites[I].cellCoord ;
      cell->floatingSites[I].relPosition = la->floatingSites[I].relPosition ;
      cell->floatingSites[I].capacity = la->floatingSites[I].capacity ;
      cell->floatingSites[I].edge = la->floatingSites[I].edge ;
      cell->floatingSites[I].terms = la->floatingSites[I].terms ;
    }
  }
#ifdef PARALLEL
  s_init_lock(&cell->floatingSitesLock);
#endif

  /*  ***** These fields are not yet initialized. Skip them *****
  cell->moveStatArray = la->moveStatArray; 
  cell->moveSelArray = la->moveSelArray; 
  cell->rangeLimiter = la->rangeLimiter; 
  */


  /* Copy all the terminals */
  cell->term0.next = NULL;		/* clear the handle for the terminal list */
  cell->term0.prec = NULL;		/* clear the handle for the terminal list */
  cell->term0.term = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.next = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.prec = NULL;		/* clear the handle for the terminal list */
  cell->floatTerm0.term = NULL;		/* clear the handle for the terminal list */
  cell->lastMovedTermLE = NULL;
  cell->net0.next = NULL;
  cell->net0.prec = NULL;
  cell->net0.net  = NULL;

  cell->floatingTerms = la->floatingTerms; 

  for (I=0;I<cell->edgesInPerim;I++) {
    cell->termsInEdge[I] = la->termsInEdge[I];
  }

  for (I=0;I<cell->edgesInPerim;I++) {
    cell->widthOfEdge[I] = la->widthOfEdge[I];
  }

  InstTerm.type =OCT_TERM;
  Instance.type = OCT_INSTANCE;
  Instance.objectId = cell->cellId;
  if (octGetById(&Instance)!=OCT_OK) FatalOctError("in masters.c");
  for (latermPtr = la->term0.next ; latermPtr!=NULL ; latermPtr = latermPtr->next) {
    laterm = latermPtr->term;
    InstTerm.contents.term.name = latermPtr->term->name;
    if (octGetByName(&Instance,&InstTerm)!=OCT_OK) FatalOctError("master: getting inst term");

    term = SH_MALLOC ( Term, 1);
    st_insert(st_terminals, (char *) InstTerm.objectId , (char *) term);

    termptr = SH_MALLOC (struct termList, 1);
    termptr->term = term;
    term->status = UNUSED;
    term->name = laterm->name;
    term->cell = cell;
    term->subnets = 0;
    term->firstSubnet = NULL;

    term->floating = laterm->floating;
    if (term->floating) {
      term->illegalEdges = laterm->illegalEdges;
      linkTerm (termptr,&cell->floatTerm0);
      createAnnealVar(AV_FLOAT_TERM, (char *) term);
    } else {
      term->curAbsCoordP = SH_MALLOC(octPoint , 1);
      term->newAbsCoordP = SH_MALLOC(octPoint , 1);
      
      term->cellCoord = laterm->cellCoord;
      term->edge = laterm->edge;
      linkTerm (termptr,&cell->term0);
    }
  }
}
