/*
 *  FILE: menu.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 */
#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "constraintDrivenVariables.h"
#include "xpa.h"
#include "xg.h"
#include "iv.h"
#include "vov.h"

extern void plotCells();

int confirm ;			/* Confirmation flag */
char 	*displayName;		/*  */
Display *display;		/*  */
int	bgConsolePixel;
int	fgConsolePixel;
int	bgPlacementPixel;
int	fgPlacementPixel;

Window 	consoleWin = 0;		/* where the current status of program is shown */
Window	placementWin = 0;		/* where the cells are displayed */
IVWindow*  ivWin;		/* The structure used by IV */

XGraph  xgCostError;
XGraph  xgAccRatio;
XGraph  xgQF;

/*  Fonts and vertical separations of lines inthat font */
#define SMALL_FONT 	"*-courier-medium-r-*-*-*-100-*"
#define SMALL_VSPACE 	12

#define BIG_FONT   	"*-courier-medium-r-*-*-*-120-*"
#define BIG_VSPACE   	18

XFontStruct *smallFont;
XFontStruct *bigFont;



/* The prefix PM_ stands for Puppy Menu */
extern int PM_resize();
extern int PM_showBB();
extern int PM_showNets();
extern int PM_showConstrainedNets();
extern int PM_showAxes();
extern int PM_showBias();
extern int PM_showWell();
extern int PM_showBins();
extern int PM_showOwner();
extern int PM_showAge();
extern int PM_redraw();
extern int PM_toggleDebug();
extern int PM_constAlpha();		/* cooling schedule */
extern int PM_adaptiveCoolSched();	/* cooling schedule */
extern int PM_empiricalCoolSched();	/* cooling schedule */
extern int PM_zeroT();
extern int PM_infT();
extern int PM_singleMove();  
extern int PM_exit();
extern int PM_save_exit();
extern int PM_centerInterchange();
extern int PM_savePlacement();
extern int PM_saveParPuppy();
extern int PM_saveParPuppyNow();
extern int PM_changeCBtype();		/* change chip boundary */

#define MENU_PANES   4
#define MENU_ENTRIES 11
PFI menu_function_array[MENU_PANES][MENU_ENTRIES] = {
    { PM_redraw, PM_resize, PM_showAge, PM_showOwner, PM_showBins, PM_showNets, PM_showConstrainedNets, PM_showAxes, PM_showBias, PM_showWell, PM_showBB },
    { PM_constAlpha, PM_empiricalCoolSched, PM_adaptiveCoolSched, PM_zeroT, PM_infT, 0, 0 },
    { PM_savePlacement, PM_saveParPuppy, PM_saveParPuppyNow, 0, 0, 0, 0 },
    { PM_singleMove, PM_save_exit, PM_exit, 0, 0, 0, 0 }
};

static xpa_entry display_menu[] = {
  {"redraw",		XPA_NOCHAR, 0, 0},
  {"recenter",		XPA_NOCHAR, 0, 0},
  {"show-age-on/off",	XPA_NOCHAR, 0, 0},
  {"show-owner-on/off",	XPA_NOCHAR, 0, 0},
  {"show-bins-on/off",	XPA_NOCHAR, 0, 0},
  {"show-nets-on/off",	XPA_NOCHAR, 0, 0},
  {"show-bad-nets-on/off",	XPA_NOCHAR, 0, 0},
  {"show-axes-on/off",	XPA_NOCHAR, 0, 0},
  {"show-bias-on/off",	XPA_NOCHAR, 0, 0},
  {"show-well-on/off",	XPA_NOCHAR, 0, 0},
  {"cell-expansion",	XPA_NOCHAR, 0, 0},
  {  0,	0, 0, 0 }
};

static xpa_entry cool_menu[] = {
  { "constant-alpha",	XPA_NOCHAR, 0, 0},
  { "empirical",	XPA_NOCHAR, 0, 0},
  { "adaptive",		XPA_NOCHAR, 0, 0},
  { "zero-T",		XPA_NOCHAR, 0, 0},
  { "infinite-T",		XPA_NOCHAR, 0, 0},
  {  0,	0, 0, 0 }
};

static xpa_entry save_menu[] = {
  { "placement",		XPA_NOCHAR, 0, 0},
  { "parameter-file",		XPA_NOCHAR, 0, 0},
  { "parameter-file-now",	XPA_NOCHAR, 0, 0},
  {  0,	0, 0, 0 }
};

static xpa_entry exit_menu[] = {
  { "single-move",	XPA_NOCHAR, 0, 0},
  { "save_and_exit",	XPA_NOCHAR, 0, 0},
  { "exit",		XPA_NOCHAR, 0, 0},
  {  0,	0, 0, 0 }
};

static xpa_entry puppy_menu[] = {
  { "Display", 		XPA_NOCHAR, "DISPLAY", 		display_menu },
  { "Cooling Schedule", XPA_NOCHAR, "COOLING SCHEDULE",	cool_menu },
  { "Save",		XPA_NOCHAR, "SAVE",		save_menu },
  { "Exit",		XPA_NOCHAR, "EXIT", 		exit_menu },
  {  0,	0, 0, 0 }
};

#ifdef OBSOLETE_CODE
struct menuEntry CommandArray[] = {
  { PM_redraw,  "Display",		"redraw",		1},
  { PM_resize,  "Display",		"recenter",		1},
  { PM_showAge,	"Display", 		"show-age",		1},
  { PM_showOwner,"Display", 		"show-cell-owner",	1},
  { PM_showBins,"Display", 		"show-bins-on/off",	1},
  { PM_showNets,"Display", 		"show-nets-on/off",	1},
  { PM_showConstrainedNets,"Display", 		"show-axes-on/off",	1},
  { PM_showAxes,"Display", 		"show-axes-on/off",	1},
  { PM_showBias,"Display", 		"show-bias-on/off",	1},
  { PM_showWell,"Display", 		"show-well-on/off",	1},
  { PM_showBB,	"Display", 		"cell-expansion",	1},
  { PM_constAlpha,"Cooling Schedule", 	"constant-alpha",	1},
  { PM_empiricalCoolSched,"Cooling Schedule","empirical",	1},
  { PM_adaptiveCoolSched,"Cooling Schedule","adaptive",		0},
  { PM_zeroT,	"Cooling Schedule", 	"zero-T",		1},
  { PM_infT,	"Cooling Schedule", 	"infinite-T",		1},
  { PM_centerInterchange,"Move Set",	"center align on/off",	0},
  { PM_changeCBtype,"Move Set",		"change-chip-boundary",	1},
  { PM_savePlacement,"Save",		"placement",		1},
  { PM_saveParPuppy, "Save",		"parameter-file",	1},
  { PM_saveParPuppyNow,"Save",		"parameter-file-now",	1},
  { PM_toggleDebug,"Debug",	 	"on/off",		0},
  { PM_singleMove, "Debug",	 	"single-move",		1},
  { PM_save_exit,"Debug",	 	"save_and_exit",	1},
  { PM_exit,	"Debug",	 	"exit",			1}
};
#endif

/* If the button is released outside the menu we must exit */
int menuEvtHandler( evt )
    XEvent *evt;
{
  return ( evt->type == ButtonRelease ) ? 1 : 0;
}

static XFontStruct* puppyLoadFont( display, name )
     Display* display;
     char* name;
{
  XFontStruct* font;
  char** fontList;
  int actual;
  
  fontList = XListFonts( display, name, 3, &actual );
  if ( !actual ) {
    char buffer[256];
    sprintf( buffer, "No such font `%s'", name );
    ivError( buffer );
    return 0;
  }
  font = XLoadQueryFont( display, fontList[0] );
  if ( font == (XFontStruct*)0) {
    char buffer[256];
    sprintf( buffer, "No such font `%s'", fontList[0] );
    ivError( buffer );
  }
  XFreeFontNames( fontList );
  return font;
}

int activateXMenu( func )
    PFI *func;
{
    extern int menuEvtHandler();
    int ret;
    static int vals[2] = { 0 , 0 };
    static int depth = 0;
    int 	*results;
    Window rootWin, childWin;
    int root_x, root_y;
    int x, y;
    unsigned int mask;

    static xpa_menu  puppyMenu = 0;
    if (puppyMenu == 0) {
	xpa_appearance appearance;

	puppyMenu = xpa_create( display, "puppy", puppy_menu, 0 , &appearance );
    }
    XQueryPointer( display, RootWindow(display,0), &rootWin, &childWin, &root_x, &root_y, &x, &y, &mask );

    ret = xpa_moded( puppyMenu, x , y, depth, vals, 0, menuEvtHandler, &results );
    if ( ret != 2  ) {
	*func = 0;
    } else {
	int pane  = results[0];
	int entry = results[1];
	*func = menu_function_array[pane][entry];
	vals[0] = results[0]; vals[1] = results[1];
	depth = 2;
    }
    return OK;
}

/* This routine is called from the inner loop and checks all the events sent by X */
void innerLoopXserver()
{
  XEvent	theEvent;
  Window	subW;		/* subWindow */
  int 	menuX;			/* x coordinate of the menu */
  int 	menuY;			/* y coordinate of the menu */
  static int 	selection = 0;	/* previous selection */
  static int 	pane = 0;	/* previous pane */
  PFI	func;			/* pointer to a function returning integer */

  if ( XDisplayActive == FALSE ) return;
  /* Loop until there are events in the queue */
  while (XPending( display ) > 0) {
    XNextEvent( display, &theEvent);

    /* Pass the event to XG or to IV to see if one of them can handle it */
    if (iv_HandleEvent( &theEvent ) == IV_OK) continue;
    if (xgHandleEvent( &theEvent ) == XG_OK) continue; 
    /* Then PUPPY has to handle this event */

    if (theEvent.type == ButtonPress) {
      activateXMenu( &func );
      if (func != (PFI) 0) {
	if ((*func)() != OK) FatalError("Problem in the selected subroutine") ;
      }
    } 
    if (theEvent.type == Expose ) {
      XExposeEvent *exposeEvent = (XExposeEvent*)&theEvent;
      if (exposeEvent->count == 0) {
	if  ( theEvent.xany.window == placementWin ) { 
	  plotCells();
	  refreshConsoleWindow();
	}
	if ( theEvent.xany.window == consoleWin ) {
	  refreshConsoleWindow();
	}
      }	  
    }
  }				/* End while */
  /* VOID FUNCTION */
  return;
}


void startAllWindows()
{
    int		graphY;		/* y coordinates for the Xgraphs */
    XWindowAttributes  	placementWinInfo;
    XWindowAttributes  	consoleWinInfo;
    XWindowAttributes	info;
    XWindowAttributes	ivWinInfo;
    XSetWindowAttributes winAttr;
    Cursor	theCursor;
    int		cursorPixel;
    extern ivWhichItem *ivChipBoundary;
    extern int	iv_changeCBtype();
    extern int	iv_changedAspectRatio();
    extern int	iv_changedExpansion();
    extern int   PM_saveParPuppy();

    cursorPixel = puppyGetColor("yellow");
    bgConsolePixel = puppyGetColor("DarkSlateGray");
    fgConsolePixel = puppyGetColor("white");
    bgPlacementPixel = puppyGetColor("LightGray");
    fgPlacementPixel = puppyGetColor("black");
  
    setColors();

    /* XGetWindowAttributes( display, RootWindow(display,0) , &info); */
    /*
     *  set up the initial parameters for placementWin
     *  This is not necessarily the best way to do it
     *  but at least it is clear
     */
    placementWinInfo.width  = 408;
    placementWinInfo.height = 400;
    placementWinInfo.x	  = 2;
    placementWinInfo.y	  = 2;
/*     placementWinInfo.bdrwidth = 1; */


    consoleWinInfo.width   = 830;
    consoleWinInfo.height  = 180;
    consoleWinInfo.x	 = placementWinInfo.x;
    consoleWinInfo.y	 = placementWinInfo.y + placementWinInfo.height + 2;
    /* consoleWinInfo.bdrwidth = 1; */


    /* INTERACTIVE VARIABLES */
    ivWin = iv_InitWindow( display, "puppy","(**) IMPORTANT (*) SOMEWHAT IMPORTANT", 12);

    if ( ivWin == (IVWindow*)0 ) FatalError("Cannot open IV window");


    iv_AddVar( ivWin, "** Current  temperature", (int *) &SM_T, IV_DOUBLE , 3, IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Starting temperature", (int *) &SM_startT, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Minimum temperature",  (int *) &SM_minimumTemperature, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
#ifndef PRODUCTION
    iv_AddVar( ivWin, "Alpha",  (int *) &SM_alpha, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
#endif
    iv_AddVar( ivWin, "** Moves per annealing vars", (int *) &movesPerAnnVar, IV_INT , IV_NO_OPT , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "** Overlap    weight", (int *)&SM_ovWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "** Net        weight", (int *)&SM_netWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "** Area       weight", (int *)&SM_areaWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "** Violations weight", (int *)&SM_conViolWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);

    if (analogMode && considerSymmetry)  
      iv_AddVar( ivWin, "** Symmetry   weight", (int *)&SM_symWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
     
    if (analogMode && considerMatching)
      iv_AddVar( ivWin, "** Matching   weight", (int *)&SM_matchWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);

    if (analogMode && considerWell)
      iv_AddVar( ivWin, "** Well Disc.  weight", (int *)&SM_wellWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);

    if (analogMode && considerConstraintDriven)
      iv_AddVar( ivWin, "** Constraint  weight", (int *)&SM_constraintWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);


#ifdef PARALLEL
    iv_AddVar( ivWin, "Cluster    weight",    (int *)&SM_clusterWeight, IV_DOUBLE , 3 , IV_NULL_FUNCTION);
#endif

    iv_AddVar( ivWin, "Shaking strength", (int *) &SM_shakeStrength, IV_INT , IV_NO_OPT , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Range limiter X", (int *) &gRangeLimiterX, IV_INT , IV_NO_OPT , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Range limiter Y", (int *) &gRangeLimiterY, IV_INT , IV_NO_OPT , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Use quality factor", (int *)  &useQualityFactor,IV_BOOLEAN , IV_NO_OPT , IV_NULL_FUNCTION);
    iv_AddVar( ivWin, "Enable subnets", (int *) &enableSubnets, IV_BOOLEAN , IV_NO_OPT , IV_NULL_FUNCTION);
    buildIvChipBoundary();
    iv_AddVar( ivWin, "Chip boundary", (int*)ivChipBoundary, IV_WHICH, 4, iv_changeCBtype );
    iv_AddVar( ivWin, "* (X/Y) of Automatic Boundary" , (int *) &desAspectRatio, IV_DOUBLE, 3, iv_changedAspectRatio);
    iv_AddVar( ivWin, "* Area of Automatic Boundary", (int *)&automaticChipArea, IV_DOUBLE, 3, iv_changedAspectRatio);
    iv_AddVar( ivWin, "* Cell Expansion" , (int *) &cellExpansion, IV_DOUBLE , 3 , iv_changedExpansion);

#ifdef OBSOLETE
    iv_AddVar( ivWin, "Debug"  , (int *) &debug, IV_BOOLEAN , IV_NO_OPT , IV_NULL_FUNCTION);
#endif

    iv_AddVar( ivWin, "Confirm"  , (int *) &confirm, IV_TOGGLE , IV_NO_OPT , IV_NULL_FUNCTION);
    /* iv_AddVar( ivWin, "Period of savings"  , (int *) &saveInterval, IV_INT , IV_NO_OPT , IV_NULL_FUNCTION); */

    iv_MapWindow( ivWin, placementWinInfo.x + placementWinInfo.width + 2 , 1 , IV_NO_OPT);
    iv_SetEraseFlag( ivWin, 1);

    XGetWindowAttributes( display, ivWin->w , &ivWinInfo);

  
    placementWinInfo.height = ivWinInfo.height;
    consoleWinInfo.y	 = placementWinInfo.y + placementWinInfo.height + 2;
    consoleWinInfo.width   = placementWinInfo.width + ivWinInfo.width + 2;


    placementWin = XCreateSimpleWindow( display, RootWindow( display, 0 ),
				       placementWinInfo.x,
				       placementWinInfo.y,
				       placementWinInfo.width,
				       placementWinInfo.height,
				       1 ,
				       fgPlacementPixel, bgPlacementPixel
				       );

    consoleWin = XCreateSimpleWindow( display, RootWindow( display, 0) ,
				     consoleWinInfo.x,
				     consoleWinInfo.y,
				     consoleWinInfo.width,
				     consoleWinInfo.height,
				     1,
				     fgConsolePixel, bgConsolePixel);

    XStoreName( display, consoleWin, "puppy_console");
    XStoreName( display, placementWin, "puppy_placement");

    XSelectInput( display, placementWin, ExposureMask | ButtonPressMask | ButtonReleaseMask );
    XSelectInput( display, consoleWin,   ExposureMask | ButtonPressMask | ButtonReleaseMask );

	    
    winAttr.override_redirect = 1;
    XChangeWindowAttributes( display, consoleWin, CWOverrideRedirect , &winAttr );
    XChangeWindowAttributes( display, placementWin, CWOverrideRedirect , &winAttr );

    XMapWindow( display, consoleWin);
    XMapWindow( display, placementWin);


    /* put the graphs below the placement Window */
    graphY = consoleWinInfo.height + consoleWinInfo.y + 1;
    xgCostError = xgNewWindow(display, "puppy", "Cost and Error","Temp","Cost",249,249,0,graphY);
    xgAccRatio = xgNewWindow(display, "puppy", "Acceptance Ratios","Temp","AR",249,249,250,graphY);
    xgQF = xgNewWindow(display, "puppy", "Selection Probability","Temp","p",330,249,500,graphY);

    /* All graphs are in logarithmic scale in the horizontal axis */
    xgSetFlags(xgCostError, XGLogXFlag);
    xgSetFlags(xgAccRatio, XGLogXFlag);
    xgSetFlags(xgQF, XGLogXFlag);

    smallFont = puppyLoadFont( display, SMALL_FONT);
    bigFont = puppyLoadFont( display, BIG_FONT);

    /* VOID FUNCTION */
}

void  startX()
{
    XDisplayActive = FALSE;
    if (graphics == TRUE) {
	if (( display = XOpenDisplay( (char*)0 ) ) != (Display *) 0 ) {
	    startAllWindows();
	    /* startXMenu(); */
	    XDisplayActive = TRUE;
	    /* XSynchronize( display, True); */ /* This is for debugging */
	} else {
	    (void) sprintf(errmsg,"Could not open DISPLAY %s. Graphics suppressed", displayName);
	    Warning(errmsg);
	    graphics = FALSE;
	}
    }
}

void endX()
{
  if ( XDisplayActive ) {
    XCloseDisplay( display );
  }
}

/* 
 * Given a standard color name,  this routine fetches the associated
 * pixel value using XGetHardwareColor.  If it can't find or allocate
 * the color,  it returns 0;
 */
int puppyGetColor(name)
    char *name;
{
    XColor def;
    XColor def_exact;
    Colormap colormap = DefaultColormap( display , 0 );

    if (XAllocNamedColor( display, colormap, name, &def, &def_exact)) {
	return def.pixel;
    } else {
	return -1;
    }
}


/* Print on placement window */
void pwprint( s, l, x, y )
    char *s;			/* the string */
    int l;			/* string length */
    int x, y;			/* String coordinates */
{
    static GC gc = 0;

    if ( gc == 0) {
	gc = XCreateGC( display, placementWin, 0, 0 );
	XSetFont( display, gc, smallFont->fid );
	XSetForeground( display, gc, fgConsolePixel );
	XSetBackground( display, gc, bgConsolePixel );
    }
    XDrawImageString( display, placementWin, gc, x, y, s , l );
}


/* Print on console window */
void cwprint( s, l, x, y )
    char *s;			/* the string */
    int l;			/* string length */
    int x, y;			/* String coordinates */
{
    static GC gc = 0;

    if ( gc == 0) {
	gc = XCreateGC( display, consoleWin, 0, 0 );
	XSetFont( display, gc, smallFont->fid );
	XSetForeground( display, gc, fgConsolePixel );
	XSetBackground( display, gc, bgConsolePixel );
    }
    XDrawImageString( display, consoleWin, gc, x, y, s , l );
}


displayProblemData()
{
    extern char *progName;
    extern char *PUPPYversion();
    extern char *annealVarStatsString();
    char str[80];
    int X = 420;
    int	Y = 0;

    (void) sprintf(str,"%s : %s", progName, PUPPYversion());
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
    (void) sprintf(str,"PLACEMENT OF: %s:%s", Chip.contents.facet.cell, Chip.contents.facet.view);
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
    (void) sprintf(str,"Cells      = %4d    Nets       = %4d",numberOfCells,numberOfNets); 
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
    (void) sprintf(str,"Float Trms = %4d    Subnets    = %4d",numberOfFloatingPins,numberOfSubnets);
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
    (void) sprintf(str,"Ann Vars Total = %d", numberOfAnnealVars); 
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
    (void) sprintf(str,"Ann Vars       = %s", annealVarStatsString());
    cwprint( str, strlen( str ), X, ( Y+=15, Y) );
}

void displaySpeed(speed)
    int speed;
{
  char str[40];
  int speedX = 10;
  int speedY = 30;

  (void) sprintf(str, "Speed %5d Processors %3d ", speed, P);
  cwprint( str, strlen( str ) , speedX, speedY );
}

void displayStepNumber(n)
    int n;
{
  char str[40];
  int StepX = 10;
  int StepY = 15;

  (void) sprintf(str,"Step %4d ", n);
  cwprint( str, strlen( str ) , StepX, StepY );
}

void displayCost(costP)
    Cost *costP;
{
    char str[40];
    int X = 10;
    int Y = 45;

    (void) sprintf(str,"Total Cost %12.2f ", costP->total);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    if (costP->total <= 0.0) return;

    (void) sprintf(str,"Ov %12.2f %3d%% %8.2g ", 
	    costP->overlap, (int)(100 * costP->overlap * SM_ovWeight / costP->total), SM_ovWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"WL %12.2f %3d%% %8.2g ",
	    costP->wireLength, (int) (100 * costP->wireLength * SM_netWeight / costP->total), SM_netWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"Ar %12.2f %3d%% %8.2g ",
	    costP->area, (int) (100 * costP->area * SM_areaWeight / costP->total), SM_areaWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"Vl %12.2f %3d%% %8.2g ",
	    (double) costP->conViol, 
		   (int)(100 * costP->conViol * SM_conViolWeight / costP->total), SM_conViolWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    if (analogMode && considerMatching) {
      (void) sprintf(str,"MA %12.2f %3d%% %8.2g ",
		     (double) costP->matching, 
		     (int)(100 * costP->matching * SM_matchWeight / costP->total), SM_matchWeight);
      cwprint( str, strlen( str ), X, Y );
    }
    Y += SMALL_VSPACE;

    if (analogMode && considerSymmetry) {
      (void) sprintf(str,"SY %12.2f %3d%% %8.2g ",
		     (double) costP->symmetry, 
		     (int)(100 * costP->symmetry * SM_symWeight   / costP->total), SM_symWeight);
      cwprint( str, strlen( str ), X, Y );
    }
    Y += SMALL_VSPACE;

    if (analogMode && considerWell) {    
      (void) sprintf(str,"WE %12.2f %3d%% %8.2g ",
		     (double) costP->well, 
		     (int)(100 * costP->well * SM_wellWeight   / costP->total), SM_wellWeight);
      cwprint( str, strlen( str ), X, Y );
    }
    Y += SMALL_VSPACE;

    if (analogMode && considerConstraintDriven) {
      (void) sprintf(str,"CO %12.2f %3d%% %8.2g ",
		     (double) costP->constraint, 
		     (int)(100 * costP->constraint * SM_constraintWeight / costP->total), 
		     SM_constraintWeight);
      cwprint( str, strlen( str ), X, Y );
    }

}

void displayError(errorP)
    Cost *errorP;
{
    char str[40];
    int X = 210;
    int Y = 45;

    (void) sprintf(str,"Total Error %12.2f ", errorP->total);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"Ov %12.2f  %12.2f  ", errorP->overlap, errorP->overlap * SM_ovWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"WL %12.2f  %12.2f  ", errorP->wireLength, errorP->wireLength * SM_netWeight);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;

    (void) sprintf(str,"Ar %12.2f  %12.2f  ", errorP->area, errorP->area * SM_areaWeight);
    cwprint( str, strlen( str ), X, Y );

    Y += SMALL_VSPACE;
    (void) sprintf(str,"Vl %12.2f  %12.2f  ", (double) errorP->conViol, errorP->conViol * SM_conViolWeight);
    cwprint( str, strlen( str ), X, Y );

    Y += SMALL_VSPACE;
    if (analogMode && considerMatching) {
      (void) sprintf(str,"MA %12.2f  %12.2f  ", (double) errorP->matching, errorP->matching * SM_matchWeight);
      cwprint( str, strlen( str ), X, Y );
    }

    Y += SMALL_VSPACE;
    if (analogMode && considerSymmetry) {
      (void) sprintf(str,"SY %12.2f  %12.2f  ", (double) errorP->symmetry, errorP->symmetry * SM_symWeight);
      cwprint( str, strlen( str ), X, Y );
    }

    Y += SMALL_VSPACE;
    if (analogMode && considerWell) {   
      (void) sprintf(str,"WE %12.2f  %12.2f  ", (double) errorP->well, errorP->well * SM_wellWeight);
      cwprint( str, strlen( str ), X, Y );
    }

    Y += SMALL_VSPACE;
    if (analogMode && considerConstraintDriven) {   
      (void) sprintf(str,"CO %12.2f  %12.2f  ", (double) errorP->constraint, errorP->constraint * SM_constraintWeight);
      cwprint( str, strlen( str ), X, Y );
    }
}

int Confirm(es)
    char *es;
{
    int		timeout = 5;

    confirm = FALSE;
    XRaiseWindow( display, consoleWin );
    XRaiseWindow( display, ivWin->w );  XSync( display, False );
    Message( es );
    Warning( "Toggle the CONFIRM flag to confirm" );
    iv_ProcessAllEvents( display );
    sleep(timeout);
    if ( XPending( display ) == 0) {
	(void) sprintf(errmsg,"Timeout: you have %d seconds to confirm.", timeout);
	Message(errmsg);
	Warning("Not confirmed");
	return NOT_OK;
    }
    iv_ProcessAllEvents( display );
    if (confirm == TRUE) {
	Message("Confirmed");
	Warning("Confirmed");
	return OK;
    }
    Message("Not confirmed");
    Warning("Not confirmed");
    return NOT_OK;
}


#define TIMEOUT_DELAY 4
#define MESSAGE_LENGTH 70

displayWarning(es)
    char *es;
{
    SHARED static GC    gc = 0;
    SHARED static int   timeout = TIMEOUT_DELAY;
    SHARED static XColor warnColor;
    SHARED static int   warnCounter = 0;
    SHARED static char  lastWarning[256];
    int i;
  

    if ( gc == 0 ) {
	gc = DefaultGC( display, 0 );
	/* XText(consoleWin, 10, 160,
	 *  lastWarning, strlen(lastWarning), bigFont->id, fgConsolePixel, warnColor.pixel); 
	 * XParseColor("gold",&warnColor);
	 * XGetHardwareColor(&warnColor);
	 */
    }
    if (es != NULL) {
	(void) sprintf(lastWarning,"WRN:%4d: %-70s ", ++warnCounter, es);
	timeout = TIMEOUT_DELAY;
    } else {
	if (--timeout == 0) {
	    for (i=0;i<MESSAGE_LENGTH;i++) lastWarning[i] = ' ';
	    lastWarning[MESSAGE_LENGTH] = '\0';
	}
    }
    lastWarning[MESSAGE_LENGTH] = '\0';	/* truncate warning */
    XDrawImageString( display, consoleWin, gc, 10 , 172, lastWarning, MESSAGE_LENGTH );
}

void displayMessage(es)
    char *es;
{
    SHARED static GC 	gc = 0;
    SHARED static XColor msgColor;
    SHARED static int   initialized = FALSE;
    SHARED static int   msgCounter = 0;
    SHARED static char  lastMsg[256];
    SHARED static int   timeout = TIMEOUT_DELAY;
    int i;

    if ( gc == 0 ) {
	gc = DefaultGC( display, 0 );
/*
	XText(consoleWin, 10, 135, lastMsg, strlen(lastMsg), bigFont->id, fgConsolePixel, msgColor.pixel);
	XParseColor("ForestGreen",&msgColor);
	XGetHardwareColor(&msgColor);
	initialized  = TRUE;
*/
    }
    if (es != NULL) {
	(void) sprintf(lastMsg,"MSG:%4d: %-70s ", ++msgCounter, es);
	timeout = TIMEOUT_DELAY;
    } else {
	if (--timeout == 0) {
	    for (i=0;i<MESSAGE_LENGTH;i++) lastMsg[i] = ' ';
	    lastMsg[MESSAGE_LENGTH] = '\0';
	}
    }
    lastMsg[MESSAGE_LENGTH] = '\0'; /* truncate message  */
    XDrawImageString( display, consoleWin, gc,  10 , 155, lastMsg, MESSAGE_LENGTH );
}


void displayTemperature(t)
    double t;
{
    char str[40];
    int X = 10;
    int Y = 300;

    if (t == INFINITE_TEMPERATURE) {
	(void) sprintf(str," Temperature INFINITE ");
    } else {
	(void) sprintf(str," Temperature %12.2f ", t);
    }
    cwprint( str, strlen( str ), X, Y );
}

void displayCoolingSchedule(alpha,attPerCell)
    double 	alpha;
    int		attPerCell;
{
    char str[40];
    int X = 10;
    int Y = 320;

    (void) sprintf(str," Cooling Schedule: alpha %.2f ", alpha);
    cwprint( str, strlen( str ), X, Y );
    Y += SMALL_VSPACE;
    (void) sprintf(str," Attempts per cell: %d ", attPerCell);
    cwprint( str, strlen( str ), X, Y );
}

void displayChipBoundaryType()
{
    char str[40];
    int X = 210;
    int Y = 30;
    extern char *ChipBoundaryTypeName[];

    (void) sprintf(str, "Boundary: %-19s ", ChipBoundaryTypeName[(int) chipBoundaryType]);
    cwprint( str, strlen( str ), X, Y );
}

void displayShakingVector()
{
    char str[40];
    int X = 210;
    int Y = 15;

    (void) sprintf(str, "Shaking vector: %3d,%3d ", SM_shakeX, SM_shakeY);
    cwprint( str, strlen( str ), X, Y );
}

void displayGraphs()
{
    extern Cost curCost;	/* defined in sa.c */
    extern Cost stepError;	/* defined in sa.c */
    extern double acceptanceRatio;
    int i;

    if (SM_T != INFINITE_TEMPERATURE && SM_T != 0.0) {
	xgSendPoint(xgCostError, SM_T, curCost.total , "Cost");
	xgSendPoint(xgCostError, SM_T, stepError.total, "Error");
	xgSendPoint(xgAccRatio, SM_T, gCellMoveSet.globalStat.acceptanceRatio, "Cell");
	xgSendPoint(xgAccRatio, SM_T, gNetMoveSet.globalStat.acceptanceRatio, "Net");
	xgSendPoint(xgAccRatio, SM_T, gFloatTermMoveSet.globalStat.acceptanceRatio, "FltTrm");
	for (i = 1 ; i < gCellMoveSet.size ; i ++) {
	    if (gCellMoveSet.moveStatArray[i].enabled) 
	      xgSendPoint(xgQF, SM_T,
			   gCellMoveSet.moveStatArray[i].relativeFrequency,
			   gCellMoveSet.moveStatArray[i].name);
	}
    }
}

/* Give a sign that something is going on */
showActivity()
{
    static GC gc= 0;
    static Window activityWin = 0;
    static int	altColor[2];
    static int	colorSelect = 0;
    static int  counter = 0;
    char   	str[16];	/* the string to be printed  */

    if (XDisplayActive) {
	if ( activityWin == 0 && consoleWin != 0 )  {
	    altColor[0] = puppyGetColor( "ForestGreen" );
	    altColor[1] = puppyGetColor( "khaki" );
	    activityWin = XCreateSimpleWindow( display, consoleWin, 420, 100, 80, 15, 0, 
					      BlackPixel(display,0), altColor[1] );
	    XMapWindow( display, activityWin );
	    gc = DefaultGC( display, 0);
	}
	colorSelect = 1 - colorSelect;
	(void) sprintf(str," OK:%3d " , ++counter);
	XSetWindowBackground( display, activityWin, altColor[colorSelect] );
	XClearWindow( display, activityWin );
	XDrawString( display, activityWin, gc, 5 , 12 , str, strlen( str ));
	XFlush( display );
    } else {
	(void) printf("."); 
	(void) fflush(stdout);	/* echo on the terminal that something is going on */
    }
}


int
PM_showBB()
{
  showBB = !showBB;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_redraw()
{
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showNets()
{
  showNets = !showNets;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showConstrainedNets()
{
  showConstrainedNets = !showConstrainedNets;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showAxes()
{
  showAxes = !showAxes;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showBias()
{
  showBias = !showBias;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showWell()
{
  showWell = !showWell;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showBins()
{
  showBins = !showBins;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showOwner()
{
  showOwner = !showOwner;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_showAge()
{
  showAge = !showAge;
  plotCells();
  refreshConsoleWindow();
  return OK;
}

int
PM_toggleDebug()
{
  char *truefalse();

  debug = !debug;
  (void) sprintf(errmsg,"Debug: %s", truefalse(debug));
  Message(errmsg);
  return OK;
}


int
PM_exit()
{
  /*  int	rv;	*/		/* return code for PQdoJob() */

  if (Confirm("Do you really want to exit ?")==OK) {
    Warning("Exiting now");
#ifdef PARALLEL  
    manageChildren(0);
#endif
    /*   while (PQdoJob(outerLoopJobQueue, &rv)==OK); * the locks are dirty. Must fix this */ /* clean up the job queue */
    endX();
    VOVend(2);
  }
  return OK;
}

int
PM_save_exit()
{
  /*  int	rv;	*/		/* return code for PQdoJob() */

  if (Confirm("Do you really want to save and exit ?")==OK) {
    Warning("Saving and exiting  now");
#ifdef PARALLEL  
    manageChildren(0);
#endif
    /*   while (PQdoJob(outerLoopJobQueue, &rv)==OK); * the locks are dirty. Must fix this */ /* clean up the job queue */
    
    saveResultsToOct(TRUE);
    saveParFile(usingParameterFile, TRUE);
    saveParFile(usingParameterFile, FALSE);
    endX();
    VOVend(0);
  }
  return OK;
}



PM_centerInterchange()
{
  char *truefalse();

  enableCenterInterchange = !enableCenterInterchange;
  (void) sprintf(errmsg,"Interchanges of centers: %s", truefalse(enableCenterInterchange));
  Message(errmsg);
  return OK;
}


PM_saveParPuppy()
{
  extern int saveParFile();

  /* can be posponed until the inner loop is finished */
  PQaddJob(outerLoopJobQueue, saveParFile);
  Message("Save parameter file added to queue");
  return OK;
}

PM_saveParPuppyNow()
{
  extern int saveParFile();

  saveParFile();
  refreshConsoleWindow();
  return OK;
}

interactiveSaveResultsToOct()
{
    /* Save as a final placement */
    saveResultsToOct(TRUE);
    return OK;
}

PM_savePlacement()
{
    extern int interactiveSaveResultsToOct();

    /* can be posponed until the inner loop is finished */
    PQaddJob(outerLoopJobQueue, interactiveSaveResultsToOct);
    Message("Save placement added to queue");
    return OK;
}




char *truefalse(flag)
    int flag;
{
  return( flag ? "TRUE" : "FALSE");
}


void syncXConnection()
{
   if (XDisplayActive == TRUE ) {
     XSync( display, False );
   }
}

/* To make sure event queue is not srewed up with  */
void childCloseXConnection()
{
  int cn = 0;			/* Connection number */
  extern Display* display;
   if (XDisplayActive == TRUE ) {
     cn = XConnectionNumber(display);
     if (close( cn ) == -1 ) {
       perror("Error for child closing X connection");
       FatalError("");
     }
   }
}

































