/*
 *  FILE: move.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: Aug.21.1987		AUTHOR: Andrea Casotto
 *
 *
 *  FUNCTIONS:
 *	MoveAnnealVar(annealVar)
 * 	UpdateAges()
 *	Execute(move)
 * 
 *	SetMoveIds()
 *	ClearMoveIds()
 *	saAccept(delta, temp)
 * 	copyNew2Cur()
 *	ExecuteCellMovement()
 *	copyCur2New()
 *	orthogonalTransform(tr)
 *	newTermPosition(cell)
 *	newPerimeter(cell)
 *	BorrowCell(cell)
 *	RestituteCells(move)
 */

#include "defs.h"
#include "sa.h"
#include "constraintDrivenVariables.h"
#include "analogVariables.h"
#include "global.h"
#include "parall.h"

/* IMPORTED VARIABLES */
extern Cost curCost;            /* <--- sa.c */

/* GLOBAL VARIABLES IN THIS MODULE ("move.c") */
int acceptedFlag;               /* exported to analogCost.c for debug */
int inconsistentError = FALSE;  /* exported to analogCost.c for debug */
static int singleStep = FALSE;	/* flag for single stepping */

int
PM_singleMove()
{
  singleStep = ! singleStep;
  return(OK);
}

int
MoveAnnealVar(annealVar)
    AnnealVar	*annealVar;
{

  int		exitInnerLoop;   /* TRUE when cooling schedule decides it is time to exit. 
				  * It can be ignored
			          * It is the value returned by this routine */
  char          *moveType;
  Move		move;	         /* move descriptor */
  double 	moveCost();
  Cost		c1,c2;		 /* for debugging purposes */
  MacroCell     *otherCell;      /* debug */
  SymmetryType type = VERTICAL; /* debug */
  int           singleCell=FALSE;/* debug */ 
  octCoord      axis=200;

  extern void   plotCells();
  extern MacroCell    *searchOtherSymmetricCell();

  if (singleStep) {
    configurationCost(&c1);
  }

  (void) selectMove(annealVar, &move); 
  if (move.type == NO_MOVE) {
    if (annealVar->type == AV_CELL) {
      RestituteCells(&move);/* restitute all cells borrowed for this move (if any) */
    }
    zeroesCost(&move.cost);
    acceptedFlag = FALSE;
  } else {
    SetMoveIds(&move);		/* to compute the cost correctly */
    moveCost(&move);
    ClearMoveIds(&move);	/* we don't need the moveIds any more */
    acceptedFlag = saAccept(move.cost.total, SM_T);
    UpdateAges(&move , acceptedFlag); /* of the cells involved in this move */

    if (move.type != CHANGE_OWNERSHIP) {
      accumulateCosts(&myProc->sumMoveCost, &move.cost);
      accumulateCosts2(&myProc->sumMoveCost2, &move.cost);
    }

    if (acceptedFlag == TRUE) {    		/* EXECUTE the MOVE ? */
      Execute(&move);				/*  */
      /*
       * If the move is not a change of ownership
       * update the statistics about the cost of the move
       */
      if (move.type != CHANGE_OWNERSHIP) {
	accumulateCosts(&myProc->deltaCost, &move.cost);
	accumulateCosts(&myProc->sumCostAccMoves, &move.cost);
	accumulateCosts2(&myProc->sumCost2AccMoves, &move.cost);
      }
    }
    if (annealVar->type == AV_CELL) {
      RestituteCells(&move);			/* restitute all cells borrowed for this move */
    }
#ifdef PARALLEL
    if (move.type==CHANGE_OWNERSHIP) {
      NEWreleaseProcess(move.procNumber);
    }
#endif
  }
  updateMoveStatArray(&move,acceptedFlag); 	/* do this in any case */

  /* Hyper paranoid check of computation of move cost */
  if (singleStep) {
    configurationCost(&c2);
    if (acceptedFlag) {
      plotCells();
      if (ABS(c2.total - c1.total - move.cost.total) > 1.0) {
	Warning("Inconsistent move cost computation");
      }
    } else { 
      if (c2.total != c1.total) {
	Warning("Strangely inconsistent cost computation");
      }
    } 
  }
  exitInnerLoop = CS_innerLoop(move.cost.total, acceptedFlag);
  return(exitInnerLoop);
}




UpdateAges(moveP , acceptedFlag)
    Move *moveP;
    int  acceptedFlag;
{
  if (acceptedFlag == FALSE) {	/* EXECUTE the MOVE ? */
    if (moveP->type == SUBNET_MOVE) {
      moveP->net->age++;
    } else if (moveP->type == TERMINAL_MOVE) {
      /* moveP->term->age++; */
    } else {
      moveP->cell->age++;		/* increment age of the cell */
      if (moveP->cell2) moveP->cell2->age++; /* and of cell2 if it exists */
    }
  } else {
    if (moveP->type == SUBNET_MOVE) {
      moveP->net->age = 0;
    } else if (moveP->type == TERMINAL_MOVE) {
      moveP->term->age = 0;
    } else if (moveP->type != CHANGE_OWNERSHIP) {
      moveP->cell->age = 0;
      if (moveP->cell2) moveP->cell2->age = 0;
    }
  }
}

/* **** move ACCEPTED **** */
Execute(moveP)
     Move *moveP;
{
  int opn;			/* other process number, temporary  */
  int movingTermIndex;
  ListElt *subnetLE;

  switch (moveP->type) {
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
  case TRANSLATE_AND_TRANSFORM:
    ExecuteCellMovement(moveP->cell);	/* must be called before lyo replace */
    lyoReplace(moveP->cell);
    binReplace(moveP->cell);
#ifdef PARALLEL
    s_lock(&myProc->curCluster.lock); /* might be useless */
    myProc->curCluster = myProc->newCluster;
    s_unlock(&myProc->curCluster.lock);
#endif
    break;
  case TRANSFORMATION:
    ExecuteCellMovement(moveP->cell);
    lyoReplace(moveP->cell);
    binReplace(moveP->cell);
#ifdef PARALLEL
    s_lock(&myProc->curCluster.lock); /* might be useless */
    myProc->curCluster = myProc->newCluster;
    s_unlock(&myProc->curCluster.lock);
#endif
    break;
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
    ExecuteCellMovement(moveP->cell);
    lyoReplace(moveP->cell);
    binReplace(moveP->cell);
    ExecuteCellMovement(moveP->cell2);
    lyoReplace(moveP->cell2);
    binReplace(moveP->cell2);
#ifdef PARALLEL
    s_lock(&myProc->curCluster.lock);
    myProc->curCluster = myProc->newCluster;
    s_unlock(&myProc->curCluster.lock);
#endif
    break;
 case TERMINAL_MOVE:
#ifdef PARALLEL
    s_lock(&moveP->cell->floatingSitesLock);
#endif
    moveP->cell->floatingSites[moveP->term->newSite].terms++;
    moveP->cell->floatingSites[moveP->term->curSite].terms--;
#ifdef PARALLEL
    s_unlock(&moveP->cell->floatingSitesLock);
#endif
    moveP->term->curSite = moveP->term->newSite;
    moveP->term->curAbsCoordP = &moveP->cell->floatingSites[moveP->term->newSite].curAbsCoord;
    moveP->term->newAbsCoordP = &moveP->cell->floatingSites[moveP->term->newSite].newAbsCoord;
    break;
 case SUBNET_MOVE:
    /* Find index of moving term in subnet */
    movingTermIndex = moveP->subnetIndex;
    subnetLE = moveP->subnetLE;
    /* Unlink from old list and link to new list: also */
    /* update the subnet counts in both new and old term */
    moveP->subnet->curTerm[movingTermIndex]->subnets--;
    unlinkElt(&(moveP->subnet->curTerm[movingTermIndex]->firstSubnet),subnetLE);
    linkEltToListHandle(&(moveP->subnet->newTerm[movingTermIndex]->firstSubnet),subnetLE);
    moveP->subnet->newTerm[movingTermIndex]->subnets++;

    /*
     * Try to find a bug:
     */
    if ( moveP->subnet->curTerm[ 1 - movingTermIndex ] == moveP->subnet->newTerm[movingTermIndex] ) {
	(void) sprintf( errmsg, "%d: Cannot execute subnet move (this is a bug)", myself);
	Warning( errmsg );	/*  Because it would attach the subnet twice to the same term */
    } else {
	moveP->subnet->curTerm[movingTermIndex] = moveP->subnet->newTerm[movingTermIndex];
    }
    break;
  case CHANGE_OWNERSHIP:
    annealVarUnlink(moveP->annealVar);
    annealVarLink(moveP->annealVar , moveP->procNumber , LAST);
    moveP->annealVar->owner = moveP->procNumber;
    moveP->cell->owner = moveP->procNumber;
    /* at this point the cell belongs to the other process */
    myProc->curNumberOfCells = myProc->newNumberOfCells;
    { myProc->curCluster = myProc->newCluster; }

    opn = moveP->procNumber;
    process[opn].curNumberOfCells = process[opn].newNumberOfCells;
    { process[opn].curCluster = process[opn].newCluster; }

    break;
  default:
    break;
  }
}

/* This routine does nothing now */
DoNotExecute(moveP)
     Move *moveP;
{
  switch (moveP->type) {
  case SUBNET_MOVE:
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
  case TRANSLATE_AND_TRANSFORM:
  case TRANSFORMATION:
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
  case TERMINAL_MOVE:
  case CHANGE_OWNERSHIP:
  case NO_MOVE:
    break;
  default:
    FatalError("Wrong move type");
    break;
  }
}



SetMoveIds(moveP)
    Move *moveP;
{
  if (moveP->type != SUBNET_MOVE && moveP->type != TERMINAL_MOVE) {
    moveP->cell->moveId = myMoveId;
    if (moveP->cell2) moveP->cell2->moveId = myMoveId;
  }
}


ClearMoveIds(moveP)
    Move *moveP;
{
  if (moveP->type != SUBNET_MOVE && moveP->type != TERMINAL_MOVE) {
    moveP->cell->moveId = 0;
    if (moveP->cell2) moveP->cell2->moveId = 0;
  }
}


/*
 *	saAccept(delta)
 *	delta is defined as Oldcost - newcost
 */
int
saAccept(delta,temp)
     double delta;
     double temp;
{
  double r;			/* a random number */

  if (delta <= 0.0 || temp == INFINITE_TEMPERATURE) {
    return (OK);
  } else if (temp==0.0) {
    return(NOT_OK);
  } else {
    r = RANDOM;
    if (exp(- delta/temp) > r ) {
      return(OK);
    } else {
      return (NOT_OK);
    }
  }
}






/*
 *   copyNew2Cur(cell)
 *   Copies the new position of the cell into the current position
 *   The terminals are also updated
 */
copyNew2Cur(cell)
     MacroCell *cell;
{
  int	I;
  struct termList  *termPtr;

  cell->curCenter.x = cell->newCenter.x;
  cell->curCenter.y = cell->newCenter.y;
  cell->curTransform = cell->newTransform;
  cell->curBbox.lowerLeft.x = cell->newBbox.lowerLeft.x; 
  cell->curBbox.lowerLeft.y = cell->newBbox.lowerLeft.y; 
  cell->curBbox.upperRight.x = cell->newBbox.upperRight.x; 
  cell->curBbox.upperRight.y = cell->newBbox.upperRight.y; 

  for (termPtr=cell->term0.next; termPtr != NULL; termPtr=termPtr->next) {
    termPtr->term->curAbsCoordP->x=termPtr->term->newAbsCoordP->x;
    termPtr->term->curAbsCoordP->y=termPtr->term->newAbsCoordP->y;
  }
  for (I = 0 ; I < cell->edgesInPerim; I++) {
    cell->curPerimeter[I].x = cell->newPerimeter[I].x;
    cell->curPerimeter[I].y = cell->newPerimeter[I].y;
  }
}


ExecuteCellMovement(cell)
     MacroCell *cell;
{
  int		I;
  struct termList  *termPtr;

  cell->curCenter.x = cell->newCenter.x;
  cell->curCenter.y = cell->newCenter.y;
  cell->curTransform = cell->newTransform;

  for (termPtr=cell->term0.next; termPtr != NULL; termPtr=termPtr->next) {
    termPtr->term->curAbsCoordP->x = termPtr->term->newAbsCoordP->x;
    termPtr->term->curAbsCoordP->y = termPtr->term->newAbsCoordP->y;
  }
  for (I = 0 ; I < cell->edgesInPerim; I++) {
    cell->curPerimeter[I].x = cell->newPerimeter[I].x;
    cell->curPerimeter[I].y = cell->newPerimeter[I].y;
  }
  for ( I=0 ; I < cell->numberOfSites ; I++ ) {
    cell->floatingSites[I].curAbsCoord.x = cell->floatingSites[I].newAbsCoord.x;
    cell->floatingSites[I].curAbsCoord.y = cell->floatingSites[I].newAbsCoord.y;
  }
}



/*
 *   copyCur2New(cell)
 *   Copies the current position of the cell into the new position
 *   The terminals are also updated
 */
void
copyCur2New(cell)
     MacroCell *cell;
{
  struct termList  *termPtr;

  cell->curCenter.x = cell->curCenter.x;
  cell->newCenter.y = cell->curCenter.y;
  cell->newTransform = cell->curTransform;
  cell->newBbox.lowerLeft.x = cell->curBbox.lowerLeft.x; 
  cell->newBbox.lowerLeft.y = cell->curBbox.lowerLeft.y; 
  cell->newBbox.upperRight.x = cell->curBbox.upperRight.x; 
  cell->newBbox.upperRight.y = cell->curBbox.upperRight.y; 
  for (termPtr=cell->term0.next; termPtr != NULL; termPtr=termPtr->next) {
    termPtr->term->newAbsCoordP->x=termPtr->term->curAbsCoordP->x;
    termPtr->term->newAbsCoordP->y=termPtr->term->curAbsCoordP->y;
  }
}



/*
 *   orthogonalTransform(tr)
 *   Given the transformation tr, returns another transformation
 *	with inverted aspect ratio
 */
octTransformType
orthogonalTransform(tr)
    octTransformType tr;
{
  switch (tr) {
  case OCT_NO_TRANSFORM:
  case OCT_MIRROR_X:
  case OCT_MIRROR_Y:
  case OCT_ROT180:
    switch ((int) ceil( (double) RANDOM * 4)) {
    case 1:
      return(OCT_ROT90);
    case 2:
      return(OCT_ROT270);
    case 3:
      return(OCT_MX_ROT90);
    case 4:
      return(OCT_MY_ROT90);
    default:
      FatalError("no such transform");
    }
    break;
  case OCT_ROT90:
  case OCT_ROT270:
  case OCT_MX_ROT90:
  case OCT_MY_ROT90:
    switch ((int) ceil(RANDOM * 4)) {
    case 1:
      return(OCT_NO_TRANSFORM);
    case 2:
      return(OCT_MIRROR_X);
    case 3:
      return(OCT_MIRROR_Y);
    case 4:
      return(OCT_ROT180);
    default:
      FatalError("no such transform");
    }
    break;
  default:
    (void) sprintf(errmsg,"illegal transformation %d\n", tr);
    FatalError(errmsg);
    break;
  }
  return((octTransformType) 0);		/* but it will never get down here */
}



/*
 * This is a critical routine. It is fully expanded for speed.
 * I'll try to speed it up a little more.
*/
int
newTermPosition(cell)
     MacroCell *cell;
{
  register struct termList 	*termPtr;
  register octCoord		cx,cy;
  register int			i;        /* integer used for update of floating sites */
  
  cx = cell->newCenter.x ;	/* for speed */
  cy = cell->newCenter.y ;	/* for speed */
 
  switch (cell->newTransform) {
 case OCT_NO_TRANSFORM : 
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx + termPtr->term->cellCoord.x;
      termPtr->term->newAbsCoordP->y = cy + termPtr->term->cellCoord.y;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx+cell->floatingSites[i].cellCoord.x;
	cell->floatingSites[i].newAbsCoord.y = cy+cell->floatingSites[i].cellCoord.y;
      }
    }
    break;
 case OCT_MIRROR_X:
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx - termPtr->term->cellCoord.x;
      termPtr->term->newAbsCoordP->y = cy + termPtr->term->cellCoord.y;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx - cell->floatingSites[i].cellCoord.x;
	cell->floatingSites[i].newAbsCoord.y = cy + cell->floatingSites[i].cellCoord.y;
      }
    }
    break;
 case OCT_MIRROR_Y:
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx + termPtr->term->cellCoord.x;
      termPtr->term->newAbsCoordP->y = cy - termPtr->term->cellCoord.y;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx + cell->floatingSites[i].cellCoord.x;
	cell->floatingSites[i].newAbsCoord.y = cy - cell->floatingSites[i].cellCoord.y;
      }
    }
    break;
 case OCT_ROT90 :
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx - termPtr->term->cellCoord.y;
      termPtr->term->newAbsCoordP->y = cy + termPtr->term->cellCoord.x;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx - cell->floatingSites[i].cellCoord.y;
	cell->floatingSites[i].newAbsCoord.y = cy + cell->floatingSites[i].cellCoord.x;
      }
    }
    break;
 case OCT_ROT180 :
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx - termPtr->term->cellCoord.x;
      termPtr->term->newAbsCoordP->y = cy - termPtr->term->cellCoord.y;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx - cell->floatingSites[i].cellCoord.x;
	cell->floatingSites[i].newAbsCoord.y = cy - cell->floatingSites[i].cellCoord.y;
      }
    }
    break;
 case OCT_ROT270 :
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx + termPtr->term->cellCoord.y;
      termPtr->term->newAbsCoordP->y = cy - termPtr->term->cellCoord.x;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx + cell->floatingSites[i].cellCoord.y;
	cell->floatingSites[i].newAbsCoord.y = cy - cell->floatingSites[i].cellCoord.x;
      }
    }
    break;
 case OCT_MX_ROT90:
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx - termPtr->term->cellCoord.y;
      termPtr->term->newAbsCoordP->y = cy - termPtr->term->cellCoord.x;
    }
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx - cell->floatingSites[i].cellCoord.y;
	cell->floatingSites[i].newAbsCoord.y = cy - cell->floatingSites[i].cellCoord.x;
      }
    }
    break;
 case OCT_MY_ROT90:
    termPtr= &(cell->term0);
    while ((termPtr=termPtr->next) != NULL) {
      termPtr->term->newAbsCoordP->x = cx + termPtr->term->cellCoord.y;
      termPtr->term->newAbsCoordP->y = cy + termPtr->term->cellCoord.x;
    }  
    if (cell->floatingTerms>0) {
      for (i=0; i<cell->numberOfSites; i++) {
	cell->floatingSites[i].newAbsCoord.x = cx + cell->floatingSites[i].cellCoord.y;
	cell->floatingSites[i].newAbsCoord.y = cy + cell->floatingSites[i].cellCoord.x;
      }
    }
    break;
 default:
    FatalError("Wrong transform in newTermPosition");
  }
}

newPerimeter(cell)
     MacroCell *cell;
{
  int 		I;
  octCoord	cx,cy;
  
  cx = cell->newCenter.x ;	/* for speed */
  cy = cell->newCenter.y ;	/* for speed */
 
  switch (cell->newTransform) {
 case OCT_NO_TRANSFORM : 
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx + cell->perimeter[I].x;
      cell->newPerimeter[I].y = cy + cell->perimeter[I].y;
    }
    break;
 case OCT_MIRROR_X:
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx - cell->perimeter[I].x;
      cell->newPerimeter[I].y = cy + cell->perimeter[I].y;
    }
    break;
 case OCT_MIRROR_Y:
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx + cell->perimeter[I].x;
      cell->newPerimeter[I].y = cy - cell->perimeter[I].y;
    }
    break;
 case OCT_ROT90 :
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx - cell->perimeter[I].y;
      cell->newPerimeter[I].y = cy + cell->perimeter[I].x;
    }
    break;
 case OCT_ROT180 :
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx - cell->perimeter[I].x;
      cell->newPerimeter[I].y = cy - cell->perimeter[I].y;
    }
    break;
 case OCT_ROT270 :
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx + cell->perimeter[I].y;
      cell->newPerimeter[I].y = cy - cell->perimeter[I].x;
    }
    break;
 case OCT_MX_ROT90:
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx - cell->perimeter[I].y;
      cell->newPerimeter[I].y = cy - cell->perimeter[I].x;
    }
    break;
 case OCT_MY_ROT90:
   for (I = 0 ; I < cell->edgesInPerim ; I++) {
      cell->newPerimeter[I].x = cx + cell->perimeter[I].y;
      cell->newPerimeter[I].y = cy + cell->perimeter[I].x;
    }  
    break;
 default:
    FatalError("Wrong transform in newPerimeter");
  }
}


#ifdef PARALLEL
/* cell is owned by another processor, but I want to move it
 * borrow the cell from the processor.
 * Before entering this routine, the owner of this cell must be 
 * stopped
 */
BorrowCell(cell)
    MacroCell	*cell;
{
  int	owner;

  owner = cell->owner;		/* ATTENTION!!  cell-owner could change */

  if (cell->status !=  OWNED) {
    return(FALSE);		/* somebody has already borrowed the cell */
  }

  /* could improve if owner == myself && status == RETURNED */
  if (owner == myself) return(TRUE); /* no need to borrow anything */

  if (NEWstopProcess(owner) == FALSE) {
    return(FALSE);
  }
  if (cell->owner != owner) {
    if (debug) Warning("Caught an ownership change on the fly. How nice !!");
    NEWreleaseProcess(owner);
    return(FALSE);		/* We stopped the wrong process */
  }    
  if (cell->status != OWNED) {	/* must check this again !! */
    NEWreleaseProcess(owner);
    return(FALSE);		/* somebody has already borrowed the cell */
  }
  cell->status = BORROWED;	/* it is BORROWED, so its owner cannot move it */
  SubtractCellToCluster(&process[owner].curCluster,cell);
  NEWreleaseProcess(owner);
  return(TRUE);
}
#endif


/*
 * cell is owned by another processor, and I tried to move it
 * Now I want to restitute the cell to its owner
 * Before entering this routine, the owner of this cell must be 
 * stopped
 */
RestituteCells(moveP)
    Move *moveP;
{
  if (moveP->cell2 && moveP->cell2->status == BORROWED) {
    moveP->cell2->status = RETURNED;
  }
}
