/*
 *  FILE: moveSet.c
 *
 *  DESCRIPTION: 
 * 
 *
 *
 *  CREATION DATE: 2.24.88		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS: 
 */
#include "defs.h"
#include "global.h"
#include "parall.h"

#define MIN_SELECTION_FREQUENCY         0.03 		/* 0.03 means 3% */


struct {
  double	relFreq;	/* relative Frequency */
  MoveType 	moveType;
  char    	*name;
  annealVarType varType;
} moveDesc[] = {
  { 0.00, NO_MOVE , 		"NO_MOVE" , 		AV_ANY },
  { 5.00, TRANSLATION ,		"TRANSLATION" , 	AV_CELL },
  { 5.00, ALIGNMENT ,		"ALIGNMENT" ,   	AV_CELL },
  { 5.00, RL_TRANSLATION ,	"RANGE_LIMITED" , 	AV_CELL },
  { 5.00, TRANSFORMATION ,	"TRANSFORMATION" , 	AV_CELL },
  { 5.00, TRANSLATE_AND_TRANSFORM, "TRANS_AND_TRANSF",	AV_CELL },
  { 3.00, CC_INTERCHANGE ,	"CC_INTERCHANGE" , 	AV_CELL },
  { 3.00, BB_INTERCHANGE ,	"BB_INTERCHANGE" , 	AV_CELL },
  { 0.00, MULTIPLE_CELL ,	"MULTIPLE_CELL" , 	AV_CELL },
  { 1.00, CHANGE_OWNERSHIP,	"CHANGE_OWNERSHIP" , 	AV_CELL },
  { 1.00, SUBNET_MOVE ,		"SUBNET_MOVE",		AV_NET },
  { 1.00, TERMINAL_MOVE ,	"TERM_MOVE"  , 		AV_FLOAT_TERM },
  { 0.00, NUMBER_OF_MOVE_TYPES , "SENTINEL(UNUSED)" ,	AV_NONE } /* must be last in the array */
};


initGlobalMoveSets()
{
  initMoveSet(&gCellMoveSet, AV_CELL);
  initMoveSet(&gNetMoveSet, AV_NET);
  initMoveSet(&gFloatTermMoveSet, AV_FLOAT_TERM);
}

int
initMoveSet(moveSet, type)
    MoveSet *moveSet;
    annealVarType type;
{
  int moveSetSize = 0;
  int moveIndex = 0;
  int i = 0 ;

  /*
   * First count the number of moves in the set
   * then allocate the arrays
   * and finally, fill them
   */

  while (moveDesc[i].moveType != NUMBER_OF_MOVE_TYPES) {
    if (moveDesc[i].varType == type || moveDesc[i].varType == AV_ANY) {
      moveSetSize++;
    }
    i += 1;
  }
  
  moveSet->size = moveSetSize;
  moveSet->moveStatArray = SH_MALLOC(struct moveStat , moveSet->size);
  moveSet->moveSelArray = SH_MALLOC(struct moveSelect , moveSet->size);
 
  i = 0;
  while (moveDesc[i].moveType != NUMBER_OF_MOVE_TYPES) {
    if (moveDesc[i].varType == type || moveDesc[i].varType == AV_ANY) {
      moveSet->moveStatArray[moveIndex].type = moveDesc[i].moveType;
      moveSet->moveStatArray[moveIndex].name = moveDesc[i].name;
      moveSet->moveStatArray[moveIndex].attempted = 0;
      moveSet->moveStatArray[moveIndex].accepted = 0;
      moveSet->moveStatArray[moveIndex].acceptanceRatio = 0.0;
      moveSet->moveStatArray[moveIndex].qf = 0.0;
      moveSet->moveStatArray[moveIndex].movingAverageArray = (double *) 0;
      moveSet->moveStatArray[moveIndex].enabled = isEnabled(moveDesc[i].moveType);
      if (moveSet->moveStatArray[moveIndex].enabled == TRUE) {
	moveSet->moveStatArray[moveIndex].relativeFrequency = moveDesc[i].relFreq;
      } else {
	moveSet->moveStatArray[moveIndex].relativeFrequency = 0.0;
      }
      
      zeroesCost(&moveSet->moveStatArray[moveIndex].sumCost);
      zeroesCost(&moveSet->moveStatArray[moveIndex].sumAbsCost);
      moveIndex++;
    }
    i += 1;
  }
  return(OK);
}

isEnabled(type)
  MoveType type;
{
  int enable = TRUE;

  switch (type) {
  case NO_MOVE:
  case MULTIPLE_CELL:
    enable = FALSE;
    break;
  case TRANSLATION:
  case ALIGNMENT:
  case RL_TRANSLATION:
    enable = enableTranslations;
    break;
  case TRANSFORMATION:
    enable = enableTransformations;
    break;
  case TRANSLATE_AND_TRANSFORM:
    enable = enableTranslations && enableTransformations;
    break;
  case CC_INTERCHANGE:
  case BB_INTERCHANGE:
  case SUBNET_MOVE:	       
  case TERMINAL_MOVE:		
    enable = TRUE;
    break;
  case CHANGE_OWNERSHIP:
#ifdef PARALLEL    
    enable = (P > 1) ?  TRUE : FALSE;
#else 
    enable = FALSE;
#endif
    break;
  default:
    FatalError("No such move type");
    break;
  }
  return(enable);
}




initMoveStatArray(array)
    struct moveStat *array;
{
  int		i;		/* counter */
  int		enable;		/* enable flag  */
  double 	rf;		/* Relative frequency, temporary */
  
  for (i=0; i<(int)NUMBER_OF_MOVE_TYPES;i++) {
    switch ((MoveType) i) {
    case NO_MOVE:
      enable = FALSE;
      rf = 0.0;
      break;
    case TRANSLATION:
    case ALIGNMENT:
    case RL_TRANSLATION:
      rf = (enableTranslations) ? 5.0 : 0.0 ;
      enable = (enableTranslations) ? TRUE : FALSE;
      break;
    case TRANSFORMATION:
      rf = (enableTransformations) ? 5.0 : 0.0 ;
      enable = enableTransformations;
      break;
    case TRANSLATE_AND_TRANSFORM:
      rf = (enableTransformations && enableTranslations) ? 5.0 : 0.0 ;
      enable = TRUE;
      break;
    case CC_INTERCHANGE:
      rf = 3.0;
      enable = TRUE;
      break;
    case BB_INTERCHANGE:
      rf = 3.0;
      enable = TRUE;
      break;
    case MULTIPLE_CELL:
      rf = 0.0;			/* not yet implemented */
      enable = FALSE;
      break;
    case CHANGE_OWNERSHIP:
#ifdef PARALLEL
      rf =  (P==1) ? 0 : 1;
      enable = TRUE;
#else
      rf = 0.0;
      enable = FALSE;
#endif
      break;
    case SUBNET_MOVE:		/* OBSOLETE */
    case TERMINAL_MOVE:		/* OBSOLETE */
      rf = 0.0;
      enable = FALSE;
      break;
    default:
      FatalError("No such move type");
      break;
    }
    array[i].relativeFrequency = rf; /* default, should be changed */
    array[i].enabled = enable;
    array[i].attempted = 0;
    array[i].accepted  = 0;
    array[i].movingAverageArray = (double *) 0;
    zeroesCost(&array[i].sumCost);
  }

}

double 
movingAverage(array,new,n)
    double **array;
    double new;			/* latest value */
    int	   n;			/* number of samples to average */
{
  int index;

  /* The array is so organized:
   * (*array)[0] = running sum
   * (*array)[1] = current index
   * (*array)[2] = max index
   * (*array)[3 .. 2+n] time series
   */
  if ((*array) == (double *) 0) {
    (*array) = SH_MALLOC(double , n+2);
    (*array)[0] = 0.0;
    (*array)[1] = 3;
    (*array)[2] = n + 2.0;
    for (index = 3 ; index < (int) (*array)[2] ; index++) {
      (*array)[index] = 0.0;		/*  */
    }
  }
  if (n + 2 != (*array)[2] )  {
    (void) sprintf(errmsg,"Cannot change interval of running average (%d hould be %d)",
		       n , (int) (*array)[2] - 2);
    FatalError(errmsg);
  }
  index = (*array)[1];
  (*array)[0] = (*array)[0] + new - (*array)[index];
  (*array)[index] = new;
  index++; 
  if (index >= (*array)[2]) index = 3;
  (*array)[1] = index;
  return( (*array)[0] / n);
}

computeQualityFactor(moveSet)
    MoveSet *moveSet;
{
  int		i;		/* a counter */
  double	qf;		/* Quality factor */
  double 	movingAverage();
  struct moveStat *statArray = moveSet->moveStatArray;

  /* compute the quality factor for each kind of move */
  for (i=0; i< moveSet->size;i++) {
    if (statArray[i].attempted > 2) {
      qf = statArray[i].sumAbsCost.total  / statArray[i].attempted;
      
      /* get a moving average of qf in order to stabilize a little better */
      statArray[i].qf = movingAverage(&statArray[i].movingAverageArray, qf, 12) ;
    }
  }
  if (debug) {
    printf("Move statistics array:\n");
    for (i=0; i < moveSet->size;i++) {
      printf("%20s: attempted/accepted %6d/%6d\n", statArray[i].name,
	     statArray[i].attempted, statArray[i].accepted);
    }
  }
}


resetMoveStatArray(moveSet)
    MoveSet *moveSet;
{
  struct moveStat *statArray = moveSet->moveStatArray;
  int 	i;
 
  for (i=0; i< moveSet->size; i++) {
    statArray[i].attempted = 0;
    statArray[i].accepted = 0;
    zeroesCost(&(statArray[i].sumCost));
    zeroesCost(&(statArray[i].sumAbsCost));
  }
}

/*
 * This function reorders the array used for the move selection so that
 * the most frequent moves types come first in the array.
 * If the flag 'useQualityFactor' is TRUE, then new frequencies for the moves are 
 * computed based on the quality factor of each type of move
 */
reorderMoveSelArray(moveSet)
    MoveSet *moveSet;
{
  struct moveStat *statArray = moveSet->moveStatArray;
  struct moveSelect *selArray = moveSet->moveSelArray;
  double	sumRelFrequencies = 0.0;
  double	minRelFreq;			/* minimum relative frequency of moves */
  double	thresh = 0.0;			/* threshold, temporary */
  int		i;				/* a counter */
  int		max;				/* used in the sorting routine */
  int		j;				/* used in the sorting routine */

  if (useQualityFactor==TRUE) {
    /* Make the frequency of selection proportional to the quality factor */
    /* However, make sure the frequency is above the minimum */
    for (i=0; i< moveSet->size;i++) {
      if (statArray[i].enabled) {
	statArray[i].relativeFrequency = statArray[i].qf;
	sumRelFrequencies += statArray[i].relativeFrequency;
      } else {
	statArray[i].relativeFrequency = 0.0;
      }
    }
    if (sumRelFrequencies == 0.0) {
      /* This happens when all QF's are 0.0, i.e. when no move of any kind is accepted*/
      /* Normalize to allow prosecution */
      sumRelFrequencies = 1.0;
    }
    minRelFreq = sumRelFrequencies * MIN_SELECTION_FREQUENCY;
    for (i=0; i<moveSet->size;i++) {
      if (statArray[i].relativeFrequency < minRelFreq && statArray[i].enabled == TRUE) {
	statArray[i].relativeFrequency = minRelFreq;
      }
    }
  }
  sumRelFrequencies = 0.0;
  for (i=0; i<moveSet->size;i++) {
    sumRelFrequencies += statArray[i].relativeFrequency;
  }
  for (i=0; i<moveSet->size;i++) {
    statArray[i].relativeFrequency /= sumRelFrequencies;
    selArray[i].freq = statArray[i].relativeFrequency;
    selArray[i].type = statArray[i].type;
    selArray[i].index = i;
  }
  /* now reorder selArray in decreasing values of freq */
  /* insertion sort */
  for (i=0; i< moveSet->size - 1;i++) {
    max = i;
    /* find largest freq */
    for (j = i ; j < moveSet->size ; j++) {
      if (selArray[j].freq > selArray[max].freq) max = j;
    }
    /* put the largest element in position `i' */
    if (i!=max) {
      SWAP(MoveType,selArray[max].type,selArray[i].type);
      SWAP(double,selArray[max].freq,selArray[i].freq);
      SWAP(int , selArray[max].index , selArray[i].index);
    }
  }

  /* set up the thresholds */
  for (i=0; i<moveSet->size;i++) {
    thresh += selArray[i].freq;
    selArray[i].thresh = thresh;
  }
  if (thresh != 1.0 && debug) {
    (void) sprintf(errmsg,"Double precision inaccuracy detected, %g should be %g,(diff = %g)",
	    thresh, 1.0, thresh-1.0);
    Warning(errmsg);
  }
  if (debug) {
    printf("Move selection array:\n");
    for (i=0; i<moveSet->size;i++) {
      printf("%2d :%20s: f %.3f : tsh %.3f : QF %8.3e : %s\n",
	     i,
	     statArray[selArray[i].index].name,
	     selArray[i].freq,
	     selArray[i].thresh,
	     statArray[selArray[i].index].qf,
	     statArray[selArray[i].index].enabled == TRUE ? "--------":"DISABLED");
    }
  }
}


/* this function is called in the inner looop, for each move */
updateMoveStatArray(moveP, acceptedFlag)
    Move	*moveP;
    int		acceptedFlag;
{
  struct moveStat *statArray;	/* moveStatArray */
  int	index;

  switch (moveP->annealVar->type) {
  case AV_CELL:
    if (useCellBasedMoveStats && moveP->annealVar->type == AV_CELL) {
      statArray = moveP->cell->moveStatArray;
    } else {
      statArray = gCellMoveSet.moveStatArray;
    }
    break;
  case AV_NET:
    statArray = gNetMoveSet.moveStatArray;
    break;
  case AV_FLOAT_TERM:
    statArray = gFloatTermMoveSet.moveStatArray;
    break;
  default:
    FatalError("in updateMoveStatArray");
    break;
  }

#ifdef PARALLEL
  s_lock(&moveStatArrayLock);
#endif
  myProc->moveAttempted++;
  index = moveP->index;		/* of the move type in the statArray */
  statArray[index].attempted++;
  if (acceptedFlag) {
    myProc->moveAccepted++;
    statArray[index].accepted++;
    accumulateCosts(&(statArray[index].sumCost), &(moveP->cost));
    accumulateAbsCosts(&(statArray[index].sumAbsCost), &(moveP->cost));
  }
#ifdef PARALLEL
  s_unlock(&moveStatArrayLock);
#endif
}

computeAccRatio(moveSet)
    MoveSet *moveSet;
{
  struct moveStat *gs;		/* global stats */
  struct moveStat *sa;		/* array of statistic */
  int	 i;

  sa = moveSet->moveStatArray;
  gs = &(moveSet->globalStat);
  gs->attempted = 0;
  gs->accepted  = 0;
  for (i = 0 ; i < moveSet->size ; i++) {
    sa[i].acceptanceRatio = (sa[i].attempted > 0) ? (double) sa[i].accepted / (double) sa[i].attempted : 0.0;
    gs->attempted += sa[i].attempted;
    gs->accepted += sa[i].accepted;
  }
  gs->acceptanceRatio = (gs->attempted > 0) ? (double) gs->accepted / (double) gs->attempted : 0.0;
}
