/*
 *    netBbox.c
 *    routines to compute the bounding box of the nets
 *>>>> ROUTINES
 *    allNetNewBbox(cell, move, cell2)
 *    curNetBbox(cell)
 */


#include "defs.h"
#include "sa.h"
#include "global.h"

/*
 *	allNetNewBbox(cell,move,cell2)
 *	computes the bounding box for all the  nets
 *	attached to cell (and cell2)
 *	given that cell has been move'd (or interchanged with cell2)
 *	No side effects other than net->newBbox
 *	
 *   MUST find a smarter way to do this. make a list of nets, with handle in moveP->nets
 */
int

#ifdef NEVER

int
allNetNewBbox(moveP)
     Move *moveP;
{
  register struct termList *termPtr;
  register struct netList  *n1,*n2;
  register struct net	     *net;
  octCoord	left,right,top,bottom;	/* bound. box edges */

  /* new version: two pointers, one for cell, the other for cell2 */
  /* we run through the list of nets, processing the nets avoiding repetitions */
  
  if (moveP->complexity <= 2) {
    n1 = moveP->cell->net0.next;
    n2 = moveP->cell2->net0.next;

    while (n1 != NULL || n2 != NULL) {
      if (n1 == NULL) {		/* list 1 is finished */
	net = n2->net; n2 = n2->next; /* take net in list 2 */
      } else if (n2 == NULL) {	/* list 2 is finished */
	net = n1->net; n1 = n1->next; /* take net in list 1 */
      } else if (n1->net->number < n2->net->number) { /* net in list 1 less than net in list 2 */
	net = n1->net; n1 = n1->next; /* take net in list 1 */
      } else if (n2->net->number < n1->net->number) { /* net in list 1 less than net in list 2 */
	net = n2->net; n2 = n2->next; /* take net in list 2 */
      } else {			/* Same net. Advance both pointers */
	net = n1->net; n1 = n1->next; n2 = n2->next;
      }
      /* compute new BB of 'net' */
      termPtr = net->term0.next;
      if (termPtr == NULL) FatalError("Error: No terminals in this net");
      if (termPtr->term->cell == moveP->cell OR termPtr->term->cell == moveP->cell2) {
	/* use new coordinates */
	left   = termPtr->term->newAbsCoordP->x;
	bottom = termPtr->term->newAbsCoordP->y;
	right  = termPtr->term->newAbsCoordP->x;
	top    = termPtr->term->newAbsCoordP->y;
      } else {
	/* use current coordinates */
	left   = termPtr->term->curAbsCoordP->x;
	bottom = termPtr->term->curAbsCoordP->y;
	right  = termPtr->term->curAbsCoordP->x;
	top    = termPtr->term->curAbsCoordP->y;
      }
      
      while ((termPtr=termPtr->next) != NULL) {
	if (termPtr->term->cell == moveP->cell OR termPtr->term->cell == moveP->cell2) {
	  /* use the NEW coordinates */
	  left   = MIN(left, termPtr->term->newAbsCoordP->x);
	  bottom = MIN(bottom,  termPtr->term->newAbsCoordP->y);
	  right  = MAX(right,  termPtr->term->newAbsCoordP->x);
	  top    = MAX(top,  termPtr->term->newAbsCoordP->y);
	} else {
	  /* use the CURRENT coordinates */
	  left   = MIN(left, termPtr->term->curAbsCoordP->x);
	  bottom = MIN(bottom,  termPtr->term->curAbsCoordP->y);
	  right  = MAX(right,  termPtr->term->curAbsCoordP->x);
	  top    = MAX(top,  termPtr->term->curAbsCoordP->y);
	}
      }
      net->newBbox[myself].lowerLeft.x = left;
      net->newBbox[myself].lowerLeft.y = bottom;
      net->newBbox[myself].upperRight.x = right;
      net->newBbox[myself].upperRight.y = top;
    }				/* while loop */
  } else {
    /* move complexity is > 2, must use a more complex algorithm to pik the nets to process */
    Warning("Move complexity > 2: code not yet implemented");
  }
}

#endif


newNetBBox(net,boxP)
    struct net *net;
    octBox *boxP;
{
  register struct termList 	*termPtr;
  register octCoord		left,right,top,bottom;	/* bound. box edges */

  /* compute new BB of 'net' */
  termPtr = net->term0.next;
  if (termPtr->term->cell->moveId == myMoveId) {
    /* use new coordinates */
    left   = termPtr->term->newAbsCoordP->x;
    bottom = termPtr->term->newAbsCoordP->y;
    right  = termPtr->term->newAbsCoordP->x;
    top    = termPtr->term->newAbsCoordP->y;
  } else {
    /* use current coordinates */
    left   = termPtr->term->curAbsCoordP->x;
    bottom = termPtr->term->curAbsCoordP->y;
    right  = termPtr->term->curAbsCoordP->x;
    top    = termPtr->term->curAbsCoordP->y;
  }
      
  while ((termPtr=termPtr->next) != NULL) {
    if (termPtr->term->cell->moveId == myMoveId ) {
      /* use the NEW coordinates */
      left   = MIN(left, termPtr->term->newAbsCoordP->x);
      bottom = MIN(bottom,  termPtr->term->newAbsCoordP->y);
      right  = MAX(right,  termPtr->term->newAbsCoordP->x);
      top    = MAX(top,  termPtr->term->newAbsCoordP->y);
    } else {
      /* use the CURRENT coordinates */
      left   = MIN(left, termPtr->term->curAbsCoordP->x);
      bottom = MIN(bottom,  termPtr->term->curAbsCoordP->y);
      right  = MAX(right,  termPtr->term->curAbsCoordP->x);
      top    = MAX(top,  termPtr->term->curAbsCoordP->y);
    }
  }
  boxP->lowerLeft.x = left;
  boxP->lowerLeft.y = bottom;
  boxP->upperRight.x = right;
  boxP->upperRight.y = top;
}

curNetBBox(net,boxP)
    struct net *net;
    octBox *boxP;
{
  register struct termList *termPtr;
  register octCoord   left, right, top, bottom;  /* bbox edges */

  termPtr= net->term0.next;
  left   = termPtr->term->curAbsCoordP->x;
  bottom = termPtr->term->curAbsCoordP->y;
  right  = termPtr->term->curAbsCoordP->x;
  top    = termPtr->term->curAbsCoordP->y;

  while ((termPtr=termPtr->next) != NULL) {
    left   = MIN(left , termPtr->term->curAbsCoordP->x);
    bottom = MIN(bottom, termPtr->term->curAbsCoordP->y);
    right  = MAX(right, termPtr->term->curAbsCoordP->x);
    top    = MAX(top , termPtr->term->curAbsCoordP->y);
  }
  boxP->lowerLeft.x = left;
  boxP->lowerLeft.y = bottom;
  boxP->upperRight.x = right;
  boxP->upperRight.y = top;
}



