/*
*	curTotOverlap(cell)
*	computes the total area of overlap for cell
*	in the current position
*	Consider that the `lyo' data structure might be changing
*	because of the effect of the other processes
*/
#include "defs.h"
#include "global.h"

#ifdef UNUSED_FUNCTION
curTotOverlap(cell)
    MacroCell *cell;
{
  octCoord 		left,right,bottom,top;
  int			totCellOverlap;
  int 			curOverlap(), newOverlap();
  MacroCell 		*lyoPtr;

  left  =cell->curBbox.lowerLeft.x;
  bottom=cell->curBbox.lowerLeft.y;
  right =cell->curBbox.upperRight.x;
  top   =cell->curBbox.upperRight.y;

  totCellOverlap = 0;

  /* **** start from PUPPY_L edge to the left **** */
  lyoPtr=cell;			/* always safe */	
  do {
    lyoPtr = lyoPtr->lyoPrec[PUPPY_L]; /* might be a strange place */
    if ( lyoPtr== gFirstCell || lyoPtr == NULL ||
	lyoPtr->curBbox.lowerLeft.y >= top ||
	lyoPtr->curBbox.upperRight.x <= left ||
	lyoPtr->curBbox.upperRight.y <= bottom ) {

	continue;
      } else {			/* **** OVERLAP !!!! */
	totCellOverlap += curOverlap(cell, lyoPtr);
      };
  } while ( lyoPtr != gFirstCell && lyoPtr != NULL &&
    (left-lyoPtr->curBbox.lowerLeft.x) < maxEdgeX &&
      lyoPtr->lyoPrec[PUPPY_L] != gFirstCell );

  /* **** start from PUPPY_L edge to the right **** */
  if (cell->lyoNext[PUPPY_L] != NULL) {
    lyoPtr = cell ;		/* safe */
    do {
      lyoPtr=lyoPtr->lyoNext[PUPPY_L];
      if ( lyoPtr== gFirstCell || lyoPtr == NULL ||
	  lyoPtr->curBbox.lowerLeft.y >= top ||
	  lyoPtr->curBbox.lowerLeft.x >= right ||
	  lyoPtr->curBbox.upperRight.y <= bottom ) {

	  continue;
	} else {		/* **** OVERLAP !!!! */
	  totCellOverlap += curOverlap(cell, lyoPtr);
	};
    } while ( lyoPtr != NULL && lyoPtr != gFirstCell &&
      lyoPtr->curBbox.lowerLeft.x < right &&
	lyoPtr->lyoNext[PUPPY_L] != NULL );
  }
  return(totCellOverlap);
}
#endif

/*
*	newTotOverlap(cell)
*	computes the total area of overlap for cell
*	in the newrent position
*/


#ifdef UNUSED_FUNCTION
newTotOverlap(cell)
    MacroCell *cell;
{
  octCoord 	left,right,bottom,top;
  int 		totCellOverlap, newOverlap();
  MacroCell 	*lyoPtr, *startLyoPtr;

  left   = cell->newBbox.lowerLeft.x;
  bottom = cell->newBbox.lowerLeft.y;
  right  = cell->newBbox.upperRight.x;
  top    = cell->newBbox.upperRight.y;

  totCellOverlap = 0;

  /* look for overlapping cells */
  /* start from old layout position of the cell */

  startLyoPtr=cell;

  /*  find the new position in the layout */
  while ( startLyoPtr->lyoNext[PUPPY_L] != NULL && 
	 left > startLyoPtr->lyoNext[PUPPY_L]->curBbox.lowerLeft.x ) {
	 startLyoPtr = startLyoPtr->lyoNext[PUPPY_L];
       }
  while ( startLyoPtr->lyoPrec[PUPPY_L] != gFirstCell &&
	 left < startLyoPtr->curBbox.lowerLeft.x) {
	 startLyoPtr = startLyoPtr->lyoPrec[PUPPY_L];
       }

  /* **** start from PUPPY_L edge to the left **** */
  lyoPtr=startLyoPtr;
  do {
    lyoPtr = lyoPtr->lyoPrec[PUPPY_L];
    if (lyoPtr == gFirstCell ||
	lyoPtr == cell ||
	lyoPtr->curBbox.lowerLeft.y >= top ||
	lyoPtr->curBbox.upperRight.x <= left ||
	lyoPtr->curBbox.upperRight.y <= bottom ) {

	continue;
      } else {			/* **** OVERLAP !!!! */
	totCellOverlap += newOverlap(cell, lyoPtr);
      };
  } while ( lyoPtr != gFirstCell && lyoPtr != NULL &&
    (left-lyoPtr->curBbox.lowerLeft.x) < maxEdgeX &&
      lyoPtr->lyoPrec[PUPPY_L] != gFirstCell );

  /* **** start from PUPPY_L edge to the right **** */
  /* check if startLyoPtr is still a good position to start */
  while ( startLyoPtr->lyoNext[PUPPY_L] != NULL && 
	 left > startLyoPtr->lyoNext[PUPPY_L]->curBbox.lowerLeft.x ) {
	 startLyoPtr = startLyoPtr->lyoNext[PUPPY_L];
       }
  while ( startLyoPtr->lyoPrec[PUPPY_L] != gFirstCell &&
	 left < startLyoPtr->curBbox.lowerLeft.x) {
	 startLyoPtr = startLyoPtr->lyoPrec[PUPPY_L];
       }
  if (startLyoPtr->lyoNext[PUPPY_L] != NULL) {
    lyoPtr = startLyoPtr ;
    do {
      lyoPtr=lyoPtr->lyoNext[PUPPY_L];
      if (lyoPtr == cell || lyoPtr == gFirstCell || lyoPtr == NULL ||
	  lyoPtr->curBbox.lowerLeft.y >= top ||
	  lyoPtr->curBbox.lowerLeft.x >= right ||
	  lyoPtr->curBbox.upperRight.y <= bottom ) {

	  continue;
	} else {		/* **** OVERLAP !!!! */
	  totCellOverlap += newOverlap(cell, lyoPtr);
	};
    } while (lyoPtr != NULL && lyoPtr != gFirstCell &&
      lyoPtr->curBbox.lowerLeft.x < right &&
	lyoPtr->lyoNext[PUPPY_L] != NULL );
  }
  return(totCellOverlap);
}
#endif

/*
 *    curOverlap(cell1 , cell2)
 *   returns the overlapping area for the two cells
 *  >>>> IMPORTANT NOTE  <<<<<
 *  cell1   uses  curBbox
 *  cell2   uses  curBbox
 */

/* TIME CRITICAL ROUTINE !! */
int
curOverlap(cell1,cell2)
    MacroCell *cell1,*cell2;
{
  octBox 		ovBox;
  register octCoord	height, width;

  if (cell1 == cell2)   return (0);

  ovBox.lowerLeft.x  = MAX (cell1->curBbox.lowerLeft.x , cell2->curBbox.lowerLeft.x);
  ovBox.upperRight.x = MIN (cell1->curBbox.upperRight.x , cell2->curBbox.upperRight.x);
  width = ovBox.upperRight.x - ovBox.lowerLeft.x;
  if (width <= 0)  return(0);

  ovBox.lowerLeft.y  = MAX (cell1->curBbox.lowerLeft.y , cell2->curBbox.lowerLeft.y);
  ovBox.upperRight.y = MIN (cell1->curBbox.upperRight.y , cell2->curBbox.upperRight.y); 
  height = ovBox.upperRight.y - ovBox.lowerLeft.y;
  if (height <= 0)  return(0);

  return (width * height);
}

/*
 *    newOverlap(cell1 , cell2)
 *   returns the overlapping area for the two cells
 *	also this routine check if cell1 == cell2, which should be 
 *	guaranteed to be false with one processor, but might be true
 *	with parallel processes
 *  >>>> IMPORTANT NOTE  <<<<<
 *  cell1   uses  newBbox
 *  cell2   uses  curBbox
 */
/* TIME CRITICAL ROUTINE !! */
int
newOverlap(cell1,cell2)
    MacroCell *cell1,*cell2;
{
  octBox		ovBox;
  register octCoord	height, width;

  if (cell1 == cell2) return(0);

  ovBox.lowerLeft.x  = MAX (cell1->newBbox.lowerLeft.x , cell2->curBbox.lowerLeft.x);
  ovBox.upperRight.x = MIN (cell1->newBbox.upperRight.x , cell2->curBbox.upperRight.x);
  width = ovBox.upperRight.x - ovBox.lowerLeft.x;
  if (width <= 0) return(0);

  ovBox.lowerLeft.y  = MAX (cell1->newBbox.lowerLeft.y , cell2->curBbox.lowerLeft.y);
  ovBox.upperRight.y = MIN (cell1->newBbox.upperRight.y , cell2->curBbox.upperRight.y); 
  height = ovBox.upperRight.y - ovBox.lowerLeft.y;
  if (height <= 0)     return(0);

  return (width * height);
}

/*
 *    newNewOverlap(cell1 , cell2)
 *   returns the overlapping area for the two cells
 *	also this routine check if cell1 == cell2, which should be 
 *	guaranteed to be false with one processor, but might be true
 *	with parallel processes
 *  >>>> IMPORTANT NOTE  <<<<<
 *  cell1   uses  newBbox
 *  cell2   uses  newBbox
 */

int
newNewOverlap(cell1,cell2)
    MacroCell *cell1,*cell2;
{
  octBox 		ovBox;
  octCoord 		height, width;

  if (cell1 == cell2) return(0);

  ovBox.lowerLeft.x  = MAX (cell1->newBbox.lowerLeft.x , cell2->newBbox.lowerLeft.x);
  ovBox.upperRight.x = MIN (cell1->newBbox.upperRight.x , cell2->newBbox.upperRight.x);
  width = ovBox.upperRight.x - ovBox.lowerLeft.x;
  if (width <= 0) return (0);

  ovBox.upperRight.y = MIN (cell1->newBbox.upperRight.y , cell2->newBbox.upperRight.y); 
  ovBox.lowerLeft.y  = MAX (cell1->newBbox.lowerLeft.y , cell2->newBbox.lowerLeft.y);
  height = ovBox.upperRight.y - ovBox.lowerLeft.y;
  if ( height < 0 )   return(0);

  return ( width * height);
}




boundaryOverlap(cellP, curOrNewFlag)
    MacroCellPointer cellP;
    int curOrNewFlag;
{
  octBox 		ovBox;
  octCoord 		height, width;
  int			areaOutsideBoundary;
  extern SHARED int	chipBoundaryOverlapFactor;      

  /* First compute the overlap between the cell and the chip boundary
   * If the cell is completely inside the boundary then the overlap area 
   * is equal to the area of the cell. We want to penalize all the area
   * which is NOT inside the boundary, that is the difference between
   * the cellP->area and the overlap area
   */
  if (curOrNewFlag == NEW) {
    ovBox.lowerLeft.x = MAX (cellP->newBbox.lowerLeft.x , chipBbox->lowerLeft.x);
    ovBox.lowerLeft.y =  MAX (cellP->newBbox.lowerLeft.y , chipBbox->lowerLeft.y);
    ovBox.upperRight.x =  MIN (cellP->newBbox.upperRight.x , chipBbox->upperRight.x);
    ovBox.upperRight.y = MIN (cellP->newBbox.upperRight.y , chipBbox->upperRight.y); 
  } else {
    ovBox.lowerLeft.x = MAX (cellP->curBbox.lowerLeft.x , chipBbox->lowerLeft.x);
    ovBox.lowerLeft.y =  MAX (cellP->curBbox.lowerLeft.y , chipBbox->lowerLeft.y);
    ovBox.upperRight.x =  MIN (cellP->curBbox.upperRight.x , chipBbox->upperRight.x);
    ovBox.upperRight.y = MIN (cellP->curBbox.upperRight.y , chipBbox->upperRight.y); 
  }
  width= ovBox.upperRight.x-ovBox.lowerLeft.x;
  height= ovBox.upperRight.y-ovBox.lowerLeft.y;
  if (width < 0 || height < 0) {
    areaOutsideBoundary =cellP->area;
  } else {
    areaOutsideBoundary =cellP->area - width * height;
  }
  return(areaOutsideBoundary * chipBoundaryOverlapFactor);
}

