/* #define PARALLEL_DEBUG  */
#include "defs.h"
/* #include <signal.h> */
#include "global.h"
#include "parall.h"

/*
 *	allocateShMvariables()
 *	Allocates in shared memory the following variables:
 *   SA
 *   process
 *   gFirstCell
 *	All the fields are initialized.
 */
int
initializeSharedVariables()
{
  PFrame.activeChild=0;
  PFrame.childrenInLimbo = 0;
  PFrame.activity= DO_NOTHING;
  PFrame.terminatedProcesses = 0;
  gFirstCell  = SH_MALLOC(MacroCell,1);
  gFirstCell->curBbox.lowerLeft.x = 0;
  gFirstCell->curBbox.lowerLeft.y = 0;
  gFirstCell->curBbox.upperRight.x = 0;
  gFirstCell->curBbox.upperRight.y = 0;

#ifdef PARALLEL
  PFrame.cpusOnLine= cpus_online();
  s_init_lock(&(PFrame.lock));
  s_init_lock(&lyoLock);
  s_init_lock(&waitLock);
  s_init_lock(&subnetMoveLock);
  s_init_lock(&moveStatArrayLock);
#endif
  outerLoopJobQueue = PQinitQueue(0);		/* use default size */
}

allocateShMvariables2()
{
  int 	I;			/* loop counter */

  process = SH_MALLOC (struct ProcessorFrame, P);
  for (I = 0; I < P; I++) {
#ifdef PARALLEL
    s_init_lock( & (process[I].mailbox.lock) );
    s_init_lock( & (process[I].curCluster.lock));
    s_init_lock( & (process[I].newCluster.lock));
#endif
    process[I].pid = 0;
    process[I].mailbox.sender = 0;
    process[I].mailbox.message = NO_MESSAGE;
    process[I].moveAttempted = 0;
    process[I].moveAccepted = 0;
    process[I].firstAnnealVar = NULL;
    process[I].lastAnnealVar = NULL;
    process[I].curNumberOfAnnealVars = 0;
    process[I].curNumberOfCells = 0;
    zeroesCost(&process[I].deltaCost);
    zeroesCost(&process[I].aveMoveCost);
    zeroesCost(&process[I].varMoveCost);
  }
  /* global variables for the parent process */
  mypid = getpid();
  myself = 0;
  myProc = &(process[0]);
  myProc->pid = mypid;
}

	    

int
manageChildren(numberOfChild) 
    int numberOfChild;
{
#ifdef PARALLEL
  
  int I;
  int pid;			/* used by parallel version */

  syncXConnection();		/* do this before forking */


  if (numberOfChild >= PFrame.cpusOnLine ) {
    Warning("Asking for more processes than CPUs");
  }
  if (numberOfChild >= MAX_CPU) {
    (void) sprintf(errmsg,"Maximum number of processes is %d", MAX_CPU);
    Warning(errmsg);
    numberOfChild = MAX_CPU - 1;
  }
  PFrame.activity=DO_NOTHING;
  if (numberOfChild > PFrame.activeChild) {
    Message("Forking processes"); 
    for (I = PFrame.activeChild+1 ; I <= numberOfChild ; I++ ) {
      pid = fork(); showActivity();
      if (pid == -1) {
	Warning("Couldn't fork");
	numberOfChild = I-1;	/* set  */
      } else {
	if (pid == 0) {
	  /* Child code */
	  mypid  = getpid();
	  myself = I;
	  myMoveId = 2 * I + 1;	/* always != 0 */
	  myProc = &(process[myself]);
	  myProc->pid = mypid;
	  childCloseXConnection(); /* To make sure event queue is not srewed up with  */
	  limbo();
	  (void) printf("Child %d returned\n",mypid); (void) fflush(stdout);
	  (void) exit(0);
	} else {
	  /* parent code */
	  /* (void) printf( " %d ",pid);  (void) fflush(stdout); */
	  PFrame.childPid[I] = pid;
	}
      }
    }
    (void) sprintf(errmsg,"Forked %d processes", numberOfChild); 
    Message(errmsg); 
  } else if (numberOfChild < PFrame.activeChild) {
    Message("Killing child processes");
    for (I = numberOfChild +1; I <= PFrame.activeChild; I++) {
      kill (PFrame.childPid[I],SIGKILL);  showActivity();
    }
  } else {
    /* the number of active children is correct */
  }
#endif
  PFrame.activeChild=numberOfChild;
  return(numberOfChild);
}



/*
 *  FUNCTION: limbo()
 *  >>>>  global variables :
 *   mypid		<set>
 *  shared variables:
 *   PFrame.activity  <used>
 */
/*  The whole function can be bypassed on uni-processor */
#ifdef PARALLEL

limbo()
{
  int	rv;			/* return value for PQdoJob */

  /*The process will loop waiting for something to do */
  s_lock(&(PFrame.lock));
  PFrame.childrenInLimbo++;
  s_unlock(&(PFrame.lock));

  while (PFrame.activity != STOP) {
    PQdoJob(outerLoopJobQueue, &rv); /* do one job from the queue */
    switch (PFrame.activity) {
  case DO_NOTHING :
      break;
  case GET_YOUR_ANNEAL_VARS:
      s_lock(&(PFrame.lock));
      PFrame.childrenInLimbo--;
      s_unlock(&(PFrame.lock));

      P_getYourAnnealVars();

      s_lock(&(PFrame.lock));
      PFrame.childrenInLimbo++;
      s_unlock(&(PFrame.lock));
      break;
  case IN_INNER_LOOP:
      s_lock(&(PFrame.lock));
      PFrame.childrenInLimbo--;
      s_unlock(&(PFrame.lock));

      P_saInnerLoop();

      s_lock(&(PFrame.lock));
      PFrame.childrenInLimbo++;
      s_unlock(&(PFrame.lock));
      break;
  case STOP:
      /* just do nothing */
      break;	
  default:
      break;
    }
  }
  /* The child is going to exit */
  s_lock(&(PFrame.lock));
  PFrame.childrenInLimbo--;
  s_unlock(&(PFrame.lock));
}

#endif


#ifdef OLDCODE
/*
 *   distibuteCells()
 *   this routine determines the owner of each cell 
 *   The cells are distributed in turn to the P processors
 * >>> AFFECTED VARIABLES
 * cell->owner
 *	for every cell.
 * SM_minNumberOfCellsPerProc   <SET>
 * SM_maxNumberOfCellsPerProc   <SET>
 *
 */
distributeCells()
{
  MacroCell 	*cell;
  int 		owner;		/* counter */

  if (debug) printf("in distributeCell()\n");
  owner = 0;
  for (cell = gFirstCell->globalNext; cell != NULL ; cell = cell->globalNext) {
    if (cell->mobility == FIXED) {
      cell->owner = P;		/* a non existing processor, but a good color */
    } else {
      cell->owner = owner++ ;
    }
    if (owner >= P) owner=0;
  }
  PFrame.terminatedProcesses = 0;
  PFrame.activity = GET_YOUR_CELLS;
  getYourCells();
  SM_minNumberOfCellsPerProc = MAX (1, (numberOfCells)/P );
  while( SM_minNumberOfCellsPerProc * P >= numberOfCells) {
    SM_minNumberOfCellsPerProc--;
  }
  SM_minNumberOfCellsPerProc = MAX (1, SM_minNumberOfCellsPerProc );
  SM_maxNumberOfCellsPerProc = SM_minNumberOfCellsPerProc;
  while( SM_maxNumberOfCellsPerProc * P <= numberOfCells) {
    SM_maxNumberOfCellsPerProc++;
  }
}
#endif


distributeAnnealingVars()
{
  MacroCell	*cell;
  ListElt	*annealVarLE;
  AnnealVar	*annealVar;
  int		movableCells = 0;
  int 		owner = 0;			/* counter */

  owner = 0;
  for (annealVarLE=ListOfAnnealVars->next;annealVarLE !=NULL;annealVarLE = annealVarLE->next) {
    annealVar = annealVarLE->pointer.annealVar;
    switch (annealVar->type) {
    case AV_CELL:
      cell = annealVar->pointer.cell;
      if (cell->mobility == FIXED) {
	cell->owner = P;	/* a non existing processor, but a good color */
	annealVar->owner = P;
      } else {
	movableCells++;
	cell->owner = owner ;
	annealVar->owner = owner++;
      }
      break;
    case AV_FLOAT_TERM:
    case AV_SUBNET:
    case AV_NET:
      annealVar->owner = owner++;
      break;
    default:
      FatalError("Wrong type in distributeAnnealVar");
    }
    if (owner >= P) owner=0;
  }

  PFrame.terminatedProcesses = 0;
  PFrame.activity = GET_YOUR_ANNEAL_VARS;
  P_getYourAnnealVars();

  SM_minNumberOfCellsPerProc = MAX (1, (movableCells)/P );
  while( SM_minNumberOfCellsPerProc * P >= movableCells) {
    SM_minNumberOfCellsPerProc--;
  }
  SM_minNumberOfCellsPerProc = MAX (1, SM_minNumberOfCellsPerProc );
  SM_maxNumberOfCellsPerProc = SM_minNumberOfCellsPerProc;
  while( SM_maxNumberOfCellsPerProc * P <= movableCells) {
    SM_maxNumberOfCellsPerProc++;
  }

  SM_minNumberOfAnnealVarsPerProc = MAX (1, (numberOfAnnealVars)/P );
  while( SM_minNumberOfAnnealVarsPerProc * P >= numberOfAnnealVars) {
    SM_minNumberOfAnnealVarsPerProc--;
  }
  SM_minNumberOfAnnealVarsPerProc = MAX (1, SM_minNumberOfAnnealVarsPerProc );
  SM_maxNumberOfAnnealVarsPerProc = SM_minNumberOfAnnealVarsPerProc;
  while( SM_maxNumberOfAnnealVarsPerProc * P <= numberOfAnnealVars) {
    SM_maxNumberOfAnnealVarsPerProc++;
  }

}





/*
 *   getYourAnnealVars()
 *   every process executes this routine looking for its own cells
 * >>> GLOBAL VARIABLES
 *   myself		(USED)
 *   myProc		(USED)
 */

P_getYourAnnealVars()
{
  ListElt	*annealVarLE;
  AnnealVar	*annealVar;


  /* This routine can always be re-entered */

  if (debug) printf("%ld: in getYourAnnealVars()\n",mypid); 

  myProc->activity = GET_YOUR_ANNEAL_VARS;
  myProc->curNumberOfCells = 0;
  zeroesCluster(&myProc->curCluster);
  myProc->firstAnnealVar = NULL; /* Just in case this is a reentering */
  myProc->lastAnnealVar = NULL;
  
  for (annealVarLE=ListOfAnnealVars;annealVarLE !=NULL;annealVarLE = annealVarLE->next) {
    annealVar = annealVarLE->pointer.annealVar;
    if (annealVar->owner == myself) {
      annealVarLink(annealVar, myself, FIRST);
      if (annealVar->type == AV_CELL) {
	annealVar->pointer.cell->owner = myself;
	annealVar->pointer.cell->status = OWNED;
	AddCellToCluster( &myProc->curCluster , annealVar->pointer.cell);
	myProc->curNumberOfCells++;
      }
    }
  }
  if (myProc->curNumberOfCells == 0) {
    Warning("I got no cells");
  } else {
    computeBaricenterAndMomentum(&myProc->curCluster);
  }
#ifdef PARALLEL
  s_lock (& (PFrame.lock));
  PFrame.terminatedProcesses ++;
  s_unlock (& (PFrame.lock));
  myProc->activity = DO_NOTHING;
  while (PFrame.terminatedProcesses != P);
#endif
  PFrame.activity = DO_NOTHING;
}

setParallelActivity(activity)
    enum Activity activity;
{
  AllChildrenInLimbo();		/* synchronization step */
  PFrame.terminatedProcesses = 0;
  PFrame.activity = activity;
}

AllChildrenInLimbo()
{
  PFrame.activity = DO_NOTHING;
  while (PFrame.childrenInLimbo < P-1) ;
  return(OK);
}

