/*
 *  FILE: power.c
 *
 *  DESCRIPTION: Interface with OCT data-base.
 *
 *
 *  CREATION DATE: Aug.9.1991          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  
 * 
 *  EXPORTED    FUNCTIONS:
 *  
 *  
 *  IMPORTED    FUNCTIONS:
 *  
 *
 */


#include "defs.h"
#include "global.h"
#include "sa.h"
#include "st.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"
#include "oh.h"
#include "oct.h"
#include "vov.h"

/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */


/* GLOBAL VARIABLES IN THIS MODULE ("power.c") */
double minPeakCurrent = INFINITY;

/* IMPORTED FUNCTIONS */


/* FUNCTION DEFINITION */

#define NUM_QUANT 27

/* FUNCTION DECLARATION */

void handlePowerNet();
void scanNets();
void scanNetsTerms();
void scanTerminals();
double readPeakCurrent();
PowerStatus isNetPowerBus();
Term *getNetPad();
void genPowerNetSet();
Net *fragmentPowerNet();
void normalizePeakcurrent();
octCoord computeTermWidth();

/* FUNCTIONS */


/* check if the net is a power bus and read its pad cell. If it is
   it is chracterized as SUPPLY, otherwise it remains SIGNAL */

void
handlePowerNet(net)
     Net *net;
{     

  if (considerPower) {	
    net->type = isNetPowerBus(net);
    if (net->type == SUPPLY) {
      net->padTerm = getNetPad(net);
      if (net->padTerm == NULL) {
	sprintf(errmsg, "Power bus %s isn't connected to any pads", net->name);
	Warning(errmsg);
	Message("No power optimization performed on this net");
	net->type == SIGNAL;
	return;
      }	  
    }
  }
}


void 
scanNets()
{

  Net   *net;

  net = netSet->globalNext;
  while (net != NULL) {
    sprintf (errmsg,"Net %s, status %d", net->name, net->status);
    Message(errmsg);
    net = net->globalNext;
  } 
  if (considerPower && powerNetSet != NULL) {
    net = powerNetSet;
    while (net != NULL) {
      sprintf (errmsg,"Power net %s, norm. curr. %f.fffff, terms: %s(%s), %s(%s)", 
	       net->name,
	       net->normPeakCurrent,
	       net->term0.next->term->name, 
	       net->term0.next->term->cell->name,
	       net->term0.next->next->term->name, 
	       net->term0.next->next->term->cell->name);
      Message(errmsg);
      net = net->globalNext;
    } 
  }
}


void
scanTerminals(P_cell)
   MacroCell  *P_cell;
{

  struct termList *termList;
  
  termList = P_cell->term0.next;
  while (termList != NULL) {
    sprintf(errmsg, "Terminal %s, position (%d,%d)", termList->term->name,
	    termList->term->curAbsCoordP->x, termList->term->curAbsCoordP->y);
    Message(errmsg);
    termList = termList->next;
  }
}


void
scanNetsTerms()
{ 

  Net   *net;
  struct termList *termList;

  net = netSet->globalNext;
  while (net != NULL) {
    fprintf (stderr,"net %s ", net->name);
    termList = net->term0.next;
    while (termList != NULL) {
      fprintf(stderr, "term %s, peak curr %f ", termList->term->name,
	      termList->term->peakCurrent);
      termList = termList->next;
    }
    fprintf(stderr, "\n",0);
    net = net->globalNext;
  }  
}


double
readPeakCurrent(instTerm, instCell)
     octObject *instTerm, *instCell;
{

  double        peakCurrent;
  octObject     termProp;

  peakCurrent = 0.0;

  if (ohGetByPropName(instTerm, &termProp, "PEAKCURRENT") != OCT_OK) {
    return peakCurrent;
  }
  peakCurrent = termProp.contents.prop.value.real;
  peakCurrent = RECT (peakCurrent);
  if (debug) {
    sprintf(errmsg, "Power term %s in %s: max current absorbed %f A",
	    instTerm->contents.term.name, 
	    instCell->contents.instance.name, 
	    peakCurrent); 
    Message(errmsg);
  }
  return peakCurrent;
}


PowerStatus
isNetPowerBus(net)
   Net  *net;
{

  struct termList *termList;
  
  termList = net->term0.next;
  while (termList != NULL) {
    if (termList->term->peakCurrent > 0.0) {
      return SUPPLY;
    }
    termList = termList->next;
  }
  return SIGNAL;
}


Term
*getNetPad(net)
   Net  *net;
{

  struct termList *termList;
  MacroCell *cell;  

  termList = net->term0.next;
  while (termList != NULL) {
    cell = termList->term->cell;
    if (cell->isPad) {
      return termList->term;
    }
    termList = termList->next;
  }
  return NULL;
}


/* the structure powerNetSet is built */

void
genPowerNetSet()
{

  Net   *net;
  Net   *genNet, *netPtr, *nextNet;
  int   numberOfPowerNets;
  int   numberOfPartitions;
  int   numberOffragments;

  if (!considerPower) {
    powerNetSet = SH_MALLOC(Net, 1);
    return;
  }

  nextNet = NULL;
  net = netSet;
  numberOfPowerNets = 0;
  numberOfPartitions = 0;
 
  while (net != NULL) {
    if (net->type == SUPPLY) {
      netPtr = net;
      numberOfPowerNets++; 
      genNet = fragmentPowerNet(netPtr, &numberOffragments);
      numberOfPartitions += numberOffragments;
      netPtr = genNet;
      while( netPtr->globalNext != NULL) netPtr = netPtr->globalNext;
      netPtr->globalNext = nextNet;
      nextNet = genNet;
    }
    net = net->globalNext;
  }
  powerNetSet = genNet;
  normalizePeakcurrent();
  sprintf(errmsg, "****** Found %d Power nets, partitioned in %d branches", 
	  numberOfPowerNets, numberOfPartitions);
  Message(errmsg);
  /*scanNets();*/
  
}


/* the power net "net" is fragmented in pieces of 2. (containing 
   the terminals which not only are connected but also have a peak 
   current greater than 0. A new structure similar to netSet is 
   built to contain the pieces and is linked to powerNet */

Net
*fragmentPowerNet(net, numberOffragments)
  Net *net;
  int *numberOffragments;
{ 
  Net *newNet, *nextNet;
  struct termList *termList;
  Term *term, *padTerm;
  int i, m = 0;
  int nets = 0;
  char *netName;

  nextNet = NULL;
  padTerm = net->padTerm;
  termList = net->term0.next;
  for (i = 0; i < net->size;i++ ) {  
    term = termList->term;
    if (term->peakCurrent > 0.0) term->type = SUPPLY;
    if (term->type == SUPPLY && term != padTerm) {
      newNet = SH_MALLOC(Net, 1);
      netName = SH_MALLOC (char, strlen(net->name) + 16);
      sprintf(netName, "%s.%d", net->name, m++);
      nets++;
      newNet->name = netName;
      newNet->weight = net->weight;
      newNet->maxLength = 0;       /* no max length */
      newNet->type = SUPPLY;
      newNet->size = 2;
      newNet->term0.next = SH_MALLOC(struct termList, 1);
      newNet->term0.next->term = padTerm;
      newNet->term0.next->next = SH_MALLOC(struct termList, 1);
      newNet->term0.next->next->term = term;
      newNet->peakCurrent = term->peakCurrent;
      minPeakCurrent = MIN(term->peakCurrent, minPeakCurrent);
      newNet->term0.next->next->next = NULL;
      newNet->globalNext = nextNet;
      nextNet = newNet; 
    }
    termList = termList->next;
  }
  net->type = SIGNAL; /* net containing power nets and signals to ground
			 becomes again of type SIGNAL */
  *numberOffragments = nets;
  return newNet;
}


void 
normalizePeakcurrent()

{
  Net *net;
  double peakCurrent;

  if (considerPower && powerNetSet != NULL) {
    net = powerNetSet;
    while (net != NULL) {
      peakCurrent = net->peakCurrent / minPeakCurrent;
      net->normPeakCurrent = peakCurrent;
      net->term0.next->term->normPeakCurrent = peakCurrent;
      net->term0.next->next->term->normPeakCurrent = peakCurrent;
      net = net->globalNext;
    } 
  }
}


octCoord
computeTermWidth(term)
     Term *term;
{

  double peakCurrent;
  int i = 0, l;
  static double quantizedWidth[NUM_QUANT];
  double width;
  double maxWidth = quantizedWidth[NUM_QUANT - 1];

  quantizedWidth[0] = 2;   /* set up this structure before I find a better one */
  quantizedWidth[1] = 3;
  quantizedWidth[2] = 4;
  quantizedWidth[3] = 5;
  quantizedWidth[4] = 6;
  quantizedWidth[5] = 8;
  quantizedWidth[6] = 10;
  quantizedWidth[7] = 12;
  quantizedWidth[8] = 14;
  quantizedWidth[9] = 17;
  for (l = 0; l < 17;l++)  quantizedWidth[10+l] = 20 + l*5;

  peakCurrent = term->peakCurrent;

  if (peakCurrent == 0.0) 
    return minWireWidth;

  width = (peakCurrent / maxCurrDensity);

  if ((width - (int)(width)) > 0) 
    width = (int)(width) + 1.0;               /* excess rounding */

  if (width < minWireWidth / lambda) 
    return (octCoord)(minWireWidth);          /* bias with minWireWidth */
  
  while (width < maxWidth) {
    if (width <= quantizedWidth[i]) {
      return (octCoord)(quantizedWidth[i] * lambda);
    } else {
      i++;
    }
  }
  return (octCoord)(maxWidth * lambda);
  
}


