/*
 *  FILE:
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 */


#ifndef PAR_INCLUDE
#define PAR_INCLUDE

#include "defs.h"
#include "analogStructures.h"
#include "constraintDrivenStructures.h"


/*
 *	NEW STRUCTURES FOR PARALLEL IMPLEMENTATION
 */

/* not all these activities are actually used */
enum Activity {DO_NOTHING, GET_YOUR_CELLS, MOVE_CELLS, READY_TO_EXIT,
		WAITING, STOP, ERROR };

struct PFrame {
    int cpusOnLine;
    int activeChild;
    int childrenInLimbo;
    int childPid[MAX_CPU];
    enum Activity activity;
#ifdef PARALLEL
    slock_t	lock;
#endif
    int terminatedProcesses;
};


/*
 * Structures used by the parallel version
 */
enum MessageType { NO_MESSAGE, STOP_TEMPORARILY, STOPPED };

struct ProcessorFrame {
    long	 	pid;
    enum Activity	activity;
    struct {
#ifdef PARALLEL
	slock_t			lock;
#endif
	int			sender;
	enum MessageType	message;
	int			messageAuxField;
	enum MessageType	reply;
    } mailbox;
    MacroCell   *FirstCell;
    MacroCell   *LastCell;
    Cost 	deltaCost;
    Cost	sumMoveCost;
    Cost	sumMoveCost2;
    Cost	aveMoveCost;
    Cost	varMoveCost;
    Cost	sumCostAccMoves;
    Cost	sumCost2AccMoves;
    int		moveAttempted;
    int		moveAccepted;
    double	newTotalArea, curTotalArea;
    octPoint	newBaricenter, curBaricenter;
    int		newNumberOfCells, curNumberOfCells;
    double	curSumXArea, newSumXArea;
    double	curSumYArea, newSumYArea;
    double	curSumX2Area, newSumX2Area;
    double	curSumY2Area, newSumY2Area;
    double	curMomentum, newMomentum;
};




#ifdef PARALLEL
typedef struct slock_t    *LockPointer;
typedef LockPointer	 *ArrayOfLocks;
#endif


#define MAX_JOBS_IN_QUEUE 20
#define DEFAULT_JOBS_IN_QUEUE 20

typedef struct queue JobQueue;

struct queue {
  int	size;			/* max number of jobs in the queue */
  PFI	*jobsArray;		/* array of jobs, used as a wrap-around ring */
  int	jobs;			/* number of jobs in the queue */
  int	first;
  int	last;
#ifdef PARALLEL
  slock_t lock;			/* lock used for this queue */
#endif
};

extern JobQueue *PQinitQueue();
extern int	PQaddJob();
extern int	PQdoJob();
#endif
