#include "defs.h"
#include "global.h"

static int	trackSpacing   = 12;


expandCell(cell)
    MacroCell *cell;
{
  /* simplified version: just expand by a fixed factor */
  int	expandArray[4];
  int 	i;
  int	edge;
  int	sumTerm;		/* sum of all the terminals of the cell */
  int	averageTerm;		/* average number of terms per edge */

  /* distinguish between box and generic rectilinear shape cell */
  if (cell->edgesInPerim == 4) {
    for (i=0;i<4;i++) {
      edge = (i - cell->lowestLeftCorner + cell->edgesInPerim) % cell->edgesInPerim;
      expandArray[edge] = (int) trackSpacing * cellExpansion * log((double)cell->termsInEdge[i] + 1.0);
    }
  } else {
    Warning("Expansion of generic rectilinear cell is a rough hack");
    /*
     * Sum up all the terminals and assume that they are equally distributed 
     * among the 4 edges of the bbox
     */
    sumTerm = 0;
    for (i=0; i< cell->edgesInPerim ; i++) {
      sumTerm += cell->termsInEdge[i];
    }
    averageTerm = sumTerm / 4;

    /* Do a symmetric expansion in all directions */
    expandArray[PUPPY_L] =
      expandArray[PUPPY_R] =
	expandArray[PUPPY_T] =
	  expandArray[PUPPY_B] = 
	    trackSpacing * cellExpansion * log( (double) averageTerm + 1.0);

    
  }
  cell->rW = expandArray[PUPPY_L] + expandArray[PUPPY_R] + cell->W;
  cell->rH = expandArray[PUPPY_T] + expandArray[PUPPY_B] + cell->H;
  offsetCenter(cell,
	       (octCoord) (expandArray[PUPPY_R]-expandArray[PUPPY_L])>>1,
	       (octCoord) (expandArray[PUPPY_T]-expandArray[PUPPY_B])>>1);
  cell->rHalfW = cell->rW >> 1;
  cell->rHalfH = cell->rH >> 1;
  cell->area =  (double) cell->rW * (double) cell->rH;
}


unexpandCell(cell)
    MacroCell *cell;
{
  cell->rW =  cell->W;
  cell->rH =  cell->H;
    
  offsetCenter(cell, -cell->centerOffset.x , -cell->centerOffset.y);
  cell->rHalfW = cell->rW >> 1;
  cell->rHalfH = cell->rH >> 1;
  cell->area =  (double) cell->rW * (double) cell->rH;
}




iv_changedExpansion()
{
  MacroCell *cell;
  extern void plotCells();

  if (cellExpansion < 0.0 ) cellExpansion = 0.0;
  if (cellExpansion > 20.0 ) cellExpansion = 20.0;

  Message("Change cell expansion (complicated operation)");

  /* This is quite a change for the shared memory representation */
  /* We must stop the children in order to maintain the consistency */
  /* of the data relative to clustering */

  AllChildrenInLimbo();
  
  totalExpandedArea = 0.0;
  for (cell = gFirstCell->globalNext ; cell != NULL ; cell = cell->globalNext) {

    unexpandCell(cell);		/* Go Back to no expansion */
    expandCell(cell);		/* Compute the expanded BBox of the cell */

    /* must now update the bbox and the bins */
    cellPlace(cell, cell->curCenter.x, cell->curCenter.y); 
    ExecuteCellMovement(cell);
    lyoReplace(cell);
    binReplace(cell);

    totalExpandedArea  += cell->area;
  }

  setParallelActivity( GET_YOUR_ANNEAL_VARS );
  P_getYourAnnealVars();
  plotCells();

  return(OK);
}

offsetCenter(cell,tx,ty)
    MacroCell *cell;
    octCoord  tx,ty;
{
  int	I;
  struct termList  *termPtr;

  cell->centerOffset.x += tx;
  cell->centerOffset.y += ty;

  for (termPtr=cell->term0.next; termPtr != NULL; termPtr=termPtr->next) {
    termPtr->term->cellCoord.x -= tx;
    termPtr->term->cellCoord.y -= ty;
  }
  for (I = 0 ; I < cell->edgesInPerim; I++) {
    cell->perimeter[I].x -= tx;
    cell->perimeter[I].y -= ty;
  }

}

double piOver4 = 0.78539816;

/* compute the angles of the subnet for all the terminals */
curNetAngles()
{
  MacroCell 		*cell;
  struct termList 	*termList; /* to run through all the terminals in a cell */
  Term			*term;
  Term			*term2;	/* the other terminal attached to the subnet */
  SubNet  		*subnetP;
  ListElt		*subnetLE;
  int			i;
  int			te;	/* term edge index */
  int			teNext; /* next corner defining the edge */
  int			termIndexInSubnet;
  struct octEdge	cellEdge;
  struct octEdge	subnetVector;
  double		angle();

  for (i = 1 ; i <= numberOfCells ; i++) {
    cell = allCellsArray[i];
    for (termList = cell->term0.next ; termList != NULL ; termList = termList->next) {
      term = termList->term;
      for (subnetLE = term->firstSubnet; subnetLE!=NULL; subnetLE = subnetLE->next) {
	subnetP = subnetLE->pointer.subnet;
	/* find the other terminal attached to the subnet */
	termIndexInSubnet =(term == subnetP->curTerm[0]) ? 0:1;
	term2 =  subnetP->curTerm[1 - termIndexInSubnet];
	
	te = term->edge;
	teNext = (te + 1) % cell->edgesInPerim;
	cellEdge.start.x = cell->curPerimeter[te].x;
	cellEdge.start.y = cell->curPerimeter[te].y;
	cellEdge.end.x = cell->curPerimeter[teNext].x;
	cellEdge.end.y = cell->curPerimeter[teNext].y;
	subnetVector.start.x = term->curAbsCoordP->x;
	subnetVector.start.y = term->curAbsCoordP->y;
	subnetVector.end.x = term2->curAbsCoordP->x;
	subnetVector.end.y = term2->curAbsCoordP->y;
	
	subnetP->curAngle[termIndexInSubnet] = angle(&cellEdge , &subnetVector) + piOver4;
      }
    }
  }
}

double
angle(v1,v2)
    struct octEdge *v1;
    struct octEdge *v2;
{
  static int		ip;		/* inner product */
  static int		ep;		/* external product */
  static double	pm;		/* product of modules of the two vectors */
  int		x1,y1,x2,y2;
  static int		m1;		/* modulus squared of v1 */
  static int		m2;		/* modulus squared of v2 */
  double	a;		/* angle, temporary in [0,pi] */

  x1 = (v1->end.x - v1->start.x);
  y1 = (v1->end.y - v1->start.y);
  x2 = (v2->end.x - v2->start.x);
  y2 = (v2->end.y - v2->start.y);
  ip = x1 * x2 + y1 * y2;
  ep = x1 * y2 - x2 * y1;
  m1 = (x1 * x1 + y1 * y1);
  m2 = (x2 * x2 + y2 * y2);
  pm = sqrt((double) m1 * (double) m2);
  if (pm == 0.0) return(0.0);
  a  = acos((double) ip /(double) pm);
  
  return( (ep >= 0) ? a : -a) ;
}
