/*
 *  FILE: sa.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 *   (void) determineInitialTemperature()
 *   simulatedAnnealing()
 *   P_saInnerLoop()
 *   readMessage()
 *   stopProcess(n)
 *   releaseProcess(n)
 *   updateSAParameters()
 *   printStatistics()
 */
#include "defs.h"
#include "global.h"
#include "sa.h"
#include "analogVariables.h"
#include "constraintDrivenVariables.h"
#include "parall.h"
#include "iv.h"

/* IMPORTED VARIABLES AND STRUCTURES */


/* IMPORTED PROCEDURES */



/* GLOBAL VARIABLES FOR SA */
Cost		curCost; 	/* cost of the current configuration */
Cost		previousCost; 	/* cost of the previous configuration */
Cost		sumDeltaCost;	/* sum of costs of all the moves */
Cost		stepError;	/* error due to parallel moves (or to bugs) */
int		StepCounter = 0;
int		movesInStep = 0; /* moves performed in the Step */
int		acceptedMovesInStep =0; /* moves accepted in the Step */
double 		acceptanceRatio = 0.0; /*  */
double		StepTime = 1.0;	/* in seconds */
int		speed = 0 ;		/* in moves per second */

#define	UNLOCKED 0
#define LOCKED   1

/*
 *	simulatedAnnealing()
 *	Simulated annealing main routine
 *	This routine is executed only by the parent process
 */
int
simulatedAnnealing()
{
    double	startTime;
    int		rv;		/* Returned value from PQdoJob */
    double	lastSaveTime = 0.0;
    int		lastSaveStep = 0;
    extern double puppyTime();	/* Used to mesure the speed of the algorithm */
    extern int  coolingSchedule;
    Cost        changeOfVirtualAxis;

    /* PERFORM SOME INITIALIZATIONS */
    StepCounter = 0;
    totalNumberOfMoves = 0;
    totalNumberOfAcceptedMoves = 0;

    if (XDisplayActive) {
	displayProblemData();
    } else {
	printHeader(stdout);
    }
    /* **** INITIALIZE RANDOM SEQUENCE ****/
    (void) sprintf(errmsg,"Random seed for this run %ld", randVar);
    Message(errmsg);

    CS_globalInit(coolingSchedule);

    configurationCost(&curCost);
    (previousCost = curCost);	/* init previousCost */

    do {
	while (PQdoJob(outerLoopJobQueue, &rv)==OK); /* help clean up the job queue */
	doubleCheck();		/* check consistency of data structure */
	if (useCellBasedMoveStats==FALSE) {

	    computeQualityFactor(&gCellMoveSet);
	    computeQualityFactor(&gNetMoveSet);
	    computeQualityFactor(&gFloatTermMoveSet);

	    reorderMoveSelArray(&gCellMoveSet);
	    reorderMoveSelArray(&gNetMoveSet);
	    reorderMoveSelArray(&gFloatTermMoveSet);

	    resetMoveStatArray(&gCellMoveSet);
	    resetMoveStatArray(&gNetMoveSet);
	    resetMoveStatArray(&gFloatTermMoveSet);

	}
	/*
	 *  before starting moving the cells together with the
	 *  other processors, make sure that all the children
	 *  are in the limbo() routine. This check is necessary
	 *  because some children might still be in P_saInnerLoop()
	 *  since the previous Step
	 */
	PFrame.terminatedProcesses = 0;
	if (PFrame.childrenInLimbo != P-1) {
	    Warning("Critical SYNCHRONIZATION ... ");
	    while (PFrame.childrenInLimbo != P-1); /* indefinite loop */
	    Warning("Exited from critical syncronization");
	}
	/* All the children are in the limbo, waiting to do something */

	/* GO!! for this Step */

	startTime = puppyTime();
	setParallelActivity( IN_INNER_LOOP );
	P_saInnerLoop();

	StepTime = puppyTime() - startTime;

	(previousCost = curCost);	/* save previous cost (struct assignment) */
      
        configurationCost( &curCost );
       
    	CS_updateTemperature( curCost.total );
	outerLoopStats();
     
        /* compute and update virtual axes stored in the global structure
         * gxSymCell (horizontal symmetries) and gySymCell (vertical symmetries) */
        if (considerVirtual)  {
          updateVirtualAxis();
        }
  
	DFwrite();
	updateGlobalRangeLimiter();
	updateSAParameters();
        outputStepDataToDisplay();
   
        /* Do an intermediate saving ? */
	if ( lastSaveStep + saveInterval <= StepCounter &&
	     puppyTime() - lastSaveTime > (10.0 * saveInterval) ) {
	    saveResultsToOct(FALSE); /* intermediate saving */
	    lastSaveTime = puppyTime();
	    lastSaveStep = StepCounter;
	}
    } while ( CS_exitOuterLoop( curCost.total ) == FALSE );
    placementPostProcessing();
    
    return OK;
}


int
outputStepDataToDisplay()
{
    extern void plotCells();
    extern IVWindow* ivWin;

    if (XDisplayActive) {
	plotCells();
	iv_UpdateVar( ivWin, IV_NO_OPT);
	refreshConsoleWindow();
	displayGraphs();
    } else {
	/* output to terminal */
	if (debug) {
	    printf("Step %4d ", StepCounter);
	    printf("previous COST = %.2f ",previousCost.total);
	    printf("new COST = %.2f ", curCost.total);
	    printf("\n");
	    printf("DeltaCost: real %.2f computed %.2f  ERROR %.2f aveErr %.2f\n",
		   previousCost.total - curCost.total,
		   sumDeltaCost.total, stepError.total,
		   (acceptedMovesInStep > 0 ? stepError.total / (double) acceptedMovesInStep: 0.0));
	    printf("TEMP = %.3f ",SM_T);
	    printf("SHKV = (%d,%d) ", SM_shakeX, SM_shakeY);
	    printf("RngL(X,Y) = (%d,%d)\n",gRangeLimiterX, gRangeLimiterY);
	    printf("totMoves %ld totAccMoves %ld\n", totalNumberOfMoves, totalNumberOfAcceptedMoves);
	    printStatistics();
	    (void) fflush(stdout);
	} else {
	    showActivity();
	}
    }
}


int
refreshConsoleWindow()
{
  displaySpeed(speed);
  displayStepNumber(StepCounter);
  displayCost(&curCost);
  displayError(&stepError);
  displayChipBoundaryType();
  displayShakingVector();
  displayMessage(NULL);
  displayWarning(NULL);
  displayProblemData();
}




/*****************/
/*  check error  */
/*****************/
/* Compute the error due to errors (or to bugs) */

outerLoopStats()
{
  int		I;		/* a counter */

  StepCounter++;
  zeroesCost(&sumDeltaCost);
  movesInStep = 0;
  acceptedMovesInStep = 0;
  for (I=0 ; I < P ; I++) {

    accumulateCosts(&sumDeltaCost, &process[I].deltaCost);
    movesInStep += process[I].moveAttempted;
    acceptedMovesInStep += process[I].moveAccepted;
  }
  speed = (StepTime==0.0) ? 0 : movesInStep / StepTime;
  
  computeAccRatio(&gCellMoveSet); 	/* Compute acceptance ratios */
  computeAccRatio(&gNetMoveSet);
  computeAccRatio(&gFloatTermMoveSet);

  acceptanceRatio =  (movesInStep==0) ? 0.0 :(double) acceptedMovesInStep / movesInStep;

  stepError.total      = curCost.total - previousCost.total - sumDeltaCost.total;
  stepError.area       = curCost.area - previousCost.area - sumDeltaCost.area;
  stepError.overlap    = curCost.overlap - previousCost.overlap - sumDeltaCost.overlap;
  stepError.wireLength = curCost.wireLength - previousCost.wireLength - sumDeltaCost.wireLength;
  stepError.cluster    = curCost.cluster - previousCost.cluster - sumDeltaCost.cluster;

  stepError.symmetry   = curCost.symmetry - previousCost.symmetry - sumDeltaCost.symmetry;
  stepError.matching   = curCost.matching - previousCost.matching - sumDeltaCost.matching;

  totalNumberOfMoves += movesInStep;
  totalNumberOfAcceptedMoves += acceptedMovesInStep;

  if (P == 1 && ABS(stepError.total) > 1E-6 && !(analogMode && considerVirtual))  {
    (void) sprintf(errmsg,"Error (%6.2g) detected in step %d", stepError.total, StepCounter);
    Warning(errmsg);
  }
}


/*
 *   P_saInnerLoop()
 *   executed by all the processes in parallel
 */
P_saInnerLoop()
{
  int		exitInnerLoop = FALSE;
  AnnealVar	*annealVar;

  myProc->activity = IN_INNER_LOOP;
  myProc->moveAttempted = 0;
  myProc->moveAccepted  = 0;
  zeroesCost(&myProc->deltaCost);
  zeroesCost(&myProc->sumMoveCost);
  zeroesCost(&myProc->sumMoveCost2);
  zeroesCost(&myProc->sumCostAccMoves);
  zeroesCost(&myProc->sumCost2AccMoves);

#ifdef OLD_CODE_TO_BE_UPDATED
  if (useCellBasedMoveStats) {
    /* For each cell owned by the processor, compute the probability of each type of move */
    for (annealVar = myProc->firstAnnealVar ; annealVar!=NULL; annealVar = annealVar->ownerNext) {
      if (annealVar->type == AV_CELL) {
	cellP = annealVar->pointer.cell;
	computeQualityFactor(cellP->moveStatArray);
	reorderMoveSelArray(cellP->moveStatArray,cellP->moveSelArray);
	resetMoveStatArray(cellP->moveStatArray);
      }
    }
  }
#endif

  CS_initInnerLoop();
  while (PFrame.activity == IN_INNER_LOOP) {

    for (annealVar = myProc->firstAnnealVar;
	 annealVar != NULL AND PFrame.activity == IN_INNER_LOOP;
	 annealVar = annealVar->ownerNext) {

#ifdef PARALLEL
	if (myProc->mailbox.message != NO_MESSAGE) {
	    readMessage();	/* Check the mailbox  */
	}
#endif      

	(void) MoveAnnealVar(annealVar);

	if (myself == 0) innerLoopXserver(); /* the parent process checks the Xeventqueue */
	/*
	 *  Check if the annealVar has changed owner and its pointers
	 *	are no longer valid for me; need to go 
	 *	to some annealVar that is mine for sure
	 */
	if (annealVar->owner != myself) {
	    annealVar = myProc->firstAnnealVar;
	}
    }
    exitInnerLoop = CS_exitInnerLoop();
    if (exitInnerLoop && myProc->activity == IN_INNER_LOOP) {
      myProc->activity = READY_TO_EXIT;
#ifdef PARALLEL
      s_lock (&(PFrame.lock));
#endif
      /* This will cause all the processes to exit even the one that might be waiting */
      if (++PFrame.terminatedProcesses == P) {
	PFrame.activity = DO_NOTHING;
      }
#ifdef PARALLEL
      s_unlock (&(PFrame.lock));
#endif
    }
  }

  /* Exiting: compute average and variance of cost of moves */
  if (myProc->moveAttempted > 1)  {
    myProc->aveMoveCost.total = myProc->sumMoveCost.total/myProc->moveAttempted;
    myProc->varMoveCost.total =
      (myProc->sumMoveCost2.total -
       myProc->moveAttempted *
       myProc->aveMoveCost.total *  myProc->aveMoveCost.total) /
	 (myProc->moveAttempted-1);
  } else {
    myProc->aveMoveCost.total = myProc->sumMoveCost.total;
    myProc->varMoveCost.total = 0;
  }
  myProc->activity = DO_NOTHING;
}




/*
 *   readMessage()
 *   if the mailbox has a message, read it and do
 *   whatever is necessary
 */
readMessage()
{
  switch (myProc->mailbox.message) {
 case NO_MESSAGE:
    return;			/* either false alarm  or job done*/
 case STOP_TEMPORARILY:
    myProc->mailbox.reply = STOPPED;
    /* idle until the processor that requested the stop has done its job */
    while (myProc->mailbox.message == STOP_TEMPORARILY);	/* Very expensive wait !! */
    myProc->mailbox.reply = NO_MESSAGE;				/* clear */
    break;
 default:
    Warning("Unknown message in my mailbox");
    break;
  }
}


/*
 *	stopProcess(n)
 *      Using the mailbox and the lock, requests process n
 *	to stop temporarily
 *	The request will fail if some other process is still
 *	keeping another process stopped, or if some process is ready
 *	to exit.
 *	It returns TRUE if the process has stopped
 *         and     FALSE if it was impossible to stop it
 *	New version, possibly faster. It only uses the 
 *	waitLock
 *
 */
#ifdef PARALLEL



/* If process n successfully stopped, it returns TRUE, otherwise it returns FALSE */

NEWstopProcess(n)
    int    n;	/* process number (not pid !!!) */
{
  if (s_clock(&myProc->mailbox.lock) == L_FAILED) {
    return(FALSE);		/* somebody is trying to communicate with me */
  } 
  /* get privileged access to mailbox of process n */
  if (s_clock(&process[n].mailbox.lock) == L_FAILED) {
    s_unlock(&myProc->mailbox.lock); /* Free my lock */
    return(FALSE);
  }
  /*
   *  I'm the only one allowed to stop this process
   *  and also the only one allowed to write into process[n].mailbox
   */
  process[n].mailbox.message=STOP_TEMPORARILY;
  /* expensive wait !! */
  while (process[n].mailbox.reply != STOPPED AND PFrame.activity == IN_INNER_LOOP) ;
  
  if (process[n].mailbox.reply == STOPPED) {
    s_unlock(&myProc->mailbox.lock); 		/* free my lock */
    return (TRUE);				/* success !! */
  } else {
    /* It's time to stop moving cells. Must quickily exit allow process[n] to continue */
    process[n].mailbox.message = NO_MESSAGE;
    s_unlock(&process[n].mailbox.lock);
    s_unlock(&myProc->mailbox.lock); 		/* free my lock */
    return (FALSE);
  }
}
#endif



/*
 *	releaseProcess(n)
 *	Using the mailbox, let the n-th process to restart
 *      Opposite function of stopProcess()
 */
#ifdef PARALLEL
NEWreleaseProcess(n)
    int 	n;	/* process number (not pid !!!) */
{
  process[n].mailbox.message = NO_MESSAGE; 	/* let the process continue */
  while (process[n].mailbox.reply == STOPPED);  /* make sure that process n has restarted */
  s_unlock(&process[n].mailbox.lock);
}
#endif




#define MAX_SHAKE_STRENGTH 10

/*
 * updateSAParameters()
 * >>>> AFFECTS
 *  SM_minNumberOfCellsPerProc
 *  SM_shakeX
 *  SM_shakeY
 */
int
updateSAParameters()
{
  /* I don't like this any more:
     if (SM_T < 30.0) SM_minNumberOfCellsPerProc = MAX(1,SM_minNumberOfCellsPerProc - 1); */
    
  SM_shakeStrength = MIN( MAX_SHAKE_STRENGTH , SM_shakeStrength);
  SM_shakeStrength = MAX( 0 , SM_shakeStrength); /* cannot be negative */
  /* rotate counterclockwise the vector (SM_shakeX,SM_shakeY) by 90 degrees */
  if (SM_shakeX != 0) {
    SM_shakeY = (SM_shakeX > 0 ? 1 : -1) * SM_shakeStrength;
    SM_shakeX = 0;
  } else {
    SM_shakeX = (SM_shakeY > 0 ? -1 : 1) * SM_shakeStrength;
    SM_shakeY = 0;
  }
}




   
/*
 *   printStatistics()
 *   for each process prints  some numbers
 */
printStatistics()
{
  int 	I;			/* counter */

  printf("--------------------------------------------------------------------------\n");
  printf("proc.  movAtt movAcc  aveCst     varCst  stdDev  cells  area  momentum\n");
  for (I = 0 ; I < P ; I++) {
    printf("%5ld ",process[I].pid);
    printf("%6d ",process[I].moveAttempted);
    printf("%6d ",process[I].moveAccepted);
    printf("%.2e ",process[I].aveMoveCost.total);
    printf("%.2e ",process[I].varMoveCost.total); 
    printf("%.2e ",sqrt(process[I].varMoveCost.total));
    printf("%6d ",process[I].curNumberOfCells);
    printf("%6.1f ",process[I].curCluster.totalArea); 
    printf("%.2e ", process[I].curCluster.momentum);
    printf("\n");
  }
  printf("--------------------------------------------------------------------------\n");
}
	
printHeader(fp)
    FILE *fp;
{
  extern char *progName;
  char *PUPPYversion();
  char *annealVarStatsString();

  if (fp!=NULL) {
    fprintf(fp,"# ***** %s : %s\n", progName, PUPPYversion());
    fprintf(fp,"# ***** PLACEMENT OF: %s:%s\n",
	    Chip.contents.facet.cell, Chip.contents.facet.view);
    fprintf(fp,"# Number of cells              = %4d\n",numberOfCells);
    fprintf(fp,"# Number of floating terminals = %4d\n",numberOfFloatingPins);
    fprintf(fp,"# Number of nets               = %4d\n",numberOfNets);
    fprintf(fp,"# Number of subnets            = %4d\n",numberOfSubnets);
    fprintf(fp,"# Number of anneal variables   = %4d = %s\n",
	    numberOfAnnealVars,
	    annealVarStatsString());
    fprintf(fp,"# Max cell area    = %g\n",maxArea);
    fprintf(fp,"# Tot cell area    = %g\n",totalArea);
  }
}
