/*
 *  FILE: selectMove.c
 *
 *  DESCRIPTION: routines for the selection of the move 
 * 
 *
 *
 *  CREATION DATE: 05/20/87		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS: 
 *  int		selectMove(cellP, moveP)
 */
#include "defs.h"
#include "global.h"
#include "parall.h"

int dbNoMove1 = 0;		/* for debugging purposes only */
int dbNoMove2 = 0;		/* for debugging purposes only */
int dbNoMove3 = 0;		/* for debugging purposes only */
int dbNoMove4 = 0;		/* for debugging purposes only */
int dbNoMove5 = 0;		/* for debugging purposes only */
int dbNoMove6 = 0;		/* for debugging purposes only */

#define MIN_RANGE_LIMITER 		4		/* in lambdas */
#define RANGE_LIMITER_DECREASE_RATE 	0.9
/* 
** Fraction of accepted translation below which the range limiter is decreased
*/
#define RANGE_LIMITER_THRESHOLD		0.25 

/* DECLARATION OF FUNCTIONS */
octTransformType newLegalTransform();


int 
selectMove(annealVar,moveP)
    AnnealVar	*annealVar;
    Move	*moveP;
{
  MacroCell		*cellP;
  struct moveSelect 	*selArray;
  double		r;		/* a random number */
  int			i;		/* a counter */
  int			giveUpFlag;


  /* Necessary initializations on moveP */
  moveP->annealVar = annealVar;
  moveP->cell = NULL;
  moveP->cell2 = NULL;		/* because it is used by RestituteCells */
  moveP->term = NULL;
  moveP->net = NULL;

  /* SHORTCUTS */
  if (annealVar->type == AV_NET && enableSubnets == FALSE) {
    moveP->index = 0;
    moveP->type = NO_MOVE;
    return(OK);
  }

  if ((annealVar->type != AV_FLOAT_TERM && annealVar->type != AV_NET) && floatingTermsOnly==TRUE) {
    moveP->index = 0;
    moveP->type = NO_MOVE;
    return(OK);
  }

  /* Do not move UNUSED floating terminals */
  if (annealVar->type == AV_FLOAT_TERM && annealVar->pointer.term->status == UNUSED) {
    moveP->index = 0;
    moveP->type = NO_MOVE;
    return(OK);
  }

  if (annealVar->owner != myself) FatalError("Should never happen");


  giveUpFlag = FALSE;		/* Becomes true when this anneal var cannot be moved */
  do {
    switch (annealVar->type) {
    case AV_CELL:
      cellP = annealVar->pointer.cell;
      if (cellP->status == BORROWED) {
	dbNoMove1++;
	moveP->index = 0;
	moveP->type = NO_MOVE;	/* somebody has borrowed this cell. Do nothing */
	return(OK);
      } else if (cellP->status == RETURNED) {
	/* The cell has now been returned to me. Get it back */
	AddCellToCluster(&myProc->curCluster, cellP);
	cellP->status = OWNED;
      } 
      moveP->cell = cellP;	/* primary cell */
      moveP->procNumber = myself;
      if (useCellBasedMoveStats) {
	FatalError("Not yet working (temp)");
	selArray = cellP->moveSelArray;
      } else {
	selArray = gCellMoveSet.moveSelArray;
      }
      break;
    case AV_FLOAT_TERM:
      moveP->term = annealVar->pointer.term;
      selArray = gFloatTermMoveSet.moveSelArray;
      break;
    case AV_NET:
      moveP->net = annealVar->pointer.net;
      selArray = gNetMoveSet.moveSelArray;
      break;
    default:
      FatalError("Anneal VAR SCREWED UP in select move");
      break;
    }


    /* Choose the type of move. Repeat until a good move is selected for this cell */
    r = RANDOM;		/* between 0 and 1 */
    i = 0;
    while ( r > selArray[i].thresh ) i++;
    moveP->type =  selArray[i].type;
    moveP->index = selArray[i].index;

    if  (annealVar->type == AV_CELL &&
	 moveP->cell->mobility==FIXED &&
	 moveP->type != CHANGE_OWNERSHIP) {
      moveP->index = 0;
      moveP->type = NO_MOVE;
      return(OK);
    }
    
    /* type chosen, now define other details of the move */

    switch (moveP->type) {
    case ALIGNMENT:
      if (selectedAlignment(moveP) != OK) continue;
      break;
    case TRANSLATION:
    case RL_TRANSLATION:
      if (selectedTranslation(moveP) != OK) continue;
      break;
    case TRANSLATE_AND_TRANSFORM:
      if (selectedTranslation(moveP) != OK) continue;
      break;
    case TRANSFORMATION: 
      if (selectedTransformation(moveP) != OK) continue;
      break;
    case CC_INTERCHANGE:
    case BB_INTERCHANGE:
      if (selectedInterchange(moveP) != OK) continue;
      break;
    case SUBNET_MOVE:
      if (selectedSubnetMove(moveP) != OK) giveUpFlag = TRUE;
      break;
    case TERMINAL_MOVE:
      if (selectedTerminalMove(moveP) != OK)  giveUpFlag = TRUE;
      break;
#ifdef PARALLEL
    case CHANGE_OWNERSHIP:
      if (selectedOwnershipChange(moveP) != OK) giveUpFlag = TRUE;
      break;
#endif
    default:
      FatalError("no move made");
      break;
    }
  } while (moveP->type == NO_MOVE && !giveUpFlag);
  return(OK);
}



selectedTranslation(moveP)
    Move *moveP;
{
    int	 transformFlag;
    MacroCell 	*cellP;
    octBox	legalBox;	/* Area in which the center is allowed to move */
    octCoord	legalBoxWidth;
    octCoord	legalBoxHeight;
    octCoord 	ncx,ncy;	/* New coordinate of center */
    double 	r;		/* A random number */
    octTransformType newTransform;
    octTransformType newLegalTransform();

    transformFlag = (moveP->type == TRANSLATE_AND_TRANSFORM);
    cellP = moveP->cell;
    cellP->moveNext = NULL;
    moveP->complexity = 1;
    /*
     *  The translation must keep the center of the cell within the chip boundary
     */
    if (moveP->type != RL_TRANSLATION) {
	legalBox = *chipBbox;
    } else {
	legalBox.lowerLeft.x = cellP->curCenter.x - gRangeLimiterX + SM_shakeX;
	legalBox.lowerLeft.y = cellP->curCenter.y - gRangeLimiterY + SM_shakeY;
	legalBox.upperRight.x = cellP->curCenter.x + gRangeLimiterX+ SM_shakeX;
	legalBox.upperRight.y = cellP->curCenter.y + gRangeLimiterY+ SM_shakeY;
	/* Chop the legal box to take into account the boundary of the chip */
	legalBox.lowerLeft.x = MAX(legalBox.lowerLeft.x, chipBbox->lowerLeft.x);
	legalBox.lowerLeft.y = MAX(legalBox.lowerLeft.y, chipBbox->lowerLeft.y);
	legalBox.upperRight.x = MIN(legalBox.upperRight.x, chipBbox->upperRight.x);
	legalBox.upperRight.y = MIN(legalBox.upperRight.y, chipBbox->upperRight.y);
    }

    legalBoxWidth = legalBox.upperRight.x - legalBox.lowerLeft.x;
    legalBoxHeight = legalBox.upperRight.y - legalBox.lowerLeft.y;
  
    if (legalBoxWidth <= 0 || legalBoxHeight <= 0) {
	/* 
	 * No legal translation is possible for cell: move the cell anywhere
	 * inside the chipbounding box
	 */
	(legalBox = *chipBbox) ;
	legalBoxWidth = legalBox.upperRight.x - legalBox.lowerLeft.x;
	legalBoxHeight = legalBox.upperRight.y - legalBox.lowerLeft.y;
    }
    /* pick a location inside the legal box. Iterate until a non-zero translation is selected */
    do {
	/* The translation can be general or just in the X or Y direction */
	r = RANDOM;
	if (r > 0.5) {
	    /* General translation */
	    ncx=  (int) (RANDOM * legalBoxWidth) + legalBox.lowerLeft.x + SM_shakeX ;
	    ncy=  (int) (RANDOM * legalBoxHeight) + legalBox.lowerLeft.y + SM_shakeY;
	} else if (r < 0.25) {
	    /* X translation */
	    ncx=  (int) (RANDOM * legalBoxWidth) + legalBox.lowerLeft.x + SM_shakeX ;
	    ncy=  cellP->curCenter.y;
	} else {
	    /* Y translation */
	    ncx=  cellP->curCenter.x;
	    ncy=  (int) (RANDOM * legalBoxHeight) + legalBox.lowerLeft.y + SM_shakeY;
	}
    } while (ncx == cellP->curCenter.x && ncy == cellP->curCenter.y);
  
    if (transformFlag) {
	newTransform= newLegalTransform(cellP);
    } else {
	newTransform = cellP->curTransform;
    }

    cellPlaceAndTransform(cellP, ncx, ncy, newTransform);
    return(OK);
}

selectedAlignment(moveP)
    Move *moveP;
{
  MacroCell 	*cellP;
  MacroCell	*neighborCellP;
  MacroCell	*lyoAdjacentCell();
  octCoord 	ncx,ncy;	/* New coordinate of center */
  octCoord	tx,ty;
  int		side;

  cellP = moveP->cell;
  cellP->moveNext = NULL;
  moveP->complexity = 1;

  /* Choose one of PUPPY_L PUPPY_R PUPPY_T PUPPY_B */
  side = (int) (floor(RANDOM * 4.0));
  
  /* Get nexty cell to the 'side' */
  neighborCellP = lyoAdjacentCell(cellP,side);
  if (neighborCellP == NULL) {
    moveP->type = NO_MOVE;
    return(NOT_OK);
  }

  switch (side) {
  case PUPPY_L:
    ty = 0;
    tx = neighborCellP->curBbox.upperRight.x - cellP->curBbox.lowerLeft.x;
    break;
  case PUPPY_R:
    ty = 0;
    tx = neighborCellP->curBbox.lowerLeft.x - cellP->curBbox.upperRight.x;
    break;
  case PUPPY_T:
    tx = 0;
    ty = neighborCellP->curBbox.lowerLeft.y - cellP->curBbox.upperRight.y;
    break;
  case PUPPY_B:
    tx = 0;
    ty = neighborCellP->curBbox.upperRight.y - cellP->curBbox.lowerLeft.y;
    break;
  default:
    FatalError("Inconsistent enumeration of edges");
    break;
  }

  /*   { char *edgeName(); printf("%8s - (%3d,%3d)\n", edgeName(side) , tx ,ty); } */
  ncx=  cellP->curCenter.x + tx;
  ncy=  cellP->curCenter.y + ty;
  cellPlaceAndTransform(cellP, ncx, ncy, cellP->curTransform);
  return(OK);
}


selectedTransformation(moveP)
    Move *moveP;
{
  MacroCell 		*cellP;
  octTransformType 	newTransform;
  octTransformType newLegalTransform();

  cellP = moveP->cell;
  cellP->moveNext = NULL;
  moveP->cell2 = NULL;
  moveP->complexity = 1;
  newTransform = newLegalTransform(cellP);
  cellPlaceAndTransform(cellP, cellP->curCenter.x, cellP->curCenter.y, newTransform);
  return(OK);
}


octTransformType newLegalTransform(cell)
    MacroCell *cell;
{
  int	newTransform;
  int	lt;			/* number of legal transforms */
  int	index;			/* of the new transformation selected */

  if (cell->legalTransformsArray == NULL) {
    /* All transforms are legal */
    newTransform= ( (int) ((double) cell->curTransform + ceil(RANDOM *7)) % 8);
  } else {
    lt = cell->numberOfLegalTransforms;
    index = (int) ( (double) cell->curTransform + ceil(RANDOM * (lt - 1 ) )) % lt;
    newTransform =  cell->legalTransformsArray[ index ];
  }
  return((octTransformType) newTransform);
}


selectedInterchange(moveP)
    Move *moveP;
{
  MacroCell	*cellP;
  MacroCell	*cell1, *cell2;	/* used for pairwise interchanges */
  int		cell2Number;	/*  */
  octPoint	nc1,nc2;	/* new centers in interchange */
  octCoord	X1,X2,Y1,Y2;	/* extension in the XY dims of the two cells to interchange */
  /***** VARIABLES for TRANSFORMATIONS ****/
  octTransformType nt1,nt2;
  octTransformType orthogonalTransform();

  if (numberOfMovingCells <= 1) {
    moveP->index = 0;
    moveP->type = NO_MOVE;	/* cannot stop the other process */
    return(NOT_OK);
  }
  cellP = moveP->cell;
  do {				/* pick up a random cell different from 'cell' */
    cell2Number = (int) floor(RANDOM * numberOfMovingCells);
  } while (varCellsArray[cell2Number] == cellP);
  if ((varCellsArray[cell2Number])->mobility == FIXED ) {
    FatalError("fixed cell in varCellsArray: should never happen");
  }
  cell1 = moveP->cell;
  cell2 = varCellsArray[cell2Number];
#ifdef PARALLEL
  if (BorrowCell(cell2) == FALSE) {
    dbNoMove3++;
    moveP->index = 0;
    moveP->type = NO_MOVE;	/* cannot stop the other process */
    return(NOT_OK);
  }
#endif
  moveP->complexity = 2;
  moveP->cell2 = cell2;
  cellP->moveNext = cell2;
  cell2->moveNext = NULL;
  moveP->subnet = NULL;
  moveP->net = NULL;
  moveP->term = NULL;
  moveP->procNumber = cell2->owner;

  if (moveP->type == CC_INTERCHANGE) {
    nc2.x = cell1->curCenter.x;
    nc2.y = cell1->curCenter.y;
    nc1.x = cell2->curCenter.x;
    nc1.y = cell2->curCenter.y;
    if (!enableTransformations || RANDOM < 0.5) { /* keep this orientation or not ? */
      /* OK, go ahead with the exchange */
      cellPlaceAndTransform(cell1, nc1.x, nc1.y, cell1->curTransform);
      cellPlaceAndTransform(cell2, nc2.x, nc2.y, cell2->curTransform);
    } else {
      /* No, exchange orientation  also */
      nt1 =  orthogonalTransform(cell1->curTransform);
      nt2 =  orthogonalTransform(cell2->curTransform);
      /* Make sure the new orientations are legal */
      if (((1 << (int) nt1) & cell1->legalTransforms)==FALSE) nt1 = cell1->curTransform;
      if (((1 << (int) nt2) & cell2->legalTransforms)==FALSE) nt2 = cell2->curTransform;
      cellPlaceAndTransform(cell1, nc1.x, nc1.y, nt1);
      cellPlaceAndTransform(cell2, nc2.x, nc2.y, nt2);
    }
  } else if (moveP->type == BB_INTERCHANGE) {
    /* try to preserve the current bbox of the two cells */
    X1 = cell1->curBbox.upperRight.x - cell1->curCenter.x;
    Y1 = cell1->curBbox.upperRight.y - cell1->curCenter.y;
    X2 = cell2->curBbox.upperRight.x - cell2->curCenter.x;
    Y2 = cell2->curBbox.upperRight.y - cell2->curCenter.y;
    if (!enableTransformations || ( X1 > Y1 && X2 > Y2)) {	/* keep the orientations */
      nt1 = cell1->curTransform;
      nt2 = cell2->curTransform;
      /* align the external edges of the two cells */
      if (cell1->curCenter.x >= cell2->curCenter.x) {
	nc1.x = cell2->curCenter.x + X2 - X1;
	nc2.x = cell1->curCenter.x + X2 - X1;
      } else {
	nc1.x = cell2->curCenter.x - X2 + X1;
	nc2.x = cell1->curCenter.x - X2 + X1;
      }
      if (cell1->curCenter.y >= cell2->curCenter.y) {
	nc1.y = cell2->curCenter.y + Y2 - Y1;
	nc2.y = cell1->curCenter.y + Y2 - Y1;
      } else {
	nc1.y = cell2->curCenter.y - Y2 + Y1;
	nc2.y = cell1->curCenter.y - Y2 + Y1;
      }
    } else {			/* change the orientations */
      nt1 =  orthogonalTransform(cell1->curTransform);
      nt2 =  orthogonalTransform(cell2->curTransform);
      /* Make sure the new transformations are legal */
      if (((1 << (int) nt1) & cell1->legalTransforms)==FALSE) nt1 = cell1->curTransform;
      if (((1 << (int) nt2) & cell2->legalTransforms)==FALSE) nt2 = cell2->curTransform;

      if (cell1->curCenter.x >= cell2->curCenter.x) {
	nc1.x = cell2->curCenter.x + X2 - Y1;
	nc2.x = cell1->curCenter.x - X1 + Y2;
      } else {
	nc1.x = cell2->curCenter.x - X2 + Y1;
	nc2.x = cell1->curCenter.x - Y2 + X1;
      }
      if (cell1->curCenter.y >= cell2->curCenter.y) {
	nc1.y = cell2->curCenter.y + Y2 - X1;
	nc2.y = cell1->curCenter.y + X2 - Y1;
      } else {
	nc1.y = cell2->curCenter.y - Y2 + X1;
	nc2.y = cell1->curCenter.y - X2 + Y1;
      }
    }
    cellPlaceAndTransform(cell1, nc1.x, nc1.y, nt1 );
    cellPlaceAndTransform(cell2, nc2.x, nc2.y, nt2 );
  } else {
    FatalError("Illegal type of interchange");
  }
  return(OK);
}


/* Select a subnet move given a net */
int
selectedSubnetMove( moveP )
    Move *moveP;
{
    /***** VARIABLES for SUBNET MOVE ****/
    Net	*net;			/* The multi-term net whose spanning tree we are trying to change. */
    SubNet *movingSubnet;	/* Pointer to the moving subnet. */
    SubNet *fixedSubnet;	/* Pointer to the fixed subnet. */
    int	movingSubnetIndex;	/*  */
    int	fixedSubnetIndex;	/*  */
    ListElt *subnetLE;		/*  */
    int	t1,t2;			/*  */
    struct termList *termPtr;	/* used to select the moving terminal */
    struct term	*termP;		/* moving terminal */
    int	found = FALSE;		/* A flag */
    int	i;			/* A counter */
    int	randomTermInNet;	/*  */
    int	avoidInfiniteLoopFlag;	/*  */

    net = moveP->net;
    /*
     * Find the next terminal in net that has more than one subnet.
     * We need to find a terminal with at least 2 subnets.
     */  

    randomTermInNet = (int)floor(RANDOM * net->size) + 1;
    termPtr = net->term0.next;	/* first terminal */
    i = 1;
    while (i++ < randomTermInNet) {
	termPtr = termPtr->next;
    }
    /* start search form this terminal */
    avoidInfiniteLoopFlag = 0;
    while ( 1 ) {
	termP = termPtr->term;
	if (termP->subnets > 1) {
	    break;		/* Found the good terminal with 2 or more subnets. */
	}
	if (termPtr->next == NULL) {
	    if (avoidInfiniteLoopFlag == 1) {
		printSubnets(net);
		FatalError("Could not find a term with two subnets");
	    } 
	    termPtr = net->term0.next; /* restart search form beginning */
	    avoidInfiniteLoopFlag = 1;
	} else {
	    termPtr = termPtr->next;
	}
    }

    /* 
     * We have the terminal to move. Somebody might try to add other subnets
     * to this terminal, while we run through the list of all the subnets
     * this might cause problems. We avoid them with a lock
     */

    /* t1 and t2 are the indexes to the two subnets used to do the move */
    t1 = (int)floor( RANDOM * termP->subnets );
    t2 = (int)floor( RANDOM * termP->subnets );
    if ( t1 == t2 ) {
	t2 = ( t2 + 1 ) % termP->subnets;
    }
    subnetLE = termP->firstSubnet;
    for ( i = 0; i <= MAX( t1, t2 ) ; i++ ) {
	if (i == t1) {
	    movingSubnet = subnetLE->pointer.subnet;
	    moveP->subnetLE = subnetLE;	/* Done for efficiency */
	}
	if (i == t2)  {
	    fixedSubnet  = subnetLE->pointer.subnet;
	}
	subnetLE = subnetLE->next;
    }
    /* 
     * Now we have the subnets. 
     */
    movingSubnetIndex = (movingSubnet->curTerm[0] == termP) ? 0 : 1;
    fixedSubnetIndex  = (fixedSubnet->curTerm[0] == termP) ? 0 : 1;
    movingSubnet->newTerm[movingSubnetIndex] = fixedSubnet->curTerm[1-fixedSubnetIndex];
    /* The other terminal does not change */
    movingSubnet->newTerm[1-movingSubnetIndex] = movingSubnet->curTerm[1-movingSubnetIndex];
    moveP->subnet = movingSubnet;
    moveP->fixedSubnet = fixedSubnet;
    moveP->subnetIndex = movingSubnetIndex;
    moveP->term = termP;
    moveP->cell = NULL;
    moveP->cell2 = NULL;
    moveP->complexity = 1;

    return OK;
}

selectedTerminalMove(moveP)
    Move *moveP;
{
  MacroCell	*cellP;		/* cell containing the floating terminal */
  Term		*termP;		/* moving terminal */
  int		newSite;

  termP = moveP->term;
  cellP = termP->cell;

  /* 
   * Select a new site at random. If that is not OK, move by 1 in the
   * direction of the selected site. Check again if that is a good site 
   */
  newSite = termP->curSite + (int) ceil((cellP->numberOfSites-1)*RANDOM);
  if (newSite >= cellP->numberOfSites) newSite -= cellP->numberOfSites;
  
  if (goodSite(cellP,termP,newSite) != OK) {
    /* Try moving by 1 in direction of newSite */
    newSite += (newSite > termP->curSite) ?  1 : -1 ;
    /* bring newSite within range */
    if (newSite < 0) newSite = cellP->numberOfSites - 1;
    if (newSite == cellP->numberOfSites) newSite = 0;
    if (goodSite(cellP, termP, newSite) != OK) {
      moveP->index = 0;
      moveP->type = NO_MOVE;
      return(NOT_OK);
    }
  }

  termP->newSite = newSite;

/*   termP->newAbsCoordP = &cellP->floatingSites[termP->newSite].curAbsCoord; */
  moveP->net = termP->net;
  moveP->complexity = 1;
  moveP->cell = termP->cell;
  moveP->cell2 = NULL;
  return(OK);
}

int goodSite(cellP,termP,site)
    MacroCell *cellP;
    Term *termP;
    int	site;
{
  if (cellP->floatingSites[site].terms >= cellP->floatingSites[site].capacity){
    return(NOT_OK);		/* the new site is already full to capacity */
  }
  if ((1 << cellP->floatingSites[site].edge) & termP->illegalEdges) {
    return(NOT_OK);		/* The selected edge was illegal */
  }
  return(OK);
}

#ifdef PARALLEL
selectedOwnershipChange(moveP)
    Move *moveP;
{
  /* can I really give up a cell ? */
  if (myProc->curNumberOfCells <= SM_minNumberOfCellsPerProc) {
    dbNoMove4++;
    moveP->index = 0;
    moveP->type = NO_MOVE;	/* give up */
    return(NOT_OK);
  } else {
    moveP->procNumber=((int) ceil(RANDOM*(P-1))+myself)%P; /* pick another process */
    if (process[moveP->procNumber].curNumberOfCells+1 > SM_maxNumberOfCellsPerProc OR
	NEWstopProcess(moveP->procNumber) == FALSE) {
      dbNoMove5++;
      moveP->index = 0;
      moveP->type = NO_MOVE;
      return(NOT_OK);
    }
  }
  moveP->cell->moveNext = NULL;
  moveP->complexity = 1;
  return(OK);
}
#endif


/*****/
updateGlobalRangeLimiter()
{
  static int translationIndex = 0; /* Init to  */
  int i = 0;			/*  */

  /* this search should be done only once */
  while (translationIndex == 0) {
    if (gCellMoveSet.moveStatArray[i].type == RL_TRANSLATION) {
      translationIndex = i;
      break;			/* to avoid */
    } else {
      if (i++ > 100) FatalError("gCellMoveSet has no RL_TRANSLATION move");
    }
  }
  if (gCellMoveSet.moveStatArray[translationIndex].attempted > 0) {
    if (gCellMoveSet.moveStatArray[translationIndex].acceptanceRatio < RANGE_LIMITER_THRESHOLD) {
      /* reduce the range limiter */
      gRangeLimiterX = MAX( MIN_RANGE_LIMITER, gRangeLimiterX * RANGE_LIMITER_DECREASE_RATE);
      gRangeLimiterY = MAX( MIN_RANGE_LIMITER, gRangeLimiterY * RANGE_LIMITER_DECREASE_RATE);
    }
  }
}


#ifdef OBSOLETE_CODE
getTermForSubnetMove(cellP,termP)      
    MacroCell 	*cellP;
    Term	**termP;
{
  struct termList *termPtr;
  int found = FALSE ;
  int searches = 0;

  if (cellP->lastMovedTermLE == NULL) {
    if (cellP->term0.next != NULL) {
      cellP->lastMovedTermLE = cellP->term0.next;
    } else if (cellP->floatingTerms > 0 ) {
      cellP->lastMovedTermLE = cellP->floatTerm0.next;
    } else {
      return(FALSE);
    }
  }
  while (found==FALSE && searches < 10) {
    termPtr = cellP->lastMovedTermLE->next;
    if (termPtr == NULL) {	/* restart the search from the beginning */
      if (cellP->floatingTerms > 0 AND RANDOM > 0.5) {
	termPtr = cellP->floatTerm0.next;
      } else {
	termPtr = cellP->term0.next;
      }
      /* If there are no terminals then we must check that .. */
      if (termPtr == NULL) {
	return(FALSE);
      }
    }
    *termP = termPtr->term;
    if ((*termP)->subnets > 1) {
      found = TRUE;
    } else {
      searches ++;
    }
    cellP->lastMovedTermLE = termPtr; /* proceed along the list */
  }
  return(found);
}

NEWgetTermForSubnetMove(net,termP)      
    Net		*net;
    Term	**termP;
{
  struct termList *termPtr;
  int found = FALSE ;

  /* We need to find a terminal with at least 2 subnets */     /* get the first */
  termPtr = net->term0.next;
  while (found == FALSE) {
    if (termPtr == NULL) FatalError("Could not find a terminal with more than two subnets");
    *termP = termPtr->term;
    if ((*termP)->subnets > 1) {
      found = TRUE;
    } else {
      termPtr = termPtr->next;
    }
  }
  return(found);
}
#endif
