/*
 *  FILE: subnets.c
 *
 *  DESCRIPTION: Routines related to creation and handling of subnets
 *
 *
 *  CREATION DATE: 08/17/87		AUTHOR: Andrea Casotto
 *
 *  FUNCTIONS: 
 *	       
 */

#include "defs.h"
#include "sa.h"
#include "global.h"

initSubnets(net)
    Net	*net;
{
  int 		i;
  struct termList *t1;
  struct termList *t2;
  SubNet	*subnetP;
  ListElt	*subnetLEcouple;

  if (net->size <= 1) {
    (void) sprintf(errmsg,"Net %s has %d terms: subnets not initialized", net->name, net->size);
    if (debug) Warning(errmsg);
    net->subnets = 0;
    net->subnetArray = NULL;
    net->subnetLEArray = NULL;
    return(-1);
  }
  net->subnets = net->size - 1;
  net->subnetArray = SH_MALLOC( SubNet , net->subnets);
  
  net->subnetLEArray = SH_MALLOC( ListElt , net->subnets * 2);

  /* Initialize all the subnets and  attach them  to the terminals of the net */
  t1 = net->term0.next;
  for (i= 0; i < net->subnets ; i++) {
    if (t1 == NULL) FatalError("Wrong terminal count for this net 1");
    t2 = t1->next;
    if (t2 == NULL) FatalError("Wrong terminal count for this net 2");
    subnetP = &net->subnetArray[i];
    subnetP->curTerm[0] = t1->term;
    subnetP->curTerm[1] = t2->term;
    subnetP->newTerm[0] = NULL;
    subnetP->newTerm[1] = NULL;
    subnetP->newAngle[0] = 0.0;
    subnetP->newAngle[1] = 0.0;
    subnetP->newAngle[0] = 0.0;
    subnetP->newAngle[1] = 0.0;
#ifdef PARALLEL
    subnetP->softLock = 0;
#endif
    subnetLEcouple = net->subnetLEArray + 2 * i;
    subnetLEcouple[0].pointer.subnet = subnetP;
    subnetLEcouple[1].pointer.subnet = subnetP;
    
    linkEltToListHandle(&t1->term->firstSubnet,&subnetLEcouple[0]);
    linkEltToListHandle(&t2->term->firstSubnet,&subnetLEcouple[1]);
    t1->term->subnets ++;
    t2->term->subnets ++;
    t1 = t1->next;
  }
  return(0);
}


printSubnets(net)
    Net *net;
{
  SubNet  	*subnet;
  int		i;
  
  
  printf("Net %s\n" , net->name);
  for (i = 0; i < net->subnets ; i++)  {
    subnet = &(net->subnetArray[i]);
    printf("%3d : (%s %s){%d} (%s %s){%d} \n", i,
	   subnet->curTerm[0]->name , subnet->curTerm[0]->cell->name , subnet->curTerm[0]->subnets,
	   subnet->curTerm[1]->name , subnet->curTerm[1]->cell->name ,subnet->curTerm[0]->subnets);
  }
  
}
