/*
 *  FILE: tap.c
 *
 *  DESCRIPTION: interface to technology
 *
 *
 *  CREATION DATE: Aug.18.1990          AUTHOR: Edoardo Charbon
 *
 *  FUNCTIONS:
 *  
 *  
 *  
 *  IMPORTED    FUNCTIONS:
 *  
 *    
 */


/* IMPORTED GLOBAL VARIABLES AND STRUCTURES */

#include "defs.h"
#include "sa.h"
#include "global.h"

#include "oct.h"
#include "tap.h"
#include "oh.h"

/* #include "strings.h" */

/* IMPORTED VARIABLES */


/* GLOBAL VARIABLES OF THIS MODULE ("analogCost.c") */


/* DECLARED ROUTINES USED IN THIS FILE THAT DO NOT RETURN INTEGERS */

void readMaxPeakCurrent();

/* IMPORTED FUNCTIONS */


/* DEFINITIONS */
#define DEFAULT_CURR_DENSITY 0.0015 /* amps / lambda */
#define DEFAULT_SHEET_RES 0.2       /* ohms / square */
#define NUM_LAYERS 2
static char* layerNames[] = {"MET1","MET2"};


/* FUNCTIONS */


void 
readTechnologyDataFromOct()
{
  readMaxPeakCurrent(&Chip);
}


/* compute max current density and min wiring width */

void
readMaxPeakCurrent(facet)
     octObject *facet;
{

  static double def = 0.1e-4;	/* amperes per lambda */
  int i;
  double currDensTable[NUM_LAYERS];
  double minwidth[NUM_LAYERS];
  double sheetResTable[NUM_LAYERS];
  double coordSize = 1e-6;

  lambda = OCT_LAMBDA;	   /* Default = 20 */
  minWireWidth = lambda;   /* 1 lambda */
  maxCurrDensity = INFINITY;  
 
  if (facet == (octObject *) 0) return;
  if (!considerPower) return;

  Message("Reading technology information");

  if ( tapGetProp( facet, TAP_LAMBDA ) ) {
    lambda = TAP_LAMBDA->value.integer;
  }
  if ( tapGetProp( facet, TAP_COORD_SIZE ) ) {
    coordSize = TAP_COORD_SIZE->value.real;
  }
  
  /* get min width and current density for each layer */
  
  for (i = 0; i < NUM_LAYERS ; i++) {
    tapLayerListElement layerList[1];
    
    layerList[0].layer.type = OCT_LAYER;
    layerList[0].layer.contents.layer.name = layerNames[i];
    
    OH_ASSERT( octGetOrCreate( facet, &layerList[0].layer) );
    minwidth[i] = tapGetMinWidth(&layerList[0].layer) / lambda ;

    if ( tapGetLyrProp( 1, layerList, TAP_SHEET_RESISTANCE ) ) {
      sheetResTable[i] = TAP_SHEET_RESISTANCE->value.real;
    } else {
      sheetResTable[i] = DEFAULT_SHEET_RES;
    }
    
    if ( tapGetLyrProp( 1, layerList, TAP_MAX_CURRENT_DENSITY ) ) {
      currDensTable[i] = TAP_MAX_CURRENT_DENSITY->value.real * coordSize * lambda;
    } else {
      currDensTable[i] = DEFAULT_CURR_DENSITY; /* Reasonable default. */
    }
    
    sprintf(errmsg,"%s curr density %lf (A/lambda), min width %d, sheet res %f (Ohm/sq)",
	    layerNames[i], currDensTable[i], minwidth[i], sheetResTable[i]);
    Message(errmsg);
    
    if (debug) {
      sprintf (errmsg, "tech directory for %s is %s", 
	       facet->contents.facet.cell, tapGetDirectory(facet));
      Message(errmsg);
    }
  }
  for (i = 0; i < NUM_LAYERS ; i++) {
    maxCurrDensity = MIN (maxCurrDensity, currDensTable[i]);
    minWireWidth = MAX (minWireWidth, minwidth[i] * lambda);
  } 
  return;

}
  
