#include "defs.h"
#include "global.h"
#include "octPlacement.h"


struct transformDecodeTable {
  char 	*name;
  int	mask;
} tdTable[] = {
  {"ALL" 		, 0xffff } , /* All transformations legal */
  {"THIS_ASPECT_RATIO"  , 1 << (int) OCT_NO_TRANSFORM | 1 << (int) OCT_MIRROR_X |
         		  1 << (int) OCT_MIRROR_Y | 1 << (int) OCT_ROT180           } ,
  {"NO_TRANSFORM"	, 1 << (int) OCT_NO_TRANSFORM },
  {"MIRROR_X"		, 1 << (int) OCT_MIRROR_X },
  {"MIRROR_Y"		, 1 << (int) OCT_MIRROR_Y },
  {"ROT90"		, 1 << (int) OCT_ROT90 },
  {"ROT180"		, 1 << (int) OCT_ROT180 },
  {"ROT270"		, 1 << (int) OCT_ROT270 }, 
  {"MX_ROT90"		, 1 << (int) OCT_MX_ROT90 },
  {"MY_ROT90"		, 1 << (int) OCT_MY_ROT90 },
  {"END_OF_TABLE"	, 0}
}; 


/*
 * "NOT 
 */
int decodeTransformConstraints(inputString)
    char *inputString;
{
  char	*s = inputString;
  char 	word[20];
  int	mask = 0 ;
  int	not_mask = 0x0;
  int	negate = FALSE;
  int	i;
  int	found;

  while ( *s!='\0' && *s == ' ') s++;	/* skip over extra leading spaces */
  while (sscanf( s, "%20s" , word) == 1) {
    while ( *s!='\0' && *s != ' ') s++;	/* skip over the word */
    while ( *s!='\0' && *s == ' ') s++;	/* skip over extra spaces */
    
    if (strcmp(word,"NOT")==0) {
      if (negate == TRUE) {
	(void) sprintf(errmsg,"Too many \"NOT\"'s in %s", inputString);
	Warning(errmsg);
	return(NOT_OK);
      }
      negate = TRUE;
      mask = 0xffff;
      not_mask = 0;
      continue;
    }
    
    found = FALSE;
    i = -1;			/* So we start from 0 in the first iteration */
    while ( ++i ,  strcmp(tdTable[i].name,"END_OF_TABLE") != 0) {
      if (strcmp(tdTable[i].name,word)==0) {
	if (negate) {
	  not_mask  |= tdTable[i].mask;
	} else {
	  mask |= tdTable[i].mask;
	}
	found = TRUE;
	break;
      }
    }
    if (found == FALSE) {
	(void) sprintf(errmsg,"Unparsable word %s in %s", word, inputString);
	Warning(errmsg);
	return(NOT_OK);
    }
  }
  /* printf( "%s -> %d\n", inputString , mask & ~not_mask ); */
  return(mask & ~not_mask);
}


char *encodeTransformConstraints(nMask)
    int nMask;			/* the negative mask */
{
  static char 	nStrValue[140];
  static char 	pStrValue[140];
  int		i = 0;
  int		tmp;
  int		pMask = ~ nMask;/* the positive mask */

  nStrValue[0] = '\0';		/* reset the string */
  pStrValue[0] = '\0';		/* reset the string */
  
  if (nMask == 0) {
    return("");			/* No constraints */
  }
  /* Decode both masks and choose the shortest */
  strcpy(nStrValue,"NOT");
  while (strcmp(tdTable[i].name,"END_OF_TABLE")!=0) {
    tmp = nMask & tdTable[i].mask; 
    if ( tmp == tdTable[i].mask) {
      strcat(nStrValue," ");
      strcat(nStrValue,tdTable[i].name);
      nMask = nMask & ~tdTable[i].mask; /* clear the bits used in this iteration */
    }
    i++;
  }

  i = 0;
  while (strcmp(tdTable[i].name,"END_OF_TABLE")!=0) {
    tmp = pMask & tdTable[i].mask; 
    if ( tmp == tdTable[i].mask) {
      strcat(pStrValue," ");
      strcat(pStrValue,tdTable[i].name);
      pMask = pMask & ~tdTable[i].mask; /* clear the bits used in this iteration */
    }
    i++;
  }

  if (strlen(nStrValue) < strlen(pStrValue)) {
    return(nStrValue);
  } else {
    return(pStrValue);
  }
}



int processLegalTransforms(cell)
    MacroCell *cell;
{
  int		mask;		/* temp. storage for cell->legalTransforms */
  int		i, j;		/* counters */
  octObject 	Instance;
  octObject	Master;

  cell->legalTransforms  = 0;
  cell->legalTransformsArray = NULL;
  cell->numberOfLegalTransforms = 8;

  Instance.objectId = cell->cellId;
  if (octGetById(&Instance)!=OCT_OK) FatalError("cellId field corrupted");

  if (octGetByName(&Instance ,&LegalTransform)!=OCT_OK) {
    Master.type = OCT_FACET;
    Master.contents.facet.facet = "interface";
    Master.contents.facet.mode = "r";
    if (octOpenMaster(&Instance,&Master)< OCT_OK) {
      FatalError("in processLegalTansform: Opening master");
    }
    if (octGetByName(&Master, &LegalTransform) != OCT_OK) {
      return(OK);		/* No property, anywhere */
    }
  }
  /* The property has been found */

  mask = decodeTransformConstraints(LegalTransform.contents.prop.value.string);
  cell->numberOfLegalTransforms = 0;
  for ( i = 0 ; i < 8 ; i++) {
    if (mask & (1 << i)) cell->numberOfLegalTransforms++;
  }
  if (cell->numberOfLegalTransforms == 0) {
    (void) sprintf(errmsg,"No legal transforms for %s", cell->name);
    Warning(errmsg);
    cell->numberOfLegalTransforms = 8;
  } else if (cell->numberOfLegalTransforms < 8) {
    cell->legalTransformsArray = SH_MALLOC( int , cell->numberOfLegalTransforms) ;
    j = 0;
    for ( i = 0 ; i < 8 ; i++) {
      if (mask & (1 << i)) {
	cell->legalTransformsArray[j++]  = i;
      }
    }
  } 
  return(OK);
}
