#include "copyright.h"
#include "sparcs.h"

/*
 *  New version of new_noncrit() that correctly handles nodes that are
 *  connected together by fixed constraints.  This is done by treating
 *  such sets of nodes as a single node in the relaxation.
 *
 *  from new_noncrit 12/16/87
 *  last modified 4/3/88
 */

clusterNoncrit(verbose)
int verbose;
{
    struct node *pn, *FnextNode();
    struct branch *pb;
    int newub, newlb, newloc;
    int leftForce;	   /*  attraction to nodes on left of current node  */
    int rightForce;	   /*  attraction to nodes on right of current node  */
    int faninMin, faninMax;
    int fanoutMin, fanoutMax;
    int faninLeftForce, faninRightForce;
    int fanoutLeftForce, fanoutRightForce;
    int numerator, denominator;
    int delta;    
    
    int debug = 0;

    FschedInit();
    if (verbose)  printf("\n");
    /**
    printf("clusterNoncrit:  initial node positions:\n");
    for(pn=pNodeList; pn!=NIL(struct node); pn=pn->cpm.list) {
	printf("    node %d:  [%d, %d, %d]\n", pn->key,
	       pn->cpm.lower, pn->cpm.loc, pn->cpm.upper);
    }
    printf("\n");
    **/

    for(pn=pNodeList; pn != NIL(struct node); pn=pn->cpm.list) {
	if (pn->cpm.upper == pn->cpm.lower)  {
	    pn->cpm.loc = pn->cpm.lower;
	} else  {
	    /* 
	     *  3/25/88  MUST start from a legal configuration!  This won't
	     *  work, since may not be a legal configuration...
	     *	
	     *    pn->cpm.loc = (pn->cpm.upper + pn->cpm.lower)/2;
	     */
	    pn->cpm.loc = pn->cpm.lower;
	    Fschedule(pn);
    	    /* printf("scheduling node %s\n", pn->name); */
	}
    }

    while ((pn=FnextNode()) != NIL(struct node))  {

	ClusterCtr++;

	/*
	 *  Mark the "count" field with pn->key, to enable the depth-first
	 *  evaluation of nodes tied to pn with fixed constraints.
	 */
	pn->cpm.active = ClusterCtr;

	if (debug)  {
	    /* fprintf(stdout, "node \"%d\": ", pn->key); */
	    fprintf(stdout, "\nnode \"%d\":\n", pn->key);
	}

	newlb = -INFINITY;
	newub =  INFINITY;	

	leftForce = rightForce = 0;
	for (pb=pn->pfanins; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
	    if (pb->pfromNode->cpm.active == ClusterCtr)  {
		if (debug)  {
		    printf("clusterNoncrit:  skipping br to proc node %d\n",
			   pn->key);
		}
		continue;
	    }
	    if (pb->lowerBound == pb->upperBound)  {
		/*
		 *  Push into the fanin node, and evaluate its bounds...
		 */
		if (pb->pfromNode->cpm.active == ClusterCtr)  {
		    /*
		    printf("clusterNoncrit:  skipping br to st node %d\n",
			   pn->key);
		    */
		    continue;
		}
		if (debug)  {
		    printf("    pushing to fixed fanin %d\n",
			   pb->pfromNode->key);
		}
		evalFanin(debug, pn, pb, &faninMin, &faninMax,
			  &faninLeftForce, &faninRightForce);
		if (debug)  {
		    printf("    offset: +%d; faninMin: %d; faninMax: %d\n",
			   pb->lowerBound, faninMin, faninMax);
		    printf(" faninLeftForce: %d; faninRightForce: %d\n",
			   faninLeftForce, faninRightForce);
		}
		newlb = SPMAX(newlb, faninMin + pb->lowerBound);
		newub = SPMIN(newub, faninMax + pb->upperBound);
		leftForce += faninLeftForce;
		rightForce += faninRightForce;
	    } else  {
		if (debug)  {
		    printf("    non-fixed fanin %d\n", pb->pfromNode->key);
		}
		newlb = SPMAX(newlb,(pb->pfromNode->cpm.loc + pb->lowerBound));
		newub = SPMIN(newub,(pb->pfromNode->cpm.loc + pb->upperBound));
		leftForce += pb->force;
	    }
	    if (debug)  {
		printf("cluster:  newlb = %d, newub = %d\n", newlb, newub);
	    }
	}
	for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	    if (pb->ptoNode->cpm.active == ClusterCtr)  {
		if (debug)  {
		    printf("clusterNoncrit:  skipping br to proc node %d\n",
			   pn->key);
		}
		continue;
	    }
	    if (pb->lowerBound == pb->upperBound)  {
		if (pb->ptoNode->cpm.active == ClusterCtr)  {
		    /*
		    printf("clusterNoncrit:  skipping br to st node %d\n",
			   pn->key);
		    */
		    continue;
		}
		/*
		 *  Push into the fanout node, and evaluate its bounds...
		 */
		if (debug)  {
		    printf("    pushing to fixed fanout %d\n",
			   pb->ptoNode->key);
		}
		evalFanout(debug, pn, pb, &fanoutMin, &fanoutMax,
			   &fanoutLeftForce, &fanoutRightForce);
		if (debug)  {
		    printf("    offset: -%d; fanoutMin: %d; fanoutMax: %d\n",
			   pb->lowerBound, fanoutMin, fanoutMax);
		    printf(" fanoutLeftForce: %d; fanoutRightForce: %d\n",
			   fanoutLeftForce, fanoutRightForce);
		}
		newlb = SPMAX(newlb, fanoutMin - pb->upperBound);
		newub = SPMIN(newub, fanoutMax - pb->lowerBound);
		leftForce += fanoutLeftForce;
		rightForce += fanoutRightForce;
	    } else  {
		if (debug)  {
		    printf("    non-fixed fanout %d\n",
			   pb->ptoNode->key);
		}
		newlb = SPMAX(newlb, (pb->ptoNode->cpm.loc - pb->upperBound));
		newub = SPMIN(newub, (pb->ptoNode->cpm.loc - pb->lowerBound));
		rightForce += pb->force;
	    }
	    if (debug)  {
		printf("cluster:  newlb = %d, newub = %d\n", newlb, newub);
	    }
	}
	if (newlb > newub)  {
	    printf("clusterNoncrit:  newlb (%d) > newub (%d)\n", newlb, newub);
	    /* exit(-1); */
	}
	if (leftForce < 0  ||  rightForce < 0)  {
	    printf("clusterNoncrit:  forces can't be negative. ");
	    printf("leftForce = %d, rightForce = %d\n", leftForce, rightForce);
	    exit(-1);
	}

	/*
	 *  Now compute the location of the node...  Basically, this amounts
	 *  to making
	 *    (free space to left node)*(leftForce) =
	 *				(free space to right node)*(rightForce)
	 */

	/**
	printf("leftForce = %d, rightForce = %d, newlb = %d, newub = %d\n",
	       leftForce, rightForce, newlb, newub);
	**/

	if (leftForce == rightForce)  {
	    newloc = (newlb + newub)/2;
	} else if (leftForce == 0  &&  rightForce != 0)  {
	    newloc = newub;
	} else if (leftForce != 0  &&  rightForce == 0)  {
	    newloc = newlb;
	} else  {
	    /*
	     *  Both forces better be nonzero, and different.
	     */
	    if (leftForce <= 0  ||  rightForce <= 0)  {
		printf("%s%s", "clusterNoncrit:  forces should be ",
		       "positive and different; leftForce = ");
		printf("%d, rightForce = %d\n",leftForce,rightForce);
		exit(-1);
	    }

	    numerator = (newub - newlb)*(rightForce);
	    denominator = leftForce + rightForce;

	    /*
	     *  The offset between the left nodes and the current node...
	     */
	    /* newloc = numerator/denominator; */
	    newloc = (numerator + denominator/2)/denominator;

	    /* printf("    delta = %d", newloc); */

	    /*
	     *  Plus the base value.
	     */
	    newloc += newlb;
	    /* printf(" newloc = %d\n", newloc); */

	}

	if (newloc < pn->cpm.lower)  {
	    printf("\nnoncrit ERROR:  newloc (%d) < pn->cpm.lower (%d)\n\n",
						    newloc, pn->cpm.lower);
	    exit(-1);
	}
	if (newloc > pn->cpm.upper)  {
	    printf("\nnoncrit ERROR:  newloc (%d) > pn->cpm.upper (%d)\n\n",
						    newloc, pn->cpm.upper);
	    exit(-1);
	}

	if (verbose || debug)  {
	    fprintf(stdout, "  old [%d, %d, %d], ",
		    pn->cpm.lower, pn->cpm.loc, pn->cpm.upper);
	    fprintf(stdout, "<-- %d, %d -->, ", leftForce, rightForce);
	    fprintf(stdout, "new [%d, %d, %d]", newlb, newloc, newub);
	}

	/*
	 *  If the current node moved, schedule its fanins and fanouts that
	 *  are not on the critical path.
	 *
	 *  Note that fixed-constraint fanins/fanouts of pn CANNOT be critial
	 *  nodes; if they were, then pn would be critical as well.
	 */
	if (newloc != pn->cpm.loc)  {
	    if (verbose)  {
		fprintf(stdout," - CHANGED\n");
	    }
	    delta = newloc - pn->cpm.loc;
	    pn->cpm.loc = newloc;

	    /*
	     *  Mark as above, but use -Iterations instead of Iterations, since
	     *  all the nodes have just been marked with Iterations.  This
	     *  ensures that everything will be visited, but only once.
	     */
	    pn->cpm.active = ++ClusterCtr;

	    for (pb=pn->pfanins; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
		if (pb->pfromNode->cpm.lower != pb->pfromNode->cpm.upper)  {
		    if (pb->lowerBound == pb->upperBound)  {
			if (pb->pfromNode->cpm.active == ClusterCtr) {
			    /*
			    printf("clusterNoncrit:  skipping branch to");
			    printf(" start node %d\n", pn->key);
			    */
			    continue;
			}
			/*
			 *  Move pb->pfromNode by delta, and schedule its non-
			 *  critical neighbors.
			 */
			updateFanin(debug, pn, pb, delta);
		    } else  {
			Fschedule(pb->pfromNode);
		    }
		}
	    }
	    for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout) {
		if (pb->ptoNode->cpm.lower != pb->ptoNode->cpm.upper)  {
		    if (pb->lowerBound == pb->upperBound)  {
			if (pb->ptoNode->cpm.active == ClusterCtr)  {
			    /*
			    printf("clusterNoncrit:  skipping branch to");
			    printf(" start node %d\n", pn->key);
			    */
			    continue;
			}
			updateFanout(debug, pn, pb, delta);
		    } else  {
			Fschedule(pb->ptoNode);
		    }
		}
	    }
	} else if (verbose)  {
	    fprintf(stdout, " - SAME\n");
	}
    }
    if (verbose)  printf("\n");
    return( OK );
}



/*
 *  12/16/87
 *  last modified 3/25/88
 */

updateFanin(debug, pstartNode, pcurBranch, delta)
int debug;
struct node *pstartNode;
struct branch *pcurBranch;
int delta;
{
    struct node *pn;
    struct branch *pb;
    
    /*
     *  The node to update is the node at the tail of pcurBranch.
     */
    pn = pcurBranch->pfromNode;
    if (debug)  {
	printf("    moving fixed fanin %d from %d to %d\n", pn->key,
	       pn->cpm.loc, pn->cpm.loc + delta);
    }
    
    if (pn->cpm.active == ClusterCtr)  {
	printf("updateFanin:  active is wrong.\n");
	exit(-1);
    } else  {
	pn->cpm.active = ClusterCtr;
    }

    pn->cpm.loc += delta;

    /*
     *  Schedule the noncritical neighbors of pn, EXCEPT for the fanout at
     *  the head of pcurBranch (which caused this routine to be invoked).
     */
    for (pb=pn->pfanins; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
	/*
	 *  Skip branch if it connects to pstartNode, to avoid looping if the
	 *  constraints happen to be cyclic.
	 */
	if (pb->pfromNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("updateFanin:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->pfromNode->cpm.lower != pb->pfromNode->cpm.upper)  {
	    if (pb->lowerBound == pb->upperBound)  {
		updateFanin(debug, pstartNode, pb, delta);
	    } else  {
		Fschedule(pb->pfromNode);
	    }
	}
    }
    for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	if (pb->ptoNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("updateFanin:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->ptoNode->cpm.lower != pb->ptoNode->cpm.upper)  {
	    if (pb->lowerBound == pb->upperBound)  {
		updateFanout(debug, pstartNode, pb, delta);
	    } else  {
		Fschedule(pb->ptoNode);
	    }
	}
    }
}



/*
 *  12/16/87
 *  last modified 3/25/88
 */

updateFanout(debug, pstartNode, pcurBranch, delta)
int debug;
struct node *pstartNode;
struct branch *pcurBranch;
int delta;
{
    struct node *pn;
    struct branch *pb;
    
    /*
     *  The node to update is the node at the HEAD of pcurBranch.
     */
    pn = pcurBranch->ptoNode;
    if (debug)  {
	printf("    moving fixed fanout %d from %d to %d\n", pn->key,
	       pn->cpm.loc, pn->cpm.loc + delta);
    }
    if (pn->cpm.active == ClusterCtr)  {
	printf("updateFanout:  active is wrong.\n");
	exit(-1);
    } else  {
	pn->cpm.active = ClusterCtr;
    }
    
    pn->cpm.loc += delta;

    /*
     *  Schedule the noncritical neighbors of pn, EXCEPT for the fanin at
     *  the tail of pcurBranch (which caused this routine to be invoked).
     */
    for (pb=pn->pfanins; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
	if (pb->pfromNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("updateFanout:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->pfromNode->cpm.lower != pb->pfromNode->cpm.upper)  {
	    if (pb->lowerBound == pb->upperBound)  {
		updateFanin(debug, pstartNode, pb, delta);
	    } else  {
		Fschedule(pb->pfromNode);
	    }
	}
    }
    for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	if (pb->ptoNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("updateFanout:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->ptoNode->cpm.lower != pb->ptoNode->cpm.upper)  {
	    if (pb->lowerBound == pb->upperBound)  {
		updateFanout(debug, pstartNode, pb, delta);
	    } else  {
		Fschedule(pb->ptoNode);
	    }
	}
    }
}
