#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/

/*
 *  Level-oriented scheduler for nodes.
 *
 *  Modified 9/23/85 for sparcs graph structs.
 *  last modified 9/22/88
 */

struct node **LFirstNode, **LLastNode;
struct node *EOL;
int SchedCount=0, SchedMaxLevels=0, ActiveNodes=0;

Lschedule(pnode)
struct node *pnode;
{
    if (pnode->cpm.level>SchedMaxLevels || pnode->cpm.level < 0)  {
	fprintf(stderr, "Lschedule: level %d at node %d > %d\n", pnode->cpm.level,
						   pnode->key, SchedMaxLevels);
	exit(0);
    }

    if (pnode->cpm.sched != NIL(struct node))  return; /*  already scheduled  */
    if (LFirstNode[pnode->cpm.level] == EOL)
	LFirstNode[pnode->cpm.level] = pnode;
    else
	LLastNode[pnode->cpm.level]->cpm.sched = pnode;
    
    LLastNode[pnode->cpm.level] = pnode;
    pnode->cpm.sched = EOL;
    SchedCount++;
    ActiveNodes++;
}

/*
 *  Processing routine for scheduler.
 */
struct node *
LnextNode(direction)
int direction;
{
    struct node *pnode;
    int level, inc_level, start_level;
    
    if (ActiveNodes <= 0)  return( NULL );	 /*  none left  */

    if (direction == REVERSE)  {
	start_level = SchedMaxLevels-1;
	inc_level = -1;
    } else  {
	start_level = 0;
	inc_level = 1;
    }
    for (level= start_level; level>=0 && level < SchedMaxLevels; level += inc_level)  {
	if (LFirstNode[level] != EOL)  {
	    pnode = LFirstNode[level];
	    LFirstNode[level] = pnode->cpm.sched;
	    pnode->cpm.sched = NIL(struct node);
	    ActiveNodes--;
	    if (LFirstNode[level] == EOL)  LLastNode[level] = EOL;

	    return( pnode );
	}
    }

    fprintf(stderr, "nextNodeLevel: lost a node somehow\n");
    fprintf(stderr, "               active = %d, schedCount = %d\n", ActiveNodes,
								  SchedCount);
    exit(1);
    /*NOTREACHED*/
}

/*
 *  Initialize the scheduler.
 */
void
LschedInit(maxLevels)
int maxLevels;
{
    int level;

    SchedCount = 0;
    ActiveNodes = 0;
    /**
    EOL = (struct node *)malloc((unsigned)sizeof(struct node));
    LFirstNode = (struct node **)malloc((unsigned)
					    maxLevels*sizeof(struct node *));
    LLastNode = (struct node **)malloc((unsigned)
					    maxLevels*sizeof(struct node *));
    **/
    EOL = ALLOC(struct node, 1);
    LFirstNode = ALLOC(struct node *, maxLevels);
    LLastNode = ALLOC(struct node *, maxLevels);
    SchedMaxLevels = maxLevels;
    for (level=0; level<SchedMaxLevels; level++)  {
	LFirstNode[level] = EOL;
	LLastNode[level] = EOL;
    }
}

/*
 *  Return the number of schedules.
 */
LschedCount()
{
    return(SchedCount);
}
