#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/

/*
 *  Sanity check for the graph.  Verify source & sink.
 *
 *  Modified 9/23/85 to work with sparcs graph data structs.
 *  Modified 5/2/86 to print according to TraceFlag.
 *
 *  last modified 4/8/87
 */

sanitize(verbose)
int verbose;
{
    struct node *pnode, *pfoNode;
    struct branch *pb, *pfanin, *pfanout;
    int allDone, faninLevel;
    extern int FschedInit(), Fschedule(), FschedCount();
    extern struct node *FnextNode();
    int i;

    if (TraceFlag)  {
	printf("\tLeveling...  ");
	fflush(stdout);
    }

    /*
     *  First make sure that there are no spurious sources/sinks.
     */
    for (i=1; i<=NumElements; i++)  {

	pnode = NodeArray[i];
	if (pnode->key == UNDEF)  continue;

	if (verbose) {
	    fprintf(stderr, "node %d :       fanins:  ", pnode->key);
	    pb = pnode->pfanins;
	    for (; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
	        fprintf(stderr, "(node %d, lb %d, ub %d) ", pb->pfromNode->key,
			pb->lowerBound, pb->upperBound);
	    }
	    fprintf(stderr, "\n fanouts: ");
	    pb = pnode->pfanouts;
	    for (; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	        fprintf(stderr, "(node %d, lb %d, ub %d) ", pb->ptoNode->key,
			pb->lowerBound, pb->upperBound);
	    }
	    fprintf(stderr, "\n");
	}
	if (pnode->cpm.numfanin == 0)  {
	    errRaise( "SPARCS", SPARCS_BUG_CODE, 
		     "sanitize:  found an extra source. (node %d)\n", pnode->key);
	}
	if (pnode->cpm.numfanout == 0) {
	    errRaise( "SPARCS", SPARCS_BUG_CODE, 
		     "sanitize:  found an extra sink. (node %d)\n", pnode->key);
	}
    }

    /*
     *  Level the graph; note that loops are errors in this case.
     */
    FschedInit();
    pSource->cpm.level = 0;
    Fschedule(pSource);
    if (verbose)  {
	fprintf(stderr,"sanitize: entering levelling code\n");
    }

    while ((pnode = FnextNode()) != NIL(struct node))  {
	/*
 	 *  Update level of current node
  	 */
	pfanin = pnode->pfanins;
	while (pfanin != NIL(struct branch))  {
	    faninLevel = (pfanin->pfromNode)->cpm.level;
	    pnode->cpm.level = SPMAX(pnode->cpm.level, faninLevel+1);
	    pfanin = pfanin->pnextFanin;
	}
	MaxLevel = SPMAX(pnode->cpm.level, MaxLevel);
	if (verbose)
	    fprintf(stderr,"sanitize: at %d level is %d\n", pnode->key,
							 pnode->cpm.level);
	/*
 	 *  Schedule fanout nodes if all their inputs have been processed
 	 */
	pfanout = pnode->pfanouts;
	while (pfanout != NIL(struct branch))  {
	    pfoNode = pfanout->ptoNode;
	    if (pfoNode->cpm.sched == NIL(struct node))  {
		if (verbose)
		    fprintf(stderr,"sanitize: proc %d,", pfoNode->key);
		allDone = TRUE;
		pfanin = pfoNode->pfanins;
		while (pfanin != NIL(struct branch))  {
		    if (pfanin->pfromNode->cpm.level == UNDEF)
			allDone = FALSE;
		    pfanin = pfanin->pnextFanin;
		}
		if (allDone == TRUE)  {
		    Fschedule(pfoNode);
		    if (verbose)  fprintf(stderr, " scheduled\n");
		} else  {
		    if (verbose)  fprintf(stderr," not scheduled\n");
		}
	    }
	    pfanout = pfanout->pnextFanout;
	}
    }
    if (verbose)  fprintf(stderr,"sanitize: finished levelling\n");
    MaxLevel++;    /* add 1 for level 0 */

    /*
     *  Walk node list & look for unprocessed (loop) nodes
     */
    allDone = TRUE;
    for (pnode=pNodeList; pnode!=NIL(struct node); pnode=pnode->cpm.list)  {
	if (pnode->cpm.level == UNDEF)  {
	    errRaise( "SPARCS", SPARCS_SANITIZE_CODE, "Some nodes are part of loops" );
	    fprintf(stderr,"sanitize: node %d is part of loop\n", pnode->key); 
	    allDone = FALSE;
	}
    }
    if (allDone == FALSE)  {
	return(ERROR);
    } else  {
        return(OK);
    }
}
