#include "copyright.h"
#include "sparcs.h"
#include "errtrap.h"
#include "vov.h"
#include "oct.h"
#include "oh.h"

/*#include "graph.h"*/

/* #define SCHED_FACTOR	7 */	/* for setting maximum schedules */
#define SCHED_FACTOR	30	/* for setting maximum schedules */

char* printElementByKey( key )
    int key;
{
    struct element el;
    octId id;
    octObject obj;

    if ( key == INFINITY ) {
	return ">>>SINK<<<<";
    }
    el = ElementByKey[key];
    id = el.ptr.po->instanceId;
    OH_ASSERT(ohGetById( &obj, id ));

    return  ohFormatName( &obj );
    
}


/*
 *  Solve the (assumed solvable) graph.
 *
 *  Modified 9/23/85 to work with sparcs graph structs.  Modified 9/25/85 to
 *  work with negative coordinates.  Modified 5/2/86 to print according to
 *  TraceFlag.  Modified 2/7/87 to remove some sloppy parts of the algorithm
 *  that could cause failures in extreme cases.
 *
 *  last modified 2/17/88
 */

solve(direction, verbose)
    directionType direction;
    int verbose;
{
    struct branch *pbr;
    struct node *pnode, *nextNode();
    int sourceLower;			/* lower bound on source location */
    extern struct node *LnextNode();
    extern int Lschedule(), LschedInit();
    extern int LschedCount(), MaxLevels;
    int newlb, newub, newloc, maxSched;
    int debug = 0;
    
    if(verbose)  fprintf(stderr,"solve: solving graph\n");

    /*
     * XXX Fix this!!
     */
    maxSched = NumNodes*NumNodes*NumNodes;
    maxSched = NumNodes * SCHED_FACTOR;
    if (maxSched < 0)  maxSched = INFINITY;

    /*
     *  Determine absolute lower bounds.
     */
    if(verbose)  fprintf(stderr,"solve: determining lower bounds\n");

    /*
     *  Note that node locations are set after this routine, in noncrit; until
     *  then their values are not determined.
     */
    if (direction == HORIZONTAL)  {
	pSource->cpm.lower =  pSource->loc.x;
	pSource->cpm.upper =  INFINITY;
	pSink->cpm.lower   = -INFINITY;
	pSink->cpm.upper   =  INFINITY;
    } else  {
	pSource->cpm.lower =  pSource->loc.y;
	pSource->cpm.upper =  INFINITY;
	pSink->cpm.lower   = -INFINITY;
	pSink->cpm.upper   =  INFINITY;
    }
    sourceLower = pSource->cpm.lower;

    /*
     *  There is a very small chance that the coordinate space is too small.
     *  If it is, either the source loc is < -INFINITY or the sink loc will
     *  turn out to be > INFINITY.
     */
    if (pSource->cpm.lower < -INFINITY)  {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "cpm_solve:  layout's left edge coordinate < -INFINITY.\n");
    }

    if (TraceFlag)  {
	printf("forward lp... ");
	fflush(stdout);
    }

    if (debug)  {
	if (TraceFlag)  printf("\n");
        printf("\nBegin solve:  pSource lower, loc, upper: %d %d %d\n",
	     		pSource->cpm.lower, pSource->cpm.loc,
			pSource->cpm.upper);
        printf("Begin solve:  pSink lower, loc, upper:   %d %d %d\n",
	     		pSink->cpm.lower, pSink->cpm.loc,
			pSink->cpm.upper);
    }

    LschedInit(MaxLevel);
    pbr = pSource->pfanouts;
    while (pbr != NIL(struct branch))  {
	Lschedule(pbr->ptoNode);
	pbr = pbr->pnextFanout;
    }


    /*
     *  No node may pass the source, BUT nodes may pass the sink.  This is
     *  due to element origins being their lower-left corners.  So, newlb
     *  is bounded by sourceLower.
     */
    while( (pnode=LnextNode(FORWARD))!=NIL(struct node) )  {

	newlb = sourceLower;
	pbr = pnode->pfanins;
	/* printf("    Processing %s\n", printElementByKey(pnode->key)); */
	while (pbr != NIL(struct branch))  {
	    newlb = SPMAX(newlb, (pbr->pfromNode->cpm.lower + pbr->lowerBound));
	    pbr = pbr->pnextFanin;
	}

	newub = sourceLower;
	pbr = pnode->pfanouts;
	while (pbr != NIL(struct branch))  {
	    newub = SPMAX (newub, (pbr->ptoNode->cpm.lower - pbr->upperBound));
	    pbr=pbr->pnextFanout;
	}
	newloc = SPMAX(pnode->cpm.lower, SPMAX(newlb, newub));
	if (newloc != pnode->cpm.lower) {
	    pnode->cpm.lower = newloc;
	    pbr = pnode->pfanins;
	    while (pbr != NIL(struct branch))  {
		if(pbr->upperBound < INFINITY)  Lschedule(pbr->pfromNode);
		pbr = pbr->pnextFanin;
	    }
	    pbr = pnode->pfanouts;
	    while (pbr != NIL(struct branch))  {
		if (pbr->lowerBound < INFINITY)  Lschedule(pbr->ptoNode);
	        pbr = pbr->pnextFanout;
	    }
	}

	if (LschedCount() >= maxSched)  {
	    errRaise( "SPARCS", SPARCS_OVERCON_CODE,
		     "overconstraint detected:\n forward pass of longest-path\n");
	}
    }
    if (TraceFlag)  {
	printf("(%d sched)", LschedCount());
	fflush(stdout);
    }

    /*
     *  The dual of the check on the size of the layout performed above.
     */
    if (pSink->cpm.lower > INFINITY)  {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "cpm_solve:  layout's right edge coordinate > INFINITY.\n");
    }
    
    /*
     *  determine absolute upper bounds
     */
    pSink->cpm.upper = pSink->cpm.lower;


    LschedInit(MaxLevel);
    pbr = pSink->pfanins;
    while (pbr!=NIL(struct branch))  {
	Lschedule(pbr->pfromNode);
	pbr=pbr->pnextFanin;
    }
    
    if (TraceFlag)  {
	printf("   reverse lp... ");
	fflush(stdout);
    }

    /*
     *  Since nodes MAY pass the sink, their max locations are bounded
     *  by infinity.
     */
    while ((pnode=LnextNode(REVERSE))!=NIL(struct node) ) {
	newlb = INFINITY;
	for (pbr = pnode->pfanins; pbr!=NIL(struct branch); pbr=pbr->pnextFanin) {
	    newlb = SPMIN(newlb, (pbr->pfromNode->cpm.upper + pbr->upperBound));
	}
	newub = INFINITY;
	for( pbr = pnode->pfanouts; pbr!=NIL(struct branch); pbr=pbr->pnextFanout) {
	    newub = SPMIN(newub, (pbr->ptoNode->cpm.upper - pbr->lowerBound));
	}
	newloc = SPMIN(pnode->cpm.upper, SPMIN(newlb, newub));
	if (newloc != pnode->cpm.upper)  {
	    pnode->cpm.upper = newloc;
	    pbr = pnode->pfanins;
	    while (pbr!=NIL(struct branch))  {
		if(pbr->lowerBound < INFINITY) Lschedule(pbr->pfromNode);
	        pbr=pbr->pnextFanin;
	    }
	    pbr = pnode->pfanouts;
	    while (pbr != NIL(struct branch))  {
		if(pbr->upperBound < INFINITY) Lschedule(pbr->ptoNode);
	        pbr = pbr->pnextFanout;
	    }
	}
	if (LschedCount() >= maxSched)  {
	    errRaise( "SPARCS", SPARCS_OVERCON_CODE, "overconstraint detected on reverse pass of longest-path\n");
	}
    }

    /*
     *  If the lower and upper bounds on the source node's location are not
     *  the same, then there is a problem.  Could occur for a very big layout
     *  where the sink's loc is just less than infinity but another node has
     *  passed it.
     */
    if (pSource->cpm.lower != pSource->cpm.upper)  {
	errRaise( "SPARCS", SPARCS_BUG_CODE, "cpm_solve:  source's min (%d) != source's max (%d)\n",
		pSource->cpm.lower, pSource->cpm.upper);
    }


    if (TraceFlag)  {
	printf("(%d sched)\n", LschedCount());
	fflush(stdout);
    }

    return(OK);
}
