#include "copyright.h"
#include "sparcs.h"

/*
 *  12/16/87
 *  last modified 3/25/87
 */

evalFanout(debug, pstartNode, pcurBranch, pmin, pmax, pleftForce, prightForce)
int debug;
struct node *pstartNode;	/* node clusterNoncrit is working on */
struct branch *pcurBranch;
int *pmin;
int *pmax;
int *pleftForce;
int *prightForce;
{
    struct node *pn;
    struct branch *pb;
    int nbrMin, nbrMax;		/* neighbor (fi/fo) min and max bounds */
    int nbrLeftForce;		/* left force on neighbor */
    int nbrRightForce;

    int newlb = -INFINITY;
    int newub =  INFINITY;	

    int leftForce = 0;
    int rightForce = 0;

    /*
     *  The node to evaluate is the node at the HEAD of pcurBranch.  Mark it
     *  with the key of the starting-point node.
     */
    pn = pcurBranch->ptoNode;
    pn->cpm.active = ClusterCtr;

    for (pb=pn->pfanins; pb!=NIL(struct branch); pb=pb->pnextFanin)  {
	/*
	 *  Skip branch if node it connects to has been reached already, to
	 *  avoid looping if the constraints happen to be cyclic.
	 */
	if (pb->pfromNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("evalFanout:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->lowerBound == pb->upperBound)  {
	    if (debug)  {
		printf("\tevalFanout:  pushing to fixed fanin %d\n",
		       pb->pfromNode->key);
	    }
	    evalFanin(debug, pstartNode, pb, &nbrMin, &nbrMax,
		      &nbrLeftForce, &nbrRightForce);
	    newlb = SPMAX(newlb, nbrMin + pb->lowerBound);
	    newub = SPMIN(newub, nbrMax + pb->upperBound);
	    leftForce += nbrLeftForce;
	    rightForce += nbrRightForce;
	} else  {
	    if (debug)  {
		printf("\tevalFanout:  non-fixed fanin %d\n",
		       pb->pfromNode->key);
	    }
	    newlb= SPMAX(newlb, (pb->pfromNode->cpm.loc + pb->lowerBound));
	    newub= SPMIN(newub, (pb->pfromNode->cpm.loc + pb->upperBound));
	    leftForce += pb->force;
	}
	if (debug)  {
	    printf("\t    evalFanout: newlb = %d, newub = %d\n", newlb, newub);
	}
    }
    for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	if (pb->ptoNode->cpm.active == ClusterCtr)  {
	    /*
	      printf("evalFanout:  skipping branch to start node %d\n",
	      pstartNode->key);
	      */
	    continue;
	}
	if (pb->lowerBound == pb->upperBound)  {
	    if (debug)  {
		printf("\tevalFanout:  pushing to fixed fanout %d\n",
		       pb->ptoNode->key);
	    }
	    evalFanout(debug, pstartNode, pb, &nbrMin, &nbrMax,
		       &nbrLeftForce, &nbrRightForce);
	    newlb = SPMAX(newlb, nbrMin - pb->upperBound);
	    newub = SPMIN(newub, nbrMax - pb->lowerBound);
	    leftForce += nbrLeftForce;
	    rightForce += nbrRightForce;
	} else  {
	    if (debug)  {
		printf("\tevalFanout:  non-fixed fanout %d\n",
		       pb->ptoNode->key);
	    }
	    newlb= SPMAX(newlb, (pb->ptoNode->cpm.loc - pb->upperBound));
	    newub= SPMIN(newub, (pb->ptoNode->cpm.loc - pb->lowerBound));
	    rightForce += pb->force;
	}
	if (debug)  {
	    printf("\t    evalFanout: newlb = %d, newub = %d\n", newlb, newub);
	}
    }
    *pmin = newlb;
    *pmax = newub;
    *pleftForce = leftForce;
    *prightForce = rightForce;
    if (debug)  {
	printf("\tevalFanout:  bounds of fanout:  [%d, %d, %d]",
	       newlb, pn->loc, newub);
	printf("  leftForce: %d  rightForce: %d\n",
	       leftForce, rightForce);
    }
}
