#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/

/*
 *  New routine to place non-critical nodes in correct locations.
 *
 *  Force values are integers >= 0.  All forces are interpreted as ATTRACTIVE;
 *  a node is attracted to its neighbors to the left and right in proportion
 *  to the total left and right forces.  The position is thus inversely
 *  proportional to the force; a strong force ==> a small spacing.
 *
 *  **** ROUNDING REMOVED 2/20/86 ****
 *  The division used to find the position is rounded.  This requires a float
 *  divide.  This rounding may be overkill.
 *
 *  2/13/86
 *  last modified 4/8/87
 */

new_noncrit(verbose)
int verbose;
{
    struct node *pn, *FnextNode();
    struct branch *pb;
    int newub, newlb, newloc;
    int leftForce;	   /*  attraction to nodes on left of current node  */
    int rightForce;	   /*  attraction to nodes on right of current node  */
    int numerator, denominator;

    /*verbose = 1;*/
    
    FschedInit();
    if (verbose)  printf("new_noncrit, before:\n");
    for (pn=pNodeList; pn != NIL(struct node); pn=pn->cpm.list) {
	if (pn->cpm.upper == pn->cpm.lower)  {
	    pn->cpm.loc=pn->cpm.lower;
	} else  {
	    pn->cpm.loc = (pn->cpm.upper+pn->cpm.lower)/2;
	    Fschedule(pn);
	}
	if (verbose)  {
	    printf("    node %d:  [%d, %d, %d]\n", pn->key, pn->cpm.lower,
		    pn->cpm.loc, pn->cpm.upper);
	}
    }
    if (verbose)  printf("\n");

    while ((pn=FnextNode()) != NULL)  {

	newlb = -INFINITY;
	newub = INFINITY;	

	for (leftForce=0, pb=pn->pfanins; pb!= NULL; pb=pb->pnextFanin)  {
	    newlb= SPMAX(newlb, (pb->pfromNode->cpm.loc + pb->lowerBound));
	    newub= SPMIN(newub, (pb->pfromNode->cpm.loc + pb->upperBound));

	    if (pb->force != UNDEF)  leftForce += pb->force;
	}
	for (rightForce=0, pb=pn->pfanouts; pb != NULL; pb=pb->pnextFanout)  {
	    newlb= SPMAX(newlb, (pb->ptoNode->cpm.loc - pb->upperBound));
	    newub= SPMIN(newub, (pb->ptoNode->cpm.loc - pb->lowerBound));

	    if (pb->force != UNDEF)  rightForce += pb->force;
	}

	/*
	 *  Now compute the location of the node...  Basically, this amounts
	 *  to making
	 *    (free space to left node)*(leftForce) =
	 *				(free space to right node)*(rightForce)
	 */
	if (leftForce!=0 && rightForce!=0)  {

	    numerator = (newub - newlb)*(rightForce);
	    denominator = leftForce + rightForce;

	    /*
	     *  The offset between the left nodes and the current node...
	     */
	    newloc = numerator/denominator;

	    /*
	     *  Round up if needed...
	     */
	    /************
	    if (((float)numerator)/((float)denominator) - newloc  >=  .5)  {
		newloc += 1;
	    }
	    ************/

	    /*
	     *  Plus the base value.
	     */
	    newloc += newlb;

	} else if (leftForce==0 && rightForce!=0)  {
	    newloc = newub;
	} else if (leftForce!=0 && rightForce==0)  {
	    newloc = newlb;
	} else  {
	    newloc = (newlb + newub)/2;     /*  leftForce = rightForce = 0  */
	}

	if (verbose)  {
	    fprintf(stdout, "noncrit:  node %d:  old [%d, %d, %d], ",
			    pn->key, pn->cpm.lower, pn->cpm.loc, pn->cpm.upper);
	    fprintf(stdout, "new [%d, %d, %d]", newlb, newloc, newub);
	}

	if (newloc < pn->cpm.lower)  {
	    printf("\nnoncrit ERROR:  newloc (%d) < pn->cpm.lower (%d)\n\n",
						    newloc, pn->cpm.lower);
	    exit(-1);
	}
	if (newloc > pn->cpm.upper)  {
	    printf("\nnoncrit ERROR:  newloc (%d) > pn->cpm.upper (%d)\n\n",
						    newloc, pn->cpm.upper);
	    exit(-1);
	}

	/*
	 *  If the current node moved, schedule its fanins and fanouts that
	 *  are not on the critical path.
	 */
	if (newloc != pn->cpm.loc)  {
	    if (verbose)  {
		fprintf(stdout," - changed\n");
	    }
	    pn->cpm.loc = newloc;
	    for (pb=pn->pfanins; pb != NULL; pb=pb->pnextFanin)  {
		if (pb->pfromNode->cpm.lower != pb->pfromNode->cpm.upper)  {
		    Fschedule(pb->pfromNode);
		}
	    }
	    for (pb=pn->pfanouts; pb != NULL; pb=pb->pnextFanout)  {
		if (pb->ptoNode->cpm.lower != pb->ptoNode->cpm.upper)  {
		    Fschedule(pb->ptoNode);
		}
	    }
	} else if (verbose)  {
	    fprintf(stdout, " - no change\n");
	}
    }
    if (verbose)  {
	printf("\nnew_noncrit, after:\n");
	for (pn=pNodeList; pn != NIL(struct node); pn=pn->cpm.list) {
	    printf("    node %d:  [%d, %d, %d]\n", pn->key, pn->cpm.lower,
		    pn->cpm.loc, pn->cpm.upper);
	}
        printf("\n");
    }
    return( OK );
}
