#include "copyright.h"
#include "sparcs.h"
#include "st.h"
#include "cache.h"
#include "oct.h"
#include "fang.h"
#include "harpoon.h"
#include "vov.h"

/*
 *  Cache the master of this instance.
 *
 *  last modified 9/22/88
 */

struct master *
cacheMaster(pinst)
octObject *pinst;
{
    octObject master, layer, frame, term;
    octGenerator layerGenId, frameGenId, termGenId;
    octStatus status, layerGenStatus, frameGenStatus, termGenStatus;
    struct master *pm;
    struct layerList *plL;
    struct termList *ptL;
    struct cacheFrame *pf;
    struct cacheTerm *pt;
    int i, layers;
    int count;
    int sparcsLayer;			/* internal layer number */
    char *masterName;

    fa_geometry geo;
    fa_box *pbox, *pfirstBox;
    char buffer[BIGBUF];

    int numPFrames = 0;
    int numTerms = 0;
    int termId = 0;

    master.type = OCT_FACET;
    layer.type = OCT_LAYER;
    term.type = OCT_TERM;

    master.contents.facet.facet = "interface";
    master.contents.facet.version = OCT_CURRENT_VERSION;
    master.contents.facet.mode = "r";
    status = octOpenMaster(pinst, &master);
    if (status != OCT_OLD_FACET && status != OCT_ALREADY_OPEN)  {
	sprintf(buffer, "cacheMaster: can't open master %s/%s/%s.",
		master.contents.facet.cell,
		master.contents.facet.view,
		master.contents.facet.facet);
	octError(buffer);
	exit(-1);
    }
    VOVinputFacet( &master );

    pm = ALLOC(struct master, 1);
    pm->players = NIL(struct layerList);
    pm->pterms  = NIL(struct termList);
    
    if (octInitGenContents(&master, OCT_LAYER_MASK, &layerGenId) != OCT_OK)  {
	octError("cacheMaster:  can't begin layer generator.");
	exit(-1);
    }
    while ((layerGenStatus=octGenerate(&layerGenId, &layer)) == OCT_OK)  {
	/*
	 *  This code a subset of procPFrames() of the stand-alone version...
	 *
	 *  Get the internal layer number out of LayerMap.
	 */
	for (i=0; i<NumLayers; i++)  {
	    if (strcmp(LayerMap[i].name, layer.contents.layer.name) == 0)  {
		sparcsLayer = i;
		break;
	    }
	}
	if (i >= NumLayers)  {
	    /*
	     *  Layer isn't in rules file.  If it is in the list of layers to
	     *  suppress warnings for, then fine.  If it is not in that list,
	     *  and if it contains frames, then complain.  In in any case
	     *  continue with the next layer.
	     */
	    if ( noWarningTest( layer.contents.layer.name ) )   goto nextLayer;

	    if ((frameGenStatus=octGenFirstContent(&layer, OCT_GEO_MASK,
						   &frame)) == OCT_GEN_DONE)  {
		;
	    } else if (frameGenStatus == OCT_OK)  {
		printf("cacheMaster:  layer %s ", layer.contents.layer.name);
		printf("contains geos but isn't in rules file; skipping.\n");
	    } else  {
		sprintf(buffer, "cacheMaster:  trying to genFirst geo on %s",
			layer.contents.layer.name);
		octError(buffer);
	    }
	    goto nextLayer;
	}

	plL = ALLOC(struct layerList, 1);
	plL->sparcsLayer = sparcsLayer;
	plL->pframes = NIL(struct cacheFrame);

	/*
	 *  Generate the protection frames on this layer.  To distinguish
	 *  t-frames from p-frames, run a containers generator on the frame
	 *  to look for a terminal definition.  Note that the geo must be
	 *  contained by 0 f-terms (pfs) or 1 f-term (tfs).
	 */
	if (octInitGenContents(&layer, OCT_GEO_MASK, &frameGenId) != OCT_OK)  {
	    octError("cacheMaster:  can't begin p-frame generator");
	    exit(-1);
	}
	while ((frameGenStatus=octGenerate(&frameGenId, &frame)) == OCT_OK)  {
	    switch (frame.type)  {
	    case OCT_BOX:
		if (octInitGenContainers(&frame, OCT_TERM_MASK, &termGenId) !=
		    OCT_OK)  {
		    octError("cacheMaster:  can't begin terminal generator.");
		    exit(-1);
		}
		count = 0;
		while ((termGenStatus=octGenerate(&termGenId,&term))==OCT_OK) {
		    count++;
		}
		if (termGenStatus != OCT_GEN_DONE)  {
		    octError("cacheMaster:  Bad exit status from term'l gen.");
		    exit(-1);
		}
		if (count == 0)  {
		    numPFrames++;
		    pf = ALLOC(struct cacheFrame, 1);
		    pf->box.lowerL.x = frame.contents.box.lowerLeft.x;
		    pf->box.lowerL.y = frame.contents.box.lowerLeft.y;
		    pf->box.upperR.x = frame.contents.box.upperRight.x;
		    pf->box.upperR.y = frame.contents.box.upperRight.y;
		    pf->pnext = plL->pframes;
		    plL->pframes = pf;
		} else if ( count > 1)  {
		    printf("cacheMaster:  frame contained by > 1 f-term.\n");
		    exit(-1);
		}
		break;
		
	    case OCT_POLYGON:
		/*
		 *  Since polygon MUST be a p-frame.  Make the polygon into
		 *  a list of boxes, by putting it into a fa_geometry via
		 *  harpoon and then calling fa_to_boxes.
		 */
		fa_init(&geo);
		harAddObjectToFang(&frame, &geo, HAR_NO_TRANSFORM);
		fa_to_boxes(&geo, &pbox);
		
		/*
		 *  Now, make a sparcs p-frame for each box in the list.
		 */
		pfirstBox = pbox;
		for (; pbox!=NIL(struct fa_box); pbox=pbox->next)  {
		    numPFrames++;
		    pf = ALLOC(struct cacheFrame, 1);
		    pf->box.lowerL.x = pbox->left;
		    pf->box.lowerL.y = pbox->bottom;
		    pf->box.upperR.x = pbox->right;
		    pf->box.upperR.y = pbox->top;
		    pf->pnext = plL->pframes;
		    plL->pframes = pf;
		}
		fa_free_boxes(pfirstBox);
		break;
		
	    default:
		printf("cacheMaster:  geo isn't box or polygon; skipping.\n");
		break;
	    }
	}
	if (frameGenStatus != OCT_GEN_DONE)  {
	    octError("cacheMaster:  Bad exit status from frame generator.");
	    exit(-1);
	}

	/*
	 *  If the layer had p-frames link it into the structure for this
	 *  master; otherwise get rid of it.
	 */
	if (plL->pframes != NIL(struct cacheFrame))  {
	    plL->pnext = pm->players;
	    pm->players = plL;
	} else  {
	    /**
	    printf("cacheMaster:  layer %s has no p-frames.\n",
		   LayerMap[sparcsLayer].name);
	    free((char *)plL);
	    **/
	}
nextLayer:
	;
    }
    if (layerGenStatus != OCT_GEN_DONE)  {
	octError("cacheMaster:  Bad exit status from layer generator.");
	exit(-1);
    }

    /*
     *  Protection frames are cached away.  Now for the terminal frames.
     */
    if (octInitGenContents(&master, OCT_TERM_MASK, &termGenId) != OCT_OK) {
	octError("cacheMaster:  can't begin f-term generator.");
	exit(-1);
    }
    while ((termGenStatus=octGenerate(&termGenId, &term))==OCT_OK)  {
	termId++;
	
	numTerms++;
	
	ptL = ALLOC(struct termList, 1);
	ptL->termId = termId;
	ptL->pterms = NIL(struct cacheTerm);

	/*
	 *  The following resembles readTFrames() of the stand-alone version.
	 */
	
	/*
	 *  Generate the terminal frames of this f-term.
	 */
	if (octInitGenContents(&term, OCT_GEO_MASK, &frameGenId) != OCT_OK)  {
	    octError("cacheMaster:  can't begin t-frame generator");
	    exit(-1);
	}
	while ((frameGenStatus=octGenerate(&frameGenId, &frame)) == OCT_OK)  {
	    if (frame.type != OCT_BOX)  {
		printf("cacheMaster:  terminal isn't a box; skipping.\n");
		continue;
	    }
	    
	    /*
	     *  Run a containers generator on the frame to find its layer.
	     */
	    if (octInitGenContainers(&frame, OCT_LAYER_MASK, &layerGenId) !=
		OCT_OK)  {
		octError("cacheMaster:  can't begin t-frame layer gen.");
		exit(-1);
	    }
	    layers = 0;
	    while ((layerGenStatus=octGenerate(&layerGenId,&layer))==OCT_OK) {
		layers++;
	    }
	    if (layerGenStatus != OCT_GEN_DONE)  {
		octError("cacheMaster:  can't end t-frame layer gen.");
		exit(-1);
	    }
	    if (layers != 1)  {
		if (layers == 0)  {
		    printf("cacheMaster: t-frame not on a layer; skipping.\n");
		} else  {
		    printf("cacheMaster:  t-frame on %d layers; skipping.\n",
			   layers);
		}
		continue;
	    }
	    
	    /*
	     *  Get the internal layer number out of LayerMap.
	     */
	    for (i=0; i<NumLayers; i++)  {
		if (strcmp(LayerMap[i].name, layer.contents.layer.name) == 0) {
		    sparcsLayer = i;
		    break;
		}
	    }
	    if (i >= NumLayers)  {
		/*
		 *  Layer isn't in rules file, so complain.
		 */
		printf("cacheMaster:  f-term is on layer %s ",
		       layer.contents.layer.name);
		printf("which isn't in rules file; skipping.\n");
		continue;
	    }
	    
	    pt = ALLOC(struct cacheTerm, 1);
	    pt->sparcsLayer = sparcsLayer;
	    pt->box.lowerL.x = frame.contents.box.lowerLeft.x;
	    pt->box.lowerL.y = frame.contents.box.lowerLeft.y;
	    pt->box.upperR.x = frame.contents.box.upperRight.x;
	    pt->box.upperR.y = frame.contents.box.upperRight.y;
	    pt->pnext = ptL->pterms;
	    ptL->pterms = pt;
	}
	if (frameGenStatus != OCT_GEN_DONE)  {
	    octError("cacheMaster:  Bad exit status from t-frame generator.");
	    exit(-1);
	}

	/*
	 *  If the term had t-frames link it into the structure for this
	 *  master; otherwise get rid of it.
	 */
	if (ptL->pterms != NIL(struct cacheTerm))  {
	    ptL->pnext = pm->pterms;
	    pm->pterms = ptL;
	} else  {
	    printf("cacheMaster:  f-term %s has no t-frames.\n",
		   term.contents.term.name);
	    free((char *)ptL);
	}
    }
    if (termGenStatus != OCT_GEN_DONE)  {
	octError("cacheMaster:  Bad exit status from f-term generator.");
	exit(-1);
    }

    if (numTerms == 0  &&  numPFrames == 0)  {
	printf("cacheMaster:  inst with no terminal or protection frames.\n");
	exit(-1);
    }
    if (numTerms == 0)  {
	printf("cacheMaster:  WARNING - inst w/out terminals.\n");
    }

    sprintf(buffer, "%s/%s", pinst->contents.instance.master,
	    pinst->contents.instance.view);
    masterName = STRSAVE(buffer);
    
    st_insert(pMasterTable, masterName, (char *)pm);
    return(pm);
}
