#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

/*
 *  This routine does a couple of quick checks to determine if the view to be
 *  spaced is in violation of policy in an obvious way.
 *
 *  Items checked:
 *	existence of INSTANCES and CONNECTORS bags
 *	existence of VIEWTYPE property with value SYMBOLIC
 *
 *  If the facet is out-of-policy a warning is printed only; execution is not
 *  aborted.
 *
 *  9/12/86
 *  last modified 1/21/88
 */

void
checkPolicy(poFacet)
octObject *poFacet;
{
    octObject oInstBag, oConBag, oProp;
    octStatus oStatus;

    oInstBag.type = OCT_BAG;
    oConBag.type = OCT_BAG;
    oProp.type = OCT_PROP;

    /*
     *  First check for properties.
     */
    oProp.objectId = oct_null_id;
    oProp.contents.prop.name = "VIEWTYPE";
    if ((oStatus=octGetByName(poFacet, &oProp)) == OCT_OK)  {
	/*
	 *  Present, so check its type and value.
	 */
	if (oProp.contents.prop.type != OCT_STRING)  {
	    printf("WARNING - input facet's VIEWTYPE property ");
	    printf("is not string-valued.\n");
	} else  {
	    if (strcmp(oProp.contents.prop.value.string, "SYMBOLIC") != 0)  {
		printf("WARNING - input facet's VIEWTYPE property ");
		printf("does not have value \"SYMBOLIC\".\n");
	    }
	}
    } else if (oStatus == OCT_NOT_FOUND)  {
        printf("WARNING - input facet has no VIEWTYPE property.\n");
    } else  {
        printf("WARNING - can't get input facet's VIEWTYPE property.\n");
	octError("");
    }

#ifdef NOT_REALLY_USEFUL 
    oProp.objectId = oct_null_id;
    oProp.contents.prop.name = "CELLCLASS";
    if ((oStatus=octGetByName(poFacet, &oProp)) == OCT_OK)  {
	/*
	 *  Present, so check its type and value.
	 */
	if (oProp.contents.prop.type != OCT_STRING)  {
	    printf("WARNING - input facet's CELLCLASS property ");
	    printf("is not string-valued.\n");
	} else  {
	    if (strcmp(oProp.contents.prop.value.string, "MODULE") != 0)  {
		printf("WARNING - input facet's CELLCLASS property ");
		printf("does not have value \"MODULE\".\n");
	    }
	}
    } else if (oStatus == OCT_NOT_FOUND)  {
        printf("WARNING - input facet has no CELLCLASS property.\n");
    } else  {
        printf("WARNING - can't generate input facet's CELLCLASS prop.\n");
	octError("");
    }
#endif

    /*
     *  Now check for the two bags.
     */
    oInstBag.contents.bag.name = "INSTANCES";
    if ((oStatus=octGetByName(poFacet, &oInstBag)) == OCT_OK)  {
	;
    } else if (oStatus == OCT_NOT_FOUND)  {
        printf("WARNING - input facet has no INSTANCES bag.\n");
    } else  {
        printf("WARNING - can't generate input facet's INSTANCES bag.\n");
	octError("");
    }

    oConBag.contents.bag.name = "CONNECTORS";
    if ((oStatus=octGetByName(poFacet, &oConBag)) == OCT_OK)  {
	;
    } else if (oStatus == OCT_NOT_FOUND)  {
        printf("WARNING - input facet has no CONNECTORS bag.\n");
    } else  {
        printf("WARNING - can't generate input facet's CONNECTORS bag.\n");
	octError("");
    }
}
