#include "copyright.h"
#include "port.h"
#include "options.h"
#include "sparcs.h"


optionStruct optionList[] = {
    { "o",  "cell:view",   "reset output cell and/or view" },
    { "r",  "file",        "reset rules file to 'file'" },
    { "t",  "",		   "print messages tracing execution" },
    { "v",  "",            "begin with a vertical spacing" },
    { "w",  "layers",      "suppress missing-rule warnings for specified layers" },
    { "C",  "",            "ignore user constraints" },
    { "I",  "n",           "stop after n iterations" },
    { "B",  "",            "locate noncritical el'ts at their bottommost positions" },
    { "L",  "",            "locate noncritical el'ts at their leftmost positions" },
    { "R",  "",            "locate noncritical el'ts at their rightmost positions" },
    { "T",  "",            "locate noncritical el'ts at their topmost positions" },
    { "O",  "",            "use Old rules file (use TAP by default )"},
    { "A",  "",         "stop when area doesn't change" },
    { "E",  "",         "stop when no elements move" },
    { "d",  "",         "debug flag (produces a lot of output)" },
    {0,0,0}
};



/* Rewritten by AC using the options package. (July 91) */
void
commandLine(argc, argv)
int argc;
char *argv[];
{
    int stopFlag = FALSE;	/* to disallow both -S and -I */
    int c;
    extern char* optarg;
    extern int optind;
    extern int SparcsUseTap;
    extern int SparcsPrintRules;

    while ( (c = optGetOption( argc, argv )) != EOF ) {
	switch ( c ) {
	case 'D': LinkSearchDepth = atoi(optarg); break;
	case 'd': SparcsPrintRules = 1; break;
	case 'e': RuleExp = atoi(optarg); break;
	case 'o': unpackFacetName(optarg, &OutCell, &OutView); break;
	case 'r': RulesFile = STRSAVE(optarg); break;
	case 't': TraceFlag = TRUE; break;
	case 'v': HorizontalFirstFlag = FALSE; break;
	case 'B': HardBottomFlag = TRUE; break;
	case 'L': HardLeftFlag = TRUE; break;
	case 'R': HardRightFlag = TRUE; break;
	case 'T': HardTopFlag = TRUE; break;
	case 'C': NoUserConstraintsFlag = TRUE; break;
	case 'O': SparcsUseTap = 0; break;

	case 'w': noWarningAdd( optarg ); break;

	case 'I':
	    if (stopFlag == TRUE)  {
		fprintf(stderr, "Warning:  -I option overrides -S option.\n");
	    }
	    IterationLimit = atoi(optarg);
	    StopCriterion = 'I';
	    break;

	case 'A': 
	    if (IterationLimit != 0)  {
		fprintf(stderr, "Warning:  -I option overrides -A option\n" );
	    } else {
		StopCriterion = 'A';
		stopFlag = TRUE;
	    }
	    break;

	case 'E':
	    if (IterationLimit != 0)  {
		fprintf(stderr, "Warning:  -I option overrides -E option\n" );
	    } else {
		StopCriterion = 'E';
		stopFlag = TRUE;
	    }
	    break;

	default: optUsage();
	}
    }

    if ( optind > argc -1 ) {
	optUsage();
    }

    /*
     *  Now read the input cell and view names.
     */
    unpackFacetName(argv[optind], &InCell, &InView);
    
    if ( InCell == 0 || InView == 0 ) {
	errRaise( "SPARCS", 1, "Illegal value for input cell and view " );
    }

    /*
     *  Take care of the default cell name (default view/facet names
     *  handled in globals.h).
     */

    if (OutCell == NIL(char))  OutCell = STRSAVE(InCell);


    return;
}


/*
 *  Modified version of unpack_facet_name of Rudell and Laidig.
 *
 *  4/17/87
 *  last modified 4/17/87
 */

unpackFacetName(name, cell, view)
char *name;		/* "cell:view:facet", which is modified */
char **cell;		/* returned */
char **view;		/* returned */
{
    char *s;

    /* Check for silly input */
    if (name == NIL(char) || strcmp(name, "") == 0) return;
	
    /* Make a copy of the input string (because we will tear it apart) */ 
    s = STRSAVE(name);

    if (s[0] != ':')  *cell = s;
    if ((s = strchr(s, ':')) != NIL(char)) {
	*s++ = '\0';
	if (s[0] != ':')  *view = s;
	if ((s = strchr(s, ':')) != NIL(char)) {
	    *s++ = '\0';
	    if (s[0] != ':')  {
		/* facet->contents.facet.facet = s; */
		printf("the facet cannot be reset at present.\n");
		exit(-1);
	    }
	    if ((s = strchr(s, ':')) != NIL(char)) {
		*s++ = '\0';
		if (s[0] != ':')  {
		    /* facet->contents.facet.version = s; */
		    printf("the version cannot be reset at present.\n");
		    exit(-1);
		}
	    }
	}
    }
}
