#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

/*
 *  Routine that checks for multiple segments on the given layer between the
 *  given terminals.  Returns TRUE if so, FALSE otherwise.
 *
 *  3/2/88
 *  last modified 5/27/88
 */

int
dupSeg(pt1, pt2, psegment, playerName)
octObject *pt1, *pt2;	/* the aterms psegment connects */
octObject *psegment;	/* the segment being checked */
char *playerName;	/* layer that psegment is on */
{
    octObject segment, term, layer;
    octStatus segStat, termStat;
    octGenerator segGenId, termGenId;

    segment.type = OCT_PATH;
    term.type = OCT_TERM;
    layer.type = OCT_LAYER;
    
    /*
     *  Gen the segments that contain pt1.
     */
    if (octInitGenContainers(pt1, OCT_PATH_MASK, &segGenId) !=	OCT_OK)  {
	octError("dupSeg:  can't begin segment generator");
	exit(-1);
    }
    while ((segStat=octGenerate(&segGenId, &segment)) == OCT_OK)  {
	if (octIdsEqual(segment.objectId, psegment->objectId))  continue;
	/*
	 *  See if the segment contains pt2.
	 */
	if (octInitGenContents(&segment, OCT_TERM_MASK, &termGenId)!=OCT_OK) {
	    octError("dupSeg:  can't begin term'l generator");
	    exit(-1);
	}
	while ((termStat=octGenerate(&termGenId, &term)) == OCT_OK)  {
	    if (octIdsEqual(term.objectId, pt2->objectId))  {
		/*
		 *  There are duplicate segments between these terminals.
		 *  See if the layers match.
		 */
		if (octGenFirstContainer(&segment,
					 OCT_LAYER_MASK, &layer) != OCT_OK)  {
		    octError("dupSeg:  can't get layer of segment");
		    exit(-1);
		}
		if (strcmp(layer.contents.layer.name, playerName) == 0)  {
		    /*
		     *  Duplicate segments on the same layer!
		     */
		    return(TRUE);
		}
	    }
	}
	if (termStat != OCT_GEN_DONE)  {
	    octError("dupSeg:  can't end terminal generator\n");
	    exit(-1);
	}
    }
    if (segStat != OCT_GEN_DONE)  {
	octError("dupSeg:  can't end segment generator\n");
	exit(-1);
    }
    return(FALSE);
}
