#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

#define POINT_BOX_INTERSECT(array, index, bx)\
	(array[index].x >= bx.lowerL.x) && (array[index].x <= bx.upperR.x) &&\
	(array[index].y >= bx.lowerL.y) && (array[index].y <= bx.upperR.y)

#define NONE	1
#define BOTH	2
#define FIRST	3
#define SECOND	4

/*
 *  Find the sparcs terminal that intersects one or both ends of the segment.
 *  If found, assign the netId to ALL the terminal frame edges that constitute
 *  the terminal's implementation.
 *
 *  Like the routine of the same name in msparcs.
 *
 *  3/17/87
 *  last modified 9/16/88
 */

int
findTFrame(points, sparcsLayer, netId, atermId, instId, po, pt)
struct octPoint points[];
int sparcsLayer;
int netId;		/* net that segment belongs to */
int atermId;		/* octId of the aterm */
int instId;		/* octId of instance that owns aterm */
struct object **po;	/* sparcs object that owns terminal; RETURNED */
struct terminal **pt;	/* pointer to intersecting sparcs term; RETURNED */
{
    struct object *pobj;
    struct terminal *pterm, *pfirstTerm;
    struct rect box;
    int intersect;
    int atermMatch;
    int layerMatch;

    intersect = NONE;
    layerMatch = FALSE;
    atermMatch = FALSE;

    /*
     *  First find the sparcs object corresponding to the oct instance that
     *  the terminal belongs to.
     */
    for (pobj=pObjectList; pobj!=NIL(struct object); pobj=pobj->pnext)  {
	if (pobj->instanceId == instId)  break;
    }
    if (pobj == NIL(struct object))  {
	printf("findTFrame:  can't find sparcs object for oct instance.\n");
	exit(-1);
    }

    /*
     *  Skip down to the head of the sublist that contains the sparcs
     *  terminal(s) of the relevant aterm.
     */
    pterm = pobj->pterminal;
    while (pterm != NIL(struct terminal))  {
	if (pterm->atermId == atermId)  {
	    atermMatch = TRUE;
	    pfirstTerm = pterm;
	    break;
	} else  {
	    pterm = pterm->pnext;
	}
    }
    if (atermMatch == FALSE)  {
	printf("\nfindTFrame:  no aterm Id match for wire.\n");
	return(intersect);
    }

    /*
     *  Now scan for a sparcs terminal on the same layer as the segment.  All
     *  sparcs terminals for the aterm are consecutive in the list.
     */
    do  {
	if (pterm->maskNumber == sparcsLayer  &&  intersect == NONE)  {
	    layerMatch = TRUE;

	    /*
	     *  Check both ends of segment for intersection with this aterm.
	     */
	    box.lowerL.x = pobj->origin.x + pterm->bb.lowerL.x;
	    box.lowerL.y = pobj->origin.y + pterm->bb.lowerL.y;
	    box.upperR.x = pobj->origin.x + pterm->bb.upperR.x;
	    box.upperR.y = pobj->origin.y + pterm->bb.upperR.y;
	    if (POINT_BOX_INTERSECT(points, 0, box))  {
		intersect = FIRST;
	    }
	    if (POINT_BOX_INTERSECT(points, 1, box))  {
		if (intersect == FIRST)  {
		    intersect = BOTH;
		} else  {
		    intersect = SECOND;
		}
	    }
	    if (intersect != NONE)  {
		*po = pobj;
		*pt = pterm;
		break;
	    }
	}
	pterm = pterm->pnext;

    } while (pterm != NIL(struct terminal)  &&  pterm->atermId == atermId);

    /*
     *  If no intersection quit; if due to no layerMatch, then report.
     */
    if (intersect == NONE)  {
	if (layerMatch == FALSE)  {
	    printf("\nfindTFrame:  no layer match for wire.\n");
	}
	return(intersect);
    }


    /*
     *  An intersection has been found.  Set the netId for each terminal frame
     *  edge.  May be set already; if so, the old netId must agree with the
     *  new one.
     */
    pterm = pfirstTerm;
    do  {
	if (pterm->plE->netId!=UNDEF && pterm->plE->netId!=netId) {
	    goto error;
	} else  {
	    pterm->plE->netId = netId;
	    pterm->plE->pt = pterm;
	}
	if (pterm->prE->netId!=UNDEF && pterm->prE->netId!=netId) {
	    goto error;
	} else  {
	    pterm->prE->netId = netId;
	    pterm->prE->pt = pterm;
	}
	if (pterm->pbE->netId!=UNDEF && pterm->pbE->netId!=netId) {
	    goto error;
	} else  {
	    pterm->pbE->netId = netId;
	    pterm->pbE->pt = pterm;
	}
	if (pterm->ptE->netId!=UNDEF && pterm->ptE->netId!=netId) {
	    goto error;
	} else  {
	    pterm->ptE->netId = netId;
	    pterm->ptE->pt = pterm;
	}

	pterm = pterm->pnext;
    } while (pterm != NIL(struct terminal) && pterm->atermId == atermId);

    return(intersect);

error:
    printf("findTFrame:  terminal on layer %s with bounding box ",
	   LayerMap[sparcsLayer].name);
    printf("(%d, %d) (%d, %d) is in more than one net; fatal.\n",
	   pterm->plE->loc, pterm->pbE->loc, pterm->prE->loc, pterm->ptE->loc);
    printf("findTFrame:  old/new netIds conflict.\n");
    exit(-1);

    /*NOTREACHED*/
}
