#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "oh.h"
#include "da.h"
#include "facet.h"


char *printbound( n )
    int n;
{
    static char buf[8][32];
    static int count = 0;
    if ( n >= INFINITY ) return "INF";
    if ( n <= -INFINITY ) return "-INF";
    
    sprintf( buf[(++count)%8], "%d", n );
    return buf[count%8];
}


int getOctElementByKey( obj, key )
    octObject *obj;
    int key;
{
    struct element el;
    octId id;

    /* Skip source (key==0) or sink (key==INFINITY)  */
    if ( key == 0 || key == INFINITY ) {
	return 0;
    }
    el = ElementByKey[key];
    if ( el.ptr.po == 0  )  return 0; /* Not a good element. */
    id = el.ptr.po->instanceId;
    OH_ASSERT(ohGetById( obj, id ));
    return 1;
}



void highlight( bag, pnode )
    octObject *bag;
    struct node *pnode;
{
    octObject obj;
    if ( getOctElementByKey( &obj, pnode->key ) ) {
	octAttachOnce( bag, &obj );
    }
}


highlightLoop( bag, pnode )
    octObject *bag;
    struct node *pnode;
{
    int count = 0;
    struct node *pn = pnode;
    
    /*
      * There may be sevaral interlacing loops. There is no guarantee that
      * the start pnode is going to be found again. We need a way to avoid
      * infinite loops.
      */
    do {
	pn->cpm.overcon = TRUE;
	if ( count++ >= 100 ) {
	    printf( "Possible interlacing loops or more than %d nodes in loop", count );
	    break;
	}
	highlight( bag, pn );
	pn = pn->oc_node;
    } while ( pn && pn != pnode );
}



void setNodeDistance( distance, pnode, dist )
    dary distance;
    struct node *pnode;
    int dist;
{
    int key = pnode->key;
    if ( key == INFINITY ) key = daLen(distance)-1;
    *daSet( int, distance, key) = dist;
}

getNodeDistance( distance, pnode )
    dary distance;
    struct node *pnode;
{
    int dist;
    if ( pnode->key == INFINITY ) {
	static int addedSinkNode = 0;
	if ( !addedSinkNode ) {
	    *daSetLast( int, distance ) = -INFINITY; /* Add another element. */
	    addedSinkNode = 1;
	}
	return *daGetLast( int, distance );
    } else {
	return *daGet( int, distance, pnode->key );
    }
}


int bfsStep( s, pnode, otherNode, dist, distance, frontier, bag )
    struct node *s;
    struct node *pnode;
    struct node *otherNode;
    int dist;
    dary distance;
    dary frontier;
    octObject *bag;
{
    octObject obj;


    if ( otherNode->cpm.overcon == TRUE ) {
	/* printf( "Node %d is already OC\n", otherNode->key ); */
	return 0;		/* Already dealt with. */
    }


    if ( otherNode == s && dist > 0 ) {
	/* printf( "Found positive loop 1 %d", dist ); */
	s->oc_node = pnode;
	highlightLoop( bag, s );
	return 1;		/*  */
    }
    if ( dist > getNodeDistance( distance, otherNode )) {
	if ( otherNode->oc_node == pnode  ) {
	    /* printf("Found positive loop2: %d\n",dist-getNodeDistance(distance,otherNode)); */
	    highlightLoop( bag, pnode );
	    return 1;
	}
	otherNode->oc_node = pnode;
	setNodeDistance( distance, otherNode, dist );
	*daSetLast( struct node*, frontier ) = otherNode;
    }
    return 0;			/* Just go on. */
}
    




int bfs( s, bag )
    struct node *s;		/* Candidate node to be in overconstraint loop. */
    octObject *bag;		/* Bag to put nodes in overconstraint loop */
{
    dary distance;
    dary frontier;
    int i;
    int front_index = 0;



    /* Skip source or sink nodes. */
    if ( s->key == 0 || s->key == INFINITY ) return 0;

    if ( s->cpm.overcon == TRUE ) {
	/* printf( "\tNode %d already in overconstraint\n", s->key ); */
	return 0;
    }

    distance = daAlloc( int, NumElements +1 ); /* Is this OK? */
    for ( i = 0; i < NumElements + 1; i++ ) *daSet(int,distance,i) = -INFINITY; 
    for ( i = 1; i < NumElements; i++ ) NodeArray[i]->oc_node = 0;
    *daSet( int, distance, s->key ) = 0; /* Set the starting node. */

    frontier = daAlloc( struct node *, 100 );
    *daSetLast( struct node*, frontier ) = s; /* Add this node to the frontier. */



    while ( 1 ) {
	int front_last = daLen( frontier );
	int dist;
	struct node *otherNode;
	
	if ( front_index >= front_last ) break;

	/* printf( "\t%3d new nodes in frontier\n", front_last - front_index ); */
	for ( ; front_index < front_last; front_index++ ) {
	    struct node *pnode = *daGet(struct node*, frontier, front_index );
	    int nd = getNodeDistance( distance, pnode );
	    struct branch *pbr = pnode->pfanins;
	    while ( pbr != 0 ) {
		if ( pbr->upperBound < INFINITY ) {

		    dist = nd - pbr->upperBound;
		    otherNode = pbr->pfromNode;

		    if ( bfsStep( s, pnode, otherNode, dist, distance, frontier, bag ) ) {
			return dist;
		    }
		}
		pbr = pbr->pnextFanin;
	    }
	    
	    pbr = pnode->pfanouts;
	    while ( pbr != 0 ) {
		if (pbr->lowerBound > -INFINITY) {
		    dist = nd + pbr->lowerBound;
		    otherNode = pbr->ptoNode;
		
		    if ( bfsStep( s, pnode, otherNode, dist, distance, frontier, bag ) ) {
			return dist;
		    }
		}
		pbr = pbr->pnextFanout;
	    }
	}
    }


    daFree( distance, 0 );
    daFree( frontier, 0 );
    
    return 0;
}


int findOverconstraints()
{
    struct node *pnode;
    struct branch *pb;
    octObject overconBag, obj;
    int count = 0;


    for (pnode=pNodeList; pnode != NULL; pnode = pnode->cpm.list)  {
	ohCreateBag( &WorkingFacet, &overconBag, "SPARCS OVERCONSTRAINTS" );
	if (pnode->cpm.lower > pnode->cpm.loc || pnode->cpm.upper < pnode->cpm.loc ) {
	    if ( bfs( pnode, &overconBag ) ) {
		count++;
		if ( getOctElementByKey( &obj, pnode->key ) ) {
		    octAttach( &overconBag, &obj );
		}
	    }
	}
	if (pnode->cpm.lower == pnode->cpm.upper ) {
	    if ( bfs( pnode, &overconBag ) ) {
		count++;
		if ( getOctElementByKey( &obj, pnode->key ) ) {
		    ohGetOrCreateBag( &WorkingFacet, &overconBag, "SPARCS CRITICAL" );
		    octAttach( &overconBag, &obj );
		}
	    }
	}
	if (pnode->cpm.lower > pnode->cpm.upper ) {
	    if ( bfs( pnode, &overconBag ) ) {
		count++;
		if ( getOctElementByKey( &obj, pnode->key ) ) {
		    ohGetOrCreateBag( &WorkingFacet, &overconBag, "SPARCS REALLY BAD" );
		    octAttach( &overconBag, &obj );
		}
	    }
	}
	if ( ohCountContents( &overconBag, OCT_ALL_MASK ) <= 1 ) octDelete( &overconBag );
    }
    fprintf( stderr, "Found %d overconstraint loops\n", count );
    
    return count;
}


findUnscheduled()
{
    struct node *pnode;
    octObject bag;

    ohCreateBag( &WorkingFacet, &bag, "SPARCS_LOOP" );
    
    for (pnode=pNodeList; pnode!=NIL(struct node); pnode=pnode->cpm.list)  {
	if (pnode->cpm.level == UNDEF)  {
	    highlight( &bag, pnode );
	}
    }

}
