#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

/*
 *  Routine to make a new object for sparcs.  Each instance in the oct
 *  description maps to an object.
 *  An object is:
 *	struct object  {
 *	    int key;
 *	    int instanceId;
 *	    struct point origin;
 *	    struct point bbUpperR;
 *	    struct point initOrigin;
 *	    struct point lastOrigin;
 *	    struct terminal *pterminal;
 *	    struct edgePtr *pfirstMaxEP;
 *	    struct edgePtr *plastMaxEP;
 *	    struct edgePtr *pfirstMidEP;
 *	    struct edgePtr *plastMidEP;
 *	    struct edgePtr *pfirstMinEP;
 *	    struct edgePtr *plastMinEP;
 *	    struct edgePtr *pminHEP;
 *	    struct edgePtr *pmidHEP;
 *	    struct edgePtr *pmaxHEP;
 *	    struct edgePtr *pminVEP;
 *	    struct edgePtr *pmidVEP;
 *	    struct edgePtr *pmaxVEP;
 *	    struct edge *pverticalEdges;
 *	    struct edge *phorizontalEdges;
 *	    char hasPFrames;
 *	    struct object *pnext;
 *	    struct node node;
 *	    char *name;
 *	    struct octTransform transform;
 *	};
 *
 *  Called by:  readOct()
 *  Calls:	-
 *
 *  1/23/86
 *  name changed from SPoctMakeObject 2/23/86
 *  last modified 9/22/88
 */

struct object *
makeObject(pinst)
octObject *pinst;
{
    struct object *pobj;

    /*
     *  Key assignment is by global counter ElementKey.
     */

    pobj = ALLOC(struct object, 1);

    pobj->key = ElementKey++;
    pobj->instanceId = pinst->objectId;

    pobj->origin.x =  INFINITY;
    pobj->origin.y =  INFINITY;
    pobj->bbUpperR.x = -INFINITY;
    pobj->bbUpperR.y = -INFINITY;

    pobj->pterminal = NIL(struct terminal);

    pobj->pfirstMaxEP = NIL(struct edgePtr);
    pobj->plastMaxEP  = NIL(struct edgePtr);
    pobj->pfirstMidEP = NIL(struct edgePtr);
    pobj->plastMidEP  = NIL(struct edgePtr);
    pobj->pfirstMinEP = NIL(struct edgePtr);
    pobj->plastMinEP  = NIL(struct edgePtr);

    pobj->pminHEP = NIL(struct edgePtr);
    pobj->pmidHEP = NIL(struct edgePtr);
    pobj->pmaxHEP = NIL(struct edgePtr);

    pobj->pminVEP = NIL(struct edgePtr);
    pobj->pmidVEP = NIL(struct edgePtr);
    pobj->pmaxVEP = NIL(struct edgePtr);

    pobj->pverticalEdges   =  NIL(struct edge);
    pobj->phorizontalEdges = NIL(struct edge);

    pobj->hasPFrames = 'F';
    pobj->pnext = NIL(struct object);
    pobj->node.key = UNDEF;

#ifdef RPC
    pobj->name = STRSAVE(pinst->contents.instance.name);
    pobj->transform.translation.x =
	pinst->contents.instance.transform.translation.x;
    pobj->transform.translation.y =
	pinst->contents.instance.transform.translation.y;
    pobj->transform.transformType =
	pinst->contents.instance.transform.transformType;
    pobj->transform.generalTransform[0][0] =
	pinst->contents.instance.transform.generalTransform[0][0];
    pobj->transform.generalTransform[0][1] =
	pinst->contents.instance.transform.generalTransform[0][1];
    pobj->transform.generalTransform[1][0] =
	pinst->contents.instance.transform.generalTransform[1][0];
    pobj->transform.generalTransform[1][1] =
	pinst->contents.instance.transform.generalTransform[1][1];
#endif

    return(pobj);
}
