#include "copyright.h"
#include "sparcs.h"
#include "tr.h"

/*
 *  XXXXX
 *  Build the initial (i.e., unsorted) edge lists for the scan-line algorithm.
 *  Struct edge and struct edgePtr coordinates are absolute, and this routine
 *  relies on the frame coordinates being absolute.  So, this routine is called
 *  before making the coordinates relative to the origin.
 *
 *  XXX Will need to be modified when treat p-frames as polygons!!!
 *
 *  Called by:	readOct()
 *  Calls:	---
 *
 *  2/14/87
 *  last modified 9/22/88
 */

int
makePFEdges(po, pstack, pbox, sparcsLayer)
struct object *po;
tr_stack *pstack;	/* transform of instance */
struct octBox *pbox;
int sparcsLayer;	/* internal mask-layer number */
{
    struct edge *pleftEdge, *prightEdge;
    struct edge *pbottomEdge, *ptopEdge;
    struct edgePtr *pleftUL, *pleftML, *pleftLL;	/* for left edge */
    struct edgePtr *prightUR, *prightMR, *prightLR;	/* for right edge */
    struct edgePtr *ptopUL, *ptopMT, *ptopUR;		/* for top edge */
    struct edgePtr *pbotLL, *pbotMB, *pbotLR;		/* for bottom edge */
    int lowerLeftX, lowerLeftY;			/* transformed box coords */
    int upperRightX, upperRightY;
    int tempX, tempY;
    int maxRule;

    maxRule = LayerMap[sparcsLayer].maxRule;
    if (maxRule == UNDEF)  {
	errRaise("SPARCS", 1, "makePFEdges:  maxRule value UNDEF for layer %s is illegal.",
		 LayerMap[sparcsLayer].name );
    }

    /*
     *  Transform the coordinates of the box.
     */
    lowerLeftX  = pbox->lowerLeft.x;
    lowerLeftY  = pbox->lowerLeft.y;
    upperRightX = pbox->upperRight.x;
    upperRightY = pbox->upperRight.y;
    tr_transform(pstack, (int32 *) &lowerLeftX, (int32 *) &lowerLeftY);
    tr_transform(pstack, (int32 *) &upperRightX, (int32 *) &upperRightY);
    
    /*
     *  If there has been a rotation/reflection the coordinates must be
     *  reordered.
     */
    if (lowerLeftX > upperRightX)  {
	tempX = upperRightX;
	upperRightX = lowerLeftX;
	lowerLeftX = tempX;
    }
    if (lowerLeftY > upperRightY)  {
	tempY = upperRightY;
	upperRightY = lowerLeftY;
	lowerLeftY = tempY;
    }
    
    /*
     *  Do the update of the bb of the object.
     */
    if (lowerLeftX < po->origin.x)  {
	po->origin.x = lowerLeftX;
    }
    if (lowerLeftY < po->origin.y)  {
	po->origin.y = lowerLeftY;
    }
    if (upperRightX > po->bbUpperR.x)  {
	po->bbUpperR.x = upperRightX;
    }
    if (upperRightY > po->bbUpperR.y)  {
	po->bbUpperR.y = upperRightY;
    }

    /*
     *  Allocate the edges and the edgePtrs.
     */
    pleftEdge = ALLOC(struct edge, 4);
    prightEdge  = pleftEdge + 1;
    ptopEdge    = pleftEdge + 2;
    pbottomEdge = pleftEdge + 3;

    pleftUL = ALLOC(struct edgePtr, 12);
    pleftML  = pleftUL + 1;
    pleftLL  = pleftUL + 2; 
    prightUR = pleftUL + 3;
    prightMR = pleftUL + 4;
    prightLR = pleftUL + 5;
    ptopUL   = pleftUL + 6;
    ptopMT   = pleftUL + 7;
    ptopUR   = pleftUL + 8;
    pbotLL   = pleftUL + 9;
    pbotMB   = pleftUL + 10;
    pbotLR   = pleftUL + 11;

    /*
     *  And fill them in.
     */
    pleftEdge->active = 'N';
    pleftEdge->key = po->key;
    pleftEdge->max = upperRightY;
    pleftEdge->maskNumber = sparcsLayer;
    pleftEdge->loc = lowerLeftX;
    pleftEdge->type = MIN;
    pleftEdge->netId = UNDEF;
    pleftEdge->masked = 'N';
    pleftEdge->pmasks = NIL(struct edge);
    pleftEdge->phardMasks = NIL(struct elist);
    pleftEdge->phardMaskedBy = NIL(struct klist);
    pleftEdge->pnext = NIL(struct edge);
    pleftEdge->pprev = NIL(struct edge);
    pleftEdge->pt = NIL(struct terminal);
    
    prightEdge->active = 'N';
    prightEdge->key = po->key;
    prightEdge->max = upperRightY;
    prightEdge->maskNumber = sparcsLayer;
    prightEdge->loc = upperRightX;
    prightEdge->type = MAX;
    prightEdge->netId = UNDEF;
    prightEdge->masked = 'N';
    prightEdge->pmasks = NIL(struct edge);
    prightEdge->phardMasks = NIL(struct elist);
    prightEdge->phardMaskedBy = NIL(struct klist);
    prightEdge->pnext = NIL(struct edge);
    prightEdge->pprev = NIL(struct edge);
    prightEdge->pt = NIL(struct terminal);
    
    /*
     *  The edges are linked to the object for updating the loc field
     *  after a spacing.
     */
    pleftEdge->plist = prightEdge;
    prightEdge->plist = po->pverticalEdges;
    po->pverticalEdges = pleftEdge;
    
    /*
     *  Create the edgePtrs.  Note that the MAX end of the edges are
     *  extended by maxRule, but NOT the MIN ends.
     *
     *  XXX  6/25/87
     *  P-frames get 3 edgePtrs now, a MID at the top of the edge as
     *  well as the MIN and MAX from before.
     */
    pleftUL->coord = upperRightY + maxRule;
    pleftUL->pedge = pleftEdge;
    
    prightUR->coord = upperRightY + maxRule;
    prightUR->pedge = prightEdge;
    
    pleftML->coord = upperRightY;
    pleftML->pedge = pleftEdge;
    
    prightMR->coord = upperRightY;
    prightMR->pedge = prightEdge;
    
    pleftLL->coord = lowerLeftY;
    pleftLL->pedge = pleftEdge;
    
    prightLR->coord = lowerLeftY;
    prightLR->pedge = prightEdge;
    
    /*
     *  And link them.
     */
    pleftUL->plist  = prightUR;
    prightUR->plist = po->pmaxHEP;
    po->pmaxHEP = pleftUL;
    
    pleftML->plist  = prightMR;
    prightMR->plist = po->pmidHEP;
    po->pmidHEP = pleftML;
    
    pleftLL->plist  = prightLR;
    prightLR->plist = po->pminHEP;
    po->pminHEP = pleftLL;
    
    
    /*
     *  Repeat for the top and bottom edges.
     */
    ptopEdge->active = 'N';
    ptopEdge->key = po->key;
    ptopEdge->max = upperRightX;
    ptopEdge->maskNumber = sparcsLayer;
    ptopEdge->loc = upperRightY;
    ptopEdge->type = MAX;
    ptopEdge->netId = UNDEF;
    ptopEdge->masked = 'N';
    ptopEdge->pmasks = NIL(struct edge);
    ptopEdge->phardMasks = NIL(struct elist);
    ptopEdge->phardMaskedBy = NIL(struct klist);
    ptopEdge->pnext = NIL(struct edge);
    ptopEdge->pprev = NIL(struct edge);
    ptopEdge->pt = NIL(struct terminal);
    
    pbottomEdge->active = 'N';
    pbottomEdge->key = po->key;
    pbottomEdge->max = upperRightX;
    pbottomEdge->maskNumber = sparcsLayer;
    pbottomEdge->loc = lowerLeftY;
    pbottomEdge->type = MIN;
    pbottomEdge->netId = UNDEF;
    pbottomEdge->masked = 'N';
    pbottomEdge->pmasks = NIL(struct edge);
    pbottomEdge->phardMasks = NIL(struct elist);
    pbottomEdge->phardMaskedBy = NIL(struct klist);
    pbottomEdge->pnext = NIL(struct edge);
    pbottomEdge->pprev = NIL(struct edge);
    pbottomEdge->pt = NIL(struct terminal);
    
    ptopEdge->plist = pbottomEdge;
    pbottomEdge->plist = po->phorizontalEdges;
    po->phorizontalEdges = ptopEdge;
    
    ptopUR->coord = upperRightX + maxRule;
    ptopUR->pedge = ptopEdge;
    
    pbotLR->coord = upperRightX + maxRule;
    pbotLR->pedge = pbottomEdge;
    
    ptopMT->coord = upperRightX;
    ptopMT->pedge = ptopEdge;
    
    pbotMB->coord = upperRightX;
    pbotMB->pedge = pbottomEdge;
    
    ptopUL->coord = lowerLeftX;
    ptopUL->pedge = ptopEdge;
    
    pbotLL->coord = lowerLeftX;
    pbotLL->pedge = pbottomEdge;
    
    ptopUR->plist  = pbotLR;
    pbotLR->plist = po->pmaxVEP;
    po->pmaxVEP = ptopUR;
    
    ptopMT->plist  = pbotMB;
    pbotMB->plist = po->pmidVEP;
    po->pmidVEP = ptopMT;
    
    ptopUL->plist  = pbotLL;
    pbotLL->plist = po->pminVEP;
    po->pminVEP = ptopUL;
}
