#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "tr.h"

/*
 *  Allocate a terminal struct and fill it in.  Input is the object, which is
 *  returned as output with the new terminal attached.
 *
 *  NOTE:  The coordinates of bb are absolute in this routine.  They are
 *	   adjusted to be relative later.
 *
 *  renamed 3/18/87 (from SPoctMakeTerminal())
 *
 *  A terminal struct is:
 *		struct terminal  {
 *		    struct rect bb;
 *		    int maskNumber;
 *		    int atermId;
 *		    struct edge *plE;
 *		    struct edge *prE;
 *		    struct edge *ptE;
 *		    struct edge *pbE;
 *		    struct terminal *pnext;
 *		};
 *
 *  Called By:	procTFrames()
 *  Calls:	-
 *
 *  1/24/86
 *  last modified 9/22/88
 */

void
makeTerminal(po, pstack, pbox, atermId, sparcsLayer)
struct object *po;		/*  object that owns terminal  */
tr_stack *pstack;		/*  transformation  */
struct octBox *pbox;		/*  oct terminal frame geo  */
int atermId;			/* octId of aterm this frame implements */
int sparcsLayer;		/*  internal layer number  */
{
    struct terminal *pt;
    int tempX, tempY;
    struct edge *pleftEdge, *prightEdge;
    struct edge *pbottomEdge, *ptopEdge;
    struct edgePtr *pleftUL, *pleftML, *pleftLL;	/* for left edge */
    struct edgePtr *prightUR, *prightMR, *prightLR;	/* for right edge */
    struct edgePtr *ptopUL, *ptopMT, *ptopUR;		/* for top edge */
    struct edgePtr *pbotLL, *pbotMB, *pbotLR;		/* for bottom edge */
    int maxRule;

    pt = ALLOC(struct terminal, 1);

    /*
     *  Fill in the terminal area, with the transform.  Note that the frame
     *  type has been checked in the caller (procTFrames()).
     */
    pt->bb.lowerL.x = pbox->lowerLeft.x;
    pt->bb.lowerL.y = pbox->lowerLeft.y;
    pt->bb.upperR.x = pbox->upperRight.x;
    pt->bb.upperR.y = pbox->upperRight.y;
    tr_transform(pstack, (int32 *) &(pt->bb.lowerL.x), (int32 *) &(pt->bb.lowerL.y));
    tr_transform(pstack, (int32 *) &(pt->bb.upperR.x), (int32 *) &(pt->bb.upperR.y));
    /*
     *  If there has been a rotation/reflection the coordinates must be
     *  reordered.
     */
    if (pt->bb.lowerL.x > pt->bb.upperR.x)  {
	tempX = pt->bb.upperR.x;
	pt->bb.upperR.x = pt->bb.lowerL.x;
	pt->bb.lowerL.x = tempX;
    }
    if (pt->bb.lowerL.y > pt->bb.upperR.y)  {
	tempY = pt->bb.upperR.y;
	pt->bb.upperR.y = pt->bb.lowerL.y;
	pt->bb.lowerL.y = tempY;
    }
    /*
     *  Update the bb of the object
     */
    if (pt->bb.lowerL.x < po->origin.x)  {
	po->origin.x = pt->bb.lowerL.x;
    }
    if (pt->bb.lowerL.y < po->origin.y)  {
	po->origin.y = pt->bb.lowerL.y;
    }
    if (pt->bb.upperR.x > po->bbUpperR.x)  {
	po->bbUpperR.x = pt->bb.upperR.x;
    }
    if (pt->bb.upperR.y > po->bbUpperR.y)  {
	po->bbUpperR.y = pt->bb.upperR.y;
    }

    pt->maskNumber = sparcsLayer;
    pt->atermId = atermId;

    /**
    pt->plE = NIL(struct edge);
    pt->prE = NIL(struct edge);
    pt->ptE = NIL(struct edge);
    pt->pbE = NIL(struct edge);
    **/
    
    /*  link this terminal into the terminal list for the object  */
    pt->pnext = po->pterminal;
    po->pterminal = pt;



    /*
     *  10/13/87
     *  The remainder of this routine used to be part of mkEdgeList().
     */

    maxRule = LayerMap[pt->maskNumber].maxRule;
    if (maxRule == UNDEF)  {
	errRaise( "SPARCS", 1, "makeTFEdges:  maxRule value UNDEF for layer %s is illegal.\n",
		 LayerMap[pt->maskNumber].name );
    }

    /*
     *  Allocate the edges and the edgePtrs.
     */
    pleftEdge = ALLOC(struct edge, 4);
    prightEdge  = pleftEdge + 1;
    ptopEdge    = pleftEdge + 2;
    pbottomEdge = pleftEdge + 3;

    /*
     *  Allocate mid edgePtrs iff there are p-frames associated with this
     *  instance.
     *
     *  XXX  Could avoid an "if" provided that code is re-arranged a bit...
     */
    if (po->hasPFrames == 'T')  {
	pleftUL = ALLOC(struct edgePtr, 12);
	pleftML  = pleftUL + 1;
	pleftLL  = pleftUL + 2; 
	prightUR = pleftUL + 3;
	prightMR = pleftUL + 4;
	prightLR = pleftUL + 5;
	ptopUL   = pleftUL + 6;
	ptopMT   = pleftUL + 7;
	ptopUR   = pleftUL + 8;
	pbotLL   = pleftUL + 9;
	pbotMB   = pleftUL + 10;
	pbotLR   = pleftUL + 11;
    } else  {
	pleftUL = ALLOC(struct edgePtr, 8);
        /**
	pleftLL  = pleftUL + 2; 
	prightUR = pleftUL + 3;
	prightLR = pleftUL + 4;
	ptopUL   = pleftUL + 5;
	ptopUR   = pleftUL + 6;
	pbotLL   = pleftUL + 7;
	pbotLR   = pleftUL + 8;
        **/
	pleftLL  = pleftUL + 1; 
	prightUR = pleftUL + 2;
	prightLR = pleftUL + 3;
	ptopUL   = pleftUL + 4;
	ptopUR   = pleftUL + 5;
	pbotLL   = pleftUL + 6;
	pbotLR   = pleftUL + 7;
    }
    

    /*
     *  First the left-hand edge, then the right-hand.
     */
    pleftEdge->active = 'N';
    pleftEdge->key = po->key;
    pleftEdge->max = pt->bb.upperR.y;
    pleftEdge->maskNumber = pt->maskNumber;
    pleftEdge->loc = pt->bb.lowerL.x;
    pleftEdge->type = MIN;
    pleftEdge->netId = UNDEF;
    pleftEdge->masked = 'N';
    pleftEdge->pmasks = NIL(struct edge);
    pleftEdge->phardMasks = NIL(struct elist);
    pleftEdge->phardMaskedBy = NIL(struct klist);
    pleftEdge->pnext = NIL(struct edge);
    pleftEdge->pprev = NIL(struct edge);
    pleftEdge->pt = NIL(struct terminal);
    
    prightEdge->active = 'N';
    prightEdge->key = po->key;
    prightEdge->max = pt->bb.upperR.y;
    prightEdge->maskNumber = pt->maskNumber;
    prightEdge->loc = pt->bb.upperR.x;
    prightEdge->type = MAX;
    prightEdge->netId = UNDEF;
    prightEdge->masked = 'N';
    prightEdge->pmasks = NIL(struct edge);
    prightEdge->phardMasks = NIL(struct elist);
    prightEdge->phardMaskedBy = NIL(struct klist);
    prightEdge->pnext = NIL(struct edge);
    prightEdge->pprev = NIL(struct edge);
    prightEdge->pt = NIL(struct terminal);
    
    pt->plE = pleftEdge;
    pt->prE = prightEdge;
    
    /*
     *  The edges are linked to the object for updating the loc field
     *  after a spacing.
     */
    pleftEdge->plist = prightEdge;
    prightEdge->plist = po->pverticalEdges;
    po->pverticalEdges = pleftEdge;
    
    /*
     *  Create and link the edgePtrs.
     */
    pleftUL->coord = pt->bb.upperR.y + maxRule;
    pleftUL->pedge = pleftEdge;
    
    prightUR->coord = pt->bb.upperR.y + maxRule;
    prightUR->pedge = prightEdge;
    
    pleftLL->coord = pt->bb.lowerL.y;
    pleftLL->pedge = pleftEdge;
    
    prightLR->coord = pt->bb.lowerL.y;
    prightLR->pedge = prightEdge;
    
    pleftUL->plist  = prightUR;
    prightUR->plist = po->pmaxHEP;
    po->pmaxHEP = pleftUL;
    
    pleftLL->plist  = prightLR;
    prightLR->plist = po->pminHEP;
    po->pminHEP = pleftLL;
    
    
    /*
     *  Repeat for the top and bottom edges.
     */
    ptopEdge->active = 'N';
    ptopEdge->key = po->key;
    ptopEdge->max = pt->bb.upperR.x;
    ptopEdge->maskNumber = pt->maskNumber;
    ptopEdge->loc = pt->bb.upperR.y;
    ptopEdge->type = MAX;
    ptopEdge->netId = UNDEF;
    ptopEdge->masked = 'N';
    ptopEdge->pmasks = NIL(struct edge);
    ptopEdge->phardMasks = NIL(struct elist);
    ptopEdge->phardMaskedBy = NIL(struct klist);
    ptopEdge->pnext = NIL(struct edge);
    ptopEdge->pprev = NIL(struct edge);
    ptopEdge->pt = NIL(struct terminal);
    
    pbottomEdge->active = 'N';
    pbottomEdge->key = po->key;
    pbottomEdge->max = pt->bb.upperR.x;
    pbottomEdge->maskNumber = pt->maskNumber;
    pbottomEdge->loc = pt->bb.lowerL.y;
    pbottomEdge->type = MIN;
    pbottomEdge->netId = UNDEF;
    pbottomEdge->masked = 'N';
    pbottomEdge->pmasks = NIL(struct edge);
    pbottomEdge->phardMasks = NIL(struct elist);
    pbottomEdge->phardMaskedBy = NIL(struct klist);
    pbottomEdge->pnext = NIL(struct edge);
    pbottomEdge->pprev = NIL(struct edge);
    pbottomEdge->pt = NIL(struct terminal);
    
    pt->ptE = ptopEdge;
    pt->pbE = pbottomEdge;
    
    ptopEdge->plist = pbottomEdge;
    pbottomEdge->plist = po->phorizontalEdges;
    po->phorizontalEdges = ptopEdge;
    
    ptopUR->coord = pt->bb.upperR.x + maxRule;
    ptopUR->pedge = ptopEdge;
    
    pbotLR->coord = pt->bb.upperR.x + maxRule;
    pbotLR->pedge = pbottomEdge;
    
    ptopUL->coord = pt->bb.lowerL.x;
    ptopUL->pedge = ptopEdge;
    
    pbotLL->coord = pt->bb.lowerL.x;
    pbotLL->pedge = pbottomEdge;
    
    ptopUR->plist  = pbotLR;
    pbotLR->plist = po->pmaxVEP;
    po->pmaxVEP = ptopUR;
    
    ptopUL->plist  = pbotLL;
    pbotLL->plist = po->pminVEP;
    po->pminVEP = ptopUL;

    /*
     *  Middle edgePtrs IFF the instance has p-frames.  If it doesn't,
     *  then it can't do masking and middle edgePtrs are not needed.
     */
    if (po->hasPFrames == 'T')  {
	pleftML->coord = pt->bb.upperR.y;
	pleftML->pedge = pleftEdge;
	
	prightMR->coord = pt->bb.upperR.y;
	prightMR->pedge = prightEdge;
	
	pleftML->plist  = prightMR;
	prightMR->plist = po->pmidHEP;
	po->pmidHEP = pleftML;
	
	ptopMT->coord = pt->bb.upperR.x;
	ptopMT->pedge = ptopEdge;
	
	pbotMB->coord = pt->bb.upperR.x;
	pbotMB->pedge = pbottomEdge;
	
	ptopMT->plist  = pbotMB;
	pbotMB->plist = po->pmidVEP;
	po->pmidVEP = ptopMT;
    }
}
