#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

/*
 *  Allocate a wire struct and fill it in.  Then, link it to the list of
 *  wires.  All wires are single-segment.  Point p1 is the lesser endpoint,
 *  and point p2 is the greater.
 *
 *  A wire struct is:
 *		struct wire  {
 *		    int key;
 *		    int pathId;
 *		    int changed;
 *		    int horizontal;
 *		    struct point p1;	    (the lesser endpoint)
 *		    struct point p2;	    (the greater endpoint)
 *		    struct object *po1;
 *		    struct object *po2;
 *		    struct terminal *pt1;
 *		    struct terminal *pt2;
 *		    int width;
 *		    int halfWidth;
 *		    int force;
 *		    int maskNumber;
 *		    struct edgePtr *pfirstMaxEP;
 *		    struct edgePtr *plastMaxEP;
 *		    struct edgePtr *pfirstMinEP;
 *		    struct edgePtr *plastMinEP;
 *		    struct wire *pnext;
 *		    struct node node;
 *		};
 *
 *  Called By:	readOct()
 *  Calls:	-
 *
 *  1/27/86
 *  renamed from SPoctMakeWire 2/26/87
 *  last modified 9/22/88
 */

struct wire *
makeWire(psegment, points, direction, sparcsLayer)
octObject *psegment;		/* object representing segment */
struct octPoint points[];
directionType direction;
int sparcsLayer;		/* internal layer number */
{
    struct wire *pw;

    /*  allocate the wire  */
    pw = ALLOC(struct wire, 1);

    /*  and fill it in  */
    pw->key = ElementKey++;
    pw->pathId = psegment->objectId;
    pw->changed = FALSE;

    /*
     *  Checks for nonzero length, non-Manhattan already performed in readOct()
     */
    if (direction == HORIZONTAL)  {
	pw->horizontal = TRUE;
	pw->p1.y = pw->p2.y = points[1].y;
	if (points[0].x <= points[1].x)  {
	    pw->p1.x = points[0].x;
	    pw->p2.x = points[1].x;
	} else  {
	    pw->p1.x = points[1].x;
	    pw->p2.x = points[0].x;
	}
    } else  {
	pw->horizontal = FALSE;
	pw->p1.x = pw->p2.x = points[1].x;
	if (points[0].y <= points[1].y)  {
	    pw->p1.y = points[0].y;
	    pw->p2.y = points[1].y;
	} else  {
	    pw->p1.y = points[1].y;
	    pw->p2.y = points[0].y;
	}
    }
    

    /*
     *  Following 4 pointers set in findTFrame()
     */
    pw->po1 = NIL(struct object);
    pw->po2 = NIL(struct object);
    pw->pt1 = NIL(struct terminal);
    pw->pt2 = NIL(struct terminal);

    pw->width = psegment->contents.path.width;
    if ((pw->width % 2) != 0)  {
	printf("makeWire:  warning - segment width is an odd number.\n");
    }
    pw->halfWidth = (pw->width)/2;

    pw->force = LayerMap[sparcsLayer].ohmsPerSquare;

    pw->maskNumber = sparcsLayer;

    /*
     *  Following 4 pointers set in makeWireEdges(), after trimWire()
     */
    pw->pfirstMaxEP = NIL(struct edgePtr);
    pw->plastMaxEP  = NIL(struct edgePtr);
    pw->pfirstMinEP = NIL(struct edgePtr);
    pw->plastMinEP  = NIL(struct edgePtr);

    pw->pnext = pWireList;
    pWireList = pw;

    pw->node.key = UNDEF;

    return(pw);
}
