#include "copyright.h"
#include "sparcs.h"

/*
 *  Create the edges and edgePtrs for a segment.  It is assumed that the wire
 *  has already been trimmed via trimWire().
 *
 *  Called By:	readOct()
 *  Calls:	-
 *
 *  11/4/87
 *  last modified 9/22/88
 */

void
makeWireEdges(pw, netId)
struct wire *pw;
int netId;
{
    struct edge *pleftEdge, *prightEdge;
    struct edge *pbottomEdge, *ptopEdge;
    struct edgePtr *plowerLeft, *plowerRight;
    struct edgePtr *pupperLeft, *pupperRight;

    if (pw->horizontal == TRUE)  {
	/*
	 *  The wire is horizontal.
	 */
	pbottomEdge = ALLOC(struct edge, 2);
	ptopEdge    = pbottomEdge + 1;

	pupperLeft = ALLOC(struct edgePtr, 4);
	plowerLeft  = pupperLeft + 1;
	pupperRight = pupperLeft + 2;
	plowerRight = pupperLeft + 3;

	pbottomEdge->active = 'N';
	pbottomEdge->key = pw->key;
	pbottomEdge->max = pw->p2.x;
	pbottomEdge->maskNumber = pw->maskNumber;
	pbottomEdge->loc = pw->p1.y - (pw->width)/2;
	pbottomEdge->type = MIN;
	pbottomEdge->netId = netId;
	pbottomEdge->masked = 'N';
	pbottomEdge->pmasks = NIL(struct edge);
	pbottomEdge->phardMasks = NIL(struct elist);
	pbottomEdge->phardMaskedBy = NIL(struct klist);
	pbottomEdge->pnext = NIL(struct edge);
	pbottomEdge->pprev = NIL(struct edge);
	pbottomEdge->pt = NIL(struct terminal);
	
	ptopEdge->active = 'N';
	ptopEdge->key = pw->key;
	ptopEdge->max = pw->p2.x;
	ptopEdge->maskNumber = pw->maskNumber;
	ptopEdge->loc = pw->p1.y + (pw->width)/2;
	ptopEdge->type = MAX;
	ptopEdge->netId = netId;
	ptopEdge->masked = 'N';
	ptopEdge->pmasks = NIL(struct edge);
	ptopEdge->phardMasks = NIL(struct elist);
	ptopEdge->phardMaskedBy = NIL(struct klist);
	ptopEdge->pnext = NIL(struct edge);
	ptopEdge->pprev = NIL(struct edge);
	ptopEdge->pt = NIL(struct terminal);
	
	/*
	 *  Note that wires are NOT extnded.
	 */
	pupperLeft->coord = pw->p1.x;
	pupperLeft->pedge = ptopEdge;
	
	plowerLeft->coord = pw->p1.x;
	plowerLeft->pedge = pbottomEdge;
	
	pupperLeft->pnext  = plowerLeft;
	if (pw->pfirstMinEP == NIL(struct edgePtr))  {
	    pw->plastMinEP = plowerLeft;
	} else  {
    printf("makeWire - fucked up wire!\n");  exit(-1);
	    plowerLeft->pnext = pw->pfirstMinEP;
	}
	pw->pfirstMinEP = pupperLeft;
	
	pupperRight->coord = pw->p2.x;
	pupperRight->pedge = ptopEdge;
	
	plowerRight->coord = pw->p2.x;
	plowerRight->pedge = pbottomEdge;
	
	pupperRight->pnext  = plowerRight;
	if (pw->pfirstMaxEP == NIL(struct edgePtr))  {
	    pw->plastMaxEP = plowerRight;
	} else  {
    printf("makeWire - fucked up wire!\n");  exit(-1);
	    plowerRight->pnext = pw->pfirstMaxEP;
	}
	pw->pfirstMaxEP = pupperRight;
    } else  {
	/*
	 *  The wire is vertical.
	 */
	pleftEdge = ALLOC(struct edge, 2);
	prightEdge = pleftEdge + 1;

	pupperLeft = ALLOC(struct edgePtr, 4);
	plowerLeft  = pupperLeft + 1;
	pupperRight = pupperLeft + 2;
	plowerRight = pupperLeft + 3;

	pleftEdge->active = 'N';
	pleftEdge->key = pw->key;
	pleftEdge->max = pw->p2.y;
	pleftEdge->maskNumber = pw->maskNumber;
	pleftEdge->loc = pw->p1.x - (pw->width)/2;
	pleftEdge->type = MIN;
	pleftEdge->netId = netId;
	pleftEdge->masked = 'N';
	pleftEdge->pmasks = NIL(struct edge);
	pleftEdge->phardMasks = NIL(struct elist);
	pleftEdge->phardMaskedBy = NIL(struct klist);
	pleftEdge->pnext = NIL(struct edge);
	pleftEdge->pprev = NIL(struct edge);
	pleftEdge->pt = NIL(struct terminal);
	
	prightEdge->active = 'N';
	prightEdge->key = pw->key;
	prightEdge->max = pw->p2.y;
	prightEdge->maskNumber = pw->maskNumber;
	prightEdge->loc = pw->p1.x + (pw->width)/2;
	prightEdge->type = MAX;
	prightEdge->netId = netId;
	prightEdge->masked = 'N';
	prightEdge->pmasks = NIL(struct edge);
	prightEdge->phardMasks = NIL(struct elist);
	prightEdge->phardMaskedBy = NIL(struct klist);
	prightEdge->pnext = NIL(struct edge);
	prightEdge->pprev = NIL(struct edge);
	prightEdge->pt = NIL(struct terminal);
	
	pupperLeft->coord = pw->p2.y;
	pupperLeft->pedge = pleftEdge;
	
	pupperRight->coord = pw->p2.y;
	pupperRight->pedge = prightEdge;
	
	pupperLeft->pnext  = pupperRight;
	if (pw->pfirstMaxEP == NIL(struct edgePtr))  {
	    pw->plastMaxEP = pupperRight;
	} else  {
    printf("makeWire - fucked up wire!\n");  exit(-1);
	    pupperRight->pnext = pw->pfirstMaxEP;
	}
	pw->pfirstMaxEP = pupperLeft;
	
	plowerLeft->coord = pw->p1.y;
	plowerLeft->pedge = pleftEdge;
	
	plowerRight->coord = pw->p1.y;
	plowerRight->pedge = prightEdge;
	
	plowerLeft->pnext  = plowerRight;
	if (pw->pfirstMinEP == NIL(struct edgePtr))  {
	    pw->plastMinEP = plowerRight;
	} else  {
    printf("makeWire - fucked up wire!\n");  exit(-1);
	    plowerRight->pnext = pw->pfirstMinEP;
	}
	pw->pfirstMinEP = plowerLeft;
    }
}
