#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "da.h"
#include "tr.h"
#include "fang.h"
#include "harpoon.h"

#ifdef RPC
#define octInitGenContents octInitGenContentsSpecial
#define octInitGenContainers octInitGenContainersSpecial
#endif

/*
 *  This routine takes a layer of an instance master and generates all of the
 *  protection frames on it.  Each p-frame is added to the sparcs object.  If
 *  OCT_POLYGONs are found, they are boxed using harpoon and fang.
 *
 *  Called by:  readOct()
 *  Calls:
 *
 *  9/19/86
 *  last modified 1/20/88
 */

void
procPFrames(player, pobj, pstack, pnumFrames)
octObject *player;	/* layer to be read for p-frames */
struct object *pobj;
tr_stack *pstack;	/* instance's transform */
int *pnumFrames;	/* incremented */
{
    octObject frame, term;
    struct octBox box;
    octGenerator frameGenId, termGenId;
    octStatus frameGenStatus, termGenStatus;
    int sparcsLayer;			/* internal layer number */
    int numTerms;
    int i;
    fa_geometry geo;
    fa_box *pbox, *pfirstBox;
    char buffer[SMALLBUF];

    term.type = OCT_TERM;

    /*
     *  Get the internal layer number out of LayerMap.
     */
    for (i=0; i<NumLayers; i++)  {
	if (strcmp(LayerMap[i].name, player->contents.layer.name) == 0)  {
	    sparcsLayer = i;
	    break;
	}
    }
    if (i >= NumLayers)  {
	/*
	 *  Layer isn't in rules file.  If it is in the list of layers to
	 *  suppress warnings for, then fine.  If it is not in that list,
	 *  and if it contains frames, then complain.  In in any case return.
	 */
	if ( noWarningTest(  player->contents.layer.name) ) return;
	
	if ((frameGenStatus=octGenFirstContent(player, OCT_GEO_MASK,
					       &frame)) == OCT_GEN_DONE)  {
	    ;
	} else if (frameGenStatus == OCT_OK)  {
	    printf("procPFrames:  layer %s ", player->contents.layer.name);
	    printf("contains geos but isn't in rules file; skipping.\n");
	} else  {
	    sprintf(buffer, "procPFrames:  trying to genFirst geo on %s",
		    player->contents.layer.name);
	    octError(buffer);
	}
	return;
    }

    /*
     *  Generate the protection frames on this layer.  To distinguish t-frames
     *  from p-frames, run a containers generator on the frame to look for a
     *  terminal definition.
     */
    if (octInitGenContents(player, OCT_GEO_MASK, &frameGenId) != OCT_OK)  {
	octError("procPFrames:  can't begin p-frame generator");
	exit(-1);
    }
    while ((frameGenStatus=octGenerate(&frameGenId, &frame)) == OCT_OK)  {
	/*
	 *  XXX  SHOULD check to insure geo is Manhattan?
	 */
	switch (frame.type)  {
	    case OCT_BOX:
		if (octInitGenContainers(&frame, OCT_TERM_MASK, &termGenId) != OCT_OK)  {
		    octError("procPFrames:  can't begin terminal generator.");
		    exit(-1);
		}
		numTerms = 0;
		while ((termGenStatus=octGenerate(&termGenId, &term)) == OCT_OK)  {
		    numTerms++;
		}
		/*
		 *  Legal values for numTerms are 0 (pf) or 1 (tf).
		 */
		if (numTerms == 0)  {
		    (*pnumFrames)++;
		    makePFEdges(pobj, pstack, &(frame.contents.box), sparcsLayer);
		} else if (numTerms > 1)  {
		    printf("procPFrames:  frame contained by > 1 f-term.\n");
		    exit(-1);
		}
		break;
		    
	    case OCT_POLYGON:
		/*
		 *  Since polygon MUST be a p-frame.  Make the polygon into
		 *  a list of boxes, by putting it into a fa_geometry via
		 *  harpoon and then calling fa_to_boxes.
		 */
		fa_init(&geo);
		harAddObjectToFang(&frame, &geo, HAR_NO_TRANSFORM);
		fa_to_boxes(&geo, &pbox);

		/*
		 *  Now, make a sparcs p-frame for each box in the list.
		 */
		pfirstBox = pbox;
		for (; pbox!=NIL(struct fa_box); pbox=pbox->next)  {
		    (*pnumFrames)++;
		    box.lowerLeft.x = pbox->left;
		    box.lowerLeft.y = pbox->bottom;
		    box.upperRight.x = pbox->right;
		    box.upperRight.y = pbox->top;
		    makePFEdges(pobj, pstack, &box, sparcsLayer);
		}
		fa_free_boxes(pfirstBox);
		break;

	    default:
		printf("procPFrames:  geo isn't box or polygon; skipping.\n");
		break;
	}
    }
    if (frameGenStatus != OCT_GEN_DONE)  {
	octError("procPFrames:  Bad exit status from frame generator.");
	exit(-1);
    }
}



static dary NoWarningLayers = 0;

noWarningAdd( layer )
    char* layer;
{
    if ( NoWarningLayers == 0 ) {
	NoWarningLayers = daAlloc( char*, 10 );
    }
    if ( noWarningTest( layer ) == 0 ) {
	*daSetLast( char*, NoWarningLayers ) = STRSAVE( layer );
    }
}

int noWarningTest( layer )
    char* layer;
{
    if ( NoWarningLayers ) {
	int i;
	for ( i = 0; i < daLen( NoWarningLayers ) ; i++ ) {
	    char* l = *daGet(char*,NoWarningLayers, i );
	    if ( strcmp( layer, l ) == 0 ) return 1; /* Yes, layer is in array. */
	}
    }
    return 0;
}

