#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "tr.h"

/*
 *  This routine takes a formal terminal of an instance master (interface
 *  facet) and generates all of the terminal frames on it.  Each t-frame
 *  is added to the sparcs object.  Terminal frames must be OCT_BOXes.
 *
 *  XXX Make return error code to readOct so that can print all errors in
 *	input before quitting.
 *  XXX Eliminate the call to makeTerminal() by including that code in this
 *	routine.
 *
 *  Called by:  readOct()
 *  Calls:
 *
 *  5/16/87
 *  last modified 11/9/87
 */

void
readTFrames(pfterm, pobj, pinst, pstack)
octObject *pfterm;	/* formal terminal to be read for p-frames */
struct object *pobj;
octObject *pinst;	/* instance impl. by curent master, to get aterms */
tr_stack *pstack;	/* instance's transform */
{
    octObject aterm;
    octObject frame;
    octGenerator frameGenId;
    octStatus frameGenStatus;
    octObject layer;
    octGenerator layerGenId;
    octStatus layerGenStatus;
    void makeTerminal();
    int sparcsLayer;			/* internal layer number */
    int i;
    int count;

    aterm.type = OCT_TERM;
    layer.type = OCT_LAYER;

    /*
     *  Get, from the instance, the aterm that corresponds to this fterm.
     */
    aterm.contents.term.name = pfterm->contents.term.name;
    if (octGetByName(pinst, &aterm) != OCT_OK)  {
	octError("readTFrames:  can't getByName aterm.\n");
	exit(-1);
    }

    /*
     *  Generate the terminal frames of pfterm.
     */
    if (octInitGenContents(pfterm, OCT_GEO_MASK, &frameGenId) != OCT_OK)  {
	octError("readTFrames:  can't begin t-frame generator");
	exit(-1);
    }
    while ((frameGenStatus=octGenerate(&frameGenId, &frame)) == OCT_OK)  {
	if (frame.type != OCT_BOX)  {
	    printf("readTFrames:  terminal isn't a box; skipping.\n");
	    continue;
	}

	/*
	 *  Run a containers generator on the frame to find its layer.
	 */
	if (octInitGenContainers(&frame,OCT_LAYER_MASK,&layerGenId)!=OCT_OK)  {
	    octError("readTFrames:  can't begin t-frame layer gen.");
	    exit(-1);
	}
	count = 0;
	while ((layerGenStatus=octGenerate(&layerGenId, &layer)) == OCT_OK)  {
	    count++;
	}
	if (layerGenStatus != OCT_GEN_DONE)  {
	    octError("readTFrames:  can't end t-frame layer gen.");
	    exit(-1);
	}
	if (count != 1)  {
	    if (count == 0)  {
	        printf("readTFrames:  t-frame not on a layer; skipping.\n");
	    } else  {
	        printf("readTFrames:  t-frame on %d layers; skipping.\n",
			count);
	    }
	    continue;
	}

	/*
	 *  Get the internal layer number out of LayerMap.
	 */
	for (i=0; i<NumLayers; i++)  {
	    if (strcmp(LayerMap[i].name, layer.contents.layer.name) == 0)  {
		sparcsLayer = i;
		break;
	    }
	}
	if (i >= NumLayers)  {
	    /*
	     *  Layer isn't in rules file, so complain.
	     */
#ifdef notdef
	    printf("readTFrames:  f-term is on layer %s ",
		    layer.contents.layer.name);
	    printf("which isn't in rules file; skipping.\n");
#endif
	    continue;
	}

	makeTerminal(pobj, pstack, &(frame.contents.box),
		     aterm.objectId, sparcsLayer);
    }
    if (frameGenStatus != OCT_GEN_DONE)  {
	octError("readTFrames:  Bad exit status from t-frame generator.");
	exit(-1);
    }
}
