#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "facet.h"

/*
 *  Read any user-supplied constraints.
 *
 *  11/2/86
 *  last modified 9/22/88
 */

void
readUserConsts()
{
    octObject firstBag;
    octObject secondBag;
    octObject constBag;
    int min, max, status;

    struct object *po1;
    struct node *pfirst, *psecond, *pn;
    
    struct userConstraint *pc;
    
    octStatus oStatus, bagGenStat2;
    octGenerator bagGenId1, bagGenId2;

    int readConst(), readBorderConst(), readBorderActiveConst();

    firstBag.type = OCT_BAG;
    secondBag.type = OCT_BAG;
    constBag.type = OCT_BAG;

    /*
     *  GetByName the XCONSTRAINTS bag.
     */
    firstBag.contents.bag.name = "XCONSTRAINTS";
    if ((oStatus=octGetByName(&WorkingFacet, &firstBag)) == OCT_NOT_FOUND)  {
	;
	/* printf("\nNo X user-constraints.\n"); */
    } else if (oStatus != OCT_OK)  {
	octError("can't getByName top-level x-constraints bag.");
	exit(-1);
    } else  {
	/*
	 *  Now look for a bag called "LEFT" or one called "RIGHT".
	 */
	if (octInitGenContents(&firstBag,OCT_BAG_MASK,&bagGenId1) != OCT_OK)  {
	    octError("readUserConsts:  can't begin constraint bag generator.");
	    exit(-1);
	}
	while (octGenerate(&bagGenId1, &secondBag) == OCT_OK)  {
    	    if (strcmp(secondBag.contents.bag.name, "LEFT") == 0)  {
		/*
		 *  Should be more bags, called "CONSTRAINT" or "ACONSTRAINT".
		 */
		if (octInitGenContents(&secondBag, OCT_BAG_MASK, &bagGenId2)
								!= OCT_OK)  {
		    octError("readUserConsts:  can't begin 2nd bag gen.");
		    exit(-1);
		}
		while ((bagGenStat2=octGenerate(&bagGenId2, &constBag))
								== OCT_OK)  {
		    if (strcmp("CONSTRAINT",
			       constBag.contents.bag.name) == 0)  {
			status = readBorderConst(&constBag, &pn,
						 &min, &max, &po1);
			if (status != OK)  continue;
			
			/*
			 *  Add to the list.
			 */
			pc = ALLOC(struct userConstraint, 1);
			pc->pfromNode = pSource;
			pc->ptoNode = pn;
			pc->min = min;
			pc->max = max;
			pc->pnext = pXConsts;
			pXConsts = pc;
		    } else  {
			printf("readUserConst:  bag named \"%s\" ",
			       constBag.contents.bag.name);
			printf("is not understood.\n");
			continue;
		    }
		}
		if (bagGenStat2 != OCT_GEN_DONE)  {
		    printf("readUserConst:  can't end 2nd bag gen.\n");
		    exit(-1);
		}
	    } else if (strcmp(secondBag.contents.bag.name, "RIGHT") == 0)  {
		/*
		 *  Should be more bags, called "CONSTRAINT"
		 */
		if (octInitGenContents(&secondBag, OCT_BAG_MASK, &bagGenId2)
								!= OCT_OK)  {
		    octError("readUserConsts:  can't begin 2nd bag gen.");
		    exit(-1);
		}
		while ((bagGenStat2=octGenerate(&bagGenId2, &constBag))
								== OCT_OK)  {
		    if (strcmp("CONSTRAINT",
			       constBag.contents.bag.name) == 0)  {
			status = readBorderConst(&constBag, &pn,
						 &min, &max, &po1);
			if (status != OK)  continue;
						
			if (min == 0)  {
			    min = po1->bbUpperR.x;
			}
			if (max == 0)  {
			    max = po1->bbUpperR.x;
			}

			/*
			 *  Add to the list.
			 */
			pc = ALLOC(struct userConstraint, 1);
			pc->pfromNode = pn;
			pc->ptoNode = pSink;
			pc->min = min;
			pc->max = max;
			pc->pnext = pXConsts;
			pXConsts = pc;
		    } else  {
			printf("readUserConst:  bag named \"%s\" ",
			       constBag.contents.bag.name);
			printf("is not understood.\n");
			continue;
		    }

		}
		if (bagGenStat2 != OCT_GEN_DONE)  {
		    printf("readUserConst:  can't end 2nd bag gen.\n");
		    exit(-1);
		}
	    } else if (strcmp(secondBag.contents.bag.name,"CONSTRAINT") == 0) {
		/*
		 *  A constraint between two instances.
		 */

		status = readConst(&secondBag, &pfirst, &psecond, &min, &max);
		if (status != OK)  continue;

		pc = ALLOC(struct userConstraint, 1);
		pc->pfromNode = pfirst;
		pc->ptoNode = psecond;
		pc->min = min;
		pc->max = max;
		pc->pnext = pXConsts;
		pXConsts = pc;

	    } else  {
		printf("readUserConst:  bag named \"%s\" is not understood.\n",
		       secondBag.contents.bag.name);
	    }
	}
    }


    /*
     *  GetByName the YCONSTRAINTS bag.
     */
    firstBag.contents.bag.name = "YCONSTRAINTS";
    if ((oStatus=octGetByName(&WorkingFacet, &firstBag)) == OCT_NOT_FOUND)  {
	;
	/* printf("\nNo Y user-constraints.\n\n"); */
    } else if (oStatus != OCT_OK)  {
	octError("can't getByName top-level y-constraints bag.");
	exit(-1);
    } else  {
	/*
	 *  Read each constraint.
	 */
	if (octInitGenContents(&firstBag,OCT_BAG_MASK,&bagGenId1) != OCT_OK)  {
	    octError("readUserConsts:  can't begin constraint bag generator.");
	    exit(-1);
	}
	while (octGenerate(&bagGenId1, &secondBag) == OCT_OK)  {
    	    if (strcmp(secondBag.contents.bag.name, "BOTTOM") == 0)  {
		/*
		 *  Should be more bags, called "CONSTRAINT"
		 */
		if (octInitGenContents(&secondBag, OCT_BAG_MASK, &bagGenId2)
								!= OCT_OK)  {
		    octError("readUserConsts:  can't begin 2nd bag gen.");
		    exit(-1);
		}
		while ((bagGenStat2=octGenerate(&bagGenId2, &constBag))
								== OCT_OK)  {
		    if (strcmp("CONSTRAINT",
			       constBag.contents.bag.name) == 0)  {
			status = readBorderConst(&constBag, &pn,
						 &min, &max, &po1);
			if (status != OK)  continue;
			
			/*
			 *  Add to the list.
			 */
			pc = ALLOC(struct userConstraint, 1);
			pc->pfromNode = pSource;
			pc->ptoNode = pn;
			pc->min = min;
			pc->max = max;
			pc->pnext = pYConsts;
			pYConsts = pc;
		    } else  {
			printf("readUserConst:  bag named \"%s\" ",
			       constBag.contents.bag.name);
			printf("is not understood.\n");
			continue;
		    }
		}
		if (bagGenStat2 != OCT_GEN_DONE)  {
		    printf("readUserConst:  can't end 2nd bag gen.\n");
		    exit(-1);
		}
	    } else if (strcmp(secondBag.contents.bag.name, "TOP") == 0)  {
		/*
		 *  Should be more bags, called "CONSTRAINT"
		 */
		if (octInitGenContents(&secondBag, OCT_BAG_MASK, &bagGenId2)
								!= OCT_OK)  {
		    octError("readUserConsts:  can't begin 2nd bag gen.");
		    exit(-1);
		}
		while ((bagGenStat2=octGenerate(&bagGenId2, &constBag))
								== OCT_OK)  {
		    if (strcmp("CONSTRAINT",
			       constBag.contents.bag.name) == 0)  {
			status = readBorderConst(&constBag, &pn,
						 &min, &max, &po1);
			if (status != OK)  continue;
			
			if (min == 0)  {
			    min = po1->bbUpperR.y;
			}
			if (max == 0)  {
			    max = po1->bbUpperR.y;
			}
			/*
			 *  Add to the list.
			 */
			pc = ALLOC(struct userConstraint, 1);
			pc->pfromNode = pn;
			pc->ptoNode = pSink;
			pc->min = min;
			pc->max = max;
			pc->pnext = pYConsts;
			pYConsts = pc;
		    } else  {
			printf("readUserConst:  bag named \"%s\" ",
			       constBag.contents.bag.name);
			printf("is not understood.\n");
			continue;
		    }
		}
		if (bagGenStat2 != OCT_GEN_DONE)  {
		    printf("readUserConst:  can't end 2nd bag gen.\n");
		    exit(-1);
		}
	    } else if (strcmp(secondBag.contents.bag.name,"CONSTRAINT") == 0) {
		/*
		 *  A constraint between two instances.
		 */

		status = readConst(&secondBag, &pfirst, &psecond, &min, &max);
		if (status != OK)  continue;

		pc = ALLOC(struct userConstraint, 1);
		pc->pfromNode = pfirst;
		pc->ptoNode = psecond;
		pc->min = min;
		pc->max = max;
		pc->pnext = pYConsts;
		pYConsts = pc;
	    } else  {
		printf("readUserConst:  bag named \"%s\" is not understood.\n",
		       secondBag.contents.bag.name);
	    }
	}
    }
}



/*
 *  Read the CONSTRAINT bag for a pairwise constraint (i.e., between two
 *  instances).  Returns OK if fine, ERROR otherwise; exits if serious oct
 *  error occurs.
 *
 *  7/10/87
 *  last modified 12/1/87
 */

int
readConst(pconstBag, pfrom, pto, pmin, pmax)
octObject *pconstBag;
struct node **pfrom, **pto;	/* returned */
int *pmin, *pmax;		/* returned */
{
    octObject firstInst;
    octObject secondInst;
    octObject minProp;
    octObject maxProp;

    struct object *po1, *po2;

    octStatus instGenStat, minStat, maxStat;
    octGenerator instGenId;

    firstInst.type = OCT_INSTANCE;
    secondInst.type = OCT_INSTANCE;
    minProp.type = OCT_PROP;
    maxProp.type = OCT_PROP;

    /*
     *  Expect to find 2 instances and 1 or 2 integer-valued properties.
     */
    if (octInitGenContents(pconstBag,OCT_INSTANCE_MASK,&instGenId) != OCT_OK) {
	octError("readConst:  can't begin inst gen.");
	exit(-1);
    }
    if ((instGenStat=octGenerate(&instGenId, &firstInst)) == OCT_OK)  {
	po1 = pObjectList;
	for (; po1!=NIL(struct object); po1=po1->pnext)  {
	    if (octIdsEqual(po1->instanceId, firstInst.objectId))  break;
	}
	if (po1 == NIL(struct object))  {
	    printf("readConst:  can't find sparcs object.");
	    exit(-1);
	}
        *pfrom = &(po1->node);
    } else if (instGenStat == OCT_GEN_DONE)  {
	printf("readConst:  first instance missing; ignoring constraint.\n");
	return(ERROR);
    } else  {
	octError("readConst:  can't get first inst.");
	exit(-1);
    }
    if ((instGenStat=octGenerate(&instGenId, &secondInst)) == OCT_OK)  {
	po2 = pObjectList;
	for (; po2!=NIL(struct object); po2=po2->pnext)  {
	    if (octIdsEqual(po2->instanceId, secondInst.objectId))  break;
	}
	if (po2 == NIL(struct object))  {
	    printf("readConst:  can't find sparcs object.");
	    exit(-1);
	}
        *pto = &(po2->node);
    } else if (instGenStat == OCT_GEN_DONE)  {
	printf("readConst:  second instance missing; ignoring constraint.\n");
	return(ERROR);
    } else  {
	octError("readConst:  can't get second inst.");
	exit(-1);
    }
    if (octGenerate(&instGenId, &secondInst) != OCT_GEN_DONE)  {
	printf("readConst:  more than two instances present;");
	printf(" ignoring constraint.\n");
	return(ERROR);
    }

    minProp.contents.prop.name = "MIN";
    maxProp.contents.prop.name = "MAX";
    minStat = octGetByName(pconstBag, &minProp);
    maxStat = octGetByName(pconstBag, &maxProp);
    if (minStat != OCT_OK  &&  minStat != OCT_NOT_FOUND)  {
	octError("readConst:  can't get minProp.");
	exit(-1);
    }
    if (maxStat != OCT_OK  &&  maxStat != OCT_NOT_FOUND)  {
	octError("readConst:  can't get maxProp.");
	exit(-1);
    }
    if (minStat == OCT_NOT_FOUND  &&  maxStat == OCT_NOT_FOUND) {
	printf("readConst:  no properties found; ignoring constraint.\n");
	return(ERROR);
    }
    if (minStat == OCT_OK)  {
	if (minProp.contents.prop.type != OCT_INTEGER)  {
	    printf("readConst:  illegal minProp type; ignoring constraint.\n");
	    return(ERROR);
	}
	*pmin = minProp.contents.prop.value.integer;
    } else  {
	*pmin = -INFINITY;
    }
    if (maxStat == OCT_OK)  {
	if (maxProp.contents.prop.type != OCT_INTEGER)  {
	    printf("readConst:  illegal maxProp type; ignoring constraint.\n");
	    return(ERROR);
	}
	*pmax = maxProp.contents.prop.value.integer;
    } else  {
	*pmax = INFINITY;
    }
    if (minStat == OCT_OK  &&  maxStat == OCT_OK)  {
	if (*pmin > *pmax)  {
	    printf("readConst:  minProp > maxProp; ignoring constraint.\n");
	    return(ERROR);
	}
    }
    return(OK);
}



/*
 *  Read the CONSTRAINT bag for a border constraint (i.e., between one instance
 *  and an edge of the BB of the cell).  Returns OK if fine, ERROR otherwise;
 *  exits if serious oct error occurs.
 *
 *  from readConst 12/1/87
 *  last modified 12/1/87
 */

int
readBorderConst(pconstBag, pn, pmin, pmax, po)
octObject *pconstBag;
struct node **pn;		/* returned */
int *pmin, *pmax;		/* returned */
struct object **po;		/* returned */
{
    octObject inst;
    octObject minProp;
    octObject maxProp;
    int count;
    struct object *po1;

    octStatus instGenStat, minStat, maxStat;
    octGenerator instGenId;

    inst.type = OCT_INSTANCE;
    minProp.type = OCT_PROP;
    maxProp.type = OCT_PROP;

    /*
     *  Expect to find 1 instance and 1 or 2 properties.
     */
    if (octInitGenContents(pconstBag,OCT_INSTANCE_MASK,&instGenId) != OCT_OK) {
	octError("readBorderConst:  can't begin inst gen.");
	exit(-1);
    }
    count = 0;
    while ((instGenStat=octGenerate(&instGenId, &inst)) == OCT_OK)  {
	count++;
    }
    if (instGenStat != OCT_GEN_DONE)  {
	octError("readBorderConst:  can't end inst generator");
	exit(-1);
    }
    if (count == 1)  {
	po1 = pObjectList;
	for (; po1!=NIL(struct object); po1=po1->pnext)  {
	    if (octIdsEqual(po1->instanceId, inst.objectId))  break;
	}
	if (po1 == NIL(struct object))  {
	    printf("readBorderConst:  can't find sparcs object.");
	    exit(-1);
	}
        *pn = &(po1->node);
	*po = po1;
    } else if (count == 0)  {
	printf("readBorderConst:  instance missing; ignoring constraint.\n");
	return(ERROR);
    } else  {
	printf("readBorderConst:  %d instances present;", count);
	printf(" ignoring constraint.\n");
	return(ERROR);
    }
    
    minProp.contents.prop.name = "MIN";
    maxProp.contents.prop.name = "MAX";
    minStat = octGetByName(pconstBag, &minProp);
    maxStat = octGetByName(pconstBag, &maxProp);
    if (minStat != OCT_OK  &&  minStat != OCT_NOT_FOUND)  {
	octError("readBorderConst:  can't get minProp.");
	exit(-1);
    }
    if (maxStat != OCT_OK  &&  maxStat != OCT_NOT_FOUND)  {
	octError("readBorderConst:  can't get maxProp.");
	exit(-1);
    }
    if (minStat == OCT_NOT_FOUND  &&  maxStat == OCT_NOT_FOUND) {
	printf("readBorderConst: no properties found; ignoring constraint.\n");
	return(ERROR);
    }
    if (minStat == OCT_OK)  {
	if (minProp.contents.prop.type != OCT_INTEGER)  {
	    printf("readBorderConst:  illegal minProp type; ");
	    printf("ignoring constraint.\n");
	    return(ERROR);
	}
	*pmin = minProp.contents.prop.value.integer;
    } else  {
	*pmin = -INFINITY;
    }
    if (maxStat == OCT_OK)  {
	if (maxProp.contents.prop.type != OCT_INTEGER)  {
	    printf("readBorderConst:  illegal maxProp type; ");
	    printf("ignoring constraint.\n");
	    return(ERROR);
	}
	*pmax = maxProp.contents.prop.value.integer;
    } else  {
	*pmax = INFINITY;
    }
    if (minStat == OCT_OK  &&  maxStat == OCT_OK)  {
	if (*pmin > *pmax)  {
	    printf("readBorderConst:  minProp > maxProp; ");
	    printf("ignoring constraint.\n");
	    return(ERROR);
	}
    }
    return(OK);
}
