#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

#define POINT_BOX_INTERSECT(array, index, bx)\
	(array[index].x >= bx.lowerL.x) && (array[index].x <= bx.upperR.x) &&\
	(array[index].y >= bx.lowerL.y) && (array[index].y <= bx.upperR.y)

#define NONE	1
#define BOTH	2
#define FIRST	3
#define SECOND	4

/*
 *  Find the sparcs terminal that intersects one or both ends of the segment.
 *  If found, assign the netId to ALL the terminal frame edges that constitute
 *  the terminal's implementation.
 *
 *  Assumes that a SINGLE terminal frame intersects one or both ends of the
 *  segment.  This could cause a problem if terminal frames were generalized
 *  such that multiple non-coincident shapes were allowed.
 *
 *  11/12/87
 *  last modified 9/19/88
 */

int
rpc_findTFrame(points, sparcsLayer, netId, instId, po, pt)
struct octPoint points[];
int sparcsLayer;
int netId;		/* net that segment belongs to */
int instId;		/* octId of instance that owns aterm */
struct object **po;	/* sparcs object that owns terminal; RETURNED */
struct terminal **pt;	/* pointer to intersecting sparcs term; RETURNED */
{
    struct object *pobj;
    struct terminal *pterm, *pfirstTerm;
    struct rect box;
    int intersect;
    int layerMatch;

    intersect = NONE;
    layerMatch = FALSE;

    /*
     *  First find the sparcs object corresponding to the oct instance that
     *  the terminal belongs to.
     */
    for (pobj=pObjectList; pobj!=NIL(struct object); pobj=pobj->pnext)  {
	if (pobj->instanceId == instId)  break;
    }
    if (pobj == NIL(struct object))  {
	printf("rpc_findTFrame:  can't find sparcs object for oct inst.\n");
	exit(-1);
    }

    /*
     *  The master caching does not allow for skipping down the terminal list
     *  according to atermId data, unlike the stand-alone vertion.  See
     *  findTFrame() for comparison.
     */
    pterm = pobj->pterminal;
    pfirstTerm = pterm;
 
    /*
     *  Now scan for a sparcs terminal on the same layer as the segment.  All
     *  sparcs terminals for a given aterm (fterm) are consecutive in the list.
     */
    while (pterm!=NIL(struct terminal))  {
	if (pterm->maskNumber == sparcsLayer  &&  intersect == NONE)  {
	    layerMatch = TRUE;

	    /*
	     *  Check both ends of segment for intersection with this aterm.
	     */
	    box.lowerL.x = pobj->origin.x + pterm->bb.lowerL.x;
	    box.lowerL.y = pobj->origin.y + pterm->bb.lowerL.y;
	    box.upperR.x = pobj->origin.x + pterm->bb.upperR.x;
	    box.upperR.y = pobj->origin.y + pterm->bb.upperR.y;
	    if (POINT_BOX_INTERSECT(points, 0, box))  {
		intersect = FIRST;
	    }
	    if (POINT_BOX_INTERSECT(points, 1, box))  {
		if (intersect == FIRST)  {
		    intersect = BOTH;
		} else  {
		    intersect = SECOND;
		}
	    }
	    if (intersect != NONE)  {
		*po = pobj;
		*pt = pterm;
		break;
	    }
	}
	pterm = pterm->pnext;
	if (pterm->atermId != pfirstTerm->atermId)  {
	    pfirstTerm = pterm;
	}
    }

    /*
     *  If no intersection quit; if due to no layerMatch, then report.
     */
    if (intersect == NONE)  {
	if (layerMatch == FALSE)  {
	    printf("\nrpc_findTFrame:  no layer match for wire.\n");
	}
	return(intersect);
    }


    /*
     *  An intersection has been found.  Set the netId for each terminal frame
     *  edge.  May be set already; if so, the old netId must agree with the
     *  new one.
     */
    pterm = pfirstTerm;
    do  {
	if (pterm->plE->netId!=UNDEF && pterm->plE->netId!=netId) {
	    printf("rpc_findTFrame:  old/new netIds conflict.\n");
	    exit(-1);
	} else  {
	    pterm->plE->netId = netId;
	    pterm->plE->pt = pterm;
	}
	if (pterm->prE->netId!=UNDEF && pterm->prE->netId!=netId) {
	    printf("rpc_findTFrame:  old/new netIds conflict.\n");
	    exit(-1);
	} else  {
	    pterm->prE->netId = netId;
	    pterm->prE->pt = pterm;
	}
	if (pterm->pbE->netId!=UNDEF && pterm->pbE->netId!=netId) {
	    printf("rpc_findTFrame:  old/new netIds conflict.\n");
	    exit(-1);
	} else  {
	    pterm->pbE->netId = netId;
	    pterm->pbE->pt = pterm;
	}
	if (pterm->ptE->netId!=UNDEF && pterm->ptE->netId!=netId) {
	    printf("rpc_findTFrame:  old/new netIds conflict.\n");
	    exit(-1);
	} else  {
	    pterm->ptE->netId = netId;
	    pterm->ptE->pt = pterm;
	}

	pterm = pterm->pnext;
    } while (pterm != NIL(struct terminal)  &&
	     pterm->atermId == pfirstTerm->atermId);

    return(intersect);
}
