#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "facet.h"

/*
 *  Routine to save some data regarding the sparcs run.  Currently the command
 *  line is saved in a property called COMMAND-LINE, and the results of the
 *  run, namely the number of iterations and the convergence status, are saved
 *  in a property called STOP-INFO.  These two (string) properties are stored
 *  on the output facet.  The facet contains a bag called "SPARCS-DATA", which
 *  in turn contains these properties (along with the critical-path data).
 *
 *  8/31/87
 *  last modified 9/1/87
 */

void
saveRunInfo(argc, argv, eltsConverged, areaConverged)
    int argc;
    char **argv;
    int eltsConverged, areaConverged;
{
    octObject dataBag;
    octObject prop;
    octStatus status;
    char *ptr, buf[BIGBUF];
    int i;
    
    dataBag.type = OCT_BAG;
    prop.type = OCT_PROP;

    /*
     *  SPARCS-DATA bag should exist since saveCP() called first, but use
     *  GetOrCreate anyway.
     */
    dataBag.contents.bag.name = "SPARCS-DATA";
    if (octGetOrCreate(&WorkingFacet, &dataBag) != OCT_OK)  {
	octError("saveRunInfo:  can't getOrCreate SPARCS-DATA bag.");
	exit(-1);
    }

    /*
     *  First do the command line.  Get rid of the path name in argv[0].
     */
    if ((ptr = strrchr(argv[0], '/')) != NULL) {
	(void) strcpy(buf, ptr+1);
    } else {
	(void) strcpy(buf, argv[0]);
    }
    for (i = 1;  i < argc;  i++)  {
	(void) strcat(buf, " ");
	(void) strcat(buf, argv[i]);
    }

    /*
     *  The properties exist if the input and output facets are the same;
     *  otherwise they must be created.
     */
    prop.contents.prop.name = "COMMAND-LINE";
    prop.contents.prop.type = OCT_STRING;
    if ((status=octGetByName(&dataBag, &prop)) == OCT_OK)  {
	prop.contents.prop.value.string = buf;
	if (octModify(&prop) != OCT_OK)  {
	    octError("saveRunInfo:  can't modify COMMAND-LINE prop.");
	    exit(-1);
	}
    } else if (status == OCT_NOT_FOUND)  {
	prop.contents.prop.value.string = buf;
	if (octCreate(&dataBag, &prop) != OCT_OK)  {
	    octError("saveRunInfo:  can't create COMMAND-LINE prop.");
	    exit(-1);
	}
    } else  {
	octError("saveRunInfo:  can't get COMMAND-LINE prop.");
	exit(-1);
    }

    /*
     *  Now do the stopping-data property.
     */
    if (StopCriterion == 'E')  {
	(void) sprintf(buf, "Elements converged after %d iterations", Iterations);
    } else if (StopCriterion == 'A')  {
	(void) sprintf(buf, "Area converged after %d iterations, ", Iterations);
	if (eltsConverged)  {
	    (void) strcat(buf, "elements also converged");
	} else  {
	    (void) strcat(buf, "elements NOT converged");
	}
    } else  {
	(void) sprintf(buf, "Iteration limit of %d reached; ", Iterations);
	if (areaConverged)  {
	    (void) strcat(buf, "area converged, ");
	} else  {
	    (void) strcat(buf, "area NOT converged, ");
	}
	if (eltsConverged)  {
	    (void) strcat(buf, "elements converged");
	} else  {
	    (void) strcat(buf, "elements NOT converged");
	}
    }

    prop.contents.prop.name = "STOP-INFO";
    if ((status=octGetByName(&dataBag, &prop)) == OCT_OK)  {
	prop.contents.prop.value.string = buf;
	if (octModify(&prop) != OCT_OK)  {
	    octError("saveRunInfo:  can't modify STOP-INFO prop.");
	    exit(-1);
	}
    } else if (status == OCT_NOT_FOUND)  {
	prop.contents.prop.value.string = buf;
	if (octCreate(&dataBag, &prop) != OCT_OK)  {
	    octError("saveRunInfo:  can't create STOP-INFO prop.");
	    exit(-1);
	}
    } else  {
	octError("saveRunInfo:  can't get STOP-INFO prop.");
	exit(-1);
    }
}
