#include "copyright.h"
#include "sparcs.h"
#include "oct.h"

/*
 *  Trim the wire endpoint coords such that the wire starts at the right edge
 *  of the left-most terminal, and goes to the left edge of the right-most
 *  terminal.  This results in wires of length > 0 when the terminals DON'T
 *  overlap, and length < 0 when the terminals DO overlap.
 *
 *  Called By:	readOct()
 *  Calls:	-
 *
 *  6/12/87
 *  last modified 11/30/87
 */

/*ARGSUSED*/
void
trimWire(pseg, pw, pt1, pt2, po1, po2)
struct octObject *pseg;
struct wire *pw;
struct terminal *pt1, *pt2;
struct object *po1, *po2;	/* owners of pt1, pt2 */
{
    struct terminal *pminTerm;
    struct terminal *pmaxTerm;
    struct object *pminObj;
    struct object *pmaxObj;
    /***
    struct octPoint path[2];
    ***/

    /*
     *  Find the left terminal - the owner of the left-most right edge,
     *  and the right terminal - the owner of the right-most left edge.
     */
    if (pw->horizontal == TRUE)  {
	/*
	 *  Terminal with left-most right edge
	 */
	if (po1->origin.x+pt1->bb.upperR.x < po2->origin.x+pt2->bb.upperR.x)  {
	    pminTerm = pt1;
	    pminObj = po1;
	} else  {
	    pminTerm = pt2;
	    pminObj = po2;
	}
	/*
	 *  And with right-most left edge
	 */
	if (po1->origin.x+pt1->bb.lowerL.x > po2->origin.x+pt2->bb.lowerL.x)  {
	    pmaxTerm = pt1;
	    pmaxObj = po1;
	} else  {
	    pmaxTerm = pt2;
	    pmaxObj = po2;
	}

	/*
	 *  First end of wire should be at right edge of pminTerm.
	 */
	pw->p1.x = pminObj->origin.x + pminTerm->bb.upperR.x;
	pw->pt1 = pminTerm;
	pw->po1 = pminObj;

	/*
	 *  If pminTerm and pmaxTerm are the same, then change pmaxTerm to
	 *  point to the other terminal.
	 */
	if (pminTerm == pmaxTerm)  {
	    if (pmaxTerm == pt1)  {
		pmaxTerm = pt2;
		pmaxObj = po2;
	    } else  {
		pmaxTerm = pt1;
		pmaxObj = po1;
	    }
	}

	pw->p2.x = pmaxObj->origin.x + pmaxTerm->bb.lowerL.x;
	pw->pt2 = pmaxTerm;
	pw->po2 = pmaxObj;
    } else  {
	/*
	 *  Terminal with bottom-most top edge
	 */
	if (po1->origin.y+pt1->bb.upperR.y < po2->origin.y+pt2->bb.upperR.y)  {
	    pminTerm = pt1;
	    pminObj = po1;
	} else  {
	    pminTerm = pt2;
	    pminObj = po2;
	}
	/*
	 *  And with top-most bottom edge
	 */
	if (po1->origin.y+pt1->bb.lowerL.y > po2->origin.y+pt2->bb.lowerL.y)  {
	    pmaxTerm = pt1;
	    pmaxObj = po1;
	} else  {
	    pmaxTerm = pt2;
	    pmaxObj = po2;
	}

	/*
	 *  First end of wire should be at top edge of pminTerm.
	 */
	pw->p1.y = pminObj->origin.y + pminTerm->bb.upperR.y;
	pw->pt1 = pminTerm;
	pw->po1 = pminObj;

	/*
	 *  If pminTerm and pmaxTerm are the same, then change pmaxTerm to
	 *  point to the other terminal.
	 */
	if (pminTerm == pmaxTerm)  {
	    if (pmaxTerm == pt1)  {
		pmaxTerm = pt2;
		pmaxObj = po2;
	    } else  {
		pmaxTerm = pt1;
		pmaxObj = po1;
	    }
	}

	pw->p2.y = pmaxObj->origin.y + pmaxTerm->bb.lowerL.y;
	pw->pt2 = pmaxTerm;
	pw->po2 = pmaxObj;
    }

    /*
     *  XXX  Debug only:  Now update the octPath with the new coordinates.
     */
    /*********
    path[0].x = pw->p1.x;
    path[0].y = pw->p1.y;
    path[1].x = pw->p2.x;
    path[1].y = pw->p2.y;

    if (octPutPoints(pseg, 2, path) != OCT_OK)  {
	octError("trimWire:  can't update path's points");
	exit(-1);
    }
    *********/

    return;
}
