#include "copyright.h"
#include "sparcs.h"
#include "errtrap.h"

/*
 *  New routine to link nodes in the graph structure.  Unlike link(), looks
 *  for an existing branch between the nodes and if found updates it.
 *
 *  Uses a new struct arc or reuses an old one depending on pNextFreeArc.
 *
 *  8/7/87
 *  last modified 9/22/88
 */

void
link(pfrom, pto, lower, upper, force)
    struct node *pfrom, *pto;
    int lower, upper, force;
{
    register struct branch *pb;
    int i = 1;
    int found = FALSE;
    
    if (force < 0)  {
	printf("link:  force = %d < 0 is illegal\n", force);
	exit(-1);
    }

    if (pfrom == pto)  {
	errRaise( "SPARCS", SPARCS_LAYOUT_CODE, "link - pfrom == pto!\n\n\tProblem object has lower left corner %d,%d",
		 pto->loc.x, pto->loc.y );
    }

    /*
     *  See if pto is already a fanout of pfrom.  If LinkSearchDepth is
     *  INFINITY, search to the end of the list of fanouts; otherwire,
     *  look down the list for LinkSearchDepth elements.
     */
    if (LinkSearchDepth == INFINITY)  {
	for (pb=pfrom->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	    if (pb->ptoNode == pto)  {
		found = TRUE;
		break;
	    }
	}
    } else  {
	for (pb=pfrom->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	    if (i > LinkSearchDepth)  {
		break;
	    } else  {
		i++;
	    }
	    if (pb->ptoNode == pto)  {
		found = TRUE;
		break;
	    }
	}
    }
    

    if (found == TRUE)  {
	/*
	 *  Update the existing branch.
	 */

	/*
	 *  Is the new upper meaningful?  If not, then do nothing.
	 */
	if (upper !=  INFINITY)  {
	    upper = upper/UNITSPERLAMBDA;
	    /*
	     *  If the old value is undefined, replace it with the new.  If
	     *  it is defined, keep the smaller of the two.
	     */
	    if (pb->upperBound == INFINITY)  {
		pb->upperBound = upper;
		pfrom->cpm.upperOutFlg = TRUE;
		pto->cpm.upperInFlg = TRUE;
		NumUppers++;
	    } else  {
		if (upper < pb->upperBound)  pb->upperBound = upper;
	    }
	}
		
	/*
	 *  If the new lower is meaningful then update.
	 */
	if (lower != -INFINITY)  {
	    lower = lower/UNITSPERLAMBDA;
	    if (pb->lowerBound == -INFINITY)  {
		pb->lowerBound = lower;
		pfrom->cpm.lowerOutFlg = TRUE;
		NumLowers++;
	    } else  {
		if (lower > pb->lowerBound)  pb->lowerBound = lower;
	    }
	}
	/*
	 *  Update the force (can just increment, since the nonsense force
	 *  has value zero).
	 */
	pb->force += force;
    } else  {
	/*
	 *  Branch wasn't found, so add one.  Reuse an old branch or allocate
	 *  a new one if necessary.
	 */
	if (pNextFreeBranch != NIL(struct branch))  {
	    pb = pNextFreeBranch;
	    pNextFreeBranch = pNextFreeBranch->pnext;
	} else  {
	    pb = ALLOC(struct branch, 1);
	    pb->pnext = pBranchList;
	    pBranchList = pb;
	}
	
	/***  6/18/87   pb->upperBound = upper; ***/
	if (upper == INFINITY)  {
	    pb->upperBound = INFINITY;
	} else  {
	    pb->upperBound = upper/UNITSPERLAMBDA;
	}
	if (upper != INFINITY)  NumUppers++;
	/***  6/18/87   pb->lowerBound = lower; ***/
	if (lower == -INFINITY)  {
	    pb->lowerBound = -INFINITY;
	} else  {
	    pb->lowerBound = lower/UNITSPERLAMBDA;
	}
	
	if (lower != -INFINITY)  NumLowers++;
	pb->proc  = 0;
	pb->force = force;
	
	pb->pfromNode = pfrom;
	pb->ptoNode = pto;
	
	pb->pnextFanout = pfrom->pfanouts;
	pfrom->pfanouts = pb;
	pfrom->cpm.numfanout++;
	
	pb->pnextFanin = pto->pfanins;
	pto->pfanins = pb;
	pto->cpm.numfanin++;
	
	/*
	 *  XXX  This looks like junk...
	 */
	if (lower > -INFINITY)  {
	    pfrom->cpm.lowerOutFlg = TRUE;
	}
	if (upper <  INFINITY)  {
	    pfrom->cpm.upperOutFlg = TRUE;
	    pto->cpm.upperInFlg = TRUE;
	    NumUppers++;
	}
	NumBranches++;
    }
}
