#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/

/*
 *  This routine is called before each spacing iteration, including the first.
 *  It resets the fields of each node, and marks the wires that are parallel
 *  to the spacing direction as UNDEF.
 *
 *  A combination of clearGraph() and activateNode(); activateNode() is thus
 *  no longer needed.
 *
 *  2/27/87
 *  last modified 3/25/88
 */

void
resetGraph(direction)
directionType direction;    /* direction of the NEXT spacing */
{
    int i;
    register struct node *pn;
    register struct object *po;
    register struct wire *pw;
    
    /*
     *  Resetting the freeBranch pointer before the first iteration is ok,
     *  since both pointers are NIL then...
     */
    pNextFreeBranch = pBranchList;
    NumNodes = 0;	/* incremented in cpm_installNode() */
    NumBranches = NumLowers = NumUppers = 0;
    /* NumSourceFanouts = NumSinkFanins = 0; */

    ClusterCtr = 0;
    
    for (i=1; i<=NumObjects; i++)  {
	pn = NodeArray[i];
	pn->key = i;			/* XXX Need to do only once */
	po = ElementByKey[i].ptr.po;
	/********  6/18/87
	  pn->loc.x = po->origin.x;
	  pn->loc.y = po->origin.y;
        ******************/
	pn->loc.x = (po->origin.x)/UNITSPERLAMBDA;
	pn->loc.y = (po->origin.y)/UNITSPERLAMBDA;
	pn->pfanouts = NIL(struct branch);
	pn->pfanins = NIL(struct branch);

	pn->cpm.lower 	= -INFINITY;
	pn->cpm.loc   	=  ZERO;
	pn->cpm.upper 	=  INFINITY;
	pn->cpm.level 	=  UNDEF;
	pn->cpm.active  = UNDEF;
	pn->cpm.lowerInFlg  =  FALSE;
	pn->cpm.upperInFlg  =  FALSE;
	pn->cpm.lowerOutFlg =  FALSE;	/* 3/18/87 */
	pn->cpm.upperOutFlg =  FALSE;
	pn->cpm.overcon     =  FALSE;
	pn->cpm.list 	=  NIL(struct node);
	pn->cpm.sched 	=  NIL(struct node);
	pn->cpm.path 	=  NIL(struct cpm_path);
	pn->cpm.numfanout 	=  ZERO;
	pn->cpm.numfanin 	=  ZERO;


	/* XXX */
	cpm_installNode(pn);


	/*
	 *  Reset linkage pointers of edges.
	 */
	if (direction == HORIZONTAL)  {
	    linkEdgePtrs(po->pmaxHEP, &(po->pfirstMaxEP), &(po->plastMaxEP));
	    linkEdgePtrs(po->pminHEP, &(po->pfirstMinEP), &(po->plastMinEP));
	    if (po->pmidHEP != NIL(struct edgePtr))  {
	        linkEdgePtrs(po->pmidHEP, &(po->pfirstMidEP),
				&(po->plastMidEP));
	    }
	} else  {
	    linkEdgePtrs(po->pmaxVEP, &(po->pfirstMaxEP), &(po->plastMaxEP));
	    linkEdgePtrs(po->pminVEP, &(po->pfirstMinEP), &(po->plastMinEP));
	    if (po->pmidVEP != NIL(struct edgePtr))  {
	        linkEdgePtrs(po->pmidVEP, &(po->pfirstMidEP),
				&(po->plastMidEP));
	    }
	}
    }
    for (i=NumObjects+1; i<=NumElements; i++)  {
	/*
	 *  Skip the horizontal wires if HORIZONTAL spacing, etc.
	 */
	pw = ElementByKey[i].ptr.pw;
	if (direction == HORIZONTAL)  {
	    if (pw->horizontal == TRUE)  {
		NodeArray[i]->key = UNDEF;
		continue;
	    }
	} else  {
	    if (pw->horizontal == FALSE)  {
		NodeArray[i]->key = UNDEF;
		continue;
	    }
	}
	pn = NodeArray[i];
	pn->key = i;
	/***************
	pn->loc.x = ElementByKey[i].ptr.pw->p1.x -
			ElementByKey[i].ptr.pw->halfWidth;
	pn->loc.y = ElementByKey[i].ptr.pw->p1.y -
			ElementByKey[i].ptr.pw->halfWidth;
	****************/
	pn->loc.x = (ElementByKey[i].ptr.pw->p1.x -
			ElementByKey[i].ptr.pw->halfWidth)/UNITSPERLAMBDA;
	pn->loc.y = (ElementByKey[i].ptr.pw->p1.y -
			ElementByKey[i].ptr.pw->halfWidth)/UNITSPERLAMBDA;
	pn->pfanouts = NIL(struct branch);
	pn->pfanins = NIL(struct branch);

	pn->cpm.lower 	= -INFINITY;
	pn->cpm.loc   	=  ZERO;
	pn->cpm.upper 	=  INFINITY;
	pn->cpm.level 	=  UNDEF;
	pn->cpm.active  = UNDEF;
	pn->cpm.lowerInFlg  =  FALSE;
	pn->cpm.upperInFlg  =  FALSE;
	pn->cpm.lowerOutFlg =  FALSE;	/* 3/18/87 */
	pn->cpm.upperOutFlg =  FALSE;
	pn->cpm.overcon     =  FALSE;
	pn->cpm.list 	=  NIL(struct node);
	pn->cpm.sched 	=  NIL(struct node);
	pn->cpm.path 	=  NIL(struct cpm_path);
	pn->cpm.numfanout 	=  ZERO;
	pn->cpm.numfanin 	=  ZERO;


	/* XXX */
	cpm_installNode(pn);
    }
}
