#include "copyright.h"

/*
 *  Structs for the graph.
 *
 *  Major modifications 7/14/86 to facilitate re-use of memory.
 *
 *  6/25/85
 *  last modified 3/25/88
 */

struct branch  {
    int lowerBound;
    int upperBound;
    int force;      		    /* attractive force */
    int proc;       		    /* used for leveling */
    struct node *pfromNode;
    struct node *ptoNode;
    struct branch *pnextFanout;	    /*  next fanout of pfrom  */
    struct branch *pnextFanin;	    /*  next fanin of pto  */
    struct branch *pnext; 	    /* NEEDED? */
};

struct cpm_path {
    struct cpm_path *next;              /* other branchPoints here */
    struct node *branchPoint;       /* pointer to this branch point */
    int lower;      			/* lower bound to branch point */
    int upper;				/* upper bound to branch point */
    int active;				/* whether dead (FALSE) or alive */
};

struct cpm_node {
    int lower;      /* lower bound on loc */
    int loc;        /* present location of node */
    int upper;      /* upper bound on loc */
    int level;      /* "level" of node from source (level0) */
    int active;	    /* used to mark nodes in clusterNoncrit */
    int lowerInFlg;	/* true if fanin branch with lower bound */
    int upperInFlg;	/* true if fanin branch with upper bound */
    int lowerOutFlg;
    int upperOutFlg;	/* true fanout branch with upper bound */
    int overcon;    /* true if node part of overconstraint loop */
    struct node *list;      /* all nodes are linked in order read */
    struct node *sched;     /* schedule pointer for event-driven */
    struct cpm_path *path;      /* path list for overconstraint check */
    int numfanout;          /* number of fanouts */
    int numfanin;           /* number of fanins */
};

struct node  {
    int key;			/* key of element corresponding to node */
    struct point loc;		/* location of node = origin of object */
    struct point newLoc;	/* new loc of node after spacing */
    struct branch *pfanouts;	/* list of fanout branches */
    struct branch *pfanins;	/* list of fanin branches */
    struct cpm_node cpm;	/* cpm-specific fields */
    struct node *oc_node;	/* Used to detect  overconstraints. */
};

/*
 *  Globals that pertain to the constraint graph
 */
extern struct node Source;
extern struct node Sink;

extern struct node *pSource;
extern struct node *pSink;
extern struct node **NodeArray;	    /*  array of ptrs to nodes except Source,
					Sink, indexed by key.  */
extern struct node *pNodeList;
extern struct branch *pBranchList;
extern struct branch *pNextFreeBranch;

extern int NumNodes;
extern int NumBranches;
extern int NumLowers;
extern int NumUppers;
extern int MaxLevel;
