#include "copyright.h"
#include "port.h"
#include "utility.h"

#undef MIN
#undef MAX

/*
 *  Basic data structs for sparcs.
 *  Renamed (was "readin.h") 6/3/86 and modified to remove hierarchical
 *  includes.
 *
 *  last modified 9/22/88
 */
#ifdef RPC
#include "oct.h"
#endif

/*
 *  Some constants and macro definitions.  (was in macros.h until 6/3/86)
 */

#define TRUE		1
#define FALSE		0
#define ZERO		0
#define INFINITY 	1000000000
#define LINELENGTH	80
#define SMALLBUF	256
#define BIGBUF		1024
#define UNDEF		-99
#define NO_RULE		-100
#define MAXLAYERS	30

#define UNITSPERLAMBDASQUARED 400

extern int UNITSPERLAMBDA;


#define SPMAX(A,B) ((A)>(B) ? (A) : (B))
#define SPMIN(A,B) ((A)<(B) ? (A) : (B))

#define STRSAVE(string) \
			strcpy(malloc((unsigned)(strlen(string)+1)),string)

/*  This adds an arc to the constraint graph.  */
#define LINK(ptail, phead, min, max, pnewArc, pnewArcList) \
		/*  New arc struct.  */ \
		pnewArc = (struct arc *) \
				malloc((unsigned)sizeof(struct arc)); \
		pnewArc->lowerBound = min; \
		pnewArc->upperBound = max; \
		pnewArc->pfrom = ptail; \
		pnewArc->pto = phead; \
		/*  arcList struct for the tail node.  */ \
		pnewArcList = (struct arcList *) \
				malloc((unsigned)sizeof(struct arcList)); \
		pnewArcList->parc = pnewArc; \
		pnewArcList->pnext = ptail->pfanout; \
		ptail->pfanout = pnewArcList; \
		/*  arcList struct for the head node.  */ \
		pnewArcList = (struct arcList *) \
				malloc((unsigned)sizeof(struct arcList)); \
		pnewArcList->parc = pnewArc; \
		pnewArcList->pnext = phead->pfanin; \
		phead->pfanin = pnewArcList;

/*  the following is 1 if bb's intersect, 0 otherwise  */
#define BB_INTERSECT(pbb1, pbb2) \
			((pbb1->lowerL.x <= pbb2->upperR.x) && \
			 (pbb1->upperR.x >= pbb2->lowerL.x) && \
			 (pbb1->lowerL.y <= pbb2->upperR.y) && \
			 (pbb1->upperR.y >= pbb2->lowerL.y))

#define COMPUTE_SIGNATURE(pchar, sig)\
			for (; *pchar!='\0'; pchar++)  {\
			    sig *= 100;\
			    sig += *pchar;\
			}

/*  The following for cpm  */
#define MIN     1
#define MAX     2
#define DUP     -1
#define NOROOM  0
#define OK      1
#define ERROR   -1
#define TRACE 1
#define TABLE 2
#define ELEMENT 4
#define RLEFT   -2
#define LEFT    -1	/* also for leftOrRight() */
#define DCARE   0
#define RIGHT   1	/* also for leftOrRight() */
#define RRIGHT  2
#define NODES   1
#define BRANCHES        2
#define FORWARD 1
#define REVERSE -1
#define ALL     9

/* Error codes */
#define SPARCS_BUG_CODE     77	/* Just something sparcs cannot handle. */
#define SPARCS_OVERCON_CODE 78	/* An overconstraint. */
#define SPARCS_LAYOUT_CODE  79	/* Possibly something is strange with the layout. */
#define SPARCS_SANITIZE_CODE  80 /* Something strange found while leveling. */

struct point {
    int x, y;
};

/*
 *  XXXX FIX THIS UP!
 */
#include "graph.h"

struct rect {
    struct point lowerL;	/*  lower left coordinate  */
    struct point upperR;	/*  upper right coordinate  */
};

/*
 *  XXX scan version only
 */
#include "scan.h"

/*
 *  All terminals are rectangular.
 *
 *  XXX Could get rid of bb, since also have ptrs to edges...
 */
struct terminal  {
    struct rect bb;		/* coords rel to origin of object */
    int maskNumber;		/* layer for this frame */
    int atermId;
    struct edge *plE;		/* left edge of terminal */
    struct edge *prE;		/* right edge of terminal */
    struct edge *ptE;		/* top edge of terminal */
    struct edge *pbE;		/* bottom edge of terminal */
    struct terminal *pnext;	/* next terminal in the linked list */
};
 
struct object  {
    int key;			/* sparcs Id number */
    int instanceId;		/* octId for corresp. octObject */
    struct point origin;	/* taken to be LL corner of bb of object  */
    struct point bbUpperR;	/* upper-rt. corner of bb (origin-relative) */
    struct point initOrigin;	/* origin from BEFORE any spacings  */
    struct point lastOrigin;	/* origin from last spacing iteration  */
    struct terminal *pterminal;	/* ptr to 1st in linked list of terminals  */
    struct edgePtr *pfirstMaxEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *plastMaxEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *pfirstMidEP;
    struct edgePtr *plastMidEP;
    struct edgePtr *pfirstMinEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *plastMinEP;	    /* XXX scan version (2/25/87) */
    struct edgePtr *pminHEP;	    /* XXX scan version (3/1/87) */
    struct edgePtr *pmidHEP;
    struct edgePtr *pmaxHEP;	    /* XXX scan version (3/1/87) */
    struct edgePtr *pminVEP;	    /* XXX scan version (3/1/87) */
    struct edgePtr *pmidVEP;
    struct edgePtr *pmaxVEP;	    /* XXX scan version (3/1/87) */
    struct edge *pverticalEdges;	    /* XXX scan version (2/25/87) */
    struct edge *phorizontalEdges;	    /* XXX scan version (2/25/87) */
    char hasPFrames;			/* 'T' or 'F' */
    struct object *pnext;	/*  objects are in a linked list  */
    struct node node;
#ifdef RPC
    char *name;
    struct octTransform transform;
#endif
};

struct wire  {		/*  all wires are single-segment  */
    int key;
    /*  net?  */
    int pathId;			/* id for oct path object */
    int changed;		/* TRUE or FALSE; used by updateOct() */
    int horizontal;		/* TRUE or FALSE */
    struct point p1;		/* first endpoint; lesser coordinates */
    struct point p2;		/* second endpoint; greater coordinates */
    struct object *po1;		/* object connected to pt. p1 */
    struct object *po2;		/* object connected to pt. p2 */
    struct terminal *pt1;	/* terminal of po1 that p1 connects to */
    struct terminal *pt2;	/* terminal of po2 that p2 connects to */
    int width;			/* obviously the wire's width */
    int halfWidth;		/* 6/12/87;  width/2 */
    int force;			/* 4/22/87;  for slack distribution */
    int maskNumber;
    struct edgePtr *pfirstMaxEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *plastMaxEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *pfirstMinEP;    /* XXX scan version (2/25/87) */
    struct edgePtr *plastMinEP;	    /* XXX scan version (2/25/87) */
    struct wire *pnext;		/*  wire segments are linked  */
    struct node node;
};

/*
 *  Added 6/16/87
 */
struct userConstraint  {
    struct node *pfromNode;
    struct node *ptoNode;
    int min, max;
    struct userConstraint *pnext;
};

/*
 *  Added 12/8/87
 */
struct pair  {
    struct node *pfromNode;
    struct node *ptoNode;
};

struct activeConstraint  {
    struct pair pair1;
    struct pair pair2;
    struct activeConstraint *pnext;
};

/*
 *  The following is to enable of referencing by key.  There is an array
 *  of 'element' structs, 'ElementByKey', numbered from 1 to NumElements
 *  (= NumObjects+NumWires).
 *  8/12/85
 */
struct element  {
    union  {
	struct object *po;	/*  if element 'key' is an object  */
	struct wire   *pw;  	/*  if element 'key' is a wire  */
    } ptr;
};

/*  for the direction of compaction  */
typedef enum  { HORIZONTAL, VERTICAL }  directionType;

/*  mask layer info  */
struct layer  {
    char *name;		    /* layer name from technology  */
    int ohmsPerSquare;	    /* for slack-distribution forces */
    int maxRule;	    /* largest spacing rule w.r.t. this layer */
};

/*
 *  Spacing rule data.
 */
struct rule  {
    int spacing;
    char mergeable;	/* 'Y' or 'N' */
};

/*
 *  Globals.  NOTE - some others are externed in graph.h and facet.h
 */
extern char *InCell;
extern char *OutCell;
extern char *Facet;
extern char *InView, *OutView;

extern char StopCriterion;	/* 'I' (IterLim), 'A' (area), 'E' (elements) */

extern int RPCSparcs;

extern char *TechRoot;
extern char *RulesFile;

extern int NumElements;
extern int NumObjects;
extern struct object *pObjectList;
extern int NumWires;
extern struct wire *pWireList;
extern int ElementKey;
extern int NumLayers;
extern int MaxBBX, MaxBBY;
extern int MinBBX, MinBBY;

extern struct edgePtr *pMaxEdgeListHead;	/* 2/15/87, scan version only */
extern struct edgePtr *pMaxEdgeListTail;	/* 3/4/87, scan version only */
extern struct edgePtr *pMidEdgeListHead;	/* 6/25/87, scan version only */
extern struct edgePtr *pMidEdgeListTail;	/* 6/25/87, scan version only */
extern struct edgePtr *pMinEdgeListHead;	/* 2/21/87, scan version only */
extern struct edgePtr *pMinEdgeListTail;	/* 3/4/87, scan version only */

extern struct element *ElementByKey;	/*  an array of element structs  */
extern struct layer LayerMap[MAXLAYERS];

/*
 *  Added 1/19/88
 */

/*  changed 6/20/87   extern int Rules[MAXLAYERS][MAXLAYERS]; */
extern struct rule Rules[MAXLAYERS][MAXLAYERS];
extern int RuleExp;

extern FILE *pout;
 
extern int Iterations;
extern int IterationLimit;

extern int ClusterCtr;

extern int LinkSearchDepth;

/*
 *  Added 6/16/87
 */
extern struct userConstraint *pXConsts;
extern struct userConstraint *pYConsts;

/*
 *  Added 12/8/87
 */
extern struct activeConstraint *pXActiveConsts;
extern struct activeConstraint *pYActiveConsts;

/*  flags  */
extern int TraceFlag;
extern int MinChangeFlag;
extern int ShowIterationsFlag;
extern int SaveIterationsFlag;
extern int VGridFlag;
extern int HorizontalFirstFlag;
extern int PrintObjectsFlag;
extern int PrintWiresFlag;
extern int PrintGraphFlag;
extern int PrintCpmGraphFlag;
extern int NoUserConstraintsFlag;
extern int ReadDB;
extern int RulesDB;
extern int UpdateDB;
extern int HardLeftFlag;
extern int HardRightFlag;
extern int HardTopFlag;
extern int HardBottomFlag;
extern int OrthogonalCPFlag;
