#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "facet.h"
#include "errtrap.h"


printErrorMessage( status )
    int status;
{
    char *command = "sparcsHelp help"; 
    fflush( stdout );
    fflush( stderr );
    
    util_csystem( command );

}

void sparcsHandler( pkg, status, msg )
    char* pkg;
    int status;
    char* msg;
{
    void printObjList(), printWireList();

    if ( strcmp( pkg, "SPARCS" ) == 0 ) {
	
	printf( "\n\nError detected by SPARCS: %s\n\n", msg );
	
	if ( status == SPARCS_OVERCON_CODE ) {
	    findOverconstraints();
	} else if ( status == SPARCS_SANITIZE_CODE ) {
	    findUnscheduled();
	}

	printf( "Saving %s ...", ohFormatName( &WorkingFacet ) );
	octFlushFacet(&WorkingFacet);

	printErrorMessage( status );

	VOVend( status );
    } else {
	errPass( pkg, status, msg );
    }
}


/*
 *  Main routine for sparcs with the scanline graph-builder.
 *
 *  1/28/86
 *  last modified 1/26/88
 */



main(argc, argv)
int argc;
char *argv[];
{
    void SPinit();
    void readInputs(), initOct();
    void printObjList(), printWireList();
    void allocGraph();
    int doSpacing();
    int width, height;
    int area, lastArea;
    int eltsConverged = FALSE;
    int areaConverged = FALSE;
    directionType direction;
    
    int h = 1;		/*  number of horizontal spacings  */
    int v = 1;		/*  number of vertical spacings  */
    

    commandLine(argc, argv);

    SPinit();


    if ( argc > 1 ) {
	VOVbegin( argc, argv );
	readDefaults();
    }
    

    
    if (TraceFlag)  {
	fprintf(pout, "Sparcs version %s initialized.\n", sparcsVersion());
	fflush(pout);
    }
    
    errPushHandler( sparcsHandler );

    octBegin();
    
    initOct();
    
    if (PrintObjectsFlag)  printObjList();
    if (PrintWiresFlag)  printWireList();
    
    allocGraph();
    
    width = (MaxBBX-MinBBX)/UNITSPERLAMBDA;
    height = (MaxBBY-MinBBY)/UNITSPERLAMBDA;
    area = width * height;
    if (TraceFlag)  {
	fprintf(pout, "\n");
	printf("Initial horizontal extent:  %d;  vertical extent:  %d;  ",
	       width, height);
	printf("area:  %d.\n", area);
	fflush(pout);
    }
    



    while ( 1 ) {
	char* dirString;
	int width, height;
	
	if ( Iterations & 0x1 ) {
	    direction = HorizontalFirstFlag ? VERTICAL : HORIZONTAL;
	} else {
	    direction = HorizontalFirstFlag ? HORIZONTAL : VERTICAL;
	}

	dirString = (direction==HORIZONTAL) ? "Horizontal": "Vertical";

	if (TraceFlag)  {
	    fprintf(pout, "\n%s spacing iteration %d.\n", dirString, Iterations);
	    fflush(pout);
	}
	eltsConverged = doSpacing(direction);
	Iterations++;
	
	lastArea = area;
	width = (MaxBBX-MinBBX)/UNITSPERLAMBDA;
	height = (MaxBBY-MinBBY)/UNITSPERLAMBDA;
	area = width * height ;
	if (TraceFlag)  {
	    printf("    New area: %5dx%5d = %d\n", width, height, area );
	    fflush(pout);
	}
	
        if (IterationLimit > 0)  {
	    if (Iterations == IterationLimit)  break;
	}
	if (Iterations > 1)  {
	    if (StopCriterion == 'A')  {
		if (lastArea == area)  {
		    areaConverged = TRUE;
		    break;
		}
	    } else if (StopCriterion == 'E')  {
		if (eltsConverged == TRUE)  break;
	    }
	}
    }


    if (TraceFlag)  {
	if (StopCriterion == 'E')  {
	    /*
	     *  Stopped when no elements moved.
	     */
	    fprintf(pout, "\nElements converged after %d iterations.\n",
		    Iterations);
	    fflush(pout);
	} else if (StopCriterion == 'A')  {
	    /*
	     *  Stopped when area remained constant.
	     */
	    fprintf(pout, "\nArea converged after %d iterations;",	
		    Iterations);
	    if (eltsConverged)  {
	        fprintf(pout, "  elements converged.\n");
	    } else  {
	        fprintf(pout, "  elements NOT converged.\n");
	    }
	    fflush(pout);
	} else  {
	    /*
	     *  Stopped when IterationLimit reached.
	     */
	    fprintf(pout, "\nIteration limit reached");
	    fprintf(pout, " (%d iterations);\n", IterationLimit);
	    if (eltsConverged)  {
	        fprintf(pout, "Elements converged,");
	    } else  {
	        fprintf(pout, "Elements NOT converged,");
	    }
	    if (areaConverged)  {
	        fprintf(pout, "  area converged.\n");
	    } else  {
	        fprintf(pout, "  area NOT converged.\n");
	    }
	    fflush(pout);
	}
    }
    
    /*     saveRunInfo(argc, argv, eltsConverged, areaConverged); */ /* Not very useful. AC Jul 91 */

    updateOct(direction);
    
    /*
     *  Compute and save the critical path for the direction orthogonal to the
     *  last spacing direction.
     */
    /********
    OrthogonalCPFlag = FALSE;
    if (OrthogonalCPFlag)  {
	if (direction == HORIZONTAL)  {
	    findCP(VERTICAL);
	    saveCP(VERTICAL);
	} else  {
	    findCP(HORIZONTAL);
	    saveCP(HORIZONTAL);
	}
    }
    ********/
    
    saveOct();
    octEnd();
    
    createInterface(&WorkingFacet);
    
    VOVend(0);			/* This calls exit(0). */
}



createInterface(facet)
    octObject *facet;
{
    char command[1024];
    
    sprintf( command, "vulcan -k %s:%s",
	    facet->contents.facet.cell,
	    facet->contents.facet.view );

    if (TraceFlag)  {
	fprintf(pout, "Creating interface facet: %s\n", command);
    }
    if ( util_csystem( command ) == 127 ) {
	errRaise( "SPARCS", 1, "cannot execute %s", command );
    }
    
}
