#include "copyright.h"
#include "sparcs.h"
#include "rpc.h"
#include "facet.h"

#define MSG_EXIT(string) \
    octError(string); \
    vemMessage(string, MSG_DISP); \
    vemPrompt(); \
    return(RPC_OK)

/*
 *  Routine to add a border constraint to a facet.  The command list (cmdList)
 *  should contain an object arg (a bag with one or more instances), followed
 *  by a text arg specifying the border.
 *
 *  2/25/88
 *  last modified 2/29/88
 */

int
addBorderConst(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct RPCArg *parg;
    lsGen gen;
    lsStatus status;
    int args;

    octObject inst;
    octObject facet;
    octObject argBag;
    octObject firstBag;
    octObject secondBag;
    octObject constBag;
    octObject minProp, maxProp;
    char buffer[SMALLBUF];
    char *pborderName;

    octGenerator instGenId;
    octStatus instGenStatus;

    inst.type = OCT_INSTANCE;
    facet.type = OCT_FACET;
    argBag.type = OCT_BAG;
    firstBag.type = OCT_BAG;
    secondBag.type = OCT_BAG;
    constBag.type = OCT_BAG;
    minProp.type = maxProp.type = OCT_PROP;

    constBag.contents.bag.name = "CONSTRAINT";
    minProp.contents.prop.name = "MIN";
    minProp.contents.prop.type = OCT_INTEGER;
    minProp.contents.prop.value.integer = 0;
    maxProp.contents.prop.name = "MAX";
    maxProp.contents.prop.type = OCT_INTEGER;
    maxProp.contents.prop.value.integer = 0;

    args = lsLength(cmdList);
    if (args !=  2)  goto usage;

    /*
     *  The first item in cmdList should be a bag of instances.
     */
    gen = lsStart(cmdList);

    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_OBJ_ARG)  goto usage;
    } else  {
	vemMessage("can't get first arg.\n", MSG_DISP);
	return(RPC_OK);
    }
	
    if (octIdsEqual(parg->argData.objArg.facet, WorkingFacet.objectId))  {
	MSG_EXIT("cursor is in the wrong window");
    }
    argBag.objectId = parg->argData.objArg.theBag;

    /* printf("argBag.objectId = %d\n", argBag.objectId); */

    /*
     *  The second arg in cmdList should be a  string, whose value must be
     *  "left", "right", "top", or "bottom".
     */
    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_TEXT_ARG)  goto usage;
    } else  {
	vemMessage("can't get second arg");
	return(RPC_OK);
    }
    if (strcmp(parg->argData.string, "left") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "LEFT";
    } else if (strcmp(parg->argData.string, "right") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "RIGHT";
    } else if (strcmp(parg->argData.string, "top") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "TOP";
    } else if (strcmp(parg->argData.string, "bottom") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "BOTTOM";
    } else  {
	sprintf(buffer, "border \"%s\" not legal\n",
		parg->argData.string);
	vemMessage(buffer, MSG_DISP);
	goto usage;
    }

    /*
     *  Get or create the top-level bag (e.g., "XCONSTRAINTS") and the second-
     *  level bag (e.g., "RIGHT").
     */
    facet.objectId = spot->facet;
    if (octGetOrCreate(&facet, &firstBag) != OCT_OK)  {
	sprintf(buffer, "can't get or create %s bag",
		firstBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    if (octGetOrCreate(&firstBag, &secondBag) != OCT_OK)  {
	sprintf(buffer, "can't get or create %s bag",
		secondBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    
    /*
     *  Generate the instances off of the bag and create a constraint
     *  for each.
     */
    if (octInitGenContents(&argBag, OCT_ALL_MASK, &instGenId) != OCT_OK)  {
	MSG_EXIT("can't begin instance generator");
    }
    while ((instGenStatus = octGenerate(&instGenId, &inst)) == OCT_OK)  {
	if (inst.type != OCT_INSTANCE)  {
	    vemMessage("\nskipping object which isn't an instance", MSG_DISP);
	    continue;
	}
	sprintf(buffer, "\nconstraining inst \"%s\" of %s:%s",
		inst.contents.instance.name,
		inst.contents.instance.master,
		inst.contents.instance.view);
	vemMessage(buffer, MSG_DISP);

	if (hasBorderConst(&inst, firstBag.contents.bag.name, &pborderName))  {
	    sprintf(buffer,
		    "\n\"%s\" already constrained to %s border; skipping",
		    inst.contents.instance.name, pborderName);
	    vemMessage(buffer, MSG_DISP);
	    continue;
	}	    
	
	if (octCreate(&secondBag, &constBag) != OCT_OK)  {
	    MSG_EXIT("trying to create CONSTRAINT bag");
	}
	if (octCreate(&constBag, &minProp) != OCT_OK)  {
	    MSG_EXIT("trying to create MIN property");
	}
	if (octCreate(&constBag, &maxProp) != OCT_OK)  {
	    MSG_EXIT("trying to create MAX property");
	}
	if (octAttach(&constBag, &inst) != OCT_OK)  {
	    MSG_EXIT("can't attach instance to bag.");
	}
    }
    vemMessage("\n", MSG_DISP);
    vemPrompt();
    if (instGenStatus != OCT_GEN_DONE)  {
	MSG_EXIT("bad exit from instance generator.");
    }
	

    if ((status=lsNext(gen, (Pointer *)&parg, 0)) != LS_NOMORE)  {
	vemMessage("can't end args gen for some reason.\n", MSG_DISP);
    }	
    return(RPC_OK);

usage:
    vemMessage("usage: instance(s) border; ", MSG_DISP);
    vemMessage("border is \"left\", \"right\", \"top\", or \"bottom\"\n",
	       MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}



/*
 *  Routine to determine whether the given instance is in a border constraint
 *  in the direction given by firstBagName (XCONSTRAINTS or YCONSTRAINTS).
 *  If it is, TRUE is returned along with the name of the border it is
 *  constrained to (e.g., RIGHT, TOP).  If not, FALSE is returned.
 *
 *  2/26/88
 *  last modified 2/26/88
 */

int
hasBorderConst(pinst, firstBagName, borderName)
struct octObject *pinst;
char *firstBagName;
char **borderName;	    /* name of border pinst is const. to; returned */
{
    octObject constBag;
    octObject borderBag;
    octGenerator constGenId, borderGenId;
    octStatus constGenStat, borderGenStat;
    
    constBag.type = borderBag.type = OCT_BAG;
    
    if (strcmp(firstBagName, "XCONSTRAINTS") == 0)  {
	/*
	 *  Look for containing bags called CONSTRAINT.
	 */
	if (octInitGenContainers(pinst, OCT_BAG_MASK, &constGenId) != OCT_OK) {
	    MSG_EXIT("hasBorderConst: can't begin CONSTRAINT bag generator");
	}
	while ((constGenStat=octGenerate(&constGenId, &constBag)) == OCT_OK)  {
	    /*
	     *  For each CONSTRAINT bag, see if it is contained by a LEFT or
	     *  RIGHT bag.  If so, pinst is already in an X border constraint
	     *  and should not be put into another.
	     */
	    if (octInitGenContainers(&constBag, OCT_BAG_MASK,
				     &borderGenId) != OCT_OK)  {
		MSG_EXIT("hasBorderConst: can't begin dir. bag generator");
	    }
	    while ((borderGenStat=octGenerate(&borderGenId,
					      &borderBag)) == OCT_OK)  {
		if (strcmp(borderBag.contents.bag.name, "RIGHT") == 0)  {
		    *borderName = borderBag.contents.bag.name;
		    return(TRUE);
		} else if (strcmp(borderBag.contents.bag.name, "LEFT") == 0)  {
		    *borderName = borderBag.contents.bag.name;
		    return(TRUE);
		}
	    }
	    if (borderGenStat != OCT_GEN_DONE)  {
		MSG_EXIT("hasBorderConst: bad exit from dir. bag generator.");
	    }
	}
	if (constGenStat != OCT_GEN_DONE)  {
	    MSG_EXIT("hasBorderConst: bad exit from CONST. bag generator.");
	}
    } else if (strcmp(firstBagName, "YCONSTRAINTS") == 0)  {
	if (octInitGenContainers(pinst, OCT_BAG_MASK, &constGenId) != OCT_OK) {
	    MSG_EXIT("hasBorderConst: can't begin CONSTRAINT bag generator");
	}
	while ((constGenStat=octGenerate(&constGenId, &constBag)) == OCT_OK)  {
	    if (octInitGenContainers(&constBag, OCT_BAG_MASK,
				     &borderGenId) != OCT_OK)  {
		MSG_EXIT("can't begin instance generator");
	    }
	    while ((borderGenStat=octGenerate(&borderGenId,
					      &borderBag)) == OCT_OK)  {
		if (strcmp(borderBag.contents.bag.name, "BOTTOM") == 0)  {
		    *borderName = borderBag.contents.bag.name;
		    return(TRUE);
		} else if (strcmp(borderBag.contents.bag.name, "TOP") == 0)  {
		    *borderName = borderBag.contents.bag.name;
		    return(TRUE);
		}
	    }
	    if (borderGenStat != OCT_GEN_DONE)  {
		MSG_EXIT("hasBorderConst: bad exit from dir. bag generator.");
	    }
	}
	if (constGenStat != OCT_GEN_DONE)  {
	    MSG_EXIT("hasBorderConst: bad exit from CONST. bag generator.");
	}
    } else  {
	MSG_EXIT("hasBorderConst: bad firstBagName argument");
    }
    return(FALSE);
}
