#include "copyright.h"
#include "sparcs.h"
#include "rpc.h"
#include "facet.h"

#define MSG_EXIT(string) \
    octError(string); \
    vemMessage(string, MSG_DISP); \
    vemPrompt(); \
    return(RPC_OK)

/*
 *  Routine to delete a border constraint.  The command list (cmdList) should
 *  should contain an object arg (a bag with one or more instances), followed
 *  by a text arg specifying the border.
 *
 *  2/27/88
 *  last modified 2/27/88
 */

int
deleteBorderConsts(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct RPCArg *parg;
    lsGen gen;
    lsStatus status;

    octObject inst;
    octObject facet;
    octObject argBag;
    octObject firstBag;
    octObject secondBag;
    octObject constBag;
    octObject minProp, maxProp;
    char buffer[SMALLBUF];
    char *pborderName;

    octGenerator instGenId;
    octStatus instGenStatus;

    inst.type = OCT_INSTANCE;
    facet.type = OCT_FACET;
    argBag.type = OCT_BAG;
    firstBag.type = OCT_BAG;
    secondBag.type = OCT_BAG;
    constBag.type = OCT_BAG;
    minProp.type = maxProp.type = OCT_PROP;

    constBag.contents.bag.name = "CONSTRAINT";
    minProp.contents.prop.name = "MIN";
    minProp.contents.prop.type = OCT_INTEGER;
    minProp.contents.prop.value.integer = 0;
    maxProp.contents.prop.name = "MAX";
    maxProp.contents.prop.type = OCT_INTEGER;
    maxProp.contents.prop.value.integer = 0;

    if (lsLength(cmdList) != 2)  goto usage;

    /*
     *  The first item in cmdList should be a bag of instances.
     */
    gen = lsStart(cmdList);

    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_OBJ_ARG)  goto usage;
    } else  {
	vemMessage("can't get first arg.\n", MSG_DISP);
	return(RPC_OK);
    }
	
    if (octIdsEqual(parg->argData.objArg.facet, WorkingFacet.objectId))  {
	MSG_EXIT("cursor is in the wrong window");
    }
    argBag.objectId = parg->argData.objArg.theBag;

    /*
     *  The second arg in cmdList should be a  string, whose value must be
     *  "left", "right", "top", or "bottom".
     */
    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_TEXT_ARG)  goto usage;
    } else  {
	vemMessage("can't get second arg");
	return(RPC_OK);
    }
    if (strcmp(parg->argData.string, "left") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "LEFT";
    } else if (strcmp(parg->argData.string, "right") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "RIGHT";
    } else if (strcmp(parg->argData.string, "top") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "TOP";
    } else if (strcmp(parg->argData.string, "bottom") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "BOTTOM";
    } else  {
	sprintf(buffer, "border \"%s\" not legal\n",
		parg->argData.string);
	vemMessage(buffer, MSG_DISP);
	goto usage;
    }

    /*
     *  Get or create the top-level bag (e.g., "XCONSTRAINTS") and the second-
     *  level bag (e.g., "RIGHT").
     */
    facet.objectId = spot->facet;
    if (octGetOrCreate(&facet, &firstBag) != OCT_OK)  {
	sprintf(buffer, "can't get or create %s bag",
		firstBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    if (octGetOrCreate(&firstBag, &secondBag) != OCT_OK)  {
	sprintf(buffer, "can't get or create %s bag",
		secondBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    
    /*
     *  Generate the instances off of the bag and delete the a constraint
     *  for each.
     */
    if (octInitGenContents(&argBag, OCT_ALL_MASK, &instGenId) != OCT_OK)  {
	MSG_EXIT("can't begin instance generator");
    }
    while ((instGenStatus = octGenerate(&instGenId, &inst)) == OCT_OK)  {
	if (inst.type != OCT_INSTANCE)  {
	    vemMessage("\nskipping object which isn't an instance", MSG_DISP);
	}
	sprintf(buffer, "\nconstraining inst \"%s\" of %s:%s",
		inst.contents.instance.name,
		inst.contents.instance.master,
		inst.contents.instance.view);
	vemMessage(buffer, MSG_DISP);

	if (hasBorderConst(&inst, firstBag.contents.bag.name, &pborderName))  {
	    sprintf(buffer,
		    "\n\"%s\" already constrained to %s border; skipping",
		    inst.contents.instance.name, pborderName);
	    vemMessage(buffer, MSG_DISP);
	    continue;
	}	    
	
	if (octCreate(&secondBag, &constBag) != OCT_OK)  {
	    MSG_EXIT("trying to create CONSTRAINT bag");
	}
	if (octCreate(&constBag, &minProp) != OCT_OK)  {
	    MSG_EXIT("trying to create MIN property");
	}
	if (octCreate(&constBag, &maxProp) != OCT_OK)  {
	    MSG_EXIT("trying to create MAX property");
	}
	if (octAttach(&constBag, &inst) != OCT_OK)  {
	    MSG_EXIT("can't attach instance to bag.");
	}
    }
    vemMessage("\n", MSG_DISP);
    vemPrompt();
    if (instGenStatus != OCT_GEN_DONE)  {
	MSG_EXIT("bad exit from instance generator.");
    }
	

    if ((status=lsNext(gen, (Pointer *)&parg, 0)) != LS_NOMORE)  {
	vemMessage("can't end args gen for some reason.\n", MSG_DISP);
    }	
    return(RPC_OK);

usage:
    vemMessage("usage: instance(s) border; ", MSG_DISP);
    vemMessage("border is \"left\", \"right\", \"top\", or \"bottom\"\n",
	       MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}



/*
 *  Routine to delete all of the border constraints from the specified border.
 *  The command list (cmdList) should contain a text arg specifying the border.
 *
 *  2/27/88
 *  last modified 2/27/88
 */

int
deleteAllBorderConsts(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct RPCArg *parg;
    lsGen gen;

    octStatus status;
    octObject inst;
    octObject facet;
    octObject firstBag;
    octObject secondBag;
    octObject constBag;
    octObject minProp, maxProp;
    char buffer[SMALLBUF];
    char *pborderName;

    octGenerator bagGenId;
    octStatus bagGenStatus;

    inst.type = OCT_INSTANCE;
    facet.type = OCT_FACET;
    firstBag.type = OCT_BAG;
    secondBag.type = OCT_BAG;
    constBag.type = OCT_BAG;
    minProp.type = maxProp.type = OCT_PROP;

    constBag.contents.bag.name = "CONSTRAINT";
    minProp.contents.prop.name = "MIN";
    maxProp.contents.prop.name = "MAX";

    if (lsLength(cmdList) != 1)  goto usage;

    gen = lsStart(cmdList);
    if (lsNext(gen, (Pointer *)&parg, 0) == LS_OK)  {
	if (parg->argType != VEM_TEXT_ARG)  goto usage;
    } else  {
	vemMessage("can't get arg");
	return(RPC_OK);
    }
    if (strcmp(parg->argData.string, "left") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "LEFT";
    } else if (strcmp(parg->argData.string, "right") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "RIGHT";
    } else if (strcmp(parg->argData.string, "top") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "TOP";
    } else if (strcmp(parg->argData.string, "bottom") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "BOTTOM";
    } else  {
	sprintf(buffer, "border \"%s\" not legal\n",
		parg->argData.string);
	vemMessage(buffer, MSG_DISP);
	goto usage;
    }

    /*
     *  Get the top-level bag (e.g., "XCONSTRAINTS") and the second-level
     *  bag (e.g., "RIGHT").
     */
    facet.objectId = spot->facet;
    if (octGetByName(&facet, &firstBag) != OCT_OK)  {
	sprintf(buffer, "can't get %s bag",
		firstBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    if (octGetByName(&firstBag, &secondBag) != OCT_OK)  {
	sprintf(buffer, "can't get %s bag",
		secondBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
    
    /*
     *  Generate each CONSTRAINT bag off of secondBag.
     */
    if (octInitGenContents(&secondBag, OCT_BAG_MASK, &bagGenId) != OCT_OK)  {
	MSG_EXIT("can't begin CONSTRAINT bag generator");
    }
    while ((bagGenStatus = octGenerate(&bagGenId, &constBag)) == OCT_OK)  {
	/*
	 *  Grab the instance and detach it from constBag.
	 */
	if (octGenFirstContent(&constBag, OCT_INSTANCE_MASK, &inst)!=OCT_OK) {
	    MSG_EXIT("can't get inst from CONSTRAINT bag\n");
	} else  {
	    sprintf(buffer, "\ndeleting constraint for inst \"%s\" of %s:%s",
		    inst.contents.instance.name,
		    inst.contents.instance.master,
		    inst.contents.instance.view);
	    vemMessage(buffer, MSG_DISP);
	}
	if (octDetach(&constBag, &inst) != OCT_OK)  {
	    MSG_EXIT("can't detach inst from CONSTRAINT bag\n");
	}

	/*
	 *  Delete the properties.
	 */
	status = octGetByName(&constBag, &minProp);
	if (status == OCT_OK)  {
	    if (octDelete(&minProp) != OCT_OK)  {
		MSG_EXIT("can't delete MIN prop\n");
	    }
	} else if (status != OCT_NOT_FOUND)  {
	    MSG_EXIT("can't get MIN prop from CONSTRAINT bag\n");
	}
	status = octGetByName(&constBag, &maxProp);
	if (status == OCT_OK)  {
	    if (octDelete(&maxProp) != OCT_OK)  {
		MSG_EXIT("can't delete MAX prop\n");
	    }
	} else if (status == OCT_NOT_FOUND)  {
	    MSG_EXIT("can't get MAX prop from CONSTRAINT bag\n");
	}

	/*
	 *  Now get rid of the CONSTRAINT bag itself.
	 */
	if (octDelete(&constBag) != OCT_OK)  {
	    MSG_EXIT("trying to delete CONSTRAINT bag");
	}
    }
    vemMessage("\n", MSG_DISP);
    vemPrompt();
    if (bagGenStatus != OCT_GEN_DONE)  {
	MSG_EXIT("bad exit from bag generator.");
    }
	
    /*
     *  The secondBag (e.g., "TOP") is no longer needed.
     */
    if (octDelete(&secondBag) != OCT_OK)  {
	sprintf(buffer, "can't delete %s bag", secondBag.contents.bag.name);
	MSG_EXIT(buffer);
    }
	
    /*
     *  The firstBag (e.g., "XCONSTRAINTS") is needed iff it still contains
     *  something.
     */
    status = octGenFirstContent(&firstBag, OCT_BAG_MASK, &constBag);
    if (status == OCT_GEN_DONE)  {
	if (octDelete(&firstBag) != OCT_OK)  {
	    sprintf(buffer, "can't delete %s bag", firstBag.contents.bag.name);
	    MSG_EXIT(buffer);
	}
    } else if (status != OCT_OK)  {
	MSG_EXIT("can't genFirst from firstBag\n");
    }
    return(RPC_OK);

usage:
    vemMessage("usage: border; ", MSG_DISP);
    vemMessage("border is \"left\", \"right\", \"top\", or \"bottom\"\n",
	       MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}
