#include "copyright.h"
#include "sparcs.h"
#include "rpc.h"
#include "facet.h"

#define MSG_EXIT(string) \
    octError(string); \
    vemMessage(string, MSG_DISP); \
    vemPrompt(); \
    return(RPC_OK)

vemSelSet leftSet = (vemSelSet)0;
vemSelSet rightSet = (vemSelSet)0;
vemSelSet topSet = (vemSelSet)0;
vemSelSet bottomSet = (vemSelSet)0;

/*
 *  Routine to display the instances constrained to a particular border of
 *  the layout.  The command list (cmdList) should contain a text arg
 *  specifying the border.
 *
 *  2/26/88
 *  last modified 3/6/88
 */

int
showBorderConsts(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct RPCArg *parg;
    lsGen gen;
    lsStatus status;

    octObject facet;
    octObject firstBag;
    octObject secondBag;
    vemSelSet selSet;
    char buffer[SMALLBUF];

    int getConsts();

    facet.type = OCT_FACET;
    firstBag.type = OCT_BAG;
    secondBag.type = OCT_BAG;

    if (lsLength(cmdList) != 1)  goto usage;

    facet.objectId = spot->facet;

    /*
     *  The arg in cmdList should be a  string, whose value must be "left",
     *  "right", "top", or "bottom".
     */
    gen = lsStart(cmdList);
    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_TEXT_ARG)  goto usage;
    } else  {
	vemMessage("can't get arg");
	return(RPC_OK);
    }
    if (strcmp(parg->argData.string, "left") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "LEFT";
	if (leftSet == (vemSelSet)0)  {
	    leftSet = vemNewSelSet(spot->facet, 65535, 65535, 65535,
				   1, 1, 4, 4, "1000 0100 0010 0001");
	    if (getConsts(&facet, &firstBag, &secondBag, leftSet) == FALSE)  {
		leftSet = (vemSelSet)0;
	    }
	} else  {
	    vemMessage("insts constrained to left border already selected\n",
		       MSG_DISP);
	    vemPrompt();
	}
	selSet = leftSet;
    } else if (strcmp(parg->argData.string, "right") == 0)  {
	firstBag.contents.bag.name = "XCONSTRAINTS";
	secondBag.contents.bag.name = "RIGHT";
	if (rightSet == (vemSelSet)0)  {
	    rightSet = vemNewSelSet(spot->facet, 65535, 65535, 0,
				    1, 1, 4, 4, "1000 0100 0010 0001");
	    if (getConsts(&facet, &firstBag, &secondBag, rightSet) == FALSE)  {
		rightSet = (vemSelSet)0;
	    }
	} else  {
	    vemMessage("insts constrained to right border already selected\n",
		       MSG_DISP);
	    vemPrompt();
	}
	selSet = rightSet;
    } else if (strcmp(parg->argData.string, "top") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "TOP"; 
	if (topSet == (vemSelSet)0)  {
	    topSet = vemNewSelSet(spot->facet, 0, 65535, 65535,
				  1, 1, 4, 4, "1000 0100 0010 0001");
	    if (getConsts(&facet, &firstBag, &secondBag, topSet) == FALSE)  {
		topSet = (vemSelSet)0;
	    }
	} else  {
	    vemMessage("insts constrained to top border already selected\n",
		       MSG_DISP);
	    vemPrompt();
	}
	selSet = topSet;
    } else if (strcmp(parg->argData.string, "bottom") == 0)  {
	firstBag.contents.bag.name = "YCONSTRAINTS";
	secondBag.contents.bag.name = "BOTTOM";
	if (bottomSet == (vemSelSet)0)  {
	    bottomSet = vemNewSelSet(spot->facet, 0, 65535, 32767,
				     1, 1, 4, 4, "1000 0100 0010 0001");
	    if (getConsts(&facet, &firstBag, &secondBag, bottomSet) == FALSE) {
		bottomSet = (vemSelSet)0;
	    }
	} else  {
	    vemMessage("insts constrained to bottom border already selected\n",
		       MSG_DISP);
	    vemPrompt();
	}
	selSet = bottomSet;
    } else  {
	sprintf(buffer, "border \"%s\" not legal\n", parg->argData.string);
	vemMessage(buffer, MSG_DISP);
	goto usage;
    }
    return(RPC_OK);

usage:
    vemMessage("usage:  border; ", MSG_DISP);
    vemMessage("border is \"left\", \"right\", \"top\", or \"bottom\"\n",
	       MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}




int
getConsts(pfacet, pfirstBag, psecondBag, selSet)
octObject *pfacet;
octObject *pfirstBag;
octObject *psecondBag;
vemSelSet selSet;
{
    octObject constBag;
    octGenerator bagGenId;
    octStatus bagGenStat;
    octStatus status;
    char buffer[SMALLBUF];
    int constraints;

    constBag.type = OCT_BAG;
    /*
     *  Get the top-level bag (e.g., "XCONSTRAINTS") and the second-level
     *  bag (e.g., "RIGHT").
     */
    if ((status=octGetByName(pfacet, pfirstBag)) == OCT_OK)  {
	if ((status=octGetByName(pfirstBag, psecondBag)) == OCT_OK)  {
	    /*
	     *  Count them up.
	     */
	    constraints = 0;
	    if (octInitGenContents(psecondBag,
				   OCT_BAG_MASK, &bagGenId) != OCT_OK)  {
		MSG_EXIT("can't begin constraint generator");
	    }
	    while ((bagGenStat=octGenerate(&bagGenId, &constBag)) == OCT_OK)  {
		constraints++;
	    }
	    if (bagGenStat != OCT_GEN_DONE)  {
		MSG_EXIT("bad exit from constraint generator.");
	    }
	} else if (status == OCT_NOT_FOUND)  {
	    sprintf(buffer, "no constraints to %s border\n",
		    psecondBag->contents.bag.name);
	    vemMessage(buffer, MSG_DISP);
	    vemPrompt();
	    return(FALSE);
	} else  {
	    sprintf(buffer, "can't get %s bag", psecondBag->contents.bag.name);
	    MSG_EXIT(buffer);
	}
    } else if (status == OCT_NOT_FOUND)  {
	sprintf(buffer, "no constraints to %s border\n",
		psecondBag->contents.bag.name);
	vemMessage(buffer, MSG_DISP);
	vemPrompt();
	return(FALSE);
    } else  {
	sprintf(buffer, "can't get %s bag\n", pfirstBag->contents.bag.name);
	MSG_EXIT(buffer);
    }
    
    /*
     *  Select the stuff in the bag.
     */
    if (vemAddSelSet(selSet, psecondBag->objectId) != VEM_OK)  {
	MSG_EXIT("can't select bag for some reason\n");
    } else  {
	sprintf(buffer, "%d constraints\n", constraints);
	vemMessage(buffer, MSG_DISP);
	vemPrompt();
    }
    return(TRUE);
}




int
clearBorderConsts(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct RPCArg *parg;
    lsGen gen;
    lsStatus status;
    char buffer[SMALLBUF];

    if (lsLength(cmdList) != 1)  goto usage;

    /*
     *  The arg in cmdList should be a  string, whose value must be "left",
     *  "right", "top", or "bottom".
     */
    gen = lsStart(cmdList);
    if ((status = lsNext(gen, (Pointer *)&parg, 0)) == LS_OK)  {
	if (parg->argType != VEM_TEXT_ARG)  goto usage;
    } else  {
	vemMessage("can't get arg");
	return(RPC_OK);
    }
    
    if (strcmp(parg->argData.string, "left") == 0)  {
	if (leftSet != (vemSelSet)0)  {
	    if (vemFreeSelSet(leftSet) != VEM_OK)  {
		vemMessage("can't free selected set for some reason\n",
			   MSG_DISP);
		vemPrompt();
	    }
	    leftSet = (vemSelSet)0;
	}
    } else if (strcmp(parg->argData.string, "right") == 0)  {
	if (rightSet != (vemSelSet)0)  {
	    if (vemFreeSelSet(rightSet) != VEM_OK)  {
		vemMessage("can't free selected set for some reason\n",
			   MSG_DISP);
		vemPrompt();
	    }
	    rightSet = (vemSelSet)0;
	}
    } else if (strcmp(parg->argData.string, "top") == 0)  {
	if (topSet != (vemSelSet)0)  {
	    if (vemFreeSelSet(topSet) != VEM_OK)  {
		vemMessage("can't free selected set for some reason\n",
			   MSG_DISP);
		vemPrompt();
	    }
	    topSet = (vemSelSet)0;
	}
    } else if (strcmp(parg->argData.string, "bottom") == 0)  {
	if (bottomSet != (vemSelSet)0)  {
	    if (vemFreeSelSet(bottomSet) != VEM_OK)  {
		vemMessage("can't free selected set for some reason\n",
			   MSG_DISP);
		vemPrompt();
	    }
	    bottomSet = (vemSelSet)0;
	}
    } else  {
	sprintf(buffer, "border \"%s\" not legal\n", parg->argData.string);
	vemMessage(buffer, MSG_DISP);
	goto usage;
    }
    return(RPC_OK);

usage:
    vemMessage("usage:  border; ", MSG_DISP);
    vemMessage("border is \"left\", \"right\", \"top\", or \"bottom\"\n",
	       MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}




int
clearAllBorderConsts(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    if (lsLength(cmdList) != 0)  goto usage;

    if (leftSet != (vemSelSet)0)  {
	if (vemFreeSelSet(leftSet) != VEM_OK)  {
	    vemMessage("can't free left selected set for some reason\n",
		       MSG_DISP);
	    vemPrompt();
	}
	leftSet = (vemSelSet)0;
    }
    if (rightSet != (vemSelSet)0)  {
	if (vemFreeSelSet(rightSet) != VEM_OK)  {
	    vemMessage("can't free right selected set for some reason\n",
		       MSG_DISP);
	    vemPrompt();
	}
	rightSet = (vemSelSet)0;
    }
    if (topSet != (vemSelSet)0)  {
	if (vemFreeSelSet(topSet) != VEM_OK)  {
	    vemMessage("can't free top selected set for some reason\n",
		       MSG_DISP);
	    vemPrompt();
	}
	topSet = (vemSelSet)0;
    }
    if (bottomSet != (vemSelSet)0)  {
	if (vemFreeSelSet(bottomSet) != VEM_OK)  {
	    vemMessage("can't free bottom selected set for some reason\n",
		       MSG_DISP);
	    vemPrompt();
	}
	bottomSet = (vemSelSet)0;
    }
    return(RPC_OK);

usage:
    vemMessage("usage:  : clear-all\n", MSG_DISP);
    vemPrompt();
    return(RPC_OK);
}
