#include "copyright.h"
#include "sparcs.h"
#include "rpc.h"

/*
 *  Main routine for the oct version of the sparcs program, with rpc.
 *
 *  Modified 10/8/87 for vem_5/rpc_5.
 *
 *  5/28/86
 *  last modified 2/12/88
 */

struct RPCSpot outSpot;	/* spot for the spaced view */
char mesgBuf[1024];

vemSelSet vCritical;
vemSelSet vNonCritical;

int
sparcs(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    void initRpcOct();
    void allocGraph();
    void updateOct();
    int doSpacing();
    int converged = FALSE;
    octObject inFacet;		/* input facet, corresp. to spot */
    struct octBox inFacetBB;	/* original bb, before spacing */
    octObject outFacet;		/* for the spaced view */
    struct octBox outFacetBB;	/* to redraw window during/after spacing */
    octStatus oStatus;
    directionType direction;
    int width, height;
    int area, lastArea;
    int eltsConverged = FALSE;
    int areaConverged = FALSE;
    
    int h = 1;		/*  number of horizontal spacings  */
    int v = 1;		/*  number of vertical spacings  */
    

    /*
     *  Get the facet where sparcs was invoked, and fill in the globals that
     *  would have been done by sp_commandLine().
     */
    inFacet.objectId = spot->facet;
    if (octGetById(&inFacet) != OCT_OK)  {
	sprintf(mesgBuf, "can't getById facet %s of %s/%s; fatal.\n",
		inFacet.contents.facet.facet, inFacet.contents.facet.cell,
		inFacet.contents.facet.view);
	vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
	RPCExit(-1);
    }
    /*
     *  Get the bb of inFacet; this is used below to redraw the window on
     *  outFacet.
     */
    if (octBB(&inFacet, &inFacetBB) != OCT_OK)  {
	sprintf(mesgBuf, "can't bounding box of %s/%s/%s; fatal.\n",
		inFacet.contents.facet.cell, inFacet.contents.facet.view,
		inFacet.contents.facet.facet);
	vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
	RPCExit(-1);
    }

    InCell = STRSAVE(inFacet.contents.facet.cell);
    InView = STRSAVE(inFacet.contents.facet.view);
    if (OutCell == NIL(char))  OutCell = STRSAVE(inFacet.contents.facet.cell);
    if (OutView == NIL(char))  OutView = STRSAVE("spaced");

    sprintf(mesgBuf, "2spacing %s/%s/%s, with octId %s0\n",
	    inFacet.contents.facet.cell, inFacet.contents.facet.view,
	    inFacet.contents.facet.facet, octFormatId(inFacet.objectId));
    vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
    vemPrompt();

    initRpcOct(spot, &outSpot);

    outFacet.objectId = outSpot.facet;
    if (octGetById(&outFacet) != OCT_OK)  {
	sprintf(mesgBuf, "can't getById facet %s of %s/%s; fatal.\n",
		outFacet.contents.facet.facet, outFacet.contents.facet.cell,
		outFacet.contents.facet.view);
	vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
	RPCExit(-1);
    }

    /*
     *  At this point the layout has been read in from oct.  The total number
     *  of elements is known, so the graph can be allocated.
     */
    allocGraph();

    width = (MaxBBX-MinBBX)/UNITSPERLAMBDA;
    height = (MaxBBY-MinBBY)/UNITSPERLAMBDA;
    area = width * height;
    if (TraceFlag)  {
	fprintf(pout, "\n");
	printf("Initial horizontal extent:  %d;  vertical extent:  %d;  ",
	       width, height);
	printf("area:  %d.\n", area);
	fflush(pout);
    }
    
    if (HorizontalFirstFlag)  {
	goto horiz;
    } else  {
	goto vert;
    }

    for (;;)  {

horiz:
	direction = HORIZONTAL;
	if (TraceFlag)  {
	    fprintf(pout, "\nHorizontal spacing iteration %d:\n", h++);
	    fflush(pout);
	}
	eltsConverged = doSpacing(direction);
	Iterations++;
	
	lastArea = area;
	area = (MaxBBX-MinBBX)/UNITSPERLAMBDA * (MaxBBY-MinBBY)/UNITSPERLAMBDA;
	if (TraceFlag)  {
	    printf("    New horizontal extent:  %d;  (vertical extent:  %d);",
		   (MaxBBX-MinBBX)/UNITSPERLAMBDA,
		   (MaxBBY-MinBBY)/UNITSPERLAMBDA);
	    printf("  New area:  %d.\n", area);
	    fflush(pout);
	}
	
        if (IterationLimit > 0)  {
	    if (Iterations == IterationLimit)  break;
	}
	if (Iterations > 1)  {
	    if (StopCriterion == 'A')  {
		if (lastArea == area)  {
		    areaConverged = TRUE;
		    break;
		}
	    } else if (StopCriterion == 'E')  {
		if (eltsConverged == TRUE)  break;
	    }
	}
	if (ShowIterationsFlag)  {
	    if (octBB(&outFacet, &outFacetBB) != OCT_OK)  {
		sprintf(mesgBuf, "can't bounding box of %s/%s/%s; fatal.\n",
			outFacet.contents.facet.cell,
			outFacet.contents.facet.view,
			outFacet.contents.facet.facet);
		vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
		RPCExit(-1);
	    }
	    outFacetBB.lowerLeft.x = SPMIN(outFacetBB.lowerLeft.x,
					    inFacetBB.lowerLeft.x);
	    outFacetBB.lowerLeft.y = SPMIN(outFacetBB.lowerLeft.y,
					    inFacetBB.lowerLeft.y);
	    outFacetBB.upperRight.x = SPMAX(outFacetBB.upperRight.x,
					    inFacetBB.upperRight.x);
	    outFacetBB.upperRight.y = SPMAX(outFacetBB.upperRight.y,
					    inFacetBB.upperRight.y);
	    vemWnQRegion(&outFacet, &outFacetBB);
	    vemWnFlush();
	}

vert:
	direction = VERTICAL;
	if (TraceFlag)  {
	    fprintf(pout, "\nVertical spacing iteration %d:\n", v++);
	    fflush(pout);
	}
	eltsConverged = doSpacing(direction);
	Iterations++;
	
	lastArea = area;
	area = (MaxBBX-MinBBX)/UNITSPERLAMBDA * (MaxBBY-MinBBY)/UNITSPERLAMBDA;
	if (TraceFlag)  {
	    printf("    (horizontal extent:  %d);  New vertical extent:  %d;",
		   (MaxBBX-MinBBX)/UNITSPERLAMBDA,
		   (MaxBBY-MinBBY)/UNITSPERLAMBDA);
	    printf("  New area:  %d.\n", area);
	    fflush(pout);
	}
	
        if (IterationLimit > 0)  {
	    if (Iterations == IterationLimit)  break;
	}
	if (Iterations > 1)  {
	    if (StopCriterion == 'A')  {
		if (lastArea == area)  {
		    areaConverged = TRUE;
		    break;
		}
	    } else if (StopCriterion == 'E')  {
		if (eltsConverged == TRUE)  break;
	    }
	}
	if (ShowIterationsFlag)  {
	    if (octBB(&outFacet, &outFacetBB) != OCT_OK)  {
		sprintf(mesgBuf, "can't bounding box of %s/%s/%s; fatal.\n",
			outFacet.contents.facet.cell,
			outFacet.contents.facet.view,
			outFacet.contents.facet.facet);
		vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
		RPCExit(-1);
	    }
	    outFacetBB.lowerLeft.x = SPMIN(outFacetBB.lowerLeft.x,
					    inFacetBB.lowerLeft.x);
	    outFacetBB.lowerLeft.y = SPMIN(outFacetBB.lowerLeft.y,
					    inFacetBB.lowerLeft.y);
	    outFacetBB.upperRight.x = SPMAX(outFacetBB.upperRight.x,
					    inFacetBB.upperRight.x);
	    outFacetBB.upperRight.y = SPMAX(outFacetBB.upperRight.y,
					    inFacetBB.upperRight.y);
	    vemWnQRegion(&outFacet, &outFacetBB);
	    vemWnFlush();
	}
    }

    if (TraceFlag)  {
	if (StopCriterion == 'E')  {
	    /*
	     *  Stopped when no elements moved.
	     */
	    fprintf(pout, "\nElements converged after %d iterations.\n",
		    Iterations);
	    fflush(pout);
	} else if (StopCriterion == 'A')  {
	    /*
	     *  Stopped when area remained constant.
	     */
	    fprintf(pout, "\nArea converged after %d iterations;",	
		    Iterations);
	    if (eltsConverged)  {
	        fprintf(pout, "  elements converged.\n");
	    } else  {
	        fprintf(pout, "  elements NOT converged.\n");
	    }
	    fflush(pout);
	} else  {
	    /*
	     *  Stopped when IterationLimit reached.
	     */
	    fprintf(pout, "\nIteration limit reached");
	    fprintf(pout, " (%d iterations);\n", IterationLimit);
	    if (eltsConverged)  {
	        fprintf(pout, "Elements converged,");
	    } else  {
	        fprintf(pout, "Elements NOT converged,");
	    }
	    if (areaConverged)  {
	        fprintf(pout, "  area converged.\n");
	    } else  {
	        fprintf(pout, "  area NOT converged.\n");
	    }
	    fflush(pout);
	}
    }

    updateOct(direction);
    saveOct();
    
    /*
     *  Re-draw the window to show the results.  The redraw area is the max
     *  of the original bb versus the spaced bb.
     */
    if (octBB(&outFacet, &outFacetBB) != OCT_OK)  {
	sprintf(mesgBuf, "can't bounding box of %s/%s/%s; fatal.\n",
		outFacet.contents.facet.cell, outFacet.contents.facet.view,
		outFacet.contents.facet.facet);
	vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
	RPCExit(-1);
    }
    outFacetBB.lowerLeft.x = SPMIN(outFacetBB.lowerLeft.x,
				    inFacetBB.lowerLeft.x);
    outFacetBB.lowerLeft.y = SPMIN(outFacetBB.lowerLeft.y,
				    inFacetBB.lowerLeft.y);
    outFacetBB.upperRight.x = SPMAX(outFacetBB.upperRight.x,
				    inFacetBB.upperRight.x);
    outFacetBB.upperRight.y = SPMAX(outFacetBB.upperRight.y,
				    inFacetBB.upperRight.y);
    vemWnQRegion(&outFacet, &outFacetBB);
    vemWnFlush();

    sprintf(mesgBuf, "2Converged.\n0");
    vemMessage(mesgBuf, MSG_DISP|MSG_NOLOG);
    vemPrompt();

    fflush(pout);
    fflush(stdout);
    /* RPCExit(0); */

    return(RPC_OK);
}



int
showCritical(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct object *po;
    struct wire *pw;

    if (spot->facet != outSpot.facet)  {
	vemMessage("Must ask for the critical path of the spaced facet.\n", 
								MSG_DISP|MSG_NOLOG);
        return(RPC_OK);
    }

    if (!vCritical) {
	vCritical = vemNewSelSet(spot->facet, 65535, 65535, 0, 1, 1, 4, 4,
				    "1000 0100 0010 0001");
    }
    for (po = pObjectList; po != NIL(struct object); po=po->pnext)  {
	if (po->node.key != UNDEF)  {
	    if (po->node.cpm.lower == po->node.cpm.upper)  {
		vemAddSelSet(vCritical, po->instanceId);
	    }
	}
    }
    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
	if (pw->node.key != UNDEF)  {
	    if (pw->node.cpm.lower == pw->node.cpm.upper)  {
		vemAddSelSet(vCritical, pw->pathId);
	    }
	}
    }
    return(RPC_OK);
}


int
clearCritical(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    if (vCritical) {
	vemClearSelSet(vCritical);
    }
    return(RPC_OK);
}


int
showNonCritical(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    struct object *po;
    struct wire *pw;

    if (spot->facet != outSpot.facet)  {
	vemMessage("Must ask for noncritical elements of spaced facet.\n", 
								MSG_DISP|MSG_NOLOG);
    }

    if (!vNonCritical) {
	vNonCritical = vemNewSelSet(spot->facet, 65535, 65535, 65535, 1, 1, 8, 8,
				    "11000000 11000000 00000000 00000000\
				     00001100 00001100 00000000 00000000");
    }
    for (po=pObjectList; po!=NIL(struct object); po=po->pnext)  {
	if (po->node.key != UNDEF)  {
	    if (po->node.cpm.lower != po->node.cpm.upper)  {
		vemAddSelSet(vNonCritical, po->instanceId);
	    }
	}
    }
    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
	if (pw->node.key != UNDEF)  {
	    if (pw->node.cpm.lower != pw->node.cpm.upper)  {
		vemAddSelSet(vNonCritical, pw->pathId);
	    }
	}
    }
    return(RPC_OK);
}



int
clearNonCritical(spot, cmdList, userOptionWord)
struct RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    if (vNonCritical) {
	vemClearSelSet(vNonCritical);
    }
    return(RPC_OK);
}
