#include "copyright.h"
#include "sparcs.h"

/*
 *  These routines are for a linear doubly-linked list with pointers to the
 *  first and last elements in the list.
 *
 *  6/17/86
 *  last modified 2/28/87
 */


/*
 *  For use when scanning a list towards its TAIL.
 *
 *  Adds an element (pnew) to a doubly-linked list.  If phead==NIL, add it at
 *  the head; if pwhere==NIL, add it at the TAIL; otherwise add it BEFORE
 *  pwhere.
 *
 *  THIS SHOULD BE A MACRO.
 *
 *  6/17/86 (as dllAdd)
 *  2/18/87 (as dllPreAdd)
 *  last modified 2/28/87
 */

void
dllPreAdd(phead, ptail, pwhere, pnew)
struct edge **phead;	/* first element in dll */
struct edge **ptail;	/* last element in dll */
struct edge *pwhere;	/* add pnew AFTER pwhere */
struct edge *pnew;	/* new element */
{
    if (*phead == NIL(struct edge))  {		    /* add at head */
	*phead = pnew;
	*ptail = pnew;
	pnew->pnext = pnew->pprev = NIL(struct edge);
	return;
    } else if (pwhere == NIL(struct edge))  {	    /* add at tail */
	(*ptail)->pnext = pnew;
	pnew->pprev = *ptail;
	pnew->pnext = NIL(struct edge);
	*ptail = pnew;
    } else  {					    /* add BEFORE pwhere */
	pnew->pnext = pwhere;
	pnew->pprev = pwhere->pprev;
	pwhere->pprev = pnew;
	if (pnew->pprev != NIL(struct edge))  {
	    pnew->pprev->pnext = pnew;
	} else  {
	    *phead = pnew;
	}
    }
}


/*
 *  For use when scanning a list towards its HEAD.
 *
 *  Adds an element (pnew) to a doubly-linked list.  If phead==NIL, add it at
 *  the head; if pwhere==NIL, add it at the HEAD; otherwise add it AFTER
 *  pwhere.
 *
 *  THIS SHOULD BE A MACRO.
 *
 *  6/17/86 (as dllAdd)
 *  2/18/87 (as dllPostAdd)
 *  last modified 2/20/87
 */

void
dllPostAdd(phead, ptail, pwhere, pnew)
struct edge **phead;	/* first element in dll */
struct edge **ptail;	/* last element in dll */
struct edge *pwhere;	/* add pnew AFTER pwhere */
struct edge *pnew;	/* new element */
{
    if (*phead == NIL(struct edge))  {		    /* add at head */
	*phead = pnew;
	*ptail = pnew;
	return;
    } else if (pwhere == NIL(struct edge))  {	    /* also add at head */
	(*phead)->pprev = pnew;
	pnew->pnext = *phead;
	*phead = pnew;
    } else  {					    /* add AFTER pwhere */
	pnew->pnext = pwhere->pnext;
	pnew->pprev = pwhere;
	pwhere->pnext = pnew;
	if (pnew->pnext != NIL(struct edge))  {
	    pnew->pnext->pprev = pnew;
	} else  {
	    *ptail = pnew;
	}
    }
}


/*
 *  Delete an element from the doubly-linked list.
 *
 *  THIS SHOULD BE A MACRO.
 *
 *  6/17/86
 *  last modified 3/1/87
 */

void
dllDelete(phead, ptail, pedge)
struct edge **phead;
struct edge **ptail;
struct edge *pedge;
{
    if (pedge->pprev == NIL(struct edge))  {	/* if first element */
	*phead = pedge->pnext;
    } else  {					/* not first element */
	pedge->pprev->pnext = pedge->pnext;
    }
    if (pedge->pnext == NIL(struct edge))  {	/* if last element */
	*ptail = pedge->pprev;
    } else  {					/* not last element */
	pedge->pnext->pprev = pedge->pprev;
    }
/* XXX needed?? (in light of dllPreAdd() setting these... */
    pedge->pnext = pedge->pprev = NIL(struct edge);
    /* free (pedge); */
}
