#include "copyright.h"
#include "sparcs.h"

/*
 *  Get the edges for each object and put them into the global list.
 *
 *  XXX can the part of the list corresp. to the instances be permanent??
 *
 *  2/15/87
 *  last modified 10/26/87
 */

linkEdges(direction)
directionType direction;
{
    struct object *po;
    struct wire *pw;
    int debug = 0;

    pMaxEdgeListHead = pMaxEdgeListTail = NIL(struct edgePtr);
    pMidEdgeListHead = pMidEdgeListTail = NIL(struct edgePtr);
    pMinEdgeListHead = pMinEdgeListTail = NIL(struct edgePtr);

    for (po=pObjectList; po!=NIL(struct object); po=po->pnext)  {
	if (pMaxEdgeListHead == NIL(struct edgePtr))  {
	    pMaxEdgeListTail = po->plastMaxEP;
	}
	po->plastMaxEP->pnext = pMaxEdgeListHead;
	pMaxEdgeListHead = po->pfirstMaxEP;
	
	if (pMinEdgeListHead == NIL(struct edgePtr))  {
	    pMinEdgeListTail = po->plastMinEP;
	}
	po->plastMinEP->pnext = pMinEdgeListHead;
	pMinEdgeListHead = po->pfirstMinEP;
	
	if (po->pfirstMidEP != NIL(struct edgePtr) &&
	    po->plastMidEP  != NIL(struct edgePtr))  {
	    if (pMidEdgeListHead == NIL(struct edgePtr))  {
		pMidEdgeListTail = po->plastMidEP;
	    }
	    po->plastMidEP->pnext = pMidEdgeListHead;
	    pMidEdgeListHead = po->pfirstMidEP;
	} else  {
	    /*
	     *  XXX
	     *  This checking can be removed after debugging is completed.
	     */
	    if (po->pfirstMidEP == NIL(struct edgePtr) &&
		po->plastMidEP  != NIL(struct edgePtr))  {
		printf("linkEdges:  inconsistent MidEP pointers.\n");
		exit(-1);
	    }
	    if (po->pfirstMidEP != NIL(struct edgePtr) &&
		po->plastMidEP  == NIL(struct edgePtr))  {
		printf("linkEdges:  inconsistent MidEP pointers.\n");
		exit(-1);
	    }
	}
    }

    for (pw=pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {
	if (direction == HORIZONTAL)  {
	    /*
	     *  Horizontal spacing ==> consider only vertical wires.
	     */
	    if (pw->horizontal == TRUE)  {
		continue;
	    } else  {
		/*
		 *  XXX  6/12/87
		 *  Only link wires with positive length.
		 */
		if (pw->p2.y > pw->p1.y)  {
		    pw->pfirstMaxEP->pnext = pw->plastMaxEP;  /* XXX 6/12/87 */
		    if (pMaxEdgeListHead == NIL(struct edgePtr))  {
			pMaxEdgeListTail = pw->plastMaxEP;
		    }
		    pw->plastMaxEP->pnext = pMaxEdgeListHead;
		    pMaxEdgeListHead = pw->pfirstMaxEP;

		    pw->pfirstMinEP->pnext = pw->plastMinEP;  /* XXX 6/12/87 */
		    if (pMinEdgeListHead == NIL(struct edgePtr))  {
			pMinEdgeListTail = pw->plastMinEP;
		    }
		    pw->plastMinEP->pnext = pMinEdgeListHead;
		    pMinEdgeListHead = pw->pfirstMinEP;
		} else  {
		    if (debug)  {
			printf("linkEdges:  skipping non-positive length wire;\n");
			printf("\t(%d, %d) (%d, %d), layer %s\n", pw->p1.x,
			    pw->p1.y, pw->p2.x, pw->p2.y,
			    LayerMap[pw->maskNumber].name);
		    }
		}
	    }
	} else  {
	    /*
	     *  Vertical spacing ==> consider only horizontal wires.
	     */
	    if (pw->horizontal == FALSE)  {
		continue;
	    } else  {
		if (pw->p2.x > pw->p1.x)  {
		    pw->pfirstMaxEP->pnext = pw->plastMaxEP;  /* XXX 6/12/87 */
		    if (pMaxEdgeListHead == NIL(struct edgePtr))  {
			pMaxEdgeListTail = pw->plastMaxEP;
		    }
		    pw->plastMaxEP->pnext = pMaxEdgeListHead;
		    pMaxEdgeListHead = pw->pfirstMaxEP;

		    pw->pfirstMinEP->pnext = pw->plastMinEP;  /* XXX 6/12/87 */
		    if (pMinEdgeListHead == NIL(struct edgePtr))  {
			pMinEdgeListTail = pw->plastMinEP;
		    }
		    pw->plastMinEP->pnext = pMinEdgeListHead;
		    pMinEdgeListHead = pw->pfirstMinEP;
		} else  {
		    if (debug)  {
			printf("linkEdges:  skipping non-positive length wire;\n");
			printf("\t(%d, %d) (%d, %d), layer %s\n", pw->p1.x,
			    pw->p1.y, pw->p2.x, pw->p2.y,
			    LayerMap[pw->maskNumber].name);
		    }
		}
	    }
	}
    }
}
