#include "copyright.h"
#include "sparcs.h"

/*
 *  Merge sort to put edge list into increasing order based on coord field.
 *  See e.g. Sedgewick for description of mergesort.
 *
 *  The list is a linear, doubly-linked list with a pointer to the first
 *  element.  There is no pointer to the last element (as in dll).
 *
 *  Returns number of elements sorted.
 *
 *  Algo:
 *
 *  3/2/87
 *  last modified 3/4/87
 */

int
mergesort(phead, ptail)
struct edgePtr **phead;		/* list header (pEdgeList) */
struct edgePtr **ptail;		/* list tail */
{
    struct edgePtr *paHead;		/* first element, first sublist */
    struct edgePtr *paTail;		/* first element, second sublist */
    struct edgePtr *pbHead;		/* first element, second sublist */
    struct edgePtr *pbTail;		/* first element, second sublist */
    struct edgePtr *pcHead;	/* first element, merge of a and b */
    struct edgePtr *pcTail;	/* last element, merge of a and b */
    struct edgePtr *pdone;	/* last element, sorted part of list */
    struct edgePtr *ptodo;	/* first element, unsorted part of list */
    struct edgePtr dummy;

    /***
    struct edgePtr *ps;
    ***/

    int count;
    /***
    int pass = 0;
    ***/
    int merges = 1;
    int sublistLength = 1;		/* length of each sublist */

    /*
     *  Add a check for lists of length 1????
     */
    dummy.pnext = *phead;


    /*
     *  Use a do-loop since must go through the list at least once.
     */
    do  {

	/*pbHead = NIL(struct edgePtr);*/    /* Needed??? */
	/*printf("inner-loop pass %d, sublistLength %d\n",
		++pass, sublistLength);*/
	/*
	 *  This loop passes through the list once, merging adjacent sublists
	 *  of length sublistLength.  Initially nothing is done, so pdone is
	 *  UNDEF and ptodo = *phead.
	 */
	pdone = &dummy;
	ptodo = dummy.pnext;
	while (ptodo != NIL(struct edgePtr))  {
	    paHead = ptodo;	/* note that cannot be NIL */
	    paTail = paHead;
	    count = sublistLength - 1;

	    /*printf("  paHead->coord = %d",
		    (paHead!=NIL(struct edgePtr)) ? paHead->coord : -1);*/

	    while (paTail != NIL(struct edgePtr)  &&  count > 0)  {
		paTail = paTail->pnext;
		count--;
	    }
	    if (paTail == NIL(struct edgePtr))  {
		/*
		 *  Reached the end of the list; there is no "b" sublist.
		 *  Patch in the "a" sublist and jump out of the while loop.
		 *  pbHead is set to NIL so that there is no accidental
		 *  merge if paHead==dummy.pnext
		 */
		pdone->pnext = paHead;
		ptodo = NIL(struct edgePtr);
		/*pbHead = NIL(struct edgePtr);*/    /* no "b" sublist */ 
	        /*printf("  paTail->coord = %d\n", -1);*/
		break;
	    }
	    /*printf("  paTail->coord = %d\n", paTail->coord);*/

	    /*
	     *  Now, paTail is at the end of the "a" sublist.  Find the "b"
	     *  sublist.
	     */

	    pbHead = paTail->pnext;	/*  CAN be NIL */
	    if (pbHead == NIL(struct edgePtr))  {
		/*
		 *  As in the case of paTail==NIL, there is no "b" sublist.
		 */
		pdone->pnext = paHead;
		ptodo = NIL(struct edgePtr);
		/*pbHead = NIL(struct edgePtr);*/    /* no "b" sublist */ 
		break;
	    }
	    pbTail = pbHead;
	    count = sublistLength - 1;

	    /*printf("  pbHead->coord = %d",
		    (pbHead!=NIL(struct edgePtr)) ? pbHead->coord : -1);*/

	    while (pbTail != NIL(struct edgePtr)  &&  count > 0)  {
		pbTail = pbTail->pnext;
		count--;
	    }
	    /*
	     *  pbTail is at the end of the "b" sublist.  If it is NIL, then
	     *  this is the last sublist for this pass.
	     */
	    if (pbTail == NIL(struct edgePtr))  {
		pbTail = *ptail;
		ptodo = NIL(struct edgePtr);
	        /*printf("  pbTail->coord = %d\n", -1);*/
	    } else  {
	        ptodo = pbTail->pnext;
	        /*printf("  pbTail->coord = %d\n", pbTail->coord);*/
	    }

	    /*
	     *  Number of merges + 1 = number of elements sorted.
	     */
	    merges++;
	    merge(paHead, paTail, pbHead, pbTail, &pcHead, &pcTail);
	    /**
	    printf("after merge:\n");
	    for (ps=pcHead; ps!=NIL(struct edgePtr); ps=ps->pnext)  {
		printf("  %d\n", ps->coord);
		if (ps == pcTail)  break;
	    }
	    **/

	    /*
	     *  Link the "c" sublist in between pdone and ptodo, and reset
	     *  pdone.
	     */
	    pdone->pnext = pcHead;
	    pcTail->pnext = ptodo;
	    pdone = pcTail;
	    if (ptodo == NIL(struct edgePtr))  *ptail = pcTail;
	}
	if (paHead == dummy.pnext)  {
	    /*
	     *  At most 2 sublists left, perhaps only 1.  In either case we
	     *  are done; if there are 2 sublists, they have just been merged,
	     *  and if there is one, it must be in order.
	     */
    	    *phead = dummy.pnext;
	    return(merges);
	}
	sublistLength += sublistLength;
    } while (1);
    /*NOTREACHED*/
}


/*
 *  Depends on the input pointers all being non-NIL!
 *
 *  3/3/87
 *  last modified 3/3/87
 */
merge(paHead, paTail, pbHead, pbTail, pcHead, pcTail)
struct edgePtr *paHead;
struct edgePtr *paTail;
struct edgePtr *pbHead;
struct edgePtr *pbTail;
struct edgePtr **pcHead;	/* head of merged sublists; returned */
struct edgePtr **pcTail;	/* tail of merged sublists; returned */
{
    struct edgePtr *pc;
    struct edgePtr dummy;

    if (paHead == NIL(struct edgePtr))  {
	printf("merge:  paHead == NIL.\n");
	exit(-1);
    }
    if (paTail == NIL(struct edgePtr))  {
	printf("merge:  paTail == NIL.\n");
	exit(-1);
    }
    if (pbHead == NIL(struct edgePtr))  {
	printf("merge:  pbHead == NIL.\n");
	exit(-1);
    }
    if (pbTail == NIL(struct edgePtr))  {
	printf("merge:  pbTail == NIL.\n");
	exit(-1);
    }

    dummy.coord = INFINITY;
    paTail->pnext = &dummy;
    pbTail->pnext = &dummy;

    if (paHead->coord <= pbHead->coord)  {
	*pcHead = paHead;
	paHead = paHead->pnext;
    } else  {
	*pcHead = pbHead;
	pbHead = pbHead->pnext;
    }
    pc = *pcHead;

    while (paHead != &dummy  ||  pbHead != &dummy)  {
	if (paHead->coord <= pbHead->coord)  {
	    pc->pnext = paHead;
	    pc = paHead;
	    paHead = paHead->pnext;
	} else  {
	    pc->pnext = pbHead;
	    pc = pbHead;
	    pbHead = pbHead->pnext;
	}
    }

    if (pbTail->coord >= paTail->coord)  {
	*pcTail = pbTail;
    } else  {
	*pcTail = paTail;
    }
}
