#include "copyright.h"
#include "sparcs.h"

/*
 *  6/30/87
 *  last modified 12/17/87
 */

void
rewalkHardTF(stateCoord, phardEdgePtr, direction)
int stateCoord;			/* current state coord */
struct hlist *phardEdgePtr;	/* hlist pointer to current edge */
directionType direction;
{
    struct edge *pedge;	    	/* edge being re-walked */
    int maskerKey;		/* key of instance that masked pedge */
    int maskerBB;		/* far edge of BB of masker of pedge */
    struct edge *pscan;	    	/* scanner to search edge struct */
    int curNet;		    	/* netId of pedge */
    char mergeable;
    struct node *pfrom;
    struct node *pto;
    struct object *pedgeObj;
    struct wire *pedgeWire;
    struct object *pscanObj;
    int edgeOrigin;
    int scanOrigin;
    int rule;
    int spacing;
    int debug = 0;

    pedge = phardEdgePtr->pedge;
    maskerKey = phardEdgePtr->hardMaskedBy;

    /*
     *  Input edge is a connected tf, from either an instance or a wire.  The
     *  element that masked it (phardEdgePtr->key) MUST be an instance!
     */
    if (maskerKey <= NumObjects)  {
	/*
	 *  XXX  Not necessarily correct for weirdly-shaped elements...
	 */
	if (direction == HORIZONTAL)  {
	    if (pedge->type == MAX)  {
		maskerBB = ElementByKey[maskerKey].ptr.po->origin.x +
				ElementByKey[maskerKey].ptr.po->bbUpperR.x;
	    } else  {
		maskerBB = ElementByKey[maskerKey].ptr.po->origin.x;
	    }
	} else  {
	    if (pedge->type == MAX)  {
		maskerBB = ElementByKey[maskerKey].ptr.po->origin.y +
				ElementByKey[maskerKey].ptr.po->bbUpperR.y;
	    } else  {
		maskerBB = ElementByKey[maskerKey].ptr.po->origin.y;
	    }
	}
    } else  {
	printf("rewalkHardTF:  masker must be an instance!\n");
	exit(-1);
    }

    if (pedge->masked == 'S'  ||  pedge->masked == 'P')  {
	printf("rewalkHardTF:  input edge masked %c; must be 'N' or 'H'\n",
		pedge->masked);
	printf("    %s edge from inst %d, on layer %s, ",
		(pedge->type == MIN) ? "MIN" : "MAX",
		pedge->key, LayerMap[pedge->maskNumber].name);
	if (pedge->netId == UNDEF)  {
	    printf("netId UNDEF\n");
	} else  {
	    printf("netId %d\n", pedge->netId);
	}
	if (pedge->key <= NumObjects)  {
	    printf("   from inst %d, origin (%d, %d)\n",
		    pedge->key,
		    ElementByKey[pedge->key].ptr.po->origin.x,
		    ElementByKey[pedge->key].ptr.po->origin.y);
	} else  {
	    printf("   from wire (%d, %d) (%d, %d)\n",
		    ElementByKey[pedge->key].ptr.pw->p1.x,
		    ElementByKey[pedge->key].ptr.pw->p1.y,
		    ElementByKey[pedge->key].ptr.pw->p2.x,
		    ElementByKey[pedge->key].ptr.pw->p2.y);
	}
	exit(-1);
    }

    if (pedge->key <= NumObjects)  {
        pedgeObj = ElementByKey[pedge->key].ptr.po;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeObj->origin.x;
	} else  {
	    edgeOrigin = pedgeObj->origin.y;
	}
    } else  {
        pedgeWire = ElementByKey[pedge->key].ptr.pw;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeWire->p1.x - pedgeWire->halfWidth;
	} else  {
	    edgeOrigin = pedgeWire->p1.y - pedgeWire->halfWidth;
	}
    }

    curNet = pedge->netId;

    if (pedge->type == MAX)  {
	/*
	 *  Must process all edges with the same loc as pedge.  So, back up to
	 *  the left to the first edge in the edge structure with the same loc
	 *  before starting the scan to the right.
	 */
	if (pedge->pprev==NIL(struct edge) || pedge->pprev->loc < pedge->loc)  {
	    pscan = pedge->pnext;
	} else  {
	    pscan = pedge->pprev;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pprev != NIL(struct edge); pscan = pscan->pprev)  {
		if (pscan->pprev->loc != pedge->loc)  break;
	    }
	}

	/*
	 *  Scan to the right, looking for MIN edges.  The ONLY edges of
	 *  interest are from the instance that masked pedge.  The search stops
	 *  in either the usual way - when a MAX rule edge on the same layer as
	 *  pedge is reached (in this case the edge will be from the masking
	 *  instance) - or when all edges of the masking instance have been
	 *  passed.
	 */
	for (; pscan != NIL(struct edge); pscan = pscan->pnext)  {
	    if (pscan->loc > maskerBB)  {
		/*
		printf("rewalkHardTF:  passed all edges of masker; done.\n");
		*/
		return;
	    }
	    if (pscan->key == maskerKey  &&  pscan->type == MIN)  {
		/*
		 *  There are other edges to skip, according to the regular
		 *  tf rules.
		 */
		if (pscan->masked != 'N')  {
		    if (debug)  {
			printf("rewalkHardTF:  skipping masked edge...\n");
		    }
		    continue;
		}
		if (pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber)  {
		    if (debug)  {
			printf("rewalkHardTF:  skipping edge from same ");
			printf("net, same layer.\n");
		    }
		    continue;
		}
		/*
		 *  And, there must be a solid overlap
		 */
		if (pscan->max <= stateCoord)  continue;
	     
		rule = getRule(pedge->maskNumber, pscan->maskNumber,
				    &mergeable);
		if (rule != NO_RULE)  {
		    /*
		     *  Skip the case of mergeable layers in the same net.
		     */
		    if (pscan->netId == curNet)  {
			if (mergeable == 'Y')  {
			    if (debug)  {
				printf("rewalkHardTF:  skipping ");
				printf("MERGEABLE edge from");
				printf(" same net, different layer.\n");
			    }
			    continue;
			} else  {
			    if (debug)  {
				printf("rewalkHardTF:  not skipping edge");
				printf(" from same net, different layer;");
				printf(" NOT mergeable.\n");
			    }
			}
		    }

		    /*
		     *  There is a rule, because the layers are not merge-
		     *  able or because the edges are from different nets,
		     *  but IFF pedge and pscan overlap sufficiently.
		     */
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough, so skip this edge.
			 */
			continue;
		    }

		    /*
		     *  Get the scan origin to determine the direction of
		     *  the branch.
		     */
		    pscanObj = ElementByKey[pscan->key].ptr.po;
		    if (direction == HORIZONTAL)  {
			scanOrigin = pscanObj->origin.x;
		    } else  {
			scanOrigin = pscanObj->origin.y;
		    }

		    /*
		     *  Note that the scan element must be an instance.
		     */
		    if (pedge->key <= NumObjects)  {
			spacing = (pedge->loc - edgeOrigin) + rule -
				    (pscan->loc - scanOrigin);
		    } else  {
			/*
			 *  XXX FIX!!
			 *
			 *  Reject if object and wire are connected.
			 */
			if (pedgeWire->po1 == pscanObj  ||
				pedgeWire->po2 == pscanObj)  {
			    continue;
			} else  {
			    spacing = pedgeWire->width + rule -
					(pscan->loc - scanOrigin);
			}
		    }

		    if (edgeOrigin < scanOrigin)  {
			/*
			 *  This should be the usual case, due to the
			 *  direction of the scan.  pedge belongs to pfrom,
			 *  pscan belongs to pto.
			 */
			if (debug)  {
			    printf("l.b. spacing constraint, ");
			    printf("%d -> %d:  %d,\n",
				    pedge->key, pscan->key, spacing);
			    printf("\tpedge: MAX, at %d, on %s;",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    printf("  pscan: MIN, at %d, on %s\n",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
			link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (scanOrigin < edgeOrigin)  {
			/*
			 *  pscan belongs to pfrom, pedge belongs to pto.
			 *  Must put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. spacing constraint, ");
			    printf("%d -> %d:  %d,\n",
				    pscan->key, pedge->key, -spacing);
			    printf("\tpedge: MAX, at %d, on %s;",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    printf("  pscan: MIN, at %d, on %s\n",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
			link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pedge->key < pscan->key)  {
			    /*
			     *  pedge belongs to pfrom, ...
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pfrom, put in spacing as
			     *  an upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MAX, at %d, on %s;",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
				printf("  pscan: MIN, at %d, on %s\n",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }
		}
	    } else if (pscan->key == maskerKey  &&  pscan->type == MAX)  {
		if (pscan->maskNumber == pedge->maskNumber)  {
		    if (pscan->max > stateCoord)  {
			if (debug)  {
			    printf("Terminating scan; found edge w/same ");
			    printf("polarity, same mask.\n");
			}
			return;
		    }
		}
	    }
	}
    } else  {
	/*
	 *  Since must process all edges with the same loc as pedge, back up to
	 *  the right to the first edge in the edge structure with the same loc
	 *  before starting the scan to the left.
	 */
	if (pedge->pnext==NIL(struct edge) || pedge->pnext->loc>pedge->loc)  {
	    pscan = pedge->pprev;
	} else  {
	    pscan = pedge->pnext;
	    for (; pscan!=NIL(struct edge) &&
		   pscan->pnext != NIL(struct edge); pscan=pscan->pnext)  {
		if (pscan->pnext->loc != pedge->loc)  break;
	    }
	}
	/*
	 *  Scan to the left, looking for MAX edges.
	 */
	for (; pscan != NIL(struct edge); pscan = pscan->pprev)  {
	    if (pscan->loc < maskerBB)  {
		/*
		printf("rewalkHardTF:  passed all edges of masker; done.\n");
		*/
		return;
	    }
	    if (pscan->key == maskerKey  &&  pscan->type == MAX)  {
		/*
		 *  Skip edges from the same element, edges from the same
		 *  net on MERGEABLE layers, and masked edges.
		 *
		 *  First weed-out the obvious ones.
		 */
		if (pscan->masked != 'N')  {
		    if (debug)  {
			printf("rewalkHardTF:  skipping masked edge...\n");
		    }
		    continue;
		}
		if (pscan->netId == curNet  &&
			pscan->maskNumber == pedge->maskNumber)  {
		    if (debug)  {
			printf("rewalkHardTF:  skipping edge from same net,");
			printf(" same layer.\n");
		    }
		    continue;
		}
		/*
		 *  And, there must be a solid overlap
		 */
		if (pscan->max <= stateCoord)  continue;

		rule = getRule(pedge->maskNumber, pscan->maskNumber,
				    &mergeable);
		if (rule != NO_RULE)  {
		    /*
		     *  Skip the case of mergeable layers in the same net.
		     */
		    if (pscan->netId == curNet)  {
			if (mergeable == 'Y')  {
			    if (debug)  {
				printf("rewalkHardTF:  skipping MERGEABLE ");
				printf("edge from same net, diff. layer.\n");
			    }
			    continue;
			} else  {
			    if (debug)  {
				printf("rewalkHardTF:  not skipping edge ");
				printf("from same net, different layer;");
				printf(" NOT mergeable.\n");
			    }
			}
		    }

		    /*
		     *  There is a rule, because the layers are not merge-
		     *  able or because the edges are from different nets,
		     *  but IFF pedge and pscan overlap sufficiently.
		     */
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough, so skip this edge.
			 */
			continue;
		    }

		    pscanObj = ElementByKey[pscan->key].ptr.po;
		    if (direction == HORIZONTAL)  {
			scanOrigin = pscanObj->origin.x;
		    } else  {
			scanOrigin = pscanObj->origin.y;
		    }

		    /*
		     *  scan element is left, edge element right.
		     */
		    if (pedge->key <= NumObjects)  {
			spacing = (pscan->loc - scanOrigin) + rule -
				    (pedge->loc - edgeOrigin);
		    } else  {
			/*
			 *  XXX FIX!!
			 *
			 *  Reject if object and wire are connected.
			 */
			if (pedgeWire->po1 == pscanObj  ||
				pedgeWire->po2 == pscanObj)  {
			    continue;
			} else  {
			    spacing = (pscan->loc - scanOrigin) + rule;
			}
		    }

		    if (scanOrigin < edgeOrigin)  {
			/*
			 *  This should be the usual case, due to the
			 *  direction of the scan.  pedge belongs to pto,
			 *  pscan belongs to pfrom.
			 */
			if (debug)  {
			    printf("l.b. spacing constraint, ");
			    printf("%d -> %d:  %d,\n",
				    pscan->key, pedge->key, spacing);
			    printf("\tpscan: MAX, at %d, on %s;",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf("  pedge: MIN, at %d, on %s\n",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
			link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (edgeOrigin < scanOrigin)  {
			/*
			 *  pscan belongs to pto, pedge belongs to pfrom.
			 *  Must put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. spacing constraint, ");
			    printf("%d -> %d:  %d,\n",
				    pedge->key, pscan->key, -spacing);
			    printf("\tpscan: MAX, at %d, on %s;",
				    pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf(" pedge: MIN, at %d, on %s\n",
				    pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
			link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pscan->key < pedge->key)  {
			    /*
			     *  pedge belongs to pto, ...
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf(" pedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
			    link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pto, put in spacing as an
			     *  upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, ");
				printf("%d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("\tpscan: MAX, at %d, on %s;",
					pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf(" pedge: MIN, at %d, on %s\n",
					pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
			    link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }
		}
	    } else if (pscan->key == maskerKey  &&  pscan->type == MIN)  {
		if (pscan->maskNumber == pedge->maskNumber)  {
		    if (pscan->max > stateCoord)  {
			if (debug)  {
			    printf("Terminating scan; found edge w/same ");
			    printf("polarity, same mask.\n");
			}
			return;
		    }
		}
	    }
	}
    }
}
