#include "copyright.h"
#include "sparcs.h"

/*
 *  Only considers edges whose max is > stateCoord.
 *
 *  6/26/87
 *  last modified 12/17/87
 */

void
rewalkPF(stateCoord, pedge, direction)
int stateCoord;		/* coord of current state; ie, MIN coord of pedge */
struct edge *pedge;	/* current edge */
directionType direction;
{
    struct edge *pscan;	    /* scanner to find edges related to cur. edge */
    struct klist *pmB;
    struct node *pfrom;
    struct node *pto;
    struct object *pedgeObj;
    struct object *pscanObj;
    struct wire *pedgeWire;
    struct wire *pscanWire;
    char mergeable;
    int edgeOrigin;
    int scanOrigin;
    int rule;
    int spacing;
    int debug = 0;

    /*
     *  pedge must not be masked.
     */
    if (pedge->masked != 'N')  {
	printf("rewalkPF:  input edge has masked = %c; must be 'N'\n",
		pedge->masked);
	exit(-1);
    }

    if (pedge->key <= NumObjects)  {
        pedgeObj = ElementByKey[pedge->key].ptr.po;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeObj->origin.x;
	} else  {
	    edgeOrigin = pedgeObj->origin.y;
	}
    } else  {
        pedgeWire = ElementByKey[pedge->key].ptr.pw;
	if (direction == HORIZONTAL)  {
	    edgeOrigin = pedgeWire->p1.x - (pedgeWire->width)/2;
	} else  {
	    edgeOrigin = pedgeWire->p1.y - (pedgeWire->width)/2;
	}
    }
    if (pedge->type == MAX)  {
	/*
	 *  Scan to the right, looking for MIN edges.  The search uncondition-
	 *  ally stops when a MAX edge on the same edge as pedge is reached.
	 *
	 *  Must process all edges with the same loc as pedge.  So, back up to
	 *  the left to the first edge in the edge structure with the same loc
	 *  before starting the scan to the right.
	 */
	if (pedge->pprev==NIL(struct edge) || pedge->pprev->loc<pedge->loc)  {
	    pscan = pedge->pnext;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pprev;
	    for (; pscan != NIL(struct edge) &&
		   pscan->pprev != NIL(struct edge); pscan = pscan->pprev)  {
		if (pscan->pprev->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}

	for (; pscan != NIL(struct edge); pscan = pscan->pnext)  {
	    /*
	     *  Don't compare two edges from the same element.
	     *  XXX OK??
		   If pscan is pf-tf (soft) masked ('S'), then it is ignored.
	     *  If pscan is pf-pf masked ('P'), then it is ignored.
	     *  If pscan is tf-tf (hard) masked ('H'), then it is ignored IFF
	     *  it it hard-masked by instance pedge->key.
	     */
	    if (pscan->key == pedge->key)  continue;
	    if (pscan->masked == 'S')  {
		if (debug)  printf("rewalkPF:  skipping pf-tf-masked edge...\n");
		continue;
	    } else if (pscan->masked == 'P')  {
		if (debug)  printf("rewalkPF:  skipping pf-pf-masked edge...\n");
		continue;
	    } else if (pscan->max <= stateCoord)  {
		if (debug)  printf("rewalkPF:  skipping nonoverlapping edge...\n");
		continue;
	    } else if (pscan->masked == 'H')  {
		if (pscan->phardMaskedBy == NIL(struct klist))  {
		    printf("rewalkPF:  pscan->phardMaskedBy shouldn't be NIL;");
		    printf("  pscan->masked == 'H'\n");
		    exit(-1);
		}
		pmB = pscan->phardMaskedBy;
		for (; pmB != NIL(struct klist); pmB = pmB->pnext)  {
		    if (pmB->key == pedge->key)  {
			if (debug)  {
			    printf("rewalkPF:  skipping hard-masked edge,");
			    printf(" masked by inst pedge->key\n");
			}
			goto nextEdge;
		    }
		}
		if (debug)
		    printf("rewalkPF:  NOT SKIPPING tf-tf-masked edge.\n");
	    }

	    if (pscan->type == MIN)  {

		/*
		 *  Potential edge from pedge's object to pscan's object.
		 */
		rule = getRule(pedge->maskNumber, pscan->maskNumber,
				&mergeable);
		if (rule != NO_RULE)  {
		    /*
		     *  Rule only if pedge and pscan overlap sufficiently.
		     */
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough for a rule, so skip this edge.
			 */
			goto nextEdge;
		    }

		    /*
		     *  Get the two origins, to determine the direction of
		     *  the branch.
		     */
		    if (pscan->key <= NumObjects)  {
			pscanObj = ElementByKey[pscan->key].ptr.po;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanObj->origin.x;
			} else  {
			    scanOrigin = pscanObj->origin.y;
			}
		    } else  {
			pscanWire = ElementByKey[pscan->key].ptr.pw;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanWire->p1.x - (pscanWire->width)/2;
			} else  {
			    scanOrigin = pscanWire->p1.y - (pscanWire->width)/2;
			}
		    }

		    /*
		     *  XXX This mess will go away if treat wires and instances
		     *      more uniformly.
		     */
		    if (pedge->key <= NumObjects)  {
		        if (pscan->key <= NumObjects)  {
			    spacing = (pedge->loc - edgeOrigin) + rule -
					(pscan->loc - scanOrigin);
		        } else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pedgeObj == pscanWire->po1  ||
				    pedgeObj == pscanWire->po2)  {
				continue;
			    } else  {
			        spacing = (pedge->loc - edgeOrigin) + rule;
			    }
			}
		    } else  {
		        if (pscan->key <= NumObjects)  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pedgeWire->po1 == pscanObj  ||
				    pedgeWire->po2 == pscanObj)  {
				continue;
			    } else  {
				spacing = pedgeWire->width + rule -
					    (pscan->loc - scanOrigin);
			    }
		        } else  {
			    spacing = pedgeWire->width + rule;
			}
		    }

		    if (edgeOrigin < scanOrigin)  {
			/*
			 *  This should be the usual case, due to the direction
			 *  of the scan.  pedge belongs to pfrom, pscan belongs
			 *  to pto.
			 */
			if (debug)  {
			    printf("l.b. spacing constraint, %d -> %d:  %d,\n",
				    pedge->key, pscan->key, spacing);
			    printf("\tpedge: MAX, at %d, on %s;", pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    
			    printf("  pscan: MIN, at %d, on %s\n", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
		        link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (scanOrigin < edgeOrigin)  {
			/*
			 *  pscan belongs to pfrom, pedge belongs to pto.  Must
			 *  put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. spacing constraint, %d -> %d:  %d,\n",
				    pscan->key, pedge->key, -spacing);
			    printf("\tpedge: MAX, at %d, on %s;", pedge->loc,
				    LayerMap[pedge->maskNumber].name);
			    
			    printf("  pscan: MIN, at %d, on %s\n", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
		        link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pedge->key < pscan->key)  {
			    /*
			     *  pedge belongs to pfrom, ...
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, %d -> %d:  %d,\n",
					pedge->key, pscan->key, spacing);
				printf("\tpedge: MAX, at %d, on %s;", pedge->loc,
					LayerMap[pedge->maskNumber].name);
				
				printf("  pscan: MIN, at %d, on %s\n", pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
		            link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pfrom, put in spacing as an
			     *  upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, %d -> %d:  %d,\n",
					pscan->key, pedge->key, -spacing);
				printf("\tpedge: MAX, at %d, on %s;", pedge->loc,
					LayerMap[pedge->maskNumber].name);
				
				printf("  pscan: MIN, at %d, on %s\n", pscan->loc,
					LayerMap[pscan->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
		            link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }
		    /* return; */
		}
	    } else if (pscan->type == MAX  &&
			pscan->maskNumber == pedge->maskNumber)  {
		/*
		 *  The edges must really overlap, and by an interval (not just
		 *  at a corner).
		 */
		if (pscan->max > stateCoord)  {
		    if (debug)  {
			printf("Terminating scan; found edge w/same ");
			printf("polarity, same mask.\n");
		    }
		    return;
		}
	    }
	nextEdge:
	    continue;
	}
    } else  {
	/*
	 *  Scan to the left, looking for MAX edges.
	 *
	 *  Since must process all edges with the same loc as pedge, back up to
	 *  the right to the first edge in the edge structure with the same loc
	 *  before starting the scan to the left.
	 */
	if (pedge->pnext==NIL(struct edge) || pedge->pnext->loc>pedge->loc)  {
	    pscan = pedge->pprev;
	    /* if (debug)  printf("    (didn't back up)\n"); */
	} else  {
	    pscan = pedge->pnext;
	    for (; pscan != NIL(struct edge) && 
		   pscan->pnext != NIL(struct edge); pscan=pscan->pnext)  {
		if (pscan->pnext->loc != pedge->loc)  break;
	    }
	    /*
	    if (debug)  {
		printf("    (backed up to %s edge, inst %d, layer %s, ",
			(pscan->type == MIN) ? "MIN" : "MAX",
			pscan->key,
			LayerMap[pscan->maskNumber].name);
		if (pscan->netId == UNDEF)  {
		    printf("netId UNDEF)\n");
		} else  {
		    printf("netId %d)\n", pscan->netId);
		}
	    }
	    */
	}
	for (; pscan != NIL(struct edge); pscan = pscan->pprev)  {
	    if (pscan->key == pedge->key)  continue;
	    if (pscan->masked == 'S')  {
		if (debug)  printf("rewalkPF:  skipping pf-tf-masked edge...\n");
		continue;
	    } else if (pscan->masked == 'P')  {
		if (debug)  printf("rewalkPF:  skipping pf-pf-masked edge...\n");
		continue;
	    } else if (pscan->max <= stateCoord)  {
		if (debug)  printf("rewalkPF:  skipping nonoverlapping edge...\n");
		continue;
	    } else if (pscan->masked == 'H')  {
		if (pscan->phardMaskedBy == NIL(struct klist))  {
		    printf("rewalkPF:  pscan->phardMaskedBy shouldn't be NIL;");
		    printf("  pscan->masked == 'H'\n");
		    exit(-1);
		}
		pmB = pscan->phardMaskedBy;
		for (; pmB != NIL(struct klist); pmB = pmB->pnext)  {
		    if (pmB->key == pedge->key)  {
			if (debug)  {
			    printf("rewalkPF:  skipping hard-masked edge,");
			    printf(" masked by inst pedge->key\n");
			}
			goto prevEdge;
		    }
		}
		if (debug)
		    printf("rewalkPF:  NOT SKIPPING tf-tf-masked edge.\n");
	    }
	    
	    if (pscan->type == MAX)  {
		/*
		 *  Potential edge from pscan's object to pedge's object.
		 */
		rule = getRule(pedge->maskNumber, pscan->maskNumber,
				&mergeable);
		if (rule != NO_RULE)  {
		    /*
		     *  Rule only if pedge and pscan overlap sufficiently.
		     */
		    if (stateCoord - pscan->max >= rule)  {
			/*
			 *  Don't overlap enough for a rule, so skip this edge.
			 */
			goto prevEdge;
		    }

		    if (pscan->key <= NumObjects)  {
			pscanObj = ElementByKey[pscan->key].ptr.po;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanObj->origin.x;
			} else  {
			    scanOrigin = pscanObj->origin.y;
			}
		    } else  {
			pscanWire = ElementByKey[pscan->key].ptr.pw;
			if (direction == HORIZONTAL)  {
			    scanOrigin = pscanWire->p1.x - (pscanWire->width)/2;
			} else  {
			    scanOrigin = pscanWire->p1.y - (pscanWire->width)/2;
			}
		    }

		    /*
		     *  scan element is left, edge element right.
		     */
		    if (pscan->key <= NumObjects)  {
		        if (pedge->key <= NumObjects)  {
			    spacing = (pscan->loc - scanOrigin) + rule -
					(pedge->loc - edgeOrigin);
		        } else  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pedgeWire->po1 == pscanObj  ||
				    pedgeWire->po2 == pscanObj)  {
				continue;
			    } else  {
			        spacing = (pscan->loc - scanOrigin) + rule;
			    }
			}
		    } else  {
		        if (pedge->key <= NumObjects)  {
			    /*
			     *  XXX FIX!!
			     *
			     *  Reject if object and wire are connected.
			     */
			    if (pscanWire->po1 == pedgeObj  ||
				    pscanWire->po2 == pedgeObj)  {
				continue;
			    } else  {
			        spacing = pscanWire->width + rule -
					    (pedge->loc - edgeOrigin);
			    }
		        } else  {
			    spacing = pscanWire->width + rule;
			}
		    }

		    if (scanOrigin < edgeOrigin)  {
			/*
			 *  This should be the usual case, due to the direction
			 *  of the scan.  pedge belongs to pto, pscan belongs
			 *  to pfrom.
			 */
			if (debug)  {
			    printf("l.b. spacing constraint, %d -> %d:  %d,\n",
				    pscan->key, pedge->key, spacing);
			    printf("  pscan: MAX, at %d, on %s;", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf("  pedge: MIN, at %d, on %s\n", pedge->loc,
			    	    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pscan->key];
			pto   = NodeArray[pedge->key];
		        link(pfrom, pto, spacing, INFINITY, 0);
		    } else if (edgeOrigin < scanOrigin)  {
			/*
			 *  pscan belongs to pto, pedge belongs to pfrom.  Must
			 *  put spacing in as an upper.
			 */
			if (debug)  {
			    printf("u.b. spacing constraint, %d -> %d:  %d,\n",
				    pedge->key, pscan->key, -spacing);
			    printf("  pscan: MAX, at %d, on %s;", pscan->loc,
				    LayerMap[pscan->maskNumber].name);
			    printf("  pedge: MIN, at %d, on %s\n", pedge->loc,
			    	    LayerMap[pedge->maskNumber].name);
			}
			pfrom = NodeArray[pedge->key];
			pto   = NodeArray[pscan->key];
		        link(pfrom, pto, -INFINITY, -spacing, 0);
		    } else  {
			/*
			 *  tie; let pfrom be the one of lesser key.
			 */
			if (pscan->key < pedge->key)  {
			    /*
			     *  pedge belongs to pto, ...
			     */
			    if (debug)  {
				printf("l.b. spacing constraint, %d -> %d:  %d,\n",
					pscan->key, pedge->key, spacing);
				printf("  pscan: MAX, at %d, on %s;", pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MIN, at %d, on %s\n", pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pscan->key];
			    pto   = NodeArray[pedge->key];
		            link(pfrom, pto, spacing, INFINITY, 0);
			} else  {
			    /*
			     *  pscan belongs to pto, put in spacing as an
			     *  upper.
			     */
			    if (debug)  {
				printf("u.b. spacing constraint, %d -> %d:  %d,\n",
					pedge->key, pscan->key, -spacing);
				printf("  pscan: MAX, at %d, on %s;", pscan->loc,
					LayerMap[pscan->maskNumber].name);
				printf("  pedge: MIN, at %d, on %s\n", pedge->loc,
					LayerMap[pedge->maskNumber].name);
			    }
			    pfrom = NodeArray[pedge->key];
			    pto   = NodeArray[pscan->key];
		            link(pfrom, pto, -INFINITY, -spacing, 0);
			}
		    }
		    /* return; */
		}
	    } else if (pscan->type == MIN  &&
			pscan->maskNumber == pedge->maskNumber)  {
		if (pscan->max > stateCoord)  {
		    if (debug)  {
			printf("Terminating scan; found edge w/same ");
			printf("polarity, same mask.\n");
		    }
		    return;
		}
	    }
	prevEdge:
	    continue;
	}
    }
}
