#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "tap.h"
#include "errtrap.h"
#include "vov.h"

int SparcsUseTap = 1;

/*
 *  Routine to open the technology file for reading.  There are three
 *  possibilities:
 *
 *	1.  If the -r command-line flag has been used, the variable RulesFile
 *	    contains the path and the filename.
 *	2.  If "sparcs.rules" is present and readable in the working directory,
 *	    use it.
 *	3.  Construct the full path as 'TechRoot/techName/"sparcs.rules"'.
 *
 *  The default for TechRoot is ~octtools/lib/technology.  techName is the value
 *  of the TECHNOLOGY property on the contents facet.
 *
 *  If the TECHNOLOGY property is missing, then RulesFile must be specified.
 *
 *  XXX Possible todo:  allow TechRoot to be reset.
 *
 *  Prior to 4/13/87, this routine known as "initRules()".
 *
 *  Called by:  initOct()
 *  Calls:	readTech()
 *
 *  4/13/87
 *  last modified 8/24/87
 */

void
openTech(facet, techName)
    octObject *facet;
    char *techName;
{
    FILE *pf;
    char buffer[2048];
    char* file;

    
    if ( SparcsUseTap ) {
	readTechFromTap( facet );
    } else {
	file = RulesFile ? util_file_expand(RulesFile) :  "sparcs.rules";

	/*
	 *  Read the rules file.  First see if the user has specified one via the
	 *  -r command-line flag or the rules-file command of the vem version.
	 */

	if ((pf=VOVfopen(file, "r")) == NIL(FILE))  {
	    (void) sprintf(buffer, "%s/%s", tapGetDirectory(facet), file);
	    file = buffer;
	    if ((pf=VOVfopen(util_file_expand(file), "r")) == NIL(FILE))  {
		errRaise("SPARCS",1,
			 "can't open rules file %s ", file );
	    }
	}
	if (TraceFlag)  {
	    printf("openTech:  rules file is \"%s\".\n", file);
	}
	RulesDB = 0;
	readTech(pf, techName);

    }
    return;

}
