#include "copyright.h"
#include "sparcs.h"
#include <pwd.h>
#include "vov.h"
/*
 *  Read the defaults file "sparcs.defaults".  The first place searched for the
 *  file is the working directory, followed by the user's home directory.  The
 *  default values from sparcs.defaults overwrite the hard-coded values.  The
 *  command-line arguments can be used in turn to overwrite the values given in
 *  sparcs.defaults.
 *
 *  4/5/86
 *  last modified 10/9/87
 */

#define BUFLEN 256

void
readDefaults()
{
    struct passwd *ppwd;
    char fileName[BUFLEN];
    FILE *pf;
    int position;
    char lineBuf[80];
    char keyword[80];
    int value;
    int debug = 0;

    /*
     *  First look for the file "sparcs.defaults" in the working directory.
     *  If it isn't found, look in the user's home directory.
     */
    if ((pf=VOVfopen("sparcs.defaults", "r")) != NIL(FILE))  {
	if (debug)  {
	    printf("readDefaults:  sparcs.defaults found in working dir\n");
	}
    } else  {
	if ((ppwd=getpwuid(getuid())) == NIL(struct passwd))  {
	    printf("readDefaults:  can't get your uid!\n");
	    exit(-1);
	}
	(void) strncpy(fileName, ppwd->pw_dir, BUFLEN-1);
	position = strlen(fileName);
	(void) strncat(fileName, "/sparcs.defaults", BUFLEN-position-1);
	if ((pf=VOVfopen(fileName, "r")) != NIL(FILE))  {
	    if (debug)  {
	        printf("readDefaults:  sparcs.defaults found in %s\n",
							    ppwd->pw_dir);
	    }
	} else  {
	    if (debug)  {
		printf("readDefaults:  no readable sparcs.defaults file");
		printf(" in working dir or home dir.\n");
	    }
	    return;
	}
    }

    /*
     *  Read each line and reset the appropriate default.  Lines are of the
     *  form:
     *		keyword  value
     *
     *  Value is ON or OFF.  Lines starting with # are comments.
     */

    while (fgets(lineBuf, 79, pf) != NULL)  {
        if (sscanf(lineBuf, "%s", keyword) != 1)  {
	    printf("readDefaults:  problem reading file %s.\n", fileName);
	    exit(-1);
	}
	if (keyword[0] == '#')  {
	    if (debug)  printf("readDefaults:  comment line...\n");
	    continue;
	} else  {
	    if (sscanf(lineBuf, "%s %d", keyword, &value) != 2)  {
		printf("readDefaults:  problem reading file %s\n",fileName);
		exit(-1);
	    }

	    /*
	     *  Decode the line.  This first set is for the user.
	     */
	    if (strcmp(keyword, "trace") == 0)  {
		if (value == TRUE)  {
		    TraceFlag = TRUE;
		} else if (value == FALSE)  {
		    TraceFlag = FALSE;
		} else  {
		    printf("readDefaults:  trace must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    /********
	    } else if (strcmp(keyword, "incremental") == 0)  {
		if (value == TRUE)  {
		    IncrementalFlag = TRUE;
		} else if (value == FALSE)  {
		    IncrementalFlag = FALSE;
		} else  {
		    printf("readDefaults:  incremental must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    ********/
	    } else if (strcmp(keyword, "vgrid") == 0)  {
		if (value == TRUE)  {
		    VGridFlag = TRUE;
		} else if (value == FALSE)  {
		    VGridFlag = FALSE;
		} else  {
		    printf("readDefaults:  vgrid must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    /*****************
	    } else if (strcmp(keyword, "onestep") == 0)  {
		if (value == TRUE)  {
		} else if (value == FALSE)  {
		} else  {
		    printf("readDefaults:  onestep must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    /*****************
	    } else if (strcmp(keyword, "hfirst") == 0)  {
		if (value == TRUE)  {
		    HorizontalFirstFlag = TRUE;
		} else if (value == FALSE)  {
		    HorizontalFirstFlag = FALSE;
		} else  {
		    printf("readDefaults:  hfirst must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    /*
	     *  The following are for debugging.
	     */
	    } else if (strcmp(keyword, "print_objects") == 0)  {
		if (value == TRUE)  {
		    PrintObjectsFlag = TRUE;
		} else if (value == FALSE)  {
		    PrintObjectsFlag = FALSE;
		} else  {
		    printf("readDefaults:  print_objects must be");
		    printf(" 1 or 0; ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "print_wires") == 0)  {
		if (value == TRUE)  {
		    PrintWiresFlag = TRUE;
		} else if (value == FALSE)  {
		    PrintWiresFlag = FALSE;
		} else  {
		    printf("readDefaults:  print_wires must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "print_graph") == 0)  {
		if (value == TRUE)  {
		    PrintGraphFlag = TRUE;
		} else if (value == FALSE)  {
		    PrintGraphFlag = FALSE;
		} else  {
		    printf("readDefaults:  print_graph must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "print_cpmgraph") == 0)  {
		if (value == TRUE)  {
		    PrintCpmGraphFlag = TRUE;
		} else if (value == FALSE)  {
		    PrintCpmGraphFlag = FALSE;
		} else  {
		    printf("readDefaults:  print_cpmgraph must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "readin_debug") == 0)  {
		if (value == TRUE)  {
		    ReadDB = TRUE;
		} else if (value == FALSE)  {
		    ReadDB = FALSE;
		} else  {
		    printf("readDefaults:  readin_debug must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "rules_debug") == 0)  {
		if (value == TRUE)  {
		    RulesDB = TRUE;
		} else if (value == FALSE)  {
		    RulesDB = FALSE;
		} else  {
		    printf("readDefaults:  rules_debug must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else if (strcmp(keyword, "update_debug") == 0)  {
		if (value == TRUE)  {
		    UpdateDB = TRUE;
		} else if (value == FALSE)  {
		    UpdateDB = FALSE;
		} else  {
		    printf("readDefaults:  update_debug must be 1 or 0;");
		    printf(" ignored.\n");
		    continue;
		}
	    } else  {
		printf("readDefaults:  keyword %s is illegal; ignored.\n",
								    keyword);
		continue;
	    }
	}
    }
    if (fclose(pf) == EOF)  {
	printf("readDefaults:  trouble closing defaults file.\n");
    }
}
