#include "copyright.h"
#include "sparcs.h"
#include "oct.h"
#include "oh.h"
#include "facet.h"
/*
 *  Control routine to do a spacing iteration.  Returns TRUE if no elements
 *  have moved, FALSE otherwise.
 *
 *  8/5/85
 *  from doHorizontalSpacing(), 2/24/87
 *  last modified 12/18/87
 */

int
doSpacing(direction)
directionType direction;
{
    void resetGraph();
    int mergesort(), procUserConstraints();
    void printGraph();
    void printCpmGraph();
    void doCpm();
    void updateOct();
    int updateX(), updateY();
    int converged;
    int numUserConstraints;
    int numMinEdgePtrs, numMaxEdgePtrs, numMidEdgePtrs;
    
    cpm_resetGlobals();
    resetGraph(direction);

    linkEdges(direction);
    
    if (TraceFlag)  {
	printf("    Building graph:\n\tSorting...  ");
	fflush(stdout);
    }
    
    numMaxEdgePtrs = mergesort(&pMaxEdgeListHead, &pMaxEdgeListTail);
    numMinEdgePtrs = mergesort(&pMinEdgeListHead, &pMinEdgeListTail);
    
    /*
     *  XXX  Error check??
     */
    if (pMidEdgeListHead != NIL(struct edgePtr)  && 
        pMidEdgeListTail != NIL(struct edgePtr))  {
        numMidEdgePtrs = mergesort(&pMidEdgeListHead, &pMidEdgeListTail);
    }

    if (TraceFlag)  {
	printf("(%d min) (%d max) (%d mid edgePtrs).\n",
	       numMinEdgePtrs, numMaxEdgePtrs, numMidEdgePtrs);
	fflush(stdout);
    }

    /***
    printf("\n\nMIN edgePtrs:\n");
    printEdgePtrList(pMinEdgeListHead);
    printf("\n\nMID edgePtrs:\n");
    printEdgePtrList(pMidEdgeListHead);
    printf("\n\nMAX edgePtrs:\n");
    printEdgePtrList(pMaxEdgeListHead);
    ***/
    
    doScan(direction);

    /*
     *  The order of the following two calls is correct; don't reverse.
     */
    bindSoSi(direction);
    constraintsForTerminals(direction);

    if (TraceFlag)  {
	printf("\tAdding forces...");
	fflush(stdout);
    }
    addForces(direction);

    if (TraceFlag)  {
	printf("  user-constraints...");
	fflush(stdout);
    }
    numUserConstraints = procUserConstraints(direction);  /* added 6/16/87 */
    if (TraceFlag)  {
	printf("  %d added.\n", numUserConstraints);
	fflush(stdout);
    }

    /*if (VGridFlag)  constraintsForGrid(direction);*/

    if (TraceFlag)  {
	printf("\t    %d Lower-bound,  %d Upper-bound\n", NumLowers,NumUppers);
	printf("\t    %d Nodes,  %d Branches,  %.2f  Branches/Node\n",
	       NumNodes, NumBranches, (float)NumBranches/(float)NumNodes);
	fflush(stdout);
    }
    
    if (PrintGraphFlag)  printGraph();

    if (PrintCpmGraphFlag)  printCpmGraph(direction);

    doCpm(direction);

    if (direction == HORIZONTAL)  {
        converged = updateX();
    } else  {
        converged = updateY();
    }

    showCriticalPath( direction );

    if (ShowIterationsFlag && !converged)  updateOct(direction);

    return(converged);
}



showCriticalPath( direction )
    directionType direction;
{
    struct node *pnode;
    octObject bag, obj;
    char *bagName;
    octGenerator gen;

    bagName = (direction == VERTICAL ) ? "SPARCS Y-CRITICAL" : "SPARCS X-CRITICAL";
    
    ohGetOrCreateBag( &WorkingFacet,&bag, bagName );
    /* If bag exists, empty it of all attachements */
    OH_ASSERT(octInitGenContents( &bag, OCT_ALL_MASK, &gen ));
    while ( octGenerate( &gen, &obj ) == OCT_OK ) {
	OH_ASSERT(octDetach( &bag, &obj ));
    }


    for (pnode=pNodeList; pnode != NULL; pnode = pnode->cpm.list)  {
	if (pnode->cpm.lower == pnode->cpm.upper ) {
	    if ( getOctElementByKey( &obj, pnode->key ) ) {
		octAttach( &bag, &obj );
	    }
	}
    }
}

