#include "copyright.h"
#include "sparcs.h"

/*
 *  To check whether wire segments have changed length, and if so update
 *  their coordinates accordingly.  If the compaction step is horizontal
 *  then the horizontal segments are processed; same for vertical.
 *
 *  TODO:  Use two lists of wire segments, one for horizontal and another
 *	   for vertical, rather than one list for both.
 *
 *  8/23/85
 *  last modified 6/22/87
 */

void
updateExtents(direction)
directionType direction;
{
    struct wire *pw;
    int deltaX, deltaY;

    switch (direction)  {

	case HORIZONTAL:
	    for (pw = pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {

		if (pw->horizontal == FALSE)
		    continue;

		/*
		 *  See if there is an object at end 1 of the segment; if so,
		 *  see if it has moved.  If it HAS moved update the endpoint
		 *  of the segment.
		 */
		if (pw->po1 != NIL(struct object))  {
		    if (pw->po1->lastOrigin.x != pw->po1->origin.x)  {
			/*  MOVED!  */
			deltaX = pw->po1->lastOrigin.x - pw->po1->origin.x;
			pw->p1.x        -= deltaX;
			pw->changed = TRUE;	/*  for updateOct()  */

			/*
			 *  Update the edgePtrs.
			 */
			pw->pfirstMinEP->coord -= deltaX;
			pw->plastMinEP->coord -= deltaX;
		    }
		}

		/*  similarly for object at pw->po2  */
		if (pw->po2 != NIL(struct object))  {
		    if (pw->po2->lastOrigin.x != pw->po2->origin.x)  {
			deltaX = pw->po2->lastOrigin.x - pw->po2->origin.x;
			pw->p2.x        -= deltaX;
			pw->changed = TRUE;	/*  for updateOct()  */

			/*
			 *  Update the edgePtrs.  Also, update the "max" field
			 *  of the two horizontal edges of the wire.
			 */
			pw->pfirstMaxEP->coord -= deltaX;
			pw->pfirstMaxEP->pedge->max -= deltaX;
			pw->plastMaxEP->coord -= deltaX;
			pw->plastMaxEP->pedge->max -= deltaX;
		    }
		}
	    }
	    break;

	case VERTICAL:
	    for (pw = pWireList; pw!=NIL(struct wire); pw=pw->pnext)  {

		if (pw->horizontal == TRUE)
		    continue;

		if (pw->po1 != NIL(struct object))  {
		    if (pw->po1->lastOrigin.y != pw->po1->origin.y)  {
			deltaY = pw->po1->lastOrigin.y - pw->po1->origin.y;
			pw->p1.y        -= deltaY;
			pw->changed = TRUE;

			/*
			 *  Update the edgePtrs.
			 */
			pw->pfirstMinEP->coord -= deltaY;
			pw->plastMinEP->coord -= deltaY;
		    }
		}

		if (pw->po2 != NIL(struct object))  {
		    if (pw->po2->lastOrigin.y != pw->po2->origin.y)  {
			deltaY = pw->po2->lastOrigin.y - pw->po2->origin.y;
			pw->p2.y        -= deltaY;
			pw->changed = TRUE;

			/*
			 *  Update the edgePtrs, and the "max" of the two
			 *  vertical edge.
			 */
			pw->pfirstMaxEP->coord -= deltaY;
			pw->pfirstMaxEP->pedge->max -= deltaY;
			pw->plastMaxEP->coord -= deltaY;
			pw->plastMaxEP->pedge->max -= deltaY;
		    }
		}
	    }
	    break;

	default:
	    fprintf(pout, "updateExtents:  illegal direction.\n");
	    exit(-1);
    }
}
