#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/
#include "grid.h"
/*#include "list.h"*/

/*
 *  Update the graph nodes and the object locations following a HORIZONTAL
 *  spacing via cpm.  To do so, scan through the cpm graph's nodes.
 *
 *  Modified 8/21/85 to update wires.  Since HORIZONTAL spacing, only VERTICAL
 *  segments show up in the graph.  Thus, the x coordinates of their center-
 *  lines must be brought up to date.
 *
 *  LATER - update lengths of the horizontal wire segments.
 *
 *  8/23/85 - calls updateExtents() to change segment lengths for the
 *	      HORIZONTAL segments.
 *
 *  9/5/85 - updates the structs that represent the virtual grid lines
 *
 *  9/6/85 - updates the global bb variables MaxBBX, MinBBX
 *
 *  9/20/85 - Modified to work with the cpm structs integrated with the sparcs
 *	      graph structs (no more file interface)
 *
 *  2/27/87 - Modified to work with scanner structs
 *
 *  7/16/85
 *  last modified 2/8/88
 */

int
updateX()
{
    extern void updateExtents();
    struct node *pn;			/*  a sparcs node  */
    struct object *po;			/*  a sparcs object  */
    struct wire *pw;
    struct gridLine *pgL;		/*  vgrid  */
    struct edge *pedge;		/* for objects (pverticalEdges) */
    struct edgePtr *peP;	/* for wires */
    int deltaX;
    int key;
    int converged = TRUE;
    int i;
    int numTransElts = 0;		/* # instances/wires that translated */

    MaxBBX = -INFINITY;
    MinBBX =  INFINITY;

    for (i=1; i<=NumElements; i++)  {

	/*  note that not all elements are in the graph  */
	pn = NodeArray[i];
	if (pn->key == UNDEF)  continue;

	key = pn->key;

	/*
	 *  NOTE THAT ONLY X IS UPDATED!!!!
	 */
	pn->newLoc.x = pn->cpm.loc;
	pn->newLoc.y = pn->loc.y;

	if (pn->key <= NumObjects)  {
	    po = ElementByKey[pn->key].ptr.po;
	    if (po == NIL(struct object))  {
		fprintf(pout, "updateX:  NIL object pointer.\n");
		return(FALSE);
	    }
	    
	    /*
	     *  NOTE THAT ONLY X IS UPDATED!!!!
	     */
	    po->lastOrigin.x = po->origin.x;
	    
	    /*  only update if location has changed...  */
	    if (po->origin.x != (pn->cpm.loc)*UNITSPERLAMBDA)  {
		converged = FALSE;
		numTransElts++;
		deltaX = po->origin.x - ((pn->cpm.loc)*UNITSPERLAMBDA);
		po->origin.x = (pn->cpm.loc)*UNITSPERLAMBDA;
#ifdef RPC
		po->transform.translation.x -= deltaX;
#endif		
		/*
		 *  Update the "loc" field of the vertical edges.
		 */
		pedge = po->pverticalEdges;
		for (; pedge!=NIL(struct edge); pedge=pedge->plist)  {
		    pedge->loc -= deltaX;
		}
		/*
		 *  Update the "coord" fields of the edgePtrs used for
		 *  vertical spacings, and the "max" field of the
		 *  associated (horizontal) edges.
		 */
		peP = po->pminVEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaX;
		}
		peP = po->pmaxVEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaX;
		    peP->pedge->max -= deltaX;
		}
		peP = po->pmidVEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaX;
		}
	    }
	    
	    if ((po->origin.x + po->bbUpperR.x) > MaxBBX)
		MaxBBX = po->origin.x + po->bbUpperR.x;
	    if (po->origin.x < MinBBX)
		MinBBX = po->origin.x;
	    
	    if (UpdateDB)  {
		fprintf(pout, "\nObject %d:\n", key);
		fprintf(pout,
			"  old node loc is %d, old object origin is %d\n",
			pn->loc.x, po->lastOrigin.x);
		fprintf(pout,
			"  new node loc is %d, new object origin is %d\n",
			pn->newLoc.x, po->origin.x);
	    }
	} else  {
	    pw = ElementByKey[pn->key].ptr.pw;
	    
	    /*
	     *  Only translate vertical wires.
	     */
	    if (pw->horizontal == TRUE)  break;
	    
	    if (pw == NIL(struct wire))  {
		fprintf(pout, "updateX:  NIL wire pointer.\n");
		return(FALSE);
	    }
	    
	    /*  only update if location has changed...  */
	    if ((pw->p1.x - pw->halfWidth) != (pn->cpm.loc)*UNITSPERLAMBDA)  {
		converged = FALSE;
		numTransElts++;
		pw->changed = TRUE;		/*  used by updateOct()  */
		deltaX = pw->p1.x - pw->halfWidth - ((pn->cpm.loc)*UNITSPERLAMBDA);
		pw->p1.x -= deltaX;
		pw->p2.x -= deltaX;
		
		/*
		 *  Update the edges
		 */
		pw->pfirstMinEP->pedge->loc -= deltaX;
		pw->plastMinEP->pedge->loc -= deltaX;
	    }
	    
	    if ((pw->p1.x + pw->halfWidth) > MaxBBX)
		MaxBBX = pw->p1.x + pw->halfWidth;
	    if ((pw->p1.x - pw->halfWidth) < MinBBX)
		MinBBX = pw->p1.x - pw->halfWidth;
	    
	    if (UpdateDB)  {
		fprintf(pout, "\nWire %d:\n", key);
		fprintf(pout, "  old node loc is %d\n", pn->loc.x);
		fprintf(pout, "  new node loc is %d.\n", pn->newLoc.x);
	    }
	}
    }

    /*  now grow/shrink the horizontal segments as needed  */
    updateExtents(HORIZONTAL);

    /*
     *  If there is a virtual grid, update it.  That is, translate the
     *  HorizontalGrid by changing the loc's of the gridlines.
     */
    if (VGridFlag)  {
	if ((pgL = pHorizontalGrid) == NIL(struct gridLine))  {
	    fprintf(pout, "updateX:  NIL pointer to VGrid.\n");
	    exit(100);
	} else  {
	    while (pgL != NIL(struct gridLine))  {
		key = pgL->pelement->key;
		deltaX = ElementByKey[key].ptr.po->lastOrigin.x -
				    ElementByKey[key].ptr.po->origin.x;
		pgL->oldLoc = pgL->loc;
		pgL->loc -= deltaX;

		pgL = pgL->pnext;
	    }
	}
    }
    if (TraceFlag)  printf("    %d Elements translated. \n", numTransElts);
	    
    return(converged);
}
