#include "copyright.h"
#include "sparcs.h"
/*#include "graph.h"*/
#include "grid.h"
/*#include "list.h"*/

/*
 *  Update the graph nodes and the object locations following a VERTICAL
 *  spacing via cpm.  Companion of updateX().
 *
 *  Updates MaxBBY, MinBBY as of 9/6/85
 *
 *  9/20/85 - Modified to work with cpm structs integrated with sparcs structs
 *	      (no more file interface)
 *
 *  7/18/85
 *  last modified 2/8/88
 */

int
updateY()
{
    void updateExtents();
    struct node *pn;			/*  a sparcs node  */
    struct object *po;			/*  a sparcs object  */
    struct wire *pw;
    struct edge *pedge;		/* for objects (phorizontalEdges) */
    struct edgePtr *peP;	/* for wires */
    struct gridLine *pgL;
    int deltaY;
    int key;
    int i;
    int numTransElts = 0;		/* # instances/wires that translated */
    int converged = TRUE;

    MaxBBY = -INFINITY;
    MinBBY =  INFINITY;

    for (i=1; i<=NumElements; i++)  {

	pn = NodeArray[i];
	if (pn->key == UNDEF)  continue;

	key = pn->key;
	if (key == INFINITY)  {
	    if (UpdateDB)
	        fprintf(pout, "\nupdateY:  skipping update of SINK node.\n");
	    continue;
	} else if (key == ZERO)  {
	    if (UpdateDB)
	        fprintf(pout, "\nupdateY:  skipping update of SOURCE node.\n");
	    continue;
	} else  {
	    pn = NodeArray[pn->key];
	}

	/*
	 *  NOTE THAT ONLY Y IS UPDATED!!!!
	 */
	pn->newLoc.x = pn->loc.x;
	pn->newLoc.y = pn->cpm.loc;

	if (pn->key <= NumObjects)  {
	    po = ElementByKey[pn->key].ptr.po;
	    if (po == NIL(struct object))  {
		fprintf(pout, "updateY:  NIL object pointer.\n");
		return(FALSE);
	    }
	    
	    /*
	     *  NOTE THAT ONLY Y IS UPDATED!!!!
	     */
	    po->lastOrigin.y = po->origin.y;
	    
	    /*  only update if location has changed...  */
	    if (po->origin.y != (pn->cpm.loc)*UNITSPERLAMBDA)  {
		converged = FALSE;
		numTransElts++;
		deltaY = po->origin.y - ((pn->cpm.loc)*UNITSPERLAMBDA);
		po->origin.y = (pn->cpm.loc)*UNITSPERLAMBDA;
#ifdef RPC
		po->transform.translation.y -= deltaY;
#endif
		/*
		 *  Update the horizontally-oriented edges.
		 */
		pedge = po->phorizontalEdges;
		for (; pedge!=NIL(struct edge); pedge=pedge->plist)  {
		    pedge->loc -= deltaY;
		}
		/*
		 *  Update the edgePtrs for horizontal spacings.  And,
		 *  update the "max" field of the vertical edges.
		 */
		peP = po->pminHEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaY;
		}
		peP = po->pmaxHEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaY;
		    peP->pedge->max -= deltaY;
		}
		peP = po->pmidHEP;
		for (; peP!=NIL(struct edgePtr); peP=peP->plist)  {
		    peP->coord -= deltaY;
		}
	    }
	    
	    if ((po->origin.y + po->bbUpperR.y) > MaxBBY)
		MaxBBY = po->origin.y + po->bbUpperR.y;
	    if (po->origin.y < MinBBY)
		MinBBY = po->origin.y;
	    
	    if (UpdateDB)  {
		fprintf(pout, "\nObject %d:\n", key);
		fprintf(pout, "  old node loc is %d, ", pn->loc.y); 
		fprintf(pout, "old object origin is %d\n",po->lastOrigin.y);
		fprintf(pout, "  new node loc is %d, ", pn->newLoc.y); 
		fprintf(pout, "new object origin is %d\n", po->origin.y);
	    }
	} else  {
	    pw = ElementByKey[pn->key].ptr.pw;
	    
	    /*
	     *  Only translate horizontal wires.
	     */
	    if (pw->horizontal == FALSE)  break;
	    
	    if (pw == NIL(struct wire))  {
		fprintf(pout, "updateY:  NIL wire pointer.\n");
		return(FALSE);
	    }
	    
	    /*  only update if location has changed...  */
	    if ((pw->p1.y - pw->halfWidth) != (pn->cpm.loc)*UNITSPERLAMBDA)  {
		converged = FALSE;
		numTransElts++;
		pw->changed = TRUE;		/*  used by updateSq()  */
		deltaY = pw->p1.y - pw->halfWidth - ((pn->cpm.loc)*UNITSPERLAMBDA);
		pw->p1.y -= deltaY;
		pw->p2.y -= deltaY;
		
		/*
		 *  Update the edges
		 */
		pw->pfirstMinEP->pedge->loc -= deltaY;
		pw->plastMinEP->pedge->loc -= deltaY;
	    }
	    
	    if ((pw->p1.y + pw->halfWidth) > MaxBBY)
		MaxBBY = pw->p1.y + pw->halfWidth;
	    if ((pw->p1.y - pw->halfWidth) < MinBBY)
		MinBBY = pw->p1.y - pw->halfWidth;
	    
	    if (UpdateDB)  {
		fprintf(pout, "\nWire %d:\n", key);
		fprintf(pout, "  old node loc is %d\n", pn->loc.y);
		fprintf(pout, "  new node loc is %d.\n", pn->newLoc.y);
	    }
	}
    }

    /*  now grow/shrink the vertical segments as needed  */
    updateExtents(VERTICAL);

    /*  if there is a virtual grid, update it.  */
    if (VGridFlag)  {
	if ((pgL = pVerticalGrid) == NIL(struct gridLine))  {
	    fprintf(pout, "updateY:  NIL VGrid pointer.\n");
	    exit(100);
	} else  {
	    while (pgL != NIL(struct gridLine))  {
		key = pgL->pelement->key;
		deltaY = ElementByKey[key].ptr.po->lastOrigin.y -
				    ElementByKey[key].ptr.po->origin.y;
		pgL->oldLoc = pgL->loc;
		pgL->loc -= deltaY;

		pgL = pgL->pnext;
	    }
	}
    }
    if (TraceFlag)  printf("    %d Elements translated. \n", numTransElts);

    return(converged);
}
