#include "copyright.h"
#include "sparcs.h"

/*
 *  Print out the graph such that cpm can read it.  Output is to a file in the
 *  working directory called "fileD#", where "file" is "InCell", "D" is "H"
 *  or "V", and "#" is the iteration number.  Cpm wants to see the graph as:
 *
 *	"fromNode  toNode  lowerBound  upperBound  force"
 *
 *  where fromNode and toNode may be strings and the other three are integers.
 *  For any of lowerBound, upperBound, and force, an undefined value is
 *  symbolized by "*".
 *
 *  7/13/85
 *  last modified 1/28/88
 */

void
printCpmGraph(direction)
directionType direction;
{
    struct node *pn;
    struct branch *pb;
    int i;
    FILE *fopen(), *pf;
    char fileName[SMALLBUF];
    void printBranch();

    if (strlen(InCell) > 73)  {
	fprintf(pout, "printCpmGraph:  fileName array overflow.\n");
	return;
    }
    if (direction == HORIZONTAL)  {
	sprintf(fileName, "%sH%d", InCell, Iterations+1);
    } else  {
	sprintf(fileName, "%sV%d", InCell, Iterations+1);
    }

    if ((pf = fopen(fileName, "w")) == NIL(FILE))  {
	fprintf(pout, "printCpmGraph:  can't open file \"%s\".\n", fileName);
	return;
    }

    /*
     *  The source node's fanouts.
     */
    pb = pSource->pfanouts;
    for (; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	printBranch(pb, pf);
    }

    /* 
     *  The fanouts of the intermediate nodes.
     */
    for (i=1; i<=NumElements; i++)  {
	pn = NodeArray[i];
	if (pn->key == UNDEF)  {
	    continue;	    /* may not have a node for each element */
	}
	for (pb=pn->pfanouts; pb!=NIL(struct branch); pb=pb->pnextFanout)  {
	    printBranch(pb, pf);
	}
    }

    if (fclose(pf) == EOF)  {
	fprintf(pout, "printCpmGraph:  can't close file.\n");
	return;
    }
}


/*
 *  11/2/87
 *  last modified 2/4/88
 */

void
printBranch(pb, pf)
struct branch *pb;
FILE *pf;
{
    if (pb->pfromNode->key == ZERO)  {
	fprintf(pf, "%s ", "s");
    } else  {
	fprintf(pf, "%d ", pb->pfromNode->key);
    }
    if (pb->ptoNode->key == INFINITY)  {
	fprintf(pf, "%s ", "t");
    } else  {
	fprintf(pf, "%d ", pb->ptoNode->key);
    }
    if (pb->lowerBound == -INFINITY)  {
	fprintf(pf, "* ");
    } else  {
	fprintf(pf, "%d ", pb->lowerBound);
    }
    if (pb->upperBound == INFINITY)  {
	fprintf(pf, "* ");
    } else  {
	fprintf(pf, "%d ", pb->upperBound);
    }
    if (pb->force == UNDEF)  {
	fprintf(pf, "*\n");
    } else  {
	fprintf(pf, "%d\n", pb->force);
    }
}
