#include "copyright.h"
#include "port.h"
#include "spider.h"

double colAssErr();
pcol COLS;
pcol newCol();
extern int MIN_SPACING;

/********************************************************************/

check_same_side_pins()
{
pin_type *pin;
int side;
int del;

for(side = TOP;side <= BOTTOM;side++)
  for(pin = FIRST_PIN(side);(pin != NULL) && (pin->next != NULL);pin = pin->next)
    if(netlist[pin->netid]->numpins > 1)
      {
	if((del = (pin->next->x - pin->next->netwidth /2.) - (pin->x + pin->netwidth /2. + MIN_SPACING)) < 0)
	  {
	    printf("(check pins): Spacing violation. %s side\n", 
		   (side==TOP)?"TOP":"BOTTOM");
	    
	    fprintf(stderr,"(SPIDER) ERROR - pins too close\n");
	    fprintf(stderr,"(SPIDER) 	PIN %s at loc %d, %d width %d\n", 
		   pin->next->pinname, 
		      (channel_types == VERTICAL)?pin->next->x:(CHAN_BOUND->ur_y - pin->next->y),
		      (channel_types == VERTICAL)?pin->next->y:pin->next->x,
		    pin->next->netwidth);

	    fprintf(stderr,"(SPIDER) 	PIN %s at loc %d, %d width %d\n", 
		   pin->pinname, 
		      (channel_types == VERTICAL)?pin->x:(CHAN_BOUND->ur_y - pin->y),
		      (channel_types == VERTICAL)?pin->y:pin->x,
		    pin->netwidth);
	    
	    del = -2 * del;
	    if(pin->netwidth > layer_width[pin->layer])
	      {
		if(pin->netwidth - layer_width[pin->layer] >= del)
		  {
/*		    printf("(spider): WARNING -- net %s narrowed from %d to ",pin->pinname,pin->netwidth);*/
		    pin->netwidth = SPquantizeWidthDOWN( pin->netwidth - del);
		    printf("%d\n",pin->netwidth);
/*		    logError("WARNING  net was narrowed to connect to cell\n",NONFATAL);*/
		    continue;
		  }
		else
		  {
		    del -= (pin->netwidth - layer_width[pin->layer]);
/*		    printf("WARNING -- net %s narrowed from %d to ",pin->pinname,pin->netwidth);*/
		    pin->netwidth = layer_width[pin->layer];
		    printf("%d\n",pin->netwidth);
		    if(pin->next->netwidth - layer_width[pin->next->layer] >= del)
		      {
/*			printf("WARNING -- net %s narrowed from %d to ",pin->next->pinname,pin->next->netwidth);*/
			pin->next->netwidth = SPquantizeWidthDOWN( pin->next->netwidth - del);
			printf("%d\n",pin->next->netwidth);
/*			logError("WARNING  net was narrowed to connect to cell\n",NONFATAL);*/
			continue;
		      }
		  }
	      }
	    else
	      {
		if(pin->next->netwidth - layer_width[pin->next->layer] >= del)
		  {
/*		    printf("WARNING -- net %s narrowed from %d to ",pin->next->pinname,pin->next->netwidth); */
		    pin->next->netwidth = SPquantizeWidthDOWN( pin->next->netwidth - del);
		    printf("%d\n",pin->next->netwidth);
/*		    logError("WARNING  net was narrowed to connect to cell\n",NONFATAL); */
		    continue;
		  }
	      }
	    
	    fprintf(stderr," (SPIDER)  Pin %s deleted\n",pin->pinname);
	    logError(error, NONFATAL);
	    
	    netlist[pin->netid]->numpins -= 1;
	    pin->netid = 0;
	  }
      }
}


/********************************************************************/

colSpacingOK()
{
pcol c;
int good;

good = TRUE;
for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
  {
    if(c->next->loc - c->loc < GRIDWIDTH)
      {
	printf("col adjustment needed -- col %d at %d\n",c->index,c->loc);
	good = FALSE;
      }
  }
return(good);
}

/********************************************************************/

insertBlankCols()
{
pcol c,cc;
int gap;
int num;
int newCols;
int loc;

for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
  {
    gap = (c->next->loc - c->next->width/2.) - (c->loc + c->width/2.) - MAX_SPACING;
    if(gap < 0)
      printf("ERROR in col assignment -- col %d at %d\n",c->index,c->loc);
    if(gap >= GRIDWIDTH)
      {
	newCols = gap/GRIDWIDTH;
	loc = c->loc + c->width/2. + MAX_SPACING + CONT_WIDTH/2.;
	for(num = newCols;num > 1;num--,loc += GRIDWIDTH)
	  {
	    cc = newCol();
	    cc->next = c->next;
	    if(cc->next != NULL)
	      cc->next->prev = cc;
	    c->next = cc;
	    cc->prev = c;
	    c = cc;
	    cc->index = MAXINT;
	    cc->width = CONT_WIDTH;
	    cc->loc = loc;
	  }
	loc = c->next->loc - c->next->width/2. - MAX_SPACING - CONT_WIDTH/2;
	cc = newCol();
	cc->next = c->next;
	if(cc->next != NULL)
	  cc->next->prev = cc;
	c->next = cc;
	cc->prev = c;
	c = cc;
	cc->index = MAXINT;
	cc->width = CONT_WIDTH;
	cc->loc = loc;
      }
  }
colSpacingOK();
}

/********************************************************************/

/* make sure that at least one blank col added to start */

padChanStart(start,end)
int start,end;
{
pcol save,c;
int gap,num,colWid,loc;


if(COLS == NULL)
  gap = end - start;
else
  {
    if(start > COLS->loc - COLS->width/2. - MAX_SPACING - CONT_WIDTH/2)
      start = COLS->loc - COLS->width/2. - MAX_SPACING - CONT_WIDTH/2;
    gap = (COLS->loc - COLS->width/2.) - start;
  }
num = 1 + (gap - (CONT_WIDTH/2. + MAX_SPACING)) / GRIDWIDTH;
colWid = (gap + CONT_WIDTH/2.) / num;
save = NULL;
for(loc = start;num > 0;num--,loc += colWid)
  {
    if(save == NULL)
      save = c = newCol();
    else
      {
	c->next = newCol();
	c->next->prev = c;
	c = c->next;
      }
    c->loc = loc;
    c->width = CONT_WIDTH;
    c->index = MAXINT;
  }
if(COLS != NULL)
  {
    c->next = COLS;
    COLS->prev = save;
  }
COLS = save;
}

/********************************************************************/

/* make sure that at least one blank col added to end */

padChanEnd(end)
int end;
{
pcol c;
int gap;
int num,colWid,loc;

for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
  {}
if(end < (c->loc + c->width/2. + MAX_SPACING + CONT_WIDTH / 2))
  end = (c->loc + c->width/2. + MAX_SPACING + CONT_WIDTH / 2);
gap = (end + CONT_WIDTH/2) - (c->loc + c->width/2.);

num = gap / GRIDWIDTH;
if(num == 0)
  logError("ERROR in determining the end columns\n",FATAL);
colWid = gap / num;
for(loc = end - (num - 1) * colWid;num > 0;num--,loc += colWid)
  {
    c->next = newCol();
    c->next->prev = c;
    c = c->next;
    c->loc = loc;
    c->width = CONT_WIDTH;
    c->index = MAXINT;
  }
}

/********************************************************************/

print_col(cols)
pcol cols;
{

printf("\n--------------------- columns -----------------------------\n");
for(;cols != NULL;cols = cols->next)
    printf("index %11d loc %d top pin %9d bot pin %9d off %5d %5d\n",cols->index,cols->loc,(cols->pin[TOP] != NULL)?cols->pin[TOP]->x:0,(cols->pin[BOTTOM] != NULL)?cols->pin[BOTTOM]->x:0,(cols->pin[TOP] != NULL)?cols->loc - cols->pin[TOP]->x:0,(cols->pin[BOTTOM] != NULL)?cols->loc - cols->pin[BOTTOM]->x:0);
fflush(stdout);
}

/********************************************************************/

print_pins()
{
/*
pin_type *pin;
*/

return;
/*
for( pin = FIRST_PIN(TOP);pin != NULL;pin = pin->next)
  printf("top pin %s at %d col %d\n",pin->pinname,pin->x,pin->col);
for( pin = FIRST_PIN(BOTTOM);pin != NULL;pin = pin->next)
  printf("bot pin %s at %d col %d\n",pin->pinname,pin->x,pin->col);
*/
}

/********************************************************************/

setUpColData()
{
int i,j;
pcol c,cols;

for(i=1,c = COLS;c != NULL;c = c->next,i++)
  {}

RC_LOC->col_loc   = CALLOC( int, i+2);
RC_LOC->col_pins   = CALLOC(pin_type **, 2);
RC_LOC->col_pins[0]= CALLOC(pin_type *, i+2);
RC_LOC->col_pins[1]= CALLOC(pin_type *, i+2);
for (j=0;j <= i+1; j++)  
  RC_LOC->col_pins[TOP][j] = RC_LOC->col_pins[BOTTOM][j] = NULL;

for(i=1,cols = COLS;cols != NULL;cols = cols->next,i++)
  {
    RC_LOC->col_loc[i] = cols->loc;
    RC_LOC->col_pins[TOP][i] = RC_LOC->col_pins[BOTTOM][i] = NULL;

    if(cols->pin[TOP] != NULL)
      {
	RC_LOC->col_pins[TOP][i] = cols->pin[TOP];
	cols->pin[TOP]->col = i;
	cols->pin[TOP] = NULL;
      }    
    if(cols->pin[BOTTOM] != NULL)
      {
	RC_LOC->col_pins[BOTTOM][i] = cols->pin[BOTTOM];
	cols->pin[BOTTOM]->col = i;
	cols->pin[BOTTOM] = NULL;
      }
    if((cols->pin[TOP] != NULL) || (cols->pin[BOTTOM] != NULL))
      logError("ERROR in translating pins to cols\n",FATAL);
  }
return(i - 1);
}

/********************************************************************/

vertex_type *getMinXVert(tvert,bvert)
vertex_type **tvert,**bvert;
{
int t_x,b_x;
vertex_type *vert1;

t_x = MAXINT;
if((*tvert) != NULL)
  t_x = (*tvert)->x;
b_x = MAXINT;
if((*bvert) != NULL)
  b_x = (*bvert)->x;

if(t_x < b_x)
  {
    vert1 = (*tvert);
    if((*tvert)->next == NULL)
      (*tvert) = NULL;
    else
      if((*tvert)->next->next == NULL)
	(*tvert) = (*tvert)->next;
      else
	(*tvert) = (*tvert)->next->next;
  }
else
  {
    vert1 = (*bvert);
    if((*bvert)->next == NULL)
      (*bvert) = NULL;
    else
      if((*bvert)->next->next == NULL)
	(*bvert) = (*bvert)->next;
      else
	(*bvert) = (*bvert)->next->next;
  }
return(vert1);
}

/********************************************************************/

countPins(side,l,r)
int side,l,r;
{
pin_type *pin;
int i;

for(pin = FIRST_PIN(side);(pin != NULL) && (pin->x <= l);pin = pin->next)
  {}
for(i = 1;
    (pin != NULL) && (pin->next != NULL) && (pin->next->x < r);
    pin = pin->next,i++)
  {}
return(i);
}

/********************************************************************/

double colAssErr(c)
pcol c;
{
double err;

for(err = 0;c != NULL;c = c->next)
  {
    if(c->pin[TOP] != NULL)
      err += (c->pin[TOP]->x - c->loc) * (c->pin[TOP]->x - c->loc);
    if(c->pin[BOTTOM] != NULL)
      err += (c->pin[BOTTOM]->x - c->loc) * (c->pin[BOTTOM]->x - c->loc);
  }
return(err);
}

/********************************************************************/

shiftPinLocs()
{
pcol c;
int moved,pass;

pass = 0;
  do {
    pass++;
    moved = FALSE;
    for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
      if((c->pin[TOP] != NULL) && (shiftPin(TOP, c) == TRUE))
	moved = TRUE;
      else
	if(c->pin[BOTTOM] != NULL)
	  moved |= shiftPin(BOTTOM, c);

    for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
      {}
    for(;(c != NULL) && (c->prev != NULL);c = c->prev)
      if((c->pin[TOP] != NULL) && (shiftPin(TOP, c) == TRUE))
	moved = TRUE;
      else
	if(c->pin[BOTTOM] != NULL)
	  moved |= shiftPin(BOTTOM, c);
    
    for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
      if((c->pin[BOTTOM] != NULL) && (shiftPin(BOTTOM, c) == TRUE))
	moved = TRUE;
      else
	if(c->pin[TOP] != NULL)
	  moved |= shiftPin(TOP, c);

    for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
      {}
    for(;(c != NULL) && (c->prev != NULL);c = c->prev)
      if((c->pin[BOTTOM] != NULL) && (shiftPin(BOTTOM, c) == TRUE))
	moved = TRUE;
      else
	if(c->pin[TOP] != NULL)
	  moved |= shiftPin(TOP, c);
  }while((moved == TRUE) && (pass <= 5));
return(pass != 1);
}

/********************************************************************/

shiftPin(side,c)
int side;
pcol c;
{
if(c->pin[side]->x != c->loc)
    return(shiftSingleCol(c,c->pin[side]->x));
return(FALSE);
}

/********************************************************************/

remove_cols()
{
pcol c,cc;

for(c = COLS;c != NULL;c = c->next)
  {
    if((c->pin[TOP] == NULL) && (c->pin[BOTTOM] == NULL))
      {
	cc = c;
	c = c->next;
	deleteCol(cc);
      }
  }
}

/********************************************************************/

movePins()
{
register pcol c;
vertex_type *tvert, *bvert, *vert1, *vert2;
int t_cnt,b_cnt,pass;
register int moved;
/*
printf("begin move pins\n");
*/
pass = 0;
do {
  tvert = CHAN_BOUND->first[TOP];
  bvert = CHAN_BOUND->first[BOTTOM];
  vert1 = getMinXVert(&tvert,&bvert);
  colAssErr(COLS);
  pass++;
  moved = FALSE;
  while((tvert != NULL) || (bvert != NULL))
    {
      vert2 = getMinXVert(&tvert,&bvert);
      t_cnt = countPins(TOP,vert1->x,vert2->x);
      b_cnt = countPins(BOTTOM,vert1->x,vert2->x);
      if(t_cnt < b_cnt)
	{
	  for(c = COLS;
	      (c != NULL) && 
	      (c->next != NULL) && 
	      (c->next->loc <= vert1->x);c = c->next)
	    {}
	  for(;(c != NULL) && 
	       (c->next != NULL) && 
	       (c->next->loc <= vert2->x);c = c->next)
	    {
	      moved |= moveSinglePin(TOP, c);
	      moved |= moveSinglePin(BOTTOM, c);
	    }
	  for(c = COLS;(c != NULL) && (c->loc < vert2->x);c = c->next)
	    {}
	  for(;(c != NULL) && 
	       (c->prev != NULL) && 
	       (c->prev->loc >= vert1->x);c = c->prev)
	    {
	      moved |= moveSinglePin(TOP, c);
	      moved |= moveSinglePin(BOTTOM, c);
	    }
	}
      else
	{
	  for(c = COLS;
	      (c != NULL) && 
	      (c->next != NULL) && 
	      (c->next->loc <= vert1->x);c = c->next)
	    {}
	  for(;(c != NULL) && 
	      (c->next != NULL) && 
	      (c->next->loc <= vert2->x);c = c->next)
	    {
	      moved |= moveSinglePin(BOTTOM, c);
	      moved |= moveSinglePin(TOP, c);
	    }
	  
	  for(c = COLS;(c != NULL) && (c->loc < vert2->x);c = c->next)
	    {}
	  for(;(c != NULL) && 
	      (c->prev != NULL) && 
	      (c->prev->loc >= vert1->x);c = c->prev)
	    {
	      moved |= moveSinglePin(BOTTOM, c);
	      moved |= moveSinglePin(TOP, c);
	    }
	}
    }
}while((moved == TRUE) && (pass <= 5));
return(pass != 1);
}

/********************************************************************/

doubleShift()
{
pcol c;
int moved;
int pass;

pass = 0;
do
  {
    moved = FALSE;
    pass++;
    for(c = COLS;c != NULL;c = c->next)
      if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
	moved |= shiftSingleCol(c,(c->pin[TOP]->x + c->pin[BOTTOM]->x) / 2);
    for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
      {}
    for(;c != NULL;c = c->prev)
      if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
	moved |= shiftSingleCol(c,(c->pin[TOP]->x + c->pin[BOTTOM]->x) / 2);
  }while(moved == TRUE);
return(pass != 1);
}

/********************************************************************/

shiftSingleCol(c,loc)
pcol c;
int loc;
{
if(loc < c->loc)
  {
    if(c->prev == NULL)
      {
	c->loc = loc;
	return(TRUE);
      }
    else
      if(c->prev->loc < c->loc - GRIDWIDTH)
	{
	  c->loc = MAX(loc ,c->prev->loc + GRIDWIDTH);
	  return(TRUE);
	}
  }
else
  if(loc > c->loc)
    {
      if(c->next == NULL)
	{
	  c->loc = loc;
	  return(TRUE);
	}
      else
	if(c->next->loc > c->loc + GRIDWIDTH)
	  {
	    c->loc = MIN(loc ,c->next->loc - GRIDWIDTH);
	    return(TRUE);
	  }
    }
return(FALSE);
}

/********************************************************************/

moveSinglePin(side, c)
register pcol c;
register int side;
{
if((c->pin[side] != NULL) && (abs(c->pin[side]->x - c->loc) > GRIDWIDTH/2))
  {
    if((c->pin[side]->x < c->loc) && (c->prev->pin[side] == NULL) &&
       (abs(c->pin[side]->x - c->prev->loc) < abs(c->pin[side]->x - c->loc)))
      {
	c->prev->pin[side] = c->pin[side];
	c->pin[side] = NULL;
	return(TRUE);
      }
    else
      if((c->pin[side]->x > c->loc) && (c->next->pin[side] == NULL) &&
	 (abs(c->pin[side]->x - c->next->loc) < abs(c->pin[side]->x - c->loc)))
	{
	  c->next->pin[side] = c->pin[side];
	  c->pin[side] = NULL;
	  return(TRUE);
	}
  }
return(FALSE);
}

/********************************************************************/

optimize()
{
int moved;
double best,new;
int i;

if(COLS == NULL)
  return;
i = 0;
best = 1E30;
do
  {
    moved = FALSE;

    moved = movePins();
    moved |= movePins2();

    new = colAssErr(COLS);
    if((new == best) || (new == 0.0))
      break;
    if(new < best)
      best = new;

    remove_cols();
    moved |= shiftPinLocs();
    
    insertBlankCols();
    if((colAssErr(COLS) == 0.0) || (moved == FALSE))
      break;

    moved = movePins();
    moved |= movePins2();
    remove_cols();
    moved |= shiftPinLocs();
    moved |= doubleShift();
    insertBlankCols();
    if(colAssErr(COLS) == 0.0)
      break;
    i++;
  }while((moved == TRUE) && (i < 5));

colSpacingOK();
}

/********************************************************************/

moveSinglePin2(side, c)
pcol c;
int side;
{
int del;

if((c->pin[!side] != NULL) && (c->pin[side] != NULL))
  {
    del = abs(c->pin[side]->x - c->pin[!side]->x);
    if((c->prev != NULL) &&
       (c->prev->pin[side] == NULL) &&
       (c->prev->pin[!side] != NULL) &&
       (abs(c->pin[side]->x - c->prev->pin[!side]->x) < del))
      {
	c->prev->pin[side] = c->pin[side];
	c->pin[side] = NULL;
	return(TRUE);
      }
    else
      if((c->next != NULL) &&
	 (c->next->pin[!side] != NULL) &&
	 (c->next->pin[side] == NULL) &&
	 (abs(c->pin[side]->x - c->next->pin[!side]->x) < del))
	{
	  c->next->pin[side] = c->pin[side];
	  c->pin[side] = NULL;
	  return(TRUE);
	}
  }
return(FALSE);
}

/********************************************************************/

movePins2()
{
pcol c;
int moved;

moved = 0; 
for(c = COLS;c != NULL;c = c->next)
  if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
    {
      moved |= moveSinglePin2(TOP, c);
      if(c->pin[TOP] != NULL)
	moved |= moveSinglePin2(BOTTOM, c);
    }
for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
  {}
for(;c != NULL;c = c->prev)
  if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
    {
      moved |= moveSinglePin2(TOP, c);
      if(c->pin[TOP] != NULL)
	moved |= moveSinglePin2(BOTTOM, c);
    }

for(c = COLS;c != NULL;c = c->next)
  if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
    {
      moved |= moveSinglePin2(BOTTOM, c);
      if(c->pin[TOP] != NULL)
	moved |= moveSinglePin2(TOP, c);
    }
for(c = COLS;(c != NULL) && (c->next != NULL);c = c->next)
  {}
for(;c != NULL;c = c->prev)
  if((c->pin[TOP] != NULL) && (c->pin[BOTTOM] != NULL))
    {
      moved |= moveSinglePin2(BOTTOM, c);
      if(c->pin[TOP] != NULL)
	moved |= moveSinglePin2(TOP, c);
    }
return(moved);
}

/********************************************************************/

convert()
{
pcol c;
pin_type *pin;

for(c = COLS;c != NULL;c = c->next)
  {
    if(c->pin[TOP] != NULL)
      {
	for(pin = FIRST_PIN(TOP);(pin != NULL) && (pin != c->pin[TOP]);pin = pin->next)
	  {}
	if(pin == NULL)
	  {
	    pin = c->pin[BOTTOM];
	    c->pin[BOTTOM] = c->pin[TOP];
	    c->pin[TOP] = pin;
	  }
      }
    if(c->pin[BOTTOM] != NULL)
      {
	for(pin = FIRST_PIN(BOTTOM);(pin != NULL) && (pin != c->pin[BOTTOM]);pin = pin->next)
	  {}
	if(pin == NULL)
	  {
	    pin = c->pin[TOP];
	    c->pin[TOP] = c->pin[BOTTOM];
	    c->pin[BOTTOM] = pin;
	  }
      }
  }
}

/********************************************************************/

deleteCol(c)
pcol c;
{
if(c == COLS)
  COLS = c->next;
if(c->prev != NULL)
  c->prev->next = c->next;
if(c->next != NULL)
  c->next->prev = c->prev;
FREE(c);
}
