#include "copyright.h"
#include "port.h"
#include "spider.h"

extern pjogged JOGGED_PINS;
static int case_num;
static term_type   routing_term;

connectpins()

/****************************************************************
 *
 *  connect_pins makes the connection between the pins along the edges
 *  of the channel to the routes inside the channel. It is assumed that
 *  the router will make connections to the top and bottom edges of the
 *  channel with LAYER1, and connections to the left and right edges
 *  with LAYER2.
 *
 ********************************************************************/

{
    int col,row;
    int subnet_layer,lay;
    int         row_1;
    int 	y1;

    int		netid;
    term_type   pin_term,term3,term4;

    char 	vianame[80];
    pin_type 	*pin;		/* the pin being connected */
    subnet_type *subnet;
    pjogged 	jog;
    int		side;

    /* connect top and bottom edge pins */

    for (side=TOP;  side <= BOTTOM;  side++)
      for (pin = FIRST_PIN(side); pin != NULL; pin = pin->next) {

	netid = pin->netid;
	pin_term.pin = pin;

	if (netlist[netid]->numpins > 1) {

	    row = get_row(side,pin);
	    col = pin->col;

	    routing_term.via = find_via(netid,row,col);
	    if(routing_term.via == NULL)
	      routing_term.via = checkForYacrScrewUp(pin,netid,row,col,side);
	    if (routing_term.via == NULL || routing_term.via->netid != netid)  
	      { 
		printf("Bad layer-map;  col,row= %d %d net= %d for pin %s\n",
		       col, row, netid, pin->pinname);
		printf("layer1,2[]= %d %d, pin:layer,col= %d %d\n",
		       ROUTING->layer1[row][col], ROUTING->layer2[row][col], 
		       pin->layer, pin->col);
		logError("Bad layer-map in CONNECTPIN. (via=NULL)",COREDUMP);
	      } 
/*
	    if((side == TOP) && (pin->y <= routing_term.via->y))
	      logError("Error in connectpin. term location (1)",COREDUMP);
	    if((side == BOTTOM) && (pin->y >= routing_term.via->y))
	      logError("Error in connectpin. term location (2)",COREDUMP);
*/

	    subnet_layer = -1;
	    for (subnet = netlist[netid]->subnetlist;
		 subnet!= NULL;
		 subnet =  subnet->next)  
	      {
		if ((subnet->col1 == col  &&  subnet->row1 == row) ||
		    (subnet->col2 == col  &&  subnet->row2 == row))  
		  {
		    if(subnet_layer == -1)
		      {
			subnet_layer = subnet->layerid;
		      }
		    else
		      if (subnet->layerid != subnet_layer) 
			{
			  subnet_layer = 0;
			  break;
			}
		  }
	      }

	/* Determine if it is a real via: */

	/*  
	 *  CASE1 -- 
	 *
	 *  --  pin is on the same layer as all connecting subnets 
	 *  ==> 
	 *      make a subnet between pin and connecting via
	 *	change the connecting via to psudo via
	 */
	 
	/*  
	 *  CASE2 -- 
	 *
	 *  --  pin is not on the same layer as all connecting subnets 
	 *      but there exist a real via at connection point
	 *  ==> 
	 *      make a subnet between pin and connecting via
	 */
	 
	/*  

	 *  CASE3 -- 
	 *
	 *  --  pin is not on the same layer as all connecting subnets 
	 *      but there is nothing in the col, row location,
	 *      in other word, we can put a via in connection point
	 *  ==> 
	 *      change the connection via to be a real via
	 *      make 1 subnet between pin and connecting via
	 */
        /*
	 *  CASE 4 --
	 *  
	 *  Pin is the result of a jog to avoid a power line.
	 *  ==>
	 *  jump from via to pin loc horizontaly in layer 2
	 */
	 
	/*  
	 *  CASE5 -- 
	 *
	 *  --  pin is not on the same layer as all connecting subnets 
	 *      and we can not put a via in connection point
	 *  ==> 
	 *      add a new via  between pin and connecting via
	 *      make 2 subnets between pin, new via and connecting via
	 */
	/*  
	 *  AND BE CAREFUL-- 
	 *
	 *  if  pin is not on the grid line (col) 
	 *  ==> 
	 *      add a new via  below the pin at middle of pin and row;
	 *      add a new via  above the row at middle of pin and row;
	 *      make 3 subnets connecte pin, new vias and connecting via;
	 *	(change case# to 5 to indicate need bld 3 new subnets)
	 */

	/*  Determine the connection cases */

	    if((pin->jogged == TRUE) && (pin->x != routing_term.via->x))
	      {
	        case_num = 4;
	        for(jog = JOGGED_PINS;
		  (jog != NULL) && (jog->pin != pin);jog = jog->next)
		  {}
 	      }
	    else
	      if (pin->layer == subnet_layer)  
		case1(row,col,subnet_layer);
	      else
		if (routing_term.via->layer1 != routing_term.via->layer2)  
		  case2();
		else
		  if (ROUTING->layer2[row][col]==0 || 
		      ROUTING->layer1[row][col]==0)  
		    case3(netid,row,col);
		  else
		    case5(netid,&row,col,subnet_layer,side);
	    
	    row = get_row(side,pin);  /* may have changed */

	    /* Compute subnets coor;  bldvias */

	    if (pin->x == routing_term.via->x) 
	      {
		/* one new subnet */
		bldsubnet(netid, 
			  pin->x, pin->y, pin->x, routing_term.via->y, 
			  pin->netwidth, pin->layer, -row, col,
			  row,col,&pin_term,&routing_term,PIN,VIA,VERTICAL);
	      }
	    else  
	      {
		if(case_num == 4)
		  {
		    row_1 = jog->row;
		    y1 = RC_LOC->row_loc[row_1];
 		    
		    if(abs(routing_term.via->x - pin->x) > pin->netwidth)
		      lay = LAYER2;
		    else
		      lay = LAYER1;

		    term3.via = bldvia("jog1", netid, pin->x, y1, col, row_1, pin->layer, lay);
		    
		    bldsubnet(netid, 
			      pin->x, pin->y, pin->x, y1, 
			      layer_width[pin->layer], pin->layer, 
			      -row,col,row_1,col,
			      &pin_term,&term3,PIN,VIA,VERTICAL);

		    if (y1 == routing_term.via->y)
		      {
			bldsubnet(netid, pin->x, y1, routing_term.via->x, y1, 
				  layer_width[lay], lay, 
				  row_1,col,row_1,routing_term.via->col,
				  &term3,&routing_term,VIA,VIA,HORIZONTAL); 
			routing_term.via->layer1 = MET1;
			routing_term.via->layer2 = MET2;
		      }
		    else
		      {
			term4.via = bldvia("jog2", netid, routing_term.via->x, y1, col, 
					   row_1, lay, LAYER1);
			
			bldsubnet(netid, 
				  pin->x, y1, routing_term.via->x, y1, 
				  layer_width[lay], lay, 
				  row_1,col,row_1,routing_term.via->col,
				  &term3,&term4,VIA,VIA,HORIZONTAL); 

			bldsubnet(netid, 
				  routing_term.via->x, y1, routing_term.via->x, routing_term.via->y, 
				  layer_width[LAYER1], LAYER1, 
				  row_1, routing_term.via->col, row,routing_term.via->col,
				  &term4,&routing_term,VIA,VIA,VERTICAL); 
		      }
		    
		  }
		else
		  {
		    if(routing_term.via->layer1 == routing_term.via->layer2)
		      lay = routing_term.via->layer1;
		    else
		      lay = pin->layer;
		    term3.via = bldvia("jog1a", netid, 
				       pin->x, routing_term.via->y, col, routing_term.via->row,
				       pin->layer, lay);
		    
		    bldsubnet(netid,
			      pin->x, pin->y, pin->x, routing_term.via->y,
			      layer_width[pin->layer], pin->layer,
			      -row,col,routing_term.via->row,col,
			      &pin_term,&term3,PIN,VIA,VERTICAL);
		    
		    bldsubnet(netid, 
			      term3.via->x, term3.via->y, routing_term.via->x, routing_term.via->y, 
			      layer_width[lay], lay, 
			      term3.via->row, term3.via->col, routing_term.via->row,routing_term.via->col,
			      &term3,&routing_term,VIA,VIA,HORIZONTAL); 
		  }
	      }
	  } 
      }
printf("Complete connectpins.c\n\n");
}



connectexits()
/*
 *  connect left, right edge pins
 */
{
int 	x1;
int         row, col;
int		netid;

term_type   term3;
char 	vianame[80];
pin_type 	*pin;		/* the pin being connected */
int		side;
pin_type **pin_map[4];

pin_map[LEFT]= CALLOC(pin_type *, NUM_ROWS(ROUTING));
pin_map[RIGHT]= CALLOC(pin_type *, NUM_ROWS(ROUTING));
FOREACH_ROW(ROUTING, row)
  {
    pin_map[LEFT][row] = pin_map[RIGHT][row] = NULL;
  }

if (CHAN_PINS->pinstatus == FLOATING)  
  {
    for (side=LEFT;  side <= RIGHT;  side++) 
      {
	if (side == LEFT)	
	  col = 1;
	else
	  col = LAST_COL(ROUTING);
	
	for (pin = FIRST_PIN(side); pin != NULL; pin = pin->next) 
	  {
	    netid = pin->netid;
	    
	    if (netlist[netid]->numpins > 1)	      
	      {
		/*  find the row location of the connection via */
		row = 0;
		while ((row <= LAST_ROW(ROUTING)) &&
		       (ROUTING->layer2[row][col] != netid))
		  row++;
		pin_map[side][row] = pin;
		
		/* Get connecting via */
		
		routing_term.via = find_via(netid,row,col);
		if(routing_term.via == NULL)
		  routing_term.via = checkForYacrScrewUp(pin,netid,row,col,side);
		
		if((routing_term.via == NULL) || 
		   (routing_term.via->netid != netid))
		  {
		    if(routing_term.via == NULL)
		      printf("Via not found\n");
		    else
		      printf("Via on wrong net %d %d\n",netid, routing_term.via->netid);
		    printf("Bad layer-map2;  col,row= %d %d net= %d for pin %s\n",
			   col, row, netid, pin->pinname);
		    printf("layer1,2[]= %d %d, pin:layer,col= %d %d\n",
			   ROUTING->layer1[row][col], ROUTING->layer2[row][col], 
			   pin->layer, pin->col);
		    logError("Bad layer-map in CONNECTPIN. (via=NULL)",COREDUMP);
		  } 
		
		/* Preset pin loc for the floating pins */
		if (pin->pintype != FIXED)		   
		  {
		    pin->x = routing_term.via->x;
		    pin->y = routing_term.via->y;
		  }
		
		if (routing_term.via->layer1 == routing_term.via->layer2)  
		  {
		    if (ROUTING->layer2[row][col]==0 || ROUTING->layer1[row][col]==0)  
		      {
			/* it is OK to put in a via */
			/* change connecting via status to a real via */
			
			routing_term.via->layer1 = LAYER1;
			routing_term.via->layer2 = LAYER2;
			
			ROUTING->layer1[row][col] = ROUTING->layer2[row][col] = netid;
		      }
		    
		    else 
		      {
			
			/* can not put in a via at col/row ==> add a new via */
			/* need add a new via and a new piece of wire */  
			
			if (side == LEFT)   
			  x1 = routing_term.via->x - GRIDWIDTH;
			else
			  x1 = routing_term.via->x + GRIDWIDTH;
			
			/* build a new via between pin and col */
			
			term3.via = bldvia("via3", netid, x1, routing_term.via->y, 
					   col, row, LAYER1, LAYER2);
			
			bldsubnet(netid, 
				  routing_term.via->x, routing_term.via->y, x1, routing_term.via->y, 
				  layer_width[routing_term.via->layer1], routing_term.via->layer1,
				  row,col,row,col,
				  &routing_term,&term3,VIA,VIA,HORIZONTAL); 
			routing_term.via = term3.via;
			
			/* Update pin loc for the floating pins */
			if (pin->pintype != FIXED)
			  {
			    pin->x = x1;
			    pin->y = routing_term.via->y;
			  }
		      }
		  }
		    /* Attach the pseudo pin (f-term) to the pseudo via (a-term) */
		    pin->aterm = routing_term.via;
	      } /* end if (netlist[netid]->numpins > 1)  */
	  }   /* end for (pin = FIRST_PIN(side); pin != NULL; pin = pin->next) */
      }
  }     /* end for (side=LEFT;  side <= RIGHT;  side++) */
for (side=LEFT;  side <= RIGHT;  side++)
  {
    FIRST_PIN(side) = NULL;
    FOREACH_ROW(ROUTING, row)
      if(pin_map[side][row] != NULL)
	{
	  if(FIRST_PIN(side) == NULL)
	    {
	      FIRST_PIN(side) = pin = pin_map[side][row];
	      pin->next = NULL;
	    }
	  else
	    {
	      pin->next = pin_map[side][row];
	      pin = pin->next;
	      pin->next = NULL;
	    }
	}
  }
    
printf("Complete connectExits.c\n\n");
}

case1(row,col,subnet_layer)
int row,col,subnet_layer;     
{
  /* pin and subnet are on the same layer */ 
 case_num = 1;
  
 if (routing_term.via->layer1 != routing_term.via->layer2)  
    {
      /* change connecting via status, 
	 since no longer need a real via */
      
      routing_term.via->layer1 = routing_term.via->layer2 = subnet_layer;
      
      /* printf("connpin: new via %s\n", routing_term.via->mastername); */
      
      if (subnet_layer == 1)      
	ROUTING->layer2[row][col] = 0;
      else 
	if (subnet_layer == 2) 
	  ROUTING->layer1[row][col] = 0;
    }
  /* end:  case1 -- bld 1 subnet & 0 via */
}

case2()
     
{
  /* pin and subnet are not on the same layer; 
     BUT there is a via on col/row */
 case_num = 2;
}

case3(netid,row,col)
     int netid;
int row,col;     
{
  /* pin and subnet are not on the same layer; 
     BUT it is OK to put in a via */
  
 case_num = 3;
 
 /* change connecting via status to a real via */
 routing_term.via->layer1 = LAYER1;
 routing_term.via->layer2 = LAYER2;
 ROUTING->layer1[row][col] = ROUTING->layer2[row][col] = netid;
}  

case5(netid,row,col,subnet_layer,side)
int *row,col,side;
int subnet_layer;
int netid;
{
  /* pin and subnet are not on the same layer; 
     AND it can not put in a via at col/row
     ==>  add a new via */
int row_1;
int y1;
term_type term;
char vianame[80];

 case_num = 5;
 
 if (side == TOP)  
   {
     row_1 = (*row) + 1;
     y1 = routing_term.via->y + 
       RC_LOC->row_width[(*row)] / 2 +
	 MAX_SPACING +
	   RC_LOC->row_width[row_1] / 2;
   }
 else
   {
     row_1 = (*row) - 1;
     y1 = routing_term.via->y - 
       RC_LOC->row_width[(*row)] / 2 - 
	 MAX_SPACING - 
	   RC_LOC->row_width[row_1] / 2;
   }

term.via = bldvia("via2", netid, routing_term.via->x, y1, col, row_1, LAYER1, LAYER2);
 
bldsubnet(netid, 
	  routing_term.via->x, y1, routing_term.via->x, routing_term.via->y, 
	  layer_width[subnet_layer], subnet_layer, 
	  row_1,col,(*row),col,&term,&routing_term,VIA,VIA,VERTICAL); 
(*row) = row_1;
RC_LOC->boundary[side][col] = row_1;
routing_term = term;
}   /* end: case5 -- bld 1 via between pin and connection point */



via_type *find_via(netid,row,col)
int row, col;
{
via_type *via;

  for (via=vialist; via != NULL; via=via->next)  
    {
      if((col == via->col) && 
	 (row == via->row) &&
	 (netid == via->netid))
	{
	  return(via);
	}
    }
return(NULL);
}


print_via()
{
via_type *via;

  for (via=vialist; via != NULL; via=via->next)  
    {
      printf("via %s -- r %d  c %d net %d\n",via->vianame,via->row,via->col,via->netid);   
    }
return(NULL);
}
