#include "copyright.h"
#include "port.h"
#include "spider.h"
#define assign_boundary(side,col,row) \
  RC_LOC->boundary[side][col] = row; 

defHchannel()

/**********************************************************************
 *  
 * 		  DEFHCHANNEL-- 
 * 		  -----------------------------------
 *	(need modify before can handle L shape channel)
 *
 *	---- Function
 * 	This routing redefine the horizontal channel boundary according to 
 *	the grid systerm and pin's status (layer#, offset ...).
 *
 *	New channel boundary will be in col/row (grid) coor.
 *
 *	Channel boundary will be shifted if pin is not on grid line or
 *	if pin is on diff layer as expected (for channel router, layer1 runs
 *	horizontal and layer2 runs vertical), hence the possibility of addit.
 *	vias.
 *
 *	---- This routine sets up the following data structure. 
 *
 *	RC_LOC->boundary[TOP][]
 *	RC_LOC->boundary[BOTTOM][]
 *
 *	---- This routine modifies the following data structure. 
 *
 *	RC_LOC->row_count
 *	pin->y    -- for top pins
 *      channel_ury
 *
 ***********************************************************************/

{
    	vertex_type *pt1, *pt2;

    	int 	row0y;   
    	int 	col;   
	int	colnum, rownum;
    	int 	side;   
	int	GHminus1;
	int old_rownum;
	int loc,r;

/* ==== Preset ====== */

        /* Allocate the memory space for new channel boundary */
        colnum = RC_LOC->col_count+10;
        RC_LOC->boundary[TOP]= CALLOC(int, colnum);
        RC_LOC->boundary[BOTTOM]= CALLOC(int, colnum);

	/* define row #0 to be the bottom edge of the channel */
	row0y = row_zero_loc() + CONT_WIDTH/2;

	/* preset channel */
        for (col=1;  col<=RC_LOC->col_count;  col++)  
	  {
	    RC_LOC->boundary[TOP][col] = -1;
	    RC_LOC->boundary[BOTTOM][col] = -1;
	  }
	
/* ==== make channel boundary in col,row coor */

  	for (side=TOP;  side<=BOTTOM;  side++)  
	  {

	    /* Preset the grid ajustment #, in case additional grid line needed 
	       (we want to shift "in" one grid)
	       case1: if pin is not on grid line 
	       case2: if pin is on layer other then LAYER1 (for YACR) 
	     */

	    if (side == TOP)  
	      { 
	        GHminus1 = 0;
	      }
	    else  
	      {
	        GHminus1 = GRIDHEIGHT - 1;
	      }


	    /* for each point of horizontal channel boundary do, 
	       compute its col,row loc */

	    old_rownum = MAXINT;
	    for(pt1 = CHAN_BOUND->first[side];pt1 != NULL;pt1 = pt1->next->next)
	      {
		pt2 = pt1->next;
		rownum = (pt2->y - ((side == TOP)?(GRIDHEIGHT/2):0) - row0y + GHminus1) / GRIDHEIGHT;
		for(col = 1;col <= RC_LOC->col_count;col++)
		  {
		    if(RC_LOC->col_pins[side][col] == NULL)
		      loc = RC_LOC->col_loc[col];
		    else
		      loc = RC_LOC->col_pins[side][col]->x;

		    if((pt1->x < loc) && (loc < pt2->x))
		      {
			assign_boundary(side,col,rownum);
		      }
		    else
		      if(loc == pt1->x)
			{
			  if(side == TOP)
			    {
			      r = MIN(rownum,(pt1 == CHAN_BOUND->first[side])?MAXINT:old_rownum);
			      assign_boundary(side,col,r);
			    }
			  else
			    {
			      r = MAX(rownum,(pt1 == CHAN_BOUND->first[side])?MININT:old_rownum);
			      assign_boundary(side,col,r);
			    }
			}
		      else
			if(loc == pt2->x)
			  {
			    if(side == TOP)
			      {
				r = MIN(rownum,
					((pt2->next == NULL)?MAXINT:
					 ((pt2->next->y - row0y + GHminus1) / GRIDHEIGHT)));
				assign_boundary(side,col,r);
			      }				
			    else
			      {
				r = MAX(rownum,
					((pt2->next == NULL)?MININT:
					 ((pt2->next->y - row0y + GHminus1) / GRIDHEIGHT)));
				assign_boundary(side,col,r);
			      }
			    
			  }
		  }
		if((pt1 == CHAN_BOUND->first[side]) && (RC_LOC->col_loc[1] < pt1->x))
		  RC_LOC->boundary[side][1] = RC_LOC->boundary[side][2];
		if((pt2->next == NULL) && (RC_LOC->col_loc[RC_LOC->col_count] > pt2->x))
		  RC_LOC->boundary[side][RC_LOC->col_count] = RC_LOC->boundary[side][RC_LOC->col_count - 1];
		old_rownum = rownum;
	      }


	    /* fix up ends of the channel */

	    if(CHAN_BOUND->first[!side]->x < CHAN_BOUND->first[side]->x)
	      {
		for(col = 1;(col <= RC_LOC->col_count) && 
		    ((RC_LOC->col_pins[side][col] == NULL) &&
		     (RC_LOC->col_loc[col] < CHAN_BOUND->first[side]->x)) ||
		    ((RC_LOC->col_pins[side][col] != NULL) && 
		     (RC_LOC->col_pins[side][col]->x < CHAN_BOUND->first[side]->x));
 		    col++)
		  {}
		rownum = RC_LOC->boundary[side][col--];
		for(;col > 0;col--)
		  {
		    assign_boundary(side,col,rownum);
		  }
	      }
	    else
	      {
		for(col = 1;(col <= RC_LOC->col_count) && 
		    (RC_LOC->boundary[side][col] == -1); 
		    col++)
		  {}
		rownum = RC_LOC->boundary[side][col--];
		for(;col > 0;col--)
		  RC_LOC->boundary[side][col] = rownum;
	      }

	    for(pt1 = CHAN_BOUND->first[side];pt1->next != NULL;pt1 = pt1->next)
	      {}
	    for(pt2 = CHAN_BOUND->first[!side];pt2->next != NULL;pt2 = pt2->next)
	      {}
	    if(pt2->x > pt1->x)
	      {
		for(col = RC_LOC->col_count;(col > 0) && 
		    ((RC_LOC->col_pins[side][col] == NULL) &&
		     (RC_LOC->col_loc[col] > pt1->x)) ||
		    ((RC_LOC->col_pins[side][col] != NULL) && 
		     (RC_LOC->col_pins[side][col]->x > pt1->x));
		    col--)
		  {}
		rownum = RC_LOC->boundary[side][col++];
		for(;col <= RC_LOC->col_count;col++)
		  {
		    assign_boundary(side,col,rownum);
		  }
	      }
	    else
	      {
		for(col = RC_LOC->col_count;(col > 0) && 
		    (RC_LOC->boundary[side][col] == -1); 
		    col--)
		  {}
		rownum = RC_LOC->boundary[side][col++];
		for(;col <= RC_LOC->col_count;col++)
		  RC_LOC->boundary[side][col] = rownum;
	      }
	  }

	offsetPins();

	fixJoggedPins();

        fix_RC_LOC_overlap();

	bottomRefY = findMinRowY();

	print_bound(TOP);
	print_bound(BOTTOM);
	printf("RC_LOC->row_count= %d;  bottomRefY= %d\n", 
	        RC_LOC->row_count, bottomRefY);
        printf("\nComplete defHchannel.c\n");
}


/**********************************************************************/

/* find the highest y location of row 0 */

findMinRowY()
{
int col,y;
int max;

max = MININT;
for(col = 1;col <= RC_LOC->col_count; col++)
  {
    y = (findBoundAtX(BOTTOM,RC_LOC->col_loc[col]) + (MAX_SPACING + 1)/2) - RC_LOC->boundary[BOTTOM][col] * GRIDHEIGHT;
    max = MAX(max,y);
  }
return(max);
}

/**********************************************************************/

offsetPins()
{
int offset_cost, layers_cost;
int side,sign;
pin_type *pinptr;

/* Preset the cost for pin layer not on LAYER1 */
if (router_type == CHANNEL_ROUTER)  
  {
    offset_cost = 1;
    layers_cost = 1;
  }
else /* router_type == SWITCHBOX_ROUTER) */  
  {
    offset_cost = 1;
    layers_cost = 0; /* for mighty, no need save row for layer change*/
  }

for (side=TOP; side <= BOTTOM;  side++)  
  {
    if (side == TOP)  
      sign = -1;
    else              
      sign =  1;
    for(pinptr = FIRST_PIN(side);pinptr != NULL;pinptr = pinptr->next)
      if(pinptr->jogged == FALSE)
	{
	  RC_LOC->boundary[side][pinptr->col] += 
	    (sign *  pinOffset(pinptr,offset_cost,layers_cost));
	}
  }
}

/**********************************************************************/

pinOffset(pinptr,offset_cost,layers_cost)
pin_type *pinptr;
int offset_cost,layers_cost;
{
int off;

off = 0;
if(pinptr->poly_width >= 0)
  {
    off++;
    if(channel_types == VERTICAL)
      {
	off += layers_cost;
	return(off);  /* no need to save a row for jogging */
      }
  }
else
  if(pinptr->layer == LAYER2)
    off += layers_cost;

if (pinptr->x != RC_LOC->col_loc[pinptr->col]) 
  off += offset_cost; 

return(off);
}

/**********************************************************************/

get_row(side,pin)
int side;
pin_type *pin;
{
int r;

r = RC_LOC->boundary[side][pin->col];
return(r);
}

/**********************************************************************/

get_col_offset(side, col)
int side;
int col;
{
int r;

if(side == TOP)
  {
    r = RC_LOC->row_count - RC_LOC->boundary[side][col];
  }
else
  {
    r = RC_LOC->boundary[side][col];
  }
return(r);
}
