#include "copyright.h"
#include "port.h"
#include "spider.h"

defrow()


/**********************************************************************
 *  
 * 		  defrow -- build horizental grid line 
 * 		  -----------------------------------
 *
 *	---- This routine sets up the following data structure. 
 *	RC_LOC->row_loc[]
 *      RC_LOC->row_width[]
 *	RC_LOC->row_net[LEFT/RIGHT][row#]
 *
 * 	RC_LOC->row_count
 *
 *   	---- This routine may modifies the following:
 *      netlist[pinptr->netid]->numpins
 *  
 ***********************************************************************/

{
    int 	side;
    int 	rownum, row_j;
    int 	remder;
    pin_type	*pinptr, *pin;


/*  Define row location */

    /* Allocate the memory space */

    RC_LOC->row_count = get_max_RC_LOC_row();
    rownum = RC_LOC->row_count + 2;
    RC_LOC->row_loc   = CALLOC( int, rownum);
    RC_LOC->row_width = CALLOC( int, rownum);
    RC_LOC->row_net[LEFT]= CALLOC(pin_type *, rownum);
    RC_LOC->row_net[RIGHT]= CALLOC(pin_type *, rownum);

    /* define row0 to be the bottom of the bottom side of channel */

    RC_LOC->row_loc[0] = row_zero_loc() + CONT_WIDTH/2;
    RC_LOC->row_width[0] = GRIDHEIGHT;

    /* pre-define row location and row width */

    for (row_j=1;  row_j <= RC_LOC->row_count;  row_j++)  
      {
	RC_LOC->row_loc[row_j] = RC_LOC->row_loc[row_j-1] + GRIDHEIGHT;
	RC_LOC->row_net[LEFT][row_j] = RC_LOC->row_net[RIGHT][row_j] = NULL;
	RC_LOC->row_width[row_j] = GRIDHEIGHT;
      }

/* Set RC_LOC->row_net, the pointer from row to pin */

    /* for each pin,  determine the row position */
    for (side=LEFT;  side <= RIGHT;  side++) {

     if((FIRST_PIN(side) != NULL) && (FIRST_PIN(side)->pintype == FIXED))  {

	for (pinptr=FIRST_PIN(side);  pinptr != NULL;  pinptr=pinptr->next)  {
	     rownum = (pinptr->y - RC_LOC->row_loc[0]) / GRIDWIDTH;
	     remder = (pinptr->y - RC_LOC->row_loc[0]) % GRIDWIDTH;
	     if (remder > GRIDWIDTH/2)  rownum++;
	     if ((rownum > RC_LOC->row_count) || (rownum < 0))
	       logError("ERROR - in assigning exits to rows\n",FATAL);

	     if (rownum < 1) rownum = 1;

	     if (RC_LOC->row_net[side][rownum] != NULL) {

		   /* error */

	    	   sprintf(error, 
		       "ERROR: Spacing violation (defrow). \007\n");
	    	   logError(error, NONFATAL);

	    	   printf(" ERROR: Spacing violation (defrow). \007\n");

		   pin = RC_LOC->row_net[side][rownum];
	    	   printf("PIN at row:  pin: %s of net: %d at loc: %d,%d\n", 
	     		        pin->pinname, pin->netid, pin->x, pin->y);

		   pin = pinptr;
	    	   printf("PIN deleted: pin: %s of net: %d at loc: %d,%d\n", 
	     		        pin->pinname, pin->netid, pin->x, pin->y);


	    	   /* Logically delete the pin from the net */
		   netlist[pinptr->netid]->numpins -= 1;
		   pinptr->netid = 0;
	     }
	     else
	     {
	     	/* set the pointer from row to pin */
	     	RC_LOC->row_net[side][rownum] = pinptr;
	     }
	}
     }
    }
    printf("(defrow): RC_LOC->row_count = %d\n", RC_LOC->row_count);
    printf("Complete defrow.c\n");

}    
