#include "copyright.h"
#include "port.h"
#include "spider.h"

defsubnets(routing,layerOne, layerTwo, layerid)
routing_type *routing;
int **layerOne;		/* one layer of the channel router output */
int **layerTwo;		/* one layer of the channel router output */
int layerid;		/* what kind of rectangle to draw */

/***********************************************************
 *
 *  defsubnet searches the array for rectangles and calls
 *  make_rectangle to make the rectangles appear in the layout of
 *  the channel.  A rectangle is drawn where there is a sequence
 *  identical numbers (!=0) along a row or column in the array.  If
 *  a number is not adjacent to an equal number in the direction
 *  being searched (along row or column) it is ignored.  This is because
 *  it will either be included in a rectangle in the other direction
 *  (along column or row) or will appear as part of a contact with
 *  the other layer, or as part of the connection to a pin along the
 *  edge of the channel.  
 *
 *  The rectangle are assumed to have a width of
 *  only one row or column, in other words, if two adjacent rows
 *  contain the same net then two parallel rectangles will be drawn--
 *  not one extra wide rectangle.  However, since the layer is searched
 *  for rectangles along both rows and columns, the two adjacent
 *  rectangles will end up connected by a short rectangle in each
 *  column (this is not necessarily the most sophisticated algorithm),
 *  This means that instead of the wide rectangle that might (or might
 *  not) have been intended, the result will be a wide rectangle with
 *  holes between the columns.
 *
 **********************************************************/

{
  int  netwidth;
  
  int  start_rect;		/* first edge of rectangle encountered */
  int  curr_net;		/* the net current rectangle belongs to */
  int  pin_row;               /* the connection row# for the pin */
  int  col1,			/* lower left x coord of rectangle */
  row1,			/* lower left y coord of rectangle */
  col2,			/* upper right x coord of rectangle */
  row2;			/* upper right y coord of rectangle */
  int  colnum, rownum;	/* loop counters */
  int  foundsubnet;
  
  /*
   *  find rectangles that lie along a row
   */
  
  FOREACH_ROW(routing, rownum)
    {
      start_rect = 1;
      curr_net = layerOne[rownum][1];
      foundsubnet = FALSE;
      
      FOREACH_COL(routing, colnum)
	{
	  if (layerOne[rownum][colnum] != curr_net) 
 	    {
	      if ((curr_net != 0) && (start_rect < colnum - 1))  
		{
		  col1 = start_rect;
		  col2 = colnum-1;
		  foundsubnet = TRUE;
		}
	      start_rect = colnum;
	    }	  
	  else 
	    if ((curr_net != 0) && (start_rect < colnum))  
	      {
		
		if ((layerTwo[rownum][colnum] == curr_net) || 
		    (colnum == LAST_COL(routing)))
		  
		  {
		    /* There is a contact on col,row  OR it is at last col */
		    
		    col1 = start_rect;
		    col2 = colnum;
		    foundsubnet = TRUE;
		    start_rect = colnum;
		  }
		
		else 
		  if (((layerOne[rownum+1][colnum] == curr_net) || (layerOne[rownum-1][colnum] == curr_net)) &&
		      (layerOne[rownum][colnum+1] == curr_net) &&
		      (layerOne[rownum][colnum-1] == curr_net) )
		    
		    {
		      /* There is a connector (of same layer) on col,row */
		      
		      col1 = start_rect;
		      col2 = colnum;
		      foundsubnet = TRUE;
		      start_rect = colnum;
		    }
		
		  else 
		    if (RC_LOC->col_pins[TOP][colnum] != NULL  &&    
			RC_LOC->col_pins[TOP][colnum]->netid == curr_net)  
		      {
			/* check if there is a connection pin at row */
			pin_row = get_row(TOP,(RC_LOC->col_pins[TOP][colnum]));
			
			if (pin_row == rownum)  
			  {
			    col1 = start_rect;
			    col2 = colnum;
			    foundsubnet = TRUE;
			    start_rect = colnum;
			  }
			else
			  {
			    if (RC_LOC->col_pins[BOTTOM][colnum] != NULL  &&  
				RC_LOC->col_pins[BOTTOM][colnum]->netid == curr_net)  
			      {
				/* check if there is a connection pin at row */
				pin_row = get_row(BOTTOM,(RC_LOC->col_pins[BOTTOM][colnum]));
				if (pin_row == rownum)  
				  {
				    col1 = start_rect;
				    col2 = colnum;
				    foundsubnet = TRUE;
				    start_rect = colnum;
				  }
			      }
			    
			  }
		      }
		
		    else 
		      if (RC_LOC->col_pins[BOTTOM][colnum] != NULL  &&  
                          RC_LOC->col_pins[BOTTOM][colnum]->netid == curr_net)  
			{
			  /* check if there is a connection pin at row */
			  pin_row = get_row(BOTTOM,(RC_LOC->col_pins[BOTTOM][colnum]));			  
			  if (pin_row == rownum)  
			    {
			      col1 = start_rect;
			      col2 = colnum;
			      foundsubnet = TRUE;
			      start_rect = colnum;
			    }
			}
	      }
	  
	  
	  if (foundsubnet == TRUE)
	    {
	      /* check for special case:
		 if this subnet is the one conn between adjacent vias,
		 avoid build two such subnets */
	      
	      if ((col2-col1) == 1  &&  layerid == LAYER1  &&
		  layerOne[rownum][col1] == layerTwo[rownum][col1]  &&
		  layerOne[rownum][col2] == layerTwo[rownum][col2])
		{
		  printf("(defs) skip subnet: %d %d %d %d\n",
			 col1, rownum, col2, rownum);
		  foundsubnet = FALSE;
		}
	    }
	  
	  if (foundsubnet == TRUE)
	    {
	      /* save the subnet information in symbolic form */
	      row1 = row2 = rownum;
	      netwidth = layer_width[layerid];
	      bldsubnet(curr_net, 0,0,0,0, netwidth, layerid,
			row1,col1,row2,col2,(term_type *)NULL,(term_type *)NULL,
			0,0,(row1 == row2)?HORIZONTAL:VERTICAL);
	      
	      foundsubnet = FALSE;
	    }
	  
	  curr_net = layerOne[rownum][colnum];
	}
    }
  
  
  /*
   *  find rectangles that lie along a column
   */
  
  FOREACH_COL(routing, colnum)
    {
      start_rect = 0;
      curr_net = layerOne[0][colnum];
      foundsubnet = FALSE;
      
      FOREACH_ROW(routing, rownum)
	{
	    if ((
		 (RC_LOC->row_count == 0) ||
		 ((RC_LOC->col_pins[TOP][colnum] != NULL) &&
		  (RC_LOC->col_pins[BOTTOM][colnum] != NULL) &&
		  (RC_LOC->col_pins[TOP][colnum]->netid == RC_LOC->col_pins[BOTTOM][colnum]->netid) &&
		  (get_row(TOP,(RC_LOC->col_pins[TOP][colnum])) ==
		   get_row(BOTTOM,(RC_LOC->col_pins[BOTTOM][colnum]))))
		 ) &&
		( RC_LOC->col_pins[TOP][colnum] != NULL ) &&
		( layerOne[rownum][colnum] == RC_LOC->col_pins[TOP][colnum]->netid) )
	    {
	      if (layerOne[rownum][colnum-1] != layerOne[rownum][colnum] &&
		  layerOne[rownum][colnum+1] != layerOne[rownum][colnum]) 
		{
		  /* case2: a short vert net conn pins at same col */
		  row1 = row2 = rownum;
		  foundsubnet = TRUE;
		  start_rect = rownum;
		  curr_net = RC_LOC->col_pins[TOP][colnum]->netid;
		}
	    }
	  else 
	    if (layerOne[rownum][colnum] != curr_net) 
	      {
		if ((curr_net != 0) && (start_rect < rownum - 1))  
		  {
		    row1 = start_rect;
		    row2 = rownum-1;
		    foundsubnet = TRUE;
		  }
		
		start_rect = rownum;
	      }
	    else 
	      if ((curr_net != 0) && (start_rect < rownum))  
		{
		  
		  if ((layerTwo[rownum][colnum] == curr_net) || 
		      (rownum == RC_LOC->row_count))
		    {
		      /* There is a contact on col,row  OR it is at last col */
		      
		      row1 = start_rect;
		      row2 = rownum;
		      foundsubnet = TRUE;
		      start_rect = rownum;
		    }
		  
		  else 
		    if (((layerOne[rownum][colnum+1] == curr_net) || 
			 (layerOne[rownum][colnum-1] == curr_net))
			&& (layerOne[rownum+1][colnum] == curr_net) 
			&& (layerOne[rownum-1][colnum] == curr_net) )
		      {
			/* There is a connector (of same layer) on col,row */
			
			row1 = start_rect;
			row2 = rownum;
			foundsubnet = TRUE;
			start_rect = rownum;
		      }
		}
	  
	  if (foundsubnet == TRUE)
	    {
	      /* check for special case:
		 if this subnet is the one conn between adjacent vias,
		 avoid build two such subnets */
	      
	      if ((row2-row1) == 1  &&  layerid == LAYER2  &&
		  layerOne[row1][colnum] == layerTwo[row1][colnum]  &&
		  layerOne[row2][colnum] == layerTwo[row2][colnum])
		{
		  foundsubnet = FALSE;
		  printf("(defs) skip subnet: %d %d %d %d\n",
			 colnum, row1, colnum, row2);
		}
	    }
	  
	  if (foundsubnet == TRUE)
	    {
	      /* save the subnet information in symbolic form */
	      col1 = col2 = colnum;
	      netwidth = layer_width[layerid];
/*
	      printf("wire %d %d  %d %d\n",col1,row1,col2,row2);
*/
	      bldsubnet(curr_net, 0,0,0,0, netwidth, layerid,
			row1,col1,row2,col2,(term_type *)NULL,(term_type *)NULL,
			0,0,(col1 == col2)?VERTICAL:HORIZONTAL);
	      
	      foundsubnet = FALSE;
	    }
	  
	  curr_net = layerOne[rownum][colnum];
	  
	}
      
    }
  
  printf("Complete defsubnets.c\n");
}
