#include <stdio.h>
#include <strings.h>
#include "spider.h"
#include "symlib.h"

#define OCT_TO_LAMBDA 20
extern int MIN_SPACING;
extern int MAX_SPACING;

extern int COMPACT_SPACE;

makeSparcsConstraints()
{
    int gap;
    char buff[500];
    int status;
    octObject inst0,inst1;
    int min,max;


    copyFacetView(channelName,"symbolic","spaced");
    if((status = openFacet(&oCellFacet,channelName,"spaced","contents",OCT_CURRENT_VERSION,"a")) < OCT_OK)
      logError("can not open oct file for channel\n",FATAL);

    if((strcmp("CPExtInst",CELL_LIST[0]->cellname) == 0) ||
       (strcmp("CPExtInst",CELL_LIST[1]->cellname) == 0))
      return(0);

    symCleanAllConstraints(&oCellFacet);
    inst0.type = OCT_INSTANCE;
    inst0.contents.instance.name = CELL_LIST[0]->cellname;
    if (octGetByName(&oCellFacet, &inst0) != OCT_OK)  
      {
	  fprintf(stderr, "compaction:  can't get instance.  Instname:");
	  fprintf(stderr, "%s\n", CELL_LIST[0]->cellname);
	  exit(-10);
      }

    inst1.type = OCT_INSTANCE;
    inst1.contents.instance.name = CELL_LIST[1]->cellname;
    if (octGetByName(&oCellFacet, &inst1) != OCT_OK)  
      {
	  fprintf(stderr, "compaction:  can't get instance.  Instname:");
	  fprintf(stderr, "%s\n", CELL_LIST[1]->cellname);
	  exit(-10);
      }

    if(channel_types == VERTICAL)
      {
	  gap = get_inst_bound(BOTTOM,&inst0) - get_inst_bound(BOTTOM,&inst1);
	  if(gap < 0)
	    {
		gap = -gap;
		min = gap - GRIDWIDTH*2;
		min = (min < 0)?0:min;
		max = gap + GRIDWIDTH*2;
		symSideConstraint(&inst1, SYM_PUT_CONST_BOTTOM, min, max);
		/*
		  symSideConstraint(&inst0, SYM_PUT_CONST_BOTTOM, 0, GRIDWIDTH/2);
		  */
	    }
	  else
	    {
		min = gap - GRIDWIDTH*2;
		min = (min < 0)?0:min;
		max = gap + GRIDWIDTH*2;
		symSideConstraint(&inst0, SYM_PUT_CONST_BOTTOM, min, max);
		/*
		  symSideConstraint(&inst1, SYM_PUT_CONST_BOTTOM, 0, GRIDWIDTH/2);
		  */
	    }
	  if(COMPACT_SPACE > 0)
	    {
		gap = get_inst_dimension(HORIZONTAL,&inst1) + 
		  COMPACT_SPACE * OCT_TO_LAMBDA;
		makeConst(SYM_HORIZONTAL,
			  &inst1,
			  &inst0,
			  gap,
			  SYM_PUT_CONST_UNDEFINED);
	    }
	  makeVertExitConstraint();
      }
    else
      {
	  gap = get_inst_bound(LEFT,&inst0) - get_inst_bound(LEFT,&inst1);
	  if(gap < 0)
	    {
		gap = -gap;
		min = gap - GRIDWIDTH*2;
		min = (min < 0)?0:min;
		max = gap + GRIDWIDTH*2;
		symSideConstraint( &inst1, SYM_PUT_CONST_LEFT,min,max);
		/*
		  symSideConstraint( &inst0, SYM_PUT_CONST_LEFT, 0, GRIDWIDTH/2);
		  */
	    }
	  else
	    {
		min = gap - GRIDWIDTH*2;
		min = (min < 0)?0:min;
		max = gap + GRIDWIDTH*2;
		symSideConstraint( &inst0, SYM_PUT_CONST_LEFT,min,max);
		/*
		  symSideConstraint( &inst1, SYM_PUT_CONST_LEFT, 0, GRIDWIDTH/2);
		  */
	    }

	  if(COMPACT_SPACE > 0)
	    {
		gap = get_inst_dimension(VERTICAL,&inst1) + 
		  COMPACT_SPACE * OCT_TO_LAMBDA;
		makeConst("y",
			  &inst1,
			  &inst0,
			  gap,
			  SYM_PUT_CONST_UNDEFINED);
	    }
	  makeHorExitConstraint();
      }
    closeOctFacet(&oCellFacet);
    sprintf(buff,"sparcs -t -S \"e\" -r %s/MOSAICOsparcs.rules %s:spaced",TECH_DIR, channelName);
    status = util_csystem(buff);
    printf("sparcs %s return code %d\n",buff,status);
    copyFacetView(channelName,"spaced","symbolic");
    return(status);
}

get_inst_bound(side, inst)
int side;
octObject *inst;
{
struct octBox box;

OCTCHKOK(octBB(inst, &box));
switch(side)
  {
  case TOP:
    return(box.upperRight.y);
  case BOTTOM:
    return(box.lowerLeft.y);
  case LEFT:
    return(box.lowerLeft.x);
  case RIGHT:
    return(box.upperRight.x);
  }
return(MAXINT);
}

get_inst_dimension(type, inst)
int type;
octObject *inst;
{
struct octBox box;

OCTCHKOK(octGetByName(&oCellFacet, inst));
OCTCHKOK(octBB(inst, &box));
switch(type)
  {
  case HORIZONTAL:
    return(box.upperRight.x - box.lowerLeft.x);
  case VERTICAL:
    return(box.upperRight.y - box.lowerLeft.y);
  }
return(MAXINT);
}

makeVertExitConstraint()
{
pin_type *pin, *pin2;
octObject via0, via1;


/* These constraints are necessary to prevent straight throughsegments from 
 *  collapsing completely into one spot. (Andrea)
 */
for(pin = FIRST_PIN(RIGHT);pin != NULL;pin = pin->next)
  {
    if(netlist[pin->netid]->numpins != 2)
      continue;

    for(pin2 = FIRST_PIN(LEFT);pin2 != NULL;pin2 = pin2->next)
      if(pin2->netid == pin->netid)
	break;
    if(pin2 == NULL)
      continue;

    via0.type = OCT_INSTANCE;
    via0.contents.instance.name = pin->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via0));

    via1.type = OCT_INSTANCE;
    via1.contents.instance.name = pin2->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via1));

    makeConst(SYM_VERTICAL,
	      &via0,
	      &via1,
	      CONT_WIDTH + MIN_SPACING,
	      SYM_PUT_CONST_UNDEFINED);
  }

for(pin = FIRST_PIN(RIGHT);pin != NULL;pin = pin->next)
  if(pin->next != NULL)
    {
      via0.type = OCT_INSTANCE;
      via0.contents.instance.name = pin->aterm->vianame;
      OCTCHKOK(octGetByName(&oCellFacet, &via0));

      via1.type = OCT_INSTANCE;
      via1.contents.instance.name = pin->next->aterm->vianame;
      OCTCHKOK(octGetByName(&oCellFacet, &via1));

      makeConst(SYM_HORIZONTAL,
		&via0,
		&via1,
		CONT_WIDTH + MIN_SPACING,
		SYM_PUT_CONST_UNDEFINED);
    }

for(pin = FIRST_PIN(LEFT);pin != NULL;pin = pin->next)
  if(pin->next != NULL)
    {
      via0.type = OCT_INSTANCE;
      via0.contents.instance.name = pin->aterm->vianame;
      OCTCHKOK(octGetByName(&oCellFacet, &via0));

      via1.type = OCT_INSTANCE;
      via1.contents.instance.name = pin->next->aterm->vianame;
      OCTCHKOK(octGetByName(&oCellFacet, &via1));

      makeConst(SYM_HORIZONTAL,
		&via0,
		&via1,
		CONT_WIDTH + MIN_SPACING,
		SYM_PUT_CONST_UNDEFINED);
    }
}

makeHorExitConstraint()
{
pin_type *pin ,*pin2, *p;
octObject via0, via1;

/* These constraints are necessary to prevent straight throughsegments from 
 *  collapsing completely into one spot. (Andrea)
 */
for(pin = FIRST_PIN(LEFT);pin != NULL;pin = pin->next)
  {
    if(netlist[pin->netid]->numpins != 2)
      continue;

    for(pin2 = FIRST_PIN(RIGHT);pin2 != NULL;pin2 = pin2->next)
      if(pin2->netid == pin->netid)
	break;
    if(pin2 == NULL)
      continue;

    via0.type = OCT_INSTANCE;
    via0.contents.instance.name = pin->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via0));

    via1.type = OCT_INSTANCE;
    via1.contents.instance.name = pin2->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via1));

    makeConst(SYM_HORIZONTAL,
	      &via0,
	      &via1,
	      CONT_WIDTH + MIN_SPACING,
	      SYM_PUT_CONST_UNDEFINED);
  }

for(pin = FIRST_PIN(LEFT);pin != NULL;pin = pin->next)
  {
    p = pin->next;
    if(p == NULL)
      continue;

    via0.type = OCT_INSTANCE;
    via0.contents.instance.name = pin->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via0));
    
    via1.type = OCT_INSTANCE;
    via1.contents.instance.name = p->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via1));
    
    makeConst(SYM_VERTICAL,
	      &via0,
	      &via1,
	      CONT_WIDTH + MIN_SPACING,
	      SYM_PUT_CONST_UNDEFINED);
  }

for(pin = FIRST_PIN(RIGHT);pin != NULL;pin = pin->next)
  {
    p = pin->next;
    if(p == NULL)
      continue;
    via0.type = OCT_INSTANCE;
    via0.contents.instance.name = pin->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via0));
    
    via1.type = OCT_INSTANCE;
    via1.contents.instance.name = p->aterm->vianame;
    OCTCHKOK(octGetByName(&oCellFacet, &via1));
    
    makeConst(SYM_VERTICAL,
	      &via0,
	      &via1,
	      CONT_WIDTH + MIN_SPACING,
	      SYM_PUT_CONST_UNDEFINED);
  }
}

makeConst(type,from_inst, to_inst,min,max)
int min, max;
int type;
octObject *from_inst, *to_inst;
{
printf("sparcs constraint of type %d from %s to %s -- min %d max %d\n",
       type,
       from_inst->contents.instance.name,
       to_inst->contents.instance.name,
       min,
       max);
symNormalConstraint(type,from_inst,to_inst,min,max);

}

call_mizer()
{
char buff[256];
int status;

sprintf(buff,"cd RTstuff;mizer %s:symbolic;cd ..;", &(channelName[8]));
status = util_csystem(buff);
printf("mizer call -- %s -- return code %d\n",buff,status);
return(status);
}

copyFacetView(cell,old_view,new_view)
char *cell,*old_view,*new_view;
{
octObject old,new;

    old.type = OCT_FACET;
    old.contents.facet.cell = cell;
    old.contents.facet.view = old_view;
    old.contents.facet.facet = "contents";
    old.contents.facet.version = OCT_CURRENT_VERSION;
 
    new.type = OCT_FACET;
    new.contents.facet.cell = cell;
    new.contents.facet.view = new_view;
    new.contents.facet.facet = "contents";
    new.contents.facet.version = OCT_CURRENT_VERSION;
    new.contents.facet.mode = "a";
    if(octCopyFacet(&new,&old) != OCT_OK)
      {
	logError("(SPIDER):WARNING problem in copying routing information",FATAL);
      }
}
