#include "copyright.h"
#include "port.h"
#include "spider.h"

/**************************************************************/

octStatus openFacet(obj,cell,view,facet,version,mode)
char *cell,*view,*facet,*mode,*version;
octObject *obj;
{
obj->type = OCT_FACET;
obj->contents.facet.cell = cell;
obj->contents.facet.view = view;
obj->contents.facet.facet = facet;
obj->contents.facet.version = version;
obj->contents.facet.mode = mode;

return(octOpenFacet(obj));
}

/**************************************************************/

char *oct_find_str_prop(obj,prop_name)
octObject *obj;
char *prop_name;
{
octObject prop;
char str[50];

prop.type = OCT_PROP;
prop.contents.prop.name = prop_name;
if(octGetByName(obj,&prop) != OCT_OK)
  {
    (void) sprintf(str,"could not find oct property %s",prop_name);
    octError(str);
    return(NULL);
  }
return(prop.contents.prop.value.string);
}

/**************************************************************/

get_layer(layer,name)
octObject *layer;
char *name;
{
layer->type = OCT_LAYER;
layer->contents.layer.name = name;
OCTCHKOK(octGetOrCreate(&oCellFacet,layer));
}

/**************************************************************/

createOctBag(bag,name)
octObject *bag;
char *name;
{
int		debug;

debug = 0;

if (debug >= 1)  printf("Start createOctBag.c\n");

bag->type = OCT_BAG;
bag->contents.bag.name = name;

if (octGetOrCreate(&oCellFacet, bag) != OCT_OK)  
  {
    printf("ERROR createOctBag:  get or create bag %s.",name);
    octError("createOctBag:  get or create bag. ");
    exit(-10);
  }

if (debug >= 1)  printf("End createOctBag.c\n");
}

/**************************************************************/

createOctNet(net,netname)
octObject *net;
char	  *netname;

{
int		debug;

debug = 0;

if (debug >= 1)  
  printf("Start createOctNet.c\n");
net->type = OCT_NET;
net->contents.net.name = netname;

if (octGetOrCreate(&oCellFacet, net) != OCT_OK)  
  {
    printf("createOctNet:  get or create net.\n netname=%s\n", netname);
    octError("createOctNet:  get or create net.");
    exit(-10);
  }

if (debug >= 1)  
  printf("End createOctNet.c\n");
}

/**************************************************************/

attach(obj1,obj2)
octObject *obj1,*obj2;
{
  if (octAttach(obj1, obj2) != OCT_OK)  
    {
      octError("attach:  attaching two objects.\n");
      exit(-10);
    }
}

/**************************************************************/

createOctInstance(inst,instname, instmaster, instview, x, y, transform)
octObject 	*inst;
char		*instname;
char		*instmaster;
char		*instview;
octCoord	x, y;
octTransformType transform;
{
int		debug;

debug = 0;

if (debug >= 1)  printf("Start createOctInstance.c\n");

/*
 *  Place an instance of the current object in the contents facet of
 *  the oct cell.
 */

inst->type = OCT_INSTANCE;
inst->contents.instance.name = instname;
inst->contents.instance.master = instmaster;
inst->contents.instance.view = instview;
inst->contents.instance.facet = "interface";
inst->contents.instance.version = OCT_CURRENT_VERSION;
inst->contents.instance.transform.translation.x = x;
inst->contents.instance.transform.translation.y = y;
inst->contents.instance.transform.transformType = transform;

if (octCreate(&oCellFacet, inst) != OCT_OK)  
  {
    octError("createOctInstance:  attaching instance to content facet.\n");
    printf("createOctInstance:  attaching instance to content facet.\n");
    printf("instname=%s, instmaster=%s, instview=%s, x,y=%d,%d\n",
	   instname, instmaster, instview, x, y);
    exit(-10);
  }
if (debug >= 1)  printf("End createOctInstance.c\n");
}

/**************************************************************/

closeOctFacet(facet)
octObject *facet;
{

if (octCloseFacet(facet) != OCT_OK)  
  {
    printf("saveOctFacet: saving contents facet for top-level cell");
    octError("saveOctFacet: saving contents facet for top-level cell");
    exit(-10);
  }
printf("End saveOctFacet\n");
}

