#include "copyright.h"
#include "port.h"
#include "spider.h"
#include "powernet.h"

int num_corners[4];
static char card[2048];
char 	channelName[80];
char 	exist_power_net;  /* flag */
int	channel_types, ordering;
pin_type *link_up_pins();
pin_type *relative_pins[2];
vertex_type *relink_boundary();

int  	pin_count = 0;     /* total num of pins counted by readin */
int  	net_count = 0;     /* total num of nets counted by readin */
int  	cell_count = 0;    /* total num of cells counted by readin */
int  	layer_count = 0;   /* total num of layers counted by readin */
cell_type	*CELL_LIST[3];

readinfile()

/**********************************************************************
 *  
 * 		  READINFILE -- read input data file
 * 		  -----------------------------------
 *
 *	---- This routine sets up the following data structure. 
 *
 *   	channelName, channel_types, ordering
 *
 *   	CHAN_BOUND->first
 *      num_corners
 *	CHAN_BOUND->ur_x, CHAN_BOUND->ur_y
 *
 *   	pin_count
 *   	net_count
 *   	cell_count
 *   	layer_count
 *
 *   	pinlist
 *   	CELL_LIST
 *  
 ***********************************************************************/

{
    vertex_type	*vertex;

    char 	cellname[100];
    char 	keyword[100];
    char 	skip[100];

    int 	cellid;
    int 	x, y; 
    int 	numpins;
    int 	status;  /* The pin list status. 	*/
			 /*     '0' - floating;      	*/
			 /*     '1' - relative floating;*/
			 /*     '2' - fixed 		*/
    int         i;
    int         debug;
    int pin_count;
    rc_loc_type *new_rc_loc();
    
    debug = 0;

    CHAN_BOUND = CALLOC(chan_bound_type, 1);
    CHAN_PINS = CALLOC(chan_pin_type, 1);
    RC_LOC = new_rc_loc();

    CHAN_BOUND->ur_x = MININT;
    CHAN_BOUND->ur_y = MININT;
    relative_pins[TOP] = NULL;
    relative_pins[BOTTOM] = NULL;

    /*  Initialize */
    for (i=TOP; i <= RIGHT; i++)  
      { 
        FIRST_PIN(i) = NULL;

	/* The first one in CHAN_BOUND->first is not used */
	CHAN_BOUND->first[i] = NULL;
	CHAN_BOUND->last[i] = NULL;
      } 


    cell_count = 0;
    layer_count = 0;
    net_count = 0;
    pin_count = 0;
    exist_power_net = FALSE;

    /*  Start reading the input file */
    rewind(inputfile);

    while (fgets(card,sizeof(card),inputfile) != NULL)
    {
    	if (debug>=1)  
	  printf("Read LINE: %s \n", card);

        if (sscanf(card,"%s", keyword) != 1)
   	    logError("Bad INPUT file.  It occurs while reading key word",FATAL);
	if (strcmp(keyword, "CHANNEL") == 0)   
	  {
	    printf("Channel name %s type %d order %d\n",
		   channelName, channel_types, ordering);
	  }
    	else 
	  if (strcmp(keyword, "PIN") == 0)   
	    {
	      
	      if (sscanf(card,"%s %d", skip, &pin_count) != 2)  
		{
		  printf("pin_count = %d \n", pin_count);
		  logError("Bad INPUT file.  It occurs while reading PIN",FATAL);
		}
	      
	      if (pin_count == 0) 
		{
		  printf("(readinfile): zero pin count. program exit.\n");
		  exit(channel_types+2);
		}
	    }
	  else 
	    if (strcmp(keyword, "NET") == 0)   
	      {
		
		if (sscanf(card,"%s %d", skip, &net_count) != 2)  
		  {
		    printf("net_count = %d \n", net_count);
		    logError("Bad INPUT file.  It occurs while reading NET",FATAL);
		  }
		netlist = CALLOC(net_type *,net_count + 1);
	      }
	
    	else 
	  if (strcmp(keyword, "CELL") == 0)   
	    {
	      if (sscanf(card,"%s %s %d %d %d", skip, cellname, &cellid, &x, &y) != 5)
		{
		  printf("Cell name %s, id %d, x,y=%d %d\n", cellname, cellid,x,y);
		  logError("Bad INPUT file.  It occurs while reading CELL",FATAL);
		}
	      
	      CELL_LIST[cellid] = CALLOC (cell_type, 1);
	      CELL_LIST[cellid]->cellname = mystrcpy( cellname);
	      CELL_LIST[cellid]->x = x;
	      CELL_LIST[cellid]->y = y;
	      
	      cell_count++;
	    }

    	else 
	  if (strcmp(keyword, "PINLIST") == 0)   
	    {
	      if (sscanf(card,"%s %s %d %d", skip, keyword, &numpins, &status) != 4)  
		{
		  printf("Pinlist %s, numpins = %d status = %d \n", 
			 keyword, numpins, status);
		  logError("Bad INPUT file.  It occurs while reading PIN", FATAL);
		}
	      
	      switch(keyword[0])
		{
		case 'T':
		  read_pin_list(TOP,numpins,status);
		  break;
		case 'B':
		  read_pin_list(BOTTOM,numpins,status);
		  break;
		case 'R':
		  read_pin_list(RIGHT,numpins,status);
		  break;
		case 'L':
		  read_pin_list(LEFT,numpins,status);
		  break;
		default:
		  printf("ERROR in pin list specification\n");
		}
	    }
    	else 
	  if (strcmp(keyword, "TOP") == 0)   
	    {
	      read_boundary(TOP);
	    }
    	else 
	  if (!strcmp(keyword, "BOTTOM"))
	    {
	      read_boundary(BOTTOM);
	    }
    	else 
	  if (strcmp(keyword, "LEFT") == 0)   
	    {
	      read_boundary(LEFT);
	    }
    	else 
	  if (strcmp(keyword, "RIGHT") == 0)   
	    {
	      read_boundary(RIGHT);
	    }
    	else 
	  {
	    (void) sprintf(card, "Bad file in input file. Last key word is %s",keyword);
	    logError(card, FATAL);
	  } 
    }


    (void) fclose(inputfile);

    if(debug)
      {
	printf("TOP channel boundary \n");
	for(vertex = CHAN_BOUND->first[TOP];vertex != NULL;vertex = vertex->next)
	  printf(": %d %d\n", vertex->x, vertex->y);
	printf("BOTTOM channel boundary \n");
	for(vertex = CHAN_BOUND->first[BOTTOM];vertex != NULL;vertex = vertex->next)
	  printf(": %d %d\n", vertex->x, vertex->y);
      }
    CELL_LIST[PSEUDOCELL] = CALLOC (cell_type, 1);

    (void) sprintf(cellname, "%s", channelName);
    CELL_LIST[PSEUDOCELL]->cellname = mystrcpy( cellname);

    printf("Complete readinfile.c\n");
}

read_boundary(side)
int side;
{
char skip[100];
int i,layer;
vertex_type *new_vertex();
vertex_type *vert;

if (sscanf(card,"%s %d %d", skip, &num_corners[side],&layer) != 3) 
  {
    printf("num_corners[%d] = %d \n", side,num_corners[side]);
    logError("Bad INPUT file.  It occurs while reading boundary info", FATAL);
  }

CHAN_BOUND->last[side] = NULL;
for (i=1;  i <= num_corners[side];  i++)  
  {
    vert = new_vertex();
    (void) fscanf(inputfile, "%d %d",&vert->x,&vert->y);

    if((side == BOTTOM) || (side == RIGHT))
      {
	vert->next = CHAN_BOUND->first[side];
	CHAN_BOUND->first[side] = vert;
	if(CHAN_BOUND->last[side] == NULL)
	  CHAN_BOUND->last[side] = vert;
      }
    else
      {
	if(CHAN_BOUND->last[side] == NULL)
	  CHAN_BOUND->first[side] = vert;
	else
	  CHAN_BOUND->last[side]->next = vert;
	CHAN_BOUND->last[side] = vert;
      }
    
    CHAN_BOUND->ur_x = MAX(CHAN_BOUND->ur_x, vert->x);
    CHAN_BOUND->ur_y = MAX(CHAN_BOUND->ur_y, vert->y);
  }

if(i > 0)
  (void) fgets(card,sizeof(card),inputfile);
}

read_pin_list(side,numpins,status)
int side,numpins,status;
{
int i;
pin_type  *newpin, *newpin2;
char 	pinname[100];
char 	netname[100];
pin_type *pinptr;
double width;
char fixed;
  
CHAN_PINS->numpins[side] = numpins;
pinptr = NULL;
for (i=1;  i <= numpins;  i++ )  
  {
    newpin = new_pin();
    newpin->next = NULL;
    
    /* read in */
    (void) fgets(card,sizeof(card),inputfile);
    
    if (sscanf(card, "%d %d %d %s %d %s %d %d %d %lf %c", 
	       &newpin->netid,
	       &newpin->x,
	       &newpin->y,
	       pinname,
	       &newpin->layer,
	       netname,
	       &newpin->nettype,
	       &newpin->netwidth,
	       &newpin->cellid,
	       &newpin->power,
	       &fixed)
	< 9)  
      {
	printf("netid= %d, x,y= %d,%d cellid= %d \n", 
	       newpin->netid, newpin->x, newpin->y, newpin->cellid);
	logError("Bad INPUT file.  It occurs while reading PIN",
		 FATAL);
      }
    
    if (newpin->cellid == -1)  
      newpin->cellid = PSEUDOCELL;
    
    newpin->pinname = mystrcpy( pinname);
    newpin->netname = mystrcpy( netname);
    newpin->pintype = status;
    
    /* compute the width of the conn net */
    if (newpin->power != 0.0) 
      {
	if(router_type == SWITCHBOX_ROUTER)
	  {
	    static int mi_flag = FALSE;

	    if(mi_flag == FALSE)
	      {
		fprintf(stderr, "(SPIDER) CHANNELS routed with MIGHTY can not have oversized lines (yet) -- \n		making all segments the size of signal wires.\n");
		mi_flag = TRUE;
	      }
	    newpin->netwidth = layer_width[newpin->layer];
	    newpin->power = 0.0;
	  }
	else
	  {
	    /* it is a VSS or VDD net */
	    if(newpin->power < 0.0)
	      newpin->power = -newpin->power;
	    width = newpin->power / SPcurrentDensity(newpin->layer);
	    newpin->netwidth = SPquantizeWidth( (int)width);
	    newpin->netwidth = MAX(newpin->netwidth,layer_width[newpin->layer]);
	    newpin->nettype  = POWERNET;  
	    exist_power_net = TRUE;
	    
	    printf("(readin): note, pin %s of net %s has width %d ",
		   newpin->pinname, newpin->netname, newpin->netwidth);
	    printf("(curr=%lg)\n", newpin->power);
	  }
      }
    else
      {
	newpin->netwidth = layer_width[newpin->layer];
      }
    
    /* update total number of layers */
    if (newpin->layer > layer_count)  
      layer_count = newpin->layer;
    
    /* add pin to the end of pin list */
    
    if(pinptr == NULL)
      {
	pinptr = newpin;
      }
    else
      {
	newpin->next = pinptr;
	pinptr = newpin;
      }

    if((fixed == 'U') || (fixed == 'R'))
      {
	newpin2 = new_pin();
	newpin2->netid = newpin->netid;
	if(relative_pins[TOP] == NULL)
	  {
	    relative_pins[TOP] = newpin2;
	  }
	else
	  {
	    newpin2->next = relative_pins[TOP];
	    relative_pins[TOP] = newpin2;
	  }
      }
    else
      {
	if((fixed == 'D') || (fixed == 'L'))
	  {
	    newpin2 = new_pin();
	    newpin2->netid = newpin->netid;
	    if(relative_pins[BOTTOM] == NULL)
	      {
		relative_pins[BOTTOM] = newpin2;
	      }
	    else
	      {
		newpin2->next = relative_pins[BOTTOM];
		relative_pins[BOTTOM] = newpin2;
	      }
	  }
      }
  }

FIRST_PIN(side) = pinptr;
}

readInChanInfo()
{
    char 	keyword[100];
    char 	skip[100];

    if (fgets(card,sizeof(card),inputfile) != NULL)
      {
	  (void) sscanf(card,"%s", keyword);
	  if (strcmp(keyword, "CHANNEL") == 0)   
	    if (sscanf(card,"%s %s %d %d", 
		       skip, channelName, &channel_types, &ordering) == 4) 
	      return;
      }

    printf("Channel name %s type %d order %d\n",
	   channelName, channel_types, ordering);
    logError("Bad INPUT file, while reading line1",FATAL);
    logError("Bad INPUT file, while reading line1",FATAL);
}

readInstNames()
{
    char 	cellname[100];
    char 	keyword[100];
    char 	skip[100];
    int 	cellid;
    int 	x, y; 

    while (fgets(card,sizeof(card),inputfile) != NULL)
      {
	if (sscanf(card,"%s", keyword) != 1)
	  logError("Bad INPUT file.  It occurs while reading key word",FATAL);
	if (strcmp(keyword, "CELL") == 0)   
	  {
	    if (sscanf(card,"%s %s %d %d %d", skip, cellname, &cellid, &x, &y) != 5)
	      {
		printf("Cell name %s, id %d, x,y=%d %d\n", cellname, cellid,x,y);
		logError("Bad INPUT file.  It occurs while reading CELL",FATAL);
	      }
	    
	    CELL_LIST[cellid] = CALLOC (cell_type, 1);
	    CELL_LIST[cellid]->cellname = mystrcpy( cellname);
	    CELL_LIST[cellid]->x = x;
	    CELL_LIST[cellid]->y = y;
	    
	    cell_count++;
	  }
      }
}
