#include "copyright.h"
#include "port.h"
#include "spider.h"

net_type        **netlist;
int refTopBoundMinY = MININT;
vertex_type	*last_vertex[4];     /* set by readinfile */

setup()

/**********************************************************************
 *  
 * 		  SETUP -- set up net list 
 * 		  -------------------------
 *
 *	---- Function
 *
 *      1. Build up data structure for net 
 *
 *	---- This routine sets up the following data structure. 
 *
 *	netlist
 *

 *         This location is the pin location before routing.
 *         From this information, one can compute the new cell location. 
 *
 *	3. Set up the following data
 *
 *         CHAN_PINS->min_x[side], CHAN_PINS->min_y[side]  - for pin
 *         CHAN_PINS->max_x[side], CHAN_PINS->max_y[side]
 *         CHAN_BOUND->min_x[side], CHAN_BOUND->min_y[side]  - for channel boundary
 *         CHAN_BOUND->max_x[side], CHAN_BOUND->max_y[side]  
 *  
 *  
 ***********************************************************************/

{
  int 	i;
  int 	side;
  pin_type 	*pinptr;
  net_type 	*netptr;
  vertex_type *ptr;
  
  
  
  /*   Define pin status,  fixed, floating, reletive */
  
  CHAN_PINS->pinstatus = FLOATING;
  if (FIRST_PIN(LEFT) != NULL) 
    {
      CHAN_PINS->pinstatus = MAX(FIRST_PIN(LEFT)->pintype, CHAN_PINS->pinstatus);
    }
  
  if (FIRST_PIN(RIGHT) != NULL) 
    {
      CHAN_PINS->pinstatus = MAX(FIRST_PIN(RIGHT)->pintype, CHAN_PINS->pinstatus);
    }
  
  
  /*   Set up the max, min channel coor */
  
  for (side=TOP;  side <= RIGHT;  side++)  
    { 
      
      CHAN_BOUND->min_x[side] = CHAN_BOUND->min_y[side] = MAXINT;
      CHAN_BOUND->max_x[side] = CHAN_BOUND->max_y[side] = MININT;
      
      for(ptr = CHAN_BOUND->first[side];(ptr != NULL);ptr = ptr->next)
	{
	  /* Save the smallest and largest x,y coor */
	  CHAN_BOUND->min_x[side] = MIN(CHAN_BOUND->min_x[side], ptr->x);
	  CHAN_BOUND->max_x[side] = MAX(CHAN_BOUND->max_x[side], ptr->x);
	  CHAN_BOUND->min_y[side] = MIN(CHAN_BOUND->min_y[side], ptr->y);
	  CHAN_BOUND->max_y[side] = MAX(CHAN_BOUND->max_y[side], ptr->y);
	}
    }
  
  /* Set up the max, min coor for pin */
  
  for (side=TOP;  side <= RIGHT;  side++)  
    {
      
      CHAN_PINS->min_x[side] = CHAN_PINS->min_y[side] = MAXINT;
      CHAN_PINS->max_x[side] = CHAN_PINS->max_y[side] = MININT;
      
      if (FIRST_PIN(side) != NULL && FIRST_PIN(side)->pintype == FIXED)  
	{
	  
          for (pinptr=FIRST_PIN(side); pinptr != NULL; pinptr=pinptr->next) 
	    {
	      CHAN_PINS->min_x[side] = MIN(CHAN_PINS->min_x[side], pinptr->x);
	      CHAN_PINS->max_x[side] = MAX(CHAN_PINS->max_x[side], pinptr->x);
	      CHAN_PINS->min_y[side] = MIN(CHAN_PINS->min_y[side], pinptr->y);
	      CHAN_PINS->max_y[side] = MAX(CHAN_PINS->max_y[side], pinptr->y);
	    }
	}
      else  
	{
	  CHAN_PINS->min_x[side] = CHAN_BOUND->min_x[side];
	  CHAN_PINS->max_x[side] = CHAN_BOUND->max_x[side];
	  CHAN_PINS->min_y[side] = CHAN_BOUND->min_y[side];
	  CHAN_PINS->max_y[side] = CHAN_BOUND->max_y[side];
	}
    }
  
  if (FIRST_PIN(TOP) == NULL && FIRST_PIN(BOTTOM) == NULL) 
    {
      CHAN_PINS->min_x[TOP] = CHAN_BOUND->min_x[TOP];
      CHAN_PINS->max_x[TOP] = CHAN_BOUND->max_x[TOP];
      CHAN_PINS->min_y[TOP] = CHAN_PINS->max_y[TOP] = CHAN_BOUND->min_y[TOP];
      
      CHAN_PINS->min_x[BOTTOM] = CHAN_BOUND->min_x[BOTTOM];
      CHAN_PINS->max_x[BOTTOM] = CHAN_BOUND->max_x[BOTTOM];
      CHAN_PINS->min_y[BOTTOM] = CHAN_PINS->max_y[BOTTOM] = CHAN_BOUND->max_y[BOTTOM];
    }  
  else 
    if (FIRST_PIN(TOP) == NULL) 
      {
	CHAN_PINS->min_x[TOP] = CHAN_PINS->min_x[BOTTOM];
        CHAN_PINS->max_x[TOP] = CHAN_PINS->max_x[BOTTOM];
	CHAN_PINS->min_y[TOP] = CHAN_PINS->max_y[TOP] = CHAN_PINS->max_y[BOTTOM];
      }
  
    else 
      if (FIRST_PIN(BOTTOM) == NULL) 
	{
	  CHAN_PINS->min_x[BOTTOM] = CHAN_PINS->min_x[TOP];
	  CHAN_PINS->max_x[BOTTOM] = CHAN_PINS->max_x[TOP];
	  CHAN_PINS->min_y[BOTTOM] = CHAN_PINS->max_y[BOTTOM] = CHAN_PINS->min_y[TOP];
	}
  
  if (FIRST_PIN(LEFT) == NULL) 
    {
      CHAN_PINS->min_x[LEFT] = CHAN_PINS->max_x[LEFT] 
	= MIN(CHAN_PINS->min_x[TOP], CHAN_PINS->min_x[BOTTOM]) - END_CLEARANCE;
      CHAN_PINS->min_y[LEFT] = MIN(CHAN_PINS->min_y[TOP], CHAN_PINS->min_y[BOTTOM]);
      CHAN_PINS->max_y[LEFT] = MAX(CHAN_PINS->max_y[TOP], CHAN_PINS->max_y[BOTTOM]);
    }
  if (FIRST_PIN(RIGHT) == NULL) 
    {
      CHAN_PINS->min_x[RIGHT] = CHAN_PINS->max_x[RIGHT] 
	= MAX(CHAN_PINS->max_x[TOP], CHAN_PINS->max_x[BOTTOM]) + END_CLEARANCE;
      CHAN_PINS->min_y[RIGHT] = MIN(CHAN_PINS->min_y[TOP], CHAN_PINS->min_y[BOTTOM]);
      CHAN_PINS->max_y[RIGHT] = MAX(CHAN_PINS->max_y[TOP], CHAN_PINS->max_y[BOTTOM]);
    }
  refTopBoundMinY = CHAN_BOUND->min_y[TOP];  
  
  /*  Build up data structure for net  */
  
  for (i=0; i <= net_count; i++)  
    {
      netlist[i] = new_net();
    }
  
  for (side=TOP;  side <= RIGHT;  side++)  
    for (pinptr=FIRST_PIN(side); pinptr != NULL; pinptr=pinptr->next) 
      {
	netptr = netlist[pinptr->netid];
	
	if (netptr->netname == NULL)  
	  {
	    /*  Found a new net */
	    
	    netptr->netname = pinptr->netname;
	    netptr->nettype = pinptr->nettype;
	    netptr->numpins = 1;
	    
	    if (pinptr->netid > net_count)  
	      {
		sprintf(error, "%s %s %s %d %s %d", 
			"Input data error, for pin: ", pinptr->pinname,
			"Total NET_COUNT=", net_count, 
			"Net_netid=", pinptr->netid);
		logError(error, FATAL);
		
	      }
	  }
	else  
	  {
	    if (pinptr->nettype == POWERNET) 
	      netptr->nettype = pinptr->nettype;
	    
	    netptr->numpins += 1;
	  }
      }

  printf("\n(setup)\n"); 
  printf("channel upper right= %d %d \n", CHAN_BOUND->ur_x, CHAN_BOUND->ur_y);
  
  printf("Complete setup.c\n");
  
}
