#include "copyright.h"
#include "port.h"
#include "spider.h"
#define iABS(a) (((a) > 0)?(a):(-(a)))

sortpins()

/**********************************************************************
 *  
 * 		  SORTPIN-- 
 * 		  -------------------------
 *
 *	---- Function
 *
 *      1. Sorts the pins in order from small to large.
 *         Currently,  only top and bottom pin list exit.  
 *	   Side pins are floating. 
 *
 *	---- This routine sets up the following data structure. 
 *
 *	FIRST_PIN(side)       -- for side = TOP, BOTTOM
 *
 **********************************************************************/
{
   pin_type    *pinptr, *pin;
   pin_type    *pinA, *pinB;
   pin_type    ***pin_map;
   pin_type    *firstpin[2];
   vertex_type *ptr;
   int	       *arrayP;
   int	       *arrayN;
   int min;
   int 	col;
   int 	col0x;
   int 	colnum;
   int 	firstcol[2]; 
   int 	side; 
   int 	sum; 
   int 	adj_x; 
   int 	indexI; 
   int 	SS;
   int col2;

/*  Preset the constants */

    adj_x = 0;

    SS = GRIDWIDTH/2;

/* Estimate total num of col, and allocate memory space */

    colnum = 
      (MAX(CHAN_PINS->max_x[TOP],CHAN_PINS->max_x[BOTTOM])-MIN(CHAN_PINS->min_x[TOP],CHAN_PINS->min_x[BOTTOM]))/SS;

    /* Allocate the memory space */
    colnum = 2 * (colnum+5);
    pin_map   = CALLOC(pin_type **, 2);
    pin_map[0]= CALLOC(pin_type *, colnum+1);
    pin_map[1]= CALLOC(pin_type *, colnum+1);

    indexI = SS+2;
    arrayP = CALLOC(int, indexI);
    arrayN = CALLOC(int, indexI);

    for (col=0; col<=GRIDWIDTH/2; col++)  
      arrayP[col] = arrayN[col] = 0;
	
    for (col=0; col<=colnum; col++)  
      {
	pin_map[0][col] = pin_map[1][col] = NULL;
      }

/* for each top, bottom pin,  store it in the map */

    col0x = MIN(CHAN_PINS->min_x[TOP],CHAN_PINS->min_x[BOTTOM]);

    for (side=TOP;  side <= BOTTOM;  side++) 
      for (pinptr=FIRST_PIN(side);  pinptr != NULL;  pinptr=pinptr->next)  
	{
	  col = (pinptr->x - col0x)/((MAX_WIDTH + MAX_SPACING)/2); 
	  
	  if(pin_map[side][col] != NULL)
	    {
	      /* error */
	      fprintf(stderr,"(SPIDER) ERROR Spacing violation: pins too close together\n");
	      pin = pin_map[side][col];
	      fprintf(stderr,"(SPIDER) PIN %s of net %d at loc %d,%d ", 
		     pin->pinname, pin->netid, 
		      (channel_types != VERTICAL)?pin->x:(CHAN_BOUND->ur_y - pin->y),
		      (channel_types != VERTICAL)?pin->y:pin->x);
	      
	      pin = pinptr;
	      fprintf(stderr,"(SPIDER) and\nPIN %s of net %d at loc %d,%d\n", 
		     pin->pinname, pin->netid, 
		      (channel_types != VERTICAL)?pin->x:(CHAN_BOUND->ur_y - pin->y),
		      (channel_types != VERTICAL)?pin->y:pin->x);
	      
	      sprintf(error,
		      "Pin %s deleted.\n",pinptr->pinname);
	      logError(error, NONFATAL);
	      
	      
	      /* Logically delete the pin from the net */
	      netlist[pinptr->netid]->numpins -= 1;
	      pinptr->netid = 0;
	    }
	  
	  else 
	    {
	      pin_map[side][col] = pinptr;
	    }
	}


/* Sort the pins */

   for (side=TOP;  side<=BOTTOM;  side++)  
     {
       
       firstpin[side] = NULL;
       
       if (FIRST_PIN(side) != NULL) 
	 { 
	   
	   col=0;
	   while (pin_map[side][col] == NULL)   col++;  
	   
	   firstpin[side] = pin_map[side][col];
	   firstcol[side] = col+1;
	 }
     }

   for (side=TOP;  side<=BOTTOM;  side++)  
     {
       
       if (FIRST_PIN(side) != NULL) 
	 { 
	   
	   pinptr = firstpin[side];
	   
	   for (col=firstcol[side];  col <= colnum;  col++)  
	     {
	       
	       pin = pin_map[side][col];
	       
	       if (pin != NULL) 
		 {
		   pin->next = NULL;
		   pinptr->next = pin;
		   pinptr = pin;
		 }
	     }
	 }
     }

   FIRST_PIN(TOP) = firstpin[TOP];
   FIRST_PIN(BOTTOM) = firstpin[BOTTOM];

   check_same_side_pins();

    /* temporary: shift top channel, try to min the pin misalignment */

    sum = 0;

   for(pinA = FIRST_PIN(TOP);(pinA != NULL);pinA = pinA->next)
     {
       for(pin = NULL,min = GRIDWIDTH/2,pinB = FIRST_PIN(BOTTOM);(pinB != NULL);pinB = pinB->next)
	 {
	   if(iABS(pinA->x - pinB->x) <= min)
	     {
	       pin = pinB;
	       min = iABS(pinA->x - pinB->x);
	     }
	 }
       if(pin != NULL)
	 {
	   pinB = pin;
	   printf("%d %d  == %d\n",pinA->x ,pinB->x,pinB->x - pinA->x);
	   sum += iABS(pinB->x - pinA->x);
	   indexI = (pinB->x - pinA->x);
	   if (indexI >= 0)  
	     arrayP[indexI] += 1;
	   else              
	     arrayN[-indexI] += 1;
	 }
     }

    /* */
    /* if top instance movable */
    if (sum != 0)  
      {
	
	indexI = 0;
	min = MAXINT;

	for(col2 = -GRIDWIDTH/2;col2 <= GRIDWIDTH/2;col2 += OUTPUT_SCALE)
	  {
	    for (sum = 0,col=0; col <= GRIDWIDTH/2;  col++)  
	      {
		if (arrayP[col] != 0)  
		  {
		    sum += (arrayP[col] * (iABS(col - col2)));
		  }
		if (arrayN[col] != 0)  
		  {
		    sum += (arrayN[col] * (iABS((-col) - col2)));
		  }
	      }
	    printf("%d -- sum %d\n",col2,sum);
	    if(sum < min)
	      {
		indexI = col2;
		min = sum;
	      }
	  }

	if (indexI != 0)   /* shift the top instance */
	  {
	    adj_x = indexI;
	    printf("sortpin): adjust top channel by %d, in x direction\n",adj_x);
	    
	    /*  Ajust top channel location */
	    
	    for (ptr=CHAN_BOUND->first[TOP]; ptr != NULL; ptr=ptr->next) 
	      {
		
		ptr->x += adj_x;
	      }
	    
	    /*  Ajust top pin location */
	    
	    for (pinptr=FIRST_PIN(TOP); pinptr!=NULL; pinptr=pinptr->next) 
	      {
		pinptr->x += adj_x;
	      }
	    
	    /* adjust global vars */
	    
	    CHAN_BOUND->ur_x  += adj_x;
	    CHAN_PINS->min_x[TOP] += adj_x;
	    CHAN_PINS->max_x[TOP] += adj_x;
	    CHAN_BOUND->min_x[TOP] += adj_x;
	    CHAN_BOUND->max_x[TOP] += adj_x;
	  }
    }

   CELL_LIST[TOP]->x += adj_x;

   /* free memory */
    FREE(pin_map);
    FREE(arrayP);
    FREE(arrayN);

   printf("\nComplete sortpins.c\n");
}
