#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "route_matrix.h"
#include "bound.h"
#include "pins.h"
#include "rc_loc.h"

/* for times (instead of getrusage) */
#ifdef hpux
#include <sys/times.h>
#include <sys/time.h>
#include <unistd.h>
#endif

/*
 *   MACRO DEFINITIONS
 *
 *   Macros are distinguished by using solely capital letters in their
 *   identifiers.  This contrasts with C defined identifiers which are strickly 
 *   lower case, and program variable and procedure names which use both upper 
 *   and lower case.
 */

/* Begin macros. */

/* Constant data type */

#define MAXNETS  500
#define MAXCOLS  1000
#define MAXROWS  200

#ifndef MAXINT
#define MAXINT   1000000000
#endif
#define MININT  -1000000000
#define INFINITY 1000000000

#define ZERO	 0
#define ONE      1
#define TWO      2

#define PSEUDOCELL      2

#define CHANNEL_ROUTER   1
#define SWITCHBOX_ROUTER 2

#define  BOOLEAN	int
#define  TRUE		1
#define  FALSE		0
#define  NOT		!
#define  AND		&&
#define  OR		||

#ifndef YES
#define YES  	1
#endif
#ifndef NO
#define NO	0
#endif
#define OK      1
#define EMPTY 	0

#define FIRST  1	
#define LAST   9
#define NEW	1
#define CURRENT	0

/* Parts of channel */
#define TOP	0
#define BOTTOM 	1
#define LEFT	2
#define RIGHT	3
#define END     4

#define HORIZONTAL  0 
#define VERTICAL    1 

#define FLOATING 		0
#define RELATIVE_FLOATING	1
#define FIXED 			2

#define PIN 			1
#define VIA 			2
#define POWERNET 		1
#define POLY 0
#define MET1 1
#define MET2 2

typedef enum {COREDUMP, FATAL, NONFATAL} error_severity_type;

/* Macro procedure that swaps two entities. */
#define  SWAP(type, a, b)   {type swapx; swapx = a; a = b; b = swapx;}

#define OCTCHKOK(octCall) \
{ \
if (octCall!=OCT_OK) \
    octError("Failed: \"octCall==OCT_OK\""); \
    }

/* Memory allocation and dealocation macros. */
extern char *p_shmalloc();
extern char *CALLOC_CHK();

/*
#define MALLOC(type,number)  (type *)CALLOC_CHK(sizeof(type),number)
*/

#define CALLOC4(type, dim) \
	(type ****)CALLOC_CHK(sizeof(type ***), (unsigned)(dim))

#define CALLOC3(type, dim) \
	(type ***)CALLOC_CHK(sizeof(type **), (unsigned)(dim))

#define CALLOC2(type, dim) \
	(type **)CALLOC_CHK(sizeof(type *), (unsigned)(dim))

#define CALLOC(type,number)  (type *)CALLOC_CHK(sizeof(type),(unsigned)(number))

/* Macro function that checks for range errors on array indices. Used for
 * debugging, should be commented out in normal operation. */
/* #define RANGE_CHECK(index,lowerbound,upperbound,message)  */
/*  if ((index) < (lowerbound))  \
 *  {   printf("Error: Array lower bound fault at location %1s.\n",message); \
 *      printf("       index = %d.\n",index);  \
 *  }  \
 *  if ((index) > (upperbound))  \
 *  {   printf("Error: Array upper bound fault at location %1s.\n",message); \
 *      printf("       index = %d.\n",index);  \
 *  }  \
 *  fflush(stdout)  
 */

typedef struct  via_struct {
	char *vianame;
	char *mastername;
	char *pinname;
	int  termid;
	int  netid;
	int  x, y;
	int  col, row;
	int  layer1, layer2; 
        struct via_struct     *next;
	struct subnet_list *subnets;
} via_type;

typedef union term_struct {
        pin_type  *pin;
        via_type  *via;
} term_type;

typedef struct subnet_struct {
	int    layerid;
	int    x1, y1;
	int    x2, y2;
	int    col1, row1;
	int    col2, row2;
	int    width;
	int    term1_status;
	int    term2_status;
	int    dir;
	term_type    term1;
	term_type    term2;
	struct subnet_struct *next;
} subnet_type;

struct subnet_list
{
subnet_type *subnet;
struct subnet_list *next;
};

typedef struct net_struct {
	char *netname;
	int  nettype;			/* '0' - regular;  '1' - power; */
	int  numpins;
	struct subnet_struct *subnetlist;
} net_type;

typedef struct  cell_struct {
	char *cellname;
	int  x, y;
	int  mobility;
} cell_type;

typedef struct pin_list *ppin_list;
struct pin_list
{
  pin_type *pin;
  int l_loc,r_loc;
  int left[2],right[2];
  ppin_list next;
};

typedef struct jogged *pjogged;
struct jogged
{
pjogged next;
pin_type *pin;  
int side;
int row;
int start,stop;
};

typedef struct col *pcol;
struct col
{
  int loc;
  int index;
  int width;
  int leftLim;
  int rightLim;
  int side;
  pin_type *pin[2];
  pcol next;
  pcol prev;
};

/*
 *  GLOBAL VARIABLE DECLARATION 
 *
 *  These variables are global to all routines.  
 *
 */

/* Begin global declarations. */

#ifdef MAIN 
#define extern 
#endif

/* The following data are set up by MAIN.C */
extern char     circuitName[80];
extern char 	channelIdname[80];
extern char     routerName[80];
extern char 	routerfile[80];
extern int      router_type;
extern int	no_compact;

/* The following data are set up by OPENFILE.C */
extern FILE 	*inputfile;
extern FILE 	*sumfile;         /* statistic summary */

/* The following data are set up by READINFILE.C */
extern char 	channelName[80];
extern char 	exist_power_net;  /* flag */
extern int      channel_types, ordering;
extern int 	num_corners[4];

extern int  	net_count;     /* total num of nets counted by readin */
extern int  	cell_count;    /* total num of cells counted by readin */
extern int  	layer_count;   /* total num of layers counted by readin */
	
extern cell_type	*CELL_LIST[3];


/* The following data are set up by SETUP.C */
extern net_type        **netlist;
extern int             refTopBoundMinY;
/* The following data are set up by YACRinWrite.C */
extern int 	bottomRefY;
extern int 	YACR_ROW_OFFSET;
extern int CHANNEL_WIDTH;
extern int CHANNEL_WIDTH_LIM;


/* The following data are set up by RESETNET.C -- net->seqnum */
extern int  	NUM_NET;

/* The following data are set up by CALLROUTER.C */
extern char     infilename[80];
extern char 	outfilename[80];
extern int  	totalDensity;

/* The following data are set up by DEFCONTACTS.C */
extern via_type	*vialist;

/* The following data are set up by GETDRC.C  */
extern int 	LAYER1;
extern int 	LAYER2;
extern char 	layer_nameTable[3][40];
extern char	*via_nameTable[3][3];
extern int 	layer_width[3];
extern int 	layer_spacing[3];

extern int  INPUT_SCALE;
extern int  OUTPUT_SCALE;

extern char *FEED_NAME;
extern char *CONT_NAME;
extern char *CONT_TERMNAME;
extern int  CONT_HEIGHT;
extern int  CONT_WIDTH;

extern int  MAX_SPACING;
extern int  MAX_WIDTH;
extern int  GRIDHEIGHT ;
extern int  GRIDWIDTH ;
extern int  SIDE_CLEARANCE ;
extern int  END_CLEARANCE ;

/* The following data are set up by NEWCORNER.C */

/* The following data are set up by DEN2.C */
extern int *col_den;

/* The following data are set up by SPIDEROCT.C */
extern octObject oCellFacet;

/* The following data are set up by TAP_STUFF.C */
extern char *TECH_DIR;

/* The following data are set up by MOVECELL.C */
extern int ADJUST;

/* temp string for assembling error messages*/
extern char 	error[128];  

/* Functions */

pin_type *new_pin();
double SPcurrentDensity();
net_type *new_net();
int **read_layer();
via_type *bldvia();
char *mystrcpy();
extern subnet_type *find_connecting_subnet();
extern via_type *checkForYacrScrewUp();
extern via_type *find_via();

#define abs(a) (((a) < 0)?(-(a)):(a))
