#include "copyright.h"
#include "port.h"
#include "errtrap.h"
#include "oct.h"
#include "tap.h"

/*
 *  FILE: spiderPower.c
 *
 *  DESCRIPTION: 
 *     this file replaces the library spiderPower.a
 *
 *  CREATION DATE: Dec 1987		AUTHOR: Faye Marron, Andrea Casotto
 *
 *  FUNCTIONS:
 * 	SPfindWidth()
 * 	SPquantizeWidth()
 * 	SPcurrentDensity()
 */
#include <sys/time.h>
#include <sys/resource.h>

#include "spider.h"

#define OCT_LAMBDA 20
#define DEFAULT_MAX_CURRENT_DENSITY 0.1e-4 /* in amperes per lambda */

#define NO 0






/* >>>>>  DATA STRUCTURES */
/* data structure for actual and psuedo pins  and also vias*/
struct node {
  int xcoord; 
  int ycoord; 

  struct edge *firsttail;
  struct edge *firsthead;

  struct node *largexson;
  struct node *smallxson;
  struct node *largeyson;
  struct node *smallyson;

  struct pw_pin_struct *pin;	/*isn't NULL only if actual or psuedo pin */
  int inqueue;			/* test for visited state of the node */
  struct edge *backedge;	/* points to the parent edge in the tree */
  struct node *qnext,*qprev;	/* double pointer for queue operations */
};

struct edge {
  struct node *headnode;
  struct node *tailnode;
  struct edge *nexthead;
  struct edge *nexttail;

  struct subnet_struct *subnet;	/* pointer to the subnet in fay's list */
  double powreq;		/* power requirements  for the subnet*/
  struct ChildPtr *leftmostson;	/* pointer to the array of children */
  struct ChildPtr *currentson;	/* points to the leftmost unvisited son */
  struct edge *nextnode;	/* used for stack */
};

struct ChildPtr {
  struct edge *sonptr;
  struct ChildPtr *nextson;
};


typedef struct pw_subnet_struct {
  int subnetid;
  int x1, y1;
  int x2, y2;
  int width;
  int layer;
  struct pw_subnet_struct *next;
} pw_subnet_type;

typedef struct pw_pin_struct {
  int pinid;
  int x, y;
  double power;
  struct pw_pin_struct *next;
} pw_pin_type;


/*
 *  Forward declaration of static functions.
 */

static int Insert();
static int InsertInG();
static int PostorderGT();
static int push();
static struct edge *pop();
static int visit();



/*
 *  GLOBAL VARIABLES IN THIS FILE
 */

static int 		gNumNodes = 0;	/* dont know what this is */
static struct node 	*gFirstNode;  /* access to the binary search tree */
static struct edge 	*gSubnetTree;	/* access to the subnet tree */


/* >>>>> ROUTINES */
SPfindWidth(pw_subnet_list, pw_pin_list)
    subnet_type *pw_subnet_list;
    pw_pin_type *pw_pin_list;
{
  int numedges = 0;
  struct node *head, *tail;
  struct node *tempnode;
  struct edge *newedge;
  struct node *source;
  subnet_type *cursubnet;
  pw_pin_type *curpin, *temppwpin, *sourcepin;
  double maxpower;
  double ptime(),stime[2];
  int debug = FALSE;

  stime[0] = ptime();

  gNumNodes = 0;
  temppwpin = pw_pin_list;
  maxpower = 0;
  sourcepin = NULL;
  while ( temppwpin != NULL ) {
    if ( temppwpin->power > maxpower ) {
      maxpower = temppwpin->power;
      sourcepin = temppwpin;
    }
    if(debug == TRUE)
      printf("(findWidth):pin:x=%d,y=%d,p=%g\n",temppwpin->x,
	     temppwpin->y, temppwpin->power);
    temppwpin = temppwpin->next;
  }
  Insert(sourcepin->x,sourcepin->y,&source);
  source->pin = sourcepin;

  cursubnet = pw_subnet_list;
  while ( cursubnet != NULL ) {
    if(debug == TRUE)
      printf("(findWidth):subnet:x1=%d,y1=%d,x2=%d,y2=%d\n",
	     cursubnet->x1,cursubnet->y1,cursubnet->x2,cursubnet->y2);

    numedges++;

    /* assume the first node is tail and the second node is head */
    /* insert the two ends in the binary search tree
       return addresses of the nodes */

    Insert(cursubnet->x1,cursubnet->y1,&tail);
    Insert(cursubnet->x2,cursubnet->y2,&head);

    newedge = (struct edge *) calloc(1,sizeof(struct edge));
    newedge->leftmostson = NULL;
    newedge->currentson = NULL;

    /* build the pointers from the new edge 
       to its head and tail nodes */

    newedge->headnode = head;
    newedge->tailnode = tail;
    newedge->subnet = cursubnet;

    if ( head->firsthead == NULL ) { 
      /* this is the firsthead edge */
      head->firsthead  = newedge;
    }
    else {
      newedge->nexthead = head->firsthead;
      head->firsthead = newedge;
    }

    if ( tail->firsttail == NULL )  {
      /* this is the firsttail edge */
      tail->firsttail = newedge;
    }
    else {
      newedge->nexttail = tail->firsttail;
      tail->firsttail = newedge;
    }
    cursubnet = cursubnet->next;
  }
  curpin = pw_pin_list;
  while (curpin != NULL ) {
    Insert(curpin->x, curpin->y, &tempnode);
    tempnode->pin = curpin;
    curpin = curpin->next;
  }
  if(debug == TRUE)
    {
      stime[1] = ptime() ;
      printf("numnodes = %d , numedges = %d\n",gNumNodes,numedges);
      printf("time to build the terminal tree = %lf\n",stime[1]-stime[0]);

      stime[0] = ptime();
    }
  MakeSubnetTree(source);
  if(debug == TRUE)
    {
      stime[1] = ptime();
      printf("time to build the subnet tree = %lf\n",stime[1]-stime[0]);
      stime[0] = ptime();
      }
  PostorderGT(gSubnetTree);
  if(debug == TRUE)
    {
      stime[1] = ptime();
      printf("time to traverse subnet tree = %lf\n",stime[1]-stime[0]);
    }
}




static Insert(xvalue, yvalue ,location)
    int xvalue, yvalue;
    struct node **location;
{

  struct node *tempnode;

  /* implement binary search tree for the nodes */

  if ( gFirstNode == NULL ) {	/* the tree is empty */
    gFirstNode = (struct node *)calloc(1,sizeof(struct node));
    gFirstNode->xcoord = xvalue;
    gFirstNode->ycoord = yvalue;
    gFirstNode->smallxson = NULL;
    gFirstNode->largexson = NULL;
    gFirstNode->smallyson = NULL;
    gFirstNode->largeyson = NULL;
    gFirstNode->backedge = NULL;
    gFirstNode->pin = NULL;
    *location = gFirstNode;
    gNumNodes++;
  }

  else {
    tempnode = gFirstNode;
    
    while ( tempnode != NULL ) {
      if ( xvalue == tempnode->xcoord ) {
	if ( yvalue == tempnode->ycoord ) {
	  *location = tempnode;
	  break;
	}
	else if (yvalue < tempnode->ycoord) {
	  if ( tempnode->smallyson == NULL) {
	    *location =(struct node *)calloc(1,sizeof(struct node));
	    tempnode->smallyson = *location;
	    (*location)->xcoord = xvalue;
	    (*location)->ycoord = yvalue;
	    (*location)->smallxson = NULL;
	    (*location)->largexson = NULL;
	    (*location)->smallyson = NULL;
	    (*location)->largeyson = NULL;
	    (*location)->backedge = NULL;
	    (*location)->pin = NULL;
	    gNumNodes++;
	    break;
	  }
	  else {
	    tempnode = tempnode->smallyson;
	  }
	}
	else if (yvalue > tempnode->ycoord) {
	  if ( tempnode->largeyson == NULL) {
	    *location =(struct node *)calloc(1,sizeof(struct node));
	    tempnode->largeyson = *location;
	    (*location)->xcoord = xvalue;
	    (*location)->ycoord = yvalue;
	    (*location)->smallxson = NULL;
	    (*location)->largexson = NULL;
	    (*location)->smallyson = NULL;
	    (*location)->largeyson = NULL;
	    (*location)->backedge = NULL;
	    (*location)->pin = NULL;
	    gNumNodes++;
	    break;
	  }
	  else {
	    tempnode = tempnode->largeyson;
	  }
	}
      }
      else if (xvalue < tempnode->xcoord) {
	if ( tempnode->smallxson == NULL) {
	  *location = (struct node *)calloc(1,sizeof(struct node));
	  tempnode->smallxson = *location;
	  (*location)->xcoord = xvalue;
	  (*location)->ycoord = yvalue;
	  (*location)->smallxson = NULL;
	  (*location)->largexson = NULL;
	  (*location)->smallyson = NULL;
	  (*location)->largeyson = NULL;
	  (*location)->backedge = NULL;
	  (*location)->pin = NULL;
	  gNumNodes++;
	  break;
	}
	else {
	  tempnode = tempnode->smallxson;
	}
      }
      else if (xvalue > tempnode->xcoord) {
	if ( tempnode->largexson == NULL) {
	  *location = (struct node *)calloc(1,sizeof(struct node));
	  tempnode->largexson = *location;
	  (*location)->xcoord = xvalue;
	  (*location)->ycoord = yvalue;
	  (*location)->smallxson = NULL;
	  (*location)->largexson = NULL;
	  (*location)->smallyson = NULL;
	  (*location)->largeyson = NULL;
	  (*location)->backedge = NULL;
	  (*location)->pin = NULL;
	  gNumNodes++;
	  break;
	}
	else {
	  tempnode = tempnode->largexson;
	}
      }
    }
  }
  return;
}


/* This procedure converts an undirected tree with a given 
source node to a directed tree such that there is one to one
correspondence between the edges of the undirected tree and
the nodes of the directed tree. The assumption is there is
only one edge connected to the sourcenode of the undirected
tree, otherwise the dircted tree generated would correspond
to the tree generated by only one of the edges connected to
the source node.
*/ 

MakeSubnetTree(source)
    struct node *source;
{
  struct node *tempnode;
  struct node *qlast;
  struct edge *tempedge;

  source->inqueue = YES;
  qlast = NULL;

  tempnode = source;
  while ( 1 ) {
    /* scan all the edges connected to the current node */

    tempedge = tempnode->firsttail;
    while (tempedge != NULL) {
      if (tempedge->headnode->inqueue == NO) { 
	if ( tempedge->headnode->pin != NULL ) {
	  /* the node is incident to a pin */
	  tempedge->powreq = tempedge->headnode->pin->power;
	}
	/* insert the previosly unvisited node in the queue */
	if ( qlast == NULL ) {
	  qlast = tempedge->headnode;
	  qlast->qnext = qlast->qprev = qlast;
	  tempedge->headnode->inqueue = YES;
	}
	else {
	  tempedge->headnode->qprev = qlast;
	  tempedge->headnode->qnext = qlast->qnext;
	  qlast->qnext->qprev = tempedge->headnode;
	  qlast->qnext = tempedge->headnode;
	  tempedge->headnode->inqueue = YES;
	}
	/* insert the current edge in the subnet graph */
	InsertInG(tempedge,tempnode->backedge);
	tempedge->headnode->backedge = tempedge;
      }

      tempedge = tempedge->nexttail;
    }
    tempedge = tempnode->firsthead;
    while (tempedge != NULL) {
      if (tempedge->tailnode->inqueue == NO) { 
	if ( tempedge->tailnode->pin != NULL ) {
	  /* the edge is incident to a pin */
	  tempedge->powreq = tempedge->tailnode->pin->power;
	}
	/* insert the node in the queue */
	if ( qlast == NULL ) {
	  qlast = tempedge->tailnode;
	  qlast->qnext = qlast->qprev = qlast;
	  tempedge->tailnode->inqueue = YES;
	}
	else {
	  tempedge->tailnode->qprev = qlast;
	  tempedge->tailnode->qnext = qlast->qnext;
	  qlast->qnext->qprev = tempedge->tailnode;
	  qlast->qnext = tempedge->tailnode;
	  tempedge->tailnode->inqueue = YES;
	}
	InsertInG(tempedge,tempnode->backedge);
	tempedge->tailnode->backedge = tempedge;
      }

      tempedge = tempedge->nexthead;
    }

    /* delete the last node in the queue */

    if ( qlast == NULL ) {
      break;
    }

    tempnode = qlast;
    if ( qlast->qnext == qlast ) {
      qlast = NULL;
    }
    else {
      qlast = qlast->qprev;
      qlast->qnext = qlast->qnext->qnext;
      qlast->qnext->qprev = qlast;
    }
  }
}

static InsertInG(newson, parent)
    struct edge *newson;
    struct edge *parent;
{
  struct ChildPtr *tempchild;

  if (parent == NULL) {
    gSubnetTree = newson;
  }
  else {
    tempchild = (struct ChildPtr *) calloc(1,sizeof(struct ChildPtr));
    tempchild->sonptr = newson;
    tempchild->nextson = parent->leftmostson;
    parent->leftmostson = parent->currentson = tempchild;
  }
}


struct edge *stckptr; /* access to the stack */

/* postorder traversing for a general tree */
static PostorderGT(treeptr)
    struct edge *treeptr;
{
  struct edge *tempptr;
  struct edge *pop();
  struct ChildPtr *childptr;

  stckptr = NULL;

  /* this function prints the tree information in postorder */
  /* algorithm is modified version of hw8 for cs160 */

  if ( treeptr == NULL ) {
    printf("tree is empty\n");
    return(-1);
  }

  /* originally all the currentson fields are the same as leftmostson */

  tempptr = treeptr;
  while ( 1 ) {
    while ( tempptr != NULL ) {
      childptr = tempptr->currentson;
      push(tempptr);
      if ( childptr != NULL ) {
	tempptr = childptr->sonptr;
      }
      else {
	tempptr = NULL;
      }
    }

    while ( 1 ) {
      tempptr = pop();

      if ( tempptr == NULL ) {
	return(0);
      }

      if (tempptr->currentson == NULL) {
	visit(tempptr);
      }
      else {
	childptr = tempptr->currentson;
	push(tempptr);
	if ( childptr != NULL ) {
	  tempptr = childptr->sonptr;
	}
	else {
	  tempptr = NULL;
	}
	break;
      }
    }
  }
}   

/* push into the stack */
static push(elmptr)
    struct edge *elmptr;
{
  elmptr->nextnode = stckptr;
  stckptr = elmptr;
  if ( elmptr->currentson != NULL ) {
    elmptr->currentson = elmptr->currentson->nextson;
  }
}

static struct edge *pop()
{
  struct edge *newptr;

  if ( stckptr == NULL )
    return ( NULL );
  else {
    newptr = stckptr;
    stckptr = stckptr->nextnode;
    return(newptr);
  }
}

static visit(elmptr)
    struct edge *elmptr;
{
  double powreq1, powreq2;
  struct ChildPtr *tempptr;
  double	maxcurd;	/* maximum current density */
  double 	SPcurrentDensity();
  int debug = FALSE;

  maxcurd = SPcurrentDensity(elmptr->subnet->layerid);

  if ( elmptr->leftmostson == NULL ) {
    /* this is a leaf subnet */
    powreq1 = powreq2 = 0.;
    if ( elmptr->headnode->pin != NULL ) {
      powreq1 = elmptr->headnode->pin->power; 
    }
    if ( elmptr->tailnode->pin != NULL ) {
      powreq2 = elmptr->tailnode->pin->power;
    }
    if(debug == TRUE)
      printf("(PostorderGT.c):leaf subnet, pin1=%g, pin2=%g\n",
	     powreq1, powreq2);
    elmptr->powreq = MAX(powreq1, powreq2);
  }
  else {
    elmptr->powreq = 0;
    tempptr = elmptr->leftmostson;
    while ( tempptr != NULL ) {
      elmptr->powreq += tempptr->sonptr->powreq;
      tempptr = tempptr->nextson;
    }
  }
  if ( (int)(elmptr->powreq / maxcurd) > elmptr->subnet->width ) {
    elmptr->subnet->width = elmptr->powreq / maxcurd;
    elmptr->subnet->width = SPquantizeWidth(elmptr->subnet->width);
  }

  if(debug == TRUE)
    printf("(PostorderGT): subnet width, cur = %d %g\n",elmptr->subnet->width,
	   elmptr->powreq);
}


/*
    ptime -- return a floating point number which represents the
    elasped processor time in seconds since the program started
*/
double ptime()
{
#ifdef hpux
  struct tms ttime;
  clock_t time;

  times(&ttime);
  time = ttime.tms_utime + ttime.tms_stime;
  time = time / sysconf(CLK_TCK);
  return((double) time);
#else
  struct rusage rusage;
  double time;

  getrusage(RUSAGE_SELF, &rusage);
  time = (float) rusage.ru_utime.tv_sec +
    (float) rusage.ru_utime.tv_usec/1.0e6;
  time += (float) rusage.ru_stime.tv_sec +
    (float) rusage.ru_stime.tv_usec/1.0e6;
  return(time);
#endif
}




int
SPquantizeWidth(w)
    int	w;			/* original width */
{
  int	m;			/* w MOD OCT_LAMBDA*/
  int	nw;			/* new width in lambda */

  /* The width is in "OCT" units: It must be rounded to the nearest LAMBDA */
  
  m = w % OCT_LAMBDA;
  if (m != 0) {
    w += OCT_LAMBDA - m;	/* round up to next OCT_LAMBDA */
  }

  /*
   * Quantize: Quantization sequence is (in lambda) :
   * 1,2,3,4,5,6,8,10,12,14,17,20,25,30,35,.... (up by 5)
   */
  nw = w / OCT_LAMBDA;		/* Work in lambda  */
  if (nw > 6) {
    if ( nw <= 14 ) {
      if (nw % 2 != 0)  nw = ((nw / 2) + 1) * 2;
    } else if ( nw <= 20) { 
      if ((nw-14) % 3 != 0)   nw = 14 + (((nw-14) / 3) + 1) * 3;
    } else if (nw < 30) {
      if (nw % 5 != 0)   nw = ((nw / 5) + 1) * 5;
    } else {
      nw = nw + 5 - nw%5;
    }
  }
  return(nw * OCT_LAMBDA);	/* return OCT units */
}


int
SPquantizeWidthDOWN(w)
    int	w;			/* original width */
{
  int	m;			/* w MOD OCT_LAMBDA*/
  int	nw;			/* new width in lambda */

  /* The width is in "OCT" units: It must be rounded to the nearest LAMBDA */
  
  m = w % OCT_LAMBDA;
  if (m != 0) {
    w -= m;	/* round down to next OCT_LAMBDA */
  }

  /*
   * Quantize: Quantization sequence is (in lambda) :
   * 1,2,3,4,5,6,8,10,12,14,17,20,25,30,35,.... (up by 5)
   */
  nw = w / OCT_LAMBDA;		/* Work in lambda  */
  if (nw > 6) {
    if ( nw <= 14 ) {
      if (nw % 2 != 0)  nw = (nw / 2) * 2;
    } else if ( nw <= 20) { 
      if ((nw-14) % 3 != 0)   nw = 14 + (((nw-14) / 3)) * 3;
    } else if (nw < 30) {
      if (nw % 5 != 0)   nw = (nw / 5) * 5;
    } else {
      nw = nw + 5 - nw%5;
    }
  }
  return(nw * OCT_LAMBDA);	/* return OCT units */
}






static double curr_denTable[3] = {0.0, 0.0, 0.0};



int SPinit( facet )
    octObject* facet;
{
    static int init = 0;

    if ( !init && facet != 0 ) {
	int i;
	octObject layer;
	double coordSize = 1.e-6;
	tapLayerListElement layerList[1];

	if( curr_denTable[MET1] == 0.0 )  {
	    if ( tapGetProp( facet, TAP_COORD_SIZE ) ) {
		coordSize = TAP_COORD_SIZE->value.real;
	    }
	    for ( i = 0; i < 3 ; i++ ) {
		layer.type = OCT_LAYER;
		layer.contents.layer.name = layer_nameTable[i];
		octGetOrCreate( facet, &layer );
		layerList[0].layer = layer;
		if ( tapGetLyrProp( 1, layerList, TAP_MAX_CURRENT_DENSITY ) ) {
		    curr_denTable[i] = TAP_MAX_CURRENT_DENSITY->value.real * coordSize;
		} else {
		    curr_denTable[i] = 0.0015;
		}
		printf("%s current density %lf (amps/oct unit)\n",
		       layer_nameTable[i], curr_denTable[i] );
	    }
	}
	init = 1;
    }
    return init;
}




/*
 * returns the maximum current density 
 * for a particular layer. If no layer is specified, then
 * some default is returned. 
 */
double SPcurrentDensity(layerNum)
    int layerNum;
{
    if ( ! SPinit( 0 ) ) {
	printf( "spider power not initialized" );
    }

    if(((layerNum != MET1) && (layerNum != MET2) &&  (layerNum != POLY)) ||
       (curr_denTable[layerNum] == 0.0)) {
	return(DEFAULT_MAX_CURRENT_DENSITY);
    }
    return curr_denTable[layerNum];
}

