#include "copyright.h"
#include "port.h"
#include "spider.h"
#include "tap.h"
#include "errtrap.h"

octStatus openFacet();
static char errBuf[80];
char *oct_find_str_prop();
char    *TECH_DIR;

char *get_tech_name()
{
    char *techName;
    octStatus stat;

    if((stat = openFacet(&oCellFacet,channelName,"preroute","contents",OCT_CURRENT_VERSION,"r")) < OCT_OK)
      {
	  sprintf(errBuf,"(tap_stuff): ERROR in opening %s\n",ohFormatName(&oCellFacet ));
	  logError(errBuf,FATAL);
      }
    if(oct_find_str_prop(&oCellFacet,"TECHNOLOGY") == NULL)
      logError("Could not find TECHNOLOGY property\n",FATAL);
    TECH_DIR = tapGetDirectory(&oCellFacet);
    printf("technology directory %s\n",TECH_DIR);

    for(techName = TECH_DIR + strlen(TECH_DIR) - 1;(techName > TECH_DIR) && (*techName != '/');techName--)
      {}
    techName++;
    printf("technology name %s\n",techName);

    return(techName);
}

set_up_width_and_spacing(layers)
    octObject layers[];
{
    layer_width[POLY] = tapGetMinWidth(&(layers[POLY]));
    layer_width[MET1] = tapGetMinWidth(&(layers[MET1]));
    layer_width[MET2] = tapGetMinWidth(&(layers[MET2]));
  
    layer_spacing[POLY] = tapGetMinSpacing(&(layers[POLY]),&(layers[POLY]));
    layer_spacing[MET1] = tapGetMinSpacing(&(layers[MET1]),&(layers[MET1]));
    layer_spacing[MET2] = tapGetMinSpacing(&(layers[MET2]),&(layers[MET2]));

    printf("%d %d %d %d %d %d\n",layer_width[POLY],layer_width[MET1],layer_width[MET2],layer_spacing[POLY],layer_spacing[MET1],layer_spacing[MET2]);

    if((layer_width[POLY] == -1) || (layer_width[MET1] == -1) || (layer_width[MET2] == -1) || (layer_spacing[POLY] == -1) || (layer_spacing[MET1] == -1) || (layer_spacing[MET2] == -1))
      logError("(tap_stuff): error in getting layer widths and spacing\n",FATAL);
}

get_connector_BB(box,layerOne,layerTwo,layerOne_dir,layerTwo_dir)
struct octBox *box;
int layerOne,layerTwo,layerOne_dir,layerTwo_dir;
{
octObject mast,conn;
octStatus stat;

get_connector_info(&mast,layerOne,layerTwo,layerOne_dir,layerTwo_dir);
/*
printf("master %s view %s\n",mast.contents.instance.master,mast.contents.instance.view);
*/

if((stat = openFacet(&conn,mast.contents.instance.master,mast.contents.instance.view,"contents",OCT_CURRENT_VERSION,"r")) < OCT_OK)
  {
    sprintf(errBuf,"(tap_stuff): ERROR in opening via master -- status %d\n",stat);
    logError(errBuf,FATAL);
  }

if(octBB(&conn,box) != OCT_OK)
  {
    sprintf(errBuf,"(tap_stuff): ERROR in getting bound box for layer %d to %d connector\n",layerOne,layerTwo);
    logError(errBuf,FATAL);
  }
/*
printf("%d %d %d %d del %d %d\n",
       box->lowerLeft.x,
       box->lowerLeft.y,
       box->upperRight.x,
       box->upperRight.y,
       box->upperRight.x - box->lowerLeft.x,
       box->upperRight.y - box->lowerLeft.y);
*/
}

set_up_layers(layers)
octObject layers[];
{
  layers[POLY].type = OCT_LAYER;
  layers[POLY].contents.layer.name = "POLY";
  OCTCHKOK(octGetOrCreate(&oCellFacet,&(layers[POLY])));
  layers[MET1].type = OCT_LAYER;
  layers[MET1].contents.layer.name = "MET1";
  OCTCHKOK(octGetOrCreate(&oCellFacet,&(layers[MET1])));
  layers[MET2].type = OCT_LAYER;
  layers[MET2].contents.layer.name = "MET2";
  OCTCHKOK(octGetOrCreate(&oCellFacet,&(layers[MET2])));
}

get_standard_connector_names()
{
    octObject master;
    char *mystrcpy();

    get_connector_info(&master,POLY,POLY,TAP_HORIZONTAL,TAP_VERTICAL);
    via_nameTable[POLY][POLY] = mystrcpy( master.contents.instance.master);

    get_connector_info(&master,POLY,MET1,TAP_HORIZONTAL,TAP_VERTICAL);
    via_nameTable[POLY][MET1] = mystrcpy( master.contents.instance.master);
    via_nameTable[MET1][POLY] = mystrcpy( master.contents.instance.master);

    via_nameTable[MET2][POLY] = mystrcpy( master.contents.instance.master);
    via_nameTable[POLY][MET2] = mystrcpy( master.contents.instance.master);
    /*
      get_connector_info(&master,POLY,MET2,TAP_HORIZONTAL,TAP_VERTICAL);
      via_nameTable[POLY][MET2] = mystrcpy( master.contents.instance.master);
      via_nameTable[MET2][POLY] = mystrcpy( master.contents.instance.master);
      */

    get_connector_info(&master,MET1,MET1,TAP_HORIZONTAL,TAP_VERTICAL);
    via_nameTable[MET1][MET1] = mystrcpy( master.contents.instance.master);

    get_connector_info(&master,MET2,MET1,TAP_HORIZONTAL,TAP_VERTICAL);
    via_nameTable[MET2][MET1] = mystrcpy( master.contents.instance.master);
    via_nameTable[MET1][MET2] = mystrcpy( master.contents.instance.master);

    get_connector_info(&master,MET2,MET2,TAP_HORIZONTAL,TAP_VERTICAL);
    via_nameTable[MET2][MET2] = mystrcpy( master.contents.instance.master);
}

get_connector_info(master,layerOne,layerTwo,layerOne_dir,layerTwo_dir)
int layerOne,layerTwo,layerOne_dir,layerTwo_dir;
octObject *master;
{
    tapLayerListElement layerlist[2];
    octStatus stat;

    layerlist[0].layer.type = OCT_LAYER;
    layerlist[0].layer.contents.layer.name = layer_nameTable[layerOne];
    OCTCHKOK(octGetOrCreate(&oCellFacet,&(layerlist[0].layer)));
    layerlist[1].layer.type = OCT_LAYER;
    layerlist[1].layer.contents.layer.name = layer_nameTable[layerTwo];
    OCTCHKOK(octGetOrCreate(&oCellFacet,&(layerlist[1].layer)));

    layerlist[0].width = layer_width[layerOne];
    layerlist[1].width = layer_width[layerTwo];
    layerlist[0].direction = layerOne_dir;
    layerlist[1].direction = layerTwo_dir;

    /*
      printf("layer %s width %d dir %d\n",
      layerlist[0].layer.contents.layer.name,
      layerlist[0].width,
      layerlist[0].direction);
      printf("layer %s width %d dir %d\n",
      layerlist[1].layer.contents.layer.name,
      layerlist[1].width,
      layerlist[1].direction);
      */
    if((stat = tapGetConnector(2,layerlist,master)) != OCT_OK) {
	errRaise("spider", 1, "getting connector-- status %d layer %d to layer %d\n",
		 stat,layerOne,layerTwo);
      }
}

closeTapFiles()
{
 OCTCHKOK(octCloseFacet(&oCellFacet));
}



get_inst_bound2(side, name)
int side;
char *name;
{
octObject inst;
struct octBox box;

inst.type = OCT_INSTANCE;
inst.contents.instance.name = name;
OCTCHKOK(octGetByName(&oCellFacet, &inst));
OCTCHKOK(octBB(&inst, &box));
switch(side)
  {
  case TOP:
    return(box.upperRight.y);
  case BOTTOM:
    return(box.lowerLeft.y);
  case LEFT:
    return(box.lowerLeft.x);
  case RIGHT:
    return(box.upperRight.x);
  }
return(MAXINT);
}
