#include "copyright.h"
#include "port.h"

/*************************************************************************
 *
 * 		            WIREWIDTH 
 * 		            ---------
 *
 *	---- Functions 
 *
 *  	Compute wire width for each wire segmnet
 *   	Updates the values for RC_LOC->row_width 
 *
 *  	---- This routine suts up following data structures:
 *  	RC_LOC->row_width[]
 *  	pw_subnet_list
 *  	pw_pin_list
 *
 *************************************************************************/
#include "spider.h"
#include "powernet.h"

wirewidth()
{
    int 	netid;		/* loop counter */
    int 	side;		/* loop counter */
    int 	delt_width;	/* the diff between subnet width & row width*/

    pin_type	*pinptr;
    subnet_type *subnet,*subnet2,
		*pw_subnet_list = NULL;
    pw_pin_type *powerpin, 
		*pw_pin_list = NULL;
    via_type    *via;
    struct subnet_list *sl,*sl2;
    int row;

    /* Pre-set */

    if (exist_power_net == TRUE)  
    {
        printf("(wirewidth): exist_power_net == TRUE\n");

	/* for each net, compute the width for each wire segment */
        for (netid=1;  netid <= net_count;  netid++) 
        {
	     if (netlist[netid]->nettype == POWERNET && 
	         netlist[netid]->numpins > 1)
	     {
	
	         printf("\n(wirewidth): compute width for net %s (netid=%d)\n", 
			  netlist[netid]->netname, netid);

	         /* get power subnet list pointer */
	         pw_subnet_list = netlist[netid]->subnetlist;  

	         /* build power pin list */
		 pw_pin_list = NULL;

                 for (side=TOP;  side <= RIGHT;  side++) { 
                      for (pinptr=FIRST_PIN(side);  
		           pinptr!=NULL;  
		           pinptr=pinptr->next)  
		      {
		           if (pinptr->netid == netid)
		           {
			       powerpin = CALLOC(pw_pin_type, 1);
			       powerpin->pinid = atoi(pinptr->pinname);
			       powerpin->x = pinptr->x;
			       powerpin->y = pinptr->y;
			       powerpin->power = pinptr->power;
			       powerpin->next  = pw_pin_list;
    
			       pw_pin_list = powerpin;
			       printf("pin %s, x,y=%d,%d, cur=%g\n",
				     pinptr->pinname, 
				     powerpin->x, powerpin->y, powerpin->power);
		           }
                      }
                 }
    
	         /* compute each subnet width */
	         SPfindWidth(pw_subnet_list, pw_pin_list);

		 /* update row width */
                 for (subnet=netlist[netid]->subnetlist; 
		      subnet!=NULL; 
		      subnet=subnet->next)  
 	         {
	             if (subnet->dir == HORIZONTAL)
	             {
		        /* 
		         * note: RC_LOC->row_width[] is defined to be 
			 * the max wire and contacts width + max spacing 
		         */
	                delt_width = subnet->width + MAX_SPACING - 
				     RC_LOC->row_width[subnet->row1];

	                if (delt_width > 0) 
			{
			   printf("(wirewidth): note. row %d -- old width %d new width %d \n",
				  subnet->row1,
				  RC_LOC->row_width[subnet->row1],
				  RC_LOC->row_width[subnet->row1] + delt_width);
			   RC_LOC->row_width[subnet->row1] += delt_width;
/*			   changed_width = TRUE; */
			}
	             }
 	         }
             }
        }

	for (via = vialist;  via != NULL;  via = via->next) 
	  {
	    for(sl = via->subnets;sl != NULL;sl = sl->next)
	      {
		subnet = sl->subnet;
		for(sl2 = via->subnets;sl2 != NULL;sl2 = sl2->next)
		  if(sl2 != sl)
		    {
		      subnet2 = sl2->subnet;
		      if((subnet->layerid != subnet2->layerid) &&
			 (subnet->dir == HORIZONTAL) && 
			 (subnet2->dir == HORIZONTAL) &&
			 ((subnet->col1 == subnet2->col2) ||
			  (subnet->col2 == subnet2->col1)))
			{
			  delt_width = MIN(subnet->width,subnet2->width) +
			    MAX_SPACING - RC_LOC->row_width[subnet->row1];
			  
			  if (delt_width > 0) 
			    {
			      printf("(wirewidth): note2. row %d -- old width %d new width %d \n",
				     subnet->row1,
				     RC_LOC->row_width[subnet->row1],
				     RC_LOC->row_width[subnet->row1] + delt_width);
			      RC_LOC->row_width[subnet->row1] += delt_width;
/*			      changed_width = TRUE; */
			    }
			}
		      else
			{
			  if((subnet->layerid != subnet2->layerid) &&
			     (subnet->dir == VERTICAL) && 
			     (subnet2->dir == VERTICAL) &&
			     ((subnet->row1 == subnet2->row1) ||
			      (subnet->row1 == subnet2->row2) ||
			      (subnet->row2 == subnet2->row1) ||
			      (subnet->row2 == subnet2->row2)))
			    {
			      if(subnet->row1 == subnet2->row2)
				row = subnet->row1;
			      else
				row = subnet->row2;
			      delt_width = MIN(subnet->width,subnet2->width) +
				MAX_SPACING - RC_LOC->row_width[row];
			      
			      if (delt_width > 0) 
				{
				  printf("(wirewidth): note3. row %d -- old width %d new width %d \n",
					 row,
					 RC_LOC->row_width[row],
					 RC_LOC->row_width[row] + delt_width);
				  RC_LOC->row_width[row] += delt_width;
/*				  changed_width = TRUE; */
				}
			    }
			}
		    }
	      }
	  }
      }
    printf("Complete wirewidth.c\n\n");
}
