#include "myU.h"
#include <stdio.h>
#include "symlib.h"

static int debug = 0;

int countAdjacentSegments( term, seg )
    octObject* term;
    octObject* seg;
{
    octGenerator genPath;
    octObject adj;
    int count= 0;

    octInitGenContainers( term, OCT_PATH_MASK, &genPath );
    while (octGenerate( &genPath, &adj )  == OCT_OK ) {
	if (octIdsEqual(adj.objectId, seg->objectId)) continue;
	count++;
    }
    return count;
}

int getAdjacentSegment( term, seg, adj, dir)
    octObject* term;
    octObject* seg;
    octObject* adj;		/* returned  */
    int dir;			/* if there are more */
{
    octGenerator genPath;
    int count= 0;

    octInitGenContainers( term, OCT_PATH_MASK, &genPath );
    while (octGenerate( &genPath, adj )  == OCT_OK ) {
	if (octIdsEqual(adj->objectId, seg->objectId)) continue;
	if(pathDirection(adj) != dir)
	  {
	    return OCT_OK;
	  }
	count++;
      }
    return OCT_NOT_FOUND;
}

octCoord length( path )
    octObject* path;
{
    octCoord l;
    int num;
    struct octPoint *points;

    ohGetPoints( path, &num, &points );
    if (num != 2 ) return -1;
    l = iabs( points[0].x - points[1].x) + iabs(points[0].y - points[1].y);
    return l;
}

UgetSegment( path, array )
    octObject* path;
    struct octPoint** array;
{
    int num;
    struct octPoint* tmp_point_array;
    struct octPoint* na;	/* new array */

    ohGetPoints( path,  &num, &tmp_point_array );
    if (num != 2 ) {
	fprintf( stderr, "More than two points in path\n");
	return -1;
    }
    na = (struct octPoint*)malloc( 2 * sizeof( struct octPoint ));
    na[0] = tmp_point_array[0];
    na[1] = tmp_point_array[1];
    *array = na;
    return 0;
}

enum Udirection foundU( center, edge1, edge2 ) 
    octObject* center;
    octObject* edge1;
    octObject* edge2;
{
    int fp1, fp2;		/* indices for the farthest points */
    enum Udirection dir1,dir2;	/* directions of the two edges */
    struct octPoint* center_points;
    struct octPoint* edge1_points;
    struct octPoint* edge2_points;
    extern enum Udirection findDirection();

    UgetSegment( center, &center_points );
    UgetSegment( edge1 , &edge1_points );
    UgetSegment( edge2 , &edge2_points );

    if(direction(center_points) == direction(edge1_points))
      {
	return U_STRAIGHT;
      }
    else
      {
	if(direction(center_points) == direction(edge2_points))
	  return U_STRAIGHT;
	else
	  {
	    fp1 = ( distance( center_points, &edge1_points[0] ) > distance ( center_points, & edge1_points[1] )) ? 0 : 1;
	    fp2 = ( distance( center_points, &edge2_points[0] ) > distance ( center_points, & edge2_points[1] )) ? 0 : 1;
	    
	    dir1 = findDirection( center_points, &edge1_points[fp1] ) ;
	    dir2 = findDirection( center_points, &edge2_points[fp2] ) ;
	    
	    if ( dir1 == U_UNDEFINED || dir2 == U_UNDEFINED ) return U_UNDEFINED;
	    if (dir1 == dir2 ) return dir1;
	  }
      }
    return U_UNDEFINED;
}

int findAllU( net, list, direction_mask, max_width, min_depth )
octObject* net;
Ulist* list;
int direction_mask;		/* Requested directions for U's */
octCoord max_width;		/* Max width of the U's */
octCoord min_depth;
{
int adj1, adj2;
octGenerator genTerm;
octObject path, path1, path2;
octObject term1, term2;
octGenerator genPath;
U* u;
int count = 0;
octCoord width = 0;
enum Udirection direction;
enum Udirection foundU();
struct octPoint *p;

octInitGenContents( net, OCT_PATH_MASK, &genPath ) ;
while ( octGenerate( &genPath, &path ) == OCT_OK ) 
  {
    octInitGenContents( &path, OCT_TERM_MASK, &genTerm) ;
    
    octGenerate( &genTerm, &term1 ) ;
    octGenerate( &genTerm, &term2 ) ;
    if ( !  symConnectorp( &term1 ) )  continue;
    if ( !  symConnectorp( &term2 ) )  continue;
    
    adj1 =  countAdjacentSegments( &term1, &path ) ;
    adj2 =  countAdjacentSegments( &term2, &path ) ;
    
    if ( (adj1 != 1) || (adj2 != 1)) 
      continue;
    
    getAdjacentSegment( &term1, &path, &path1, 0 );
    getAdjacentSegment( &term2, &path, &path2, 0 );
    
    direction =  foundU ( &path, &path1, &path2 ) ;
    switch(direction)
      {
      case U_RIGHT:
      case U_LEFT:
      case U_UP:
      case U_DOWN:
	if((1<<(int)direction)&(int)direction_mask)
	  {
	    width = length( & path );
	    if ( max_width <= 0 || width <= max_width ) 
	      {
		u = createU( &path,  direction, width );
		u->Uc1 = term1.contents.term.instanceId;
		u->Uc2 = term2.contents.term.instanceId;
		if (debug) fprintf( stderr, "Found U with direction %d, width %d\n" , (int)direction, width);
		addUlist( list, u );
		count++;
	      }
	  }
	break;
      case U_STRAIGHT:
	if(get_continious_path_terms(&path,&term1,&term2) == FALSE)
	  continue;
	
	OCTCHKOK(octGetById(&term1));
	OCTCHKOK(octGetById(&term2));
	
	getAdjacentSegment( &term1, &path, &path1, pathDirection(&path));
	getAdjacentSegment( &term2, &path, &path2, pathDirection(&path));
	
	direction =  foundU ( &path, &path1, &path2 ) ;
	if((direction != U_UNDEFINED) &&
	   (direction != U_STRAIGHT) &&
	   (1<<(int)direction)&(int)direction_mask ) 
	  {
	    width = length( & path );
	    if ( max_width <= 0 || width <= max_width ) 
	      {
		u = createU( &path,  direction, width );
		u->Uc1 = term1.contents.term.instanceId;
		u->Uc2 = term2.contents.term.instanceId;
		if (debug) fprintf( stderr, "Found U with direction %d, width %d\n" , (int)direction, width);
		addUlist( list, u );
		count++;
	      }
	  }
	break;
      default:
	;
      }
  }
return count;
}
