#include <stdio.h>

#include "fang.h"
#include "oct.h"
#include "harpoon.h"
#include "oh.h"

static int debug = 0;

/*
 * Returns 0 if intersection is empty, otherwise it returns 1
 * and the intersection box has the bounding box of the intersection
 * Temporarily, also check intersection with the layer PLACE,
 * which is the union of all layers.
 */
int instanceBoxIntersect( inst, box, layername, intersectionBox )
    octObject* 		inst;
    struct octBox* 	box;
    char* 		layername;
    struct octBox 	*intersectionBox; /* RETURNED value */
{
    octObject master;		/* interface facet of master of inst */
    octObject layer;		/* the layer in master with name layername */
    octObject layerPlace;	/* The layer PLACE */
    fa_geometry geobox;		/* fang representation of box */
    fa_geometry geolayer;	/* fang representation of geometry on layer */
    int ret = 0;		/* return code */

    ohOpenMaster( &master, inst, "interface", "r" );
    if ( ohGetByLayerName( &master, &layer, layername ) == OCT_OK || 
	 ohGetByLayerName( &master, &layerPlace, "PLACE" ) == OCT_OK )  {

	fa_init( &geobox );
	fa_add_box( &geobox, box->lowerLeft.x, box->lowerLeft.y, box->upperRight.x, box->upperRight.y );

	fa_init( &geolayer );
	if ( octGetById( &layer ) == OCT_OK ) {
	    harAddContentsToFang( &layer, &geolayer, &(inst->contents.instance.transform) );
	}
	if ( octGetById( &layerPlace ) == OCT_OK )  {
	    harAddContentsToFang( &layerPlace, &geolayer, &(inst->contents.instance.transform) );
	}

	if ( debug ) {
	    fprintf( stderr, "Intersecting geometries:\n" );
	    fa_dump_geo( &geobox );
	    fa_dump_geo( &geolayer );
	}
	if ( fa_intersect_p( &geobox, &geolayer ) ) {
	    {
		fa_geometry geointersection;
		fa_box ib;	/* intersection bbox */

		fa_init( &geointersection );
		fa_and( &geobox, &geolayer, &geointersection );
		fa_bb( &geointersection, &ib );
		intersectionBox->lowerLeft.x = ib.left;
		intersectionBox->lowerLeft.y = ib.bottom;
		intersectionBox->upperRight.x = ib.right;
		intersectionBox->upperRight.y = ib.top;

		ret = 1;	/* YES, there is an intersection */
		if ( debug ) {
		    fprintf( stderr, "Intersection\n" );
		    fa_dump_geo( &geointersection );
		}
		fa_free_geo( &geointersection );
	    }
	    
	}
	fa_free_geo( &geobox );
	fa_free_geo( &geolayer );
    }
    /* For speed reasons, do not free this facet */
    /* octFreeFacet( &master ); */
    return ret;
}
