/*
 *  FILE: main.c
 *
 *  DESCRIPTION: main routine for the program utify
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR: Andrea Casotto
 *  MODIFY DATE:  June 1989
 *
 *  FUNCTIONS: 
 */
#include "copyright.h"
#include "port.h"
#include "options.h"
#include "myU.h"
#include "oh.h"

#define NIL( type )   ((type*)0)
#define STRSAV(s)		strcpy(MALLOC(char,strlen(s)+1), s)
#define MALLOC(type,number)  (type *)malloc((unsigned)(sizeof(type)*(number)))
/*
 *   >>>  GLOBAL VARIABLES DECLARATION
 */
extern octObject connector_bag;

short	verbose = FALSE;


/* ARGSUSED */

optionStruct optionList[] = {
    {"w", "integer",   "Max U width to be considered"},
    {"l", "integer",   "Min U length to be considered"},
    {"v", "",                  "Verbose mode"},
    {"o", "cell:view",        "Specify output cell and view"},
    {OPT_RARG,             "","cell:view"},
    { 0, 0 , 0}
};



main(argc,argv)
    int argc;
    char *argv[];
{
    char *output_cell, *output_view;
    char *input_cell, *input_view;
    int c;
    octObject	Chip;		/* The chip on which utify runs */
    char		buf[256];
    int max_width = 0;
    int min_length = 0;

    sprintf( buf, "Starting %s: shorten all U-routes", argv[0]);
    Message( buf );

    /* parse command line options*/
    output_cell = NIL(char);
    output_view = NIL(char);
    while ((c = optGetOption(argc,argv)) != EOF) {
	switch(c) {
	case 'w':
	    max_width = atoi( optarg);
	    break;
	case 'l':
	    min_length = atoi( optarg );
	    break;
	case 'o':
	    output_cell = STRSAV(optarg);
	    output_view = strchr(output_cell, ':');
	    if (output_view != NIL(char)) {
		*output_view++ = '\0';
	    }
	    break;
	case 'v':
	    verbose = TRUE;
	    break;
	default:
	    optUsage(argc, argv);
	    break;
	}
    }

    /* the remaining arguments are argv[optind ... argc-1] */
    if (optind >= argc) {
	optUsage(argc, argv);
    } else {
	input_cell = STRSAV(argv[optind]);
	optind++;
	input_view = strchr(input_cell, ':');
	if (input_view != NIL(char)) {
	    *input_view++ = '\0';
	} else {
	    input_view = "symbolic";
	}
    }
    /* check for extraneous arguments */
    if (optind < argc)    optUsage(argc, argv);

    octBegin();


    if (output_cell != NIL(char)) {
	if (output_view == NIL(char)) output_view = "utify";
	utify_copyFacets(input_cell , input_view , output_view);
    } else {
	output_view = input_view;
	output_cell = input_cell;
    }

    Chip.type = OCT_FACET;
    Chip.contents.facet.cell = output_cell;
    Chip.contents.facet.view = output_view;
    Chip.contents.facet.facet = "contents";
    Chip.contents.facet.version = OCT_CURRENT_VERSION;
    Chip.contents.facet.mode = "a"; 
    if (octOpenFacet(&Chip) < OCT_OK ) {
	(void) sprintf(buf,"Opening %s", ohFormatName(&Chip));
	FatalError(buf);
    }

    connector_bag.type = OCT_BAG;
    connector_bag.contents.bag.name = "CONNECTORS";
    OCTCHKOK(octGetOrCreate(&Chip,&connector_bag));

    doUreduction( &Chip, max_width, min_length );

  
    OH_ASSERT(octFlushFacet(&Chip));
    octEnd();
    exit(0);			/* successfull */
}


void showActivity( n )
    int n;
{
    static int count  = 0;
    static int line = 10;

    if ( ( ++count % n ) == 0) {
	fprintf( stderr, "." );
	if (line-- == 0) {
	    fprintf( stderr, "\n" );
	    line = 10;
	}
    }
}

int  doUreduction( facet, max_width, min_length ) 
    octObject *facet;
    int max_width, min_length;
{
    int debug = 1;
    octObject net;
    octGenerator genNet;
    int some_U_reduced = 0;
    Ulist* v_list;		/* List of vertical U's */
    Ulist* h_list;		/* List of horizontal U's */
    UlistElement* ptr ;
    U* u;
    char msg[256];
    int pass = 1;		/* counter of iterations */

    v_list = createUlist();
    h_list = createUlist();

    do {
	some_U_reduced = 0;
	/* 
	 * Do vertical U's first
	 */
	Message( "Vertical U's" );
	clearUlist( v_list );
	OH_ASSERT( octInitGenContents( facet, OCT_NET_MASK, &genNet )) ;
	while ( octGenerate( &genNet, &net ) == OCT_OK ) {
	    findAllU( &net, v_list, U_UP_MASK | U_DOWN_MASK, max_width, min_length );
	}
	sprintf( msg , "Found %d vertical  U's", v_list->count );
	Message( msg );

	for ( ptr = v_list->first; ptr != 0 ; ptr = ptr->next ) {
	    u = ptr->u;
	    if ( push_U( facet, u ) != 0 ) {
		some_U_reduced++;
	    }
	    showActivity( 1 );
	}
	if (debug) octFlushFacet( facet );
	/*
	 * Now do horizontal U's
	 */

	Message( "Horizontal U's" );
	clearUlist( h_list );
	OH_ASSERT( octInitGenContents( facet, OCT_NET_MASK, &genNet )) ;
	while ( octGenerate( &genNet, &net ) == OCT_OK ) {
	    findAllU( &net, h_list, U_RIGHT_MASK | U_LEFT_MASK, max_width, min_length );
	}
	sprintf( msg , "Found %d horizontal  U's", h_list->count );
	Message( msg );

	for ( ptr = h_list->first; ptr != 0 ; ptr = ptr->next ) {
	    u = ptr->u;
	    if ( push_U( facet, u ) != 0 ) {
		some_U_reduced++;
	    }
	    showActivity( 1 );
	}
	if (debug) octFlushFacet( facet );

	sprintf( msg, "%d U's reduced in pass %d", some_U_reduced, pass );
	Message( msg );
	pass++;
    } while (some_U_reduced );


    sprintf( msg , "Done reducing" );
    Message( msg );

    return ;
}

FatalError(es)
     char *es;			/* error string */
{
  (void) fprintf(stderr,"Fatal Error: %s\n",es);
  exit(-1);
}


Warning(es)			/* continuable error */
     char *es;			/* error string */
{
  (void) fprintf(stderr, "utify:warning: %s\n", es);
}

Message(es)			
     char *es;			/* error string */
{
  (void) fprintf(stderr, "utify:message: %s\n", es); 
}


utify_copyFacets ( cellName , oldView , newView )
    char *cellName, *oldView ,*newView;
{
  octObject  	Input, Output;
  char buf[256];

  (void) sprintf(buf,"Copying %s:%s into %s:%s", cellName, oldView , cellName, newView);
  Message(buf);

  Input.type = OCT_FACET;
  Input.contents.facet.cell = cellName;
  Input.contents.facet.view = oldView;
  Input.contents.facet.facet = "contents";
  Input.contents.facet.version = OCT_CURRENT_VERSION;

  Output.type = OCT_FACET;
  Output.contents.facet.cell = cellName;
  Output.contents.facet.view = newView;
  Output.contents.facet.facet = "contents";
  Output.contents.facet.version = OCT_CURRENT_VERSION;

  if (octCopyFacet(&Output,&Input) != OCT_OK) FatalError("Copying contents facet");
  Input.contents.facet.facet = "interface";
  Output.contents.facet.facet = "interface";
  if (octCopyFacet(&Output,&Input) != OCT_OK) {
      Warning("Problems copying interface facet. Does it exist ?");
  }
}
