#include "port.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"

extern int ground_reqd;	/* Imported from power.c */
extern int supply_reqd;
extern double peak_vdd_i;
extern double peak_gnd_i;

/*
 *  cleanup_cell: remove any 'left-overs' from a previous routing
 */

int cleanup_cell(facet, int_facet, interface_present)
octObject *facet, *int_facet;
int interface_present;		/* == 1 => interface facet provided */
{
    octObject net, path, inst, bag, layer;
    octObject inter_term, content_term,prop;
    octObject interface;
    octGenerator gen, gen1,genTerm;
    octStatus status;
    char *name;
    int n;

    if (interface_present) {
        interface = *int_facet;
    }
    if (!interface_present &&
	    ohOpenInterface(&interface,facet,"a") < OCT_OK) {
	(void) printf("Unable to open interface of the facet \n");
	return 1;
    }
    /*
     * Look for the old properties in the contents facet
     * and translate them into the correct properties in the
     * interface
     */

    OH_ASSERT(octInitGenContents(facet,OCT_TERM_MASK,&genTerm));
    while( octGenerate(&genTerm,&content_term)  == OCT_OK){
	if (octIdIsNull(content_term.contents.term.instanceId)) {
	    status = ohGetByPropName(&content_term,&prop,"WOLFE-TOP");
	    if (status == OCT_OK && prop.contents.prop.type == OCT_REAL){
		OH_ASSERT(octDelete(&prop));
		OH_ASSERT_DESCR(ohGetByTermName(&interface,&inter_term,
		content_term.contents.term.name),"Getting terminal in interface");
		OH_ASSERT(ohCreatePropReal(&inter_term,&prop,"TERM_RELATIVE_POSITION",
		prop.contents.prop.value.real));
		OH_ASSERT(ohCreatePropStr(&inter_term,&prop,"TERM_EDGE","TOP"));
		continue;
	    }
	    status = ohGetByPropName(&content_term,&prop,"WOLFE-BOTTOM");
	    if ( status == OCT_OK && prop.contents.prop.type == OCT_REAL){
		OH_ASSERT(octDelete(&prop));
		OH_ASSERT_DESCR(ohGetByTermName(&interface,&inter_term,
		content_term.contents.term.name),"Getting terminal in interface");
		OH_ASSERT(ohCreatePropReal(&inter_term,&prop,"TERM_RELATIVE_POSITION",
		prop.contents.prop.value.real));
		OH_ASSERT(ohCreatePropStr(&inter_term,&prop,"TERM_EDGE","BOTTOM"));
		continue;
	    }
	    status = ohGetByPropName(&content_term,&prop,"WOLFE-LEFT");
	    if ( status == OCT_OK && prop.contents.prop.type == OCT_REAL){
		OH_ASSERT(octDelete(&prop));
		OH_ASSERT_DESCR(ohGetByTermName(&interface,&inter_term,
		content_term.contents.term.name),"Getting terminal in interface");
		OH_ASSERT(ohCreatePropReal(&inter_term,&prop,"TERM_RELATIVE_POSITION",
		prop.contents.prop.value.real));
		OH_ASSERT(ohCreatePropStr(&inter_term,&prop,"TERM_EDGE","LEFT"));
		continue;
	    }
	    status = ohGetByPropName(&content_term,&prop,"WOLFE-RIGHT");
	    if ( status == OCT_OK && prop.contents.prop.type == OCT_REAL){
		OH_ASSERT(octDelete(&prop));
		OH_ASSERT_DESCR(ohGetByTermName(&interface,&inter_term,
		content_term.contents.term.name),"Getting terminal in interface");
		OH_ASSERT(ohCreatePropReal(&inter_term,&prop,"TERM_RELATIVE_POSITION",
		prop.contents.prop.value.real));
		OH_ASSERT(ohCreatePropStr(&inter_term,&prop,"TERM_EDGE","RIGHT"));
		continue;
	    }
	}
    }
    if (!interface_present) OH_ASSERT(octCloseFacet(&interface));

    /*
     *  Generate over all of the nets to:
     *	    1) delete all net implementations
     *	    2) drop any trivial nets
     *	    3) delete all connectors and feedthrus
     *	    4) delete all paths on MET1 and MET2
     */
    OH_ASSERT(octInitGenContents(facet, OCT_NET_MASK, &gen));
    while ((status = octGenerate(&gen, &net)) == OCT_OK) {
	name = net.contents.net.name;

	/* Generate all paths connected to the net and delete them */
	OH_ASSERT(octInitGenContents(&net, OCT_PATH_MASK, &gen1));
	while ((status = octGenerate(&gen1, &path)) == OCT_OK) {
	    OH_ASSERT(octDelete(&path));
	}
	OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

	/* Check for a trivial net */
	n = ohCountContents(&net, OCT_TERM_MASK);
	if (n < 2) {
	    (void) fprintf(stderr, 
		"Net \"%s\" with fewer than 2 terminals deleted\n", name);
	    OH_ASSERT(octDelete(&net));
	}
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);


    /* 
     *  Delete any feed-thru cells (from the WOLFE-FEED bag) 
     */
    OH_ASSERT(ohGetOrCreateBag(facet, &bag, "WOLFE-FEED"));
    OH_ASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen));
    while ((status = octGenerate(&gen, &inst)) == OCT_OK) {
	OH_ASSERT(octDelete(&inst));
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);


    /* 
     *  Delete the routing instances (from the CONNECTORS bag)
     */
    OH_ASSERT(ohGetOrCreateBag(facet, &bag, "CONNECTORS"));
    OH_ASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen));
    while ((status = octGenerate(&gen, &inst)) == OCT_OK) {
	OH_ASSERT(octDelete(&inst));
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

	if (ohGetByLayerName(facet, &layer, hlayer_name) == OCT_OK) {
		OH_ASSERT(octInitGenContents(&layer, OCT_PATH_MASK, &gen));
		while ((status = octGenerate(&gen, &path)) == OCT_OK) {
		OH_ASSERT(octDelete(&path));
		}
		OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
	}

	if (ohGetByLayerName(facet, &layer, vlayer_name) == OCT_OK) {
	    OH_ASSERT(octInitGenContents(&layer, OCT_PATH_MASK, &gen));
	    while ((status = octGenerate(&gen, &path)) == OCT_OK) {
		OH_ASSERT(octDelete(&path));
	    }
	    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
	}

    /*
     * For all the instances place them at the origin and 
     * make the transform the same as that of the master.
     */
    if ( ohGetByBagName( facet, &bag, "INSTANCES") != OCT_OK){
	(void) printf( "No INSTANCES bag in %s\n", ohFormatName(facet));
	return 1;
    }
    OH_ASSERT( octInitGenContents( &bag, OCT_INSTANCE_MASK, &gen));
    while ( (status = octGenerate( &gen, &inst)) == OCT_OK){
	inst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
	inst.contents.instance.transform.translation.x = 0;
	inst.contents.instance.transform.translation.y = 0;
	(void) octModify( &inst);
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    return 0;
}

/*
*
* update_interface: Routine for updating the new interface facet  
*
*/

int update_interface(facet,newfacet)
octObject *facet, *newfacet;
{
    octObject term,term1,term2,prop,gnd_term,power_term;
    octObject bag,bag1,newbag,newterm;
    octGenerator genBag,genBagTerm,genTerm;
    octStatus s, UpdateContents();

    /* Copy all the properties and Bags to the newfacet */

    s = UpdateContents(facet,newfacet, OCT_BAG_MASK || OCT_PROP_MASK);
    if (s != OCT_OK) return 1;

    OH_ASSERT(octInitGenContents(facet,OCT_TERM_MASK,&genTerm));
    while (octGenerate( &genTerm,&term) == OCT_OK) {
	OH_ASSERT(ohGetByTermName(newfacet,&newterm,term.contents.term.name));
	s = UpdateContents(&term,&newterm,OCT_PROP_MASK);
	if ( s != OCT_OK) return 1;
	if (ohGetByPropName(&term,&prop,"TERMTYPE") == OCT_OK) {
	    if(strcmp(prop.contents.prop.value.string,"SUPPLY")==0) {
		power_term = term;
	    } else if(strcmp(prop.contents.prop.value.string,"GROUND") == 0) {
		gnd_term = term;
	    } else {
		/* Update the location of the terminal */
		/* XXX */
	    }
	}

	OH_ASSERT(octInitGenContents(&term,OCT_BAG_MASK,&genBag));
	while (octGenerate(&genBag,&bag) == OCT_OK) {
	    if (ohGetByPropName(&bag,&prop,"WOLFE_VISITED") != OCT_OK) {
		/* Create the bag and fix the containers */
		newbag = bag;
		OH_ASSERT(octCreate(&newterm,&newbag));
		s = UpdateContents(&bag,&newbag,OCT_PROP_MASK);
		if(s != OCT_OK) return 1;
		OH_ASSERT(octInitGenContainers(&bag,OCT_TERM_MASK,&genBagTerm));
		while (octGenerate(&genBagTerm,&term1) == OCT_OK) {
		    if (!octIdsEqual(term1.objectId, term.objectId)) {
			OH_ASSERT(ohGetByTermName(newfacet,&term2,term1.contents.term.name));
			OH_ASSERT(octAttach(&term2,&newbag));
		    }
		}
		OH_ASSERT(ohCreatePropStr(&bag,&prop,"WOLFE_VISITED",""));
	    }
	}
    }

    /* Check to see that the new power terminals have the 
       required properties */
    if (supply_reqd >= 1) {
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Vdd"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","SUPPLY"));
	if( (supply_reqd == 1) && ( power_position == BOTH)) {
	    OH_ASSERT(ohGetOrCreateBag(newfacet,&bag,"JOINEDTERMS"));
	    OH_ASSERT(ohCreateBag(&bag,&bag1,"JOINED"));
	    OH_ASSERT(octAttach( &bag1,&term));
	    OH_ASSERT(octAttach( &bag1,&power_term));
	}
    }
    if (supply_reqd == 2) {
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Vdd1"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","SUPPLY"));
	/* Create the joined terms bag */
	OH_ASSERT(ohGetOrCreateBag(newfacet,&bag,"JOINEDTERMS"));
	OH_ASSERT(ohCreateBag(&bag,&bag1,"JOINED"));
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Vdd"));
	OH_ASSERT(octAttach( &bag1,&term));
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Vdd1"));
	OH_ASSERT(octAttach( &bag1,&term));
    }
    if( ground_reqd >= 1){
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Gnd"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","GROUND"));
	if( (ground_reqd == 1) && ( power_position == BOTH)){
	    OH_ASSERT(ohGetOrCreateBag(newfacet,&bag,"JOINEDTERMS"));
	    OH_ASSERT(ohCreateBag(&bag,&bag1,"JOINED"));
	    OH_ASSERT(octAttach( &bag1,&term));
	    OH_ASSERT(octAttach( &bag1,&gnd_term));
	}
    }
    if( ground_reqd == 2 ){
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Gnd1"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","GROUND"));
	/* Create the joined terms bag */
	OH_ASSERT(ohGetOrCreateBag(newfacet,&bag,"JOINEDTERMS"));
	OH_ASSERT(ohCreateBag(&bag,&bag1,"JOINED"));
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Gnd"));
	OH_ASSERT(octAttach( &bag1,&term));
	OH_ASSERT(ohGetByTermName(newfacet,&term,"wolfe_Gnd1"));
	OH_ASSERT(octAttach( &bag1,&term));
    }
    return 0;
}


static double get_term_prop_real( term, name )
    octObject *term;
    char* name;
{
    octObject prop;
    if ( ohGetByPropName( term, &prop, name ) == OCT_OK ) {
	 return prop.contents.prop.value.real;
    } else if ( term->contents.term.instanceId != oct_null_id )  {
	octObject formal;
	ohFindFormal( &formal, term );
	return get_term_prop_real( &formal, name );
    } else {
	return 0.0;
    }
}

static void update_peak_current( contents, interface, term )
    octObject* contents;
    octObject* interface;
    octObject* term;
{
    octGenerator genTerm;
    octObject net, prop, actual_term;
    double peak = 0.0;

    OH_ASSERT(octGetByName( contents, term ));
    if ( octGenFirstContainer( term, OCT_NET_MASK, &net ) == OCT_OK ) {
	OH_ASSERT(octInitGenContents( &net, OCT_TERM_MASK, &genTerm ));
	while ( octGenerate( &genTerm, &actual_term ) == OCT_OK ) {
	    peak += get_term_prop_real( &actual_term, "PEAKCURRENT" );
	}
    
    }
    OH_ASSERT(octGetByName( interface, term ));
    ohCreateOrModifyPropReal( term, &prop, "PEAKCURRENT", peak );
    printf( "%s: estimated %s Amps.\n", ohFormatName( term ), ohFormatName( &prop ) );
}

int
new_update_interface(contents, facet)
octObject *contents;
octObject *facet;
{
    octObject term, prop;
    
    /* Check to see that the new power terminals have the 
       required properties */

    if (supply_reqd == 1) {
	if (power_position == BOTH){
	    OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Vdd1"));
	} else {
	    OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Vdd"));
	}
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","SUPPLY"));
    } else if (supply_reqd == 2) {
	OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Vdd"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","SUPPLY"));
	OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Vdd1"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","SUPPLY"));
    }
    if( ground_reqd == 1){
	if (power_position == BOTH){
	    OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Gnd1"));
	} else {
	    OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Gnd"));
	}
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","GROUND"));
    } else if( ground_reqd == 2 ){
	OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Gnd"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","GROUND"));
	OH_ASSERT(ohGetByTermName(facet,&term,"wolfe_Gnd1"));
	OH_ASSERT(ohCreateOrModifyPropStr(&term,&prop,"TERMTYPE","GROUND"));
    }

    {
	octGenerator genTerm;
	octObject term, prop;
	octInitGenContents( facet, OCT_TERM_MASK, &genTerm );
	while ( octGenerate( &genTerm, &term ) == OCT_OK ) {
	    if ( ohGetByPropName( &term, &prop, "TERMTYPE" ) == OCT_OK ) {
		char* termtype = prop.contents.prop.value.string;
		if ( strcmp( termtype, "GROUND" ) == 0 || strcmp( termtype, "SUPPLY" ) == 0) {
		    update_peak_current( contents, facet, &term );
		}
	    }
	}
    }
    return 0;
}


octStatus UpdateContents(src,dest,mask)
octObject *src,*dest;
octObjectMask mask;
{
    octGenerator gen;
    octObject newobj, obj;
    octStatus s;

    if( (s = octInitGenContents(src,mask,&gen)) != OCT_OK) return s;
    while( (s = octGenerate(&gen,&obj)) == OCT_OK){
	newobj = obj;
	if ( obj.type == OCT_TERM){
	    if (( s = octAttach(dest,&newobj)) != OCT_OK) return s;
	}
	else if ( obj.type == OCT_PROP){
	    if (( s = octCreate(dest,&newobj)) != OCT_OK) return s;
	}
	else if ( obj.type == OCT_BAG){
	    if( (s = octCreate(dest,&newobj)) != OCT_OK) return s;
	    /* recursively call the replace routine */
	    if ((s = UpdateContents(&obj,&newobj,OCT_PROP_MASK || OCT_TERM_MASK || OCT_BAG_MASK)) != OCT_OK) return s;
	}
    }
    if( s == OCT_GEN_DONE) s = OCT_OK;
    return s;
}
