#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"
#include "nle.h"

int
flatten_hierarchy(facet)
octObject *facet;
{
    octObject interface, master, inst, prop, bag;
    octGenerator genInst;
    int count, module_flag, interface_present;

    OH_ASSERT_DESCR(ohGetByBagName(facet, &bag, "INSTANCES"),
				  "Missing  INSTANCES bag");
    OH_ASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &genInst));
    while ( octGenerate(&genInst, &inst) == OCT_OK) {
        /* Flatten if it has more than 1 sub_instance */

        module_flag = 0;
        if (ohOpenMaster(&interface, &inst, "interface", "r") >= OCT_OK){
            /* Look for CELLCLASS == MODULE */
	    interface_present = 1;
            OH_ASSERT(ohGetByPropName( &interface, &prop, "CELLCLASS"));
            if ( strcmp( prop.contents.prop.value.string,"MODULE") == 0){
                module_flag = 1;
		if (ohOpenMaster(&master, &inst, "contents", "a") < OCT_OK) {
		    octError("Failed to open master for cleanup\n");
		    return 1;
		}
	    }
        } else {
	    interface_present = 0;
	    if (ohOpenMaster(&master, &inst, "contents", "a") < OCT_OK) {
		octError("Failed to open master for cleanup\n");
		return 1;
	    }
            /* Count the instances in the master */
            count = ohCountContents(&master, OCT_INSTANCE_MASK);
            module_flag = ( count > 1 ? 1 : 0);
        }

        if ( module_flag ){
            if ( ohGetByPropName(&master, &prop, "WOLFE_FLAT") != OCT_OK){

                if (cleanup_cell(&master, (interface_present ? &interface :
			(octObject *)NULL), interface_present) != 0) {
                    (void) printf("Failed during cleanup of %s\n",ohFormatName(&inst));
                    return 1;
                }
                OH_ASSERT( ohCreatePropStr(&master, &prop, "WOLFE_FLAT",""));
            }
            if ( nleFlatInstance(&inst, NIL(octObject), NLE_SOFT) != NLE_OK) {
                return 1;
            }
        }
    }
    return 0;
}
