/* sccsid = @(#)main.c	1.7 wolfe (ITD) 7/30/90   */
#include "port.h"
#include "utility.h"
#include "errtrap.h"
#include "options.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"
#include "vulcan.h"

extern int feed_count;			/* imported from read_place.c */
extern int imp_feed_count;		/* imported from write_cell.c */
extern int supply_reqd;			/* Imported from power.c */
extern int ground_reqd;			/* Imported from power.c */

/* local to this module */
static int run_wolfe();
static int delete_files = 1;
static char *machine = NIL(char);
static int do_timberwolf = 1;
static int hierarchy = 0;
static char *sccsid = "sccsid = @(#)main.c	1.95 wolfe (ITD) 8/14/91";
static char *sccsversion = "version 1.95 last edited 8/14/91";
#define RUN(command, status) {\
    if (debug > 1) (void) printf("EXEC: %s\n", command);	\
    status = util_csystem(command);				\
    if (debug > 1) (void) printf("status = %d\n", status);	\
    if (status != 0) {					\
	(void) fprintf(stderr, "Error status %d from command:\n    %s\n",\
	    status, command);				\
	return 1;					\
    }							\
}


optionStruct optionList[] = {
    {"f",	0,		"turn on fast mode"},
    {"h",	0,		"flatten the hierarchy"},
    {"m",	"machine",	"run on another machine"},
    {"i",	0,		"add implicit"},
    {"o",	"[cell]:view",	"output cell:view"},
    {"r",	"rows",		"force TW to use a fixed number of rows"},
    {"t",	"file",		"alternate rules file location"},
    {"e",	0,		"extend terminals"},
    /* {"x",	0,		"do not run TimberWolf"}, */
    {"d",	0,		"turn on debugging"},
    {"v",	0,		"turn all full debugging"},
    {"R",       "#",            "Set RowSep"},
    {"I",       "#",            "Set Implicit Feedthru range"},
    {"g",       "yn",           "Should TWSC do route?"},
    {"H",       "#",            "Horizontal net weight"},
    {"V",       "#",            "Vertical net weight"},
    {"P",       "#",            "Power position (LEFT, RIGHT, or BOTH)"},
    {"W",       "#",            "Power rail width in lambdas"},
    { OPT_RARG,	"cell:view",	"facet defaults to `contents'" },
    {0, 0, 0}};


abortWolfe()
{
    fprintf(stderr, "aborting run....\n");
    VOVend(-1);
}


main(argc, argv)
int argc;
char **argv;
{
    octObject interface, facet, outfacet;
    octStatus status;
    struct octBox bb;
    char *input_name, *output_name, *cellroot, *cell, *view;
    char prefix[1024], *rules_file;
    char command[1024];
    int k;
    double sugg_row_sep, actual_length;
    int option;
    extern char *optarg;
    extern int optind;

#ifdef bsd4_2
    setlinebuf(stdout);
    setlinebuf(stderr);
#endif

    /* parse command line options*/
    output_name = NIL(char);
    rules_file = "scmos/wolfe.rules";
    fprintf (stderr, "wolfe %s\n", sccsversion);
    
    while ((option = optGetOption(argc, argv)) != EOF) {
	switch(option) {
	    case 'd':
		debug = 1;
		break;
	    case 'f':
		fast_mode = 1;
		break;
	    case 'i':
		Add_implicit = 0;
		break;
	    case 'm':
		machine = util_strsav(optarg);
		break;
	    case 'o':
		output_name = util_strsav(optarg);
		break;
	    case 'r':
		num_rows = atoi(optarg);
		break;
	    case 't':
		rules_file = util_strsav(optarg);
		break;
	    case 'h':
		hierarchy = 1;
		break;
	    case 'e':
		extend_terms = 1;
		break;
	    case 'x':
		errRaise( "wolfe", 1, "Option -x is no longer used (it never worked)" );
		do_timberwolf = 0;
		break;
	    case 'v':
		debug = 99;
		delete_files = 0;
		break;
	    case 'R':
		rowSep = atof (optarg);
		break;
	    case 'I':
		implicit_feed_thru_range = atof (optarg);
		break;
	    case 'g':
		if (*optarg == 'n') {
		    do_global_route = 0;
		} 
		break;
	    case 'H':
		h_net_weight = atof (optarg);
		break;
	    case 'V':
		v_net_weight = atof (optarg);
		break;
	    case 'P':
		if (strcmp (optarg, "RIGHT") == 0)
		  power_position = RIGHT;
	        else if ( strcmp(optarg,"LEFT") == 0)
		  power_position = LEFT;
		else if ( strcmp(optarg,"BOTH") == 0)
		  power_position = BOTH;
		else
		  printf("Unknown side for Power Terminal - Using LEFT \n");
		break;
	    case 'W':
		power_width = atoi (optarg);
		break;
	    default:
		optUsage();
		break;
	}
    }

    /*
     *  Finish the command line argument parsing 
     */
    if (optind != argc-1) {
	optUsage();
    } else {
	input_name = argv[optind];
    }

    if ( hierarchy && ( output_name == NIL(char))) {
	(void) printf("output view required with -h option\n");
	abortWolfe();
    }

    VOVbegin( argc, argv );
    octBegin();
    
    /*
     *  Get the input and output facet names
     */
    ohUnpackDefaults(&facet, "a", ":logic:contents");
    OH_ASSERT(ohUnpackFacetName(&facet, input_name));
    cell = facet.contents.facet.cell;
    view = facet.contents.facet.view;

    /*
     *  Unpack the output name (if present), and make a copy of the facet
     */
    if (output_name != NIL(char)) {
	outfacet = facet;
	if (ohUnpackFacetName(&outfacet, output_name) == OCT_OK) {
	    

	    /* Copy the contents facet */
	    facet.contents.facet.facet = "contents";
	    outfacet.contents.facet.facet = "contents";
	    VOVinputFacet( &facet );
	    VOVoutputFacet( &outfacet );
	    if (octCopyFacet(&outfacet, &facet) != OCT_OK) {
		octError("copying the contents facet");
		abortWolfe();
	    }

	    /* Copy the interface facet */
	    facet.contents.facet.facet = "interface";
	    outfacet.contents.facet.facet = "interface";
	    VOVinputFacet( &facet );
	    VOVoutputFacet( &outfacet );
	    if (octCopyFacet(&outfacet, &facet) != OCT_OK) {
		octError("copying the interface facet");
		abortWolfe();
	    }

	    facet = outfacet;
	    facet.contents.facet.facet = "contents";
	    cell = outfacet.contents.facet.cell;
	    view = outfacet.contents.facet.view;
	}
    }

    /* 
     *  Open the facet 
     */
    VOVinputFacet( &facet );
    VOVoutputFacet( &facet );
    if ((status = octOpenFacet(&facet)) < OCT_OK) {
	octError("can not open the facet");
	abortWolfe();
    } else if (status == OCT_NEW_FACET) {
	(void) fprintf(stderr, "%s does not exist\n", ohFormatName(&facet));
	abortWolfe();
    } else if (status == OCT_INCONSISTENT) {
	ohPrintInconsistent(&facet, stdout);
	VOVend(-1);
    }

    /*
     *  Read the rules file (layers, etc.)
     */
    if (read_rules_file(rules_file,&facet) != 0) abortWolfe();


    /* 
     *  Make-up the prefix to be used for the TW files 
     */
    cellroot = strrchr(cell, '/');
    cellroot = (cellroot == NIL(char)) ? cell : cellroot+1;
    (void) sprintf(prefix, "/usr/tmp/wlf.XXXXXX"); /* Make short because some UNIX (e.g. HP-UX) have a 14 character limint on file names. */
    (void) mktemp(prefix);

    /* 
     *  Get the width of the feedthru cell (make sure its around as well ...)
     */
    if (ohOpenFacet(&outfacet, feedthru_cell, feedthru_view, 
		"contents", "r") < OCT_OK) {
	octError("Cannot open the feedthrough cell\n");
	abortWolfe();
    }
    VOVinputFacet( &outfacet );
    OH_ASSERT(octTB(&outfacet, &bb));
    feedThruWidth = (bb.upperRight.x - bb.lowerLeft.x) / units_per_lambda;

    /* 
     *  Read any parameters from the facet 
     */
    read_parameters(&facet);

    /* 
     *  Cleanup the facet 
     */
    if ( hierarchy && ( output_name != NIL(char)) ) {
	if (flatten_hierarchy(&facet) != 0) abortWolfe();
    } else if (cleanup_cell(&facet, NIL(octObject), 0) != 0) abortWolfe();

    /* 
     * Make the names unique
     */
    ohUniqNames(&facet, OCT_INSTANCE_MASK);
    ohUniqNames(&facet, OCT_NET_MASK);

    /* 
     *  Write the timberwolf files 
     */
    if (write_cell_file(prefix, &facet) != 0) abortWolfe();
    if (write_block_file(prefix, &facet) != 0) abortWolfe();
    if (write_param_file(prefix, &facet) != 0) abortWolfe();
    if (write_netweight_file(prefix, &facet) != 0) abortWolfe();

    /* 
     *  Run timberwolf 
     */
    if (do_timberwolf && (run_wolfe(cell, view, prefix) != 0)) {
	(void) fprintf(stderr, "Error running TimberWolf\n");
	abortWolfe();
    }

    /* Report timberwolf channel densities */
    if (echo_output_file(prefix) != 0) abortWolfe();

    /* 
     *  Read the timberwolf files (cell placement, and pin locations)
     */
    if (read_place_file(prefix, &facet) != 0) abortWolfe();
    if ( do_global_route && (read_pin_file(prefix, &facet) != 0) ) abortWolfe();

    /* 
     *  Route the channels
     */
    if (route_channels(&facet) != 0) abortWolfe();

    /*
     *  Final cleanup -- connect power and ground, and cleanup the
     *  formal terminal attachments
     */
    if (do_global_route && (connect_vdd_gnd(&facet)) != 0) {
	(void) printf("Unable to route power and ground\n");
    }

    /* 
     *  Get the final bbox and Net Statistics 
     */
    if (debug > 0) {
	octCoord xdim, ydim;
	OH_ASSERT(octBB(&facet, &bb));
	xdim = (bb.upperRight.x - bb.lowerLeft.x)/units_per_lambda; 
	ydim = (bb.upperRight.y - bb.lowerLeft.y)/units_per_lambda;
	(void) printf("Final bbox = %d by %d\n", xdim, ydim);
	(void) printf("Target Aspect Ratio = %3.2lf\n", TargetAspectRatio);
	(void) printf("Actual Aspect Ratio = %3.2lf\n", (double) xdim / ydim);
	if (net_length(&facet,&actual_length) ){
	/* Subtract the approximate vdd-gnd length */
	k = (power_position == BOTH) ? 2 : 1;
	actual_length -= 2*(num_rows *(xdim - units_per_lambda*(power_width-vlayer_pitch)) + k*ydim);
	(void) printf("Actual Net Length = %d\n", (int) actual_length/units_per_lambda);
	}
	(void) printf("Number of explicit feedthru cells added = %d\n", feed_count);
	(void) printf("Number of implicit feedthru positions available = %d\n", imp_feed_count);
	sugg_row_sep = rowSep * TargetAspectRatio * (double)ydim / (double)xdim;
	(void) printf("Suggested value for rowSep = %1.2lf\n",
	sugg_row_sep);
    }
    /*
     *  Delete the TimberWolf input files
     */
    if (delete_files) {
	if( debug == 1){
	/* save the .par .out .pin .pl1 file */
        (void) sprintf(command, 
	    "rm -f %s.{blk,cel,cf?,net,pl2,sav,sv2,twf} >& /dev/null", prefix);
	}
	else{
	(void) sprintf(command,
	"rm -f %s.{par,out,pl?,pin,blk,cel,cf?,net,sav,sv2,twf} >& /dev/null", prefix);
	}
	(void) util_csystem(command);
#ifdef TW5_0
	(void) sprintf(command, "rm -f rowlen.dat >& /dev/null");
	(void) util_csystem(command);
#endif
    }

    if ( do_global_route ) {
	discard_pin_table();
    }
    discard_block_table();

    /* Flush the facet so that the user gets the contents right
       even though the interface may not be created */
    if (octFlushFacet(&facet) < OCT_OK) {
	octError("flushing the contents facet");
	abortWolfe();
    }

    /*
     * XXX BUG: must close and reopen the interface to
     * get the new supply and ground terminals.
     * In the new version of OCT they should appear 
     * automatically
     */
    if (ohOpenInterface(&interface, &facet, "r") < OCT_OK) {
	(void) printf("Cannot open the interface\n");
	octError("");
	abortWolfe();
    }

    while (octGetById(&interface) == OCT_OK) {
	(void) octCloseFacet(&interface);
    }

    if (ohOpenInterface(&interface, &facet, "a") < OCT_OK) {
	(void) printf("Cannot open the interface\n");
	octError("");
	abortWolfe();
    }

    (void) vulcan(&facet, &interface);
    new_update_interface(&facet, &interface);

    if (octCloseFacet(&interface) < OCT_OK) {
	octError("closing the interface");
	abortWolfe();
    }

    if (octCloseFacet(&facet) < OCT_OK) {
	octError("closing the facet");
	abortWolfe();
    }

    octEnd();
    VOVend(0);
}



/*
 *  conventions:
 *
 *  Assume cell = foo/bar/xyz
 *  Assume view = logic
 *  Then cellroot = xyz
 *
 *  If running on the local machines, the TW files are in
 *  foo/bar/xyz/logic/xyz.cel
 *
 *  If running on a remote machine, the TW files are copied to
 *  /usr/tmp/wolfe/xyz.logic/xyz.cel
 *
 *  Note that the foo/bar part has been lost !  Hence, multiple runs of
 *  cells with the same cellname/viewname from different directories
 *  collide !
 */
static int run_wolfe(cell, view, prefix)
char *cell;
char *view;
char *prefix;
{
    char command[1024], cmd0[1024], cmd0a[1024], cmd1[1024], cmd2[1024];
    char *user, *mach, *cellroot, *mach_rcp, *mach_rsh;
    int status;

    cellroot = strrchr(cell, '/');
    cellroot = (cellroot == NIL(char)) ? cell : cellroot+1;

    if (machine == NIL(char)) {
#ifdef TW4_2
#ifndef TW5_0
	if (fallback){
		(void) sprintf(command, "TimberWolfSC %s/%s/%s", cell, view, cellroot);
	}else{
		(void) sprintf(command, "TimberWolfSC-4 %s/%s/%s", cell, view, cellroot);
	}
#else
	(void) sprintf(command, "TimberWolfSC-5 %s/%s/%s", cell, view, cellroot);
#endif
#endif
	(void) sprintf(command, "TimberWolfSC-4 %s", prefix);
	RUN(command, status);

    } else {
	/* Hack the format user@machine into format for rsh and rcp*/
	mach_rsh = machine;
	mach_rcp = machine;
	if (strchr(machine, '@') != NIL(char)) {
	    user = machine;
	    mach = strchr(machine, '@');
	    *mach++ = '\0';
	    (void) sprintf(command, "%s -l %s", mach, user);
	    mach_rsh = util_strsav(command);
	    (void) sprintf(command, "%s.%s", mach, user);
	    mach_rcp = util_strsav(command);
	}

	/* create remote directory (delete any existing with the same name) */
	(void) sprintf(cmd0, "mkdir /usr/tmp/wolfe");
	(void) sprintf(cmd0a, "chmod ugo+w /usr/tmp/wolfe");
	(void) sprintf(cmd1, "rm -rf /usr/tmp/wolfe/%s.%s", cellroot, view);
	(void) sprintf(cmd2, "mkdir /usr/tmp/wolfe/%s.%s", cellroot, view);
	(void) sprintf(command, "rsh %s -n >&/dev/null \"%s;%s;%s;%s\"",
			mach_rsh, cmd0, cmd0a, cmd1, cmd2);
	RUN(command, status);

	/* copy the files to this directory on the remote machine */
	(void) sprintf(command, 
	    "rcp %s/%s/%s.{blk,cel,par,net} %s:/usr/tmp/wolfe/%s.%s",
	    cell, view, cellroot, mach_rcp, cellroot, view);
	RUN(command, status);

	/* execute TimberWolf on the remote machine */
#ifdef TW4_2
#ifndef TW5_0
	(void) sprintf(command, 
	    "rsh %s -n TimberWolfSC-4 /usr/tmp/wolfe/%s.%s/%s",
	    mach_rsh, cellroot, view, cellroot); 
#else
	(void) sprintf(command, 
	    "rsh %s -n TimberWolfSC-5 /usr/tmp/wolfe/%s.%s/%s",
	    mach_rsh, cellroot, view, cellroot); 
#endif
#endif
	RUN(command, status);

	/* copy the files back to this machine */
	(void) sprintf(command, 
	    "rcp %s:/usr/tmp/wolfe/%s.%s/%s.\\{pl1,pin,out\\} %s/%s",
	    mach_rcp, cellroot, view, cellroot, cell, view);
	RUN(command, status);
    }
#ifdef TW4_2
    (void) printf("\n");
#endif
    return 0;
}


int net_length(facet, tot_length)
octObject *facet;
double *tot_length;
{
    double length, hypot();
    int32 num_points;
    struct octPoint *points;
    octGenerator gen, genNet;
    octObject net,path;
    octStatus status;
    octCoord dx, dy;


    /*
     * Compute the length of all paths attached to the net
     */
    *tot_length = 0.0;
    OH_ASSERT(octInitGenContents(facet, OCT_NET_MASK, &genNet));
    while ((status = octGenerate(&genNet, &net)) == OCT_OK) {
	OH_ASSERT(octInitGenContents(&net, OCT_PATH_MASK, &gen));
	while ((status = octGenerate(&gen, &path)) == OCT_OK) {
	    ohGetPoints(&path, &num_points, &points);
	    if (num_points != 2) return 0;	  /* Policy violation */

	    dx = points[0].x - points[1].x;
	    dy = points[0].y - points[1].y;
	    if (dx == 0) {
		length = dy < 0 ? -dy : dy;
	    } else if (dy == 0) {
		length = dx < 0 ? -dx : dx;
	    } else {
		length = hypot((double) dx, (double) dy);
	    }
	    *tot_length += length;
	}
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);


    return 1;
}
