/* sccsid = @(#)route.c	1.7 wolfe (ITD) 7/30/90   */
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"

extern int *timberwolf_density;		/* from parse.l */
int total_tw_density = 0, total_density = 0, total_tracks = 0;

static int grid_channel();
static void call_yacr();
static void set_offsets();
static void place_contact();
static void place_path();
static void trace_net();

extern octStatus get_connector();

#define DIR_LEFT	0
#define DIR_RIGHT	1
#define DIR_UP		2
#define DIR_DOWN	3

#define HORZ	0
#define VERT	1

/*
 *  x_center[j] and y_center[i] map an (i,j) location into octCoords
 */
static octCoord *x_center, *y_center;

/* 
 * x_offset and y_offset are the offsets so that the edges of the wires
 * are on a lambda grid 
 */
static octCoord x_offset, y_offset;

/*
 *  vlayer and hlayer are the layer objects within the current routing 
 *  channel; also, connectors is the CONNECTORS bag within the current
 *  routing channel
 */
static octObject vlayer, hlayer, connectors, instances;

/*
 *  top_terms[j], bottom_terms[j] give the actual terminal to which a
 *  path connects to on the top and bottom edges of the channel; ditto
 *  for left_terms[i] and right_terms[i]
 */
static octObject **top_terms, **bottom_terms, **left_terms, **right_terms;


/*
 *  The size of the routing grid
 */
static int ncols, nrows;

#define SLOP	0				/* was 2 */
#define SLOP2	1.5*hlayer_pitch*units_per_lambda + hlayer_width*units_per_lambda/2


int route_channels(facet)
octObject *facet;
{
    octCoord height, max_height, chan_height;
    int **horz, **vert;
    int i, j, *top, *bottom, *left, *right, *width;

    /* Loop once for each channel (there is one more channel than rows) */
    height = 0;
    set_offsets();
    for(i = 1; i <= num_rows + 1; i++) {

	if (do_global_route) {
		/* Apply a grid to the channel */
		grid_channel(i, vlayer_pitch*units_per_lambda, 
				&top, &bottom, &left, &right, &width, &ncols);

		/* Call yacr */
		call_yacr(&horz, &vert, top, bottom, left, right, i);
	}
	chan_height = (nrows + SLOP) * hlayer_pitch * units_per_lambda;

	/* Relocate the cells on the top edge (unless topmost channel) */
	if (i != num_rows + 1) {
	    relocate_block(block_list[i], height + chan_height, &max_height);
	}

	if (do_global_route) {
		/* Map the yacr output into the symbolic policy spec */
		if (ncols > 0) {
			route_nets(facet, i, horz, vert, width, height);
			for(j = 0; j <= nrows+1; j++) {
			FREE(horz[j]);
			FREE(vert[j]);
			}
			FREE(horz);
			FREE(vert);
		}
	}

	height += max_height + chan_height + (hlayer_width + 1) *units_per_lambda;

	if (do_global_route) {
		FREE(top); 
		FREE(bottom); 
		FREE(left); 
		FREE(right);
		FREE(width);
	}
    }

    if (debug > 0) {
	(void) printf("TOTALS: TW density=%d, YACR density=%d, YACR tracks=%d\n",
	    total_tw_density, total_density, total_tracks); 
    }
    return 0;
}

/*
 *  Remember the following have 'file scope':
 *     nrows, ncols, total_tw_density, total_density, total_tracks
 */

static void
call_yacr(horz, vert, top, bottom, left, right, chan)
int ***horz, ***vert;
int *top, *bottom, *left, *right;
int chan;
{
    int density, cycles, ncol;

    if (ncols > 0) {
	/* Route the channel with yacr */
	if (yacr(top, bottom, left, right, ncols, horz, vert, &nrows, &ncol, 
			&density, &cycles) != 0) {
	    (void) fprintf(stderr, "Problem with yacr\n");
	    exit(1);
	}

	if (ncols != ncol) {
	    (void) fprintf(stderr, "YACR added column, cannot complete routing\n");
	    (void) fprintf(stderr, "Re-run with '-v' flag for more detail\n");
	    exit(1);
	}
    } else {
	density = cycles = nrows = 0;
    }

    if (debug > 0) {
	(void) printf("Channel %d: TW density=%d, YACR density=%d, YACR tracks=%d cyclic?=%s\n",
	    chan, timberwolf_density[chan], density, nrows, 
	    cycles != 0 ? "yes" : "no");
    }
    total_tw_density += timberwolf_density[chan];
    total_density += density;
    total_tracks += nrows;
}



relocate_block(list, height, max_height)
block_record_t *list;
octCoord height;
octCoord *max_height;
{
    octObject cell;
    struct octBox bb;
    block_record_t *p;

    /* Relocate the cells, and find height of largest cell in the block */
    *max_height = 0;
    for(p = list; p != NIL(block_record_t); p = p->next) {

	/* relocate this standard cell */
	OH_ASSERT(ohGetById(&cell, p->objectId));
	cell.contents.instance.transform.translation.y += height;
	OH_ASSERT(octModify(&cell));

	/* Keep running track of the highest cell seen so far */
	OH_ASSERT(octTB(&cell, &bb));
	*max_height = MAX(*max_height, bb.upperRight.y - bb.lowerLeft.y);
    }
}

/*
 *  route_nets: map the yacr output (horz, vert) into symbolic layout
 *
 *  Global variables:
 *	pin_t channel[channel #][location] is a linked list of pins for
 *	    each channel
 *
 *  We are quite gross, and make use of several variables which are
 *  global to this module.
 *
 *	vlayer, hlayer are Oct objects for the layers in the current
 *	    facet (exported to place_path)
 *	connectors is the CONNECTORS bag for the current facet (exported
 *	    to place_contact)
 *	top_terms, bottom_terms, left_terms, right_terms are arrays of
 *	    Oct objects which are the terminals on each side (exported to
 *	    trace_net)
 *	x_center, y_center map the grid coordinates to Oct coordinates 
 *	    (exported to place_pin, place_path, place_contact)
 *	nrows, ncols identify number of rows and columns
 *
 *  The strategy is as follows:
 *	1) The formal terminals along each channel edge are created first;
 *	The actual terminal which is attached to the formal terminal is
 *	saved in the array "top_terms" (for example) for later reference.
 *
 *	2) Walk along the top and bottom edges calling trace_net
 */
 
route_nets(facet, chan_num, horz, vert, width, height)
octObject *facet;
int chan_num;
int **horz;
int **vert;
int *width;
octCoord height;
{
    pin_t *p;
    int i, j, k, power_conn_avail;
    octCoord p_width, v_sep, temp;
    struct octPoint *t;
    octStatus status;
    octObject inst, power_conn;

    /* Create the layers */
    OH_ASSERT(ohGetOrCreateLayer(facet, &vlayer, vlayer_name));
    OH_ASSERT(ohGetOrCreateLayer(facet, &hlayer, hlayer_name));

    /* HACK: get the connector for the power terminals */
    status  = get_connector(vlayer, hlayer, hlayer_width*units_per_lambda, power_width*units_per_lambda, &power_conn);
    power_conn_avail = (status == OCT_OK? 1 : 0);

    /* Get the connectors bag for the routing instances */
    OH_ASSERT(ohGetOrCreateBag(facet, &connectors, "CONNECTORS"));
    OH_ASSERT(ohGetOrCreateBag(facet, &instances, "INSTANCES"));

    /* create array of centers for each column
     *
     *  the left pins (i.e., x_center[0]) are at zero because we relocated
     *  all cells by min_xposition;
     *  the right pins (i.e., x_center[ncols+1]) are at max_xposition
     *  (to which a full column has been added already)
     */
    if( power_position == BOTH) k=2;
    else k=1;

    p_width = power_width * units_per_lambda;
    v_sep = vlayer_pitch * units_per_lambda;
    x_center = ALLOC(octCoord, ncols+2);
    x_center[0] = (extend_terms ? 0-k*(p_width + v_sep) : 0-(k*p_width +(k-1)*v_sep - hlayer_width * (units_per_lambda / 2)));

    /* Make sure the left edge of the contact lines up with a lambda grid */
    temp = width[0] / 2 - (hv_contact_width * units_per_lambda) / 2 ;
    if ( temp % units_per_lambda == 0){
	x_center[1] = width[0] / 2;
    } else{
	x_center[1] = (octCoord)(floor((double)(temp / units_per_lambda))) + hv_contact_width * units_per_lambda / 2;
    }
    for(j = 2; j <= ncols; j++) {
	x_center[j] = x_center[j-1] + (width[j-2]+1)/2 + width[j-1]/2;
    }
    x_center[ncols+1] = (extend_terms ? max_xposition + k *(p_width +v_sep) : max_xposition+k * p_width +(k-1)*v_sep - hlayer_width * (units_per_lambda / 2));


    /* create array of centers for each row; remember that row 0 corresponds
     * to the center of a top terminal, and row nrows+1 corresponds to the
     * center of a bottom terminal
     */
    y_center = ALLOC(octCoord, nrows+2);
    temp = height + (hlayer_pitch*units_per_lambda / 2) - SLOP2;
    if ( ((temp - (hv_contact_height *units_per_lambda)/2) % units_per_lambda) == 0){
    	y_center[nrows+1] = temp;
    } else{
	y_center[nrows+1] = (octCoord) ceil((double)((temp) + (units_per_lambda / 2)));
    }
    for(j = nrows; j >= 0; j--) {
	y_center[j] = y_center[j+1] + hlayer_pitch*units_per_lambda;
    }


    /*
     *  Step 1: create all of the pins and connect them to the cells
     */
    top_terms = ALLOC(octObject *, ncols + 2);
    for(j = 0; j <= ncols+1; j++) {
	top_terms[j] = NIL(octObject);
    }
    for(p = channel[chan_num][TOP]; p != NIL(pin_t); p = p->next) {
	if ((j = p->column) == -1) continue;	/* pin was dropped */
	top_terms[j+1] = ALLOC(octObject, 1);

	/* Need to handle TOP of the highest channel separately */
	if (chan_num == num_rows+1) {
	    OH_ASSERT(ohGetById(top_terms[j+1], p->termId));

	    /* Move the formal terminal impl. to its proper place */
	    OH_ASSERT(ohGetById(&inst, p->instId));
	    t = &inst.contents.instance.transform.translation;
	    if ( strcmp(inst.contents.instance.master,hv_contact_name) == 0){
	    	t->x = x_center[j+1];
	    	t->y = y_center[0];
	    } else{
		t->x = x_center[j+1] - x_offset;
	    	t->y = y_center[0] - y_offset;
	    }
	    OH_ASSERT(octModify(&inst));
	} else {
		OH_ASSERT(ohGetById(top_terms[j+1],p->termId));
	}
    }


    bottom_terms = ALLOC(octObject *, ncols + 2);
    for(j = 0; j <= ncols+1; j++) {
	bottom_terms[j] = NIL(octObject);
    }
    for(p = channel[chan_num][BOTTOM]; p != NIL(pin_t); p = p->next) {
	if ((j = p->column) == -1) continue;	/* pin was dropped */
	bottom_terms[j+1] = ALLOC(octObject, 1);

	/* Need to handle BOTTOM of lowest channel separately */
	if (chan_num == 1) {
	    OH_ASSERT(ohGetById(bottom_terms[j+1], p->termId));

	    /* Move the formal terminal impl. to its proper place */
	    OH_ASSERT(ohGetById(&inst, p->instId));
	    t = &inst.contents.instance.transform.translation;
	    if ( strcmp(inst.contents.instance.master,hv_contact_name) == 0){
	    	t->x = x_center[j+1];
	    	t->y = y_center[nrows+1];
	    } else{
		t->x = x_center[j+1] - x_offset;
	    	t->y = y_center[nrows+1] - y_offset;
	    }
	    OH_ASSERT(octModify(&inst));
	} else {
		OH_ASSERT(ohGetById(bottom_terms[j+1],p->termId));
	}
    }

    left_terms = ALLOC(octObject *, nrows + 2);
    for(i = 0; i <= nrows+1; i++) {
	left_terms[i] = NIL(octObject);
    }
    for(p = channel[chan_num][LEFT]; p != NIL(pin_t); p = p->next) {
	/* ouch! find this group on left edge */ 
	for(i = 1; i <= nrows; i++) {
	    if (horz[i][0] == p->group) {
		left_terms[i] = ALLOC(octObject, 1);
		OH_ASSERT(ohGetById(left_terms[i], p->termId));

		/* Move the formal terminal impl. to its proper place */
		OH_ASSERT(ohGetById(&inst, p->instId));
		t = &inst.contents.instance.transform.translation;
		/*
		 * NOTE: If the connector corresponds to the connector
		 * required for power nets connecterd to signal terminals
		 * we want it flush with the power/ground rail
		 */
		if(power_conn_avail && !strcmp(inst.contents.instance.master, 
			power_conn.contents.instance.master)){
		    /* Align connector with the power strip on left */
		    t->x = 0 - p_width/2;
		    t->y = y_center[i];
		} else {
		    t->x = x_center[0] ;
		    if (!strcmp(inst.contents.instance.master,hv_contact_name)){
			    t->y = y_center[i];
		    } else{
			    t->y = y_center[i] - y_offset;
		    }
		}
		OH_ASSERT(octModify(&inst));
		break;
	    }
	}
    }

    right_terms = ALLOC(octObject *, nrows + 2);
    for(i = 0; i <= nrows+1; i++) {
	right_terms[i] = NIL(octObject);
    }
    for(p = channel[chan_num][RIGHT]; p != NIL(pin_t); p = p->next) {
	/* ouch! find this group on left edge */ 
	for(i = 1; i <= nrows; i++) {
	    if (horz[i][ncols+1] == p->group) {
		right_terms[i] = ALLOC(octObject, 1);
		OH_ASSERT(ohGetById(right_terms[i], p->termId));

		/* Move the formal terminal impl. to its proper place */
		OH_ASSERT(ohGetById(&inst, p->instId));
		t = &inst.contents.instance.transform.translation;
		if(power_conn_avail && !strcmp(inst.contents.instance.master, 
			power_conn.contents.instance.master)){
		    t->x = max_xposition + p_width/2;
		    t->y = y_center[i];
		} else {
		    t->x = x_center[ncols+1] ;
		    if (!strcmp(inst.contents.instance.master,hv_contact_name)){
			    t->y = y_center[i];
		    } else{
			    t->y = y_center[i] - y_offset;
		    }
		}
		OH_ASSERT(octModify(&inst));
		break;
	    }
	}
    }


    /*
     *  Step 2: chase the nets
     */
    chase_nets(facet, horz, vert, chan_num, TOP);
    chase_nets(facet, horz, vert, chan_num, BOTTOM);

    FREE(x_center);
    FREE(y_center);
    for(i = 0; i <= ncols+1; i++) {
	if (top_terms[i] != NIL(octObject)) FREE(top_terms[i]);
    }
    for(i = 0; i <= ncols+1; i++) {
       if (bottom_terms[i] != NIL(octObject)) FREE(bottom_terms[i]);
    }
    for(i = 0; i <= nrows+1; i++) {
       if (left_terms[i] != NIL(octObject)) FREE(left_terms[i]);
    }
    for(i = 0; i <= nrows+1; i++) {
       if (right_terms[i] != NIL(octObject)) FREE(right_terms[i]);
    }
    FREE(top_terms); 
    FREE(bottom_terms); 
    FREE(left_terms); 
    FREE(right_terms);
}

/*
 *  chase_nets: setup calls for trace_net for each pin
 */

chase_nets(facet, horz, vert, chan_num, chan_loc)
octObject *facet;
int **horz;
int **vert;
int chan_num;
int chan_loc;
{
    octObject net, *array;
    pin_t *p;
    int j, start_col, start_row, current_row, direction;

    start_row = (chan_loc == TOP) ? 0 : nrows+1;
    current_row = (chan_loc == TOP) ? 1 : nrows;
    direction = (chan_loc == TOP) ? DIR_DOWN : DIR_UP;

    /* walk across TOP chasing nets */
    for(p = channel[chan_num][chan_loc]; p != NIL(pin_t); p = p->next) {

	/* See if the pin was dropped during gridding (sigh) */
	if ((j = p->column) == -1) continue;
	start_col = j+1;
	array = (chan_loc == TOP)?top_terms[start_col]:bottom_terms[start_col];

	/* See if the pin has already been visited by trace_net */
	if (vert[start_row][start_col] != p->group) continue;

	/* Find the net */
	OH_ASSERT(ohGetById(&net, p->netId));

	/* Trace the net */
	vert[start_row][start_col] = 0;
	trace_net(facet, vert, horz, VERT, p->group, start_row, start_col,
		    current_row, start_col, direction, &net, array);
    }
}

static void
trace_net(facet, layer, other_layer, layer_type, net_num, 
			si, sj, i, j, dir, net, endpoint)
octObject *facet;
int **layer;
int **other_layer;
int layer_type;			/* layer type (HORZ or VERT) */
int net_num;
int si, sj;			/* where current segment started */
int i, j;			/* current position */
int dir;
octObject *net;			/* the oct net we are tracing */
octObject *endpoint;		/* the object at the other end of the path */
{
    octObject term;
    int other_layer_net, ni, nj, need_contact, need_same_layer_contact;
    char *contact_name, *contact_view;

    layer[i][j] = 0;		/* destructive walk */

    /* Check for reached top pin */
    if (i == 0) {
	if (top_terms[j] == NIL(octObject)) {
/*	    (void) fprintf(stderr, "Huh ?  Me thinks yacr screwed up\n");*/
	} else {
	    place_path(si, sj, i, j, layer_type, net, endpoint, top_terms[j]);
	}
	return;
    /* Check for reached bottom pin */
    } else if (i == nrows+1) {
	if (bottom_terms[j] == NIL(octObject)) {
/*	    (void) fprintf(stderr, "Huh ?  Me thinks yacr screwed up\n");*/
	} else {
	    place_path(si, sj, i, j, layer_type, net, endpoint, bottom_terms[j]);
	}
	return;
    /* Check for reached side pin */
    } else if (j == 0) {
	if (left_terms[i] == NIL(octObject)) {
/*	    (void) fprintf(stderr, "Huh ?  Me thinks yacr screwed up\n");*/
	} else {
	    place_path(si, sj, i, j, layer_type, net, endpoint, left_terms[i]);
	}
	return;
    } else if (j == ncols+1) {
	if (right_terms[i] == NIL(octObject)) {
/*	    (void) fprintf(stderr, "Huh ?  Me thinks yacr screwed up\n");*/
	} else {
	    place_path(si, sj, i, j, layer_type, net, endpoint, right_terms[i]);
	}
	return;
    }

    need_contact = (other_layer[i][j] == net_num);
    need_same_layer_contact = 
      ((dir != DIR_LEFT) && layer[i][j-1] == net_num) ||
      ((dir != DIR_RIGHT) && layer[i][j+1] == net_num) ||
      ((dir != DIR_UP) && layer[i-1][j] == net_num) ||
      ((dir != DIR_DOWN) && layer[i+1][j] == net_num);

    /* See if net changes layer */
    if (need_contact) {
	place_contact(facet, hv_contact_name, hv_contact_view, i, j, &term);
	place_path(si, sj, i, j, layer_type, net, endpoint, &term);
	si = i;
	sj = j;
	endpoint = &term;
    /* See if net changes direction on the current layer */
    } else if (need_same_layer_contact) {
	contact_name = (layer_type == HORZ) ? h_contact_name : v_contact_name;
	contact_view = (layer_type == HORZ) ? h_contact_view : v_contact_view;
	place_contact(facet, contact_name, contact_view, i, j, &term);
	place_path(si, sj, i, j, layer_type, net, endpoint, &term);
	si = i;
	sj = j;
	endpoint = &term;
    }


    /* Check in direction we are moving */
    ni = i; 
    nj = j;
    switch (dir) {
	case DIR_LEFT: nj--; break; 
	case DIR_RIGHT: nj++; break;
	case DIR_UP: ni--; break;
	case DIR_DOWN: ni++; break;
    }
    if (layer[ni][nj] == net_num) {
	trace_net(facet, layer, other_layer, layer_type, net_num, 
	    /*start*/si,sj, /*current*/ni,nj, dir, net, endpoint);
    }


    /* Check other layer */
    other_layer_net = other_layer[i][j];
    if (other_layer_net == net_num && other_layer[i][j-1] == net_num) {
	other_layer[i][j] = 0;
	trace_net(facet, other_layer, layer, 1 - layer_type, net_num, 
		    /*start*/ i, j, /*current*/ i, j-1, 
		    /*direction*/ DIR_LEFT, net, &term);
    }
    if (other_layer_net == net_num && other_layer[i][j+1] == net_num) {
	other_layer[i][j] = 0;
	trace_net(facet, other_layer, layer, 1 - layer_type, net_num, 
		    /*start*/ i, j, /*current*/ i, j+1, 
		    /*direction*/ DIR_RIGHT, net, &term);
    }
    if (other_layer_net == net_num && other_layer[i-1][j] == net_num) {
	other_layer[i][j] = 0;
	trace_net(facet, other_layer, layer, 1 - layer_type, net_num, 
		    /*start*/ i, j, /*current*/ i-1, j, 
		    /*direction*/ DIR_UP, net, &term);
    }
    if (other_layer_net == net_num && other_layer[i+1][j] == net_num) {
	other_layer[i][j] = 0;
	trace_net(facet, other_layer, layer, 1 - layer_type, net_num, 
		    /*start*/ i, j, /*current*/ i+1, j, 
		    /*direction*/ DIR_DOWN, net, &term);
    }


    /* Check this layer */
    if (dir != DIR_LEFT && layer[i][j-1] == net_num) {
	trace_net(facet, layer, other_layer, layer_type, net_num, 
	    /*start*/i,j, /*current*/i,j-1, DIR_LEFT, net, &term);
    }
    if (dir != DIR_RIGHT && layer[i][j+1] == net_num) {
	trace_net(facet, layer, other_layer, layer_type, net_num, 
	    /*start*/i,j, /*current*/i,j+1, DIR_RIGHT, net, &term);
    }
    if (dir != DIR_UP && layer[i-1][j] == net_num) {
	trace_net(facet, layer, other_layer, layer_type, net_num, 
	    /*start*/i,j, /*current*/i-1,j, DIR_UP, net, &term);
    }
    if (dir != DIR_DOWN && layer[i+1][j] == net_num) {
	trace_net(facet, layer, other_layer, layer_type, net_num, 
	    /*start*/i,j, /*current*/i+1,j, DIR_DOWN, net, &term);
    }
}

/*
 *  place_path: create a path from grid coordinates (si, sj) to (i,j)
 *  including all attachments needed by the OCT symbolic policy.
 *
 *  This includes:
 *	path attached to a layer
 *	path attached to the net it implements
 *	path contains the terminals at its endpoints
 *
 *  Be careful and check the following ----
 *	Endpoint on left or right side is connected to power rail...
 *	The vertical connector is to a feedthru terminal
 *	The path is manhatten and the edges of path are on a lambda grid
 */

static void
place_path(si, sj, i, j, layer, net, endpoint1, endpoint2)
int si, sj;			/* IN: starting point for the path */
int i, j;			/* IN: ending point for the path */
int layer;			/* IN: layer (HORZ or VERT) */
octObject *net;			/* IN: net which path is implementing */
octObject *endpoint1;		/* IN: object at first end of the path */
octObject *endpoint2;		/* IN: object at other end of the path */
{
    octObject path, feedinst;
    struct octBox bbox1, bbox2;
    int is_1_feedthru, is_2_feedthru;
    struct octPoint points[2];
    octCoord x1, x2;
    int MAGIC = -12345;

    /* create a path and attach it to the net */
    path.type = OCT_PATH;
    path.contents.path.width = (layer == HORZ) ? hlayer_width : vlayer_width;
    path.contents.path.width *= units_per_lambda;
    OH_ASSERT(octCreate(net, &path));

    /* place the points for the path from (si,sj) to (i,j) */
    points[0].x = x_center[sj] - x_offset;
    points[0].y = y_center[si] - y_offset;
    points[1].x = x_center[j] - x_offset;
    points[1].y = y_center[i] - y_offset;

    /* Now check to see if any of these need to be modified */
    OH_ASSERT(ohTerminalBB(endpoint1,&bbox1));
    OH_ASSERT(ohTerminalBB(endpoint2,&bbox2));

    if (layer == HORZ){
	/*
	 * Checks for power connections to the sides  -----
	 */
	if (si == i){ /* Horizontal segment */
	    x1 = (bbox1.upperRight.x + bbox1.lowerLeft.x)/2;
	    x2 = (bbox2.upperRight.x + bbox2.lowerLeft.x)/2;
	    points[0].x = x1;
	    points[1].x = x2;
	}
    } else {
	is_1_feedthru = is_2_feedthru = 0;
	if((bbox1.lowerLeft.y < MAGIC) && (bbox1.upperRight.y > MAGIC)){
	    if( si == nrows+1)
	       points[0].y = y_center[nrows+1]; 
	    if ( si == 0)    points[0].y = y_center[0];
	    /*
	     * Translate the connector instance containing the 
	     * feedthru terminal 
	     */
	    OH_ASSERT(octGenFirstContainer(endpoint1,OCT_INSTANCE_MASK,&feedinst));
	    feedinst.contents.instance.transform.translation.y = points[0].y;
	    feedinst.contents.instance.transform.translation.x -= x_offset;
	    is_1_feedthru = 1;
	    OH_ASSERT(octModify(&feedinst));
	    OH_ASSERT(ohTerminalBB(endpoint1,&bbox1));
	}
	if((bbox2.lowerLeft.y < MAGIC) && (bbox2.upperRight.y > MAGIC)){
	    if( i == nrows+1)
	       points[1].y = y_center[nrows+1]; 
	    if ( i == 0)  points[1].y = y_center[0];
	    /*
	     * Translate the instance containing the 
	     * feedthru terminal 
	     */
	    OH_ASSERT(octGenFirstContainer(endpoint2,OCT_INSTANCE_MASK,&feedinst));
	    feedinst.contents.instance.transform.translation.y = points[1].y;
	    feedinst.contents.instance.transform.translation.x -= x_offset;
	    is_2_feedthru = 1;
	    OH_ASSERT(octModify(&feedinst));
	    OH_ASSERT(ohTerminalBB(endpoint2,&bbox2));
	}

	/*
	 * Now check to see if one of the pins lies outside the channel ---
	 * Since we add feedthrus bottom up we need worry about the path
	 * to the feedthru only when bbox is on the top
	 */
	if( (y_center[0] < bbox1.lowerLeft.y) ||
		(y_center[nrows+1] > bbox1.upperRight.y) ){
	    points[0].x = (bbox1.lowerLeft.x + bbox1.upperRight.x) /2;
	    if (is_1_feedthru == 0 && (bbox1.upperRight.x-bbox1.lowerLeft.x != path.contents.path.width)){
		points[0].x -= x_offset;
	    }
	    points[0].y = (bbox1.lowerLeft.y + bbox1.upperRight.y) /2;
	}
	if( (y_center[0] < bbox2.lowerLeft.y) ||
		(y_center[nrows+1] > bbox2.upperRight.y) ){
	    points[1].x = (bbox2.lowerLeft.x + bbox2.upperRight.x) /2;
	    if (is_2_feedthru == 0 && bbox2.upperRight.x-bbox2.lowerLeft.x != path.contents.path.width){
		points[1].x -= x_offset;
	    }
	    points[1].y = (bbox2.lowerLeft.y + bbox2.upperRight.y) /2;
	}
    }
    OH_ASSERT(octPutPoints(&path, 2, points));

    /* attach the endpoints to the path segment */
    OH_ASSERT(octAttach(&path, endpoint1));
    OH_ASSERT(octAttach(&path, endpoint2));

    /* Attach the path to a layer */
    OH_ASSERT(octAttach(layer == HORZ ? &hlayer : &vlayer, &path));
}

/*
 *  place_contact: place a routing contact at grid position (i, j)
 *
 *  Return the (single !) actual terminal attached to the contact instance
 */

static void
place_contact(facet, contact_name, contact_view, i, j, term)
octObject *facet;		/* IN: facet to attach things to */
char *contact_name;		/* IN: cell name of the contact */
char *contact_view;		/* IN: cell view of the contact */
int i, j;			/* IN: grid coordinates for the contact */
octObject *term;		/* OUT: actual terminal of contact instance */
{
    octObject contact;
    char name[20];
    octCoord x_position, y_position;
    static int contact_num = 0;

    (void) sprintf(name, "c%d", contact_num++);

    /* If it is a routing instance - ie. same layers then it
    need to be relocated since the wires are aligned to the edge */
    if ( strcmp(contact_name,hv_contact_name) != 0){
	x_position = x_center[j] - x_offset;
	y_position = y_center[i] - y_offset;
    } else{
	x_position = x_center[j];
	y_position = y_center[i];
    }

    OH_ASSERT(ohCreateInstance(facet, &contact, name, contact_name, contact_view,
		x_position, y_position, OCT_NO_TRANSFORM));
    OH_ASSERT(octAttach(&connectors, &contact));

    OH_ASSERT_DESCR(octGenFirstContent(&contact, OCT_TERM_MASK, term),
	"attempting to find a terminal on a routing contact");
}

/*
 *  Don't even try to understand it, its black magic ...
 */
static int 
grid_channel(i, min_grid_spacing, top_p, bottom_p, 
			left_p, right_p, width_p, ncol)
int i;			/* IN: channel number */
int min_grid_spacing;	/* IN: minimum vertical grid spacing */
int **top_p;		/* OUT: array of top pins */
int **bottom_p;		/* OUT: array of bottom pins */
int **left_p;		/* OUT: array of left pins */
int **right_p;		/* OUT: array of right pins */
int **width_p;		/* OUT: width of each column */
int *ncol;		/* OUT: total number of columns */
{
    pin_t *ptop, *pbottom, *next_pin, *next_opp_pin;
    pin_t infinity, *p;
    int ngrid, curpos, cnt;
    int *top, *bottom, *width;
    octObject inst, term;

    /* Create the left pin list */
    cnt = 0;
    for(p = channel[i][LEFT]; p != NIL(pin_t); p = p->next) 
	cnt++;
    *left_p = ALLOC(int, cnt+1);
    (*left_p)[0] = cnt;
    cnt = 1;
    for(p = channel[i][LEFT]; p != NIL(pin_t); p = p->next) 
	 (*left_p)[cnt++] = p->group;

    /* Create the right pin list */
    cnt = 0;
    for(p = channel[i][RIGHT]; p != NIL(pin_t); p = p->next) 
	cnt++;
    *right_p = ALLOC(int, cnt+1);
    (*right_p)[0] = cnt;
    cnt = 1;
    for(p = channel[i][RIGHT]; p != NIL(pin_t); p = p->next) 
	 (*right_p)[cnt++] = p->group;

    ptop = channel[i][TOP];
    pbottom = channel[i][BOTTOM];
    infinity.position = 999999;
    infinity.next = (pin_t *) NULL;
    top = ALLOC(int, 4000);
    bottom = ALLOC(int, 4000);
    width = ALLOC(int, 4000);
    ngrid = 0;
    curpos = 0;

    next_pin = pbottom;
loop:
    /* 
     *  next_pin points to next pin (either top or bottom)
     *  next_opo_pin points to next next pin on the opposite side
     */
    if (ptop == NIL(pin_t)) {
      if (pbottom == NIL(pin_t)) {
        *top_p = top;
        *bottom_p = bottom;
        *width_p = width;
        *ncol = ngrid;
        return;
      } else {
        next_pin = pbottom;
        next_opp_pin = &infinity;
      }
    } else {
      if (pbottom == NIL(pin_t)) {
        next_pin = ptop;
        next_opp_pin = &infinity;
      } else {
        /* Find which pin comes first */
        if (ptop->position <= pbottom->position) {
          next_pin = ptop;
          next_opp_pin = pbottom;
        } else {
          next_pin = pbottom;
          next_opp_pin = ptop;
        }
      }
    }
      OH_ASSERT(ohGetById(&inst, next_pin->instId));
      OH_ASSERT(ohGetById(&term, next_pin->termId));
    if (curpos > next_pin->position) {
      OH_ASSERT(ohGetById(&inst, next_pin->instId));
      OH_ASSERT(ohGetById(&term, next_pin->termId));
      (void) fprintf(stderr, 
         "ERROR: Cannot completely grid channel #%d, pin %s of %s dropped\n",
          i, term.contents.term.name, inst.contents.instance.name);
      (void) fprintf(stderr, "\t Route this pin manually\n");
      next_pin->column = -1;
      if (next_pin == ptop) {
        ptop = ptop->next;
      } else {
        pbottom = pbottom->next;
      }
      goto loop;
    }


    /* Delineate grid lines until reaching this pin */
    for(; curpos <= next_pin->position; curpos += min_grid_spacing) {
	top[ngrid] = 0;
	bottom[ngrid] = 0;
	width[ngrid] = min_grid_spacing;
	ngrid++;
    }

    /* Place the next pin in the last grid created above */
    if (next_pin == ptop) {
	top[ngrid-1] = ptop->group;
    } else {
	bottom[ngrid-1] = pbottom->group;
    }
    next_pin->column = ngrid-1;

    /* See if the last grid line was wide enough */
    if (curpos < next_pin->position + next_pin->width) {
	width[ngrid-1] += (next_pin->position + next_pin->width) - curpos;
	curpos = next_pin->position + next_pin->width;
    }
    if (next_pin == ptop) {
	ptop = ptop->next;
    } else {
	pbottom = pbottom->next;
    }

    /* See if next pin from opposite side fits in here */
    if (next_opp_pin->position < curpos) {
	if (next_opp_pin == ptop) {
	    top[ngrid-1] = ptop->group;
	} else {
	    bottom[ngrid-1] = pbottom->group;
	}
	next_opp_pin->column = ngrid-1;
	if (curpos < next_opp_pin->position + next_opp_pin->width) {
	    width[ngrid-1] += (next_opp_pin->position + next_opp_pin->width)
				- curpos;
	    curpos = next_opp_pin->position + next_opp_pin->width;
	}
	if (next_opp_pin == ptop) {
	    ptop = ptop->next;
	} else {
	    pbottom = pbottom->next;
	}
    }

    goto loop;
}

/* Set the offsets for the wires */
static void
set_offsets()
{
    if ( ((hv_contact_width - hlayer_width) % 2) == 0){
	x_offset = 0;
	y_offset = 0;
    } else{
	x_offset = (hv_contact_width - vlayer_width) * units_per_lambda / 2;
	y_offset = (hv_contact_height - hlayer_width) * units_per_lambda / 2;
    }
}
