/* @(#)tb.c	1.4 wolfe (ITD) 3/21/90   */
/*   
 *       octTB 
 *       Author : Barry Boes
 *       Date   : 3-14-89
 *       Background :  The previous versions of Wolfe used the bounding
 *       box of a cell for the cells placement.  But, placing the cell
 *      by its bounding box is incorrect, since is causes the power
 *      and ground rails not to abut properly.  So, this problem was
 *      previously solved by clipping the overhanging n and p-well
 *      from the edges of the cells, causing the parts to be placed
 *      correctly, but forever causing the cells to lose this 
 *      overhanging well, which is incorrect.  So, this routine is an
 *      addition used to cause placement of the parts by the layer
 *      OVERLAP defined for the parts.
 */

#define OVERLAP_IS_USED 2
#define OVERLAP_NOT_USED 1


#include "oct.h" 
#include <stdio.h>

int octTB(object, tb)
octObject *object;
struct octBox *tb;
{
  int retval;
  octObject *tobject, layer, master, box, ov_layer, tmpobj;
  struct octBox tmpBB1, tmpBB2;
  int tempOverlap;

  retval = OCT_OK;
  
  if (object->type == OCT_INSTANCE)
  {
    master.type = OCT_FACET;
    master.contents.facet.mode = "r";
    master.contents.facet.facet = "contents";
    master.contents.facet.cell = object->contents.instance.master;
    master.contents.facet.view = object->contents.instance.view;
    master.contents.facet.version = OCT_CURRENT_VERSION;
    VOVinputFacet( &master );
    if (octOpenFacet (&master) < OCT_OK)
    {
      fprintf (stderr, "wolfe ERROR: could not open master of instance %s:%s\n",
	      object->contents.instance.name,
	      object->contents.instance.view);
      return (-1);
    }
    ov_layer.type = OCT_LAYER;
    ov_layer.contents.layer.name = "OVERLAP";
    if (octGetByName (&master, &ov_layer) != OCT_OK)
     if (octBB (object, tb) != OCT_OK)
     {
        fprintf (stderr, "wolfe ERROR: could not read either the OVERLAP layer\n");
	fprintf (stderr, "or the bounding box for instance %s:%s:%s\n",
	        master.contents.facet.cell,
	        master.contents.facet.view,
	        master.contents.facet.facet);
      return (-1);
      }
      else
	return (OCT_OK);
  }
  else
  {
    if (object->type != OCT_FACET)
    {
      fprintf (stderr, "wolfe ERROR: can only get the OVERLAP box for a\n");
      fprintf (stderr, "        facet or instance.\n");
      return (-1);
    }
    /*
     *  Locate the layer OVERLAP
     */
    ov_layer.contents.layer.name = "OVERLAP";
    ov_layer.type = OCT_LAYER;
    if ((octGetByName (object, &ov_layer)) != OCT_OK)
     if (octBB (object, tb) != OCT_OK)
     {
        fprintf (stderr, "error: could not get either the OVERLAP layer\n");
	fprintf (stderr, "or the bounding box for cell %s:%s:%s\n",
	        object->contents.facet.cell,
	        object->contents.facet.view,
	        object->contents.facet.facet);
        return (-1);
     }
     else
       return (OCT_OK);
  }
  if (octGenFirstContent (&ov_layer, OCT_GEO_MASK, &box) != OCT_OK)
  {
    fprintf (stderr, "wolfe : Error obtaining a geometry on the OVERLAP layer\n");
    if (object->type == OCT_FACET)
      fprintf (stderr, "         of the cell %s:%s.\n",
             object->contents.facet.cell,
             object->contents.facet.view);
    else
      fprintf (stderr, "         of the cell %s:%s.\n",
             object->contents.instance.master,
             object->contents.instance.view);
    fprintf (stderr, "         If your cells don't use the OVERLAP layer for\n");
    fprintf (stderr, "         placement, please delete the OVERLAP layer from\n");
    fprintf (stderr, "         your cells.  If your cells do use the OVERLAP layer,\n");
    fprintf (stderr, "         please insure that they have been installed properly.\n");
    fprintf (stderr, "         For more information, see wolfe (5).\n");
    return (-1);
  }
  if (box.type == OCT_POLYGON || box.type == OCT_BOX)
  {
    tmpobj.type = OCT_BOX;
    if (octBB (&box, &tmpobj.contents.box) != OCT_OK)
    {
      fprintf (stderr, "wolfe Error: could not generate OVERLAP box on geometry.\n");
      return (-1);
    }
    if (object->type == OCT_INSTANCE)
    {
      octTransformGeo (&tmpobj, &object->contents.instance.transform);
    }
    tb->lowerLeft.x = tmpobj.contents.box.lowerLeft.x;
    tb->lowerLeft.y = tmpobj.contents.box.lowerLeft.y;
    tb->upperRight.x = tmpobj.contents.box.upperRight.x;
    tb->upperRight.y = tmpobj.contents.box.upperRight.y;
  }
  else
  {
    fprintf (stderr, "wolfe Error: could not generate any geometries on the OVERLAP layer.\n");
    return (-1);
  }
  return (OCT_OK);
}
