/*
 *  Wolfe, Version 1.1: Interface Between Oct and TimberWolf/Yacr for
 *  macro-cell generation using a 'standard-cell' approach
 *
 *  Author: Richard Rudell
 *	    205 Cory Hall
 *	    Univ. of California
 *	    Berkeley, CA 94720
 *
 *  Revision History:
 *	Version 1.0: Sept. 1986
 *	    - First cut, never released
 *
 *	Version 1.1: Feb. 1987
 *	    - Formal terminal handling 'cleaned-up'
 *		If a formal terminal has a connector attached to it, this
 *		connector is used as the implementation; otherwise, a 
 *		default connector (usually M1-M2 contact) is created.
 *
 *	    - Improvements to TimberWolfSC allow pins on all sides
 *
 *	Version 1.2: Sep. 1987
 *	    - ifdef's TW3_3 and TW4_2 to specify timberwolf versions
 *		differences:
 *		    TW4_2 has no att.per.cell parameter in .par file;
 *		    added a -f for 'fast' mode and replaced -a and -s
 */

/*
 *  Channel edges: top, bottom, left and right 
 */
#define TOP	0
#define BOTTOM	1
#define LEFT	2
#define RIGHT	3
#define BOTH	4
extern char *print_location[5];		/* maps integer to a name */


/*
 *  The pin struct records information for each pin on each edge of 
 *  each channel.  This is created from the TimberWolf '.pin' file in
 *  read_pin_file (read_pins.c).  The first index is the channel number,
 *  and the second index is the location (TOP, BOTTOM, etc.) within the
 *  channel.
 */
typedef struct pin_struct pin_t;
struct pin_struct {
    int position;		/* pin location (lower left corner) */
    int width;			/* pin width */
    int group;			/* which group pin belongs to */
    int column;			/* which column pin lies in */
    octId instId;
    octId termId;
    octId netId;
    pin_t *next;
};
extern pin_t ***channel;


/* 
 *  Linked list of which instances are in which blocks.  This is created from
 *  the Timberwolf '.pl1' file in read_place_file (read_place.c).  The index
 *  is the block number.
 */
typedef struct block_record_struct block_record_t;
struct block_record_struct {
    octId objectId;
    block_record_t *next;
};
extern block_record_t **block_list;

/*
 *  Record minimum x position assigned by TW
 */
extern int min_xposition;
extern int max_xposition;

/* 
 *  TimberWolfSC parameters
 */
extern int num_rows;
extern double rowSep;
extern double indent;
extern double h_net_weight;
extern double v_net_weight;
#ifdef TW3_3
extern int att_per_cell;
extern double do_global_route_cell_swaps;
#endif

#ifdef TW4_2
extern int fallback;
#endif
extern int Add_implicit;
extern int fast_mode;
extern int addFeeds;
#ifndef TW5_0
extern double implicit_feed_thru_range;
#endif
extern int do_global_route;
extern double TargetAspectRatio;

extern int complex_feeds;


/*
 *  Design rules (read in options.c from a 'turd' file)
 */
extern int units_per_lambda;
extern char *feedthru_cell;
extern char *feedthru_view;
extern int feedThruWidth;
extern int hlayer_width;
extern int vlayer_width;
extern int hlayer_pitch;
extern int vlayer_pitch;
extern char *hlayer_name;
extern char *vlayer_name;
extern char *h_contact_name;
extern char *h_contact_view;
extern char *v_contact_name;
extern char *v_contact_view;
extern char *hv_contact_name;
extern char *hv_contact_view;
extern int  hv_contact_width;
extern int  hv_contact_height;
extern char *default_terminal_cell;
extern char *default_terminal_view;
extern int extend_terms;
extern int power_width;
extern int power_position;	/* Vdd on left or right side */


/*  MISC  */
extern int debug;		/* debug parameter */

