/* sccsid = @(#)write_cell.c	1.7 wolfe (ITD) 7/30/90   */
/* Modified     9/25/87 to add the additional
properties WOLFE_LEFT , WOLFE_RIGHT, WOLFE_TOP
WOLFE_BOTTOM for compatability with the manual
and earlier versions                 
Also added FEEDTHRU information for the cells.

                        Kanwar Jit Singh    */

#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "st.h"
#include "wolfe.h"

static int write_pad_info();
static void write_power_terminals();
int imp_feed_count;

int write_cell_file(prefix, facet)
char *prefix;
octObject *facet;
{
    struct octBox cell_bb;
    octObject inst_master, inst, inst_bag, bag, prop;
    octGenerator gen;
    octCoord x_center, y_center;
    octStatus status;
    FILE *fp;
    char filename[1024];
    int cell_num, class, position;
    char fixed[20], left_or_right[20] ;



    /* Open for write the TimberWolf cell file */
    (void) sprintf(filename, "%s.cel", prefix);
    if ((fp = fopen(filename, "w")) == NULL) {
	perror(filename);
	return 1;
    }
    cell_num = 0;
    imp_feed_count = 0;

    /* Generate all instances from INSTANCES bag */
    OH_FAIL(ohGetByBagName(facet, &inst_bag, "INSTANCES"),
	    "Can't find INSTANCES bag connected to", facet);
    OH_ASSERT(octInitGenContents(&inst_bag, OCT_INSTANCE_MASK, &gen));
    while ((status = octGenerate(&gen, &inst)) == OCT_OK) {

	/* Check for a WOLFE-CLASS attached to the instance */
	class = 0;
	status = ohGetByNameContainer(&inst, &bag, OCT_BAG, "WOLFE-CLASS");
	/* leave the WOLFE-ROW in for upward compatibility even though its 
	   dumb to only be able to specify the row 
	   */
	if (status == OCT_OK) {
	    OH_FAIL(octGenFirstContent(&bag, OCT_PROP_MASK, &prop),
		    "Cannot find WOLFE-ROW or WOLFE-Place property in a WOLFE-CLASS bag", &bag);
	    if (strcmp(prop.contents.prop.name, "WOLFE-ROW") == 0
		&& prop.contents.prop.type == OCT_INTEGER) {
		class = prop.contents.prop.value.integer;
		strcpy(fixed,"nonfixed");
		position = 0 ;
		strcpy(left_or_right,"left");
	    }
	    if (strcmp(prop.contents.prop.name, "WOLFE-PLACE") == 0
		&& prop.contents.prop.type == OCT_STRING) {
		sscanf(prop.contents.prop.value.string,
		       "%s %d %s %d", fixed,&position,left_or_right, &class);
	    }
	}

	/*
	 *  Write out the TimberWolf header for the cell
	 */
	cell_num++;
	(void) fprintf(fp, "cell %d %s\n", 
		       cell_num, inst.contents.instance.name);
	if (class != 0) {
	    (void) fprintf(fp, 
			   "initially %s %d from %s of block %d\n", fixed, position, left_or_right, class);
	}

	/* 
	 * Grab bounding box of the cell from the instance, determine the
	 * funny coordinates needed for TimberWolf, and write it out
	 */

	if ( ohOpenMaster(&inst_master, &inst, "contents", "r") < OCT_OK){
	    perror(filename);
	    return 1;
	}
	if (octTB(&inst_master, &cell_bb) != OCT_OK) { 
	    /*if (octBB(&inst, &cell_bb) != OCT_OK) { */
	    (void) fprintf(stderr, "Failed on %s\n", ohFormatName(&inst));
	    octError("need bounding box for inst\n");
	    return 1;
	}
	x_center = (cell_bb.upperRight.x + cell_bb.lowerLeft.x) / 2;
	y_center = (cell_bb.upperRight.y + cell_bb.lowerLeft.y) / 2;
	(void) fprintf(fp, "left %d right %d bottom %d top %d\n",
		       (cell_bb.lowerLeft.x - x_center) / units_per_lambda,
		       (cell_bb.upperRight.x - x_center) / units_per_lambda,
		       (cell_bb.lowerLeft.y - y_center) / units_per_lambda, 
		       (cell_bb.upperRight.y - cell_bb.lowerLeft.y) / units_per_lambda
		       + (cell_bb.lowerLeft.y - y_center)/units_per_lambda );

	write_pins(fp, &inst, cell_bb);
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

    if (write_pad_info(facet, fp) != 0) {
	return 1;
    }

    if (ferror(fp) || (fclose(fp) == EOF)) {
	(void) fprintf(stderr, "Fatal error writing file %s\n", filename);
	return 1;
    } else {
	return 0;
    }
}

struct feed_thru_list{
    octCoord feed_thru_pos;
    struct feed_thru_list *next;
    } ;

write_pins(fp, inst, cell_bb)
FILE *fp;
octObject *inst;
struct octBox cell_bb;
{
    octObject temp_facet,net, aterm, aterm1, fterm, fterm1, joined,
    prop, acc_prop, instmaster, currterm, jointerm;
    octGenerator gen;
    octStatus status;
    octCoord xc, yc, pin_width, x_center, y_center, top, bottom;
    struct octBox interface_bb,pin_bb;
    int i, j, nterms, feedthru_p, hi_R;
    octId *terms, temp_id;
    char *name, *dummy, *netname;
    static int first = 1;
    static st_table *feed_table;
    struct feed_thru_list *feed_thru_ptr, *feed_ptr;
    struct feed_thru_list *get_feed_thru();

    /* Re-compute the cell location */
    x_center = (cell_bb.upperRight.x + cell_bb.lowerLeft.x) / 2;
    y_center = (cell_bb.upperRight.y + cell_bb.lowerLeft.y) / 2;
    bottom = (cell_bb.lowerLeft.y - y_center)/units_per_lambda ;
    top = (cell_bb.upperRight.y - cell_bb.lowerLeft.y)/units_per_lambda + bottom ;

    /* Find the octId of the master of the instance in storing
    the data inside the symbol table */
    if (ohOpenMaster(&temp_facet,inst,"interface","r") < OCT_OK){
        octError("Opening the master of instance \n");
        return;
    }
    temp_id = temp_facet.objectId;
    OH_ASSERT(octTB(&temp_facet,&interface_bb));

    /* Scan if the cell has been visited before-- If yes then
    recover the feedthru info from the corresp structure , else
    save the cell and feedthru information in the symbol table*/
    if( Add_implicit){
	if( first){
	feed_table = st_init_table(st_numcmp,st_numhash);
	first = 0;
	}
	if (st_lookup(feed_table, (char *) temp_id, &dummy) == 1){
	    feed_thru_ptr = (struct feed_thru_list *) dummy;
	} 
	else {
	/* get the feed thru info and add it to the symbol table*/
	    feed_thru_ptr = get_feed_thru(&temp_facet,interface_bb);
	    if( debug > 2){
		(void) printf(" FEEDTHRU INFO DERIVED FOR %s \n",ohFormatName(inst));
	    }
	    (void) st_insert(feed_table, (char *) temp_id, (char *) feed_thru_ptr);
	}
	/* at this stage the feed thru info is in feed_thru_ptr */

	j=0;
	for ( feed_ptr = feed_thru_ptr; feed_ptr != NIL(struct feed_thru_list); feed_ptr = feed_ptr->next){
	    imp_feed_count++;
	    (void)fprintf(fp,"pin name WOLFE-TOP-FEED%d signal TW_PASS_THRU %d %d \n",
	    ++j,feed_ptr->feed_thru_pos / units_per_lambda,
	    top);
	    (void)fprintf(fp,"equiv name WOLFE-BOT-FEED%d %d %d \n",
	    j,feed_ptr->feed_thru_pos / units_per_lambda,
	    bottom);
	}
    }

    /* Gen the terminals into an array; watching for JOINED is tough */
    ohContentsToArray(inst, OCT_TERM_MASK, &terms, &nterms);

    for(i = 0; i < nterms; i++) {
    /* See if terminal has already been output (because of JOINED) */
	if (octIdIsNull(terms[i])) continue;
	OH_ASSERT(ohGetById(&aterm, terms[i]));

	/* Find the net on the actual terminal */
	status = octGenFirstContainer(&aterm, OCT_NET_MASK, &net);
	if (status == OCT_GEN_DONE) continue;
	OH_ASSERT(status);

	/* Find the formal terminal in the interface */
	OH_ASSERT_DESCR(ohFindFormal(&fterm, &aterm),
	    "failed finding matching formal terminal in interface");

	/* Check if it is a power supply or ground pin */
	status = ohGetByPropName(&fterm, &prop, "TERMTYPE");
	if (status == OCT_OK && prop.contents.prop.type == OCT_STRING) {
	    name = prop.contents.prop.value.string;
	    if (strcmp(name, "GROUND")==0 || strcmp(name, "SUPPLY")==0){
	    continue;
	    }
	}
	OH_ASSERT(status == OCT_NOT_FOUND ? OCT_OK : status);
    /* check if this pin is a declared feedthru or high resistance */
    status = ohGetByPropName(&fterm, &prop, "FEEDTHRU");
       if (status == OCT_OK) feedthru_p = 1;
         else feedthru_p = 0;
    
    status = ohGetByPropName(&fterm, &prop, "HI_R");
       if (status == OCT_OK) hi_R = 1;
         else hi_R = 0;

	/* Get the pin bbox */
	status = ohTerminalBB(&aterm, &pin_bb);
	if (status != OCT_OK) {
	    (void) fprintf(stderr, "Failed finding bounding box for %s\n",
	    ohFormatName(&aterm));
	    OH_ASSERT(status);
	}

	/* output the pin information */
	pin_width = pin_bb.upperRight.x - pin_bb.lowerLeft.x;
	xc = pin_bb.lowerLeft.x - x_center + pin_width/2;
	yc = (pin_bb.lowerLeft.y > y_center) ? top : bottom;

    if (feedthru_p) netname = "TW_PASS_THRU";
      else netname = net.contents.net.name;

/* Now see if we can get an ACCESSIBILITY property for the terminal */

    acc_prop.type = OCT_PROP;
    acc_prop.contents.prop.name = "ACCESSIBILITY";
    if (octGetByName (&fterm, &acc_prop) == OCT_OK)
    {
     if (strcmp (acc_prop.contents.prop.value.string, "TOP_ONLY") == 0)
     {
       (void) fprintf(fp, "pin name %s signal %s %d %d\n",
       aterm.contents.term.name, netname, 
       xc , top );
     }
     else if (strcmp (acc_prop.contents.prop.value.string, "BOTTOM_ONLY") == 0)
     {
       (void) fprintf(fp, "pin name %s signal %s %d %d\n",
       aterm.contents.term.name, netname, 
       xc , bottom );
     }
     else if (strcmp (acc_prop.contents.prop.value.string, "BOTH") == 0)
     {
        (void) fprintf(fp, "pin name WOLFE-TOP-%s signal %s %d %d\n",
        aterm.contents.term.name, netname, 
        xc , top );
        (void) fprintf(fp, "equiv name WOLFE-BOT-%s %d %d\n",
        aterm.contents.term.name, 
        xc , bottom );
     }
     else
     {
       fprintf (stderr, "ACCESSIBILITY property %s unrecognized.\n", 
		prop.contents.prop.value.string);
       exit (-1);
     }
   }
   else
     {
	if (pin_bb.lowerLeft.y == cell_bb.lowerLeft.y ||
		  pin_bb.upperRight.y == cell_bb.upperRight.y) {
	    (void) fprintf(fp, "pin name %s signal %s %d %d\n",
	    aterm.contents.term.name, net.contents.net.name, 
	    xc / units_per_lambda, yc);

	} else {
	    /* pin is interior to cell; assume accessible from all directions */
	    (void) fprintf(fp, "pin name WOLFE-TOP-%s signal %s %d %d\n",
	    aterm.contents.term.name, net.contents.net.name, 
	    xc / units_per_lambda, top);
	    (void) fprintf(fp, "equiv name WOLFE-BOT-%s %d %d\n",
	    aterm.contents.term.name, 
	    xc / units_per_lambda, bottom);
	}
      }


	/* Look for a JOINED bag */

        if (ohOpenFacet (&instmaster, inst->contents.instance.master,
                         "physical", "interface", "r") < OCT_OK)
        {
	  char buffer[1024];
          (void) sprintf(buffer, "Cannot open the cell %s:physical:interface\n",
                         inst->contents.instance.master);
	  octError(buffer);
        }
	VOVinputFacet( &instmaster );
        if (ohGetByTermName (&instmaster, &currterm,
                             aterm.contents.term.name)
            != OCT_OK)
        {
	  char buffer[1024];
          (void) sprintf(buffer, "Could not find the terminal %s\n",
                         aterm.contents.term.name);
	  octError(buffer);
          (void) sprintf(buffer, "In the cell %s:physical:interface.\n",
                         inst->contents.instance.master);
	  octError(buffer);
        }
	status = ohGetByNameContainer(&currterm, &joined, OCT_BAG, "JOINED");
	if (status == OCT_NOT_FOUND) continue;
	OH_ASSERT(status);


	/* all terminals in bag 'joined' are equivalent to term */
	OH_ASSERT(octInitGenContents(&joined, OCT_TERM_MASK, &gen));
	while ((status = octGenerate(&gen, &jointerm)) == OCT_OK) {
	    if (!octIdsEqual(jointerm.objectId, currterm.objectId)) {
              if (ohGetByTermName (inst, &aterm1, jointerm.contents.term.name)
                  != OCT_OK) {
	        char buffer[1024];
                (void) sprintf(buffer, "Could not find the terminal %s\n",
                               jointerm.contents.term.name);
	        octError(buffer);
                (void) sprintf(buffer, "In the cell %s:physical:contents.\n",
                               inst->contents.instance.name);
	        octError(buffer);
              }

	    /* Output the equiv pin information */
	    OH_ASSERT(ohTerminalBB(&aterm1, &pin_bb));
	    pin_width = pin_bb.upperRight.x - pin_bb.lowerLeft.x;
	    xc = pin_bb.lowerLeft.x - x_center + pin_width/2;
	    yc = (pin_bb.lowerLeft.y > y_center) ? top : bottom;
	    (void) fprintf(fp, "equiv name %s %d %d\n",
		aterm1.contents.term.name,
		xc / units_per_lambda, yc);

	    /* Remove the terminal from the terminals to be processed */
	    for(j = 0; j < nterms; j++) {
		if (octIdsEqual(terms[j], aterm1.objectId)) {
		terms[j] = oct_null_id;
		break;
		}
	    }
	    }
	}
	OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    }
    FREE(terms);
}


int box_compare(box1,box2)
octObject *box1, *box2;
{
    struct octBox bb1, bb2;
    OH_ASSERT(octBB(box1,&bb1));
    OH_ASSERT(octBB(box2,&bb2));
    return bb1.lowerLeft.x - bb2.lowerLeft.x;
}

/*
 * Scan the interface facet of the cell and find out the
 * projections of the v_layer on the x_axis. Then depending on
 * the v_layer pitch find the empty spaces that can 
 * accomodate the feed thrus. Form a linked list and return
 * the positions relative to the cell center
 */

struct feed_thru_list *get_feed_thru(interface,cell_bb)
octObject *interface;
struct octBox cell_bb;
{
    octCoord x_center,max_xpos, v_pitch;
    struct octBox bbox;
    octGenerator genBox;
    int i = 0, feed_count = 0,temp_count;
    octObject vlayer, *temp_array;
    struct feed_thru_list *feed, *new_feed;

    OH_ASSERT(ohGetOrCreateLayer(interface,&vlayer,vlayer_name));
    temp_count = ohCountContents(&vlayer,OCT_GEO_MASK);

    /* Store the boxes on v-layer in a temporary array */
    temp_array = ALLOC(octObject, temp_count);
    if ( debug > 2){
        (void) printf("Number of geometries on %s = %d\n",vlayer_name,temp_count);
    }
    OH_ASSERT( octInitGenContents(&vlayer,OCT_GEO_MASK,&genBox));
    while( octGenerate(&genBox,&(temp_array[i++])) == OCT_OK){
    }


    qsort((char *)temp_array, temp_count, sizeof(octObject ), box_compare);

    /* Now that the boxes are sorted we can scan it serially */
    x_center = (cell_bb.upperRight.x + cell_bb.lowerLeft.x) /2 ;
    max_xpos = cell_bb.lowerLeft.x;    /* Initialising */
    feed = NIL(struct feed_thru_list);
    v_pitch = (vlayer_pitch * units_per_lambda);
    for ( i = 0; i < temp_count; i++){
    OH_ASSERT(octBB(temp_array+i,&bbox));
    if ( debug > 2){
    /* Print the lower left x positions of the boxes */
        (void) printf("The %dth box has ll pos = %d\n",i+1,bbox.lowerLeft.x);
    }
    while( max_xpos + 5 * v_pitch/4 <= bbox.lowerLeft.x){
                new_feed = ALLOC(struct feed_thru_list,1);
                new_feed->next = feed;
                new_feed->feed_thru_pos =
                  (max_xpos-x_center)+v_pitch/2;
                feed = new_feed;
                feed_count++;
                max_xpos += v_pitch;
     }
/*    max_xpos = MAX((bbox.upperRight.x + v_pitch/4) , max_xpos); */
    while (max_xpos < bbox.upperRight.x + v_pitch/4)
      max_xpos += v_pitch;
    if( i == temp_count-1){
    /* compare with the end of the bbox of the cell */
    while( max_xpos + 5 * v_pitch/4 <= cell_bb.upperRight.x){
                new_feed = ALLOC(struct feed_thru_list,1);
                new_feed->next = feed;
                new_feed->feed_thru_pos = (max_xpos-x_center)+v_pitch/2;
                feed = new_feed;
                feed_count++;
                max_xpos += v_pitch;
     }
    }
    }

    if( debug > 2){
        (void) printf("%d Feed-Thrus found \n",feed_count);
    }
    FREE(temp_array);
     return feed;
}


/*
 *  Formal terminal positions are specified by placing properties in
 *  the interface facet of a cell.
 *
 *  TERM-FLOATING-EDGE is a string-valued property attached to a terminal
 *  in the interface facet with value "LEFT", "RIGHT", "TOP", or "BOTTOM" 
 *  giving the proper side for a terminal.
 *
 *  TERM-RELATIVE-POSITION is a real-valued property (with value between
 *  0 and 1) giving the position relative position for the terminal along
 *  the specified edge.
 */

static int write_pad_info(facet, fp)
octObject *facet;
FILE *fp;
{
    octObject net, term, fterm, interface, cnt, sideprop;
    octGenerator gen;
    octStatus status;
    float frac;
    int count[4], placed[4], pos, pad_num;
    char *side;

    status = ohOpenInterface(&interface, facet, "a");
    if (status < OCT_OK) {
    (void) fprintf(stderr, "Unable to open interface facet of %s\n",
        ohFormatName(&interface));
    octError(NIL(char));
    exit(1);
    }

    /*
     *  Generate over the formal terminals, counting # terminals per side
     */
    count[TOP] = count[BOTTOM] = count[LEFT] = count[RIGHT] = 0;
    OH_ASSERT(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &fterm)) == OCT_OK) {
	if ( get_terminal_position(&fterm, &pos, &frac) )
	    count[pos]++;
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);


    /*
     *  Generate over the formal terminals writing the cell information
     *  to the cell file.
     */
    pad_num = 0;
    placed[TOP] = placed[BOTTOM] = placed[LEFT] = placed[RIGHT] = 0;
    OH_ASSERT(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &fterm)) == OCT_OK) {
      frac = (float) 0;
        if ( get_terminal_position(&fterm, &pos, &frac)){

	    /*
	     * Not a power/ground terminal: If nop position specified
	     * hack for uniform distribution of pins 
	     */
	    if (frac == -1) {
		frac = (float) (placed[pos] / count[pos]);
	    }
/*
 * Start of Code added by Barry Boes to keep two formal terminals
 * on the same net and same side of the cell from being generated.
 * This is done because of problems TW has with this configuration.
 */
/*            cnt.type = OCT_BAG;
            cnt.contents.bag.name = "JOINED";
*/
            /* fprintf (stderr, "pin name %s ", fterm.contents.term.name); */
/*
            if ((octGetContainerByName (&fterm, &cnt)) ==  OCT_OK)
*/
            if ((octGenFirstContainer (&fterm, OCT_NET_MASK, &cnt)) == OCT_OK)
            {
		/* fprintf (stderr, "\n"); */
              switch (pos)
              {
                case LEFT :
                  side = "LEFT";
                  break;
                case RIGHT :
                  side = "RIGHT";
                  break;
                case TOP :
                  side = "TOP";
                  break;
                case BOTTOM :
                  side = "BOTTOM";
                  break;
                }
              if ((ohGetByPropName (&cnt, &sideprop, side)) == OCT_OK)
              {
		  /* fprintf (stderr, "got a side %s\n", side); */
                continue;
              }
              OH_ASSERT (ohCreatePropStr (&cnt, &sideprop, side,
                                          side));
              /* fprintf (stderr, "made a side %s\n", side); */
              octMarkTemporary (&sideprop);
            }
            else
            fprintf (stderr, "no JOINED\n");
            
	    status = ohTerminalNet(&fterm, &net);
	    if (status != OCT_OK) {
		(void) fprintf(stderr, "Cannot find a net for %s\n",
		ohFormatName(&fterm));
		octError("ohTerminalNet");
	    } else {
		(void) fprintf(fp, "pad %d WOLFE-PAD orient 0\n", pad_num);
		(void) fprintf(fp, "padside %c\n", print_location[pos][0]);
		(void) fprintf(fp, "left %d right %d bottom %d top %d\n",
		-hlayer_width, hlayer_width, -hlayer_width, hlayer_width);
		(void) fprintf(fp, "sidespace %6.4f\n", frac);
		(void) fprintf(fp, "pin name %s signal %s %d %d\n",
		fterm.contents.term.name, net.contents.net.name, 0, 0);
		pad_num++;
		placed[pos]++;
	    }
	}
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

    /* 
     * XXX Hack to support terminals conneced to SUPPLY or GROUND nets
     */
    write_power_terminals(facet, fp, &pad_num, "SUPPLY");
    write_power_terminals(facet, fp, &pad_num, "GROUND");

    OH_FAIL(octCloseFacet(&interface), 
    "Error closing interface facet", &interface);

    return 0;
}


/* Return 1 => not a power/gnd formal terminal */
int
get_terminal_position(fterm, pos_p, frac_p)
octObject *fterm;
int *pos_p;
float *frac_p;
{
    octObject term, prop, facet, interface;
    octStatus status;
    char *val;

    /* Make sure this is the interface facet formal terminal */
    octGetFacet(fterm, &facet);
    if (strcmp(facet.contents.facet.facet, "interface") != 0) {
    status = ohOpenInterface(&interface, &facet, "a");
    if (status < OCT_OK) {
        (void) fprintf(stderr, "Unable to open interface facet of %s\n",
        ohFormatName(&facet));
        octError(NIL(char));
        exit(1);
    }
    if(ohGetByTermName(&interface,&term,fterm->contents.term.name) != OCT_OK){
        (void) printf("Connot find %s in interface \n",
        ohFormatName(fterm));
        *pos_p = LEFT;
        *frac_p = (float) 0;
        return 0;
    }
    (void) octCloseFacet(&interface);
    }
    else {
    term = *fterm;
    }

    status = ohGetByPropName(&term, &prop, "TERMTYPE");
    if ((status == OCT_OK) && (prop.contents.prop.type == OCT_STRING)){
        val = prop.contents.prop.value.string;

	if (!strcmp(val,"SUPPLY")){
	    *pos_p = power_position; 
	    *frac_p = 0.9; /* Power terminals on the top */
	    if (power_position == BOTH) *pos_p = LEFT;
	    return 0;
	} else if (!strcmp(val,"GROUND")){
	    *pos_p = RIGHT;
	    *frac_p = 0.1; /* Ground terminals on the bottom */
	    if (power_position == RIGHT) *pos_p = LEFT;
	    return 0;
	}
    }

    status = ohGetByPropName(&term, &prop, "TERM_EDGE");
    if ((status == OCT_OK) && (prop.contents.prop.type == OCT_STRING)) {
        val = prop.contents.prop.value.string;
        status = ohGetByPropName(&term, &prop, "TERM_RELATIVE_POSITION");
        if (status == OCT_OK && prop.contents.prop.type == OCT_REAL)
            *frac_p = (float) prop.contents.prop.value.real;
          else
            if (status == OCT_OK && prop.contents.prop.type ==
                OCT_STRING)
            {
              sscanf (prop.contents.prop.value.string, "%f", frac_p);
            }
         if (strcmp(val, "LEFT") == 0) {
            *pos_p = LEFT;
            return 1;
        } else if (strcmp(val, "RIGHT") == 0) {
            *pos_p = RIGHT;
            return 1;
        } else if (strcmp(val, "TOP") == 0) {
            *pos_p = TOP;
            return 1;
        } else if (strcmp(val, "BOTTOM") == 0) {
            *pos_p = BOTTOM;
            return 1;
        }
    }

    *pos_p = random() % 4;
    *frac_p = (float) (random() % 65536) / 65536;
    (void) fprintf(stderr, 
    "Randomly assigning formal terminal %s to %s at %4.3f\n",
    term.contents.term.name, print_location[*pos_p], *frac_p);
    OH_ASSERT(ohCreateOrModifySprop(&term, &prop, "TERM_EDGE", 
            print_location[*pos_p]));
    OH_ASSERT(ohCreateOrModifyRprop(&term, &prop, "TERM_RELATIVE_POSITION", 
            *frac_p));

    return 1;
}

static void
write_power_terminals(facet, fp, pad_num, string)
octObject *facet;
FILE *fp;
int *pad_num;	/* Unique identifier for the pad */
char *string;	/* Type of the power rail */
{
    float frac;
    octGenerator gen;
    octStatus status;
    int add_supply=0, position, formal_supply=0;
    octObject term, power_term, fterm, prop, net, inst;
    octObject inst_interface;

    find_power_net(facet,&net,string);
    OH_ASSERT(octInitGenContents(&net, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &term)) == OCT_OK) {
	if (!octIdIsNull(term.contents.term.instanceId)) {
	    /* Is this instance a power/gnd terminal on that instance ?? */
	    inst.objectId = term.contents.term.instanceId;
	    OH_ASSERT(octGetById(&inst));
	    OH_ASSERT(ohOpenMaster(&inst_interface, &inst, "interface", "r"));
	    OH_ASSERT(ohGetByTermName(&inst_interface, &fterm, term.contents.term.name));
	    status = ohGetByPropName(&fterm, &prop, "TERMTYPE");
	    if (status == OCT_OK && prop.contents.prop.type == OCT_STRING){
		if (strcmp(prop.contents.prop.value.string, string)){
		    /* A signal pin is connected to SUPPLY/GND rail */
		    (void)fprintf(stdout,"MESSAGE: signal term \"%s\" on inst \"%s\" connected to %s net \"%s\"\n", 
		    fterm.contents.term.name, inst.contents.instance.name,
		    string, net.contents.net.name);
		    add_supply = 1;
		}
	    }
	} else {
	    formal_supply = 1;
	    power_term = term; /* Need it for the "pin name" field */
	}
    }

    /* Need to make sure that the terminal for the supply reaches the side */
    if (add_supply){
	(*pad_num)++;
	if (!strcmp(string,"SUPPLY")){
	    position = power_position; 
	    frac = 0.9; /* Power terminals on the top */
	    if (power_position == BOTH) position = LEFT;
	} else if (!strcmp(string,"GROUND")){
	    position = RIGHT;
	    frac = 0.1; /* Ground terminals on the bottom */
	    if (power_position == RIGHT) position = LEFT;
	} else {
	    (void)fprintf(stderr,"Unknown option in write_power_terminal\n");
	    return;
	}
	(void) fprintf(fp, "pad %d WOLFE-PAD orient 0\n", *pad_num);
	(void) fprintf(fp, "padside %c\n", print_location[position][0]);
	(void) fprintf(fp, "left %d right %d bottom %d top %d\n",
	-hlayer_width, hlayer_width, -hlayer_width, hlayer_width);
	(void) fprintf(fp, "sidespace %6.4f\n", frac);
	(void) fprintf(fp, "pin name %s signal %s %d %d\n",
	(formal_supply ? power_term.contents.term.name : string),
	net.contents.net.name, 0, 0);
    }
}
