/* Modified to handle criticality of the nets 
								9/25/87
							Kanwar Jit Singh */
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"


/* ARGSUSED */
int write_netweight_file(prefix, facet)
char *prefix;
octObject *facet;		/* May need this eventually */
{
    char filename[1024];
    FILE *fp;
	double net_crit;
	double h_crit, v_crit;
	octObject net, prop;
	octGenerator genNet;
	octStatus status;

    (void) sprintf(filename, "%s.net", prefix);
    if ((fp = fopen(filename, "w")) == NULL) {
	(void) fprintf(stderr, "Unable to open \"%s\"\n", filename);
	return 1;
    }

    (void) fprintf(fp, "allnets HVweights %4.2lf %4.2lf\n", 
	h_net_weight, v_net_weight);

	/* Read the values of NETCRITICALITY for the nets
	and translate them into positive values 	*/
	OH_ASSERT( octInitGenContents(facet,OCT_NET_MASK,&genNet));
	while ((status = octGenerate(&genNet,&net)) == OCT_OK){
		/* Check for the CRITICALITY PROPERTY */
		status = ohGetByPropName(&net,&prop,"NETCRITICALITY");
		if ((status == OCT_OK) && (prop.contents.prop.type == OCT_REAL)){
			net_crit = (double)prop.contents.prop.value.real;

			/* Scale the criticality so that 0 corresponds to h_net_weight
			1000  corresponds to 10 * h_net_weight
			-1000 corresponds to 0.1 * h_net_weight 			*/
			if( net_crit > 0){
			h_crit = (net_crit *9 / 1000 + 1)*h_net_weight ;
			v_crit = (net_crit *9 / 1000 + 1)*v_net_weight ;
			}
			else{
			h_crit = ((net_crit * 0.9) / 1000.0 + 1)*h_net_weight ;
			v_crit = ((net_crit * 0.9) / 1000.0 + 1)*v_net_weight ;
			}
			(void) fprintf(fp,"net %s HVweights %4.2lf %4.2lf ",
			net.contents.net.name,h_crit,v_crit);
			status = ohGetByPropName(&net,&prop,"NETTYPE");
			if ((status == OCT_OK) && (prop.contents.prop.type == OCT_STRING)){
				if(strcmp(prop.contents.prop.value.string,"CLOCK") == 0)
				(void) fprintf(fp,"ignore");
			}
			(void) fprintf(fp,"\n");
		}
	}


    /* Return true if any errors writing the file */
    return ferror(fp) || (fclose(fp) == EOF);
}
